#!/usr/bin/perl -W
#--------------------------------------------------------------
# These are MONO & UNDULATOR related subs for beamline controls Perl scripts
#
#     				Author: Sergey Stepanov
#
# Version-2.0:  2001/02/05 using Pezca
# Version-3.0:  2005/03/28 using CA-monitors array
# Version-4.0:  2008/04/16 added functions to read mono offsets from mySQL
# Version-4.1:  2013/08/21 changed mySQL table from EpicsBeamlineAttributes to Config
# Version-4.2:  2013/10/11 added calcMonoTime and runMonoIDscan
# Version-4.3:  2014/03/28 switched calls to motor_subs.pl from passing arrays by value to passing by pointer
# Version-4.4:  2015/11/17 added mono_temperature and mono_stabilization_time
# Version-4.5:  2016/09/01 implemented "IDsteps" option in "runMono" (mono is moving continuously, ID in steps and trying to keep)
# Version-4.6:  2021/02/06 added feedback interaction to the &runMono multiStep and IDsteps modes.
# Version-4.7:  2021/08/16 added &getMonoHKL() and &setMonoHKL().
# Version-4.8:  2021/08/21 added undulator 5th and 7th harmonics and  switch between 111 and 333.
#--------------------------------------------------------------
# List of functions:
# ^^^^^^^^^^^^^^^^^
#  1.       $ActEn = &getEnergy([$monoAxisPV]);
#  2.       $ActEn = &getIDEnergy($undulatorPV);
#  3.       $RqsEn = &getIDEnergyRqs($undulatorPV);
#  4.      $ActGap = &getIDGap($undulatorPV);
#  5. $ActHarmonic = &getIDHarmonic($undulatorPV);
#  6.      $RqsGap = &getIDGapRqs($undulatorPV);
#  7       $status = &setIDHarmonic($undulatorPV,$harmonic);   		#set harmonic only
#  8       $status = &isIDwritable($undulatorPV);              		#1=yes,0=no,2="no APS shutter permit",-1=error
#  9       $status = &setID($undulatorPV,$energy,[$harmonic]); 		#set the values, but not start
# 10.      $status = &startID($undulatorPV,$energy,[$harmonic]);	#set the value and start
# 11.      $status = &trackID($undulatorPV);
# 12.      $status = &abortID($undulatorPV);
# 13.      $status = &runID($undulatorPV,$energy,[$harmonic]);
# 14.         $gap = &E2gap($undulatorPV,$energy,[$harmonic,$restore]);
# 15.    ($status,
#       $beginGap,
#         $endGap) = &calcIDSpeed($undulatorPV,$beginEnergy,$endEnergy,$scanTime);
# 16.      $status = &startIDScan($undulatorPV,$beginGap,$endGap,$scanTime);
# 17.      $status = &abortIDScan($undulatorPV);
# 18.      $status = &runMono($monoAxis, $monoAssy, $E_end, 
#                          [$linkID, $undulatorPV, $IDoffsetEnd,	#if $IDoffset is specified, it overrides SQL data
#                           $syncMode, $syncStep, $switchHarmonic,
#                           $attenuation, $switchMono, 
#                           \$enmono13, \@harmEn, \@harmNum, 
#                           \@Elower, \@Eupper, \@Olower, \@Oupper]);
# 19.      $status = &runMonoIDscan($monoAxisPV,$monoAssyPV,$energy,$undulatorPV,$IDoffset,
#                          [$IDharmonic,$attenuation]);
# 20.      $status = &getVariableIDoffsets(\%EnergiesIJ_and_Offsets_Energies);
#                                                            		#array of hashes, contains $records[$i]->{name}->data
#                                                            		# Possible names (not all may be defined):
#                                                            		# h13,h35,h57,m13
#                                                            		# eLo1,eUp1,oLo1,oUp1
#                                                            		# eLo3,eUp3,oLo3,oUp3
#                                                            		# eLo5,eUp5,oLo5,oUp5
#                                                            		# eLo7,eUp7,oLo7,oUp7
# 21.    $status = &getIDparams(\$enmono13,\@harmEn,\@harmNum,
#                               \@Elower,\@Olower,\@Eupper,\@Oupper);	# a more-friendly wrapper around &getVariableIDoffsets
# 22.    $offset = &calcIDoffset($energy,[$switchID,			#$enmono13 is not used, but passed for reading SQL
#                                \$enmono13,\@harmEn,\@harmNum,
#                                \@Elower,\@Eupper,\@Olower,\@Oupper]);
# 23.    $status = &getLaneOffsets(\@laneVLenergy, \@laneVLposition, \@laneVLmaterial,
#                                  \@laneHDenergy, \@laneHDposition, \@laneHDmaterial);
# 24.     $time = &calcMonoTime($monoAssy, $energy);
# 25.     $temp = &mono_temperature($E,$E13);				#also uses global $beamline
# 26.     $time = &mono_stabilization_time($E_i,$E_f,$E13);		#also uses global $beamline
# 27.     $hkl  = &getMonoHKL();					#$hkl = '111' or '333'
# 28.   $status = &setMonoHKL($hkl);					#$hkl = '111' or '333'
#--------------------------------------------------------------
# use Pezca;
# do 'motor_subs.pl';

# use Math::Trig;
# use warnings::unused;
  use strict;
# no strict 'vars';
# no strict 'subs';
# no strict 'refs';

#=================================================================

# sub getEnergy ([$monoAxisPV]) {
sub getEnergy (;$) {

### This routine returns actual Mono energy

  our ($beamline);
  my ($script, $monoAxisPV, $status, $energy);
  $script     = 'getEnergy';
  $monoAxisPV = $_[0];

  if (!defined $beamline) {print STDOUT '*** '.$script.': no beamline defined.'."\n"; return 0;}
  if (!defined $monoAxisPV) {$monoAxisPV = $beamline.':MO:E';}
  else {
     if ($monoAxisPV !~ /^$beamline/) {
        if ($monoAxisPV =~ /^:/) {$monoAxisPV = $beamline.':'.$monoAxisPV;}
        else                     {$monoAxisPV = $beamline.$monoAxisPV;}
     }
  }
### Get Actual Energy
  $energy = &getAxisAct($monoAxisPV,\$status);
  if ($status) {return 0;}
  else         {return $energy;}
}

#=================================================================

# sub getIDEnergy ($undulatorPV) {
sub getIDEnergy ($) {

### This routine returns actual Undulator energy

  my ($script, $undulatorPV, $status, $energy);
  $script      = 'getIDEnergy';
  $undulatorPV = $_[0];

### Get Actual undulator Energy
  ($status,$energy) = Pezca::GetDouble($undulatorPV.':Energy');
  if ($status) {
     &report_only($script, $status, $undulatorPV.':Energy');
     return 0.0;
  } else {
     return $energy;
  }
}

#=================================================================

# sub getIDEnergyRqs ($undulatorPV) {
sub getIDEnergyRqs ($) {

### This routine returns requested Undulator energy

  my ($script, $undulatorPV, $status, $energy);
  $script      = 'getIDEnergyRqs';
  $undulatorPV = $_[0];

### Get Requested undulator Energy
  ($status,$energy) = Pezca::GetDouble($undulatorPV.':EnergySet');
  if ($status) {
     &report_only($script, $status, $undulatorPV.':EnergySet');
     return 0.0;
  } else {
     return $energy;
  }
}

#=================================================================

# sub getIDGap ($undulatorPV) {
sub getIDGap ($) {

### This routine returns actual Undulator gap

  my ($script, $undulatorPV, $status, $gap);
  $script      = 'getIDGap';
  $undulatorPV = $_[0];

### Get Actual undulator Gap
  ($status,$gap) = Pezca::GetDouble($undulatorPV.':Gap');
  if ($status) {
     &report_only($script, $status, $undulatorPV.':Gap');
     return 0.0;
  } else {
     return $gap;
  }
}

#=================================================================

# sub getIDHarmonic ($undulatorPV) {
sub getIDHarmonic ($) {

### This routine returns actual Undulator harmonic

  my ($script, $undulatorPV, $status, $harmonic);
  $script      = 'getIDHarmonic';
  $undulatorPV = $_[0];

  if (!defined $undulatorPV) {$undulatorPV = '';}
  if (length($undulatorPV) == 0) {
     print STDOUT '!!! '.$script.': no ID PV. Harmonic is set to 1'."\n\n";
     return 1;
  }
### Get Actual undulator Harmonic
  ($status,$harmonic) = Pezca::GetDouble($undulatorPV.':HarmonicValue');
  if ($status) {
     &report_only($script, $status, $undulatorPV.':HarmonicValue');
     return 0;
  } else {
     return $harmonic;
  }
}

#=================================================================

# sub getIDGapRqs ($undulatorPV) {
sub getIDGapRqs ($) {

### This routine returns requested Undulator gap

  my ($script, $undulatorPV, $status, $gap);
  $script      = 'getIDGapRqs';
  $undulatorPV = $_[0];

### Get Requested undulator Gap
  ($status,$gap) = Pezca::GetDouble($undulatorPV.':GapSet');
  if ($status) {
     &report_only($script, $status, $undulatorPV.':GapSet');
     return 0.0;
  } else {
     return $gap;
  }
}

#=================================================================

# sub setIDHarmonic ($undulatorPV, $harmonic) {
sub setIDHarmonic ($$) {

### This routine sets requested Undulator harmonic

  my ($script, $undulatorPV, $harmonic, @harmonicList, $status);
  $script      = 'setIDHarmonic';
  $undulatorPV = shift(@_);
  $harmonic    = shift(@_);
  $status = 0;
  @harmonicList = (1, 3, 5);

  if (defined $harmonic) {
     if (!(grep(/^$harmonic$/, (@harmonicList)))) {
        print STDOUT '*** '.$script.': incorrect ID harmonic = '.$harmonic."\n";
        print STDOUT '*** '.$script.': ID harmonic must be one of '."(@{harmonicList})\n";
        return -1;
     }
     $status = Pezca::PutDoubleOldCa($undulatorPV.':HarmonicValue',$harmonic);
     &report_only($script, $status, $undulatorPV.':HarmonicValue');
  }
  return $status;
}

#=================================================================

# sub isIDwritable($undulatorPV) {              #0=no,1=yes,2="no APS shutter permit",-1=error
sub isIDwritable ($) {

### This routine checks if Undulator is writable from given computer:

  my ($script, $undulatorPV, $harmonic, $status, $statusDescr);
  my ($permitPV, $permit);
  $script      = 'isIDwritable';
  $undulatorPV = shift(@_);

  $permitPV = 'ACIS:ShutterPermit';
  ($status,$permit) = Pezca::GetDouble($permitPV);			# 1=permit, 0=no permit
  if ($status) {
     &report_only($script, $status, $permitPV);
     $permit = 0;  
     return -1;								# error
  }
  elsif ($permit != 1) {
     print STDOUT '*** '.$script.': no APS shutter permit at the time.'."\n";
     return 2;								# no APS shutter permit
  }

  $harmonic = &getIDHarmonic($undulatorPV);
  if ($harmonic == 0) {return -1;}
  $status = &setIDHarmonic($undulatorPV,$harmonic);
  if ($status == 3) {
     $statusDescr = &ezcaStatusCode($status);
     print STDOUT '*** '.$script.': no write access to ID='.$undulatorPV
                        .' (status='.$status.', "'.$statusDescr.'")'."\n";
     return 0;								# ID not writable
  }
  elsif ($status != 0) {
     return -1;								# error
  }
  else {
     return 1;								# ID is writable
  }
}

#=================================================================

# sub setID ($undulatorPV, $energy, [$harmonic]) {
sub setID ($$;$) {

### This routine sets requested Undulator energy  (and harmonic, if given)

  my ($script, $undulatorPV, $energy, $harmonic, $status);
  $script      = 'setID';
  $undulatorPV = shift(@_);
  $energy      = shift(@_);
  $harmonic    = shift(@_);
  $status = 0;

  if (defined $harmonic) {
     $status = &setIDHarmonic ($undulatorPV, $harmonic);
     if ($status) {return $status;}
  }
  $status = Pezca::PutDoubleOldCa($undulatorPV.':EnergySet',$energy);
  &report_only($script, $status, $undulatorPV.':EnergySet');
  return $status;
}

#=================================================================

# sub startID ($undulatorPV, $energy, [$harmonic]) {
sub startID ($$;$) {

### This routine starts Undulator motion

  my ($script, $undulatorPV, $energy, $harmonic, $status);
  $script      = 'startID';
  $undulatorPV = shift(@_);
  $energy      = shift(@_);
  $harmonic    = shift(@_);

  $status = &setID($undulatorPV,$energy,$harmonic);
  if ($status) {return $status;}

  $status = Pezca::PutDoubleOldCa($undulatorPV.':Start',1);
  &report_only($script, $status, $undulatorPV.':Start');
  return $status;
}

#=================================================================

# sub trackID ($undulatorPV) {
sub trackID ($) {

### This routine tracks Undulator motion and returns status

  our ($IGNORE_ERRORS, $wait_max, $nonstop);
  my ($script, $undulatorPV, $status, $ignore_errors_saved);
  my ($diff_max, $wait_step, $wait_, $arrived, $busy);
  my ($current_time, $energyAct, $energyRqs, $diff);
  $script      = 'trackID';
  $undulatorPV = $_[0];

  if (!defined $wait_max)      {$wait_max      = 60;}
  if (!defined $nonstop)       {$nonstop       = 0;}
  if (!defined $IGNORE_ERRORS) {$IGNORE_ERRORS = 0;}

  $ignore_errors_saved = $IGNORE_ERRORS;
  $IGNORE_ERRORS = 1;			# ignore errors in this routine

  $diff_max  = 0.01;			# 10 eV
  $wait_step = 0.1;			# 100 ms
  $wait_     = 0;
  $arrived   = 0;

  $status = &addCaMonitor($undulatorPV.':Busy');
  if ($status) {
     $IGNORE_ERRORS = $ignore_errors_saved;
     return $status;
  }

  while ($arrived < 3 && $wait_ < $wait_max) {
     select(undef,undef,undef,$wait_step);	# sleep($wait_step);
     $wait_ += $wait_step;
     ($status,$busy) = Pezca::GetDouble($undulatorPV.':Busy');
     if ($busy != 1) {$arrived++;}
     else            {$arrived=0;}
  }

### Check for timeout:
  if ($wait_ >= $wait_max) {
     $current_time = scalar localtime;
     print STDOUT '*** '.$script.': '.$current_time."\n";
     print STDOUT '*** '.$script.': '.$undulatorPV.' has not arrived after '.$wait_."s\n";
     if ($nonstop == 1) {
        print STDOUT '*** '.$script.': ... aborting motion and continuing script.'."\n";
        &abortID($undulatorPV);
        return 1;
     } else {
        &abortID($undulatorPV);
        &press_key_exit();
     }
  }

### Check for position:
  $energyAct = &getIDEnergy($undulatorPV);
  $energyRqs = &getIDEnergyRqs($undulatorPV);
  $diff = abs ($energyAct - $energyRqs);
  if ($diff > $diff_max) {
     $current_time = scalar localtime;
     print STDOUT '*** '.$script.': '.$current_time."\n";
     print STDOUT '*** '.$script.': '.$undulatorPV.':Energy: pos.mismatch = '.$diff."\n",
                  '*** '.$script.': requested = '.$energyRqs."\n",
                  '*** '.$script.': actual    = '.$energyAct."\n",
                  '*** '.$script.': allowance = '.$diff_max."\n";
     return 1;
  }

  return 0;
}

#=================================================================

# sub abortID ($undulatorPV) {
sub abortID ($) {

### This routine aborts Undulator motion

  my ($script, $undulatorPV, $status);
  $script      = 'abortID';
  $undulatorPV = $_[0];

  $status = Pezca::PutDoubleOldCa($undulatorPV.':Stop',1);
  &report_only($script, $status, $undulatorPV.':Stop');
  return $status;
}

#=================================================================

# sub runID ($undulatorPV, $energy, [$harmonic]) {
sub runID ($$;$) {

### This routine starts and tracks Undulator motion

  my ($script, $undulatorPV, $energy_id, $harmonic, $status);
  $script      = 'runID';
  $undulatorPV = shift(@_);
  $energy_id   = shift(@_);
  $harmonic    = shift(@_);

  $status = &startID($undulatorPV,$energy_id,$harmonic);
  if ($status) {return $status;}

  $status = &trackID($undulatorPV);
  if ($status) {return $status;}
}

#=================================================================

# sub E2gap ($undulatorPV, $energy, [$harmonic,$restore]) {
sub E2gap ($$;$$) {

### This routine returns Undulator gap for given energy

  my ($script, $undulatorPV, $energy, $harmonic, $restore);
  my ($status, $gap, $Enow, $Hnow);
  $script      = 'E2gap';
  $undulatorPV = shift(@_);
  $energy      = shift(@_);
  $harmonic    = shift(@_);
  $restore     = shift(@_);   if (!defined $restore) {$restore=0;}

  $Enow = &getIDEnergyRqs($undulatorPV);
  if (defined $harmonic) {$Hnow = &getIDHarmonic($undulatorPV);}

# print STDOUT 'Setting ID='.$undulatorPV.' to energy='.$energy."\n";
  $status = &setID($undulatorPV,$energy,$harmonic);
  if ($status) {return 0;}
  select(undef,undef,undef,0.4);
  $gap = &getIDGapRqs($undulatorPV);
# print STDOUT 'readback gap='.$gap."\n";

### Cleanup:
  if    ($restore =~ /^restore$/i ||
         $restore =~ /^cleaunp$/i ||
         $restore eq '1') {
     $status = &setID($undulatorPV,$Enow,$Hnow);
  }
  elsif ($restore ne '0') {
     print STDOUT '!!! '.$script.': unexpected restore parameter ['.$restore.']'."\n";
  }
  return $gap;
}

#=================================================================

# sub calcIDSpeed($undulatorPV,$beginEnergy,$endEnergy,$scanTime) {
sub calcIDSpeed ($$$$) {

# This routine calculates needed speed of undulator for on-the-fly-scans.
# It returns ($status,$beginGap,$endGap).
# +---------------------------------------------------------------+
# + ATTENTION: $beginEnergy,$endEnergy are UNDULATOR calibration, +
# + i.e. they should include undulator-to-mono offset.            +
# +---------------------------------------------------------------+

  my ($script, $undulatorPV, $beginEnergy, $endEnergy, $scanTime);
  my ($status, $beginGap, $endGap, $IDGAPMINSPEED, $IDGAPMAXSPEED);
  my ($currEnergy, $speedGap);
  $script      = 'calcIDSpeed';
  $undulatorPV = shift(@_);
  $beginEnergy = shift(@_);
  $endEnergy   = shift(@_);
  $scanTime    = shift(@_);

  $IDGAPMINSPEED = 0.01;			# mm per sec
  $IDGAPMAXSPEED = 16.;				# mm per sec

  $currEnergy = &getIDEnergyRqs($undulatorPV);	# remember requested ID energy
  if ($currEnergy <= 0.) {return 1, 0, 0;}
  $beginGap = &E2gap($undulatorPV,$beginEnergy);
  if ($beginGap <= 0.)   {return 1, 0, 0;}
  $endGap   = &E2gap($undulatorPV,$endEnergy);
  if ($endGap <= 0.)     {return 1, 0, 0;}
  $status = &setID($undulatorPV,$currEnergy);	# restore requested ID energy
  if ($status)           {return 1, 0, 0;}

# print STDOUT 'beginEnergy='.$beginEnergy
#            .'  endEnergy='.$endEnergy
#            .' currEnergy='.$currEnergy
#            .' beginGap='.$beginGap
#            .' endGap='.$endGap."\n\n";

  if ($scanTime <= 0) {
     print STDOUT '*** '.$script.': Zero motion time'."\n";
     $status = 1;
     $endGap = $beginGap;
     return $status,$beginGap,$endGap;
  }

  $speedGap = abs ($endGap - $beginGap) / $scanTime;

  if($speedGap == 0) {
     print STDOUT '*** '.$script.': resultant ID scan range is zero.'."\n";
     print STDOUT '*** '.$script.': Can this computer write into ID PVs?'."\n";
     $status = 1;
     return $status,$beginGap,$endGap;
  }
  elsif($speedGap < $IDGAPMINSPEED) {
     printf STDOUT '*** '.$script.': resultant   ID speed=%-8.3f'."\n",$speedGap;
     printf STDOUT '*** '.$script.': min.allowed ID speed=%-8.3f'."\n",$IDGAPMINSPEED;
     $scanTime = abs ($endGap - $beginGap) / $IDGAPMINSPEED;
     printf STDOUT '*** '.$script.': Decrease scan time to %-8.3fs'."\n",$scanTime;
     $status = 1;
     return $status,$beginGap,$endGap;
  }
  elsif($speedGap > $IDGAPMAXSPEED) {
     printf STDOUT '*** '.$script.': resultant   ID speed=%-8.3f'."\n",$speedGap;
     printf STDOUT '*** '.$script.': min.allowed ID speed=%-8.3f'."\n",$IDGAPMAXSPEED;
     $scanTime = abs ($endGap - $beginGap) / $IDGAPMAXSPEED;
     printf STDOUT '*** '.$script.': Increase scan time to %-8.3fs'."\n",$scanTime;
     $status = 1;
     return $status,$beginGap,$endGap;
  }
  else  {
     $status = 0;
     return $status,$beginGap,$endGap;
  }
}

#=================================================================

# sub startIDScan($undulatorPV,$beginGap,$endGap,$scanTime) {
sub startIDScan($$$$) {

  my ($script, $undulatorPV, $beginGap, $endGap, $scanTime);
  my ($IDLATENCY, $ON, $OFF, $wait_step, $wait_max, $wait_);
  my ($prgState, $SCANNING, $status);
  $script      = 'startIDScan';
  $undulatorPV = shift(@_);
  $beginGap    = shift(@_);
  $endGap      = shift(@_);
  $scanTime    = shift(@_);

### Time to wait after ID scan started (data according to Mohan):
  $IDLATENCY = 2;

  $ON        = 1;
  $OFF       = 0;
  $wait_step = 0.05;			# 50 ms
  $wait_max  = 5;
  $wait_     = 0;
  $prgState  = 0;			# 0=waiting,1=premove,2=scanning,3=finishing
  $SCANNING  = 2;
  $status    = 0;

### Disable SyncScan mode (?? - copied from fast scans):
  $status = Pezca::PutDoubleOldCa($undulatorPV.':SyncScanMode',$OFF);
  &report_only($script, $status, $undulatorPV.':SyncScanMode');
  if ($status) {return 1;}

### Enter start Gap:
  $status = Pezca::PutDoubleOldCa($undulatorPV.':SSStartGap',$beginGap);
  &report_only($script, $status, $undulatorPV.':SSStartGap');
  if ($status) {return 1;}

### Enter end Gap:
  $status = Pezca::PutDoubleOldCa($undulatorPV.':SSEndGap',$endGap);
  &report_only($script, $status, $undulatorPV.':SSEndGap');
  if ($status) {return 1;}

### Enter synchronized scan time:
  $status = Pezca::PutDoubleOldCa($undulatorPV.':SSTime',$scanTime);
  &report_only($script, $status, $undulatorPV.':SSTime');
  if ($status) {return 1;}

### Enable SyncScan mode:
  $status = Pezca::PutDoubleOldCa($undulatorPV.':SyncScanMode',$ON);
  &report_only($script, $status, $undulatorPV.':SyncScanMode');
  if ($status) {return 1;}

### We setup a CA-monitor on the undulator Program State
### to see if the undulator is in the premove state yet or
### in the scanning state already. After that we need to
### wait for 2 more seconds for undulator to reach the
### starting point and then we can start scanning mono.

  $status = &addCaMonitor($undulatorPV.':SSState');

### Start SyncScan:
  $status = Pezca::PutDoubleOldCa($undulatorPV.':SSStart',1);
  &report_only($script, $status, $undulatorPV.':SSStart');
  if ($status) {return 1;}

### Wait until ID gets into scanning mode:
  do {
     select(undef,undef,undef,$wait_step);	# sleep($wait_step);
     $wait_  += $wait_step;
     ($status,$prgState) = Pezca::GetDouble($undulatorPV.':SSState');
  } while ($prgState != $SCANNING && $wait_ < $wait_max);

  if ($wait_ >= $wait_max) {
     print STDOUT '*** '.$script.': '.$undulatorPV.' did not reach SCANNING mode after '.$wait_."s\n";
     print STDOUT '*** '.$script.': ... aborting undulator and continuing script'."\n";
     &abortIDScan($undulatorPV);
     return 1;
  }

### At this point the ID has accelerated to the scanning mode.
### Now we need to wait for about 2 seconds (according to Mohan)
### while it reaches the starting energy position.
### After that we can start the scaler and the monochromator

  select(undef,undef,undef,$IDLATENCY);		# sleep;
  return 0;
}

#=================================================================

# sub abortIDScan($undulatorPV) {
sub abortIDScan($) {

  my ($script, $undulatorPV, $status, $OFF);
  $script      = 'abortIDScan';
  $undulatorPV = shift(@_);

  $status = 0;
  $OFF    = 0;

# Stop SyncScan:
  $status = Pezca::PutDoubleOldCa($undulatorPV.':Stop',1);
  &report_only($script, $status, $undulatorPV.':Stop');

### Disable SyncScan mode:
  $status = Pezca::PutDoubleOldCa($undulatorPV.':SyncScanMode',$OFF);
  &report_only($script, $status, $undulatorPV.':SyncScanMode');

  select(undef,undef,undef,0.1);		# sleep;
  return $status;
}

#=================================================================

# sub runMono($monoAxis, $monoAssy, $E_end, 
#            [$linkID, $undulatorPV, $IDoffsetEnd,			#if $IDoffset is specified, it overrides SQL data
#             $syncMode, $syncStep, $switchHarmonic,
#             $attenuation, $switchMono, 
#             \$enmono13, \@harmEn, \@harmNum, 
#             \@Elower, \@Eupper, \@Olower, \@Oupper]);

sub runMono($$$;$$$$$$$$\$\@\@\@\@\@\@) {

### This routine drives Mono to a given $energy and optionally drives
### undulator to the same postion plus $IDoffset. 
### It may also switch ID harmonics between 1,3,5,7
### and mono Bragg reflection between 111 and 333.
### Attenuation is maintained if passed.

  our ($beamline, $IGNORE_ERRORS, $diff_max, $report);
  my ($script, $monoAxisPV, $monoAssyPV, $energy, $linkID);
  my ($undulatorPV, $IDoffset, $syncMode, $syncStep);
  my ($IDharmonic, $attenuation, $syncStep_default);
  my ($IDharm0, $attenuation_resulted);
  my ($e_id, $motionTime, $diff_max_saved, $diff_max_energy);
  my ($energy_id, $energy0, $energy_id0, $IDoffset0);
  my ($e_, $runprg, $id_step, $nsteps, $i, $j, $status, $IDahead);
  my ($IDbusy, $wait_, $wait_step, $wait_max);
  my ($mostab_state, $mostab_mode);
  my ($switchMono, $switchID);
  my ($enmono13_ref, $harmEn_ref, $harmNum_ref);
  my ($Elower_ref, $Eupper_ref, $Olower_ref, $Oupper_ref);
  my (@events, $nevents, $nharm, $E_j, $E0_j);
  my ($step_j, $hkl, $IDoffsetEnd);

  $script       = 'runMono';
  $monoAxisPV   = shift(@_);
  $monoAssyPV   = shift(@_);
  $energy       = shift(@_);
  $linkID       = shift(@_);	if (!defined $linkID)      {$linkID = 1;}
# if ($linkID) {
  $undulatorPV  = shift(@_);	if (!defined $undulatorPV) {$linkID = 0;}
  $IDoffsetEnd  = shift(@_);							# overrides SQL data at end if used
  $syncMode     = shift(@_);							# noSync(or none)/multistep/IDsteps/IDscan/
  $syncStep     = shift(@_);							# default = 0.25 keV
  $switchID     = shift(@_);	if (!defined $switchID)    {$switchID = 1;}
# }
  $attenuation  = shift(@_);							#attenuation will be maintained if passed
  $switchMono   = shift(@_);	if (!defined $switchMono) {$switchMono = 0;}
### SQL data:
  $enmono13_ref = shift(@_);							#energy when mono needs to be switched (optional) 
  $harmEn_ref   = shift(@_);							#$E_low, 1to3, 3to5, 5to7
  $harmNum_ref  = shift(@_);							#  1       3    5     7
  $Elower_ref   = shift(@_);							#mono-ID offset data for harm. 1,3,5,7
  $Eupper_ref   = shift(@_);							#mono-ID offset data for harm. 1,3,5,7
  $Olower_ref   = shift(@_);							#mono-ID offset data for harm. 1,3,5,7
  $Oupper_ref   = shift(@_);							#mono-ID offset data for harm. 1,3,5,7

  if (!defined $IGNORE_ERRORS) {$IGNORE_ERRORS=0;}
  if (!defined $report)        {$report=0;}
  if (!defined $diff_max)      {$diff_max_saved = 1;}
  else                         {$diff_max_saved = $diff_max;}
  $diff_max_energy = 0.01;							# 10 eV
  $syncStep_default = 0.25;							# step in keV

  if (!defined $syncMode)           {$syncMode = 'noSync';}
  if ($syncMode =~ /^none$/i)       {$syncMode = 'noSync';}
  if ($syncMode !~ /^noSync$/i    &&
      $syncMode !~ /^multistep$/i &&
      $syncMode !~ /^IDsteps$/i   &&
      $syncMode !~ /^IDscan$/i)   {
     print STDOUT '!!! '.$script.': incorrect sync mode="'.$syncMode.'". Using "noSync".'."\n";
     $syncMode = 'noSync';
  }
  if (!defined $syncStep || $syncStep == 0) {
     $syncStep = $syncStep_default;						# step in keV
     if ($syncMode =~ /^multistep/i) {
        print STDOUT $script.'(multiStep): using default energy change step='.$syncStep.' keV'."\n";
     }
  }
  $syncStep = abs($syncStep);

  $status = &addCaMonitor($monoAssyPV.':RunPrg'); if ($status) {return $status;}
  $status = &addCaMonitor($monoAxisPV.':ActPos'); if ($status) {return $status;}
  if ($linkID) {
     $status = &addCaMonitor($undulatorPV.':Busy'); if ($status) {return $status;}
  }
#
  $energy0 = &getAxisAct($monoAxisPV, \$status);				# initial energy
  if ($status) {return $status;}

  @events = ();
  $events[0]->{'energy'} = $energy0;
  $events[0]->{'type'}   = 'beg';
  $events[0]->{'data'}   = '';
  $events[1]->{'energy'} = $energy;
  $events[1]->{'type'}   = 'end';
  $events[1]->{'data'}   = '';
  $nevents = @events;
  
  $status = &getIDparams($enmono13_ref,$harmEn_ref,$harmNum_ref,
                         $Elower_ref,$Olower_ref,$Eupper_ref,$Oupper_ref);	# a more-friendly wrapper around &getVariableIDoffsets
  if ($status) {return $status;}
  $nharm = @$harmEn_ref;
  ### Make the events array (events are changing ID harmonic or mono hkl or arriving at end)
  if ($linkID && $switchID) {
     for ($i=1; $i<$nharm; $i++) {
        if (($energy-@$harmEn_ref[$i])*(@$harmEn_ref[$i]-$energy0)>=0) {
           $events[$nevents]->{'energy'} = @$harmEn_ref[$i];
           $events[$nevents]->{'type'}   = 'harmonic';
           if ($energy >= $energy0) {$events[$nevents]->{'data'} = @$harmNum_ref[$i];}
           else                     {$events[$nevents]->{'data'} = @$harmNum_ref[$i-1];}
           $nevents++;
        }
     }
  }
  if ($switchMono && ($energy-$$enmono13_ref)*($$enmono13_ref-$energy0)>=0) {
     $events[$nevents]->{'energy'} = $$enmono13_ref;
     $events[$nevents]->{'type'}   = 'hkl';
     if ($energy >= $energy0) {$events[$nevents]->{'data'} = '333';}
     else                     {$events[$nevents]->{'data'} = '111';}
     $nevents++;
  }
  if ($energy >= $energy0) {@events = sort{$a->{'energy'} <=> $b->{'energy'}} @events;}
  else                     {@events = sort{$b->{'energy'} <=> $a->{'energy'}} @events;}
  if ($nevents > 2) {
     printf STDOUT $script.': energy change from %.3f keV to %.3f keV will have %d stop(s):'."\n",$energy0,$energy,($nevents-2);
     for ($j=1; $j<$nevents-1; $j++) {
       printf STDOUT '   at E=%.3f keV switching %8s to %s'."\n", $events[$j]->{'energy'}, $events[$j]->{'type'}, $events[$j]->{'data'};
     }
  }
  $hkl = &getMonoHKL();
  if ($linkID) {$IDharmonic = &getIDHarmonic($undulatorPV);}

  if ($nevents > 2 || $syncMode =~ /^multistep$/i || $syncMode =~ /^IDsteps$/i) {
     if (defined $beamline && defined &function_exists('getMostabState')) {
        $mostab_state = &getMostabState($beamline);			#0=run, 1=pause, -1=error
        $mostab_mode = &getMostabMode($beamline);				#0=software, 1=hybrid, 2=hardware
     } else {
        $mostab_state = -1;
        $mostab_mode = -1;
        if (!defined $beamline) {
           print STDOUT '!!! '.$script.': global variable "beamline" is undefined.'."\n";
        }
        if (!defined &function_exists('getMostabState')) {
           print STDOUT '!!! '.$script.': function "getMostabState" is undefined.'."\n"
                              .'Please source "mostab_subs.pl"'."\n";
        }
     }
  }

  for ($j=1; $j<$nevents; $j++) {  # = = = = = = = = = = = = = = = = = = = = = =# loop over harmonic and hkl changes

     ### End energy for this interval
     $E0_j = $events[$j-1]->{'energy'};
     $E_j  = $events[$j]->{'energy'};

     if ($linkID) { #- - - - ID LINKED - - - - - - - - - - - - - - - - - - - -

        ### ID offset at the end of this interval and previous intervals
        $IDoffset0 = &calcIDoffset($E0_j, $switchID,
                                   $enmono13_ref, $harmEn_ref, $harmNum_ref,
                                   $Elower_ref, $Eupper_ref, $Olower_ref, $Oupper_ref);
        $IDoffset  = &calcIDoffset($E_j, $switchID,
                                   $enmono13_ref, $harmEn_ref, $harmNum_ref,
                                   $Elower_ref, $Eupper_ref, $Olower_ref, $Oupper_ref);
#       printf STDOUT $script.': j=%d E0_j=%.3f IDoffset0=%.3f E_j=%.3f IDoffset=%.3f'."\n",$j,$E0_j,$IDoffset0,$E_j,$IDoffset;

        if ($syncMode =~ /^noSync$/i) { #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
           ### Start Mono:
           $status = &startMotor ($monoAxisPV, $E_j, $monoAssyPV, 'checkAuto');  # autoMode=forceAuto/checkAuto
           if ($status) {return $status;}
           ### Start Undulator:
           $energy_id = $E_j + $IDoffset;
           $status = &startID($undulatorPV,$energy_id);
           if ($status) {$linkID = 0;}
           ### wait for Mono to finish:
           $wait_step = 0.25;							# in seconds
           do {
              select(undef,undef,undef,$wait_step);				# sleep($wait_step);
              if (defined $attenuation) {
                 ($status,$e_) = Pezca::GetDouble($monoAxisPV.':ActPos');
                 if ($status) {$e_ = $E_j;}
                 ($status,$attenuation_resulted) = &setAttenuation2($attenuation,$e_,0);
                 if ($status) {
                    printf STDOUT "\n".'*** '.$script.': Failed to set attenuation to'
                                 .' %.2gx at energy=%.3f keV'."\n",$attenuation, $e_;
                    undef $attenuation;
                 }
              }
              ($status,$runprg) = Pezca::GetDouble($monoAssyPV.':RunPrg');
              if ($status) {$runprg = 0;}
           } while ($runprg != 0);
           ### Mono motion time presuming that ID does not slow things down:
           $motionTime = &calcMotorTime($monoAssyPV, $monoAxisPV, $E_j);
           $diff_max = $diff_max_energy;
           $status = &trackMotor ($monoAssyPV, $motionTime, $monoAxisPV);
           $diff_max = $diff_max_saved;
           if ($linkID == 1) {
              if ($status) {&abortID($undulatorPV);}
              else         {&trackID($undulatorPV);}
           }
        }
#
        elsif ($syncMode =~ /^multistep$/i) { #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#          print STDOUT $script.': running in multiStep mode'."\n";
           $nsteps = int(abs(($E_j - $E0_j)/$syncStep))+1;
           $step_j = ($E_j - $E0_j) / $nsteps;
           if ($mostab_state == 0) {						#0=run, 1=pause, -1=error
#             print STDOUT $script.'(multiStep): centering DAC on P2 piezo before energy change...'."\n";
              &mostabAutoShiftP2($beamline);					#center DAC before move
           }
           $id_step = ($IDoffset - $IDoffset0) / $nsteps;
           for ($i=1; $i<=$nsteps; $i++) {
              $e_ = $E0_j + $step_j*$i;
              ### Start Mono:
              $status = &startMotor ($monoAxisPV, $e_, $monoAssyPV, 'checkAuto');	#autoMode=forceAuto/checkAuto
              if ($status) {return $status;}
              ### Start Undulator:
              $e_id = $e_ + ($IDoffset0 + $id_step*$i);
              $status = &startID($undulatorPV,$e_id);
              if ($status) {$linkID = 0;}
              ### Mono motion time presuming that ID does not slow things down:
              $motionTime = &calcMotorTime($monoAssyPV, $monoAxisPV, $E_j);
              ### wait for Mono to finish:
              $diff_max = $diff_max_energy;
              $status = &trackMotor ($monoAssyPV, $motionTime, $monoAxisPV);
              $diff_max = $diff_max_saved;
              if ($linkID) {
                 if ($status) {&abortID($undulatorPV);}
                 else         {&trackID($undulatorPV);}
              }
              if (defined $attenuation) {
                 ($status,$attenuation_resulted) = &setAttenuation2($attenuation,$e_,0);
                 if ($status) {
                    printf STDOUT "\n".'*** '.$script.': Failed to set attenuation to'
                                 .' %.2gx at energy=%.3f keV'."\n",$attenuation, $e_;
                    undef $attenuation;
                 }
              }
           }
        }
#
        elsif ($syncMode =~ /^IDsteps$/i) { #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#          print STDOUT $script.': running in IDsteps mode'."\n";
           if ($mostab_state == 0) {
#             print STDOUT $script.'(IDsteps): centering DAC on P2 piezo before energy change...'."\n";
              &mostabAutoShiftP2($beamline);					#center DAC before move
           }
           $energy_id = $E_j + $IDoffset;
           $IDahead   = 0.1;							#in keV
           if ($E_j < $E0_j) {$IDahead = - $IDahead;}
           $wait_step = 0.02;							# 50 ms
           $wait_max  = 5;
           $status = &startMotor ($monoAxisPV, $E_j, $monoAssyPV, 'checkAuto'); # autoMode=forceAuto/checkAuto
           if ($status) {return $status;}
           do {
              ($status,$e_) = Pezca::GetDouble($monoAxisPV.':ActPos');
              ### Start Undulator:
              if ($E_j != $E0_j) {
                 ### This is planned ID position for the next step ($IDahead accounts for mono in motion)
                 $e_id = $e_ + $IDoffset0 + ($IDoffset-$IDoffset0)*($e_-$E0_j)/($E_j-$E0_j) + $IDahead;
              } else {
                 $e_id = $e_ + $IDoffset;
              }
              if    ($E_j >= $E0_j && $e_id > $energy_id) {$e_id = $energy_id;}
              elsif ($E_j <  $E0_j && $e_id < $energy_id) {$e_id = $energy_id;}
              printf STDOUT $script.'/IDsteps: running ID to %.3f keV'."\n",$e_id;
              $status = &startID($undulatorPV,$e_id);
              if ($status) {$linkID = 0;}
              if (defined $attenuation) {
                 ($status,$attenuation_resulted) = &setAttenuation2($attenuation,$e_,0);
                 if ($status) {
                    printf STDOUT "\n".'*** '.$script.': Failed to set attenuation to'
                                 .' %.2gx at energy=%.3f keV'."\n",$attenuation, $e_;
                    undef $attenuation;
                 }
              }
              $wait_ = 0;
              do {
                 select(undef,undef,undef,$wait_step);				# sleep($wait_step);
                 $wait_ += $wait_step;
                 ($status,$IDbusy) = Pezca::GetDouble($undulatorPV.':Busy');
              } while ($IDbusy && $wait_ < $wait_max);
              ($status,$runprg) = Pezca::GetDouble($monoAssyPV.':RunPrg');
              if ($status) {$runprg = 0;}
           } while ($runprg != 0);
           $motionTime = &calcMotorTime($monoAssyPV, $monoAxisPV, $E_j);
           $status = &trackMotor ($monoAssyPV, $motionTime, $monoAxisPV);
        }
#
        elsif ($syncMode =~ /^IDscan$/i) { #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
           $status = &runMonoIDscan($monoAxisPV, $monoAssyPV, $E_j,
                                    $undulatorPV, $IDoffset, $IDharmonic);
           if ($status) {
              print STDOUT '!!! '.$script.': changing energy with sync ID scan'
                                 .' failed. Switching to non-sync.'."\n";
              $syncMode = 'noSync';
              $status = &runMono($monoAxisPV, $monoAssyPV, $E_j, 
                                 $linkID, $undulatorPV, undef,			# $IDoffsetEnd is undef, use from SQL
                                 $syncMode, $syncStep, 0,			# $switchID=0
                                 $attenuation, 0,				# $switchMono=0
                                 $enmono13_ref, $harmEn_ref, $harmNum_ref, 
                                 $Elower_ref, $Eupper_ref, $Olower_ref, $Oupper_ref);
              if ($status) {return $status;}
           }
        }
#
     } else {  #- - - - ID IS NOT LINKED - - - - - - - - - - - - - - - - - - -
#
        $status = &startMotor ($monoAxisPV, $E_j, $monoAssyPV, 'checkAuto');	# autoMode=forceAuto/checkAuto
        if ($status) {return $status;}
        ### wait for Mono to finish:
        $wait_step = 0.25;							# in seconds
        do {
           select(undef,undef,undef,$wait_step);				# sleep($wait_step);
           if (defined $attenuation) {
              ($status,$e_) = Pezca::GetDouble($monoAxisPV.':ActPos');
              if ($status) {$e_ = $E_j;}
              ($status,$attenuation_resulted) = &setAttenuation2($attenuation,$e_,0);
              if ($status) {
                 printf STDOUT "\n".'*** '.$script.': Failed to set attenuation to'
                              .' %.2gx at energy=%.3f keV'."\n",$attenuation, $e_;
                 undef $attenuation;
              }
           }
           ($status,$runprg) = Pezca::GetDouble($monoAssyPV.':RunPrg');
           if ($status) {$runprg = 0;}
        } while ($runprg != 0);
        $motionTime = &calcMotorTime($monoAssyPV, $monoAxisPV, $E_j);
        $diff_max = $diff_max_energy;
        $status = &trackMotor ($monoAssyPV, $motionTime, $monoAxisPV);
        $diff_max = $diff_max_saved;
        if ($status) {return $status;}
     } #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

     ### After-interval-move operation (switch harmonic, switch hkl, final energy tweak)

     if ($linkID && $switchID && $events[$j]->{'type'} eq 'harmonic' && $events[$j]->{'data'} ne $IDharmonic) {
        ### Figure out ID harmonic change ###
        ### If Mostab is running, pause it for the step when harmonic is changed
        if ($mostab_state == 0) {					#if running
           if ($mostab_mode == 2) {&setIntensityFeedbackFPGAstate($beamline,'Paused','string');}
           else                   {&pauseMostab($beamline);}
        }
        printf STDOUT $script.' changing ID harmonic from %d to %d at energy=%.2f keV'."\n",
                                                 $IDharmonic,$events[$j]->{'data'},$E_j;
        $IDharmonic = $events[$j]->{'data'};
        $status = &setIDHarmonic($undulatorPV,$IDharmonic);
        $IDoffset = &calcIDoffset($E_j,0,
                                 $enmono13_ref,$harmEn_ref,$harmNum_ref,
                                 $Elower_ref,$Eupper_ref,$Olower_ref,$Oupper_ref);
        $status = &runID($undulatorPV,$E_j+$IDoffset);
        ### If feedback is not paused:
        if ($mostab_state == 0) {						#if running
           ### Restore Mostab after the step where we changed the harmonic
           if ($mostab_mode == 2) {&setIntensityFeedbackFPGAstate($beamline,'Active','string');}
           else                   {&restoreMostab($beamline); &waitMostabLock ($beamline);}
        }
     }
     elsif ($switchMono && $events[$j]->{'type'} eq 'hkl' && $events[$j]->{'data'} ne $hkl) {
        ### Figure out Bragg HKL change ###
        ### If Mostab is running, pause it for the step when harmonic is changed
        if ($mostab_state == 0) {					#if running
           if ($mostab_mode == 2) {&setIntensityFeedbackFPGAstate($beamline,'Paused','string');}
           else                   {&pauseMostab($beamline);}
        }
        printf STDOUT $script.' changing Bragg reflection from %s to %s at energy=%.2f keV'."\n",
                                                 $hkl,$events[$j]->{'data'},$E_j;
        $hkl = $events[$j]->{'data'};
        $status = &setMonoHKL($hkl);
        if ($mostab_state == 0) {						#if running
           ### Restore Mostab after the step where we changed the harmonic
           if ($mostab_mode == 2) {&setIntensityFeedbackFPGAstate($beamline,'Active','string');}
           else                   {&restoreMostab($beamline); &waitMostabLock ($beamline);}
        }
     }
     elsif ($linkID && $events[$j]->{'type'} eq 'end') {
        if ($syncMode =~ /^IDsteps$/i || $syncMode =~ /^IDscan$/i) {
           ### Sync mono and ID at the destinations -- just in case ###
           $switchID   = 0;
           $switchMono = 0;
           $syncMode   = 'noSync';
           $syncStep   = 0;
           $status = &runMono($monoAxisPV, $monoAssyPV, $E_j, 
                              $linkID, $undulatorPV, $IDoffsetEnd,
                              $syncMode, $syncStep, $switchID,
                              $attenuation, $switchMono,
                              $enmono13_ref, $harmEn_ref, $harmNum_ref, 
                              $Elower_ref, $Eupper_ref, $Olower_ref, $Oupper_ref);
           if ($status) {return $status;}
        }
        elsif (defined $IDoffsetEnd) {
           $status = &runID($undulatorPV,$E_j+$IDoffsetEnd);
           if ($status) {return $status;}
        }
     }
  } # = = = = = = = = = = = = = = = = = = end of loop over events
  return $status;
}

#=================================================================

# sub runMonoIDscan($monoAxisPV,$monoAssyPV,$energy,$undulatorPV,$IDoffset,[$IDharmonic]) {
sub runMonoIDscan($$$$$;$) {

### This routine drives Mono to a given $energy and simultaneously scans
### undulator to the same postion plus $IDoffset. The undulator speed is
### calculated in such a way that both mono and undulator shoud arrive to
### their destinations simultaneously

  our ($IGNORE_ERRORS, $diff_max);
  my ($script, $monoAxisPV, $monoAssyPV, $energy);
  my ($undulatorPV, $IDoffset, $IDharmonic, $status);
  my ($motionTime, $diff_max_saved, $diff_max_energy);
  my ($id0, $id1, $beginGap, $endGap);
  $script      = 'runMonoIDscan';
  $monoAxisPV  = shift(@_);
  $monoAssyPV  = shift(@_);
  $energy      = shift(@_);
  $undulatorPV = shift(@_);
  $IDoffset    = shift(@_);
  $IDharmonic  = shift(@_);

# if (!defined $IDoffset)      {$IDoffset       = 0;}
  if (!defined $IGNORE_ERRORS) {$IGNORE_ERRORS  = 0;}
  if (!defined $diff_max)      {$diff_max_saved = 1;}
  else                         {$diff_max_saved = $diff_max;}
  $diff_max_energy = 0.01;					# 10ev
  $status = 0;

  if (defined $IDharmonic) {
     $status = &setIDHarmonic($undulatorPV,$IDharmonic);
     if ($status) {return $status;}
  }
  $motionTime = &calcMonoTime($monoAssyPV, $energy);
  $id0 = &getIDEnergy($undulatorPV);
  $id1 = $energy + $IDoffset;
  ($status,$beginGap,$endGap) = &calcIDSpeed($undulatorPV,$id0,$id1,$motionTime);
  if ($status) {return $status;}
  $status = &startIDScan($undulatorPV,$beginGap,$endGap,$motionTime);
  if ($status) {return $status;}
  $diff_max = $diff_max_energy;
  $status = &runMotor ($monoAssyPV, $monoAxisPV, $energy, 'checkAuto');		# autoMode=forceAuto/checkAuto
  $diff_max = $diff_max_saved;							# restore
  if ($status) {
     &abortIDScan($undulatorPV);
     return $status;
  }
  $status = &trackID($undulatorPV);
  $status = &abortIDScan($undulatorPV);
  return $status;
}

#=================================================================

# sub getVariableIDoffsets(\%EnergiesIJ_and_Offsets_Energies) {		#fills array of hashes, contains $records[$i]->{name}->data
#                                                            		# Possible names (not all may be defined):
#                                                            		# h13,h35,h57,m13
#                                                            		# eLo1,eUp1,oLo1,oUp1
#                                                            		# eLo3,eUp3,oLo3,oUp3
#                                                            		# eLo5,eUp5,oLo5,oUp5
#                                                            		# eLo7,eUp7,oLo7,oUp7
sub getVariableIDoffsets(\%) {

# use DBI;

### Usage:
# $status = &getVariableIDoffsets(\@energy13_and_Offsets_Energies);

  our ($report, $mysqlHost, $mysqlUser, $mysqlPass, $mysqlDb);
  my ($script, $hashRef, $mysqlTbl, @ids, $status);
  my ($dsn, $dbh, $sth, $rawref, $id, $val, $i);
  my ($idmap, $h, $h1, $h2);

  $script    = 'getVariableIDoffsets';
  $hashRef   = shift(@_);
  if (!defined $report) {$report=0;}

  $mysqlTbl  = 'Config';
  @ids       =('offset.h3h1Cutoff(keV)',		# energy switching ID 1 -> 3		#h13
               'offset.h5h3Cutoff(keV)',		# energy switching ID 3 -> 5		#h35
               'offset.h7h5Cutoff(keV)',		# energy switching ID 5 -> 7		#h57
               'offset.111to333(keV)',			# energy switching mono 111 to 333	#m13
               'offset.h1_eLo(keV)',			#-- E_lower         (1st ID harmonic)	#eLo1
               'offset.h1_eUp(keV)',			# | E_upper         (1st ID harmonic)	#eUp1
               'offset.h1_oLo(keV)',			# | Offset(E_lower) (1st ID harmonic)	#oLo1
               'offset.h1_oUp(keV)',			#-- Offset(E_upper) (1st ID harmonic)	#oUp1
               'offset.h3_eLo(keV)',			#--					#eLo3
               'offset.h3_eUp(keV)',			# | same for 3rd ID harmonic		#eUp3
               'offset.h3_oLo(keV)',			# |					#oLo3
               'offset.h3_oUp(keV)',			#--					#oUp3
               'offset.h5_eLo(keV)',			#--					#eLo5
               'offset.h5_eUp(keV)',			# | same for 5th ID harmonic		#eUp5
               'offset.h5_oLo(keV)',			# |					#oLo5
               'offset.h5_oUp(keV)',			#--					#oUp5
               'offset.h7_eLo(keV)',			#--					#eLo7
               'offset.h7_eUp(keV)',			# | same for 7th ID harmonic		#eUp7
               'offset.h7_oLo(keV)',			# |					#oLo7
               'offset.h7_oUp(keV)');			#--					#oUp7

### Read the parameters from the JBluIce MySQL database:

# print STDOUT 'mysqlHost='.$mysqlHost.' mysqlDb='.$mysqlDb."\n";
# print STDOUT 'mysqlUser='.$mysqlUser.' mysqlPass='.$mysqlPass."\n";

  $dsn = 'DBI:mysql:host='.$mysqlHost.';database='.$mysqlDb;
  $dbh = DBI->connect($dsn,$mysqlUser,$mysqlPass,{PrintError=>0,});
  if (!defined $dbh) {
     print STDOUT '*** '.$script.': Cannot connect to mySQL server '.$mysqlHost."\n";
     $status = -1;
     return $status;
  }
  foreach $id (@ids) {
     $idmap = $id;
     $idmap =~ s/^offset\.//;
     if ($idmap =~ /^h\d_.*\(keV\)$/) {			#h5_eUp(keV) or h5_oUp(keV)
        $idmap =~ s/\(keV\)$//;				#remove (keV) at the end
        $h = substr $idmap,1,1;				#ID harmonic number
        $idmap =~ s/^h\d_//;				#remove h1_ or h3_ or h5_ or h7_
        $idmap .= $h;					#append harmnic number
     }
     elsif ($idmap =~ /Cutoff\(keV\)$/) {		#h3h1Cutoff(keV)
        $idmap =~ s/Cutoff\(keV\)$//;			#remove Cutoff(keV) at the end
        $h1 = substr $idmap,3,1;
        $h2 = substr $idmap,1,1;
        $idmap = 'h'.$h1.$h2;
     }
     elsif ($idmap =~ /^111to333\(keV\)$/) {
        $idmap = 'm13';
     }
     else {
        print STDOUT '*** '.$script.': Cannot map mySQL ID='.$id.' into internal name.'."\n";
        $status = -1;
        return $status;
     }
     $sth = $dbh->prepare('SELECT * FROM '.$mysqlTbl.' WHERE ID="'.$id.'"');
     $status = $sth->execute();				#status is undefined in case of error
     if (!defined $status) {
        print STDOUT '*** '.$script.': Failed executing mySQL SELECT on ID='.$id
                                          .' in table='.$mysqlTbl
                                   .' on mySQL server='.$mysqlHost."\n"
                    .'*** '.$script.': '.$sth->errstr."\n";
        $status = -1;
        return $status;
     }
     $rawref = $sth->fetchrow_hashref();
     if (!defined $rawref) {
        if ($report) {
           print STDOUT '!!! '.$script.': Cannot find ID="'.$id.'" in table "'
                              .$mysqlTbl.'" on mySQL server '.$mysqlHost."\n";
        }
        next;
     }
     if (! defined $rawref->{'Value'}) {
        if ($report) {
           print STDOUT '!!! '.$script.': No value for ID="'.$id.'" in table "'
                              .$mysqlTbl.'" on mySQL server '.$mysqlHost."\n";
        }
        next;
     }
     $val = $rawref->{'Value'};
     if ($report) {print STDOUT 'id='.$id.' val='.$val."\n";}
     $$hashRef{$idmap} = $val;
  }
  $sth->finish();
  $dbh->disconnect();
  $status = 0;
  return $status;
}

#=================================================================

# $status = &getIDparams(\$enmono13,\@harmEn,\@harmNum,
#                        \@Elower,\@Olower,\@Eupper,\@Oupper);			# a more-friendly wrapper around &getVariableIDoffsets
sub getIDparams(\$\@\@\@\@\@\@) {
  my ($script, $status, $data_passed, %IDdata);
  my ($enmono13_ref, $enmono13, $i);
  my ($harmEn_ref, $harmNum_ref, $Elower_ref, $Eupper_ref, $Olower_ref, $Oupper_ref);
  my (@harmEn,     @harmNum,     @Elower,     @Eupper,     @Olower,     @Oupper);

  $script = 'getIDparams';
  $enmono13_ref = shift(@_);							#energy when mono needs to be switched between 111 and 333(optional) 
  $harmEn_ref   = shift(@_);							#$E_low, 1to3, 3to5, 5to7
  $harmNum_ref  = shift(@_);							#  1       3    5     7
  $Elower_ref   = shift(@_);							#mono-ID offset data for harm. 1,3,5,7
  $Olower_ref   = shift(@_);							#mono-ID offset data for harm. 1,3,5,7
  $Eupper_ref   = shift(@_);							#mono-ID offset data for harm. 1,3,5,7
  $Oupper_ref   = shift(@_);							#mono-ID offset data for harm. 1,3,5,7

  $enmono13 = 0;
  @harmEn   = (0);
  @harmNum  = (1);
  @Elower   = ();
  @Eupper   = ();
  @Olower   = ();
  @Oupper   = ();
  $data_passed = 1;

  if (defined $harmEn_ref  && @$harmEn_ref  > 0) {@harmEn   = @$harmEn_ref;}   else {$data_passed = 0;}
  if (defined $harmNum_ref && @$harmNum_ref > 0) {@harmNum  = @$harmNum_ref;}  else {$data_passed = 0;}
  if (defined $Elower_ref  && @$Elower_ref  > 0) {@Elower   = @$Elower_ref;}   else {$data_passed = 0;}
  if (defined $Eupper_ref  && @$Eupper_ref  > 0) {@Eupper   = @$Eupper_ref;}   else {$data_passed = 0;}
  if (defined $Olower_ref  && @$Olower_ref  > 0) {@Olower   = @$Olower_ref;}   else {$data_passed = 0;}
  if (defined $Oupper_ref  && @$Oupper_ref  > 0) {@Oupper   = @$Oupper_ref;}   else {$data_passed = 0;}
  if (defined $enmono13_ref)                     {$enmono13 = $$enmono13_ref;} else {$data_passed = 0;}
# print STDOUT $script.': data passed = '.$data_passed."\n";
  if (! $data_passed) {
     ### If no harmonic switch energies are passed, request from SQL 'blc2004/Config' table:
     $status = &getVariableIDoffsets(\%IDdata);
     if ($status) {
        print STDOUT '*** '.$script.': cannot read mono-ID data from SQL'."\n";
        return $status;
     }
     if (defined $IDdata{'h13'} && $IDdata{'h13'}>0) {push @harmEn,$IDdata{'h13'}; push @harmNum,3;}
     if (defined $IDdata{'h35'} && $IDdata{'h35'}>0) {push @harmEn,$IDdata{'h35'}; push @harmNum,5;}
     if (defined $IDdata{'h57'} && $IDdata{'h57'}>0) {push @harmEn,$IDdata{'h57'}; push @harmNum,7;}
     foreach $i (qw(1 3 5 7)) { 						#0=1, 1=3, 2=5, 3=7
        if (defined $IDdata{'eLo'.$i} && defined $IDdata{'oLo'.$i} &&
            defined $IDdata{'eUp'.$i} && defined $IDdata{'oUp'.$i}) {
           push (@Elower,$IDdata{'eLo'.$i});
           push (@Olower,$IDdata{'oLo'.$i});
           push (@Eupper,$IDdata{'eUp'.$i});
           push (@Oupper,$IDdata{'oUp'.$i});
        } else {
           push (@Elower,0);
           push (@Olower,0);
           push (@Eupper,0);
           push (@Oupper,0);
        }
     }
     if (defined $IDdata{'m13'}) {$enmono13 = $IDdata{'m13'};} else {$enmono13 = 0;}
     $$enmono13_ref = $enmono13;							#energy when mono needs to be switched (optional) 
     @$harmEn_ref   = @harmEn;							#$E_low, 1to3, 3to5, 5to7
     @$harmNum_ref  = @harmNum;							#  1       3    5     7
     @$Elower_ref   = @Elower;							#mono-ID offset data for harm. 1,3,5,7
     @$Eupper_ref   = @Eupper;							#mono-ID offset data for harm. 1,3,5,7
     @$Olower_ref   = @Olower;							#mono-ID offset data for harm. 1,3,5,7
     @$Oupper_ref   = @Oupper;							#mono-ID offset data for harm. 1,3,5,7
  }
  return 0;
}

#=================================================================

# sub calcIDoffset($energy, $switchID, 
#                  \$enmono13, \@harmEn, \@harmNum, 
#                  \@Elower, \@Eupper, \@Olower, \@Oupper) {
sub calcIDoffset($;$\$\@\@\@\@\@\@) {

### Usage:
# $offset = &calcIDoffset($energy,$switchID,
#                         \$enmono13,\@harmEn,\@harmNum,\@Elower,\@Eupper,\@Olower,\@Oupper);

  our ($undulator);
  my ($script, $energy, $switchID, $offset, $status, $IDharmonic);
  my ($enmono13_ref, $harmEn_ref, $harmNum_ref);
  my ($Elower_ref, $Eupper_ref, $Olower_ref, $Oupper_ref);
  my ($nharm, $i, $k);
  $script    = 'calcIDoffset';
  $energy       = shift(@_);
  $switchID     = shift(@_); if (!defined $switchID) {$switchID=1;}
  $enmono13_ref = shift(@_);							#energy when mono needs to be switched (optional) 
  $harmEn_ref   = shift(@_);							#$E_low, 1to3, 3to5, 5to7
  $harmNum_ref  = shift(@_);							#  1       3    5     7
  $Elower_ref   = shift(@_);							#mono-ID offset data for harm. 1,3,5,7
  $Eupper_ref   = shift(@_);							#mono-ID offset data for harm. 1,3,5,7
  $Olower_ref   = shift(@_);							#mono-ID offset data for harm. 1,3,5,7
  $Oupper_ref   = shift(@_);							#mono-ID offset data for harm. 1,3,5,7

### I case mySQL data NOT passed:
  $status = &getIDparams($enmono13_ref,$harmEn_ref,$harmNum_ref,
                         $Elower_ref,$Olower_ref,$Eupper_ref,$Oupper_ref);	# a more-friendly wrapper around &getVariableIDoffsets
  if ($status) {
     print STDOUT '*** '.$script.': error getting ID params from SQL.'."\n";
     return 0;									#return zero ID offset
  }

  if ($switchID) {
     ### Find at what harmonic we are going to be at $energy
     $IDharmonic = @$harmNum_ref[0];
     $nharm = @$harmEn_ref;							#0th element is E=0
     for ($i=1; $i<$nharm; $i++) {     
        if (@$harmEn_ref[$i] > 0 && $energy < @$harmEn_ref[$i]) {
           $IDharmonic = @$harmNum_ref[$i-1];
           last;
        }
     }
  } else {
     ### Get and use current harmonic
     if (!defined $undulator) {return 0;}
     $IDharmonic = &getIDHarmonic($undulator);
  }
  $i = int(($IDharmonic+1)/2 - 1);						#$i: 0=h1, 1=h3, 2=h5, 3=h7
  if ($i < @$Eupper_ref) {
     if (@$Eupper_ref[$i] != @$Elower_ref[$i]) {
           $k = (@$Oupper_ref[$i] - @$Olower_ref[$i]) 
              / (@$Eupper_ref[$i] - @$Elower_ref[$i]);
           $offset = @$Olower_ref[$i] + ($energy - @$Elower_ref[$i]) * $k;
        } else {
           $offset  = 0.5 * (@$Olower_ref[$i] + @$Oupper_ref[$i]);
        }
  } else {
     $offset  = 0;
  }
  return $offset;
}

#=================================================================

# sub getLaneOffsets(\@laneVLenergy, \@laneVLposition, \@laneVLmaterial,
#                    \@laneHDenergy, \@laneHDposition, \@laneHDmaterial) {
sub getLaneOffsets(\@\@\@\@\@\@) {

# use DBI;

### Usage:
# $status = getLaneOffsets(\@laneVLenergy, \@laneVLposition, \@laneVLmaterial,
#                          \@laneHDenergy, \@laneHDposition, \@laneHDmaterial);

  our ($mysqlHost, $mysqlUser, $mysqlPass, $mysqlDb);
  my ($script,   $VLenergy, $VLposition, $VLmaterial);
  my ($mysqlTbl, $HDenergy, $HDposition, $HDmaterial);
  my ($dsn, $dbh, $sth, $rawref);
  my ($id, $val, $count, $status);
  $script     = 'getLaneOffsets';
  $VLenergy   = shift(@_);
  $VLposition = shift(@_);
  $VLmaterial = shift(@_);
  $HDenergy   = shift(@_);
  $HDposition = shift(@_);
  $HDmaterial = shift(@_);

  $mysqlTbl   = 'EnergyMove';

### Read the parameters from the JBluIce MySQL database:

# print STDOUT 'mysqlHost='.$mysqlHost.' mysqlDb='.$mysqlDb."\n";
# print STDOUT 'mysqlUser='.$mysqlUser.' mysqlPass='.$mysqlPass."\n";

  $dsn = 'DBI:mysql:host='.$mysqlHost.';database='.$mysqlDb;
  $dbh = DBI->connect($dsn,$mysqlUser,$mysqlPass,{PrintError=>0,});
  if (!defined $dbh) {
     print STDOUT '*** '.$script.': Cannot connect to mySQL server '.$mysqlHost."\n";
     return -1;
  }
  $sth = $dbh->prepare('SELECT * FROM '.$mysqlTbl);
  $status = $sth->execute();			#status is undefined in case of error
     if (!defined $status) {
        print STDOUT '*** '.$script.': Failed executing mySQL "SELECT *"'
                                          .' on table='.$mysqlTbl
                                     .', mySQL server='.$mysqlHost."\n"
                    .'*** '.$script.': '.$sth->errstr."\n";
        return -1;
     }

  $count = 0;
  no strict 'refs';
  while ($rawref = $sth->fetchrow_hashref()) {
     $id = $rawref->{'id'};
     if ($id =~ /^laneVL/) {
        $val=$rawref->{'energy'};   if (defined $val) {push (@$VLenergy,$val);}
        $val=$rawref->{'position'}; if (defined $val) {push (@$VLposition,$val);}
        $val=$rawref->{'material'}; if (defined $val) {push (@$VLmaterial,$val);}
        $count++;
     }
     elsif ($id =~ /^laneHD/) {
        $val=$rawref->{'energy'};   if (defined $val) {push (@$HDenergy,$val);}
        $val=$rawref->{'position'}; if (defined $val) {push (@$HDposition,$val);}
        $val=$rawref->{'material'}; if (defined $val) {push (@$HDmaterial,$val);}
        $count++;
     }
     else {
        print STDOUT '!!! '.$script.': unexpected record id='.$id
                           .' in table='.$mysqlTbl
                           .' at mySQL host='.$mysqlHost."\n";
     }
  }
# print STDOUT "There were $count rows returned.\n";
  if ($count == 0) {
     print STDOUT '*** '.$script.': Cannot find table '.$mysqlTbl
                        .' on mySQL server '.$mysqlHost."\n";
     return -1;
  }
  $sth->finish();
  $dbh->disconnect();
  return 0;
}

#=================================================================

# sub calcMonoTime($monoAssy, $energy) (		#returns calculated $motionTime to $energy
sub calcMonoTime($$) {

  our ($report, $MONO_ENERGY, $MONO_LAMBDA, $MONO_THETA);
  my ($script, $monoAssy, $energy, $status, $motionTime);
  my (@axes, @drives, @motors, @theta, $speed, $ESinTheta);
  my ($accelTimeAssy, $accelTimeAxis, $accelTime);

  $monoAssy = shift(@_);
  $energy   = shift(@_);
  if (!defined $report) {$report=0;}
  $motionTime = 0;

  $script = 'calcMonoTime';
  $status = &getAssyDataFromTsub($monoAssy,\@axes,\@drives,\@motors);
  if ($status) {return -abs($status);}
  $MONO_ENERGY=$axes[0]; $MONO_LAMBDA=$axes[1]; $MONO_THETA=$drives[0];


### This will work, but it is a bit of overkill and contains hardcoded
### mono energy & theta EPICS PVs
# $motionTime = &calcMotorTime ($monoAssy, $axes[0], $energy);
# if ($report) {printf STDOUT 'Mono Motion time (method-1) = %.3f'."\n",$motionTime;}

  $ESinTheta = &getESinTheta($monoAssy);
  $theta[0] = &getAxisAct($MONO_THETA,\$status);
  if ($status) {return -abs($status);}
  $theta[1] = &E2theta($monoAssy,$energy,$ESinTheta);

  $speed = &getAxisSpeed($monoAssy, $MONO_THETA);
### Calculate motion time for Theta:
  if ($speed) {$motionTime = abs(($theta[1] - $theta[0]) / $speed);}
  else        {return -1;}

### Query assembly acceleration time:
  $accelTimeAssy = &getAssyAccelTime($monoAssy, \$status);

### Query axis acceleration time:
  $accelTimeAxis = &getMtrAccelTime($motors[0], \$status);

  $accelTime = 0.002*&maxval(($accelTimeAssy,$accelTimeAxis));

  if ($report) {printf STDOUT 'Mono Motion time (method-2) = %.3f'."\n",$motionTime;}
  $motionTime += $accelTime;
### ATTENTION: in principle, we should also
###            add the backlash motion time!
### Add backlash acceleration:
  $motionTime += $accelTime;		#recalculate $accelTime from milliseconds

  if ($report) {printf STDOUT 'Mono Motion time (method-2+accel) = %.3f'."\n",$motionTime;}
  return $motionTime;
}

#=================================================================

# $temp = &mono_temperature($E,$E13);				#also uses global $beamline
sub mono_temperature($$) {
  our ($beamline);
  my ($script, $E, $E_tr, $T);
  my ($f11, $f12, $f13,  $f21, $f22, $f23);

  $script = 'mono_stabilization_temperature';

  $E    = shift(@_);						#ID energy
  $E_tr = shift(@_);						#energy of transition between 1st & 3rd ID harmonics

  if (!defined $E_tr || $E_tr == 0) {return 0;}
  if (!defined $beamline) {die $script.': undefined beamline';}
  if ($beamline !~ /^23[io]{1}$/) {return 0;}			#we only have data for 23o or 23i

### We should move this into mySQL:
  if    ($beamline eq '23i') {
     ($f11, $f12, $f13) = (-0.0537550, 0.45428, 86.042);	#copied from IDB measurements
     ($f21, $f22, $f23) = (-0.0081632, 0.14274, 87.913);	#copied from IDB measurements
#    ($f11, $f12, $f13) = (0., 0., 0.);
#    ($f21, $f22, $f23) = (0., 0., 0.);
  }
  elsif ($beamline eq '23o') {
     ($f11, $f12, $f13) = (-0.0537550, 0.45428, 86.042);	#measured for 1st harmonic
     ($f21, $f22, $f23) = (-0.0081632, 0.14274, 87.913);	#measured for 3rd harmonic
  }
  else {
      return 0;
  }

  if ($E < $E_tr) {$T = $f11*$E*$E + $f12*$E + $f13;}		#1st harmonic
  else            {$T = $f21*$E*$E + $f22*$E + $f23;}		#3rd harmonic
  return $T;
}

#=================================================================

# $time = &mono_stabilization_time($E_i, $E_f, $E13);		#also uses global $beamline
sub mono_stabilization_time($$$) {
  our ($beamline);
  my ($script, $E1, $E2, $E_tr, $T1, $T2);
  my ($tau, $sigma, $twait);

  $script = 'mono_stabilization_time';
  $E1   = shift(@_);						#initial ID energy
  $E2   = shift(@_);						#final ID energy
  $E_tr = shift(@_);						#energy of transition between 1st & 3rd ID harmonics

  if (!defined $beamline) {die $script.': undefined beamline';}
  if ($beamline !~ /^23[io]{1}$/) {return 0;}			#we only have data for 23o or 23i

### We should move this into mySQL as
### it may be beamline-specific:
  $tau   = 48.09;						#temperature relaxation rate constant (may be beamline-dependent)
  $sigma = 0.2;							#temperature error band (deviation from equlibrium when temperature can be condidered relaxed)

  $T1 = &mono_temperature($E1,$E_tr);
  $T2 = &mono_temperature($E2,$E_tr);
  if (abs($T2 - $T1) < $sigma) {return 0;}			#nothing to wait for

  $twait = $tau * log(abs($T2 - $T1)/$sigma);
  return $twait;
}

#=================================================================

# $hkl = &getMonoHKL();						#returns hkl string, e.g. 111 or 333
sub getMonoHKL() {
  our ($beamline);
  my ($script, $monoset_PV, $monoset, $status);
  my ($monodesc_PV, $monodesc);
  
  $script = 'getMonoHKL';
  if (!defined $beamline) {die $script.': undefined beamline';}
  if ($beamline !~ /^23[iob]{1}$/) {return 0;}			#we only have data for 23o, 23i, 23b

  $monoset_PV = $beamline.':MO:En:Chkl';
  $status = &addCaMonitor($monoset_PV);

  ($status,$monoset) = Pezca::GetString($monoset_PV);		#expect set1 to set9
  if ($status) {
     &report_only($script, $status, $monoset_PV);
     return '';
  }

  $monodesc_PV = $beamline.':MO:En:'.$monoset.'.DESC';
# $status = &addCaMonitor($monodesc_PV);			#monitors on DESC cannot be set

  ($status,$monodesc) = Pezca::GetString($monodesc_PV);		#expect "Si 111"
  if ($status) {
     &report_only($script, $status, $monodesc_PV);
     return '';
  }
  if    ($monodesc =~ / 111\s*$/) {return '111';}
  elsif ($monodesc =~ / 333\s*$/) {return '333';}
  else {
     print STDOUT '*** '.$script.' unrecognized reflection ['.$monodesc.']'."\n";
     return '';
  }
}

#=================================================================

# $status = &setMonoHKL($hkl);						#expects hkl string, e.g. 111 or 333
sub setMonoHKL() {
  our ($beamline);
  my ($script, $hkl, $status, $monoassy_PV, $i);
  my ($monoaxis_PV, $monoEn);
  my ($monoset_PV,  $monoset);
  my ($monodesc_PV, $monodesc);
  
  $script = 'setMonoHKL';
  $hkl    = shift(@_);
  
  if (!defined $hkl) {$hkl = '';}
  if ($hkl ne '111' && $hkl ne '333') {
     print STDOUT '*** '.$script.' incorrect reflection ['.$hkl.']'."\n";
     return -1;
  }
  if (!defined $beamline) {die $script.': undefined beamline';}
  if ($beamline !~ /^23[iob]{1}$/) {return 0;}			#we only have data for 23o, 23i, 23b

  $monoassy_PV = $beamline.':MO:En';
  $monoaxis_PV = $beamline.':MO:E';
  $monoset_PV  = $monoassy_PV.':Chkl';

# $monoEn = &getAxisAct($monoaxis_PV,\$status);
  $monoEn = &getAxisRqs($monoaxis_PV,\$status);
  if ($status) {return $status;}

  for ($i=1; $i<10; $i++) {
     $monodesc_PV = $monoassy_PV.':set'.$i.'.DESC';
     ($status,$monodesc) = Pezca::GetString($monodesc_PV);		#expect "Si 111"
     if ($status) {&report_only($script, $status, $monodesc_PV); return -1;}
     if ($monodesc =~ / ${hkl}\s*$/) {
        $status = &setAssyManual ($monoassy_PV);
        print STDOUT $script.': setting mono Bragg reflection to hkl='.$hkl.'.'."\n";
        $status = Pezca::PutStringOldCa($monoset_PV, 'set'.$i);
        if ($status) {&report_only($script, $status, $monoset_PV); return -1;}
        select(undef,undef,undef,1);					# sleep($wait_step);
### Run mono to the energy before the change:
        printf STDOUT $script.': running mono to initial energy=%.3f keV.'."\n", $monoEn;
        $status = &runMotor ($monoassy_PV, $monoaxis_PV, $monoEn, 'forceAuto');
        return $status;
      }
  }
  print STDOUT '*** '.$script.' unrecognized reflection ['.$hkl.']'."\n";
  return -1;
}
