#!/usr/bin/perl -W
#--------------------------------------------------------------
# These are MOSTAB (beamline intensity and positional feedback)
#       related subs for beamline controls Perl scripts
#
#     		Author: Sergey Stepanov
#
# Version-1.0:  2007/11/15 using Pezca
# Version-1.1:  2021/02/06 added ShiftP2 and RecallP2
#--------------------------------------------------------------
# List of functions:
# ^^^^^^^^^^^^^^^^^
# 1.      $present = &isMostabPresent ($beamline);
# 2.      $present = &isPostabPresent ($beamline);
# 3.      $running = &checkMostabRunning ($beamline); 			#1=run, 0=no, -1=error
# 4. $mostab_state = &getMostabState ($beamline); 			#get intensity feedback state: 0=run, 1=pause, -1=error
# 5. $postab_state = &getPostabState ($beamline); 			#get positional feedback state: 0=run, 3=pause, -1=error
# 6.        $mode  = &getMostabMode ($beamline);			#get intensity feedback mode: 0=software, 1=hybrid, 2=hardware
# 7.       $status = &mostabAutoShiftP2 ($beamline,[$dac_p2_PV]);	#center DAC & FPGA by shifting P2 only when they are far from center
# 8.       $status = &mostabShiftP2 ($beamline);			#center DAC & FPGA by shifting P2
# 9.       $status = &mostabRecallP2 ($beamline);			#recall P2 position from MemPos
#10.       $status = &pauseMostab ($beamline);
#11.       $status = &pausePostab ($beamline);
#12.       $status = &restoreMostab ($beamline);			#this does not wait for lock
#13.       $status = &restorePostab ($beamline);			#this does not wait for lock
#14.       $status = &waitMostabLock ($beamline,[$wait_max]);    	#default $wait_max=120s; returns -1=error 0=OK
#15.       $status = &setPointsPostab ($beamline,$postab_rqsH,$postab_rqsV);
#16.       $p2_opt = &getMostabSearchOption ($beamline); 		#get intensity feedback search option: 0=search with DAC, 1=search with P2
#17.       $status = &setMostabSearchOption ($beamline,$p2_opt);	#set intensity feedback search option: 0=search with DAC, 1=search with P2
###--> Functions for FPGA:
#18.       $state  = &getIntensityFeedbackFPGAstate($beamline,['string']);		#0=Paused,1=Active,2=Zero,3=Center
#19.       $status = &setIntensityFeedbackFPGAstate($beamline,$state,['string']);	#0=Paused,1=Active,2=Zero,3=Center
#20.       $state  = &getVertPosFeedbackFPGAstate($beamline,['string']);		#0=Paused,1=Active,2=Zero,3=Center
#21.       $status = &setVertPosFeedbackFPGAstate($beamline,$state,['string']);		#0=Paused,1=Active,2=Zero,3=Center
#--------------------------------------------------------------
# use Pezca;
# do 'mostab_subs.pl';

# use warnings::unused;
  use strict;
# no strict 'vars';
# no strict 'subs';
# no strict 'refs';

#=================================================================

# sub isMostabPresent ($beamline) {
sub isMostabPresent ($) {

### This routine reads returns 1=present, 0=no.

  my ($beamline, $script, @beamlinelist);

  $script   = 'isMostabPresent';
  $beamline = shift(@_);
  $beamline =~ s/:$//;
  @beamlinelist = qw/23i 23o/;
  if (grep(/^${beamline}$/, @beamlinelist)) {return 1;}
  else                                      {return 0;}
}

#=================================================================

# sub isPostabPresent ($beamline) {
sub isPostabPresent ($) {

### This routine reads returns 1=present, 0=no.

  my ($beamline, $script, @beamlinelist);

  $script   = 'isPostabPresent';
  $beamline = shift(@_);
  $beamline =~ s/:$//;
  @beamlinelist = qw/23i 23o/;
  if (grep(/^${beamline}$/, @beamlinelist)) {return 1;}
  else                                      {return 0;}
}

#=================================================================

# sub checkMostabRunning ($beamline) {
sub checkMostabRunning ($) {

### This routine reads returns 1=run, 0=no, -1=error.

  my ($beamline, $script, $mostab_status_PV, $mostab_status);
  my ($status, $wait_step, $wait_time, $wait_max);

  $script   = 'checkMostabRunning';
  $beamline = shift(@_);
  $beamline =~ s/:$//;
  if (! &isMostabPresent($beamline)) {return 0;}
  $mostab_status_PV = $beamline.':mostab:status';
  $wait_max = 5;
  $wait_step = 0.05;
  ($status,$mostab_status) = Pezca::GetDouble($mostab_status_PV);
  if ($status) {
     &report_only($script, $status, $mostab_status_PV);
     print STDOUT "\n".'*** '.$script.': failure='.$status
                      .' reading status of Intensity Feedback'."\n\n";
     return -1;
  }
  elsif ($mostab_status == 0) {return 0;}

  ### Status seems to be OK, check it:
  $status = Pezca::PutDoubleOldCa($mostab_status_PV, 0);
  if ($status) {
     &report_only($script, $status, $mostab_status_PV);
     print STDOUT "\n".'*** '.$script.': failure='.$status
                      .' writing status of Intensity Feedback'."\n\n";
     return -1;
  }
  $wait_time = 0;
  do {
     select(undef,undef,undef,$wait_step);	# sleep
     $wait_time += $wait_step;
     ($status,$mostab_status) = Pezca::GetDouble($mostab_status_PV);
  } while ($mostab_status == 0 && $wait_time < $wait_max);

  return $mostab_status;
}

#=================================================================

# sub getMostabState ($beamline) {
sub getMostabState ($) {

### This routine reads current Mostab state (0=run, 1=pause).
### Returns 0=run, 1=pause, -1=error.

  my ($beamline, $script, $mostab_pause_PV, $status, $mostab_state);

  $script   = 'getMostabState';
  $beamline = shift(@_);
  $beamline =~ s/:$//;
  if (! &isMostabPresent($beamline)) {return 1;}

  $mostab_pause_PV = $beamline.':mostab:pause';

  ($status,$mostab_state) = Pezca::GetDouble($mostab_pause_PV);
  if ($status) {
     &report_only($script, $status, $mostab_pause_PV);
     print STDOUT "\n".'*** '.$script.': failure='.$status
                      .' reading status of Intensity Feedback'."\n\n";
     return -1;
  } else {
     return $mostab_state;
  }
}

#=================================================================

# sub getPostabState ($beamline) {
sub getPostabState ($) {

### This routine reads current Postab state (0=run, 3=pause).
### Returns 0=run, 3=pause, -1=error.

  my ($beamline, $script, $postab_pause_PV, $status, $postab_state);

  $script   = 'getPostabState';
  $beamline = shift(@_);
  $beamline =~ s/:$//;
  if (! &isPostabPresent($beamline)) {return 3;}

  $postab_pause_PV = $beamline.':postab:pause';

  ($status,$postab_state) = Pezca::GetDouble($postab_pause_PV);
  if ($status) {
     &report_only($script, $status, $postab_pause_PV);
     print STDOUT "\n".'*** '.$script.': failure='.$status
                      .' reading status of Positional Feedback'."\n\n";
     return -1;
  }
  return $postab_state;
}

#=================================================================

# sub getMostabMode ($beamline) {
sub getMostabMode ($) {

### This routine reads current Mostab Mode (0=software, 1=hybrid, 2=hardware).
### In case of error presumes Mode=software.

  my ($beamline, $script, $mostab_mode_PV, $status, $mostab_mode);

  $script   = 'getMostabMode';
  $beamline = shift(@_);
  $beamline =~ s/:$//;
  if (! &isMostabPresent($beamline)) {return 0;}

  $mostab_mode_PV = $beamline.':feedback:uselockin';

  ($status,$mostab_mode) = Pezca::GetDouble($mostab_mode_PV);
  if ($status) {
     &report_only($script, $status, $mostab_mode_PV);
     print STDOUT "\n".'!!! '.$script.': failure='.$status
                      .' reading mode of Intensity Feedback'."\n\n";
     $mostab_mode = 0;							#presume software
  }
  return $mostab_mode;
}

#=================================================================

# sub mostabAutoShiftP2 ($beamline,[$dac_p2_PV]) {			#center DAC & FPGA by shifting P2 only when needed
sub mostabAutoShiftP2 ($;$) {

### This routine returns 1=error, 0=OK.

  my ($script, $beamline, $dac_p2_PV, $fpga_p2_PV, $status);
  my ($dac,  $DAC_l,  $DAC_h, $needed, $mode);
  my ($fpga, $FPGA_l, $FPGA_h);

  $script    = 'mostabAutoShiftP2';
  $beamline  = shift(@_);	$beamline =~ s/:$//;
  if (! &isMostabPresent($beamline)) {return 0;}

  $dac_p2_PV = shift(@_);
  if (!defined $dac_p2_PV) {$dac_p2_PV = 'DAC1_1';}			#default
  if ($dac_p2_PV !~ /^$beamline/) {$dac_p2_PV = $beamline.':'.$dac_p2_PV;}
  $fpga_p2_PV = $beamline.':feedback:FPGA:voltageOut';

  ($DAC_l, $DAC_h)   = (2.5, 7.5);
  ($FPGA_l, $FPGA_h) = (250, 750);

  if (&checkMostabRunning($beamline) != 1) {return 1;} 			#1=run, 0=no, -1=error
  if (&getMostabState($beamline) != 0) {return 1;} 			#get intensity feedback state: 0=run, 1=pause, -1=error
  $mode = &getMostabMode($beamline);					#0=software, 1=hybrid, 2=hardware

  $needed = 0;
  if ($mode != 2) {							#not HW feedback
     ($status,$dac) = Pezca::GetDouble($dac_p2_PV);			#expected in range [0:10]
     if ($status) {&report_only($script,$status,$dac_p2_PV); return 1;}
     if ($dac < $DAC_l || $dac > $DAC_h) {
        $needed = 1;
        printf STDOUT $script.': DAC=%.2f is not in range [%.2f-%.2f]. Will call P2 shift.'."\n",$dac,$DAC_l,$DAC_h;
     }
  } else {								#HW (FPGA) feedback
     ($status,$fpga) = Pezca::GetDouble($fpga_p2_PV);			#expected in range [0:1000]
     if ($status) {&report_only($script,$status,$fpga_p2_PV); return 1;}
     if ($fpga < $FPGA_l || $fpga > $FPGA_h) {
        $needed = 1;
        printf STDOUT $script.': FPGA=%.2f is not in range [%.2f-%.2f]. Will call P2 shift.'."\n",$fpga,$FPGA_l,$FPGA_h;
     }
  }
  if ($needed) {$status = &mostabShiftP2($beamline);}
  else         {$status = 0;}  
  return $status;
}

#=================================================================

# sub mostabShiftP2 ($beamline) {			#center DAC & FPGA by shifting P2
sub mostabShiftP2 ($) {

### This routine returns 1=error, 0=OK.

  my ($beamline, $script, $mostab_opt_PV);
  my ($status, $wait_max);

  $script   = 'mostabShiftP2';
  $beamline = shift(@_);
  $beamline =~ s/:$//;
  if (! &isMostabPresent($beamline)) {return 0;}

  if (&checkMostabRunning($beamline) != 1) {return 1;} 			#1=run, 0=no, -1=error
  if (&getMostabState($beamline) != 0) {return 1;} 			#get intensity feedback state: 0=run, 1=pause, -1=error

  $mostab_opt_PV = $beamline.':mostab:opt';
  $wait_max = 25;

### Press "Shift P2 (center DAC)" button:
  $status = Pezca::PutDoubleOldCa($mostab_opt_PV, 1);
  if ($status) {&report_only($script,$status,$mostab_opt_PV); return 1;}

  if (&waitMostabLock($beamline,$wait_max) != 0) {return 1;}
  else                                           {return 0;}
}

#=================================================================

# sub mostabRecallP2 ($beamline) {			#recall P2 position from MemPos
sub mostabRecallP2 ($) {

### This routine returns 1=error, 0=OK.

  my ($beamline, $script, $mostab_recall_PV);
  my ($status, $wait_max);

  $script   = 'mostabRecallP2';
  $beamline = shift(@_);
  $beamline =~ s/:$//;
  if (! &isMostabPresent($beamline)) {return 0;}

  if (&checkMostabRunning($beamline) != 1) {return 1;} 			#1=run, 0=no, -1=error
  if (&getMostabState($beamline) != 0) {return 1;} 			#get intensity feedback state: 0=run, 1=pause, -1=error

  $mostab_recall_PV = $beamline.':mostab:opt';
  $wait_max = 25;

### Press "Recall P2 from MemPos" button:
  $status = Pezca::PutDoubleOldCa($mostab_recall_PV, 1);
  if ($status) {&report_only($script,$status,$mostab_recall_PV); return 1;}

  if (&waitMostabLock($beamline,$wait_max) != 0) {return 1;}
  else                                           {return 0;}
}

#=================================================================

# sub pauseMostab ($beamline) {
sub pauseMostab ($) {

### This routine pauses mostab if needed.
### Returns 0=OK, -1=error.

  my ($beamline, $script, $mostab_pause_PV, $status);
  my ($mostab_state, $pausable);
# my ($mode, $mostab_descr_PV, $descr);

  $script   = 'pauseMostab';
  $beamline = shift(@_);
  $beamline =~ s/:$//;
  if (! &isMostabPresent($beamline)) {return 0;}

  $mostab_pause_PV = $beamline.':mostab:pause';
# $mostab_descr_PV = $beamline.':mostab:status.DESC';

  $mostab_state = &getMostabState ($beamline);				#(0=run, 1=pause, -1=error)
  if    ($mostab_state == -1) {return $mostab_state;}			#error
  elsif ($mostab_state == 1)  {return 0;}				#already paused
  else {
#    $mode = &getMostabMode ($beamline);				#(0=software, 1=hybrid, 2=hardware)
#    if ($mode == 2) {							#if hardware mode
#       ($status,$descr) = Pezca::GetString($mostab_descr_PV);
#       if ($status) {
#          &report_only($script, $status, $mostab_descr_PV);
#          $descr = 'unknown';
#       }
#       if ($descr =~ /pause/i) {$pausable=1;}
#       else                    {$pausable=0;}
#    } else {
        $pausable=1;
#    }
     if ($pausable) {
        print STDOUT "\n".'Pausing Intensity Feedback'."\n";
        $mostab_state  = 1;						#pause
        $status = Pezca::PutDouble($mostab_pause_PV, $mostab_state);
        if ($status) {
           &report_only($script, $status, $mostab_pause_PV);
           print STDOUT "\n".'*** '.$script.': failure='.$status
                            .' pausing Intensity Feedback'."\n\n";
           return -1;
        }
        select(undef,undef,undef,0.5);					#sleep
     } else {
        print STDOUT "\n".'!!! '.$script.': Hardware-mode'
                         .' Intensity Feedback is unpausable.'."\n";
     }
     return 0;
  }
}

#=================================================================

# sub pausePostab ($beamline) {
sub pausePostab ($) {

### This routine pauses postab if needed:
### Returns 0=OK, -1=error.

  my ($beamline, $script, $postab_pause_PV, $status, $postab_state);

  $script   = 'pausePostab';
  $beamline = shift(@_);
  $beamline =~ s/:$//;
  if (! &isPostabPresent($beamline)) {return 0;}

  $postab_pause_PV = $beamline.':postab:pause';

  $postab_state = &getPostabState ($beamline);				#(0=run, 1=pause, -1=error)
  if    ($postab_state == -1) {return $postab_state;}			#error
  elsif ($postab_state == 3)  {return 0;}				#already paused
  else {
     print STDOUT "\n".'Pausing Positional Feedback'."\n";
     $postab_state  = 3;						#pause
     $status = Pezca::PutDouble($postab_pause_PV, $postab_state);
     if ($status) {
        &report_only($script, $status, $postab_pause_PV);
        print STDOUT "\n".'*** '.$script.': failure='.$status
                         .' pausing Positional Feedback'."\n\n";
        return -1;
     }
     select(undef,undef,undef,0.5);					#sleep
     return 0;
  }
}

#=================================================================

# sub restoreMostab ($beamline) {
sub restoreMostab ($) {

### This routine restores mostab if needed.
### Returns -1=error 0=OK

  my ($beamline, $script, $mostab_pause_PV, $status, $mostab_state);

  $script   = 'restoreMostab';
  $beamline = shift(@_);
  $beamline =~ s/:$//;
  if (! &isMostabPresent($beamline)) {return 0;}

  $mostab_pause_PV = $beamline.':mostab:pause';

  $mostab_state = &getMostabState ($beamline);				#(0=run, 1=pause, -1=error)
  if    ($mostab_state == -1) {return $mostab_state;}			#error
  elsif ($mostab_state != 1)  {return 0;}				#not paused
  else {
     print STDOUT "\n".'Restoring Intensity Feedback'."\n";
     $mostab_state  = 0;				# run
     $status = Pezca::PutDouble($mostab_pause_PV, $mostab_state);
     if ($status) {
        &report_only($script, $status, $mostab_pause_PV);
        print STDOUT "\n".'*** '.$script.': failure='.$status
                         .' restoring Intensity Feedback'."\n\n";
        return -1;
     }
     select(undef,undef,undef,0.5);					#sleep
     return 0;
  }
}

#=================================================================

# sub restorePostab ($beamline) {
sub restorePostab ($) {

### This routine restores postab if needed.
### Returns -1=error 0=OK

  my ($beamline, $script, $postab_pause_PV, $status, $postab_state);

  $script   = 'restorePostab';
  $beamline = shift(@_);
  $beamline =~ s/:$//;
  if (! &isPostabPresent($beamline)) {return 0;}

  $postab_pause_PV = $beamline.':postab:pause';

  $postab_state = &getPostabState ($beamline);				#(0=run, 1=pause, -1=error)
  if    ($postab_state == -1) {return $postab_state;}			#error
  elsif ($postab_state != 1)  {return 0;}				#not paused
  else {
     print STDOUT "\n".'Restoring Positional Feedback'."\n";
     $postab_state  = 0;						#restore
     $status = Pezca::PutDouble($postab_pause_PV, $postab_state);
     if ($status) {
        &report_only($script, $status, $postab_pause_PV);
        print STDOUT "\n".'*** '.$script.': failure='.$status
                         .' restoring Positional Feedback'."\n\n";
        return -1;
     }
     select(undef,undef,undef,0.5);					#sleep
     return 0;
  }
}

#=================================================================

# sub waitMostabLock ($beamline,$wait_max) {
sub waitMostabLock ($;$) {

### This script tracks mostab locking:
### Returns -1=error 0=OK

  my ($beamline, $wait_max, $script);
  my ($wait_step, $wait_time, $status);
  my ($mostab_lock_pv,   $mostab_lock);
  my ($mostab_pause_pv,  $mostab_pause);
  my ($mostab_status_pv, $mostab_status);

  $script    = 'waitMostabLock';
  $beamline = shift(@_);
  $wait_max = shift(@_);
  $beamline =~ s/:$//;
  if (! &isMostabPresent($beamline)) {return 0;}

  if (!defined $wait_max) {$wait_max = 120.0}				#default (in seconds)
  $wait_step = 0.2;							#poll intervals when waiting
  $mostab_lock_pv   = $beamline.':mostab:lock';
  $mostab_pause_pv  = $beamline.':mostab:pause';
  $mostab_status_pv = $beamline.':mostab:status';

  print STDOUT "\n".'Waiting for intensity lock (up to '
                   .$wait_max.' seconds)'."\n";
  $status = &addCaMonitor($mostab_lock_pv);
  if ($status) {
     print STDOUT "\n".'*** '.$script.': failure=%d setting [%s] PV monitor'."\n\n", $status,$mostab_lock_pv;
     return -1;
  }
  $status = &addCaMonitor($mostab_pause_pv);
  if ($status) {
     print STDOUT "\n".'*** '.$script.': failure=%d setting [%s] PV monitor'."\n\n", $status,$mostab_pause_pv;
     return -1;
  }
  $status = &addCaMonitor($mostab_status_pv);
  if ($status) {
     print STDOUT "\n".'*** '.$script.': failure=%d setting [%s] PV monitor'."\n\n", $status,$mostab_status_pv;
     return -1;
  }
### Wait for intensity lock:
  $mostab_lock   = 0;							#0=not-locked, 1=locked
  $mostab_pause  = 0;							#0=Run, 1=Pause
  $mostab_status = 1;							#0=exiting, 1=active, 2=doing.operation, 3=waiting, 4=lost
  $wait_time = 0;
  do {
     select(undef,undef,undef,$wait_step);	# sleep
     $wait_time += $wait_step;
     ($status,$mostab_lock)   = Pezca::GetDouble($mostab_lock_pv);	#0=not-locked, 1=locked
     ($status,$mostab_pause)  = Pezca::GetDouble($mostab_pause_pv);	#0=Run, 1=Pause
     ($status,$mostab_status) = Pezca::GetDouble($mostab_status_pv);	#0=exiting, 1=active, 2=doing.operation, 3=waiting, 4=lost
  } while ($mostab_lock  == 0 && 
           $mostab_pause == 0 &&
           $mostab_status > 0 &&
#          $mostab_status < 4 &&
           $wait_time < $wait_max);

  if ($wait_time > $wait_max) {
     print STDOUT "\n".'!!! Intensity lock not established within timeout!!!'."\n\n";
     return -1;
  } 
  elsif ($mostab_status == 0) {
     print STDOUT "\n".'!!! Intensity feedback is not running, lock cannot be established!!!'."\n\n";
     return -1;
  } 
  elsif ($mostab_pause == 1) {
     print STDOUT "\n".'!!! Intensity feedback is paused, lock cannot be established!!!'."\n\n";
     return -1;
  } 
  else {
     print STDOUT '...Lock established OK!'."\n";
     return 0;
  }
}

#=================================================================

# sub setPointsPostab ($beamline,$postab_rqsH,$postab_rqsV) {
sub setPointsPostab ($$$) {

### This script sets new hold points for positional feedback:
### Returns Nonzero=error 0=OK

  my ($beamline, $postab_rqsH, $postab_rqsV, $postab_rpvH, $postab_rpvV);
  my ($script, $status1, $status2);

  $script      = 'setPointsPostab';
  $beamline    = shift(@_);
  $postab_rqsH = shift(@_);
  $postab_rqsV = shift(@_);
  $beamline    =~ s/:$//;
  if (! &isPostabPresent($beamline)) {return 0;}

  $postab_rpvH = $beamline.':postab:set_h';
  $postab_rpvV = $beamline.':postab:set_v';

  $status1 = Pezca::PutDouble($postab_rpvH, $postab_rqsH);
  if ($status1) {
     &report_only($script, $status1, $postab_rpvH);
     print STDOUT "\n".'*** '.$script.': failure='.$status1
                      .' setting H-point'."\n\n";
  }

  $status2 = Pezca::PutDouble($postab_rpvV, $postab_rqsV);
  if ($status2) {
     &report_only($script, $status2, $postab_rpvV);
     print STDOUT "\n".'*** '.$script.': failure='.$status2
                      .' setting H-point'."\n\n";
  }
  return $status1+$status2;
}

#=================================================================

# sub getMostabSearchOption ($beamline) {
sub getMostabSearchOption ($) {

### This routine reads current Mostab search option (0=search with DAC, 1=search with P2).
### Returns 0=search with DAC, 1=search with P2, -1=error.

  my ($beamline, $script, $p2_opt_PV, $status, $p2_opt);

  $script   = 'getMostabSearchOption';
  $beamline = shift(@_);  $beamline =~ s/:$//;
  if (! &isMostabPresent($beamline)) {return 0;}

  $p2_opt_PV = $beamline.':mostab:search_p2';

  ($status,$p2_opt) = Pezca::GetDouble($p2_opt_PV);
  if ($status) {
     &report_only($script, $status, $p2_opt_PV);
     print STDOUT "\n".'*** '.$script.': failure='.$status
                      .' reading the Intensity Feedback search option'."\n\n";
     return -1;
  }
  return $p2_opt;
}

#=================================================================

# sub setMostabSearchOption ($beamline,$p2_opt) {
sub setMostabSearchOption ($$) {

### This routine sets mostab search option (0=search with DAC, 1=search with P2).
### Returns 0=OK, -1=error.

  my ($beamline, $script, $p2_opt_PV, $status, $p2_opt);

  $script   = 'setMostabSearchOption';
  $beamline = shift(@_);  $beamline =~ s/:$//;
  $p2_opt   = shift(@_);
  if (! &isMostabPresent($beamline)) {return 0;}

  if ($p2_opt ne '0' && $p2_opt ne '1') {
     print STDOUT "\n".'*** '.$script.': incorrect Intensity'
                      .' Feedback search option='.$p2_opt."\n\n";
     return -1;
  }

  $p2_opt_PV = $beamline.':mostab:search_p2';

  $status = Pezca::PutDouble($p2_opt_PV, $p2_opt);
  if ($status) {
     &report_only($script, $status, $p2_opt_PV);
     print STDOUT "\n".'*** '.$script.': failure='.$status
          .' setting the Intensity Feedback search option='.$p2_opt."\n\n";
     return -1;
  }
  return 0;
}

#=================================================================

# sub getIntensityFeedbackFPGAstate($beamline,['string'] {		#0=Paused,1=Active,2=Zero,3=Center
sub   getIntensityFeedbackFPGAstate($;$) {

  my ($script, $beamline, $datatype, $fpgaPV, $status, $state);

  $script   = 'getIntensityFeedbackFPGAstate';
  $beamline = shift(@_);  $beamline =~ s/:$//;
  $datatype = shift(@_);  if (!defined $datatype) {$datatype = 'double';}
  if (! &isMostabPresent($beamline)) {return 0;}

  $fpgaPV = $beamline.':feedback:FPGA:feedBackMode';

  if ($datatype =~ /^string$/i) {($status, $state) = Pezca::GetString($fpgaPV);}
  else                          {($status, $state) = Pezca::GetDouble($fpgaPV);}
  if ($status) {
     &report_only ($script, $status, $fpgaPV);
     if ($datatype =~ /^string$/i) {$state = 'Disconnected';}
     else                          {$state= -1;}
  }
  return $state;
}

#=================================================================

# sub setIntensityFeedbackFPGAstate($beamline,$state,['string'] {		#0=Paused,1=Active,2=Zero,3=Center
sub   setIntensityFeedbackFPGAstate($$;$) {

  my ($script, $beamline, $datatype, $fpgaPV, $status, $state);
  my (@states, $nstates, @match, $nmatch);

  $script   = 'setIntensityFeedbackFPGAstate';
  $beamline = shift(@_);  $beamline =~ s/:$//;
  $state    = shift(@_);
  $datatype = shift(@_);  if (!defined $datatype) {$datatype = 'double';}
  if (! &isMostabPresent($beamline)) {return 0;}

  @states  = qw(Paused Active Zero Center);
  $nstates = @states;

  $fpgaPV = $beamline.':feedback:FPGA:feedBackMode';

  if ($datatype =~ /^string$/i) {
     @match = grep(/^${state}$/i, @states);
     $nmatch = @match;
     if ($nmatch < 1) {
        print STDOUT '*** '.$script.': incorrect requested state='.$state."\n"
                   . '*** '.$script.': Allowed states are: ',"(@{states})\n";
        return -1;
     }
     $state = $match[0];
     $status = Pezca::PutString($fpgaPV,$state);
  } else {
     if ($state < 0 || $state > $nstates-1) {
	print STDOUT '*** '.$script.': incorrect requested state='.$state."\n"
                   . '*** '.$script.': Allowed states are: (0-',$nstates-1,')'."\n";
	return -1;
     }
     $status = Pezca::PutDouble($fpgaPV,$state);
  }
  &report_only ($script, $status, $fpgaPV);
  return $status;
}

#=================================================================

# sub getVertPosFeedbackFPGAstate($beamline,['string'] {		#0=Paused,1=Active,2=Zero,3=Center
sub   getVertPosFeedbackFPGAstate($;$) {

  my ($script, $beamline, $datatype, $fpgaPV, $status, $state);

  $script   = 'getVertPosFeedbackFPGAstate';
  $beamline = shift(@_);  $beamline =~ s/:$//;
  $datatype = shift(@_);  if (!defined $datatype) {$datatype = 'double';}
  if (! &isPostabPresent($beamline)) {return 0;}

  $fpgaPV = $beamline.':feedback2:FPGA:feedBackMode';

  if ($datatype =~ /^string$/i) {($status, $state) = Pezca::GetString($fpgaPV);}
  else                          {($status, $state) = Pezca::GetDouble($fpgaPV);}
  if ($status) {
     &report_only ($script, $status, $fpgaPV);
     if ($datatype =~ /^string$/i) {$state = 'disconnected';}
     else                          {$state = -1;}
  }
  return $state;
}

#=================================================================

# sub setVertPosFeedbackFPGAstate($beamline,$state,['string'] {		#0=Paused,1=Active,2=Zero,3=Center
sub   setVertPosFeedbackFPGAstate($$;$) {

  my ($script, $beamline, $datatype, $fpgaPV, $status, $state);
  my (@states, $nstates, @match, $nmatch);

  $script   = 'setVertPosFeedbackFPGAstate';
  $beamline = shift(@_);  $beamline =~ s/:$//;
  $state    = shift(@_);
  $datatype = shift(@_);  if (!defined $datatype) {$datatype = 'double';}
  if (! &isPostabPresent($beamline)) {return 0;}

  @states  = qw(Paused Active Zero Center);
  $nstates = @states;

  $fpgaPV = $beamline.':feedback2:FPGA:feedBackMode';

  if ($datatype =~ /^string$/i) {
     @match = grep(/^${state}$/i, @states);
     $nmatch = @match;
     if ($nmatch < 1) {
        print STDOUT '*** '.$script.': incorrect requested state='.$state."\n"
                   . '*** '.$script.': Allowed states are: ',"(@{states})\n";
        return -1;
     }
     $state = $match[0];
     $status = Pezca::PutString($fpgaPV,$state);
  } else {
     if ($state < 0 || $state > $nstates-1) {
        print STDOUT '*** '.$script.': incorrect requested state='.$state."\n"
                   . '*** '.$script.': Allowed states are: (0-',$nstates-1,')'."\n";
        return -1;
     }
     $status = Pezca::PutDouble($fpgaPV,$state);
  }
  &report_only ($script, $status, $fpgaPV);
  return $status;
}
