#!/usr/bin/perl -W
#--------------------------------------------------------------
# These are Shutter related subs for beamline controls Perl scripts
#
#     		Author: Sergey Stepanov
#
# Version-1.0:  2004/11/12 using Pezca
# Version-1.1:  2009/11/12 moved hutch shutters PVs into beamline.pl
# Version-1.2   2015/03/16 added setGmcaShutterNames, setupShutterMotor, trackShutter;
#                          added tracking to openShutter and closeShutter
#--------------------------------------------------------------
# List of functions:
# ^^^^^^^^^^^^^^^^^
# 0.		      &setGmcaShutterNames(\$shutterCMD,\$shutterRBK,\$shutter_assy);	#fill missing names (GMCA only)
# 1.        $status = &setupShutterMotor($shutter_assy);
# 2.       $allowed = &checkShutterOpeningAllowed($shutter_assy,[$shutter_axis]);	#1=allowed,0=disabled,-1=error
# 2.        $status = &openShutter ($shutterCMD, [$shutterRBK,$shutter_assy]);		#0=OK, -1=failure
# 3.        $status = &closeShutter ($shutterCMD, [$shutterRBK,$shutter_assy]);		#0=OK, -1=failure
# 4.        $status = &trackShutter($shutter_destination,$shutterRBK);			#0=OK, -1=failure
# 5. $open_or_close = &getShutter($shutterRBK);						#1=open,0=close,-1=error
# 6.        $status = &openEndStationShutter($beamline);				#0=OK
# 7.        $status = &closeEndStationShutter($beamline);       			#0=OK
# 8. $open_or_close = &getEndStationShutter($beamline);					#1=open,0=close,-1=not.responding
# 9.        $status = &openCshutter($beamline);						#0=OK
#10.        $status = &closeCshutter($beamline);       					#0=OK
#11. $open_or_close = &getCshutter($beamline);			 			#1=open/not present,0=close,-1=not.responding (GMCA ID-in only)
#12.        $status = &openFrontEndShutter($beamline);					#0=OK
#13.        $status = &closeFrontEndShutter($beamline);       				#0=OK
#14. $open_or_close = &getFrontEndShutter($beamline);					#1=open,0=close,-1=not.responding
#15.     $SRcurrent = &getSRcurrent([\*$status,$LUN]);					#get APS beam current
#16.         $yesno = &getShutterPermit([\*$status,$LUN]);					#get APS shutter permit
#--------------------------------------------------------------
# use Pezca;
# do 'shutter_subs.pl';

# use warnings::unused;
  use strict;
# no strict "vars";
# no strict "subs";
# no strict "refs";

#=================================================================

# sub setGmcaShutterNames(\$shutterCMD,\$shutterRBK,\$shutter_assy) {	#fill missing names (GMCA only)
sub setGmcaShutterNames (\$\$\$) {

  our ($beamline, $shutter, $shutter_readback);				#this is from beamline.pl
  my ($shutterCMD_ref,$shutterRBK_ref,$shutter_assy_ref);

  $shutterCMD_ref   = shift(@_);
  $shutterRBK_ref   = shift(@_);
  $shutter_assy_ref = shift(@_);

  if ($beamline =~ /^23/) {						#this function is GMCA-specific
     if (!defined $$shutterCMD_ref && defined $shutter) {
        $$shutterCMD_ref = $shutter;
     }
     if (!defined $$shutterRBK_ref && defined $shutter_readback) {
        $$shutterRBK_ref = $shutter_readback;
     }
     if (!defined $$shutter_assy_ref) {
        $$shutter_assy_ref = $beamline.':SH:Ps';			#GMCA motor name
     }
     if (defined $$shutterCMD_ref && $$shutterCMD_ref !~ /^${beamline}/) {
        $$shutterCMD_ref = $beamline.':'.$$shutterCMD_ref;
     }
     if (defined $$shutterRBK_ref && $$shutterRBK_ref !~ /^${beamline}/) {
        $$shutterRBK_ref = $beamline.':'.$$shutterRBK_ref;
     }
  }
}

#=================================================================

# sub setupShutterMotor($shutter_assy) {
sub setupShutterMotor ($) {

  my ($script, $shutter_assy, $status, $script_main);

  $script       = 'setupShutterMotor';
  $shutter_assy = shift(@_);

  $status = 0;
  if (defined $shutter_assy) {
     if (defined &function_exists('abortAssy')) {
        $status += &abortAssy ($shutter_assy);
        $status += &setAssyAuto ($shutter_assy);
     } else {
        $script_main = &makeScriptName();
	print STDOUT "\n".'*** '.$script_main.': "abortAssy" is undefined in '.$script.'.'
	            ."\n".'*** '.$script_main.': Please include "motor_subs.pl".'."\n\n";
     }
  }
  return $status;
}

#=================================================================

# sub checkShutterOpeningAllowed($shutter_assy,[$shutter_axis]) {
sub checkShutterOpeningAllowed ($;$) {

  my ($script, $shutter_assy, $shutter_axis, $status, $script_main);
  my (@axes, @drives, @motors, $openProtectCalc_PV);
  my ($protect, @inputs, $input, $pv, $val);

  $script       = 'checkShutterOpeningAllowed';
  $shutter_assy = shift(@_);
  $shutter_axis = shift(@_);

  if (!defined $shutter_axis) {
     @axes   = ();
     @drives = ();
     @motors = ();
     if (defined &function_exists('getAssyDataFromTsub')) {
        $status = &getAssyDataFromTsub($shutter_assy,\@axes,\@drives,\@motors);
        if ($status) {return -1;}
     } else {
        $script_main = &makeScriptName();
	print STDOUT "\n".'*** '.$script_main.': "getAssyDataFromTsub" is undefined in '.$script.'.'
	            ."\n".'*** '.$script_main.': Please include "motor_subs.pl".'."\n\n";
	return -1;
     }
     $shutter_axis = $motors[0];
  }
  if ($shutter_assy !~ /:$/) {$shutter_assy .=':';}
  if ($shutter_axis !~ /:$/) {$shutter_axis .=':';}
  $openProtectCalc_PV = $shutter_axis.'OpenProtectCalc';
# $status = Pezca::PutDouble($openProtectCalc_PV.'.PROC',1);
# if ($status) {&report_only($script, $status, $openProtectCalc_PV.'.PROC'); return -1;}
  ($status,$protect) = Pezca::GetDouble($openProtectCalc_PV);
  if ($status) {&report_only($script, $status, $openProtectCalc_PV); return -1;}
  if ($protect == 0) {return 1;}
  else {
     print STDOUT '*** '.$script.': Shutter opening is disabled by the detector protection CALC.'."\n"
		 .'    CALC record ['.$openProtectCalc_PV.'] dump:'."\n";
     ($status,$val) = Pezca::GetString($openProtectCalc_PV.'.CALC');
     if ($status) {&report_only($script, $status, $openProtectCalc_PV.'.CALC'); $val='ERR';}
     print STDOUT '    CALC: '.$val."\n";
     @inputs = qw(A B C D E F);
     foreach $input (@inputs) {
	($status,$pv) = Pezca::GetString($openProtectCalc_PV.'.INP'.$input);
        if ($status) {&report_only($script, $status, $openProtectCalc_PV.'.INP'.$input); $pv='ERR';}
	($status,$val) = Pezca::GetDouble($openProtectCalc_PV.'.'.$input);
        if ($status) {&report_only($script, $status, $openProtectCalc_PV.'.'.$input); $val='ERR';}
	printf STDOUT '       '.$input.':  %-40s '.$val."\n",$pv;
     }
     return 0;
  }
}

#=================================================================

# sub openShutter ($shutterCMD,[$shutterRBK,$shutter_assy]) {           #returns 0=OK, -1=failure
sub openShutter ($;$$) {

  our ($shutter_open, $report);					        #this is from beamline.pl
  my ($script, $shutterCMD, $shutterRBK, $shutter_assy, $status);
  my ($attempt, $maxattempts);

  $script  = 'openShutter';
  $shutterCMD   = shift(@_);
  $shutterRBK   = shift(@_);
  $shutter_assy = shift(@_);
  if (!defined $report) {$report=0;};

  &setGmcaShutterNames(\$shutterCMD,\$shutterRBK,\$shutter_assy);

  $maxattempts = 3;
  $attempt     = 0;

  if (defined $shutterCMD) {
     if ($report) {print STDOUT 'Opening shutter '.$shutterCMD."\n";}
     do {
        $attempt++;
        if (defined $shutter_assy) {&setupShutterMotor($shutter_assy);}
        $status = Pezca::PutDoubleOldCa($shutterCMD,$shutter_open);
        if ($status) {&report_only($script, $status, $shutterCMD); return -1;}          #return failure
        if (defined $shutterRBK) {$status = &trackShutter($shutter_open,$shutterRBK);}  #0=OK, -1=failure
	else                     {select(undef,undef,undef,0.050);}			#give shutter time to finish
        if ($status && $attempt < $maxattempts) {
           print STDOUT '!!! '.$script.': retrying...'."\n";
	}
     } while ($status && $attempt < $maxattempts);
     if ($status && defined $shutter_assy) {
	&checkShutterOpeningAllowed($shutter_assy);
     }
     return $status;                                                    #this is 0 or status from trackShutter
  } else {
     if ($report) {print STDOUT '!!! '.$script.': Shutter undefined -- opening skipped.'."\n";}
     return 0;                                                          #OK (no shutter)
  }
}

#=================================================================

# sub closeShutter ($shutterCMD,[$shutterRBK,$shutter_assy]) {          #returns 0=OK, -1=failure
sub closeShutter ($;$$) {

  our ($shutter_closed, $report);				        #this is from beamline.pl
  my ($script, $shutterCMD, $shutterRBK, $shutter_assy, $status);
  my ($attempt, $maxattempts);

  $script  = 'closeShutter';
  $shutterCMD   = shift(@_);
  $shutterRBK   = shift(@_);
  $shutter_assy = shift(@_);
  if (!defined $report) {$report=0;};

  &setGmcaShutterNames(\$shutterCMD,\$shutterRBK,\$shutter_assy);

  $maxattempts  = 3;
  $attempt      = 0;

  if (defined $shutterCMD) {
     if ($report) {print STDOUT 'Closing shutter '.$shutterCMD."\n";}
     do {
        $attempt++;
        if (defined $shutter_assy) {&setupShutterMotor($shutter_assy);}
        $status = Pezca::PutDoubleOldCa($shutterCMD,$shutter_closed);
        if ($status) {&report_only($script, $status, $shutterCMD); return -1;}           #return failure                   
        if (defined $shutterRBK) {$status = &trackShutter($shutter_closed,$shutterRBK);} #0=OK, -1=failure
	else                     {select(undef,undef,undef,0.050);}			 #give shutter time to finish
        if ($status && $attempt < $maxattempts) {
           print STDOUT '!!! '.$script.': retrying...'."\n";
	}
     } while ($status && $attempt < $maxattempts);
     return $status;                                                    #this is 0 or status from trackShutter
  } else {
     if ($report) {print STDOUT '!!! '.$script.': Shutter undefined -- closing skipped.'."\n";}
     return 0;                                                          #OK (no shutter)
  }
}

#=================================================================

# sub trackShutter($shutter_destination,$shutterRBK) {		        #returns Shutter condition: 0=OK, -1=failure
sub trackShutter ($$) {

  our ($shutter_open, $shutter_closed);				        #this is from beamline.pl
  my ($script, $shutter_dest, $shutterRBK, $status, $shutter_state);
  my ($wait_time, $wait_max, $wait_sleep, $diff, $dest);
# my ($start);

  $script        = 'trackShutter';
  $shutter_dest = shift(@_);
  $shutterRBK   = shift(@_);

  if (!defined $shutterRBK) {
     print STDOUT '*** '.$script.': undefined PV "$shutterRBK"'."\n\n";
     $shutter_state = -1;
     return -1;                                                         #error
  }

# $start      = time();
# $wait_max   = 0.500;						        #500ms
# $wait_sleep = 0.010;						        #10ms -- this may be too short (not sleeping)
  $wait_max   = 1.000;						        #1000ms (2019.02)
  $wait_sleep = 0.050;						        #10ms   (2019.02)
  $wait_time  = 0.;
  $status = &addCaMonitor($shutterRBK);
  &report_only($script, $status, $shutterRBK);

  do {
     select(undef,undef,undef,$wait_sleep);			        #sleep 10ms
     ($status, $shutter_state) = Pezca::GetDouble($shutterRBK);
     if ($status) {
        &report_only($script, $status, $shutterRBK);
        $shutter_state = -1;
	return -1;                                                      #error
     }
#    $wait_time = time() - $start;
     $wait_time += $wait_sleep;
     $diff = abs($shutter_state-$shutter_dest);
  } while ($wait_time < $wait_max && $diff > 0.01);
  if ($diff < 0.01) {return 0;}                                         #OK
  else {
     if    ($shutter_dest == $shutter_open)   {$dest = 'OPEN';}
     elsif ($shutter_dest == $shutter_closed) {$dest = 'CLOSE';}
     else                                     {$dest = '???';}
     print STDOUT '*** '.$script.': timeout on waiting shutter to '.$dest.' (state='.$shutter_state.', dest='.$shutter_dest.', wait='.$wait_time.')/'."\n";
#    printf STDOUT '*** '.$script.': wait_time=%.3fs rqs=%d state=%d diff=%.3f'."\n",
#                                   $wait_time,$shutter_dest,$shutter_state,$diff;
     return -1;                                                         #error
  }
}

#=================================================================

# sub getShutter ($shutterRBK) {				# returns Shutter condition: 1=open, 0=close
sub getShutter ($) {

  our ($shutter_open, $report);                                 #this is from beamline.pl
  my ($script, $shutterRBK, $status, $shutter_state);

  $script     = 'getShutter';
  $shutterRBK = shift(@_);
  if (!defined $report) {$report=0;};

  if (defined $shutterRBK) {
### This must be defined in beamline.h:
     if (!defined $shutter_open) {
        &press_key_exit('*** '.$script.': global variable "$shutter_open" is undefined.');
     }
     ($status, $shutter_state) = Pezca::GetDouble($shutterRBK);
     if ($status) {&report_only($script, $status, $shutterRBK); return -1;}	#failed
     if ($shutter_state == $shutter_open) {return 1;}				#opened
     else                                 {return 0;}				#closed
  } else {
     if ($report) {print STDOUT '!!! '.$script.': Shutter RBK is undefined -- presuming open.'."\n";}
     return 1;									#opened
  }
}

#=================================================================

# sub openEndStationShutter($beamline) {			# returns $status: 0=success, 1=failure
sub openEndStationShutter($) {

  our ($beamline);						#this is from beamline.pl
  our ($HUTCHshutterOPE,  $HUTCHshutterCLO);			#this is from beamline.pl
  our ($HUTCHshutterOpen, $HUTCHshutterClosed);			#this is from beamline.pl
  our ($HUTCHshutterSTS);					# this is from beamline.pl
  my ($script, $bl, $status, $host);
  my ($wait_max, $wait_, $wait_step, $shutterStatus);

  $script    = 'openEndStationShutter';
  $bl        = shift(@_);
  if (!defined $bl) {$bl = $beamline;}
  if ($bl =~ /^23d$/)     {return 0;}
  if ($bl !~ /^23[iob]$/) {return 1;}
  $wait_max  = 10;						#max seconds for shutter to open
  $wait_step = 0.2;

### Endshutter operation is only allowed from certain hosts:
  $host = $ENV{HOSTNAME};
  if (!defined $host) {$host=`hostname`; $host=~s/\n//g;}
  $host = lc($host);
  $host =~ s/.gmca.aps.anl.gov$//;
# if ($bl eq '23i' && $host ne 'bl1ws3') {return 1;}
# if ($bl eq '23o' && $host ne 'bl2ws3') {return 1;}
  if ($bl eq '23b' && $host ne 'bl3ws3') {return 1;}

  if (!defined $HUTCHshutterSTS) {
     print STDOUT '!!! '.$script.': shutter PV "HUTCHshutterSTS" is undefined'."\n"; 
     return -1;
  }
  $status = &addCaMonitor($HUTCHshutterSTS);
  if ($status) {
     &report_only($script, $status, $HUTCHshutterSTS);
     return 1;
  }
  ($status,$shutterStatus) = Pezca::GetDouble($HUTCHshutterSTS);
  if ($shutterStatus == $HUTCHshutterOpen) {return 0;}

### Send "Open" command:
  $status = Pezca::PutDoubleOldCa($HUTCHshutterOPE,1);
  if ($status) {
     &report_only($script, $status, $HUTCHshutterOPE);
     return 1;
  }

### Wait to complete:
  $wait_ = 0;
  do {
     select(undef,undef,undef,$wait_step);
     $wait_ += $wait_step;
     ($status,$shutterStatus) = Pezca::GetDouble($HUTCHshutterSTS);
  } while ($shutterStatus != $HUTCHshutterOpen && $wait_ < $wait_max);

  if ($shutterStatus != $HUTCHshutterOpen) {
     print STDOUT '!!! '.$script.': Could not open endstation shutter for beamline '.$bl."\n";
     print STDOUT '      CMD='.$HUTCHshutterOPE.' RBK='.$HUTCHshutterSTS."\n";
     return 1;
  } else {
     return 0;
  }
}

#=================================================================

# sub closeEndStationShutter($beamline) {			# returns $status: 0=success, 1=failure
sub closeEndStationShutter($) {

  our ($beamline);						#this is from beamline.pl
  our ($HUTCHshutterOPE,  $HUTCHshutterCLO);			#this is from beamline.pl
  our ($HUTCHshutterOpen, $HUTCHshutterClosed);			#this is from beamline.pl
  our ($HUTCHshutterSTS);					# this is from beamline.pl
  my ($script, $bl, $status, $host);
  my ($wait_max, $wait_, $wait_step, $shutterStatus);

  $script    = 'closeEndStationShutter';
  $bl        = shift(@_);
  if (!defined $bl) {$bl = $beamline;}
  if ($bl =~ /^23d$/)     {return 0;}
  if ($bl !~ /^23[iob]$/) {return 1;}
  $wait_max  = 10;						#max seconds for shutter to open
  $wait_step = 0.2;

### Endshutter operation is only allowed from certain hosts:
  $host = $ENV{HOSTNAME};
  if (!defined $host) {$host=`hostname`; $host=~s/\n//g;}
  $host = lc($host);
  $host =~ s/.gmca.aps.anl.gov$//;
# if ($bl eq '23i' && $host ne 'bl1ws3') {return 1;}
# if ($bl eq '23o' && $host ne 'bl2ws3') {return 1;}
  if ($bl eq '23b' && $host ne 'bl3ws3') {return 1;}

  if (!defined $HUTCHshutterSTS) {
     print STDOUT '!!! '.$script.': shutter PV "HUTCHshutterSTS" is undefined'."\n"; 
     return -1;
  }
  $status = &addCaMonitor($HUTCHshutterSTS);
  if ($status) {
     &report_only($script, $status, $HUTCHshutterSTS);
     return 1;
  }
  ($status,$shutterStatus) = Pezca::GetDouble($HUTCHshutterSTS);
  if ($shutterStatus == $HUTCHshutterClosed) {return 0;}

### Send "Close" command:
  $status = Pezca::PutDoubleOldCa($HUTCHshutterCLO,1);
  if ($status) {
     &report_only($script, $status, $HUTCHshutterCLO);
     return 1;
  }

### Wait to complete:
  $wait_ = 0;
  do {
     select(undef,undef,undef,$wait_step);
     $wait_ += $wait_step;
     ($status,$shutterStatus) = Pezca::GetDouble($HUTCHshutterSTS);
  } while ($shutterStatus != $HUTCHshutterClosed && $wait_ < $wait_max);

  if ($shutterStatus != $HUTCHshutterClosed) {
     print STDOUT '!!! '.$script.': Could not close endstation shutter for beamline '.$bl."\n";
     print STDOUT '      CMD='.$HUTCHshutterCLO.' RBK='.$HUTCHshutterSTS."\n";
     return 1;
  } else {
     return 0;
  }
}

#=================================================================

# sub getEndStationShutter($beamline) {				# returns $status: 0=closed, 1=open, -1=failure
sub getEndStationShutter($) {

  our ($beamline);						#this is from beamline.pl
  our ($HUTCHshutterOPE,  $HUTCHshutterCLO);			#this is from beamline.pl
  our ($HUTCHshutterOpen, $HUTCHshutterClosed);			#this is from beamline.pl
  our ($HUTCHshutterSTS);					# this is from beamline.pl
  my ($script, $bl, $status, $shutterStatus);

  $script = 'getEndStationShutter';
  $bl     = shift(@_);
  if (!defined $bl) {$bl = $beamline;}
  if ($bl !~ /^23[iob]$/) {return 1;}

  if (!defined $HUTCHshutterSTS) {
     print STDOUT '!!! '.$script.': shutter PV "HUTCHshutterSTS" is undefined'."\n"; 
     return -1;
  }
  ($status,$shutterStatus) = Pezca::GetDouble($HUTCHshutterSTS);
  if ($status) {
    &report_only($script, $status, $HUTCHshutterSTS);
    return -1;
  }

  if ($shutterStatus == $HUTCHshutterOpen) {return 1;}
  else                                     {return 0;}
}

#=================================================================

# sub openCshutter($beamline) {					# returns $status: 0=success, 1=failure
sub openCshutter($) {

  our ($beamline);						#this is from beamline.pl
  our ($CshutterOPE,  $CshutterCLO);				#this is from beamline.pl
  our ($CshutterOpen, $CshutterClosed, $CshutterSTS);		#this is from beamline.pl
  my ($script, $bl, $status, $host);
  my ($wait_max, $wait_, $wait_step, $shutterStatus);

  $script    = 'openCshutter';
  $bl        = shift(@_);
  if (!defined $bl) {$bl = $beamline;}
  if ($bl =~ /^23d$/)     {return 0;}				#for debugging in the lab
  if ($bl !~ /^23i$/)     {return 1;}
  $wait_max  = 10;						#max seconds for shutter to open
  $wait_step = 0.2;

  if (!defined $CshutterSTS) {
     print STDOUT '!!! '.$script.': shutter PV "CshutterSTS" is undefined'."\n"; 
     return -1;
  }
  $status = &addCaMonitor($CshutterSTS);
  if ($status) {
     &report_only($script, $status, $CshutterSTS);
     return 1;
  }
  ($status,$shutterStatus) = Pezca::GetDouble($CshutterSTS);
  if ($shutterStatus == $CshutterOpen) {return 0;}

### C-shutter operation is only allowed from certain hosts:
  $host = $ENV{HOSTNAME};
  if (!defined $host) {$host=`hostname`; $host=~s/\n//g;}
  $host = lc($host);
  $host =~ s/.gmca.aps.anl.gov$//;
  if ($host ne 'bl1ws2'      && 
      $host ne 'bl1ws3'      && 
      $host ne 'bl1keithley' && 
      $host ne 'bl1upper'    &&
      $host ne 'directnet') {return 1;}

### Send "Open" command:
  $status = Pezca::PutDoubleOldCa($CshutterOPE,1);
  if ($status) {
     &report_only($script, $status, $CshutterOPE);
     return 1;
  }

### Wait to complete:
  $wait_ = 0;
  do {
     select(undef,undef,undef,$wait_step);
     $wait_ += $wait_step;
     ($status,$shutterStatus) = Pezca::GetDouble($CshutterSTS);
  } while ($shutterStatus != $CshutterOpen && $wait_ < $wait_max);

  if ($shutterStatus != $CshutterOpen) {
     print STDOUT '!!! '.$script.': Could not open C-hutch shutter for beamline '.$bl."\n";
     print STDOUT '      CMD='.$CshutterOPE.' RBK='.$CshutterSTS."\n";
     return 1;
  } else {
     return 0;
  }
}

#=================================================================

# sub closeCshutter($beamline) {				# returns $status: 0=success, 1=failure
sub closeCshutter($) {

  our ($beamline);						#this is from beamline.pl
  our ($CshutterOPE,  $CshutterCLO);				#this is from beamline.pl
  our ($CshutterOpen, $CshutterClosed, $CshutterSTS);		#this is from beamline.pl
  my ($script, $bl, $status, $host);
  my ();
  my ($wait_max, $wait_, $wait_step, $shutterStatus);

  $script    = 'closeCshutter';
  $bl        = shift(@_);
  if (!defined $bl) {$bl = $beamline;}
  if ($bl =~ /^23d$/)     {return 0;}				#for debugging in the lab
  if ($bl !~ /^23i$/)     {return 1;}
  $wait_max  = 10;						#max seconds for shutter to open
  $wait_step = 0.2;

  if (!defined $CshutterSTS) {
     print STDOUT '!!! '.$script.': shutter PV "CshutterSTS" is undefined'."\n"; 
     return -1;
  }
  $status = &addCaMonitor($CshutterSTS);
  if ($status) {
     &report_only($script, $status, $CshutterSTS);
     return 1;
  }
  ($status,$shutterStatus) = Pezca::GetDouble($CshutterSTS);
  if ($shutterStatus == $CshutterClosed) {return 0;}

### C-shutter operation is only allowed from certain hosts:
  $host = $ENV{HOSTNAME};
  if (!defined $host) {$host=`hostname`; $host=~s/\n//g;}
  $host = lc($host);
  $host =~ s/.gmca.aps.anl.gov$//;
  if ($host ne 'bl1ws2'      && 
      $host ne 'bl1ws3'      && 
      $host ne 'bl1keithley' && 
      $host ne 'bl1upper'    &&
      $host ne 'directnet') {return 1;}

### Send "Close" command:
  $status = Pezca::PutDoubleOldCa($CshutterCLO,1);
  if ($status) {
     &report_only($script, $status, $CshutterCLO);
     return 1;
  }

### Wait to complete:
  $wait_ = 0;
  do {
     select(undef,undef,undef,$wait_step);
     $wait_ += $wait_step;
     ($status,$shutterStatus) = Pezca::GetDouble($CshutterSTS);
  } while ($shutterStatus != $CshutterClosed && $wait_ < $wait_max);

  if ($shutterStatus != $CshutterClosed) {
     print STDOUT '!!! '.$script.': Could not close C-hutch shutter for beamline '.$bl."\n";
     print STDOUT '      CMD='.$CshutterCLO.' RBK='.$CshutterSTS."\n";
     return 1;
  } else {
     return 0;
  }
}

#=================================================================

# sub getCshutter($beamline);		 			#1=open/not present,0=close,-1=not.responding (GMCA ID-in only)
sub getCshutter($) {

  our ($beamline);						#this is from beamline.pl
  our ($CshutterOPE,  $CshutterCLO);				#this is from beamline.pl
  our ($CshutterOpen, $CshutterClosed, $CshutterSTS);		#this is from beamline.pl
  my ($script, $bl, $status, $shutterStatus);

  $script = 'getCshutter';
  $bl     = shift(@_);
  if (!defined $bl) {$bl = $beamline;}
  if ($bl !~ /^23i$/) {return 1;}

  if (!defined $CshutterSTS) {
     print STDOUT '!!! '.$script.': shutter PV "CshutterSTS" is undefined'."\n"; 
     return -1;
  }
  ($status,$shutterStatus) = Pezca::GetDouble($CshutterSTS);
  if ($status) {
    &report_only($script, $status, $CshutterSTS);
    return -1;
  }

  if ($shutterStatus == $CshutterOpen) {return 1;}
  else                                 {return 0;}
}


#=================================================================

# sub openFrontEndShutter($beamline) {			# returns $status: 0=success, 1=failure
sub openFrontEndShutter($) {

  our ($beamline);						#this is from beamline.pl
  our ($FEshutterOPE,  $FEshutterCLO);				#this is from beamline.pl
  our ($FEshutterPOS,  $FEshutterSTS);				#this is from beamline.pl
  our ($FEshutterOpen, $FEshutterClosed);			#this is from beamline.pl
  my ($script, $bl, $status, $host);
  my ($wait_max, $wait_, $wait_step, $shutterStatus, $shutterPos);

  $script    = 'openFrontEndShutter';
  $bl        = shift(@_);
  if (!defined $bl) {$bl = $beamline;}
  if ($bl =~ /^23d$/)    {return 0;}				#for debugging in the lab
  if ($bl !~ /^23[io]$/) {return 1;}
  $wait_max  = 10;						#max seconds for shutter to open
  $wait_step = 0.2;

  if (!defined $FEshutterSTS) {
     print STDOUT '!!! '.$script.': shutter PV "FEshutterSTS" is undefined'."\n"; 
     return -1;
  }
  $status = &addCaMonitor($FEshutterSTS);
  if ($status) {
     &report_only($script, $status, $FEshutterSTS);
     return 1;
  }
  ($status,$shutterStatus) = Pezca::GetDouble($FEshutterSTS);
  if ($shutterStatus == $FEshutterOpen) {return 0;}

### FE shutter operation is only allowed from certain hosts:
  $host = $ENV{HOSTNAME};
  if (!defined $host) {$host=`hostname`; $host=~s/\n//g;}
  $host = lc($host);
  $host =~ s/.gmca.aps.anl.gov$//;
  if ($host ne 'bl1ws2' &&  $host ne 'bl2ws2' && $host ne 'bl1keithley' && $host ne 'bl1upper' &&
      $host ne 'bl1ws3' &&  $host ne 'bl2ws3' && $host ne 'bl2keithley' && $host ne 'bl2upper' &&
      $host ne 'directnet') {return 1;}

### Send "Open" command:
  $status = Pezca::PutDoubleOldCa($FEshutterOPE,1);
  if ($status) {
     &report_only($script, $status, $FEshutterOPE);
     return 1;
  }

### Wait to complete:
  $wait_ = 0;
  do {
     select(undef,undef,undef,$wait_step);
     $wait_ += $wait_step;
     ($status,$shutterStatus) = Pezca::GetDouble($FEshutterSTS);
  } while ($shutterStatus != $FEshutterOpen && $wait_ < $wait_max);

  if ($shutterStatus != $FEshutterOpen) {
     print STDOUT '!!! '.$script.': Could not open FE shutter for beamline '.$bl."\n";
     print STDOUT '      CMD='.$FEshutterOPE.' RBK='.$FEshutterSTS."\n";
     if (defined $FEshutterPOS) {
        ($status,$shutterPos) = Pezca::GetString($FEshutterPOS);
        if ($status) {$shutterPos='CAerr';}
        print STDOUT '      Current FE shutter state is: '.$shutterPos."\n";
     }
     return 1;
  } else {
     return 0;
  }
}

#=================================================================

# sub closeFrontEndShutter($beamline) {			# returns $status: 0=success, 1=failure
sub closeFrontEndShutter($) {

  our ($beamline);						#this is from beamline.pl
  our ($FEshutterOPE,  $FEshutterCLO);				#this is from beamline.pl
  our ($FEshutterPOS,  $FEshutterSTS);				#this is from beamline.pl
  our ($FEshutterOpen, $FEshutterClosed);			#this is from beamline.pl
  my ($script, $bl, $status, $host);
  my ($wait_max, $wait_, $wait_step, $shutterStatus, $shutterPos);

  $script    = 'closeFrontEndShutter';
  $bl        = shift(@_);
  if (!defined $bl) {$bl = $beamline;}
  if ($bl =~ /^23d$/)    {return 0;}				#for debugging in the lab
  if ($bl !~ /^23[io]$/) {return 1;}
  $wait_max  = 10;						#max seconds for shutter to open
  $wait_step = 0.2;

  if (!defined $FEshutterSTS) {
     print STDOUT '!!! '.$script.': shutter PV "FEshutterSTS" is undefined'."\n"; 
     return -1;
  }
  $status = &addCaMonitor($FEshutterSTS);
  if ($status) {
     &report_only($script, $status, $FEshutterSTS);
     return 1;
  }
  ($status,$shutterStatus) = Pezca::GetDouble($FEshutterSTS);
  if ($shutterStatus == $FEshutterClosed) {return 0;}

### FE shutter operation is only allowed from certain hosts:
  $host = $ENV{HOSTNAME};
  if (!defined $host) {$host=`hostname`; $host=~s/\n//g;}
  $host = lc($host);
  $host =~ s/.gmca.aps.anl.gov$//;
  if ($host ne 'bl1ws2' &&  $host ne 'bl2ws2' && $host ne 'bl1keithley' && $host ne 'bl1upper' &&
      $host ne 'bl1ws3' &&  $host ne 'bl2ws3' && $host ne 'bl2keithley' && $host ne 'bl2upper' &&
      $host ne 'directnet') {return 1;}

### Send "Close" command:
  $status = Pezca::PutDoubleOldCa($FEshutterCLO,1);
  if ($status) {
     &report_only($script, $status, $FEshutterCLO);
     return 1;
  }

### Wait to complete:
  $wait_ = 0;
  do {
     select(undef,undef,undef,$wait_step);
     $wait_ += $wait_step;
     ($status,$shutterStatus) = Pezca::GetDouble($FEshutterSTS);
  } while ($shutterStatus != $FEshutterClosed && $wait_ < $wait_max);

  if ($shutterStatus != $FEshutterClosed) {
     print STDOUT '!!! '.$script.': Could not close FE shutter for beamline '.$bl."\n";
     print STDOUT '      CMD='.$FEshutterCLO.' RBK='.$FEshutterSTS."\n";
     if (defined $FEshutterPOS) {
        ($status,$shutterPos) = Pezca::GetString($FEshutterPOS);
        if ($status) {$shutterPos='CAerr';}
        print STDOUT '!!! '.$script.': Current FE shutter state is: '.$shutterPos."\n";
     }
     return 1;
  } else {
     return 0;
  }
}

#=================================================================

# sub getFrontEndShutter($beamline) {				# returns $status: 0=closed, 1=open, -1=failure
sub getFrontEndShutter($) {

  our ($beamline);						#this is from beamline.pl
  our ($FEshutterOPE,  $FEshutterCLO);				#this is from beamline.pl
  our ($FEshutterPOS,  $FEshutterSTS);				#this is from beamline.pl
  our ($FEshutterOpen, $FEshutterClosed);			#this is from beamline.pl
  my ($script, $bl, $status, $shutterStatus);

  $script = 'getFrontEndShutter';
  $bl     = shift(@_);
  if (!defined $bl) {$bl = $beamline;}
  if ($bl !~ /^23[iob]$/) {return 1;}

  if (!defined $FEshutterSTS) {
     print STDOUT '!!! '.$script.': shutter PV "FEshutterSTS" is undefined'."\n"; 
     return -1;
  }
  ($status,$shutterStatus) = Pezca::GetDouble($FEshutterSTS);
  if ($status) {
    &report_only($script, $status, $FEshutterSTS);
    return -1;
  }

  if ($shutterStatus == $FEshutterOpen) {return 1;}
  else                                  {return 0;}
}

#=================================================================

 # sub getSRcurrent([\*$status,$LUN]) {				#returns $SRcurrent
sub getSRcurrent(;\$$) {

  my ($script, $statref, $LUN, $SRcurrentPV, $SRcurrent, $status);

  $script  = 'getSRcurrent';
  $statref = shift(@_);
  $LUN     = shift(@_);

  $SRcurrentPV = 'S:SRcurrentAI';

### Measure synchrotron current:
  ($status, $SRcurrent) = Pezca::GetDouble($SRcurrentPV);
  if ($status) {
     &report_only($script,$status,$SRcurrentPV,$LUN);
     $SRcurrent = 100.;						#should we take "0"?
  }
  if (defined $statref) {$$statref = $status};
  return $SRcurrent;
}

#=================================================================

 # sub getShutterPermit([\*$status,$LUN]) {				#returns $SRcurrent
sub getShutterPermit(;\$$) {

  my ($script, $statref, $LUN, $permitPV, $permit, $status);

  $script  = 'getShutterPermit';
  $statref = shift(@_);
  $LUN     = shift(@_);

  $permitPV = 'ACIS:ShutterPermit';

### Query APS shutters permit (0=no, 1=yes):
  ($status, $permit) = Pezca::GetDouble($permitPV);
  if ($status) {
     &report_only($script,$status,$permitPV,$LUN);
     $permit = 1;						#should we take "0"?
  }
  if (defined $statref) {$$statref = $status};
  return $permit;
}
