#!/usr/bin/perl -W
#--------------------------------------------------------------
# These are VIDEO (Axis and IQeye camera servers) related subs
#              for beamline controls Perl scripts
#
#     				Author: Sergey Stepanov
#
# Version-1.0:  2008/02/05 using LWP::UserAgent, Image::Magick and DBI
# Version-1.1:  2008/03/25 added imageCrossection with GnuPlot
# Version-2.0:  2008/04/02 changed imageCrossection: no tmp data in GnuPlot
#                          added wrappers for individual Horiz. & Vert. sections.
# Version-2.1:  2010/08/10 added getCameraSnap2 and saveCameraSnap2 that
#                          do not need mySQL and hires/lores terminology
# Version-3.0:  2014/01/13 added functions to work with PZT presets
#--------------------------------------------------------------
# List of functions:
# ^^^^^^^^^^^^^^^^^
#  1. $status = &getCameraList(\@cameranames);			#get available camera names from mySql
#
#  2. $status = &getCameraType($camera,\$type,\$ip);		#where camera=HiRes/LoRes/Hutch
#                                                               #returns camera type=iqeye/axis/axis1/...
#                                                               #and camera IP (reads mySQL)
#
#  3. $status = &getCameraTypes(\@types,\@ips);			#camera types and IPs,
#                                                               #data order: HiRes/LoRes/Hutch
#
#  4. $status = &getCameraSnap(\$image,$camera,[$zoom|$downsample,
#                            $type,$ip,$compression,$colorlevel,$rotation]);
#                                                               #returns \$image in form of ImageMagick structure
#                                                               #where camera=HiRes/LoRes/Hutch
#
# 4b. $status = &getCameraSnap2(\$image,$cameraType,$ip,
#                             [$zoom|$downsample,$compression,$colorlevel,$rotation]);
#                                                               #returns \$image in form of ImageMagick structure
#                                                               #where cameraType=iqeye/axis/axis1/.../axis4
#
#  5. $status = &saveCameraSnap($file,$camera,[$display,$zoom|$downsample,
#                            $type,$ip,$compression,$colorlevel,$rotation]);
#                                                               #where camera=HiRes/LoRes/Hutch
#
# 5a. $status = &saveCameraSnap2($file,$cameraType,$ip,[$display,
#                            $zoom|$downsample,,$compression,$colorlevel,$rotation]);
#                                                               #where cameraType=iqeye/axis/axis1/.../axis4
#
#  6. $status = &imageCrossectionH(\$image,$hsectionType,\@hsection,
#                                              [$gnuplot,$destroy]);
#
#  7. $status = &imageCrossectionV(\$image,$vsectionType,\@vsection,
#                                              [$gnuplot,$destroy]);
#
#  8. $status = &imageCrossection(\$image,$hsectionType,\@hsection,
#                                         $vsectionType,\@vsection,
#                                            [$gnuplot,$destroy]);
#                                                               #takes \$image in form of ImageMagick structure
#                                                               #$sectionType=nnn|first|last|center|integral|none|undef
#                                                               #\@section is reference to array where data is returned
#                                                               #$gnuplot=0|1 $destroy=0|1
#  9.           &msleep_gnuplot([$milliseconds]);		#default is 1ms
#
# 10. $status = &skipTopup([$topupInterval,$topupAllowance]);
#
# 11. $status = &getPZTcameraPresets(\@presetNames,[$url,$cameraNumber]);	#if $url is not given, takes hutch PZT camera  [AXIS only]
#
# 12. $status = &gotoPZTcameraPreset($preset,[$url,$cameraNumber]);		#if $url is not given, takes hutch PZT camera  [AXIS only]
#
# 13. $status = &openCameraVideoStream($url,[$camera,$resolution,            	# open in firefox                              [AXIS only]
#                                      $compression,$colorlevel,$clock,$date,
#                                      $quad,$text,$rotation,$req_fps]);
# 14. $status = &openCameraHomepage($url)			            	# open in firefox
#--------------------------------------------------------------
# use warnings::unused;
  use strict;
# use LWP::UserAgent;           # World-Wide Web library for Perl (libwww-perl)
# use Image::Magick;            # ImageMagick API for Perl
# use DBI;			# mySQL interface (for getting camera data)
# use File::Spec;               # one of the two (for displaying image with external prg)
## use Cwd 'abs_path';          # one of the two (for displaying image with external prg)
## use List::Util qw(min max);	# min/max functions (for zooming and cropping) -- replaced by minval
# use Time::HiRes qw(usleep);

#=================================================================

# sub getCameraList(\@cameras) {
sub getCameraList (\@) {				#also returns $status

### This routine returns names of all available cameras in the beamline database

  our ($mysqlHost, $mysqlDb, $mysqlUser, $mysqlPass);
  my ($script, $camerasRef, $mysqlTbl);
  my ($dsn, $dbh, $sth, $nrecords, $hashref, $i);

  $script     = 'getCameraList';
  $camerasRef = shift(@_);
  $mysqlTbl = 'CameraIP';

  @$camerasRef = ();
  $dsn = 'DBI:mysql:host='.$mysqlHost.';database='.$mysqlDb;
  $dbh = DBI->connect($dsn,$mysqlUser,$mysqlPass,{PrintError=>0,});
  if (!defined $dbh) {
     print STDOUT '*** '.$script.': Cannot connect to mySQL server '.$mysqlHost."\n";
     return -1;
  }
  $sth = $dbh->prepare('SELECT * FROM '.$mysqlTbl);
  $nrecords = $sth->execute();			#nrecords works for MySQL only. For SQLite result is def/undef only.
  if (!defined $nrecords) {
     print STDOUT '*** '.$script.': Failed reading table='.$mysqlTbl
                                .', DB='.$mysqlDb
                                .', mySQL server='.$mysqlHost."\n"
                 .'*** '.$script.': '.$sth->errstr."\n";
     $sth->finish();
     $dbh->disconnect();
     return -1;
  }
  elsif ($nrecords == 0) {
     print STDOUT '*** '.$script.': No records in table ='.$mysqlTbl
                                .', DB='.$mysqlDb
                                .', mySQL server='.$mysqlHost."\n";
     $sth->finish();
     $dbh->disconnect();
     return -1;
  }
# print STDOUT $script.': found '.$nrecords.' beamline cameras:'."\n";
  for ($i=0; $i<$nrecords; $i++) {
     $hashref = $sth->fetchrow_hashref();
     if (! defined $hashref->{'name'}) {
        print STDOUT '*** '.$script.': No field "name" in table='.$mysqlTbl
                                   .', DB='.$mysqlDb
                                   .', mySQL server '.$mysqlHost."\n";
        $sth->finish();
        $dbh->disconnect();
        return -1;
     }
#    print STDOUT '    '.$hashref->{'name'}."\n";
     push(@$camerasRef,$hashref->{'name'});
  }
  $sth->finish();
  $dbh->disconnect();
  return 0;
}

#=================================================================

# sub getCameraType($camera,\$type,\$ip) {
sub getCameraType ($\$\$) {

### This routine determines camera type (iqeye, axis, axis1, ... axis4)
### for a given camera (HighRes, LowRes, Hutch)

  our ($mysqlHost, $mysqlDb, $mysqlUser, $mysqlPass);
  my ($script, $camera, $typeRef, $ipRef, $mysqlTbl, @cameraList);
  my ($dsn, $dbh, $sth, $status, $nrecords, $keyword, $cam, @val, $nval);

  $script   = 'getCameraType';
  $camera   = shift(@_);
  $typeRef  = shift(@_);
  $ipRef    = shift(@_);
  $mysqlTbl = 'CameraIP';

  $status = &getCameraList(\@cameraList);
  if ($status) {
     print STDOUT '*** '.$script.': failed to fetch the list of available cameras'."\n";
     return -1;
  }

  $keyword = undef;
  foreach $cam (@cameraList) {
     if    ($cam =~ /^${camera}$/i)  {$keyword = $cam; last;}
     elsif ($camera =~ /^hires$/i
	    && $cam =~ /^HighRes$/i) {$keyword = $cam; last;}
     elsif ($camera =~ /^lores$/i
	    && $cam =~ /^LowRes$/i)  {$keyword = $cam; last;}
  }
  if (!defined $keyword)  {
     print STDOUT '*** '.$script.': unrecognized camera "'.$camera.'"'."\n";
     return -1;
  }

  $dsn = 'DBI:mysql:host='.$mysqlHost.';database='.$mysqlDb;
  $dbh = DBI->connect($dsn,$mysqlUser,$mysqlPass,{PrintError=>0,});
  if (!defined $dbh) {
     print STDOUT '*** '.$script.': Cannot connect to mySQL server '.$mysqlHost."\n";
     return -1;
  }
  $sth = $dbh->prepare('SELECT * FROM '.$mysqlTbl.' WHERE name="'.$keyword.'" ');
  $nrecords = $sth->execute();			#nrecords works for MySQL only. For SQLite result is def/undef only.
  if (!defined $nrecords) {
     print STDOUT '*** '.$script.': Failed reading name="'.$keyword.'"'."\n",
                  '***          in table '.$mysqlDb.'.'.$mysqlTbl."\n",
                  '***          at mySQL server '.$mysqlHost."\n";
     $sth->finish();
     $dbh->disconnect();
     return -1;
  }
  elsif ($nrecords == 0) {
     print STDOUT '*** '.$script.': No records found for name="'.$keyword.'"'."\n",
                  '***          in table '.$mysqlDb.'.'.$mysqlTbl."\n",
                  '***          at mySQL server '.$mysqlHost."\n";
     $sth->finish();
     $dbh->disconnect();
     return -1;
  }
  elsif ($nrecords > 1) {
     print STDOUT '*** '.$script.': Multiple records ('.$nrecords.') found for name="'.$keyword.'"'."\n",
                  '***          in table '.$mysqlDb.'.'.$mysqlTbl."\n",
                  '***          at mySQL server '.$mysqlHost."\n",
                  '***          '.$sth->errstr."\n";
     $sth->finish();
     $dbh->disconnect();
     return -1;
  }
  @val = $sth->fetchrow_array();
  $nval = @val;
  if ($nval != 3) {
     print STDOUT '*** '.$script.': Incorrect record ['."@val".'] for name="'.$keyword.'"'."\n",
                  '***          in table '.$mysqlDb.'.'.$mysqlTbl."\n",
                  '***          at  mySQL server '.$mysqlHost."\n";
     $sth->finish();
     $dbh->disconnect();
     return -1;
  }
  ($$typeRef, $$ipRef) = ($val[1],$val[2]);
  $sth->finish();
  $dbh->disconnect();
  return 0;
}

#=================================================================

# sub getCameraTypes(\@type,\@ip) {
sub getCameraTypes (\@\@) {

### This routine determines camera type (iqeye, axis, axis1, ... axis4)
### for all beamline cameras (HighRes, LowRes, Hutch)

  my ($script, $typeRef, $ipRef, @cameraList);
  my ($status, $cam, $type_i, $ip_i);

  $script   = 'getCameraTypes';
  $typeRef  = shift(@_);
  $ipRef    = shift(@_);

  $status = &getCameraList(\@cameraList);
  if ($status) {
     print STDOUT '*** '.$script.': failed to fetch the list of available cameras'."\n";
     return -1;
  }

  foreach $cam (@cameraList) {
     $status = &getCameraType($cam,\$type_i,\$ip_i);
     if ($status) {return $status;}
#    print 'camera='.$cam.'  type='.$type_i.'  ip='.$ip_i."\n";
     push(@$typeRef,$type_i);
     push(@$ipRef,$ip_i);
  }
  return 0;
}


#=================================================================


# sub getCameraSnap(\$image,$camera,[$zoom|$downsample,$type,$ip,$compression,$colorlevel,$rotation]) {
sub getCameraSnap (\$$;$$$$$$) {

### This routine grabs snapshot from a given camera and returns it to the
### calling program as an ImageMagick structure.
### Here:
### $camera = HighRes/HiRes, LowRes/LoRes, or Hutch
### $zoom/$downsample applies to IQeye cameras only. Zoom must be in range
### 1 to 4 and refers to image, not cropped in the camera. In case any zoom
### is requested (even $zoom=1) the image is resized to 640x480; otherwice
### the maximum resolutiuon is applied.
### Negative "zoom" value is treated as downsample and must be -1,-2,-3,-4.
### $type=iqeye|axis|axisn and ip are optional to save time on multiple calls.
### when not given, they are taken from mySQL.

  our ($IQEYE_MAXX, $IQEYE_MAXY);	#speedup: define IQeye cropping image in calling program
  my ($script, $imageRef, $camera);
  my ($zoom, $downsample, $type, $ip);
  my ($compression, $colorlevel, $rotation);
  my ($compressionDefault, $colorlevelDefault, $rotationDefault);
  my (@compressionList, @colorlevelList, @rotationList);
  my ($status, $cameraNo, $iqeyeMaxX, $iqeyeMaxY);
  my ($useragent, $response, $string, $img, $width, $height);
  my ($zoomSaved, $zoomX, $zoomY, $xcor, $ycor, $savX, $savY);
  my ($axisSavX,  $axisSavY, $iqeyeSavX, $iqeyeSavY, $zoomMax, $downsampleMax);

  $script      = 'getCameraSnap';
  $imageRef    = shift(@_);		#where to store ImageMagick structure
  $camera      = shift(@_);		#HiRes,HighRes,LoRes,LowRes,Hutch
  $zoom        = shift(@_);		#range: [+1 to +4]; iqeye camera only
# $downsample  = shift(@_);		#range: [-1 to -4]; iqeye camera only
  $type        = shift(@_);		#iqeye, axis, axis1, ... axis4
  $ip          = shift(@_);		#camera IP address
  $compression = shift(@_);		#axis: can be 0, 25, 50, 75, 100
  $colorlevel  = shift(@_);		#axis: can be ??? /0 = grayscale, 100 = full color/
  $rotation    = shift(@_);		#axis: can be 0, 90, 180, 270

  @compressionList = (0, 25, 50, 75, 100);
  @colorlevelList = (100, 0);
  @rotationList = (0, 90, 180, 270);
  $compressionDefault = $compressionList[0];
  $colorlevelDefault  = $colorlevelList[0];
  $rotationDefault    = $rotationList[0];
  if (!defined $compression) {$compression = $compressionDefault;}
  if (!defined $colorlevel)  {$colorlevel  = $colorlevelDefault;}
  if (!defined $rotation)    {$rotation    = $rotationDefault;}
  if (! grep(/^${compression}$/, @compressionList)) {
    print STDOUT '!!! '.$script.': incorrect compression='.$compression.".\n",
                 '   Using default='.$compressionDefault."\n";
    $compression = $compressionDefault;
  }
  if (! grep(/^${colorlevel}$/, @colorlevelList)) {
    print STDOUT '!!! '.$script.': incorrect colorlevel='.$colorlevel.".\n",
                 '   Using default='.$colorlevelDefault."\n";
    $colorlevel = $colorlevelDefault;
  }
  if (! grep(/^${rotation}$/, @rotationList)) {
    print STDOUT '!!! '.$script.': incorrect rotation='.$rotation.".\n",
                 '   Using default='.$rotationDefault."\n";
    $rotation = $rotationDefault;
  }

  if ((defined $IQEYE_MAXX) && (defined $IQEYE_MAXY)) {
     ($iqeyeMaxX, $iqeyeMaxY) = ($IQEYE_MAXX, $IQEYE_MAXY);	  #IQeye: cropped image
  } else {
     ($iqeyeMaxX, $iqeyeMaxY) = (2560, 1920);			  #IQeye: default image (4xVGA);
  }
  ($axisSavX,  $axisSavY)  = (704, 480);			  #image size we want for Axis
  ($iqeyeSavX, $iqeyeSavY) = (640, 480);			  #image size we *MAY* want for IQeye
  $zoomMax = &minval($iqeyeMaxX/$iqeyeSavX,$iqeyeMaxY/$iqeyeSavY); #needs List::Util
  $downsampleMax = 2*int($zoomMax/2);
  if ($downsampleMax < 1) {$downsampleMax=1;}

  if    ($camera =~ /^HighRes$/i || $camera =~ /^HiRes$/i) {$camera = 'HighRes';}
  elsif ($camera =~ /^LowRes$/i  || $camera =~ /^LoRes$/i) {$camera = 'LowRes';}
  elsif ($camera =~ /^Hutch$/i)                            {$camera = 'Hutch';}

### Attention: there is a difference here: if zoom is not specified,
### IQeye snapshot will target MAXIMUM resolution, but if zoom is given
### (even zoom=1), the snapshot will target 640x480 resolution.
### With downsample (passed as "negative" zoom), resolutiuon will be Max/downsample.
  if (!defined $zoom)  {$zoom = 0;}

  if(!defined $type || !defined $ip) {
     $status = &getCameraType($camera,\$type,\$ip);
     if ($status) {return -1;}
  }
  if ($type =~ /^axis/i) {
     $cameraNo = $type;
     $cameraNo =~ s/^axis[:-]?//i;
     if (length($cameraNo) == 0) {
        $cameraNo = 1;
     }
     elsif ($cameraNo !~ /^[1-4]?$/) {
        print STDOUT '*** '.$script.': incorrect AXIS server input ['.$cameraNo.']'."\n";
        return -1;
     }
     $type ='axis';
  }
  elsif ($type =~ /^iqeye/i) {
     $cameraNo = '';
     $type     ='iqeye';
  }
  else  {
     print STDOUT '*** '.$script.': unsupported camera type ['.$type.']'."\n";
     return -1;
  }

  if ($type eq 'iqeye') {
     if ($zoom > 0) {
        if ($zoom < 1 || $zoom > $zoomMax) {
           print STDOUT '!!! '.$script.': digital zoom ['.$zoom.'] is out of [1-'.$zoomMax.'] range. Ignoring...'."\n";
           $zoom = 1;
        }
        ($savX, $savY) = ($iqeyeSavX, $iqeyeSavY);		#image size we want
#	print STDOUT "IQeye: maxX=${iqeyeMaxX}  zoom=${zoom} savX=${savX}\n";
        $downsample = int($iqeyeMaxX/($zoom*$savX));
        $downsample = 2*int($downsample/2);
	if ($downsample < 1) {$downsample = 1;}
     }
     elsif ($zoom < 0) {
        $downsample = -$zoom;		#when negative value passed, it is considered downsample
        $zoom = 0;
	$downsample = 2*int($downsample/2);
	if ($downsample < 1) {$downsample = 1;}
        if ($downsample < 1 || $downsample > $downsampleMax) {
           print STDOUT '!!! '.$script.': downsample ['.$downsample.'] not in [1-'.$downsampleMax.'] range. Ignoring...'."\n";
           $downsample = 1;
        }
        ($savX, $savY) = (undef, undef); 			#unused
     }
     elsif ($zoom == 0) {
        ($savX, $savY) = (undef, undef); 			#unused
        $downsample = 1;
     }
  }
  elsif ($type eq 'axis') {
### These parameters are unsupported with Axis:
     $zoom = undef;
     $downsample = undef;
     ($savX, $savY) = ($axisSavX, $axisSavY);			#image size we want
  }

  $useragent = new LWP::UserAgent;
  $useragent->timeout(10);					# timeout in 60 seconds
  $useragent->agent('Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.8.1.8) Gesko/20071008 Firefox/2.0.0.8');

  if ($type eq 'iqeye') {
     $string = 'http://'.$ip.'/now.jpg'
              .'?snap=pre'			#new picture before taking
              .'&ds='.$downsample               #downsample: 1,2,4,8,16,VGA,QVGA,QQVGA
#             .'&jq=85'                         #**unsupported on 7xx** JPEG quality: 8-92
#             .'&mo=1'                          #**unsupported on 7xx** make monochrome (1/0)
#             .'&sh=0'                          #**unsupported on 7xx** sharpen by 0-100
              ;
  }
  elsif ($type eq 'axis') {
     $string = 'http://'.$ip.'/axis-cgi/jpg/image.cgi'
              .'?camera='.$cameraNo
	      .'&resolution='.$savX.'x'.$savY
              .'&compression='.$compression
              .'&colorlevel='.$colorlevel
              .'&rotation='.$rotation
              ;
  }

  $response = $useragent->get($string);
  if (! $response->is_success) {
     print STDOUT '*** '.$script.': HTTP request failed!'."\n",
                  'Request:  '.$string."\n",
                  'Response: '.$response->status_line."\n";
     return -1;
  }
  $img = $response->content;

### Read the image from memory:
  $$imageRef=Image::Magick->new(magick=>'JPG');	#allocate image structure
  $$imageRef->BlobToImage($img);			#read image from block of bytes
  $width  = $$imageRef->Get('columns');
  $height = $$imageRef->Get('rows');
  print STDOUT $camera.' camera: Image size is '.$width.'x'.$height."\n";

  if ($type eq 'iqeye' && $zoom > 0) {
     if ($width*$downsample < $iqeyeMaxX || $height*$downsample < $iqeyeMaxY) {
        print STDOUT '!!! '.$script.': IQeye image is cropped in the camera.'."\n";
### Recalculate zoom to the maximum value with given cropping:
	$zoomSaved = $zoom;
	$zoom = $zoom * $downsample * &minval($width/$iqeyeMaxX, $height/$iqeyeMaxY);
	$iqeyeMaxX  = $width*$downsample;
	$iqeyeMaxY  = $height*$downsample;
        $downsample = &minval(int($iqeyeMaxX/($zoom*$savX)),int($iqeyeMaxY/($zoom*$savY)));
	$downsample = 2*int($downsample/2);
	if ($downsample < 1) {$downsample = 1;}
        print  STDOUT '!!! '.$script.': Cropped image size='.$iqeyeMaxX.'x'.$iqeyeMaxY.".\n";
        printf STDOUT '!!! '.$script.': zoom will be reduced from %3.2f to %3.2f'."\n"
                                    .'    (proportionally to cropping)'."\n",
                                     $zoomSaved, $zoom;
        print  STDOUT '!!! '.$script.': Retaking with downsample='.$downsample.' ...'."\n";
        $string = 'http://'.$ip.'/now.jpg'
                 .'?snap=pre'			#new picture before taking
                 .'&ds='.$downsample		#downsample: 1,2,4,8,16,VGA,QVGA,QQVGA
                 ;
        $response = $useragent->get($string);
        if (! $response->is_success) {
           print STDOUT '*** '.$script.': HTTP request failed!'."\n",
                  'Request: '.$string."\n",
                  'Response: '.$response->status_line."\n";
           return -1;
        }
        $img = $response->content;
        $$imageRef->BlobToImage($img);			#read image from block of bytes
        $width  = $$imageRef->Get('columns');
        $height = $$imageRef->Get('rows');
     }
     $zoomX = int($width/$zoom);			# part we want to see
     $zoomY = int($height/$zoom);                       # part we want to see
     $xcor  = int(($width - $zoomX)/2);
     $ycor  = int(($height - $zoomY)/2);
     if ($xcor > 0 || $ycor > 0) {
### Crop image to zoomed area:
        $status = $$imageRef->Crop(geometry=>'72x72', width=>$zoomX, height=>$zoomY, x=>$xcor, y=>$ycor, fuzz=>1);
#       $status = $$imageRef->Crop(geometry=>'$zoomXx$zoomY+$xcor+$ycor');
        if ($status) {
           print STDOUT '*** '.$script.' image cropping from '.$width.'x'.$height
                                                       .' to '.$zoomX.'x'.$zoomY.' failed'."\n";
	   return -1;
        } else {
           print STDOUT $camera.' camera: Image cropped from '.$width.'x'.$height
                                                       .' to '.$zoomX.'x'.$zoomY."\n";
        }
     }
### Resample to 640x480:
### Available filters:
### Point,Box,Triangle,Hermite,Hanning,Hamming,Blackman,Gaussian,Quadratic,Cubic,Catrom,Mitchell,Lanczos,Bessel,Sinc
#    $status = $$imageRef->AdaptiveResize(geometry=>'72x72', width=>640, height=>480);
     $status = $$imageRef->Resize(geometry=>'72x72', width=>$savX, height=>$savY, filter=>'Hermite', support=>1);
     if ($status) {
         print STDOUT '*** '.$script.' image resizing from '.$zoomX.'x'.$zoomY
                                                      .' to '.$savX.'x'.$savY.' failed'."\n";
	return -1;
     } else {
        print STDOUT $camera.' camera: Image resized from '.$zoomX.'x'.$zoomY
                                                     .' to '.$savX.'x'.$savY."\n";
     }
  }

  return 0;
}


#=================================================================


# sub getCameraSnap2(\$image,$cameraType,$ip,[$zoom|$downsample,$compression,$colorlevel,$rotation]) {
sub getCameraSnap2 (\$$$;$$$$) {

### This routine grabs snapshot from a given camera and returns it to the
### calling program as an ImageMagick object.
### Here:
### $cameraType = iqeye|axis|axisn (n=1,...4)
### $ip = camera IP
### $zoom/$downsample applies to IQeye cameras only. Zoom must be in range
### 1 to 4 and refers to image, not cropped in the camera. In case any zoom
### is requested (even $zoom=1) the image is resized to 640x480; otherwice
### the maximum resolutiuon is applied.
### Negative "zoom" value is treated as downsample and must be -1,-2,-3,-4.

  our ($IQEYE_MAXX, $IQEYE_MAXY);	#speedup: define IQeye cropping image in calling program
  my ($script, $imageRef, $camera, $ip, $zoom, $downsample);
  my ($compression, $colorlevel, $rotation);
  my ($compressionDefault, $colorlevelDefault, $rotationDefault);
  my (@compressionList, @colorlevelList, @rotationList);
  my ($status, $cameraNo, $iqeyeMaxX, $iqeyeMaxY);
  my ($useragent, $response, $string, $img, $width, $height);
  my ($zoomSaved, $zoomX, $zoomY, $xcor, $ycor, $savX, $savY);
  my ($axisSavX,  $axisSavY, $iqeyeSavX, $iqeyeSavY, $zoomMax, $downsampleMax);

  $script      = 'getCameraSnap2';
  $imageRef    = shift(@_);		#where to store ImageMagick structure
  $camera      = shift(@_);		#iqeye, axis, axis1, ... axis4
  $ip          = shift(@_);		#camera IP address
  $zoom        = shift(@_);		#range: [+1 to +4]; iqeye camera only
# $downsample  = shift(@_);		#range: [-1 to -4]; iqeye camera only
  $compression = shift(@_);		#axis: can be 0, 25, 50, 75, 100
  $colorlevel  = shift(@_);		#axis: can be ??? /0 = grayscale, 100 = full color/
  $rotation    = shift(@_);		#axis: can be 0, 90, 180, 270

  @compressionList = (0, 25, 50, 75, 100);
  @colorlevelList = (100, 0);
  @rotationList = (0, 90, 180, 270);
  $compressionDefault = $compressionList[0];
  $colorlevelDefault  = $colorlevelList[0];
  $rotationDefault    = $rotationList[0];
  if (!defined $compression) {$compression = $compressionDefault;}
  if (!defined $colorlevel)  {$colorlevel  = $colorlevelDefault;}
  if (!defined $rotation)    {$rotation    = $rotationDefault;}
  if (! grep(/^${compression}$/, @compressionList)) {
    print STDOUT '!!! '.$script.': incorrect compression='.$compression.".\n",
                 '   Using default='.$compressionDefault."\n";
    $compression = $compressionDefault;
  }
  if (! grep(/^${colorlevel}$/, @colorlevelList)) {
    print STDOUT '!!! '.$script.': incorrect colorlevel='.$colorlevel.".\n",
                 '   Using default='.$colorlevelDefault."\n";
    $colorlevel = $colorlevelDefault;
  }
  if (! grep(/^${rotation}$/, @rotationList)) {
    print STDOUT '!!! '.$script.': incorrect rotation='.$rotation.".\n",
                 '   Using default='.$rotationDefault."\n";
    $rotation = $rotationDefault;
  }

  if ((defined $IQEYE_MAXX) && (defined $IQEYE_MAXY)) {
     ($iqeyeMaxX, $iqeyeMaxY) = ($IQEYE_MAXX, $IQEYE_MAXY);	  #IQeye: cropped image
  } else {
     ($iqeyeMaxX, $iqeyeMaxY) = (2560, 1920);			  #IQeye: default image (4xVGA);
  }
  ($axisSavX,  $axisSavY)  = (704, 480);			  #image size we want for Axis
  ($iqeyeSavX, $iqeyeSavY) = (640, 480);			  #image size we *MAY* want for IQeye
  $zoomMax = &minval($iqeyeMaxX/$iqeyeSavX,$iqeyeMaxY/$iqeyeSavY);	  #needs List::Util
  $downsampleMax = 2*int($zoomMax/2);
  if ($downsampleMax < 1) {$downsampleMax=1;}

### Attention: there is a difference here: if zoom is not specified,
### IQeye snapshot will target MAXIMUM resolution, but if zoom is given
### (even zoom=1), the snapshot will target 640x480 resolution.
### With downsample (passed as "negative" zoom), resolutiuon will be Max/downsample.
  if (!defined $zoom)  {$zoom = 0;}

  if ($camera =~ /^axis/i) {
     $cameraNo = $camera;
     $cameraNo =~ s/^axis[:-]?//i;
     if (length($cameraNo) == 0) {
        $cameraNo = 1;
     }
     elsif ($cameraNo !~ /^[1-4]?$/) {
        print STDOUT '*** '.$script.': incorrect AXIS server input ['.$cameraNo.']'."\n";
        return -1;
     }
     $camera ='axis';
  }
  elsif ($camera =~ /^iqeye/i) {
     $cameraNo = '';
     $camera   ='iqeye';
  }
  else  {
     print STDOUT '*** '.$script.': unsupported camera type ['.$camera.']'."\n";
     return -1;
  }

  if ($camera eq 'iqeye') {
     if ($zoom > 0) {
        if ($zoom < 1 || $zoom > $zoomMax) {
           print STDOUT '!!! '.$script.': digital zoom ['.$zoom.'] is out of [1-'.$zoomMax.'] range. Ignoring...'."\n";
           $zoom = 1;
        }
        ($savX, $savY) = ($iqeyeSavX, $iqeyeSavY);		#image size we want
#	print STDOUT "IQeye: maxX=${iqeyeMaxX}  zoom=${zoom} savX=${savX}\n";
        $downsample = int($iqeyeMaxX/($zoom*$savX));
        $downsample = 2*int($downsample/2);
	if ($downsample < 1) {$downsample = 1;}
     }
     elsif ($zoom < 0) {
        $downsample = -$zoom;					#when negative value passed, it is considered downsample
        $zoom = 0;
	$downsample = 2*int($downsample/2);
	if ($downsample < 1) {$downsample = 1;}
        if ($downsample < 1 || $downsample > $downsampleMax) {
           print STDOUT '!!! '.$script.': downsample ['.$downsample.'] not in [1-'.$downsampleMax.'] range. Ignoring...'."\n";
           $downsample = 1;
        }
        ($savX, $savY) = (undef, undef); 			#unused
     }
     elsif ($zoom == 0) {
        ($savX, $savY) = (undef, undef); 			#unused
        $downsample = 1;
     }
  }
  elsif ($camera eq 'axis') {
### These parameters are unsupported with Axis:
     $zoom = undef;
     $downsample = undef;
     ($savX, $savY) = ($axisSavX, $axisSavY);			#image size we want
  }

  $useragent = new LWP::UserAgent;
  $useragent->timeout(10);					# timeout in 60 seconds
  $useragent->agent('Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.8.1.8) Gesko/20071008 Firefox/2.0.0.8');

  if ($camera eq 'iqeye') {
     $string = 'http://'.$ip.'/now.jpg'
              .'?snap=pre'			#new picture before taking
              .'&ds='.$downsample               #downsample: 1,2,4,8,16,VGA,QVGA,QQVGA
#             .'&jq=85'                         #**unsupported on 7xx** JPEG quality: 8-92
#             .'&mo=1'                          #**unsupported on 7xx** make monochrome (1/0)
#             .'&sh=0'                          #**unsupported on 7xx** sharpen by 0-100
              ;
  }
  elsif ($camera eq 'axis') {
     $string = 'http://'.$ip.'/axis-cgi/jpg/image.cgi'
              .'?camera='.$cameraNo
	      .'&resolution='.$savX.'x'.$savY
              .'&compression='.$compression
              .'&colorlevel='.$colorlevel
              .'&rotation='.$rotation
              ;
  }

  $response = $useragent->get($string);
  if (! $response->is_success) {
     print STDOUT '*** '.$script.': HTTP request failed!'."\n",
                  'Request:  '.$string."\n",
                  'Response: '.$response->status_line."\n";
     return -1;
  }
  $img = $response->content;

### Read the image from memory:
  $$imageRef=Image::Magick->new(magick=>'JPG');	#allocate image structure
  $$imageRef->BlobToImage($img);			#read image from block of bytes
  $width  = $$imageRef->Get('columns');
  $height = $$imageRef->Get('rows');
  print STDOUT $camera.' camera: Image size is '.$width.'x'.$height."\n";

  if ($camera eq 'iqeye' && $zoom > 0) {
     if ($width*$downsample < $iqeyeMaxX || $height*$downsample < $iqeyeMaxY) {
        print STDOUT '!!! '.$script.': IQeye image is cropped in the camera.'."\n";
### Recalculate zoom to the maximum value with given cropping:
	$zoomSaved = $zoom;
	$zoom = $zoom * $downsample * &minval($width/$iqeyeMaxX, $height/$iqeyeMaxY);
	$iqeyeMaxX  = $width*$downsample;
	$iqeyeMaxY  = $height*$downsample;
        $downsample = &minval(int($iqeyeMaxX/($zoom*$savX)),int($iqeyeMaxY/($zoom*$savY)));
	$downsample = 2*int($downsample/2);
	if ($downsample < 1) {$downsample = 1;}
        print  STDOUT '!!! '.$script.': Cropped image size='.$iqeyeMaxX.'x'.$iqeyeMaxY.".\n";
        printf STDOUT '!!! '.$script.': zoom will be reduced from %3.2f to %3.2f'."\n"
                                    .'    (proportionally to cropping)'."\n",
                                     $zoomSaved, $zoom;
        print  STDOUT '!!! '.$script.': Retaking with downsample='.$downsample.' ...'."\n";
        $string = 'http://'.$ip.'/now.jpg'
                 .'?snap=pre'			#new picture before taking
                 .'&ds='.$downsample		#downsample: 1,2,4,8,16,VGA,QVGA,QQVGA
                 ;
        $response = $useragent->get($string);
        if (! $response->is_success) {
           print STDOUT '*** '.$script.': HTTP request failed!'."\n",
                  'Request: '.$string."\n",
                  'Response: '.$response->status_line."\n";
           return -1;
        }
        $img = $response->content;
        $$imageRef->BlobToImage($img);			#read image from block of bytes
        $width  = $$imageRef->Get('columns');
        $height = $$imageRef->Get('rows');
     }
     $zoomX = int($width/$zoom);			# part we want to see
     $zoomY = int($height/$zoom);                       # part we want to see
     $xcor  = int(($width - $zoomX)/2);
     $ycor  = int(($height - $zoomY)/2);
     if ($xcor > 0 || $ycor > 0) {
### Crop image to zoomed area:
        $status = $$imageRef->Crop(geometry=>'72x72', width=>$zoomX, height=>$zoomY, x=>$xcor, y=>$ycor, fuzz=>1);
#       $status = $$imageRef->Crop(geometry=>'$zoomXx$zoomY+$xcor+$ycor');
        if ($status) {
           print STDOUT '*** '.$script.' image cropping from '.$width.'x'.$height
                                                       .' to '.$zoomX.'x'.$zoomY.' failed'."\n";
	   return -1;
        } else {
           print STDOUT $camera.' camera: Image cropped from '.$width.'x'.$height
                                                       .' to '.$zoomX.'x'.$zoomY."\n";
        }
     }
### Resample to 640x480:
### Available filters:
### Point,Box,Triangle,Hermite,Hanning,Hamming,Blackman,Gaussian,Quadratic,Cubic,Catrom,Mitchell,Lanczos,Bessel,Sinc
#    $status = $$imageRef->AdaptiveResize(geometry=>'72x72', width=>640, height=>480);
     $status = $$imageRef->Resize(geometry=>'72x72', width=>$savX, height=>$savY, filter=>'Hermite', support=>1);
     if ($status) {
        print STDOUT '*** '.$script.' image resizing from '.$zoomX.'x'.$zoomY
                                                     .' to '.$savX.'x'.$savY.' failed'."\n";
	return -1;
     } else {
        print STDOUT $camera.' camera: Image resized from '.$zoomX.'x'.$zoomY
                                                     .' to '.$savX.'x'.$savY."\n";
     }
  }

  return 0;
}


#=================================================================


# sub saveCameraSnap($file,$camera,[$display,$zoom|$downsample,$type,$ip,$compression,$colorlevel,$rotation]) {
sub saveCameraSnap ($$;$$$$$$$) {

### This routine grabs snapshot from a given camera (using getCameraSnap)
### and then saves it in a file.
### Here:
### $camera = HighRes/HiRes, LowRes/LoRes, or Hutch
### $display = 0 or 1 (call or not eog/irfan)
### $zoom/$downsample applies to IQeye cameras only. Zoom must be in range
### 1 to 4 and refers to image, not cropped in the camera. In case any zoom
### is requested (even $zoom=1) the image is resized to 640x480; otherwice
### the maximum resolutiuon is applied.
### Negative "zoom" value is treated as downsample and must be -1,-2,-3,-4.
### $type=iqeye|axis|axisn and ip are optional to save time on multiple calls.
### when not given, they are taken from mySQL.

  my ($script, $file, $camera, $display, $zoom, $type, $ip);
  my ($compression, $colorlevel, $rotation, @formatList);
  my ($status, $image, $name, $ext, $imgFormat, $fullpath, $eog);

  $script      = 'saveCameraSnap';
  $file        = shift(@_);
  $camera      = shift(@_);		#HiRes,HighRes,LoRes,LowRes,Hutch
  $display     = shift(@_);		#flag: 1=yes,0=no
  $zoom        = shift(@_);		#range: +1:+4; iqeye camera only
# $downsample  = shift(@_);		#range: -1:-4; iqeye camera only
  $type        = shift(@_);		#iqeye, axis, axis1, ... axis4
  $ip          = shift(@_);		#camera IP address
  $compression = shift(@_);		#axis: can be 0, 25, 50, 75, 100
  $colorlevel  = shift(@_);		#axis: can be ??? /0 = grayscale, 100 = full color/
  $rotation    = shift(@_);		#axis: can be 0, 90, 180, 270

### External image viewer:
  if ($^O !~ /Win/i) {
    $eog    = &Which('eog');					# eye of Gnome
    if (!defined $eog) {$eog = &Which('eom');}			# eye of MATE
  } else {
    if    (-f 'C:/Progra~1/IrfanView/i_view32.exe') {$eog = 'C:/Progra~1/IrfanView/i_view32.exe';}
    elsif (-f 'C:/Progra~2/IrfanView/i_view32.exe') {$eog = 'C:/Progra~2/IrfanView/i_view32.exe';}
    elsif (-f 'C:/Progra~1/IrfanView/i_view64.exe') {$eog = 'C:/Progra~1/IrfanView/i_view64.exe';}
  }

### Supported image formats (determined from file name):
  @formatList  = ('jpg', 'jpeg', 'tif', 'tiff', 'png', 'gif', 'bmp');

### Find image format from the file name:
  ($name,$ext) = &analyzeFilename($file);
  if (!grep(/^${ext}$/i, @formatList)) {
     print STDOUT '*** '.$script.': unsupported image format ['.$ext.']'."\n";
     return -1;
  }
  $imgFormat = $ext;

  if (!defined $display) {$display = 0;}
  if ($display < 0 || $display > 1) {
     print STDOUT '!!! '.$script.': display flag ['.$display.'] must be 0 or 1. Ignoring...'."\n";
     $display = 0;
  }

  $status = &getCameraSnap(\$image,$camera,$zoom,$type,$ip,$compression,$colorlevel,$rotation);
  if ($status) {
     print STDOUT '*** '.$script.': call of getCameraSnap returned status='.$status."\n";
     return $status;
  }


### Set image compression. This can be one of:
### None, BZip, Fax, Group4, JPEG, JPEG2000, LosslessJPEG, LZW, RLE, Zip
### If lossless compression is required, then TIFF must use Zip or LZW.
### For PNG the default compression parameter is ZIP.
  if ($imgFormat =~ /^tif/i) {
#   $image->Set(compression=>'None');		        # this is default
    $image->Set(compression=>'Zip');
#   $image->Set(compression=>'LZW');
  }

  print STDOUT 'Saving image as '.$file."\n";
  $status = &ask_to_overwrite($file);			#this will ask for overwriting, unless NONINTERACTIVE=1
  if ($status) {$image->Write($file);}
  else         {return -1;}

  if ($display) {
     if ($^O !~ /Win/i) {
	if (defined $eog) {system($eog.' "'.$file.'" &');}
        else  {print STDOUT '!!! '.$script.': no image viewer found in the system path!'."\n";}
    }
     else {
        $fullpath = File::Spec->rel2abs($file);
#       $fullpath = abs_path($file);
	if    (defined $eog) {system(1, $eog, $fullpath);}
	elsif (-f 'C:/WINDOWS/system32/shimgvw.dll') {system(1, 'rundll32.exe','C:/WINDOWS/system32/shimgvw.dll,ImageView_Fullscreen', $fullpath);}
        else  {print STDOUT '!!! '.$script.': no image viewer found in the system path!'."\n";}
     }
  }

  return 0;
}


#=================================================================


# sub saveCameraSnap2($file,$cameraType,$ip,[$display,$zoom|$downsample,$compression,$colorlevel,$rotation]) {
sub saveCameraSnap2 ($$$;$$$$$) {

### This routine grabs snapshot from a given camera (using getCameraSnap)
### and then saves it in a file.
### Here:
### $cameraType = iqeye|axis|axisn (n=1,...4)
### $ip = camera IP
### $display = 0 or 1 (call or not eog/irfan)
### $zoom/$downsample applies to IQeye cameras only. Zoom must be in range
### 1 to 4 and refers to image, not cropped in the camera. In case any zoom
### is requested (even $zoom=1) the image is resized to 640x480; otherwice
### the maximum resolutiuon is applied.
### Negative "zoom" value is treated as downsample and must be -1,-2,-3,-4.

  my ($script, $file, $camera, $ip, $display, $zoom);
  my ($compression, $colorlevel, $rotation, @formatList);
  my ($status, $image, $name, $ext, $imgFormat, $fullpath, $eog);

  $script      = 'saveCameraSnap2';
  $file        = shift(@_);
  $camera      = shift(@_);		#iqeye, axis, axis1, ... axis4
  $ip          = shift(@_);		#camera IP address
  $display     = shift(@_);		#flag: 1=yes,0=no
  $zoom        = shift(@_);		#range: +1:+4; iqeye camera only
# $downsample  = shift(@_);		#range: -1:-4; iqeye camera only
  $compression = shift(@_);		#axis: can be 0, 25, 50, 75, 100
  $colorlevel  = shift(@_);		#axis: can be ??? /0 = grayscale, 100 = full color/
  $rotation    = shift(@_);		#axis: can be 0, 90, 180, 270

### External image viewer:
  if ($^O !~ /Win/i) {
    $eog    = &Which('eog');					# eye of Gnome
    if (!defined $eog) {$eog = &Which('eom');}			# eye of MATE
  } else {
    if    (-f 'C:/Progra~1/IrfanView/i_view32.exe') {$eog = 'C:/Progra~1/IrfanView/i_view32.exe';}
    elsif (-f 'C:/Progra~2/IrfanView/i_view32.exe') {$eog = 'C:/Progra~2/IrfanView/i_view32.exe';}
    elsif (-f 'C:/Progra~1/IrfanView/i_view64.exe') {$eog = 'C:/Progra~1/IrfanView/i_view64.exe';}
  }

### Supported image formats (determined from file name):
  @formatList  = ('jpg', 'jpeg', 'tif', 'tiff', 'png', 'gif', 'bmp');

### Find image format from the file name:
  ($name,$ext) = &analyzeFilename($file);
  if (!grep(/^${ext}$/i, @formatList)) {
     print STDOUT '*** '.$script.': unsupported image format ['.$ext.']'."\n";
     return -1;
  }
  $imgFormat = $ext;

  if (!defined $display) {$display = 0;}
  if ($display < 0 || $display > 1) {
     print STDOUT '!!! '.$script.': display flag ['.$display.'] must be 0 or 1. Ignoring...'."\n";
     $display = 0;
  }

  $status = &getCameraSnap2(\$image,$camera,$ip,$zoom,$compression,$colorlevel,$rotation);
  if ($status) {
     print STDOUT '*** '.$script.': call of getCameraSnap2 returned status='.$status."\n";
     return $status;
  }


### Set image compression. This can be one of:
### None, BZip, Fax, Group4, JPEG, JPEG2000, LosslessJPEG, LZW, RLE, Zip
### If lossless compression is required, then TIFF must use Zip or LZW.
### For PNG the default compression parameter is ZIP.
  if ($imgFormat =~ /^tif/i) {
#   $image->Set(compression=>'None');		        # this is default
    $image->Set(compression=>'Zip');
#   $image->Set(compression=>'LZW');
  }

  print STDOUT 'Saving image as '.$file."\n";
  $status = &ask_to_overwrite($file);			#this will ask for overwriting, unless NONINTERACTIVE=1
  if ($status) {$image->Write($file);}
  else         {return -1;}

  if ($display) {
     if ($^O !~ /Win/i) {
	if (defined $eog) {system($eog.' "'.$file.'" &');}
        else  {print STDOUT '!!! '.$script.': no image viewer found in the system path!'."\n";}
    }
     else {
        $fullpath = File::Spec->rel2abs($file);
#       $fullpath = abs_path($file);
	if    (defined $eog) {system(1, $eog, $fullpath);}
	elsif (-f 'C:/WINDOWS/system32/shimgvw.dll') {system(1, 'rundll32.exe','C:/WINDOWS/system32/shimgvw.dll,ImageView_Fullscreen', $fullpath);}
        else  {print STDOUT '!!! '.$script.': no image viewer found in the system path!'."\n";}
     }
  }

  return 0;
}


#=================================================================


# sub imageCrossectionH(\$image,$hsectionType,\@hsection,
#                              [$gnuplot,$destroy]) {
sub imageCrossectionH (\$$\@$$) {

### This routine analyzes ImageMagick image: it draws horizontal crossection
### of the image and optionally plots it with GnuPlot.
### Here:
### \$image = ImageMagick image (ATTENTION: this sub converts it into grayscale)
### $hsectionType = nnn|first|last|center|integral|none|undef  (undef=none)
### \@hsection, is reference to the array in which data is returned
### $gnuplot = 0|1 is flag to regust GnuPlot
### $destroy = 0|1 is flag to destroy image object and free memory

  my ($script, $imageRef, $hsectionType, $hsectionRef);
  my ($gnuplot, $destroy, $status);

  $script       = 'imageCrossectionH';
  $imageRef     = shift(@_);
  $hsectionType = shift(@_);		#nnn|first|last|center|integral|none|undef  (undef=none)
  $hsectionRef  = shift(@_);
  $gnuplot      = shift(@_);		#0|1|undef
  $destroy      = shift(@_);		#0|1|undef

  $status = &imageCrossection($imageRef, $hsectionType, $hsectionRef,
                                            undef,         undef,
                                            $gnuplot,      $destroy);
  return $status;
}


#=================================================================


# sub imageCrossectionV(\$image,$vsectionType,\@vsection,
#                              [$gnuplot,$destroy]) {
sub imageCrossectionV (\$$\@$$) {

### This routine analyzes ImageMagick image: it draws vertical crossection
### of the image and optionally plots it with GnuPlot.
### Here:
### \$image = ImageMagick image (ATTENTION: this sub converts it into grayscale)
### $vsectionType = nnn|first|last|center|integral|none|undef  (undef=none)
### \@vsection, is reference to the array in which data is returned
### $gnuplot = 0|1 is flag to regust GnuPlot
### $destroy = 0|1 is flag to destroy image object and free memory

  my ($script, $imageRef, $vsectionType, $vsectionRef);
  my ($gnuplot, $destroy, $status);

  $script       = 'imageCrossectionH';
  $imageRef     = shift(@_);
  $vsectionType = shift(@_);		#nnn|first|last|center|integral|none|undef  (undef=none)
  $vsectionRef  = shift(@_);
  $gnuplot      = shift(@_);		#0|1|undef
  $destroy      = shift(@_);		#0|1|undef

  $status = &imageCrossection($imageRef,    undef,         undef,
                                         $vsectionType, $vsectionRef,
                                           $gnuplot,      $destroy);
  return $status;
}


#=================================================================


# sub imageCrossection(\$image,$hsectionType,\@hsection,
#                              $vsectionType,\@vsection,
#                              [$gnuplot,$destroy]) {
sub imageCrossection (\$$\@$\@;$$) {

### This routine analyzes ImageMagick image: it draws horizontal and/or
### vertical crossection of the image and optionally plots it with GnuPlot.
### Here:
### \$image = ImageMagick image (ATTENTION: this sub converts it into grayscale)
### $hsectionType = nnn|first|last|center|integral|none|undef  (undef=none)
### $vsectionType = nnn|first|last|center|integral|none|undef  (undef=none)
### \@hsection, \@vsection is reference to the arrays in which data is returned
### $gnuplot = 0|1 is flag to regust GnuPlot
### $destroy = 0|1 is flag to destroy image object and free memory

  our $GNUPLOT_VERSION;
  my ($script, $imageRef, $destroy, $status);
  my ($hsectionType, $hsectionRef, $vsectionType, $vsectionRef);
  my ($gnuplot, $gnuopen, $gnuexe, $printPRG, @gnuargs);
  my ($pid, $milliseconds, $sleep_time);
  my ($width, $height, $maxlightH, $maxlightV, $light);
  my (@img_pixels, $i, $j, $printed);

  $script       = 'imageCrossection';
  $imageRef     = shift(@_);
  $hsectionType = shift(@_);		#nnn|first|last|center|integral|none|undef  (undef=none)
  $hsectionRef  = shift(@_);
  $vsectionType = shift(@_);		#nnn|first|last|center|integral|none|undef  (undef=none)
  $vsectionRef  = shift(@_);
  $gnuplot      = shift(@_);		#0|1|undef
  $destroy      = shift(@_);		#0|1|undef

  @gnuargs = ('-persist', '-raise');
  $gnuopen = 0;
  $milliseconds = 1;

  if (!defined $hsectionType) {$hsectionType = 'none';}
  if (!defined $vsectionType) {$vsectionType = 'none';}
  if (!defined $gnuplot)      {$gnuplot      = 0;}
  if (!defined $destroy)      {$destroy      = 0;}

  if ( ($hsectionType !~ /^[0-9]{1,3}$/) &&
       ($hsectionType !~ /^first$/i)     &&
       ($hsectionType !~ /^center$/i)    &&
       ($hsectionType !~ /^last$/i)      &&
       ($hsectionType !~ /^integral$/i)  &&
       ($hsectionType !~ /^none$/i) ) {
     print STDOUY '*** '.$script.': incorrect H-section type specification: ['.$hsectionType.']'."\n";
     return -1;
  }

  if ( ($vsectionType !~ /^[0-9]{1,3}$/) &&
       ($vsectionType !~ /^first$/i)     &&
       ($vsectionType !~ /^center$/i)    &&
       ($vsectionType !~ /^last$/i)      &&
       ($vsectionType !~ /^integral$/i)  &&
       ($vsectionType !~ /^none$/i) ) {
     print STDOUY '*** '.$script.': incorrect V-section type specification: ['.$vsectionType.']'."\n";
     return -1;
  }

### Get image size:
  $width  = $$imageRef->Get('columns');
  $height = $$imageRef->Get('rows');
  print STDOUT 'Image:  width='.$width.'  height='.$height."\n";

### Convert the image to grayscale:
  print STDOUT '!!! '.$script.': converting image to grayscale'."\n";
  $$imageRef->Quantize(colorspace=>'gray');

  if ($hsectionType !~ /^none$/i) {                          #crossection requested
     if    ($hsectionType =~ /^first$/i)    {$hsectionType = 0;}
     elsif ($hsectionType =~ /^center$/i)   {$hsectionType = int(0.5*$height+0.5);}
     elsif ($hsectionType =~ /^last$/i)     {$hsectionType = $height-1;}
     elsif ($hsectionType =~ /^integral$/i) {print STDOUT '-> Integrated horizontal section selected'."\n";}
     else {
        if ($hsectionType < 0 || $hsectionType > $height-1) {
           printf STDOUT '*** '.$script.': Horizontal section pixel='.$hsectionType.' not in range 0-%d'."\n",$height-1;
           if    ($hsectionType < 0)         {$hsectionType=0;}
           elsif ($hsectionType > $height-1) {$hsectionType=$height-1;}
           print STDOUT '...Using pixel='.$hsectionType."\n";
        }
     }
     if ($hsectionType =~ /^integral$/i) {			#integral crossection
        for ($i=0; $i<$height; $i++) {
           @img_pixels = $$imageRef->GetPixels(width=>$width, height=>1, map=>'R', x=>0, y=>$i);
           if ($i == 0) {@$hsectionRef  = @img_pixels;}
           else         {for ($j=0; $j<$width; $j++) {$$hsectionRef[$j] += $img_pixels[$j];}}
        }
     } else {							#differential crossection
### On ImageMagick image pixels are counted from left top corner,
### while we need to count from left bottom corner:
        $hsectionType = $height-1 - $hsectionType;
        @$hsectionRef = $$imageRef->GetPixels(width=>$width, height=>1, map=>'R', x=>0, y=>$hsectionType);
     }
     if ($gnuplot) {
        print STDOUT '-> Plotting section=horizontal, length='.$width.', taken at y='.$hsectionType."\n";
        $maxlightH = 0;
        for ($j=0; $j<$width; $j++) {
           $light = $$hsectionRef[$j];
           if ($light > $maxlightH) {$maxlightH = $light;}
        }
        $maxlightH = int($maxlightH+1.5);
### Plot data using GnuPlot:
        $gnuopen = &open_gnuplot(0,$width,undef,\@gnuargs);
        if ($gnuopen) {
           if ($vsectionType !~ /^none$/i) {
              print GNUPLOT 'set multiplot'."\n";
              print GNUPLOT 'set origin 0.0, 0.0'."\n";
              print GNUPLOT 'set size   0.5, 1.0'."\n";
           }
           print  GNUPLOT 'unset grid'."\n";
           print  GNUPLOT 'unset ylabel'."\n";
           print  GNUPLOT 'unset ytics'."\n";
           print  GNUPLOT 'set xlabel "Horizontal pixels"'."\n";
           printf GNUPLOT 'set yrange [0:%d]'."\n",$maxlightH;
### This plots data without a data file (see the gnuplot 'help special-filenames' command):
### This is similar to << in unix shell script.  The data are entered
### as though they are being read from a file, one data point per record.
### The letter "e" at the start of the first column terminates data entry.
           print  GNUPLOT 'plot "-" title "Horizontal Section" with lines'."\n";
           for ($j=0; $j<$width; $j++) {
              $light = int($$hsectionRef[$j]);
              print GNUPLOT $j.'  '.$light."\n";
	      if ($j == 10*int($j/10)) {&msleep_gnuplot($milliseconds);}
           }
           print GNUPLOT 'e'."\n";			#end of data
        }
     }
  }

  if ($vsectionType !~ /^none$/i) {
     if    ($vsectionType =~ /^first$/i)    {$vsectionType = 0;}
     elsif ($vsectionType =~ /^center$/i)   {$vsectionType = int(0.5*$width+0.5);}
     elsif ($vsectionType =~ /^last$/i)     {$vsectionType = $width-1;}
     elsif ($vsectionType =~ /^integral$/i) {print STDOUT '-> Integrated vertical section selected'."\n";}
     else {
        if ($vsectionType < 0 || $vsectionType > $width-1) {
           printf STDOUT '*** '.$script.': Vertical section pixel='.$vsectionType.' not in range 0-%d'."\n",$width-1;
           if    ($vsectionType < 0)        {$vsectionType=0;}
           elsif ($vsectionType > $width-1) {$vsectionType=$width-1;}
           print STDOUT '...Using pixel='.$vsectionType."\n";
        }
     }
     if ($vsectionType =~ /^integral$/i) {			#integral crossection
        for ($i=0; $i<$width; $i++) {
           @img_pixels = $$imageRef->GetPixels(width=>1, height=>$height, map=>"R", x=>$i, y=>0);
           if ($i == 0) {@$vsectionRef  = @img_pixels;}
           else         {for ($j=0; $j<$height; $j++) {$$vsectionRef[$j] += $img_pixels[$j];}}
        }
     } else {							#differential crossection
        @$vsectionRef = $$imageRef->GetPixels(width=>1, height=>$height, map=>"R", x=>$vsectionType, y=>0);
     }
     if ($gnuplot) {
        print STDOUT '-> Plotting section=vertical, length='.$height.', taken at x='.$vsectionType."\n";
        $maxlightV = 0;
        for ($j=0; $j<=$height-1; $j++) {
           $light = $$vsectionRef[$j];
           if ($light > $maxlightV) {$maxlightV = $light;}
        }
        $maxlightV = int($maxlightV+1.5);
### Plot data using GnuPlot:
        if ($hsectionType !~ /^none$/i && $gnuopen) {
           print GNUPLOT 'set origin 0.5, 0.0'."\n";
           print GNUPLOT 'set size   0.5, 1.0'."\n";
           print GNUPLOT 'set xrange [0:'.$height.']'."\n";
        } else {
           $gnuopen = &open_gnuplot(0,$height,undef,\@gnuargs);
        }
        if ($gnuopen) {
           print  GNUPLOT 'unset grid'."\n";
           print  GNUPLOT 'unset ylabel'."\n";
           print  GNUPLOT 'unset ytics'."\n";
           print  GNUPLOT 'set xlabel "Vertical pixels"'."\n";
           printf GNUPLOT 'set yrange [0:%d]'."\n",$maxlightV;
### This plots data without a data file (see the gnuplot 'help special-filenames' command):
### This is similar to << in unix shell script.  The data are entered
### as though they are being read from a file, one data point per record.
### The letter "e" at the start of the first column terminates data entry.
           print  GNUPLOT 'plot "-" title "Vertical Section" with lines'."\n";
           for ($j=0; $j<=$height-1; $j++) {
### On ImageMagick image, the pixels are counted from left top
### corner, while we need to count from left bottom corner:
              $light = int($$vsectionRef[$height-1-$j]);
              print  GNUPLOT $j.'  '.$light."\n";
 	      if ($j == 10*int($j/10)) {&msleep_gnuplot($milliseconds);}
           }
           print GNUPLOT 'e'."\n";			#end of data
        }
     }
  }

  if ($destroy) {
     $$imageRef = ();          # delete all the images but retain the Image::Magick object
     undef $imageRef;          # destroy Image::Magick object
  }

###-------------------------------------------
### Send to printer:
  if ($gnuopen) {
#    &print_gnuplot();	      # this does not work with multiplot and the '-' file
     $printed = 0;
     $sleep_time = 2;
     $gnuexe = &gnuplot_executable();
     if (!defined $gnuexe) {return 0;}
     if ($^O !~ /Win/i) {$printPRG = 'lpr';    }    # UNIX
     else               {$printPRG = 'gsprint';}    # Windows

     print STDOUT 'Print plot [n]?_';
     if ($^O !~ /Win/i) {$_ = uc <STDIN>; chomp $_; print STDOUT "\n";}
     else               {$_ = getc STDIN;           print STDOUT "$_\n";}

     &close_gnuplot();	     			#close gnuplot window
     if ($_ =~ /^y$/i) {
        $printed = 1;
        print STDOUT 'Creating temp file "gnu.ps"'."\n";
        $pid = open(GNUPLOT, '|-', $gnuexe);
        if (defined $pid) {
           select GNUPLOT; $|=1;			# set unbuffered output
           print GNUPLOT 'set terminal postscript'."\n";
           print GNUPLOT 'set output "gnu.ps"'."\n";
           print GNUPLOT 'set tics out'."\n";
           print GNUPLOT 'set mxtics 10'."\n";
           print GNUPLOT 'set mytics 10'."\n";
           print GNUPLOT 'unset grid'."\n";
           print GNUPLOT 'unset ylabel'."\n";
           print GNUPLOT 'unset ytics'."\n";
           if ($GNUPLOT_VERSION >= 4) {print GNUPLOT 'set style data linespoints'."\n";}
           else                       {print GNUPLOT 'set data style linespoints'."\n";}
           if ($hsectionType !~ /^none$/i) {
              if ($vsectionType !~ /^none$/i) {
                 print GNUPLOT 'set multiplot'."\n";
                 print GNUPLOT 'set origin 0.0, 0.0'."\n";
                 print GNUPLOT 'set size   0.5, 1.0'."\n";
              }
              printf GNUPLOT 'set xrange [0:%d]'."\n",$width;
              printf GNUPLOT 'set yrange [0:%d]'."\n",$maxlightH;
              print  GNUPLOT 'set xlabel "Horizontal pixels"'."\n";
### This plots data without a data file (see the gnuplot 'help special-filenames' command):
### This is similar to << in unix shell script.  The data are entered
### as though they are being read from a file, one data point per record.
### The letter "e" at the start of the first column terminates data entry.
              print  GNUPLOT 'plot "-" title "Horizontal Section" with lines'."\n";
              for ($j=0; $j<$width; $j++) {
                 $light = int($$hsectionRef[$j]);
                 print GNUPLOT $j.'  '.$light."\n";
 	         if ($j == 10*int($j/10)) {&msleep_gnuplot($milliseconds);}
              }
              print GNUPLOT 'e'."\n";			#end of data
           }
           if ($vsectionType !~ /^none$/i) {
              if ($hsectionType !~ /^none$/i && $gnuopen) {
                 print GNUPLOT 'set origin 0.5, 0.0'."\n";
                 print GNUPLOT 'set size   0.5, 1.0'."\n";
              }
              printf GNUPLOT 'set xrange [0:%d]'."\n",$height;
              printf GNUPLOT 'set yrange [0:%d]'."\n",$maxlightV;
              print  GNUPLOT 'set xlabel "Vertical pixels"'."\n";
### This plots data without a data file (see the gnuplot 'help special-filenames' command):
### This is similar to << in unix shell script.  The data are entered
### as though they are being read from a file, one data point per record.
### The letter "e" at the start of the first column terminates data entry.
              print  GNUPLOT 'plot "-" title "Vertical Section" with lines'."\n";
              for ($j=0; $j<=$height-1; $j++) {
### On ImageMagick image, the pixels are counted from left top
### corner, while we need to count from left bottom corner:
                 $light = int($$vsectionRef[$height-1-$j]);
                 print  GNUPLOT $j.'  '.$light."\n";
 	         if ($j == 10*int($j/10)) {&msleep_gnuplot($milliseconds);}
              }
              print GNUPLOT 'e'."\n";			#end of data
	   }
           select (undef, undef, undef, $sleep_time);	# sleep($sleep_time);
           &close_gnuplot();	     #close gnu.ps
           print STDOUT 'Sending plot to printer with the "'.$printPRG.'" command...'."\n";
           $status = `${printPRG} gnu.ps`;
           if ($status eq '1' || $status eq '-1') {
	      print STDOUT 'Plot printing failed with status = '.$status."\n";
           }
        } else {
           print STDOUT  '*** '.$script.': cannot open pipe to '.$gnuexe.": $!";
        }
     }
  }
  return 0;
}


#=================================================================


# sub msleep_gnuplot([$milliseconds]) {
sub msleep_gnuplot (;$) {
  my $milliseconds = shift(@_);
  if (!defined $milliseconds) {$milliseconds=1;}
### Use one of the two:
  if (exists &usleep) {				#this is imported as "use Time::HiRes qw(usleep);"
     usleep(1000*$milliseconds);
  } else {
     select (undef, undef, undef, 0.001*$milliseconds);
  }
}


#=================================================================


#sub skipTopup([$topupInterval,$topupAllowance]) {
sub skipTopup(;$$) {

  my ($script, $topupInterval, $topupAllowance, $topupIntervalPV);
  my ($time2topupPV, $time2topup1, $time2topup2);
  my ($time2topup1_prev, $time2topup2_prev);
  my ($timeSinceTopupPV, $timeSinceTopup);
  my ($topup, $status, $check_interval);

### This routine waits +-2s (or $topupAllowance) around the APS beam topup
### to ensure that a short operation we are going to do will not interfere
### with the topup.

  $script = 'skipTopup';
  $topupInterval  = shift(@_);
  $topupAllowance = shift(@_);					#+-interval around topup which we want to avoid
  if (!defined $topupAllowance) {$topupAllowance = 2;}		#in seconds
  if ($topupAllowance < 0)      {$topupAllowance =-$topupAllowance;}

  $time2topupPV     = 'Mt:TopUpTime2Inject';
  $topupIntervalPV  = 'S:TopUpInterval';
  $timeSinceTopupPV = 'XFD:TimeSinceTopUp';
  $check_interval   = 0.25;			# This should exceed scan rate of at least one of $time2topupPV or $timeSinceTopupPV

  if (!defined $topupInterval)  {
     ($status, $topupInterval) = Pezca::GetDouble($topupIntervalPV);
     if ($status) {
        &report_only($script, $status, $topupIntervalPV);
        $topupInterval = 120;					#in seconds
        print STDOUT '!!! '.$script.': presuming topup interval is '.$topupInterval.'s.'."\n";
     }
  }
  elsif ($topupInterval < 0)  {
     $topupInterval =-$topupInterval;
  }

  $status = &addCaMonitor($time2topupPV);
  if ($status) {
     &report_only($script, $status, $time2topupPV);
     print STDOUT '*** '.$script.': topups will be ignored'."\n";
     return 1;
  }
  $status = &addCaMonitor($timeSinceTopupPV);
  if ($status) {
     &report_only($script, $status, $timeSinceTopupPV);
     print STDOUT '*** '.$script.': topups will be ignored'."\n";
     return 1;
  }

  $time2topup1_prev = -1.;
  $time2topup2_prev = -1.;
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  do {
     ($status, $time2topup1) = Pezca::GetDouble($time2topupPV);
     if ($status) {
        &report_only($script, $status, $time2topupPV);
        print STDOUT '*** '.$script.': topups will be ignored'."\n";
        return 1;
     }
     ($status, $timeSinceTopup) = Pezca::GetDouble($timeSinceTopupPV);
     if ($status) {
        &report_only($script, $status, $timeSinceTopupPV);
        print STDOUT '*** '.$script.': topups will be ignored'."\n";
        return 1;
     }
     $time2topup2 = $topupInterval - $timeSinceTopup;

     if ($time2topup1 == $time2topup1_prev && $time2topup2 == $time2topup2_prev) {
        print STDOUT '*** '.$script.': APS topup timer is not running.'."\n";
        return 1;
     }
     elsif ($topupInterval < $time2topup1) {
        printf STDOUT '*** '.$script.': APS time to topup (%.0fs) exceeds'
                     .' topup interval (%.0fs)'."\n", $time2topup1, $topupInterval;
        return 1;
     }

     if (&minval($time2topup1,$time2topup2) < $topupAllowance || $timeSinceTopup < $topupAllowance) {
	### we are around the topup. Sleep a bit and check again.
        $topup = 1;
        printf STDOUT '  Waiting for topup (time2topup=%1.0fs, allowance='
                     .$topupAllowance.'s, interval='.$topupInterval.'s)...'."\n", $time2topup1;
        select (undef, undef, undef, $check_interval);
     } else {
	### we are far from the topup. Can exit now.
        $topup = 0;
        return 0;
     }
     $time2topup1_prev = $time2topup1_prev;
     $time2topup2_prev = $time2topup2_prev;
  } while ($topup);
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  return 0;
}


#=================================================================


# 10. $status = &getPZTcameraPresets(\@presetNames,[$url,$cameraNumber]);	#if $url is not given, takes hutch PZT camera
# sub getPZTcameraPresets(\@presetNames,[$url,$cameraNumber]) {
sub getPZTcameraPresets(\@;$$) {

# use LWP::UserAgent;		#World-Wide Web library for Perl (libwww-perl)

  our ($axisCameraURL2);			#axisCameraURL2 is hutch PZT camera
  my ($script, $presetsRef, $url, $camera);
  my ($ua, $response, $string, $data, @lines, $line, @words, $nwords);

  $script     = 'getPZTcameraPresets';
  $presetsRef = shift(@_);
  $url        = shift(@_);	if (!defined $url)    {$url    = $axisCameraURL2;}
  $camera     = shift(@_);	if (!defined $camera) {$camera = 1;}

  @$presetsRef = ();
  if (!defined $url) {
     print STDOUT '*** '.$script.': no URL specified'."\n";
     return 1;
  }

  #print STDOUT 'Requesting postion presets from URL='.$url.' camera='.$camera."\n";
  ### This is for Axis PZT cameras only:
  $string = 'http://'.$url.'/axis-cgi/com/ptz.cgi'
           .'?query=presetpositionall'
           .'&camera='.$camera
            ;
  #print STDOUT "\n".$string."\n\n";
  $ua = LWP::UserAgent->new;
  $response = $ua->get($string);

#----------------------------------------------------------------------
# HTTP status codes:
# ^^^^^^^^^^^^^^^^^
# 100 - Continue                      404 - Not Found
# 101 - Switching Protocols           405 - Method Not Allowed
# 200 - OK                            406 - Not Acceptable
# 201 - Created                       407 - Proxy Authentication Required
# 202 - Accepted                      408 - Request Timeout
# 203 - Non-Authoritative Information 409 - Conflict
# 204 - No Content                    410 - Gone
# 205 - Reset Content                 411 - Length Required
# 206 - Partial Content               412 - Precondition Failed
# 300 - Multiple Choices              413 - Request Entity Too Large
# 301 - Moved Permanently             414 - Request URI Too Long
# 302 - Found                         415 - Unsupported Media Type
# 303 - See Other                     416 - Requested Range Not Satisfiable
# 304 - Not Modified                  417 - Expectation Failed
# 305 - Use Proxy                     500 - Internal Server Error
# 306 - No Longer Used                501 - Not Implemented
# 307 - Temporary Redirect            502 - Bad Gateway
# 400 - Bad Request                   503 - Service Unavailable
# 401 - Not Authorised                504 - Gateway Timeout
# 402 - Payment Required              505 - HTTP Version Not Supported
# 403 - Forbidden
#----------------------------------------------------------------------
  if ($response->is_success) {
     $data = $response->content;
     $data =~ s/^Preset Positions for .+[\n\r]//;
     @lines = split(/[\n\r]/,$data);
     foreach $line (@lines) {
        chomp $line;
	$line =~ s/\s+//g;
	if (length($line) > 0) {
	   #print STDOUT 'line:  '.$line."\n";
           @words = split('=',$line);
           $nwords = @words;
           if ($nwords != 2) {
              print STDOUT '!!! '.$script.': unexpected preset ['.$line.'] returned by '.$url."\n";
              return -1;
           }
	   push @$presetsRef, $words[1];
	}
     }
     return 0;
  } else {
     print STDOUT '!!! '.$script.': could not get presets from '.$url
                 .' -- HTTP request failed with status: '.$response->status_line."\n";
     return -1;
  }
}


#=================================================================


# 11. $status = &gotoPZTcameraPreset($preset,[$url,$cameraNumber]);		#if $url is not given, takes hutch PZT camera
# sub gotoPZTcameraPreset($preset,[$url,$cameraNumber]) {
sub gotoPZTcameraPreset($;$$) {

# use LWP::UserAgent;		#World-Wide Web library for Perl (libwww-perl)

  our ($axisCameraURL2);			#axisCameraURL2 is hutch PZT camera
  my ($script, $preset, $url, $camera, $status);
  my ($ua, $response, $string, @presets, $npresets, $i);

  $script = 'gotoPZTcameraPreset';
  $preset = shift(@_);
  $url    = shift(@_);		if (!defined $url)    {$url    = $axisCameraURL2;}
  $camera = shift(@_);		if (!defined $camera) {$camera = 1;}

  if (!defined $url) {
     print STDOUT '*** '.$script.': no URL specified'."\n";
     return 1;
  }

#### Request all possible presets:
  $status = &getPZTcameraPresets(\@presets,$url,$camera);
  if ($status || !grep(/^$preset$/i,@presets)) {
     print STDOUT '*** '.$script.': preset='.$preset
                        .' is invalid for url='.$url
                        .' camera='.$camera."\n";
     return $status;
  }

  $npresets = @presets;
  for ($i=0; $i<$npresets; $i++) {
     if ($presets[$i] =~ /^$preset$/i) {
	$preset = $presets[$i];
        #print STDOUT 'Sending URL='.$url.' camera='.$camera.' to preset='.$preset."\n";
        ### This is for Axis PZT cameras only:
        $string = 'http://'.$url.'/axis-cgi/com/ptz.cgi'
                 .'?gotoserverpresetname='.$preset
#                .'?gotoserverpresetno='.($i+1)
                 .'&camera='.$camera
                 ;
        #print STDOUT "\n".$string."\n\n";
        $ua = LWP::UserAgent->new;
        $response = $ua->get($string);
        if (! $response->is_success) {
           print STDOUT '*** '.$script.': failed to send url='.$url
                              .' camera='.$camera
                              .' to preset='.$preset."\n";
           return 1;
        } else {
	   return 0;
	}
     }
  }
  print STDOUT '*** '.$script.': to preset='.$preset
                     .' not found for url='.$url
                     .' camera='.$camera."\n";
  return -1;
}


#=================================================================


# 12. $status = &openCameraVideoStream($url,[$camera,$resolution,            	# open in firefox
#                                      $compression,$colorlevel,$clock,$date,
#                                      $quad,$text,$rotation,$req_fps]);
sub openCameraVideoStream($;$$$$$$$$$$) {

  my ($script, $url, $camera, $resolution, $compression, $colorlevel);
  my ($clock, $date, $quad, $text, $rotation, $fps);
  my ($firefox, $str, $status);

  $script = 'openCameraVideoStream';
  $url         = shift(@_);
  $camera      = shift(@_);     if (!defined $camera) {$camera=1;}
  $resolution  = shift(@_);
  $compression = shift(@_);
  $colorlevel  = shift(@_);
  $clock       = shift(@_);
  $date        = shift(@_);
  $quad        = shift(@_);
  $text        = shift(@_);
  $rotation    = shift(@_);
  $fps         = shift(@_);

  if ($^O !~ /^Win/) {$firefox=&Which('firefox');}		#Linux
  else               {$firefox=&Which('firefox.exe');}		#Windows

  if (!defined $firefox) {
     print STDOUT '*** '.$script.': cannot Firefox.'."\n";
     return 1;
  }
  if (!defined $url) {
     print STDOUT '*** '.$script.': URL is undefined.'."\n";
     return 1;
  }
  if (! grep(/^${camera}$/,(1,2,3,4))) {
     print STDOUT '*** '.$script.': incorrect camera='.$camera."\n";
     return 1;
  }

  $str = 'http://'.$url.'/axis-cgi/mjpg/video.cgi?camera='.$camera;

  if (defined $resolution) {
     if (! grep(/^${resolution}$/,('704x480','352x240','176x112'))) {
        print STDOUT '*** '.$script.': incorrect resolution='.$resolution."\n";
        return 1;
     } else {
	$str .='&resolution='.$resolution;
     }
  }
  if (defined $compression) {
     if (! grep(/^${compression}$/,(0,25,50,75,100))) {
        print STDOUT '*** '.$script.': incorrect compression='.$compression."\n";
        return 1;
     } else {
	$str .='&compression='.$compression;
     }
  }
  if (defined $colorlevel) {
     if (! &is_int($colorlevel) || $colorlevel < 0 || $colorlevel > 100) {
        print STDOUT '*** '.$script.': incorrect colorlevel='.$colorlevel."\n";
        return 1;
     } else {
	$str .='&colorlevel='.$colorlevel;
     }
  }
  if (defined $clock) {
     if (! grep(/^${clock}$/,(0,1))) {
        print STDOUT '*** '.$script.': incorrect clock='.$clock."\n";
        return 1;
     } else {
	$str .='&clock='.$clock;
     }
  }
  if (defined $date) {
     if (! grep(/^${date}$/,(0,1))) {
        print STDOUT '*** '.$script.': incorrect date='.$date."\n";
        return 1;
     } else {
	$str .='&date='.$date;
     }
  }
  if (defined $quad) {
     if (! grep(/^${quad}$/,(0,1))) {
        print STDOUT '*** '.$script.': incorrect quad='.$quad."\n";
        return 1;
     } else {
	$str .='&quad='.$quad;
     }
  }
  if (defined $text) {
     if (! grep(/^${text}$/,(0,1))) {
        print STDOUT '*** '.$script.': incorrect text='.$text."\n";
        return 1;
     } else {
	$str .='&text='.$text;
     }
  }
  if (defined $rotation) {
     if (! grep(/^${rotation}$/,(0,1))) {
        print STDOUT '*** '.$script.': incorrect rotation='.$rotation."\n";
        return 1;
     } else {
	$str .='&rotation='.$rotation;
     }
  }
  if (defined $fps) {
     if (! &is_int($fps) || $fps < 1 || $fps > 30) {
        print STDOUT '*** '.$script.': incorrect fps='.$fps."\n";
        return 1;
     } else {
	$str .='&fps='.$fps;
     }
  }

  ### This is for Axis cameras/servers only:
  if ($^O !~ /^Win/) {$status = system($firefox.' -new-window '.$str.' &');}	#Linux
  else               {$status = system(1,$firefox,'-new-window',$str);}		#Windows
  return $status;
}


#=================================================================


# 13. $status = &openCameraHomepage($url);					#open in firefox
sub openCameraHomepage($) {

  my ($script, $url, $firefox, $status);

  $script = 'openCameraHomepage';
  $url         = shift(@_);

  if ($^O !~ /^Win/) {$firefox=&Which('firefox');}		#Linux
  else               {$firefox=&Which('firefox.exe');}		#Windows

  if (!defined $firefox) {
     print STDOUT '*** '.$script.': cannot find Firefox.'."\n";
     return 1;
  }
  if (!defined $url) {
     print STDOUT '*** '.$script.': URL is undefined.'."\n";
     return 1;
  }

  if ($^O !~ /^Win/) {$status = system($firefox.' -new-window '.'http://'.$url.' &');}	#Linux
  else               {$status = system(1,$firefox,'-new-window','http://'.$url);}	#Windows
  return $status;
}
