#!/usr/bin/perl -W
#--------------------------------------------------------------
# This example PERL script:
# - calls the X0h program from https://x-server.gmca.aps.anl.gov/
# - gets the scattering factors chi_0r and chi_0i
# - saves them into a file as a function of energy
#
# The script can use either the LWP PERL module or wget (external program).
# In the later case it can run on Microsoft Windows under Cygwin
# (https://www.cygwin.com).
#
#     		Author: Sergey Stepanov
#
# Version-1.0:  2005/12/31  getX0h_v1.pl
# Version-1.1:  2006/11/08  getX0h_v1.pl
# Version-2.0:  2012/06/28  getX0h_v2.pl
# Version-3.0:  2021/08/16  FWHM_by_E.pl - Explore FWHM*E for Silicon 111 and 333
#--------------------------------------------------------------
  use strict;
  use LWP::Simple;		# World-Wide Web library for Perl (libwww-perl)
  use LWP::Protocol::https;

  select STDOUT; $|=1;		# set unbuffered output

  my ($url, %FORM, $request, $file, @E, $nE, $dE);
  my ($key, $buffer, @content, $i, $string);
  my ($fwhm_x_E, $keyword, $found);
#--------------Input parameters---------------------------
### Output file:
  $file = 'FWHM_by_E.dat';

### Energy range and the number of energy points
  $fwhm_x_E = 400;		# 400/E is roughly the rocking curve FWHM in microrad at a given energy for Si-111
  $fwhm_x_E = 80;		# 80/E is roughly the rocking curve FWHM in microrad at a given energy for Si-333
  @E  = (4, 20); 		# start & end energy
  $nE = 51;			# number of energy points
# $keyword = 'FWHMurad_sg=';	# for 
  $keyword = 'FWHM2urad_sg=';	
### The paramters can also be passed
### as command line arguments:
# @E  = ($ARGV[0], $ARGV[1]);	# start & end energy
# $nE = $ARGV[2];		# number of energy ppints

### Energy step:
  if ($nE > 1) {$dE = ($E[1]-$E[0])/($nE-1);} else {$dE = 0;}

  $url = 'https://x-server.gmca.aps.anl.gov/cgi/x0h_form.exe';
###--------------Parameters of X0h CGI interface--------------
  $FORM{'xway'} = 2;		# 1=wavelength, 2=energy, 3=characteristic line
# $FORM{'wave'} = 0;   		# works with xway=2 or xway=3 (see below)
# $FORM{'line'} = 'Cu-Ka1';	# works with xway=3 only
  $FORM{'line'} = '';		# works with xway=3 only

### Target:
  $FORM{'coway'} = 0;		# 0=crystal, 1=other material, 2=chemicalformula
  $FORM{'code'} = 'Silicon';	# specify crystal (works with coway=0 only)
  $FORM{'amor'} = '';		# specify other material (coway=1 only)
  $FORM{'chem'} = '';		# specify chemical formula (coway=2 only)
  $FORM{'rho'}  = '';		# specify density in g/cm3 (coway=2 only)

### Bragg reflection:
# ($FORM{'i1'},$FORM{'i2'},$FORM{'i3'}) = (1, 1, 1);
  ($FORM{'i1'},$FORM{'i2'},$FORM{'i3'}) = (3, 3, 3);

### Database Options for dispersion corrections df1, df2:
### -1 - Automatically choose DB for f',f"
###  0 - Use X0h data (5-25 keV or 0.5-2.5 A) -- recommended for Bragg diffraction.
###  2 - Use Henke data (0.01-30 keV or 0.41-1240 A) -- recommended for soft x-rays.
###  4 - Use Brennan-Cowan data (0.03-700 keV or 0.02-413 A)
###  6 - Use Windt data (0.01-100 KeV or 0.12-1240 A)
###  8 - Use Chantler/NIST data (0.01-450 KeV or 0.28-1240 A)
### 10 - Compare results for all of the above sources.
  $FORM{'df1df2'} = -1;

### Output options:
  $FORM{'modeout'} = 1;		# 0=html, 1=quasytext
  $FORM{'detail'}  = 0;		# 0=no coords, 1=print coords
###-----------------------------------------------------------

### Open output file for writing:
  open (DAT,' > ',$file) || die 'Cannot open '.$file."\n";	# overwrite DAT file
  select DAT; $|=1;						# set unbuffered output

### Print data header:
  print DAT '#Energy,      FWHM,        400/E'."\n";

  for ($i=0; $i<$nE; $i++) { 			# Energy loop
     $FORM{'wave'} = $E[0] + $dE * $i;
     $request = '';
     foreach $key (keys %FORM) {$request .= '&'.$key.'='.$FORM{$key};}
     $request =~ s/^\&/\?/;			# replace 1st "&" by "?"

     $buffer = get($url.$request);		# request X0h data from server:
     @content = split /[\r\n]/, $buffer;	# split response into lines

     undef $found;
     foreach $string (@content) {		# loop over page lines
        if ($string =~ m/${keyword}/) {		# if the line contains "${keyword}"
           $string =~ s/^.*${keyword}//g;	# erase everything before the data
           $string =~ s/\s*$//;			# erase all spaces after data
           $found = $string;
           last; 
        }
     }
     if (defined $found) {
#       printf DAT '%10.3f,  %s,  %10.3f'."\n", $FORM{'wave'}, $found, $fwhm_x_E/$FORM{'wave'};
        printf DAT '%10.3f,  %10.3f'."\n", $FORM{'wave'}, $found*$FORM{'wave'};
     }
  }
