#!/usr/bin/perl -W
#--------------------------------------------------------------
#
# This script implements SONICC scan, then reads WWW-served SONIC images
# (before v.3.1 these were TXT-formatted SONICC images, which it was
#  converting into PNG), displays them on the screen and sends mouse
#  clicks on the image to the GM/CA point&clik server.
#
#                       Author: Sergey Stepanov
#
# Version-1.0:  2012/10/28	initial implementation using PDL::IO::Pic and NetPBM
# Version-1.1:  2012/10/31	added abort and standby buttons, added image rotate flag
# Version-1.2:  2012/10/31	added second image
# Version-1.3:  2012/11/12	added image resampling (magnification),
#                               added reading/reloading calibration from user's file,
#                               added showing image intensity,
#                               added re-click (1st click after scan requests crystal
#                               shift while repeated clicks first undo prev. shift)
# Version-1.4:  2012/11/14	added logfile (one per directory)
# Version-1.5:  2013/02/25	added outer scan over crystal positions along the beam
# Version-1.6:  2013/03/07	added Sonic IN/OUT buttons
# Version-1.7:  2013/03/12	added focus offset
# Version-1.8:  2013/06/12	changed to work with sample environment server
# Version-1.9:  2014/03/28      changed functions calls to arrays pointers (preparation to Perl-5.16)
# Version-2.0:  2015/02/24	added automatic determining if image scaling needs be applied
# Version-2.1:  2015/03/03	changed to three modalities of SONICC scan
# Version-2.2:  2015/03/04	added strick syntax checking of sonicc.conf
# Version-2.3:  2015/07/28	made sonicc.conf mandatory and moved it to /mnt/software/scripts (should be world-writable)
# Version-3.0   2016/03/29      replaced 2 fixed diplay windows with a loop over windows;
#                               updated GUI (added MEDM buttons, added colors);
#                               added continuous and single-shot modes;
#                               added reading JPEGs from WWW instead of TXT-file exchange
# Version-3.1   2016/07/29	implemented reading continouos image stream with point-and-click during scan
# Version-3.2   2017/03/27	disabled continuous scanning button (see $enable_continuous)
# Version-3.3   2017/04/12	added option to turn off sample environment lights when bringing SONICC;
#                               added drawing scale box and OAV center on the images
# Version-3.4   2019/03/15	moved $focusOffsetSHG & $focusOffsetTPE into respective @parsSHG & @parsTPE arrays;
#                               switched to loading @parsSHG & @parsTPE arrays from EPICS instead of config file
#--------------------------------------------------------------
# Subs: &main
#       &monitorPVs();
#       &process_click($widget,$xpixel,$ypixel);			#widget is transferred automatically
#       &display_coords($widget,$xpixel,$ypixel,$imindx);		#widget is transferred automatically
#	&drawScaleBox(\$image,$mode,$width,$height);
#       &updateLogBox($msg,[$tag]);
#       &retrieve_gonio_positions(\@positions,[$log_flag]);
#       &soniccMove($dest);
#       &calculate_soniccPos_tag();
#       &soniccWaitForIN();
#       &switch_sonicc_RL('Remote'|'Local');
#       &send_SONICC(cmd,[$debug]);
#       &query_SONICC($cmd,[$debug]);
#       &request_SONICC_to_load_scan_config()
#       &update_calibration_label($soniccMode);
#       &report_CAerr($status,$pv);
#       &open_sonicc_medm();
#       &open_calibr_medm();
#       &disable_buttons();
#       &enable_buttons();
#       &set_wrk_dir();
#       &scan_init();
#       &start_zscan('single'|'zscan');
#       &command_SONICC_take_image();
#       &command_SONICC_start_continuous();
#       &abort_scan();
#       &sonicc_standby();
#       &get_sonicc_states();
#       &toggle_step_entry();
#       &load_soniccConf('new'|'reload');
#       &prev_image();
#       &next_image();
#       &replace_images();
#       &toggle_enhance();
#       &read_www_images($imstep,$label);
#	&display_image($imstep,$imindx);
#       &make_directory($directory,\$msg);
#       &save_images4user();
#       &goodbye();
#--------------------------------------------------------------
  use LWP::UserAgent;           						#World-Wide Web library for Perl
  use Tk;
  use Tk::BrowseEntry;
  use Tk::widgets qw/JPEG PNG ROText/;
  use File::Path qw/mkpath/;
  use File::Copy;
  use Image::Magick;            						#ImageMagick API for Perl
  use MIME::Base64;								#encode_base64() function
  use Pezca;
# use warnings::unused;
  use strict;

  our ($NONINTERACTIVE, $USE_ZENITY_FOR_EXIT, $IGNORE_ERRORS, $beamline, $HAVE_XYZ, $nRetry);
  our ($BYE_IN_PROGRESS, $SUB_IN_PROGRESS, $SCAN_IN_PROGRESS, $IMAGE_LOADED);
  our (@parsSHG, @parsTPE, $soniccConf, $configFullPath);
  our ($mw, $repeater, $fBtns, $fLog, $fImage, $fOpt, $btnWidth);
  our ($fScans, $onceBtn, $contBtn, $zscanBtn, $dirLbl, $dirBtn);
  our ($fTools, $getstateBtn, $standbyBtn);
# our ($commitBtn);
  our ($saveimgBtn, $rdconfBtn, $medmSoniccBtn, $medmCalibrBtn, $localBtn, $remoteBtn);
  our ($fAbort, $abortBtn);
  our ($logBox, $fScale, @sclRadio, @scnRadio, $scnLbox);
  our (@imLabelNm, @imHeader, @imDisplay, $imLabelXY, @header);
  our (@scnList, $fStep, $entryStep, $lblStep, $unitStep);
  our ($scanPts, $scanStep, $lastScan_pts, $currImage);
  our ($calibrLbl, @enhanceLabels, @scanLabels, @scanSwitchers, @channels);
  our ($soniccMode, $nModes, $nextBtn, $prevBtn);
  our ($windowsImageDir, $userHome, $userSaveDir);
  our ($iDisplay, $nDisplay);
  our ($sonicCmdPV, $sonicRspPV, $clickPV);
  our (@gonioPVs, @gonioAct, $nPVs, $xyAssyPV, @xyAxesPV);
  our ($soniccAssy_PV, $soniccAxis_PV, $sonicc_tolerance);
  our ($sonicc_in_PV, $sonicc_out_PV, $sonicc_in_pos, $sonicc_out_pos);
  our ($fInOut, $btnIN, $btnOUT, $lblInOut);
  our (@modeConfigs, $windowsConfigPath, $soniccPosition);
  our ($PEZCA, $script, $debug);
  our ($btnbkgcolor, $bpresscolor, $fborderwidth, $frelief);
  our ($greencolor, $goldcolor, $redcolor, $yellowcolor);
  our (@images, $flipv, $fliph, $rotate, $magnify, $grayscale, $enhance);
  our ($imgExt, $imgType, $useragent, $layout);
  our (@lightsPVs, @lightsVal);
  our (@shgPVs, @tpePVs, $epics_cfg);
#      	$images[$img]{mode}							#$img=0,..$lastScan_pts-1; this is SONICC mode of image
#      	$images[$img]{label}							#$img=0,..$lastScan_pts-1; this is label of image
#      	$images[$img]{x}                        				#$img=0,..$lastScan_pts-1; -+
#      	$images[$img]{y}                        				#$img=0,..$lastScan_pts-1;  + these are gonio/crystal
#      	$images[$img]{h}                        				#$img=0,..$lastScan_pts-1;  + coordinates when the image was acquired
#      	$images[$img]{omega}                    				#$img=0,..$lastScan_pts-1; -+
#      	$images[$img]{'www'.$i}    						#$img=0,..$lastScan_pts-1, $i=0,..$nDisplay-1;
#      	$images[$img]{'magic'.$i}    						#$img=0,..$lastScan_pts-1, $i=0,..$nDisplay-1;
#      	$images[$img]{'enhance'.$i}    						#$img=0,..$lastScan_pts-1, $i=0,..$nDisplay-1;
#      	$images[$img]{'pixels'.$i}($x,$y)	    				#$img=0,..$lastScan_pts-1, $i=0,..$nDisplay-1; this is matrix ncol x nlines
#      	$images[$img]{'ncols'.$i}    						#$img=0,..$lastScan_pts-1, $i=0,..$nDisplay-1; these are image sizes
#      	$images[$img]{'nlines'.$i},						#$img=0,..$lastScan_pts-1, $i=0,..$nDisplay-1; these are image sizes
#      	$images[$img]{'min'.$i},						#$img=0,..$lastScan_pts-1, $i=0,..$nDisplay-1; this is image min intensity
#      	$images[$img]{'max'.$i}    						#$img=0,..$lastScan_pts-1, $i=0,..$nDisplay-1; this is image max intensity

  my ($msg, $status, $ipaddr, $pi2, $degree2rad, @beamlinelist, @sonicclist, $arg);
  my ($line, @configlocations, $enable_continuous);
  my ($b_, $w_, $h_, $imindx, $i, $j);
  my (@parsPVs, $pv);

  if (!defined $ENV{PEZCA}) {
     die 'Please define the PEZCA environment. It must point to the PERL scripts repository';
  }
  $PEZCA=$ENV{PEZCA};
  $PEZCA =~ s/\\/\//g;
  if ($PEZCA !~ /\/$/) {$PEZCA .='/';}

# $beamline = '23b';
  do $PEZCA.'Common/motor_subs.pl';
  do $PEZCA.'Common/common_subs.pl';
  if (defined $beamline) {do $PEZCA.'/Common/beamline_'.$beamline.'.pl';}
  else                   {do $PEZCA.'Common/beamline.pl';}

  select(STDERR);  $|=1;							#Flush output after each print
  select(STDOUT);  $|=1;							#Flush output after each print

  $script = &makeScriptName();

  $NONINTERACTIVE      = 1;
  $USE_ZENITY_FOR_EXIT = 1;
  $IGNORE_ERRORS       = 1;
  $BYE_IN_PROGRESS     = 0;
  $SUB_IN_PROGRESS     = 0;
  $SCAN_IN_PROGRESS    = 0;							#1=oneShot,2=Cont.,3=zScan,4=sonicMotion
  $IMAGE_LOADED        = 0;
  $debug               = 0;							#set to 1 in order to see commands sent to SONICC
  $enable_continuous   = 0;

  $btnbkgcolor = '#0097FF';
  $bpresscolor = '#00D8FF';
  $greencolor  = '#00BC00';
  $goldcolor   = '#D0D000';
  $redcolor    = '#FF0000';
  $yellowcolor = '#E0E000';

  $nRetry  = 5;
  &setEpicsRetry();
  Pezca::AutoErrorMessageOff();

######################################################################################
  @beamlinelist  = ('23i:', '23o:', '23b:', '23d:');
  @sonicclist    = ('23o:');
  foreach $arg (@ARGV) {
     if (grep(/^$arg:?$/, @beamlinelist)) {
        $arg=~s/:$//g;
        if ($beamline ne $arg) {
           $beamline = $arg;
           print STDOUT "\n".'!!! '.$script.': Cross operation for beamline='.$beamline."\n";
           do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
	}
     } else {
        print STDOUT '!!! '.$script.': unrecornized command line argument '.$arg.'. Ignoring!'."\n";
     }
  }
  if (! grep(/^$beamline:?$/, @sonicclist)) {
     &zenity_error_exit($script.': SONICC is not present at beamline='.$beamline);
  }
######################################################################################
  $imgExt      = 'jpg';								# jpg or png
  $ipaddr       = '192.168.1.50';						# SONICC CA server IP
  if (!defined $ENV{EPICS_CA_ADDR_LIST}) {$ENV{EPICS_CA_ADDR_LIST} = $ipaddr;}
  else    {if ($ENV{EPICS_CA_ADDR_LIST} !~ /$ipaddr/i) {$ENV{EPICS_CA_ADDR_LIST} .= ' '.$ipaddr;}}

  $clickPV      = $beamline.':bi:pointclick';
  $sonicCmdPV   = $beamline.':SONICC:cmd';
  $sonicRspPV   = $beamline.':SONICC:rsp';

  $windowsImageDir   = 'X:/ArgonneData/';					#do not forget slash at the end!
  $windowsConfigPath = $windowsImageDir.'ConfigurationFiles/';			#do not forget slash at the end!
  $windowsImageDir   =~ s/\//\\/g;						#change "/" to "\"
  $windowsConfigPath =~ s/\//\\/g;						#change "/" to "\"
  if (defined $ENV{HOME}) {
     $userHome    = $ENV{HOME};
     $userSaveDir = $userHome.'/sonicc';					#directory where images are saved for user records
     if (! -e $userSaveDir) {
	$status = &make_directory($userSaveDir,\$msg);
	if ($status) {&updateLogBox($msg);}
     }
  }

  @enhanceLabels= ('As taken', 'Enhanced contrast');
  @scanLabels   = ('SHG TPEF BF', 'TP-UVF FLUO BF');
  @channels     = ([qw(A  D  B)],  [qw(C   D   B)]);				#2D array (web server channels to read images)
  $nModes       = @scanLabels;
  @scanSwitchers= qw(SONICC TPE-UVF);						#cmd to send to SONICC for switching modes
  @modeConfigs  = qw(SONICC_SDImageConfig_SHG.txt SONICC_SDImageConfig_TPEUVF.txt);
  @header       = ([split /\s+/,$scanLabels[0]], [split /\s+/,$scanLabels[1]]);	#2D array
  $nDisplay     = @{$header[0]};						#how many image displays we have on the screen ($nDisplay=3;)

  $soniccMode   = 0;								#0=SHG/TPEF/BF, 1=TPE-UVF/FLUO/BF
  $magnify      = 1;								#0.5,1,2
  $flipv        = 0;								#0=no, 1=yes
  $fliph        = 0;								#0=no, 1=yes
  $rotate       = 0;								#0,90,180,270
  $grayscale    = 0;								#0=no, 1=yes
  $enhance      = 0;								#default intensity scale flag (0=as is,1=enhanced)
  $layout       = 'Horizontal';							#Horizontal/Vertical

  $soniccConf   = 'sonicc.conf';                                                #this when $epics_cfg=0;
  @configlocations = ('/mnt/software/scripts/', $PEZCA.'Data/');                #this when $epics_cfg=0;
######################################################################################
### PVs:
  if (defined $HAVE_XYZ) {@gonioPVs = ('GO:CX', 'GO:CY', 'GO:CZ', 'GO:OM');}
  else                   {@gonioPVs = ('GO:CX', 'GO:CY', 'GO:SH', 'GO:OM');}
  $xyAssyPV = $beamline.':GO:Cr';
  @xyAxesPV = ($beamline.':'.$gonioPVs[0], $beamline.':'.$gonioPVs[1]);
  $nPVs = @gonioPVs;
  for ($i=0; $i<$nPVs; $i++) {
     $gonioPVs[$i]=$beamline.':'.$gonioPVs[$i].':ActPos';
     $status = &addCaMonitor($gonioPVs[$i]);
#    &report_only($script,$status,$gonioPVs[$i]);       			#already reported
  }
# $soniccAssy_PV        = $beamline.':JET:Ps';
# $soniccAxis_PV        = $beamline.':JET:P';
  $soniccAssy_PV        = $beamline.':SNC:Ps';
  $soniccAxis_PV        = $beamline.':SNC:P';
  $sonicc_in_PV     = $soniccAxis_PV.':RqsPosMem1';
  $sonicc_out_PV    = $soniccAxis_PV.':RqsPosMem2';
  $sonicc_tolerance = 0.2; 							#max |act-rqs| in mm
  $status = &setAssyMonitors ($soniccAssy_PV);
  $status = &setAxisMonitors ($soniccAxis_PV);
  ($status, $sonicc_in_pos)  = Pezca::GetDouble($sonicc_in_PV);
  if ($status) {&report_only($script,$status,$sonicc_in_PV);  $sonicc_in_pos=0;}
  ($status, $sonicc_out_pos) = Pezca::GetDouble($sonicc_out_PV);
  if ($status) {&report_only($script,$status,$sonicc_out_PV); $sonicc_out_pos=0;}
###            backlight ringlight HiRes  LowRes
  @lightsPVs = qw(DAC1_8  DAC1_4  DAC1_6  DAC1_5);
  for ($i=0; $i<@lightsPVs; $i++) {
     $lightsPVs[$i] = $beamline.':'.$lightsPVs[$i];
     $status = &addCaMonitor($lightsPVs[$i]);
  }
  @parsPVs = qw(omegaOffset xcenter_px ycenter_px xpixel2um ypixel2um axes_rotation focusOffset);
  @shgPVs = ();
  @tpePVs = ();
  $epics_cfg = 1;
  for ($i=0; $i<@parsPVs; $i++) {
     $pv = $beamline.':SONICC:'.$parsPVs[$i].':';
     $status = &addCaMonitor($pv.'SHG');
     if ($status) {$epics_cfg = 0; last;}
     push(@shgPVs,$pv.'SHG');
     $status = &addCaMonitor($pv.'TPE');
     if ($status) {$epics_cfg = 0; last;}
     push(@tpePVs,$pv.'TPE');
  }
### Load user-specific @parsSHG and @parsTPE from EPICS or config file (it must exist!)
  $status = &load_soniccConf('new');
  if ($status) {
     if ($epics_cfg) {
        &zenity_error_exit($script.': failed to load SONICC calibration from EPICS PVs'."\n");
     } else {
        &zenity_error_exit($script.': failed to load SONICC calibration from'."\n"
                       .'file ['.$soniccConf.'] at any of these locations:'."\n"
                       ."(@{configlocations})");
     }
  }

  @images = ();									#2D-array or hashes, [$iStep][$display]
  $imgType = uc($imgExt);
  if ($imgType =~ /JPG/) {$imgType = 'JPEG';}

  $pi2 = 2.0 * atan2(1.,1.);							# atan2(1)=pi/4; pi/2=2*atan2(1)
  $degree2rad = $pi2 / 90.0;							# (pi/2)/90.

  $fborderwidth = 0; $frelief = 'flat';						# this is normal setting
# $fborderwidth = 2; $frelief = 'groove';					# this is for debugging (comment out for normal operations)
######################################################################################
### Create a GUI:
  $mw         = MainWindow->new;

###===========================
  $fImage     = $mw->Frame(-borderwidth=>$fborderwidth,
                          -relief=>$frelief,                               )->grid(-row=>0, -column=>0, -padx=>10, -pady=>10);
  $line = 0;
  $b_ = int(40);								# base for calculating the initial size of image placeholders
  if ($b_ < 10) {$b_ = 10;}
  $calibrLbl  = $fImage->Label(-text =>'',                                 )->grid(-row=>$line++, -column=>0, -columnspan=>$nDisplay);
  if ($layout =~ /^Vertical$/i) {
### Place one image under the other
     $w_ = int(2.*$b_);
     $h_ = int(.5*$b_);
     for ($iDisplay=0; $iDisplay<$nDisplay; $iDisplay++) {
	$i = $iDisplay+1;
        $imHeader[$iDisplay]  = $fImage->Label(-text=>$header[$soniccMode][$iDisplay],
                                                                           )->grid(-row=>$line++, -column=>0, -columnspan=>$nDisplay);
        $imDisplay[$iDisplay] = $fImage->Label(-text=>'image-'.$i,
                                               -width=>$w_,
                                               -height=>$h_,
                                               -cursor=>'tcross',
                                               -borderwidth=>1,
                                               -relief=>'groove',
                                               -background=>'white',       )->grid(-row=>$line++, -column=>0, -columnspan=>$nDisplay);
        $imLabelNm[$iDisplay] = $fImage->Label(-text =>'',                 )->grid(-row=>$line++, -column=>0, -columnspan=>$nDisplay);
     }
  } else {	#Horizontal
### Place images side-by-side
     $w_ = int($b_);
     $h_ = int($b_);
     for ($iDisplay=0; $iDisplay<$nDisplay; $iDisplay++) {
	$i = $iDisplay+1;
        $imHeader[$iDisplay]  = $fImage->Label(-text=>$header[$soniccMode][$iDisplay],
                                                                           )->grid(-row=>$line,   -column=>$iDisplay);
        $imDisplay[$iDisplay] = $fImage->Label(-text=>'image-'.$i,
                                               -width=>$w_,
                                               -height=>$h_,
                                               -cursor=>'tcross',
                                               -borderwidth=>1,
                                               -relief=>'groove',
                                               -background=>'white',       )->grid(-row=>$line+1, -column=>$iDisplay);
        $imLabelNm[$iDisplay] = $fImage->Label(-text =>'',                 )->grid(-row=>$line+2, -column=>$iDisplay);
     }
     $line += 3;
  }
  $imLabelXY  = $fImage->Label(-text =>''                                  )->grid(-row=>$line++, -column=>0, -columnspan=>$nDisplay);

  $prevBtn    = $fImage->Button(-text=>'<< Prev.',
                                -background=>$btnbkgcolor,
                                -activebackground=>$btnbkgcolor,
                                -width=>12,
                                -state=>'disabled',
                                -command=>\&prev_image,                    )->grid(-row=>$line, -column=>0, -pady=>5, -padx=>5, -sticky=>'w',);
  $nextBtn    = $fImage->Button(-text=>'Next >>',
                                -background=>$btnbkgcolor,
                                -activebackground=>$btnbkgcolor,
                                -width=>12,
                                -state=>'disabled',
                                -command=>\&next_image,                    )->grid(-row=>$line, -column=>$nDisplay-1, -pady=>5, -padx=>5, -sticky=>'e',);
  if ($nDisplay < 3) {
### Add line for linear/log container
     $line++;
     $fScale  = $fImage->Frame(-borderwidth=>$fborderwidth,
                               -relief=>$frelief,                          )->grid(-row=>$line++, -column=>0, -columnspan=>$nDisplay, -padx=>5, -pady=>5);
  } else {
### Scqueeze linear/log container between '<< Prev.' and 'Next >>'
     $fScale  = $fImage->Frame(-borderwidth=>$fborderwidth,
                               -relief=>$frelief,                          )->grid(-row=>$line++, -column=>1, -columnspan=>$nDisplay-2, -padx=>5, -pady=>5);
  }
  for ($i=0; $i<2; $i++) {
     $sclRadio[$i] = $fScale->Radiobutton(-text=>$enhanceLabels[$i],
                                          -variable=>\$enhance,
                                          -value   =>$i,
		                          -state   =>'disabled',
                                          -command =>\&toggle_enhance,     )->grid(-row=>0, -column=>$i,);
  }
###===========================
  $fBtns      = $mw->Frame(-borderwidth=>$fborderwidth,
                           -relief=>$frelief,                              )->grid(-row=>0, -column=>1, -sticky=>'n',);
  $btnWidth = 24;
  $line = 0;
  $dirLbl     = $fBtns->Label(-text =>$userSaveDir,                        )->grid(-row=>$line++,);
  $dirBtn     = $fBtns->Button(-text=>'Set Save Directory',
                               -background=>$btnbkgcolor,
                               -activebackground=>$btnbkgcolor,
                               -width=>$btnWidth-1,
                               -command=>\&set_wrk_dir,                    )->grid(-row=>$line++,);
###---------------------------
  $fInOut     = $fBtns->Frame(-borderwidth=>$fborderwidth,
                              -relief=>$frelief,                           )->grid(-row=>$line++,);
  $btnIN      = $fInOut->Button(-text=>'IN',
                                -background=>$btnbkgcolor,
                                -activebackground=>$btnbkgcolor,
                                -width=>int($btnWidth/3)-5,
                                -command=>[\&soniccMove,  'IN'],           )->grid(-row=>0, -column=>0, -pady=>1, -padx=>0,);
  $soniccPosition = 'Querying';
  $lblInOut   = $fInOut->Label(-textvariable=>\$soniccPosition,
                              -background=>'#000000',
			      -foreground=>$yellowcolor,
			      -width=>int($btnWidth/3)+4,                  )->grid(-row=>0, -column=>1, -pady=>2, -padx=>0, -sticky=>'ns');
  $btnOUT     = $fInOut->Button(-text=>'OUT',
                                -background=>$btnbkgcolor,
                                -activebackground=>$btnbkgcolor,
                                -width=>int($btnWidth/3)-5,
                                -command=>[\&soniccMove, 'OUT'],           )->grid(-row=>0, -column=>2, -pady=>1, -padx=>0,);
###---------------------------
  $fOpt       = $fBtns->Frame(-borderwidth=>2,
                             -relief=>'ridge',                             )->grid(-row=>$line++, -pady=>5, -padx=>5, -sticky=>'we');
  $j = 0;       $fOpt->Label(-text =>'MODES',
                             -foreground=>'red',                           )->grid(-row=>$j++, -column=>0, -pady=>1);
  for ($i=0; $i<$nModes; $i++) {
     $scnRadio[$i] = $fOpt->Checkbutton(-text=>$scanLabels[$i],
                                        -variable=>\$soniccMode,
                                        -onvalue=>$i,
					-offvalue=>1-$i,
		                        -command =>\&request_SONICC_to_load_scan_config,
			                -indicatoron=>0,
                                        -background=>$btnbkgcolor,
                                        -selectcolor=>$bpresscolor,
                                        -activebackground=>$btnbkgcolor,
			                -width=>$btnWidth-4,               )->grid(-row=>$j++, -column=>0);
  }
###---------------------------
  $fScans     = $fBtns->Frame(-borderwidth=>2,
                              -relief=>'ridge',                            )->grid(-row=>$line++, -pady=>5, -padx=>5, -sticky=>'we');
  $j = 0;       $fScans->Label(-text =>'SCANS',
                             -foreground=>'red',                           )->grid(-row=>$j++, -column=>0,-columnspan=>2, -pady=>1);
  $onceBtn    = $fScans->Button(-text=>'Take One Shot',
                                -background=>$btnbkgcolor,
                                -activebackground=>$btnbkgcolor,
                                -width=>$btnWidth-4,
                                -command=>[\&start_zscan, 'single'],       )->grid(-row=>$j++,-column=>0,-pady=>2,-padx=>1,);
  if ($enable_continuous) {
     $contBtn = $fScans->Button(-text=>'Start Continuous',
                                -background=>$btnbkgcolor,
                                -activebackground=>$btnbkgcolor,
                                -width=>$btnWidth-4,
                                -command=>\&command_SONICC_start_continuous, )->grid(-row=>$j++,-column=>0,-pady=>2,-padx=>1,);
  }
  $scanStep      = 25;
  @scnList       = (); for ($i=1; $i<=41; $i+=2) {push @scnList, $i;}
  $scanPts       = $scnList[0];
  $lastScan_pts  = 0;
  $currImage     = 0;
  $scnLbox       = $fScans->BrowseEntry(-label=>'Sample Z-shifts ',	#see http://search.cpan.org/dist/Tk/pod/BrowseEntry.pod
                                       -choices=>\@scnList,
                                       -variable=>\$scanPts,
#                                      -state=>'readonly',
                                       -width=>4,
                                       -listwidth=>20,
                                       -listheight=>scalar @scnList,
                                       -browsecmd=>\&toggle_step_entry,    )->grid(-row=>$j++,-column=>0,-columnspan=>2,-pady=>5,-padx=>5,);
  $fStep      = $fScans->Frame(-borderwidth=>$fborderwidth,
                               -relief=>$frelief,                          )->grid(-row=>$j++,-column=>0,-columnspan=>2,-pady=>5,-padx=>5,);
  $lblStep    = $fStep->Label(-text =>'Sample Z-step ',                    )->pack(-side=>'left',-pady=>2,-padx=>2,);
  $entryStep  = $fStep->Entry(-textvariable=>\$scanStep,
			      -state => 'disabled',
                              -width=>4,
			      -justify=>'right',                           )->pack(-side=>'left',-pady=>2,-padx=>2,);
  $unitStep   = $fStep->Label(-text =>'um',                                )->pack(-side=>'left',-pady=>2,-padx=>2,);

  $zscanBtn    = $fScans->Button(-text=>' Start Z-Scan ',
                                 -background=>$btnbkgcolor,
                                 -activebackground=>$btnbkgcolor,
                                 -width=>$btnWidth-4,
                                 -command=>[\&start_zscan, 'zscan'],       )->grid(-row=>$j++,-column=>0,-pady=>2,-padx=>1,);
###---------------------------
  $fAbort     = $fBtns->Frame(-borderwidth=>0,
                              -relief=>'flat',                             )->grid(-row=>$line++,-pady=>0,-padx=>0,-sticky=>'we',);
  $j = 0;
  $abortBtn   = $fAbort->Button(-text=>'Abort / Stop',
                                -background=>$goldcolor,
                                -activebackground=>$goldcolor,
                                -width=>$btnWidth-3,
                                -command=>\&abort_scan,                    )->grid(-row=>$j++,-pady=>2*17,-padx=>5,);
###---------------------------
  $fTools     = $fBtns->Frame(-borderwidth=>2,
                              -relief=>'ridge',                            )->grid(-row=>$line++,-pady=>5,-padx=>5,-sticky=>'we',);
  $j = 0;       $fTools->Label(-text =>'TOOLS',
                               -foreground=>'red',                         )->grid(-row=>$j++,-column=>0,-columnspan=>2,-pady=>1);
  $getstateBtn = $fTools->Button(-text=>' State? ',
                                -background=>$btnbkgcolor,
                                -activebackground=>$btnbkgcolor,
                                -width=>int($btnWidth/2-4),
                                -command=>\&get_sonicc_states,             )->grid(-row=>$j, -column=>0, -pady=>1);
  $standbyBtn = $fTools->Button(-text=>' Standby ',
                                -background=>$btnbkgcolor,
                                -activebackground=>$btnbkgcolor,
                                -width=>int($btnWidth/2-4),
                                -command=>\&sonicc_standby,                )->grid(-row=>$j++, -column=>1, -pady=>1);
  $remoteBtn  = $fTools->Button(-text=>' Remote ',
                                -background=>$btnbkgcolor,
                                -activebackground=>$btnbkgcolor,
                                -width=>int($btnWidth/2-4),
                                -command=>[\&switch_sonicc_RL,'Remote'],   )->grid(-row=>$j, -column=>0, -pady=>1);
  $localBtn    = $fTools->Button(-text=>' Local ',
                                -background=>$btnbkgcolor,
                                -activebackground=>$btnbkgcolor,
                                -width=>int($btnWidth/2-4),
                                -command=>[\&switch_sonicc_RL,'Local'],    )->grid(-row=>$j++, -column=>1, -pady=>1);
# $commitBtn  = $fTools->Button(-text=>'Commit HW',
#                               -background=>$btnbkgcolor,
#                               -activebackground=>$btnbkgcolor,
#                               -width=>$btnWidth-3,
#                               -command=>[\&send_SONICC,'Commit_HW',1],   )->grid(-row=>$j++,-column=>0,-columnspan=>2,-pady=>1);
  $saveimgBtn = $fTools->Button(-text=>'Save images',
                                -background=>$btnbkgcolor,
                                -activebackground=>$btnbkgcolor,
                                -width=>$btnWidth-3,
                                -command=>\&save_images4user,              )->grid(-row=>$j++,-column=>0,-columnspan=>2,-pady=>1);
  $rdconfBtn  = $fTools->Button(-text=>'Reload Config',
                                -background=>$btnbkgcolor,
                                -activebackground=>$btnbkgcolor,
                                -width=>$btnWidth-3,
                                -command=>[\&load_soniccConf, 'reload'],   )->grid(-row=>$j++,-column=>0,-columnspan=>2,-pady=>1);
  $medmSoniccBtn= $fTools->Button(-text=>'Open SONICC MEDM',
                                -background=>$btnbkgcolor,
                                -activebackground=>$btnbkgcolor,
                                -width=>$btnWidth-3,
                                -command=>\&open_sonicc_medm,              )->grid(-row=>$j++,-column=>0,-columnspan=>2,-pady=>1);
  $medmCalibrBtn= $fTools->Button(-text=>'Open Calibr. MEDM',
                                -background=>$btnbkgcolor,
                                -activebackground=>$btnbkgcolor,
                                -width=>$btnWidth-3,
                                -command=>\&open_calibr_medm,              )->grid(-row=>$j++,-column=>0,-columnspan=>2,-pady=>1);
###===========================
  $fLog = $mw->Frame(-borderwidth=>$fborderwidth,
                     -relief=>$frelief,                                    )->grid(-row=>1, -column=>0, -columnspan=>2, -sticky=>'nwe');
# $logBox = $fLog->Scrolled('Listbox',
# $logBox = $fLog->Scrolled('Text',                                       			#see: http://docstore.mik.ua/orelly/perl3/tk/ch06_02.htm
  $logBox = $fLog->Scrolled('ROText',                                       			#see: http://docstore.mik.ua/orelly/perl3/tk/ch06_02.htm
                            -scrollbars=>'se',
                            -height=>'10',
			    -font=>[-weight=>'normal', -size=>12, -family=>'Courier'],		#see: http://docstore.mik.ua/orelly/perl3/tk/ch03_01.htm
 			    -foreground=>'#000077',
                            -background=>'#FFFFFF',                        )->pack(-side => 'top',-fill => 'x');
  $logBox->Subwidget("xscrollbar")->configure(-background=>$btnbkgcolor,);
  $logBox->Subwidget("yscrollbar")->configure(-background=>$btnbkgcolor,);
###===========================

### Set up binding so that, when the left mouse button (1)
### is clicked on the image window ($imDisplay), &(process_click) is
### called and is given the x and y coordinates of the click.
### See: http://oreilly.com/catalog/mastperltk/chapter/ch15.html

  for ($imindx=0; $imindx<$nDisplay; $imindx++) {
     $imDisplay[$imindx]->bind('<1>',      [\&process_click, Ev('x'), Ev('y')]);		#1st argument $widget is passed automatically
     $imDisplay[$imindx]->bind('<Motion>', [\&display_coords, Ev('x'), Ev('y'), $imindx]);	#1st argument $widget is passed automatically
     $imDisplay[$imindx]->bind('<Leave>',  sub{$imLabelXY->configure(-text=>'');});
  }

  $mw->protocol('WM_DELETE_WINDOW', \&goodbye);
# $mw->protocol('WM_DELETE_WINDOW', [\&goodbye]);
# $mw->bind('<Destroy>', [\&goodbye]);			#this does not work
# $mw->bind('<KeyPress-Escape>', [\&goodbye]);

  $mw->resizable( 0, 0 );
  $mw->focus();
  &switch_sonicc_RL('Remote');
  &request_SONICC_to_load_scan_config();

  $repeater      = $mw->repeat(100,\&monitorPVs);

  $useragent = new LWP::UserAgent;
  $useragent->timeout(3);						# timeout in nn seconds

  MainLoop;
  &goodbye;

#==========================================================================
sub monitorPVs() {

  ### monitor PVs: this function is called
  ### every 100ms by $mw->repeat:
  &calculate_soniccPos_tag();				#currently we only update SONICC position here, but we can add more monitors
  if (Exists($mw)) {$mw->update;}
}

#==========================================================================
### Called on chosen event on the main window.
#sub process_click($widget,$xpixel,$ypixel) {
sub process_click($$$) {
  my ($widget, $x, $y, $x_, $y_, $z_, $phi, $status, $msg);
  my (@pars, $dSampleX, $dSampleY, $dSampleH);
  my ($omegaOffset, $xcenter_px, $ycenter_px, $xpixel2um, $ypixel2um, $axes_rotation, $focusOffset);

  ($widget, $x, $y) = @_;
  if (!$IMAGE_LOADED) {return;}
  ### One can only click in continuous mode or when no SONICC running:
  if ($SCAN_IN_PROGRESS == 1 || $SCAN_IN_PROGRESS == 3) {
     &updateLogBox('!!! Ignoring clicks while Single-shot or Z-scan are in progress');
     return;
  }
  $msg = $x.' '.$y;
# $imLabelXY->configure(-text=>$msg);
  &updateLogBox('Clicked at '.$msg);
  if (!defined $images[$currImage]{mode}) {
     &updateLogBox('*** Undefined soniccMode of currently displayed image');
     return;
  }
  if    ($images[$currImage]{mode} eq '0') {@pars=@parsSHG;}
  elsif ($images[$currImage]{mode} eq '1') {@pars=@parsTPE;}
  else {&updateLogBox('*** Unknown soniccMode='.$images[$currImage]{mode}); return;}
  if (!defined $magnify) {$magnify=1;}
### @pars = omegaOffset xcenter(px) ycenter(px) xpixel2um ypixel2um axes_rotation focusOffset
  ($omegaOffset, $xcenter_px, $ycenter_px, $xpixel2um, $ypixel2um, $axes_rotation, $focusOffset) = @pars;
  if ($magnify > 1) {
     $xcenter_px = $xcenter_px * $magnify;
     $ycenter_px = $ycenter_px * $magnify;
     $xpixel2um  = $xpixel2um  / $magnify;
     $ypixel2um  = $ypixel2um  / $magnify;
  }
### @pars = omegaOffset xcenter(px) ycenter(px) xpixel2um ypixel2um axes_rotation focusOffset
  if (! $rotate) {
     $axes_rotation = $axes_rotation * $degree2rad;
     $x_ = ( ($x - $xcenter_px) * cos($axes_rotation)
         +   ($y - $ycenter_px) * sin($axes_rotation) ) * $xpixel2um;
     $y_ = ( ($x - $xcenter_px) * sin($axes_rotation)
         +   ($y - $ycenter_px) * cos($axes_rotation) ) * $ypixel2um;
  } else {
     $axes_rotation = ($axes_rotation+90) * $degree2rad;
     $x_ = -( ($x - $ycenter_px) * cos($axes_rotation)
         +    ($y - $xcenter_px) * sin($axes_rotation) ) * $ypixel2um;
     $y_ =  ( ($x - $ycenter_px) * sin($axes_rotation)
         +    ($y - $xcenter_px) * cos($axes_rotation) ) * $xpixel2um;
  }
  $phi = $images[$currImage]{omega} +  $omegaOffset;    			# omega+omegaOffset
  $phi = $phi - 360.*int($phi/360);
  if ($phi < 0.) {$phi += 360.;}
  $phi = $phi * $degree2rad;
  $x_  = 0.001 * $x_;								# um to mm
  $y_  = 0.001 * $y_;								# um to mm
  if ($SCAN_IN_PROGRESS == 0) {
     ### When we are clicking, scan is stopped and sample is shifted
     ### back by -$focusOffset and not where we recorded positions
     ### while acquiring this particular image:
     $z_ = -0.001 * $focusOffset;
  }										# um to mm
  elsif ($SCAN_IN_PROGRESS == 2) {
     ### No need for z-shift in the continuous scan mode:
     ### the sample will be shifted after scan is stopped.
     $z_ = 0;
  }
  $dSampleX = sin($phi) * $y_ + cos($phi) * $z_;				# in mm
  $dSampleY = cos($phi) * $y_ - sin($phi) * $z_;				# in mm
  $dSampleH =-$x_;								# in mm

  $msg = sprintf 'ABS %.2f %.2f %.2f %.1f um', 1000*($images[$currImage]{x}+$dSampleX),
					       1000*($images[$currImage]{y}+$dSampleY),
					       1000*($images[$currImage]{h}+$dSampleH),
					       $images[$currImage]{omega};		# omega
  &updateLogBox('Sending P&C cmd: '.$msg);
  if (length($msg) < 40) {
     $status = Pezca::PutStringOldCa($clickPV,$msg);
     if ($status) {&report_CAerr($status,$clickPV);}
     else         {&updateLogBox('Sent to P&C server: '.$msg);}
  } else {
     &updateLogBox('Too long P&C cmd');
     $status = -1;
  }
}

#--------------------------------------------------------------------------
#sub display_coords($widget,$xpixel,$ypixel,$imindx) {
sub display_coords($$$$) {
  my ($widget, $x, $y, $imindx, $intensity, $msg);

  ($widget, $x, $y, $imindx) = @_;
  if (! Exists($mw)) {return;}
  if ($IMAGE_LOADED &&
      defined $images[$currImage]{'ncols'.$imindx} &&
      defined $images[$currImage]{'nlines'.$imindx}) {
     if ($x >= 0 && $x < $images[$currImage]{'ncols'.$imindx}  &&
         $y >= 0 && $y < $images[$currImage]{'nlines'.$imindx}) {
        $msg = 'x='.$x.' y='.$y;
#       $intensity = at($images[$currImage]{'intensity'.$imindx},($x,$y));
        $intensity = $images[$currImage]{'pixels'.$imindx}[$x][$y];
        if (defined $intensity) {$msg .= ' Intensity='.$intensity;}
	else                    {$msg .= ' Intensity=???';}
     } else {
        $msg = '';
     }
     if (Exists($imLabelXY)) {$imLabelXY->configure(-text=>$msg);}
     if (Exists($mw))        {$mw->update;}
  }
}

#--------------------------------------------------------------------------
#sub drawScaleBox(\$image,$mode,$width,$height) {
sub drawScaleBox(\$$$$) {
  my ($imgref, $mode, $width, $height);
  my ($scaleSize, $scaleColor, $scalex_px, $scaley_px, $wrkpixel);
  my ($xpixel2um, $ypixel2um, $off, $fontsize, $label, $xc, $yc);

  ($imgref, $mode, $width, $height) = @_;
  $scaleSize  = 100;									#50um
  $scaleColor = 'red';
  $off        = 5;									#offset from end in px
  $fontsize   = 11;									#in px
  $label      = $scaleSize.' um';
  if    ($mode eq '0') {
     $xc        = $parsSHG[1];
     $yc        = $parsSHG[2];
     $xpixel2um = $parsSHG[3];
     $ypixel2um = $parsSHG[4];
  }
  elsif ($mode eq '1') {
     $xc        = $parsTPE[1];
     $yc        = $parsTPE[2];
     $xpixel2um = $parsTPE[3];
     $ypixel2um = $parsTPE[4];
  }
  else {
     &updateLogBox('*** drawScaleBox: Unknown soniccMode='.$mode);
     return;
  }
  if ($magnify > 1) {
     $xpixel2um  = $xpixel2um  / $magnify;
     $ypixel2um  = $ypixel2um  / $magnify;
  }
  if ($rotate) {								#swap x and y
     $wrkpixel  = $xpixel2um;
     $xpixel2um = $ypixel2um;
     $ypixel2um = $wrkpixel;
  }
  $scalex_px = int($scaleSize / $xpixel2um + 0.5);
  $scaley_px = int($scaleSize / $ypixel2um + 0.5);

  $$imgref->Draw(stroke=>$scaleColor,
                 primitive=>'rectangle',
		 points=>$off.','.($height-$off).' '.($off+$scalex_px).','.($height-$off-$scaley_px));
  if ($width > $fontsize*length($label)+4 && $height > $fontsize+4) {
### Try to fit label inside the box:
     $$imgref->Draw(stroke=>$scaleColor,primitive=>'text', text=>$label,
		   pointsize=>$fontsize,
                   points=>($off+2).','.($height-$off-2));
  } else {
### Draw label above the box:
     $$imgref->Draw(stroke=>$scaleColor,primitive=>'text', text=>$label,
		   pointsize=>$fontsize,
                   points=>($off+2).','.($height-$off-$scaley_px-2));
  }
### Draw center crosshair
# $xc = int(0.5*$width+0.5);
# $yc = int(0.5*$height+0.5);
  $$imgref->Draw(stroke=>$scaleColor,
                 primitive=>'line',
		 points=>($xc-$off).','.$yc.' '.($xc+$off).','.$yc);
  $$imgref->Draw(stroke=>$scaleColor,
                 primitive=>'line',
		 points=>$xc.','.($yc-$off).' '.$xc.','.($yc+$off));
}

#--------------------------------------------------------------------------
#sub updateLogBox($msg,[$tag]) {
sub updateLogBox($;$) {
  my ($msg, $tag);
  $msg = shift(@_);
  $tag = shift(@_);
  if (!defined $tag) {				#auto-detect tag, if not specified
     if ($msg =~ /^\*\*\*/ ||
         $msg =~ / error/  ||
         $msg =~ /Error/) {$tag = 'error';}
     else                 {$tag = 'info';}
  }
  if (Exists($mw) && defined $logBox && Exists($logBox)) {
     $logBox->insert('end', $msg."\n", $tag);
     if    ($tag eq 'error') {$logBox->tagConfigure($tag, -foreground=>$redcolor);}
     elsif ($tag eq 'warn')  {$logBox->tagConfigure($tag, -foreground=>$goldcolor);}
     $logBox->see('end');
     if (Exists($mw)) {$mw->update;}
  } else {
     print STDOUT $msg."\n";
  }
}

#--------------------------------------------------------------------------
#sub retrieve_gonio_positions(\@positions,[$log_flag]) {
sub retrieve_gonio_positions(\@;$) {
  my ($status, $i, $msg, $pos_ref, $log_flag);

  $pos_ref  = shift(@_);
  $log_flag = shift(@_);	if (!defined $log_flag) {$log_flag = 0;}

  if ($log_flag) {&updateLogBox('Retrieving gonio positions (mm):');}
  for ($i=0; $i<$nPVs; $i++) {
     ($status, @$pos_ref[$i]) = Pezca::GetDouble($gonioPVs[$i]);
     if ($status) {
        &report_only($script,$status,$gonioPVs[$i]);
        &updateLogBox('*** ERROR='.$status.' reading PV='.$gonioPVs[$i]);
	return $status;
     }
     if ($log_flag) {
        $msg = sprintf '%s=%.3f', $gonioPVs[$i], @$pos_ref[$i];
        &updateLogBox($msg);
     }
  }
  return 0;
}

#--------------------------------------------------------------------------
#sub soniccMove($dest) {
sub soniccMove($) {
  my ($dest, @args, $status, $i);
  $dest = shift(@_);
  if     ($dest =~ /^IN$/i) {
     if ($soniccPosition eq 'IN') {
        for ($i=0; $i<@lightsPVs; $i++) {
           ($status, $lightsVal[$i]) = Pezca::GetDouble($lightsPVs[$i]);	#remember before turning off
           if ($status) {&report_CAerr($status,$lightsPVs[$i]);}
        }
     }
     &updateLogBox('Turning OFF sample illumination lights');
     for ($i=0; $i<@lightsPVs; $i++) {
        $status = Pezca::PutDoubleOldCa($lightsPVs[$i],0);			#turn light off
        if ($status) {&report_CAerr($status,$lightsPVs[$i]);}
     }
#    @args = ('soniccIN', 'collimatorOUT', 'backlight=off', 'ringlight=off');
     @args = ('soniccIN', 'collimatorOUT');
  }
  elsif ($dest =~ /^OUT$/i) {
     &updateLogBox('Turning ON sample illumination lights');
     for ($i=0; $i<@lightsPVs; $i++) {
	if (defined $lightsVal[$i]) {
           $status = Pezca::PutDoubleOldCa($lightsPVs[$i],$lightsVal[$i]);	#turn lights back on
           if ($status) {&report_CAerr($status,$lightsPVs[$i]);}
        }
     }
     @args = ('IN');								#collimator-In + move SONICC out
  }
  else {
     &updateLogBox('*** Incorrect SONICC destination='.$dest);
     return;
  }
  &updateLogBox('Moving SONICC '.$dest);
  $status = &setSampleEnv(\@args,'nowait');
  if ($status) {&updateLogBox('*** Failed to move SONICC '.$dest);}
  else         {&updateLogBox('Successfully sent command to move SONICC '.$dest);}
  return $status;
}

#--------------------------------------------------------------------------
sub calculate_soniccPos_tag() {
  my ($inPos, $runPrg, $actPos, $rqsPos, $ampEna, $status, $prevPos);
  my ($direction, $percent);

  $prevPos = $soniccPosition;

  if (Exists($mw)) {$mw->update;} else {return;}
  $inPos = &getInPos ($soniccAssy_PV,\$status);
  if ($status) {
     &updateLogBox('*** Error reading SONICC inpos.');
     $soniccPosition = 'Unknown';
     if ($prevPos ne $soniccPosition && Exists($lblInOut)) {
        $lblInOut->configure(-foreground=>$yellowcolor);
     }
     if (Exists($mw)) {$mw->update;} return;
  }
  $runPrg = &getRunPrg ($soniccAssy_PV,\$status);
  if ($status) {
     &updateLogBox('*** Error reading SONICC runprg.');
     $soniccPosition = 'Unknown';
     if ($prevPos ne $soniccPosition && Exists($lblInOut)) {
        $lblInOut->configure(-foreground=>$yellowcolor);
     }
     if (Exists($mw)) {$mw->update;} return;
  }
  $actPos = &getAxisAct ($soniccAxis_PV,\$status);
  if ($status) {
     &updateLogBox('*** Error reading SONICC actpos.');
     $soniccPosition = 'Unknown';
     if ($prevPos ne $soniccPosition && Exists($lblInOut)) {
        $lblInOut->configure(-foreground=>$yellowcolor);
     }
     if (Exists($mw)) {$mw->update;} return;
  }
  $rqsPos = &getAxisRqs ($soniccAxis_PV,\$status);
  if ($status) {
     &updateLogBox('*** Error reading SONICC rqspos.');
     $soniccPosition = 'Unknown';
     if ($prevPos ne $soniccPosition && Exists($lblInOut)) {
        $lblInOut->configure(-foreground=>$yellowcolor);
     }
     if (Exists($mw)) {$mw->update;} return;
  }
  $ampEna = &checkMotorAmp ($soniccAxis_PV,0,\$status);
  if ($status) {
     &updateLogBox('*** Error reading SONICC motor amplifier.');
     $soniccPosition = 'Unknown';
     if ($prevPos ne $soniccPosition && Exists($lblInOut)) {
        $lblInOut->configure(-foreground=>$yellowcolor);
     }
     if (Exists($mw)) {$mw->update;} return;
  }

  if (abs($rqsPos-$sonicc_in_pos) < abs($rqsPos-$sonicc_out_pos)) {
     $direction = 'IN';
     $percent   = 100*abs(($actPos-$sonicc_out_pos)
                / ($sonicc_out_pos-$sonicc_in_pos));
  } else {
     $direction = 'OUT';
     $percent   = 100*abs(($actPos-$sonicc_in_pos)
                / ($sonicc_out_pos-$sonicc_in_pos));
  }
  $percent = int($percent+0.5);

  if ($inPos != 1 || $runPrg == 1) {
     if ($ampEna == 0) {$soniccPosition = 'stuck@middle';}
     else              {$soniccPosition = 'run'.$direction.' '.$percent.'%';}			#runOUT 45%
     if ($prevPos ne $soniccPosition && Exists($lblInOut)) {
        $lblInOut->configure(-foreground=>$yellowcolor);
     }
  } else {
     if    (abs($actPos-$sonicc_in_pos)  < $sonicc_tolerance) {
        $soniccPosition = 'IN';
        if ($prevPos ne $soniccPosition && Exists($lblInOut)) {
           $lblInOut->configure(-foreground=>$greencolor);
        }
     }
     elsif (abs($actPos-$sonicc_out_pos) < $sonicc_tolerance) {
        $soniccPosition = 'OUT';
        if ($prevPos ne $soniccPosition && Exists($lblInOut)) {
           $lblInOut->configure(-foreground=>$redcolor);
        }
     }
     else {
       $soniccPosition = 'seats@middle';
       if ($prevPos ne $soniccPosition && Exists($lblInOut)) {
          $lblInOut->configure(-foreground=>$yellowcolor);
       }
     }
  }
  if (Exists($mw)) {$mw->update;}
  return;
}

#--------------------------------------------------------------------------
sub soniccWaitForIN() {
  my ($status, $inPos, $amp);
  my ($wait_, $wait_step, $wait_max, $w_, $wait_print, $msg);

  if ($soniccPosition eq 'IN') {return $soniccPosition ;}

  &updateLogBox('SONICC is not IN. Moving there...');
  $status = &soniccMove ('IN');
  $w_    = 0;
  $wait_ = 0;
  $wait_max = 120;
  $wait_step = 0.2;
  $wait_print = 10;
  $SCAN_IN_PROGRESS = 4;
  do {
     select (undef, undef, undef, $wait_step);
     &calculate_soniccPos_tag();
     $inPos = &getInPos ($soniccAssy_PV);
     $amp = &checkMotorAmp ($soniccAxis_PV);
     if ($inPos eq 1) {$inPos = 'Positioned';} else {$inPos = 'Moving';}
     if ($amp   eq 1) {$amp   = 'ON';}         else {$amp   = 'OFF';}
     $wait_ += $wait_step;
     $w_    += $wait_step;
     if ($w_ > $wait_print) {
        $w_ = 0;
        $msg = sprintf '%ds waiting for SONICC instrument to move IN.'
                      .' Current State="%s", Pos=%s, Amp=%s',
                      $wait_, $soniccPosition, $inPos, $amp;
        &updateLogBox($msg);
     }
  } while ($wait_          < $wait_max &&
           $soniccPosition ne 'IN'     &&
           $amp            ne 'OFF'    &&
           ($inPos         eq 'Moving' || $wait_ < 10) &&
           $SCAN_IN_PROGRESS);
  if ($soniccPosition ne 'IN') {
     if    ($wait_ >= $wait_max) {&updateLogBox('*** Timeout on moving SONICC IN.');}
     elsif ($amp   eq 'OFF')     {&updateLogBox('*** SONICC motor amplifier is OFF.');}
     elsif (!$SCAN_IN_PROGRESS)  {&updateLogBox('*** SONICC motion aborted.');}
     elsif ($inPos ne 'Moving')  {&updateLogBox('*** SONICC stopped moving due to unknown reason.');}
     else                        {&updateLogBox('*** Failed to move SONICC IN.');}
  } else {
     &updateLogBox('SONICC has been moved IN.');
  }
  $SCAN_IN_PROGRESS = 0;
  return $soniccPosition;
}

#--------------------------------------------------------------------------
#sub switch_sonicc_RL(Remote|Local) {
sub switch_sonicc_RL($) {
  my ($state, $status);
  $state = shift(@_);
  if ($state ne 'Remote' && $state ne 'Local') {
     &updateLogBox('*** Incorrect SONICC state='.$state.' -- expecting "Remote" or "Local"');
     return -1;
  }
  &updateLogBox('Switching SONICC to '.$state.' mode');
  $status = &send_SONICC('Control_State '.$state, $debug);
  return $status;
}

#--------------------------------------------------------------------------
#sub send_SONICC(cmd,[$debug]) {						#returns $status
sub send_SONICC($;$) {
  my ($cmd, $status, $debug);
  $cmd   = shift(@_);
  $debug = shift(@_);  if (!defined $debug) {$debug=0;}
  if ($debug) {&updateLogBox('Sending "'.$cmd.'" to SONICC');}
  $status = &PezcaPutStringWF($sonicCmdPV,$cmd);
  if ($status) {
     &report_CAerr($status,$sonicCmdPV);
     &updateLogBox('*** Command "'.$cmd.'" failed');
  }
  return $status;
}

#--------------------------------------------------------------------------
#sub query_SONICC($cmd,[$debug]) {						#returns $status,$rsp
sub query_SONICC($;$) {
  my ($cmd, $rsp, $status, $debug);
  $cmd = shift(@_);
  $debug = shift(@_);  if (!defined $debug) {$debug=0;}
  $status = &send_SONICC($cmd,$debug);
  if ($status == 0) {
     select (undef, undef, undef, 0.2);
     ($status,$rsp) = &PezcaGetStringWF($sonicRspPV);
     if ($status) {
        &report_CAerr($status,$sonicRspPV);
        &updateLogBox('*** Failed to get response to "'.$cmd.'"');
     } else {
        if ($debug) {&updateLogBox('SONICC responded: "'.$rsp.'"');}
     }
  }
  return $status, $rsp;
}

#--------------------------------------------------------------------------
#sub request_SONICC_to_load_scan_config() {
sub request_SONICC_to_load_scan_config() {
  my ($iDisplay, $status);
  for ($iDisplay=0; $iDisplay<$nDisplay; $iDisplay++) {
     if (defined $imHeader[$iDisplay]) {$imHeader[$iDisplay]->configure(-text=>$header[$soniccMode][$iDisplay]);}
  }
  &updateLogBox('Requesting SONICC to load '.$scanLabels[$soniccMode].' config '.$windowsConfigPath.$modeConfigs[$soniccMode]);
  $status = &send_SONICC('Imaging_Config '.$windowsConfigPath.$modeConfigs[$soniccMode], 1);
  return $status;
}

#--------------------------------------------------------------------------
#sub update_calibration_label($soniccMode) {
sub update_calibration_label($) {
  my ($soniccMode, $msg);

  $soniccMode = shift(@_);
  if ($soniccMode == 0) {							#SGH mode
### 0=omegaOffset 1=xc(px) 2=yc(px) 3=xpixel2um 4=ypixel2um, 5=axes_rotation, 6=focusOffset
     $msg = 'xc='.$parsSHG[1].' yc='.$parsSHG[2]
        .' xp2u='.$parsSHG[3].' yp2u='.$parsSHG[4]
        .' a='.$parsSHG[5].' focus='.$parsSHG[6];
  }
  elsif ($soniccMode == 1) {							#TPE-UVF mode
     $msg = 'xc='.$parsTPE[1].' yc='.$parsTPE[2]
        .' xp2u='.$parsTPE[3].' yp2u='.$parsTPE[4]
        .' a='.$parsTPE[5].' focus='.$parsTPE[6];
  }
  else {
     $msg = '';
  }
  if (defined $calibrLbl && Exists($calibrLbl)) {$calibrLbl->configure(-text=>$msg);}
}

#--------------------------------------------------------------------------
#sub report_CAerr($status,$pv) {
sub report_CAerr($$) {
  my ($status, $pv, $statusDescr);
  $status  = shift(@_);
  $pv      = shift(@_);
  if ($status) {
     $statusDescr  = &ezcaStatusCode($status);
     &updateLogBox('*** EZCA error='.$status.' ['.$statusDescr.'] PV=['.$pv.']');
  }
}

#--------------------------------------------------------------------------
sub open_sonicc_medm() {
  our ($CAQTDM);							#from beamline.pl
  my ($adl, $medmprg, $macro, $string, $status);
  if (!defined $ENV{FASTSCANS}) {
     &updateLogBox('*** Environment FASTSCANS is undefined');
     return;
  }
  if (!defined $CAQTDM) {$CAQTDM=0;}

  $adl = $ENV{FASTSCANS}.'/../gmcaApp/adl/sonicc.adl';

  if ($CAQTDM == 0) {$medmprg = 'medm';}
  else              {$medmprg = 'startDM'; $adl =~ s/\.adl$/\.ui/;}

  if (-f $adl) {
     $macro = 'bln='.$beamline.':,assy='.$soniccAssy_PV.':,axis='.$soniccAxis_PV.':';
     $string = $medmprg.' -attach -macro "'.$macro.'" -x '.$adl.' >& /dev/null &';
     $status = system($string);
     if ($status) {&updateLogBox('*** Error starting MEDM: status='.$status);}
  } else {
     &updateLogBox('*** MEDM file "'.$adl.'" not found.');
  }
}

#--------------------------------------------------------------------------
sub open_calibr_medm() {
  our ($CAQTDM);							#from beamline.pl
  my ($adl, $medmprg, $macro, $string, $status);
  if (!defined $ENV{FASTSCANS}) {
     &updateLogBox('*** Environment FASTSCANS is undefined');
     return;
  }
  if (!defined $CAQTDM) {$CAQTDM=0;}

  $adl = $ENV{FASTSCANS}.'/../gmcaApp/adl/soniccServer.adl';

  if ($CAQTDM == 0) {$medmprg = 'medm';}
  else              {$medmprg = 'startDM'; $adl =~ s/\.adl$/\.ui/;}

  if (-f $adl) {
     $macro = 'ioc='.$beamline.':ioc3,P='.$beamline.':SONICC';
     $string = $medmprg.' -attach -macro "'.$macro.'" -x '.$adl.' >& /dev/null &';
     $status = system($string);
     if ($status) {&updateLogBox('*** Error starting MEDM: status='.$status);}
  } else {
     &updateLogBox('*** MEDM file "'.$adl.'" not found.');
  }
}

#--------------------------------------------------------------------------
sub disable_buttons() {
  my ($i);
  if ($SCAN_IN_PROGRESS == 1) {
     $onceBtn->configure(-relief=>'sunken');					#make scan button pressed
     if ($enable_continuous) {
        $contBtn->configure(-state=>'disabled');				#disable
     }
     $zscanBtn->configure(-state=>'disabled');					#disable
  }
  elsif ($SCAN_IN_PROGRESS == 2) {
     if ($enable_continuous) {
        $contBtn->configure(-relief=>'sunken');					#make scan button pressed
     }
     $onceBtn->configure(-state=>'disabled');					#disable
     $zscanBtn->configure(-state=>'disabled');					#disable
  }
  elsif ($SCAN_IN_PROGRESS == 3) {
     $zscanBtn->configure(-relief=>'sunken');					#make scan button pressed
     $onceBtn->configure(-state=>'disabled');					#disable
     if ($enable_continuous) {
        $contBtn->configure(-state=>'disabled');				#disable
     }
  }
  elsif ($SCAN_IN_PROGRESS == 4) {
     $onceBtn->configure(-state=>'disabled');					#disable
     if ($enable_continuous) {
        $contBtn->configure(-state=>'disabled');				#disable
     }
     $zscanBtn->configure(-state=>'disabled');					#disable
  }
  $dirBtn->configure(-state=>'disabled');					#disable directory change
  if ($SCAN_IN_PROGRESS == 3) {
     $scnLbox->configure(-state=>'disabled');					#disable choice of scan pts
     $entryStep->configure(-state=>'disabled');					#disable changing scan step
  }
  for ($i=0; $i<2; $i++) {
     $sclRadio[$i]->configure(-state=>'disabled');				#disable log/linear radio
  }
  for ($i=0; $i<$nModes; $i++) {
     $scnRadio[$i]->configure(-state=>'disabled');				#disable scantype radio
  }
  $prevBtn->configure(-state=>'disabled');					#disable image browsing buttons
  $nextBtn->configure(-state=>'disabled');		       			#disable image browsing buttons

  $localBtn->configure(-state=>'disabled');					#disable swithching SONICC to local
  $remoteBtn->configure(-state=>'disabled');					#disable swithching SONICC to remote

  $saveimgBtn->configure(-state=>'disabled');					#disable saving images
  $rdconfBtn->configure(-state=>'disabled');					#disable reload config
}

#--------------------------------------------------------------------------
sub enable_buttons() {
  my ($i);
  $onceBtn->configure(-relief=>'raised');					#'unpress' scan button
  $onceBtn->configure(-state=>'normal');					#re-enable scan button
  if ($enable_continuous) {
     $contBtn->configure(-relief=>'raised');					#'unpress' scan button
     $contBtn->configure(-state=>'normal');					#re-enable scan button
  }
  $zscanBtn->configure(-relief=>'raised');					#'unpress' scan button
  $zscanBtn->configure(-state=>'normal');					#re-enable scan button
  $dirBtn->configure(-state=>'normal');						#re-enable directory choice
  $scnLbox->configure(-state=>'normal');					#re-enable choice of scan pts
  $entryStep->configure(-state=>'normal');					#re-enable changing scan step
  if ($IMAGE_LOADED) {
     for ($i=0; $i<2; $i++) {
        $sclRadio[$i]->configure(-state=>'normal');				#re-enable log/linear radio
     }
  }
  for ($i=0; $i<$nModes; $i++) {
     $scnRadio[$i]->configure(-state=>'normal');				#re-enable scantype radio
  }
  if ($lastScan_pts > 1) {
     if ($currImage > 0) {
        $prevBtn->configure(-state=>'normal');					#re-enable image browsing buttons
     }
     if ($currImage < $lastScan_pts-1) {
        $nextBtn->configure(-state=>'normal');					#re-enable image browsing buttons
     }
  }

  $localBtn->configure(-state=>'normal');					#disable swithching SONICC to local
  $remoteBtn->configure(-state=>'normal');					#disable swithching SONICC to remote

  $saveimgBtn->configure(-state=>'normal');					#disable saving images
  $rdconfBtn->configure(-state=>'normal');					#disable reload config
}

#--------------------------------------------------------------------------
sub set_wrk_dir() {
### This program sets the directory under user HOME where user can
### save images for his records.

  my ($dir, $msg, $status, $iniDir);

  if ($SUB_IN_PROGRESS) {return;}
  $SUB_IN_PROGRESS=1;

  if (!defined $userSaveDir) {$userSaveDir = '~/sonicc';}
  $iniDir = $userSaveDir;
### See: http://search.cpan.org/dist/Tk/pod/chooseDirectory.pod
  $dir = $mw->chooseDirectory(-initialdir=>$iniDir,
# 			      -mustexist=>0,		#unsupported on RHEL6 Linux
   			      -title=>'Choose a Directory');

  if (defined $dir) {
     if ($dir !~ /\/$/) {$dir .= '/';}						#append "/"
     if (! -e $dir) {
	$status = &make_directory($dir,\$msg);
	if ($status) {
           &updateLogBox($msg);
           $SUB_IN_PROGRESS=0;
           return;
        }
     }
     elsif (! -d $dir) {
	$msg = '*** Selection "'.$dir.'" is not a directory.';
        &updateLogBox($msg);
	$SUB_IN_PROGRESS=0; return;
     }
     $msg = 'New working directory is "'.$dir.'".';
     &updateLogBox($msg);
     if ($iniDir ne $dir) {
        $dirLbl->configure(-text=>$dir);
     }
     $userSaveDir = $dir;
  }
  $SUB_IN_PROGRESS=0;
}

#--------------------------------------------------------------------------
sub scan_init() {
  my ($status, $state);
  $SUB_IN_PROGRESS=1;

  $state = &soniccWaitForIN();
  if ($state ne 'IN') {$SUB_IN_PROGRESS=0; return 1;}

  &updateLogBox('Setting SONICC State '.$scanLabels[$soniccMode]);
  $status = &send_SONICC('SONICC_State '.$scanSwitchers[$soniccMode], $debug);
  if ($status) {$SUB_IN_PROGRESS=0; return 1;}
  select (undef, undef, undef, 0.2);
  &updateLogBox('Setting SONICC Resonator ON');
  $status = &send_SONICC('Resonator_State On', $debug);
  if ($status) {$SUB_IN_PROGRESS=0; return 1;}
  select (undef, undef, undef, 0.2);
  return 0;
}

#--------------------------------------------------------------------------
#sub start_zscan('single'|'zscan') {
sub start_zscan($) {
  my ($operation, $scan_points, $status, $nimages, $offPts);
  my ($waited, $iStep, $zpos, $focusOffset);
  my ($msg, $i, $ntaken);

  $operation = shift(@_);
  if    ($operation =~ /^single$/) {$scan_points = 1;}
  elsif ($operation =~ /^zscan$/)  {$scan_points = $scanPts;}
  else                             {&updateLogBox('*** Incorrect Scan Mode: '.$operation); return;}

  if ($SUB_IN_PROGRESS) {return;}

  $status = &scan_init();
  if ($status) {return;}

  if    ($operation =~ /^single$/) {$SCAN_IN_PROGRESS = 1;}			#1=oneShot,2=Cont.,3=zScan
  elsif ($operation =~ /^zscan$/)  {$SCAN_IN_PROGRESS = 3;}                     #1=oneShot,2=Cont.,3=zScan
  &disable_buttons();

  $scan_points = 2*int(($scan_points-1)/2)+1;
  $status = &retrieve_gonio_positions(\@gonioAct, 1);				#this is in mm
  if ($status) {
     &updateLogBox('*** Failed to get Gonio positions. Quitting SONICC Scan');
     $SCAN_IN_PROGRESS = 0;							#1=oneShot,2=Cont.,3=zScan
     $SUB_IN_PROGRESS = 0;
     &enable_buttons();
     return;
  }
  if    ($soniccMode == 0) {$focusOffset = $parsSHG[6];}			# $soniccMode=0=SHG
  elsif ($soniccMode == 1) {$focusOffset = $parsTPE[6];}			# $soniccMode=1=TPE-UVF
  else {
     $focusOffset = 0; &updateLogBox('*** Incorrect soniccMode='.$soniccMode);
  }

  &updateLogBox('Starting SONICC Scan');
  $offPts = int(($scan_points-1)/2);
  $zpos = -$offPts*$scanStep + $focusOffset;
  $msg = 'Shifting the sample by '.$zpos.'um to the';
  if ($scan_points > 1) {$msg .= ' initial Z-position';}
  else                  {$msg .= ' SONICC focal plane';}
  if ($zpos != 0) {
     &updateLogBox($msg.' ("+/-" is down/up stream)');
     $msg = '2Z '.$zpos.' 0';
     $status = Pezca::PutStringOldCa($clickPV,$msg);
     if ($status) {&report_CAerr($status,$clickPV);}
     else         {&updateLogBox('Sent to P&C server: '.$msg);}
     $waited = &waitMotor($xyAssyPV,10,1);			     		#waited: value=how-long-waited, 0=timeout
  }

  @images       = ();								#clear hash array of image objects and associated information
  $lastScan_pts = 0;								#remember conditions at which the images were taken (for later preview).
										#we increment it in the loop, so that if scan aborted, we can still see what has been collected
  for ($iStep=0; $iStep<$scan_points; $iStep++) { #######################################
                                                                                        #
     #################                                                                  #
     ### Take image  #                                                                  #
     #################                                                                  #
     $status = &command_SONICC_take_image();                                            #
     if ($status) {return 1;}                                                           #
                                                                                        #
     #####################                                                              #
     ### Display images: #                                                              #
     #####################                                                              #
     ### 1. Construct image label                                                       #
     if ($scan_points > 1) {                                                            #
        $zpos = ($iStep - int(($scan_points-1)/2))*$scanStep;                           #
	$zpos = 'z='.$zpos.'um';                                                        #
     } else {                                                                           #
        $zpos = '';                                                                     #
     }                                                                                  #
     $nimages = 0;                                                                      #
     ### 2. Read images if available and display them:                                  #
     $ntaken = &read_www_images($lastScan_pts,$zpos);                                   #
     if ($ntaken == 0) {                                                                #
        &updateLogBox('!!! read_www_images returned no images');                        #
        next;                                                                           #
     }                                                                                  #
     for ($i=0; $i<$nDisplay; $i++) {                                                   #
        &display_image($lastScan_pts,$i);                                               #
        $imLabelNm[$i]->configure(-text=>$zpos);                                        #
     }                                                                                  #
     $currImage = $lastScan_pts;                                                        #
     $lastScan_pts++;						                        #one more set of images is taken
     $IMAGE_LOADED = 1;                                                                 #
     for ($i=0; $i<2; $i++) {$sclRadio[$i]->configure(-state=>'normal');}               #
     &update_calibration_label($soniccMode);                                            #
     $mw->focus();                                                                      #
                                                                                        #
     ###########################                                                        #
     ### Move to new Z-position:                                                        #
     ###########################                                                        #
     if ($iStep < $scan_points-1) {                                                     #
        &updateLogBox('Shifting the sample downstream by '.$scanStep.'um.');            #
        $msg = '2Z '.$scanStep.' 0';                                                    #
        $status = Pezca::PutStringOldCa($clickPV,$msg);                                 #
        if ($status) {&report_CAerr($status,$clickPV);}                                 #
        else         {&updateLogBox('Sent to P&C server: '.$msg);}                      #
        $waited = &waitMotor ($xyAssyPV,10,1);		     				#waited: value=waited, 0=timeout
     }                                                                                  #
  } #####################################################################################
  &sonicc_standby();
  ### Return the sample back to original Z-position:
  &updateLogBox('Shifting the sample back to the orinal Z-position.');
  $status = &runNMotors ($xyAssyPV, 2, \@xyAxesPV, [$gonioAct[0],$gonioAct[1]], 'forceAuto');
  if ($status) {&updateLogBox('*** Failed to return sample to the initial Z-position.');}
  else         {&updateLogBox('Sample Z-position is restored.');}
  &enable_buttons();
  &updateLogBox('SONICC Z-Scan finished');
  $SCAN_IN_PROGRESS=0;
  $SUB_IN_PROGRESS=0;
}

#--------------------------------------------------------------------------
sub command_SONICC_take_image() {
  my ($status, $wait_, $wait_step, $wait_max, $w_, $wait_print, $progress);

  &updateLogBox('Taking SONICC image');
  $status = &send_SONICC('Scan_Once '.$windowsImageDir, 1);
  if ($status) {
     &enable_buttons(); $SCAN_IN_PROGRESS=0; $SUB_IN_PROGRESS=0; return 1;
  }
  $w_    = 0;
  $wait_ = 0;
  $wait_max = 180;
  $wait_step = 0.2;
  $wait_print = 2;
  $progress = '';
  ### Wait for image to be taken:
  do {
     select (undef, undef, undef, $wait_step);
     $wait_ += $wait_step;
     $w_    += $wait_step;
     ($status,$progress) = &query_SONICC('Scan_Status');
     if ($status) {
        &updateLogBox('Aborting SONICC Scan');
        &send_SONICC('Scan_Abort', $debug);
        #$zscanBtn->configure(-state=>'normal');
        &enable_buttons(); $SCAN_IN_PROGRESS=0; $SUB_IN_PROGRESS=0; return 1;
     }
     if ($w_ > $wait_print) {
        $w_ = 0;
        if (&is_int($progress)) {$progress .= 'line(s) or sweeps(s) done';}
        &updateLogBox($wait_.'s waiting for SONICC instrument completion, '
                      .'progress: '.$progress);
     }
     $mw->update;
  } while ($wait_ < $wait_max && $progress ne 'Scan Complete' && $SCAN_IN_PROGRESS);

  if ($progress ne 'Scan Complete' && $SCAN_IN_PROGRESS) {				#scan timeout
     &updateLogBox('*** Aborting SONICC Instrument on timeout');
     &send_SONICC('Scan_Abort', $debug);
     &enable_buttons(); $SCAN_IN_PROGRESS=0; $SUB_IN_PROGRESS=0; return 1;
  }
  elsif (! $SCAN_IN_PROGRESS) {								#scan aborted
     &enable_buttons(); $SUB_IN_PROGRESS=0; return 1;
  }
  else {
     &updateLogBox('SONICC Image Taken'); return 0;
  }
}

#--------------------------------------------------------------------------
sub command_SONICC_start_continuous() {
  my ($status, $wait_, $waited, $wait_step, $wait_max);
  my ($focusOffset, $progress, $i, $ntaken, $label);

  if ($SUB_IN_PROGRESS) {return;}

  $status = &scan_init();
  if ($status) {return;}

  $SCAN_IN_PROGRESS = 2;							#1=oneShot,2=Cont.,3=zScan
  &disable_buttons();

  if    ($soniccMode == 0) {$focusOffset = $parsSHG[6];}			# $soniccMode=0=SHG
  elsif ($soniccMode == 1) {$focusOffset = $parsTPE[6];}			# $soniccMode=1=TPE-UVF
  else {
     $focusOffset = 0; &updateLogBox('*** Incorrect soniccMode='.$soniccMode);
  }

  &updateLogBox('Starting SONICC continuous image acquisition');
  $status = &send_SONICC('Scan_Cont '.$windowsImageDir, 1);
  if ($status) {
     &updateLogBox('*** Failed to initiate continouos SONICC acquisition');
     $SCAN_IN_PROGRESS = 0;
     $SUB_IN_PROGRESS = 0;
     &enable_buttons();
     return 1;
  }
  if ($focusOffset != 0) {
     $msg = 'Shifting the sample by '.$focusOffset
          . ' um to the SONICC focal plane'
	  . '("+/-" is down/up stream)';
     &updateLogBox($msg);
     $msg = '2Z '.$focusOffset.' 0';
     $status = Pezca::PutStringOldCa($clickPV,$msg);
     if ($status) {&report_CAerr($status,$clickPV);}
     else         {&updateLogBox('Sent to P&C server: '.$msg);}
     $waited = &waitMotor($xyAssyPV,10,1);			     		#waited: value=how-long-waited, 0=timeout
  }

  $wait_step = 0.1;
  $wait_max  = 0.5;
  for ($wait_=0; $wait_<$wait_max; $wait_+=$wait_step) {
     select (undef, undef, undef, $wait_step);
     $mw->update;
  }

  ($status,$progress) = &query_SONICC('Scan_Status');
  if ($status) {
     &updateLogBox('Aborting SONICC continuous image acquisition');
     &send_SONICC('Scan_Abort', $debug);
     #$zscanBtn->configure(-state=>'normal');
     &enable_buttons(); $SCAN_IN_PROGRESS=0; $SUB_IN_PROGRESS=0; return 1;
  }
  elsif ($progress =~ /Already scanning/i) {
     &updateLogBox('!!! '.$progress);
     return 0;
  }
  elsif ($progress =~ /Command not valid in Local mode/) {
     &updateLogBox('*** '.$progress);
     return 1;
  }
  $lastScan_pts = 0;
  $currImage = $lastScan_pts;
  $label = 'continuous';
  do {  ###########################################################################
     ($status,$progress) = &query_SONICC('Scan_Status');                          #
     if ($status) {                                                               #
        &updateLogBox('*** Error='.$status.' reading SONICC status');             #
        $SCAN_IN_PROGRESS = 0;                                                    #
     }                                                                            #
     elsif ($progress =~ /complete/i) {                                           #
        &updateLogBox('!!! SONICC reports that continuous scan is complete');     #
        $SCAN_IN_PROGRESS = 0;                                                    #
     }                                                                            #
     if ($SCAN_IN_PROGRESS == 2) {                                                #
        $ntaken = &read_www_images($lastScan_pts,$label);                         #
        if ($ntaken > 0) {                                                        #
           for ($i=0; $i<$nDisplay; $i++) {&display_image($lastScan_pts,$i);}     #
           $IMAGE_LOADED = 1;                                                     #
           for ($i=0; $i<2; $i++) {$sclRadio[$i]->configure(-state=>'normal');}   #
           &update_calibration_label($soniccMode);                                #
           $mw->focus();                                                          #
	} else {                                                                  #
	   &updateLogBox('!!! read_www_images returned no images');               #
        }                                                                         #
     }                                                                            #
  } while ($SCAN_IN_PROGRESS == 2);  ##############################################

# &abort_scan();
  &sonicc_standby();
  ### Return the sample back to original Z-position:
  if ($focusOffset != 0) {
     $focusOffset = - $focusOffset;
     $msg = 'Shifting the sample back by '.$focusOffset
          . ' um to the SONICC focal plane'
	  . '("+/-" is down/up stream)';
     &updateLogBox($msg);
     $msg = '2Z '.$focusOffset.' 0';
     $status = Pezca::PutStringOldCa($clickPV,$msg);
     if ($status) {&report_CAerr($status,$clickPV);}
     else         {&updateLogBox('Sent to P&C server: '.$msg);}
     $waited = &waitMotor($xyAssyPV,10,1);			     		#waited: value=how-long-waited, 0=timeout
  }
  &enable_buttons();
  &updateLogBox('SONICC Continuous Scan ended');
  $SCAN_IN_PROGRESS = 0;
  $SUB_IN_PROGRESS = 0;
}

#--------------------------------------------------------------------------
sub abort_scan() {
  my ($status);
  if (! $SCAN_IN_PROGRESS) {return;}
  $SCAN_IN_PROGRESS=0;
  &updateLogBox('Aborting SONICC Scan');
  $status  = &send_SONICC('Scan_Abort', $debug);
  $status += &send_SONICC('Scan_Stop', $debug);
  return $status;
}

#--------------------------------------------------------------------------
sub sonicc_standby() {
  my ($status);
  &abort_scan();
  $SCAN_IN_PROGRESS=0;
  &updateLogBox('Switching SONICC to standby');
  $status  = &send_SONICC('SONICC_State Standby', $debug);
  return $status;
}

#--------------------------------------------------------------------------
sub get_sonicc_states() {
  my ($status, $cmd, $rsp, @pars, $par);
  @pars = qw(SONICC Control Resonator);
  foreach $par (@pars) {
     $cmd = $par.'_State Status';
     ($status,$rsp) = &query_SONICC($cmd);
     if ($status == 0) {
        &updateLogBox('--> '.$par.' state is "'.$rsp.'"');
     }
  }
}

#--------------------------------------------------------------------------
sub toggle_step_entry() {
  if (defined $entryStep && Exists($entryStep)) {
     if ($scanPts > 1) {$entryStep->configure(-state=>'normal');}
     else              {$entryStep->configure(-state=>'disabled');}
  }
}

#--------------------------------------------------------------------------
sub load_soniccConf($) {
  my ($type, $status, $candidate, $i, $val);
  if ($SCAN_IN_PROGRESS) {
     &updateLogBox('*** Cannot load config while scan in progress');
     return 1;
  }
  $type = shift(@_);
  if (!defined $type || $type ne 'reload') {$type = 'new';}

  if ($epics_cfg) {
     &updateLogBox('Loading SONICC calibration from EPICS PVs:');
     for ($i=0; $i<@shgPVs; $i++) {
        ($status, $val) = Pezca::GetDouble($shgPVs[$i]);
        if ($status) {&report_only($script, $status, $shgPVs[$i]);}
        else         {$parsSHG[$i] = $val;}
        ($status, $val) = Pezca::GetDouble($tpePVs[$i]);
        if ($status) {&report_only($script, $status, $tpePVs[$i]);}
        else         {$parsTPE[$i] = $val;}
     }
     &updateLogBox('Config loaded:'."\n"
     .'     (omegaOffset xcenter ycenter  xpx2um  ypx2um, axes_rotation focusOffset)'."\n"
     .'SHG: '."(@{parsSHG})\n"
     .'TPE: '."(@{parsTPE})");
     &update_calibration_label($soniccMode);
     return 0;
  }     

### ---
### The file sonnic.conf is an interpreted Perl script and it should look like this
### (here omegaOffset=0 for HiRes camera, omegaOffset=90 for loRes camera):
###       omegaOffset xcenter(px) ycenter(px) xpixel2um ypixel2um axes_rotation focusOffset
###    index:   0        1            2          3          4          5            6
#@parsSHG = ( 0,         299,       277,        2.50,       2.73,      3.42,         -25);
#@parsTPE = ( 0,         299,       277,        2.50,       2.73,      3.42          -75);
### ---

### Find which sonicc.conf to load:
  if ($type eq 'new' || !defined $configFullPath) {
     foreach $candidate (@configlocations) {
        if (-f $candidate.$soniccConf) {	     					#if file exists
           $configFullPath = $candidate.$soniccConf;				#Sergey: changed to global file 2015/07/28
	   last;
	}
     }
  }
  if (defined $configFullPath) {
     &updateLogBox('Loading SONICC calibration from "'.$configFullPath.'":');
### ---
### If the "do" command can read the file but cannot compile it, it returns
### "undef" and sets an error message in $@ .
### If the "do" command cannot read the file, it returns "undef"
### and sets $! to the error.
### Always check $@ first, as compilation could fail in a way that also sets $! .
### If the file is successfully compiled, the "do" command returns the value
### of the last expression evaluated.
#       $status = do $configFullPath;
### ---
     $status = eval `cat $configFullPath`;					#Eval does the same as "do", but it provide more strict syntax checking
     if ($@) {								#'$@' is error message
        &updateLogBox('*** Could not eval '.$configFullPath.': '.$@);
        return 1;
     }
     elsif (!defined $status) {
        &updateLogBox('*** Failed to eval '.$configFullPath);
        return 1;
     }
     else {
        &updateLogBox('Config loaded:'."\n"
        .'     (omegaOffset xcenter ycenter  xpx2um  ypx2um, axes_rotation focusOffset)'."\n"
        .'SHG: '."(@{parsSHG})\n"
        .'TPE: '."(@{parsTPE})");
        &update_calibration_label($soniccMode);
        return 0;
     }
  } else {
     &zenity_error_exit($script.': no calibration file ['.$soniccConf.']'."\n"
                       .'found in any of these locations:'."\n"
                       ."(@{configlocations})");
  }
}

#--------------------------------------------------------------------------
sub prev_image() {
  if ($lastScan_pts < 2) {
     $prevBtn->configure(-state=>'disabled');					#disable image browsing buttons
     $nextBtn->configure(-state=>'disabled');
     return;
  }
  if ($currImage <= 0) {
     $currImage = 0;
     $prevBtn->configure(-state=>'disabled');					#disable image browsing buttons
     $nextBtn->configure(-state=>'normal');
     return;
  }
  $currImage--;
  if ($currImage == 0) {
     $prevBtn->configure(-state=>'disabled');					#disable prev. image browsing button
  }
  if ($currImage < $lastScan_pts-1) {
     $nextBtn->configure(-state=>'normal');					#enable next image browsing button
  }
  &replace_images();
}

#--------------------------------------------------------------------------
sub next_image() {
  if ($lastScan_pts < 2) {
     $prevBtn->configure(-state=>'disabled');					#disable image browsing buttons
     $nextBtn->configure(-state=>'disabled');
     return;
  }
  if ($currImage >= $lastScan_pts-1) {
     $currImage = $lastScan_pts-1;
     $nextBtn->configure(-state=>'disabled');					#disable image browsing buttons
     $prevBtn->configure(-state=>'normal');
     return;
  }
  $currImage++;
  if ($currImage == $lastScan_pts-1) {
     $nextBtn->configure(-state=>'disabled');					#disable next image browsing button
  }
  if ($currImage > 0) {
     $prevBtn->configure(-state=>'normal');					#enable prev. image browsing button
  }
  &replace_images();
}

#--------------------------------------------------------------------------
sub replace_images() {
  my ($i, $label);
  $label = $images[$currImage]{label};
  for ($i=0; $i<$nDisplay; $i++) {
     &display_image($currImage,$i);
     $imLabelNm[$i]->configure(-text=>$label);
  }
}

#--------------------------------------------------------------------------
sub toggle_enhance() {									#switch  between enhanced and normal presentations
  my ($i);
# if ($SUB_IN_PROGRESS || !$IMAGE_LOADED) {return;}
# $SUB_IN_PROGRESS = 1;

  &updateLogBox('Switching to presentation="'.$enhanceLabels[$enhance].'"');
  for ($i=0; $i<$nDisplay; $i++) {
     &display_image($currImage,$i);
  }
# $SUB_IN_PROGRESS=0;
}

#--------------------------------------------------------------------------
#sub read_www_images($imstep,$label) {								#returns the number of taken images
sub read_www_images($$) {
  my ($imstep, $label, $imindx, $url, $response, $ntaken, @XYZO);
  my ($width, $height, $new_width, $new_height, $msg, $i);
  my ($status, $imi, $jmi, $mi, $ima, $jma, $ma, $pix);

  $imstep = shift(@_);
  $label  = shift(@_);

### Reading images from the web
  if ($SCAN_IN_PROGRESS != 2) {&updateLogBox('Reading images from WWW');}					#if not continuous scan

  $ntaken = 0;
  for ($imindx=0; $imindx<$nDisplay; $imindx++) {
     if (defined $images[$imstep]{'www'.$imindx})     {undef $images[$imstep]{'www'.$imindx};}			#LWP image object
     if (defined $images[$imstep]{'magic'.$imindx})   {undef $images[$imstep]{'magic'.$imindx};}		#ImageMagick image object
     if (defined $images[$imstep]{'enhance'.$imindx}) {undef $images[$imstep]{'enhance'.$imindx};}		#ImageMagick image object
     if (defined $images[$imstep]{'pixels'.$imindx})  {undef $images[$imstep]{'pixels'.$imindx};}
     if (defined $images[$imstep]{'ncols'.$imindx})   {undef $images[$imstep]{'ncols'.$imindx};}
     if (defined $images[$imstep]{'nlines'.$imindx})  {undef $images[$imstep]{'nlines'.$imindx};}
     if (defined $images[$imstep]{'min'.$imindx})     {undef $images[$imstep]{'min'.$imindx};}
     if (defined $images[$imstep]{'max'.$imindx})     {undef $images[$imstep]{'max'.$imindx};}
     if ($SCAN_IN_PROGRESS != 2) {
### Examples (single shot): 
### http://192.168.1.50:4000/ChannelALast.jpg                           #SHG mode, SHG
### http://192.168.1.50:4000/ChannelDLast.jpg                           #SHG mode, TPEF
### http://192.168.1.50:4000/ChannelBLast.jpg                           #SHG mode, BF
### http://192.168.1.50:4000/ChannelCLast.jpg                           #TP mode, TP-UVF
### http://192.168.1.50:4000/ChannelDLast.jpg                           #TP mode, FLUO
### http://192.168.1.50:4000/ChannelBLast.jpg                           #TP mode, BF
        $url = 'http://'.$ipaddr.':4000/Channel'                        #single shot
             . $channels[$soniccMode][$imindx].'Last.'.lc($imgExt);
     } else {
### Examples (continuous): 
### http://192.168.1.50:4000/ChannelA.jpg                               #SHG mode, SHG
### http://192.168.1.50:4000/ChannelD.jpg                               #SHG mode, TPEF
### http://192.168.1.50:4000/ChannelB.jpg                               #SHG mode, BF
### http://192.168.1.50:4000/ChannelC.jpg                               #TP mode, TP-UVF
### http://192.168.1.50:4000/ChannelD.jpg                               #TP mode, FLUO
### http://192.168.1.50:4000/ChannelB.jpg                               #TP mode, BF
        $url = 'http://'.$ipaddr.':4000/Channel'                                                                #continuous
             . $channels[$soniccMode][$imindx].'.'.lc($imgExt);
     }
     $msg = 'Sending request for image='.$header[$soniccMode][$imindx].' to URL='.$url;
     &updateLogBox($msg);
     $response = $useragent->get($url);					#read image from the www
     if (! $response->is_success()) {
        $msg = '*** Failed image request for mode='.$soniccMode
             . ', channel='.$header[$soniccMode][$imindx].', URL='.$url."\n"
             . '*** WWW response: '.$response->status_line();
	&updateLogBox($msg);
	next;
     }
     elsif ($response->content_type() !~ /image/) {
        $msg = '*** Incorrect image content type='.$response->content_type()
             . ', channel='.$header[$soniccMode][$imindx].', URL='.$url;
	&updateLogBox($msg);
	next;
     }
     elsif ($response->content_length() < 1000) {
        $msg = '*** Incorrect image content length='.$response->content_length()
             . ', channel='.$header[$soniccMode][$imindx].', URL='.$url;
	&updateLogBox($msg);
	next;
     }
     else {
        $images[$imstep]{'www'.$imindx} = $response->content;
	$ntaken++;
     }
     if (Exists($mw)) {$mw->update;}
  }
  if ($ntaken == 0) {return $ntaken;}

  ### Save sample coordinates & SONICC mode
  $images[$imstep]{mode}  = $soniccMode;
  $images[$imstep]{label} = $label;
  &retrieve_gonio_positions(\@XYZO, 0);										#this is in mm
  ($images[$imstep]{x},$images[$imstep]{y},$images[$imstep]{h},$images[$imstep]{omega}) = @XYZO;

  if ($SCAN_IN_PROGRESS != 2) {&updateLogBox('Converting images to ImageMagic objects');}			#if not continuous scan
### All PerlMagick methods return an undefined string context upon success.
### If any problems occur, the error is returned as a string with an embedded
### numeric status code. A status code less than 400 is a warning. This means
### that the operation did not complete but was recoverable to some degree.
### A numeric code greater or equal to 400 is an error and indicates the
### operation failed completely.
  $ntaken = 0;
  for ($imindx=0; $imindx<$nDisplay; $imindx++) {
     if (defined $images[$imstep]{'www'.$imindx}) {
        $images[$imstep]{'magic'.$imindx} = Image::Magick->new(magick=>$imgType);				#allocate image structure
        $status = $images[$imstep]{'magic'.$imindx}->BlobToImage($images[$imstep]{'www'.$imindx});
        if ($status) {
           $msg = '*** Error=['.$status.'] converting web blob to image for'
                . ' mode='.$soniccMode.', channel='.$header[$soniccMode][$imindx];
	   &updateLogBox($msg);
	   next;
        }
        if ($imgType =~ /^JPEG$/) {$images[$imstep]{'magic'.$imindx}->Set(quality=>85);}
        $width  = $images[$imstep]{'magic'.$imindx}->Get('columns');
        $height = $images[$imstep]{'magic'.$imindx}->Get('rows');
        if (!defined $width || !defined $height) {
           $msg = '*** Error getting image sizes for mode='
                . $soniccMode.', channel='.$header[$soniccMode][$imindx];
	   &updateLogBox($msg);
	   next;
	}
        $images[$imstep]{'ncols'.$imindx}  = $width;
        $images[$imstep]{'nlines'.$imindx} = $height;
	$ntaken++;
        if (Exists($mw)) {$mw->update;}
     }
  }
  if ($ntaken == 0) {return $ntaken;}

### Postprocessing images
  if ($SCAN_IN_PROGRESS != 2) {
     if ($grayscale)                    {&updateLogBox('Converting images to grayscale');}
     if ($magnify != 1 && $magnify > 0) {&updateLogBox('Rescaling images by factor='.$magnify);}		#if not continuous scan
     if ($rotate)                       {&updateLogBox('Rotating images by angle='.$rotate.' degr.');}
     if ($flipv)                        {&updateLogBox('Flipping images vertically');}
     if ($fliph)                        {&updateLogBox('Flipping images horizontally');}
     if ($enhance)                      {&updateLogBox('Enhancing images contrast');}
  }
  $ntaken = 0;
  for ($imindx=0; $imindx<$nDisplay; $imindx++) {
     if (defined $images[$imstep]{'magic'.$imindx}) {
	### Converting images to grayscale
	if ($grayscale) {
           $status = $images[$imstep]{'magic'.$imindx}->Set(colorspace=>'Gray');				#convert to grayscale (works with older and newer versions of Magick)
           if ($status) {
              $msg = '!!! Error=['.$status.'] converting to Grayscale image for mode='
                   . $soniccMode.', channel='.$header[$soniccMode][$imindx];
	      &updateLogBox($msg);
           }
           if (Exists($mw)) {$mw->update;}
	}
        ### Scaling (magnifying/demagnifying) images
        if ($magnify != 1 && $magnify > 0) {
           if ($magnify == 2) {
              $status = $images[$imstep]{'magic'.$imindx}->Magnify();
           }
           elsif ($magnify == 0.5) {
              $status = $images[$imstep]{'magic'.$imindx}->Minify();
           }
           else  {
              $new_width  = int($width*$magnify+0.5);
              $new_height = int($height*$magnify+0.5);
#             $status = $images[$imstep]{'magic'.$imindx}->Resize(geometry=>$width.'x'.$height, width=>$new_width, height=>$new_height, filter=>'Lanczos');
#             $status = $images[$imstep]{'magic'.$imindx}->Resize(geometry=>$width.'x'.$height, width=>$new_width, height=>$new_height, filter=>'Point');
              $status = $images[$imstep]{'magic'.$imindx}->Resize(geometry=>$width.'x'.$height, width=>$new_width, height=>$new_height, filter=>'Box');
           }
           if ($status) {
              $msg = '*** Error=['.$status.'] rescaling image for mode='
                   . $soniccMode.', channel='.$header[$soniccMode][$imindx];
	      &updateLogBox($msg);
	      next;
	   }
           $width  = $images[$imstep]{'magic'.$imindx}->Get('columns');
           $height = $images[$imstep]{'magic'.$imindx}->Get('rows');
           if (!defined $width || !defined $height) {
              $msg = '*** Error getting rescaled image sizes for mode='
                   . $soniccMode.', channel='.$header[$soniccMode][$imindx];
	      &updateLogBox($msg);
	      next;
	   }
           $images[$imstep]{'ncols'.$imindx}  = $width;
           $images[$imstep]{'nlines'.$imindx} = $height;
           if (Exists($mw)) {$mw->update;}
	}
        ### Rotating images
	if ($rotate) {
           $status = $images[$imstep]{'magic'.$imindx}->Rotate(degrees=>$rotate);
           if ($status) {
              $msg = '*** Error=['.$status.'] rotating image for mode='
                   . $soniccMode.', channel='.$header[$soniccMode][$imindx];
	      &updateLogBox($msg);
	      next;
	   }
           $width  = $images[$imstep]{'magic'.$imindx}->Get('columns');
           $height = $images[$imstep]{'magic'.$imindx}->Get('rows');
           if (!defined $width || !defined $height) {
              $msg = '*** Error getting rotated image sizes for mode='
                   . $soniccMode.', channel='.$header[$soniccMode][$imindx];
	      &updateLogBox($msg);
	      next;
	   }
           $images[$imstep]{'ncols'.$imindx}  = $width;
           $images[$imstep]{'nlines'.$imindx} = $height;
           if (Exists($mw)) {$mw->update;}
	}
        ### Flipping images horizontally
        if ($fliph) {
           $status = $images[$imstep]{'magic'.$imindx}->Flop();
           if ($status) {
              $msg = '*** Error=['.$status.'] flipping horizontally image for mode='
                   . $soniccMode.', channel='.$header[$soniccMode][$imindx];
	      &updateLogBox($msg);
	      next;
	   }
           if (Exists($mw)) {$mw->update;}
	}
        ### Flipping image vertically
        if ($flipv) {
           $status = $images[$imstep]{'magic'.$imindx}->Flip();
           if ($status) {
              $msg = '*** Error=['.$status.'] flipping vertically image for mode='
                   . $soniccMode.', channel='.$header[$soniccMode][$imindx];
	      &updateLogBox($msg);
	      next;
	   }
           if (Exists($mw)) {$mw->update;}
	}
	### Get intensities:
        $imi = 0;  $jmi = 0;  $mi = 16777215+1;									#maximum 24-bit number (16,777,215) + 1
        $ima = 0;  $jma = 0;  $ma = -1;
        for ($i=0; $i<$height; $i++) {
           ### Read a line from the image:
	   $images[$imstep]{'pixels'.$imindx}[$i] =
            [$images[$imstep]{'magic'.$imindx}->GetPixels
                                           (map=>'I',y=>$i,height=>1,x=>0,width=>$width,normalize=>0)];		#[] is important (array reference)
           for ($j=0; $j<$width; $j++) {
 	      $pix = $images[$imstep]{'pixels'.$imindx}[$i][$j];
	      if    ($pix < $mi) {$mi=$pix; $imi=$i; $jmi=$j;}
	      elsif ($pix > $ma) {$ma=$pix; $ima=$i; $jma=$j;}
           }
	}
	$msg = 'Image for mode='.$soniccMode.', channel='.$header[$soniccMode][$imindx]."\n"
             . '     max='.$ma.' at ('.$ima.','.$jma.')'."\n"
             . '     min='.$mi.' at ('.$imi.','.$jmi.')';
	&updateLogBox($msg);
	$images[$imstep]{'min'.$imindx} = $mi;
	$images[$imstep]{'max'.$imindx} = $ma;
	### Produce enhanced-mode image if requested:
	if ($enhance) {
	   $images[$imstep]{'enhance'.$imindx} = $images[$imstep]{'magic'.$imindx}->Clone;			#clone image
           $status = $images[$imstep]{'enhance'.$imindx}->
                                SigmoidalContrast(contrast=>10,'mid-point'=>0.5*($mi+$ma));			#works
           if ($status) {
              $msg = '*** Error=['.$status.'] enhancing image for mode='
                   . $soniccMode.', channel='.$header[$soniccMode][$imindx];
	      &updateLogBox($msg);
	   }
	}
	$ntaken++;
        if (Exists($mw)) {$mw->update;}
     }
  }
  return $ntaken;
}

#--------------------------------------------------------------------------
# sub display_image($imstep,$imindx) {
sub display_image($$) {									#display image on the GUI screen
  my ($imstep, $imindx, $imgSclBox, $img64, $photo, $status, $mi, $ma, $msg);

  $imstep = shift(@_);
  $imindx = shift(@_);

  if (defined $images[$imstep]{'magic'.$imindx}) {
     if ($enhance) {
        if (!defined $images[$imstep]{'enhance'.$imindx}) {
           $images[$imstep]{'enhance'.$imindx} = $images[$imstep]{'magic'.$imindx}->Clone;		#clone image
           $mi = $images[$imstep]{'min'.$imindx};
           $ma = $images[$imstep]{'max'.$imindx};
           $status = $images[$imstep]{'enhance'.$imindx}->SigmoidalContrast(contrast=>10,'mid-point'=>0.5*($mi+$ma));
           if ($status) {
              $msg = '*** Error=['.$status.'] enhancing image for mode='
                   . $soniccMode.', channel='.$header[$soniccMode][$imindx];
              &updateLogBox($msg);
           }
        }
	$imgSclBox = $images[$imstep]{'enhance'.$imindx}->Clone;					#clone image
     } else {
	$imgSclBox = $images[$imstep]{'magic'.$imindx}->Clone;						#clone image
     }
     &drawScaleBox(\$imgSclBox,
                   $images[$imstep]{mode},
		   $images[$imstep]{'ncols'.$imindx},							#width
                   $images[$imstep]{'nlines'.$imindx});							#height
     $img64 = encode_base64($imgSclBox->ImageToBlob());
     undef $imgSclBox;
     $photo = $imDisplay[$imindx]->Photo(-format=>$imgType,-data=>$img64);
     $imDisplay[$imindx]->configure(-image=>$photo,-width=>0,-height=>0);
     $mw->focus();
  } else {
     $msg = '!!! image for step='.$imstep.' imdex='.$imindx.' is undefined';
     &updateLogBox($msg);
  }
}

#--------------------------------------------------------------------------
#sub make_directory($directory,\$msg) {
sub make_directory($\$) {
  my ($directory, $msg_ref, $mkErr, $dir, $str, $diag);

  $directory = shift(@_);
  $msg_ref   = shift(@_);

  $$msg_ref = '';
  if (! -e $directory) {
     mkpath($directory, {error => \$mkErr});
     if (@$mkErr) {
        foreach $diag (@$mkErr) {
           ($dir, $str) = %$diag;
           if ($dir eq '') {$$msg_ref .= '*** General error: "'.$str.'"'."\n";}
           else            {$$msg_ref .= '*** Problem creating "'.$dir.'": '.$str."\n";}
        }
        $$msg_ref .= '*** Directory "'.$directory.'" does not exist and could not be created.';
	return 1;
     }
  }
  return 0;
}

#--------------------------------------------------------------------------
#sub save_images4user() {
sub save_images4user() {
  my ($imstep, $imindx, $status, $file, $msg, $soniccMode, $label);
  my (@ct, $ss, $mm, $hh, $DD, $MM, $YY);

  if (!defined $userSaveDir) {
      &updateLogBox('*** Cannot save images because userSaveDir is undefined');
      return;
  }
  if (! -e $userSaveDir) {
     $status = &make_directory($userSaveDir,\$msg);
     if ($status) {
	$msg .= "\n".'*** Cannot save images';
	&updateLogBox($msg);
	return;
     }
  }

  @ct = localtime(time);
  ($ss,$mm,$hh,$DD,$MM,$YY) = ($ct[0],$ct[1],$ct[2],$ct[3],$ct[4]+1,$ct[5]+1900);
  for ($imstep=0; $imstep<$lastScan_pts; $imstep++) {
     $soniccMode = $images[$imstep]{mode};
     $label      = $images[$imstep]{label};
     for ($imindx=0; $imindx<$nDisplay; $imindx++) {
        $file = sprintf '%04d.%02d.%02d-%02d.%02d.%02d_%s', $YY,$MM,$DD,$hh,$mm,$ss, $header[$soniccMode][$imindx];
	if (length($label) > 0) {$file .= '_'.$label;}
        $file .= '.'.lc($imgExt);
	if (-f $userSaveDir.'/'.$file) {
	   &updateLogBox('*** File "'.$file.'/'.$userSaveDir.'" already exists. Skipping...');
	} else {
           if ($enhance && defined $images[$imstep]{'enhance'.$imindx}) {
              $status = $images[$imstep]{'enhance'.$imindx}->Write($userSaveDir.'/'.$file);
           } else {
              $status = $images[$imstep]{'magic'.$imindx}->Write($userSaveDir.'/'.$file);
           }
           if ($status) {&updateLogBox('*** Failed to save "'.$file.'" to "'.$userSaveDir.'"');}
	   else         {&updateLogBox('Saved "'.$file.'" to "'.$userSaveDir.'"');}
	}
     }
  }
  return;
}

#--------------------------------------------------------------------------
sub goodbye() {
  require Tk::ErrorDialog;
  if (defined $BYE_IN_PROGRESS && $BYE_IN_PROGRESS) {return;}
  $BYE_IN_PROGRESS = 1;
  if ($SCAN_IN_PROGRESS) {&abort_scan();}
  &updateLogBox('Goodbye!'."\n");
  &clearAllCaMonitors();
# $mw->destroy;
# Tk::exit;
  exit 0;
}
