#!/usr/bin/perl -W
#--------------------------------------------------------------
#
# This script is basically a proxy for the SONICC EPICS server. 
# While the SONICC server runs directly on a Windows PC controlling
# SONIC, this one can run on Keithley or any other computer with access
# to video network though which the SONICC computer is accessible. 
#
# Once the command is received via the CMD PV, the server performs the 
# following operations:
# 1. Brings SONICC in and turns off any lights as needed.
# 2. Shifts sample upstream or downstream according to the SONICC focus offset.
# 3. Sends series of command to SONICC in order to take image in the requested mode.
# 4. Reports to JBluIce of any errors if present.
# 5. Retuns status to a caller telling if the command was success on failure.
#    In case of success the caller can read respective image from SONICC web server
#    and transform it according to calibrations available in EPICS.
# 6. Shifts focus back (optionally?)
# 7. Returns SONICC back and restores lights intensity.
# 
#
#                       Author: Sergey Stepanov
#
# Version-1.0:  2012/03/18	initial implementation re-using code from SONICCscan.pl
#--------------------------------------------------------------
# Subs:
#                         &updateLogBox($msg,[$tag]);
#                         &report_CAerr($status,$pv);
#               $status = &get_crystalPositions(\@positions);
#       $soniccPosition = &get_soniccPosition();
#               $status = &get_lightsIntensity(\@intensities);
#               $status = &set_lightsIntensity('ON'|'OFF',[\@intensities]);
#               $status = &soniccMove('IN'|'OUT');
#       $soniccPosition = &soniccWaitForIN();
#               $status = &load_sonicc_calibration();
#               $status = &send_SONICC(cmd,[$debug]);
#       ($status, $rsp) = &query_SONICC($cmd,[$debug]);
#               $status = &switch_sonicc_RL('Remote'|'Local');
#               $status = &request_SONICC_to_load_scan_config()
#               $status = &sonicc_scan_init();
#               $status = &do_sonicc_scan([$stayIN]);
#               $status = &command_SONICC_take_image();
#               $status = &sonicc_abort();
#               $status = &sonicc_standby();
#                         &goodbye();
#--------------------------------------------------------------
  use strict;
  use warnings;
# use warnings::unused;
  use Pezca;

  our ($NONINTERACTIVE, $USE_ZENITY_FOR_EXIT, $ZENITY_VIA_BLUICE);
  our ($IGNORE_ERRORS, $beamline, $HAVE_XYZ, $nRetry);
  my ($PEZCA, $script, $debug);
  my (@parsSHG, @parsTPE, @shgPVs, @tpePVs, @parsPVs, $pv);
  my (@modeSwitchers, @modeLabels, $soniccMode, $nModes);
  my ($windowsImageDir, $windowsConfigPath, @modeConfigs);
  my ($serverCmdPV, $soniccCmdPV, $soniccRspPV, $clickPV);
  my ($crystalAssyPV, @crystalAxesPV, @crystalAct, $nPVs);
  my ($soniccAssy_PV, $soniccAxis_PV, $sonicc_tolerance);
  my ($sonicc_in_PV, $sonicc_out_PV, $sonicc_in_pos, $sonicc_out_pos);
  my (@lightsPVs, @lightsVal);
  my ($msg, $status, $ipaddr, @beamlinelist, @sonicclist, $i);
  my ($sleep_interval, $connection_lost, $bluice, @allowed_hosts);
  my ($gotcmd, @cmdlist, @validlist, $invalid, @words, @cmd, $w, $stayIN);
  my ($stamp, $string, $start_time, $end_time);

  if (!defined $ENV{PEZCA}) {
     die 'Please define the PEZCA environment. It must point to the PERL scripts repository';
  }
  $PEZCA=$ENV{PEZCA};
  $PEZCA =~ s/\\/\//g;
  if ($PEZCA !~ /\/$/) {$PEZCA .='/';}

# $beamline = '23o';
  do $PEZCA.'Common/motor_subs.pl';
  do $PEZCA.'Common/common_subs.pl';
  if (defined $beamline) {do $PEZCA.'/Common/beamline_'.$beamline.'.pl';}
  else                   {do $PEZCA.'Common/beamline.pl';}

# select(STDERR);  $|=1;							#Flush output after each print
# select(STDOUT);  $|=1;							#Flush output after each print
  *STDERR->autoflush();
  *STDOUT->autoflush();

  $script = &makeScriptName();

  $NONINTERACTIVE      = 1;
  $IGNORE_ERRORS       = 1;
  $debug               = 0;							#set to 1 in order to see commands sent to SONICC
  $bluice              = 1;
# $USE_ZENITY_FOR_EXIT = 1;							#move it down so that multiple instances do not display BluIce popups
# $ZENITY_VIA_BLUICE   = $bluice;						#move it down so that multiple instances do not display BluIce popups

  $USE_ZENITY_FOR_EXIT = 0;							#no popups, silent exit on multiple instances
  @allowed_hosts = qw(bl1keithley bl2keithley bl3keithley);
  &check_Nrunning_and_host_allowed($script,\@allowed_hosts);
        
  $nRetry  = 5;
  &setEpicsRetry();
  Pezca::AutoErrorMessageOff();

######################################################################################
  @beamlinelist  = ('23i:', '23o:', '23b:', '23d:');
  @sonicclist    = ('23o:');
  if (! grep(/^$beamline:?$/, @sonicclist)) {
     &zenity_error_exit($script.': SONICC is not present at beamline='.$beamline);
  }
######################################################################################
  $USE_ZENITY_FOR_EXIT = 1;
  $ZENITY_VIA_BLUICE   = $bluice;						#from not display popups via BluIce

  $ipaddr       = '192.168.1.50';						# SONICC CA server IP
  if (!defined $ENV{EPICS_CA_ADDR_LIST}) {$ENV{EPICS_CA_ADDR_LIST} = $ipaddr;}
  else    {if ($ENV{EPICS_CA_ADDR_LIST} !~ /$ipaddr/i) {$ENV{EPICS_CA_ADDR_LIST} .= ' '.$ipaddr;}}

  $serverCmdPV = $beamline.':SONICC:serverCMD';                                 #this is CMD on Keithley server (server CMD)
  $soniccCmdPV = $beamline.':SONICC:cmd';                                       #this is CMD on SONICC computer (device CMD)
  $soniccRspPV = $beamline.':SONICC:rsp';                                       #this is RSP on SONICC computer (device RSP)
  $clickPV     = $beamline.':bi:pointclick';

  $windowsImageDir   = 'X:/ArgonneData/';					#do not forget slash at the end!
  $windowsConfigPath = $windowsImageDir.'ConfigurationFiles/';			#do not forget slash at the end!
  $windowsImageDir   =~ s/\//\\/g;						#change "/" to "\"
  $windowsConfigPath =~ s/\//\\/g;						#change "/" to "\"

  @modeSwitchers = ('SONICC', 'TPE-UVF');					#cmd to send to SONICC for switching modes
  $nModes        = @modeSwitchers;
  @modeLabels    = ('SHG TPEF BF', 'TP-UVF FLUO BF');
  @modeConfigs   = ('SONICC_SDImageConfig_SHG.txt', 'SONICC_SDImageConfig_TPEUVF.txt');
  for ($i=0; $i<$nModes; $i++) {$modeConfigs[$i] = $windowsConfigPath.$modeConfigs[$i];}

  $soniccMode   = 0;								#0=SHG/TPEF/BF, 1=TPE-UVF/FLUO/BF

######################################################################################
### PVs:
  if (defined $HAVE_XYZ) {@crystalAxesPV = ('GO:CX', 'GO:CY', 'GO:CZ');}
  else                   {@crystalAxesPV = ('GO:CX', 'GO:CY', 'GO:SH');}
  $nPVs = @crystalAxesPV;
  for ($i=0; $i<$nPVs; $i++) {$crystalAxesPV[$i] = $beamline.':'.$crystalAxesPV[$i];}
  $crystalAssyPV = $beamline.':GO:Cr';
  $status = &setAssyMonitors ($crystalAssyPV);
  for ($i=0; $i<$nPVs; $i++) {$status = &setAxisMonitors($crystalAxesPV[$i]);}

  $soniccAssy_PV    = $beamline.':SNC:Ps';
  $soniccAxis_PV    = $beamline.':SNC:P';
  $sonicc_in_PV     = $soniccAxis_PV.':RqsPosMem1';
  $sonicc_out_PV    = $soniccAxis_PV.':RqsPosMem2';
  $sonicc_tolerance = 0.2; 							#max |act-rqs| in mm
  $status = &setAssyMonitors ($soniccAssy_PV);
  $status = &setAxisMonitors ($soniccAxis_PV);
  ($status, $sonicc_in_pos)  = Pezca::GetDouble($sonicc_in_PV);
  if ($status) {&report_only($script,$status,$sonicc_in_PV);  $sonicc_in_pos=0;}
  ($status, $sonicc_out_pos) = Pezca::GetDouble($sonicc_out_PV);
  if ($status) {&report_only($script,$status,$sonicc_out_PV); $sonicc_out_pos=0;}

###            backlight ringlight HiRes  LowRes
  @lightsPVs = qw(DAC1_8  DAC1_4  DAC1_6  DAC1_5);
  for ($i=0; $i<@lightsPVs; $i++) {
     $lightsPVs[$i] = $beamline.':'.$lightsPVs[$i];
     $status = &addCaMonitor($lightsPVs[$i]);
  }
  @parsPVs = qw(omegaOffset ycenter_px xcenter_px ypixel2um xpixel2um axes_rotation focusOffset);
  @shgPVs = ();
  @tpePVs = ();
  for ($i=0; $i<@parsPVs; $i++) {
     $pv = $beamline.':SONICC:'.$parsPVs[$i].':';
     $status = &addCaMonitor($pv.'SHG');
     if ($status) {last;}
     push(@shgPVs,$pv.'SHG');
     $status = &addCaMonitor($pv.'TPE');
     if ($status) {last;}
     push(@tpePVs,$pv.'TPE');
  }
  if ($status) {&zenity_error_exit($script.': failed to connect to SONICC calibration PVs');}
### Load user-specific @parsSHG and @parsTPE from EPICS:
  $status = &load_sonicc_calibration();
  if ($status) {&zenity_error_exit($script.': failed to load SONICC calibration from EPICS');}

### Main loop -- wait for command:
  $sleep_interval  = 0.05;
  $connection_lost = 0;
  @cmdlist = qw(takeSHG takeTPE withdraw stayIN);
  @validlist = ('accepted', 'finished,', 'status=1', 'status=0');
  $status = &PezcaPutStringWF($serverCmdPV,' ');
  &report_CAerr($status,$serverCmdPV);
CMD:
  while(1) { ##############################################################################
### Wait for command:                                                                     #
     $gotcmd = 0;						#no command yet           #
     $stamp = scalar localtime;                                                           #
     &updateLogBox($stamp.'  Waiting for SONICC server command...');                      #
# --------------------------------------------------------	  		          #
     do {                                                                                 #
        select(undef,undef,undef,$sleep_interval);	         #short sleep             #
        ($status,$string) = &PezcaGetStringWF($serverCmdPV);                              #
        if ($status) {                                           #PV read failed          #
           &report_CAerr($status,$serverCmdPV);                                           #
           $gotcmd = 0;                                                                   #
      	   $string = ' ';                                                                 #
      	   if ($connection_lost == 0) {                          #1st loss of connection  #
              $stamp = scalar localtime;                                                  #
              &updateLogBox($stamp.' PV connection lost!','error');                       #
	      $connection_lost = 1;                                                       #
	   }                                                                              #
           sleep 5;                                       	 #wait before trying again#
        }                                                                                 #
        else {                                                   #PV read OK              #
	   if ($connection_lost) {                               #connection restored     #
	      $connection_lost = 0;                                                       #
              $stamp = scalar localtime;                                                  #
              &updateLogBox($stamp.' PV connection restored. Restoring CA-monitors.');    #
	      &resetAllCaMonitors('list');                                                #
              $stamp = scalar localtime;                                                  #
              &updateLogBox($stamp.' CA monitors are reconnected.');                      #
	   }                                                                              #
           $string =~ s/\s+/ /g;                                 #compress spaces         #
           $string =~ s/^\s+//g;                                 #remove start spaces     #
           $string =~ s/\s+$//g;                                 #remove end spaces       #
 	   if (length($string) > 0 || $string =~ /^\s+$/) {                               #
              @words = split(/\s+/,$string);                                              #
              @cmd = ();                                                                  #
              $invalid = 0;                                                               #
              foreach $w (@words) {                                                       #
                 if    (grep(/^$w$/i,@cmdlist))     {push(@cmd,$w);}                      #
                 elsif (! grep(/^$w$/i,@validlist)) {$invalid++;}                         #
              }                                                                           #
              if ($invalid > 0) {                                                         #
                 $stamp = scalar localtime;                                               #
                 &updateLogBox($stamp.' Incorrect keywords in CMD=['.$string.']'."\n"     #
                                     .'Valid commands: '."[@cmdlist]",'warn');            #
                 if (@cmd == 0) {                                                         #
                    $status = &PezcaPutStringWF($serverCmdPV,' ');                        #
                    &report_CAerr($status,$serverCmdPV);                                  #
                    next CMD;                                                             #
                 }                                                                        #
              }                                                                           #
              if (@cmd > 0) {                                                             #
                 $gotcmd = 1;                                 	 #got valid command       #
                 $status = &PezcaPutStringWF($serverCmdPV,'accepted');                    #
                 &report_CAerr($status,$serverCmdPV);                                     #
              }                                                                           #
           }                                                                              #
        }                                                                                 #
     } while ($gotcmd != 1);                                                              #
# --------------------------------------------------------			          #
     $stamp = scalar localtime;                                                           #
     &updateLogBox($stamp.'  ...Got command: "'.$string.'"');                             #
     $start_time = time();                                                                #
     $soniccMode = -1;                                                                    #
     $stayIN     = 0;                                                                     #
     foreach $w (@words) {                                                                #
        if    ($w =~ /^takeSHG$/i)  {$soniccMode = 0;}                                    #
        elsif ($w =~ /^takeTPE$/i)  {$soniccMode = 1;}                                    #
        elsif ($w =~ /^withdraw$/i) {$stayIN     = 0;}                                    #
        elsif ($w =~ /^stayIN$/i)   {$stayIN     = 1;}                                    #
     }                                                                                    #
     if    ($soniccMode >= 0) {$status = &do_sonicc_scan($stayIN);}                       #
     elsif (! $stayIN)        {$status = &soniccMove('OUT');}                             #
     else                     {$status = 999;}                          #unexpected       #
     $end_time = time();                                                                  #
     $stamp = scalar localtime;                                                           #
     $msg = sprintf ' Elapsed time=%.1f seconds',($end_time-$start_time);                 #
     &updateLogBox($stamp.$msg);                                                          #
     $status = &PezcaPutStringWF($serverCmdPV,'finished, status='.$status);               #
     &report_CAerr($status,$serverCmdPV);                                                 #
                                                                                          #
  } #######################################################################################


  &goodbye;

#-------------------------------------------#sub updateLogBox($msg,[$tag]) {
sub updateLogBox($;$) {
  my ($msg, $tag);
  $msg = shift(@_);
  $tag = shift(@_);
  print STDOUT $msg."\n";
  if ($bluice) {
     if (!defined $tag) {				#auto-detect tag, if not specified
        ### Possible tags: note,warn,error,popup_note,popup_error,direct
        if    ($msg =~ /^\*\*\*/ ||
               $msg =~ / error/  ||
               $msg =~ /Error/)      {$tag = 'error'; $tag =~ s/^\*\*\*//;}
        elsif ($msg =~ /^\!\!\!/ ||
               $msg =~/Warn/)        {$tag = 'warn'; $tag =~ s/^\!\!\!//;}
        else                         {$tag = 'note';}
     }
     &tellBluIce($msg,$tag);
  }
}

#--------------------------------------------------------------------------
#sub report_CAerr($status,$pv) {
sub report_CAerr($$) {
  my ($status, $pv, $statusDescr);
  $status  = shift(@_);
  $pv      = shift(@_);
  if ($status) {
     $statusDescr  = &ezcaStatusCode($status);
     &updateLogBox('*** EZCA error='.$status.' ['.$statusDescr.'] PV=['.$pv.']');
  }
}

#--------------------------------------------------------------------------
#sub get_crystalPositions(\@positions) {                                #returns $status
sub get_crystalPositions(\@) {
  my ($status, $i, $msg, $pos_ref);

  $pos_ref  = shift(@_);

  &updateLogBox('Retrieving gonio crystal positions (mm):');
  for ($i=0; $i<$nPVs; $i++) {
#    @$pos_ref[$i] = &getAxisAct ($axis,\$status); 
     ($status, @$pos_ref[$i]) = Pezca::GetDouble($crystalAxesPV[$i].':ActPos');
     if ($status) {
        &report_CAerr($status,$crystalAxesPV[$i].':ActPos');
	return $status;
     }
     if ($debug) {
        $msg = sprintf '%s=%.3f', $crystalAxesPV[$i].':ActPos', @$pos_ref[$i];
        &updateLogBox($msg);
     }
  }
  return 0;
}

#--------------------------------------------------------------------------
sub get_soniccPosition() {                                                   #returns $soniccPosition (text tag)
  my ($soniccPosition, $inPos, $runPrg, $actPos, $rqsPos, $ampEna);
  my ($direction, $percent, $status);

  $inPos = &getInPos ($soniccAssy_PV,\$status);
  if ($status) {
     &updateLogBox('*** Error reading SONICC inpos.');
     $soniccPosition = 'Unknown';
     return $soniccPosition;
  }
  $runPrg = &getRunPrg ($soniccAssy_PV,\$status);
  if ($status) {
     &updateLogBox('*** Error reading SONICC runprg.');
     $soniccPosition = 'Unknown';
     return $soniccPosition;
  }
  $actPos = &getAxisAct ($soniccAxis_PV,\$status);
  if ($status) {
     &updateLogBox('*** Error reading SONICC actpos.');
     $soniccPosition = 'Unknown';
     return $soniccPosition;
  }
  $rqsPos = &getAxisRqs ($soniccAxis_PV,\$status);
  if ($status) {
     &updateLogBox('*** Error reading SONICC rqspos.');
     $soniccPosition = 'Unknown';
     return $soniccPosition;
  }
  $ampEna = &checkMotorAmp ($soniccAxis_PV,0,\$status);
  if ($status) {
     &updateLogBox('*** Error reading SONICC motor amplifier.');
     $soniccPosition = 'Unknown';
     return $soniccPosition;
  }

  if (abs($rqsPos-$sonicc_in_pos) < $sonicc_tolerance) {                #requested is close to "in"
     $direction = 'IN';
     $percent   = 100*abs(($actPos-$sonicc_out_pos)
                / ($sonicc_out_pos-$sonicc_in_pos));
  }
  if (abs($rqsPos-$sonicc_out_pos) < $sonicc_tolerance) {               #requested is close to "out"
     $direction = 'OUT';
     $percent   = 100*abs(($actPos-$sonicc_in_pos)
                / ($sonicc_out_pos-$sonicc_in_pos));
  } 
  else {
     $direction = 'UNKNOWN';
     $percent   = 0;
  }
  $percent = int($percent+0.5);

  if ($inPos != 1 || $runPrg == 1) {
     if ($ampEna == 0) {$soniccPosition = 'stuck@middle';}
     else              {$soniccPosition = 'running '.$direction.', '.$percent.'% done';}			#running OUT, 45% done
  } else {
     if    (abs($actPos-$sonicc_in_pos)  < $sonicc_tolerance) {
        $soniccPosition = 'IN';
     }
     elsif (abs($actPos-$sonicc_out_pos) < $sonicc_tolerance) {
        $soniccPosition = 'OUT';
     }
     else {
       $soniccPosition = 'seats@middle';
     }
  }
  return $soniccPosition;
}

#--------------------------------------------------------------------------
#sub get_lightsIntensity(\@intensities) {                               #returns $status
sub get_lightsIntensity(\@) {
  my ($intensities_ref, $soniccPosition, $status, $statusAll, $i);

  $intensities_ref = shift(@_);
  $soniccPosition  = &get_soniccPosition();

  $statusAll = 0;
  if ($soniccPosition eq 'OUT') {                                       #remember before bringing SONICC in & turning off
     for ($i=0; $i<@lightsPVs; $i++) {
        ($status, @$intensities_ref[$i]) = Pezca::GetDouble($lightsPVs[$i]);
        if ($status) {&report_CAerr($status,$lightsPVs[$i]); $statusAll++;}
     }
  } 
# else {for ($i=0; $i<@lightsPVs; $i++) {@$intensities_ref[$i] = undef;}}      #cannot remember if not out
  return $statusAll;
}

#--------------------------------------------------------------------------
#sub set_lightsIntensity('ON'|'OFF',[\@intensities]) {                  #returns $status
sub set_lightsIntensity($;\@) {
  my ($onoff, $intensities_ref, $status, $statusAll, $i);

  $onoff           = shift(@_);
  $intensities_ref = shift(@_);

  if ($onoff eq 'ON') {
     &updateLogBox('Turning ON sample illumination lights');
     $statusAll = 0;
     for ($i=0; $i<@lightsPVs; $i++) {
        if (defined $lightsVal[$i]) {
           $status = Pezca::PutDoubleOldCa($lightsPVs[$i], @$intensities_ref[$i]);	#turn lights on
           if ($status) {&report_CAerr($status,$lightsPVs[$i]); $statusAll++;}
        }
     }
  } 
  elsif ($onoff eq 'OFF') {
     &updateLogBox('Turning OFF sample illumination lights');
     $statusAll = 0;
     for ($i=0; $i<@lightsPVs; $i++) {
        $status = Pezca::PutDoubleOldCa($lightsPVs[$i], 0);	                        #turn lights off
        if ($status) {&report_CAerr($status,$lightsPVs[$i]); $statusAll++;}
     }
  } 
  else {
     &updateLogBox('*** Incorrect Beamline Lights switch mode='.$onoff);
     $statusAll = -1;
  }
  return $statusAll;
}

#--------------------------------------------------------------------------
#sub soniccMove($dest) {                                                #returns $status
sub soniccMove($) {
  my ($dest, $soniccPosition, $status);
  $dest = shift(@_);
  $soniccPosition = &get_soniccPosition();

  &updateLogBox('Moving SONICC '.$dest);
  if ($dest =~ /^IN$/i) {
     $status = &get_lightsIntensity(\@lightsVal);
     $status = &setSampleEnv(['soniccIN','collimatorOUT'],'nowait');
  }
  elsif ($dest =~ /^OUT$/i) {
     $status = &setSampleEnv(['IN'],'nowait');                          #collimator-In + move SONICC out
  }
  else {
     &updateLogBox('*** Incorrect SONICC destination='.$dest);
     return 1;
  }
  if ($status) {&updateLogBox('*** Failed to move SONICC '.$dest);}
  else         {&updateLogBox('Successfully sent command to move SONICC '.$dest);}

  if    ($dest =~ /^IN$/i)  {&set_lightsIntensity('OFF');}
  elsif ($dest =~ /^OUT$/i) {&set_lightsIntensity('ON',\@lightsVal);}
  return $status;
}

#--------------------------------------------------------------------------
sub soniccWaitForIN() {                                                 #returns $soniccPosition
  my ($status, $soniccPosition, $inPos, $amp);
  my ($wait_, $wait_step, $wait_max, $w_, $wait_print, $msg);

  $soniccPosition = &get_soniccPosition();

  if ($soniccPosition eq 'IN') {return $soniccPosition;}

  &updateLogBox('SONICC is not IN. Moving there...');
  $status = &soniccMove ('IN');
  if ($status) {return $soniccPosition;}

  $w_    = 0;
  $wait_ = 0;
  $wait_max = 120;
  $wait_step = 0.2;
  $wait_print = 10;
  do {
     select (undef, undef, undef, $wait_step);
     $soniccPosition = &get_soniccPosition();
     $inPos = &getInPos ($soniccAssy_PV);
     $amp = &checkMotorAmp ($soniccAxis_PV);
     if ($inPos eq 1) {$inPos = 'Positioned';} else {$inPos = 'Moving';}
     if ($amp   eq 1) {$amp   = 'ON';}         else {$amp   = 'OFF';}
     $wait_ += $wait_step;
     $w_    += $wait_step;
     if ($w_ > $wait_print) {
        $w_ = 0;
        $msg = sprintf '%ds waiting for SONICC instrument to move IN.'
                      .' Current State="%s", Pos=%s, Amp=%s',
                      $wait_, $soniccPosition, $inPos, $amp;
        &updateLogBox($msg);
     }
  } while ($wait_          < $wait_max &&
           $soniccPosition ne 'IN'     &&
           $amp            ne 'OFF'    &&
           ($inPos         eq 'Moving' || $wait_ < 10));
  if ($soniccPosition ne 'IN') {
     if    ($wait_ >= $wait_max) {&updateLogBox('*** Timeout on moving SONICC IN.');}
     elsif ($amp   eq 'OFF')     {&updateLogBox('*** SONICC motor amplifier is OFF.');}
     elsif ($inPos ne 'Moving')  {&updateLogBox('*** SONICC stopped moving due to unknown reason.');}
     else                        {&updateLogBox('*** Failed to move SONICC IN.');}
  } else {
     &updateLogBox('SONICC has been moved IN.');
  }
  return $soniccPosition;
}

#--------------------------------------------------------------------------
sub load_sonicc_calibration() {                                         #returns $status
  my ($status, $statusAll, $changed, $i, $val);
  &updateLogBox('Loading SONICC calibration from EPICS PVs:');
  $statusAll = 0;
  $changed   = 0;
  for ($i=0; $i<@shgPVs; $i++) {
     ($status, $val) = Pezca::GetDouble($shgPVs[$i]);
     if ($status) {&report_CAerr($status, $shgPVs[$i]); $statusAll++;}
     else {
        if (! defined $parsSHG[$i] || $parsSHG[$i] != $val) {$changed++;}
        $parsSHG[$i] = $val;
     }
     ($status, $val) = Pezca::GetDouble($tpePVs[$i]);
     if ($status) {&report_CAerr($status, $tpePVs[$i]); $statusAll++;}
     else {
        if (! defined $parsTPE[$i] || $parsTPE[$i] != $val) {$changed++;}
        $parsTPE[$i] = $val;
     }
  }
  if ($statusAll || $changed) {
     &updateLogBox('Config loaded with '.$statusAll.' errors:'."\n"
                 . '     (omegaOffset xcenter ycenter  xpx2um  ypx2um, axes_rotation focusOffset)'."\n"
                 . 'SHG: '."(@{parsSHG})\n"
                 . 'TPE: '."(@{parsTPE})");
  }
  return $statusAll;
}

#--------------------------------------------------------------------------
#sub send_SONICC(cmd,[$debug]) {						#returns $status
sub send_SONICC($;$) {
  my ($cmd, $status, $debug);
  $cmd   = shift(@_);
  $debug = shift(@_);  if (!defined $debug) {$debug=0;}
  if ($debug) {&updateLogBox('Sending "'.$cmd.'" to SONICC');}
  $status = &PezcaPutStringWF($soniccCmdPV,$cmd);
  if ($status) {
     &report_CAerr($status,$soniccCmdPV);
     &updateLogBox('*** Command "'.$cmd.'" failed');
  }
  return $status;
}

#--------------------------------------------------------------------------
#sub query_SONICC($cmd,[$debug]) {						#returns $status,$rsp
sub query_SONICC($;$) {
  my ($cmd, $rsp, $status, $debug);
  $cmd = shift(@_);
  $debug = shift(@_);  if (!defined $debug) {$debug=0;}
  $status = &send_SONICC($cmd,$debug);
  if ($status == 0) {
     select (undef, undef, undef, 0.2);
     ($status,$rsp) = &PezcaGetStringWF($soniccRspPV);
     if ($status) {
        &report_CAerr($status,$soniccRspPV);
        &updateLogBox('*** Failed to get response to "'.$cmd.'"');
     } else {
        if ($debug) {&updateLogBox('SONICC responded: "'.$rsp.'"');}
     }
  }
  return $status, $rsp;
}

#--------------------------------------------------------------------------
#sub switch_sonicc_RL(Remote|Local) {                                   #returns $status
sub switch_sonicc_RL($) {
  my ($state, $status);
  $state = shift(@_);
  if ($state ne 'Remote' && $state ne 'Local') {
     &updateLogBox('*** Incorrect SONICC state='.$state.' -- expecting "Remote" or "Local"');
     return -1;
  }
  &updateLogBox('Switching SONICC to '.$state.' mode');
  $status = &send_SONICC('Control_State '.$state, $debug);
  return $status;
}

#--------------------------------------------------------------------------
#sub request_SONICC_to_load_scan_config() {                             #returns $status
sub request_SONICC_to_load_scan_config() {
  my ($status);
  &updateLogBox('Requesting SONICC to load '.$modeLabels[$soniccMode].' config '.$modeConfigs[$soniccMode]);
  $status = &send_SONICC('Imaging_Config '.$modeConfigs[$soniccMode], $debug);
  return $status;
}

#--------------------------------------------------------------------------
sub sonicc_scan_init() {                                                #returns $status
  my ($status);

  &updateLogBox('Switching SONICC to Remote');
  $status = &switch_sonicc_RL('Remote');
  if ($status) {return 1;}
  select (undef, undef, undef, 0.2);

  $status = &request_SONICC_to_load_scan_config();
  if ($status) {return 1;}
  select (undef, undef, undef, 0.2);

  &updateLogBox('Setting SONICC State '.$modeLabels[$soniccMode]);
  $status = &send_SONICC('SONICC_State '.$modeSwitchers[$soniccMode], $debug);
  if ($status) {return 1;}
  select (undef, undef, undef, 0.2);
  &updateLogBox('Setting SONICC Resonator ON');
  $status = &send_SONICC('Resonator_State On', $debug);
  if ($status) {return 1;}
  select (undef, undef, undef, 0.2);
  return 0;
}

#--------------------------------------------------------------------------
#sub do_sonicc_scan([$stayIN]) {                                        #returns $status
sub do_sonicc_scan(;$) {                                                #returns $status
  my ($stayIN, $status, $waited, $focusOffset, $msg);
  
  $stayIN = shift(@_);
  if (!defined $stayIN) {$stayIN = 0;}

  if (&soniccWaitForIN() ne 'IN') {
     &updateLogBox('*** Failed to bring SONICC IN. Quitting SONICC scan');
     return 1;
  }

  if (&sonicc_scan_init()) {
     &updateLogBox('*** Failed to initialize SONICC. Quitting SONICC scan');
     return 1;
  }

  if (&get_crystalPositions(\@crystalAct)) {                                     #this is in mm
     &updateLogBox('*** Failed to get Gonio Crystal positions. Quitting SONICC Scan');
     return 1;
  }

  $status = &load_sonicc_calibration();
  if ($status) {&zenity_error_exit($script.': failed to reload SONICC calibration from EPICS');}

  if    ($soniccMode == 0) {$focusOffset = $parsSHG[6];}			# $soniccMode=0: SHG
  elsif ($soniccMode == 1) {$focusOffset = $parsTPE[6];}			# $soniccMode=1: TPE-UVF
  else {
     $focusOffset = 0; &updateLogBox('*** Incorrect soniccMode='.$soniccMode);
  }

  if ($focusOffset != 0) {
     $msg = 'Shifting the sample by '.$focusOffset.'um to the SONICC focal plane';
     &updateLogBox($msg.' ("+/-" is down/up stream)');
     $msg = '2Z '.$focusOffset.' 0';
     $status = Pezca::PutStringOldCa($clickPV,$msg);
     if ($status) {&report_CAerr($status,$clickPV);}
     else         {&updateLogBox('Sent to P&C server: "'.$msg.'"');}
     $waited = &waitMotor($crystalAssyPV,10,1);			     	#waited: value=how-long-waited, 0=timeout
  }

###################################
### Command SONICC to take image  #
###################################
  $status = &command_SONICC_take_image();
  if ($status) {&updateLogBox('*** Commanding SONICC to take image failed');}
  else         {&updateLogBox('SONICC Image is taken');}
  &sonicc_standby();

### Return the sample back to original Z-position:
  if ($focusOffset != 0) {
     &updateLogBox('Shifting the sample back to the original Z-position.');
#    $status = &runNMotors   ($crystalAssyPV, 3, \@crystalAxesPV, \@crystalAct, 'forceAuto');
     $status = &startNMotors ($crystalAssyPV, 3, \@crystalAxesPV, \@crystalAct, 'forceAuto');
     if ($status) {&updateLogBox('*** Failed to return sample to the initial Z-position.');}
     else         {&updateLogBox('Sample Z-position is restored.');}
  }
  if (! $stayIN) {&soniccMove('OUT');}
  else           {&updateLogBox('Keeping SONICC IN as requested.');}
  return $status;                                                       #this is status from &command_SONICC_take_image()
}

#--------------------------------------------------------------------------
sub command_SONICC_take_image() {                                       #returns $status
  my ($status, $wait_, $wait_step, $wait_max, $w_, $wait_print, $progress);

  &updateLogBox('Taking SONICC image');
  $status = &send_SONICC('Scan_Once '.$windowsImageDir, $debug);
  if ($status) {return 1;}
  $w_    = 0;
  $wait_ = 0;
  $wait_max = 180;
  $wait_step = 0.2;
  $wait_print = 2;
  $progress = '';
  ### Wait for image to be taken:
  do {
     select (undef, undef, undef, $wait_step);
     $wait_ += $wait_step;
     $w_    += $wait_step;
     ($status,$progress) = &query_SONICC('Scan_Status');
     if ($status) {
        &sonicc_abort();
        return 1;
     }
     if ($w_ > $wait_print) {
        $w_ = 0;
        if (&is_int($progress)) {$progress .= 'line(s) or sweeps(s) done';}
        &updateLogBox($wait_.'s waiting for SONICC instrument completion, '
                      .'progress: '.$progress);
     }
  } while ($wait_ < $wait_max && $progress ne 'Scan Complete');

  if ($progress ne 'Scan Complete') {				#scan timeout
     &updateLogBox('*** Aborting SONICC Instrument on timeout');
     &sonicc_abort();
     return 1;
  }
  else {
     &updateLogBox('SONICC Image Taken'); 
     return 0;
  }
}

#--------------------------------------------------------------------------
sub sonicc_abort() {                                                    #returns $status
  my ($status);
  &updateLogBox('Aborting SONICC Scan');
  $status  = &send_SONICC('Scan_Abort', $debug);
  $status += &send_SONICC('Scan_Stop', $debug);
  return $status;
}

#--------------------------------------------------------------------------
sub sonicc_standby() {                                                  #returns $status
  my ($status);
  &sonicc_abort();
  &updateLogBox('Switching SONICC to standby');
  $status  = &send_SONICC('SONICC_State Standby', $debug);
  return $status;
}

#--------------------------------------------------------------------------
sub goodbye() {
  &sonicc_abort();
  &updateLogBox('Goodbye!'."\n");
  &clearAllCaMonitors();
  exit 0;
}
