#!/usr/bin/perl -W
# _CALLED_FROM_BLUICE_
#--------------------------------------------------------------
#
#                Author: Sergey Stepanov
#
# Version-1.0:  2007/11/11 using Pezca
# Version-2.0:  2007/11/19 moved lane change parameters into mySQL
# Version-3.0:  2007/12/05 changed to use variable ID offsets with data from mySQL
# Version-3.1:  2009/10/12 added confirmation dialog for big energy changes
# Version-3.2:  2009/11/23 added skipping optimization when MOSTAB fails
# Version-3.3:  2010/10/22 added multiple attempts to move energy if failed
# Version-3.4:  2011/01/07 added global flag for opening popups via BluIce
# Version-3.5:  2011/02/07 added restoring mostab/postab and killing mirrors on abort
# Version-3.6:  2011/12/08 changed popup after optimization is complete to a no-popup
#                          if initially the pin was not detected by the sensor
# Version-3.7:  2013/01/29 added turning on P2 search option in Mostab at really big energy changes
# Version-3.8:  2013/10/07 added lockin flag (do not pause lockin-type feedback)
# Version-3.9:  2013/10/15 added options to sync mono & ID during energy change: noSync/multistep/IDscan
# Version-4.0:  2014/01/27 added setting energy for Pilatus
# Version-4.1:  2014/03/06 added option to keep shutter open with lockin;
#                          removed beam optimization with lockin
# Version-4.2   2014/03/28 changed functions calls to arrays pointers (preparation to Perl-5.16)
# Version-4.3   2014/10/06 cleaned up interaction with lockin feedback ($feedback_mode=2)
# Version-4.4   2015/05/14 changed the way external scripts are called (added tracing);
#                          added killing external scripts on abort
# Version-4.5   2015/05/15 added recoveryTime command line option (sleeping before optimization)
# Version-4.6   2015/10/30 added lower cap of 120s on automatic recoveryTime if ID harmonic changed.
# Version-4.7   2015/11/18 changed automatic recoveryTime calc to temperature recovery time based on Bob's studes;
#                          added new cmdline parameter dThetaAutoAlign
# Version-4.8   2016/02/02 deprecated honorLockin flag and allowed pausing in the intensity feedback HW mode
# Version-4.9   2016/07/19 added attenuation to the call of runMono. Now, if move is multistep, attenuation
#                          is preseved even during the run (not only restored at the end);
#                          disabled pausing lock-in intensity feedback when sync is requested
# Version-5.0:  2016/09/02 added one more option to sync mono & ID during energy change: IDsteps
# Version-5.1:  2016/10/31 added bigEnergyChangeThresholdKeV command line parameter
# Version-5.2   2016/11/07 added protection from calculating zero alignment range;
#                          added setting new energy for Eiger
# Version-5.3   2017/10/19 converted pin sensor readout into robot_subs functions
# Version-5.4   2019/02/01 small bug fixes
# Version-5.5   2019/02/08 added controls for CRL @ IDD, v1: correct CRL downstream position.
#                          no full support yet (read positions from SQL and interpolate).
#                          Needs $PEZCA/data/Be.dat
# Version-5.6   2019/02/28 added controls for CRL @ IDD, v2: interpolate SQL data; now three options for 
#                          adjusting the CRL downstream position: shiftCRL=calc|SQLabs|SQLrel
# Version-6.0   2020/09/28 added transfocator
# Version-6.1   2020/10/14 added transfocator refocusing to reference with/without shift
# Version-6.2   2020/10/15 added optimization threshold $shift_big on shifting transfocator or CRL box 
# Version-6.3   2020/10/21 added sample auto-shifting if present
# Version-6.4   2020/10/27 changed sample auto-shifting from the point&click server request to direct moving
# Version-6.5   2020/10/29 added checking that focus is set with a popup
# Version-6.6   2020/10/30 added transfocator horizontal misalignment angle ($TF_H_misalignRad) 
# Version-6.7   2020/11/17 added transfocator using saved alignment data with horizontal misalignment angle 
# Version-6.8   2020/11/25 added added upstream/downstream effective shift of transfocator when changing lenses 
# Version-6.9   2021/02/06 added AlignScanStretch=[auto|none|value] and added &read_jbluice_config_flags for alignMiniBeam.pl
# Version-7.0:  2021/02/14 changed to set TF focus via the TF server.
# Version-7.1:  2021/08/18 adapted to simplified &getTFserverState()
# Version-7.2:  2021/08/22 added 5th and 7th ID harmonics and 111 <-> 333 switch
# Version-7.3:  2021/10/04 added $check_TF to speedup the script and otherwise &excludeCRLradius for the same purpose
# Version-7.4:  2022/03/03 added using / not using HFM and VFM individually
#--------------------------------------------------------------
# ATTENTION:
# To install MySQL support under Windows (ActivePerl)
# ppm>install DBI
# ppm>install DBD-mysql
#--------------------------------------------------------------
# use Math::Trig;					#needed for mono Theta <-> Energy calculations
  use DBI;
  use Pezca;
  use Time::HiRes qw(gettimeofday tv_interval);		#needed for measuring total time
  use warnings::unused;
  use strict;

  our ($IGNORE_ERRORS, $NONINTERACTIVE, $ATTENUATION_USE_FACTOR);
  our ($USE_ZENITY_FOR_EXIT, $ZENITY_VIA_BLUICE, $JBLUICE_SERVER);
  our ($beamline, $undulator, $shutter, $shutter_readback, $HAVE_XYZ);
  our ($have_KBM, $have_CRL, $have_TF);			#see beamline.pl
  our (@CRLconfig);					#cached array of the latest CRL or TF config (what is normally read from SQL)
  my ($PEZCA, $script, $syntax, @beamlinelist, $status, $xterm, $gterm, $mterm, $scr);
  my ($E_min, $E_max, $E_big, $Theta_big, $Theta_bigbig);
  my ($switchID, $switchMono, $motionTime, $E_drvl, $E_drvh);
  my ($enmono13, @harmEn, @harmNum, @Elower, @Eupper, @Olower, @Oupper);
  my ($E_beg, $E_end, $E_step, $H_beg, $linkID, $ID_beg, $ID_end);
  my ($Theta_beg, $Theta_end, $Theta_step, $ESinTheta, $msg, $actpos);
  my ($monoAssy, $monoAxis, $laneAssyVL, $laneAxisVL, $laneAssyHD, $laneAxisHD);
  my ($harmOff_beg, $harmOff_ssb, $harmOff_end, $HLname, $nHL, $nUH);
  my (@laneEnVL, @lanePosVL, @laneMatVL, $nVL, $VL_beg, $VLname);
  my (@laneEnHD, @lanePosHD, @laneMatHD, $nHD, $HD_beg, $HDname);
  my ($gapCheckPV, $gapCheck, $diff, $i, $mostab_initial, $postab_initial);
  my ($narg, $arg, $attenuation, $attenuation_resulted);
  my ($align, $align_needed, $align_stretch, @alignFlags);
  my ($opt_specified, $bln_specified, $lid_specified, $blu_specified);
  my ($syn_specified, $shtr_specified, $crl_specified, $mbr_specified);
  my ($hfm_specified, $vfm_specified, $hdm_specified, $mir_specified);
  my ($iUH_beg, $iVL_beg, $iHD_beg, $hkl_beg);
  my ($iUH_ssb, $iVL_ssb, $iHD_ssb, $hkl_ssb);
  my ($iUH_end, $iVL_end, $iHD_end, $hkl_end);
  my ($use_HFM, $use_VFM, $use_HDM, $have_HDM);
  my ($pinsensor_PV, $pinsensor_present, $pin_mounted);
  my ($attempt, $attempts, $use_xterm, $nelm, $t0, $elapsed);
  my ($feedback_mode, @feedbackLabels);
  my ($mostab_disabled, $postab_disabled, $gapcheck_disabled);
  my (@prg, $str, $mostab_wait_max, $mostab_wait_extra, $wait_, $zzz);
  my ($mostab_search_ini, $mostab_search_opt, $mirrors_enabled);
  my ($align_factor, $align_scanmode, $tell_bluice, $gnuplot, $medm);
  my ($detector, $detectorPV, $syncMode, $syncStep, $shutterOpt);
  my ($recoveryTime, $rt_specified, $da_specified, $bt_specified);
  my ($use_CRL, $nCRL, $radiusCRL, $materialCRL);
  my ($CRL_drvl, $CRL_drvh, $TF_drvl, $TF_drvh, $TF_drv_center);
  my ($focus_beg, $focus_end, $focus_rqs, $focus_shift);
  my ($align_Mode, $ali_specified, $ar_specified, @hasharray, $nhash);
  my ($CRL_StB_assy, $CRL_dwstB_PV, $CRL_dwstB_beg, $CRL_dwstB_end);
  my ($use_TF, @use_TF_labels, $device);
  my ($nrecords, $shift_val, $shift_big, $ts_specified);
  my ($TF_axis_PV, $TF_assy_PV, $TF_X_beg, $TF_X_end);
  my ($TF_axsH_PV, $TF_assH_PV, $TF_H_beg, $TF_H_end);
  my ($TF_H_shift, $TF_H_misalignRad, $ma_specified);
  my ($TFserver_sts, $TFserver_ref, $TFserver_rqs, $TFrangeMax);
  my ($TFsrchMode, @InOut_end, $check_TF);
  my ($protectSample, $ps_specified, $moveSample, $sample_moved);
  my ($xyzAssyPV, $zAssyPV, @xyzAxesPV, @xyzIni, $nxyz, $direction);
  my ($omegaAxisPV, $omega, $degree2rad, $md_specified, $gn_specified);
  my ($E_exclude, $r_exclude);				#any lenses with this or smaller radius will be excluded an smaller energies

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  $IGNORE_ERRORS          = 1;							#GLOBAL: do not break script if error occurred
  $NONINTERACTIVE         = 1;							#GLOBAL: do not wait for any answers
  $ATTENUATION_USE_FACTOR = 1;							#GLOBAL: how to treat attenuation value
  $USE_ZENITY_FOR_EXIT    = 1;							#GLOBAL: popup exit messages
 
  do $PEZCA.'/Common/motor_subs.pl';
  do $PEZCA.'/Common/mono_subs.pl';
  do $PEZCA.'/Common/crl_subs.pl';
  do $PEZCA.'/Common/mostab_subs.pl';
  do $PEZCA.'/Common/shutter_subs.pl';
  do $PEZCA.'/Common/attenuation_subs_v2.pl';
  do $PEZCA.'/Common/ccd_subs.pl';
  do $PEZCA.'/Common/robot_subs.pl';
  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  select STDERR; $|=1;								#set unbuffered output
  select STDOUT; $|=1;								#set unbuffered output

  $script = &makeScriptName();
  $scr    = 'EnMove';
  $syntax = "\n\n".'Syntax: '.$script.' RqsEnergy'                                       ."\n"
          . ' [beamline] [bigEnergyChangeThresholdKeV=xx] [noID]'                        ."\n"
          . ' [noSync|multiStep[=nnn]|IDsteps|IDscan] [recoveryTime=xx|auto]'            ."\n"
          . ' [noAlign|autoAlign|forceAlign] [dThetaAutoAlign=xx]'                       ."\n"
          . ' [switchMonoBragg|keepMonoBragg]'                                           ."\n"
          . ' [AlignScanStretch=auto|none|value] [alignWithKBM|alignWithCRL|alignWithTF]'."\n"
          . ' [TFshiftAutoAlign=xx] [TFmisalignmentAngleH=xx]'                           ."\n"
          . ' [noHFM|useHFM] [noVFM|useVFM] [noHDM|useHDM] [noMirrors|useMirrors]'       ."\n"
          . ' [noshiftCRL|shiftCRL=calc|=SQLabs|=SQLrel|=none]'                          ."\n"
          . ' [noTransfocator|useTransfocator[=rel|=auto|=reference]]'                   ."\n"
          . ' [ignoreSample|warnSample|moveSample=xx[h|v]|moveSampleNoWarn=xx[h|v]]'     ."\n"
          . ' [keepShutter|closeShutter|openShutter] [bluice|nobluice]'                  ."\n"
          . ' [medm|nomedm] [gnuplot|gnuplot=image|nognuplot]'                           ."\n"
                                                                                         ."\n";
  @beamlinelist = ('23i:', '23o:', '23b:', '23d:');
# ##############################
  ($E_min, $E_max)   = (3.5, 35.);						#this is only if Energy (DRVL,DRVH) is not set
  $E_big             = 0.5;							#what to consider BIG step (move into mySQL?)
  $Theta_big         = 1.;							#big theta change when re-alligning may be needed
  $Theta_bigbig      = 3.;							#big theta change when P2 search may be needed
  $shift_big         = 2;							#transfocator downstream shift when optimization may be needed
  $use_xterm         = 0;
  $mostab_wait_max   = 240;							#in seconds
  $mostab_wait_extra = 120;							#used when p2 search is on
  $align_scanmode    = 'fastScan';						#fastScan or StepScan or '' (default)
  $shutterOpt        = 'close';							#close/keep/open
  $recoveryTime      = 0;							# waiting time for mono temperature to recover before beam alignment
  @feedbackLabels    = ('software', 'hybrid', 'hardware(lockin)');
  $use_HFM           = 1;							#this can be changed per command line or per beamline below
  $use_VFM           = 1;							#this can be changed per command line or per beamline below
  $use_HDM           = 1;							#this can be changed per command line or per beamline below
  $protectSample     = 1;							# sample protection: 0=ignore, 1=warn, 2=auto-move 3=auto-move without warning
  $moveSample        = 0;							# vertical distance for moving up in mm
  $direction         = 'UP';							#UP, DOWN, OUTBOARD, INBOARD
  $TF_H_misalignRad  = 0.;							#TF misalignment angle in radians (H shift per downstream shift)
  $align_stretch     = 'auto';							#auto|none|value
  $check_TF          = 1;							#flag to pre-check if TF can focus at new energy before proceeding. Disabling may speed the script quite a bit!
  $E_exclude         = 26;							#exclude small-radius lenses below this energy in keV (need to match TF server!)
  $r_exclude         = 50;							#any lenses with this or smaller radius will be excluded an smaller energies (need to match TF server!)
  @use_TF_labels     = ('noTransfocator', 					# $use_TF=0
                        'useTransfocator=relative (preserve start focus)',	# $use_TF=1
                        'useTransfocator=auto (rel2start or reference)',	# $use_TF=2
                        'useTransfocator=reference (preserve reference focus)');# $use_TF=2

# ##############################

  ($iUH_beg, $iVL_beg, $iHD_beg, $hkl_beg) = (0, 0, 0, '');
  ($iUH_ssb, $iVL_ssb, $iHD_ssb, $hkl_ssb) = (0, 0, 0, '');
  ($iUH_end, $iVL_end, $iHD_end, $hkl_end) = (0, 0, 0, '');

  if ($^O !~ /Win/i) {
     $gterm = &Which('gnome-terminal');
     $mterm = &Which('mate-terminal');
     $xterm = &Which('xterm');
     $gterm = `which gnome-terminal 2>/dev/null`; chomp $gterm;
     $mterm = `which mate-terminal 2>/dev/null`;  chomp $mterm;
     if    (defined $gterm) {$xterm = $gterm.' --hide-menubar --disable-factory -e';}
     elsif (defined $mterm) {$xterm = $mterm.' --hide-menubar --disable-factory -e';}
     elsif (defined $xterm) {$xterm = $xterm.' +cm -sb -rightbar -sl 5000 -e';}
     else                   {die '*** No usable terminal found';}
  } else                    {$xterm = 'cmd /c';}

# $ARGV[0] = '14.';
# $ARGV[1] = '23o';
# $ARGV[2] = 'noID';
# $ARGV[2] = 'noAlign';

  $narg = @ARGV;
  print STDOUT "\n".'Date: '.scalar(localtime)."\n";
  if ($narg > 0) {print STDOUT $script.': started with cmdline = '."@ARGV\n\n";}
  else           {print STDOUT $script.': started with no cmdline arguments.'."\n\n";}
  if ($narg == 0) {
     &press_key_exit('*** '.$script.': no energy in the cmdline'.$syntax,1);
  }
  $E_end = $ARGV[0];
  if (length($E_end) == 0 || ! &is_float($E_end) || $E_end <= 0) {
     &press_key_exit('*** '.$script.': requested energy='.$E_end
                    .' must be a positive number.'."\n".$syntax,1);
  }

#---------------------------------------------------------------------------
### Scan for optional command line arguments:
  for ($i=1; $i<$narg; $i++) {
     $arg = $ARGV[$i];
     $arg =~ s/^-{0,2}//;									#trim leading dashes
#    if (grep {$arg =~ /^$_:?$/} @beamlinelist) {						#use this when beamlinelist does not contain ':'
     if (grep(/^$arg:?$/, @beamlinelist)) {
        if (!defined $bln_specified) {
           $bln_specified = 1;
           $arg =~ s/:$//;
           if ($arg ne $beamline) {
              print STDOUT "\n".'!!! '.$script.': Cross-beamline operation:'
                             .' controlling '.$arg.' from '.$beamline."\n\n";
              $beamline = $arg;
              do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
           }
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^noID$/i || $arg =~ /^linkID$/i) {
        if (!defined $lid_specified) {
           $lid_specified = 1;
           if    ($arg =~ /^linkID$/i) {$linkID = 1;}
           elsif ($arg =~ /^noID$/i)   {$linkID = 0;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^forceAlign$/i || $arg =~ /^forceOptimize$/i || $arg =~ /^align=force$/i ||
            $arg =~ /^autoAlign$/i  || $arg =~ /^autoOptimizeA$/i || $arg =~ /^align=auto$/i  ||
            $arg =~ /^noAlign$/i    || $arg =~ /^noOptimize$/i    || $arg =~ /^align=no$/i) {
        if (!defined $opt_specified) {
           $opt_specified = 1;
           if    ($arg =~ /^forceAlign$/i || $arg =~ /^forceOptimize$/i || $arg =~ /^align=force$/i) {$align = 2;}
           elsif ($arg =~ /^autoAlign$/i  || $arg =~ /^autoOptimize$/i  || $arg =~ /^align=auto$/i)  {$align = 1;}	#this is redundant as it is default
           elsif ($arg =~ /^noAlign$/i    || $arg =~ /^noOptimize$/i    || $arg =~ /^align=no$/i)    {$align = 0;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^dThetaAutoAlign=/i) {
        if (!defined $da_specified) {
           $da_specified = 1;
           $Theta_big = lc($arg);
           $Theta_big =~ s/^dThetaAutoAlign=//i;
           if (! &is_decimal($Theta_big) || $Theta_big < 0) {
              &press_key_exit('*** '.$script.': incorrect command line argument-'
                                                .$i.' ['.$ARGV[$i].']!'.$syntax,1);
           }
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^AlignScanStretch=/i) {				#[auto|none|value]
        if (!defined $ar_specified) {
           $ar_specified = 1;
           $align_stretch = lc($arg);
           $align_stretch =~ s/^AlignScanStretch=//i;
           if    ($align_stretch =~ /^auto$/) {$align_stretch='auto';}
           elsif ($align_stretch =~ /^none$/) {$align_stretch='none';}
           elsif (&is_decimal($align_stretch)) {
              $align_factor = abs($align_stretch);
              if ($align_factor == 0) {$align_factor=1;}
              $align_stretch = 'fixed';
           }
           else {                          
              &press_key_exit('*** '.$script.': incorrect command line argument-'
                                                .$i.' ['.$ARGV[$i].']!'.$syntax,1);
           }
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^TFshiftAutoAlign=/i) {
        if (!defined $ts_specified) {
           $ts_specified = 1;
           $shift_big = lc($arg);
           $shift_big =~ s/^TFshiftAutoAlign=//i;
           if (! &is_decimal($shift_big) || $shift_big < 0) {
              &press_key_exit('*** '.$script.': incorrect command line argument-'
                                                .$i.' ['.$ARGV[$i].']!'.$syntax,1);
           }
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
      elsif ($arg =~ /^TFmisalignmentAngleH=/i) {
        if (!defined $ma_specified) {
           $ma_specified = 1;
           $TF_H_misalignRad = lc($arg);
           $TF_H_misalignRad =~ s/^TFmisalignmentAngleH=//i;
           if (! &is_float($TF_H_misalignRad) || abs($TF_H_misalignRad) > 0.05) {		#not more then 5% lateral/downstream shift is reasonable
              &press_key_exit('*** '.$script.': incorrect command line argument-'
                                                .$i.' ['.$ARGV[$i].']!'.$syntax,1);
           }
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
    elsif ($arg =~ /^bigEnergyChangeThresholdKeV=/i) {
        if (!defined $bt_specified) {
           $bt_specified = 1;
           $E_big = lc($arg);
           $E_big =~ s/^bigEnergyChangeThresholdKeV=//i;
           if (&is_decimal($E_big) != 1 || $E_big < 0) {
              &press_key_exit('*** '.$script.': incorrect command line argument-'
                                                .$i.' ['.$ARGV[$i].']!'.$syntax,1);
           }
           elsif ($E_big > 2) {
              &press_key_exit('*** '.$script.': energy threshold exceeds 2 keV in argument-'
                                                .$i.' ['.$ARGV[$i].']!'.$syntax,1);
           }
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^switchMonoBragg$/i || $arg =~ /^keepMonoBragg$/i) {
        if (!defined $mbr_specified) {
           $mbr_specified = 1;
           if    ($arg =~ /^switchMonoBragg$/i) {$switchMono = 1;}
           elsif ($arg =~ /^keepMonoBragg$/i)   {$switchMono = 0;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^keepShutter$/i  ||
            $arg =~ /^closeShutter$/i ||
            $arg =~ /^openShutter$/i) {
        if (!defined $shtr_specified) {
           $shtr_specified = 1;
           if    ($arg =~ /^keepShutter$/i)  {$shutterOpt = 'keep';}
           elsif ($arg =~ /^closeShutter$/i) {$shutterOpt = 'close';}
           elsif ($arg =~ /^closeShutter$/i) {$shutterOpt = 'open';}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^noSync$/i    ||
            $arg =~ /^multiStep$/i || $arg =~ /^multiStep=/i ||
            $arg =~ /^IDsteps$/i   || $arg =~ /^IDscan$/i) {
        if (!defined $syn_specified) {
           $syn_specified = 1;
           if    ($arg =~ /^noSync$/i)    {$syncMode = 'noSync';}
           elsif ($arg =~ /^multiStep$/i) {$syncMode = 'multiStep';}				#default step=0.25 keV (see &runMono)
           elsif ($arg =~ /^multiStep=/i) {$syncMode = 'multiStep';
              $syncStep = $arg;  $syncStep =~ s/^multiStep=//i;
              if (! &is_float($syncStep)) {
                 &press_key_exit('*** '.$script.': incorrect command line argument-'
                                                   .$i.' ['.$ARGV[$i].']!'.$syntax,1);
              }
           }
           elsif ($arg =~ /^IDsteps$/i)  {$syncMode = 'IDsteps';}
           elsif ($arg =~ /^IDscan$/i)   {$syncMode = 'IDscan';}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^shiftCRL=/i || $arg =~ /^noshiftCRL$/i) {
        if (!defined $crl_specified) {
           $crl_specified = 1;
           if    ($arg =~ /^noshiftCRL$/i)     {$use_CRL=0;}					#this is redundant as it is default
           elsif ($arg =~ /^shiftCRL=none$/i)  {$use_CRL=0;}					#this is redundant as it is default
           elsif ($arg =~ /^shiftCRL=calc/i)   {$use_CRL=1; $use_TF=0;}
           elsif ($arg =~ /^shiftCRL=SQLabs/i) {$use_CRL=2; $use_TF=0;}
           elsif ($arg =~ /^shiftCRL=SQLrel/i) {$use_CRL=3; $use_TF=0;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^useTransfocator/i || $arg =~ /^noTransfocator$/i) {
        if (!defined $crl_specified) {
           $crl_specified = 1;									#same flag as for CRL (cannot use both)
           if    ($arg =~ /^noTransfocator$/i)                   {$use_TF=0;}			#this is redundant as it is default
           elsif ($arg =~ /^useTransfocator$/i)                  {$use_TF=1; $use_CRL=0;}
           elsif ($arg =~ /^useTransfocator=rel$/i)              {$use_TF=1; $use_CRL=0;}
           elsif ($arg =~ /^useTransfocator=auto$/i)             {$use_TF=2; $use_CRL=0;}
           elsif ($arg =~ /^useTransfocator=reference$/i)        {$use_TF=3; $use_CRL=0;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^alignWithKBM$/i || $arg =~ /^alignWithCRL$/i || $arg =~ /^alignWithTF$/i) {
        if (!defined $ali_specified) {
           $ali_specified = 1;
           if    ($arg =~ /^alignWithKBM$/i) {$align_Mode = 'KBM';}
           elsif ($arg =~ /^alignWithCRL$/i) {$align_Mode = 'CRL';}
           elsif ($arg =~ /^alignWithTF$/i)  {$align_Mode = 'TF';}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^useHFM$/i || $arg =~ /^noHFM$/i) {
        if (!defined $hfm_specified) {
           $hfm_specified = 1;
           if    ($arg =~ /^useHFM$/i) {$use_HFM = 1;}
           elsif ($arg =~ /^noHFM$/i)  {$use_HFM = 0;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^useVFM$/i || $arg =~ /^noVFM$/i) {
        if (!defined $vfm_specified) {
           $hfm_specified = 1;
           if    ($arg =~ /^useVFM$/i) {$use_VFM = 1;}
           elsif ($arg =~ /^noVFM$/i)  {$use_VFM = 0;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^useHDM$/i || $arg =~ /^noHDM$/i) {
        if (!defined $hdm_specified) {
           $hdm_specified = 1;
           if    ($arg =~ /^useHDM$/i) {$use_HDM = 1;}
           elsif ($arg =~ /^noHDM$/i)  {$use_HDM = 0;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^useMirrors$/i || $arg =~ /^noMirrors$/i) {
        if (!defined $mir_specified) {
           $mir_specified = 1;
           if    ($arg =~ /^useMirrors$/i) {$use_HFM = 1; $use_VFM = 1; $use_HDM = 1;}
           elsif ($arg =~ /^noMirrors$/i)  {$use_HFM = 0; $use_VFM = 0; $use_HDM = 0;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^bluice$/i || $arg =~ /^nobluice$/i) {
        if (!defined $blu_specified) {
           $blu_specified = 1;
           if    ($arg =~ /^bluice$/i)   {
              $tell_bluice = 1; $gnuplot='nognuplot'; $medm='nomedm';
           }
           elsif ($arg =~ /^nobluice$/i) {
              $tell_bluice = 0; $gnuplot='gnuplot'; $medm='medm';
           }
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^medm$/i || $arg =~ /^nomedm$/i) {
        if (!defined $md_specified) {
           $md_specified = 1;
           $medm = lc($arg);
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^gnuplot$/i || $arg =~ /^gnuplot=image$/i || $arg =~ /^nognuplot$/i) {
        if (!defined $gn_specified) {
           $gn_specified = 1;
           $gnuplot = lc($arg);
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^recoveryTime=/i) {
        if (!defined $rt_specified) {
           $rt_specified = 1;
           $recoveryTime = lc($arg);
           $recoveryTime =~ s/^recoveryTime=//i;
           if ($recoveryTime !~ /^auto$/ &&
               (&is_decimal($recoveryTime) != 1 || $recoveryTime < 0)) {
              &press_key_exit('*** '.$script.': incorrect command line argument-'
                                                .$i.' ['.$ARGV[$i].']!'.$syntax,1);
           }
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^ignoreSample$/i || 
            $arg =~ /^warnSample$/i || 
            $arg =~ /^moveSample=/i ||
            $arg =~ /^moveSampleNoWarn=/i) {
        if (!defined $ps_specified) {								#this will exit automatically if $NONINTERACTIVE=1
           $ps_specified = 1;
           if ($arg =~ /^ignoreSample$/i) {
              $protectSample = 0;								#0=ignore sample
              $moveSample    = 0;
           } 
           elsif ($arg =~ /^warnSample$/i) {
              $protectSample = 1;								#1=warn that sample needs to be moved
              $moveSample    = 0;
           } 
           elsif ($arg =~ /^moveSample=/i) {
              $protectSample = 2;								#2=move sample automatically (after asking)
              $moveSample = lc($arg);
              $moveSample =~ s/^moveSample=//i;
           }
           elsif ($arg =~ /^moveSampleNoWarn=/i) {
              $protectSample = 3;								#3=move sample automatically without asking
              $moveSample = lc($arg);
              $moveSample =~ s/^moveSampleNoWarn=//i;
           }
           else {
              print STDOUT '!!! '.$script.': unexpected sample protection option! Ignoring.'."\n\n";
              $protectSample = 1;								#1=warn that sample needs to be moved
              $moveSample    = 0;
           }
           if ($protectSample > 1) {								#2=move sample(ask), 3=move sample without asking
              if    ($moveSample =~ /h$/i) {$direction = 'OUTBOARD'; chop $moveSample;}
              elsif ($moveSample =~ /v$/i) {$direction = 'UP';       chop $moveSample;}
              else                         {$direction = 'UP';}
              if (! &is_float($moveSample) || $moveSample == 0 || abs($moveSample) > 50) {	#this will exit automatically if $NONINTERACTIVE=1
                 &press_key_exit('*** '.$script.': Incorrect cmdline: moveSample'
                                       .' must specify a positive number in mm.'
                                       .' It cannot exceed 50mm.'.$syntax);
              }
              if    ($direction eq 'UP'       && $moveSample < 0) {$direction = 'DOWN';}
              elsif ($direction eq 'OUTBOARD' && $moveSample < 0) {$direction = 'INBOARD';}
           }
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif (
        $arg =~ /^honorLockin$/i) {print STDOUT '!!! '.$script.': the honorLockin flag is deprecated!'."\n\n";
     }
     else {
        &press_key_exit('*** '.$script.': unrecognized command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
     }
  }
#---------------------------------------------------------------------------

### Configuration parameter -- change if beamline hardware changes:
  if (!defined $have_KBM) {$have_KBM = 1;}							#see beamline.pl
  if (!defined $have_CRL) {$have_CRL = 0;}							#see beamline.pl
  if (!defined $have_TF)  {$have_TF  = 0;}							#see beamline.pl

# if ($beamline eq '23i') {$use_VFM  = 0;}							#no mirror (2022.03)
  if ($beamline ne '23o') {$use_HDM  = 0;}							#no mirrors

### Defaults for refocusing with CRL or Transfocator
  if (!defined $use_CRL) {$use_CRL = $have_CRL;}						#0=no,1=calc,2=SQLabs,3=SQLrel  (only if $have_CRL=1)
  if (!defined $use_TF)  {$use_TF  = $have_TF;}							#0=no,1=rel,2=auto,3=ref,4=refNoShift (transfocator refocus/shift; only if $have_TF=1) 

### Post processing (if not given in the command line):
  if (!defined $switchMono)  {$switchMono  = 0;}						#do not switch between 111 and 333 automatically
  if (!defined $align)       {$align       = 1;}						#auto choose based on dTheta
  if (!defined $linkID)      {$linkID      = 1;}
  if (!defined $tell_bluice) {$tell_bluice = 1;}						#tell, but still plot!
  if (!defined $gnuplot)     {$gnuplot     = 'gnuplot';}
  if (!defined $medm)        {$medm        = 'medm';}

  if (!defined $align_Mode) {
     if    ($have_KBM) {$align_Mode = 'KBM';}							#'KBM', 'CRL', or 'TF'
     elsif ($have_TF)  {$align_Mode = 'TF';}
     elsif ($have_CRL) {$align_Mode = 'CRL';}
     else              {$align_Mode = '';}
  }
  if (! $have_KBM) {
     $use_HFM = 0;
     $use_VFM = 0;
     if ($align_Mode eq 'KBM') {
        print STDOUT '!!! '.$script.': Beam alignment with KBMs requested,'
                    .' but no KBMs in the beam! Alignment is disabled.'."\n\n";
        $align = 0;
     }
  }
  if (! $have_CRL) {
     if ($use_CRL > 0) {
        print STDOUT '!!! '.$script.': CRLs motion requested, but no CRLs'
                    .' at the beamline! Ignoring.'."\n\n";
        $use_CRL = 0;
     }
     if ($align_Mode eq 'CRL') {
        print STDOUT '!!! '.$script.': Beam alignment with CRLs requested,'
                    .' but no CRLs at the beamline! Alignment is disabled.'."\n\n";
        $align = 0;
     }
  }
  if (! $have_TF) {
     if ($use_TF > 0) {
        print STDOUT '!!! '.$script.': Transfocator refocusing requested,'
                    .' but no TF at the beamline! Ignoring.'."\n\n";
        $use_TF = 0;
     }
     if ($align_Mode eq 'TF') {
        print STDOUT '!!! '.$script.': Beam alignment with Transfocator requested,'
                    .' but no TF at the beamline! Alignment is disabled.'."\n\n";
        $align = 0;
     }
  }
  if ($tell_bluice) {$ZENITY_VIA_BLUICE = 1;}
#---------------------------------------------------------------------------
  $shift_val  = 0;										#transfocator/CRL shift value
  $monoAssy   = $beamline.':MO:En';
  $monoAxis   = $beamline.':MO:E';
  if ($beamline ne '23b') {
    $laneAssyVL = $beamline.':KB:Vy';
    $laneAxisVL = $beamline.':KB:VL';
    $laneAssyHD = $beamline.':HD:St';
    $laneAxisHD = $beamline.':HD:SL';
  } else {
    $laneAssyVL = $beamline.':VF:Vy';								#use vertical focusing instead of VKB
    $laneAxisVL = $beamline.':VF:VL';								#use vertical focusing instead of VKB
    $laneAssyHD = $beamline.':VC:Vy';								#use white beam mirror instead of HDM
    $laneAxisHD = $beamline.':VC:VL';								#use white beam mirror instead of HDM
  }
  if (defined $shutter) {
     if ($shutter !~ /^${beamline}:/) {$shutter   = $beamline.':'.$shutter;}
  }
  if (defined $shutter_readback) {
     if ($shutter_readback !~ /^${beamline}:/) {
                                $shutter_readback = $beamline.':'.$shutter_readback;}
  }
  $xyzAssyPV = $beamline.':GO:Cr';
  if (defined $HAVE_XYZ) {
     $zAssyPV   = $xyzAssyPV;
     @xyzAxesPV = ($beamline.':GO:CX',$beamline.':GO:CY',$beamline.':GO:CZ');
  } else {
     $zAssyPV = $beamline.':GO:St';
     if ($beamline eq '23b') {$zAssyPV .= 'B';}
     @xyzAxesPV = ($beamline.':GO:CX',$beamline.':GO:CY',$beamline.':GO:SH');
  }
  $nxyz =  @xyzAxesPV; 
  $omegaAxisPV = $beamline.':GO:OM';
  $degree2rad = atan2(1.,1.) / 45.;								#atan2(1)=pi/4 = 45.

  $pinsensor_PV = &getPinSensorPV($beamline);
  if (defined $pinsensor_PV) {$pinsensor_present = 1;}
  else                       {$pinsensor_present = 0;}
  $pin_mounted = 1;										#presume pin is present
# if ($beamline eq '23b') {$pinsensor_present = 0;}
  $gapCheckPV   = $beamline.':bi:enableEnergyGapCheck';
  if ($beamline eq '23b' || $beamline eq '23d') {$linkID = 0;}
  if ($beamline eq '23o') {$E_max = 20.;}
  $VLname       = 'VFM';
  $HLname       = 'HFM';									#Those are not changed since HFM lane motion is unstable
  $HDname       = 'HDM';
  if ($beamline eq '23b') {$HDname = 'VCM';}

  ($detector,$status) = &getCCDname();								#see ccd_subs.pl: mySQL-> setialNo -> name
  if ($status) {
     $detector = 'marccd';
     print STDOUT '!!! '.$script.': detector detection failed with status='.$status."\n"
                 .'!!! '.$script.': presuming detector='.$detector."\n";
  } else {
     print STDOUT 'Beamline detector = '.$detector."\n";
  }

### Check if request is in range:
  ($E_drvl, $E_drvh) = &getAxisDRVLH ($monoAxis);
  if ($E_drvl != 0 || $E_drvh != 0) {($E_min,$E_max) = ($E_drvl, $E_drvh);}

  if ($E_end < $E_min || $E_end > $E_max) {
     if ($tell_bluice) {&tellBluIce($scr.': Energy is not in range','error');}
     &zenity_error_exit('Requested energy='.$E_end.' is not in range ['
                                   .$E_min.'-'.$E_max.'] keV'."\n".$syntax,1);
  }
  if ($shutterOpt =~ /^close$/i) {&closeShutter ($shutter,$shutter_readback);}
  if ($shutterOpt =~ /^open$/i)  {&openShutter ($shutter,$shutter_readback);}

### ------------------------------------------------------------------------
### Determine lane switch and harmonic switch parameters;
### read them from BluIce mySQL database:
###              ^^^^^^^^^^^^^^^^
### Read mySQL 'blc2004/Config' table:
  $status = &getIDparams(\$enmono13,\@harmEn,\@harmNum,
                         \@Elower,\@Olower,\@Eupper,\@Oupper);			# a more-friendly wrapper around &getVariableIDoffsets
  if ($status) {
     &zenity_error_exit('Cannot read Mono-ID Offsets data from mySQL',1);
  }

### Read mySQL 'blc2004/EnergyMove' table:
  $status = &getLaneOffsets(\@laneEnVL, \@lanePosVL, \@laneMatVL,
                            \@laneEnHD, \@lanePosHD, \@laneMatHD);
  if ($status) {
     if ($tell_bluice) {&tellBluIce($scr.': mySQL/lane error','error');}
     &zenity_error_exit('Cannot read Lane Offsets data from mySQL',1);
  }

### Data below has migrated into mySQL 'EnergyMove' table:
# if    ($beamline eq '23i' || $beamline eq '23d') {
#    @laneEnVL  = ($E_min,  9.35,   20.);
#    @lanePosVL = (5.67,   -4.33,  15.67);
#    @laneMatVL = ('SiO2', 'Rh',   'Pt');
#    @laneEnHD  = ();
#    @lanePosHD = ();
#    @laneMatHD = ();
# }
# elsif ($beamline eq '23o') {
#    @laneEnVL  = ($E_min, 9.35);
#    @lanePosVL = (-4.33,  5.67);
#    @laneMatVL = ('SiO2', 'Rh');
#    @laneEnHD  = ($E_min, 16.);
#    @lanePosHD = (-4.33,  5.67);
#    @laneMatHD = ('Rh',  'Pt');
# }
# elsif ($beamline eq '23b') {		# These are fake. Data to be given by Derek.
#    @laneEnVL  = ($E_min, 9.35);       # place here VFM data (focus.mirror)
#    @lanePosVL = (-4.33,  5.67);       # place here VFM data (focus.mirror)
#    @laneMatVL = ('SiO2', 'Rh');       # place here VFM data (focus.mirror)
#    @laneEnHD  = ($E_min, 9.35);       # place here VCM data (white beam mirror)
#    @lanePosHD = (-4.33,  5.67);       # place here VCM data (white beam mirror)
#    @laneMatHD = ('SiO2', 'Rh');       # place here VCM data (white beam mirror)
# }

  $nHL = 0;											#Those are not changed since HFM lane motion is unstable
  $nVL = @lanePosVL;
  $nHD = @lanePosHD;	if ($nHD > 0) {$have_HDM=1;} else {$have_HDM=0;}
  $nUH = @harmNum;
  print STDOUT $script.': #### Current mySQL parameters ####'."\n";
  if ($nUH > 0) {
     print STDOUT '  -- ID harmonic changes:'."\n";
     for ($i=1; $i<$nUH; $i++) {
        printf STDOUT '                 at E=%5.2f keV  harmonic: %d to %d'."\n",
                                  $harmEn[$i], $harmNum[$i-1], $harmNum[$i];
     }
     print STDOUT '  -- ID offsets:'."\n";
     for ($i=0; $i<$nUH; $i++) {
        printf STDOUT '      ID_harm=%d  at E=%5.2f keV  offset=%5.3f keV'."\n"
                      .'                 at E=%5.2f keV  offset=%5.3f keV'."\n",
                     $harmNum[$i], $Elower[$i], $Olower[$i], $Eupper[$i], $Oupper[$i];
     }
  }
  if (defined $enmono13 && $enmono13 > 0) {
     print STDOUT '  -- Mono Bragg reflection change from 111 to 333:'."\n";
     print STDOUT '                 at E='.$enmono13.' keV'."\n";
  }
  if ($nVL > 0 && $use_VFM) {									#Vertical focusing mirror lane changes
     print STDOUT '  -- '.$VLname.' lane changes:'."\n";
     for ($i=1; $i<$nVL; $i++) {
        print STDOUT '      at E='.$laneEnVL[$i].'  position: '
                               .$lanePosVL[$i-1].' -> '.$lanePosVL[$i]."\n"
                    .'                              material: '
                               .$laneMatVL[$i-1].' -> '.$laneMatVL[$i]."\n";
     }
  }
  if ($enmono13 > 0 && $switchMono) {								#Switching between 111 and 333 Bragg reflections
     print STDOUT '  -- Mono Bragg reflection changes:'."\n"
                 .'      at E='.$enmono13.'  Bragg planes from (111) to (333).'."\n";
  } else {
     $switchMono = 0;
     $enmono13   = 0;
  }

  if ($nHL > 0 && $use_HFM) {									#Horizontal focusing mirror lane changes not implemented
     ### HFM lane change is not implemented and the
     ### thresholds are not included in mySQL because the
     ### HFM lane motion is wobbly. The angle of 3mrad
     ### permits enough X-ray reflection on the same lane
     ### in the whole beamline energy range and the 3rd
     ### harmonic is suppressed by the VFM.
     print STDOUT '  -- '.$HLname.' lane changes are not implemented.'."\n";
     $nHL = 0;
  }
  if ($have_HDM && $use_HDM) {									#Horizontal deflecting mirror lane changes
     print STDOUT '  -- '.$HDname.' lane changes:'."\n";
     for ($i=1; $i<$nHD; $i++) {
        print STDOUT '      at E='.$laneEnHD[$i].'  position: '
                               .$lanePosHD[$i-1].' -> '.$lanePosHD[$i]."\n"
                    .'                              material: '
                               .$laneMatHD[$i-1].' -> '.$laneMatHD[$i]."\n";
     }
  }
  print STDOUT $script.': #### End of current mySQL parameters #### '."\n\n";


  $mostab_disabled   = 0;
  $postab_disabled   = 0;
  $gapcheck_disabled = 0;
  $mirrors_enabled   = 0;
###------------------------------------------------------------------------
  $SIG{HUP}  = 'signal_handler';								#(1) hangup (see signum.h)
  $SIG{INT}  = 'signal_handler';								#(2) interrupt (see signum.h)
  $SIG{QUIT} = 'signal_handler';								#(3) quit (see signum.h)
# $SIG{KILL} = 'signal_handler';								#(9) this is what is sent by kill -9 -- unblockable
  $SIG{TERM} = 'signal_handler';								#(15) terminate -- sent by kill without paramaters
# $SIG{STOP} = 'signal_handler';								#(19) stop -- unblockable
### ------------------------------------------------------------------------
### Determine where are we now (at start):
  $E_beg = &getAxisAct ($monoAxis);
  if ($linkID) {$ID_beg = &getIDEnergy($undulator);}
  else         {$ID_beg = $E_beg;}

  $ESinTheta  = &getESinTheta($monoAssy);
  $Theta_beg  = &E2theta($monoAssy,$E_beg,$ESinTheta);
  $Theta_end  = &E2theta($monoAssy,$E_end,$ESinTheta);
  $Theta_step = abs($Theta_end - $Theta_beg);

  if ($tell_bluice) {&tellBluIce($scr.': started');}
  select(undef,undef,undef,0.05);         							#short sleep

  ($status,$attenuation) = &getAttenuation2($E_beg);
  if ($status) {$attenuation = 0;}

  if ($linkID) {
     ($status,$gapCheck) = Pezca::GetDouble($gapCheckPV);
     if ($status) {
        &report_only($script, $status, $gapCheckPV);
        print STDOUT '!!! '.$script.': Presuming no gap check in JBluIce.'."\n\n";
        $gapCheck = 0;
     }
  } else {$gapCheck = 0;}

  if ($linkID) {
     if ($tell_bluice) {&tellBluIce($scr.': getting ID harmonic');}
     $H_beg  = &getIDHarmonic($undulator);
  } else {
     $H_beg  = 1;
  }
  if ($nVL > 0  && $use_VFM) {$VL_beg = &getAxisAct ($laneAxisVL);}
  if ($have_HDM && $use_HDM) {$HD_beg = &getAxisAct ($laneAxisHD);}

  $diff = 1.E+10;
  for ($i=0; $i<$nUH; $i++) {
     if (abs($H_beg-$harmNum[$i]) < $diff) {
        $diff = abs($H_beg-$harmNum[$i]);
        $iUH_beg = $i;
     }
  }
  if ($use_VFM) {
     $diff = 1.E+10;
     for ($i=0; $i<$nVL; $i++) {
        if (abs($VL_beg-$lanePosVL[$i]) < $diff) {
           $diff = abs($VL_beg-$lanePosVL[$i]);
           $iVL_beg = $i;
        }
     }
  }
  if ($use_HDM) {
     $diff = 1.E+10;
     for ($i=0; $i<$nHD; $i++) {
        if (abs($HD_beg-$lanePosHD[$i]) < $diff) {
           $diff = abs($HD_beg-$lanePosHD[$i]);
           $iHD_beg = $i;
        }
     }
  }

### Determine where SHOULD we be now (at start):
  for ($i=0; $i<$nUH; $i++) {if ($E_beg > $harmEn[$i]) {$iUH_ssb = $i;}}
  if ($use_VFM) {
     for ($i=0; $i<$nVL; $i++) {if ($E_beg > $laneEnVL[$i]) {$iVL_ssb = $i;}}
  }
  if ($use_HDM) {
     for ($i=0; $i<$nHD; $i++) {if ($E_beg > $laneEnHD[$i]) {$iHD_ssb = $i;}}
  }

### Determine where should we be at end:
  for ($i=0; $i<$nUH; $i++) {if ($E_end >= $harmEn[$i]) {$iUH_end = $i;}}
  if ($use_VFM) {
     for ($i=0; $i<$nVL; $i++) {if ($E_end >= $laneEnVL[$i]) {$iVL_end = $i;}}
  }
  if ($use_HDM) {
     for ($i=0; $i<$nHD; $i++) {if ($E_end >= $laneEnHD[$i]) {$iHD_end = $i;}}
  }
  $hkl_beg = &getMonoHKL();
  if ($enmono13 > 0) {
     if ($E_beg > $enmono13) {$hkl_ssb = '333';} else {$hkl_ssb = '111';}
     if ($E_end > $enmono13) {$hkl_end = '333';} else {$hkl_end = '111';}
  } else {
     $hkl_ssb = $hkl_beg;
     $hkl_end = $hkl_beg;
  }

### Determine if the start positions differ from what they should be:
  if ($iUH_beg != $iUH_ssb && ($linkID)) {
    print STDOUT '!!! '.$script.': initial ID harmonic is '.
                 $H_beg.' instead of '.$harmNum[$iUH_ssb]."\n";
  }
  if ($use_VFM && $iVL_beg != $iVL_ssb) {
    printf STDOUT '!!! '.$script.': initial '.$VLname.' lane pos. is '.
                 '%.2f instead of '.$lanePosVL[$iVL_ssb]."\n", $VL_beg;
  }
  if ($use_HDM && $iHD_beg != $iHD_ssb) {
     printf STDOUT '!!! '.$script.': initial '.$HDname.' lane pos. is '.
                   '%.2f instead of '.$lanePosHD[$iHD_ssb]."\n", $HD_beg;
  }

### If it is a small move, we allow some hysteresis:
  $E_step = $E_end - $E_beg;
  if (abs($E_step) < $E_big) {
     $switchMono = 0;
     if ($iUH_beg == $iUH_ssb) {$iUH_end = $iUH_beg;} 						#"if" added 2016.10.31 - without it may go beyond the switch endlessly while now only once.
     if ($use_VFM) {
        if ($iVL_beg == $iVL_ssb) {$iVL_end = $iVL_beg;}					#"if" added 2016.10.31 - without it may go beyond the switch endlessly while now only once.
     }
     if ($use_HDM) {
        if ($iHD_beg == $iHD_ssb) {$iHD_end = $iHD_beg;}					#"if" added 2016.10.31 - without it may go beyond the switch endlessly while now only once.
     }
  }

### Determine destination ID offset, offset at start and where it should be at start:
  if ($linkID) {
     if ($Eupper[$iUH_ssb] != $Elower[$iUH_ssb]) {
        $harmOff_ssb = $Olower[$iUH_ssb] + (      $E_beg      - $Elower[$iUH_ssb])
                                         * ($Oupper[$iUH_ssb] - $Olower[$iUH_ssb])
                                         / ($Eupper[$iUH_ssb] - $Elower[$iUH_ssb]);
     } else {
        $harmOff_ssb = 0.5 * ($Olower[$iUH_ssb] + $Oupper[$iUH_ssb]);
     }
     if ($Eupper[$iUH_end] != $Elower[$iUH_end]) {
        $harmOff_end = $Olower[$iUH_end] + (      $E_end      - $Elower[$iUH_end])
                                         * ($Oupper[$iUH_end] - $Olower[$iUH_end])
                                         / ($Eupper[$iUH_end] - $Elower[$iUH_end]);
     } else {
        $harmOff_end = 0.5 * ($Olower[$iUH_end] + $Oupper[$iUH_end]);
     }
     $harmOff_beg = $ID_beg - $E_beg;
### Round to 1eV:
     $harmOff_ssb = int(1000*$harmOff_ssb)/1000.;
     $harmOff_end = int(1000*$harmOff_end)/1000.;
     $harmOff_beg = int(1000*$harmOff_beg)/1000.;
  } else {
     $harmOff_ssb = 0.0;
     $harmOff_end = 0.0;
     $harmOff_beg = 0.0;
  }
  $ID_end = $E_end + $harmOff_end;

  if (abs($harmOff_beg-$harmOff_ssb) > 5.0) {							#more than 5eV
     printf STDOUT '!!! '.$script.': initial ID offset is '.
                   '%.3f keV instead of %.3f keV'."\n", $harmOff_beg, $harmOff_ssb;
  }

### [Advanced] move:

### 0. Check if ID is writable:
  if ($linkID) {
     if ($tell_bluice) {&tellBluIce($scr.': checking ID writable');}
     $status  = &isIDwritable($undulator);
     if ($status == 0) {
        if ($tell_bluice) {&tellBluIce($scr.': no ID control allowed','error');}
        &zenity_error_exit('Failed to control Insertion Device from this computer.'."\n"
                          .'Only few computers are allowed to control the ID'."\n"
                          .'Typically these are ws2 and ws3. If you are changing'."\n"
                          .'energy from correct computer, then verify that the'."\n"
                          .'ID control is not disabled by the APS because of'."\n"
                          .'beam dump or other ring issues.',1);				#1=report to STDERR
      }
     elsif ($status == 2) {
        if ($tell_bluice) {&tellBluIce($scr.': no APS shutter permit','error');}
        &zenity_error_exit('No APS shutter permit at the time.',1);				#1=report to STDERR
      }
     elsif ($status != 1) {
        if ($tell_bluice) {&tellBluIce($scr.': error checking ID control permission','error');}
        &zenity_error_exit('PV access error while checking if ID is controlable.',1);		#1=report to STDERR
      }
  }

### Conditions for alignment needed (will be more related to CRL/Transfocator):
  if ($align == 1) {										#when auto-align
     if ($Theta_step > $Theta_big) {
        $align_needed = 1;
        printf STDOUT $script.': auto-align needed because'
                     .' QB_step=%.2f > QB_threshold=%.2f'."\n",
                     $Theta_step,$Theta_big;
     }
     if ($linkID  && $iUH_beg != $iUH_end) {
        $align_needed = 1;
        print STDOUT $script.': auto-align needed because ID harmonic is changing'."\n";
     }
     if ($use_VFM && $iVL_beg != $iVL_end) {
        $align_needed = 1;
        print STDOUT $script.': auto-align needed because VFM lane is changing'."\n";
     }
     if ($use_HDM && $iHD_beg != $iHD_end) {
        $align_needed = 1;
        print STDOUT $script.': auto-align needed because HDM lane is changing'."\n";
     }
     if ($switchMono && $hkl_beg ne $hkl_end) {
        $align_needed = 1;
        print STDOUT $script.': auto-align needed because Bragg reflection will be changing '.$hkl_beg.' to '.$hkl_end."\n";
     }
  }
  elsif ($align == 2) {$align_needed = 1;}							#when forced align
  else                {$align_needed = 0;}							#when no align

### ---------------------------------------------------
### 1a. Calculate upstream/downstream movement of CRL box to preserve focus:
  if ($use_CRL) {
     $device = 'CRL-old';
     $nrecords = &readCRLconfigSQLpopulated($device,\@CRLconfig);
     if ($nrecords == 1) {
        $nCRL        = $CRLconfig[0]{'Number'};
        $radiusCRL   = $CRLconfig[0]{'Radius'};
        $materialCRL = $CRLconfig[0]{'Material'};
        $CRL_StB_assy = $beamline.':CRL:StB';
        $CRL_dwstB_PV = $beamline.':CRL:ZB';
        $CRL_dwstB_beg = &getAxisAct($CRL_dwstB_PV);
     } else {
        $str = ': Unexpected number of SQL records='.$nrecords.' for device='.$device."\n"
             . 'Request to shift CRL will be ignored.';
        print STDOUT $script.$str."\n";
        if ($tell_bluice) {&tellBluIce($scr.$str);}
        $use_CRL = 0;
     }
  }
  if ($use_CRL) {
     if ($use_CRL == 1) {									#shiftCRL=calc
        $focus_beg = &calcCRLfocus($E_beg,$radiusCRL,$nCRL,$materialCRL);
        $focus_end = &calcCRLfocus($E_end,$radiusCRL,$nCRL,$materialCRL);
     } else {											#shiftCRL=SQLabs|SQLrel
        $nhash = &readCRLboxSavedPosSQL(\@hasharray);
        $status = &interpolateCRLboxPos($E_beg,'downstream',\$focus_beg,\@hasharray);		#$focus_beg is CRL position
        if ($status) {$focus_beg = 0;}
        $status = &interpolateCRLboxPos($E_end,'downstream',\$focus_end,\@hasharray);		#$focus_end is CRL position
        if ($status) {$focus_end = 0;}
     }
     if ($focus_beg > 0 && $focus_end > 0) {
        if ($use_CRL == 1) {									#shiftCRL=calc (relative shift)
           $focus_shift   = $focus_end-$focus_beg;
           $CRL_dwstB_end = $CRL_dwstB_beg - $focus_shift;
           $str = sprintf ': Adjusting CRL upstream/downstream position'
                         .' by %.2fmm to preserve focus.',-$focus_shift;
        }
        elsif ($use_CRL == 3) {									#shiftCRL=SQLrel (relative shift)
           $focus_shift   = - ($focus_end-$focus_beg);
           $CRL_dwstB_end = $CRL_dwstB_beg - $focus_shift;
           $str = sprintf ': Adjusting CRL upstream/downstream position'
                         .' by %.2fmm to preserve focus.',-($focus_end-$focus_beg);
        } else {
           $CRL_dwstB_end = $focus_end;
           $focus_shift   = -($CRL_dwstB_end - $CRL_dwstB_beg);
        }
        $str = sprintf ': Adjusting downstream CRL position from %.2fmm'
                      .' to %.2fmm by %.3fmm to preserve focus.',
                         $CRL_dwstB_beg,$CRL_dwstB_end,-$focus_shift;
        print STDOUT $script.$str."\n";
        if ($tell_bluice) {&tellBluIce($scr.$str);}
        if ($use_CRL == 1) {
           $str = sprintf ': focus_calc_old=%.2fmm, focus_calc_new=.%2fmm,'
                         .' radius=%.1fum, nCRL=%d',
                          $focus_beg, $focus_end, $radiusCRL, $nCRL;
           print STDOUT $script.$str."\n";
           if ($tell_bluice) {&tellBluIce($scr.$str);}
        } else {
           if (abs($CRL_dwstB_beg-$focus_beg) > 0.1) {
              $str = sprintf ': initial CRL position=%.2fmm differs from'
                            .' SQL data=%.2fmm by %.2fmm',
                            $CRL_dwstB_beg,$focus_beg,($CRL_dwstB_beg-$focus_beg);
              print STDOUT '!!! '.$script.$str."\n";
              if ($tell_bluice) {&tellBluIce($scr.$str,'warn');}
           }     
        }
        ($CRL_drvl, $CRL_drvh) = &getAxisDRVLH ($CRL_dwstB_PV);
        if ($CRL_drvl != 0 || $CRL_drvh != 0) {
           if ($CRL_dwstB_end < $CRL_drvl || $CRL_dwstB_end > $CRL_drvh) {
              $str = sprintf 'CRL downstream destination=%.2f is not'
                            .' within soft limits [%.2f:%.2f].',
                            $CRL_dwstB_end,$CRL_drvl,$CRL_drvh;
              $str .= "\n".'Sample may not be in focus.'.
                      "\n".'Do you still want to proceed with the energy change?';
              if (&zenity_question($str,'Yes','No') == 0) {&signal_handler('quit',2);}		#this will return "512" to calling perl script
              if    ($CRL_dwstB_end < $CRL_drvl) {$CRL_dwstB_end = $CRL_drvl;}
              elsif ($CRL_dwstB_end > $CRL_drvh) {$CRL_dwstB_end = $CRL_drvh;}
           }
        }
        $shift_val = $CRL_dwstB_end - $CRL_dwstB_beg;
        $str = sprintf ': calculated CRL box travel=%.1fmm',$shift_val;
        print STDOUT $script.$str."\n";
        if ($tell_bluice) {&tellBluIce($scr.$str);}
        ### Additional condition for re-alignment
        if ($align == 1 && $use_CRL && abs($shift_val) > $shift_big) {
           $align_needed = 1;
           printf STDOUT $script.': auto-align is needed because'
                        .' CRL_shift=%.3f > CRL_threshold=%.3f'."\n",
                        abs($shift_val),$shift_big;
        }
     } else {
        if ($use_CRL == 1) {
           $str = ': Failed calculating CRL focal distance.'
                 .' CRLs will not be moved.';
        } else {
           $str = ': Failed determining CRL downstream position from SQL.'
                 .' CRLs will not be moved.';
        }
        print STDOUT '*** '.$script.$str."\n";
        if ($tell_bluice) {&tellBluIce($scr.$str,'error');}
#       $use_CRL = 0;										#we cannot disable CRLs because they needs to be passed to alighnent
        $CRL_dwstB_end = $CRL_dwstB_beg;
     }
  }												#CRL box will be moved after changing energy


### ---------------------------------------------------
### 1b. Calculate new configuration of transfocator lenses In/Out and
###     calculate its upstream/downstream shift to preserve focus:
###     (do not change anything yet before we finish moving mono)
  if ($use_TF > 0) {										#0=no,1=rel,2=auto,3=ref,4=refNoShift
     ### Retrieve transfocator configuration from SQL
     $device = 'CRL-2020';
     $nrecords = &readCRLconfigSQLpopulated($device,\@CRLconfig);
     if ($nrecords > 0) {
        if ($E_end <= $E_exclude) {				#exclude small lenses with radius <= $r_exclude
           $nrecords = &excludeCRLradius($r_exclude,$device);
        }
        $TF_assy_PV = $beamline.':CRL:Cl';
        $TF_axis_PV = $beamline.':CRL:X';
        $TF_assH_PV = $beamline.':CRL:Ct';
        $TF_axsH_PV = $beamline.':CRL:Y';
        $TF_X_beg   = &getAxisAct($TF_axis_PV);
        $TF_H_beg   = &getAxisAct($TF_axsH_PV);
        ($TF_drvl, $TF_drvh) = &getAxisDRVLH ($TF_axis_PV);
        if ($TF_drvl == 0 && $TF_drvh == 0) {
           $str = 'Transfocator upstream/downstream motion soft limits are not set.'."\n"
                . 'Proceed without refocusing (beam focus will be lost)?';
           if (&zenity_question($str,'Yes','No') == 0) {&signal_handler('quit',2);}		#this will return "512" to calling perl script
           $use_TF = 0;
        } else {
           ### This is the center of transfocator motion interval:
           $TF_drv_center = 0.5*($TF_drvl + $TF_drvh);					#center of Up/Dn stream interval
        }
     } else {											#nrecords=0
        $str = 'Unexpected number of SQL records='.$nrecords.' for device='.$device."\n"
             . 'Proceed without refocusing (beam focus will be lost)?';
        if (&zenity_question($str,'Yes','No') == 0) {&signal_handler('quit',2);}		#this will return "512" to calling perl script
        $use_TF = 0;
     }
  }
  if ($use_TF > 0) {										#1=rel2=auto,3=ref,4=refNoShift
     ### Read all current parameters of transfocator server:
     ### ($TFserver_sts: 0=stopped, 1=processing, 2=ready)
     ($TFserver_sts,$TFserver_ref,$TFserver_rqs,
      $focus_beg,$TF_X_beg,
      $TFsrchMode,$TFrangeMax) = &getTFserverState();
     if ($TFserver_sts != 2 || $TFserver_ref < 100 || $focus_beg < 100) {
        $str = sprintf 'Transfocator server not ready:'."\n"
                      . 'Server_running=%d, focus_ref=%.2fmm, focus_act=%.2fmm.'."\n\n"
                      . 'Proceed without refocusing (beam focus will be lost)?',
                         $TFserver_sts,$TFserver_ref,$focus_beg;
        if (&zenity_question($str,'Yes','No') == 0) {&signal_handler('quit',2);}		#this will return "512" to calling perl script
        $use_TF = 0;
     }
     if ($use_TF == 2) {									#2=auto
        ### ($TFserver_sts: 0=stopped, 1=processing, 2=ready)
        if ($TFserver_sts == 2 && $TFserver_ref > 100) {$use_TF = 3;}				#3=ref
        else                                           {$use_TF = 1;}				#1=rel
     }
  }
  if ($use_TF > 0) {
     ### Requested focus at end energy:
     if    ($use_TF == 1) {$focus_rqs = $focus_beg;}						#1=rel
     elsif ($use_TF > 2)  {$focus_rqs = $TFserver_ref;}						#3=ref
     else                 {$focus_rqs = -1; $use_TF = 0;}					#0=no TF (we should not be here!)
     $str = sprintf ': initial actual focus=%.1fmm'
                   .', TFserver reference focus=%.1fmm'
                   .', new-energy target focus=%.1fmm'
                   .', TF X-drive pos.=%.1fmm',
                    $focus_beg, $TFserver_ref, $focus_rqs, $TF_X_beg;
     print STDOUT $script.$str."\n";
     if ($tell_bluice) {&tellBluIce($scr.$str);}
     if ($check_TF) {
        ### Calc actual focus for new energy using current $focus_rqs as request:
        $str = sprintf ': Calculating TF lens combination for the new energy = %.3f keV',$E_end;
        print STDOUT $script.$str."\n";
        ($focus_end,@InOut_end) = &matchTransfocatorFocus4 ($focus_rqs,$E_end,$device,\$msg,
                                                            $TFsrchMode,$TFrangeMax,$TF_axis_PV);
        if ($focus_end <= 0) {
           $str  = sprintf 'Failed to calculate Transfocator focus'
                          .' for end energy: focus=%.2f'."\n", $focus_end;
           if (length($msg) > 0) {$str .= $msg."\n";}
           $str .= 'Proceed without refocusing (beam focus will be lost)?';
           if (&zenity_question($str,'Yes','No') == 0) {&signal_handler('quit',2);}		#this will return "512" to calling perl script
           $use_TF = 0;
        } else {
           $str = sprintf ': calculated new energy lenses combination focus=%.1fmm',$focus_end;
           print STDOUT $script.$str."\n";
           if ($tell_bluice) {&tellBluIce($scr.$str);}
        }
     } else {
        $focus_end = $focus_rqs;
     }
  }
  if ($use_TF > 0 && $check_TF) {
     ### Calc $TF_X_end as a difference between the requested and calculated focuses:
     $focus_shift = $focus_rqs - $focus_end;
     $TF_X_end = $TF_drv_center + $focus_shift;
     if    ($TF_X_end < $TF_drvl) {$TF_X_end = $TF_drvl;}
     elsif ($TF_X_end > $TF_drvh) {$TF_X_end = $TF_drvh;}
     ### Transfocator travel distance to new energy:
     $shift_val = $TF_X_end - $TF_X_beg;
     $str = sprintf ': calculated TF travel=%.1fmm',$shift_val;
     print STDOUT $script.$str."\n";
     if ($tell_bluice) {&tellBluIce($scr.$str);}
     ### Additional condition for alignment:
     if ($align == 1 && $use_TF) {
        if    (abs($shift_val) > $shift_big) {
           $align_needed = 1;
           printf STDOUT $script.': TF auto-align is needed because'
                        .' |TF_shift|=%.3f > TF_threshold=%.3f'."\n",
                        abs($shift_val),$shift_big;
        }
     }
  }	#Transfocator will be reconfigured and moved after changing energy


### ---------------------------------------------------
### 1c. Check if sample could be fried: ask operator to remove or move if needed
  if (($align == 1 && $align_needed)         ||							#when auto-align
      ($align == 2)                          ||							#when forced align
      (&getShutter($shutter_readback) == 1)) {
     &protect_sample_fried();									#this will exit if no consent
  }

  $t0 = [gettimeofday];										#start measuring time

### ---------------------------------------------------
### 2a. Disable beamline feedback:
### What mode are we running (0=software,1=hybrid,2=hardware(lockin))
  $feedback_mode = &getMostabMode ($beamline);
  if ($feedback_mode < 0 || $feedback_mode > 2) {
     printf STDOUT '*** '.$script.': incorrect intensity feedback mode=%d, using mode=0'."\n", $feedback_mode;
     $feedback_mode = 0;
  }
  printf STDOUT $script.': intensity feedback mode = %s'."\n", 
                        $feedbackLabels[$feedback_mode];

  if ($feedback_mode <  2) {
     if (!defined $syncMode) {$syncMode = 'noSync';}						#default for feedback software/hybrid feedback
### We may still want to keep sync modes for updating attenuation during energy change
#    if ($syncMode !~ /^noSync$/i) {
#       print STDOUT "\n"
#          .'!!! '.$script.': The requested ID syncMode='.$syncMode.' is ignored because'."\n"
#          .'the intensity feedback is not in the hardware mode mode and will be paused.'."\n"
#          .'The "noSync" mode will be used'."\n";
#       $syncMode = 'noSync';									#no sense to sync mono with ID if feedback will be turned off (soft or hybrid mode)
#    }
  }
  elsif ($feedback_mode == 2) {									#if lockin feedback mode
     if (!defined $syncMode) {$syncMode = 'multiStep';}						#default for lockin feedback
#    if (!defined $syncMode) {$syncMode = 'IDsteps';}						#default for lockin feedback
#    if (!defined $syncMode) {$syncMode = 'IDscan';}						#default for lockin feedback
#    if (!defined $syncMode) {$syncMode = 'noSync';}						#default for lockin feedback
  }
  else {
     print STDOUT "\n\n"
       .'!!! '.$script.': Unexpected Mostab mode='.$feedback_mode
                                .'. Presuming "0" (soft feedback)'."\n\n\n";
     $feedback_mode = 0;
     $syncMode = 'noSync';
  }
  print STDOUT $script.': the monochromator-undulator'
              .' synchronization mode will be: '.$syncMode."\n";

  $mostab_initial = &getMostabState ($beamline);						#0=run, 1=pause, -1=error
  if ($mostab_initial == -1) {
     $str = ': Cannot talk to MOSTAB.';
     print STDOUT '*** '.$script.$str."\n";
     if ($tell_bluice) {&tellBluIce($scr.$str,'error');}
     $mostab_initial = 1;									#presume paused if failure
  }
  $postab_initial = &getPostabState ($beamline);
  if ($postab_initial == -1) {
     $str = ': Cannot talk to POSTAB.';
     print STDOUT '*** '.$script.$str."\n";
     if ($tell_bluice) {&tellBluIce($scr.$str,'error');}
     $postab_initial = 3;									#presume paused if failure
  }

  if (($linkID && $iUH_beg != $iUH_end) || ($Theta_step > $Theta_big)) {
     if ($mostab_initial != 1 && ($feedback_mode != 2 || $syncMode eq 'noSync')) {		#if not paused
        $str = ': Pausing I0 feedback';
        print STDOUT $script.$str."\n";
        if ($tell_bluice) {&tellBluIce($scr.$str);}
        $status = &pauseMostab ($beamline);
        $mostab_disabled = 1;
     }
     if (($use_VFM && $iVL_beg != $iVL_end) || 
         ($use_HDM && $iHD_beg != $iHD_end)) {
        if ($postab_initial != 3) {								#if not paused
           $str = ': Pausing positional feedback';
           print STDOUT $script.$str."\n";
           if ($tell_bluice) {&tellBluIce($scr.$str);}
           $status = &pausePostab ($beamline);
           $postab_disabled = 1;
        }
     }
  }
  else {
     if ((! $linkID || $iUH_beg == $iUH_end) && $Theta_step <= $Theta_big) {
        $str = sprintf ': no feedback pausing required since no ID harmonic'
             .' change and dTheta=%.3f < threshold=%.3f',$Theta_step, $Theta_big;
        print STDOUT $script.$str."\n";
        if ($tell_bluice) {&tellBluIce($scr.$str);}
     }
     elsif ((! $use_VFM || $iVL_beg == $iVL_end) &&
            (! $use_HDM || $iHD_beg == $iHD_end)) {
        $str = ': no positional feedback pausing required since no mirror lane change';
        print STDOUT $script.$str."\n";
        if ($tell_bluice) {&tellBluIce($scr.$str);}
     }
  }

### ---------------------------------------------------
### 2b. Enable and start moving mirrors:
  if ($use_VFM && $iVL_end != $iVL_beg) {
     $mirrors_enabled = 1;
     $status = &configAssy ($laneAssyVL);
     if ($status) {&signal_handler('quit');}
     $str = sprintf ': Starting motion of %s from %.2f to %.2f (from %s to %s lane)...',
            $VLname, $VL_beg, $lanePosVL[$iVL_end], $laneMatVL[$iVL_beg], $laneMatVL[$iVL_end];
     print STDOUT $script.$str."\n"; 
     if ($tell_bluice) {&tellBluIce($scr.$str);}
     $status = &startNMotors ($laneAssyVL, 1, [$laneAxisVL],
                             [$lanePosVL[$iVL_end]], 'checkAuto');
     if ($status) {
### Report error to BluIce:
        $str = ': Failed starting ['.$VLname.'], status='.$status;
        print STDOUT '*** '.$script.$str."\n"; 
        if ($tell_bluice) {&tellBluIce($scr.$str,'error');}
     }
  }
  if ($use_HDM && $iHD_end != $iHD_beg) {
     $mirrors_enabled = 1;
     $status = &configAssy ($laneAssyHD);
     if ($status) {&signal_handler('quit');}
     $str = sprintf ': Starting motion of %s from %.2f to %.2f (from %s to %s lane)...',
            $HDname, $HD_beg, $lanePosHD[$iHD_end], $laneMatHD[$iHD_beg], $laneMatHD[$iHD_end];
     print STDOUT $script.$str."\n"; 
     if ($tell_bluice) {&tellBluIce($scr.$str);}
     $status = &startNMotors ($laneAssyHD, 1, [$laneAxisHD], [$lanePosHD[$iHD_end]], 'checkAuto');
     if ($status) {
### Report error to BluIce:
        $str = ': Failed starting ['.$HDname.'], status='.$status;
        print STDOUT '*** '.$script.$str."\n"; 
        if ($tell_bluice) {&tellBluIce($scr.$str,'error');}
     }
  }

### ---------------------------------------------------
### 3. Run mono:
  if ($gapCheck) {
     print STDOUT $script.': Disabling JBluIce gap check.'."\n";
     $gapcheck_disabled = 1;
     $status = Pezca::PutDouble($gapCheckPV,0);
     &report_only($script, $status, $gapCheckPV);
  }
  $attempts = 3;
  $attempt = 0;
  do {
     $attempt++;
     $status = &configAssy($monoAssy);
     if ($status) {&signal_handler('quit');}
     if ($attempt == 1) {
        $status = &setAxisHiSpeed($monoAssy,$monoAxis);						#just in case set the speed to max
        if ($status) {&signal_handler('quit');}
     }
     if ($tell_bluice) {
        &tellBluIce($scr.': starting energy move ('.$attempt.'/'.$attempts.')');
     }
     printf STDOUT $script.': Starting motion of Energy from %.3f to %.3f...'."\n",
                                                             $E_beg, $E_end;
     printf STDOUT '(Bragg angle from %.2f to %.2f -- by %.2f)'."\n",
                                        $Theta_beg, $Theta_end, $Theta_step;
     if ($linkID) {
        printf STDOUT '(ID energy: from %.3f keV to %.3f keV)'."\n",
                                                  $ID_beg, $ID_end;
        printf STDOUT '(ID offset: from %.3f keV to %.3f keV)'."\n",
                                                  $harmOff_beg, $harmOff_end;
        if ($harmNum[$iUH_beg] != $harmNum[$iUH_end]) {
           print STDOUT '(ID harmonic: from '.$harmNum[$iUH_beg]
                                      .' to '.$harmNum[$iUH_end].')'."\n";
           if ($tell_bluice) {
              &tellBluIce($scr.': ID harmonic '.$harmNum[$iUH_beg]
                                       .' -> '.$harmNum[$iUH_end]);
           }
        } else {
           print  STDOUT '(ID harmonic = '.$harmNum[$iUH_end].")\n";
        }
     }
     if ($iUH_end != $iUH_beg) {$switchID = 1;} 
     else                      {$switchID = 0;}
     $status = &runMono($monoAxis, $monoAssy, $E_end, 
                        $linkID, $undulator, undef,			# $IDoffsetEnd=undef, use SQL
                        $syncMode, $syncStep, $switchID,
                        $attenuation, $switchMono,
                        \$enmono13, \@harmEn, \@harmNum, 
                        \@Elower, \@Eupper, \@Olower, \@Oupper);
     if ($status) {
        $str = sprintf ': Failed to move Energy to %.3f: status=%d (attempt %d/%d)', 
                                             $E_end, $status, $attempt, $attempts;
        print STDOUT '*** '.$script.$str."\n";
        if ($tell_bluice) {&tellBluIce($scr.$str,'error');}
     }
  } while ($status && $attempt < $attempts);

### ---------------------------------------------------
### Restore gap check:
  if ($gapCheck) {
     print STDOUT $script.': Re-enabling BluIce gap check.'."\n";
     $gapcheck_disabled = 0;
     $status = Pezca::PutDouble($gapCheckPV,1);
     &report_only($script, $status, $gapCheckPV);
  }

  if ($tell_bluice) {&tellBluIce($scr.': energy motion complete.');}
  select(undef,undef,undef,0.05);         #short sleep

### ---------------------------------------------------
### Set attenuation to the value of the old energy:
  if ($attenuation > 0) {
     print STDOUT $script.': Restoring attenuation...'."\n";
     if ($tell_bluice) {&tellBluIce($scr.': restoring attenuation');}
     ($status,$attenuation_resulted) = &setAttenuation2($attenuation,$E_end);
     if ($status) {
        $str = sprintf ': Failed to set attenuation to %.2gx at new energy',$attenuation;
        print STDOUT '*** '.$script.$str."\n";
        if ($tell_bluice) {&tellBluIce($scr.$str,'warn');}
     }
  }

### ---------------------------------------------------
### Move CRL box to preserve focus:
  if ($use_CRL && abs($shift_val) > 0) {
     $str = sprintf ': Starting motion of CRL box, PV=%s from %.3f to %.3f...',
                      $CRL_dwstB_PV, $CRL_dwstB_beg, $CRL_dwstB_end;
     print STDOUT $script.$str."\n";
     if ($tell_bluice) {&tellBluIce($scr.$str);}
     $status = &startNMotors ($CRL_StB_assy, 1, [$CRL_dwstB_PV], [$CRL_dwstB_end], 'checkAuto');
     if ($status) {
        ### Report error to BluIce:
        $str = sprintf ': Failed starting assy=[%s] and axis=[%s] to %.2f, status=%d', 
                       $CRL_StB_assy, $CRL_dwstB_PV, $CRL_dwstB_end, $status;
        print STDOUT '*** '.$script.$str."\n";
        if ($tell_bluice) {&tellBluIce($scr.$str,'error');}
#       $use_CRL = 0;										#we cannot disable CRLs because it needs to be passed to alighnent
        $CRL_dwstB_end = $CRL_dwstB_beg;
     }
  }

### ---------------------------------------------------
### Set Transfocator focus at new energy to match the focus value of the old or requested energy:
  if ($use_TF > 0) {
     $str = ': Reconfiguring Transfocator lenses In/Out and X-pos.';
     print STDOUT $script.$str."\n";
     if ($tell_bluice) {&tellBluIce($scr.$str);}
     ($focus_end,$TF_X_end,$msg) = &setTFserverFocus($focus_rqs);				#we can also add max wait time here
     if (abs($focus_rqs - $focus_end) > 1 || $focus_end <= 0 || length($msg) > 0) {
        $str = sprintf '%s *** %s: error setting lenses for Transfocator focus:'."\n"
                      .'Requested focal distance=%.1fmm,'."\n"
                      .'Actual    focal distance=%.1fmm.'."\n"
                      .'Focusing mode: %s.'."\n"
                      .'When retuning beamline from %.2f keV to %.2f keV.',
                        scalar(localtime), $script, $focus_rqs, $focus_end, 
                        $use_TF_labels[$use_TF], $E_beg, $E_end;
        if (length($msg) > 0) {$str .= "\n".'--------------'."\n".$msg;}
        print STDOUT $str."\n";
        &zenity_error($str,'nowait');
     } else {
        $str = sprintf ': Transfocator is reconfigured to new energy:'
                      .' rqs=%.1fmm, act=%.1fmm.'."\n"
                      .' Focusing mode = %s',
                        $focus_rqs, $focus_end, $use_TF_labels[$use_TF];
        print STDOUT scalar(localtime).' '.$script.$str."\n";
        if ($tell_bluice) {&tellBluIce($scr.$str);}
        ### We preliminary calculated $shift_val before energy change.
        ### Now, let us repeat the calculation in case if final 
        ### $TF_X_end deviate from the preliminary calculations:
        $shift_val = $TF_X_end - $TF_X_beg;
        $str = sprintf ': final TF travel=%.1fmm',$shift_val;
        print STDOUT $script.$str."\n";
        if ($tell_bluice) {&tellBluIce($scr.$str);}
        if ($align == 1 && $check_TF == 0 && $align_needed == 0 && abs($shift_val) > $shift_big) {
### We did not check for TF shift because of $check_TF=0 and thus we
### may have overlooked that alignment may be needed and sample may
### need to be removed. What can we do now? -- SS 2021/10/04
           $align_needed = 1;
           printf STDOUT $script.': TF auto-align is needed because'
                        .' |TF_shift|=%.3f > TF_threshold=%.3f'."\n",
                        abs($shift_val),$shift_big;
        }
     }
     if (abs($shift_val) > 0 && abs($TF_H_misalignRad) > 0) {
        $TF_H_shift = - $shift_val * $TF_H_misalignRad;
        printf STDOUT $script.': Transfocator lateral compensation due to'
                             .' shift along beam = %.1fum'."\n",
                             1000*$TF_H_shift;
        $TF_H_end = $TF_H_beg + $TF_H_shift;
        $str = sprintf ': Laterally moving Transfocator to compensate for misalignment,'
                      .' PV=%s from %.3fmm to %.3fmm (by %.1fum) ...',
                       $TF_axsH_PV, $TF_H_beg, $TF_H_end, 1000*$TF_H_shift;
        print STDOUT $script.$str."\n";
        if ($tell_bluice) {&tellBluIce($scr.$str);}
        $status = &runNMotors ($TF_assH_PV, 1, [$TF_axsH_PV], [$TF_H_end], 'checkAuto');
        if ($status) {
           ### Report error to BluIce:
           $str = sprintf ': Failed starting assy=[%s] and axis=[%s] to %.2f, status=%d', 
                          $TF_assy_PV, $TF_axis_PV, $TF_H_end, $status;
           print STDOUT '*** '.$script.$str."\n";
           if ($tell_bluice) {&tellBluIce($scr.$str,'error');}
        }
     }
  }

### ---------------------------------------------------
### 4a. Wait for mirrors:
  if ($use_HDM && $iHD_end != $iHD_beg) {
     print STDOUT $script.': Waiting for ['.$HDname.']...'."\n";
     $motionTime = &calcMotorTime($laneAssyHD, $laneAxisHD, $lanePosHD[$iHD_end]);
     $msg = '';
     $status = &trackMotor($laneAssyHD, $motionTime, $laneAxisHD, \$msg);
     if ($status) {
### Report error to BluIce:
        if ($tell_bluice) {&tellBluIce($scr.': '.$HDname.' track error','warn');}
        $actpos = &getAxisAct($laneAxisHD);
        $str  = ': Tracking '.$HDname.' lane change detected a problem, status='.$status."\n";
        $str .= sprintf 'Requested position=%.2fmm, actual position=%.2fmm.'."\n",$lanePosHD[$iHD_end],$actpos;
        $str .= '*** Incorrect mirror lane may result in significantly lowered beam intensity ***'."\n";
        $str .= $msg;
#       print STDOUT '*** '.$script.$str."\n";
#       &zenity_error('*** '.$script.$str,undef,'nowait');
        &zenity_error('*** '.$script.$str);
     }
  }
  if ($use_VFM && $iVL_end != $iVL_beg) {
     print STDOUT $script.': Waiting for ['.$VLname.']...'."\n";
     $motionTime = &calcMotorTime($laneAssyVL, $laneAxisVL, $lanePosVL[$iVL_end]);
     $msg = '';
     $status = &trackMotor($laneAssyVL, $motionTime, $laneAxisVL, \$msg);
     if ($status) {
### Report error to BluIce:
        if ($tell_bluice) {&tellBluIce($scr.': '.$VLname.' track error','warn');}
        $actpos = &getAxisAct($laneAxisVL);
        $str  = ': Tracking '.$VLname.' detected a problem, status='.$status."\n";
        $str .= sprintf 'Requested position=%.2fmm, actual position=%.2fmm.'."\n",$lanePosVL[$iVL_end],$actpos;
        $str .= '*** Incorrect mirror lane may result in significantly lowered beam intensity ***'."\n";
        $str .= $msg;
#       print STDOUT '*** '.$script.$str."\n";
#       &zenity_error('*** '.$script.$str,undef,'nowait');
        &zenity_error('*** '.$script.$str);
     }
  }

### ---------------------------------------------------
### 4b1. Wait for CRL box:
  if ($use_CRL) {
     if ($CRL_dwstB_beg ne $CRL_dwstB_end) {
        printf STDOUT $script.': Waiting for assy=[%s] and axis=[%s]...'."\n", 
                               $CRL_StB_assy, $CRL_dwstB_PV;
        $motionTime = &calcMotorTime($CRL_StB_assy, $CRL_dwstB_PV, $CRL_dwstB_end);
        $msg = '';
        $status = &trackMotor($CRL_StB_assy, $motionTime, $CRL_dwstB_PV, \$msg);
        if ($status) {
### Report error to BluIce:
           if ($tell_bluice) {&tellBluIce($scr.': '.$CRL_dwstB_PV.' track error','warn');}
           $str = ': Tracking '.$CRL_dwstB_PV.' detected a problem, status='.$status."\n".$msg;
#          print STDOUT '*** '.$script.$str."\n";
#          &zenity_error('*** '.$script.$str,undef,'nowait');
           &zenity_error('*** '.$script.$str);
        }
     }
  }

### ---------------------------------------------------
### 4b2. Wait for Transfocator:
  if ($use_TF) {
     if ($use_TF < 4 && ($TF_X_beg ne $TF_X_end)) {
        print STDOUT $script.': Waiting for ['.$TF_assy_PV.']...'."\n";
        $motionTime = &calcMotorTime($TF_assy_PV, $TF_axis_PV, $TF_X_end);
        $msg = '';
        $status = &trackMotor($TF_assy_PV, $motionTime, $TF_axis_PV, \$msg);
        if ($status) {
### Report error to BluIce:
           if ($tell_bluice) {&tellBluIce($scr.': '.$TF_axis_PV.' track error','warn');}
           $str = ': Tracking '.$TF_axis_PV.' detected a problem, status='.$status."\n".$msg;
#          print STDOUT '*** '.$script.$str."\n";
#          &zenity_error('*** '.$script.$str,undef,'nowait');
           &zenity_error('*** '.$script.$str);
        }
     }
  }

### ---------------------------------------------------
### 5. Tell new calibration to Pilatus or Eiger:
  if (abs($E_step) > 0.5 && ($detector =~ /^pilatus$/i || $detector =~ /^eiger$/i)) {
### Change Pilatus or Eiger calibration if the energy changed by more than 500eV
     $str = ': sending energy update command to '.$detector.' ...';
     print STDOUT $script.$str."\n";
     if ($tell_bluice) {&tellBluIce($scr.$str);}
     $str = int(1000*$E_end+0.5);								#round to 1eV
     if ($detector =~ /^pilatus$/i) {
        $detectorPV = $beamline.':fpe:DetCmdStr';
        ($status,$nelm) = Pezca::GetDouble($detectorPV.'.NELM');
        if ($status != 6) {
           $status = &PezcaPutStringWF($detectorPV,'SetEnergy '.$str,'oldca');
        }
     }
     elsif ($detector =~ /^eiger$/i) {
        $detectorPV = $beamline.':fpe:PhotonEnergy';
        $status = Pezca::PutDoubleOldCa($detectorPV,$str);
     }
     if ($status) {
#       &report_only($script, $status, $detectorPV);
        $str = ': Failed setting '.$detector.' for new energy (PV='
              .$detectorPV.'). Most likely FPE is off.';
        print STDOUT '!!! '.$script.$str."\n\n";
        if ($tell_bluice) {&tellBluIce($scr.$str,'warn');}
     }
  }

### ---------------------------------------------------
### 6. Disable mirrors:
  if ($mirrors_enabled && ($beamline eq '23i' || $beamline eq '23o')) {
     print STDOUT $script.': Disabling mirrors motors...'."\n";
     if ($tell_bluice) {&tellBluIce($scr.': disabling mirrors');}
     $mirrors_enabled = 0;
     @prg = ('perl', '-W', $PEZCA.'/kill_kbm.pl', $beamline);
     if ($^O !~ /Win/i) {push @prg, '&'; $str=join(' ',@prg); $status=system($str);}
     else               {$status=system(1,@prg);}
     if ($status) {
        print STDOUT "\n".'*** '.$script.': Failed to execute kill_kbm.pl'."\n";
        if ($tell_bluice) {&tellBluIce($scr.': cannot disable mirrors','warn');}
     }
  }

### ---------------------------------------------------
### 7. Re-enable feedback:
  if (($linkID && $iUH_beg != $iUH_end) || $Theta_step > $Theta_big) {
     if ($mostab_initial != 1 && $mostab_disabled != 0) {					#if not paused initially (#0=run, 1=pause, -1=error)
        $mostab_disabled = 0;
        $mostab_search_ini = &getMostabSearchOption ($beamline);
        if ($mostab_search_ini != 1) {$mostab_search_ini = 0;}
        $mostab_search_opt = $mostab_search_ini;
        if ($mostab_search_ini == 0 && $Theta_step > $Theta_bigbig) {
           print STDOUT $script.': Turning ON search with P2 in the I0'
                                                        .' feedback...'."\n";
           $status = &setMostabSearchOption ($beamline,1);
           if ($status == 0) {$mostab_search_opt = 1;}
        }
        print STDOUT $script.': Restoring I0 feedback...'."\n";
        if ($tell_bluice) {&tellBluIce($scr.': restoring I0 feedback');}
        $status = &restoreMostab ($beamline);
        if ($status) {
           print STDOUT "\n".'*** '.$script.': Failed to restart MOSTAB'."\n";
           if ($tell_bluice) {&tellBluIce($scr.': MOSTAB failed','warn');}
           $align = -1;
        }
        if ($align != -1) {
           print STDOUT $script.': Waiting for I0 lock...'."\n";
           if ($tell_bluice) {&tellBluIce($scr.': waiting I0 lock');}
           $wait_ = $mostab_wait_max+$mostab_wait_extra*$mostab_search_opt;
           $status = &waitMostabLock ($beamline, $wait_);
           if ($status) {
              print STDOUT "\n".'*** '.$script.': Failed to get MOSTAB lock'."\n";
              if ($tell_bluice) {&tellBluIce($scr.': I0 lock failed','warn');}
              $align = -1;
           } else {
              print STDOUT $script.': I0 locked.'."\n";
              if ($tell_bluice) {&tellBluIce($scr.': I0 locked');}
           }
        }
        if ($mostab_search_ini == 0 && $mostab_search_opt == 1) {
           print STDOUT $script.': Turning OFF search with P2 in the I0 feedback...'."\n";
           $status = &setMostabSearchOption ($beamline,0);
        }
     }
     if (($use_VFM && $iVL_beg != $iVL_end) ||
         ($use_HDM && $iHD_beg != $iHD_end)) {
        if ($postab_initial != 3) {								#if positional feedback was not in "pause" before the script turned it off (soft/hybrid mode)
           print STDOUT $script.': Restoring positional feedback...'."\n";
           if ($tell_bluice) {&tellBluIce($scr.': restoring pos.feedback');}
           $postab_disabled = 0;
           $status = &restorePostab ($beamline);
           if ($status) {
              print STDOUT "\n".'*** '.$script.': Failed to restart POSTAB'."\n";
              if ($tell_bluice) {&tellBluIce($scr.': POSTAB failed','warn');}
#             $align = -1;
           }
        }
     }
  }
  else {
     if ((! $linkID || $iUH_beg == $iUH_end) && ($Theta_step <= $Theta_big)) {
        printf STDOUT $script.': no feedback restore required since no'
             .' ID harmonic change and dTheta=%.3f < threshold=%.3f'."\n",
                                                    $Theta_step, $Theta_big;
     }
     elsif ((! $use_VFM || $iVL_beg == $iVL_end) &&
            (! $use_HDM || $iHD_beg == $iHD_end)) {
        print STDOUT $script.': no positional feedback restore required'
                    .' since no mirror lane change'."\n";
     }
  }

### ---------------------------------------------------
### 8. Performing optimization:

  if (($align == 1 && $align_needed) || ($align == 2)) {
     if ($recoveryTime eq 'auto') {
        $recoveryTime = 0;
        ### Si (111) Bragg angles:
        ###     6keV=19.2 9keV=12.7 12keV=9.5 15keV=7.6 18keV=6.3
        ### Method-1 of calculating recovery time: use change of Bragg
        ### angle to make a proportion knowing that going from 6keV to
        ### 12keV corresponds to  Bragg angle change by ~10 degr and
        ### sleep time of 150s.
        $recoveryTime = int(150*($Theta_step/10.));
        if ($linkID) {
           if ($iUH_beg != $iUH_end && $recoveryTime < 120.) {$recoveryTime = 120.;}
        }
        #
        ### Method-2 of calculating recovery time: scale the time
        ### to the undulator gap change (need to read gaps):
#       $recoveryTime = int(150*abs($Gap_step/($Gap_12-$Gap_6)));
        #
        ### Method-3 of calculating recovery time: scale the time to
        ### the crystal temperature change (need to read thermocouples):
#       $recoveryTime = int(150*abs($T_step/($T_12-$T_6)));
        #
        ### Method-4 of calculating recovery time: scale the time to
        ### the crystal temperature change based on Bob's temperature
        ### rates curves
        if ($linkID) {
           $recoveryTime = &mono_stabilization_time($ID_beg, $ID_end, $harmEn[1]);	#energy of ID 1<->3 transition
        }
     }
     if ($recoveryTime > 0) {
        printf STDOUT $script.': Waiting for %ds to let the monochromator'
                             .' settle thermo stability...'."\n",$recoveryTime;
#       $recoveryBin = 0.1;
#       $recoveryCnt = int($recoveryTime/$recoveryBin);
#       for ($i=0; $i<$recoveryCnt; $i++) {select(undef, undef, undef, $recoveryBin);}
        $status = &zenity_progress($recoveryTime,'Waiting monochromator to settle thermo stability');
        print STDOUT $script.': ... done waiting for monochromator settling.'."\n";
     }
     print STDOUT $script.': Aligning beam...'."\n";
     if ($tell_bluice) {&tellBluIce($scr.': aligning beam');}
     if ($align_stretch eq 'auto') {
        if ($Theta_big > 0 && $Theta_step > 0) {
           $align_factor = sprintf '%.1f', ($Theta_step/$Theta_big);
           if ($align_factor < 1) {$align_factor = 1;}
        } else {
           $align_factor = 1;
        }
        if ($align_factor > 1) {
           printf STDOUT '!!! '.$script.': as the Bragg angle changed by %.2fdegr., '
             .'the alignment range is increased by %.1f times compared to normal.'."\n",
                                                       $Theta_step, $align_factor;
        }
     }
     elsif ($align_stretch eq 'none')  {$align_factor=1;}
     if (!defined $align_factor) {$align_factor=1;}
     if ($align_factor == 0)     {$align_factor=1;}
     $align_factor = abs($align_factor);

     $status = &read_jbluice_config_flags('alignMinibeamFlags', \@alignFlags);
### Remove flags that we want to replace with our settings:
     @alignFlags = grep(! /^${beamline}$/i, @alignFlags);
     @alignFlags = grep(! /bluice$/i,       @alignFlags);		#tellBluIce|bluice|nobluice
     @alignFlags = grep(! /sample/i,        @alignFlags);		#ignoreSample|warnSample|moveSample=xx[h|v]|moveSampleNoWarn=xx[h|v]
     @alignFlags = grep(! /gnuplot/i,       @alignFlags);		#gnuplot|gnuplot=image|nognuplot
     @alignFlags = grep(! /medm$/i,         @alignFlags);		#medm|nomedm
     @alignFlags = grep(! /scan$/i,         @alignFlags);		#fastScan|stepScan
     @alignFlags = grep(! /KBM$/i,          @alignFlags);		#useKBM|noKBM
     @alignFlags = grep(! /CRL$/i,          @alignFlags);		#useCRL|noCRL
     @alignFlags = grep(! /TF$/i,           @alignFlags);		#useTF|noTF
### Insert our flags:
     push(@alignFlags, $beamline);
     push(@alignFlags, $gnuplot);
     push(@alignFlags, $medm);
     push(@alignFlags, $align_scanmode);
     if (! grep(/^autoPinholeOrGuardSlits$/i, @alignFlags)) {push(@alignFlags,'autoPinholeOrGuardSlits');}

     if ($align_factor != 1) {
        @alignFlags = grep(! /^HalfRange/i, @alignFlags);		#HalfRangeV|HalfRangeVfactor|HalfRangeH|HalfRangeHfactor
        push(@alignFlags, 'HalfRangeHfactor='.$align_factor);
        push(@alignFlags, 'HalfRangeVfactor='.$align_factor);
     }
     if    ($align_Mode eq 'KBM') {push(@alignFlags,'useKBM');}
     elsif ($align_Mode eq 'CRL') {push(@alignFlags,'useCRL');}
     elsif ($align_Mode eq 'TF')  {push(@alignFlags,'useTF');}
     if ($tell_bluice)            {push(@alignFlags,'bluice');}
     else                         {push(@alignFlags,'nobluice');}
     @prg = ('perl', '-W', $PEZCA.'/alignMiniBeam.pl');
     if ($use_xterm) {unshift(@prg,$xterm);}
     print STDOUT $script.': Starting alignMiniBeam.pl '."@alignFlags\n";
     push(@prg,@alignFlags);
     $status = &execute_external_script(\@prg);
     if ($status) {
        print STDOUT "\n".'*** '.$script.': Failed executing alignMiniBeam.pl'."\n";
        if ($tell_bluice) {&tellBluIce($scr.': failed to align beam','warn');}
     } else {
        $msg = ': Beam optimization is complete.';
        print STDOUT $script.$msg."\n";
        if ($tell_bluice) {&tellBluIce($scr.$msg,'warn'); select(undef,undef,undef,0.2);}	#short sleep
     }
  }
  else {
     if ($align == 0) {
        print  STDOUT $script.': beam aligning is disabled by cmdline option'."\n";
     }
     elsif ($align == -1) {
        print  STDOUT $script.': beam aligning is skipped because MOSTAB/POSTAB failed'."\n";
        if ($tell_bluice) {&tellBluIce($scr.': beam align skipped','warn');}
     }
     else {
        print  STDOUT $script.': no beam aligning required since:'."\n";
        printf STDOUT $script.':   -- dTheta=%.3f < threshold=%.3f'."\n", $Theta_step, $Theta_big;
        if ($use_HFM || $use_VFM || $use_HDM) {print  STDOUT $script.':   -- no lane/harmonic change'."\n";}
        if ($tell_bluice) {&tellBluIce($scr.': beam aligning not needed','warn');}
     }
  }

  $elapsed = tv_interval ($t0);									#measure time (in seconds) - exclude sample returning time

### ---------------------------------------------------
### 9. Returning the sample if needed:

  if (($align == 1 && $align_needed)         ||							#when auto-align
      ($align == 2)                          ||							#when forced align
      (&getShutter($shutter_readback) == 1)) {
     if ($pin_mounted) {									#if pin was originally present
        $zzz = $ZENITY_VIA_BLUICE;
        if (! defined $JBLUICE_SERVER || ! $tell_bluice) {$ZENITY_VIA_BLUICE = 0;}
        if ($protectSample == 1) {								#1=warn, 2=auto-move 3=auto-move without warning
           $msg = 'Energy change completed.'."\n"
                . 'You can put crystals into the beam.'."\n";
#          &zenity_info($msg, 'nowait');
           &zenity_info($msg);
        }
        elsif ($protectSample > 1 && defined $sample_moved && $moveSample != 0) {		#2=auto-move (first ask) 3=auto-move without warning
           if    ($direction eq 'UP')       {$direction = 'DOWN';}
           elsif ($direction eq 'DOWN')     {$direction = 'UP';}
           elsif ($direction eq 'OUTBOARD') {$direction = 'INBOARD';}
           elsif ($direction eq 'INBOARD')  {$direction = 'OUTBOARD';}
           $msg = sprintf 'Moving sample back (%s) by %.1fum.',$direction,abs(1000*$moveSample);
           if ($protectSample == 2) {								#2=auto-move (first ask)
              $msg = 'Energy change completed.'."\n".$msg;
              &zenity_info($msg, 'nowait');
           } else {										#3=auto-move without warning
              print STDOUT $script.': '.$msg."\n";
              if ($tell_bluice) {&tellBluIce($scr.': '.$msg);}
           }
           if ($xyzAssyPV eq $zAssyPV) {$i=3;} else {$i=2;}
           $status = &startNMotors ($xyzAssyPV, $i, \@xyzAxesPV, \@xyzIni);
           if ($status) {
              print STDOUT '!!! '.$script.': Failed to put the pin back into the beam'
                               .' (assy='.$xyzAssyPV.' status='.$status.')!'."\n\n";
              if ($tell_bluice) {&tellBluIce($scr.': Failed to put the pin back into the beam!','warn');}
           }
           if ($i < 3) {
              $status = &startNMotors ($zAssyPV, 1, [$xyzAxesPV[2]], [$xyzIni[2]]);
              if ($status) {
                 print STDOUT '!!! '.$script.': Failed to put the pin back into the beam'
                                  .' (assy='.$zAssyPV.' status='.$status.')!'."\n\n";
                 if ($tell_bluice) {&tellBluIce($scr.': Failed to put the pin back into the beam!','warn');}
              }
           }
        }
        else {
           $msg = ': Energy change is complete. You can put crystals into the beam.';
           print STDOUT $script.$msg."\n";
           if ($tell_bluice) {&tellBluIce($scr.$msg);}
        }
        $ZENITY_VIA_BLUICE = $zzz;
     } else {
        $msg = ': Energy change is complete. You can put crystals into the beam.';
        print STDOUT $script.$msg."\n";
        if ($tell_bluice) {&tellBluIce($scr.$msg);}
     }
  }

  $msg = sprintf ': Energy change from %.3f keV to %.3f keV finished at [%s].'
        .' Elapsed time=%.2fmin', $E_beg, $E_end, scalar(localtime), $elapsed/60.;
  print STDOUT $script.$msg."\n";
  if ($tell_bluice) {&tellBluIce($scr.$msg);}

  &clearAllCaMonitors();
  exit 0;

#=================================================================

sub protect_sample_fried() {

  if (($align == 1 && $align_needed)         ||							#when auto-align
      ($align == 2)                          ||							#when forced align
      (&getShutter($shutter_readback) == 1)) {
     print STDOUT $script.': Checking if sample protection is needed...'."\n\n";
  } else {return;}

  if ($pinsensor_present) {
     ($status,$pin_mounted) = &pinMounted($pinsensor_PV);
     if ($status) {
        print STDOUT '!!! '.$script.': Presuming pin is mounted.'."\n\n";
        $pin_mounted = 1;									#pin is mounted
     }
  } else {
     $pin_mounted = 1;										#pin is mounted
  }

  if ($pin_mounted) {
     if ($protectSample == 1) {									#1=warn sample
        if (($align == 1 && $align_needed) || ($align == 2)) {
           $str = sprintf
           'Requested energy change %.3f keV -> %.3f keV will require beam'."\n"
          .'optimization, but pin sensor reports that there is a pin on'."\n"
          .'the goniometer. Please make sure sample is not in the beam!'."\n"
          .'Do you want to start the energy change?',
           $E_beg, $E_end;
        }
        elsif (&getShutter($shutter_readback) == 1) {
           $str = sprintf
           'Requested energy change %.3f keV -> %.3f keV will run with fast'."\n"
          .'shutter open, but pin sensor reports that there is a pin on'."\n"
          .'the goniometer. Please make sure sample is not in the beam!'."\n"
          .'Do you want to start the energy change?',
           $E_beg, $E_end;
        } 
        else {
           $str = 'Unexpected condition for protecting sample!'."\n"
                 .'Do you want to start the energy change?';
        }
        if (&zenity_question($str,'Yes','No') == 0) {&signal_handler('quit',2);}		#this will return "512" to calling perl script
     }
     elsif ($protectSample > 1) {								##2=auto-move (first ask) 3=auto-move without warning
        if ($protectSample == 2) {								#2=auto-move sample (first ask)
           $str = sprintf
                  'ATTENTION: there is a pin mounted on the goniometer!'."\n"
                . 'To protect it from radiation, I am going to retract'."\n"
                . 'the pin %s by %.3fmm (%.1fum)'."\n"
                . 'Press "OK" to retract the pin or "Cancel" to proceed.'."\n"
                . 'with energy change without retracting the pin.', $direction, abs($moveSample), abs(1000*$moveSample);
           if (&zenity_question($str,'OK','Cancel') == 0) {
              $moveSample = 0;
              $str = 'Are you sure you want to proceed with the sample in the beam?'."\n"
                   . 'Press "Yes" to proceed or "No" to abort changing energy.';
              if (&zenity_question($str,'Yes','No') == 0) {&signal_handler('quit',2);}		#this will return "512" to calling perl script
           }
        }
        if ($moveSample != 0) {
           for ($i=0; $i<$nxyz; $i++) {$xyzIni[$i] = &getAxisAct($xyzAxesPV[$i]);}
           $omega = &getAxisAct($omegaAxisPV);
           ### Move the sample out of the beam:
           if ($tell_bluice) {&tellBluIce($scr.': moving sample '.$direction,'warn');}
           if ($direction =~ /^U/ || $direction =~ /^D/) {
              $status = &runNMotors ($xyzAssyPV, 2, [$xyzAxesPV[0],$xyzAxesPV[1]],
                                     [$xyzIni[0]+$moveSample*sin($omega*$degree2rad),
                                      $xyzIni[1]+$moveSample*cos($omega*$degree2rad)]);
           } else {
              $status = &runNMotors ($zAssyPV, 1, [$xyzAxesPV[2]],
                                     [$xyzIni[2]+$moveSample]);
           }
           if ($status) {
              $msg = 'Error='.$status.' while moving sample out of the beam.';
              print STDOUT '*** '.$script.': '.$msg."\n";
              if ($tell_bluice) {&tellBluIce($scr.': '.$msg,'error');}
              &zenity_error_exit('Failed moving sample '.$direction.'.'."\n"
                                .'Energy not changed (script aborted).',1);
           } else {
              $sample_moved = 1;
           }
#       } else {
#          ### Commented out because ($moveSample=0) is possible if we allowed not moving the pin
#          &zenity_error_exit('Moving sample '.$direction.' requred, but specified distance is zero.'."\n"
#                               .'Energy not changed (script aborted).',1);
        }
     }
  }
  else {
     if ($tell_bluice) {
        if (($align == 1 && $align_needed) || ($align == 2)) {
           $msg = $scr.': going to optimize beam';
        }
        elsif (&getShutter($shutter_readback) == 1) {
           $msg = $scr.': going to change energy with shutter open';
        }
        else {
           $msg = $scr.': going to change energy with *UNEXPECTED* case';
        }
        &tellBluIce($msg,'warn');
        select(undef,undef,undef,0.3);         							#short sleep
        &tellBluIce($scr.': expect no sample mounted','warn');
        select(undef,undef,undef,1);         							#sleep
     }
  }
  return;
}

#=================================================================

sub signal_handler ($;$) {
  our ($ABORT_IN_PROCESS);
# local($sig) = @_;
  my $sig  = shift(@_);
  my $code = shift(@_);
  if ($sig !~ /^quit$/) {print STDOUT "\n".'*** '.$script.': Caught a SIG'.$sig."\n";}

  if (defined $ABORT_IN_PROCESS) {return;}
  $ABORT_IN_PROCESS = 1;
  &abort_external_script();

  if (defined $monoAssy) {
     print STDOUT $script.': Aborting Energy motion.'."\n" ;
     &abortAssy ($monoAssy, 0);
     if ($linkID) {&abortID($undulator);}
  }
  if ($gapcheck_disabled) {
     print STDOUT $script.': Re-enabling JBluIce gap check.'."\n";
     $status = Pezca::PutDoubleOldCa($gapCheckPV,1);
  }
  if ($protectSample > 1 && defined $sample_moved && $moveSample != 0) {			#2=auto-move(first ask) 3=auto-move without warning
     if    ($direction eq 'UP')       {$direction = 'DOWN';}
     elsif ($direction eq 'DOWN')     {$direction = 'UP';}
     elsif ($direction eq 'OUTBOARD') {$direction = 'INBOARD';}
     elsif ($direction eq 'INBOARD')  {$direction = 'OUTBOARD';}
     printf STDOUT $script.': moving sample back (%s) by %.1fum.'."\n",
                   $direction,abs(1000*$moveSample);
     if ($xyzAssyPV eq $zAssyPV) {
        $status = &startNMotors ($xyzAssyPV, $nxyz, \@xyzAxesPV, \@xyzIni);
     } else {
        $status = &startNMotors ($xyzAssyPV, 2, \@xyzAxesPV, \@xyzIni);
        $status = &startNMotors ($zAssyPV, 1, [$xyzAxesPV[2]], [$xyzIni[2]]);
     }
  }
  if ($mostab_disabled) {
     print STDOUT $script.': Re-enabling intensity feedback.'."\n";
     &restoreMostab ($beamline);
  }
  if ($postab_disabled) {
     print STDOUT $script.': Re-enabling positional feedback.'."\n";
     &restorePostab ($beamline);
  }
  if ($mirrors_enabled) {
     @prg = ('perl', '-W', $PEZCA.'/kill_kbm.pl', $beamline);
     if ($^O !~ /Win/i) {push @prg, '&'; $str=join(' ',@prg); system($str);}
     else               {system(1,@prg);}
  }
  print STDOUT $script.': Closing monitors and exiting.'."\n" ;
  &clearAllCaMonitors();
  if (defined $code) {exit $code;}								#this will return "256*n" to calling perl script (when n is small)
  else               {exit 1;}									#this will return "256" to calling perl script
}
