#!/usr/bin/perl -w
#------------------------------------------------------------------------
# This script sets the offsets for 4 CAEN QuadEM counters.
# 1. It checks if the front-end shutter is closed. if yes, proceeds to step-4
# 2. Records current WBSL size
# 3. Fully closes WBSL to make sure there is nothing to count in the CAEN channels
# 4. Sets the CAEN offsets 
# 5. Re-opens WBSL to the recorded size if the front-end shutter was not closed.
#
#                                Author: Sergey Stepanov
#
# Version-1.0 2022/11/08: initial impelentation
#--------------------------------------------------------------
  use strict;
  use Pezca;
  use warnings::unused;

  our ($beamline, $NONINTERACTIVE, $IGNORE_ERRORS);
  my ($script, $PEZCA, $status, $syntax, $i);
  my ($arg, @beamlinelist, $bl_specified);
  my ($caen, $wbsl_assy, @wbsl, @wopen, @wclosed, @wtarget);
  my ($pv_off, $pv_cmp, $val1, $val2, @vh, @l_limit, $drvh);
  my ($mostab_initial, $mostab_disabled);

  select(STDERR);  $|=1;					#flush output after each print
  select(STDOUT);  $|=1;					#flush output after each print

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  do $PEZCA.'/Common/caen_subs.pl';
  do $PEZCA.'/Common/shutter_subs.pl';
  do $PEZCA.'/Common/motor_subs.pl';
  do $PEZCA.'/Common/mostab_subs.pl';
  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  $script = &makeScriptName();
  $syntax = "\n\n".'Syntax: '.$script.' [beamline]'."\n";
  @beamlinelist  = ('23i:', '23o:', '23b:', '23d:');

  foreach $arg (@ARGV) {
     $arg =~ s/^-{0,2}//;							#trim leading dashes
     if (grep(/^$arg:?$/, @beamlinelist)) {
        if (defined $bl_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $bl_specified = 1;
        $arg =~ s/:$//;								#trim ':'
        if ($arg ne $beamline) {
           print STDOUT "\n".'!!! '.$script.': Cross-beamline operation:'
                        .' controlling '.$arg.' from '.$beamline."\n\n";
           $beamline = $arg;
           do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
        }
     }
     else {									#this will exit automatically if $NONINTERACTIVE=1
        &press_key_exit('*** '.$script.': Unrecognized cmdline parameter "'
                       .$arg.'"'.$syntax);
     }
  }										#end of loop over cmdline arguments

  $SIG{'HUP'}  = 'signal_handler';			#(1) hangup (see signum.h)
  $SIG{'INT'}  = 'signal_handler';			#(2) interrupt (see signum.h)
  $SIG{'QUIT'} = 'signal_handler';			#(3) quit (see signum.h)
# $SIG{'KILL'} = 'signal_handler';			#(9) this is what is sent by kill -9 -- unblockable
  $SIG{'TERM'} = 'signal_handler';			#(15) terminate -- sent by kill without paramaters
# $SIG{'STOP'} = 'signal_handler';			#(19) stop -- unblockable

  $NONINTERACTIVE = 1;
  $IGNORE_ERRORS = 1;

  $caen      = $beamline.':AH501';
  $wbsl_assy = $beamline.':WB:Ap';
  $wbsl[0]   = $beamline.':WB:SV';
  $wbsl[1]   = $beamline.':WB:SH';
  @vh = (0,1);						#close v or h or both
  $status = 0;
  for ($i=0; $i<2; $i++) {($wclosed[$i],$drvh) = &getAxisDRVLH($wbsl[$i],\$status);}

  print STDOUT $script.': checking front-end-shutter state.'."\n";
  if (&getFrontEndShutter($beamline) != 0) {					#1=open,0=close,-1=not.responding
     print STDOUT $script.': front-end-shutter is not closed. Closing WBSL.'."\n";

     ### Save current WBSL sizes:
     for ($i=0; $i<2; $i++) {
       $wopen[$i] = &getAxisAct ($wbsl[$i],\$status);
     }
     printf STDOUT $script.': initial  WBSL sizes V=%.3fmm H=%.3fmm.'."\n",$wopen[0],$wopen[1];

     &pause_feedback();

     ### Closing WBSL
     for ($i=0; $i<2; $i++) {
        if ($vh[$i] == 1) {$wtarget[$i] = $wclosed[$i];}
        else              {$wtarget[$i] = $wopen[$i];}
     }
     printf STDOUT $script.': closing WBSL to V=%.3fmm H=%.3fmm.'."\n",$wtarget[0],$wtarget[1];
     $status = &runNMotors ($wbsl_assy, 2, \@wbsl, \@wtarget, 'forceAuto');
     if ($status) {
        print STDOUT '*** '.$script.': failed to close WBSL.'."\n";
        ### Check if we hit a lower limit:
        for ($i=0; $i<2; $i++) {($l_limit[$i],$drvh) = &checkMotorLimits($wbsl[$i]);}
        if ($l_limit[0] == 0 && $l_limit[1] == 0) {
           print STDOUT '*** '.$script.': failed to close WBSL. Trying to re-open and exiting.'."\n";
           $status = &runNMotors ($wbsl_assy, 2, \@wbsl, \@wopen, 'forceAuto');
           ### Re-enable feedback:
           &restore_feedback();
           &clearAllCaMonitors();
          exit 1;
        } else {
           print STDOUT '*** '.$script.': still continuing because WBSL size is at a lower limit.'."\n";
        }
     }
  } else {
     print STDOUT $script.': front-end-shutter is closed -- no need to close WBSL.'."\n";
  }

  if (&getCAEN($caen) != 1) {
     &startCAEN($caen);
     select(undef,undef,undef,0.5);					# sleep($wait);
  }

  for ($i=1; $i<=4; $i++) {
     ### Change offsets and print them along with the old ones:
     $pv_off = $caen.':CurrentOffset'.$i;
     $pv_cmp = $caen.':ComputeCurrentOffset'.$i.'.PROC';

     ($status,$val1) = Pezca::GetDouble($pv_off);
     if ($status) {&report_only($script, $status, $pv_off); $val1=0;}

     $status = Pezca::PutDoubleOldCa($pv_cmp, 1);
     if ($status) {&report_only($script, $status, $pv_cmp);}
     select(undef,undef,undef,0.3);					# sleep($wait);

     ($status,$val2) = Pezca::GetDouble($pv_off);
     if ($status) {&report_only($script, $status, $pv_off); $val2=0;}

     printf STDOUT 'CAEN channel-%d:  %9.0f => %9.0f'."\n", $i, $val1, $val2;
  }

  if (defined $wopen[0] && defined $wopen[1]) {
     printf STDOUT $script.': re-opening WBSL to V=%.3fmm H=%.3fmm.'."\n",$wopen[0],$wopen[1];
     $status = &runNMotors ($wbsl_assy, 2, \@wbsl, \@wopen, 'forceAuto');
     if ($status) {print STDOUT '*** '.$script.': failed to re-open WBSL, status='.$status."\n";}
     ### Re-enable feedback:
     &restore_feedback();
  }
  &clearAllCaMonitors();
  exit $status;


#=================================================================

sub pause_feedback() {
  $mostab_disabled = 0;							# mostab has not been paused
  $mostab_initial = &getMostabState ($beamline);			# mostab state: 0=run, 1=pause, -1=error
  if ($mostab_initial == -1) {
     print STDOUT $script.': Cannot talk to MOSTAB.'."\n";
     $mostab_initial = 1;						# presume paused if failure
  }
  if ($mostab_initial != 1) {						# if not paused
     print STDOUT $script.': Pausing I0 feedback'."\n";
     $status = &pauseMostab ($beamline);
     $mostab_disabled = 1;
  }
}


#=================================================================

sub restore_feedback() {
  my $status = 0;
  if (defined $mostab_initial && defined $mostab_disabled) {
     ### $mostab_initial:  0=run, 1=pause, -1=error
     if ($mostab_initial != 1 && $mostab_disabled != 0) {		# if not paused initially (#0=run, 1=pause, -1=error)
        $mostab_disabled = 0;
        print STDOUT $script.': Restoring I0 feedback...'."\n";
        $status = &restoreMostab ($beamline);
        if ($status) {
           print STDOUT "\n".'*** '.$script.': Failed to restart MOSTAB'."\n";
        }
     }
  }
  return $status;
}


#=================================================================

sub signal_handler($) {
  my $signame = shift;
# local($signame) = @_;
  print STDOUT "\n".$script.': *** Caught a SIG'.$signame."\n";
  if (defined $wopen[0] && defined $wopen[1]) {
     &abortAssy ($wbsl_assy);
     select(undef,undef,undef,0.3);					# sleep($wait);
     &runNMotors ($wbsl_assy, 2, \@wbsl, \@wopen, 'forceAuto');
  }
  &restore_feedback();
  &clearAllCaMonitors();
  print STDOUT $script.': exiting!'."\n\n";
  exit 1;
}
