#!/usr/bin/perl -W
#------------------------------------------------------------------------
# This script converts Canberra MCA ROI to Canberra TCA discrimination
#
#
#     		Author: Sergey Stepanov
#
# Version-1.0:  2021/02/03 using Pezca
#------------------------------------------------------------------------
  use Pezca;
# use warnings::unused;
  use strict;

  our ($beamline);
  my ($PEZCA, $script, $cmdline, $syntax, $narg, $arg, $i, @beamlinelist);
  my ($bln_specified, $medm_specified, $reset_specified);
  my ($ROI_l, $ROI_h, $low, $high, $medm, $reset, $mca, $adc, $tca, $msg);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  do $PEZCA.'/Common/fluo_subs.pl';
  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  select STDERR; $|=1;					#set unbuffered output
  select STDOUT; $|=1;					#set unbuffered output

#################################
### These defaults can be
### overwritten using CMD line:
  $medm  = 0;						#0=no adc/tca MEDM, 1=show adc/tca MEDM
  $reset = 0;
#################################

  @beamlinelist = ('23i:', '23o:', '23b:', '23d:');

  $script = &makeScriptName();
  $script =~ s/\.pl$//i;
  $cmdline = '[beamline] [medm] [reset]'."\n";
           ;
  $syntax = "\n\n".'Syntax: '.$script.'.pl '.$cmdline."\n";

  $narg = @ARGV;
  if ($narg > 0) {&reportout('Started with cmdline = '."@ARGV\n");}
  else           {&reportout('Started with no cmdline arguments.'."\n");}

### ==================================================================== begin of command line parameters
  for ($i=0; $i<$narg; $i++) {
     $arg = $ARGV[$i];
     $arg =~ s/^-{0,2}//;					#trim leading dashes
#    if (grep {$arg =~ /^$_:?$/} @beamlinelist) {		#use this wheh beamlinelist does not contain ':'
     if (grep(/^$arg:?$/, @beamlinelist)) {
        if (!defined $bln_specified) {
           $bln_specified = 1;
           $arg =~ s/:$//;
           if ($arg ne $beamline) {
              &reportout('Cross-beamline operation: controlling '.$arg.' from '.$beamline);
              $beamline = $arg;
              do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
           }
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "beamline": '.$ARGV[$i].$syntax);
        }
     }
     elsif ($arg =~ /^medm$/i || $arg =~ /^nomedm$/i) {
        if (!defined $medm_specified) {
           $medm_specified = 1;
           if    ($arg =~ /^medm$/i)   {$medm = 1;}
           elsif ($arg =~ /^nomedm$/i) {$medm = 0;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "medm": '.$ARGV[$i].$syntax);
        }
     }
     elsif ($arg =~ /^reset$/i || $arg =~ /^noreset$/i) {
        if (!defined $reset_specified) {
           $reset_specified = 1;
           if    ($arg =~ /^reset$/i)   {$reset = 1;}
           elsif ($arg =~ /^noreset$/i) {$reset = 0;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "medm": '.$ARGV[$i].$syntax);
        }
     }
     else {
        &press_key_exit('*** '.$script.': Incorrect command line: '.$ARGV[$i].$syntax);
     }
  }
### ==================================================================== end of command line parameters

  $mca = $beamline.':aim_adc1';    &reportout('MCA='.$mca);
  $adc = $beamline.':aim_adc1';    &reportout('ADC='.$adc);
  $tca = $beamline.':aim_tca1';    &reportout('TCA='.$tca);

  if ($medm) {
     &openMedmCanberraMCA ($mca);
     &openMedmCanberraTCA ($tca);
  }

  if ($reset) {
### clear MCA ROI
     &reportout('Resetting MCA ROI-0');
     ($ROI_l,$ROI_h) = (-1,-1);
     &setMCA_ROI($mca,$ROI_l,$ROI_h);
     &reportout('MCA/ROI-0: low='.$ROI_l.', high='.$ROI_h);
### reset TCA
     &reportout('Resetting TCA');
     $low = 5.;							#lower threshold for SCA-1
     &setupMCA_TCA($tca,$low);
     &reportout('TCA: low='.$low.',  high=100');
  } else {
     &reportout('Reading MCA ROI');
     ($ROI_l, $ROI_h) = &checkMCA_ROI($mca);
     &reportout('MCA: ROI_low='.$ROI_l.',  RO_high='.$ROI_h);
     if ($ROI_l == $ROI_h) {&press_key_exit('*** '.$script.': MCA ROI is not set.');}
     &reportout('Setting TCA discrimination from MCA ROI');
     &setDiscrFromROI($mca,$adc,$tca);
     &reportout('Reading back TCA discrimination');
     ($low,$high) = &getMCA_TCA($tca);
     $msg = sprintf 'Discr.low=%.3f, Discr_high=%.3f',$low, $high;
     &reportout($msg);
  }
  &clearAllCaMonitors();
  exit;

#==========================================================================

#sub reportout($msg,[$msgtype]) {
sub reportout($;$) {
  my ($msg,$msgtype);
  $msg       = shift(@_);
  $msgtype   = shift(@_);			# $msgtype=note,warn,error,popup_note,popup_error,direct

  if (!defined $msgtype)   {$msgtype   = 'note';}

  $msg       = $script.': '.$msg;
  if    ($msgtype =~ /warn/) {$msg = '!!! '.$msg;}
  elsif ($msgtype =~ /err/)  {$msg = '*** '.$msg;}
  print STDOUT $msg."\n";
}
