#!/usr/bin/perl -W
#--------------------------------------------------------------
#             This script checks automounter gripper:
# 1. Runs all motors to their limits where they are normally homed
# 2. Executes search for index pulse
# 3. Compares actual reading vs expected "0" and prints deviation
# 4. Returns all motors to their actual positions.
#
#     		     Author: Sergey Stepanov
#
# Version 1.00 (2019.11.08): initial implementation.
#--------------------------------------------------------------
  use strict;
  use Pezca;

  our ($beamline, $nonstop, $robot, $USE_ZENITY_FOR_EXIT);

  my ($PEZCA, $script, $msg, $status, $i);
  my ($assy, $pmac, @axes, @drives, @motors, @mtNumbers);
  my ($naxes, $ndrv, $nmtr, $iV, $iH, $iR);
  my (@inimtr, @inidrv, @curmtr, @scale, @egu);
  my ($searchStepsV, $searchStepsH, $searchStepsR);
  my (@beamlinelist, $narg, $arg, $popup);
  my (@deviat, $mtrPos, $summaryPV);
  
  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  do $PEZCA.'/Common/home_subs.pl';
  do $PEZCA.'/Common/motor_subs.pl';
  do $PEZCA.'/Common/robot_subs.pl';
  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  select STDOUT; $|=1;			# set unbuffered output
  select STDERR; $|=1;			# set unbuffered output
  &setEpicsRetry();

  $SIG{HUP}  = 'signal_handler';	#(1) hangup (see signum.h)
  $SIG{INT}  = 'signal_handler';	#(2) interrupt (see signum.h)
  $SIG{QUIT} = 'signal_handler';	#(3) quit (see signum.h)
# $SIG{KILL} = 'signal_handler';	#(9) this is what is sent by kill -9 -- unblockable
  $SIG{TERM} = 'signal_handler';	#(15) terminate -- sent by kill without paramaters
# $SIG{STOP} = 'signal_handler';	#(19) stop -- unblockable

  $script = &makeScriptName();
# $script =~ s/\.pl$//i;

### Parse command line
  @beamlinelist = ('23i:', '23o:', '23b:', '23d:');
  $popup = 1;
  $narg = @ARGV;
  for ($i=0; $i<$narg; $i++) {
     $arg = $ARGV[$i];
     ### Check if the default beamline name is overridden:
     if (grep(/^$arg:?$/, @beamlinelist)) {
        $arg =~ s/:$//;
        if ($beamline ne $arg) {
           print STDOUT "\n".'!!! '.$script.': Cross-beamline operation for '.$arg.' from '.$beamline."\n\n";
           $beamline = $arg;
           do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
        }
     }
     elsif ($arg =~ /^(no)?popup$/i) {
        if ($arg =~ /^popup$/i) {$popup=1;}
        else                    {$popup=0;}
     }
     else {
        &press_key_exit('*** '.$script.': unrecognized command line argument ['.$arg.'].');
     }
  }

### Defaults (may be changed by cmdline options):
  if ($popup) {$USE_ZENITY_FOR_EXIT = 1;}
  else        {$USE_ZENITY_FOR_EXIT = 0;}
  $nonstop = 2;                 	#must be global (do not abort at timeouts)

  if (($beamline ne '23i') && 
      ($beamline ne '23o') && 
      ($beamline ne '23d')) {&press_key_exit('*** '.$script.': this script cannot run at '.$beamline);}

  $assy = 'RB:Gr';
  $summaryPV = $beamline.':'.$assy.':Sync.DESC';
  $status = Pezca::PutStringOldCa($summaryPV,'');
  &report_only($script,$status,$summaryPV);

### Search configNN.tcl for given $assy
  $status = &getAssyDataFromConfigTcl ($beamline,$assy,\$pmac,\@axes,\@drives,\@motors,\@mtNumbers);
  $assy = $beamline.':'.$assy;
  $naxes = @axes;
  $ndrv = @drives;
  $nmtr = @motors;
  if ($naxes != 0 || $ndrv != 3 || $nmtr != 3) {
     &press_key_exit('*** '.$script.': incorrect number of axes, drives, or motors: naxes='.$naxes.' ndrives='.$ndrv.' nmotors='.$nmtr);
  }

  if (! &getHomed($assy)) {&press_key_exit('*** '.$script.': gripper [.'.$assy.'] is not homed.');}
  
  if ($robot !~ /^${beamline}/) {$robot = $beamline.':'.$robot;}
  print STDOUT 'Pausing robot ['.$robot.']'."\n";
  &pauseRobot($robot);

  if (&abortAssy($assy)) {&press_key_exit('*** '.$script.': cannot stop gripper assembly ['.$assy.'].');}

  print STDOUT 'assembly='.$assy."\n";
  for  ($i=0; $i<$ndrv; $i++) {
     $inimtr[$i] = &getAxisAct($motors[$i]);
     $inidrv[$i] = &getAxisAct($drives[$i]);
     $scale[$i]  = &getDriveScale($drives[$i]);
     $egu[$i]    = &getDriveEGU($drives[$i]);
     $deviat[$i] = 0.;
     printf STDOUT 'drive=%s inidrv=%-8.4f  mtr=%s inimtr=%.3f'."\n", $drives[$i], $inidrv[$i], $motors[$i], $inimtr[$i];
  }
  ($iV) = grep {$drives[$_] =~ /V:$/} 0..$#drives;
  ($iH) = grep {$drives[$_] =~ /H:$/} 0..$#drives;
  ($iR) = grep {$drives[$_] =~ /R:$/} 0..$#drives;
# printf STDOUT 'Vertical=%s Horizontal=%s Rotation=%s'."\n",$drives[$iV],$drives[$iH],$drives[$iR];
# exit;   

  if (&jogGripperUp($motors[$iV]))       {&press_key_exit('*** '.$script.': error jogging gripper UP.');}
  if (&jogGripperOutboard($motors[$iH])) {&press_key_exit('*** '.$script.': error jogging gripper Outboard.');}
  if (&rotateGripperUp($motors[$iR]))    {&press_key_exit('*** '.$script.': error rotating gripper UP.');}

  $searchStepsV = int(abs(150/$scale[$iV]));                         #150mm
  $searchStepsH = int(abs(100/$scale[$iH]));                         #100mm
  $searchStepsR = int(abs(110/$scale[$iR]));                         #110degr

###------------
  print STDOUT "\n".'-------------------------------------'."\n"
              .'Checking homing for gripper rotation...'."\n";
  $status = &jogMotorProc ($motors[$iR], '-'.$searchStepsR.'^0', '^');
  if ($status) {print STDOUT '*** Searching index failed for '.$motors[$iR]."\n";}
  else {
     $mtrPos = &getAxisAct ($motors[$iR]);
     $deviat[$iR] = $scale[$iR]*$mtrPos;
     printf STDOUT 'Homing of gripper rotation is off my %.4fdegr.'."\n",$deviat[$iR];
  }
  if (&rotateGripperUp($motors[$iR]))    {&press_key_exit('*** '.$script.': error rotating gripper UP.');}

###------------
  print STDOUT "\n".'-------------------------------------'."\n"
              .'Checking homing for gripper horizontal...'."\n";
  $status = &jogMotorProc ($motors[$iH], '+'.$searchStepsH.'^0', '^');
  if ($status) {print STDOUT '*** Searching index failed for '.$motors[$iH]."\n";}
  else {
     $mtrPos = &getAxisAct ($motors[$iH]);
     $deviat[$iH] = $scale[$iH]*$mtrPos;
     printf STDOUT 'Homing of gripper horizontal is off my %.4fmm'."\n",$deviat[$iH];
  }
  if (&jogGripperOutboard($motors[$iH])) {&press_key_exit('*** '.$script.': error jogging gripper Outboard.');}

###------------
  print STDOUT "\n".'-------------------------------------'."\n"
              .'Checking homing for gripper vertical...'."\n";
  $status = &jogMotorProc ($motors[$iV], '-'.$searchStepsV.'^0', '^');
  if ($status) {print STDOUT '*** Searching index failed for '.$motors[$iV]."\n";}
  else {
     $mtrPos = &getAxisAct ($motors[$iV]);
     $deviat[$iV] = $scale[$iV]*$mtrPos;
     printf STDOUT 'Homing of gripper vertical is off my %.4fmm'."\n",$deviat[$iV];
  }
  if (&jogGripperUp($motors[$iV]))       {&press_key_exit('*** '.$script.': error jogging gripper UP.');}

### Restore initial positions:
  print STDOUT "\n".'Restoring initial gripper positions.'."n";
  $status = &jogMotorProc ($motors[$iR], $inimtr[$iR], '=');
  if ($status) {print STDOUT '***  Returning motor [%s] to inipos=%d failed with status=%d.'."\n",$motors[$iR],$inimtr[$iR],$status;}
  $status = &jogMotorProc ($motors[$iH], $inimtr[$iH], '=');
  if ($status) {print STDOUT '***  Returning motor [%s] to inipos=%d failed with status=%d.'."\n",$motors[$iH],$inimtr[$iH],$status;}
  $status = &jogMotorProc ($motors[$iV], $inimtr[$iV], '=');
  if ($status) {print STDOUT '***  Returning motor [%s] to inipos=%d failed with status=%d.'."\n",$motors[$iV],$inimtr[$iV],$status;}

  print STDOUT "\n".'======================================'."\n";
  $msg = sprintf 'V=%.4f H=%.4f R=%.4f',$deviat[$iV],$deviat[$iH],$deviat[$iR];
  if (length($msg) > 39) {$msg = sprintf 'V=%.3f H=%.3f R=%.3f',$deviat[$iV],$deviat[$iH],$deviat[$iR];}
  if (length($msg) > 39) {$msg = sprintf '%.3f %.3f %.3f',      $deviat[$iV],$deviat[$iH],$deviat[$iR];}
  if (length($msg) > 39) {$msg = sprintf '%.2f %.2f %.2f',      $deviat[$iV],$deviat[$iH],$deviat[$iR];}
  else                   {$msg = substr($msg,0,39);}
  $status = Pezca::PutStringOldCa($summaryPV,$msg);
  &report_only($script,$status,$summaryPV);
  $msg = 'Summary of deviations:'."\n"; 
  for  ($i=0; $i<$ndrv; $i++) {$msg .= sprintf 'drive=%s deviation=%.4f%s'."\n",$drives[$i],$deviat[$i],$egu[$i];}
  if ($USE_ZENITY_FOR_EXIT) {&zenity_info($msg);}
  else                      {print STDOUT $msg;}

  &clearAllCaMonitors();
  print STDOUT 'All done.'."\n";
  exit 0;   


#=================================================================

sub jogGripperUp ($) {
  ### Pre-jog gripper UP to avoid collisions:
  my ($status, $mtr);
  $mtr = shift(@_);  if ($mtr !~ /:$/) {$mtr .= ':';}
  print STDOUT "\n".'Pre-jogging robot gripper UP to avoid collisions...'."\n";
  $status = &jogMotor($mtr, 'J+', $mtr.'PsLimSet', 1);
  if ($status) {
     $msg = '*** '.$script.': pre-jogging robot gripper UP failed. Please investigate!';
     &press_key_exit($msg);
  }
  return $status;
}

sub jogGripperOutboard ($) {
  ### Move gripper outboard to avoid collisions:
  my ($status, $mtr);
  $mtr = shift(@_);  if ($mtr !~ /:$/) {$mtr .= ':';}
  print STDOUT "\n".'Pre-jogging robot gripper OUTBOARD to avoid collisions...'."\n";
  $status = &jogMotor($mtr, 'J-', $mtr.'NgLimSet', 1);
  if ($status) {
     $msg = '*** '.$script.': pre-jogging robot gripper OUTBOARD failed. Please investigate!';
     &press_key_exit($msg);
  }
  return $status;
}

sub rotateGripperUp ($) {
  ### Pre-rotate gripper UP to avoid collisions:
  my ($status, $mtr);
  $mtr = shift(@_);  if ($mtr !~ /:$/) {$mtr .= ':';}
  print STDOUT "\n".'Pre-rotating robot gripper UP to avoid collisions...'."\n";
  $status = &jogMotor($mtr, 'J+', $mtr.'PsLimSet', 1);
  if ($status) {
     $msg = '*** '.$script.': pre-rotating robot gripper UP failed. Please investigate!';
     &press_key_exit($msg);
  }
  return $status;
}


#=================================================================

sub signal_handler {
# local($sig) = @_;
  my $sig = shift;
  $USE_ZENITY_FOR_EXIT = 0;
  &press_key_exit($script.': *** Caught a SIG'.$sig);
}

