#!/usr/bin/perl -W
#--------------------------------------------------------------
#             This script checks automounter gripper:
# 1. Runs all motors to their limits where they are normally homed
# 2. Executes search for index pulse
# 3. Compares actual reading vs expected "0" and prints deviation
# 4. Returns all motors to their actual positions.
#
#     		     Author: Sergey Stepanov
#
# Version 1.00 (2019.11.08): initial implementation (checkGripper.pl).
# Version 1.01 (2019.12.09): added checking Dewar and renamed the script to checkRobotHoming.pl
#--------------------------------------------------------------
  use strict;
  use Pezca;

  our ($beamline, $nonstop, $robot, $USE_ZENITY_FOR_EXIT);

  my ($PEZCA, $script, $msg, $status, $i);
  my (@beamlinelist, $narg, $arg, $popup);
  my ($mtrPos, $summaryPV);
  my ($assyGr, $pmacGr, @axesGr, @drivesGr, @motorsGr, @mtrnumGr);
  my ($assyDr, $pmacDr, @axesDr, @drivesDr, @motorsDr, @mtrnumDr);
  my ($naxesGr, $ndrivesGr, $nmotorsGr);
  my ($naxesDr, $ndrivesDr, $nmotorsDr);
  my (@inimtrGr, @inidrvGr, @deviatGr, @eguGr, @scaleGr);
  my ($inimtrDr, $inidrvDr, $deviatDr, $eguDr, $scaleDr, $offsetDr);
  my ($searchStepsV, $searchStepsH, $searchStepsR, $iV, $iH, $iR);
  my ($searchStepsD);
  
  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  do $PEZCA.'/Common/home_subs.pl';
  do $PEZCA.'/Common/motor_subs.pl';
  do $PEZCA.'/Common/robot_subs.pl';
  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  select STDOUT; $|=1;			# set unbuffered output
  select STDERR; $|=1;			# set unbuffered output
  &setEpicsRetry();

  $SIG{HUP}  = 'signal_handler';	#(1) hangup (see signum.h)
  $SIG{INT}  = 'signal_handler';	#(2) interrupt (see signum.h)
  $SIG{QUIT} = 'signal_handler';	#(3) quit (see signum.h)
# $SIG{KILL} = 'signal_handler';	#(9) this is what is sent by kill -9 -- unblockable
  $SIG{TERM} = 'signal_handler';	#(15) terminate -- sent by kill without paramaters
# $SIG{STOP} = 'signal_handler';	#(19) stop -- unblockable

  $script = &makeScriptName();
# $script =~ s/\.pl$//i;

### Parse command line
  @beamlinelist = ('23i:', '23o:', '23b:', '23d:');
  $popup = 1;
  $narg = @ARGV;
  for ($i=0; $i<$narg; $i++) {
     $arg = $ARGV[$i];
     ### Check if the default beamline name is overridden:
     if (grep(/^$arg:?$/, @beamlinelist)) {
        $arg =~ s/:$//;
        if ($beamline ne $arg) {
           print STDOUT "\n".'!!! '.$script.': Cross-beamline operation for '.$arg.' from '.$beamline."\n\n";
           $beamline = $arg;
           do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
        }
     }
     elsif ($arg =~ /^(no)?popup$/i) {
        if ($arg =~ /^popup$/i) {$popup=1;}
        else                    {$popup=0;}
     }
     else {
        &press_key_exit('*** '.$script.': unrecognized command line argument ['.$arg.'].');
     }
  }

### Defaults (may be changed by cmdline options):
  if ($popup) {$USE_ZENITY_FOR_EXIT = 1;}
  else        {$USE_ZENITY_FOR_EXIT = 0;}
  $nonstop = 2;                 	#must be global (do not abort at timeouts)

  if (($beamline ne '23i') && 
      ($beamline ne '23o') && 
      ($beamline ne '23d')) {&press_key_exit('*** '.$script.': this script cannot run at '.$beamline);}

  $assyGr = 'RB:Gr';
  $summaryPV = $beamline.':'.$assyGr.':Sync.DESC';
  $status = Pezca::PutStringOldCa($summaryPV,'');
  &report_only($script,$status,$summaryPV);

### Search configNN.tcl for given $assyGr
  $status = &getAssyDataFromConfigTcl ($beamline,$assyGr,\$pmacGr,\@axesGr,\@drivesGr,\@motorsGr,\@mtrnumGr);
  $assyGr = $beamline.':'.$assyGr;
  $naxesGr = @axesGr;
  $ndrivesGr = @drivesGr;
  $nmotorsGr = @motorsGr;
  if ($naxesGr != 0 || $ndrivesGr != 3 || $nmotorsGr != 3) {
     &press_key_exit('*** '.$script.': incorrect number of Gripper axes, drives, or motors: naxes='.$naxesGr
                                                                                       .' ndrives='.$ndrivesGr
                                                                                       .' nmotors='.$nmotorsGr);
  }

### Search configNN.tcl for given $assyDr
  $assyDr = 'RB:Dr';
  $status = &getAssyDataFromConfigTcl ($beamline,$assyDr,\$pmacDr,\@axesDr,\@drivesDr,\@motorsDr,\@mtrnumDr);
  $assyDr = $beamline.':'.$assyDr;
  $naxesDr = @axesDr;
  $ndrivesDr = @drivesDr;
  $nmotorsDr = @motorsDr;
  if ($naxesDr != 0 || $ndrivesDr != 1 || $nmotorsDr != 1) {
     &press_key_exit('*** '.$script.': incorrect number of Dewar-phy axes, drives, or motors: naxes='.$naxesDr
                                                                                         .' ndrives='.$ndrivesDr
                                                                                         .' nmotors='.$nmotorsDr);
  }

  if (! &getHomed($assyGr)) {&press_key_exit('*** '.$script.': gripper [.'.$assyGr.'] is not homed.');}
  if (! &getHomed($assyDr)) {&press_key_exit('*** '.$script.': Dewar-phy [.'.$assyDr.'] is not homed.');}
  
  if ($robot !~ /^${beamline}/) {$robot = $beamline.':'.$robot;}
  print STDOUT 'Pausing robot ['.$robot.']'."\n";
  &pauseRobot($robot);

  if (&abortAssy($assyGr)) {&press_key_exit('*** '.$script.': cannot stop gripper assembly ['.$assyGr.'].');}
  if (&abortAssy($assyDr)) {&press_key_exit('*** '.$script.': cannot stop Dewar-phy assembly ['.$assyDr.'].');}

  print STDOUT 'Gripper assembly='.$assyGr."\n";
  for  ($i=0; $i<$ndrivesGr; $i++) {
     $inimtrGr[$i] = &getAxisAct($motorsGr[$i]);
     $inidrvGr[$i] = &getAxisAct($drivesGr[$i]);
     $scaleGr[$i]  = &getDriveScale($drivesGr[$i]);
     $eguGr[$i]    = &getDriveEGU($drivesGr[$i]);
     $deviatGr[$i] = 0.;
     printf STDOUT 'drive=%s inipos=%-8.4f  mtr=%s inipos=%.3f'."\n", $drivesGr[$i], $inidrvGr[$i], $motorsGr[$i], $inimtrGr[$i];
  }
  ($iV) = grep {$drivesGr[$_] =~ /V:$/} 0..$#drivesGr;
  ($iH) = grep {$drivesGr[$_] =~ /H:$/} 0..$#drivesGr;
  ($iR) = grep {$drivesGr[$_] =~ /R:$/} 0..$#drivesGr;

  print STDOUT 'Dewar-phy assembly='.$assyDr."\n";
  $inimtrDr = &getAxisAct($motorsDr[0]);
  $inidrvDr = &getAxisAct($drivesDr[0]);
  $scaleDr  = &getDriveScale($drivesDr[0]);
  $offsetDr = &getDriveOffset($drivesDr[0]);
  $eguDr    = &getDriveEGU($drivesDr[0]);
  $deviatDr = 0.;
  printf STDOUT 'drive=%s inipos=%-8.4f  mtr=%s inipos=%.3f'."\n", $drivesDr[0], $inidrvDr, $motorsDr[0], $inimtrDr;

  if (&jogGripperUp($motorsGr[$iV]))       {&press_key_exit('*** '.$script.': error jogging gripper UP.');}
  if (&jogGripperOutboard($motorsGr[$iH])) {&press_key_exit('*** '.$script.': error jogging gripper Outboard.');}
  if (&rotateGripperUp($motorsGr[$iR]))    {&press_key_exit('*** '.$script.': error rotating gripper UP.');}

  $searchStepsV = int(abs(150/$scaleGr[$iV]));                         #150mm
  $searchStepsH = int(abs(100/$scaleGr[$iH]));                         #100mm
  $searchStepsR = int(abs(110/$scaleGr[$iR]));                         #110degr
  $searchStepsD = int(abs(361/$scaleDr));                              #361degr

###------------
  print STDOUT "\n".'-------------------------------------'."\n"
              .'Checking homing for Dewar rotation...'."\n";
  $status = &jogMotorProc ($motorsDr[0], '-'.$searchStepsD.'^0', '^');
  if ($status) {print STDOUT '*** Searching index failed for '.$motorsDr[0]."\n";}
  else {
     $mtrPos = &getAxisAct ($motorsDr[0]);
     $deviatDr = $scaleDr*$mtrPos;
     printf STDOUT 'Homing of Dewar rotation is off my %.4fdegr.'."\n",$deviatDr;
  }

###------------
  print STDOUT "\n".'-------------------------------------'."\n"
              .'Checking homing for gripper rotation...'."\n";
  $status = &jogMotorProc ($motorsGr[$iR], '-'.$searchStepsR.'^0', '^');
  if ($status) {print STDOUT '*** Searching index failed for '.$motorsGr[$iR]."\n";}
  else {
     $mtrPos = &getAxisAct ($motorsGr[$iR]);
     $deviatGr[$iR] = $scaleGr[$iR]*$mtrPos;
     printf STDOUT 'Homing of gripper rotation is off my %.4fdegr.'."\n",$deviatGr[$iR];
  }
  if (&rotateGripperUp($motorsGr[$iR]))    {&press_key_exit('*** '.$script.': error rotating gripper UP.');}

###------------
  print STDOUT "\n".'-------------------------------------'."\n"
              .'Checking homing for gripper horizontal...'."\n";
  $status = &jogMotorProc ($motorsGr[$iH], '+'.$searchStepsH.'^0', '^');
  if ($status) {print STDOUT '*** Searching index failed for '.$motorsGr[$iH]."\n";}
  else {
     $mtrPos = &getAxisAct ($motorsGr[$iH]);
     $deviatGr[$iH] = $scaleGr[$iH]*$mtrPos;
     printf STDOUT 'Homing of gripper horizontal is off my %.4fmm'."\n",$deviatGr[$iH];
  }
  if (&jogGripperOutboard($motorsGr[$iH])) {&press_key_exit('*** '.$script.': error jogging gripper Outboard.');}

###------------
  print STDOUT "\n".'-------------------------------------'."\n"
              .'Checking homing for gripper vertical...'."\n";
  $status = &jogMotorProc ($motorsGr[$iV], '-'.$searchStepsV.'^0', '^');
  if ($status) {print STDOUT '*** Searching index failed for '.$motorsGr[$iV]."\n";}
  else {
     $mtrPos = &getAxisAct ($motorsGr[$iV]);
     $deviatGr[$iV] = $scaleGr[$iV]*$mtrPos;
     printf STDOUT 'Homing of gripper vertical is off my %.4fmm'."\n",$deviatGr[$iV];
  }
  if (&jogGripperUp($motorsGr[$iV]))       {&press_key_exit('*** '.$script.': error jogging gripper UP.');}

### Restore initial positions:
  print STDOUT "\n".'Restoring initial Dewar and gripper positions.'."n";
  $status = &jogMotorProc ($motorsDr[0], $inimtrDr, '=');
  if ($status) {
     printf STDOUT '***  Returning motor [%s] to inipos=%d failed with status=%d.'."\n",
                                         $motorsDr[0],$inimtrDr,$status;
  }
  $status = &jogMotorProc ($motorsGr[$iR], $inimtrGr[$iR], '=');
  if ($status) {
     printf STDOUT '***  Returning motor [%s] to inipos=%d failed with status=%d.'."\n",
                                         $motorsGr[$iR],$inimtrGr[$iR],$status;
  }
  $status = &jogMotorProc ($motorsGr[$iH], $inimtrGr[$iH], '=');
  if ($status) {
     printf STDOUT '***  Returning motor [%s] to inipos=%d failed with status=%d.'."\n",
                                         $motorsGr[$iH],$inimtrGr[$iH],$status;
  }
  $status = &jogMotorProc ($motorsGr[$iV], $inimtrGr[$iV], '=');
  if ($status) {
     printf STDOUT '***  Returning motor [%s] to inipos=%d failed with status=%d.'."\n",
                                         $motorsGr[$iV],$inimtrGr[$iV],$status;
  }

  print STDOUT "\n".'======================================'."\n";
### Should not exceed 39char:
### V=123.123 H=123.123 R=123.123 D=123.123
  $msg = sprintf 'V=%.3f H=%.3f R=%.3f D=%.3f',$deviatGr[$iV],$deviatGr[$iH],$deviatGr[$iR],$deviatDr;
  if (length($msg) > 39) {$msg = sprintf '%.3f %.3f %.3f %.3f', $deviatGr[$iV],$deviatGr[$iH],$deviatGr[$iR],$deviatDr;}
  if (length($msg) > 39) {$msg = sprintf '%.2f %.2f %.2f %.3f', $deviatGr[$iV],$deviatGr[$iH],$deviatGr[$iR],$deviatDr;}
  else                   {$msg = substr($msg,0,39);}
  $status = Pezca::PutStringOldCa($summaryPV,$msg);
  &report_only($script,$status,$summaryPV);
  $msg = 'Summary of deviations:'."\n"; 
  for  ($i=0; $i<$ndrivesGr; $i++) {$msg .= sprintf 'drive=%s deviation=%.4f%s'."\n",$drivesGr[$i],$deviatGr[$i],$eguGr[$i];}
                                    $msg .= sprintf 'drive=%s deviation=%.4f%s'."\n",$drivesGr[0], $deviatDr,    $eguDr;
  if ($USE_ZENITY_FOR_EXIT) {&zenity_info($msg);}
  else                      {print STDOUT $msg;}

  &clearAllCaMonitors();
  print STDOUT 'All done.'."\n";
  exit 0;   


#=================================================================

sub jogGripperUp ($) {
  ### Pre-jog gripper UP to avoid collisions:
  my ($status, $mtr);
  $mtr = shift(@_);  if ($mtr !~ /:$/) {$mtr .= ':';}
  print STDOUT "\n".'Pre-jogging robot gripper UP to avoid collisions...'."\n";
  $status = &jogMotor($mtr, 'J+', $mtr.'PsLimSet', 1);
  if ($status) {
     $msg = '*** '.$script.': pre-jogging robot gripper UP failed. Please investigate!';
     &press_key_exit($msg);
  }
  return $status;
}

sub jogGripperOutboard ($) {
  ### Move gripper outboard to avoid collisions:
  my ($status, $mtr);
  $mtr = shift(@_);  if ($mtr !~ /:$/) {$mtr .= ':';}
  print STDOUT "\n".'Pre-jogging robot gripper OUTBOARD to avoid collisions...'."\n";
  $status = &jogMotor($mtr, 'J-', $mtr.'NgLimSet', 1);
  if ($status) {
     $msg = '*** '.$script.': pre-jogging robot gripper OUTBOARD failed. Please investigate!';
     &press_key_exit($msg);
  }
  return $status;
}

sub rotateGripperUp ($) {
  ### Pre-rotate gripper UP to avoid collisions:
  my ($status, $mtr);
  $mtr = shift(@_);  if ($mtr !~ /:$/) {$mtr .= ':';}
  print STDOUT "\n".'Pre-rotating robot gripper UP to avoid collisions...'."\n";
  $status = &jogMotor($mtr, 'J+', $mtr.'PsLimSet', 1);
  if ($status) {
     $msg = '*** '.$script.': pre-rotating robot gripper UP failed. Please investigate!';
     &press_key_exit($msg);
  }
  return $status;
}


#=================================================================

sub signal_handler {
# local($sig) = @_;
  my $sig = shift;
  $USE_ZENITY_FOR_EXIT = 0;
  &press_key_exit($script.': *** Caught a SIG'.$sig);
}

