#!/usr/bin/perl -W
# _CALLED_FROM_BLUICE_
#------------------------------------------------------------------------
# This is a PERL implementation of ENERGY scans with fluorescent detector.
# (the scans over ENERGY by moving monochromator). The details of this
# program are described in:
# Stepanov et.al. J. Appl. Crystallogr., v.44, No 4, p.772-778 (2011).
#
#
#     		Author: Sergey Stepanov
#
# Version-1.0:  2004/11/12 using Pezca
# Version-1.1:  2005/03/27 -- TCA controls added; more monitored flags added to motion sub.
# Version-2.0:  2008/12/08 -- added reading scan range data from mySQL,
#                          -- added scan-for-dummies MCA preset (data from mySQL),
#                          -- removed internal Edge DB; see v.1.1 for it.
# Version-3.0:  2009/01/27 -- added option to do step scans as scanrec TBL scans
# Version-4.0:  2009/01/29 -- added option to do on-the-fly HD-sync scans with Struck
# Version-4.1:  2009/02/02 -- cleaned up v.4.0: some operations moved out into subs.
# Version-4.2:  2009/02/03 -- converted on-the-fly scan into 3-band mode
# Version-4.3:  2009/05/14 -- read edge from JBluIce if not given in the cmdline
# Version-4.4:  2009/06/15 -- added Struck cleanup and restoring Struck dwell
# Version-4.5:  2009/10/03 -- added putting beamstop/collimator IN and checking
#                             how far are we initially from the requested edge.
# Version-4.6:  2009/10/13 -- added running to requested energy with bigEnergyMove
# Version-5.0:  2010/01/18 -- added adaptive mode: search for chemical shift with
#                             Levenberg-Marquardt fitting (see Perl PDL)
# Version-6.0:  2010/03/29 -- changed to use "signal" as a key to find I0 Struck
#                             input and I0 Keithley channel from mySQL; also
#                             switched to using I0 offsets from mySQL
# Version-6.1:  2010/05/13 -- added switch between fixed(DB)/adaptive/user-specified
#                             edge position based on a PV and/or command line (preferred)
# Version-6.2:  2010/06/03 -- added pausing intensity feedback; made moving ID the default
# Version-6.3:  2010/09/21 -- changed to use scancenter PV
# Version-6.4:  2010/11/06 -- added i0=none (to suppress normalization)
# Version-6.5:  2011/01/06 -- added global flag for opening popups via BluIce
# Version-6.6:  2011/01/07 -- renamed from madScan.pl to edgeScan.pl
# Version-6.7:  2012/08/08 -- added flag: [no]waitFeedbackRestore
# Version-6.8:  2012/08/08 -- added posting 'final' JBluIce array so that
#                             JBluIce does not need to wait until the scan
#                             finishes cleaning up and restoring beamline config.
# Version-6.9:  2013/04/24 -- changed to work with sampleEnvSrv
# Version-7.0:  2013/04/26 -- changed to work with $ENSCAN_FLAGS
# Version-7.1:  2014/03/28 -- changed functions calls to arrays pointers (preparation to Perl-5.16)
# Version-7.2:  2014/06/18 -- added stopping Struck as Joerger
# Version-7.3:  2014/06/27 -- changed bringing collimator & beamstop IN AFTER moving energy to the edge
# Version-7.4:  2016/08/02 -- added extra logging for the "edgeShift=user" mode
# Version-7.5:  2016/08/09 -- added "multistep" to big energy move; changed returning to E_ini with bigEnergyMove
# Version-7.6:  2016/08/16 -- moved STDOUT printing into &reportout() function
# Version-7.7:  2018/02/05 -- added @bigEnergyFlags and reading them from JBluIce Config Table when called from JBluIce
# Version-7.8:  2019/11/06 -- added command line option to bring IN pid diode instead of beamstop
# Version-7.9:  2020/02/21 -- added global switch for using Falcon instead of Canberra
# Version-8.0:  2020/10/12 -- added restoring energy for transfocator
# Version-8.1:  2020/11/18 -- refined @bigEnergyFlags
# Version-8.2:  2021/06/05 -- added scanrecFly mode (for Falcon if the SCA output does not work)
# Version-8.3:  2021/08/26 -- adapted to new runMono which accounts for the 5th and 7th harmonic
#------------------------------------------------------------------------
# no strict 'vars';
# no strict 'subs';
# no strict 'refs';				#commenred out 2009/01/26
  use PDL;                                      #Perl data language
  use PDL::Math;				#extended math functions (atan, erf,...)
  use PDL::Fit::LM;                             #Levenberg-Marquardt fitting
  use DBI;					#mySQL interface (for scan and MCA data)
  use Time::HiRes qw(gettimeofday tv_interval); #needed for measuring total time
  use Pezca;
# use warnings::unused;
  use strict;

  our ($USE_FALCON);				#global from beamline.pl
  our ($USE_ZENITY_FOR_EXIT, $ZENITY_VIA_BLUICE, $NONINTERACTIVE, $report);
  our ($beamline, $undulator, $joergerName, $struckName, $struck_ext);
  our ($shutter, $shutter_readback, $shutter_closed);
  our ($scanrecName, $scanrecAbort);
  our ($have_TF, $have_CRL);
  our ($ENSCAN_FLAGS);				#global ENflyScan flags
  our $EZCAFLOAT;
  our ($enmono13, @harmEn, @harmNum, @Elower, @Eupper, @Olower, @Oupper);		#make global to store SQL data between calls

  my ($PEZCA, $script, $scr, $cmdline, $syntax, $narg, $arg, $i, @beamlinelist);
  my ($moveID, $linkID, $scalerNo, $i0_channel, $Flu_channel);
  my ($cnt_time, $mca_setmode, $writelog, $usegnuplot, $runchooch, $choochprg);
  my ($bln_specified, $i0_specified, $flu_specified, $count_specified, $ge_specified);
  my ($mca_specified, $logfile_specified, $gnuplot_specified, $chooch_specified);
  my ($moveID_specified, $bluice_specified, $edge_specified, $medm_specified);
  my ($fileroot_specified, $fileroot, $filehandle, $scanmode_specified);
  my ($waitFBR_specified, $bring_specified, $bringPinDiode);
  my ($outfile, $outpost, $outefs, $tmpfile, $logfile, $autoplot);
  my ($Edge, $Edge_element, $Edge_KLMN, @Offsets, $tell_bluice);
  my ($mononame, $assembly, $axisEnergy, $scaler, $fileBkg);
  my ($mca, $adc, $tca, $mca_trig);
  my ($struck, $gotoEdge, @bigEnergyFlags);
  my ($status, $EdgeEnergy, $EdgeEnergySQL, $ScanCenter);
  my (@E_step, @E_pm, $FluoEnergy, $FluoPMwidth);
  my ($N_step, $N_pm, $nE, $E_start, $E_end, @E, $startScalersPV);
  my ($logscale, $gnuopen, $E_KeV, $E1, $E2, $E_chemshift);
  my ($E_ini, $E_undini, $dE, $nE_accu, $E_edge);
  my ($mcaNMAX, $ROI_l, $ROI_h, $gain, $offset, $low, $Off);
  my (@ct, $year, $mon, $monN, $mday, $hour, $min, $datestamp, $biScanCenterPV);
  my ($bipostPV, $biEdgePV, $biEdgeModePV, $biEdgeShiftMaxPV);
  my (@En_array, @Dt_array, @I0_array, $EdgeShiftMax, $shiftmax_specified);
  my ($fakeflag, $fakefile, @fakearray, $fakenpts, $fake_specified);
  my ($scanMode, $scanrec, $scanAbort, $medm, $msg);
  my (@inputs, $ninputs, @ROIs);
  my ($ESinTheta, $Theta_big, $Theta_ini, $Theta_edge, $dwell_ini, $mtrOut);
  my ($gapCheckPV, $gapCheckState);
  my (@prg, $t0, $elapsed, $adaptive, $adapt_specified);
  my ($i0_name, $timestampDB, $epochDB, $kchannel, $klabelDB, $koffsetDB);
  my (@signals, @darks, $kgain, $igain, $koffset, $klabel);
  my ($epoch, $bkg_age_hours, $offset_delta);
  my ($feedback_specified, $pauseFeedback, $waitFBrestore, $mostab_initial);
  my ($IDoffset, $syncMode, $syncStep, $switchID, $attenuation, $switchMono); 
# my ($motionTime);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  do $PEZCA.'/Common/shutter_subs.pl';
  do $PEZCA.'/Common/keithley_subs.pl';
  do $PEZCA.'/Common/joerger_subs_v2.pl';
  do $PEZCA.'/Common/motor_subs.pl';
  do $PEZCA.'/Common/mono_subs.pl';
  do $PEZCA.'/Common/mostab_subs.pl';
  do $PEZCA.'/Common/fluo_subs.pl';
  do $PEZCA.'/Common/struck_subs_v2.pl';
  do $PEZCA.'/Common/scanrec_subs.pl';
  do $PEZCA.'/Common/gnuplot_subs.pl';
  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  select STDERR; $|=1;					#set unbuffered output
  select STDOUT; $|=1;					#set unbuffered output
  $SIG{'HUP'}  = 'signal_handler';			#(1) hangup (see signum.h)
  $SIG{'INT'}  = 'signal_handler';			#(2) interrupt (see signum.h)
  $SIG{'QUIT'} = 'signal_handler';			#(3) quit (see signum.h)
# $SIG{'KILL'} = 'signal_handler';			#(9) this is what is sent by kill -9 -- unblockable
  $SIG{'TERM'} = 'signal_handler';			#(15) terminate -- sent by kill without paramaters
# $SIG{'STOP'} = 'signal_handler';			#(19) stop -- unblockable

#################################
### These defaults can be
### overwritten using CMD line:
  $tell_bluice    = 1;		#1=inform BluIce 0=don't
  $moveID         = 1; 		#1=move ID to the upper scan range + IDoffset
  $scalerNo       = 1;		#scaler1=Struck, scaler2(3)=Joerger
  $i0_name        = 'I0';	#mySQL tag for I0 ('I0', 'Pin', 'Beamstop') or 'none'
# $i0_name        = 'none';	#mySQL tag for I0 ('I0', 'Pin', 'Beamstop') or 'none'
  if (defined $USE_FALCON && $USE_FALCON) {
     $Flu_channel = 4;          #Falcon ROI-1 to Struck input (fast scans only)
  } else {
     $Flu_channel = 24;		#default Struck input for fluorescence with Canberra (fast scans only)
  }
  $cnt_time       = 2;		#scaler count time
  $mca_setmode    = 'auto';     #'optimize'=set MCA according to mySQL 'ready'=presume MCA is ready, 'auto'=optimize if not ready
  $writelog       = 1;		#set to '0' if log is not needed
  $usegnuplot     = 1;		#set to '0' if GnuPlot is not needed
  $runchooch      = 1;		#set to '0' if chooch is not needed
  $scanMode       = 1;		#0=step/internal, 1=scanrec/TBL, 2=fly, 3=scanrec/TBL/noMCA, 4=scanrecFly, 10=auto
  $medm           = 0;          #0=no scanrec MEDM, 1=show scanrec MEDM
  $fakeflag       = 0;		#use fake data for debugging (comment out when unneeded)
  $gotoEdge       = 1;		#0=gotoEdgeNone, 1=gotoEdge with BigEnergyMove, 2=gotoEdgeSmall
  $adaptive       = 0;		#1=use adaptive scan to accomodate chemical shift of the edge (LM fit) 2=take edge energy from PV (user-specified edge)
  $EdgeShiftMax   = 50;         #maximum allowed chemical shift of the edge from a tabulated value
  $pauseFeedback  = 1;		#1=pause mostab (if running), 0=ignore
  $waitFBrestore  = 'wait';	#'wait/nowait' intensity feedback restore
  $bringPinDiode  = 0;		#1=bring PinDiode, 0=bring Beamstop
#################################

  $fakefile     = $PEZCA.'/Data/fake_MAD_Se_K.dat';		#use any chooch data file
  $choochprg    = '/mnt/software/px/bin/chooch';
  @beamlinelist = ('23i:', '23o:', '23b:', '23d:');
  $report       = 0;		#global var used by some subs as print flag

  if ($tell_bluice) {$NONINTERACTIVE = 1; $USE_ZENITY_FOR_EXIT=1;}
  else              {$NONINTERACTIVE = 0; $USE_ZENITY_FOR_EXIT=0;}

  $script = &makeScriptName();
  $script =~ s/\.pl$//i;
  $scr    = 'edgeScan';
  $cmdline = 'Element-Edge [beamline] [i0=name|none] [flu=nn|struckFlu=nn]'."\n"
           . ' [count=N] [mca=optimize|ready|auto] [logfile|nologfile]'."\n"
           . ' [gnuplot|nognuplot] [medm|nomedm] [chooch|nochooch]'."\n"
           . ' [moveID|noID] [bluice|nobluice] [fileroot=name]'."\n"
           . ' [scan=step|scanrecTBL|scanrecTBLdiscr|scanrecFly|fly|auto] [fake|nofake]'."\n"
           . ' [gotoEdge|gotoEdgeSmall|gotoEdgeNone]'."\n"
           . ' [edgeShift=none/adaptive/user]'."\n"
           . ' [edgeShiftMax=value(eV)]'."\n"
           . ' [pauseFeedback|ignoreFeedback] [[no]waitFeedBackrestore]'."\n"
           . ' [bringPinDiode|bringBeamStop]'."\n"
           ;
  $syntax = "\n\n".'Syntax: '.$script.'.pl '.$cmdline."\n";

  $narg = @ARGV;
  &reportout("\n".'Date: '.scalar(localtime)."\n");
  if ($narg > 0) {&reportout('Started with cmdline = '."@ARGV\n");}
  else           {&reportout('Started with no cmdline arguments.'."\n");}

### ==================================================================== begin of command line parameters
  for ($i=0; $i<$narg; $i++) {
     $arg = $ARGV[$i];
     $arg =~ s/^-{0,2}//;					#trim leading dashes
#    if (grep {$arg =~ /^$_:?$/} @beamlinelist) {		#use this wheh beamlinelist does not contain ':'
     if (grep(/^$arg:?$/, @beamlinelist)) {
        if (!defined $bln_specified) {
           $bln_specified = 1;
           $arg =~ s/:$//;
           if ($arg ne $beamline) {
              &reportout('Cross-beamline operation: controlling '.$arg.' from '.$beamline);
              $beamline = $arg;
              do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
           }
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "beamline": '.$ARGV[$i].$syntax);
        }
     }
     elsif ($arg =~ /^i0=/i) {
        if (!defined $i0_specified) {
           $i0_specified = 1;
           $i0_name = $arg;
           $i0_name =~ s/^i0=//i;
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "i0": '.$ARGV[$i].$syntax);
        }
     }
     elsif ($arg =~ /^flu=/i || $arg =~ /^struckFlu=/i) {
        if (!defined $flu_specified) {
           $flu_specified = 1;
           $Flu_channel = $arg;
           if    ($arg =~ /^struckFlu=/i) {$Flu_channel =~ s/^struckFlu=//i;}
           elsif ($arg =~ /^flu=/i)       {$Flu_channel =~ s/^flu=//i;}
           if (! &is_int($Flu_channel) || $Flu_channel < 4 || $Flu_channel > 32) {
              &press_key_exit('*** '.$script.': Struck input for fluorescence "'
                             .$ARGV[$i].'" must be an integer in range [4-32].'
                             .$syntax);
           }
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "flu" or "struckFlu": '.$ARGV[$i].$syntax);
        }
     }
     elsif ($arg =~ /^count=/i) {
        if (!defined $count_specified) {
           $count_specified = 1;
           $cnt_time = $arg;
           $cnt_time =~ s/^count=//i;
           if (! &is_float($cnt_time) || $cnt_time < 0.1 || $cnt_time > 30) {
              &press_key_exit('*** '.$script.': Count time "'.$ARGV[$i]
                             .'" must be a number in range [0.1-30].'.$syntax);
           }
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "count": '.$ARGV[$i].$syntax);
        }
     }
     elsif ($arg =~ /^mca=/i) {
        if (!defined $mca_specified) {
           $mca_specified = 1;
           $mca_setmode = $arg;
           $mca_setmode =~ s/^mca=//i;
           if (($mca_setmode !~ /^optimize$/i) &&
               ($mca_setmode !~ /^ready$/i)    &&
               ($mca_setmode !~ /^auto$/i)) {
              &press_key_exit('*** '.$script.': Incorrect command line: '.$ARGV[$i].$syntax);
           }
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "mca": '.$ARGV[$i].$syntax);
        }
     }
     elsif ($arg =~ /^logfile$/i || $arg =~ /^nologfile$/i) {
        if (!defined $logfile_specified) {
           $logfile_specified = 1;
           if    ($arg =~ /^logfile$/i)   {$writelog = 1;}
           elsif ($arg =~ /^nologfile$/i) {$writelog = 0;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "logfile": '.$ARGV[$i].$syntax);
        }
     }
     elsif ($arg =~ /^gnuplot$/i || $arg =~ /^nognuplot$/i) {
        if (!defined $gnuplot_specified) {
           $gnuplot_specified = 1;
           if    ($arg =~ /^gnuplot$/i)   {$usegnuplot = 1;}
           elsif ($arg =~ /^nognuplot$/i) {$usegnuplot = 0;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "gnuplot": '.$ARGV[$i].$syntax);
        }
     }
     elsif ($arg =~ /^medm$/i || $arg =~ /^nomedm$/i) {
        if (!defined $medm_specified) {
           $medm_specified = 1;
           if    ($arg =~ /^medm$/i)   {$medm = 1;}
           elsif ($arg =~ /^nomedm$/i) {$medm = 0;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "medm": '.$ARGV[$i].$syntax);
        }
     }
     elsif ($arg =~ /^chooch$/i || $arg =~ /^nochooch$/i) {
        if (!defined $chooch_specified) {
           $chooch_specified = 1;
           if    ($arg =~ /^chooch$/i)   {$runchooch = 1;}
           elsif ($arg =~ /^nochooch$/i) {$runchooch = 0;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "chooch": '.$ARGV[$i].$syntax);
        }
     }
     elsif ($arg =~ /^moveID$/i || $arg =~ /^noID$/i) {
        if (!defined $moveID_specified) {
           $moveID_specified = 1;
           if    ($arg =~ /^moveID$/i)   {$moveID = 1;}
           elsif ($arg =~ /^noID$/i)     {$moveID = 0;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "moveID": '.$ARGV[$i].$syntax);
        }
     }
     elsif ($arg =~ /^bluice$/i || $arg =~ /^nobluice$/i) {
        if (!defined $bluice_specified) {
           $bluice_specified = 1;
           if    ($arg =~ /^bluice$/i)   {$tell_bluice=1; $NONINTERACTIVE=1; $USE_ZENITY_FOR_EXIT=1;}
           elsif ($arg =~ /^nobluice$/i) {$tell_bluice=0; $NONINTERACTIVE=0; $USE_ZENITY_FOR_EXIT=0;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "bluice": '.$ARGV[$i].$syntax);
        }
     }
     elsif ($arg =~ /^fileroot=/i) {
        if (!defined $fileroot_specified) {
           $fileroot_specified = 1;
           $fileroot = $ARGV[$i];
           $fileroot =~ s/^-{0,2}fileroot=//i;
           if (length($fileroot) == 0) {undef $fileroot;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "mca": '.$ARGV[$i].$syntax);
        }
     }
     elsif ($arg =~ /^scan=/i) {
        if (!defined $scanmode_specified) {
           $scanmode_specified = 1;
           $arg =~ s/^scan=//i;
           if    ($arg =~ /^step$/i)            {$scanMode = 0;}
           elsif ($arg =~ /^scanrecTBL$/i)      {$scanMode = 1;}
           elsif ($arg =~ /^fly$/i)             {$scanMode = 2;}
           elsif ($arg =~ /^scanrecTBLdiscr$/i) {$scanMode = 3;}
           elsif ($arg =~ /^scanrecFly$/i)      {$scanMode = 4;}
           elsif ($arg =~ /^auto$/i)            {$scanMode = 10;}
           else {
              &press_key_exit('*** '.$script.': Incorrect command line: '.$ARGV[$i].$syntax);
           }
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "scan": '.$ARGV[$i].$syntax);
        }
     }
     elsif ($arg =~ /^fake$/i || $arg =~ /^nofake$/i) {
        if (!defined $fake_specified) {
           $fake_specified = 1;
           if    ($arg =~ /^fake$/i)   {$fakeflag = 1;}
           elsif ($arg =~ /^nofake$/i) {$fakeflag = 0;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "fake": '.$ARGV[$i].$syntax);
        }
     }
     elsif ($arg =~ /^gotoEdge$/i || $arg =~ /^gotoEdgeSmall$/i || $arg =~ /^gotoEdgeNone$/i) {
        if (!defined $ge_specified) {
           $ge_specified = 1;
           if    ($arg =~ /^gotoEdgeNone$/i)  {$gotoEdge = 0;}
           elsif ($arg =~ /^gotoEdge$/i)      {$gotoEdge = 1;}
           elsif ($arg =~ /^gotoEdgeSmall$/i) {$gotoEdge = 2;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "gotoEdge": '.$ARGV[$i].$syntax);
        }
     }
     elsif ($arg =~ /^edgeShift=/i) {
        if (!defined $adapt_specified) {
           $adapt_specified = 1;
           if    ($arg =~ /^edgeShift=fixed/i)    {$adaptive=0;}	#0=scan is not adaptive
           elsif ($arg =~ /^edgeShift=adaptive/i) {$adaptive=1;}	#1=use adaptive scan to accomodate chemical shift of the edge (LM fit)
           elsif ($arg =~ /^edgeShift=user/i)     {$adaptive=2;}	#2=take edge energy from PV or cmdline (user-specified edge)
           else {
              &press_key_exit('*** '.$script.': Incorrect command line: '.$ARGV[$i].$syntax);
           }
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "edge=fixed|adaptive|user": '.$ARGV[$i].$syntax);
        }
     }
     elsif ($arg =~ /^edgeShiftMax=/i) {
        if (!defined $shiftmax_specified) {
           $shiftmax_specified = 1;
           $EdgeShiftMax = $arg;
           $EdgeShiftMax =~ s/^edgeShiftMax=//i;
           if (! &is_float($EdgeShiftMax) || $EdgeShiftMax <= 0. || $EdgeShiftMax >= 50) {
              &press_key_exit('*** '.$script.': command line parameter EdgeShiftMax "'
                              .$ARGV[$i].'" must be a positive number in range [0-50].'.$syntax);
           }
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "edgeShiftMax": '.$ARGV[$i].$syntax);
        }
     }
     elsif ($arg =~ /^pauseFeedback$/i || $arg =~ /^ignoreFeedback$/i) {
        if (!defined $feedback_specified) {
           $feedback_specified = 1;
           if    ($arg =~ /^pauseFeedback$/i)  {$pauseFeedback = 1;}
           elsif ($arg =~ /^ignoreFeedback$/i) {$pauseFeedback = 0;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "pauseFeedback|ignoreFeedback": '.$ARGV[$i].$syntax);
        }
     }
     elsif ($arg =~ /^nowaitFeedbackRestore$/i || $arg =~ /^waitFeedbackRestore$/i) {
        if (!defined $waitFBR_specified) {
           $waitFBR_specified = 1;
           if    ($arg =~ /^nowaitFeedbackRestore$/i)  {$waitFBrestore = 'nowait';}
           elsif ($arg =~   /^waitFeedbackRestore$/i)  {$waitFBrestore = 'wait';}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "[no]waitFeedbackRestore": '.$ARGV[$i].$syntax);
        }
     }
     elsif ($arg =~ /^bringPinDiode/i || $arg =~ /^bringBeamStop/i) {
        if (!defined $bring_specified) {
           $bring_specified = 1;
           if    ($arg =~ /^bringPinDiode/i) {$bringPinDiode = 1;}
           elsif ($arg =~ /^bringBeamStop/i) {$bringPinDiode = 0;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "bringPinDiode|bringBeamStop": '.$ARGV[$i].$syntax);
        }
     }
     else {
        if (!defined $edge_specified) {
           $edge_specified = 1;
           $Edge = $arg;
           ($Edge_element, $Edge_KLMN) = split(/-/,$Edge);
           if ((!defined $Edge_element) || (!defined $Edge_KLMN)) {
              &press_key_exit('*** '.$script.': Incorrect command line: '
                              .$ARGV[$i].$syntax);
           }
           elsif (length($Edge_element) < 1 ||
                  length($Edge_element) > 2 ||
                  length($Edge_KLMN) < 1 ||
                  length($Edge_KLMN) > 2) {
              &press_key_exit('*** '.$script.': Incorrect command line: '
                              .$ARGV[$i].$syntax);
           }
        } else {
           &press_key_exit('*** '.$script.': duplicate command line parameter'
                          .' "Element-Edge": '.$ARGV[$i].$syntax);
        }
     }
  }
### ==================================================================== end of command line parameters

### Configuration parameter -- change if beamline hardware changes:
  if (!defined $have_CRL) {$have_CRL = 0;}					#see beamline.pl
  if (!defined $have_TF)  {$have_TF  = 0;}					#see beamline.pl

  if ($tell_bluice) {$ZENITY_VIA_BLUICE = 1;}
  if ($fakeflag && $moveID)  {
     &reportout('Fake mode: disabling undulator.','warn');
     $moveID = 0;
  }
  if ($fakeflag && $pauseFeedback)  {
     &reportout('Fake mode: disabling I0 feedback interaction.','warn');
     $pauseFeedback = 0;
  }

  if ($have_CRL || $have_TF) {
     if ($gotoEdge == 2) {
        $msg = 'Going to the edge switched from small steps to using bigEnergyMove because beamline has a CRL device';
        &reportout($msg,'warn');
        $gotoEdge = 1;
     }
  }

###########################################################################
### Postprocessing of cmdline arguments and preparing to data aquisition:

  &reportout('======================== Initializing Edge Scan ========================');

  &reportout('Reading list of intensity signals from mySQL...');
  $status = &getDarks4KeithleyList(\@signals);
  if ($status) {
     &zenity_error_exit('*** '.$script.': failed to get the list of signals from mySQL');
  }
  if (!grep(/^${i0_name}$/, @signals) && $i0_name !~ /^none$/i) {
     &zenity_error_exit('*** '.$script.': incorrect signal='.$i0_name
                              .'-- must be one of '."(@{signals})");
  }
  if ($i0_name !~ /^none$/i) {
     &reportout('Reading darks for '.$i0_name.' from mySQL...');
     $status = &getDarks4KeithleyGains($i0_name,\$timestampDB,\$epochDB,
                                    \$kchannel,\$klabelDB,\$koffsetDB,
                                    \$i0_channel,\@darks);
     if ($status) {
        &zenity_error_exit('*** '.$script.': failed to get data for '
                                 .$i0_name.' from mySQL');
     }
     $kgain   = &keithley_read_gain($kchannel);
     $klabel  = &keithley_read_label($kchannel);
     $koffset = &keithley_read_offset($kchannel);
     $epoch   = time();
     $bkg_age_hours = ($epoch - $epochDB)/3600.;
     $offset_delta  =  abs($koffset-$koffsetDB)
                    / &findMaxRef([abs($koffset),abs($koffsetDB),1]);

     if ($bkg_age_hours > 24 || $offset_delta > 0.01 || $klabel ne $klabelDB) {
        $msg = sprintf '%s darks age=%.2f hours offset change=%.2f%s label(mySQL)=%s label(current)=%s',
                       $i0_name, $bkg_age_hours, 100*$offset_delta, '%', $klabelDB, $klabel;
        &reportout($msg,'warn');
        if    ($bkg_age_hours > 24)  {&reportout($i0_name.' darks data is more than 24h old. Re-measuring...','warn');}
        elsif ($offset_delta > 0.01) {&reportout($i0_name.' offset changed by more than 1%. Re-measuring...','warn');}
        elsif ($klabel ne $klabelDB) {&reportout($i0_name.' Keithley label changed. Re-measuring...','warn');}
        @prg = ('perl', '-W', $PEZCA.'/measureDark1.pl', $beamline, $i0_name);
        $status = &execute_external_script(\@prg);
        if ($status) {&zenity_error_exit ('*** '.$script.': Failed to acquire '.$i0_name.' darks');}
        &reportout($i0_name.' darks are acquired.');
### Re-read updated I0 darks:
        $status = &getDarks4KeithleyGains($i0_name,\$timestampDB,\$epochDB,
                                          \$kchannel,\$klabelDB,\$koffsetDB,
                                          \$i0_channel,\@darks);
        if ($status) {&zenity_error_exit('*** '.$script.': failed to get data for '.$i0_name.' from mySQL');}
     }
  }

  if ($NONINTERACTIVE) {$autoplot = 1;}				#auto-print GNUplot after scan
  else                 {$autoplot = 0;}				#do not auto-print GNUplot

# if ($tell_bluice) {
#    &reportout('JBluIce interaction mode. Setting nochooch, nognuplot','warn');
#    $runchooch  = 0;						#do not run chooch in server mode
#    $usegnuplot = 0;						#do not plot in server mode
# }
  $status = &read_jbluice_config_flags('bigEnergyMove_flags', \@bigEnergyFlags);
  if ($status) {@bigEnergyFlags = ('multiStep', 'recoveryTime=auto');}
  if (! grep(/bluice$/i,@bigEnergyFlags)) {
     if ($tell_bluice) {unshift(@bigEnergyFlags, 'bluice');}
     else              {unshift(@bigEnergyFlags, 'nobluice');}
  }
  if (! grep(/^$beamline$/,@bigEnergyFlags)) {unshift(@bigEnergyFlags, $beamline);}
  @bigEnergyFlags = grep(! /^keepShutter$/i,@bigEnergyFlags);	#remove keepShutter
  if ($moveID == 0) {
     if (! grep(/^noID$/i,@bigEnergyFlags))  {unshift(@bigEnergyFlags, 'noID');}
  }

# &reportout("bigEnergyMove flags: (@bigEnergyFlags)");

  $linkID = 1;							#this is for energy pre-move
  if (length($undulator) == 0 || $beamline =~ /^23b/ || $beamline =~ /^23d/) {
    $linkID = 0;						#this is for energy pre-move
    $moveID = 0;						#this is for moving undulator to high scan end
    $pauseFeedback = 0;						#no feedback at BM yet
  }

### Auto scan mode:
  if ($scanMode == 10) {
     if ($cnt_time > 1.6772) {$scanMode = 1;}			#scanrecTBL
     else                    {$scanMode = 2;}			#fly
  }

  if    ($scanMode == 0) {@inputs = ();}			#step
  elsif ($scanMode == 1) {@inputs = ();}			#scanrecTBL
  elsif ($scanMode == 2) {@inputs = ($Flu_channel);}		#fly
  elsif ($scanMode == 3) {@inputs = ($Flu_channel);}		#scanrecTBLdiscr
  elsif ($scanMode == 4) {@inputs = (); $i0_name='none';}	#scanrecFly (no I0 records, no use of Joerger
  else  {&press_key_exit('*** '.$script.': incorrect scanMode='.$scanMode);}
  if ($i0_name !~ /^none$/i) {unshift (@inputs, $i0_channel);}
  $ninputs = @inputs;
  @ROIs   = (0);			       			#we only work with ROI-0 (hardcoded here)

### Fake data (for debugging purposes):
  if (!defined $fakeflag) {$fakeflag = 0;}
  if ($fakeflag) {
     $fakenpts = &readFakeData ($fakefile, \@fakearray);
#    &reportout('Fake array from '.$fakefile.' (npts='.$fakenpts.')','warn');
#    $msg ='';  for ($i=0; $i<$fakenpts; $i++) {$msg .= '  '.$fakearray[$i];}
#    &reportout(&msg."\n");
  }

  &reportout('+----------------------------------------------------------+');
  &reportout('| This is an Edge (Energy) scan with fluorescence detector |');
  &reportout('+--------------------------------------------------------- +');
  $mononame      = 'MO';
  $assembly      = 'En';
  $axisEnergy    = 'E';
  $assembly      = $beamline.':'.$mononame.':'.$assembly;
  $axisEnergy    = $beamline.':'.$mononame.':'.$axisEnergy;
  if (defined $USE_FALCON && $USE_FALCON) {
     $mca        = $beamline.':falcon:';
  } else {          
     $mca        = $beamline.':aim_adc1';
     $adc        = $beamline.':aim_adc1';
     $tca        = $beamline.':aim_tca1';
  }
  $gapCheckPV    = $beamline.':bi:enableEnergyGapCheck';
  $gapCheckState = 0;						#presume it is disabled
  $scaler        = $joergerName.$scalerNo;
  $scanrec       = $scanrecName;
  $scanAbort     = $scanrecAbort;
  $struck        = $struckName;

  if ($scaler    !~ /^${beamline}:/)  {$scaler    = $beamline.':'.$$scaler;}
  if ($scanrec   !~ /^${beamline}:/)  {$scanrec   = $beamline.':'.$scanrec;}
  if ($scanAbort !~ /^${beamline}:/)  {$scanAbort = $beamline.':'.$scanAbort;}
  if ($struck    !~ /^${beamline}:/)  {$struck    = $beamline.':'.$struck;}
  if (defined $shutter) {
     if ($shutter !~ /^${beamline}:/) {$shutter   = $beamline.':'.$shutter;}
  }
  if (defined $shutter_readback) {
     if ($shutter_readback !~ /^${beamline}:/) {
                                $shutter_readback = $beamline.':'.$shutter_readback;}
  }

### Use this sequence record to automate opening shutter followed by
### a short ~50ms delay followed by starting Joerger scaler [and MCA]:
  if ($scanMode == 0 || $scanMode == 1) {
### Start Joerger and MCA when:
### scanMode=0 (step/internal) or 
### scanMode=1 (scanrec/TBL)   or
### This is a sequence (seq) EPICS record for triggering:
###  23o:scanrec:startScalers.LNK1  =>  23o:SH:mp:bo PP MS
###  23o:scanrec:startScalers.LNK2  =>  23o:falcon:EraseStart PP MS
###  23o:scanrec:startScalers.LNK2  =>  23o:scaler1.CNT PP MS
###  23o:scanrec:startScalers.DOL1  1
###  23o:scanrec:startScalers.DOL2  1
###  23o:scanrec:startScalers.DOL3  1
### See: https://epics.anl.gov/EpicsDocumentation/AppDevManuals/RecordRef/Recordref-34.html
### Question: should we use multiple ScanRec triggers instead of this record?
     $startScalersPV = $beamline.':scanrec:startScalers';
     if (defined $USE_FALCON && $USE_FALCON) {
#       $mca_trig = $mca.'mca1.ERST';					#not working
        $mca_trig = $mca.'EraseStart';					#working, Falcon
     } else {
        $mca_trig = $mca.'.ERST';					#working, Canberra
     }
     $status = Pezca::PutStringOldCa($startScalersPV.'.LNK2', $mca_trig.' PP MS');
     &report_only($script, $status, $startScalersPV.'.LNK2');
     $startScalersPV .= '.PROC';
### Alternative:
#    $startScalersPV = $mca_trig;
#    $settleTime = 0;
#    $autowaitNr = 2;							#wait for MCA and scaler (see scanrec help)
#    &setupSRtriggers ($scanrec, [$mca_trig,$scaler.'.CNT',$shutter], [1,1,1], $settleTime, $autowaitNr);
  }
  elsif ($scanMode == 4) {
### scanMode=4 (scanrec/Fly):
     undef $startScalersPV;					#this is set in &setupSRdetectors_mca
  }
  elsif ($scanMode == 3) {
### Start Joerger only (no MCA) when
### scanMode=3 (scanrecTBLdiscr):
     $startScalersPV = $beamline.':scanrec:startScaler.PROC';
  }

### In the scanMode=0 (fly) we use Struck.
### In all other cases we use Joerger:
  if ($scanMode == 2) {$fileBkg = 'struck_Bkg.dat';}
  else                {$fileBkg = 'joerger'.$scalerNo.'_Bkg.dat';}
#-----------------------------------------------------------------
### JBluIce interaction PVs:
  $biEdgePV         = $beamline.':bi:SCAN:edge';		#where JBluIce posts Element-Edge
# $biEdgeEnergyPV   = $beamline.':bi:SCAN:edgeEnergy';		#where JBluIce posts Edge energy in eV
  $biScanCenterPV   = $beamline.':bi:SCAN:scanCenter';		#where JBluIce posts scan center in eV
  $bipostPV         = $beamline.':bi:SCAN:fastscanData';	#$flag,$nptH,@xH,@fluo
  $biEdgeModePV     = $beamline.':bi:SCAN:edgeMode';
  $biEdgeShiftMaxPV = $beamline.':bi:SCAN:edgeShiftMax';
#-----------------------------------------------------------------
  if (!defined $edge_specified || !defined $Edge_element || !defined $Edge_KLMN) {
     if ($tell_bluice) {
        &reportout('Element-Edge parameter is not specified. Using Element-Edge selected in JBluIce','warn');
        ($status,$Edge) =  Pezca::GetString($biEdgePV);
        if ($status) {&zenity_error_exit ('*** '.$script.': PVget failed on '.$biEdgePV);}
        ($Edge_element, $Edge_KLMN) = split(/-/,$Edge);
        if ((!defined $Edge_element) || (!defined $Edge_KLMN)) {
           &zenity_error_exit ('*** '.$script.': Incorrect Element-Edge data in JBluIce selection');
        }
     } else {
        &zenity_error_exit ('*** '.$script.': Element-Edge parameter is not specified.'.$syntax);
     }
  }
  if (!defined $adapt_specified) {
     if ($tell_bluice) {
        &reportout('Edge edgeShift parameter (none/adaptive/user) is not specified.'."\n"
                  .'Using edgeMode selected in JBluIce (PV='.$biEdgeModePV.')','warn');
        ($status,$arg) =  Pezca::GetDouble($biEdgeModePV);
        if ($status) {
            &report_only($script, $status, $biEdgeModePV);
            if ($tell_bluice) {&tellBluIce($scr.': cannot read '.$biEdgeModePV,'error');}
            $arg = $adaptive;							#apply default
        }
        $adaptive = $arg;
        if ($adaptive < 0 || $adaptive > 2) {
           &zenity_error_exit ('*** '.$script.': Incorrect edgeShift parameter (none/adaptive/user) in JBluIce selection');
        }
     } else {
        &reportout('Using default edgeShift='.$adaptive,'warn');
     }
  }

  if (!$tell_bluice && $adaptive > 1) {				#2=take edge energy from PV (user-specified edge)
     $adaptive = 1;
     &reportout('EdgeShift changed to '.$adaptive.' because no source of edge position without JBluIce','warn');
  }
  if ($adaptive != 0 && !defined $shiftmax_specified) {
     if ($tell_bluice) {
        &reportout('Edge edgeShiftMax parameter (eV) is not specified.'."\n"
                  .'Using JBluIce selection (PV='.$biEdgeShiftMaxPV.')','warn');
        ($status,$arg) =  Pezca::GetDouble($biEdgeShiftMaxPV);
        if ($status) {
           &report_only($script, $status, $biEdgeShiftMaxPV);
           if ($tell_bluice) {&tellBluIce($scr.': cannot read '.$biEdgeShiftMaxPV,'error');}
           $arg = $EdgeShiftMax;	#apply default
        }
        $EdgeShiftMax = $arg;
        &reportout('edgeShiftMax='.$EdgeShiftMax.' eV');
     } else {
        &reportout('Using default edgeShiftMax='.$EdgeShiftMax.' eV','warn');
     }
  }
#-----------------------------------------------------------------
### Read scan data from mySQL (energy in eV):
  $status = &getEdgeDataSQL($Edge_element,$Edge_KLMN,\$EdgeEnergySQL,\@E_step,\@E_pm,\$FluoEnergy,\$FluoPMwidth);
  if ($status) {
     $msg = 'No mySQL data for element='.$Edge_element.', Edge='.$Edge_KLMN;
     if ($tell_bluice) {&tellBluIce($scr.': '.$msg,'error');}
     &zenity_error_exit ('*** '.$script.': '.$msg);
  }
  $EdgeEnergy = $EdgeEnergySQL;
  $N_step = @E_step;
  $N_pm   = @E_pm;

  if ($adaptive == 2) {						#2=take edge energy from ScanCenterPV (user-specified edge)
     &reportout('EdgeShift=user. Reading Scan Center PV ('
                .$biScanCenterPV.') to determine EdgeShift.');
     ($status,$ScanCenter) =  Pezca::GetDouble($biScanCenterPV);
     if ($status) {
        &report_only($script, $status, $biScanCenterPV);
        if ($tell_bluice) {&tellBluIce($scr.': cannot read '.$biScanCenterPV,'error');}
        $ScanCenter = $EdgeEnergy;
     }
     $E_chemshift = $ScanCenter - $EdgeEnergy;
     $msg = sprintf 'ScanCenter=%.2feV  Edge=%.2feV  EdgeShift=%.2feV',
                             $ScanCenter, $EdgeEnergy, $E_chemshift;
     &reportout($msg);
     if (abs($E_chemshift) > $EdgeShiftMax) {
        $msg = sprintf 'User-specified edge=%.2feV is deviated'
                      .' from tabular edge=%.2feV by shift=%.2feV.'."\n"
                      .'This exceeds allowed maximum=%.2feV.',
                      $ScanCenter, $EdgeEnergy, $E_chemshift, $EdgeShiftMax;
        if (&zenity_question($msg."\n".'Continue?','Yes','No') != 1) {
#          &zenity_error_exit ('*** '.$script.': '.$msg);
           &clearAllCaMonitors();
           exit 1;
        }
     }
     elsif (abs($E_chemshift) > $E_step[0]) {
        $msg = sprintf $scr.': user-specified edge shift=%.2feV', $E_chemshift;
        &reportout($msg,'warn');
     }
     $EdgeEnergy = $ScanCenter;
  } else {
     $ScanCenter = $EdgeEnergy;
  }
  $E_chemshift = 0;

### Calculate step scan points (energy in eV):
  $status = &makeEnergyArrayRef($EdgeEnergy,
                                \@E_pm, \@E_step,
                                \@E, \$msg);
  if ($status) {&zenity_error_exit ('*** '.$script.': '.$msg);}

  $nE = @E;							#energy in eV
  if ($nE > 999 && $tell_bluice) {
     $msg = 'Number of pts in the scan='.$nE.' exceeds EPICS array. Truncating to 999';
     if ($tell_bluice) {&tellBluIce($scr.': '.$msg,'warn');}
     &zenity_error ('!!! '.$script.': '.$msg);
     splice (@E, 999, $nE-999);
     $nE = 999;
  }
  $E_start = $E[0];						#energy in eV
  $E_end   = $E[$nE-1];                                 	#energy in eV
  if ($tell_bluice) {
     $cmdline = sprintf 'setGraphXLimit %.2f %.2f', $E_start, $E_end;
     &reportout('Sending "'.$cmdline.'" to JBluIce..');
     $status = &tellBluIce($cmdline,'direct');
  }

###########################################################################
### Process input & prepare to scan:

### Output scan info:
  &reportout('# -----------------------------------------');
  &reportout('# Monochromator: '.$axisEnergy);
### In the scanMode=0 (fly) we use Struck.
### In all other cases we use Joerger:
  if ($scanMode == 2) {&printStruckInputs(\*STDOUT, $struck, $ninputs, @inputs);}
  else                {&printJoergerInputs(\*STDOUT, $scaler, $ninputs, @inputs);}
  $msg = sprintf '# Element=%s, Edge=%s, Edge_energy=%.2feV',$Edge_element,$Edge_KLMN,$EdgeEnergySQL;
  &reportout($msg);
  $msg = sprintf '# E_start=%.2feV, E_end=%.2feV, NPTS='.$nE."\n",$E_start,$E_end;
  &reportout($msg);
  &reportout('# Regions = '.$N_step.' (at each side)');
  &reportout('# Regions Steps = ('."@{E_step}".') eV');
  &reportout('# Regions sizes = ('."@{E_pm}".') eV');
  $msg = '# Scanning mode: ';
  if    ($scanMode == 0) {$msg .= 'StepScan/script';}
  elsif ($scanMode == 1) {$msg .= 'TableScan/scanrec';}
  elsif ($scanMode == 2) {$msg .= 'FlyScan/Struck with ADC/TCA discrimination (no MCA)';}
  elsif ($scanMode == 3) {$msg .= 'TableScan/scanrec with ADC/TCA discrimination (no MCA)';}
  elsif ($scanMode == 4) {$msg .= 'FlyScan/scanrec';}
  else                   {$msg .= '*** UNKNOWN'; $scanMode=0;}
  &reportout($msg);
  if    ($adaptive == 1) {&reportout('# Adaptive scan: YES');}
  elsif ($adaptive == 2) {&reportout('# Adaptive scan: NO, but use user-specified edge');}
  else                   {&reportout('# Adaptive scan: NO');}

  if ($NONINTERACTIVE == 0) {&ask_to_continue();}

### 1. Check if requested energy is close to where we are now

### Record initial positions:
  $E_ini = &getAxisAct($axisEnergy);				#record position before scan start
  $msg = sprintf 'Recorded initial mono position E=%.2feV', 1000.*$E_ini;
  &reportout($msg);
  if ($moveID) {
     $E_undini = &getIDEnergy($undulator);
     $msg = sprintf 'Recorded initial undulator position E=%.2feV', 1000.*$E_undini;
     &reportout($msg);
  }

### Read mySQL 'blc2004/Config' table:
  $status = &getIDparams(\$enmono13,\@harmEn,\@harmNum,
                         \@Elower,\@Olower,\@Eupper,\@Oupper);			# a more-friendly wrapper around &getVariableIDoffsets
  if ($status) {
     &zenity_error_exit('Cannot read Mono-ID Offsets data from mySQL',1);
  }
  $IDoffset    = undef;
  $syncMode    = undef,
  $syncStep    = undef;
  $switchID    = undef;
  $attenuation = undef;
  $switchMono  = undef; 

### Initially move the mono [and optionally the undulator] to the edge:
### $gotoEdge: 0=gotoEdgeNone, 1=gotoEdge with BigEnergyMove, 2=gotoEdgeSmall
  $E_edge = $EdgeEnergy/1000.;
  if ($gotoEdge == 1) {						#1=gotoEdge with BigEnergyMove
### Any energy changes are allowed:
     $msg = sprintf 'Pre-running energy to the edge=%.2feV with bigEnergyMove.', $EdgeEnergy;
     &reportout($msg);
     @prg = ('perl', '-W', $PEZCA.'/bigEnergyMove.pl', $E_edge, @bigEnergyFlags);
     $status = &execute_external_script(\@prg);
     if ($status) {
        if ($status != 512 && $status != 2) {			#'exit 1' gives 256 & 'exit 2' gives 512
           $msg = 'Error='.$status.' changing energy with bigEnergyMove.pl';
           if ($tell_bluice) {&tellBluIce($scr.': '.$msg,'error');}
           &zenity_error_exit ('*** '.$script.': '.$msg);
        } else {
           &reportout('bigEnergyMove was aborted by operator. Exiting','error');
           &clearAllCaMonitors();
           exit 1;
        }
     }
  }
  elsif ($gotoEdge == 2) {					#2=gotoEdgeSmall
### Only small energy changes are allowed:
     $msg = sprintf 'Pre-running energy to the edge=%.2feV with smallMove.', $EdgeEnergy;
     &reportout($msg);
     $Theta_big = 1.;						#big theta change when re-alligning may be needed
     $ESinTheta = &getESinTheta($assembly);
     $Theta_ini = &E2theta($assembly,$E_ini,$ESinTheta);
     $Theta_edge = &E2theta($assembly,$EdgeEnergy/1000.,$ESinTheta);
     if (abs($Theta_edge-$Theta_ini) > $Theta_big) {
        $msg = sprintf 'Current energy=%.3fkeV is too far from the requested'
                      .' edge energy=%.3fkeV.', $E_ini,$EdgeEnergy/1000.;
        if ($tell_bluice) {&tellBluIce($scr.': '.$msg,'error');}
        &zenity_error_exit($script.': '.$msg."\n"
                           .'Please remove your sample and change the energy first'."\n"
                           .'The sample needs to be removed because the energy'."\n"
                           .'change may require beam optimization.');
     } else {
        $status = &runMono($axisEnergy, $assembly, $E_edge,
                           $linkID, $undulator, $IDoffset,
                           $syncMode, $syncStep, $switchID,
                           $attenuation, $switchMono,
                           \$enmono13, \@harmEn, \@harmNum, 
                           \@Elower, \@Eupper, \@Olower, \@Oupper);
        if ($status) {
           &reportout('Small energy pre-move failed. Exiting.','error');
           &clearAllCaMonitors();
           exit 1;
        }
     }
  }
  else {							#0=gotoEdgeNone
     &reportout('No pre-moving energy to the edge was requested.','warn');
  }

### 2. Bring collimator & beamstop IN, pindiode and backlight OUT
  if ($bringPinDiode) {
     &reportout('Moving collimator/pindiode IN and beamstop/backlight OUT'
               .' with sampleEnvSrv');
     @prg = ($beamline, 'collimatorIN', 'beamstopOUT', 'pindiodeIN', 'backlightOUT');
  } else {
     &reportout('Moving collimator/beamstop IN and pindiode/backlight OUT'
               .' with sampleEnvSrv');
#    @prg = ('IN', $beamline, 'pindiodeOUT', 'backlightOUT');
     @prg = ($beamline, 'collimatorIN', 'beamstopIN', 'pindiodeOUT', 'backlightOUT');
  }
  $status = &setSampleEnv(\@prg);
  if ($status) {
     &reportout('Failure='.$status.' from sampleEnv','error');
     &zenity_error_exit($script.': failed on sampleEnv');
  }

  $t0 = [gettimeofday];						#start measuring time

  if ($scanMode == 2) {						#if hw-sync fly scan
     if (length($fileBkg) > 0) {&reportout('Reading Struck scaler offsets from '.$fileBkg);}
     else                      {&reportout('No Struck scaler offsets file is specified. Setting offsets to zero.');}
     &reportout('Reading Struck scaler offsets from '.$fileBkg);
     @Offsets = &getStruckOffsets($fileBkg,$struck);
### Get initial Struck dwell:
     $dwell_ini = &getStruckDwell ($struck);
  } else {
### Read the scaler background offsets:
     if (length($fileBkg) > 0) {&reportout('Reading Joerger scaler offsets from '.$fileBkg);}
     else                      {&reportout('No Joerger scaler offsets file is specified. Setting offsets to zero.');}
     @Offsets = &getJoergerOffsets($fileBkg,$scaler);
  }

  if ($i0_name !~ /^none$/i) {
     $igain = &keithley_gain_index($kgain);
     $Offsets[$i0_channel-1] = $darks[$igain-3];			#replace by mySQL data
     $Off = $Offsets[$i0_channel-1];
  } else {
     $Off = 0;
  }

  if ($scanMode != 2) {
### set Joerger/Struck scaler:
     &reportout('Setting up Joerger scaler '.$scaler);
     &setJoerger($scaler, $cnt_time);
  }
### set MCA (even if we do not use it in scan, we still may have to optimize it and to get ROI)
  &reportout('Setting up MCA '.$mca);
  &setMCA($mca,$cnt_time);

### MCA ROI modes:
### 'optimize' = set MCA according to mySQL,
### 'ready'    = presume MCA is ready,
### 'auto'     = optimize only if not ready (if current ROI are all zero)
  if ($mca_setmode =~ /^auto$/i) {
### !!!! This may create danger of using old ROI from some other sample:
     &reportout('Checking ROI for MCA '.$mca);
     ($ROI_l, $ROI_h) = &checkMCA_ROI($mca);
     &reportout('MCA: ROI_low='.$ROI_l.',  RO_high='.$ROI_h);
     if ($ROI_l == $ROI_h) {
        &reportout('MCA has no ROI set. Will optimize MCA to tabular values.','warn');
        $mca_setmode = 'optimize';
     }
  }

### Optionally set MCA ROI (if requested MCA optimization):
  if ($mca_setmode =~ /^optimize$/i) {
     $mcaNMAX = &MCAnmax($mca);
     $E1 = $FluoEnergy - $FluoPMwidth;
     $E2 = $FluoEnergy + $FluoPMwidth;
     if (defined $USE_FALCON && $USE_FALCON) {
       ($ROI_l, $ROI_h) = &setFalconROI($mca,$E1,$E2);
        &reportout('Setting up MCA '.$mca.' ROI: low='.$ROI_l.' high='.$ROI_h);
     } else {
        &reportout('Optimizing MCA '.$mca.' for energy range ['.$E1.' : '.$E2.']');
        ($ROI_l, $ROI_h) = &optimizeMCA_ADC($E1,$E2,$mcaNMAX,$adc);
        &reportout('Setting up MCA '.$mca.' ROI: low='.$ROI_l.' high='.$ROI_h);
        &setMCA_ROI($mca,$ROI_l,$ROI_h);
     }
  }
  elsif ($mca_setmode =~ /^ready$/i) {
### Check MCA ROI
     &reportout('Checking ROI for MCA '.$mca.' :');
     ($ROI_l, $ROI_h) = &checkMCA_ROI($mca);
     &reportout('MCA: ROI_low='.$ROI_l.',  ROI_high='.$ROI_h);
     if ($ROI_l == 0 && $ROI_h == 0) {
        $msg = 'Cannot record fluorescence because ROI is not set';
        if ($tell_bluice) {&tellBluIce($scr.': '.$msg,'error');}
        &zenity_error_exit ('*** '.$script.': '.$msg);
     }
  }

### Check MCA ADC
  if (defined $adc) {
     &reportout('Reading ADC '.$adc.' :');
     ($gain, $offset) = &getMCA_ADC($adc);
     &reportout('ADC: gain='.$gain.',  offset='.$offset);
  }

### set MCA TCA
  if (defined $tca) {
     &reportout('Setting up TCA '.$tca.' :');
     $low = 5.;							#lower threshold for SCA-1
     &setupMCA_TCA($tca,$low);
     &reportout('TCA: low='.$low.',  high=100');
  }

### For on-the-fly scans with Struck set ADC & TCA discrimination
### according to the ROI:
  if ($scanMode == 2 || $scanMode == 3) {
     if (defined $adc && defined $tca) {&setDiscrFromROI($mca,$adc,$tca);}
  }
#-----------------------------------------------------------------
### This is ($sec,$min,$hour,$mday,$mon,$year$,$wday,$yday,$isdst):
  @ct = localtime(time);	#($sec,$min,$hour,$mday,$mon,$year$,$wday,$yday,$isdst)
  ($min,$hour,$mday,$monN,$year) = ($ct[1],$ct[2],$ct[3],$ct[4]+1,$ct[5]+1900);
  $mon  = ('Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec')[$ct[4]];
  $datestamp = sprintf '%04d-%02d-%02d_%02d-%02d',$year,$monN,$mday,$hour,$min;
  if (!defined $fileroot)  {$fileroot = $Edge.'_'.$datestamp;}
  $outfile = $fileroot.'.dat';
  $outpost = $fileroot.'.ps';
  $outefs  = $fileroot.'.efs';
  $logfile = $fileroot.'.log';
#-----------------------------------------------------------------
# $writelog = 1;						#2010.11.06: uncomment this for debugging
  if ($writelog) {
     &reportout('Opening log file: '.$logfile);
### Compose log filename:
     $filehandle = open (LOG,'> '.$logfile);
     if (defined $filehandle) {
        select LOG; $|=1; 					#set unbuffered output
        print  LOG '# Edge: '.$Edge."\n";
        print  LOG '# Monochromator: '.$axisEnergy."\n";
        printf LOG '# E_start=%.2feV, E_end=%.2feV, NPTS='.$nE."\n",$E_start,$E_end;
        print  LOG '# Regions = '.$N_step.' (at each side)'."\n";
        print  LOG '# Regions sizes = ('."@{E_pm}".') eV'."\n";
        print  LOG '# Regions Steps = ('."@{E_step}".') eV'."\n";
        if ($adaptive == 1) {
           print LOG '# Adaptive scan: may shift scan regions upon detecting chemical shift'."\n";
        }
        if ($scanMode == 2) {
           &printStruckInputs(\*LOG, $struck, $ninputs, @inputs);
           &printStruckOffsets(\*LOG, $fileBkg, $struck, \@Offsets, $ninputs, \@inputs);
        } else {
           &printJoergerInputs(\*LOG, $scaler, $ninputs, @inputs);
           &printJoergerOffsets(\*LOG, $fileBkg, $scaler, \@Offsets, $ninputs, \@inputs);
        }
        print  LOG '# MCA: ROI_low='.$ROI_l.',  RO_high='.$ROI_h."\n";
        print  LOG '# ADC: Gain='.$gain.',  Offset='.$offset."\n";
        print  LOG '# TCA: low='.$low.',  high=100'."\n";
        print  LOG "#\n";
     } else {
        &reportout('Cannot open LOG file='.$logfile,'error');
        $writelog = 0;
     }
  }
#-----------------------------------------------------------------
  if ($scanMode != 0) {
     if ($usegnuplot) {
        &reportout('Gnuplot is not supported with scan modes other than "step"','warn');
        $usegnuplot = 0;
     }
  }
  if ($usegnuplot) {
     $tmpfile = 'gnu.tmp';
     $logscale = 0;						#1=yes, 0=no
     $gnuopen = &open_gnuplot($E_start,$E_end,$logscale);	#this returns global GNUPLOT
     if ($gnuopen) {
### Open temp file for Gnuplot:
        $filehandle = open (GNUTMP,'> '.$tmpfile);		#overwrite GNUTMP file
        if (defined $filehandle) {
           select GNUTMP; $|=1;					#set unbuffered output
        } else {
           &reportout('Cannot open temp file for gnuplot ['.$tmpfile.']','error');
           &close_gnuplot();
           $gnuopen = 0;
        }
     }
  } else {
     $gnuopen = 0;
  }

  if ($tell_bluice) {&postBluIceArray ($bipostPV);}		#clear data array

#-----------------------------------------------------------------
### set Mono assembly into automatic start/stop
  &reportout('Setting up mono '.$assembly);
  &configAssy($assembly);
  &setAssyMonitors ($assembly);
  &setAxisMonitors ($axisEnergy);

### Pause BluIce energy gap warning (if needed):
  ($status, $gapCheckState) = Pezca::GetDouble($gapCheckPV);
  &report_only($script, $status, $gapCheckPV);
  if ($status) {$gapCheckState = 0;}				#presume paused
  if ($gapCheckState != 0) {
     &reportout('Pausing JBluIce Energy Gap warning');
     $status = Pezca::PutDoubleOldCa($gapCheckPV, 0);
     &report_only($script, $status, $gapCheckPV);
  }

  if ($moveID) {
     $E_KeV = $E_end/1000.;
     $IDoffset = &calcIDoffset($E_KeV, 0, 
                              \$enmono13, \@harmEn, \@harmNum, 
                              \@Elower, \@Eupper, \@Olower, \@Oupper);
     $E_KeV = $E_KeV + $IDoffset;
     $msg = sprintf 'Running undulator='.$undulator.' to the upper end'
                          .' of scan range: E=%.2feV', 1000.*$E_KeV;
     &reportout($msg);
     &runID($undulator,$E_KeV);
     $IDoffset = undef;
     select(undef,undef,undef,1.0);				#sleep a bit;
  }

  if ($pauseFeedback) {
     $mostab_initial = &getMostabState ($beamline);		#0=run, 1=pause, -1=error
     if ($mostab_initial == -1) {				#ERROR
        &reportout('Cannot talk to MOSTAB.','error');
        $pauseFeedback = 0;
     }
     elsif ($mostab_initial == 1) {				#PAUSE
        &reportout('No need to pause intensity feedback since it is stopped','warn');
        $pauseFeedback = 0;
     }
     else {
        &reportout('Pausing I0 feedback');
        $status = &pauseMostab ($beamline);
        if ($status) {
           &reportout('Error pausing I0 feedback','error');
           $pauseFeedback = 0;
        }
     }
  }
### Enable mono pitch:
# $status = Pezca::PutStringOldCa($beamline.':MO:mp2:StrCmd', 'O0 J/');
# &report_only($script, $status, $beamline.':MO:mp2:StrCmd');

#/////////////////////////////////////////////////////////////////
### do energy scan:
  @En_array = ();
  @Dt_array = ();
  @I0_array = ();
  $nE_accu  = 0;
#/////////////////////////////////////////////////////////////////
  if ($adaptive == 1) {						#ADAPTIVE MODE
### First do a fixed-step scan with the biggest step (1eV):
### (The @E array is not needed and can be overwritten)
     $dE = $E_step[$N_step-1];					#the biggest step
     $nE = int(abs(($E_end - $E_start)/$dE) + 1.5);
     if ($nE > 999 && $tell_bluice) {
        $msg = 'Number of requested pts='.$nE.' exceeds EPICS array. Truncating to 999';
        if ($tell_bluice) {&tellBluIce($scr.': '.$msg,'warn');}
        &zenity_error ('!!! '.$script.': '.$msg);
        $nE = 999;
     }
     @E = ();
     for ($i=0; $i<$nE; $i++) {$E[$i]=$E_start+$dE*$i;}
     $msg = sprintf 'Starting rough pass of Energy Scan with '.$nE
                   .' points in the energy range [%.2f:%.2f]..',
                   $E[0], $E[$nE-1];
     &reportout($msg);
     if ($scanMode == 0) {					#0=internal table step scan
        &do_internal_table_step_scan(\@E);
     }
     elsif ($scanMode == 1 || $scanMode == 3) {			#1=scanrec/TBL scan with MCA ROI, 3=scanrec/TBL scan without MCA
        &do_scanrec_table_step_scan(\@E);
     }
     elsif ($scanMode == 4) {					#4=scanrec/Fly scan with MCA ROI
        &do_1band_scanrec_fly_scan('RoughPass',$E[0],$E[$nE-1],$nE);
        ### Swap energy-reversed points and remove duplicates:
        $nE_accu = &removeDuplicates (\@En_array, \@Dt_array,
                                      \@I0_array, $dE);
     }
     elsif ($scanMode == 2) {					#2=Struck on-the-fly scan
        $ENSCAN_FLAGS->{'skip mtr link'}   = 1;
        $ENSCAN_FLAGS->{'skip mtr unlink'} = 1;
        &do_1band_struck_fly_scan('RoughPass',$E[0],$E[$nE-1],$nE);
        ### Swap energy-reversed points and remove duplicates:
        $nE_accu = &removeDuplicates (\@En_array, \@Dt_array,
                                      \@I0_array, $dE);
     }								#end of scan types

### Now, find the chemical shift with Marquardt-Levenberg
### non-linear least squares fitting:
     $E_chemshift = &calcChemShift(\@En_array, \@Dt_array,
                                   $EdgeEnergy, $dE);
     if (abs($E_chemshift) > $EdgeShiftMax) {
        $msg = sprintf '1st-pass fitted edge=%.2feV is deviated from'
                      .' tabular edge=%.2feV by shift=%.2feV exceeding'
                      .' the allowed maximum=%.2feV. Fit is ignored!',
                         $EdgeEnergy+$E_chemshift, $EdgeEnergy,
                         $E_chemshift, $EdgeShiftMax;
        &reportout($msg,'error');
        $E_chemshift = 0;
     } else {
        $msg = sprintf '1st-pass fitted edge shift=%.2feV', $E_chemshift;
        &reportout($msg,'warn');
     }

### Calculate step scan points (energy in eV) for the central (fine-sampled)
### part of the scan interval with the account for chemical shift:
     splice (@E_step, $N_step-1, 1);				#erase big-step interval from the list
     splice (@E_pm,   $N_step-1, $N_pm-$N_step+1);		#erase big-step interval from the list
     $status = &makeEnergyArrayRef($EdgeEnergy+$E_chemshift,
                                   \@E_pm, \@E_step,
                                   \@E, \$msg);
     if ($status) {&zenity_error_exit ('*** '.$script.': '.$msg);}

     $nE = @E;							#energy in eV
     if ($nE > 999 && $tell_bluice) {
        $msg = 'Number of scan pts='.$nE.' exceeds EPICS array. Truncating to 999';
        if ($tell_bluice) {&tellBluIce($scr.': '.$msg,'warn');}
        &zenity_error ('!!! '.$script.': '.$msg);
        $i = $nE - 999;
        if ($i < 0) {$i = 0;}
        splice (@E, $i, $nE-$i);
        $nE = $i;
     }
     if ($scanMode == 0) {					#0=internal table step scan
        &do_internal_table_step_scan(\@E);
     }
     elsif ($scanMode == 1 || $scanMode == 3) {			#1=scanrec/TBL scan with MCA ROI, 3=scanrec/TBL scan without MCA
        &do_scanrec_table_step_scan(\@E);
     }
     elsif ($scanMode == 4) {					#4=scanrec/Fly scan with MCA ROI
        $i = int(abs(($E[$nE-1]-$E[0])/$E_step[0])+1.5);
        &do_1band_scanrec_fly_scan('FinePass',$E[0],$E[$nE-1],$i);
     }
     elsif ($scanMode == 2) {					#2=Struck on-the-fly scan
        $i = int(abs(($E[$nE-1]-$E[0])/$E_step[0])+1.5);
        &do_1band_struck_fly_scan('FinePass',$E[0],$E[$nE-1],$i);
     }                                                          #end of scan types
### Swap energy-reversed points and remove duplicates:
     $nE_accu = &removeDuplicates (\@En_array, \@Dt_array, \@I0_array, $E_step[0]);
  }
#/////////////////////////////////////////////////////////////////
  else {							#non-ADAPTIVE MODE
     $msg = sprintf 'Starting Energy scan with '.$nE.' points in'
                   .' energy range [%.2f:%.2f]..', $E[0], $E[$nE-1];
     &reportout($msg);
     if ($scanMode == 0) {					#0=internal table step scan
        &do_internal_table_step_scan(\@E);
     }
     elsif ($scanMode == 1 || $scanMode == 3) {			#1=scanrec/TBL scan with MCA ROI, 3=scanrec/TBL scan without MCA
        &do_scanrec_table_step_scan(\@E);
     }
     elsif ($scanMode == 4) {					#1=scanrec/Fly scan with MCA ROI
        &do_3band_scanrec_fly_scan();
     }
     elsif ($scanMode == 2) {					#2=Struck on-the-fly scan
        &do_3band_struck_fly_scan();
     }                                                          #end of scan types
  }
#/////////////////////////////////////////////////////////////////

  if ($tell_bluice) {&postBluIceArray ($bipostPV, \@En_array, \@Dt_array, 'final');}
  if ($writelog)  {close(LOG);   }
  if ($gnuopen)   {close(GNUTMP);}
  $elapsed = tv_interval ($t0);     				#measure time (in seconds)
  $msg = sprintf 'Scan(s) finished! Total time spent=%.2fmin', $elapsed/60.;
  &reportout($msg);
# - - - - - - - - - - - - - - - - - -

### Restore initial configuration of Struck, MCA and energy motor;
### then return energy motor and undulator to the initial positions:
  &signal_handler('cleanup');
# - - - - - - - - - - - - - - - - - -
  if ($runchooch) {
### Open file for Chooch:
     &reportout('Opening file for Chooch: '.$outfile);
     $filehandle = open (CHOOCH,'> '.$outfile);			#overwrite CHOOCH file
     if (defined $filehandle) {
        select CHOOCH; $|=1;					#set unbuffered output
        $nE = @En_array;
        $E_end = $En_array[$nE-1];
        printf CHOOCH 'Edge='.$Edge.' Start=%.1f End=%.1f Cntr=%.1f'."\n",
                                      $E_start,  $E_end, $EdgeEnergy;
        print  CHOOCH $nE."\n";
        for ($i=0; $i<$nE; $i++) {
           printf CHOOCH "%12.5f \t %12.5f\n", $En_array[$i], $Dt_array[$i];
        }
        close(CHOOCH);
### Execute chooch here:
        if (-x $choochprg) {
           @prg = ($choochprg,
                   '-x',
                   '-e', $Edge_element,
                   '-a', $Edge_KLMN,
                   '-o', $outefs,
                   '-p', $outpost,
                   $outfile);
           &reportout('Starting Chooch: '."@prg");
           $status = system(@prg);
        } else {
           $status = 1;
        }
        if ($status) {
           &reportout('Error='.$status.' starting Chooch.','error');
        }
     } else {
        &reportout('Cannot write Chooch file '.$outfile,'error');
        $runchooch = 0;
     }
     if ($NONINTERACTIVE == 0) {&press_key_continue();}
 }
# - - - - - - - - - - - - - - - - - -
### Restore Intensity feedback (MOSTAB) if it was originally running:
  &restoreFeedback($waitFBrestore);

  if ($gnuopen) {
     &print_gnuplot ($autoplot);				#this also closes gnuplot
#    &close_gnuplot ();
  }
  &clearAllCaMonitors();
  exit 0;

##################################################################

#==========================================================================

sub signal_handler(;$) {
  our ($ABORT_IN_PROCESS);
  my ($signame, $msg);
  $signame = shift;
  if ($signame !~ /^quit$/i && $signame !~ /^cleanup$/i) {
     &reportout('Caught a SIG'.$signame,'error');
  }

  if (defined $ABORT_IN_PROCESS) {return;}
  $ABORT_IN_PROCESS = 1;
  &abort_external_script();

  if ($signame !~ /^cleanup$/i) {$report = 1;}

  if (defined $shutter && defined $beamline) {
     if ($shutter !~ /^${beamline}/) {$shutter = $beamline.':'.$shutter;}
     &closeShutter($shutter,$shutter_readback);
  }
  if (defined $axisEnergy && defined $assembly && defined $E_ini) {
     if    ($scanMode == 0) {	#-------------------------------#0=internal table step scan
        &abortAssy ($assembly);
     }
     elsif ($scanMode == 1) {					#1=scanrec/TBL scan with MCA ROI
        if (defined $scanAbort) {&abortSR($scanAbort,$scanrec);}
        &abortAssy ($assembly);
     }
     elsif ($scanMode == 2) {	#-------------------------------#2=Struck on-the-fly scan
        if (defined $struck) {
           &abortStruck ($struck);
           &setStruckReadRate ($struck,'2 second');
           &eraseStruck ($struck);
           if (defined $dwell_ini) {
              &setStruckDwell ($struck, $dwell_ini);		#Restore initial Struck dwell:
           }
        }
        &abortAssy ($assembly);
        if (defined $mtrOut) {
          &unlinkMtrOut($assembly, $mtrOut);
          &restoreAxis($assembly, $axisEnergy);		#restore speed and motion prg
        }
        if (defined $USE_FALCON && $USE_FALCON) {
           $mcaNMAX = &MCAnmax($mca);
           &setFalconSCA($mca,0,$mcaNMAX-1);
        } else {
           &setupMCA_ADC($adc);
           &setMCA_TCA($tca,$low,100);
        }
     }
     elsif ($scanMode == 3) {	#-------------------------------#3=scanrec/TBL scan without MCA
        if (defined $scanAbort) {&abortSR($scanAbort,$scanrec);}
        &abortAssy ($assembly);
        if (defined $USE_FALCON && $USE_FALCON) {
           $mcaNMAX = &MCAnmax($mca);
           &setFalconSCA($mca,0,$mcaNMAX-1);
        } else {
           &setupMCA_ADC($adc);
           &setMCA_TCA($tca,$low,100);
        }
     }                          #-------------------------------#end of scan mode options
#    $msg = sprintf 'Returning Mono to E_ini=%.2feV',1000.*$E_ini;
#    &reportout($msg);
#    &startNMotors ($assembly, 1, [$axisEnergy], [$E_ini], 'checkAuto');
#    $motionTime = 30;
#    if (defined $undulator && defined $E_undini && defined $moveID) {
#       if ($moveID && length($undulator) > 0) {
#          $msg = sprintf 'Returning undulator='.$undulator.' to E_ini=%.2feV', 1000.*$E_undini;
#          &reportout($msg);
#          &runID($undulator,$E_undini);
#       }
#    }
#    &trackMotor ($assembly, $motionTime, $axisEnergy);
     if ($have_TF || $have_CRL) {
        $msg = sprintf 'Returning beamline Energy to E_edge=%.2feV before going to E_ini=%.2feV', 1000.*$E_edge, 1000.*$E_ini;
        &reportout($msg);
        $status = &runMono($axisEnergy, $assembly, $E_edge,
                           $linkID, $undulator, $IDoffset,
                           $syncMode, $syncStep, $switchID,
                           $attenuation, $switchMono,
                           \$enmono13, \@harmEn, \@harmNum, 
                           \@Elower, \@Eupper, \@Olower, \@Oupper);
        if ($status) {
           $msg = sprintf 'Returning beamline Energy to E_edge=%.2feV failed. It may screw up CRL focus.', 1000.*$E_edge;
           &reportout('Small energy pre-move failed. Exiting.','error');
           &reportout($msg,'warn');
        }
     }
     $msg = sprintf 'Returning beamline Energy to E_ini=%.2feV with bigEnergyMove', 1000.*$E_ini;
     &reportout($msg);
     @prg = ('perl', '-W', $PEZCA.'/bigEnergyMove.pl', $E_ini, @bigEnergyFlags);
     $status = &execute_external_script(\@prg);
  }

  if (defined $gapCheckPV && defined $gapCheckState) {
     if ($gapCheckState != 0) {
        &reportout('Restoring JBluIce Energy Gap warning');
        $status = Pezca::PutDoubleOldCa($gapCheckPV, $gapCheckState);
        &report_only($script, $status, $gapCheckPV);
     }
  }

### Restore Intensity feedback (MOSTAB) if it was originally running:
  if ($signame =~ /^cleanup$/i) {&restoreFeedback($waitFBrestore);}
  else                          {&restoreFeedback('nowait');}


  if ($signame =~ /^cleanup$/i) {return;}
  &clearAllCaMonitors();
  exit 1;
}

#==========================================================================

# $status = &makeEnergyArrayRef($E_center,\@E_pm,\@E_step,\@E,\$message);
sub makeEnergyArrayRef ($\@\@\@\$) {
### Fills @E and returns $status.
  my ($E_center, $E_pm_ref, $E_step_ref, $E_ref, $messageRef);
  my ($script, $msg, $N_pm, $N_step, $g, $i, $j);
  my ($E_start, $E_end, $dE, $E_npts, $nE);
  $script     = 'makeEnergyArrayRef';
  $E_center   = shift(@_);
  $E_pm_ref   = shift(@_);
  $E_step_ref = shift(@_);
  $E_ref      = shift(@_);
  $messageRef = shift(@_);

  @$E_ref = ();				#initialize (clear) the positions array
  $N_pm   = @$E_pm_ref;
  $N_step = @$E_step_ref;
  &reportout('Making energy positions table with '.$N_pm.' step sizes');
  if ($N_pm != $N_step && $N_pm != $N_step+1) {
     $$messageRef = 'Incorrect number of energy steps or scan'
                   .' intervals: N_step='.$N_step.', N_pm='.$N_pm;
     &reportout($$messageRef,'error');
     return -1;
#    &press_key_exit();
  }
### Test energy intervals for reasonable values:
  for ($i=0; $i<$N_pm; $i++) {
     $g = @$E_pm_ref[$i];
     if ($g <= 0.) {
        $$messageRef = 'Negative E_pm('.$i.')='.$g;
        &reportout($$messageRef,'error');
        return -1;
#       &press_key_exit();
     }
  }
### Test energy steps for reasonable values:
  for ($i=0; $i<$N_step; $i++) {
     $g = @$E_step_ref[$i];
     if ($g <= 0.) {
        $$messageRef = 'Negative E_step('.$i.')='.$g;
        &reportout($$messageRef,'error');
        return -1;
#       &press_key_exit();
     }
  }
### Test energy center for reasonable value:
  if ($E_center < 4000 || $E_center > 35000) {
     $$messageRef = 'E_center='.$E_center.'eV not in reasonable range [4,000-35,000]eV';
     &reportout($$messageRef,'error');
     return -1;
#    &press_key_exit();
  }
  $nE = 0;
### Create scan intervals at the left of the peak, v1:
# for ($i=0; $i<$N_step; $i++) {
#    $E_start = $E_center;
#    for ($j=0; $j<$N_step-$i; $j++) {$E_start = $E_start - @$E_pm_ref[$j];}
#    $E_end  = $E_start + @$E_pm_ref[$N_step-$i-1];
#    $dE     = @$E_step_ref[$N_step-$i-1];
#    $E_npts = int( ($E_end - $E_start) / $dE);
#    for ($j=0; $j<$E_npts; $j++) {@$E_ref[$nE] = $E_start + $j * $dE;  $nE++;}
# }
### Create scan intervals at the left of the peak, v2:
  for ($i=$N_step; $i>0; $i--) {
     $E_start = $E_center;
     for ($j=0; $j<$i; $j++) {$E_start = $E_start - @$E_pm_ref[$j];}
     $E_end  = $E_start + @$E_pm_ref[$i-1];
     $dE     = @$E_step_ref[$i-1];
     $E_npts = int( ($E_end - $E_start) / $dE);
     for ($j=0; $j<$E_npts; $j++) {@$E_ref[$nE] = $E_start + $j * $dE;  $nE++;}
  }
### Create scan intervals at the right of the peak:
  for ($i=0; $i<$N_step; $i++) {
     $E_start = $E_center;
     for ($j=0; $j<$i; $j++) {$E_start = $E_start + @$E_pm_ref[$j];}
     $E_end  = $E_start + @$E_pm_ref[$i];
### Exception for the last assymmetric interval:
     if ($i == $N_step-1 && $N_pm > $N_step) {$E_end  = $E_start + @$E_pm_ref[$i+1];}
     $dE     = @$E_step_ref[$i];
     $E_npts = int( ($E_end - $E_start) / $dE);
     for ($j=0; $j<$E_npts; $j++) {@$E_ref[$nE] = $E_start + $j * $dE;  $nE++;}
  }

### Remove duplicate points, if exist:
  for ($i=1; $i<$nE; $i++) {
     if (@$E_ref[$i] <= @$E_ref[$i-1]) {
        $msg = sprintf 'Duplicate scan point removed at i=%5d,  E=%.2f', $i, @$E_ref[$i];
        &reportout($msg,'warn');
        splice(@$E_ref, $i, 1);
        $nE--;
     }
  }

### Debugging:
# open (TST,'>  /tmp/EnArray.dat') || die "Can't open EnArray.dat\n";
# for ($i=0; $i<$nE; $i++) {
#   printf TST '%5d.  %.2f'."\n", $i, @$E_ref[$i];
# }
# close(TST);
# press_key_exit();

  $messageRef = '';
  return 0;
}

#==========================================================================

# $status = &postBluIceArray($bipostPV,[\@En_array,\@Dt_array,'force|final|normal']);
sub postBluIceArray ($;\@\@$) {

  my ($script, $bipostPV, $En_array_ref, $Dt_array_ref);
  my (@biarray, $nE, $nD, $status, $force);

  $script       = 'postBluIceArray';
  $bipostPV     = shift(@_);
  $En_array_ref = shift(@_);
  $Dt_array_ref = shift(@_);
  $force        = shift(@_);	if (!defined $force) {$force='normal';}

  if (defined $En_array_ref && defined $Dt_array_ref) {
     if    ($force =~ /^final$/i) {@biarray = (3);}	#tell JBluIce that this is final array
     elsif ($force =~ /^force$/i) {@biarray = (2);}	#force BluIce to refresh the whole plot
     else                         {@biarray = (1);}
     $nE = @$En_array_ref;
     $nD = @$Dt_array_ref;
     if ($nE != $nD) {
        &reportout('Positions-Data arrays mismatch. Pos='.$nE.'  Dat='.$nD,'warn');
#       $nE = &minval($nE,$nD);
     }
     push (@biarray, $nE);
     push (@biarray, @$En_array_ref);
     push (@biarray, @$Dt_array_ref);
#    print "@{biarray}\n";
  } else {
     @biarray = (0, 0);					#clear data array
  }

# $nbiarray = @biarray;
# &reportout('Posting JBluIce array:','note','');
# my $msg = '';
# for ($i=0; $i<$nbiarray; $i++) {$msg .= '  '.$biarray[$i];}
# &reportout($msg,'note','');

  $status = Pezca::PutList($bipostPV, 'DOUBLE', @biarray);
  if ($status) {
     &reportout('Error='.$status.' posting JBluIce data into PV='.$bipostPV,'error');
  }
  return $status;
}

#==========================================================================

# $fakenpts = &readFakeData($fakefile,\@fakearray);
sub readFakeData ($\@) {

  my ($script, $fakefile, $fakearray_ref);
  my ($filehandle, $line, $i, $string, $fakenpts, $energy);

  $script        = 'readFakeData';
  $fakefile      = shift(@_);
  $fakearray_ref = shift(@_);

# &reportout('Opening fake data file: '.$fakefile);
  $filehandle = open (FAKE,'< '.$fakefile);
  if (!defined $filehandle) {
     &press_key_exit('*** '.$script.': failed to open fake file='.$fakefile."!\n");
  }
  $line = 0;
  $i = 0;
  @$fakearray_ref = ();
  while (<FAKE>) {
     $line++;
     $string = $_;						#read line
     chomp $string;						#trail CR
     if    ($line == 1) {
        if ($string !~ /^Chooch/) {
           &reportout('Fake file='.$fakefile.' does not seem to be the Chooch file!','warn');
        }
     }
     elsif ($line == 2) {
        $fakenpts=$string;
     }
     else {
        ($energy,@$fakearray_ref[$i]) = split(/\s+/,$string);  	#split into words
        @$fakearray_ref[$i] = @$fakearray_ref[$i] * 1.E8;	#magnify to typical intensity
        $i++;
     }
  }
  if ($fakenpts != $i) {
     &reportout('Inconsistent structure of fake file='.$fakefile
               .' -- number of data lines does not match header data.','warn');
     $fakenpts = $i;
  }
  close(FAKE);
# &reportout('Fake array from '.$fakefile.' (npts='.$fakenpts.')','note','');
# my $msg = '';
# for ( $i=0; $i<$fakenpts; $i++) {$msg .= '  '.@$fakearray_ref[$i];}
# &reportout($msg,'note','');
  return $fakenpts;
}

#==========================================================================

# &loadFakeData($pt,\$Dt,\$I0,\@fakearray,$Off);
sub loadFakeData ($\$\$\@$) {

### Replace real data with fake for debugging purposes

  my ($script, $pt, $Dt_ref, $I0_ref, $fakearray_ref, $Off, $fakenpts);

  $script        = 'loadFakeData';
  $pt            = shift(@_);
  $Dt_ref        = shift(@_);
  $I0_ref        = shift(@_);
  $fakearray_ref = shift(@_);
  $Off           = shift(@_);

  $fakenpts = @$fakearray_ref;

  if ($pt < $fakenpts) {$$Dt_ref = @$fakearray_ref[$pt];}
  else                 {$$Dt_ref = @$fakearray_ref[$fakenpts-1];}
  $$I0_ref = 100. + $Off;
}

#==========================================================================

# $actPts = &removeDuplicates(\@En,\@Dt,\@I0,$step);
sub removeDuplicates (\@\@\@$) {

### Compress on-the-fly scan data by removing duplicate and out-of-order pts

  my ($script, $En_ref, $Dt_ref, $I0_ref, $step, $actPts, $actPtsNew, $sorted);

  $script  = 'removeDuplicates';
  $En_ref  = shift(@_);
  $Dt_ref  = shift(@_);
  $I0_ref  = shift(@_);
  $step    = shift(@_);

  $actPts = @$En_ref;

### Sort the arrays in the order of increasing energy:
  $sorted = &sort_by_x('incr', $En_ref, $Dt_ref, $I0_ref);
  if ($sorted > 0) {
     &reportout('Scan energy points have been re-ordered.','warn');
  }

### Remove duplicate points:
  $actPtsNew = &remove_duplicates(0.25, $step, $En_ref, $Dt_ref, $I0_ref);

  if ($tell_bluice && ($sorted || $actPtsNew != $actPts)) {
     &reportout('Posting scan data ordered over energy and free from duplicates');
     &postBluIceArray ($bipostPV, $En_ref, $Dt_ref, 'force');
  }
  return $actPtsNew;
}

#==========================================================================

# &do_internal_table_step_scan(\@E)
sub do_internal_table_step_scan (\@) {
#
  my ($E_ref, $act1, $act2, @I0, $sum_data, $net_data);
  my ($pt, $dead_time, $I0_corr, $E_eV, $nE, $msg);

  $E_ref = shift(@_);
  $nE = @$E_ref;

  for ($pt=1; $pt <= $nE; $pt++) {	#===============================
     $E_eV  = @$E_ref[$pt-1];
     $E_KeV = $E_eV / 1000.;					#in KeV

     if ($pt == 1) {
        $msg = sprintf 'Moving mono to starting position E=%.2f eV', $E_eV;
        &reportout($msg);
     }
     &runMotor ($assembly, $axisEnergy, $E_KeV, 'checkAuto');
     $act1 = &getAxisAct ($axisEnergy); 			#ActPos just after stop
     $act1 = 1000.*$act1;					#recalc in eV

### Run measurements:
     if (defined $startScalersPV) {
        $status = Pezca::PutDoubleOldCa($startScalersPV, 1);
        &report_only($script, $status, $startScalersPV);
     } else {
        &openShutter($shutter,$shutter_readback);
        select(undef,undef,undef,0.05);				#sleep(50ms);
        &startJoerger($scaler);
        &startMCA($mca);
     }
     &waitJoerger($scaler,$cnt_time);
     &waitMCA($mca);
     &closeShutter($shutter,$shutter_readback);
     $act2 = &getAxisAct ($axisEnergy);				#ActPos after measurements
     $act2 = 1000.*$act2;					#recalc in eV

     if ($i0_name !~ /^none$/) {@I0 = &readJoerger ($scaler, $cnt_time, 1, $i0_channel);}
     else                      {@I0 = (1.);}
     ($sum_data, $net_data) = &readMCA_ROI($mca);
     $dead_time = &readMCA_deadTime($mca);
     if ($fakeflag) {&loadFakeData($pt,\$sum_data,\$I0[0],\@fakearray,$Off);}
     $I0_corr = $I0[0] - $Off;
     if ($I0_corr <= 0.) {$I0_corr = 1.;}			#avoid zero/negative intensity
     if ($I0[0] > $Off) {
        push (@En_array, 0.5*($act1+$act2));
        push (@Dt_array, $sum_data/$I0_corr);
        $msg = sprintf '%5d. E=%-9.2f  I0=%-8.0f   MCA=%-8.0f   MCA/I0=%-12.5g  DeadTime=%-7.3f',
                      $pt, $En_array[$nE_accu], $I0_corr, $sum_data, $Dt_array[$nE_accu], $dead_time;
        &reportout($msg);
        if ($tell_bluice && $nE_accu>=1) {&postBluIceArray ($bipostPV, \@En_array, \@Dt_array);}
        if ($gnuopen) {
           printf GNUTMP '%12.5f    %12.5g'."\n", $En_array[$nE_accu], $Dt_array[$nE_accu];
           if ($nE_accu >= 1 &&  -s $tmpfile > 100) {
### Plot with GnuPlot:  plot "file" using 1:2, "file" using 1:3, .....
              print GNUPLOT "\n";				#to avoid PERL warnings
              print GNUPLOT 'plot "'.$tmpfile.'"'."\n";
           }
        }
        $nE_accu++;
     } else {
        $msg = sprintf '%-5d. E=%.2f  ==> Incorrect I0=%-8.0f', $pt, $E_eV, $I0[0]-$Off;
        &reportout($msg,'error');
     }
     if ($writelog) {
        if ($pt == 1) {						#LOG header
           print LOG '#   N/Total      E(rqs)     E(act1)     E(act2)',
                     '         I0        MCA(ROI)     MCA/I0    MCA DeadTime',"\n";
        }
        printf LOG   "%5d/%2d    %10.3f  %10.3f  %10.3f   %8d.   %8.0f   %12.5g   %7.3f\n",
                     $pt, $nE, $E_eV, $act1, $act2,
                     $I0_corr, $sum_data, $sum_data/$I0_corr, $dead_time;
     }
  }	#===============================================================
  return;
}

#==========================================================================

# &do_scanrec_table_step_scan(\@E)
sub do_scanrec_table_step_scan(\@) {
#
  my ($E_ref, @EK, $nE, $settleTime, $afterScan, $wait_, $wait_step);
  my ($npts, $pt, $pb, @En_scanrec, @I0_scanrec, @Dt_scanrec, $busy, $msg);

  $E_ref = shift(@_);
  $nE = @$E_ref;

  &setJoergerTime($scaler, $cnt_time+0.2);			#add time to be sure it finishes later than MCA

### Setup scan record:
  &reportout('Setting up scan record ['.$scanrec.']...');
  $settleTime = 0;
  $afterScan  = 0;
  @EK=();  for ($pt=0; $pt<$nE; $pt++) {$EK[$pt]=@$E_ref[$pt]/1000.}
  &clearSR ($scanrec);
  &setupSRmotorsCommon ($scanrec, $nE, $settleTime, $afterScan);
  &setupSRmotorTBL ($scanrec, 1, $axisEnergy.':ActPos', $axisEnergy.':RqsPos', \@EK);
  &setupSRmotorLIN ($scanrec, 2, 'time', $shutter, $shutter_closed, $shutter_closed, 0);
  if    ($scanMode == 1) {						#scanrecTBL
     &setupSRdetectors_scaler_mca ($scanrec, $scaler, $mca, \@inputs, \@ROIs, $startScalersPV);
  }
  elsif ($scanMode == 3) {						#scanrecTBLdiscr
     &setupSRdetectors_pv ($scanrec, $scaler, \@inputs);
     Pezca::PutStringOldCa($scanrec.'.T1PV',$startScalersPV);
  }
  &setupSRafterScan ($scanrec, $shutter, $shutter_closed, 0);
# &setupSRrefreshrate ($scanrec, $cnt_time);			#for partial arrays
  &setupSRrefreshrate ($scanrec, 0.1);				#for partial arrays, 0.1 is the minumum
### Setting monitors on motors motion and scan record execution:
  $status = &addCaMonitor($scanrec.'.BUSY');
  $status = &addCaMonitor($scanrec.'.CPT');
### Start scan record:
  &reportout('Starting TBL scan with scan record ['.$scanrec.']...');
  &startSR ($scanrec);
  if ($medm) {
     if    ($scanMode == 1) {						#scanrecTBL
        &openSRmedm_scaler_mca($scanrec, \@inputs, \@ROIs, undef, 'realtime');
     }
     elsif ($scanMode == 3) {						#scanrecTBLdiscr
        &openSRmedm_pv ($scanrec, \@inputs, undef, 'realtime');
     }
  }
  $busy = 1;
  $wait_ = 0;
  $wait_step = $cnt_time + 1;
  while ($busy == 1) {						#while scan record is running
     select(undef,undef,undef,$wait_step);			#sleep($wait_step);
     $wait_ += $wait_step;
     ($status,$busy) = Pezca::GetDouble($scanrec.'.BUSY');
     ($status,$npts) = Pezca::GetDouble($scanrec.'.CPT');
### Read partial scanrec arrays here:
     if ($tell_bluice && $npts > 2) {
        ($status,@En_scanrec) = Pezca::GetList($scanrec.'.P1CA','double',$npts);
        &report_only($script, $status, $scanrec.'.P1CA');
        if ($i0_name !~ /^none$/i) {
           ($status,@I0_scanrec) = Pezca::GetList($scanrec.'.D01CA',$EZCAFLOAT,$npts);
           &report_only($script, $status, $scanrec.'.D01CA');
           ($status,@Dt_scanrec) = Pezca::GetList($scanrec.'.D02CA',$EZCAFLOAT,$npts);
           &report_only($script, $status, $scanrec.'.D02CA');
        } else {
           for ($pt=0; $pt<$npts; $pt++) {$I0_scanrec[$pt] = 1.}
           ($status,@Dt_scanrec) = Pezca::GetList($scanrec.'.D01CA',$EZCAFLOAT,$npts);
           &report_only($script, $status, $scanrec.'.D01CA');
        }
        if ($npts > 999) {$npts = 999;}
        for ($pt=0; $pt<$npts; $pt++) {
           $pb = $pt;
           $En_array[$pb] = $En_scanrec[$pt] * 1000.;
           $I0_array[$pb] = $I0_scanrec[$pt];			#here: @inputs=($i0_channel,$Flu_channel)
           $Dt_array[$pb] = $Dt_scanrec[$pt];			#here: @inputs=($i0_channel,$Flu_channel)
           if ($fakeflag) {
              &loadFakeData($pb,\$Dt_array[$pb],\$I0_array[$pb],\@fakearray,$Off);
           }
           if ($I0_array[$pb] > $Off) {$I0_array[$pb] = $I0_array[$pb] - $Off;}
           else                       {$I0_array[$pb] = 1.;}
           $Dt_array[$pb] = $Dt_array[$pb] / $I0_array[$pb];
        }
        &postBluIceArray ($bipostPV, \@En_array, \@Dt_array);
     }
#    &reportout($npts.' of '.$nE.' points done. Scanning...');
     print STDOUT $script.': '.$npts.' of '.$nE.' points done. Scanning...'."\n";
  }
  $msg = sprintf 'Scanrec scan finished. Time=%.1fs', $wait_;
  &reportout($msg);
### Read final scanrec arrays here:
  $npts = &waitSRdata($scanrec);
  ($status,@En_scanrec) = Pezca::GetList($scanrec.'.P1RA','double',$npts);
  &report_only($script, $status, $scanrec.'.P1CA');
  if ($i0_name !~ /^none$/i) {
     ($status,@I0_scanrec) = Pezca::GetList($scanrec.'.D01DA',$EZCAFLOAT,$npts);
     &report_only($script, $status, $scanrec.'.D01CA');
     ($status,@Dt_scanrec) = Pezca::GetList($scanrec.'.D02DA',$EZCAFLOAT,$npts);
     &report_only($script, $status, $scanrec.'.D02CA');
  } else {
     for ($pt=0; $pt<$npts; $pt++) {$I0_scanrec[$pt] = 1.}
     ($status,@Dt_scanrec) = Pezca::GetList($scanrec.'.D01DA',$EZCAFLOAT,$npts);
     &report_only($script, $status, $scanrec.'.D01CA');
  }
  if ($tell_bluice && $npts > 999) {
     $msg = 'Number of actual pts='.$npts.' exceeds EPICS array. Truncating to 999';
     if ($tell_bluice) {&tellBluIce($scr.': '.$msg,'warn');}
     &zenity_error ('!!! '.$script.': '.$msg);
     $npts = 999;
  }
  for ($pt=0; $pt<$npts; $pt++) {
     $pb = $pt;
     $En_array[$pb] = $En_scanrec[$pt] * 1000.;
     $I0_array[$pb] = $I0_scanrec[$pt];				#here: @inputs=($i0_channel,$Flu_channel)
     $Dt_array[$pb] = $Dt_scanrec[$pt];				#here: @inputs=($i0_channel,$Flu_channel)
     if ($fakeflag) {&loadFakeData($pb,\$Dt_array[$pb],\$I0_array[$pb],\@fakearray,$Off);}
     if ($I0_array[$pb] > $Off) {$I0_array[$pb] = $I0_array[$pb] - $Off;}
     else                       {$I0_array[$pb] = 1.;}
     $Dt_array[$pb] = $Dt_array[$pb] / $I0_array[$pb];
     if ($writelog) {
        if ($pb == 0) {						#create LOG header
           print LOG '#   N/Total          E           I0           MCA/I0',"\n";
        }
        printf LOG '%5d/%2d    %10.3f   %12.5g   %12.5g'."\n",
                   $pb+1, $npts, $En_array[$pb], $I0_array[$pb], $Dt_array[$pb];
     }
  }
  $nE_accu = $npts;
  $npts = @En_array;
  print STDOUT $script.': posting final array of length '.$npts."\n";
# if ($tell_bluice) {&postBluIceArray ($bipostPV, \@En_array, \@Dt_array, 'final');}
  return;
}


#==========================================================================

# &do_3band_scanrec_fly_scan();
sub do_3band_scanrec_fly_scan () {
#
  my (@E3B_start, @E3B_end, @E3B_step, $E_npts, $band, $scantag);

  $E3B_start[0] = $EdgeEnergy - $E_pm[0] - $E_pm[1] - $E_pm[2];
  $E3B_end[0]   = $E3B_start[0] + $E_pm[2];
  $E3B_step[0]  = $E_step[2];
#
  $E3B_start[1] = $E3B_end[0];
  $E3B_end[1]   = $EdgeEnergy + $E_pm[0] + $E_pm[1];
  $E3B_step[1]  = $E_step[0];
#
  $E3B_start[2] = $E3B_end[1];
  $E3B_end[2]   = $E3B_start[2] + $E_pm[3];
  $E3B_step[2]  = $E_step[2];
#
  &reportout('Doing 3-band scanrec-synchronized on-the-fly scan...');
  for ($band=0; $band<3; $band++) {
     $E_npts = int(abs(($E3B_end[$band]-$E3B_start[$band])/$E3B_step[$band])+1.5);
     $scantag = sprintf 'Band-%d/3',$band+1;
     &do_1band_scanrec_fly_scan ($scantag, $E3B_start[$band], $E3B_end[$band], $E_npts);
  }								#end of loop over the bands

### Swap energy-reversed points and remove duplicates:
  $nE_accu = &removeDuplicates (\@En_array, \@Dt_array, \@I0_array, $E_step[0]);
  return;
}


#==========================================================================

# &do_1band_scanrec_fly_scan($scantag, $E_start, $E_end, $E_npts)
sub do_1band_scanrec_fly_scan ($$$$) {
#
  our ($EXTRA_MOTION_TIME);	#extra allowance for tracking because
  our ($IGNORE_ERRORS);
  my ($scantag, $E_start_eV, $E_end_eV, $E_npts);
  my ($wait_, $wait_step);
  my ($npts, $pt, $pb, $msg);
  my ($srDeadTime, $mcaOverhead, $scanRange, $scanTimeReal);
  my ($busy, $l_speed, $h_speed, $scanMode, $settleTime, $afterScan);
  my (@En_scanrec, @I0_scanrec, @Dt_scanrec);

  $scantag    = shift(@_);
  $E_start_eV = shift(@_);
  $E_end_eV   = shift(@_);
  $E_npts     = shift(@_);

  $srDeadTime  = 0.005;						#ScanRec needs some time to finish/erase
  $mcaOverhead = 0.100;						#MCA needs some time to finish/erase [2008/09/17]

  $msg = sprintf 'Setting up ['.$scantag.'] fly scan with '.$E_npts
                .' points in range [%.2f-%.2f]eV...', $E_start_eV, $E_end_eV;
  &reportout($msg);
### set motor:
  &reportout('Setting up motor ['.$axisEnergy.']');
  &configAssy ($assembly);
  &restoreAxis ($assembly, $axisEnergy);			#restore high-speed

### set Scan Record:
  &reportout('Setting up scanrec ['.$scanrec.']');
  $settleTime = 0;
  $afterScan  = 0;
  $scanMode   = 2;						#2=on-the-fly
  &clearSR ($scanrec);
  &setupSRmotorsMore ($scanrec, $E_npts, 
                      $scanMode, $settleTime, $afterScan, 1, 
                      [$axisEnergy.':ActPos'], [$axisEnergy.':RqsPos'], 
                      [$E_start_eV/1000.], [$E_end_eV/1000.]);
  &setupSRdetectors_mca ($scanrec, $mca, \@ROIs);
  &setupSRafterScan ($scanrec, $shutter, $shutter_closed, 0);
  &setupSRrefreshrate ($scanrec, 0.1);				#for partial arrays, 0.1 is the minumum

### Setting monitors on motors motion and scan record execution:
  $status = &addCaMonitor($scanrec.'.BUSY');
  $status = &addCaMonitor($scanrec.'.CPT');
  $msg = sprintf 'Moving motor ['.$axisEnergy.'] to the starting position (%.3g)'."\n",$E_start_eV/1000.;
  &reportout($msg);

### Increase expected motion time (just in case):
  $EXTRA_MOTION_TIME = 600;
  $status = &runMotor ($assembly, $axisEnergy, $E_start_eV/1000.);
  undef $EXTRA_MOTION_TIME;
  if ($status) {
     $msg = 'Failed moving ['.$axisEnergy.'] to the starting position.';
     &zenity_error ('!!! '.$script.': '.$msg);
  }

### Calculate scan speed for the motor:
  &reportout('Calculating slew-mode motor speed for ['.$axisEnergy.']');
  $scanRange = abs ($E_end_eV - $E_start_eV)/1000.;
  $IGNORE_ERRORS = 1;
  $scanTimeReal = ($cnt_time + $srDeadTime + $mcaOverhead) * &maxval(1,($E_npts - 1));
  ($l_speed,$h_speed) = &calcAxisSpeed ($assembly,$axisEnergy,$scanRange,$scanTimeReal);
  $IGNORE_ERRORS = 0;
  if ($l_speed > $h_speed) {
     $msg = sprintf 'Calculated low speed=%.3f exceeds max drive speed=%.3f',$l_speed,$h_speed;
     &reportout('Calculating slew-mode motor speed for ['.$axisEnergy.']');
  }

### Setting slow speed for scanning:
  $msg = sprintf 'Setting up motor ['.$axisEnergy.'] speed=%.3g & motionPRG=slew'."\n",$l_speed;
  &reportout($msg);
  &setAxisSlewScan ($assembly, $axisEnergy, $l_speed);

  if ($medm) {&openSRmedm_mca ($scanrec, \@ROIs, undef, 'realtime');}

  &openShutter ($shutter,$shutter_readback);

### Start scan record:
  &reportout('Starting FLY scan with scan record ['.$scanrec.']...');
  &startSR ($scanrec);
  $busy = 1;
  $wait_ = 0;
  $wait_step = $cnt_time + 1;
  while ($busy) {						#while scan record is running
     select(undef,undef,undef,$wait_step);			#sleep($wait_step);
     $wait_ += $wait_step;
     ($status,$busy) = Pezca::GetDouble($scanrec.'.BUSY');
     ($status,$npts) = Pezca::GetDouble($scanrec.'.CPT');
### Read partial scanrec arrays here:
     if ($tell_bluice && $npts > 2) {
        ($status,@En_scanrec) = Pezca::GetList($scanrec.'.P1CA','double',$npts);
        &report_only($script, $status, $scanrec.'.P1CA');
        for ($pt=0; $pt<$npts; $pt++) {$I0_scanrec[$pt] = 1.}
        ($status,@Dt_scanrec) = Pezca::GetList($scanrec.'.D01CA',$EZCAFLOAT,$npts);
        &report_only($script, $status, $scanrec.'.D01CA');
        if ($npts > 999) {$npts = 999;}
        for ($pt=0; $pt<$npts; $pt++) {
           $pb = $pt;
           $En_array[$pb] = $En_scanrec[$pt] * 1000.;
           $I0_array[$pb] = $I0_scanrec[$pt];			#here: @inputs=($i0_channel,$Flu_channel)
           $Dt_array[$pb] = $Dt_scanrec[$pt];			#here: @inputs=($i0_channel,$Flu_channel)
           if ($fakeflag) {
              &loadFakeData($pb,\$Dt_array[$pb],\$I0_array[$pb],\@fakearray,$Off);
           }
           if ($I0_array[$pb] > $Off) {$I0_array[$pb] = $I0_array[$pb] - $Off;}
           else                       {$I0_array[$pb] = 1.;}
           $Dt_array[$pb] = $Dt_array[$pb] / $I0_array[$pb];
        }
        &postBluIceArray ($bipostPV, \@En_array, \@Dt_array);
     }
  }
  $msg = sprintf 'Scanrec scan finished. Time=%.1fs', $wait_;
  &reportout($msg);

  &closeShutter ($shutter,$shutter_readback);
  &abortSR ($scanAbort, $scanrec);
  &abortAssy ($assembly, 1, $axisEnergy);		# also restore speed and motion prg.

### Restore speed:
### Read final scanrec arrays here:
  $npts = &waitSRdata($scanrec);
  ($status,@En_scanrec) = Pezca::GetList($scanrec.'.P1RA','double',$npts);
  &report_only($script, $status, $scanrec.'.P1CA');
  for ($pt=0; $pt<$npts; $pt++) {$I0_scanrec[$pt] = 1.}
  ($status,@Dt_scanrec) = Pezca::GetList($scanrec.'.D01DA',$EZCAFLOAT,$npts);
  &report_only($script, $status, $scanrec.'.D01CA');
  if ($tell_bluice && $npts > 999) {
     $msg = 'Number of actual pts='.$npts.' exceeds EPICS array. Truncating to 999';
     if ($tell_bluice) {&tellBluIce($scr.': '.$msg,'warn');}
     &zenity_error ('!!! '.$script.': '.$msg);
     $npts = 999;
  }
  for ($pt=0; $pt<$npts; $pt++) {
     $pb = $pt;
     $En_array[$pb] = $En_scanrec[$pt] * 1000.;
     $I0_array[$pb] = $I0_scanrec[$pt];				#here: @inputs=($i0_channel,$Flu_channel)
     $Dt_array[$pb] = $Dt_scanrec[$pt];				#here: @inputs=($i0_channel,$Flu_channel)
     if ($fakeflag) {&loadFakeData($pb,\$Dt_array[$pb],\$I0_array[$pb],\@fakearray,$Off);}
     if ($I0_array[$pb] > $Off) {$I0_array[$pb] = $I0_array[$pb] - $Off;}
     else                       {$I0_array[$pb] = 1.;}
     $Dt_array[$pb] = $Dt_array[$pb] / $I0_array[$pb];
     if ($writelog) {
        if ($pb == 0) {						#create LOG header
           print LOG '#   N/Total          E           I0           MCA/I0',"\n";
        }
        printf LOG '%5d/%2d    %10.3f   %12.5g   %12.5g'."\n",
                   $pb+1, $npts, $En_array[$pb], $I0_array[$pb], $Dt_array[$pb];
     }
  }
  $nE_accu = $npts;
# if ($tell_bluice) {&postBluIceArray ($bipostPV, \@En_array, \@Dt_array, 'final');}
  return;
}

#==========================================================================

# &do_3band_struck_fly_scan();
sub do_3band_struck_fly_scan () {
#
  my (@E3B_start, @E3B_end, @E3B_step, $E_npts, $band, $scantag);

  $E3B_start[0] = $EdgeEnergy - $E_pm[0] - $E_pm[1] - $E_pm[2];
  $E3B_end[0]   = $E3B_start[0] + $E_pm[2];
  $E3B_step[0]  = $E_step[2];
#
  $E3B_start[1] = $E3B_end[0];
  $E3B_end[1]   = $EdgeEnergy + $E_pm[0] + $E_pm[1];
  $E3B_step[1]  = $E_step[0];
#
  $E3B_start[2] = $E3B_end[1];
  $E3B_end[2]   = $E3B_start[2] + $E_pm[3];
  $E3B_step[2]  = $E_step[2];
#
  &reportout('Doing 3-band hardware-synchronized on-the-fly scan...');
  $ENSCAN_FLAGS->{'skip mtr link'}   = 1;
  $ENSCAN_FLAGS->{'skip mtr unlink'} = 1;
  for ($band=0; $band<3; $band++) {
     $E_npts = int(abs(($E3B_end[$band]-$E3B_start[$band])/$E3B_step[$band])+1.5);
     $scantag = sprintf 'Band-%d/3',$band+1;
     &do_1band_struck_fly_scan ($scantag, $E3B_start[$band], $E3B_end[$band], $E_npts);
  }								#end of loop over the bands

### Swap energy-reversed points and remove duplicates:
  $nE_accu = &removeDuplicates (\@En_array, \@Dt_array, \@I0_array, $E_step[0]);
  return;
}

#==========================================================================

# &do_1band_struck_fly_scan($scantag, $E_start, $E_end, $E_npts)
sub do_1band_struck_fly_scan ($$$$) {
#
  my ($scantag, $E_start_eV, $E_end_eV, $E_npts, @struckData);
  my ($wait_, $wait_step, $acquiring, $inpos, $runprg, $positioned);
  my ($updateRate, $delayTime, $cts2egu, $struckPs, $struckNg, $dwell);
  my ($axisTheta, $beginThetaAct, $npts, $pt, $pb, $i, $msg);
  $scantag    = shift(@_);
  $E_start_eV = shift(@_);
  $E_end_eV   = shift(@_);
  $E_npts     = shift(@_);

  $msg = sprintf 'Setting up ['.$scantag.'] fly scan with '.$E_npts
                .' points in range [%.2f-%.2f]eV...', $E_start_eV, $E_end_eV;
  &reportout($msg);
  ($status, $msg, $assembly,
   $axisEnergy, $axisTheta, $beginThetaAct,
   $ESinTheta, $cts2egu, $mtrOut,
   $struckPs, $struckNg, $dwell) =
                       &ENflyScan_setup($E_start_eV/1000.,$E_end_eV/1000.,$E_npts,
                                        $cnt_time,undef,undef,undef,$scaler);
  if ($status) {
     $msg =~ s/^\*\*\* ${script}: //;
     if ($tell_bluice) {&tellBluIce($scr.': '.$msg,'error');}
     &zenity_error_exit ('*** '.$script.': '.$msg);
  }
  if ($nE_accu == 0) {						#the very first scan
     &setStruckReadRate ($struck,'Passive');
     if ($medm) {
        for ($i=0; $i<$ninputs; $i++) {&openStruckMedm($struck,$struck_ext,$inputs[$i]);}
     }
  }
# ($status,$msg) = &ENflyScan_wait ($assembly,$axisEnergy,$mtrOut,$dwell,
#                                   $E_start_eV/1000.,$E_end_eV/1000.,$E_npts);		#waiting scan to finish
  $wait_      = 0;
  $wait_step  = 0.01;
  $acquiring  = 1;
  $inpos      = 0;
  $runprg     = 1;
  $positioned = 0;
  $delayTime  = $dwell;
  $updateRate = 4.*$dwell/$wait_step;
  if ($delayTime < 0.1) {$delayTime = 0.1;}			#delay motors start in order to catch all counts
  $i = 0;
  while (1) {
     select(undef,undef,undef,$wait_step);			#sleep($wait_step);
     $wait_ += $wait_step;
     ($status,$acquiring) = Pezca::GetDouble($struck.':Acquiring');
     ($status,$inpos)     = Pezca::GetDouble($assembly.':InPos');
     ($status,$runprg)    = Pezca::GetDouble($assembly.':RunPrg');
     if ($inpos == 1) {$positioned++;}
     else             {$positioned=0;}
     $i++;
     if (($acquiring!=1 || ($runprg==0 && $positioned>=3)) && $wait_> 5*$delayTime) {
        $msg = sprintf 'Scan ['.$scantag.'] finished! '
                      .'(acquiring='.$acquiring
                      .' mtrRunPrg='.$runprg
                      .' mtrInPos='.$positioned
                      .' time=%.1fs)', $wait_;
        &reportout($msg);
        &closeShutter ($shutter,$shutter_readback);
        &abortAssy ($assembly);
        &abortStruck ($struck);
        &restoreAxis($assembly, $axisEnergy);			#restore speed and motion prg
        last;
     }
     elsif ($tell_bluice && $i > $updateRate) {
        $i = 0;
        ($status,$npts) = &ENflyScan_read ($axisTheta,$beginThetaAct,$ESinTheta,
                                           $cts2egu,$struckPs,$struckNg,
                                           \@inputs,\@Offsets,\@struckData,'flywithread');
        &reportout('Scan ['.$scantag.'] points done '.$npts.'/'.$E_npts);
        if (($nE_accu + $npts) > 999) {$npts = 999 - $nE_accu;}
        for ($pt=0; $pt<$npts; $pt++) {
           $pb = $nE_accu + $pt;
           $En_array[$pb] = $struckData[0][$pt] * 1000.;
#          $Ti_array[$pb] = $struckData[1][$pt];
           if ($i0_name !~ /^none$/i) {
              $I0_array[$pb] = $struckData[2][$pt];		#here: @inputs=($i0_channel,$Flu_channel)
              $Dt_array[$pb] = $struckData[3][$pt];		#here: @inputs=($i0_channel,$Flu_channel)
           } else {
              $I0_array[$pb] = 1.;				#here: @inputs=($Flu_channel)
              $Dt_array[$pb] = $struckData[2][$pt];		#here: @inputs=($Flu_channel)
              $struckData[3][$pt] = 0;
           }
           if ($fakeflag) {
              &loadFakeData($pb,\$Dt_array[$pb],\$I0_array[$pb],\@fakearray,$Off);
           }
           if ($I0_array[$pb] <= 0.) {$I0_array[$pb] = 1.;}
           $Dt_array[$pb] = $Dt_array[$pb] / $I0_array[$pb];
        }
        &postBluIceArray ($bipostPV, \@En_array, \@Dt_array);
     }
  }								#end of scan waiting
### Read final data:
  ($status,$npts) = &ENflyScan_read ($axisTheta,$beginThetaAct,$ESinTheta,
                                     $cts2egu,$struckPs,$struckNg,
                                     \@inputs,\@Offsets,\@struckData,'final');

  if ($tell_bluice && ($nE_accu + $npts) > 999) {
     $pb = $nE_accu + $npts;
     $msg = 'Number of actual pts='.$pb.' exceeds EPICS array. Truncating to 999';
     if ($tell_bluice) {&tellBluIce($scr.': '.$msg,'warn');}
     &zenity_error ('!!! '.$script.': '.$msg);
     $npts = 999 - $nE_accu;
     if ($npts < 0) {$npts = 0;}
  }
  for ($pt=0; $pt<$npts; $pt++) {
     $pb = $nE_accu + $pt;
     $En_array[$pb] = $struckData[0][$pt] * 1000.;
#    $Ti_array[$pb] = $struckData[1][$pt];
     if ($i0_name !~ /^none$/i) {
        $I0_array[$pb] = $struckData[2][$pt];			#here: @inputs=($i0_channel,$Flu_channel)
        $Dt_array[$pb] = $struckData[3][$pt];			#here: @inputs=($i0_channel,$Flu_channel)
     } else {
        $I0_array[$pb] = 1.;					#here: @inputs=($Flu_channel)
        $Dt_array[$pb] = $struckData[2][$pt];			#here: @inputs=($Flu_channel)
        $struckData[3][$pt] = 0;
     }
     if ($fakeflag) {
        &loadFakeData($pb,\$Dt_array[$pb],\$I0_array[$pb],\@fakearray,$Off);
     }
     if ($I0_array[$pb] <= 0.) {$I0_array[$pb] = 1.;}
     $Dt_array[$pb] = $Dt_array[$pb] / $I0_array[$pb];
     if ($writelog) {
        if ($pb == 0) {
           print LOG '#  N/Total          E             I0           MCA/I0'
                    .'   NStruck       Struck[0]      Struck[2]       Struck[3]'
                    ."\n";
        }
        printf LOG '%5d/%-5d    %10.3f   %12.5g   %12.5g   '
                  .'%5d         %10.3f   %12.5g   %12.5g'
                  ."\n",
                   $pb+1, $nE_accu+$npts, $En_array[$pb],         $I0_array[$pb],      $Dt_array[$pb],
                   $pt,                $struckData[0][$pt]*1000., $struckData[2][$pt], $struckData[3][$pt]
                   ;
     }
  }
  if ($writelog) {print LOG '#'."\n";}
  $nE_accu += $npts;
# if ($tell_bluice) {&postBluIceArray ($bipostPV, \@En_array, \@Dt_array, 'final');}
  return;
}

#==========================================================================

# $E_chemshift = &calcChemShift(\@En_array, \@Dt_array, $EdgeEnergy, $roughStep);
sub calcChemShift (\@\@$$) {

### Find the chemical shift with Marquardt-Levenberg
### non-linear least squares fitting:

  our ($Rough_Data_Inflection_Width, $Rough_Data_Background);
  my ($En_ref, $Dt_ref, $edge, $step, $weight, $chemshift);
  my ($ymin, $ymax, $Rough_Data_Amplitude, $msg);
  my ($fitedge, $fitAmplitude);
  my ($xpdl, $ypdl, $inipar, $yfit, $parfit, $covarianceMatrix, $iterations);

  $En_ref = shift(@_);
  $Dt_ref = shift(@_);
  $edge   = shift(@_);
  $step   = shift(@_);

  $chemshift = 0.;
  $xpdl = pdl ($En_ref);
  $ypdl = pdl ($Dt_ref);

  $ymin = min $ypdl;
  $ymax = max $ypdl;
  ($Rough_Data_Amplitude,$Rough_Data_Background) = ($ymax-$ymin,$ymin);
  $Rough_Data_Inflection_Width = $step;				#this is typically 1eV
  $weight = 1;							#weights of different points in fitting
  $inipar = pdl [$edge,$Rough_Data_Amplitude];
  ($yfit, $parfit, $covarianceMatrix, $iterations) =
                        lmfit ($xpdl, $ypdl, $weight, \&madfit_func_atan_amplitude,
                        $inipar, {Maxiter => 300, Eps => 1e-4});
  if (defined $parfit) {
     ($fitedge, $fitAmplitude) = list $parfit;
     $chemshift = $fitedge - $edge;
     $msg = sprintf 'LMFIT absorption edge=%.2feV', $fitedge;
     &reportout($msg);
     $msg = sprintf 'LMFIT chemical  shift=%.2feV', $chemshift;
     &reportout($msg);
  } else {
     &reportout('LMFIT failed to find absorption edge','error');
     &reportout('LMFIT number of iterations='.$iterations,'error');
  }
  return $chemshift;
}

#==========================================================================

sub madfit_func_atan_amplitude ($$$$) {
  our ($Rough_Data_Inflection_Width, $Rough_Data_Background);
  our ($Amplitude,$Background,$iniwidth);
  my ($PI, $xarg, $par, $ym, $dyda, $edge, $height);
  my ($width, $bkg, $arg, @dy);
  $PI = 3.14159265358979;
  ($width,$bkg) = ($Rough_Data_Inflection_Width,$Rough_Data_Background);
  ($xarg,$par,$ym,$dyda) = @_;
  ($edge,$height) = map {$par->slice("($_)")} (0..1);
  (@dy)           = map {$dyda->slice(",($_)")} (0..1);
  $arg = 2.*($xarg-$edge)/$width;
  $ym .= $height*(0.5+atan2($arg,1.0)/$PI) + $bkg;
  $dy[0] .= (-2./$width)*($height/$PI)/(1.+$arg*$arg);
  $dy[1] .= 0.5+atan2($arg,1.0)/$PI;
  return;
}

#==========================================================================

sub restoreFeedback ($) {

  my $waitflag = shift(@_);
  if (!defined $waitflag) {$waitflag = 'wait';}

### Kill mono pitch:
# $status = Pezca::PutStringOldCa($beamline.':MO:mp2:StrCmd', 'K');
# &report_only($script, $status, $beamline.':MO:mp2:StrCmd');
  if ($pauseFeedback && defined $mostab_initial) {
     &reportout('Restoring I0 feedback');
     $status = &restoreMostab ($beamline);
     if ($status) {
        &reportout('Failed to restore I0 feedback','error');
     } else {
        if ($waitflag =~ /^wait$/i) {
           &reportout('Waiting for I0 lock...');
           $status = &waitMostabLock ($beamline);
           if ($status) {
              &reportout('Failed to get I0 lock','error');
           } else {
              &reportout('Got MOSTAB confirmation that I0 is locked');
           }
        } else {
           &reportout('Not waiting for I0 to lock','warn');
        }
     }
  }
  return;
}

#==========================================================================

#sub reportout($msg,[$msgtype,$msgbluice]) {
sub reportout($;$$) {
  my ($msg,$msgtype,$msgbluice);
  $msg       = shift(@_);
  $msgtype   = shift(@_);			# $msgtype=note,warn,error,popup_note,popup_error,direct
  $msgbluice = shift(@_);			# optional different message to BluIce

  if (!defined $msgtype)   {$msgtype   = 'note';}
  if (!defined $msgbluice) {$msgbluice = $msg;}

  $msg       = $script.': '.$msg;
  $msgbluice = $scr.': '.$msg;
  $msgbluice =~ s/\n*$//;			# trim end-of-lines

  if    ($msgtype =~ /warn/) {$msg = '!!! '.$msg;}
  elsif ($msgtype =~ /err/)  {$msg = '*** '.$msg;}

  print STDOUT $msg."\n";
  if ($tell_bluice && length($msgbluice)>0) {&tellBluIce($msgbluice,$msgtype);}
  return;
}
