#!/usr/bin/perl -W
# _CALLED_FROM_BLUICE_
#--------------------------------------------------------------
# BluIce call: fluoRaster.pl 23i mcaTime=0.5 file=/home/gmca/bluice_temp/fluoRaster.log offsetH=-0.1:0.1:20 offsetV=-0.095:0.095:20 gnuplot=n gnulog=n medm=n bluice mcaROI=auto
# BluIce track: /var/tmp/bluice-gmca> tail -f gmca-2009-03-29-12-16-23-6674.log | grep fluoRaster.pl
#--------------------------------------------------------------
# This script implements 2D fluorescence rastering of sample crystal in focused
# X-ray beam. The aim is to locate a small crystal(s) by mapping florescence
# yield depending on 2D position. Depending on command-line options, the script
# uses different scanning algorithms including step scanning over 2D grid, step
# scanning over specified set of polygons (2D grid with excluded spots), fly
# scanning with EPICS scan record or hardware-synchronized fly scanning. The
# scans are always done in horizontal direction by moving gonio support, while
# in the vertical direction the sample is shifted step-wise with XY positioner
# (both X and Y motors are involved to keep the sample in the focal plane). Any
# horizontal scanning option can be combined with shuttle scanning when the
# sample is alternatively moved left to right on one pass and right to left on
# the next one. The details of this program are described in Stepanov et.al.
# J. Appl. Crystallogr., v.44, No 4, p.772-778 (2011).
#
#       		Author: Sergey Stepanov
#
# Version-1.0:  2007/07/13 using Pezca
# Version-2.0:  2008/09/18 added arbitrary omega, passing many parameters via
#                          cmdline, and sending messages to BluIce log
# Version-2.1:  2008/09/18 replaced cmdline input offset=min:max:step -> offset=min:max:npts
# Version-3.0:  2009/02/06 added option to auto-set MCA for given edge:
#                          mcaROI=ready|BluIce|auto|Element-Edge|EeVmin:EeVmax|rough
#                          added option to do Hardware-synchronized fly scans
#                          in addition to Scanrec-syncronized fly scans
#                          or automatically choose between the two depending
#                          on the counting time per point.
# Version-3.1:  2009/06/15 added Struck cleanup and restoring Struck dwell
# Version-3.2:  2009/06/17 added backup/restore of previous mtrOut link
# Version-3.3:  2009/07/29 removed Joerger scaler from the stepMCA mode (new PV required!)
# Version-3.4:  2009/10/14 added 3D plot to png/ps/pdf
# Version-3.5:  2009/10/21 changed the way limits are treated: now the whole grid borders
#                          and the number of cells is expected
# Version-4.0:  2011/01/07 renamed from findFluorescenceFast.pl to fluoRaster.pl,
#                          added global flag for opening popups via BluIce,
#                          renamed cmdline argument 'tellBluIce=y|n' to 'bluice|nobluice'
# Version-4.1:  2011/01/14 added polygon and shuttle scanning functionalities
# Version-4.2:  2011/01/26 added automatic switch from polygon mode to fly/step if
#                          the polygon is a rectangle; added skipping lines in the
#                          polygon mode if there are no scan points in the line.
# Version-4.3:  2011/02/05 added PDL-based interpolation of on-the-fly scans data
#                          to the centers of grid cells.
# Version-4.4:  2012/09/25 added check for 64-bit OS at bitwise operations (polygon scan)
# Version-5.0:  2013/05/03 added crystal-XYZ or gonio-SH switch
# Version-5.1:  2014/03/28 changed functions calls to arrays pointers (preparation to Perl-5.16)
# Version-5.2:  2014/06/18 added stopping Struck as Joerger
# Version-5.3:  2015/08/03 added shrinking too wide fluorescence range as needed
# Version-5.4:  2019/11/06 added bringing PinDiode IN using get/set sampleEnv
# Version-5.5:  2020/02/21 added global switch for using Falcon instead of Canberra
#--------------------------------------------------------------
  use strict;
  use Time::HiRes qw(gettimeofday tv_interval); #needed for measuring total time
  use DBI;					#mySQL interface (for scan and MCA data)
# use Math::Trig;				#needed for omega sin/cos calculations (conflicts with PDL: both provide trigonometric functions)
  use PDL;					#interpolation of on-the-fly scan positions
  use PDL::GSL::INTERP; 			#interpolation of on-the-fly scan positions
  use Config;					#determine 32bit vs 64bit OS fir bit operations
  use Pezca;

  our ($USE_FALCON);				#global from beamline.pl
  our ($NONINTERACTIVE, $USE_ZENITY_FOR_EXIT, $ZENITY_VIA_BLUICE);
  our ($IGNORE_ERRORS, $IGNORE_MOTION_ESTIMATE);
  our ($beamline, $shutter, $joergerName, $HAVE_XYZ);
  our ($scanrecName, $scanrecAbort, $struckName, $struck_ext);
  our ($SRSCAN_FLAGS, $HDSCAN_FLAGS);		#global flags for scans
  our (@x1SR,@y1SR);
  our (@x1HD,@y1HD);
  our $EZCALONG;
  my ($PEZCA, $file, $filepart, @offsetsXY, @offsetsSH, $struck, @struckData);
  my ($stepXY, $stepSH, $mcaTime, $Flu_channel, $motionTime);
  my ($tell_bluice, $gnuplot, $logscale, $phiOffset, @mcaROI, $nROI);
  my ($script, $scr, $cmdline, $syntax, $narg, $arg, $i, $j, @beamlinelist);
  my ($bln_specified, $fil_specified, $hhh_specified, $vvv_specified);
  my ($tim_specified, $roi_specified, $gnu_specified, $glo_specified);
  my ($mdm_specified, $blu_specified, $phi_specified, $scn_specified);
  my ($flu_specified, $shu_specified, $int_specified, $shuttle, $direction);
  my ($degree2rad, $scanrec, $scanAbort, $assyXY, @axesXY, $assySH, $axisSH);
  my ($axisOmega, $mca, $ptsXY, $ptsSH, $maxPts, $omega, $phi);
  my (@iniXY, $iniSH, @rangeXY, @rangeSH, $scanMode, $scanTime, $fullTime);
  my (@ct, $minute, $hour, $mday, $mon, $year, @xData, @yData, $value);
  my ($status, $actPts, $posXY, $posXY_print, $filehandle);
  my ($pt_min, $y_min, $pt_max, $y_max, $gnuopen, $gnudest, $absoluteFlu);
  my ($absoluteSH, $absoluteXY, $ROI, @ROIs, @destXY, @tmparr, $ntmp, $tmpval);
  my ($energy, $E, $roi_lo, $roi_hi, $adc, $tca, $mcaNMAX, $low, $high);
  my ($bipostPV, $biflag, @biarray, $wait_time, $wait_step, $wait_max);
  my ($mcaMode, $Edge, $Edge_element, $Edge_KLMN, $biEdgePV);
  my ($EdgeEnergy, @E_step, @E_pm, $FluoEnergy, $FluoPMwidth);
  my ($cts2egu, $formula, $mtrOut, $struckPs, $struckNg, $t0, $elapsed);
  my ($scalerNo, $scaler, $startScalersPV, @inputs, $settleTime);
  my ($dwell_ini, $linked, $pmac, $pcsPrev, $formulaPrev);
# my ($currentRunPV, $currentRun);
  my ($selectedCellsPV, @selectedCells34, $os);
  my (@selectedCells, $numCells, $num, $n, $range1, $range2);
  my (@positions, $npositions, $medm, $interpolate, $cell_min);
  my (@rangeSH_middles, $ptsSH_middles, $x_pdl, $y_pdl, $spl);
  my ($tooWide, $E1, $E2);
  my ($pinInitiallyOUT, %states);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  do $PEZCA.'/Common/joerger_subs_v2.pl';
  do $PEZCA.'/Common/motor_subs.pl';
  do $PEZCA.'/Common/shutter_subs.pl';
  do $PEZCA.'/Common/fluo_subs.pl';
  do $PEZCA.'/Common/struck_subs_v2.pl';
  do $PEZCA.'/Common/scanrec_subs.pl';
  do $PEZCA.'/Common/gnuplot_subs.pl';
  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  select STDOUT; $|=1;			#set unbuffered output
  select STDERR; $|=1;			#set unbuffered output

########## EDIT DEFAULTS ###########
  @offsetsXY      = (-0.15, 0.15);	#scan offsets from current position
  $ptsXY          = 6;
##$stepXY         = 0.05;
  @offsetsSH      = (-0.3, 0.3);	#scan offsets from current position
  $ptsSH          = 13;
##$stepSH         = 0.05;
  $mcaTime        = 0.5;
  if (defined $USE_FALCON && $USE_FALCON) {
     $Flu_channel = 4;			#Falcon ROI-1 to Struck input (HD & discrim. scans only)
  } else {
     $Flu_channel = 24;			#default Struck/Scaler input for fluorescence (HD & discrim. scans only)
  }
  $tell_bluice    = 1;			#1=inform BluIce 0=not
  $gnuplot        = 0;			#1=plot with gnuplot, 0=not, 2=image(png,ps,pdf)
  $logscale       = 0;			#for gnuplot
  $medm           = 1;			#1=show scanrec MEDM, 0=not
  $phiOffset      = 0.0;		#offset between phi and omega
  $mcaMode        = 'auto';		#ready|BluIce|E_eV|Edge|auto|rough; auto is either ready or BluIce depending on ROI set
  $scanMode       = 'auto';		#flyHD or flySR depending on $mcaTime vs 1.6772
  $scalerNo       = 1;			#scaler1=Struck, scaler2(3)=Joerger (step scans only)
  $settleTime     = 1;			#positioner settle time for step-scan modes (in seconds)
  $shuttle        = 0;			#shuttle-type scanning
  $interpolate    = 1;			#interpolation of on-the-fly scans
  $cell_min       = 18;			#below this size turn of shuttle rastering
###### END OF EDIT AREA ############

  if ($tell_bluice) {$NONINTERACTIVE = 1; $USE_ZENITY_FOR_EXIT=1;}
  else              {$NONINTERACTIVE = 0; $USE_ZENITY_FOR_EXIT=0;}

  $script = &makeScriptName();
  $script =~ s/\.pl$//i;
  $scr = $script;			#short name (for tellBluIce)
  @beamlinelist = ('23i:', '23o:', '23b:', '23d:');
#---------------------------------------------------------------------------
### Read cmdline parameters:
  $syntax = "\n\n".'Syntax: '.$script.'.pl [beamline] [file=xxx]'
          . ' [mcaROI=ready|BluIce|auto|Element-Edge|EeVmin:EeVmax|rough]'
          . ' [mcaTime=xx] [offsetH=min:max:npts] [offsetV=min:max:npts]'
#         . ' [mcaTime=xx] [offsetH=min:max:step] [offsetV=min:max:step]'
          . ' [flu=nn|struckFlu=nn] [gnuplot=y|n|png|ps|pdf] [gnulog=y|n]'
          . ' [medm=y|n] [bluice|nobluice] [phiOffset=xx] [shuttle=y|n]'
          . ' [scan=flyHD|flySR|stepMCA|stepDiscr|polygonMCA|polygonDiscr|auto]'
          . ' [interpolate=y|n]'
          . "\n";
  $narg = @ARGV;
  print STDOUT "\n".'Date: '.scalar(localtime)."\n";
  if ($narg > 0) {print STDOUT "\n\n".$script.': started with cmdline = '."@ARGV\n\n\n";}
  else           {print STDOUT "\n\n".$script.': started with no cmdline arguments.'."\n\n\n";}
  for ($i=0; $i<$narg; $i++ ) {
     $arg = lc $ARGV[$i];
     $arg =~ s/^-{0,2}//;					#trim leading dashes
#    if (grep {$arg =~ /^$_:?$/} @beamlinelist) {		#use this wheh beamlinelist does not contain ':'
     if (grep(/^$arg:?$/, @beamlinelist)) {
        if (defined $bln_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate beamline parameter in the cmdline.'.$syntax);
        }
        $bln_specified=1;
        $arg =~ s/:$//;						#trim ':'
        if ($arg ne $beamline) {
           print STDOUT "\n".'!!! '.$script.': Cross-beamline operation: controlling '.$arg.' from '.$beamline."\n\n";
           $beamline = $arg;
           do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
        }
     }
     elsif ($arg =~ /^scan=/i) {
        if (defined $scn_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate scan type argument in the cmdline.'.$syntax);
        }
        $scn_specified=1;
        $scanMode = $ARGV[$i];
        $scanMode =~ s/^scan=//i;
        if ($scanMode !~ /^flyHD$/i &&
            $scanMode !~ /^flySR$/i &&
            $scanMode !~ /^stepMCA$/i &&
            $scanMode !~ /^stepDiscr$/i &&
            $scanMode !~ /^contourMCA$/i &&
            $scanMode !~ /^contourDiscr$/i &&
            $scanMode !~ /^polygonMCA$/i &&
            $scanMode !~ /^polygonDiscr$/i &&
            $scanMode !~ /^auto$/i) {
           &zenity_error_exit ('*** '.$script.': incorrect argument '.$ARGV[$i].' in the cmdline.'.$syntax);
        }
        if    ($scanMode =~ /^contourMCA$/i)   {$scanMode = 'polygonMCA';}
        elsif ($scanMode =~ /^contourDiscr$/i) {$scanMode = 'polygonDiscr';}
     }
     elsif ($arg =~ /^file=/i) {
### File is needed in two cases: either if you want to save data or if you
### want Gnuplot to plot data. If file is not specified, then:
### (a) if no gnuplot option requested and tell_bluice=1, then no data is written
### (b) if either gnuplot=1 or tell_bluice=1, then default filename is taken
        if (defined $fil_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate filename argument '.$ARGV[$i].' in the cmdline.'.$syntax);
        }
        $fil_specified=1;
        $file = $ARGV[$i];
        $file =~ s/^file=//i;
     }
     elsif ($arg =~ /^mcaROI=/i) {
### mcaROI=ready|BluIce|auto|Element-Edge|EeVmin:EeVmax|rough
### ready: expect that ROI,gain,offset is preset by the caller (e.g. BluIce)
### BluIce: take Element-Edge data from JBluIce xxx:bi:SCAN:edge PV; then optimize MCA for respective mySQL data
### auto: if ROI is set, behave as 'ready' mode, otherwice use 'BluIce' mode
### Element-Edge: read element and edge; then optimize MCA for respective mySQL data
### EeVmin:EeVmax: read expected fluorescence peak E_eV_min and E_eV_max; then optimize MCA for this peak
### rough: set MCA to measure everything just below X-ray energy
### --- these are marked as $mcaMode=ready|BluIce|auto|Edge|E_eV|rough
        if (defined $roi_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate ROI argument '.$ARGV[$i].' in the cmdline.'.$syntax);
        }
        $roi_specified=1;
        $arg =~ s/^mcaROI=//i;
        if    ($arg =~ /^ready$/i)  {$mcaMode = 'ready';}
        elsif ($arg =~ /^BluIce$/i) {$mcaMode = 'BluIce';}
        elsif ($arg =~ /^auto$/i)   {$mcaMode = 'auto';}
        elsif ($arg =~ /^rough$/i)  {$mcaMode = 'rough';}
        elsif ($arg =~ /\-/) {
           $mcaMode = 'Edge';
           $Edge = $ARGV[$i];
           $Edge =~ s/^mcaROI=//i;
           @mcaROI = split(/\-/,$Edge);
           $nROI = @mcaROI;
           if ($nROI != 2) {
              &zenity_error_exit ('*** '.$script.': Incorrect ROI argument '.$ARGV[$i].' in the cmdline.'.$syntax);
           }
           ($Edge_element, $Edge_KLMN) = ($mcaROI[0],$mcaROI[1]);
           if (length($Edge_element) < 1 ||
               length($Edge_element) > 2 ||
               length($Edge_KLMN) < 1 ||
               length($Edge_KLMN) > 2) {
              &zenity_error_exit ('*** '.$script.': Incorrect ROI argument '.$ARGV[$i].' in the cmdline.'.$syntax);
           }
        }
        elsif ($arg =~ /\:/) {
           $mcaMode = 'E_eV';
           @mcaROI = split /\:/, $arg;
           $nROI = @mcaROI;
           if ($nROI != 2) {
              &zenity_error_exit ('*** '.$script.': Incorrect ROI argument '.$ARGV[$i].' in the cmdline.'.$syntax);
           }
           for ($j=0; $j<$nROI; $j++) {
              if (! &is_decimal($mcaROI[$j]) || $mcaROI[$j] <= 0) {
                 &zenity_error_exit ('*** '.$script.': Incorrect ROI='.$mcaROI[$j].' in the cmdline.'.$syntax);
              }
              if ($mcaROI[$j] < 35.)  {$mcaROI[$j] = 1000.*$mcaROI[$j];}	#convert to eV
           }
        }
        else {
           &zenity_error_exit ('*** '.$script.': Incorrect ROI argument '.$ARGV[$i].' in the cmdline.'.$syntax);
        }
     }
     elsif ($arg =~ /^mcaTime=/i) {
        if (defined $tim_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate mcaTime argument '.$ARGV[$i].' in the cmdline.'.$syntax);
        }
        $tim_specified=1;
        $mcaTime = $arg;
        $mcaTime =~ s/^mcaTime=//i;
        if (! &is_decimal($mcaTime) || $mcaTime < 0.5 || $mcaTime > 10.0) {
           &zenity_error_exit ('*** '.$script.': mcaTime='.$mcaTime.' must be a number in range [0.5:10].'.$syntax);
        }
     }
     elsif ($arg =~ /^flu=/i || $arg =~ /^struckFlu=/i) {
        if (!defined $flu_specified) {
           $flu_specified = 1;
           $Flu_channel = $arg;
           if    ($arg =~ /^struckFlu=/i) {$Flu_channel =~ s/^struckFlu=//i;}
           elsif ($arg =~ /^flu=/i)       {$Flu_channel =~ s/^flu=//i;}
           if (! &is_int($Flu_channel) || $Flu_channel < 4 || $Flu_channel > 32) {
              &zenity_error_exit ('*** '.$script.': Struck input='.$ARGV[$i].' for fluorescence must be an integer in range [4-32]'.$syntax);
           }
        } else {
           &zenity_error_exit ('*** '.$script.': duplicate cmdline parameter "flu" or "struckFlu": '.$ARGV[$i].$syntax);
        }
     }
     elsif ($arg =~ /^offsetH=/i) {
        if (defined $hhh_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate offsetH argument '.$ARGV[$i].' in the cmdline.'.$syntax);
        }
        $hhh_specified=1;
        $arg =~ s/^offsetH=//i;
        @tmparr = split /\:/, $arg;
        $ntmp = @tmparr;
        if ($ntmp != 3) {
           &zenity_error_exit ('*** '.$script.': Incorrect offsetH argument '.$ARGV[$i].' in the cmdline.'.$syntax);
        }
        for ($j=0; $j<$ntmp; $j++) {
           if (! &is_float($tmparr[$j])) {
              &zenity_error_exit ('*** '.$script.': Incorrect offsetH('.$j.')='.$tmparr[$j].' in the cmdline.'.$syntax);
           }
        }
        @offsetsSH = ($tmparr[0], $tmparr[1]);		#grid offsets from current position
##      $stepSH    = $tmparr[2];
        $ptsSH     = $tmparr[2];			#number of cells in the grid
        if (! &is_int($ptsSH) || $ptsSH < 1) {
           &zenity_error_exit ('*** '.$script.': nptsH='.$ptsSH.' in "'.$ARGV[$i].'" must be a positive integer.'.$syntax);
        }
     }
     elsif ($arg =~ /^offsetV=/i) {
        if (defined $vvv_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate offsetV argument '.$ARGV[$i].' in the cmdline.'.$syntax);
        }
        $vvv_specified=1;
        $arg =~ s/^offsetV=//i;
        @tmparr = split /\:/, $arg;
        $ntmp = @tmparr;
        if ($ntmp != 3) {
           &zenity_error_exit ('*** '.$script.': Incorrect offsetV argument '.$ARGV[$i].' in the cmdline.'.$syntax);
        }
        for ($j=0; $j<$ntmp; $j++) {
           if (! &is_float($tmparr[$j])) {
              &zenity_error_exit ('*** '.$script.': Incorrect offsetV('.$j.')='.$tmparr[$j].' in the cmdline.'.$syntax);
           }
        }
        @offsetsXY = ($tmparr[0], $tmparr[1]);		#grid offsets from current position
##      $stepXY    = $tmparr[2];
        $ptsXY     = $tmparr[2];			#number of cells in the grid
        if (! &is_int($ptsXY) || $ptsXY < 1) {
           &zenity_error_exit ('*** '.$script.': nptsXY='.$ptsXY.' in the cmdline "'.$ARGV[$i].'" must be positive integer.'.$syntax);
        }
     }
     elsif ($arg =~ /^gnuplot=/i) {
        if (defined $gnu_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate gnuplot argument '.$ARGV[$i].' in the cmdline.'.$syntax);
        }
        $gnu_specified=1;
        $arg =~ s/^gnuplot=//i;
        if    ($arg eq '1' || $arg =~ /^y$/i || $arg =~ /^yes$/i) {$gnuplot = 1; $gnudest = 'screen';}
        elsif ($arg eq '0' || $arg =~ /^n$/i || $arg =~ /^no$/i)  {$gnuplot = 0; $gnudest = 'none';}
        elsif ($arg =~ /^png$/i) {$gnuplot = 2; $gnudest = 'png';}
        elsif ($arg =~ /^ps$/i)  {$gnuplot = 2; $gnudest = 'ps';}
        elsif ($arg =~ /^pdf$/i) {$gnuplot = 2; $gnudest = 'pdf';}
        else {
           &zenity_error_exit ('*** '.$script.': Incorrect flag gnuplot='.$arg.' in the cmdline.'.$syntax);
        }
     }
     elsif ($arg =~ /^gnulog=/i) {
        if (defined $glo_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate gnulog argument '.$ARGV[$i].' in the cmdline.'.$syntax);
        }
        $glo_specified=1;
        $arg =~ s/^gnulog=//i;
        if    ($arg eq '1' || $arg =~ /^y$/i || $arg =~ /^yes$/i) {$logscale = 1;}
        elsif ($arg eq '0' || $arg =~ /^n$/i || $arg =~ /^no$/i)  {$logscale = 0;}
        else {
           &zenity_error_exit ('*** '.$script.': Incorrect flag gnulog='.$arg.' in the cmdline.'.$syntax);
        }
     }
     elsif ($arg =~ /^medm=/i) {
        if (defined $mdm_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate medm argument '.$ARGV[$i].' in the cmdline.'.$syntax);
        }
        $mdm_specified=1;
        $arg =~ s/^medm=//i;
        if    ($arg eq '1' || $arg =~ /^y$/i || $arg =~ /^yes$/i) {$medm = 1;}
        elsif ($arg eq '0' || $arg =~ /^n$/i || $arg =~ /^no$/i)  {$medm = 0;}
        else {
           &zenity_error_exit ('*** '.$script.': Incorrect flag medm='.$arg.' in the cmdline.'.$syntax);
        }
     }
     elsif ($arg =~ /^shuttle=/i) {
        if (defined $shu_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate shuttle argument '.$ARGV[$i].' in the cmdline.'.$syntax);
        }
        $shu_specified=1;
        $arg =~ s/^shuttle=//i;
        if    ($arg eq '1' || $arg =~ /^y$/i || $arg =~ /^yes$/i) {$shuttle = 1;}
        elsif ($arg eq '0' || $arg =~ /^n$/i || $arg =~ /^no$/i)  {$shuttle = 0;}
        else {
           &zenity_error_exit ('*** '.$script.': Incorrect flag shuttle='.$arg.' in the cmdline.'.$syntax);
        }
     }
     elsif ($arg =~ /^interpolate=/i) {
        if (defined $int_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate interpolate argument '.$ARGV[$i].' in the cmdline.'.$syntax);
        }
        $int_specified=1;
        $arg =~ s/^interpolate=//i;
        if    ($arg eq '1' || $arg =~ /^y$/i || $arg =~ /^yes$/i) {$interpolate = 1;}
        elsif ($arg eq '0' || $arg =~ /^n$/i || $arg =~ /^no$/i)  {$interpolate = 0;}
        else {
           &zenity_error_exit ('*** '.$script.': Incorrect flag interpolate='.$arg.' in the cmdline.'.$syntax);
        }
     }
     elsif ($arg =~ /^BluIce$/i || $arg =~ /^noBluIce$/i) {
        if (defined $blu_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate BluIce|noBluIce argument '.$ARGV[$i].' in the cmdline.'.$syntax);
        }
        $blu_specified=1;
        if    ($arg =~ /^BluIce$/i)   {$tell_bluice = 1;}
        elsif ($arg =~ /^noBluIce$/i) {$tell_bluice = 0;}
        if ($tell_bluice) {$NONINTERACTIVE=1; $USE_ZENITY_FOR_EXIT=1;}
        else              {$NONINTERACTIVE=0; $USE_ZENITY_FOR_EXIT=0;}
     }
     elsif ($arg =~ /^phiOffset=/i) {
        if (defined $phi_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate phiOffset argument '.$ARGV[$i].' in the cmdline.'.$syntax);
        }
        $phi_specified=1;
        $phiOffset = $arg;
        $phiOffset =~ s/^phiOffset=//i;
        if (! &is_float($phiOffset) || $phiOffset < -180. || $phiOffset > 180.) {
           &zenity_error_exit ('*** '.$script.': phiOffset='.$phiOffset.' must be a number in range [-180:180].'.$syntax);
        }
     }
     else {
        &zenity_error_exit ('*** '.$script.': unrecognized argument '.$ARGV[$i].' in the cmdline.'.$syntax);
     }
  }

  if ($tell_bluice) {$ZENITY_VIA_BLUICE = 1;}		#we set it here because $syntax is too long for JBluIce message

###########################################################################
  $IGNORE_MOTION_ESTIMATE = 1;

  $degree2rad = atan2(1.,1.) / 45.0;			#atan2(1)=pi/4

  if (!defined $HAVE_XYZ) {$HAVE_XYZ=0;}
  $scanrec   = $scanrecName;
  $scanAbort = $scanrecAbort;
  $struck    = $struckName;
  $scaler    = $joergerName.$scalerNo;
  if ($scanrec   !~ /^${beamline}/) {$scanrec   = $beamline.':'.$scanrec;}
  if ($scanAbort !~ /^${beamline}/) {$scanAbort = $beamline.':'.$scanAbort;}
  if ($struck    !~ /^${beamline}/) {$struck    = $beamline.':'.$struck;}
  if ($scaler    !~ /^${beamline}/) {$scaler    = $beamline.':'.$scaler;}

  $assyXY    = $beamline.':GO:Cr';
  @axesXY    = ($beamline.':GO:CX',$beamline.':GO:CY');

  if (!$HAVE_XYZ) {
     $assySH = $beamline.':GO:St';
     if ($beamline eq '23b') {$assySH .= 'B';}
     $axisSH = $beamline.':GO:SH';
  } else {
     $assySH = $assyXY;
     $axisSH = $beamline.':GO:CZ';
  }

  $axisOmega = $beamline.':GO:OM';			#for checking Omega (must be 0 or 90
# $assyOmega = $beamline.':GO:Om';

  if (defined $USE_FALCON && $USE_FALCON) {
     $mca    = $beamline.':falcon:';
  } else {          
     $mca    = $beamline.':aim_adc1';
     $adc    = $beamline.':aim_adc1';
     $tca    = $beamline.':aim_tca1';
  }
  $energy    = $beamline.':MO:E';
  $bipostPV  = $beamline.':bi:SCAN:fastscanData';      #$flag,$ptV,$nptV,$xV,$nptH,@xH,@fluo
  $biEdgePV  = $beamline.':bi:SCAN:edge';	       #where BluIce posts Element-Edge


### ----------------PATCH----------------------------
### This is temporary patch for testing step scans
### It should be removed after testing is complete
#  $scanMode = 'stepMCA';
## $scanMode = 'stepDiscr';
#  print STDOUT "\n\n\n\n".'!!! '.$script.': temporary patched to use scanMode='.$scanMode."\n\n\n\n";
#  if ($tell_bluice) {&tellBluIce($scr.': PATCHED to STEP MODE','warn');}
### ----------------PATCH----------------------------


###-------------------------------------------------------------------------
### Read polygon exclusions:
  if ($scanMode =~ /^polygon/i) {
     $os = $Config{longsize}*8;
#    $currentRunPV = $beamline.':bi:r_runs:currentRun';
#    ($status,$currentRun) = Pezca::GetDouble($currentRunPV);
#    &report_exit($script,$status,$currentRunPV);
#    $selectedCellsPV = $beamline.':bi:r_r'.$currentRun.':selectedCells';
     $selectedCellsPV = $beamline.':bi:r_runs:selectedCells';
     print STDOUT $script.': scanMode='.$scanMode.' -- getting list of polygon cells from '.$selectedCellsPV.'  (OS='.$os.' bit)'."\n";
     if ($os eq 32) {
        ($status,@selectedCells34) = Pezca::GetList($selectedCellsPV,'LONG',34);	#this is as many as it could be
     }
     elsif ($os eq 64) {
        ($status,@selectedCells34) = Pezca::GetList($selectedCellsPV,'DOUBLE',34);	#this is as many as it could be
     }
     else {
        &zenity_error_exit ('*** '.$script.': unexpected OS bitness = '.$os."\n");
     }
     &report_exit($script,$status,$selectedCellsPV);
     print STDOUT $script.': List of polygon cells: ('."@{selectedCells34})\n";
### ---------------------------------------------------
### The structure of this array of 34 elements (according to Mark):
### Number of cells: selectedCells34[0]
###
### cell 1 selected:  0 != (selectedCells34[1] & (1<<30))
### cell 2 selected:  0 != (selectedCells34[1] & (1<<29))
### ...
### cell 30 selected: 0 != (selectedCells34[1] & (1<<1 ))
### cell 31 selected: 0 != (selectedCells34[1] & (1<<0 ))
### cell 32 selected: 0 != (selectedCells34[2] & (1<<30))
### cell 33 selected: 0 != (selectedCells34[2] & (1<<29))
### ...
### ---------------------------------------------------
     #my $yyy = '/usr/local/epics/base/bin/linux-x86_64/caget '.$beamline.':bi:r_runs:selectedCells';
     #my $xxx = `$yyy`;
     #print STDOUT "\n\n".'From caget:'."\n".$xxx."\n\n";
     $numCells = int($selectedCells34[0]+0.1);
     printf STDOUT $script.': numCells='.$numCells.' ptsSH='.$ptsSH.' ptsXY='.$ptsXY.'  grid=ptsSH*ptsXY=%d'."\n",$ptsSH * $ptsXY;
     if ($numCells != $ptsSH * $ptsXY) {
        &zenity_error_exit ('*** '.$script.': Number of polygon cells = '.$numCells."\n".
                            'Stored in '.$selectedCellsPV.' differs from the grid size:'."\n".
                            'X='.$ptsSH.' x Y='.$ptsXY);
     }
     $i = 0;			#cell index
     $n = 0;			#long number index
     do {
        $n++;
        if ($n>33 || !defined $selectedCells34[$n]) {
           &zenity_error_exit ('*** '.$script.': Out of data on selectedCells'."\n".
                            'array stored in '.$selectedCellsPV.':'."\n".
                            'number(long)='.$n.', cells='.$i.'/'.$numCells);
        }
        $num = int($selectedCells34[$n]+0.1);	#long number (we use bits 0-30; bit 31 is ignored)
        #if ($os eq 64) {$num = &tos($num);}	#restrict num to 32 bit (see &tos below in this file)
        #printf STDOUT $script.'[DEBUG]: n='.$n.'  selectedCells34[n]='.$num.' (hex=0x%x  bin=%0'.$os.'b)'."\n",$num,$num;
        for ($j=30; $j>=0; $j--) {
           if (($num & (1 << $j)) != 0) {$selectedCells[$i] = 1;}
           else                         {$selectedCells[$i] = 0;}
           #print STDOUT $script.'[DEBUG]: i='.$i.'  selectedCells[i]='.$selectedCells[$i]."\n";
           $i++;
        }
     } while ($i<$numCells);
     #print STDOUT "\n\n".$script.': selected cells=('."@{selectedCells})\n\n\n";
     #&signal_handler();	#temporary stop for debugging
     $n = 0;                    #number of excluded cells
     for ($i=0; $i<$numCells; $i++) {
        if ($selectedCells[$i] == 0) {$n++;}
     }
     if ($n == 0) {		#no exclusions - we work on rectangle
        print STDOUT $script.': no excluded cells in the '.$scanMode.' scan mode'."\n";
        if ($scanMode =~ /^polygonMCA$/i) {
           print STDOUT $script.': switching to stepMCA'."\n";
           $scanMode = 'stepMCA';
        }
        elsif ($scanMode =~ /^polygonDiscr$/i) {
           if ($mcaTime > 1.6772) {$scanMode = 'flySR';}
           else                   {$scanMode = 'flyHD';}
        }
        print STDOUT $script.': switching to '.$scanMode."\n";
     } else {
        print STDOUT $script.': scanning mode='.$scanMode.' -- '.$n.' of '.$numCells.' are marked for skipping'."\n";
     }
  }
###-------------------------------------------------------------------------


  $low = 5.;  						#default discriminations
  $high= 100.; 						#default discriminations
  if    ($scanMode =~ /^auto$/i) {
     if ($mcaTime > 1.6772) {$scanMode = 'flySR';}
     else                   {$scanMode = 'flyHD';}
  }
  elsif ($scanMode =~ /^flyHD$/i) {
     if ($mcaTime > 1.6772) {
        if ($tell_bluice) {&tellBluIce($scr.': cnt time lowered','warn');}
        print STDOUT "\n".'!!! '.$script.': Counting time per point decreased from '.$mcaTime.'s to 1.67s'."\n";
        $mcaTime = 1.67;
     }
  }

###-------------------------------------------------------------------------
### The scan intervals are in screen shifts, we move in the opposite direction:
  for ($i=0; $i<2; $i++) {
     $offsetsSH[$i] = -$offsetsSH[$i];
  }

### If needed, invert the scan intervals to the correct order, so that we
### work along backlash:
  if ($offsetsXY[1] < $offsetsXY[0]) {
    $tmpval = $offsetsXY[1];
    $offsetsXY[1] = $offsetsXY[0];
    $offsetsXY[0] = $tmpval;
  }
  if ($offsetsSH[1] < $offsetsSH[0]) {
    $tmpval = $offsetsSH[1];
    $offsetsSH[1] = $offsetsSH[0];
    $offsetsSH[0] = $tmpval;
  }

### Scan steps (VERTICAL):
  if ($ptsXY > 1) {
### In step mode we need to scan between the middles of
### the cells ($ptsXY is the number of sells, not the grid lines):
    $stepXY = ($offsetsXY[1] - $offsetsXY[0]) / $ptsXY;
    $offsetsXY[1] -= 0.5*$stepXY;
    $offsetsXY[0] += 0.5*$stepXY;
  }
  else {
     $stepXY = 0;
  }
  if ($ptsSH < 5 && $scanMode =~ /^fly/i) {
### We cannot do fly scans with few steps -- it is inefficient:
     if ($tell_bluice) {&tellBluIce($scr.': switched to stepscan: pts='.$ptsSH,'warn');}
     print STDOUT "\n".'!!! '.$script.': switched scanmode from '
                      .$scanMode.' to stepDiscr because npts='.$ptsSH.' < 5'."\n";
     $scanMode = 'stepDiscr';
  }

  if ($ptsSH > 1) {
     $stepSH = ($offsetsSH[1] - $offsetsSH[0]) / $ptsSH;
     if ($scanMode =~ /^fly/i) {                                          #           grid
### In fly mode we need to scan the whole grid because the scan reports   #                    
### data corresponding to the middles of the cells:                       #  flyscan
        $ptsSH++;
     } else {
### In step/polygon mode we need to scan between the middles of           #           grid
### the cells ($ptsSH is the number of cells, not the grid lines):        #                    
        $offsetsSH[1] -= 0.5*$stepSH;                                     #    stepscan
        $offsetsSH[0] += 0.5*$stepSH;
     }
  }
  else {
    $stepSH = 0;
    $offsetsSH[0] = 0.5*($offsetsSH[1] + $offsetsSH[0]);
    $offsetsSH[1] = $offsetsSH[0];
  }

#---------------------------------------------------------------------------
  if ($scanMode =~ /MCA$/i) {
### Use this sequence record to automate opening shutter followed by a
### short ~50ms delay and then followed by starting MCA:
     $startScalersPV = $beamline.':scanrec:startMCA.PROC';
     ($status,$value) = Pezca::GetDouble($startScalersPV);
     if ($status == 5) {
        print STDOUT "\n".'!!! '.$script.': no access to MCA start PV='.$startScalersPV."\n";
### Use this sequence record to automate opening shutter followed by a
### short ~50ms delay and then followed by starting MCA and Joerger scalers:
### (we do not need Joerger, but there is no sequence record for MCA only)
        $startScalersPV = $beamline.':scanrec:startScalers.PROC';
     }
  }
  elsif ($scanMode =~ /Discr$/i) {
### Use this sequence record to automate opening shutter followed by a
### short ~50ms delay and then followed by starting Joerger scaler only (no MCA):
     $startScalersPV = $beamline.':scanrec:startScaler.PROC';
  }

### Warning for step scans: need Tim Mooney/Louis Muir patch to PMAC DB:
  if ($scanMode !~ /^fly/i) {
     print STDOUT "\n".'!!! '.$script.': step scans require Tim Mooney patch to PMAC DB.'."\n";
     ($status,$value) = Pezca::GetDouble($startScalersPV);
     if ($status == 5) {
        &zenity_error_exit ('*** '.$script.': no access to scaler(s) start PV='.$startScalersPV."\n");
     }
  }

#---------------------------------------------------------------------------
  if ($scanMode =~ /^flyHD$/i) {
     $maxPts = &getStruckNmax($struck, $struck_ext);
     if ($ptsSH > $maxPts) {
        if ($tell_bluice) {&tellBluIce($scr.': npts > '.$maxPts,'error');}
        &zenity_error_exit ('*** '.$script.': Number of scan points='.$ptsSH.' exceeds Struck capacity='.$maxPts);
     }
  }
  else {
     $maxPts = &SRmaxPts($scanrec);
     if ($ptsSH > $maxPts) {
        if ($tell_bluice) {&tellBluIce($scr.': npts > '.$maxPts,'error');}
        &zenity_error_exit ('*** '.$script.': Number of scan points='.$ptsSH.' exceeds scanrec capacity='.$maxPts);
     }
  }

  if ($scanMode =~ /^flyHD$/i) {		#read virtual motor linking parameters
     ($cts2egu,$formula,$mtrOut,$struckPs,$struckNg) =
                                                   &searchMotorList($assySH,$axisSH);
     if (!defined $cts2egu) {
        if ($tell_bluice) {&tellBluIce($scr.': no mtr data in DB','error');}
        &zenity_error_exit ('*** '.$script.': No HD-synchronized data in motorList for assy='.$assySH.', axis='.$axisSH);
     }
### Get initial Struck dwell:
     $dwell_ini = &getStruckDwell ($struck);
### Check if mtrOut was previously linked and if so, backup linking data:
     $linked = &isMtrOutLinked($assySH,$mtrOut);
     if ($linked < 0) {
        $pmac = &getPmac($assySH);
        $pcsPrev = -$linked;
        $formulaPrev = &getLinkedFormula($pmac,$pcsPrev,$mtrOut);
     }
  }

###-------------------------------------------------------------------------
### Current positions:
  $omega     = &getAxisAct ($axisOmega);
  $phi       = $omega + $phiOffset;
  @iniXY     = (&getAxisAct($axesXY[0]),&getAxisAct($axesXY[1]));
  $iniSH     = &getAxisAct ($axisSH);

### Range of scanning:
  @rangeXY   = (       $offsetsXY[0],        $offsetsXY[1]);
  @rangeSH   = ($iniSH+$offsetsSH[0], $iniSH+$offsetsSH[1]);
  if ($scanMode =~ /^fly/i) {
     @rangeSH_middles = ($rangeSH[0]+0.5*$stepSH, $rangeSH[1]-0.5*$stepSH);
     $ptsSH_middles = $ptsSH - 1;
  }

  if ($scanMode =~ /^fly/i) {$scanTime  =  $mcaTime        * $ptsSH;}
  else                      {$scanTime  = ($mcaTime + 0.5) * $ptsSH;}	#add 0.5s on moving 1 step horizontally
  $fullTime  = ($scanTime + 0.5) * $ptsXY;				#add 0.5s on moving 1 step vertically

### mcaROI=ready|BluIce|auto|Element-Edge|EeVmin:EeVmax|rough
### ready: expect that ROI,gain,offset is preset by the caller (e.g. BluIce)
### BluIce: take Element-Edge data from JBluIce xxx:bi:SCAN:edge PV; then optimize MCA for respective mySQL data
### auto: if ROI is set, behave as 'ready' mode, otherwice use 'BluIce' mode
### Element-Edge: read element and edge; then optimize MCA for respective mySQL data
### EeVmin:EeVmax: read expected fluorescence peak E_eV_min and E_eV_max; then optimize MCA for this peak
### rough: set MCA to measure everything just below X-ray energy
### --- these are marked as $mcaMode=ready|BluIce|auto|Edge|E_eV|rough
  if ($mcaMode =~ /^auto$/i) {
     ($roi_lo, $roi_hi) = &checkMCA_ROI($mca);
     if ($roi_lo == -1 || $roi_hi == -1) {$mcaMode = 'BluIce';}	#ROI not set
     else                                {$mcaMode = 'ready';}	#ROI set
  }
  if ($mcaMode =~ /^ready$/i) {
     ($roi_lo, $roi_hi) = &checkMCA_ROI($mca);
     if ($roi_lo == -1 || $roi_hi == -1) {			#ROI not set
        if ($tell_bluice) {&tellBluIce($scr.': no ROI set','error');}
        &zenity_error_exit ('*** '.$script.': MCA ROI is not set.'."\n".
                            'Please make sure that ROI is setup properly by doing an excitation'."\n".
                            'scan in BluIce and then selecting a fluorescence peak with the bars.');
     }
     else {
        @mcaROI = ($roi_lo, $roi_hi);
     }
  }
  elsif ($mcaMode =~ /^BluIce$/i) {
     ($status,$Edge) =  Pezca::GetString($biEdgePV);
     if ($status) {
        &zenity_error_exit ('*** '.$script.': PVget failed on '.$biEdgePV);
     }
     @mcaROI = split(/-/,$Edge);
     $nROI = @mcaROI;
     if ($nROI != 2) {
        &zenity_error_exit ('*** '.$script.': Incorrect Element-Edge='.$Edge.' in PV='.$biEdgePV);
     }
     ($Edge_element, $Edge_KLMN) = ($mcaROI[0],$mcaROI[1]);
     if (length($Edge_element) < 1 ||
         length($Edge_element) > 2 ||
         length($Edge_KLMN) < 1 ||
         length($Edge_KLMN) > 2) {
        &zenity_error_exit ('*** '.$script.': Incorrect Element-Edge='.$Edge.' in PV='.$biEdgePV);
     }
  }
  elsif ($mcaMode =~ /^rough$/i) {
     $E = &getAxisAct ($energy);
     $E = 1000.*$E;						#convert to eV
     $mcaROI[0] = 50.;						#minimum MCA range
     $mcaROI[1] = $E - 100.;					#maximum MCA range
  }

  if ($mcaMode =~ /^BluIce$/i || $mcaMode =~ /^Edge$/i) {
     if ($tell_bluice) {&tellBluIce($scr.': using ROI from mySQL');}
     $status = &getEdgeDataSQL($Edge_element,$Edge_KLMN,\$EdgeEnergy,\@E_step,\@E_pm,\$FluoEnergy,\$FluoPMwidth);
     if ($status) {
        if ($tell_bluice) {&tellBluIce($scr.': no mySQL data','error');}
        &zenity_error_exit ('*** '.$script.': No mySQL data for element='.$Edge_element.', Edge='.$Edge_KLMN);
     }
     @mcaROI = ($FluoEnergy-$FluoPMwidth, $FluoEnergy+$FluoPMwidth);
  }

  if ($mcaMode !~ /^ready$/i) {					#for all MCA modes,except 'ready'
     $mcaNMAX = &MCAnmax($mca);
     printf STDOUT $script.': Optimizing MCA for fluorescence in range [%.1f : %.1f]eV'."\n",$mcaROI[0],$mcaROI[1];
     if ($tell_bluice) {&tellBluIce($scr.': optimizing MCA');}
     $tooWide = 1;						#adjust lower energy threshold if energy range is too wide for MCA
     if (defined $adc) {
        ($roi_lo,$roi_hi,$E1,$E2) = &optimizeMCA_ADC($mcaROI[0],$mcaROI[1],$mcaNMAX,$adc,$tooWide);
        if (defined $E1 && defined $E2) {
           $mcaROI[0] = $E1;
           $mcaROI[1] = $E2;
           printf STDOUT '!!! '.$script.': Fluorescence range was too wide for MCA. Adjusted to [%.1f : %.1f]eV'."\n",$mcaROI[0],$mcaROI[1];
           if ($tell_bluice) {&tellBluIce($scr.': fluorescence range was too wide for MCA -- reduced','warn');}
        }
     }
     printf STDOUT $script.': Setting MCA ROI: lo=%d, hi=%d'."\n",$roi_lo,$roi_hi;
     if ($tell_bluice) {&tellBluIce($scr.': setting MCA ROI');}
     &setMCA_ROI($mca,$roi_lo,$roi_hi);
  }

### For on-the-fly scans with Struck set ADC & TCA discrimination
### according to the ROI:
  if (defined $USE_FALCON && $USE_FALCON) {
     if ($scanMode =~ /^flyHD$/i || $scanMode =~ /Discr$/i) {
        &setFalconROI2SCA($mca);
     } else {
        $mcaNMAX = &MCAnmax($mca);
        &setFalconSCA($mca,0,$mcaNMAX-1);
     }
  } else {
     if ($scanMode =~ /^flyHD$/i || $scanMode =~ /Discr$/i) {
        &setDiscrFromROI($mca,$adc,$tca);
     } else {
        &setMCA_ADC_discr($adc,$low,$high);			#discr=[5-100]
        &setMCA_TCA($tca,$low,$high);				#discr=[5-100]
     }
  }

  if ($scanMode =~ /MCA$/i && $startScalersPV =~ /startScalers/) {
### This scaler is not used in the MCA mode, but the scan record will trigger
### it and wait for it when $startScalersPV =~ /startScalers/
     print STDOUT $script.': Setting up scaler ['.$scaler.']'."\n";
     &setJoerger($scaler,$mcaTime);
  }

  $gnuopen = 0;
  $absoluteFlu = -1;
  print STDOUT '-------------------------------------------------------------'."\n";
  print STDOUT $script.': This script will be doing 2D scan of '.$axisSH.' and '.$assyXY."\n"
             . 'while looking for fluorescence signal on ROI-0 of '.$mca."\n";
  print STDOUT '-------------------------------------------------------------'."\n";
  print  STDOUT $script.': 2D scan parameters:'."\n";
  printf STDOUT $script.': '.$scanMode.' scan of '.$axisSH.' from %.3f to %-.3f  step=%-.3f  npts='.$ptsSH."\n",@rangeSH,$stepSH;
  printf STDOUT $script.': Step-type scan of '.$assyXY.' from %.3f to %-.3f  step=%-.3f  npts='.$ptsXY."\n",@rangeXY,$stepXY;
  printf STDOUT $script.': MCA counting time = '.$mcaTime.'s,  full scan time='.$fullTime.'s (%.1fmin)'."\n",$fullTime/60.;
  printf STDOUT $script.': Initial positions: SH='.'%-4.3f  X=%-4.3f  Y=%-4.3f'."\n",$iniSH,@iniXY;
  if    ($scanMode =~ /^flySR$/i)        {print STDOUT $script.': Scanning algorithm: on-the-fly scan with MCA and scan record'."\n";}
  elsif ($scanMode =~ /^flyHD$/i)        {print STDOUT $script.': Scanning algorithm: on-the-fly scan with Struck and hardware synchronization'."\n";}
  elsif ($scanMode =~ /^stepMCA$/i)      {print STDOUT $script.': Scanning algorithm: step scan with MCA and scan record'."\n";}
  elsif ($scanMode =~ /^stepDiscr$/i)    {print STDOUT $script.': Scanning algorithm: step scan with amplifier discrimination and scan record'."\n";}
  elsif ($scanMode =~ /^polygonMCA$/i)   {print STDOUT $script.': Scanning algorithm: polygon scan with MCA and scan record'."\n";}
  elsif ($scanMode =~ /^polygonDiscr$/i) {print STDOUT $script.': Scanning algorithm: polygon scan with amplifier discrimination and scan record'."\n";}
  print  STDOUT $script.': MCA ROI setup mode='.$mcaMode."\n";
  print  STDOUT "\n";
  &ask_to_continue();

  $t0 = [gettimeofday];						#start measuring time

  $ROI    = 0;
  @ROIs   = ($ROI);
  @inputs = ($Flu_channel);

###--------------------------------------------------------------
  $SIG{'HUP'}  = 'signal_handler';			#(1) hangup (see signum.h)
  $SIG{'INT'}  = 'signal_handler';			#(2) interrupt (see signum.h)
  $SIG{'QUIT'} = 'signal_handler';			#(3) quit (see signum.h)
# $SIG{'KILL'} = 'signal_handler';			#(9) this is what is sent by kill -9 -- unblockable
  $SIG{'TERM'} = 'signal_handler';			#(15) terminate -- sent by kill without paramaters
# $SIG{'STOP'} = 'signal_handler';			#(19) stop -- unblockable

###---------------------------------------------------------------
### Determine Pindiode position:
  $pinInitiallyOUT = 1;                         #presume worse (pindiode is out and needs to be put IN)
  $status = &getSampleEnv(\%states);
  if (! $status && defined $states{'Pindiode'} && $states{'Pindiode'} =~ /^in$/i) {$pinInitiallyOUT = 0;}

  if ($pinInitiallyOUT) {
     print STDOUT "\n".$script.': Bringing Pindiode IN using sampleEnvSrv'."\n";
     $status = &setSampleEnv([$beamline, 'pindiodeIN', 'beamstopOUT', 'backlightOUT']);
     if ($status) {
        print STDOUT '*** '.$script.': failed to bring Pindiode IN.'."\n\n";
        if ($tell_bluice) {&tellBluIce($scr.': failed to bring Pindiode IN','error');}
#       &signal_handler();
     }
  }

###---------------------------------------------------------------
# if (!defined $file) {$file = '';}
  if (!defined $file || length($file) == 0) {
### File is needed in two cases: either if you want to save data or if you
### want Gnuplot to plot data. If file is not specified, then:
### (a) if no gnuplot option requested and tell_bluice=1, then no data is written
### (b) if either gnuplot=1 or tell_bluice=0, then default filename is taken
     if (! $tell_bluice || $gnuplot) {
        @ct  = localtime(time);				#($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)
        ($minute,$hour,$mday,$mon,$year) = ($ct[1],$ct[2],$ct[3],$ct[4]+1,$ct[5]+1900);
        $file = sprintf 'fluoraster_'.$year.'.%02d.%02d_%02d.%02d',$mon,$mday,$hour,$minute;
     }
  }
  if (defined $file) {
     $filehandle = open (DAT,'> '.$file.'.dat');	#open OUT file for writing
     if (!defined $filehandle) {
        print STDOUT '*** '.$script.': Could not open output file "'.$file.'.dat": '."$!\n\n";
        if ($tell_bluice) {&tellBluIce($scr.': outfile error','error');}
        &signal_handler();
     }
### write file header:
     printf DAT '# '.$scanMode.' fluorescence scan at omega = %.3f'."\n",$omega;
     printf DAT '# '.$scanMode.' scan of '.$axisSH.' from %.3f to %-.3f  step=%-.3f  npts='.$ptsSH."\n",@rangeSH,$stepSH;
     printf DAT '# Step-type  scan of '.$assyXY.' from %.3f to %-.3f  step=%-.3f  npts='.$ptsXY."\n",@rangeXY,$stepXY;
     print  DAT '# MCA counting time = '.$mcaTime.'s, MCA ROI = '."@{mcaROI}\n";
     print  DAT '# Vert(XY)   Horz(SH)   FluoROI'."\n";
  }

  if ($gnuplot == 1) {$gnuopen = &open_gnuplot ($rangeSH[0], $rangeSH[1], $logscale);}
  else               {$gnuopen = 0;}

  if ($tell_bluice) {
     @biarray = (0);
     $status = Pezca::PutList($bipostPV,'DOUBLE',@biarray);
     if ($status) {&zenity_error_exit ('*** '.$script.': PVput failed on '.$bipostPV);}
#    &addCaMonitor($bipostPV);			#this does not work!!! (SS)
  }

  if ($shuttle && abs($stepSH) < 0.001*$cell_min) {
     print STDOUT '!!! '.$script.': turning off shuttle rastering option because of small cell size <'.$cell_min.'um'."\n";
  }
  $HDSCAN_FLAGS->{'skip calc speed'}    = 1;    #this is on 2nd run
  $HDSCAN_FLAGS->{'skip mtr link'}      = 1;    #this is on 2nd run
  $HDSCAN_FLAGS->{'skip mtr unlink'}    = 1;
  $HDSCAN_FLAGS->{'skip move to start'} = 1;
  $HDSCAN_FLAGS->{'return to start'}    = 0;	#this is default
  $HDSCAN_FLAGS->{'skip set slew'}      = 0;	#this is default
  $HDSCAN_FLAGS->{'skip restore mprg'}  = 0;	#this is default
  $HDSCAN_FLAGS->{'skip set speed'}     = 0;	#this is default
  $HDSCAN_FLAGS->{'skip restore speed'} = 0;	#this is default

  if (! $shuttle) {				#no-shuttle mode
     if (!$HAVE_XYZ) {
        $HDSCAN_FLAGS->{'return to start'} = 1;	#after scan HDflyScan will send motor to starting position (no tracking) - fly scans
     } else {
### When we have XYZ, horizontal axis is returned together with XY
        $HDSCAN_FLAGS->{'return to start'} = 0;	#this is default
     }
  } else {					#shuttle mode
     if (!$HAVE_XYZ) {
### This is to speedup operations when SH is uncoupled from XY:
        $HDSCAN_FLAGS->{'skip set slew'}      = 1;
        $HDSCAN_FLAGS->{'skip restore mprg'}  = 1;
        $HDSCAN_FLAGS->{'skip set speed'}     = 1;
        $HDSCAN_FLAGS->{'skip restore speed'} = 1;
     } else {
### When we have XYZ, we have to restore speed & motion prg for fast & precise vertical shift
        $HDSCAN_FLAGS->{'return to start'} = 0;	#this is default
     }
  }
### For scan record settings are the same as for HD-scan:
  $SRSCAN_FLAGS = $HDSCAN_FLAGS;

  $direction  = 1;
  $motionTime = 60;				#for how long axisSH may be returing to start
  if ($tell_bluice) {$IGNORE_ERRORS = 1;}
############################ BEGIN 2D SCAN #################################################
XYPOS:                                                                                      #
  for ($i=0; $i<$ptsXY; $i++) {                                                             #
     $posXY = $rangeXY[0] + $stepXY * $i;                                                   #
     $posXY_print = sprintf '%.3f', $posXY;                                                 #
        										    #
     if ($scanMode !~ /^polygon/i) {                                                        #
        if ($direction > 0) {$range1=$rangeSH[0]; $range2=$rangeSH[1];}    		    #
        else                {$range1=$rangeSH[1]; $range2=$rangeSH[0];}    		    #
     } else {   									    #
        $npositions = &makePositions(\@positions,\@rangeSH,$ptsSH,                          #
                                     \@selectedCells,$direction);                           #
#       print STDOUT $script.': for H='.$posXY_print.' X-positions('.$npositions.') =';     #
#       for ($j=0; $j<$npositions; $j++) {printf STDOUT '  %.3f', $positions[$j];}          #
#       print STDOUT "\n";                                                                  #
        if ($npositions == 0) {next XYPOS;}     					    #
        $range1 = $positions[0]; $range2 = $positions[$npositions-1];                       #
     }          									    #
     if ($shuttle) {$direction = -$direction;}          #invert direction of next scanning  #
        										    #
     printf STDOUT $script.': %d/'.$ptsXY.'. Vert. Shifting '                               #
                          .$assyXY.' to '.$posXY_print."\n", $i+1;                          #
     @destXY = (($iniXY[0]+$posXY*sin($phi*$degree2rad)),                                   #
                ($iniXY[1]+$posXY*cos($phi*$degree2rad)));                                  #
     if (!$HAVE_XYZ) {                                                                      #
        if ($i == 0) {&startMotor ($axisSH, $range1, $assySH, 'checkAuto');}                #
        $status = &runNMotors ($assyXY,2,\@axesXY,\@destXY,'checkAuto');                    #
     } else {                                                                               #
        $status = &runNMotors ($assyXY,3,[@axesXY,$axisSH],[@destXY,$range1],'checkAuto');  #
     }                                                                                      #
     if ($status) {                                                                         #
        &zenity_error_exit ('*** '.$script.': moving assembly '.$assyXY.' failed');         #
     }                                                                                      #
     if (!$HAVE_XYZ && ($i == 0 || !$shuttle)) {                                            #
### When !$shuttle, motor is sent back to start by the scan without tracking                #
        $status = &trackMotor ($assySH, $motionTime, $axisSH);                              #
        if ($status) {                                                                      #
           &zenity_error_exit ('*** '.$script.': moving assembly '.$assySH.' failed');      #
        }                                                                                   #
     }                                                                                      #
        										    #
     print STDOUT $script.': Doing '.$scanMode.' scan with motor '.$axisSH."\n";            #
     if ($tell_bluice) {&tellBluIce($scr.': scan @H='.$posXY_print);}                       #
                                                                                            #
     if ($scanMode =~ /^flySR$/i) {                                                         #
### Do on-the-fly scan. It returns GLOBALS: @x1SR,@y1SR,...@y{$nROIs}SR:                    #
        ($status, $actPts) =                                                                #
          &SRflyScanMCA ($assySH,$axisSH,$range1,$range2,$ptsSH,                            #
                         $scanTime,$medm,$mca,\@ROIs);                                      #
     }                                                                                      #
     elsif ($scanMode =~ /^flyHD$/i) {                                                      #
### Do on-the-fly scan. It returns GLOBALS: @x1HD,@y1HD,...@y{$ninputs}HD:                  #
        ($status,$actPts) =                                                                 #
          &HDflyScan ($assySH,$axisSH,$range1,$range2,$ptsSH,                               #
                      $scanTime,$mtrOut,$cts2egu,$formula,$struckPs,$struckNg,              #
                      $medm,'',1,\@inputs,$scaler);                                         #
     }                                                                                      #
     elsif ($scanMode =~ /^stepMCA$/i) {                                                    #
### Do step scan with MCA. It returns GLOBALS: @x1SR,@y1SR,...@y{$nROIs}SR:                 #
        ($status, $actPts) =                                                                #
          &SRstepScanMCA ($assySH,$axisSH,$range1,$range2,$ptsSH,$mcaTime,                  #
                      $medm,$mca,$startScalersPV,\@ROIs,undef,$settleTime);                 #
     }                                                                                      #
     elsif ($scanMode =~ /^stepDiscr$/i) {                                                  #
### Do step scan with MCA. It returns GLOBALS: @x1SR,@y1SR,...@y{$nROIs}SR:                 #
        ($status, $actPts) =                                                                #
          &SRstepScan ($assySH,$axisSH,$range1,$range2,$ptsSH,$mcaTime,                     #
                      $medm,$scaler,$startScalersPV,'',\@inputs,undef,$settleTime);         #
     }                                                                                      #
     elsif ($scanMode =~ /^polygonMCA$/i) {                                                 #
### Do table scan with MCA. It returns GLOBALS: @x1SR,@y1SR,...@y{$nROIs}SR:                #
        ($status, $actPts) =                                                                #
          &SRtableScanMCA ($assySH,$axisSH,\@positions,$mcaTime,                            #
                      $medm,$mca,$startScalersPV,\@ROIs,undef,$settleTime);                 #
     }                                                                                      #
     elsif ($scanMode =~ /^polygonDiscr$/i) {                                               #
### Do table scan with MCA. It returns GLOBALS: @x1SR,@y1SR,...@y{$nROIs}SR:                #
        ($status, $actPts) =                                                                #
          &SRtableScan ($assySH,$axisSH,\@positions,$mcaTime,                               #
                      $medm,$scaler,$startScalersPV,'',\@inputs,undef,$settleTime);         #
     }                                                                                      #
     if ($status) {                                                                         #
        print STDOUT $script.': ...Scan failed with status='.$status."\n";                  #
        if ($tell_bluice) {                                                                 #
           &tellBluIce($scr.': scan failed','error');                                       #
        }                                                                                   #
        if ($status == 666) {                                                               #
           &zenity_error($script.': too short time per point -- must exceed 0.1s');         #
        }                                                                                   #
        elsif ($status == 777) {                                                            #
           &zenity_error($script.': speed exceeds maximum - increase time or npts');        #
        }                                                                                   #
        &signal_handler();                                                                  #
     } else {                                                                               #
        if ($scanMode =~ /^fly/i && $interpolate) {                                         #
           #Convert fly scan data arrays into piddles:                                      #
           if ($scanMode =~ /^flyHD$/i) {                                                   #
              &sort_by_x('incr',\@x1HD,\@y1HD);                                             #
              $actPts = &remove_duplicates(0.2,$stepSH,\@x1HD,\@y1HD);                      #
              $x_pdl = pdl(\@x1HD);                                                         #
              $y_pdl = pdl(\@y1HD);                                                         #
           } else {                                                                         #
              &sort_by_x('incr',\@x1SR,\@y1SR);                                             #
              $actPts = &remove_duplicates(0.2,$stepSH,\@x1SR,\@y1SR);                      #
              $x_pdl = pdl(\@x1SR);                                                         #
              $y_pdl = pdl(\@y1SR);                                                         #
           }                                                                                #
           #Create spline object (we use Akima-spline):                                     #
           $spl = PDL::GSL::INTERP->init('akima',$x_pdl,$y_pdl);                            #
           #Determine interpolated values for the middles of the grid:                      #
           for ($j=0; $j<$ptsSH_middles; $j++) {                                            #
              $xData[$j] = $rangeSH_middles[0] + $stepSH*$j;                                #
              $yData[$j] = $spl->eval($xData[$j],{Extrapolate=>1});                         #
           }                                                                                #
           $actPts = $ptsSH_middles;                                                        #
        } else {                                                                            #
           if ($scanMode =~ /^flyHD$/i) {@xData = @x1HD; @yData = @y1HD;}                   #
           else                         {@xData = @x1SR; @yData = @y1SR;}                   #
        }                                                                                   #
     }                                                                                      #
     if ($tell_bluice) {                                                                    #
        &tellBluIce($scr.': done @H='.$posXY_print);                                        #
     }                                                                                      #
     $HDSCAN_FLAGS->{'skip calc speed'} = 1;                                                #
     $SRSCAN_FLAGS->{'skip calc speed'} = 1;                                                #
     if (!$shuttle && !$HAVE_XYZ) {                                                         #
        $HDSCAN_FLAGS->{'return to start'} = 1;                                             #
        $SRSCAN_FLAGS->{'return to start'} = 1;                                             #
     }                                                                                      #
#-----------------------------------------------------------------                          #
     if ($gnuopen) {                                                                        #
### Open partial file for writing (for GnuPlot):                                            #
        $filepart = sprintf $file.'_%03d.dat', $i+1;                                        #
        $filehandle = open (OUT,'> '.$filepart);	#open OUT file for writing          #
        if (!defined $filehandle) {                                                         #
           print STDOUT '!!! '.$script.': Could not open partial output file ['             #
                                                        .$filepart.']'."\n\n";              #
           $gnuopen = 0;                                                                    #
           close_gnuplot();                                                                 #
        } else {                                                                            #
### write file header:                                                                      #
           printf OUT '# On-the-fly fluorescence scan at '                                  #
                      .$assyXY.'=%.3f'."\n", $posXY;                                        #
           printf OUT '# Continuous scan of %s from %.3f to %-.3f'                          #
                      .' step=%-.3f npts=%d'."\n",                                          #
                       $axisSH,@rangeSH,$stepSH,$ptsSH;                                     #
           print  OUT '# MCA counting time='.$mcaTime."s\n";                                #
        }                                                                                   #
     }                                                                                      #
        										    #
     if (defined $file) {                                                                   #
### Write data into a file(s):                                                              #
        for ($j=0; $j<$actPts; $j++) {                                                      #
           printf DAT ' %8.4f   %8.4f   %10.4g '."\n", $xData[$j], $posXY, $yData[$j];      #
           if ($gnuopen) {printf OUT ' %8.4f   %10.4g '."\n",  $xData[$j], $yData[$j];}     #
        }                                                                                   #
        print DAT "\n";					#write space line for GnuPlot-3D    #
        if ($gnuopen) {close(OUT);}                                                         #
     }                                                                                      #
        										    #
### Analyze intensity:                                                                      #
     ($pt_min,$y_min,$pt_max,$y_max) = &arrayStatsRef(\@yData);                             #
     if ($y_max > $absoluteFlu) {                                                           #
        $absoluteFlu = $y_max;                                                              #
        $absoluteXY  = $posXY;                                                              #
        $absoluteSH  = $xData[$pt_max];                                                     #
     }                                                                                      #
### Post data to BluIce:                                                                    #
     if ($tell_bluice) {                                                                    #
        if ($i > 0) {                                                                       #
           $wait_time = 0;                                                                  #
           $wait_step = 0.05;                                                               #
           $wait_max  = 5;                                                                  #
           do {                                         #wait for BluIce to clear prev.post #
              select(undef,undef,undef,$wait_step);	#sleep($wait_step);                 #
              $wait_time += $wait_step;                                                     #
              ($status,@biarray) = Pezca::GetList($bipostPV,'DOUBLE',2);                    #
              if (!defined $biarray[0]) {$biarray[0]=1;}                                    #
           } while ($biarray[0] == 1 && $wait_time < $wait_max);                            #
           if ($wait_time >= $wait_max) {                                                   #
              print STDOUT '!!! '.$script.': timeout waiting for data read ',               #
                                                'ready flag from JBluIce'."\n";             #
              &tellBluIce($scr.': post allow timeout','error');                             #
           }                                                                                #
        }                                                                                   #
        @biarray = (1);                                                                     #
        push (@biarray, $i+1);                                                              #
        push (@biarray, $ptsXY);                                                            #
        push (@biarray, $posXY);                                                            #
        push (@biarray, $actPts);                                                           #
        push (@biarray, @xData);                                                            #
        push (@biarray, @yData);                                                            #
#       print "@{biarray}\n";                                                               #
        $status = Pezca::PutList($bipostPV,'DOUBLE',@biarray);                              #
        if ($status) {                                                                      #
           &tellBluIce($scr.': array post failed','error');                                 #
        }                                                                                   #
     }                                                                                      #
### Plot the file:                                                                          #
     if ($gnuopen) {                                                                        #
        if ($i==1) {print  GNUPLOT 'plot   "'.$filepart.'" using 1:2 with lines'."\n";}     #
        else       {print  GNUPLOT 'replot "'.$filepart.'" using 1:2 with lines'."\n";}     #
     }                                                                                      #
  }                                                                                         #
############################ END OF 2D SCAN ################################################
  if (defined $file) {close(DAT);}

  if ((defined $absoluteXY) && (defined $absoluteSH)) {
     printf STDOUT $script.': Maximum fluorescence intensity in ROI=%-5.4g'."\n",$absoluteFlu;
     printf STDOUT $script.': Found at XY=%-4.3f  SH=%-5.3f'."\n",$absoluteXY,$absoluteSH;
  }

  if ($scanMode !~ /^flyHD$/i) {
     &abortSR($scanAbort);
  }
  print STDOUT $script.': ... Reverting motors to initial positions...'."\n";
  &abortAssy ($assySH);
  if ($scanMode =~ /^fly/i) {
     &restoreAxis ($assySH, $axisSH);			#also restore speed and motion prg.
     if ($scanMode =~ /^flyHD$/i) {
        print STDOUT $script.': ... Unlinking virtual motor mtrOut='.$mtrOut.'...'."\n";
        &unlinkMtrOut($assySH,$mtrOut);
        if (defined $formulaPrev) {
           print STDOUT $script.': ... Restoring previous link of mtrOut='.$mtrOut.' to PCS='.$pcsPrev."\n";
           &linkMtrOut_Pmac($pmac,$pcsPrev,$mtrOut,$formulaPrev,'sync');
        }
     }
  }
# &runMotor ($assySH,$axisSH,$iniSH,'checkAuto');
# &runNMotors ($assyXY,2,\@axesXY,\@iniXY,'checkAuto');
  if (!$HAVE_XYZ) {
     &startMotor ($axisSH,$iniSH,$assySH,'checkAuto');
     &startNMotors ($assyXY,2,\@axesXY,\@iniXY,'checkauto');
  } else {
     &startNMotors ($assyXY,3,[@axesXY,$axisSH],[@iniXY,$iniSH],'checkauto');
  }

  if ($scanMode =~ /^flyHD$/i || $scanMode =~ /Discr$/i) {
     if (defined $adc && defined $tca) {
        print STDOUT $script.': ... Removing ADC & TCA discriminations...'."\n";
        &setMCA_ADC_discr($adc,$low,$high);
        &setMCA_TCA ($tca,$low,$high);
     }
  }
  if ($scanMode =~ /^flyHD$/i) {
### Restore initial Struck dwell:
     print STDOUT $script.': ... Restoring Struck'."\n";
     $status = &setStruckDwell ($struck, $dwell_ini);
     &eraseStruck ($struck);
  }

  $elapsed = tv_interval ($t0);     				#measure time (in seconds)
  printf STDOUT $script.': ...Scan(s) finished! Total time spent=%.2fmin'."\n",$elapsed/60.;

  if ($gnuopen) {
     if ($NONINTERACTIVE) {&print_gnuplot($NONINTERACTIVE);}
     else                 {&print_gnuplot();}
  }

###-----------------------------------------------------------------
### Plot 3D map:
  if ($gnuplot) {
     print STDOUT $script.': printing 3D map: destination='.$gnudest."\n";
     $gnuopen = &open_gnuplot();
     if ($gnuopen) {
        if ($gnuplot == 2) {
           if ($gnudest =~ /^png$/i) {
              print GNUPLOT 'set terminal png small size 640,480'."\n";
              print GNUPLOT 'set output "'.$file.'.png"'."\n";
           }
           elsif ($gnudest =~ /^ps$/i || $gnudest =~ /^pdf$/i) {
              print GNUPLOT 'set terminal postscript portrait enhanced color blacktext'."\n";
              print GNUPLOT 'set output "'.$file.'.ps"'."\n";
           }
        }
        print  GNUPLOT 'set pm3d map'."\n";
        print  GNUPLOT 'set pm3d clip1in'."\n";
#       print  GNUPLOT 'set pm3d corners2color geomean'."\n";
#       print  GNUPLOT 'set pm3d flush begin'."\n";
#       print  GNUPLOT 'set pm3d flush end'."\n";
#       print  GNUPLOT 'set colorbox noborder'."\n";
#       print  GNUPLOT 'unset cblabel'."\n";
#       print  GNUPLOT 'unset cbtics'."\n";
        print  GNUPLOT 'unset key'."\n";
        print  GNUPLOT 'unset grid'."\n";
        print  GNUPLOT 'set xlabel "Horizontal pos. [um]"'."\n";
        print  GNUPLOT 'set ylabel "Vertical pos. [um]"'  ."\n";
#       printf GNUPLOT 'set xrange [%.3f : %.3f]'."\n",$rangeSH[0],$rangeSH[1];
#       printf GNUPLOT 'set yrange [%.3f : %.3f]'."\n",$rangeXY[0],$rangeXY[1];
        printf GNUPLOT 'set xrange [%.3f : %.3f]'."\n",$rangeSH[1],$rangeSH[0];
        printf GNUPLOT 'set yrange [%.3f : %.3f]'."\n",$rangeXY[1],$rangeXY[0];
        if ($logscale) {print GNUPLOT 'set logscale z'."\n";}
        print  GNUPLOT 'splot "'.$file.'.dat"'."\n";
        if    ($gnuplot == 1) {
           sleep(5);
           print  GNUPLOT 'set pm3d interpolate 16,16'."\n";
           print  GNUPLOT 'replot'."\n";
           &print_gnuplot($NONINTERACTIVE);
        }
        elsif ($gnuplot == 2) {
           print  GNUPLOT 'set pm3d interpolate 16,16'."\n";
           if    ($gnudest =~ /^png$/i) {print GNUPLOT 'set output "'.$file.'_x16.png"'."\n";}
           elsif ($gnudest =~ /^ps$/i ||
                  $gnudest =~ /^pdf$/i) {print GNUPLOT 'set output "'.$file.'_x16.ps"'."\n";}
           print  GNUPLOT 'replot'."\n";
           &close_gnuplot();
           if ($gnudest =~ /^pdf$/i) {
              $status = system ('ps2pdf', $file.'.ps', $file.'.pdf');
              if ($status) {print STDOUT '*** '.$script.': failed to convert PS to PDF'."\n";}
              else         {unlink $file.'.ps';}		#delete postscript file
              $status = system ('ps2pdf', $file.'_x16.ps', $file.'_x16.pdf');
              if ($status) {print STDOUT '*** '.$script.': failed to convert x16 PS to PDF'."\n";}
              else         {unlink $file.'_x16.ps';}		#delete postscript file
           }
        }
     }
     else  {
        print STDOUT '!!! '.$script.': failed to open pipe to Gnuplot'."\n";
     }
  }

###-----------------------------------------------------------------
  print STDOUT $script.': All done!'."\n";
  &clearAllCaMonitors();
  exit;

#=================================================================

### $npositions = &makePositions(\@positions,\@rangeSH,$ptsSH,\@selectedCells,$direction);
sub makePositions(\@\@$\@$) {
  my ($positionsRef, $rangeRef, $pts, $selectedRef, $direction, $npositions);
  my ($step, $i, @tmp);
  $positionsRef = shift(@_);
  $rangeRef     = shift(@_);
  $pts          = shift(@_);
  $selectedRef  = shift(@_);
  $direction    = shift(@_);
  if ($pts > 1) {$step = (@$rangeRef[1]-@$rangeRef[0])/($pts-1.);}
  else          {$step = 0.0;}
  @$positionsRef=();
  $npositions = 0;
  for ($i=0; $i<$pts; $i++) {
     if (@$selectedRef[$i]) {
       @$positionsRef[$npositions] = @$rangeRef[0] + $step*$i;
       $npositions++;
     }
  }
  ### Clear this row of selected points:
  splice @$selectedRef, 0, $pts;

  ### Invert the positions array if needed:
  if ($direction < 0) {
    for ($i=0; $i<$npositions; $i++) {$tmp[$npositions-1-$i] = @$positionsRef[$i];}
    for ($i=0; $i<$npositions; $i++) {@$positionsRef[$i]     = $tmp[$i];}
  }

  return $npositions;
}

#=================================================================

sub tos {
### Try to correct to 32 bit (4294967295=0xffffffff, 2147483648=0x80000000)
  my ($num) = @_;
# if ($num >  4294967295) {$num = $num - 4294967296;}
# if ($num < -2147483648) {$num = $num + 4294967296;}
  if ($num >  0xffffffff) {$num = $num - 0xffffffff;}
  if ($num < -0x80000000) {$num = $num + 0xffffffff;}
  if ($num >= 0) {
     if ($num >> 31) {$num = $num - 2**32;}
  }
  return $num;
}

#=================================================================

sub signal_handler {
  our ($ABORT_IN_PROCESS);
  my $sig = shift;

  if (defined $sig) {print STDOUT "\n".'*** '.$script.': Caught a SIG'.$sig."\n";}
  else              {print STDOUT "\n".'*** '.$script.': Exiting on failed scan'."\n";}

  if (defined $ABORT_IN_PROCESS) {return;}
  $ABORT_IN_PROCESS = 1;
  &abort_external_script();

  $IGNORE_ERRORS = 1;
  if ($shutter !~ /^${beamline}/) {$shutter = $beamline.':'.$shutter;}
  print STDOUT $script.': Closing shutter ['.$shutter.']'."\n";
  &closeShutter ($shutter);

  if (defined $assySH && defined $assyXY) {
     if ($scanMode =~ /^flyHD$/i) {
        if (defined $struck) {
           print STDOUT $script.': Aborting Struck ['.$struck.']'."\n";
           &abortStruck ($struck);
#          &setStruckReadRate ($struck,'2 second');
           &eraseStruck ($struck);
           if (defined $dwell_ini) {&setStruckDwell ($struck, $dwell_ini);}
        }
     }
     else {
        if (defined $scanrec) {
           print STDOUT $script.': Aborting scan record ['.$scanrec.']'."\n";
           &abortSR($scanAbort);
        }
     }
     print STDOUT $script.': Stopping/restoring assembly ['.$assyXY.']'."\n";
     &abortAssy ($assyXY);
     print STDOUT $script.': Stopping/restoring motor ['.$axisSH.']'."\n";
     &abortAssy ($assySH);
     if ($scanMode =~ /^fly/i) {
        &restoreAxis ($assySH, $axisSH);	#also restore speed and motion prg.
        if ($scanMode =~ /^flyHD$/i) {
           if (defined $mtrOut) {
              print STDOUT $script.': Unlinking virtual motor ['.$mtrOut.'] from assembly ['.$assySH.']'."\n";
              &unlinkMtrOut($assySH,$mtrOut);
              if (defined $formulaPrev && defined $pcsPrev) {
                 print STDOUT $script.': Restoring previous link of mtrOut='.$mtrOut.' to PCS='.$pcsPrev."\n";
                 &linkMtrOut_Pmac($pmac,$pcsPrev,$mtrOut,$formulaPrev,'sync');
              }
           }
        }
     }
     print STDOUT $script.': Reverting motors to initial positions'."\n";
#    &runMotor ($assySH,$axisSH,$iniSH,'checkAuto');
#    &runNMotors ($assyXY,2,\@axesXY,\@iniXY,'checkAuto');
     if (!$HAVE_XYZ) {
        if (defined $iniSH)    {&startMotor($axisSH,$iniSH,$assySH,'checkAuto');}
        if (defined $iniXY[1]) {&startNMotors($assyXY,2,\@axesXY,\@iniXY,'checkauto');}
     } else {
        if (defined $iniSH && defined $iniXY[1]) {
           &startNMotors($assyXY,3,[@axesXY,$axisSH],[@iniXY,$iniSH],'checkauto');
        }
     }

     if ($scanMode =~ /^flyHD$/i || $scanMode =~ /Discr$/i) {
        if (defined $adc && defined $tca && defined $low && defined $high) {
           print STDOUT $script.': Removing ADC and TCA discriminations'."\n";
           &setMCA_ADC_discr($adc,$low,$high);
           &setMCA_TCA ($tca,$low,$high);
        }
     }
     if ($gnuopen) {&close_gnuplot();}
  }
  if (defined $pinInitiallyOUT && $pinInitiallyOUT) {
     print STDOUT $script.': Moving Pin Diode OUT and Beamstop IN using sampleEnvSrv'."\n";
     $status = &setSampleEnv([$beamline, 'pindiodeOUT', 'beamstopIN']);
  }

  print STDOUT $script.': Clearing all CA monitors'."\n";
  &clearAllCaMonitors();
  print STDOUT $script.': Done with exiting!'."\n";
  exit 1;
}
