#!/usr/bin/perl -W
#--------------------------------------------------------------
#       This script checks if Dectris FURKA is running and
#              if yes, returns current FURKA user
#
#     		Author: Sergey Stepanov
#
# Version-1.0:  2014/01/25 initial implemensshcmdtation for Pilatus FURKA
# Version-2.0:  2016/05/31 added support for Eiger-style FURKA
# Version-3.0:  2017/12/07 added check if FURKA is enabled.
#                          if not, returns furka_user=furka_is_disabled
# Version-4.0:  2021/07/13 added getCCDfpehost (all functions are internal!)
# Version-5.0:  2022/08/16 added support for furka on Mint
# Version-5.1:  2022/09/23 added error catching and handling of the backticks
#--------------------------------------------------------------
  use strict;

  our ($beamline);
  my ($host, $furka, $mode, $cmd, $msg, $str, @arr, $narr, $status);
  my ($script, $bl, $uid, $gid, $partial_path, $fhost, $fenabled);
  my ($pscmd, $getent, $chkcfg, $sshcmd, $debug, $stderr, @bmlist);
  my ($strmwr, $streamwriter, $user);

  select STDOUT; $|=1;			# set unbuffered output
  select STDERR; $|=1;			# set unbuffered output
  $stderr = 1;				# output error to STDERR

# $debug = 1;
  if (!defined $debug) {$debug=0;}

  $script = $0;
  $script =~ s/\\/\//g;			#change \ to /
  $script =~ s/^.*\///g;		#remove everything before last /

  $pscmd  = '/bin/ps -eo pid,cmd'
          .' | /bin/grep -iv "grep\|edit\|term\|onsole\|bash\|csh\|rsync\|\<defunct\>"'
          .' | /bin/grep ';
  $getent = '/usr/bin/getent passwd ';
# $chkcfg = '/usr/sbin/chkconfig --list furka 2>/dev/null | /usr/bin/grep -c 5:on';	#returns 0/1
  $chkcfg = 'systemctl is-active furka 2>/dev/null';					##this does not produce output, but status
# $chkcfg = '/usr/sbin/chkconfig --level=5 furka 2>/dev/null';				#this does not produce output, but status
  $strmwr = 'systemctl is-active streamwriter 2>/dev/null';				#returns active/inactive/unknown/
  $sshcmd = '/usr/bin/ssh -aTxq -o BatchMode=yes -o StrictHostKeyChecking=no ';
  @bmlist = qw(23d 23i 23o 23b);

### Current OS user:
  $user = $ENV{USER};
  if (!defined $user) {$user = 'current';}

#---------sshcmd
### Determine host and beamline
  $host = $ENV{HOSTNAME};
  if (!defined $host && $^O !~ /Win/i) {$host=`hostname`; $host=~s/\n//g;}
  if (!defined $host) {
     $str = '*** '.$script.': HOSTNAME is undefined!';
     &zenity_error_exit ($str,$stderr);
  }
  $host = lc($host);
  $host =~ s/(-40g)*(\.gmca\.aps\.anl\.gov)*$//;
  if ($debug) {print STDERR $script.': host='.$host."\n";}
  if    (defined $ENV{BEAMLINE})                   {$bl = $ENV{BEAMLINE};}
  elsif ($host =~ /^bl1/)                          {$bl = 1;}
  elsif ($host =~ /^bl2/ || $host =~ /^directnet/) {$bl = 2;}
  elsif ($host =~ /^bl3/)                          {$bl = 3;}
  else  {
     $str = '*** '.$script.': Cannot determine beamline for host='.$host;
     &zenity_error_exit ($str,$stderr);
  }
  $beamline = $bmlist[$bl];
#---------

  if    ($bl == 1) {
     ($fhost,$status) = &getCCDfpehost();				# uses cainfo to get host (see below)
     if ($debug) {print STDERR $script.': FPEhost=['.$fhost.']'."\n";}
     if ($status || $fhost eq 'unknown') {$fhost = 'bl'.$bl.'ppu';}	# Pilatus PPU
     $furka  = '/usr/local/dectris/bin/furka_dectris';
     if ($fhost =~ /^bl1ws11/) {					# this is host for CdTe Eiger
        $furka = 'python2.7 /etc/init.d/furka start';
#       print STDOUT 'furka_user=furka_is_disabled'."\n";
        print STDOUT 'furka_user=streamwriter'."\n";			#we presume that on CdTe furka is disabled and streamwriter is running
        exit;
     }
  }
  elsif ($bl == 2) {
     ($fhost,$status) = &getCCDfpehost();				# uses cainfo to get host (see below)
     if ($status || $fhost eq 'unknown') {$fhost = 'bl'.$bl.'epu';}	# Eiger EPU
     $furka = 'python2.7 /etc/init.d/furka start';
  }
  else  {
     $str = '*** '.$script.': unsupported beamline='.$bl.' (host='.$host.')';
     &zenity_error_exit ($str,$stderr);
  }
  $fhost =~ s/(-40g)*(\.gmca\.aps\.anl\.gov)*$//;
  if ($debug) {print STDERR $script.': FPEhost=['.$fhost.']'."\n";}

  $pscmd .= '"'.$furka.'"';

### If the script is running not on $fhost, but on a remote host, we need to ssh:
  if ($host !~ /^${fhost}/) {
     $mode    = 'remote';
     $status = &checkSSHconnection($fhost);
#    $status = &checkSSHconnection('px0');				#uncomment for testing ssh error reporting
     $sshcmd .= $fhost.' ';
     $pscmd   = $sshcmd.$pscmd;
     $getent  = $sshcmd.$getent;
     $chkcfg  = $sshcmd.$chkcfg;
  } else {
     $mode = 'local';
  }

  $cmd = $chkcfg;
  if ($debug) {print STDERR $script.': Running cmd='.$cmd."\n";}
  $fenabled = `${cmd}`;
  if ($debug) {print STDERR $script.': Return code='.${^CHILD_ERROR_NATIVE}.' ('.$?.')'."\n";}
  if (${^CHILD_ERROR_NATIVE}) {
     $str = '*** '.$script.' command:'."\n".$cmd."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.'.';
     if ($mode eq'remote') {$str .= "\n".'Please verify passwordless ssh connection to '.$fhost.' for the '.$user.' user.';}
     &zenity_error_exit ($str,$stderr);
  }
  chomp $fenabled;
  if ($debug) {print STDERR $script.': Response: '.$fenabled."\n";}
  if ($fenabled =~ /^active/) {
     if ($debug) {print STDERR $script.': furka is enabled'."\n";}
  }
  elsif ($fenabled =~ /^inactive/) {
     if ($debug) {print STDERR $script.': furka is disabled'."\n";}
     if ($bl == 2) {
        if($mode eq 'remote') {$strmwr = $sshcmd.$strmwr;}
        if ($debug) {print STDERR $script.': Running cmd='.$strmwr."\n";}
        $streamwriter = `${strmwr}`;					#returns active/inactive/unknown/
        if ($debug) {print STDERR $script.': Return code='.${^CHILD_ERROR_NATIVE}.' ('.$?.')'."\n";}
        if (${^CHILD_ERROR_NATIVE}) {
           $str = '*** '.$script.' command:'."\n".$cmd."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.'.';
           if ($mode eq'remote') {$str .= "\n".'Please verify passwordless ssh connection to '.$fhost.' for the '.$user.' user.';}
           &zenity_error_exit ($str,$stderr);
        }
        chomp $streamwriter;
        if ($debug) {print STDERR $script.': Response: '.$streamwriter."\n";}
        if ($streamwriter =~ /^active/) {print STDOUT 'furka_user=streamwriter'."\n"; exit;}
     }
     print STDOUT 'furka_user=furka_is_disabled'."\n";
     exit;
  }
  else {
     print STDERR $script.': unexpected response to furka enabled/disabled query:'."\n"
                  .'CMD: '.$cmd."\n"
                  .'RSP: '.$fenabled."\n";
  }
  $cmd = $pscmd;
  if ($debug) {print STDERR $script.': Running cmd='.$cmd."\n";}
  $msg = `${cmd}`;
  if ($debug) {print STDERR $script.': Return code='.${^CHILD_ERROR_NATIVE}.' ('.$?.')'."\n";}
  if (${^CHILD_ERROR_NATIVE}) {
     $str = '*** '.$script.' command:'."\n".$cmd."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.'.';
     if ($mode eq'remote') {$str .= "\n".'Please verify passwordless ssh connection to '.$fhost.' for the '.$user.' user.';}
     &zenity_error_exit ($str,$stderr);
  }
  @arr = split(/\n/,$msg);
  $narr = @arr;
  if    ($narr == 0) {
     print STDERR '!!! '.$script.': FURKA is not running'."\n";
     print STDOUT 'furka_user=none'."\n";
     exit;
  }
  elsif ($narr >  1) {
     print STDOUT 'furka_user=multiple'."\n";
     $str = '!!! '.$script.': Multiple ('.$narr.') instances of FURKA found'."\n";
     &zenity_error_exit ($str,$stderr);
  }
  chomp $msg;								#remove line end
  $str = $msg;								#save $msg for record
  if ($fhost eq 'bl1ppu') {						# Pilatus PPU
     if ($msg =~ /^.*${furka} --user /) {
        $msg =~ s/^.*${furka} --user //;
        ($uid,$gid) = split(/:/,$msg);
        if (!defined $uid || $uid !~ /^\d*$/ ||
            !defined $gid || $gid !~ /^\d*$/) {
           $str = '!!! '.$script.': Unexpected FURKA string ['.$str.']'."\n"
                        .'--- in response to ['.$cmd.']'."\n"
                        .'--- when parsing ['.$msg.'] as uid:gid'."\n";
           if (length($msg) == 0)  {$str .= '-- no uid:gid in the string!'."\n";}
           if (!defined $uid) {
              $str .= '-- uid is undefined!'."\n";
           } else {
              $str .= 'uid=['.$uid.']'."\n";
              if ($uid !~ /^\d*$/) {$str .= '-- uid=['.$uid.'] is not numeric!'."\n";}
              else                 {$str .= 'uid=['.$uid.']'."\n";}
           }
           if (!defined $gid) {
              $str .= '-- gid is undefined!'."\n";
           } else {
              $str .= 'gid=['.$gid.']'."\n";
              if ($gid !~ /^\d*$/) {$str .= '-- gid=['.$gid.'] is not numeric!'."\n";}
              else                 {$str .= 'gid=['.$gid.']'."\n";}
           }
           &zenity_error_exit ($str,$stderr);
        } else {
           $cmd = $getent.$uid;
           if ($debug) {print STDERR $script.': Running cmd='.$cmd."\n";}
           $msg = `${cmd}`;						#Expecting: 'user0:x:600:600:user0:/mnt/share1/user0:/bin/bash'
           if ($debug) {print STDERR $script.': Return code='.${^CHILD_ERROR_NATIVE}.' ('.$?.')'."\n";}
           if (${^CHILD_ERROR_NATIVE}) {
              $str = '*** '.$script.' command:'."\n".$cmd."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.'.';
              if ($mode eq'remote') {$str .= "\n".'Please verify passwordless ssh connection to '.$fhost.' for the '.$user.' user.';}
              &zenity_error_exit ($str,$stderr);
           }
           @arr = split (/:/,$msg);
           $narr = @arr;
           if ($narr != 7) {print STDERR '!!! '.$script.': Unexpected getent data ['.$msg.'] for uid='.$uid."\n"; exit;}
           print STDOUT 'furka_user='.$arr[0]."\n";
           exit;
        }
     } else {
        $str = '!!! '.$script.': Unexpected FURKA process ['.$msg.']'."\n"
                        .'--- in response to ['.$cmd.']'."\n"
                        .'--- Expected=[pid '.$furka.' --user uid:gid]'."\n";
        if ($msg =~ /${furka}/) {$str .= 'It is possible that FURKA was started without user specification when restarting GRIMSEL'."\n";}
        &zenity_error_exit ($str,$stderr);
     }
  }
  elsif ($fhost eq 'bl2epu') {						#Eiger EPU
     if ($msg =~ /^.*${furka} -u /) {
        $msg =~ s/^.*${furka} -u //;
        ($uid,$gid,$partial_path) = split(/:/,$msg);
        if (!defined $uid || $uid !~ /^\d*$/ ||
            !defined $gid || $gid !~ /^\d*$/ ||
            !defined $partial_path) {
           $str = '!!! '.$script.': Unexpected FURKA string ['.$str.']'."\n"
                        .'--- in response to ['.$cmd.']'."\n"
                        .'--- when parsing ['.$msg.'] as uid:gid:partial_path'."\n";
           if (length($msg) == 0)  {$str .= '-- no uid:gid:partial_path in the string!'."\n";}
           if (!defined $uid) {
              $str .= '-- uid is undefined!'."\n";
           } else {
              $str .= 'uid=['.$uid.']'."\n";
              if ($uid !~ /^\d*$/) {$str .= '-- uid=['.$uid.'] is not numeric!'."\n";}
              else                 {$str .= 'uid=['.$uid.']'."\n";}
           }
           if (!defined $gid) {
              $str .= '-- gid is undefined!'."\n";
           } else {
              $str .= 'gid=['.$gid.']'."\n";
              if ($gid !~ /^\d*$/) {$str .= '-- gid=['.$gid.'] is not numeric!'."\n";}
              else                 {$str .= 'gid=['.$gid.']'."\n";}
           }
           if (!defined $partial_path) {
              $str .= '-- partial_path is undefined!'."\n";
           }
           &zenity_error_exit ($str,$stderr);
        } else {
           $cmd = $getent.$uid;
           if ($debug) {print STDERR $script.': Running cmd='.$cmd."\n";}
           $msg = `${cmd}`;						#Expecting: 'user0:x:600:600:user0:/mnt/share1/user0:/bin/bash'
           if ($debug) {print STDERR $script.': Return code='.${^CHILD_ERROR_NATIVE}.' ('.$?.')'."\n";}
           if (${^CHILD_ERROR_NATIVE}) {
              $str = '*** '.$script.' command:'."\n".$cmd."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.'.';
              if ($mode eq'remote') {$str .= "\n".'Please verify passwordless ssh connection to '.$fhost.' for the '.$user.' user.';}
              &zenity_error_exit ($str,$stderr);
           }
           @arr = split (/:/,$msg);
           $narr = @arr;
           if ($narr != 7) {print STDERR '!!! '.$script.': Unexpected getent data ['.$msg.'] for uid='.$uid."\n"; exit;}
           print STDOUT 'furka_user='.$arr[0]."\n";
           exit;
        }
     } else {
        $str = '!!! '.$script.': Unexpected FURKA process ['.$msg.']'."\n"
                        .'--- in response to ['.$cmd.']'."\n"
                        .'--- Expected=[pid '.$furka.' -u uid:gid:partial_path]'."\n";
        if ($msg =~ /${furka}/) {$str .= 'It is possible that FURKA was started without user specification';}
        &zenity_error_exit ($str,$stderr);
     }
  }


#=================================================================

#sub zenity_error_exit ($message,[$stderr]) {
sub zenity_error_exit ($;$) {

### Open graphical window to report

  my ($msg, $type, $stderr);
  my ($LUN, $status, $path, @cmd, $string);

  $msg    = shift(@_);                                          #pointer to message
  $stderr = shift(@_);						#0=stdout or 1=stderr
  if (!defined $stderr) {$stderr = 0;}

  $type   = 'error';
  $status = 0;
  $msg =~ s/\&/and/g;
# $msg =~ s/</\"/g;
# $msg =~ s/>/\"/g;
  $msg =~ s/</\</g;
  $msg =~ s/>/\>/g;
  @cmd = ("--title='".$script."'", '--'.$type, "--text='".$msg."'");
  push @cmd, '--no-wrap';						#this helps to reduce excessive white space below text
  push @cmd, '--ellipsize';						#this fixes the high window size with long texts
  push @cmd, ' >/dev/null', '2>&1';					#redirect both stdout and stderr (preferred)
  $path = '/usr/bin/zenity';
  if ($^O !~ /Win/i && -f $path) {					#Linux with zenity installed
### zenity --error --text="bla-bla-\rbla" --title="Beamstop-IN" --window-icon="/usr/share/pixmaps/applet-critical.png"
     unshift @cmd, $path;						#prepend with 'path'
     push @cmd, '&';
     $string = join(' ',@cmd);						#merge from array
     $status = system($string);						#only string form allows no-wait
     if ($status) {print STDOUT "\n".' !!! Error='.$status.' running '.$path."!\n\n";}
  }

  if ($stderr) {$LUN = \*STDERR;}
  else         {$LUN = \*STDOUT;}
  print $LUN "\n".'*** '.$script.':'."\n".$msg."\n\n";
  exit 1;
}

#=================================================================

# sub getCCDfpehost() {
sub getCCDfpehost() {			#returns ($fpehost,$status)

### This routine returns the name of beamline FPE host (e.g. bl1ppu, bl2epu, or bl3mar)

  our ($beamline);
  my ($script, $pv, $fpehost, $cmd, $str);

  $script   = 'getCCDfpehost';

  ### Here we need any PV hosted by FPE and common for both detectors:
# $pv = $beamline.':fpe:DetCmdStr';					#not present on Eiger2
# $pv = $beamline.':fpe:AcquireTime';					#not present on bl1ppu
  $pv = $beamline.':4:HOSTNAME';
  $cmd = 'cainfo -w 1 '.$pv.' | grep "^    Host:" 2>/dev/null';
  if ($debug) {print STDERR $script.': Running cmd='.$cmd."\n";}
  $fpehost = `$cmd`;
  if ($debug) {print STDERR $script.': Return code='.${^CHILD_ERROR_NATIVE}.' ('.$?.')'."\n";}
  if (${^CHILD_ERROR_NATIVE}) {
     $str = '*** '.$script.' command:'."\n".$cmd."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.'.';
     &zenity_error_exit ($str,$stderr);
  }
  $fpehost =~ s/^\s+Host:\s+//;
  $fpehost =~ s/:\d+\n*$//;
  if ($fpehost !~ /\.gmca\.aps\.anl\.gov$/) {return ('unknown',1);}
  else                                      {return ($fpehost,0);}
}

#=================================================================

# sub checkSSHconnection($host) {
sub checkSSHconnection($) {			#returns ($status)

### This routine verifies passwordless connection to the specified host

  our ($beamline);
  my ($host, @cmd, $str);

  $host = shift(@_);

  if ($debug) {print STDERR "\n".$script.': checking connection to '.$host.' for user='.$user."\n";}
  @cmd = ('/usr/bin/ssh','-q','-o','BatchMode=yes','-o','StrictHostKeyChecking=no',$host,'/bin/date','>','/dev/null');
  if ($debug) {print STDERR $script.': '.join(' ',@cmd)."\n";}
  $status = system(@cmd) >> 8;
  if ($status > 0)  {
     $str = '*** '.$script.': '."\n"
          . '    Passwordless SSH connection to host='.$host.' for user='.$user.' failed with code='.$status.'.'."\n\n"
          . '    The test command was:'."\n"
          . '    '.join(' ',@cmd)."\n\n"
          . '    Cannot continue until fixed.'."\n\n"
          . '    Please verify the SSH settings and then re-run '.$script."\n"
          . '    If in trouble, please contact Sergey Stepanov.';
     &zenity_error_exit ($str,$stderr);
  }
  else {if ($debug) {print STDERR $script.': connection to '.$host.' is OK.'."\n\n";}}
  return $status;
}
