#!/usr/bin/perl -W
#--------------------------------------------------------------
# This example PERL script:
# - calls the X0h program from https://x-server.gmca.aps.anl.gov/
# - gets the refration index delta (n = 1 - delta - i*eta)
# - saves them into a file as a function of energy
#
# The script can use either the LWP PERL module or wget (external program).
# In the later case it can run on Microsoft Windows under Cygwin
# (https://www.cygwin.com).
#
#     		Author: Sergey Stepanov
#
# Version-1:    2005/12/31
# Version-1.1:  2006/11/08
# Version-1.2:  2021/06/22      added comments for output file
#--------------------------------------------------------------
  use LWP::Simple;	        # World-Wide Web library for Perl (libwww-perl)
  use LWP::Protocol::https;
  use strict;
  use warnings;

  select STDOUT; $|=1;		# set unbuffered output

  my ($file, $E1, $E2, $dE, $E, $n, $url, $xway, $wave, $line);
  my ($coway, $code, $amor, $chem, $rho, $i1, $i2, $i3);
  my ($df1df2, $modeout, $detail, $address, $buffer);
  my (@content, $ncon, $x, $i, $j, $script);

#--------------Input parameters---------------------------
### Energy range and the number of energy points
  $E1 = 1.;		        # start energy
  $E2 = 60.;		        # end energy
  $dE = .1;                     # energy step
  $n = int(($E2-$E1)/$dE+1.5);  # number of energy pts

  $url = 'https://x-server.gmca.aps.anl.gov/cgi/x0h_form.exe';
###--------------Parameters of X0h CGI interface--------------
  $xway = 2;		        # 1 - wavelength, 2 - energy, 3 - line type
  $wave = 0;        	        # works with xway=2 or xway=3 (not used here, see #E below
# $line = "Cu-Ka1";	        # works with xway=3 only
  $line = '';   	        # works with xway=3 only

### Target:
  $coway = 0;   	        # 0 - crystal, 1 - other material, 2 - chemicalformula
### Crystal
  $code = 'Beryllium';          # works with coway=0 only
### Other material:
  $amor = '';   	        # works with coway=1 only
### Chemical formula:
  $chem = '';   	        # works with coway=2 only
### and density (g/cm3):
  $rho = '';		        # works with coway=2 only

### Miller indices:
  ($i1, $i2, $i3) = (0, 0, 0);

### Database Options for dispersion corrections df1, df2:
### -1 - Automatically choose DB for f',f"
###  0 - Use X0h data (5-25 keV or 0.5-2.5 A) -- recommended for Bragg diffraction.
###  2 - Use Henke data (0.01-30 keV or 0.4-1200 A) -- recommended for soft x-rays.
###  4 - Use Brennan-Cowan data (0.03-700 keV or 0.02-400 A)
###  6 - Use Windt data (0.01-100 KeV or 0.12-1240 A)
###  8 - Use Chantler data (0.01-450 KeV or 0.28-1240 A)
### 10 - Compare results for all of the above sources.
  $df1df2 = 2;							# use Henke for lower energy
# $df1df2 = 4;							# use Cowan for higher energy
# $df1df2 = 8;							# use Chantler for higher energy

### Output options:
  $modeout = 1;	                # 0 - html out, 1 - quasy-text out with keywords
  $detail  = 0;         	# 0 - don't print coords, 1 = print coords

### Output file:
  $file = $code.'.dat';
###-----------------------------------------------------------

### Open output file for writing:
  open (DAT,'>',$file) || die 'Cannot open '.$file."\n";	# overwrite DAT file
  select DAT; $|=1;					        # set unbuffered output

### Print data header:
  print DAT '#Energy(KeV),      delta'."\n";
  $script = $0 =~ s/^.*[\/\\]//gr;		#remove everything before last "/" or "\"
  print DAT '### Produced by '.$script.' on '.scalar(localtime)."\n";
  print DAT '### coway='.$coway;
  if    ($coway == 0) {print DAT ' code='.$code;}
  elsif ($coway == 1) {print DAT ' amor='.$amor;}
  elsif ($coway == 2) {print DAT ' chem='.$chem.' rho='.$rho;}
  print DAT ' df1df2='.$df1df2."\n";

  for ($i=0; $i<$n; $i++) { 	# Energy loop

     $E = sprintf '%.3f', ($E1 + $dE * $i);

     $address = $url
                .'?xway='.$xway
                .'&wave='.$E
                .'&line='.$line
                .'&coway='.$coway
                .'&code='.$code
                .'&amor='.$amor
                .'&chem='.$chem
                .'&rho='.$rho
                .'&i1='.$i1
                .'&i2='.$i2
                .'&i3='.$i3
                .'&df1df2='.$df1df2
                .'&modeout='.$modeout
                .'&detail='.$detail;

### Request X0h data from the server:
     $buffer = get($address);
     @content = split /\r?\n/, $buffer;	# split page into lines

     $ncon = @content;   			# number of lines on page

     print STDOUT '   '.$E;
     print DAT    '   '.$E;
     for ($j=0; $j<$ncon; $j++) {		# loop over page lines
        $x = $content[$j];
        chomp $x;				# strip LF/CR
        if ( $x =~ m/delta=/ ) {		# if contains "delta="
      	   $x =~ s/^.*delta=//g;		# erase everything before the data
           print STDOUT ',  '.$x,"\n";
           print DAT    ',  '.$x;
        }
     }
     print DAT "\n";
  }
