#!/usr/bin/perl -W
#--------------------------------------------------------------
# This script measures background (dark) intensity in Joerger scaler
# inputs 1-16 (1-32 for newer models or Struck working as Joerger).
# The intensity is measured with closed shutter and recorded into
# joerger$_Bkg.dat, where $=1,2,3... depending on the scaler number
# at the beamline.
#
#     		Author: Sergey Stepanov
#
# Version-2.0:  2001/02/05 using Pezca
# Version-2.1:  2013/05/30 added copying scaler-1 offests to Struck offsets
#--------------------------------------------------------------
  use strict;
  use Pezca;

  our ($beamline, $joergerName, $shutter, $shutter_readback);
  our (@excludes);					#may define externally in beamline.pl
  my ($PEZCA, $cnt_time, $narg, $outfile, $scaler, $scaler_no, $syntax);
  my ($maxJoerger, @Intensity, @inputs, $ninputs, $nexcludes, @files);
  my ($MBSH_ini, $shutter_ini, $handle, $script, $status, $i, $inp);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  do $PEZCA.'/Common/joerger_subs_v2.pl';
  do $PEZCA.'/Common/shutter_subs.pl';
  do $PEZCA.'/Common/motor_subs.pl';
  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  select STDOUT; $|=1;					# set unbuffered output

  $script   = &makeScriptName();
  $cnt_time = 10;					# joerger count time (s)
  $syntax   = "\n".'Syntax: '.$script.' <joerger_scaler_number>'."\n\n";
#----------------------------------------------------------------
  $narg = @ARGV;
  if ($narg != 1) {die $syntax;}
  if ($ARGV[0] !~ /^[01234]$/) {die $syntax;}

  $scaler_no = $ARGV[0];
  $outfile = 'joerger'.$scaler_no.'_Bkg.dat';
  $scaler  = $joergerName.$scaler_no;

  if ($scaler           !~ /^${beamline}:/) {$scaler           = $beamline.':'.$scaler;}
  if ($shutter          !~ /^${beamline}:/) {$shutter          = $beamline.':'.$shutter;}
  if ($shutter_readback !~ /^${beamline}:/) {$shutter_readback = $beamline.':'.$shutter_readback;}
#----------------------------------------------------------------
### This is currentlty for all scaler-1 on;y
  if ($scaler_no == 1) {
#    if (!defined @excludes) {                          #if not defined in beamline.pl
     if (! @excludes) {                                 #if not defined in beamline.pl
        if    ($beamline =~ /^23i/) {@excludes = (25, 26, 29, 30, 31, 32);}
        elsif ($beamline =~ /^23o/) {@excludes = (25, 26, 27, 28, 29, 30, 31, 32);}
        elsif ($beamline =~ /^23b/) {@excludes = (4, 5, 6, 7, 8, 9);}
        else                        {@excludes = ();}
     }
  }
  if (!grep(/^1$/,@excludes)) {push (@excludes,1);}	# timer channel
  @excludes = sort(@excludes);
#----------------------------------------------------------------
### Determine the number of inputs:
  $maxJoerger = &JoergerNCH($scaler);

### Compile the list of all inputs:
  @inputs = ();
  for ($i=0; $i<$maxJoerger; $i++) {$inputs[$i] = $i+1;}
  $ninputs = @inputs;
#----------------------------------------------------------------
### Get shutter status:
  $shutter_ini = &getShutter($shutter_readback);
  $MBSH_ini    = &getEndStationShutter($beamline);

  if ($shutter_ini == 1) {				# if shutter is initially open
     print STDOUT $script.': closing shutter '.$shutter."\n";
     &closeShutter($shutter,$shutter_readback);
  }
  if ($MBSH_ini == 1) {
     print STDOUT $script.': closing endstation shutter'."\n";
     $status = &closeEndStationShutter($beamline);
     if ($status) {&press_key_exit('*** '.$script.': failed to close endstation shutter');}
  }

### Set scaler:
  print STDOUT $script.': setting '.$scaler."\n";
  &setJoerger ($scaler, $cnt_time);

  print STDOUT $script.': measuring offsets with '.$scaler.' for '.$cnt_time.'s'."\n";
  @Intensity = &runJoerger ($scaler, $cnt_time, $ninputs, @inputs);

  print STDOUT $script.': excluding offsets for scaler inputs ('."@{excludes})\n";
  $nexcludes = @excludes;
  for ($i=0; $i<$nexcludes; $i++) {
     $inp = $excludes[$i];
     if (grep(/^${inp}$/,@inputs)) {$Intensity[$inp-1]=0;}
  }

  if ($MBSH_ini == 1) {
     print STDOUT $script.': opening endstation shutter'."\n";
     $status = &openEndStationShutter($beamline);
     if ($status) {
        print STDOUT '!!! '.$script.': failed to open endstation shutter'."\n";
     }
  }
  if ($shutter_ini == 1) {				# if shutter was initially open
     print STDOUT $script.': opening shutter '.$shutter."\n";
     &openShutter ($shutter,$shutter_readback);
  }

  @files = ($outfile);
  if ($scaler_no == 1) {push(@files,'struck_Bkg.dat');}		# scaler-1 is same as Struck

  foreach $outfile (@files) {
### Open file for writing:
     print STDOUT $script.': Opening file ['.$outfile.']'."\n";
     $handle = open (DAT,'> '.$outfile) || die 'Cannot open file ['.$outfile.']'."\n";	# overwrite DAT file
     if (!defined $handle) {
        &press_key_exit('*** '.$script.':  could not open file ['.$outfile.']');
     }
     print STDOUT $script.': writing data'."\n";
     for ($i=0; $i<$ninputs; $i++) {printf DAT '%d'."\n",int($Intensity[$i]);}
     close(DAT);
  }

  &clearAllCaMonitors();
  print STDOUT $script.': ..Finished!'."\n";
  exit;

