#!/usr/bin/perl -W
#--------------------------------------------------------------
# This GUI script is aimed at emulating listMotors2.tcl.
# It lists all beamline motor per PMAC along with their numbers.
#
#                   Author: Sergey Stepanov
#
# Version 1.0  2019/02/24 initial implementation.
# Version 1.1  2020/02/07 bug fixes for non-continuous motors and
#                         added configNN.tcl check for duplicate motors in PMAC
#--------------------------------------------------------------
  use Tk;
  use Pezca;
  use strict;
  use warnings;

  our ($NONINTERACTIVE, $USE_ZENITY_FOR_EXIT, $IGNORE_ERRORS, $beamline);
  my ($PEZCA, $script, $syntax, $arg, $narg, @beamlinelist);
  my (@list, $nassy, $nmotors, $i, $j, $k, $m, $f);
  my (@plst, $npl, $str, $mmax);
  my (@pmacs, $npmacs, $pmac);
  my ($redcolor, $pinkcolor, $bluecolor, $whitecolor);
  my ($mw, @blufr, @frame);

  if (!defined $ENV{PEZCA}) {
     die 'Please define the PEZCA environment. It must point to the PERL scripts depository';
  }
  $PEZCA = $ENV{PEZCA};
  $PEZCA =~ s/\\/\//g;
  if ($PEZCA !~ /\/$/) {$PEZCA .='/';}

# $beamline = '23i';

  do $PEZCA.'Common/motor_subs.pl';
  do $PEZCA.'Common/common_subs.pl';
  if (!defined $beamline) {do $PEZCA.'Common/beamline.pl';}
  else                    {do $PEZCA.'Common/beamline_'.$beamline.'.pl';}

  select(STDERR);  $|=1;				# Flush output after each print
  select(STDOUT);  $|=1;				# Flush output after each print

  $script = &makeScriptName();
  $syntax = "\n\n".'Syntax: '.$script.' [beamline]'."\n\n";
  @beamlinelist  = ('23i:', '23o:', '23b:', '23d:');
  $narg = @ARGV;
  if ($narg > 0) {print STDOUT $script.': started with cmdline = '."@ARGV\n\n";}
  else           {print STDOUT $script.': started with no cmdline arguments.'."\n\n";}

  $NONINTERACTIVE      = 1;
  $USE_ZENITY_FOR_EXIT = 1;
  $IGNORE_ERRORS       = 1;

  foreach $arg (@ARGV) {
     if (grep(/^${arg}:?$/, @beamlinelist)) {
        $arg =~ s/:$//;						#trim ':'
        if ($arg ne $beamline) {
           print STDOUT "\n".'!!! '.$script.': Cross-beamline operation: controlling '.$arg.' from '.$beamline."\n\n";
           $beamline = $arg;
           do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
	}
     } else {
        &press_key_exit('*** '.$script.': Unrecognized cmdline parameter "'.$arg.'"'.$syntax);
     }
  }

### Fiil @list - array of hashes with comprehensive information about all assemblies from configNN.tcl
### Keywords: label, assy, type, pmac, pcs, naxes, axis1,... ndrives, drive1,... nmotors, motor1, mtr1_no,...
  $nassy = &generateComponentList($beamline,\@list);
  if ($nassy < 0) {
     &press_key_exit('*** '.$script.': error='.$nassy.' from generateComponentList');
  }
  @pmacs = ();
  for ($i=0; $i<$nassy; $i++) {
     $pmac = $list[$i]->{'pmac'};
     if (! grep {/^$pmac$/} @pmacs) {push @pmacs,$pmac;}
  }
  @pmacs = sort @pmacs;
  $npmacs = @pmacs;
# for ($i=0; $i<$npmacs; $i++) {print STDOUT $pmacs[$i],"\n";}
  
  $redcolor    = '#FF0000';
  $pinkcolor   = '#FFC0CE';
  $bluecolor   = '#5959E5';
  $whitecolor  = '#FFFFFF';
# $greencolor  = '#00BC00';
# $goldcolor   = '#D0D000';
# $yellowcolor = '#E0E000';
# $lightcolor  = '#FDFDFD';
### Create a GUI:
  $mw = MainWindow->new;
                $mw->Label(-text =>'Beamline='.$beamline, 
                           -anchor=>'w',
                           -borderwidth=>1,
                           -relief=>'groove',
                           -background=>$redcolor,
                           -foreground=>$whitecolor,)->grid(-row=>0, -column=>0, -pady=>0, -padx=>0, -columnspan=>$npmacs, -sticky=>'nsew',);

  for ($j=0; $j<$npmacs; $j++) { ####################################### LOOP OVER PMACS
     $blufr[$j] = $mw->Frame(-borderwidth=>2,
                           -relief=>'flat',
                           -background=>$bluecolor,)->grid(-row=>1, -column=>$j, -pady=>0, -padx=>0, -sticky=>'nsew',);
     $frame[$j] = $blufr[$j]->Frame(-borderwidth=>0,
                           -relief=>'flat')->pack(-side => 'top',-pady=>0,-padx=>0,-fill=>'both',-expand =>1,);
     $frame[$j]->Label(-text =>$pmacs[$j], 
                           -anchor=>'center',
                           -borderwidth=>0,
                           -relief=>'flat',
                           -background=>$pinkcolor,)->pack(-side => 'top',-pady=>0,-padx=>0,-fill=>'x');

     @plst = ();                                        #list of motor PVs and numbers for given PMAC
     $f = 0;
     for ($i=0; $i<$nassy; $i++) {
        if ($list[$i]->{'pmac'} eq $pmacs[$j]) {
           $nmotors = $list[$i]->{'nmotors'};
           for ($k=0; $k<$nmotors; $k++) {
              $m = $k+1;  
              $plst[$f]{'motor'} = $list[$i]->{'motor'.$m};
              $plst[$f]{'mtrno'} = $list[$i]->{'mtr'.$m.'_no'};
              $f++;
           }
        }
     }
     $npl = @plst;
     @plst = sort {$a->{'mtrno'} <=> $b->{'mtrno'}} @plst;

### Check pmac's motors list for duplicate motor numbers:
     if ($npl > 1) {
        for ($i=1; $i<$npl; $i++) {
	   if ($plst[$i]{'mtrno'} == $plst[$i-1]{'mtrno'}) {
	      $str = '*** '.$script.': Motors '.$plst[$i]{'motor'}
	      .' and '.$plst[$i-1]{'motor'}.' in '.$pmacs[$j]
	      .' have the same number='.$plst[$i]{'mtrno'}
	      .'. Check config'.$beamline.'.tcl!';
	      &zenity_error($str);
	   }
        }
     } 

     if ($plst[$npl-1]->{'mtrno'} > 16) {$mmax = 32;}
     else                               {$mmax = 16;}
     $f = 0;
     for ($i=0; $i<$mmax; $i++) {
        $m = $i+1;
        if ($m == $mmax) {
           $str = sprintf '%02d %s', $m, 'reserved out';
        }
        elsif ($f < $npl && $m == $plst[$f]->{'mtrno'}) {
           $str = sprintf '%02d %s', $m, $plst[$f]->{'motor'};
           $f++;
        }
        else {
           $str = sprintf '%02d', $m;
        }                
#       $frame[$j]->Label(-text=>$str,                                  #this does not allow copying 
#                          -anchor=>'w',
#                          -borderwidth=>0,
#                          -relief=>'flat',)->pack(-side => 'top',-pady=>0,-padx=>0,-fill=>'x');
        $frame[$j]->Entry(-text=>$str, 
                          -state=>'readonly',
                          -readonlybackground=>'',
                          -width=>-1,
                          -borderwidth=>0,
                          -relief=>'flat',)->pack(-side => 'top',-pady=>0,-padx=>0,-fill=>'x');
     }
  } #################################################################### END OF LOOP OVER PMACS

  $mw->protocol('WM_DELETE_WINDOW', \&goodbye);

  $mw->resizable( 0, 0 );
  $mw->focus();

  $mw->repeat(1000,\&refresh_gui);

  MainLoop;

#==========================================================================
sub refresh_gui() {
  if (Exists($mw)) {$mw->update;}
}

#==========================================================================
sub goodbye() {
# require Tk::ErrorDialog;
  &clearAllCaMonitors();
# $mw->destroy;
# Tk::exit;
  exit 0;
}
