#!/usr/bin/perl -W
#--------------------------------------------------------------
# This script displays NX session information for given session number
#
# ATTENTION: this can only run via sudo -- add it to /etc/sudoers
#
#         	Author: Sergey Stepanov
#
# Version-1.0: 2021/02/21 initial implementation (parsing NX sessionId files or Redis DB)
#--------------------------------------------------------------
  use strict;
  use POSIX;
  use warnings::unused;

  my ($script, $host, @nxfolders, $folder, $cmd, $status, $line);
  my ($nxdir, $nxserver, $nxsesfile, $serverVersion, $serverMajor);
  my ($nxid, $found, @seslist, @seslines, @folderlist, $arg, $syntax);

  select STDOUT; $|=1;					#set unbuffered output
  select STDERR; $|=1;					#set unbuffered output

  $script = $0 =~ s/^.*[\/\\]//gr;
  $script =~ s/\.pl$//i;
  $syntax = "\n".'Syntax: '.$script.' sessionId [running|closed|failed]'."\n";
  @folderlist = qw(running closed failed);
#---------------------------------------------------------------------
  if ($^O =~ /Win/i) {die $script.': cannot run under Windows.'."\n";}

  if    (defined $ENV{HOSTNAME})  {$host = $ENV{HOSTNAME};}
  elsif (defined $ENV{HOST})      {$host = $ENV{HOST};}
  else                            {$host=`hostname`; $host=~s/\n//g;}
  if (!defined $host) {die $script.': undefined HOST'."\n";}

  $host = lc($host);
  $host =~ s/[.-]\Qgmca.aps.anl.gov\E$//;
  if ($host !~ /bl[123]ws[256]$/ && $host ne 'px0') {die $script.': unlisted host='.$host."\n";}

### ----------------------------------------------------------
### /etc/sudoers should have
# ALL ALL = (root) NOPASSWD:  /home/gmca/epics/pezca/nxdb.pl
### ----------------------------------------------------------
  if (geteuid() ne '0') {				#if not root
     if (exec('/usr/bin/sudo','-n',$0,@ARGV)) {die $script.': failed restarting with sudo.'."\n";}
  }

  @nxfolders = ();
  foreach $arg (@ARGV) {
     if    (length($arg) >= 32)           {$nxid = $arg;}
     elsif (grep(/^$arg$/i, @folderlist)) {push(@nxfolders,lc($arg));}
     else {die $script.': Incorrect command line argument ['.$arg.'].'.$syntax;}
  }
  if (!defined $nxid) {die $script.': please specify NX session id'.$syntax;}
  if (@nxfolders == 0) {@nxfolders = @folderlist;}

  $nxdir     = '/usr/NX';
  $nxserver  = $nxdir.'/bin/nxserver';

### Find the version of nxserver:
  if (! -e $nxserver) {die $script.': cannot find NX Server at '.$nxserver."\n";}
  $serverVersion = `$nxserver --version 2>/dev/null | grep ' - Version ' | grep -v ' WARNING: '`; 
  $serverVersion =~ s/^.* - Version //;
  chomp $serverVersion;
  $serverMajor = $serverVersion;
  $serverMajor =~ s/\..*//g;
  if ($serverMajor !~ /^\d*$/) {die $script.': error determining NX server generation: '.$serverMajor."\n";}

  $found = 0;
  foreach $folder (@nxfolders) {
     if ($serverMajor > 6) {
        $cmd = $nxserver.' --redis cmd smembers '.$folder;
        @seslist = split(/[\s\n]+/,`$cmd 2>/dev/null`);
        if (grep(/^${nxid}$/,@seslist) > 0) {$found=$folder; last;}
     } else {
        if (-e $nxdir.'/var/db/'.$folder.'/sessionId{'.$nxid.'}') {$found=$folder; last;}
     }
  }
  if ($found eq '0') {print STDOUT $script.': session ID='.$nxid.' not found.'."\n"; exit;} 

  print STDOUT $script.': session ID='.$nxid.' found in folder='.$found."\n"
              .'------------------------------------------------------'."\n";

### ========================================================================
### Read all information for given session from sessionId file or Redis:
### ========================================================================
  if ($serverMajor > 6) {
     $nxsesfile = 'Redis/'.$found.'.'.$nxid;
     $status = &read_nx_redis($nxserver,$found,$nxid,\@seslines);
  } else {
     $nxsesfile = $nxdir.'/var/db/'.$found.'/sessionId{'.$nxid.'}';
     $status = &read_entire_file($nxsesfile,\@seslines);
  }

  foreach $line (@seslines) {print STDOUT $line,"\n";}
  exit;

#==========================================================================

# sub read_entire_file($file,\@lines) {
sub read_entire_file($\@) {
# use Fcntl;					#not used (defines O_RDONLY constant (O_RDONLY=0,O_WRONLY=1,O_RDWR=2))
  my ($file, $linesref, $buffer, $status);

  $file     = shift(@_);
  $linesref = shift(@_);

  @$linesref = ();
  $status = sysopen(FH, $file, 0);		#0 is used instead of O_RDONLY defined in "use Fcntl";
  if (!defined $status) {print STDOUT '*** '.$script.': Could not open '.$file."\n"; return 1;}
  sysread(FH, $buffer, -s FH);
  close FH;
  @$linesref = split(/\r?\n/,$buffer);
  return 0;
}

#==========================================================================

# sub read_nx_redis($nxserver,$sesfolder,$session,\@lines) {
sub read_nx_redis($$$\@) {
  my ($nxserver, $sesfolder, $session, $linesref);
  my ($cmd, $buffer, @buf, $nbuf, $i);

  $nxserver  = shift(@_);
  $sesfolder = shift(@_);
  $session   = shift(@_);
  $linesref  = shift(@_);

  @$linesref = ();
  if ($sesfolder ne 'running' && 
      $sesfolder ne 'closed'  && 
      $sesfolder ne 'failed') {
     print STDOUT '*** '.$script.': incorrect NX Redis folder ['.$sesfolder.']'."\n";
     return 1;
  }
  $cmd = $nxserver.' --redis cmd hgetall '.$sesfolder.'.'.$session;
  $buffer = `$cmd 2>/dev/null`;
  if (${^CHILD_ERROR_NATIVE}) {
     print STDOUT '*** '.$script.': Error executing '.$cmd."\n";
     return 1;
  }
  @buf = split(/\s+/,$buffer);
  $nbuf = @buf;
  if ($nbuf == 0 || $nbuf != 2*int($nbuf/2)) {
     print STDOUT '*** '.$script.': incorrect words number='.$nbuf.' in the output of '.$cmd."\n";
     return 1;
  }
  for ($i=0; $i<$nbuf-1; $i+=2) {
     push @$linesref,$buf[$i].'='.$buf[$i+1];
  }
# foreach $buffer (@$linesref) {print STDOUT $buffer,"\n";}; exit;
  return 0;
}
