#!/usr/bin/perl -W
#--------------------------------------------------------------
# This script displays nxplayer version of current user when
# the user is logged with nxplayer to ws2, ws5, ws6 or px
#
# ATTENTION: this can only run via sudo -- add it to /etc/sudoers
#
#         	Author: Sergey Stepanov
#
# Version-1.0: 2017/04/03 initial implementation (parsing NX sessionId files for player version)
# Version-2.0: 2017/11/21 added checking if remote IP is in sshd_config AllowedRemote
# Version-2.1: 2018/02/17 added more debugging options
# Version-2.2: 2018/04/13 extended to work properly with web player
# Version-2.3: 2018/10/11 adopted to NX version 6.3.6
# Version-2.4: 2020/08/10 fixed interpreting outside name of workstation,
#                         allowed in debugging mode to request closed sessions
# Version-2.5  2021/01/22 added support for NX-7 (Redis database)
# Version-2.6  2021/02/07 added extracting foreignAddres from RealConnected in case of web sessions.
#--------------------------------------------------------------
  use strict;
# use Net::CIDR::Lite;					#this is loaded below via &try_load_module()					
# use warnings::unused;

  my ($script, $user, $home, $host, @lines, $line, @words);
  my ($debug, $nxdir, $nxserver, $serverVersion, $serverMajor);
  my ($nxuser, $nxip, $nxid, $nxclient, $nxOS, $nxbrowser, $nxsession);
  my ($nxsesfile, $nxsesfile2, $nxsesreal, $nxclienttype, $nxrunning);
  my ($str, $loginAllowed, $found, $i);
  my (@now, $log, $loghandle, $status, @seslines, @seslines2, $sesfolder);

  select STDOUT; $|=1;					#set unbuffered output
  select STDERR; $|=1;					#set unbuffered output

  $script = $0 =~ s/^.*[\/\\]//gr;
  $script =~ s/\.pl$//i;

# $debug = 1;
# $sesfolder = 'closed';					#for debugging
# $ENV{NX_SESSION_ID} = '8E34A18D3089E377D02C6A8692EEB9BB';	#for debugging outside NX session (set ID of real closed session)
  if (!defined $debug) {$debug = 0;}

#---------------------------------------------------------------------
  if ($^O =~ /Win/i) {die $script.': cannot run under Windows.'."\n";}

  if    (defined $ENV{HOSTNAME})  {$host = $ENV{HOSTNAME};}
  elsif (defined $ENV{HOST})	  {$host = $ENV{HOST};}
  else                            {$host=`hostname`; $host=~s/\n//g;}
  if (!defined $host) {
     if ($debug) {print STDOUT $script.': undefined HOST'."\n";}
     exit 1;
  }
  $host = lc($host);
  $host =~ s/[.-]\Qgmca.aps.anl.gov\E$//;
  if ($host !~ /bl[123]ws[256]$/ && $host ne 'px0') {
     if ($debug) {print STDOUT $script.': unlisted host='.$host."\n";}
     exit 1;
  }

  if    (defined $ENV{USER})      {$user = $ENV{USER};}
  elsif (defined $ENV{USERNAME})  {$user = $ENV{USERNAME};}
  else                            {$user = 'unknown';}

  if    (defined $ENV{HOME})      {$home = $ENV{HOME};}
  else                            {$home = 'unknown';}

### When initial user is not root, but he issued "su",
### the environment "USER" is preserved, but "HOME" is not.
  if ($user ne 'root' && $home !~ /^\/root/) {
### ----------------------------------------------------------
### /etc/sudoers should have
# Defaults env_keep += "NX_SESSION_ID"
# Defaults !requiretty
# ALL ALL = (root) NOPASSWD:  /home/gmca/epics/pezca/nxversion.pl
# ALL ALL = (root) NOPASSWD: /usr/bin/perl -W /home/gmca/epics/pezca/nxversion.pl
### ----------------------------------------------------------
#    exec('/usr/bin/sudo', 'NX_SESSION_ID='.$ENV{NX_SESSION_ID}, '-n', $0);
     $status = exec('/usr/bin/sudo', '-n', $0);
     exit 0;
  }
  if ($user eq 'root' && defined $ENV{SUDO_USER}) {$user = $ENV{SUDO_USER};}

#=====================================================================
# From this place we run as root (via sudo)
#---------------------------------------------------------------------
  if (!defined $ENV{NX_SESSION_ID}) {
     if ($debug) {print STDOUT $script.': undefined NX_SESSION_ID. This script should be called from NX session.'."\n";}
     exit 1;
  }

  if    ($debug == 1) {
     *LOG = *STDOUT;
  }
  elsif ($debug > 1) {
     @now = localtime(time);				# This is (sec,min,hour,mday,mon,year,wday,yday,isdst)
     $log = sprintf '/tmp/nxversion_%04d.%02d.%02d_%02d.%02d.log',$now[5]+1900,$now[4]+1,$now[3],$now[2],$now[1];
     if (defined $log) {
        $loghandle = open (LOG,' > '.$log);
        if (!defined $loghandle) {
           print STDERR $script.': cannot open logfile ['.$log.']. Switching to STDERR.'."\n\n";
           *LOG = *STDERR;
        } else {
           chmod 0666, $log;
           select LOG; $|=1;				#set unbuffered output
        }
     } else {
        *LOG = *STDERR;
     }
  }

  if (! defined $sesfolder) {$sesfolder = 'running';}
  $nxdir     = '/usr/NX';
  $nxserver  = $nxdir.'/bin/nxserver';
  $nxrunning = $nxdir.'/var/db/'.$sesfolder;
  $nxsession = $ENV{NX_SESSION_ID};
  $nxsesfile = $nxrunning.'/sessionId{'.$nxsession.'}';

### When we run via sudo, user becomes root, but we above we have taken the original name from $ENV{SUDO_USER}:
  if ($debug > 1) {print LOG 'host='.$host.' user='.$user.' home='.$home.' session='.$nxsession.' sessionFile='.$nxsesfile."\n";}     

### ========================================================================
### Find the version of nxserver:
### ========================================================================
  if (! -e $nxserver) {
     if ($debug) {print STDOUT $script.': cannot find NX Server at '.$nxserver."\n";}
     exit 1;
  }

  $serverVersion = `$nxserver --version 2>/dev/null | grep ' - Version ' | grep -v ' WARNING: '`; 
  $serverVersion =~ s/^.* - Version //;
  chomp $serverVersion;

  if (!defined $serverVersion) {
     if ($debug) {print STDOUT $script.': cannot find installed NX server'."\n";}
     exit 1;
  } else {
     if ($debug > 1) {print LOG 'Installed NX server version='.$serverVersion."\n";}
     $serverMajor = $serverVersion;
     $serverMajor =~ s/\..*//g;
     if ($debug > 1) {print LOG 'Installed NX server Generation='.$serverMajor."\n";}
     if ($serverMajor !~ /^\d*$/) {
        if ($debug > 1) {print LOG '!!! Error determining NX server Generation='.$serverMajor.'. Assuming 6.'."\n";}
        $serverMajor = 6;
     }
  }

### ========================================================================
### Read all information for the current session for sessionId or Redis:
### ========================================================================
  if ($serverMajor > 6 || ! -e $nxsesfile) {
     if ($debug > 1 && $serverMajor <= 6) {
        print LOG 'NX session file='.$nxsesfile.' does not exits. Assuming NX7/Redis'."\n";
     }
     $status = &read_nx_redis($nxserver,$sesfolder,$nxsession,\@seslines);
     $nxsesfile='Redis/'.$sesfolder.'.'.$nxsession;
  } else {
     $status = &read_entire_file($nxsesfile,\@seslines);
  }
  if ($status) {
     if ($debug) {print LOG 'Error determining session information for '.$sesfolder.' and '.$nxsession."\n";}     
     exit 1;
  } else {
     if ($debug > 1) {
        print LOG '-------------------------------- Session file:'."\n";
        foreach $line (@seslines) {print LOG $line."\n";}
        print LOG '-------------------------------- End Session file'."\n";
     }
  }

### ========================================================================
### Check if current session exists among today's NX logins:
### ========================================================================
  
  $status = &try_load_module('Net::CIDR::Lite');			#this will exit if no module found  (see below)

  if ($sesfolder eq 'running') {					#this is real running session
     $str = `$nxserver --list | grep '^[0-9]'`;
     if ($debug > 1) {print LOG 'Output of "'.$nxserver.' --list":'."\n".$str."\n";}
     @lines = split(/[\n\r]/,$str);
     foreach $line (@lines) {
        @words = split(/\s+/,$line);
        if ($debug > 1) {print LOG 'Line=['.$line.'].'."\n";}
        if (@words == 5 && $words[0] =~ /^[0-9]*$/) {
           $nxid   = $words[3];
           if ($nxid ne $nxsession) {
              if ($debug > 1) {print LOG 'Skipping not-matching nxid='.$nxid.' nxsession='.$nxsession."\n";} 
              next;
           } else {
              $nxuser = $words[1];
              $nxip   = $words[2];    
              $found  = 1;
              last;
           }
        }
     }
     if (! defined $found) {
        if ($debug) {print STDOUT $script.': no proper record in the output of "nxserver --list":'."\n".$str."\n";}
        exit 1;
     }
     elsif (! defined $nxuser || length($nxuser) < 1) {
        if ($debug) {print STDOUT $script.': no Username in the output of "nxserver --list":'."\n".$str."\n";}
        exit 1;
     }
     elsif (! defined $nxip || length($nxip) < 1) {
        if ($debug) {print STDOUT $script.': no Remote IP in the output of "nxserver --list":'."\n".$str."\n";}
        exit 1;
     }
  } else {								#this is in case of debugging
     $nxid   = &findkey2('sessionId',$nxsesfile,\@seslines);
     $nxuser = &findkey2('userName',$nxsesfile,\@seslines);
     $nxip   = &findkey2('foreignAddres',$nxsesfile,\@seslines);
     if (length($nxid) == 0 || $nxid ne $nxsession) {
        if ($debug) {print STDOUT $script.': unexpected nxid='.$nxid.' in '.$nxsesfile.', nxsession='.$nxsession."\n";}
        exit 1;
     }
     elsif (length($nxuser) == 0) {
        if ($debug > 1) {print LOG 'No userName in '.$nxsesfile.', nxsession='.$nxsession."\n";}
        $nxuser = '-';
     }
     elsif (length($nxip) == 0) {
        if ($debug > 1) {print LOG 'No foreignAddres in '.$nxsesfile.', nxsession='.$nxsession."\n";}
        $nxip = '-';
     }
  }
        
  if ($nxuser ne $user && $debug > 1) {print LOG '!!! nxuser='.$nxuser.' script_user='.$user.' are not matching.'."\n";}
  $nxclienttype = &findkey2('clientType',$nxsesfile,\@seslines);
  if ($debug > 1) {print LOG 'nxclienttype(clientType)='.$nxclienttype."\n";}
  if ($nxclienttype eq 'nxwebclient') {
     if ($debug > 1) {print LOG 'Client is "nxwebclient".'."\n";}
     $nxbrowser = &findkey2('clientBrowser',$nxsesfile,\@seslines);
     if ($debug > 1) {print LOG 'nxbrowser(clientBrowser)='.$nxbrowser."\n";}
     if (length($nxbrowser) > 0) {$nxOS='web(browser:'.$nxbrowser.')';}
     else                        {$nxOS='web';}
     if ($nxip eq '-') {
        if ($debug > 1) {print LOG 'Getting remote IP of "nxwebclient" as "foreignAddres" in the session file.'."\n";}
        $nxip = &findkey2('foreignAddres',$nxsesfile,\@seslines);
        if ($debug > 1) {print LOG 'nxip(foreignAddress)='.$nxip."\n";}
        if (length($nxip) == 0) {
           if ($debug > 1) {print LOG 'Getting RealConnected session ID.'."\n";}
           ### In NX-7 "RealConnected" does not show up in its Redis database immediately. We need to wait until it shows up:
           $i=0;
           do {
              $i++;
              $nxsesreal = &findkey2('RealConnected',$nxsesfile,\@seslines);
              if ($debug > 1) {print LOG 'RealConnected('.$i.')='.$nxsesreal."\n";}
              if (length($nxsesreal) == 0) {
                 sleep(1);
                 if ($serverMajor > 6) {&read_nx_redis($nxserver,$sesfolder,$nxsession,\@seslines);}
                 else                  {&read_entire_file($nxsesfile,\@seslines);}
              }
           } while (length($nxsesreal) == 0 && $i<=10);
           if (length($nxsesreal) > 0) {
              if ($serverMajor > 6) {
                 $nxsesfile2 = 'Redis/'.$sesfolder.'.'.$nxsesreal;
                 $status = &read_nx_redis($nxserver,$sesfolder,$nxsesreal,\@seslines2);
              } else {
                 $nxsesfile2 = $nxrunning.'/sessionId{'.$nxsesreal.'}';
                 $status = &read_entire_file($nxsesfile,\@seslines);
              }
              if ($status) {
                 if ($debug) {print LOG 'Failed reading data from RealConnected='.$nxsesreal."\n";}
                 exit 1;
              } else {
                 if ($debug > 1) {print LOG 'Getting remote IP of "RealConnected".'."\n";}
                 $nxip = &findkey2('foreignAddres',$nxsesfile2,\@seslines2);
                 if ($debug > 1) {print LOG 'nxip(foreignAddress)='.$nxip."\n";}
              }
           }
        }
     }
  } else {								#regular client
     if ($debug > 1) {print LOG 'Client is a regular nxplayer (session.'."\n";}
     $nxOS = &findkey2('clientPlatform',$nxsesfile,\@seslines);
     if ($debug > 1) {print LOG 'nxOS(clientPlatform)='.$nxOS."\n";}
     if ($nxip eq '-' && $debug > 1) {					#we should not need this in real operation
        if ($debug > 1) {print LOG 'No IP. Getting remote IP of regular client as "foreignAddres" in the session file.'."\n";}
        $nxip = &findkey2('foreignAddres',$nxsesfile,\@seslines);
        if ($debug > 1) {print LOG 'nxip(foreignAddress)='.$nxip."\n";}
     }
  }
  if ($nxip eq '-') {
     if ($debug) {print LOG '!!! Missing nxip='.$nxip.' (presumably disconnected session).'."\n";}
     exit 1;
  }
  if ($debug > 1) {print LOG 'nxuser='.$nxuser.'  nxip='.$nxip.'  nxid='.$nxid."\n";}
  $nxclient = &findkey2('clientVersion',$nxsesfile,\@seslines);
  if ($debug > 1) {print LOG 'nxclient(clientVersion)='.$nxclient."\n";}
        
  $loginAllowed = &checkLoginAllowed($user,$nxip);
  $str = 'nxclient='.$nxclient.' OS='.$nxOS.' nxserver='.$serverVersion.' ip='.$nxip.' loginAllowed='.$loginAllowed."\n";
  if ($debug > 1) {print LOG $str;}
  print STDOUT $str;
  exit;

#==========================================================================
sub findkey($$) {				#$key,$sessionfile
  my ($key, $sessionfile, $str, $val);
  ($key,$sessionfile) = @_;
  $str = '/usr/bin/grep -hs "^'.$key.'=" '.$sessionfile;
  $val = `$str`;
# if ($debug > 1) {print LOG 'findkey: search=['.$str.'], result=['.$val.'].'."\n";}
  $val =~ s/$key=//;
  $val =~ s/[\n\r]//g;
  return $val;
}

#==========================================================================
sub findkey2($$\@) {				#$key,$file,\@buf
  my ($key, $file, $bufref, @rez, $nrez, $val);
  ($key,$file,$bufref) = @_;
  if (defined $file && length($file) > 0) {$file = ' in ['.$file.']';} else {$file = '';}
  @rez = grep(/^${key}=/,@$bufref);
  $nrez = @rez;
  if ($nrez > 0) {
     $val = $rez[0];
     $val =~ s/$key=//;
     $val =~ s/[\n\r]//g;
     if ($nrez > 1 && $debug > 1) {
        print LOG 'findkey2: search for ['.$key.']'.$file.' returned multiple results (using first):'."\n@rez\n";
     }
  } else {
     if ($debug > 1) {
        print LOG 'findkey2: search for ['.$key.']'.$file.' did not find anything.'."\n";
     }
     $val = '';
  }
  return $val;
}

#==========================================================================

sub checkLoginAllowed($$) {				#$user,$ip
  my ($user, $ip, $sshcfg, @buf, $key);
  my ($str, @words, $nwords, $word, $usr, $adr);
  my ($cidr, @cidr_list, $nranges, @parts, $nparts);
  my ($ipmin, $ipmax, $found, $i, $restrict);

  ($user,$ip) = @_;

  if (!defined $ip || length($ip)==0 || $ip eq '-') {return 0;}
  
  $sshcfg = '/etc/ssh/sshd_config';
  if (! -e $sshcfg) {
     if ($debug > 1) {print LOG $script.': file "'.$sshcfg.'" does not exist.'."\n";}
     return 0;
  }
  if (! -r $sshcfg) {
     if ($debug > 1) {print LOG $script.': file "'.$sshcfg.'" is not readable.'."\n";}
     return 0;
  }
  if (&read_entire_file($sshcfg,\@buf)) {
     if ($debug > 1) {print LOG $script.': cannot open "'.$sshcfg.'" for reading.'."\n";}
     return 0;
  }
  $key = 'AllowUsers';
  $cidr = Net::CIDR::Lite->new;
  $restrict = 0;
  $nranges = 0;
  foreach $str (@buf) {
     if ($str !~ /$key/) {next;}
     $str =~ s/DO_NOT_MODIFY_AFTER_HERE//g;	# remove divider
     $str =~ s/\t/ /g;				# replace tabs by spaces
     $str =~ s/\s+/ /g;				# compress spaces ("+" means match 1 or more times)
     $str =~ s/^ *//g;				# remove beginning-of-the-line spaces
     $str =~ s/ *$//g;				# remove end-of-the-line spaces
     $str =~ s/^#.*$//g;			# remove comment lines
     if ($str =~ /^${key}/) {
        $restrict = 1;				#there is at least one AllowUsers statement
        $str =~ s/^${key} //;
        @words = split(/\s+/,$str);
        $nwords = @words;
        foreach $word (@words) {
           if ($word =~ /\@/) {($usr,$adr) = split(/\@/,$word);}
           else               {$usr = '*'; $adr = $word;}
           if ($usr eq $user || $usr eq '*') {
              if    ($adr =~ /^([0-9]{1,3}\.){1,3}[0-9]{1,3}\/{1}[0-9]{1,2}$/) {			#record is CIDR: 1.2.3/24
                 if ($debug > 1) {print LOG 'Found CIDR=',$adr,"\n";}
                 $cidr->add($adr);
                 $nranges++;
              }
              elsif ($adr =~ /^([0-9]{1,3}\.){1,3}[0-9]{1,3}$/) {			#record is IP or range: 1.2.3 (which means 1.2.3.1-1.2.3.255)
                 @parts = split(/\./,$adr);
                 $nparts = @parts;
                 if    ($nparts < 4 && $nparts > 1) {
                    $ipmin = $ipmax = $adr;
                    for ($i=0; $i<(4-$nparts); $i++) {$ipmin .= '.1'; $ipmax .= '.255';}
                    if ($debug > 1) {print LOG 'Found RANGE=',$ipmin.'-'.$ipmax,"\n";}
                    $cidr->add_range($ipmin.'-'.$ipmax);
                    $nranges++;
                 } 
                 elsif ($nparts == 4) {
                    if ($debug > 1) {print LOG 'Found IP=',$adr,"\n";}
                    $cidr->add_ip($adr);
                    $nranges++;
                 }
                 else {
                    if ($debug > 1) {print LOG $script.': Unexpected '.$key.' address='.$adr."\n";}
                 }
              }
              elsif ($adr =~ /^([0-9]{1,3}\.){3}\*$/) {					#record is IP range: 1.2.3.*
                 $adr =~ s/\*$//;
                 $ipmin = $adr.'1';
                 $ipmax = $adr.'255';
                 if ($debug > 1) {print LOG 'Found RANGE=',$ipmin.'-'.$ipmax,"\n";}
                 $cidr->add_range($ipmin.'-'.$ipmax);
                 $nranges++;
              }
              elsif ($adr =~ /^([0-9]{1,3}\.){2}\*\.\*$/) {				#record is IP range: 1.2.*.*
                 $adr =~ s/\*\.\*$//;
                 $ipmin = $adr.'1.1';
                 $ipmax = $adr.'255.255';
                 if ($debug > 1) {print LOG 'Found RANGE=',$ipmin.'-'.$ipmax,"\n";}
                 $cidr->add_range($ipmin.'-'.$ipmax);
                 $nranges++;
              }
           }
        }   
     }
  }
  if ($nranges > 0) {
     if ($debug > 1) {
        @cidr_list = $cidr->list();
        print LOG 'CIDR list:'."\n";
        foreach $i (@cidr_list) {print LOG '   '.$i."\n";}
     }
     $found = $cidr->find($ip);								#returns 1 or empty string
     if ($found ne '1') {$found = 0;}
  } else {
     $found = 1 - $restrict;
  }
  if ($debug > 1) {print LOG 'nranges='.$nranges.'  restrict='.$restrict.' found='.$found."\n";}
  return $found;
}

#=================================================================

sub try_load_module($) {
  my $module = shift(@_);
  eval('use '.$module);
  if ($@) {
     if ($debug) {print STDOUT $script.': failed to load module="'.$module.'"'."\n";}
     die '*** '.$script.': module "'.$module.'" failed to load.';
  }
}

#==========================================================================

# sub read_entire_file($file,\@lines) {
sub read_entire_file($\@) {
# use Fcntl;					#not used (defines O_RDONLY constant (O_RDONLY=0,O_WRONLY=1,O_RDWR=2))
  my ($file, $linesref, $buffer, $status);

  $file     = shift(@_);
  $linesref = shift(@_);

  @$linesref = ();
  $status = sysopen(FH, $file, 0);		#0 is used instead of O_RDONLY defined in "use Fcntl";
  if (!defined $status) {print STDOUT '*** '.$script.': Could not open '.$file."\n"; return 1;}
  sysread(FH, $buffer, -s FH);
  close FH;
  @$linesref = split(/\r?\n/,$buffer);
  return 0;
}

#==========================================================================

# sub read_nx_redis($nxserver,$sesfolder,$session,\@lines) {
sub read_nx_redis($$$\@) {
  my ($nxserver, $sesfolder, $session, $linesref);
  my ($cmd, $buffer, @buf, $nbuf, $i);

  $nxserver  = shift(@_);
  $sesfolder = shift(@_);
  $session   = shift(@_);
  $linesref  = shift(@_);

  @$linesref = ();
  if ($sesfolder ne 'running' && 
      $sesfolder ne 'closed'  && 
      $sesfolder ne 'failed') {
     print STDOUT '*** '.$script.': incorrect NX Redis folder ['.$sesfolder.']'."\n";
     return 1;
  }
  $cmd = $nxserver.' --redis cmd hgetall '.$sesfolder.'.'.$session;
  $buffer = `$cmd 2>/dev/null`;
  if (${^CHILD_ERROR_NATIVE}) {
     print STDOUT '*** '.$script.': Error executing '.$cmd."\n";
     return 1;
  }
  @buf = split(/\s+/,$buffer);
  $nbuf = @buf;
  if ($nbuf == 0 || $nbuf != 2*int($nbuf/2)) {
     print STDOUT '*** '.$script.': incorrect words number='.$nbuf.' in the output of '.$cmd."\n";
     return 1;
  }
  for ($i=0; $i<$nbuf-1; $i+=2) {
     push @$linesref,$buf[$i].'='.$buf[$i+1];
  }
# foreach $buffer (@$linesref) {print STDOUT $buffer,"\n";}; exit;
  return 0;
}
