#!/usr/bin/perl -W
# _CALLED_FROM_BLUICE_
#---------------------------------------------------------------------------
# This script attempts to optimize fluorescence signal by finding
# correct attenuation and moving the fluorescence detector in/out.
# It also optimizes I0 signal by setting appropriate gain on the
# Keithley amplifier.
#
#		Author: Sergey Stepanov
#
# Steps:
#    1. Go to some eV above the absorption edge selected in BluIce
#    2. Move fluorescence detector to the sample (to the closest distance)
#    3. Set initial attenuation (by default, set max attenuation
#    4. By varying attenuation and detector distance, optimize fluorescence 
#       detector deadtime to be in range [0.8-1.2] of target.
#    5. Optimize Keithley gain on I0.
#
#
# Subroutines:
#    sub set_conditions()
#    sub measure_signal()
#    sub signal_handler([$signal])
#---------------------------------------------------------------------------
# Version-1.0:  2008/03/13 initial release using Pezca
# Version-1.1:  2008/03/31 fork capture added.
# Version-1.2:  2008/07/21 tell BluIce option added
# Version-1.3:  2009/02/13 changed to use zenity for all exits
# Version-1.4:  2009/02/17 added posting optimized values as PVs
# Version-1.5:  2009/03/31 added check of SR current.
#                          added checking I0 before throwing in attenuation
#                          and increasing Keithley gain if needed.
#                          added recalculating Keithley gain according to
#                          the attenuation change imposed by MCA optimization.
# Version-1.6:  2009/05/14 added option to pre-drive energy to the absorption
#                          edge selected in BluIce (Using bigEnergyMove.pl)
# Version-1.7:  2009/05/20 added recording initial attenuation and I0 keithley
#                          gain if initially the detector is in parking pos.
# Version-1.8:  2009/07/21 added power cycling the detector at start
# Version-1.9:  2009/10/03 added putting beamstop/collimator IN and checking
#                          how far are we initially from the requested edge.
# Version-2.0:  2009/10/13 restored going to edge with bigEnergyMove
# Version-3.0:  2010/03/26 changed to use "signal" as a key to find I0 Struck
#                          input and I0 Keithley channel from mySQL; also
#                          added measuring I0 with measureDark1.pl if needed
# Version-3.1:  2010/04/13 fixed possible infinite in the fluo intensity optimization
# Version-3.2:  2010/04/14 added resetting MCA ADC and TCA to whole range before
#                          doing optimization
# Version-3.3:  2011/01/07 added global flag for opening popups via BluIce
# Version-4.0:  2011/07/08 made detector power cycling optional; changed optimization
#                          to start with the closest detector distance;
#                          Combining fluorescence optimization with Keithley optimization
#                          was planned, but not implemented since it may delay the procedure:
#                          it would cause extra [SLOW] gain changes because of varying
#                          attenuation for the MCA.
# Version-4.1:  2013/04/24 changed to work with sampleEnvSrv
# Version-4.2:  2014/02/01 added startup confirmation to prevent accidental starting from EPICS toolbar
# Version-4.3:  2014/03/28 changed functions calls to arrays pointers (preparation to Perl-5.16)
# Version-4.4:  2016/10/31 added checks for negative I0 after subtracting dark;
#                          added checking darks for overflow and restricting max Keithley gain;
#                          changed search for I0 initial gain to start from 1E3 instead of current gain
# Version-4.5:  2019/11/06 added command line option to bring IN pid diode instead of beamstop
# Version-4.6:  2020/02/21 added global switch for using Falcon instead of Canberra
# Version-4.7:  2020/10/13 added support for transfocator at the beamline (safer energy changes)
# Version-4.8:  2020/10/22 added $aboveEdgeEV instead of default 150eV and changed IDD default to 50eV
# Version-4.9:  2020/11/18 added @bigEnergyFlags
# Version-5.0:  2020/11/19 added $attenuation_max and optimizing distance if max attenuation reached
# Version-5.1:  2021/05/07 changed to 3 branches depending on starting with 1=OK, 2=High, 3=Low deadtime.
# Version-5.2:  2021/07/14 added $darkmode=zero|real and search mode MCA-nocount; 
#                          increased $mca_minimum, decreased max attenuation;
#                          added two command line parameters: MCAmin=xx and I0min=xx
# Version-5.3:  2021/08/26 adapted to new runMono which accounts for the 5th and 7th harmonic
# Version-5.4:  2021/10/27 added gnuplot and input for max_iterations
#---------------------------------------------------------------------------
  use strict;
  use Pezca;
# use Math::Trig;							#needed for mono Theta <-> Energy calculations
  use DBI;								#mySQL DB interface (for attenuation, edge energy & I0 darks)
  use warnings::unused;

  our ($USE_FALCON);							#global from beamline.pl
  our ($beamline, $shutter, $shutter_readback);
  our ($joergerName, $struckName);
  our ($have_KBM, $have_TF, $have_CRL);
  our ($undulator, $ATTENUATION_USE_FACTOR);
  our ($NONINTERACTIVE, $USE_ZENITY_FOR_EXIT, $ZENITY_VIA_BLUICE);
  our ($enmono13, @harmEn, @harmNum, @Elower, @Eupper, @Olower, @Oupper);		#make global to store SQL data between calls

  my ($PEZCA, $script, $scr, $cmdline, $syntax, $cmd, $msg, $msg2, $tell_bluice);
  my ($i0, $cntTime, $scalerNo, $attenuation_start, $deadtime_goal, $struck);
  my ($narg, @processes, $nprocesses, $yesno);
  my (@beamlinelist, $bl_specified, $dt_specified);
  my ($i0_specified, $ct_specified, $at_specified, $om_specified);
  my ($bi_specified, $pc_specified, $fldPosLo, $fldPosHi);
  my ($oi_specified, $ge_specified, $gotoEdge, $biEdgePV, $saveTriggPV);
  my ($pn_specified, $bringPinDiode);
  my ($am_specified, $attenuation_max);
  my ($im_specified, $mm_specified, $gn_specified, $it_specified, $itm);
  my ($scaler_min, $scaler_max, $mca_minimum, $I0_minimum, $I0_optimum);
  my ($optimizeMCA, $optimizeI0, $mca, $tca, $adc, @mca_deadtime, @mca_total);
  my ($fldAssyPV, $fldPosPV, $fldPos_ini, @fldPos, $fldAct);
  my ($scaler, $fileBkg, $bkg_age_hours, $I0_cts, $low, $darkmode);
  my ($attenuation_ini, @attenuation, $success, $iii, $iii_max);
  my ($status, $i, $arg, $mode, $forkflag, $fff);
  my ($start_time, $end_time, $powerPV, $powerState, $powercycle, $hasoverfill);
  my ($kchannel, $kgain_ini, $kgain_max, $kgain, $igain, $igain_max, @gains);
  my ($energyAsPV, $energyPV, $postEnPV, $postAtPV, $postDiPV, $postGaPV);
  my ($energy_act, $attenuation_act, $distance_act);
  my ($SRcurrentPV, $SRcurrent, $I0_atstart, $I0_is_upstream_fast_shutter);
  my ($Edge, $Edge_element, $Edge_KLMN, $EdgeEnergy, $aboveEdgeEV, $ae_specified);
  my (@E_step, @E_pm, $FluoEnergy, $FluoPMwidth);
  my ($parkEnPV,  $parkAtPV, $parkGaPV, $parkDiPV, $l_limit, $h_limit);
  my ($wasParked, $parkAttn, $parkGain, $parkDist, $allowance);
  my (@prg, $E_ini, $ESinTheta, @bigEnergyFlags);
  my ($Theta_ini, $Theta_edge, $Theta_big, $ID_ini, $IDoffset, $linkID);
  my ($i0_name, $timestampDB, $epochDB, @darks, $dark, @signals);
  my ($klabelDB, $koffsetDB, $klabel, $koffset, $epoch, $offset_delta);
  my ($motionNpts, $motitonRange, $motionStep);
  my ($syncMode, $syncStep, $switchID, $attenuation, $switchMono);
  my ($gnuplot, $gnuopen, @gnuargs, $gnuimage, @ct, $displaytime, @x, @y);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  do $PEZCA.'/Common/motor_subs.pl';
  do $PEZCA.'/Common/shutter_subs.pl';
  do $PEZCA.'/Common/joerger_subs_v2.pl';
  do $PEZCA.'/Common/struck_subs_v2.pl';
  do $PEZCA.'/Common/fluo_subs.pl';
  do $PEZCA.'/Common/attenuation_subs_v2.pl';
  do $PEZCA.'/Common/keithley_subs.pl';
  do $PEZCA.'/Common/gnuplot_subs.pl';
  do $PEZCA.'/Common/mono_subs.pl';
  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  select STDOUT; $|=1;							#set unbuffered output
  $SIG{INT}  = 'signal_handler';
  $SIG{QUIT} = 'signal_handler';

  @beamlinelist = ('23i:', '23o:', '23b:', '23d:');
  $script = &makeScriptName();
  $script =~ s/\.pl$//i;
  $scr    = 'optimizeFluo';
  $cmdline = ' [beamline] [noMCA] [noI0] [deadtime=xx]'        ."\n"
           . ' [gotoEdge|gotoEdgeSmall|gotoEdgeNone]'          ."\n"
           . ' [i0=name] [countTime=xx] [iterationsMax=nn]'    ."\n"
           . ' [bluice|nobluice] [powercycle|nopowercycle]'    ."\n"
           . ' [bringPinDiode|bringBeamStop] [aboveEdgeEV=xx]' ."\n"
           . ' [attenuationStart=nn] [attenuationMax=xx]'      ."\n"
           . ' [MCAmin=xx] [I0min=xx]'                         ."\n"
           . ' [gnuplot|gnuplot=image|nognuplot]'             ."\n"
           ;
  $syntax = "\n\n".'Syntax: '.$script.'.pl'.$cmdline."\n";
#----------------------------------------------------------------
### Protection from multiple instances:

  if ($^O !~ /Win/i) {
#    $cmd = '/bin/ps -ef'
     $cmd = '/bin/ps -eo user:19,pid,cmd'
           .' | /bin/grep -iv "grep\|edit\|geany\|term\|onsole\|bash\|csh\|pezca23id"'
           .' | /bin/grep optimizeFluorescence';
     $msg = `$cmd`;
     @processes = split(/\n/, $msg);
     $nprocesses = @processes;
     print STDOUT $script.': nprocesses = '.$nprocesses."\n";
     if ($nprocesses > 1) {
        $msg = '*** '.$script.': '.$nprocesses.' instance(s) of script are running!'
                     ."\n".$msg;
        &zenity_error_exit ($msg, undef,'nowait');
     }
  }
#----------------------------------------------------------------

###------------------------------------------------------------------------
### Defaults (if no cmdline arguments are given):
### These can be overwritten with cmdline options:
  $I0_is_upstream_fast_shutter = 0;					#1=no need to open shutter when measuring I0
  $i0_name                     = 'I0';					#mySQL tag for I0 ('I0', 'Pin', 'Beamstop')
  $cntTime                     = 0.5;					#scaler and MCA counting time
  $scalerNo                    = 1;					#scaler1=Struck, scaler3=Joerger
# $attenuation_max             = 500.;					#max attenuation (cannot define before we know beamline)
# $attenuation_start           = $attenuation_max;			#starting attenuation (cannot defined before we know beamline)
  $deadtime_goal               = 10;					#this is expressed in '%'
  $iii_max                     = 20;					#maximum iterations
  $tell_bluice                 = 1;					#1=inform BluIce 0=don't
  $powercycle                  = 0;					#1=powercycle detector, 0=don't
  $bringPinDiode               = 0;					#1=bring PinDiode, 0=bring Beamstop
  $scaler_min                  = 5;					#min allowed struck input
  $scaler_max                  = 32;					#max allowed struck input
# $mca_minimum                 = 100;					#minimum total intensity that can be considered not noise (need to know MCA)
  $I0_minimum                  = 1000;					#minimum I0 that can be considered not noise
  $I0_optimum                  = 0.95E+6;				#range is 0-10V times V2F converter (1E5)
  $I0_optimum                  = $I0_optimum/2.;			#take the middle
  $optimizeMCA                 = 1;
  $optimizeI0                  = 1;
  $motionNpts                  = 5;					#detector steps to move backward if fluorescence too high
  $tell_bluice                 = 1;
  $darkmode                    = 'zero';				#zero or real
  $ATTENUATION_USE_FACTOR      = 1;					#USE 'x' only!!!
  $USE_ZENITY_FOR_EXIT         = 1;
  $gotoEdge                    = 0;					#before optimizing, drive E to edge chosen in BluIce: 1=always, 2=only if small dE
  $gnuplot                     = 0;					#0=no gnuplot, 1=gnuplot for MCA deadtime vs counts 2=gnuplot saved as image
  $displaytime                 = 30;					#how long to display the plot
  if ($tell_bluice) {$NONINTERACTIVE=1; $ZENITY_VIA_BLUICE=1;}		#preliminary
  else              {$NONINTERACTIVE=0; $ZENITY_VIA_BLUICE=0;}		#preliminary
### end of Defaults (if no cmdline arguments are given)
###------------------------------------------------------------------------

###------------------------------------------------------------------------
### Parse cmdline arguments:

  $narg = @ARGV;
  print STDOUT "\n".'Date: '.scalar(localtime)."\n";
  if ($narg > 0) {print STDOUT "\n\n".$script.': started with cmdline = '."@ARGV\n\n\n";}
  else           {print STDOUT "\n\n".$script.': started with no cmdline arguments.'."\n\n\n";}

  for ($i=0; $i<$narg; $i++ ) {
     $arg = $ARGV[$i];
     $arg =~ s/^-{0,2}//;						#trim up to 2 leading dashes
     if (grep(/^$arg:?$/, @beamlinelist)) {
        if (defined $bl_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate beamline parameter '
                              .'in the cmdline.'.$syntax);
        }
        $bl_specified=1;
        $arg =~ s/:$//;							#trim ':'
        if ($arg ne $beamline) {
           $beamline = $arg;
           print STDOUT "\n".'!!! '.$script.': Cross-beamline calibration for '
                       .$beamline."!\n\n";
           do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
        }
     }
     elsif ($arg =~ /^noMCA$/i) {
        if (defined $om_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate noMCA in the cmdline.'
                              .$syntax, undef,'nowait');
        }
        $om_specified=1;
        $optimizeMCA = 0;
     }
     elsif ($arg =~ /^noI0$/i) {
        if (defined $oi_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate noI0 in the cmdline.'
                              .$syntax, undef,'nowait');
        }
        $oi_specified=1;
        $optimizeI0 = 0;
     }
     elsif ($arg =~ /^gotoEdge$/i || $arg =~ /^gotoEdgeSmall$/i) {
        if (defined $ge_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate gotoEdge'
                              .' in the cmdline.'.$syntax, undef,'nowait');
        }
        $ge_specified=1;
        if    ($arg =~ /^gotoEdge$/i)      {$gotoEdge = 1;}
        elsif ($arg =~ /^gotoEdgeSmall$/i) {$gotoEdge = 2;}
        elsif ($arg =~ /^gotoEdgeNone$/i)  {$gotoEdge = 0;}
     }
     elsif ($arg =~ /^aboveEdgeEV=/i) {
        if (defined $ae_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate aboveEdgeEV parameter'
                              .' in the cmdline.'.$syntax, undef,'nowait');
        }
        $ae_specified=1;
        $aboveEdgeEV = $arg;
        $aboveEdgeEV =~ s/^aboveEdgeEV=//i;
        if (! &is_decimal($aboveEdgeEV) || $aboveEdgeEV < 10 || $aboveEdgeEV > 200) {
           &zenity_error_exit ('*** '.$script.': Incorrect aboveEdgeEV in the cmdline'
                              .' -- must be a number in range 10-200 eV.'
                              .$syntax, undef,'nowait');
        }
     }
     elsif ($arg =~ /^deadTime=/i) {
        if (defined $dt_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate deadTime parameter'
                              .' in the cmdline.'.$syntax, undef,'nowait');
        }
        $dt_specified=1;
        $deadtime_goal = $arg;
        $deadtime_goal =~ s/^deadTime=//i;
        if (! &is_decimal($deadtime_goal) || $deadtime_goal <= 1 || $deadtime_goal > 50) {
           &zenity_error_exit ('*** '.$script.': Incorrect deadTime in the cmdline'
                              .' -- must be a positive number in range 0-50 percent.'
                              .$syntax, undef,'nowait');
        }
     }
     elsif ($arg =~ /^countTime=/i) {
        if (defined $ct_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate countTime parameter'
                              .' in the cmdline.'.$syntax, undef,'nowait');
        }
        $ct_specified=1;
        $cntTime = $arg;
        $cntTime =~ s/^countTime=//i;
        if (! &is_float($cntTime) || $cntTime <= 0) {
           &zenity_error_exit ('*** '.$script.': Incorrect cmdline: countTime'
                              .' must be a positive number in seconds.'
                              .$syntax, undef,'nowait');
        }
     }
     elsif ($arg =~ /^iterationsMax=/i) {
        if (defined $it_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate iterationsMax parameter'
                              .' in the cmdline.'.$syntax, undef,'nowait');
        }
        $it_specified=1;
        $itm = $arg;
        $itm =~ s/^iterationsMax=//i;
        if (! &is_int($itm) || $itm <= $iii_max || $itm > 100) {
           &zenity_error_exit ('*** '.$script.': Incorrect iterationsMax in the cmdline'
                              .' -- must be a positive number in range '.$iii_max.'-100 percent.'
                              .$syntax, undef,'nowait');
        }
        $iii_max = $itm;
     }
     elsif ($arg =~ /^i0=/i) {
        if (defined $i0_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate I0 channel in'
                              .' the cmdline.'.$syntax, undef,'nowait');
        }
        $i0_specified=1;
        $i0_name = $arg;
        $i0_name =~ s/^i0=//i;
     }
     elsif ($arg =~ /^attenuationStart=/i) {
        if (defined $at_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate attenuationStart'
                              .' parameter in the cmdline.'.$syntax, undef,'nowait');
        }
        $at_specified=1;
        $attenuation_start = $arg;
        $attenuation_start =~ s/^attenuationStart=//i;
        if (! &is_float($attenuation_start) || $attenuation_start < 50) {
           &zenity_error_exit ('*** '.$script.': Incorrect cmdline: '.$ARGV[$i]."\n"
                              .'Starting attenuation factor must be a number and'
                              .' at least 50.'.$syntax, undef,'nowait');
        }
     }
     elsif ($arg =~ /^attenuationMax=/i) {
        if (defined $am_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate attenuationMax'
                              .' parameter in the cmdline.'.$syntax, undef,'nowait');
        }
        $am_specified=1;
        $attenuation_max = $arg;
        $attenuation_max =~ s/^attenuationMax=//i;
        if (! &is_float($attenuation_max) || $attenuation_max < 100) {
           &zenity_error_exit ('*** '.$script.': Incorrect cmdline: '.$ARGV[$i]."\n"
                              .'Maximum attenuation factor must be a number and'
                              .' at least 100.'.$syntax, undef,'nowait');
        }
     }
     elsif ($arg =~ /^MCAmin=/i) {
        if (defined $mm_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate MCAmin'
                              .' parameter in the cmdline.'.$syntax, undef,'nowait');
        }
        $mm_specified=1;
        $mca_minimum = $arg;
        $mca_minimum =~ s/^MCAmin=//i;
        if (! &is_int($mca_minimum) || $mca_minimum < 100) {
           &zenity_error_exit ('*** '.$script.': Incorrect cmdline: '.$ARGV[$i]."\n"
                              .'Minimum MCA count rate must be a number and'
                              .' at least 100.'.$syntax, undef,'nowait');
        }
     }
     elsif ($arg =~ /^I0min=/i) {
        if (defined $im_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate I0min'
                              .' parameter in the cmdline.'.$syntax, undef,'nowait');
        }
        $im_specified=1;
        $I0_minimum = $arg;
        $I0_minimum =~ s/^I0min=//i;
        if (! &is_int($I0_minimum) || $I0_minimum < 1000) {
           &zenity_error_exit ('*** '.$script.': Incorrect cmdline: '.$ARGV[$i]."\n"
                              .'Minimum I0 count rate must be a number and'
                              .' at least 1000.'.$syntax, undef,'nowait');
        }
     }
     elsif ($arg =~ /^bluice/i || $arg =~ /^nobluice/i) {
        if (defined $bi_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate bluice|nobluice'
                              .' in the cmdline.'.$syntax, undef,'nowait');
        }
        $bi_specified=1;
        if    ($arg =~ /^bluice/i)   {$tell_bluice = 1;}
        elsif ($arg =~ /^nobluice/i) {$tell_bluice = 0;}
     }
     elsif ($arg =~ /^powercycle/i || $arg =~ /^nopowercycle/i) {
        if (defined $pc_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate powercycle|nopowercycle'
                               .' in the cmdline.'.$syntax, undef,'nowait');
        }
        $pc_specified=1;
        if    ($arg =~ /^powercycle/i)   {$powercycle = 1;}
        elsif ($arg =~ /^nopowercycle/i) {$powercycle = 0;}
     }
     elsif ($arg =~ /^bringPinDiode/i || $arg =~ /^bringBeamStop/i) {
        if (defined $pn_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate bringPinDiode|bringBeamStop'
                               .' in the cmdline.'.$syntax, undef,'nowait');
        }
        $pn_specified=1;
        if    ($arg =~ /^bringPinDiode/i) {$bringPinDiode = 1;}
        elsif ($arg =~ /^bringBeamStop/i) {$bringPinDiode = 0;}
     }
     elsif ($arg =~ /^gnuplot$/i || $arg =~ /^gnuplot=image$/i || $arg =~ /^nognuplot$/i) {
        if (defined $gn_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &zenity_error_exit ('*** '.$script.': Duplicate gnuplot|nonognuplot'
                              .' in the cmdline.'.$syntax, undef,'nowait');
        }
        $gn_specified=1;
        if    ($arg =~ /^gnuplot=image$/i) {$gnuplot = 2;}
        elsif ($arg =~ /^gnuplot$/i)       {$gnuplot = 1;}
        else                               {$gnuplot = 0;}
     }
     else {
        &zenity_error_exit ('*** '.$script.': Unrecognized cmdline argument: "'
                           .$ARGV[$i].'".'.$syntax, undef,'nowait');
     }
  }

  if ($tell_bluice) {$NONINTERACTIVE = 1; $ZENITY_VIA_BLUICE = 1;}	#final
  else              {$NONINTERACTIVE = 0; $ZENITY_VIA_BLUICE = 0;}	#final

  if (! $NONINTERACTIVE) {
     $msg = 'This script will optimize fluorescense detector signal'."\n"
          . 'by changing detector distance and beam attenuation.'."\n"
          . 'Continue?';
     if ($USE_ZENITY_FOR_EXIT) {
        $yesno = &zenity_question($msg,'Yes','No');
        if ($yesno !~ /^1$/) {
           print STDOUT '   ...Exiting'."\n";
           &clearAllCaMonitors();
           exit;
        }
     } else {
        &ask_to_continue($msg);
     }
  }

###------------------------------------------------------------------------
### Postprocessing of cmdline arguments and preparing to data aquisition:

### Configuration parameter -- change if beamline hardware changes:
  if (!defined $have_KBM) {$have_KBM = 1;}				#see beamline.pl
  if (!defined $have_TF)  {$have_TF  = 0;}				#see beamline.pl
  if (!defined $have_CRL) {$have_CRL = 0;}				#see beamline.pl

  if ($have_CRL || $have_TF) {
     if ($gotoEdge == 2) {
        print STDOUT '!!! '.$script. ': Going to the edge switched from small'
        .' steps to using bigEnergyMove because beamline has a CRL device'."\n";
        $gotoEdge = 1;
     }
  }
  if (!defined $aboveEdgeEV) {
     if ($have_CRL || $have_TF) {$aboveEdgeEV = 50;}
     else                       {$aboveEdgeEV = 150;}
  }
  if (!defined $attenuation_max) {
     if ($have_KBM) {$attenuation_max = 200.;}				#max attenuation
     else           {$attenuation_max = 100.;}				#max attenuation
  }
  if (!defined $attenuation_start) {
     $attenuation_start = $attenuation_max;				#starting attenuation
  }
  elsif ($attenuation_start > $attenuation_max) {
     print STDOUT '!!! '.$script. ': requested start attenuation='.$attenuation_start.'x'
     .' exceeds max attenuation='.$attenuation_max.'x -- correcting.'."\n";
     $attenuation_start = $attenuation_max;
  }

  $status = &getDarks4KeithleyList(\@signals);
  if ($status) {
     &zenity_error_exit('*** '.$script.': failed to get list of'
                       .' signals from mySQL.', undef,'nowait');
  }
  if (! grep(/^${i0_name}$/, @signals)) {
     &zenity_error_exit('*** '.$script.': incorrect signal='.$i0_name
                       .' -- must be one of '."(@{signals}).", undef,'nowait');
  }

  $gnuopen = 0;
  if ($gnuplot == 2) {
     $gnuimage = '/mnt/software/gmca_shared/logfiles/';
     if (! -e $gnuimage) {$gnuimage = './';}
     @ct = localtime();							# ($sec,$min,$hour,$mday,$mon,$year,...)
     $gnuimage .= sprintf $beamline.'_optifluo_%04d-%02d-%02d_%02d-%02d.png',($ct[5]+1900),($ct[4]+1),$ct[3],$ct[2],$ct[1];
  }

  $SRcurrentPV= 'S:SRcurrentAI';
  $fldAssyPV  = $beamline.':FLU:Di';
  $fldPosPV   = $beamline.':FLU:D';
  if (defined $USE_FALCON && $USE_FALCON) {
     $mca     = $beamline.':falcon:';
     if (!defined $mca_minimum) {$mca_minimum = 4000;}			#minimum total intensity that can be considered not noise
  } else {          
     $mca     = $beamline.':aim_adc1';
     $tca     = $beamline.':aim_tca1';
     $adc     = $beamline.':aim_adc1';
     if (!defined $mca_minimum) {$mca_minimum = 100;}			#minimum total intensity that can be considered not noise
  }
  $energyPV   = $beamline.':MO:E';
  $energyAsPV = $beamline.':MO:En';
  $biEdgePV   = $beamline.':bi:SCAN:edge';				#where BluIce posts Element-Edge
#
  $parkEnPV   = $beamline.':bi:SCAN:nonFluLastEnergy';
  $parkAtPV   = $beamline.':bi:SCAN:nonFluLastAttenu';
  $parkGaPV   = $beamline.':bi:SCAN:nonFluLastGainI0';
  $parkDiPV   = $beamline.':bi:SCAN:fluParkDistance';
#
  $postEnPV   = $beamline.':bi:SCAN:fluOptLastEnergy';
  $postAtPV   = $beamline.':bi:SCAN:fluOptLastAttenu';
  $postGaPV   = $beamline.':bi:SCAN:fluOptLastGainI0';
  $postDiPV   = $beamline.':bi:SCAN:fluOptLastDistan';
### IOC save trigger:
  $saveTriggPV = $beamline.':ioc2:saveTrigger.PROC';
#
  $scaler  = $joergerName.$scalerNo;
# $fileBkg = 'joerger'.$scalerNo.'_Bkg.dat';
  $fileBkg = '';
  if ($scaler  !~ /^${beamline}:/)    {$scaler    = $beamline.':'.$scaler;}
### scaler1=Struck, scaler3=Joerger
### If the Joerger scaler is Struck, we may need to stop Struck
  if ($scalerNo == 1) {
     $struck = $struckName;
     if ($struck !~ /^${beamline}/)   {$struck    = $beamline.':'.$struck;}
  }
  if (defined $shutter) {
     if ($shutter !~ /^${beamline}:/) {$shutter   = $beamline.':'.$shutter;}
  }
  if (defined $shutter_readback) {
     if ($shutter_readback !~ /^${beamline}:/) {
                                $shutter_readback = $beamline.':'.$shutter_readback;}
  }
### Digital IO which controls the Ketek detector power:
  if ($beamline =~ /^23i/ || $beamline =~ /^23o/) {
     $powerPV = $beamline.':pmac21:acc65e:bo12';
  }

  if (&checkFalcon2BluiceConsistency()) {
     &zenity_error_exit('*** '.$script.': PEZCA and BluIce choices about MCA mismatch', undef,'nowait');
  }

  $IDoffset    = undef;
  $syncMode    = undef;
  $syncStep    = undef;
  $switchID    = undef;
  $attenuation = undef;
  $switchMono  = undef;
###-------------------------------------------------------------------------
### Bring collimator & beamstop IN, pindiode and backlight OUT (2009/10/09):

  if ($bringPinDiode) {
     print STDOUT "\n".'Moving collimator/pindiode IN and '
                 .'beamstop/backlight OUT with sampleEnvSrv'."\n";
     @prg = ($beamline, 'collimatorIN', 'beamstopOUT', 'pindiodeIN', 'backlightOUT');
  } else {
     print STDOUT "\n".'Moving collimator/beamstop IN and '
                 .'pindiode/backlight OUT with sampleEnvSrv'."\n";
#    @prg = ('IN', $beamline, 'pindiodeOUT', 'backlightOUT');
     @prg = ($beamline, 'collimatorIN', 'beamstopIN', 'pindiodeOUT', 'backlightOUT');
  }
  $status = &setSampleEnv(\@prg);
  if ($status) {
     print STDOUT "\n".'*** '.$script.': failure='.$status
                      .' from sampleEnvSrv.'."\n\n";
     if ($tell_bluice) {&tellBluIce($scr.': sampleEnv error','error');}
     &zenity_error_exit($script.': failed on sampleEnv.', undef,'nowait');
  }

###-------------------------------------------------------------------------
  $start_time = time();

### Power up or optionally power cycle Ketek detector:
  if ($optimizeMCA && defined $powerPV && $powercycle) {
### Get power status of Ketek detector:
     ($status, $powerState) = Pezca::GetDouble($powerPV);
     if ($status) {
### In case of error we presume that Ketek power is not computer-controlled
        &report_only($script,$status,$powerPV);
        undef $powerPV;
     }
     else  {
        if ($powerState == 1) {
           if ($powercycle) {
              print STDOUT $script.': Power-cycling Ketek detector'."\n";
              if ($tell_bluice) {&tellBluIce($scr.': turning off Ketek');}
              $status = Pezca::PutDouble($powerPV,0);
              &report_only($script,$status,$powerPV);
           }
        }
        elsif ($powerState == 0) {
           print STDOUT $script.': Powering up Ketek detector'."\n";
           if ($tell_bluice) {&tellBluIce($scr.': powering up Ketek');}
           $status = Pezca::PutDouble($powerPV,1);
           &report_only($script,$status,$powerPV);
        }
     }
  }

### Measure synchrotron current:
  ($status, $SRcurrent) = Pezca::GetDouble($SRcurrentPV);
  if ($status) {
     &report_only($script,$status,$SRcurrentPV);
     $SRcurrent = 100.;
  }
  if ($SRcurrent < 20.) {
     if ($tell_bluice) {&tellBluIce($scr.': low SR current','error');}
     $msg = sprintf '*** '.$script.': low SR current=%.1f mA.'
                   .' Not expecting X-ray intensity.'."\n",$SRcurrent;
     &zenity_error_exit ($msg, undef,'nowait');
  }

### Optionally go to some eV above the absorption edge selected in BluIce:
  $E_ini = &getAxisAct($energyPV);					#record position before start

### Determine flags for bigEnergyMove
  if ($gotoEdge == 1) {
     $status = &read_jbluice_config_flags('bigEnergyMove_flags', \@bigEnergyFlags);
     if ($status) {@bigEnergyFlags = ('multiStep', 'recoveryTime=auto');}
     if (! grep(/bluice$/i,@bigEnergyFlags)) {
        if ($tell_bluice) {unshift(@bigEnergyFlags, 'bluice');}
        else              {unshift(@bigEnergyFlags, 'nobluice');}
     }
     if (! grep(/^$beamline$/,@bigEnergyFlags)) {unshift(@bigEnergyFlags, $beamline);}
     @bigEnergyFlags = grep(! /^keepShutter$/i,@bigEnergyFlags);	#remove keepShutter
  }

  if ($gotoEdge) {
     ($status,$Edge) =  Pezca::GetString($biEdgePV);
     if ($status) {
        if ($tell_bluice) {&tellBluIce($scr.': get edge failed','error');}
        &zenity_error_exit ('*** '.$script.': PVget failed on '.$biEdgePV,
                            undef,'nowait');
     }
     ($Edge_element, $Edge_KLMN) = split(/-/,$Edge);
     if ((!defined $Edge_element) || (!defined $Edge_KLMN)) {
        if ($tell_bluice) {&tellBluIce($scr.': parse edge failed','error');}
        &zenity_error_exit ('*** '.$script.': Incorrect Element-Edge data'
                           .' in the BluIce selection.', undef,'nowait');
     }
### Read energy data (in eV) from mySQL:
     $status = &getEdgeDataSQL($Edge_element, $Edge_KLMN, \$EdgeEnergy,
                               \@E_step, \@E_pm, \$FluoEnergy, \$FluoPMwidth);
     if ($status) {
        if ($tell_bluice) {&tellBluIce($scr.': no edge data in SQL','error');}
        &zenity_error_exit ('*** '.$script.': No mySQL data for element='
                           .$Edge_element.', Edge='.$Edge_KLMN, undef,'nowait');
     }
### Drive energy to some eV above the edge:
     $EdgeEnergy = $EdgeEnergy + $aboveEdgeEV;
     printf STDOUT $script.': driving energy to %.1feV (%.1feV above the edge).'."\n",$EdgeEnergy,$aboveEdgeEV;
     $EdgeEnergy = 0.001*$EdgeEnergy;					#convert into KeV
     if ($gotoEdge == 1) {
### Any energy changes are allowed:
        @prg = ('perl', '-W', $PEZCA.'/bigEnergyMove.pl', $EdgeEnergy, @bigEnergyFlags);
        $status = &execute_external_script(\@prg);
        if ($status) {
           if ($status != 512 && $status != 2) {			#'exit 1' corresponds to 256 and 'exit 2' corresponds to 512
              if ($tell_bluice) {&tellBluIce($scr.': cannot go to Edge','error');}
              &zenity_error_exit ('*** '.$script.': failed to change energy'
                                 .' with bigEnergyMove.pl'."\n"
                                 .'Error code='.$status, undef,'nowait');
           } else {
              if ($tell_bluice) {&tellBluIce($scr.': go to Edge aborted','error');}
              &clearAllCaMonitors();
              exit 1;
           }
        }
     }
     elsif ($gotoEdge == 2) {						#small move
### Only small energy changes are allowed:
        $Theta_big = 1.;						#big theta change when re-alligning may be needed
        $ESinTheta = &getESinTheta($energyAsPV);
        $Theta_ini = &E2theta($energyAsPV,$E_ini,$ESinTheta);
        $Theta_edge = &E2theta($energyAsPV,$EdgeEnergy,$ESinTheta);
        if (abs($Theta_edge-$Theta_ini) > $Theta_big) {
           if ($tell_bluice) {&tellBluIce($scr.': too far from edge','error');}
           $msg = sprintf $script.': Current energy=%.2fkeV is too far from'
                                 .' the requested edge energy=%.2fkeV.'."\n",
                                 $E_ini, $EdgeEnergy;
           &zenity_error_exit($msg."\n"
                        .'Please remove your sample and change the energy first'."\n"
                        .'The sample needs to be removed because the energy'."\n"
                        .'change may require beam optimization.', undef,'nowait');
        } else {
           if (defined $undulator) {
              if (length($undulator) > 0) {
                 $ID_ini = &getIDEnergy($undulator);
                 $IDoffset = $ID_ini - $E_ini;
                 $linkID = 1;
              } else {
                 $linkID = 0;
              }
           } else {
              $linkID = 0;
           }
           $status = &runMono($energyPV, $energyAsPV, $EdgeEnergy, 
                              $linkID, $undulator, $IDoffset,		# $gotoEdge=2 (small move)
                              $syncMode, $syncStep, $switchID,
                              $attenuation, $switchMono,
                              \$enmono13, \@harmEn, \@harmNum, 
                              \@Elower, \@Eupper, \@Olower, \@Oupper);
           if ($status) {
              if ($tell_bluice) {&tellBluIce($scr.': failed driving En','error');}
              &zenity_error_exit ('*** '.$script.': failure='.$status
                                 .' while driving energy to '.$EdgeEnergy
                                 . 'KeV.', undef,'nowait');
           }
        }
     }
  }

  $energy_act = &getAxisAct($energyPV);

### Verify that the process succeeded:
  if ($gotoEdge) {
     if (abs($energy_act - $EdgeEnergy) > 0.1) {
        if ($tell_bluice) {&tellBluIce($scr.': goto Edge failed','error');}
        &zenity_error_exit ('*** '.$script.': driving to '.$Edge
                           .' failed: we are not at the edge.', undef,'nowait');
     }
  }

### Read scaler/keithley data and respective darks for I0 from mySQL:
  print STDOUT $script.': reading '.$i0_name.' darks.'."\n";
  $status = &getDarks4KeithleyGains($i0_name,\$timestampDB,\$epochDB,
                                    \$kchannel,\$klabelDB,\$koffsetDB,
                                    \$i0,\@darks);
  if ($status) {
     if ($tell_bluice) {&tellBluIce($scr.': reading darks from mySQL failed','error');}
     &zenity_error_exit('*** '.$script.': failed to get data for '.$i0_name
                       .' from mySQL.', undef,'nowait');
  }
### Verify if I0 darks need to be re-measured and re-do the measurements if needed:
  $klabel  = &keithley_read_label($kchannel);
  $koffset = &keithley_read_offset($kchannel);
  $epoch = time();
  $bkg_age_hours = ($epoch - $epochDB)/3600.;
  $offset_delta  =  abs($koffset-$koffsetDB)
                 / &findMaxRef([abs($koffset),abs($koffsetDB),1]);

  printf STDOUT $script.': '.$i0_name.' darks are %.2f hours old.'."\n", $bkg_age_hours;
  if ($bkg_age_hours > 24 || $offset_delta > 0.01 || $klabel ne $klabelDB) {
     print STDOUT '!!! '.$script.': '.$i0_name.' conditions changed (darks old or new offset or new label).'."\n";
     print STDOUT $script.': Acquiring fresh '.$i0_name.' darks...'."\n";
     if ($tell_bluice) {&tellBluIce($scr.': acquiring '.$i0_name.' darks');}
     @prg = ('perl', '-W', $PEZCA.'/measureDark1.pl', $beamline, $i0_name, $darkmode);
     $status = &execute_external_script(\@prg);
     if ($status) {
        if ($tell_bluice) {&tellBluIce($scr.': acquiring '.$i0_name.' darks failed','error');}
        &zenity_error_exit ('*** '.$script.': Failed to acquire '.$i0_name
                           .' darks with measureDark1.pl.', undef,'nowait');
     }
     print STDOUT $script.': darks are acquired.'."\n";
     if ($tell_bluice) {&tellBluIce($scr.': '.$i0_name.' darks acquired');}
### Re-read updated I0 darks:
     $status = &getDarks4KeithleyGains($i0_name,\$timestampDB,\$epochDB,
                                       \$kchannel,\$klabelDB,\$koffsetDB,
                                       \$i0,\@darks);
     if ($status) {
        if ($tell_bluice) {&tellBluIce($scr.': re-reading darks from mySQL failed','error');}
        &zenity_error_exit('*** '.$script.': failed to get data for '.$i0_name
                          .' from mySQL.', undef,'nowait');
     }
  }

### Determine max Keithley gain where darks are not overflowing:
### (Keithley gains can very from 1E3 to 1E10, but we do not use 1E10)
  for ($i=9; $i>=3; $i--) {if ($darks[$i-3] < $I0_optimum/2.) {last;}}
  $igain_max = $i;
  $kgain_max = 10**$i;
  printf STDOUT $script.': maximum possible '.$i0_name.' gain is %1.0e.'."\n",$kgain_max;
  if ($igain_max < 6) {
     &zenity_error_exit('*** '.$script.': something is wrong with '.$i0_name
                       .' darks -- they overflow at 1E'.$igain_max.'.', undef,'nowait');
  }

### Setup scaler and MCA:
  print STDOUT $script.': Setting up Joerger scaler'."\n";
  if ($tell_bluice) {&tellBluIce($scr.': setting up Joerger');}
  &setJoerger ($scaler,$cntTime,$struck);
  if ($optimizeMCA) {
     print STDOUT $script.': Setting MCA='.$mca."\n";
     if ($tell_bluice) {&tellBluIce($scr.': setting up MCA='.$mca);}
     &setMCA_fastest($mca,$cntTime);
     print STDOUT $script.': Removing MCA='.$mca.' ROI'."\n";
     if ($tell_bluice) {&tellBluIce($scr.': removing MCA='.$mca.' ROI');}
     &setMCA_ROI ($mca, -1, -1);
     if (defined $adc) {
        print STDOUT $script.': Resetting MCA ADC='.$adc.' to the whole range'."\n";
        if ($tell_bluice) {&tellBluIce($scr.': resetting MCA ADC='.$adc);}
        $low = 5;
#       &setupMCA_ADC ($adc, 256, 0, $low);				#old: this is about 180 keV
        &setupMCA_ADC ($adc, 1024, 0, $low);				#2021.10.27: this is about 45 keV
     }
     if (defined $tca) {
        print STDOUT $script.': Removing discrimination in MCA TCA='.$tca."\n";
        if ($tell_bluice) {&tellBluIce($scr.': resetting MCA TCA='.$tca);}
        &setupMCA_TCA ($tca, $low);
     }
  }

### Open EndStation shutter:
  if ($tell_bluice) {&tellBluIce($scr.': opening hutch shutter');}
  $status = &openEndStationShutter($beamline);
  if ($status) {
     if ($tell_bluice) {&tellBluIce($scr.': shutter open failed','error');}
     &zenity_error_exit ('*** '.$script.': Failed to open endstation shutter.',
                         undef,'nowait');
  }

### Get initial Keithley gain:
  print STDOUT $script.': Getting initial Keithley gain for I0'."\n";
  $kgain_ini = &keithley_read_gain($kchannel);
  printf STDOUT $script.': Keithley channel='.$kchannel
                       .', Initial Keithley gain=%1.0e'."\n", $kgain_ini;

### Measure I0 before throwing in attenuation and raise I0 gain if needed:
### ATTENTION: since the initial attenuation is unknown (may be low) this may be
### risky for samples when I0 is measured downstream fast shutter (not our case).
  $igain = 2;
  $I0_atstart = 0;
  while ($igain < $igain_max && $I0_atstart < $I0_minimum) {
     $igain++;
     if ($igain == 5) {$igain = 6;}					#skip resonant gain
     $kgain = 10**$igain;
     $status = &keithley_set_gain($kchannel, $kgain);
     if ($status) {
        $msg = sprintf '*** '.$script.': Failed to set '.$i0_name.' gain=%1.0e', $kgain;
        &signal_handler('quit');
     }
     if (! $I0_is_upstream_fast_shutter) {
        $status = &openShutter ($shutter,$shutter_readback);		#0=OK, -1=failure
        if ($status) {
           $msg = '*** '.$script.': Failed to open fast shutter. Is beamstop or pin diode in place?';
           &signal_handler('quit');
        }
     }
     ($I0_atstart) = &runJoerger ($scaler, $cntTime, 1, ($i0));
     if (! $I0_is_upstream_fast_shutter) {&closeShutter ($shutter,$shutter_readback);}
     $dark = $darks[$igain-3];
     $I0_atstart -= $dark;						#subtract dark current
     if ($I0_atstart < 1) {
        printf STDOUT '!!! '.$script.': Negative or zero starting I0=%1.0f, at Keithley gain=%1.0e.'."\n",
                      $I0_atstart, $kgain;
        if ($dark > 0) {
           printf STDOUT '-- detected with dark=%1.0f. Try re-measuring Darks.'."\n",$dark;
        } else {
           print STDOUT '-- May be too high initial attenuation or too low Keithley gain'."\n";
        }
        $I0_atstart = 1;						#we need at least 1 pulse
     }
  }
### If raising Keithley gain did not succeed
### to bring acceptable I0:
  if ($I0_atstart < $I0_minimum) {
     if ($tell_bluice) {&tellBluIce($scr.': low I0 at start','warn');}
     $msg = sprintf '*** '.$script.': low starting I0=%1.0f. Failed'
                   .' to increase by raising Keithley gain up to %1.0e.'
                   ."\n",$I0_atstart, $kgain_max;
     &signal_handler('quit');
  }
  $kgain_ini = $kgain;							#corrected gain without attenuation
  $parkGain = $kgain_ini;

### Query limits of motion and parking distance of fluorescence detector:
  print STDOUT $script.': Getting motion limits of fluorescence detector'."\n";
  ($fldPosLo,$fldPosHi) = &getAxisDRVLH($fldPosPV);
  print STDOUT $script.': Getting parking position of fluorescence detector'."\n";
  ($status,$parkDist) = Pezca::GetDouble($parkDiPV);
  &report_exit($script,$status,$parkDiPV);
  if ($parkDist == 0.0 && $fldPosHi == 0.0 && $fldPosLo == 0.0) { #undefined?
     $msg = '*** '.$script.': Neither fluorescense detector parking'
                          .' distance nor high soft limit is defined.';
     &signal_handler('quit');
  }
  elsif ($parkDist > $fldPosHi) {
     print STDOUT '!!! '.$script.': Detector parking distance is too far. '
                 .'Setting to high soft limit.'."\n";
     $parkDist = $fldPosHi;
  }
  elsif ($parkDist < $fldPosLo || $parkDist == 0) {
     print STDOUT '!!! '.$script.': Detector parking distance is unset. '
                 .'Setting to high soft limit.'."\n";
     $parkDist = $fldPosHi;
  }
### In the operations that follow below, the soft limit will be
### used not as a drive limit, but distance optimization limit:
  $allowance = 1.0;							#1mm
  $fldPosHi = $parkDist - $allowance;					#1mm less

### Get initial position of the detector and determine if it is in parking:
  print STDOUT $script.': Getting initial fluorescence detector position'."\n";
  $fldPos_ini = &getAxisAct($fldPosPV);
  printf STDOUT $script.': initial detector position is %.2fmm'."\n",$fldPos_ini;

  if (abs($fldPos_ini - $parkDist) < 0.5*$allowance) {
     print STDOUT $script.': Detector was initially parked:'
                         .' will be storing BL parameters.'."\n";
     if ($tell_bluice) {&tellBluIce($scr.': detector parked');}
     $wasParked = 1;
  } else {
     print STDOUT $script.': Detector was NOT initially parked'."\n";
     if ($tell_bluice) {&tellBluIce($scr.': detector NOT parked');}
     $wasParked = 0;
  }

### Move fluorescence detector to the sample (to the closest distance)
  if ($optimizeMCA) {
     printf STDOUT $script.': Moving fluorescence detector to the'
                          .' closest distance=%.2fmm'."\n",$fldPosLo;
     $status = &runMotor ($fldAssyPV, $fldPosPV, $fldPosLo, 'checkAuto');
     if ($status) {
### Check if we hit hard limit:
        ($l_limit,$h_limit) = &checkMotorLimits ($fldPosPV);
        if ($l_limit == 1 && $h_limit == 0) {
           $fldAct = &getAxisAct ($fldPosPV);
           printf STDOUT '!!! '.$script.': fluorescence detector hit lower hard'
                                       .' limit at %.3fmm'."\n",$fldAct;
           printf STDOUT '!!! '.$script.': -- which is above the lower soft'
                                       .' limit of %.3fmm'."\n",$fldPosLo;
           $fldPosLo = $fldAct - $allowance;
        } else {
           $msg = sprintf '*** '.$script.': failed to drive fluorescence'
                                        .' detector to %.2fmm',$fldPosLo;
           &signal_handler('quit');
        }
     }
     $motitonRange = $fldPosHi - $fldPosLo;
     $motionStep = $motitonRange/($motionNpts+1);
  }

### Measure initial beam attenuation:
  if ($optimizeMCA || $wasParked) {
     print STDOUT $script.': Getting initial beam attenuation'."\n";
     ($status,$attenuation_ini) = &getAttenuation2($energy_act);
     if ($status) {
        if ($tell_bluice) {&tellBluIce($scr.': get attenuation failed','error');}
        $msg = '*** '.$script.': Failed to get current attenuation.';
        &signal_handler('quit');
     }
     printf STDOUT $script.': initial beam attenuation is %.1fx'."\n",$attenuation_ini;
     $parkAttn = $attenuation_ini;
  }

### Wait for Ketek detector to power up and warm up (if powered/cycled before):
  if ($optimizeMCA && defined $powerPV && $powercycle) {
     $end_time = time() - $start_time;
     if ($powerState == 1 && $powercycle) {				#initially was on, then powered down
        if ($end_time < 10) {
           printf STDOUT $script.': waiting for %ds before powering'
                        .' up Ketek detector'."\n", 10-$end_time;
           if ($tell_bluice) {&tellBluIce($scr.': waiting Ketek up');}
           sleep(10-$end_time);
        }
        print STDOUT $script.': Powering up Ketek detector'."\n";
        if ($tell_bluice) {&tellBluIce($scr.': powering up Ketek');}
        $status = Pezca::PutDouble($powerPV,1);
        &report_only($script,$status,$powerPV);
        $end_time = 0;
     }
     if ($end_time < 5 && ($powercycle || $powerState == 0)) {
        printf STDOUT $script.': waiting for %ds to warm up Ketek detector'
                     ."\n",5-$end_time;
        if ($tell_bluice) {&tellBluIce($scr.': waiting Ketek warm');}
        sleep(5-$end_time);
     }
  }

###------------------------------------------------------------------------
### Begin interactive process of finding the best attenuation / detector distance:
###
### We start with $attenuation_start, which is <= $attenuation_max and closest distance
### 1. If $mca_deadtime is in the range [0.8:1.2]*$deadtime_goal, we are OK
### 2. If $mca_deadtime > 1.2*$deadtime_goal, we increase attenuation to $attenuation_max and then start increasing $fldPos
### 3. If $mca_deadtime < 0.8*$deadtime_goal, we decrease attenuation to bring it into range.

  $success  = 0;
  $iii      = 0;
  $msg      = '';

  if ($optimizeMCA) { #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     $fldPos[0]      = $fldPosLo;
     $attenuation[0] = $attenuation_start;
     $gains[0]       = $kgain_ini;
     $mode           = 'MCA-initial';
     $forkflag       = 0;

#########
OPTIMIZE:
#########
     while (! $success) {
        if ($iii > $iii_max) {
           if ($tell_bluice) {&tellBluIce($scr.': iterations limit','error');}
           $msg .= '*** '.$script.': Iterations limit in optimizing'
                                 .' fluorescence is reached!';
           &signal_handler('quit');
        }
        if ($tell_bluice) {&tellBluIce($scr.': MCA optimization step='.$iii);}
        printf STDOUT "\n".$script.': -----------> Iteration = %d  (mode='.$mode.')'."\n",$iii;
#       if ($iii > 0) {&press_key_continue();}

### Set attenuation, detector distance, I0 keythley gain from the $iii element of arrays:
        &set_conditions();
### Measure I0, MCA signal and deadtime:
        ($I0_cts,$mca_deadtime[$iii],$mca_total[$iii]) = &measure_signal();

### 1. If $mca_deadtime is in the range [0.8:1.2]*$deadtime_goal, we are OK
        if ($mca_deadtime[$iii] >= 0.8*$deadtime_goal &&
            $mca_deadtime[$iii] <= 1.2*$deadtime_goal) {
           if ($mca_total[$iii] < $mca_minimum) {
              if ($tell_bluice) {&tellBluIce($scr.': fluor. deadtime OK at low signal.','error');}
              $msg .= '*** '.$script.': fluorescence detector deadtime is OK even at low fluor. counts.';
              &signal_handler('quit');
           }
           $mode    = 'MCA-infork';
           $success = 1;
           if ($tell_bluice) {&tellBluIce($scr.': SUCCESS (MCA deadtime in range)');}
           print STDOUT "\n".$script.': SUCCESS (MCA deadtime in range)!'."\n\n";
           last OPTIMIZE;
        }
        if ($iii == 0) {
### No-count is a special mode for MCA: when there is no useful signal 
### (e.g. when shutter is closed, no beam or too high attenuation),
### then deadtime is not low, but opposite: tends to reach 100%.
           if    ($mca_total[$iii] < $mca_minimum)                {$mode = 'MCA-nocount';}
           elsif ($mca_deadtime[$iii] > 1.2*$deadtime_goal) {$mode = 'MCA-high';}
           elsif ($mca_deadtime[$iii] < 0.8*$deadtime_goal) {$mode = 'MCA-low';}
        }
        if ($mca_deadtime[$iii] <= 0.1) {$mca_deadtime[$iii] = 0.1;}	#no zero deadtime allowed


### ------------------
### 2. In the MCA-nocount mode we decrease attenuation to bring the rate above threshold.
        if ($mode eq 'MCA-nocount') {
           if ($mca_total[$iii] > $mca_minimum) {
              if    ($mca_deadtime[$iii] > 1.2*$deadtime_goal) {$mode = 'MCA-high';}
              elsif ($mca_deadtime[$iii] < 0.8*$deadtime_goal) {$mode = 'MCA-low';}
              if ($tell_bluice) {&tellBluIce($scr.': MCA count rate made above min)');}
              printf STDOUT "\n".$script.': MCA count rate=%.0f has been made above min=%.0f,'
                           .' deadtime=%.1f%s. Switching to mode=%s'."\n\n",
                            $mca_total[$iii], $mca_minimum, $mca_deadtime[$iii], '%', $mode;
#             next OPTIMIZE;
           }
           elsif ($attenuation[$iii] > 1) {
              $fff=10;
              if ($tell_bluice) {&tellBluIce($scr.': decreasing attenuation by '.$fff.'x');}
              print STDOUT $script.':      Decreasing attenuation by factor of '.$fff."\n";
              $attenuation[$iii+1] = $attenuation[$iii] / $fff;	#decrease attenuation
              $fldPos[$iii+1]      = $fldPos[$iii];			#retain detector distance
              $gains[$iii+1]       = $gains[$iii];			#retain Keithley gain
              if ($attenuation[$iii+1] < 1) {$attenuation[$iii+1] = 1;}
              $iii++;
              next OPTIMIZE;
           }
           elsif ($fldPos[$iii] > $fldPosLo) {
              $msg = sprintf '%.2fmm',$motionStep;
              if ($tell_bluice) {&tellBluIce($scr.': decreasing fluor. detector distance by '.$msg);}
              print STDOUT $script.':      Decreasing fluorescence detector distance by '.$msg."\n";
              $attenuation[$iii+1] = $attenuation[$iii];		#retain attenuation
              $fldPos[$iii+1]      = $fldPos[$iii] - $motionStep;	#increase detector distance
              $gains[$iii+1]       = $gains[$iii];			#retain Keithley gain
              if ($fldPos[$iii+1] < $fldPosLo) {$fldPos[$iii+1] = $fldPosLo;}
              $iii++;
              next OPTIMIZE;
           }
           else {
              if ($tell_bluice) {&tellBluIce($scr.': fluor. signal low with no attenuation and min distance.','error');}
              $msg .= '*** '.$script.': fluorescence detector deadtime is low even with no attenuation and at minimal distance.';
              &signal_handler('quit');
           }                 
        }
### ------------------
### 3. In the MCA-high mode we increase attenuation until it reaches $attenuation_max
### and then if it does not help, we start increasing $fldPos
        if ($mode eq 'MCA-high') {
           if ($mca_deadtime[$iii] > 1.2*$deadtime_goal) {
              if ($mca_total[$iii] < $mca_minimum) {
                 if ($tell_bluice) {&tellBluIce($scr.': fluor. deadtime high at low signal.','error');}
                 $msg .= '*** '.$script.': fluorescence detector deadtime is high even at low fluor. counts.';
                 &signal_handler('quit');
              }
              elsif ($forkflag) {					#if forkflag is set, then we arrived here with low step, should be OK
### We have achieved  deadtime < 0.8*$deadtime_goal and returned back to
### deadtime > 1.2*$deadtime_goal. It means, we probably cannot do better:
                 $success = 1;
                 if ($tell_bluice) {&tellBluIce($scr.': SUCCESS (MCA deadtime roughly in range)');}
                 print STDOUT "\n".$script.': SUCCESS (MCA deadtime is roughly in range)!'."\n\n";
                 last OPTIMIZE;
              }
              elsif ($attenuation[$iii] < $attenuation_max) {
                 $fff=4;
                 if ($tell_bluice) {&tellBluIce($scr.': increasing attenuation by '.$fff.'x');}
                 print STDOUT $script.':      Increasing attenuation by factor of '.$fff."\n";
                 $attenuation[$iii+1] = $attenuation[$iii] * $fff;	#increase attenuation
                 $fldPos[$iii+1]      = $fldPos[$iii];			#retain detector distance
                 $gains[$iii+1]       = $gains[$iii];			#retain Keithley gain
                 if ($attenuation[$iii+1] > $attenuation_max) {$attenuation[$iii+1] = $attenuation_max;}
                 $iii++;
                 next OPTIMIZE;
              }
              elsif ($fldPos[$iii] < $fldPosHi) {
                 $msg = sprintf '%.2fmm',$motionStep;
                 if ($tell_bluice) {&tellBluIce($scr.': increasing fluor. detector distance by '.$msg);}
                 print STDOUT $script.':      Increasing fluorescence detector distance by '.$msg."\n";
                 $attenuation[$iii+1] = $attenuation[$iii];		#retain attenuation
                 $fldPos[$iii+1]      = $fldPos[$iii] + $motionStep;	#increase detector distance
                 $gains[$iii+1]       = $gains[$iii];			#retain Keithley gain
                 if ($fldPos[$iii+1] > $fldPosHi) {$fldPos[$iii+1] = $fldPosHi;}
                 $iii++;
                 next OPTIMIZE;
              }
              else {
                 if ($tell_bluice) {&tellBluIce($scr.': fluor. deadtime high at max attenuation and distance.','error');}
                 $msg .= '*** '.$script.': fluorescence detector deadtime is high even at max attenuation and distance.';
                 &signal_handler('quit');
              }                 
           }
           if ($mca_deadtime[$iii] < 0.8*$deadtime_goal) {
              if (! $forkflag) {
                 $forkflag = 1;
                 if ($tell_bluice) {&tellBluIce($scr.': deadtime is in fork');}
                 print STDOUT $script.':      Deadtime is in fork.'."\n";
              }
### First time we jumped from high deadtime to low. Could overdue attenuation
### with 10x step -- try to backup:
              if ($attenuation[$iii] > 1) {
                 $fff=2;
                 if ($tell_bluice) {&tellBluIce($scr.': decreasing attenuation by '.$fff.'x');}
                 print STDOUT $script.':      Decreasing attenuation by factor of '.$fff."\n";
                 $attenuation[$iii+1] = $attenuation[$iii] / $fff;	#decrease attenuation
                 $fldPos[$iii+1]      = $fldPos[$iii];			#retain detector distance
                 $gains[$iii+1]       = $gains[$iii];			#retain Keithley gain
                 if ($attenuation[$iii+1] < 1) {$attenuation[$iii+1] = 1;}
                 $iii++;
                 next OPTIMIZE;
              }
              elsif ($fldPos[$iii] > $fldPosLo) {
                 $msg = sprintf '%.2fmm',$motionStep;
                 if ($tell_bluice) {&tellBluIce($scr.': decreasing fluor. detector distance by '.$msg);}
                 print STDOUT $script.':      Decreasing fluorescence detector distance by '.$msg."\n";
                 $attenuation[$iii+1] = $attenuation[$iii];		#retain attenuation
                 $fldPos[$iii+1]      = $fldPos[$iii] - $motionStep;	#increase detector distance
                 $gains[$iii+1]       = $gains[$iii];			#retain Keithley gain
                 if ($fldPos[$iii+1] < $fldPosLo) {$fldPos[$iii+1] = $fldPosLo;}
                 $iii++;
                 next OPTIMIZE;
              }
              else {
                 if ($tell_bluice) {&tellBluIce($scr.': fluor. deadtime low with no attenuation and min distance.','error');}
                 $msg .= '*** '.$script.': fluorescence detector deadtime is low even with no attenuation and at minimal distance.';
                 &signal_handler('quit');
              }                 
           }
        }

### ------------------
### 4. In the MCA-low mode we decrease attenuation to bring it into range.
        if ($mode eq 'MCA-low') {
           if ($mca_deadtime[$iii] < 0.8*$deadtime_goal) {
              if ($forkflag) {						#if forkflag is set, then we arrived here with low step, should be OK
### We have achieved  deadtime > 1.2*$deadtime_goal and returned back to
### deadtime < 0.8*$deadtime_goal. It means, we probably cannot do better:
                 $success = 1;
                 if ($tell_bluice) {&tellBluIce($scr.': SUCCESS (MCA deadtime roughly in range)');}
                 print STDOUT "\n".$script.': SUCCESS (MCA deadtime is roughly in range)!'."\n\n";
                 last OPTIMIZE;
              }
              elsif ($attenuation[$iii] > 1) {
                 $fff=4;
                 if ($tell_bluice) {&tellBluIce($scr.': decreasing attenuation by '.$fff.'x');}
                 print STDOUT $script.':      Decreasing attenuation by factor of '.$fff."\n";
                 $attenuation[$iii+1] = $attenuation[$iii] / $fff;	#decrease attenuation
                 $fldPos[$iii+1]      = $fldPos[$iii];			#retain detector distance
                 $gains[$iii+1]       = $gains[$iii];			#retain Keithley gain
                 if ($attenuation[$iii+1] < 1) {$attenuation[$iii+1] = 1;}
                 $iii++;
                 next OPTIMIZE;
              }
              elsif ($fldPos[$iii] > $fldPosLo) {
                 $msg = sprintf '%.2fmm',$motionStep;
                 if ($tell_bluice) {&tellBluIce($scr.': decreasing fluor. detector distance by '.$msg);}
                 print STDOUT $script.':      Decreasing fluorescence detector distance by '.$msg."\n";
                 $attenuation[$iii+1] = $attenuation[$iii];		#retain attenuation
                 $fldPos[$iii+1]      = $fldPos[$iii] - $motionStep;	#increase detector distance
                 $gains[$iii+1]       = $gains[$iii];			#retain Keithley gain
                 if ($fldPos[$iii+1] < $fldPosLo) {$fldPos[$iii+1] = $fldPosLo;}
                 $iii++;
                 next OPTIMIZE;
              }
              else {
                 if ($tell_bluice) {&tellBluIce($scr.': fluor. deadtime low with no attenuation and min distance.','error');}
                 $msg .= '*** '.$script.': fluorescence detector deadtime is low even with no attenuation and at minimal distance.';
                 &signal_handler('quit');
              }                 
           }
           if ($mca_deadtime[$iii] > 1.2*$deadtime_goal) {
              if (! $forkflag) {
                 $forkflag = 1;
                 if ($tell_bluice) {&tellBluIce($scr.': deadtime is in fork');}
                 print STDOUT $script.':      Deadtime is in fork.'."\n";
              }
### First time we jumped from low deadtime to high. Could overdue attenuation
### with 10x step -- try to backup:
              if ($mca_total[$iii] < $mca_minimum) {
                 if ($tell_bluice) {&tellBluIce($scr.': fluor. deadtime high at low signal.','error');}
                 $msg .= '*** '.$script.': fluorescence detector deadtime is high even at low fluor. counts.';
                 &signal_handler('quit');
              }
              elsif ($attenuation[$iii] < $attenuation_max) {
                 $fff=2;
                 if ($tell_bluice) {&tellBluIce($scr.': increasing attenuation by '.$fff.'x');}
                 print STDOUT $script.':      Increasing attenuation by factor of '.$fff."\n";
                 $attenuation[$iii+1] = $attenuation[$iii] * $fff;	#increase attenuation
                 $fldPos[$iii+1]      = $fldPos[$iii];			#retain detector distance
                 $gains[$iii+1]       = $gains[$iii];			#retain Keithley gain
                 if ($attenuation[$iii+1] > $attenuation_max) {$attenuation[$iii+1] = $attenuation_max;}
                 $iii++;
                 next OPTIMIZE;
              }
              elsif ($fldPos[$iii] < $fldPosHi) {
                 $msg = sprintf '%.2fmm',$motionStep;
                 if ($tell_bluice) {&tellBluIce($scr.': increasing fluor. detector distance by '.$msg);}
                 print STDOUT $script.':      Increasing fluorescence detector distance by '.$msg."\n";
                 $attenuation[$iii+1] = $attenuation[$iii];		#retain attenuation
                 $fldPos[$iii+1]      = $fldPos[$iii] + $motionStep;	#increase detector distance
                 $gains[$iii+1]       = $gains[$iii];			#retain Keithley gain
                 if ($fldPos[$iii+1] > $fldPosHi) {$fldPos[$iii+1] = $fldPosHi;}
                 $iii++;
                 next OPTIMIZE;
              }
              else {
                 if ($tell_bluice) {&tellBluIce($scr.': fluor. deadtime high at max attenuation and distance.','error');}
                 $msg .= '*** '.$script.': fluorescence detector deadtime is high even at max attenuation and distance.';
                 &signal_handler('quit');
              }                 
           }
        }
     }									### end of OPTIMIZE loop
#########
#END OF OPTIMIZE
#########
     $attenuation_act = $attenuation[$iii];
     $distance_act    = $fldPos[$iii];					#strictly speaking this is RqsPos!

### Plot deadtime vs mca_total
     if ($gnuplot && $iii>2) {
        @gnuargs = ('-persist', '-raise');
        $gnuopen = &open_gnuplot(undef,undef,undef,\@gnuargs);
        if ($gnuopen) {
           @x = ();
           @y = ();
           for ($i=0; $i<$iii; $i++) {
              $x[$i] = $mca_total[$i];
              $y[$i] = $mca_deadtime[$i];
           }
           &sort_by_x('incr',\@x,@y);					#see common_subs.pl
           if ($gnuplot == 2) {
              print STDOUT "\n".$script.': Plotting '.$iii.' data points into '.$gnuimage."\n";
              print GNUPLOT 'set terminal png small size 800,600 enhanced'."\n";
              print GNUPLOT 'set output "'.$gnuimage.'"'."\n";
           }
           print  GNUPLOT 'set title "MCA Dead Time vs Total Counts"'."\n";
           print  GNUPLOT 'set key font ",8"'."\n";
           print  GNUPLOT 'set xlabel "Total counts (cts/s)"'."\n";
           print  GNUPLOT 'set ylabel "Deadtime (percent)"'."\n";
           print  GNUPLOT 'set nogrid'."\n";
           print  GNUPLOT 'plot "-" using 1:2 notitle with linespoints pt 7 ps 1'."\n";
           for ($i=0; $i<$iii; $i++) {
              printf GNUPLOT '%.1f,  %.1f'."\n",$x[$i],$y[$i];
              if ($i == 10*int($i/10)) {&gnuplot_filess_delay(2);}	#sleep 2 milliseconds
           }
           print GNUPLOT 'e'."\n";					#end of data
           if ($gnuplot == 2) {&close_gnuplot(); $gnuopen=0;}
        } else {
           $msg .= "\n".'*** '.$script.': failed to open pipe to GnuPlot!'."n";
#          &signal_handler('quit');
        }
     }

  } #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ end optimizing MCA


#--------------------------------------------------------------------------
### Now adjust keithley gain for I0:
  if ($optimizeI0) {
     $mode = 'keithley';
     if ($iii == 0) {							#MCA optimization was skipped
        $I0_cts = $I0_atstart;						#signal has not been measured
     } else {								#we are here after MCA optimization
### During MCA optimization we may have changed attenuation.
### Adjust Keithley gain accordingly:
        $gains[$iii+1] = $gains[$iii]
                       * int($attenuation_act/$attenuation_ini);
        if ($gains[$iii+1] < 1E-3) {$gains[$iii+1] = 1E-3;}
        $igain = &keithley_gain_index($gains[$iii+1]);			#e.g. 1E4 -> 4
        if ($igain == 5) {						#avoid resonance on pin diodes
           if ($gains[$iii+1] > $gains[$iii]) {$igain = 6;}
           else                               {$igain = 4;}
        }
        elsif ($igain > $igain_max) {
          $igain = $igain_max;
        }
        $gains[$iii+1] = 10**$igain;
        if ($gains[$iii+1] != $kgain_ini) {
           printf STDOUT "\n".$script.': Pre-adjusting Keithley gain from %1.0e'
                        .' to %1.0e according to changed attenuation'
                        ."\n",$kgain_ini,$gains[$iii+1];
           if ($tell_bluice) {&tellBluIce($scr.': preadjusting I0 gain');}
           $status = &keithley_set_gain($kchannel, $gains[$iii+1]);
           if ($status) {
              $msg = sprintf "\n".'*** '.$script.': Failed to set '.$i0_name.' gain=%1.0e'."\n", $gains[$iii+1];
              &signal_handler('quit');
           }
###        $kgain_ini = $gains[$iii+1];					#SS: commented 2016.10.31 -- not clear for what
           $attenuation[$iii+1] = $attenuation[$iii];
           $fldPos[$iii+1]      = $fldPos[$iii];
           $iii++;
### Measure I0:
           ($I0_cts,$mca_deadtime[$iii],$mca_total[$iii]) = &measure_signal();
        }
     }

     printf STDOUT $script.': Starting optimization of Keithley gain '
                  .'(initial=%1.0e)'."\n",$gains[$iii];
     if ($tell_bluice) {&tellBluIce($scr.': optimizing I0 gain');}

     printf STDOUT $script.':      I0 (optimum)=%1.0f'."\n",$I0_optimum;
     if ($I0_cts < 0.5*$I0_optimum || $I0_cts+$dark > $I0_optimum) {
        printf STDOUT $script.': I0=%1.0f is not close to optimal=%1.0f'."\n",
                      $I0_cts, $I0_optimum;
        print  STDOUT $script.': Keithley gain adjustment needed'."\n";
        $hasoverfill = 0;
        $iii_max += $iii;
        $msg = '';
#########
KEITHLEY:
#########
        while(1) {
           $iii++;
           if ($iii > $iii_max) {
              if ($tell_bluice) {&tellBluIce($scr.': iterations limit reached','error');}
              $msg .= "\n".'*** '.$script.': Iterations limit in optimizing I0 is reached!'."\n";
              &signal_handler('quit');
           }
           if ($tell_bluice) {&tellBluIce($scr.': MCA optimization step='.$iii);}
           printf STDOUT "\n".$script.': -----------> Iteration = %d  (mode='.$mode.')'."\n",$iii;
##### Too low I0:
           if ($I0_cts < 0.5*$I0_optimum) {
              if ($hasoverfill) {
### There is no sense to increase gain because the higher gain overfills Keithley:
                 if ($tell_bluice) {&tellBluIce($scr.': I0 gain success');}
                 print STDOUT $script.': Keithley gain adjustment converged.'."\n";
                 last KEITHLEY;
              }
              $gains[$iii] = $gains[$iii-1] * int($I0_optimum/&maxval($I0_cts,1));
              if ($gains[$iii] < 1E-3) {$gains[$iii] = 1E-3;}
              $igain = &keithley_gain_index($gains[$iii]);		#e.g. 1E4 -> 4
              if ($igain == 5) {$igain = 6;}				#skip resonant gain
              if (10**$igain <= $gains[$iii-1]) {
### Calculated increased gain not really increasing:
                 if ($tell_bluice) {&tellBluIce($scr.': I0 gain done');}
                 print STDOUT $script.': Calculated increased gain not really increasing.'."\n",
                              $script.': Keithley gain adjustment complete.'."\n";
                 if ($I0_cts < $I0_minimum) {
                    if ($tell_bluice) {&tellBluIce($scr.': I0 low','error');}
                    $msg = "\n".'*** '.$script.': too low I0 at maximum Keithley gain.';
                    &signal_handler('quit');
                 }
                 last KEITHLEY;
              }
              $gains[$iii] = 10**$igain;
              if ($tell_bluice) {&tellBluIce($scr.': raising I0 gain');}
              $msg = sprintf $script.': Low I0 signal: setting Keithley'
                           .' gain=%1.0e'."\n", $gains[$iii];
              print STDOUT $msg;
           }
##### Too high I0:
           elsif ($I0_cts+$dark > $I0_optimum) {
              $hasoverfill = 1;
              $gains[$iii] = $gains[$iii-1] / 10;
              $igain = &keithley_gain_index($gains[$iii]);		#e.g. 1E4 -> 4
              if ($igain == 5) {$igain = 4;}				#skip resonant gain
              $gains[$iii] = 10**$igain;
              if ($tell_bluice) {&tellBluIce('decreasing I0 gain');}
              $msg = sprintf $script.': High I0 signal: setting Keithley'
                           .' gain=%1.0e'."\n", $gains[$iii];
              print STDOUT $msg;
           }
           else {
##### I0 OK:
              if ($tell_bluice) {&tellBluIce($scr.': I0 gain adjusted.');}
              print STDOUT $script.': I0 is brought to normal range. '
                          .'Keithley gain adjustment complete.'."\n";
              last KEITHLEY;
           }
           $status = &keithley_set_gain($kchannel, $gains[$iii]);
           if ($status) {
              $msg = sprintf "\n".'*** '.$script.': Failed to set '.$i0_name.' gain=%1.0e', $gains[$iii];
              &signal_handler('quit');
           }
### Re-measure I0:
           ($I0_cts,$mca_deadtime[$iii],$mca_total[$iii]) = &measure_signal();
        }
     } else {
        if ($tell_bluice) {&tellBluIce($scr.': I0 gain is OK');}
        print STDOUT $script.': I0 is within normal range. '
                    .'Keithley gain does not need to be adjusted.'."\n";
     }
  }

### Record parked values:
  if ($wasParked) {
     print STDOUT $script.': Detector was initially parked: '
                 .'recording attenuation and Keithley gain.'."\n";
     if ($tell_bluice) {&tellBluIce($scr.': Storing park values');}
     $status = Pezca::PutDouble($parkEnPV, $energy_act);
     &report_only($script,$status,$postEnPV);
     $status = Pezca::PutDouble($parkAtPV, $parkAttn);
     &report_only($script,$status,$parkAtPV);
     $status = Pezca::PutDouble($parkGaPV, $parkGain);
     &report_only($script,$status,$parkGaPV);
  }
  else {
     print STDOUT $script.': Detector was NOT initially parked: '
                 .'NOT recording attenuation and Keithley gain.'."\n";
     if ($tell_bluice) {&tellBluIce($scr.': Skipping park values');}
  }
### Record found values:
  if ($optimizeMCA) {
     if ($tell_bluice) {&tellBluIce($scr.': Storing opt.values');}
     print STDOUT $script.': Storing optimized values for future reference...'."\n";
     $status = Pezca::PutDouble($postEnPV, $energy_act);
     &report_only($script,$status,$postEnPV);
     $status = Pezca::PutDouble($postAtPV, $attenuation_act);
     &report_only($script,$status,$postAtPV);
     $status = Pezca::PutDouble($postDiPV, $distance_act);
     &report_only($script,$status,$postDiPV);
     if ($optimizeI0) {
        $kgain = &keithley_read_gain($kchannel);
        $status = Pezca::PutDouble($postGaPV, $kgain);
        &report_only($script,$status,$postGaPV);
     }
  }
  if ($wasParked || $optimizeMCA) {
### Now force IOC to autosave the new offset:
     print STDOUT $script.': Triggering IOC2 autosave'."\n";
     $status = Pezca::PutDouble($saveTriggPV, 1);
     &report_only($script, $status, $saveTriggPV);
  }

  if ($gotoEdge) {
     printf STDOUT $script.': Restoring initial energy E_ini=%.3f KeV'."\n",$E_ini;
     if ($linkID) {printf STDOUT ' ..... and undulator ID_ini=%.3f KeV'."\n",$ID_ini;}
     if ($tell_bluice) {&tellBluIce($scr.': restoring Energy');}
     $status = 0;
     if ($gotoEdge == 1) {
        @prg = ('perl', '-W', $PEZCA.'/bigEnergyMove.pl', $E_ini, @bigEnergyFlags);
        $status = &execute_external_script(\@prg);
     }
     elsif ($gotoEdge == 2) {						#small move
        $status = &runMono($energyPV, $energyAsPV, $E_ini, 
                           $linkID, $undulator, $IDoffset,
                           $syncMode, $syncStep, $switchID,
                           $attenuation, $switchMono,
                           \$enmono13, \@harmEn, \@harmNum, 
                           \@Elower, \@Eupper, \@Olower, \@Oupper);
     }
     if ($status) {
        print STDOUT "\n".'*** '.$script.': Restoring energy failed.'."\n\n";
        if ($tell_bluice) {&tellBluIce($scr.': energy err','error');}
     }
  }

  if ($tell_bluice) {&tellBluIce($scr.': all done');}

  print STDOUT "\n".$script.': All Done!'."\n\n";
  $end_time = time();
  $msg = sprintf $script.': Elapsed time = %1.0f seconds'."\n",($end_time-$start_time);
  print STDOUT $msg."\n";
# &press_key_exit($msg);
  &clearAllCaMonitors();
  if ($gnuplot && $gnuopen) {
     $SIG{INT}  = sub {die 'Exit on kill request';};
     $SIG{QUIT} = sub {die 'Exit on kill request';};
     printf STDOUT $script.': This window will auto-close in %1.0fs.'."\n",$displaytime;
     print  STDOUT $script.': Press Ctrl+C to close now.'."\n";
     sleep($displaytime);
  } 
  exit 0;


#==========================================================================

sub set_conditions() {							#Set attenuation, detector distance, I0 keithley gain from the $iii element of arrays

  my ($gains_prev, $status);
  my ($attenuation_act, $attenuation_prev, $fldPos_prev);

  if ($iii > 0) {
     $attenuation_prev = $attenuation[$iii-1];
     $fldPos_prev      = $fldPos[$iii-1];
     $gains_prev       = $gains[$iii-1];
  } else {
     $attenuation_prev = &getAttenuation2($energy_act);
     $fldPos_prev      = &getAxisRqs ($fldPosPV);
     $gains_prev       = &keithley_read_gain($kchannel);
  }

  printf STDOUT $script.': %3d. Setting up detector conditions: attenuation=%.1fx'
                       .'  distance=%.2fmm, I0_gain=%1.0e ...'."\n",
                        $iii, $attenuation[$iii], $fldPos[$iii], $gains[$iii];

  if ($fldPos[$iii] != $fldPos_prev) {
     $status = &runMotor ($fldAssyPV, $fldPosPV, $fldPos[$iii], 'checkAuto');
     if ($status) {
        ($l_limit,$h_limit) = &checkMotorLimits ($fldPosPV);
        if ($fldPos[$iii] > $fldPos_prev && $h_limit == 1 && $l_limit == 0) {
           $fldAct = &getAxisAct ($fldPosPV);
           printf STDOUT '!!! '.$script.': fluorescence detector upper hard'
                                       .' limit is at %.3fmm'."\n", $fldAct;
           printf STDOUT '!!! '.$script.': -- which is below the upper soft'
                                       .' limit of %.3fmm'."\n", $fldPosHi;
           $fldPosHi = $fldAct;
        }
        elsif ($fldPos[$iii] < $fldPos_prev && $h_limit == 0 && $l_limit == 1) {
           $fldAct = &getAxisAct ($fldPosPV);
           printf STDOUT '!!! '.$script.': fluorescence detector lower hard'
                                       .' limit is at %.3fmm'."\n", $fldAct;
           printf STDOUT '!!! '.$script.': -- which is below the lower soft'
                                       .' limit of %.3fmm'."\n", $fldPosLo;
           $fldPosLo = $fldAct;
        }
        else {
           printf STDOUT "\n".'*** '.$script.': failed to drive fluorescence'
                                            .' detector to %.2fmm'."\n",
                                             $fldPos[$iii];
           &signal_handler('quit');
        }
        $fldPos[$iii] = &getAxisAct ($fldPosPV);
        printf STDOUT "\n".'*** '.$script.': fluorescence detector is at'
                                    .' distance=%.2f'."\n", $fldPos[$iii];
     }
  }

  if ($attenuation[$iii] != $attenuation_prev) {
     ($status,$attenuation_act) = &setAttenuation2($attenuation[$iii],$energy_act);
     if ($status) {
        printf STDOUT "\n".'*** '.$script.': Failed to set attenuation=%.1fx'
                                         ."\n", $attenuation[$iii];
        &signal_handler('quit');
     }
#    $attenuation[$iii] = $attenuation_act;				#this may screw up logic when attenuation reaches the limit
  }

  if ($gains[$iii] != $gains_prev) {
     $status = &keithley_set_gain($kchannel,$gains[$iii]);
     if ($status) {
        $msg = sprintf "\n".'*** '.$script.': Failed to set I0_gain=%1.0e', $gains[$iii];
        &signal_handler('quit');
     }
  }

  return;
}


#==========================================================================

sub measure_signal() {

  my ($mca_deadtime, $mca_total, $mca_npts, $I0_cts);

  $igain = &keithley_gain_index($gains[$iii]);				#e.g. 1E4 -> 4
  if ($mode ne 'keithley') {
#    $msg = sprintf "\n".$script.': %3d. Measuring fluorescence at'
#                              .' attenuation=%.1fx position=%.2fmm ...'."\n",
#                              $iii, $attenuation[$iii], $fldPos[$iii];
     $msg = sprintf $script.':      Measuring fluorescence at attenuation=%.1fx position=%.2fmm ...'."\n",
                                    $attenuation[$iii], $fldPos[$iii];
     print STDOUT $msg;
     &openShutter ($shutter,$shutter_readback);
     &startMCA($mca);
     &startJoerger ($scaler);
     &waitMCA($mca);
     &waitJoerger ($scaler);
     &closeShutter ($shutter,$shutter_readback);
     $mca_deadtime = &readMCA_deadTime_from_live($mca);
     ($mca_total, $mca_npts) = &readMCA_total ($mca);
#    $I0_cts = 0;
     ($I0_cts) = &readJoerger ($scaler, $cntTime, 1, ($i0));
     $dark = $darks[$igain-3];
     $I0_cts -= $dark;							#subtract dark current
     $msg2 = sprintf $script.':      MCA_deadTime=%.1f%s (target=%.1f%s)'
                          .' MCA_totalCounts=%1.0f (expected minimum=%1.0f)'
                          .' I0=%1.0f'."\n",
                           $mca_deadtime, '%', $deadtime_goal, '%',
                           $mca_total, $mca_minimum, $I0_cts;
     print STDOUT $msg2;
  } else {
#    $msg = sprintf "\n".$script.': %3d. Measuring I0 at Keithley gain=%1.0e...'."\n",
#                               $iii, $gains[$iii];
     $msg = sprintf $script.':      Measuring I0 at Keithley gain=%1.0e...'."\n", 
                                    $gains[$iii];
     print STDOUT $msg;
     if (! $I0_is_upstream_fast_shutter) {&openShutter ($shutter,$shutter_readback);}
     ($I0_cts) = &runJoerger ($scaler, $cntTime, 1, ($i0));
     if (! $I0_is_upstream_fast_shutter) {&closeShutter ($shutter,$shutter_readback);}
     $dark = $darks[$igain-3];
     $I0_cts -= $dark;							#subtract dark current
     $mca_deadtime = 0;
     $mca_total = 0;
     $msg2 = sprintf $script.':      I0=%1.0f (expected minimum=%1.0f,'
                            .' optimum=%1.0f)'."\n",
                             $I0_cts, $I0_minimum, $I0_optimum;
     print STDOUT $msg2;
  }
  $msg .= $msg2;
  if ($I0_cts < 1 && $dark > 0) {
     printf STDOUT "\n".'*** '.$script.': too high '.$i0_name.' dark = %1.0f at'
                  .' gain=%1.0e. Darks may need to be re-measured.'
                  .' Setting '.$i0_name.'=1 for now.'."\n", $dark,$gains[$iii];
     $I0_cts = 1;
  }
  return $I0_cts, $mca_deadtime, $mca_total;
}


#==========================================================================

#sub signal_handler([$signal]) {
sub signal_handler(;$) {
  our ($ABORT_IN_PROCESS);
  my ($sig, $attenuation_act);
  $sig = shift;

  if ($sig ne 'quit') {print STDOUT "\n".'*** '.$script.': Caught a SIG'.$sig."\n\n";}
  else                {print STDOUT "\n".'!!! '.$script.': Cleaning up before quit'."\n\n";}

  if (defined $ABORT_IN_PROCESS) {return;}
  $ABORT_IN_PROCESS = 1;
  &abort_external_script();

  if (defined $shutter) {
    if ($shutter !~ /^${beamline}/) {$shutter = $beamline.':'.$shutter;}
    print STDOUT $script.': Closing shutter ['.$shutter."]\n";
    &closeShutter ($shutter,$shutter_readback);
  }

  if (defined $fldAssyPV && defined $fldPosPV) {
     print STDOUT $script.': Stopping motor '.$fldPosPV."\n";
     &abortAssy ($fldAssyPV);
     if (defined $fldPos_ini) {
        printf STDOUT $script.': returning fluorescence detector to %.3f'."\n",$fldPos_ini;
        &startNMotors ($fldAssyPV, 1, [$fldPosPV], [$fldPos_ini], 'checkAuto');
     }
  }

### Restore Ketek power state:
  if (defined $powerState) {
     if ($powerState == 0) {
        print STDOUT $script.': Powering down Ketek (restoring power-off)'."\n";
        $status = Pezca::PutDouble($powerPV,$powerState);
     }
  }

### Restore Keithley gain
  if (defined $kgain_ini) {
     printf STDOUT $script.': Restoring initial Keithley gain=%1.0e for '.$i0_name."\n",$kgain_ini;
  }

### Restore attenuation:
  if (defined $attenuation_ini && defined $energy_act) {
     printf STDOUT $script.': Restoring attenuation to %.1fx'."\n",$attenuation_ini;
     ($status,$attenuation_act) = &setAttenuation2($attenuation_ini,$energy_act);
     if ($status) {print STDOUT "\n".'*** '.$script.': Restoring attenuation failed.'."\n\n";}
  }

### Restore energy:
  if (defined $gotoEdge && defined $E_ini) {
     if ($gotoEdge == 1) {
        if (@bigEnergyFlags == 0) {@bigEnergyFlags = ('multiStep', 'recoveryTime=auto');}
        @prg = ('perl', '-W', $PEZCA.'/bigEnergyMove.pl', $E_ini, @bigEnergyFlags);
        $status = &execute_external_script(\@prg);
     }
     elsif ($gotoEdge == 2) {
        printf STDOUT $script.': Restoring initial energy E_ini=%.3f KeV'."\n",$E_ini;
        if ($linkID) {printf STDOUT ' ..... and undulator ID_ini=%.3f KeV'."\n",$ID_ini;}
        if ($sig ne 'quit') {
        if ($linkID) {&startID($undulator,$E_ini+$IDoffset);}
           $status = &startMotor ($energyPV,$E_ini,$energyAsPV,'checkAuto');
        } else {
           $status = &runMono($energyPV, $energyAsPV, $E_ini, 
                              $linkID, $undulator, $IDoffset,
                              $syncMode, $syncStep, $switchID,
                              $attenuation, $switchMono,
                              \$enmono13, \@harmEn, \@harmNum, 
                              \@Elower, \@Eupper, \@Olower, \@Oupper);
        }
        if ($status) {print STDOUT "\n".'*** '.$script.': Restoring energy failed.'."\n\n";}
     }
  }

  print STDOUT $script.': Clearing all CA monitors'."\n";
  &clearAllCaMonitors();
  print STDOUT $script.': Done with exiting!'."\n";
  if ($sig eq 'quit' && defined $msg) {&zenity_error($msg, undef,'nowait');}
  &clearAllCaMonitors();
  if ($gnuplot && $gnuopen) {
     $SIG{INT}  = sub {die 'Exit on kill request';};
     $SIG{QUIT} = sub {die 'Exit on kill request';};
     printf STDOUT $script.': This window will auto-close in %1.0fs.'."\n",$displaytime;
     print  STDOUT $script.': Press Ctrl+C to close now.'."\n";
     sleep($displaytime);
  } 
  exit 1;
}
