#!/usr/bin/perl -W
# _CALLED_FROM_BLUICE_
#--------------------------------------------------------------
# This script moves fluorescence detector back to parking position
# and restores original beam attenuation and keithley gain
#
#     		Author: Sergey Stepanov
#
# Version-1.0:  2009/05/22 initial release using Pezca
# Version-1.1:  2009/07/10 changed order: finish tracking all the way back before removing attenuation
# Version-1.2:  2020/11/18 @bigEnergyFlags
#--------------------------------------------------------------
  use strict;
  use Pezca;
  use DBI;				#mySQL DB interface (for attenuation control)

  our ($beamline, $shutter);
  our ($NONINTERACTIVE, $USE_ZENITY_FOR_EXIT, $ATTENUATION_USE_FACTOR);

  my ($PEZCA, $script, $syntax, $cmd, $msg, $tell_bluice);
  my ($narg, $arg, @processes, $nprocesses, @prg);
  my (@beamlinelist, $bl_specified, $re_specified);
  my ($mcaAssyPV, $mcaPosPV, $mcaPosLo, $mcaPosHi);
  my ($attenuation_act, $kchannel, $keithley_label_i0);
  my ($status, $start_time, $end_time, $motionTime, $i);
  my ($energyPV, $restoreEnergy, $energyLo, $energyHi);
  my ($parkDiPV, $parkAtPV, $parkGaPV, $parkEnPV);
  my ($parkDist, $parkAttn, $parkGain, $parkEnergy, @bigEnergyFlags);


  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  do $PEZCA.'/Common/motor_subs.pl';
  do $PEZCA.'/Common/shutter_subs.pl';
  do $PEZCA.'/Common/attenuation_subs_v2.pl';
  do $PEZCA.'/Common/keithley_subs.pl';
  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  select STDOUT; $|=1;			# set unbuffered output
  $SIG{INT}  = 'signal_handler';
  $SIG{QUIT} = 'signal_handler';

###------------------------------------------------------------------------
### Defaults (if no cmdline arguments are given):
### These can be overwritten with cmdline options:
  $keithley_label_i0 = 'I0';		#keithley amplifier label for i0
  $tell_bluice       = 1;		#1=inform BluIce 0=don't
  $restoreEnergy     = 0;		#1=restore energy recorded at unparking time
### end of Defaults (if no cmdline arguments are given)
###------------------------------------------------------------------------
  @beamlinelist = ('23i', '23o', '23b', '23d');
  $script  = &makeScriptName();
  $syntax = "\n\n".'Syntax: '.$script .' [beamline] [restoreEnergy|keepEnergy]';

  if ($tell_bluice) {$NONINTERACTIVE = 1; $USE_ZENITY_FOR_EXIT=1;}
  else              {$NONINTERACTIVE = 0; $USE_ZENITY_FOR_EXIT=0;}
#----------------------------------------------------------------
### Protection from multiple instances:
  if ($^O !~ /Win/i) {
#    $cmd = '/bin/ps -ef'
     $cmd = '/bin/ps -eo user:19,pid,cmd'
          .' | /bin/grep -iv "grep\|edit\|term\|onsole\|bash\|csh\|pezca23id"'
          .' | /bin/grep parkFluoDetector';
     $msg = `${cmd}`;
     @processes = split(/\n/,$msg);
     $nprocesses = @processes;
     print STDOUT 'nprocesses = '.$nprocesses."\n";
     if ($nprocesses > 1) {
        $msg =  '*** '.$script.': '.$nprocesses.' instance(s) of script are running!'
                      ."\n".$msg;
        &zenity_error_exit ($msg);
     }
  }
#----------------------------------------------------------------

  $narg = @ARGV;
  print STDOUT "\n".'Date: '.scalar(localtime)."\n";
  if ($narg > 0) {print STDOUT $script.': started with cmdline = '."@ARGV\n\n";}
  else           {print STDOUT $script.': started with no cmdline arguments.'."\n\n";}

  $ATTENUATION_USE_FACTOR = 1;		#USE 'x' only!!!

###------------------------------------------------------------------------
### Parse cmdline arguments:

  for ($i=0; $i<$narg; $i++ ) {
     $arg = lc $ARGV[$i];
     if (grep(/^$arg$/, @beamlinelist)) {
        if (defined $bl_specified) {
           &zenity_error_exit ('***'.$script.': Duplicate beamline parameter in the cmdline.'.$syntax);
        }
        $bl_specified=1;
        if ($arg ne $beamline) {
           $beamline = $arg;
           print STDOUT "\n".'!!! '.$script.': Operation for beamline='.$beamline."\n\n";
           do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
        }
     }
     elsif ($arg =~ /^-{0,2}restoreEnergy$/i || $arg =~ /^-{0,2}keepEnergy$/i) {
        if (defined $re_specified) {
           &zenity_error_exit ('*** '.$script.': Duplicate restoreEnergy|keepEnergy flag in the cmdline.'.$syntax);
        }
        $re_specified = 1;
        if    ($arg =~ /^-{0,2}restoreEnergy$/i) {$restoreEnergy = 1;}
        elsif ($arg =~ /^-{0,2}keepEnergy$/i)    {$restoreEnergy = 0;}
     }
     else {
        &zenity_error_exit ('*** '.$script.': Unrecognized cmdline argument: '.$ARGV[$i].$syntax);
     }
  }

###------------------------------------------------------------------------
### Postprocessing of cmdline arguments and preparing to data aquisition:

  $start_time = time();
#
  $mcaAssyPV  = $beamline.':FLU:Di';
  $mcaPosPV   = $beamline.':FLU:D';
  $energyPV   = $beamline.':MO:E';
#
  $parkEnPV   = $beamline.':bi:SCAN:nonFluLastEnergy';
  $parkAtPV   = $beamline.':bi:SCAN:nonFluLastAttenu';
  $parkGaPV   = $beamline.':bi:SCAN:nonFluLastGainI0';
  $parkDiPV   = $beamline.':bi:SCAN:fluParkDistance';
#
  if (defined $shutter) {
     if ($shutter !~ /^${beamline}/) {$shutter = $beamline.':'.$shutter;}
  }

### Read parking data:
  print STDOUT 'Reading parking distance for fluorescent detector from '.$parkDiPV."\n";
  ($status,$parkDist) = Pezca::GetDouble($parkDiPV);
  if ($status) {
     $msg = '*** '.$script.': failed to read parking distance PV='.$parkDiPV;
     if ($tell_bluice) {&tellBluIce($msg,'error');}
     &report_exit($script,$status,$parkDiPV);
  }
  print STDOUT 'Reading beam attenuation for the detector parked mode from '.$parkAtPV."\n";
  ($status,$parkAttn) = Pezca::GetDouble($parkAtPV);
  if ($status) {
     $msg = '*** '.$script.': failed to read stored attenuation PV='.$parkAtPV;
     if ($tell_bluice) {&tellBluIce($msg,'error');}
     &report_exit($script,$status,$parkAtPV);
  }
  print STDOUT 'Reading I0 gain for the detector parked mode from '.$parkGaPV."\n";
  ($status,$parkGain) = Pezca::GetDouble($parkGaPV);
  if ($status) {
     $msg = '*** '.$script.': failed to read stored I0 gain PV='.$parkGaPV;
     if ($tell_bluice) {&tellBluIce($msg,'error');}
     &report_exit($script,$status,$parkGaPV);
  }

### Optionally restore energy to the state before detector was unparked:
  if ($restoreEnergy) {
     print STDOUT 'Reading X-ray energy of the time when the detector was parked from '.$parkEnPV."\n";
     ($status,$parkEnergy) = Pezca::GetDouble($parkEnPV);
     if ($status) {
        $msg = '*** '.$script.': failed to read stored energy PV='.$parkEnPV;
        if ($tell_bluice) {&tellBluIce($msg,'error');}
        &report_exit($script,$status,$parkEnPV);
     }
     print STDOUT 'Reading X-ray energy limits from EPICS'."\n";
     ($energyLo,$energyHi) = &getAxisDRVLH($energyPV);
     if (($parkEnergy < $energyLo || $parkEnergy > $energyHi) &&
                   ($energyLo > 0 || $energyHi > 0)) {
        $msg = sprintf '*** '.$script.': cannot restore Mono Energy because'
                             .' ParkEnergy=%.3f not within mono motion limits',
                             $parkEnergy;
        if ($tell_bluice) {&tellBluIce($msg,'error');}
        &zenity_error_exit ($msg);
     }
### Find flags for bigEnergyMove
     $status = &read_jbluice_config_flags('bigEnergyMove_flags', \@bigEnergyFlags);
     if ($status) {@bigEnergyFlags = ('multiStep', 'recoveryTime=auto');}
     if (! grep(/bluice$/i,@bigEnergyFlags)) {
        if ($tell_bluice) {unshift(@bigEnergyFlags, 'bluice');}
        else              {unshift(@bigEnergyFlags, 'nobluice');}
     }
     if (! grep(/^$beamline$/,@bigEnergyFlags)) {unshift(@bigEnergyFlags, $beamline);}
     @bigEnergyFlags = grep(! /^keepShutter$/i,@bigEnergyFlags);	#remove keepShutter
### Execute bigEnergyMove
     @prg = ('perl', '-W', $PEZCA.'/bigEnergyMove.pl', $parkEnergy, @bigEnergyFlags);
     $status = &execute_external_script(\@prg);
     if ($status) {
        $msg = '*** '.$script.': failed to start bigEnergyMove.pl for driving energy';
        if ($tell_bluice) {&tellBluIce($msg,'error');}
        &zenity_error_exit ($msg);
     }
  }

### Query motion limits of fluorescence detector:
  print STDOUT 'Getting motion limits of fluorescent detector from EPICS'."\n";
  ($mcaPosLo,$mcaPosHi) = &getAxisDRVLH($mcaPosPV);
  if ($parkDist == 0.0 && $mcaPosHi == 0.0 && $mcaPosLo == 0.0) { #undefined?
     &zenity_error_exit ('*** '.$script.': Neither fluorescense detector parking distance nor high drive limit is defined');
  }
  elsif ($parkDist > $mcaPosHi) {
     print STDOUT '!!! Detector parking distance is beyond the motion limit -- adjusting!'."\n";
     $parkDist = $mcaPosHi;
  }

### Starting detector motion to the parking position:
  print STDOUT 'Starting moving detector to the parked position'."\n";
  $motionTime = &calcMotorTime($mcaAssyPV, $mcaPosPV, $parkDist);
  $status = &startMotor ($mcaPosPV, $parkDist, $mcaAssyPV, 'checkAuto');
  if ($status) {
     $msg = '*** '.$script.': failed to start fluor.detector motion to parking position';
     if ($tell_bluice) {&tellBluIce($msg,'error');}
     &zenity_error_exit ($msg);
  }

### Tracking detector motion to the parking position:
  print STDOUT 'Tracking detector motion to the parked position'."\n";
  $status = &trackMotor ($mcaAssyPV, $motionTime, $mcaPosPV);
  if ($status) {
     $msg = '*** '.$script.': Tracking of detector motion to the park'
                          .' position received error='.$status;
     if ($tell_bluice) {&tellBluIce($msg,'error');}
     &zenity_error_exit ($msg);
  }

### Restoring beam attenuation:
  print STDOUT 'Setting beam attenuation for the detector parked mode'."\n";
  ($status,$attenuation_act) = &setAttenuation1($parkAttn);
  if ($status) {
     $msg = '*** '.$script.': Failed to set beam attenuation.';
     if ($tell_bluice) {&tellBluIce($msg,'error');}
     &zenity_error_exit ($msg);
  }

### Restoring Keithley gain on I0:
  print STDOUT 'Setting I0 Keithley gain for the detector parked mode (label='.$keithley_label_i0.').'."\n";
  $kchannel = &keithley_find_amp($keithley_label_i0);
  if ($kchannel == 0) {
     $msg = '*** '.$script.': Failed to find Keithley Amp with label='.$keithley_label_i0;
     if ($tell_bluice) {&tellBluIce($msg,'error');}
     &zenity_error_exit ($msg);
  }
  print STDOUT 'Keithley channel = '.$kchannel."\n";
  &keithley_set_gain($kchannel, $parkGain);

  print STDOUT "\n".'All Done!'."\n\n";
  $end_time = time();
  printf STDOUT 'Elapsed time = %2.0f seconds'."\n",($end_time-$start_time);
  &press_key_exit();
# &clearAllCaMonitors();
# exit 0;


#=================================================================

sub signal_handler {
  our ($ABORT_IN_PROCESS);
  my $sig = shift;
  print STDOUT "\n".' *** Caught a SIG'.$sig."\n\n";

  if (defined $ABORT_IN_PROCESS) {return;}
  $ABORT_IN_PROCESS = 1;
  &abort_external_script();

  if (defined $shutter) {
    if (!($shutter =~ /^${beamline}/)) {$shutter = $beamline.':'.$shutter;}
    print STDOUT 'Closing shutter ['.$shutter.']'."\n";
    &closeShutter ($shutter);
  }

  if (defined $mcaAssyPV && defined $mcaPosPV) {
     print STDOUT 'Stopping motor '.$mcaPosPV."\n";
     &abortAssy ($mcaAssyPV);
  }

  print STDOUT 'Clearing all CA monitors'."\n";
  &clearAllCaMonitors();
  print STDOUT 'Done with exiting!'."\n";
  exit -1;
}
