#!/usr/bin/perl -W
# _USED_BY_BLUICE_
#--------------------------------------------------------------
# This script is a "state-notation-like" server handling motion
# of Gonio head for JBluIce. The server receives three parameters
# via an auxiliary string PV:
#
# [1] cameraNr (0=lowres 1=hires)
# [2] required x in pixels
# [3] required y in pixels
#
# It also reads current camera zoom and gonio omega that are
# involved into calculating pixels->mm and geometry transformations
# respectively.
#
# In case of problems the script aborts motion and sends
# error to JBluIce
#
#     		     Author: Sergey Stepanov
#
# Version 1.0 /2006.05.10/: initial implementation.
# Version 1.1 /2007.01.15/: added "checkAuto" to starting gonio support.
#                           added list of allowed hosts.
# Version 2.0 /2007.02.06/: added optional monitoring of automounter "READY" status.
# Version 2.1 /2007.06.11/: added 3-click centering
# Version 2.2 /2007.06.26/: added posting pixels2mm into the click string description
# Version 3.0 /2007.06.26/: changed to read px2mm from EPICS calc string instead of
#                           internal calculation;
#                           removed posting pixels2mm into the click string description
#                           since it is no longer needed.
# Version 4.0 /2008.01.18/: added support for IQeye video camera (different aspect ratio).
#                           Attn: no support for digital zoom yet. To support digital
#                           zoom, we need to add zoom factor to click string and change
#                           the click string PV fron string to waveform (longer space)
# Version 4.1 /2008.10.13/: increased a bit the estimated motion in order to prevent timouts.
# Version 5.0 /2009.03.23/: rewrote 3-click mode to "real-3-click".
# Version 5.1 /2009.06.15/: changed click monitor from NewMonitorValueString to non-empty string
# Version 5.2 /2009.08.17/: added date stamp to the printouts
# Version 5.3 /2009.08.26/: moved clearing msg string up to avoid delays
# Version 5.4 /2009.10.01/: minor code cleanup (mostly quotes)
# Version 5.5 /2009.10.01/: fixed clearing string if robot is paused
# Version 5.6 /2010.03.04/: added skipping clicks if frame being acquired
# Version 5.7 /2010.07.12/: changed 3-click centering for low-res: now adding -90 to phi.
# Version 5.8 /2010.12.08/: minor fixes in messages sent to BluIce log
# Version 5.9 /2010.12.15/: comment on 3-click added; one tiny bug in 3-click checks fixed.
# Version 6.0 /2012.03.30/: clicks undo/redo added
# Version 7.0 /2012.10.27/: added SONICC mode: (camera=2, dx_um, dy_um, phiOffset)
# Version 7.1 /2012.11.14/: added SONICC re-click (camera=2R, dx_um, dy_um, phiOffset)
# Version 7.2 /2013.01.11/: added SONICC upstream-downstream shift (camera=2Z, dz_um)
#                           added command-line help
# Version 8.0 /2013.05.03/: added global switch to XYZ instead of XY+SH
# Version-8.1 /2014.03.28/: changed functions calls to arrays pointers (preparation to Perl-5.16)
# Version-8.2 /2016.03.28/: added absolute motion (camera=ABS x_um, y_um, z_ym, omega) -- mainly for SONICC
# Version-8.3 /2018.03.14/: switched to cleaning clickPV only after the operation is finished.
#--------------------------------------------------------------
# use Math::Trig;		# needed for sin/cos on XY positioner calcs
  use Time::HiRes;              # needed for short sleeps
# use List::Util qw(min max);
  use Pezca;
# use warnings::unused;
  use strict;

  our ($NONINTERACTIVE, $IGNORE_ERRORS, $report, $nonstop);
  our ($beamline, $robot, $HAVE_XYZ);
  my ($PEZCA, $script, $syntax, @beamlinelist, $watchrobot, $narg, $arg);
  my ($host, @hosts_allowed, $instances, @processes, $nprocesses, $i);
  my ($clickPV, $p2mLoVPV, $p2mLoHPV, $p2mHiVPV, $p2mHiHPV, $phiOffset);
  my ($assyXY, @axesXY, $assySH, $axisSH, $omegaAxisPV, $omegaAssyPV);
  my ($robotStatusPV, $robotPausePV, $robotStatus, $robotOK, $robotPause);
  my ($cameraOrientation, $verticalOrientation, $horizontalOrientation);
  my ($loResOrientation, $hiResOrientation, $otherOrientation);
  my ($omega, $phi, $angle, $nwords, $rotation);
  my ($status, $statusAll, $motionTimeXY, $motionTimeSH, $motionTimeMax);
  my ($degree2rad, $scale, $calib_Height, $sleep_interval, $stop_interval);
  my ($elapsed, $connection_lost, $click, $camera, $runprg, $filter);
  my ($clickType, $xpixel, $ypixel, $dxpixel, $dypixel, $dxmm, $dymm);
  my ($zpixel, $dzmm, $alongBeam, $naxis);
  my ($p2mVPV, $p2mHPV, $factorV, $factorH);
  my ($imageWidth, $imageWidthMin, $imageHeight, $imageHeightMin);
  my ($dSampleX,   $dSampleY,   $dSampleH);
  my ($ActSampleX, $ActSampleY, $ActSampleH);
  my ($NewSampleX, $NewSampleY, $NewSampleH, $newOmega, $dOmega);
  my (@cameraLabel, $t0, $clickString, $clickStrBak, @cmd);
  my ($clickInd, @camera3C, @scale3C, @factorH3C, @factorV3C, @phiOffset3C);
  my (@dxmm3C, @dymm3C, @phi3C, $det, $rc, $rs, $radius, $phi0, $y0, $pi2);
  my ($dx1, $dx2, $dx3, $dy1, $dy2, $dy3, $ph1, $ph2, $ph3, $stamp);
  my ($fpe_startPV, $fpe_operstatusPV, $fpe_start, $fpe_operstatus);
  my (@undostack, @redostack, $stackmax, @rqs, $nelem, $undo, $crossrun);
  my ($reclick, $stacklength, @acts);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;
  do $PEZCA.'/Common/motor_subs.pl';
  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  select STDOUT; $|=1;			# set unbuffered output

  $NONINTERACTIVE = 1;	 		# to exit from script with no prompting
  $IGNORE_ERRORS  = 1;			# ignore EZCA errors
  $report  = 0;				# set to 0 if no printouts are needed
  $nonstop = 2;				# do not abort!

  $script = &makeScriptName();
  $script =~ s/\.pl$//i;
  $syntax = "\n\n".'Syntax: '.$script.' [beamline] [help] [watchrobot|norobot]'."\n";
  @beamlinelist = ('23i:', '23o:', '23b:', '23d:');
  $watchrobot   = 1;			# default robot communication if no command line
  $clickPV      = ':bi:pointclick';
  $narg = @ARGV;
  for ($i=0; $i<$narg; $i++) {
     $arg = $ARGV[$i];
     $arg =~ s/^-{0,2}//;					#trim leading dashes
     if (grep(/^$arg:?$/, @beamlinelist)) {
        $arg =~ s/:$//;
        if ($beamline ne $arg) {
           $beamline = $arg;
           print STDOUT "\n".'!!! '.$script.': Cross operation for beamline='.$beamline."\n";
           print STDOUT '!!! This should never be used for other than debugging!'."\n";
           $crossrun = 1;
           do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
        }
     }
     elsif ($arg =~ /^watchrobot$/i) {$watchrobot = 1;}
     elsif ($arg =~ /^norobot$/i)    {$watchrobot = 0;}
     elsif ($arg =~ /^help$/i)       {&click_syntax_help();}
     else {
        $NONINTERACTIVE = 0;
        &press_key_exit("\n".'!!! '.$script.': Incorrect command line argument "'.$ARGV[$i].'"'.$syntax);
        exit 1;
     }
  }

### Protection from running on multiple computers:
  if (!defined $crossrun || $crossrun == 0) {
     $host = $ENV{HOSTNAME};
     if (!defined $host) {$host=`hostname`; $host=~s/\n//g;}
     if (!defined $host) {print STDOUT '!!! HOSTNAME is undefined!'."\n\n"; $host='unknown';}
     $host = lc($host);
     $host =~ s/.gmca.aps.anl.gov$//;
     if    ($beamline eq '23i') {@hosts_allowed = ('bl1keithley', 'keithley1');}
     elsif ($beamline eq '23o') {@hosts_allowed = ('bl2keithley', 'keithley2');}
     elsif ($beamline eq '23b') {@hosts_allowed = ('bl3keithley', 'keithley3');}
     elsif ($beamline eq '23d') {@hosts_allowed = ('www', 'px0', 'sergey');}
     else                       {@hosts_allowed = 'undefined';}

     if (! grep(/^$host$/, @hosts_allowed)) {
        print STDOUT "\n".'*** '.$script.': cannot run on '.$host."\n";
        print STDOUT "\n".'*** '.$script.': host(s) allowed for ',
                    'beamline='.$beamline.": \'@{hosts_allowed}\'.\n\n";
        $NONINTERACTIVE = 0;
        &press_key_exit();
        exit 1;
     }
  }

### Protection from multiple instances:
  if ($^O !~ /Win/i) {
#    $filter = '/bin/ps -ef | '
     $filter = '/bin/ps -eo user:19,pid,cmd | '
             . '/bin/grep -iv "grep\|edit\|term\|onsole\|bash\|csh" | '
             . '/bin/grep pointclickserver';
     $instances  = `$filter`;
     @processes  = split(/\n/,$instances);
     $nprocesses = @processes;
     $nprocesses--;					# ignore this script
     if ($nprocesses > 0) {
        print STDOUT "\n".'*** '.$script.': '.$nprocesses.' instance(s) of '
                   .'pointclickserver already running!'."\n\n".$instances."\n\n";;
        $NONINTERACTIVE = 0;
#       &press_key_exit();
        exit;
     }
  }

  $clickPV  = $beamline.$clickPV;

### See: Tools -> BluIce Configuration -> Video px2mm calibrations.
### The calibrations are calculated by a calc record using the
### interpolation formula:
###
### factor = 0.001*[A + exp[-(K-C)/D]],  where K is zoom motor position & the
###                                      rest are the camera and V/H specific
###                                      constants.
###
  $p2mLoVPV = $beamline.':bi:px2mm_v_lo';
  $p2mLoHPV = $beamline.':bi:px2mm_h_lo';
  $p2mHiVPV = $beamline.':bi:px2mm_v_hi';
  $p2mHiHPV = $beamline.':bi:px2mm_h_hi';

  $assyXY   = $beamline.':GO:Cr';
  @axesXY   = ($beamline.':GO:CX', $beamline.':GO:CY');

  if (defined $HAVE_XYZ) {
     $axisSH = $beamline.':GO:CZ';
     push (@axesXY, $axisSH);
  } else {
     if ($beamline ne '23b') {$assySH = $beamline.':GO:St';}
     else                    {$assySH = $beamline.':GO:StB';}
     $axisSH = $beamline.':GO:SH';
  }
  $naxis = @axesXY;

  $omegaAxisPV  = $beamline.':GO:OM';
  $omegaAssyPV  = $beamline.':GO:Om';

  $fpe_startPV      = $beamline.':FPE:start';
  $fpe_operstatusPV = $beamline.':FPE:operationStatus';

  if ($robot !~ /^${beamline}/) {
     $robot = $beamline.':'.$robot;
  }
  $robotStatusPV  = $robot.'sta';			# this is string-type
  $robotPausePV   = $robot.'pause';			# 0=ready, 1=pause
  $robotOK        = 'SMR: ready';			# preferred status

  $verticalOrientation   = 'Vertical';			# constant
  $horizontalOrientation = 'Horizontal';		# constant
  $otherOrientation      = 'Other';			# constant
  $loResOrientation = $verticalOrientation;		# lo-camera orientation
  $hiResOrientation = $horizontalOrientation;		# hi-camera orientation

  print STDOUT 'This is point-and-click server for '.$assyXY."\n";

  $pi2 = 2.0 * atan2(1.,1.);			# atan2(1)=pi/4; pi/2=2*atan2(1)
  $degree2rad = $pi2 / 90.0;			# (pi/2)/90.

  $scale = 1;					# default calibration scale ($imageHeight/$calib_Height)
  $calib_Height = 480.;				# calibration is for this image size!!!
  $imageWidthMin  = 160;			# 160
  $imageHeightMin = 120;			# 120

  @undostack = @redostack = ();
  $stackmax = 10;

#----------------------------------------------------------------
  $SIG{'HUP'}  = 'signal_handler';			#(1) hangup (see signum.h)
  $SIG{'INT'}  = 'signal_handler';			#(2) interrupt (see signum.h)
  $SIG{'QUIT'} = 'signal_handler';			#(3) quit (see signum.h)
# $SIG{'KILL'} = 'signal_handler';			#(9) this is what is sent by kill -9 -- unblockable
  $SIG{'TERM'} = 'signal_handler';			#(15) terminate -- sent by kill without paramaters
# $SIG{'STOP'} = 'signal_handler';			#(19) stop -- unblockable

### Set monitors to make the things faster:
  $status = &addCaMonitor($clickPV, 'string');

  $status = &addCaMonitor($p2mLoVPV);
  $status = &addCaMonitor($p2mLoHPV);
  $status = &addCaMonitor($p2mHiVPV);
  $status = &addCaMonitor($p2mHiHPV);

  $status = &addCaMonitor($omegaAxisPV.':ActPos');

  $status = &setAssyMonitors($assyXY);
  $status = &addCaMonitor($assyXY.':AccelTime');
  $status = &addCaMonitor($assyXY.':AccelCurve');
  for ($i=0; $i<$naxis; $i++) {
     $status = &setAxisMonitors($axesXY[$i]);
     $status = &addCaMonitor($axesXY[$i].':speed');
  }

  if (!defined $HAVE_XYZ) {
     $status = &setAssyMonitors($assySH);
     $status = &addCaMonitor($assySH.':AccelTime');
     $status = &addCaMonitor($assySH.':AccelCurve');
     $status = &setAxisMonitors($axisSH);
     $status = &addCaMonitor($axisSH.':speed');
  }

  $status = &addCaMonitor($fpe_startPV);
  $status = &addCaMonitor($fpe_operstatusPV);

  if ($watchrobot) {
    $status = &addCaMonitor($robotStatusPV, 'string');
    $status = &addCaMonitor($robotPausePV);
    $robotStatus = 'unknown';
    $robotPause  = 0;
  } else {
    $robotStatus = $robotOK;
    $robotPause  = 1;
  }

  &setEpicsMessages('off');
  $sleep_interval  = 0.05;
  $stop_interval   = -8.0;					#set to <=0 to disable periodical aborts to assyXY & assySH
  $elapsed         = 0.0;
  $connection_lost = 0;
  @cameraLabel = ('LowRes', 'HiRes', 'SONICC', 'Absolute');
  $t0 = [Time::HiRes::gettimeofday];			 	# start measuring time
  $clickInd = 0;						# reset 3-click index
CLICK:
  while(1) { ###############################################################################
### Wait for click:                                                                         #
    $click = 0;						        #no click yet               #
    &clear_click();						# clear click string        #
    $stamp = scalar localtime;                                                              #
    print STDOUT "\n".$script.': '.$stamp.'  Waiting for Click...'."\n";                    #
# --------------------------------------------------------			            #
    do { #################################################################################  #
       select(undef,undef,undef,$sleep_interval);	        #short sleep             #  #
       ($status,$clickString) = Pezca::GetString($clickPV);                              #  #
       if ($status) {                                           #Connection lost         #  #
          if ($connection_lost == 0) {                                                   #  #
             &report_only($script,$status,$clickPV);                                     #  #
             $stamp = scalar localtime;                                                  #  #
             print STDOUT "\n".'*** '.$script.': '.$stamp.'  PV connection lost!'."\n";  #  #
#            &tellBluIce('P&C: connection lost','error');                                #  #
             $connection_lost = 1;                                                       #  #
          }                                                                              #  #
          $click = 0;                                           #Connection lost         #  #
          $clickString = '';                                    #Clear the string        #  #
          $clickStrBak = $clickString;                          #Make a copy             #  #
          sleep 5;                                              #wait before trying again#  #
       }                                                                                 #  #
       elsif (length($clickString) == 0) {                      #Nothing entered yet     #  #
### No click:                                                                            #  #
          $clickStrBak = $clickString;                          #make a copy             #  #
          if ($connection_lost) {                                                        #  #
             $stamp = scalar localtime;                                                  #  #
             print STDOUT "\n".$script.': '.$stamp.'  PV connection restored. ',         #  #
                                         'Reconnecting all monitors...'."\n";            #  #
             &tellBluIce('P&C: re-connecting PVs');                                      #  #
             $connection_lost = 0;                                                       #  #
             &resetAllCaMonitors('list');                                                #  #
             $stamp = scalar localtime;                                                  #  #
             print STDOUT "\n".$script.': '.$stamp.'  Monitors are reconnected'."\n";    #  #
          }                                                                              #  #
          $click = 0;                                                                    #  #
          if ($stop_interval > 0) {                                                      #  #
             $elapsed = Time::HiRes::tv_interval ($t0);         #time from start (secnds)#  #
             if ($elapsed > $stop_interval) {                                            #  #
### Periodically send abort to the resting stage                                         #  #
### (just in case JBluIce lost it's callback)                                            #  #
                $runprg = &getRunPrg($assyXY);                                           #  #
                if ($runprg==0) {                                                        #  #
                   $stamp = scalar localtime;                                            #  #
                   print STDOUT $script.': '.$stamp.'  Sending periodical ',             #  #
                                                    'Abort to '.$assyXY."\n";            #  #
                   &abortAssy ($assyXY,0);                                               #  #
                }                                                                        #  #
                if (!defined $HAVE_XYZ) {                                                #  #
                   $runprg = &getRunPrg($assySH);                                        #  #
                   if ($runprg==0) {                                                     #  #
                      $stamp = scalar localtime;                                         #  #
                      print STDOUT $script.': '.$stamp.'  Sending periodical ',          #  #
                                                       'Abort to '.$assySH."\n";         #  #
                      &abortAssy ($assySH,0);                                            #  #
                   }			                                                 #  #
                }                                                                        #  #
                $t0 = [Time::HiRes::gettimeofday];       	#restart measuring time  #  #
             }                                                                           #  #
          }                                                                              #  #
       }                                                                                 #  #
       else {                                                                            #  #
### Got a click:                                                                         #  #
          if ($connection_lost) {                                                        #  #
             $stamp = scalar localtime;                                                  #  #
             print STDOUT "\n".$script.': '.$stamp.'  PV connection restored. ',         #  #
                                         'Reconnecting all monitors...'."\n";            #  #
             &tellBluIce('P&C: re-connecting PVs');                                      #  #
             $connection_lost = 0;                                                       #  #
             &resetAllCaMonitors();                                                      #  #
          }                                                                              #  #
### Click string format:                                                                 #  #
### ^^^^^^^^^^^^^^^^^^^                                                                  #  #
### $camera  $xpixel  $ypixel  $imageWidth  $imageHeight  [R]   #5-6 words               #  #
### $camera  $dx_um  $dy_um  [R]  um                            #3-4 words + um          #  #
### redo                                                        #1 word (redo)           #  #
### undo                                                        #1 word (undo)           #  #
### ------> where:                                                                       #  #
### camera: 0=LowRes, 1=HiRes                                                            #  #
### imageWidth x imageHeight = (704x480) or (352x240) for Axis camera                    #  #
### imageWidth x imageHeight = (640x480) or (320x240) for IQeye camera                   #  #
### rotation: R=rotate sample by 90deg. after processing click (default: no rotation)    #  #
### um: can be at any position. When click in um (microns),                              #  #
###     parameters $imageWidth and $imageHeight are not used.                            #  #
###                                                                                      #  #
### Additional click string for SONICC (camera=2):                                       #  #
### ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^                                        #  #
### 2 $dx_um $dy_um $phiOffset                                  #4 wrds (SONICC)         #  #
### 2R $dx_um $dy_um $phiOffset                                 #4 wrds (SONICC/reclick) #  #
### 2Z $dz_um $phiOffset                                        #3 wrds (SONICC/z-shift) #  #
### ABS x_mm y_mm z_mm omega                                    #5 wrds (SONICC/absolute)#  #
### ABS x_mm y_mm z_mm omega um                                 #6 wrds (SONICC/absolute)#  #
### ------> where:                                                                       #  #
### phiOffset is the angular direction to the SONICC camera                              #  #
###     (normally same as the HiRes camera; then phiOffset=0).                           #  #
### 2R requests re-click.                                                                #  #
### 2Z requests sample shift downstream(+) or upstream(-) the beam.                      #  #
###                                                                                      #  #
          $clickString =~ s/\s+$//g;                            #remove end spaces       #  #
          $clickStrBak = $clickString;                          #make a backup           #  #
          if ($clickString =~ /um/i) {$clickType = 'um';  $clickString =~ s/um//gi;}     #  #
          else                       {$clickType = 'px';}                                #  #
                                                                                         #  #
          @cmd = split /\s+/, $clickString;                	#parse click string      #  #
                                                                                         #  #
          $undo = 0;						#0=none,1=undo,2=redo    #  #
          $reclick = 0;						#0=none,1=reclick        #  #
          $alongBeam = 0;       				#0=no,1=yes              #  #
          $nwords = @cmd;                                                                #  #
          if (($nwords == 6 && $cmd[0] =~ /^[01]$/)  ||         #pixels shift + rotation #  #
              ($nwords == 5 && $cmd[0] =~ /^[01]$/)  ||         #pixels shift no rotation#  #
              ($nwords == 4 && $cmd[0] =~ /^[01]$/ && $clickType eq 'um') ||    #dx_um+R #  #
              ($nwords == 3 && $cmd[0] =~ /^[01]$/ && $clickType eq 'um') ||    #dx_um   #  #
              ($nwords == 4 && $cmd[0] =~ /^2R?$/i)  ||         #SONICC / SONICC-reclick #  #
              ($nwords == 3 && $cmd[0] =~ /^2Z$/i)   ||         #SONICC up/down stream   #  #
              ($nwords == 5 && $cmd[0] =~ /^ABS$/i)  ||         #Absolute coordinates    #  #
              ($nwords == 1 && $cmd[0] =~ /^(REDO|UNDO)$/i)) {                           #  #
             &clear_click();				 	#clear click string      #  #
             if (($nwords == 5 || $nwords == 6) && $cmd[0] =~ /^[01]$/) {                #  #
                ($camera, $xpixel, $ypixel, $imageWidth, $imageHeight, $rotation) = @cmd;#  #
             }                                                                           #  #
             elsif (($nwords == 3 || $nwords == 4) && $clickType eq 'um') {              #  #
                ($camera, $xpixel, $ypixel, $rotation) = @cmd;                           #  #
                ($imageWidth, $imageHeight) = (0,0);		#not used in um mode     #  #
             }                                                                           #  #
             elsif ($nwords == 4 && $cmd[0] =~ /^2R?$/i) {      #SONICC & SONICC-reclick #  #
                ($camera, $xpixel, $ypixel, $phiOffset) = @cmd;                          #  #
                $rotation = 'N';                                #not used with SONICC    #  #
                $clickType = 'um';                                                       #  #
                ($imageWidth, $imageHeight) = (0,0);		#not used in um mode     #  #
                if ($camera =~ /^2R$/i) {                                                #  #
                   $reclick = 1;                                                         #  #
                   $camera  = 2;                                                         #  #
                }                                                                        #  #
             }                                                                           #  #
             elsif ($nwords == 3 && $cmd[0] =~ /^2Z$/i) {       #SONICC up/down stream   #  #
                ($camera, $zpixel, $phiOffset) = @cmd;                                   #  #
                $camera    = 2;                                                          #  #
                $alongBeam = 1;       				#translate up/down stream#  #
                $rotation  = 'N';                               #not used with SONICC    #  #
                $clickType = 'um';                                                       #  #
                ($xpixel,     $ypixel)      = (0,0);		#no x/y shift            #  #
                ($imageWidth, $imageHeight) = (0,0);		#not used in um mode     #  #
             }                                                                           #  #
             elsif ($nwords == 5 && $cmd[0] =~ /^ABS$/i) {      #absolute coordinates    #  #
                ($camera, $NewSampleX, $NewSampleY, $NewSampleH, $newOmega) = @cmd;      #  #
                if ($clickType eq 'um') {                                                #  #
                   $NewSampleX = 0.001*$NewSampleX;                                      #  #
                   $NewSampleY = 0.001*$NewSampleY;                                      #  #
                   $NewSampleH = 0.001*$NewSampleH;                                      #  #
                }                                                                        #  #
                $camera    = 3;                                                          #  #
                $rotation  = 'N';                               #not used with Absolute  #  #
                $clickType = 'mm';                                                       #  #
                $phiOffset = 0;                                                          #  #
                ($xpixel,     $ypixel)      = (0,0);		#no x/y shift            #  #
                ($imageWidth, $imageHeight) = (0,0);		#not used in um mode     #  #
             }                                                                           #  #
             elsif ($nwords == 1) {                                                      #  #
                if    ($cmd[0]=~/^UNDO$/i) {$undo=1;}                                    #  #
                elsif ($cmd[0]=~/^REDO$/i) {$undo=2;}                                    #  #
                ($camera, $xpixel, $ypixel, $rotation) = (0,0,0,'N');                    #  #
                ($imageWidth, $imageHeight) = (0,0);		#not used in undo mode   #  #
             }                                                                           #  #
             if (!defined $rotation) {$rotation = 'N'}          #R=rotate N=normal (no)  #  #
             if ($rotation ne 'R' && $rotation ne 'N') {        #R=rotate N=normal (no)  #  #
                $stamp = scalar localtime;                                               #  #
                print STDOUT '*** '.$script.': '.$stamp.'  Wrong 3-click rotation ',     #  #
                                                           'flag='.$rotation."\n";       #  #
                &tellBluIce('P&C: wrong 3-click rotation flag','warn');                  #  #
                $rotation = 'N';                                #R=rotate N=normal (no)  #  #
             }                                                                           #  #
             $scale = $imageHeight/$calib_Height;        	#scale to calibration    #  #
             $click = 1;                                 	#got a click!            #  #
                                                                                         #  #
### Now check for possible collisions: if frame                                          #  #
### is being acquired & robot not ready:                                                 #  #
             $click = &checkCollisionsFree();                                            #  #
          } else {                                                                       #  #
### Incorrect number of words in the click string:                                       #  #
             if (length($clickString) > 0) {                                             #  #
                $stamp = scalar localtime;                                               #  #
                print STDOUT '*** '.$script.': '.$stamp                                  #  #
                             .'  Incorrect click command: "'                             #  #
                             .$clickStrBak.'"'."\n";                                     #  #
                &tellBluIce('P&C: wrong click "'.$clickStrBak.'"','error');              #  #
             }                                                                           #  #
             $click = 0;                                        #ignore incorrect click  #  #
             &clear_click();				        #clear click string      #  #
          }                                                                              #  #
       }                                                                                 #  #
    } while ($click != 1); ###############################################################  #
# --------------------------------------------------------			            #
    $stamp = scalar localtime;                                                              #
    print STDOUT $script.': '.$stamp.'  ...Got click: "'.$clickStrBak.'"'."\n";             #
    &processing_click();					#processing click string    #
                                                                                            #
    if    ($undo == 1) {                                                                    #
       &undoPositions();                                                                    #
       $undo = 0;                                                                           #
       next CLICK;                                                                          #
    }                                                                                       #
    elsif ($undo == 2) {                                                                    #
       &redoPositions();                                                                    #
       $undo = 0;                                                                           #
       next CLICK;                                                                          #
    }                                                                                       #
                                                                                            #
    if ($camera !~ /^[0123]$/) {                    #0=LowRes 1=HiRes 2=SONICC 3=Absolute   #
       $stamp = scalar localtime;                                                           #
       print STDOUT '*** '.$script.': '.$stamp.'  Incorrect camera='.$camera."\n";          #
       &tellBluIce('P&C: wrong video camera='.$camera,'error');                             #
       next CLICK;                                                                          #
    }                                                                                       #
    if ($camera eq '0') {                                       #LowRes camera              #
       $cameraOrientation = $loResOrientation;                                              #
       $p2mVPV = $p2mLoVPV;                                                                 #
       $p2mHPV = $p2mLoHPV;                                                                 #
    }                                                                                       #
    elsif ($camera eq '1') {                                    #HiRes camera               #
       $cameraOrientation = $hiResOrientation;                                              #
       $p2mVPV = $p2mHiVPV;                                                                 #
       $p2mHPV = $p2mHiHPV;                                                                 #
    }                                                                                       #
    elsif ($camera eq '2') {                                    #SONIC                      #
       $cameraOrientation = $otherOrientation;                                              #
       $p2mVPV = 'UNUSED';                                                                  #
       $p2mHPV = 'UNUSED';                                                                  #
    }                                                                                       #
    elsif ($camera eq '3') {                                    #Absolute                   #
       $cameraOrientation = $otherOrientation;                                              #
       $p2mVPV = 'UNUSED';                                                                  #
       $p2mHPV = 'UNUSED';                                                                  #
    }                                                                                       #
                                                                                            #
    if (! $reclick) { #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#        #
### Get Omega:                                                                     #        #
       $omega = &getAxisAct ($omegaAxisPV,\$status);                               #        #
       if ($status) {                                                              #        #
          $stamp = scalar localtime;                                               #        #
          print STDOUT '*** '.$script.': '.$stamp.'  Cannot read omega position. ' #        #
                                                        .'Setting omega=0.'."\n";  #        #
          &tellBluIce('P&C: cannot get Omega','error');                            #        #
          $omega = 0.;                                                             #        #
          next CLICK;                                                              #        #
       }                                                                           #        #
    } else {  #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~reclick~~~~~~~~~~~#        #
       $stacklength = @undostack;                                                  #        #
       if ($stacklength < 4) {                                                     #        #
          $stamp = scalar localtime;                                               #        #
          print STDOUT '*** '.$script.': '.$stamp                                  #        #
                      .'  Empty undo stack for reclick'."\n";                      #        #
          &tellBluIce('P&C: empty undo for reclick','error');                      #        #
          next CLICK;                                                              #        #
       }                                                                           #        #
       #          X                Y           SH or Z       $omega                #        #
       @acts = ($undostack[0],$undostack[1],$undostack[2],$undostack[3]);          #        #
       $omega = $acts[3];                                                          #        #
    }  #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#        #
                                                                                            #
    if ($clickType eq 'px') {  #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#    #
### Test input for consistency:                                                        #    #
### $camera: 0=LowRes 1=HiRes 2=SONIC 3=Absolute                                       #    #
       if ($imageWidth < $imageWidthMin) {                                             #    #
          $stamp = scalar localtime;                                                   #    #
          print STDOUT '*** '.$script.': '.$stamp.'  Too small camera-'.$camera        #    #
                      .' imageWidth='.$imageWidth.' < '.$imageWidthMin.'.'."\n"        #    #
                      .'Please reduce digital zoom'."\n";                              #    #
          &tellBluIce('P&C: too small imgWidth='.$imageWidth.'<'.$imageWidthMin        #    #
                     .' Reduce zoom.','error');                                        #    #
          next CLICK;                                                                  #    #
       }                                                                               #    #
       if ($imageHeight < $imageHeightMin) {                                           #    #
          $stamp = scalar localtime;                                                   #    #
          print STDOUT '*** '.$script.': '.$stamp.'  Too small camera-'.$camera        #    #
                      .' imageHeight='.$imageHeight.' < '.$imageHeightMin.'.'."\n"     #    #
                      .'Please reduce digital zoom'."\n";                              #    #
          &tellBluIce('P&C: too small imgHeight='.$imageHeight.'<'.$imageHeightMin     #    #
                     .' Reduce zoom.','error');                                        #    #
          next CLICK;                                                                  #    #
       }                                                                               #    #
       if ($xpixel < 0 || $xpixel > $imageWidth) {                                     #    #
          $stamp = scalar localtime;                                                   #    #
          print STDOUT '*** '.$script.': '.$stamp.'  Xpixel='.$xpixel                  #    #
                      .' not in the image range [0-'.$imageWidth.']'."\n";             #    #
          &tellBluIce('P&C: wrong Xpixel='.$xpixel,'error');                           #    #
          next CLICK;                                                                  #    #
       }                                                                               #    #
       if ($ypixel < 0 || $ypixel > $imageHeight) {                                    #    #
          $stamp = scalar localtime;                                                   #    #
          print STDOUT '*** '.$script.': '.$stamp.'  Ypixel='.$ypixel                  #    #
                     .' not in the image range [0-'.$imageHeight.']'."\n";             #    #
          &tellBluIce('P&C: wrong Ypixel='.$ypixel,'error');                           #    #
          next CLICK;                                                                  #    #
       }                                                                               #    #
### Request pixels -> mm for current camera and zoom:                                  #    #
### +------------------------------------------+                                       #    #
### | This reads the result of the scan record |                                       #    #
### | calculation. So, any changes to camera   |                                       #    #
### | calibration are in effect immediately    |                                       #    #
### | (i.e. point&click does not need restart).|                                       #    #
### +------------------------------------------+                                       #    #
       ($status, $factorV) = Pezca::GetDouble($p2mVPV);                                #    #
       if ($status) {                                                                  #    #
          &report_only($script, $status, $p2mVPV);                                     #    #
          $stamp = scalar localtime;                                                   #    #
          print STDOUT '*** '.$script.': '.$stamp.'  Cannot read V-scale PV: '         #    #
                                                 .$p2mVPV."\n";                        #    #
          &tellBluIce('P&C: cannot get Vscale','error');                               #    #
          $factorV = 0.;                                                               #    #
          next CLICK;                                                                  #    #
       }                                                                               #    #
       ($status, $factorH) = Pezca::GetDouble($p2mHPV);                                #    #
       if ($status) {                                                                  #    #
          &report_only($script, $status, $p2mHPV);                                     #    #
          $stamp = scalar localtime;                                                   #    #
          print STDOUT '*** '.$script.': '.$stamp.'  Cannot read H-scale PV: '         #    #
                                                 .$p2mHPV."\n";                        #    #
          &tellBluIce('P&C: cannot get Hscale','error');                               #    #
          $factorH = 0.;                                                               #    #
          next CLICK;                                                                  #    #
       }                                                                               #    #
### Deviation of click point from the image center:                                    #    #
       $dxpixel = $xpixel - int(($imageWidth+0.0)/2.0);                                #    #
       $dypixel = $ypixel - int(($imageHeight+0.0)/2.0);                               #    #
       printf STDOUT $script.': Image shift (px):  dx=%.2f  dy=%.2f'.                  #    #
                             '  camera=%s (phiOffset='.$phiOffset.")\n",               #    #
                                         $dxpixel, $dypixel, $cameraLabel[$camera];    #    #
### Recalculate pixels into mm:                                                        #    #
       $dxmm = $factorH * $dxpixel / $scale;                                           #    #
       $dymm = $factorV * $dypixel / $scale;                                           #    #
       $dzmm = 0;                                                                      #    #
    }                                                                                  #    #
    else { # if ($clickType eq 'um' || $clickType eq 'mm') #~~~~~~~~~~~~~~~~~~~~~~~~~~~#    #
       if ($camera ne '3') {                                           #if not Absolute#    #
          if (!$alongBeam) {                                                           #    #
             $dxmm = -0.001*$xpixel;                                      # um -> mm   #    #
             $dymm = +0.001*$ypixel;                                      # um -> mm   #    #
             $dzmm = 0;                                                   # um -> mm   #    #
          } else {                                                                     #    #
             $dxmm = 0;                                                   # um -> mm   #    #
             $dymm = 0;                                                   # um -> mm   #    #
             $dzmm = +0.001*$zpixel;                                      # um -> mm   #    #
          }                                                                            #    #
       } else {  #($camera eq '3')                                     #Absolute       #    #
          $dxmm = 0;                                                      # not used   #    #
          $dymm = 0;                                                      # not used   #    #
          $dzmm = 0;                                                      # not used   #    #
       }                                                                               #    #
    } #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#    #
    if ($camera ne '3') {                                                #if not Absolute   #
       print STDOUT $script.': Image shift (mm):  ';                                        #
       if (!$alongBeam) {printf STDOUT 'dx=%.3f  dy=%.3f'."\n", $dxmm, $dymm;}              #
       else             {printf STDOUT 'dz=%.3f'."\n", $dzmm;}                              #
      										            #
       if ($cameraOrientation eq $horizontalOrientation) {                                  #
          ### Hi-Res camera                                                                 #
          $phiOffset = 0.0;			     # offfset between phi and omega        #
       }                                                                                    #
       elsif ($cameraOrientation eq $verticalOrientation) {                                 #
          ### Lo-Res camera                                                                 #
          $phiOffset = 90.0;			     # offfset between phi and omega        #
       }                                                                                    #
       elsif ($cameraOrientation eq $otherOrientation) {                                    #
          ### SONICC: offset between phi and omega passed as a part of the click:           #
          print STDOUT $script.': other camera orientation, phi='.$phiOffset."\n";          #
       }                                                                                    #
       else {                                                                               #
          $phiOffset = 0.0;                                                                 #
          $stamp = scalar localtime;                                                        #
          print STDOUT '*** '.$script.': '.$stamp.' Unknown camera orientation!'."\n";	    #
          &tellBluIce('P&C: unknown camera','error');                                       #
          next CLICK;                                                                       #
       }                                                                                    #
    }                                                                                       #
    $phi = $omega + $phiOffset;                                                             #
    $phi = $phi - 360.*int($phi/360);                                                       #
    if ($phi < 0.) {$phi += 360.;}                                                          #
      										            #
### If 3-click mode:								            #
    if ($rotation eq 'R' && $clickType eq 'px') { #%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%#    #
       $camera3C[$clickInd]    = $cameraOrientation;                                   #    #
       $phiOffset3C[$clickInd] = $phiOffset;                                           #    #
       $scale3C[$clickInd]     = $scale;                                               #    #
       $factorH3C[$clickInd]   = $factorH;                                             #    #
       $factorV3C[$clickInd]   = $factorV;                                             #    #
       $dxmm3C[$clickInd]      = $dxmm;                                                #    #
       $dymm3C[$clickInd]      = $dymm;                                                #    #
       $phi3C[$clickInd]       = $phi;                                                 #    #
       if ($clickInd > 0) {						               #    #
          if ($camera3C[$clickInd]  ne   $camera3C[$clickInd-1] ||                     #    #
              $phiOffset3C[$clickInd] != $phiOffset3C[$clickInd-1] ||                  #    #
              abs(1.-$scale3C[$clickInd-1]/$scale3C[$clickInd]) > 0.01 ||              #    #
              abs(1.-$factorH3C[$clickInd-1]/$factorH3C[$clickInd]) > 0.01 ||          #    #
              abs(1.-$factorV3C[$clickInd-1]/$factorV3C[$clickInd]) > 0.01) {          #    #
             $camera3C[0]    = $camera3C[$clickInd];                                   #    #
             $phiOffset3C[0] = $phiOffset3C[$clickInd];                                #    #
             $scale3C[0]     = $scale3C[$clickInd];                                    #    #
             $factorH3C[0]   = $factorH3C[$clickInd];                                  #    #
             $factorV3C[0]   = $factorV3C[$clickInd];                                  #    #
             $dxmm3C[0]      = $dxmm3C[$clickInd];                                     #    #
             $dymm3C[0]      = $dymm3C[$clickInd];                                     #    #
             $phi3C[0]       = $phi3C[$clickInd];                                      #    #
             $clickInd = 0;				# params changed; start over   #    #
             $stamp = scalar localtime;                                                #    #
             print STDOUT '!!! '.$script.': '.$stamp.'  3-click counter is reset '     #    #
                                        .'because of new click conditions'."\n";       #    #
             &tellBluIce('P&C: 3-click counter reset','warn');                         #    #
          }                                                                            #    #
          elsif (abs($phi3C[$clickInd]-$phi3C[$clickInd-1]) < 5.) {     #d(phi)<5degr. #    #
             $camera3C[$clickInd-1]    = $camera3C[$clickInd];                         #    #
             $phiOffset3C[$clickInd-1] = $phiOffset3C[$clickInd];                      #    #
             $scale3C[$clickInd-1]     = $scale3C[$clickInd];                          #    #
             $factorH3C[$clickInd-1]   = $factorH3C[$clickInd];                        #    #
             $factorV3C[$clickInd-1]   = $factorV3C[$clickInd];                        #    #
             $dxmm3C[$clickInd-1]      = $dxmm3C[$clickInd];                           #    #
             $dymm3C[$clickInd-1]      = $dymm3C[$clickInd];                           #    #
             $phi3C[$clickInd-1]       = $phi3C[$clickInd];                            #    #
             $clickInd--;				# click @ same omega: overwrite#    #
             $stamp = scalar localtime;                                                #    #
             print STDOUT '!!! '.$script.': '.$stamp.'  3-click: previous click',      #    #
                                                         ' is overwritten'."\n";       #    #
             &tellBluIce('P&C: 3-click overwritten','warn');                           #    #
          }                                                                            #    #
       }                                                                               #    #
    } else { #%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%#    #
       $rotation = 'N';                                 # no 3-click with arrows (mm)  #    #
       $clickInd = 0;                                   # params changed; start over   #    #
    } #%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%#    #
                     								            #
    if ($rotation ne 'R') { #%%%%%%%%%%% begin of normal (none-3-click) processing %%%%#    #
       $phi = $phi * $degree2rad;                                                      #    #
### Calculate XY motions (since we only see the projection of the vertical             #    #
### deviation of the crystal, the vertical shift of this projection could              #    #
### potenticall be compensated by multiple ways. The choice of this particalular       #    #
### one is to keep the sample in the focal plane, i.e. don't move it away or           #    #
### towards the camera):                                                               #    #
       if (!$alongBeam) {                                                              #    #
          $dSampleX =  sin($phi) * $dymm;                                              #    #
          $dSampleY =  cos($phi) * $dymm;                                              #    #
       } else {      								       #    #
          $dSampleX =  cos($phi) * $dzmm;               # to test: is "+" downstream?  #    #
          $dSampleY = -sin($phi) * $dzmm;               # to test: is "+" downstream?  #    #
       }                                                                               #    #
### Calculate horizontal motion:                                                       #    #
       $dSampleH = -$dxmm;                                                             #    #
                                                                                       #    #
    } else {                #%%%%%%%%%%%%%% begin of 3-click processing %%%%%%%%%%%%%%%#    #
### 3-click operation:                                                                 #    #
       if ($clickInd < 2) {                                                            #    #
### Clicks 1 and 2 of 3 ($clickInd = 0,1):                                             #    #
          printf STDOUT $script.': [3C] got click-%d of 3, '                           #    #
                               .'rotating omega by +90degr.'."\n",                     #    #
                                 $clickInd+1;                                          #    #
          $angle = $omega + 90.;                                                       #    #
          $angle = $angle - 360.*int($angle/360);                                      #    #
          if ($angle < 0.) {$angle += 360.;}                                           #    #
          &pushPositionsInStack();                                                     #    #
          $status = &runMotor ($omegaAssyPV, $omegaAxisPV, $angle, 'checkAuto');       #    #
          if ($status) {                                                               #    #
             $stamp = scalar localtime;                                                #    #
             print STDOUT '*** '.$script.': '.$stamp.'  Rotating omega by +90degr. '   #    #
                                             .'failed with status='.$status."\n";      #    #
             &tellBluIce('P&C: omega+90 failed','error');                              #    #
          }                                                                            #    #
          print STDOUT $script.': [3C] finished rotating omega by +90degr.'."\n";      #    #
       }                                                                               #    #
       else {                                                                          #    #
### Last click:                                                                        #    #
### ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~        #    #
###  The calculations of vertical shift are performed according to the        ~        #    #
###  formula stating that at any phi the vertical position is:                ~        #    #
###                                                                           ~        #    #
###  $y = $radius * cos($phi-$phi0) + $y0                                     ~        #    #
###     = $radius * (cos($phi)*cos($phi0) + sin($phi)*sin($phi0) +$y0;        ~        #    #
###                                                                           ~        #    #
###  This gives:                                                              ~        #    #
###  $dy1 = $radius * (cos($phi1)*cos($phi0) + sin($phi1)*sin($phi0)) + $y0;  ~        #    #
###  $dy2 = $radius * (cos($phi2)*cos($phi0) + sin($phi2)*sin($phi0)) + $y0;  ~        #    #
###  $dy3 = $radius * (cos($phi3)*cos($phi0) + sin($phi3)*sin($phi0)) + $y0;  ~        #    #
###                                                                           ~        #    #
###  Let us denote $rc = $radius * cos($phi0),   $rs = $radius * sin($phi0)   ~        #    #
###  Then:                                                                    ~        #    #
###  cos($phi1)*$rc + sin($phi1)*$rs + $y0 = $dy1;                            ~        #    #
###  cos($phi2)*$rc + sin($phi2)*$rs + $y0 = $dy2;                            ~        #    #
###  cos($phi3)*$rc + sin($phi3)*$rs + $y0 = $dy3;                            ~        #    #
###                                                                           ~        #    #
###  Let us subtract (1)-(2) and (1-3):                                       ~        #    #
###                                                                           ~        #    #
###  (cos($phi1)-cos($phi2))*$rc + (sin($phi1)-sin($phi2))*$rs = $dy1-$dy2;   ~        #    #
###  (cos($phi1)-cos($phi3))*$rc + (sin($phi1)-sin($phi3))*$rs = $dy1-$dy3;   ~        #    #
###                                                                           ~        #    #
###         |cos($phi1)-cos($phi2)   sin($phi1)-$sin(phi2)|                   ~        #    #
###  $det = |                                             |                   ~        #    #
###         |cos($phi1)-cos($phi3)   sin($phi1)-$sin(phi3)|                   ~        #    #
###                                                                           ~        #    #
###                                                                           ~        #    #
###         |$dy1-$dy2   sin($phi1)-$sin(phi2)|                               ~        #    #
###   $rc = |                                 | / $det                        ~        #    #
###         |$dy1-$dy3   sin($phi1)-$sin(phi3)|                               ~        #    #
###                                                                           ~        #    #
###                                                                           ~        #    #
###         |cos($phi1)-$sin(phi2)   $dy1-$dy2|                               ~        #    #
###   $rs = |                                 | / $det                        ~        #    #
###         |cos($phi1)-$sin(phi3)   $dy1-$dy3|                               ~        #    #
###                                                                           ~        #    #
###                                                                           ~        #    #
###   $y0 = $dy1 - $rc*cos($ph1) - $rs*sin($ph1)                              ~        #    #
### ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~        #    #
          $dx1=$dxmm3C[0];  $dy1=$dymm3C[0];  $ph1=$phi3C[0]*$degree2rad;              #    #
          $dx2=$dxmm3C[1];  $dy2=$dymm3C[1];  $ph2=$phi3C[1]*$degree2rad;              #    #
          $dx3=$dxmm3C[2];  $dy3=$dymm3C[2];  $ph3=$phi3C[2]*$degree2rad;              #    #
                                                                                       #    #
          $det = (cos($ph1)-cos($ph2))*(sin($ph1)-sin($ph3))                           #    #
               - (cos($ph1)-cos($ph3))*(sin($ph1)-sin($ph2));                          #    #
          if ($det < 1.E-8) {                                                          #    #
             print STDOUT $script.': [3C] rotation matrix is singular'."\n",;          #    #
             &tellBluIce('P&C: singular matrix','error');                              #    #
             $clickInd = 0;                                                            #    #
             next CLICK;                                                               #    #
          }                                                                            #    #
          $rc = (($dy1-$dy2)*(sin($ph1)-sin($ph3))                                     #    #
              -  ($dy1-$dy3)*(sin($ph1)-sin($ph2))) / $det;                            #    #
          $rs = ((cos($ph1)-cos($ph2))*($dy1-$dy3)                                     #    #
              -  (cos($ph1)-cos($ph3))*($dy1-$dy2)) / $det;                            #    #
          $radius = sqrt($rc*$rc + $rs*$rs);                                           #    #
          $phi0 = atan2($rs,$rc);			# in radians                   #    #
          $y0 = $dy1 - $rc*cos($ph1) - $rs*sin($ph1);                                  #    #
          print STDOUT "\n".$script.': [3C] sample deviations based on 3 clicks:'."\n";#    #
          printf STDOUT 'radius=%.3f  phi0=%.3f'."\n", $radius, $phi0/$degree2rad;     #    #
          printf STDOUT 'y0=%.3f   r*cos(phi0)=%.3f   r*sin(phi0)=%.3f'."\n\n",        #    #
                                                                   $y0,$rc,$rs;        #    #
          $dSampleX = ($y0 + $rs);                                                     #    #
          $dSampleY = ($y0 + $rc);                                                     #    #
          $dSampleH = -($dx1+$dx2+$dx3)/3.0;                                           #    #
       }                                                                               #    #
    } #%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% end of any processing/calculating coords %%#    #
                                                                                            #
                                                                                            #
    if ($rotation eq 'N' || $clickInd == 2) { ## %%% normal click or last 3-click %%% ##    #
### Do motion:                                                                         #    #
                                                                                       #    #
       if (! $reclick) {                                                               #    #
### Get current positions of all three drives:                                         #    #
          $statusAll=0;                                                                #    #
          $ActSampleX = &getAxisAct ($axesXY[0],\$status);   $statusAll += $status;    #    #
          $ActSampleY = &getAxisAct ($axesXY[1],\$status);   $statusAll += $status;    #    #
          $ActSampleH = &getAxisAct ($axisSH,\$status);      $statusAll += $status;    #    #
          $omega      = &getAxisAct ($omegaAxisPV,\$status); $statusAll += $status;    #    #
          if ($statusAll) {                                                            #    #
             $stamp = scalar localtime;                                                #    #
             print STDOUT '*** '.$script.': '.$stamp.'  Error getting current'         #    #
                                .' position(s). May drive to wrong place'."\n";        #    #
             &tellBluIce('P&C: error getting current positions','error');              #    #
          }                                                                            #    #
       } else {                                                         #reclick       #    #
          ($ActSampleX,$ActSampleY,$ActSampleH,$omega) = @acts;                        #    #
       }                                                                               #    #
                                                                                       #    #
       if ($camera ne '3') {                                            #non-Absolute  #    #
          ### Calculate new requested positions:                                       #    #
          $NewSampleX = $ActSampleX + $dSampleX;                                       #    #
          $NewSampleY = $ActSampleY + $dSampleY;                                       #    #
          $NewSampleH = $ActSampleH + $dSampleH;                                       #    #
          $newOmega   = $omega;                                                        #    #
       } else {                                                         #Absolute      #    #
          $dSampleX = $NewSampleX - $ActSampleX;                                       #    #
          $dSampleY = $NewSampleY - $ActSampleY;                                       #    #
          $dSampleH = $NewSampleH - $ActSampleH;                                       #    #
          $dxmm = $dSampleX;                                                           #    #
          $dymm = $dSampleY;                                                           #    #
          $dzmm = $dSampleH;                                                           #    #
       }                                                                               #    #
       $dOmega = $newOmega - $omega;                                                   #    #
                                                                                       #    #
       printf STDOUT $script.': Sample shift:'."\n"                                    #    #
                    .'                   dx=%.3f mm'."\n"                              #    #
                    .'                   dy=%.3f mm'."\n"                              #    #
                    .'                   dh=%.3f mm'."\n"                              #    #
                    .'               domega=%.3f degr'."\n",                           #    #
                     $dSampleX, $dSampleY, $dSampleH, $dOmega;                         #    #
       if ($dymm != 0 || $dxmm != 0 || $dzmm != 0 || $dOmega != 0) {                   #    #
          if (! $reclick) {                                                            #    #
             &pushPositionsInStack([$ActSampleX,$ActSampleY,$ActSampleH,$omega]);      #    #
          }                                                                            #    #
          @rqs = ();                                                                   #    #
          if (!defined $HAVE_XYZ) {                                                    #    #
             if ($dymm != 0 ||                                                         #    #
                 $dzmm != 0) {($rqs[0],$rqs[1]) = ($NewSampleX,$NewSampleY);}          #    #
             else            {($rqs[0],$rqs[1]) = (undef,      undef);}                #    #
             if ($dxmm != 0) {$rqs[2] = $NewSampleH;}                                  #    #
             else            {$rqs[2] = undef;}                                        #    #
          } else {                                                                     #    #
             if ($dymm != 0 ||                                                         #    #
                 $dzmm != 0 ||                                                         #    #
                 $dxmm != 0) {@rqs = ($NewSampleX,$NewSampleY, $NewSampleH);}          #    #
             else            {@rqs = (  undef,        undef,      undef);}             #    #
          }                                                                            #    #
          if ($dOmega != 0) {$rqs[3] = $newOmega;}                                     #    #
          else              {$rqs[3] = undef;}	  		#omega not changed     #    #
          &runAll(\@rqs,[$ActSampleX,$ActSampleY,$ActSampleH,$omega]);                 #    #
       }                                                                               #    #
                                                                                       #    #
    } #%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% end of motion part %%%%%%%%%%%%%%%%%%#    #
### Abort motion to let JBluIce recover from missed callback:                               #
    &abortAssy ($assyXY,0);			                                            #
    if (!defined $HAVE_XYZ) {&abortAssy ($assySH,0);}			                    #
    if ($rotation eq 'R') {					#3-click                    #
       $clickInd++;                                                                         #
       if ($clickInd > 2) {$clickInd = 0;}                                                  #
    }                                                                                       #
    $stamp = scalar localtime;                                                              #
    print STDOUT $script.': '.$stamp.'  Click processed.'."\n";                             #
  } #########################################################################################
  &clear_click();						# clear click string
  &clearAllCaMonitors();
  exit;

#=================================================================

sub clear_click () {
  $clickString = '';
  $status = Pezca::PutString($clickPV,$clickString);			# clear click string
  if ($status) {&report_only($script,$status,$clickPV);}
}

#=================================================================

sub processing_click () {
  $clickString = 'Processing click';
  $status = Pezca::PutString($clickPV,$clickString);			# clear click string
  if ($status) {&report_only($script,$status,$clickPV);}
}

#=================================================================

sub pushPositionsInStack (;\@) {
  my ($act_ref, $status, @acts);
  $act_ref  = shift(@_);
  if (!defined $act_ref) {
#    print STDOUT '!!! pushPositionsInStack: did not receive current positions. Querying from hardware'."\n";
     $status = &getActPositions (\@acts);
     if ($status) {return;}
     @$act_ref = @acts;
#    print STDOUT 'pushPositionsInStack: receieved these positions '."(@${act_ref}\n";
  }
# $nelem = splice(@undostack,0,0,@$act_ref);			#adds to beginning
  $nelem = unshift(@undostack,@$act_ref);			#adds to beginning
  if ($nelem > 4*$stackmax) {splice(@undostack,-4,4);}		#removes last 4 elements
  @redostack = ();
# print STDOUT "\n".$script.': pushPositionsInStack called on stack=.'."\n";
# print STDOUT 'Current undo stack has '.$nelem.' elements: '."(@undostack)\n";
# print STDOUT 'Current redo stack is empty:'."(@redostack)\n\n";
}

#=================================================================

sub getActPositions (\@) {
  my ($act_ref, $actpos, @axes, $axis, $status);
  $act_ref  = shift(@_);
  if (!defined $HAVE_XYZ) {@axes = (@axesXY, $axisSH, $omegaAxisPV);}
  else                    {@axes = (@axesXY, $omegaAxisPV);}
  @$act_ref = ();
  foreach $axis (@axes) {
     $actpos = &getAxisAct ($axis,\$status);
     if ($status) {
        $stamp = scalar localtime;
        print STDOUT '*** '.$script.': '.$stamp.'  error='.$status
                           .' while getting position of '.$axis."\n";
        &tellBluIce('P&C: '.$axis.' read error','warn');
        return 1;
     }
     push(@$act_ref, $actpos);				    #adds to the end
  }
# print STDOUT 'getActPositions: got these positions '."(@${act_ref})\n";
  return 0;
}

#=================================================================

sub undoPositions () {

  my (@act, $status);

  $nelem = @undostack;

  if ($nelem >= 4) {
     $status = &getActPositions (\@act);
     if (! $status) {
#       splice(@redostack,0,0,@act);			#adds to beginning
        unshift(@redostack,@act);			#adds to beginning
     }
     @rqs = splice(@undostack,0,4);			#remove 1st four eleements
     print STDOUT $script.': '.$stamp.'  Undoing click'."\n";
     &tellBluIce('P&C: Undoing click');
     &runAll(\@rqs);
  } else {
     print STDOUT '!!! '.$script.': '.$stamp.'  Undo buffer is empty.'."\n";
     &tellBluIce('P&C: Undo buffer empty','warn');
  }
# $nelem = @undostack;
# print STDOUT "\n".$script.': undoPositions called.'."\n";
# print STDOUT 'Current undo stack has '.$nelem.' elements: '."(@undostack)\n";
# $nelem = @redostack;
# print STDOUT 'Current redo stack has '.$nelem.' elements: '."(@redostack)\n\n";
}

#=================================================================

sub redoPositions () {

  my (@act, $status);

  $nelem = @redostack;

  if ($nelem >= 4) {
     $status = &getActPositions (\@act);
     if (! $status) {
#       splice(@undostack,0,0,@act);			#adds to beginning
        unshift(@undostack,@act);			#adds to beginning
     }
     @rqs = splice(@redostack,0,4);			#remove 1st four eleements
     print STDOUT $script.': '.$stamp.'  Redoing click'."\n";
     &tellBluIce('P&C: Redoing click');
     &runAll(\@rqs);
  } else {
     print STDOUT '!!! '.$script.': '.$stamp.'  Redo buffer is empty.'."\n";
     &tellBluIce('P&C: Redo buffer empty','warn');
  }
# $nelem = @undostack;
# print STDOUT "\n".$script.': redoPositions called.'."\n";
# print STDOUT 'Current undo stack has '.$nelem.' elements: '."(@undostack)\n";
# $nelem = @redostack;
# print STDOUT 'Current redo stack has '.$nelem.' elements: '."(@redostack)\n\n";
}

#=================================================================

sub runAll (\$;\$) {
  my ($rqsref, $actref, @acts, $status, @requested, $def, $i);

  $rqsref = shift(@_);						#rqs positions of: (@axesXY,$axisSH,$omegaAxis)
  $actref = shift(@_);						#act positions of: (@axesXY,$axisSH,$omegaAxis)
  if (defined $actref) {@acts = @$actref;}
  else {
    $acts[0] = &getAxisAct ($axesXY[0],  \$status);
    $acts[1] = &getAxisAct ($axesXY[1],  \$status);
    $acts[2] = &getAxisAct ($axisSH,     \$status);
    $acts[3] = &getAxisAct ($omegaAxisPV,\$status);
  }

  if (defined @$rqsref[0]) {printf STDOUT $script.': '.$axesXY[0]  .': %.3f --> %.3f'."\n",$acts[0],@$rqsref[0];}
  else                     {printf STDOUT $script.': '.$axesXY[0]  .': %.3f --> %.3f'."\n",$acts[0],$acts[0];}
  if (defined @$rqsref[1]) {printf STDOUT $script.': '.$axesXY[1]  .': %.3f --> %.3f'."\n",$acts[1],@$rqsref[1];}
  else                     {printf STDOUT $script.': '.$axesXY[1]  .': %.3f --> %.3f'."\n",$acts[1],$acts[1];}
  if (defined @$rqsref[2]) {printf STDOUT $script.': '.$axisSH     .': %.3f --> %.3f'."\n",$acts[2],@$rqsref[2];}
  else                     {printf STDOUT $script.': '.$axisSH     .': %.3f --> %.3f'."\n",$acts[2],$acts[2];}
  if (defined @$rqsref[3]) {printf STDOUT $script.': '.$omegaAssyPV.': %.3f --> %.3f'."\n",$acts[3],@$rqsref[3];}
  else                     {printf STDOUT $script.': '.$omegaAssyPV.': %.3f --> %.3f'."\n",$acts[3],$acts[3];}

  @requested = ();
  $def = 1;
  for ($i=0; $i<$naxis; $i++) {
     push(@requested,$$rqsref[$i]);
     if (!defined @$rqsref[0]) {$def=0;}
  }
### Start motion:
  if ($def) {
     $stamp = scalar localtime;
     print  STDOUT $script.': '.$stamp.'  Starting '.$assyXY.'...'."\n";
     $status = &startNMotors ($assyXY, $naxis, \@axesXY, \@requested);
     if ($status) {
### Report error to JBluIce:
        $stamp = scalar localtime;
        print STDOUT '*** '.$script.': '.$stamp.'  Cannot start XY: status='.$status."\n";
        &tellBluIce('P&C: Gonio-Cr failed','error');
     }
  }
  if (!defined $HAVE_XYZ && defined @$rqsref[2]) {
     $stamp = scalar localtime;
     print STDOUT $script.': '.$stamp.'  Starting '.$assySH.'...'."\n";
     $status = &startMotor ($axisSH, @$rqsref[2], $assySH, 'checkAuto');
     if ($status) {
### Report error to JBluIce:
        $stamp = scalar localtime;
        print STDOUT '*** '.$script.': '.$stamp.'  Cannot start SH: status='.$status."\n";
        &tellBluIce('P&C: Gonio-SH failed','error');
     }
  }

  if (defined @$rqsref[3]) {
     $stamp = scalar localtime;
     print  STDOUT $script.': '.$stamp.'  Running '.$omegaAssyPV.'...'."\n";
     $status = &runMotor ($omegaAssyPV, $omegaAxisPV, @$rqsref[3], 'checkAuto');
     if ($status) {
        $stamp = scalar localtime;
        print STDOUT '*** '.$script.': '.$stamp.'  Cannot start Omega: status='.$status."\n";
        &tellBluIce('P&C: Omega failed','error');
     }
  }

### Estimate motion time:
  $motionTimeMax = 0;
  if ($def) {
     $motionTimeXY = &calcNMotorsTime($assyXY, $naxis, \@axesXY, \@requested);
     if ($motionTimeMax < $motionTimeXY) {$motionTimeMax = $motionTimeXY;}
  }
  if (!defined $HAVE_XYZ && defined @$rqsref[2]) {
     $motionTimeSH = &calcMotorTime($assySH, $axisSH, $NewSampleH);
     if ($motionTimeMax < $motionTimeSH) {$motionTimeMax = $motionTimeSH;}
  }
  $motionTimeMax += 3;					#add a few seconds to avoid timouts

### Track XY positioner motion:
  if ($def) {
     $status = &trackNMotors($assyXY,$motionTimeMax,$naxis,\@axesXY);
     if ($status) {
### Report error to JBluIce:
        $stamp = scalar localtime;
        print STDOUT '*** '.$script.': '.$stamp.'  Gonio-Cr timeout, status='.$status."\n";
        &tellBluIce('P&C: Gonio-Cr timeout','warn');
     }
     $stamp = scalar localtime;
     print STDOUT $script.': '.$stamp.'  Finished tracking '.$assyXY."\n";
  }

### Track support horizontal motion:
  if (!defined $HAVE_XYZ && defined @$rqsref[2]) {
     $status = &trackMotor($assySH,$motionTimeMax,$axisSH);
     if ($status) {
### Report error to JBluIce:
        $stamp = scalar localtime;
        print STDOUT '*** '.$script.': '.$stamp.'  Gonio-SH timeout, status='.$status."\n";
        &tellBluIce('P&C: Gonio-SH timeout','warn');
     }
     $stamp = scalar localtime;
     print STDOUT $script.': '.$stamp.'  Finished tracking '.$assySH."\n";
  }
### Abort motion to let JBluIce recover from missed callback(s):
# &abortAssy ($assyXY,0);
# if (!defined $HAVE_XYZ) {&abortAssy ($assySH,0);}
# &abortAssy ($omegaAssyPV,0);
}

#=================================================================

sub checkCollisionsFree() {

  my $checkmonitor = 0;						#flag to check for dead monitors (was "1" before 3/2012)
  $fpe_start      = 0;
  $fpe_operstatus = 0;
  $robotStatus    = $robotOK;
  $robotPause     = 1;

### Check if frame is being acquired:
  ($status, $fpe_start) = Pezca::GetDouble($fpe_startPV);
  if ($status) {                           			#Connection lost
     &report_only($script,$status,$fpe_startPV);
     $fpe_start = 0;               				#fpe not avail., ignore
  }
### added 2010.12.08: check for lost callback (may be removed later)####
  if ($fpe_start && $checkmonitor) {                                   #
     &resetCaMonitor($fpe_startPV);                                    #
     ($status, $fpe_start) = Pezca::GetDouble($fpe_startPV);	#reread the value after resetting the movitor
     if ($status) {                                                    #
        $fpe_start = 0;                                                #
     }                                                                 #
     elsif ($fpe_start == 0) {					#is it different after resetting monitor?
        $stamp = scalar localtime;                                     #
        print STDOUT '!!! '.$script.': '.$stamp                        #
                           .'  Possible lost callback on'              #
                           .$fpe_startPV."\n";                         #
        &tellBluIce('P&C: possible lost callback on '                  #
                           .$fpe_startPV,'warn');                      #
     }                                                                 #
  }                                                                    #
### end of added 2010.12.08: check for lost callback ###################

  ($status, $fpe_operstatus) = Pezca::GetDouble($fpe_operstatusPV);
  if ($status) {						#Connection lost
     &report_only($script,$status,$fpe_operstatusPV);
     $fpe_operstatus = 0;					#fpe not avail., ignore
  }
### added 2010.12.08: check for lost callback (may be removed later)####
  if ($fpe_operstatus && $checkmonitor) {                              #
     &resetCaMonitor($fpe_operstatusPV);                               #
     ($status, $fpe_operstatus) = Pezca::GetDouble($fpe_operstatusPV); #reread the value after resetting the movitor
     if ($status) {                                                    #
        $fpe_operstatus = 0;                                           #
     }                                                                 #
     elsif ($fpe_operstatus == 0) {				#is it different after resetting monitor?
        $stamp = scalar localtime;                                     #
        print STDOUT '!!! '.$script.': '.$stamp                        #
                           .'  Possible lost callback on'              #
                           .$fpe_operstatusPV."\n";                    #
        &tellBluIce('P&C: possible lost callback on '                  #
                           .$fpe_operstatusPV,'warn');                 #
     }                                                                 #
  }                                                                    #
### end of added 2010.12.08: check for lost callback ###################

  if ($fpe_start || $fpe_operstatus == 1) {
     $stamp = scalar localtime;
     print STDOUT '!!! '.$script.': '.$stamp.'  Request ignored '
                        .'because FPE is acquiring a frame: start='
                        .$fpe_start.' operation='.$fpe_operstatus."\n";
     &tellBluIce('P&C: request ignored: FPE is acquiring a frame. '
                 	.$fpe_startPV.'='.$fpe_start.' '
                 	.$fpe_operstatusPV.'='.$fpe_operstatus,
                        'error');
     $click = 0; return 0;                 			#ignore click due to FPE
  }

### Now check if robot is ready:
  if ($watchrobot) {
     ($status, $robotStatus) = Pezca::GetString($robotStatusPV);
     if ($status) {                           			#Connection lost
        &report_only($script,$status,$robotStatusPV);
        $robotStatus = $robotOK;				#robot not avail.,ignore
     }
     ($status, $robotPause) = Pezca::GetDouble($robotPausePV);
     if ($status) {                           			#Connection lost
        &report_only($script,$status,$robotPausePV);
        $robotPause  = 1;					#robot not avail.,ignore
     }
     if ($robotStatus !~ /${robotOK}$/i      &&
         $robotStatus !~ /warm_up/i          &&			#this will change @new gripper
         $robotStatus !~ /pin_base warm-up/i &&			#this is part of mount, gonio is already in place
         $robotPause  != 1) {
### Ignore click if robot is not ready and not paused:
        $stamp = scalar localtime;
        print STDOUT '!!! '.$script.': '.$stamp.'  Request ignored '
                    .'because robot is not in the ready/paused '
     	       .'state: state='.$robotStatus."\n";
        &tellBluIce('P&C: request ignored: robot is working. '
                    .$robotStatusPV.'='.$robotStatus,'warn');
        $click = 0; return 0;					#ignore click due robot
     }
  }
  return 1;							#OK
}

#=================================================================

sub click_syntax_help () {
  $NONINTERACTIVE = 0;
  print STDOUT "\n".'This script is a point-and-click server. It shifts the sample in the'.
               "\n".'beamline microscopes field of view per mouse click on the image.'.
               "\n".'The click data is passed to the server via the "'.$beamline.$clickPV.'" PV.'."\n\n";
  print STDOUT 'Acceptable click commands:'."\n"
              .'--------------------------'."\n"
              .'camera xpixel ypixel imageWidth imageHeight [R]'."\n"
              .'camera dx_um dy_um um [R]'."\n"
              .'redo'."\n"
              .'undo'."\n"
              .' --> where:'."\n"
              .'  camera: 0=LowRes, 1=HiRes'."\n"
              .'  imageWidth x imageHeight = (704x480) or (352x240) for Axis camera.'."\n"
              .'  imageWidth x imageHeight = (640x480) or (320x240) for IQeye camera.'."\n"
              .'  rotation: R=rotate sample by 90 degr. after processing the click.'."\n"
              .'  um: can be at any position. When click is in um (microns),'."\n"
              .'      the parameters imageWidth and imageHeight are not used.'."\n\n";
  print STDOUT 'Additional click commands for SONICC (camera=2):'."\n"
              .'------------------------------------------------'."\n"
              .'2  dx_um dy_um phiOffset'."\n"
              .'2R dx_um dy_um phiOffset'."\n"
              .'2Z dz_um phiOffset'."\n"
              .' --> where:'."\n"
              .'  phiOffset is the angular direction to the SONICC camera'."\n"
              .'      (normally same as the HiRes camera; then phiOffset=0).'."\n"
              .'  2R requests re-click.'."\n"
              .'  2Z requests sample shift downstream(+) or upstream(-) the beam.'."\n\n";
  &press_key_exit("\n".'Command line:'.$syntax);
}

#=================================================================

sub signal_handler ($) {
# local($sig) = @_;
  my $sig = shift;

  $stamp = scalar localtime;
  print STDOUT "\n".'*** '.$script.': '.$stamp.'  Caught a SIG'.$sig."\n";
  &clear_click();						# clear click string
  &clearAllCaMonitors();
  print STDOUT $script.': Finished!'."\n";
  exit;
}

