#!/usr/bin/perl -w
#----------------------------------------------------------------
# This script provides a wrapper around the Mark Rivers EPID
# program in order to provide positional feedback.
# Inputs come from the GS BPMs.
# Output voltages go to the piezo crystals in the HFM and VFM.
#
# Author:	dwy
# Version:	2.04	2006/04/05
#		2.05	2006/06/20 	- Deals with attenuator changes
#					- Minor fixes
#		2.06	2006/06/21	- More bug fixes, work being done on IDout
#               2.07	2006/09/29	- Removed option for choosing H or V feedback
#		3.01	2006/10/02	- Break apart the code; create more efficient organiation
#----------------------------------------------------------------

use Pezca;
use strict;
use List::Util;

# Global variables are declared here:

   # Configuration variables
      our ($beamline, $outdir, $outfile);
      our ($gs_up_ch, $gs_dn_ch, $gs_ob_ch, $gs_ib_ch, $vV_check_ch, $hV_check_ch, $vV_ch, $hV_ch);
      our ($rate, $lo_cutoff, $hi_cutoff, $dac_adc_err, $low_sr_cutoff, $good_lo, $good_hi);
   # Misc variables
      our ($host, $user, $zero_measured, $stpt_measured, $message_time);
      our ($filehandle, $init_time, $adj_result, $quitting, $aborted);
      our ($last_pos_v, $last_pos_h, @last_arr_v, @last_arr_h);
   # Variables for PV names and values
      # General
      our ($postab_pv, $pid_pv, $pid_v_pv, $pid_h_pv, $pid_calc_pv, $pid_calc2_pv);
      # Status checks
      our ($shutter_permit_pv, $shutter_a_pv, $shutter_white_pv, $shutter_mono_pv, $ID_mov_pv);
      our ($mbsh_open_pv, $mbsh_close_pv, $E_pv, $E_mov_pv, $E_ID, $E_ID_pv);
      our ($shutter_permit, $shutter_a, $shutter_white, $shutter_mono, $ID_mov);
      our ($mbsh_open, $mbsh_close, $E, $E_mov);
      # Keithleys
      our ($kl_gain_up_pv, $kl_gain_dn_pv, $kl_gain_ob_pv, $kl_gain_ib_pv);
      our ($kl_gain_up_rqs_pv, $kl_gain_dn_rqs_pv, $kl_gain_ob_rqs_pv, $kl_gain_ib_rqs_pv);
      our ($kl_off_up_pv, $kl_off_dn_pv, $kl_off_ob_pv, $kl_off_ib_pv);
      our ($kl_off_up_rqs_pv, $kl_off_dn_rqs_pv, $kl_off_ob_rqs_pv, $kl_off_ib_rqs_pv);
      our ($kl_gain, $kl_gain_rqs);
      our ($kl_off_up, $kl_off_dn, $kl_off_ob, $kl_off_ib);
      our ($kl_off_up_rqs, $kl_off_dn_rqs, $kl_off_ob_rqs, $kl_off_ib_rqs);
      our ($kl_pv, $kl, $kl_set_pv, $kl_set);
      # Attenuators
      our ($att_pv, $att);
      our ($att_00_pv, $att_01_pv, $att_02_pv, $att_03_pv);
      our ($att_04_pv, $att_05_pv, $att_06_pv, $att_07_pv);
      our ($att_08_pv, $att_09_pv, $att_10_pv, $att_11_pv);
      our ($att_12_pv, $att_13_pv, $att_14_pv, $att_15_pv);
      # DAC
      our ($vV_pv, $vV, $hV_pv, $hV);
      # ADC
      our ($gs_up, $gs_up_pv, $gs_up_name, $gs_up_name_pv, $gs_up_rate_pv);
      our ($gs_dn, $gs_dn_pv, $gs_dn_name, $gs_dn_name_pv, $gs_dn_rate_pv);
      our ($gs_ob, $gs_ob_pv, $gs_ob_name, $gs_ob_name_pv, $gs_ob_rate_pv);
      our ($gs_ib, $gs_ib_pv, $gs_ib_name, $gs_ib_name_pv, $gs_ib_rate_pv);
      our ($vV_check, $vV_check_pv, $vV_check_name, $vV_check_name_pv, $vV_check_rate_pv);
      our ($hV_check, $hV_check_pv, $hV_check_name, $hV_check_name_pv, $hV_check_rate_pv);
      # Postab PVs
      our (@pvs);
      our ($postab_status_pv, $postab_status, $postab_stline_pv, $postab_stline);
      our ($postab_sttime_pv, $postab_sttime, $postab_pause_pv, $postab_pause);
      our ($postab_pause_v_pv, $postab_pause_v, $postab_pause_h_pv, $postab_pause_h);
      our ($postab_zero_pv, $postab_zero, $postab_stpt_pv, $postab_stpt);
      our ($postab_zero_chk_pv, $postab_stpt_chk_pv, $postab_active_pv, $postab_active);
      our ($postab_query_pv, $postab_query, $postab_stop_pv, $postab_stop);
      our ($postab_mode_pv, $postab_mode, $sr_current_pv, $sr_current);
      our ($postab_set_v_pv, $postab_set_v, $postab_set_h_pv, $postab_set_h);
      our ($postab_del_v_pv, $postab_del_v, $postab_del_h_pv, $postab_del_h);
      our ($message1_pv, $message2_pv, $message3_pv, $message4_pv, $message5_pv);
      our ($message1, $message2, $message3, $message4, $message5);
      our ($postab_kl_rqs_pv, $postab_kl_rqs);
      our ($postab_v_driver_pv, $postab_v_driver);
      # Mostab PVs
      our ($mostab_pv, $mostab_pause_pv, $mostab_pause, $mostab_pause_init);
      # PID control PVs
      our ($v_on_pv, $v_on, $h_on_pv, $h_on, $v_pid_rate_pv, $v_pid_rate, $h_pid_rate_pv, $h_pid_rate);
      our ($v_stpt_pv, $h_stpt_pv, $v_rdbkpv_pv, $v_rdbkpv, $h_rdbkpv_pv, $h_rdbkpv);
      our ($v_ctrlpv_pv, $v_ctrlpv, $h_ctrlpv_pv, $h_ctrlpv);
      # PID parameter PVs
      our ($v_kp_pv, $v_kp, $h_kp_pv, $h_kp);
      our ($v_ki_pv, $v_ki, $h_ki_pv, $h_ki);
      our ($v_kd_pv, $v_kd, $h_kd_pv, $h_kd);
      our ($v_i_pv, $v_i, $h_i_pv, $h_i);
      our ($v_dt_pv, $h_dt_pv);
      our ($v_ll_pv, $v_ll, $h_ll_pv, $h_ll);
      our ($v_hl_pv, $v_hl, $h_hl_pv, $h_hl);
      # PID calc 1 PVs
      our ($calc_rate_pv, $calc_rate);
      our ($up_0_cmt_pv, $up_0_cmt, $up_0_ipv_pv, $up_0_ipv, $up_0_clc_pv, $up_0_clc, $up_0_val_pv, $up_0_val, $up_0_out_pv, $up_0_out);
      our ($dn_0_cmt_pv, $dn_0_cmt, $dn_0_ipv_pv, $dn_0_ipv, $dn_0_clc_pv, $dn_0_clc, $dn_0_val_pv, $dn_0_val, $dn_0_out_pv, $dn_0_out);
      our ($ob_0_cmt_pv, $ob_0_cmt, $ob_0_ipv_pv, $ob_0_ipv, $ob_0_clc_pv, $ob_0_clc, $ob_0_val_pv, $ob_0_val, $ob_0_out_pv, $ob_0_out);
      our ($ib_0_cmt_pv, $ib_0_cmt, $ib_0_ipv_pv, $ib_0_ipv, $ib_0_clc_pv, $ib_0_clc, $ib_0_val_pv, $ib_0_val, $ib_0_out_pv, $ib_0_out);
      our ($up_cmt_pv, $up_cmt, $up_ipv_pv, $up_ipv, $up_clc_pv, $up_clc, $up_val_pv, $up_val, $up_out_pv, $up_out);
      our ($dn_cmt_pv, $dn_cmt, $dn_ipv_pv, $dn_ipv, $dn_clc_pv, $dn_clc, $dn_val_pv, $dn_val, $dn_out_pv, $dn_out);
      our ($ob_cmt_pv, $ob_cmt, $ob_ipv_pv, $ob_ipv, $ob_clc_pv, $ob_clc, $ob_val_pv, $ob_val, $ob_out_pv, $ob_out);
      our ($ib_cmt_pv, $ib_cmt, $ib_ipv_pv, $ib_ipv, $ib_clc_pv, $ib_clc, $ib_val_pv, $ib_val, $ib_out_pv, $ib_out);
      our ($vs_cmt_pv, $vs_cmt, $vs_ipv_pv, $vs_ipv, $vs_clc_pv, $vs_clc, $vs_val_pv, $vs_val, $vs_out_pv, $vs_out);
      our ($vp_cmt_pv, $vp_cmt, $vp_ipv_pv, $vp_ipv, $vp_clc_pv, $vp_clc, $vp_val_pv, $vp_val, $vp_out_pv, $vp_out);
      our ($hs_cmt_pv, $hs_cmt, $hs_ipv_pv, $hs_ipv, $hs_clc_pv, $hs_clc, $hs_val_pv, $hs_val, $hs_out_pv, $hs_out);
      our ($hp_cmt_pv, $hp_cmt, $hp_ipv_pv, $hp_ipv, $hp_clc_pv, $hp_clc, $hp_val_pv, $hp_val, $hp_out_pv, $hp_out);
      our ($vsp_cmt_pv, $vsp_cmt, $vsp_ipv_pv, $vsp_ipv, $vsp_clc_pv, $vsp_clc, $vsp_val_pv, $vsp_val, $vsp_out_pv, $vsp_out);
      our ($hsp_cmt_pv, $hsp_cmt, $hsp_ipv_pv, $hsp_ipv, $hsp_clc_pv, $hsp_clc, $hsp_val_pv, $hsp_val, $hsp_out_pv, $hsp_out);
      our ($v_lo_cmt_pv, $v_lo_cmt, $v_lo_ipv_pv, $v_lo_ipv, $v_lo_clc_pv, $v_lo_clc, $v_lo_val_pv, $v_lo_val, $v_lo_out_pv, $v_lo_out);
      our ($h_lo_cmt_pv, $h_lo_cmt, $h_lo_ipv_pv, $h_lo_ipv, $h_lo_clc_pv, $h_lo_clc, $h_lo_val_pv, $h_lo_val, $h_lo_out_pv, $h_lo_out);
      # PID calc 2 PVs
      our ($calc2_rate_pv, $calc2_rate);
      our ($sr_cmt_pv, $sr_cmt, $sr_ipv_pv, $sr_ipv, $sr_clc_pv, $sr_clc, $sr_val_pv, $sr_val, $sr_out_pv, $sr_out);
      our ($sr_cmp_cmt_pv, $sr_cmp_cmt, $sr_cmp_ipv_pv, $sr_cmp_ipv, $sr_cmp_clc_pv, $sr_cmp_clc, $sr_cmp_val_pv, $sr_cmp_val, $sr_cmp_out_pv, $sr_cmp_out);
      our ($vV_cmt_pv, $vV_cmt, $vV_ipv_pv, $vV_ipv, $vV_clc_pv, $vV_clc, $vV_val_pv, $vV_val, $vV_out_pv, $vV_out);
      our ($vV_adc_cmt_pv, $vV_adc_cmt, $vV_adc_ipv_pv, $vV_adc_ipv, $vV_adc_clc_pv, $vV_adc_clc, $vV_adc_val_pv, $vV_adc_val, $vV_adc_out_pv, $vV_adc_out);
      our ($vV_err_cmt_pv, $vV_err_cmt, $vV_err_ipv_pv, $vV_err_ipv, $vV_err_clc_pv, $vV_err_clc, $vV_err_val_pv, $vV_err_val, $vV_err_out_pv, $vV_err_out);
      our ($hV_cmt_pv, $hV_cmt, $hV_ipv_pv, $hV_ipv, $hV_clc_pv, $hV_clc, $hV_val_pv, $hV_val, $hV_out_pv, $hV_out);
      our ($hV_adc_cmt_pv, $hV_adc_cmt, $hV_adc_ipv_pv, $hV_adc_ipv, $hV_adc_clc_pv, $hV_adc_clc, $hV_adc_val_pv, $hV_adc_val, $hV_adc_out_pv, $hV_adc_out);
      our ($hV_err_cmt_pv, $hV_err_cmt, $hV_err_ipv_pv, $hV_err_ipv, $hV_err_clc_pv, $hV_err_clc, $hV_err_val_pv, $hV_err_val, $hV_err_out_pv, $hV_err_out);
      our ($act_rqs_cmt_pv, $act_rqs_cmt, $act_rqs_ipv_pv, $act_rqs_ipv, $act_rqs_clc_pv, $act_rqs_clc, $act_rqs_val_pv, $act_rqs_val, $act_rqs_out_pv, $act_rqs_out);
      our ($v_rqs_cmt_pv, $v_rqs_cmt, $v_rqs_ipv_pv, $v_rqs_ipv, $v_rqs_clc_pv, $v_rqs_clc, $v_rqs_val_pv, $v_rqs_val, $v_rqs_out_pv, $v_rqs_out);
      our ($h_rqs_cmt_pv, $h_rqs_cmt, $h_rqs_ipv_pv, $h_rqs_ipv, $h_rqs_clc_pv, $h_rqs_clc, $h_rqs_val_pv, $h_rqs_val, $h_rqs_out_pv, $h_rqs_out);
      our ($v_on_cmt_pv, $v_on_cmt, $v_on_ipv_pv, $v_on_ipv, $v_on_clc_pv, $v_on_clc, $v_on_val_pv, $v_on_val, $v_on_out_pv, $v_on_out);
      our ($h_on_cmt_pv, $h_on_cmt, $h_on_ipv_pv, $h_on_ipv, $h_on_clc_pv, $h_on_clc, $h_on_val_pv, $h_on_val, $h_on_out_pv, $h_on_out);
      our ($V_limit_cmt_pv, $V_limit_cmt, $V_limit_ipv_pv, $V_limit_ipv, $V_limit_clc_pv, $V_limit_clc, $V_limit_val_pv, $V_limit_val, $V_limit_out_pv, $V_limit_out);
      our ($H_limit_cmt_pv, $H_limit_cmt, $H_limit_ipv_pv, $H_limit_ipv, $H_limit_clc_pv, $H_limit_clc, $H_limit_val_pv, $H_limit_val, $H_limit_out_pv, $H_limit_out);

# Mode definitions:
   # 0  = Startup;
   # 1  = Feedback paused;
   # 2  = Vertical feedback on (horizontal paused);
   # 3  = Horizontal feedback on (vertical paused);
   # 4  = 2-D feedback on;
   # 5  = Measure offsets;
   # 6  = Measure setpoints;
   # 7  = Wait for beam (component check);
   # 8  = Live KL change;
   # 99 = Stop script;

# Status definitions:
   # 0 = Stopped;
   # 1 = Vert feedback loop running;
   # 2 = Horz feedback loop running;
   # 3 = Both feedback loops running;
   # 4 = Other;

# Postab PV names:
   # Postab PVs begin with the {beamline}:postab: and end with one of the following:
   #
   # active -		Is postab active?
   # status -		See above.  Status is generally on, off, or other.
   # statusline -	Text description of what's actually happening.
   # statustime -	Time of last status update.
   # pause -		Under user control.  Pauses both horz and vert feedback.
   # pause_h -		Pause just horz feedback (i.e. vertical active)
   # pause_v -		Pause just vert feedback (i.e. horizontal active)
   # mode - 		See above.  Mode will be used for major functions (e.g. pause or active feedback),
   #			but not for minor ones (e.g. DAC check)
   # zero -		Request that offsets be measured
   # stpt -		Request that setpoints be measured
   # zero_chk -		Have offsets been measured?
   # stpt_chk -		Have setpoints been measured?
   # stop -		Completely stop feedback
   # set_v -		Vertical setpoint
   # set_vInc		Increment to vertical setpoint
   # set_h		Horizontal setpoint
   # set_hInc		Increment to horizontal setpoint
   # message{x} -	Text. Message fields 1 through 5.

# Additional PV names:
   # In addition to the above PVs created just for this script, feedback calculations
   # are done in a pre-existing, dedicated calculation record.  The PV names take the
   # form {beamline}:SoftPID1_incalc.xxxx or {beamline}:SoftPID2_incalc.xxxx.

# Subs
   # The following subs are used in this script:
   #
   # init_vars		- sets initial variable values
   # logfile		- opens a logfile
   # script_check	- checks for existing script
   # init_PVs		- sets initial PV values
   # pause		- pauses; watches for changes and input
   # offsets		- measures offsets
   # setpoints		- measures setpoints and offsets
   # feedback		- feedback operational, both directions
   # feedback_h		- feedback operational, horizontal
   # feedback_v		- feedback operational, vertical
   # pause_no_beam	- take action if beam is lost
   # stop_script	- shut everything down
   # putPV		- an elaborate 'caput'
   # getPV		- an elaborate 'caget'
   # set_monitors	- set PV monitors
   # clear_monitors	- clear PV monitors
   # adj_monitor	- actually does the setting or clearing of the monitors
   # pv_check		- watch PVs for user input; take appropriate action
   # att_check		- watch for attenuator changes
   # kl_check		- watch for changes to Keithley gains
   # kl_set		- set Keithley gains
   # mode		- sets PVs based on a given mode
   # dac_check		- verify that actual output matches requested value for DAC within a given tolerance
   # message		- write messages to output
   # signal_handler	- take action on keyboard break, etc.

my $CommonPEZCA;
if    (defined $ENV{PEZCA})     { $CommonPEZCA = $ENV{PEZCA}.'/Common/';}
elsif (defined $ENV{FASTSCANS}) {$CommonPEZCA = $ENV{FASTSCANS}.'/../pezca/Common/';}
else                            {$CommonPEZCA = 'Common/';}
if  (-f $CommonPEZCA.'beamline.pl') {do $CommonPEZCA.'beamline.pl';}
else                                {die 'Cannot find '.$CommonPEZCA.'beamline.pl';}

  select STDERR; $|=1;                  # set unbuffered output
  select STDOUT; $|=1;                  # set unbuffered output

&init_vars;		# initialize variables
&logfile;		# start logfile; check for pre-existing scripts; initialize feedback
&script_check;		# check for pre-existing scripts
&init_PVs;		# set initial values of PVs, set monitors
$postab_mode = 1;	# after initialization, feedback goes into a paused state

while ($postab_mode != 99) {
   if ($postab_mode == 1)    {print STDOUT 'pause'."\n";         &pause;}
   if ($postab_mode == 4)    {print STDOUT 'feedback'."\n";      &feedback;}
   if ($postab_mode == 5)    {print STDOUT 'offsets'."\n";       &offsets;}
   if ($postab_mode == 6)    {print STDOUT 'setpoints'."\n";     &setpoints;}
   if ($postab_mode == 7)    {print STDOUT 'pause_no_beam'."\n"; &pause_no_beam;}
   if (($postab_mode != 1) &&
       ($postab_mode != 4) &&
       ($postab_mode != 5) &&
       ($postab_mode != 6) &&
       ($postab_mode != 7) &&
       ($postab_mode != 99)) {print STDOUT 'else pause'."\n";    &pause;}
}

&stop_script(1);
exit;

# ---------------------------------------------------------------

sub pause {

   # Declare local vars

   my (@pvs);

   &message("Paused.");
   &mode($postab_mode);
   while ($postab_mode == 1) {
      (@pvs) = &pv_check(10);
      if ($pvs[3] == 1) {print STDOUT "@pvs\n"; $postab_mode = 99; return;}	# exit
      if ($pvs[1] == 1) {print STDOUT "@pvs\n"; $postab_mode = 5;  return;}	# offsets
      if ($pvs[2] == 1) {print STDOUT "@pvs\n"; $postab_mode = 6;  return;}	# setpoints
      if ($pvs[0] == 0) {print STDOUT "@pvs\n"; $postab_mode = 4;  return;}	# feedback active
      select(undef,undef,undef,0.5);
      &active_light;
   }
}

# ---------------------------------------------------------------

sub offsets {

   # Declare local variables here:
   my ($init_shutter_mono, $set_result);

   # Set postab PVs to correct values for mode
   &mode($postab_mode);
   &message('Starting - offsets.');
   &active_light;

   # Adjust offsets to approximately 0.025 V
   &message('Adjusting KL offsets.');
   $set_result = &kl_set(1);
   &active_light;

   # Pause or exit if terminated early
   if ($set_result == 0) {
      &message('Offset measurement terminated early.');
      &message('Offsets not recorded.');
      $zero_measured = 0;
      &putPV($postab_zero_chk_pv,$zero_measured,1);
      (@pvs) = &pv_check(10);
      if ($pvs[3] == 1) {$postab_mode = 99; return;}		# stop
      else              {$postab_mode = 1;  return;}		# pause
   }

   # Check for user input
   (@pvs) = &pv_check(50);
   if ($pvs[3] == 1) {$postab_mode = 99; return;}		# stop
   if ($pvs[0] == 3) {$postab_mode = 1;  return;}		# pause = 3; pause
   if ($pvs[2] == 1) {$postab_mode = 6;}			# setpoints
   if ($pvs[1] == 1) {$postab_mode = 5;}			# offsets; ignore here
   if ($pvs[0] == 0) {$postab_mode = 4;}			# pause = 0; run


   # Get current Keithley gains and attenuator values:
   &kl_check;							# This requires KL server
   &att_check($att);						# This sub is currently only used here
   &active_light;

   # Change flag
   $zero_measured = 1;
   &putPV($postab_zero_chk_pv,$zero_measured,1);
   $postab_zero = 0;
   &putPV($postab_zero_pv,$postab_zero,1);
   &message('Finished - offsets.');
   if ($postab_mode == 5) {$postab_mode = 1;}
}

# ---------------------------------------------------------------

sub setpoints {

   # Declare local variables
   my ($init_shutter_mono, $set_result);
   my (@status_pvs);
   my ($permit, $err, $timeout);
   my ($sr_err, $permit_err, $a_err, $wbsh_err, $mbsh_err, $E_mov_err, $ID_mov_err);

   &mode($postab_mode);
   &message('Starting - setpoints.');
   &active_light;

   # Optimize Keithleys
   &message('Checking KL gains and offsets.');
   $set_result = &kl_set(0);
   &active_light;

   # If gains and offsets not handled properly...
   if ($set_result == 0) {
      &message('Setpoint measurement terminated early.');
      &message('Setpoints not recorded.');
      $zero_measured = 0;
      $stpt_measured = 0;
      &putPV($postab_zero_chk_pv,$zero_measured,1);
      &putPV($postab_stpt_chk_pv,$stpt_measured,1);
   }
   (@pvs) = &pv_check(60);
   if ($pvs[3] == 1) {$postab_mode = 99; return;}		# stop
   if ($pvs[0] == 3) {$postab_mode = 1;  return;}		# pause = 3; pause
   if ($pvs[2] == 1) {$postab_mode = 6;}			# setpoints
   if ($pvs[1] == 1) {$postab_mode = 5;}			# offsets; ignore here
   if ($pvs[0] == 0) {$postab_mode = 4;}			# pause = 0; run

   # Verify and/or wait for open shutter
   &message('Checking for open MBSH.');
   &active_light;
   $shutter_mono = &getPV($shutter_mono_pv,1);
   if ($shutter_mono) {
      &message('Shutter is closed!');
      $timeout = 0;
      if (($beamline eq '23i') || ($beamline eq '23o')) {
         &message('Opening shutter.');
         Pezca::PutDouble($mbsh_open_pv,1);			# Added due to unknown EPICS problem.  PutDouble returns error, but error isn't obvious
         #&putPV($mbsh_open_pv,1,1);
	 &timeout(0);
         $shutter_mono = &getPV($shutter_mono_pv,1);
	 while (($shutter_mono == 1) && ($timeout == 0)) {
	    $shutter_mono = &getPV($shutter_mono_pv,1);
            (@pvs) = &pv_check(61);
            if ($pvs[3] == 1) {$postab_mode = 99; return;}	# stop
            if ($pvs[0] == 3) {$postab_mode = 1;  return;}	# pause = 3; pause
            if ($pvs[2] == 1) {$postab_mode = 6;}		# setpoints
            if ($pvs[1] == 1) {$postab_mode = 5;}		# offsets; ignore here
            if ($pvs[0] == 0) {$postab_mode = 4;}		# pause = 0; run
	    $timeout = &timeout(30);
            if ($timeout) {&message('Error opening MBSH.');}
	    select(undef,undef,undef,0.5);
	    &active_light;
	 }
      }
      if ((($beamline ne '23i') && ($beamline ne '23o')) || ($timeout)) {
         &message('Please open MBSH manually!');
	 while ($shutter_mono == 1) {
	    $shutter_mono = &getPV($shutter_mono_pv,1);
            (@pvs) = &pv_check(61);
            if ($pvs[3] == 1) {$postab_mode = 99; return;}	# stop
            if ($pvs[0] == 3) {$postab_mode = 1;  return;}	# pause = 3; pause
            if ($pvs[2] == 1) {$postab_mode = 6;}		# setpoints
            if ($pvs[1] == 1) {$postab_mode = 5;}		# offsets; ignore here
            if ($pvs[0] == 0) {$postab_mode = 4;}		# pause = 0; run
	    select(undef,undef,undef,0.5);
	    &active_light;
	 }
      }
   }
   &message('MBSH is open.');
   &active_light;

   # Check for bad intensity
   $v_lo_val = &getPV($v_lo_val_pv,1);
   $h_lo_val = &getPV($h_lo_val_pv,1);

   # If intensities are out of range...
   if ($v_lo_val||$h_lo_val) {
      # Check first to see if intensity is too high
      $up_val = &getPV($up_val_pv,1);
      $dn_val = &getPV($dn_val_pv,1);
      $ob_val = &getPV($ob_val_pv,1);
      $ib_val = &getPV($ib_val_pv,1);
      if (($up_val > $good_hi) || ($up_val > $good_hi) || ($up_val > $good_hi) || ($up_val > $good_hi)) {
         &message('Keithley gain is too high');
         &message('Please lower gain');
         &message('Re-measure offsets, setpoints');
         $postab_mode = 1;
         return;
      }

      # Check for SR current, shutter problems, etc.
      $E_mov		= &getPV($E_mov_pv,1);
      if ($E_mov)	{$E_mov_err = 0;}
      else		{$E_mov_err = 1;}
      $ID_mov_err	= &getPV($ID_mov_pv,1);
      $mbsh_err		= &getPV($shutter_mono_pv,1);
      $wbsh_err		= &getPV($shutter_white_pv,1);
      $a_err		= &getPV($shutter_a_pv,1);
      $permit		= &getPV($shutter_permit_pv,1);
      if ($permit)	{$permit_err = 0;}
      else		{$permit_err = 1;}
      $sr_err		= &getPV($sr_cmp_val_pv,1);
      $err = $E_mov_err + $ID_mov_err + $mbsh_err + $wbsh_err + $a_err + $permit_err + $sr_err;
      if ($err > 0)    {
         if ($E_mov_err)  {&message('DCM is moving.');}
         if ($ID_mov_err) {&message('ID is moving.');}
         if ($mbsh_err)   {&message('MBSH is closed.');}
         if ($wbsh_err)   {&message('IBSH is closed.');}
         if ($a_err)      {&message('FES is closed.');}
         if ($permit_err) {&message('No FES permit.');}
         if ($sr_err)     {&message('No ring current.');}
         &message('Please bring beam to the endstation.');
         &message('Then measure setpoints.');
         $postab_mode = 1;
	 &active_light;
         return;
      }

      # Low intensity
      &message('Intensity out of range');
      &message('Cannot record setpoints');
      &message('Is Keithley gain correct?');
      $stpt_measured = 0;
      &putPV($postab_stpt_chk_pv,$stpt_measured,1);
      $postab_zero = 0;
      $postab_stpt = 0;
      &putPV($postab_zero_pv,$postab_zero,1);
      &putPV($postab_stpt_pv,$postab_stpt,1);
      &message('Pausing.');
      $postab_mode = 1;
      return;
   }

   # Measure setpoint
   &message('Recording setpoints.');
   &active_light;
   $vp_val   = &getPV($vp_val_pv,1);
   $hp_val   = &getPV($hp_val_pv,1);
   $postab_set_v = $vp_val;
   $postab_set_h = $hp_val;
   &putPV($postab_set_v_pv,$postab_set_v,0);
   &putPV($postab_set_h_pv,$postab_set_h,0);
   $last_pos_v = &getPV($vV_pv,1);
   $last_pos_h = &getPV($hV_pv,1);
   shift @last_arr_v;
   shift @last_arr_h;
   push @last_arr_v, $last_pos_v;
   push @last_arr_h, $last_pos_h;

   # Change flag
   $zero_measured = 1;
   $stpt_measured = 1;
   &putPV($postab_zero_chk_pv,$zero_measured,1);
   &putPV($postab_stpt_chk_pv,$stpt_measured,1);
   $postab_zero = 0;
   $postab_stpt = 0;
   &putPV($postab_zero_pv,$postab_zero,1);
   &putPV($postab_stpt_pv,$postab_stpt,1);
   &message('Finished - setpoints.');
   &kl_check;						# This requires KL server
   &att_check($att);

   # Return feedback to paused state
   $postab_mode = 1;
}

# ---------------------------------------------------------------

sub feedback {

   # Declare local variables
   my ($del_kl, $del_att, $adj_result, $dac_err);

   &message('Preparing for active feedback.');

   # Check for KL changes, setpoints, and offsets
   $del_kl = &kl_check;
   if ($del_kl) {
      &message('Keithley gains have changed!');
      &message('Measure offsets, setpoints');
      &message('Then start feedback');
      $zero_measured = 0;
      $stpt_measured = 0;
      &putPV($postab_zero_chk_pv,$zero_measured,1);
      &putPV($postab_stpt_chk_pv,$stpt_measured,1);
      $postab_mode = 1;
      return;
   }
   $zero_measured = &getPV($postab_zero_chk_pv,1);
   if ($zero_measured == 0) {
      &message('Offsets have not been measured!');
      &message('Measure offsets, setpoints');
      &message('Then start feedback');
      $stpt_measured = 0;
      &putPV($postab_stpt_chk_pv,$stpt_measured,1);
      $postab_mode = 1;
      return;
   }
   $stpt_measured = &getPV($postab_stpt_chk_pv,1);
   if ($stpt_measured == 0) {
      &message('Setpoints have not been measured.');
      &message('Measure setpoints first');
      &message('Then start feedback');
      &putPV($postab_stpt_chk_pv,$stpt_measured,1);
      $postab_mode = 1;
      return;
   }

   # Check intensity
   &message('Checking for intensity.');
   $v_lo_val = &getPV($v_lo_val_pv,1);
   $h_lo_val = &getPV($h_lo_val_pv,1);
   if ($v_lo_val||$h_lo_val) {
      &message('Intensity out of range!');
      &message('Feedback will investigate.');
      $postab_mode = 7;
      return;
   }

   # If using DCM for vertical, check mostab here
   if ($postab_v_driver) {&check_mostab(0);}

   # Start feedback
   &message('Feedback starting now!');
   &mode($postab_mode);

   # While feedback is running...
   while(1) {
      # Take a short break - is 0.25 sec too long?
      select(undef,undef,undef,0.25);

      # Check for user input
      (@pvs) = &pv_check(40);
      if ($pvs[3] == 1) {
         if ($postab_v_driver) {&check_mostab(2)};
	 $postab_mode = 99;
	 return;
      }                     # stop
      if ($pvs[0] == 3) {
         if ($postab_v_driver) {&check_mostab(2)};
	 $postab_mode = 1;
	 return;
      }                     # pause = 3; pause
      if (($pvs[2] == 1) || ($pvs[1] == 1)) {
         $postab_zero = 0;
         $postab_stpt = 0;
         &putPV($postab_zero_pv,$postab_zero,1);
         &putPV($postab_stpt_pv,$postab_stpt,1);
         &message('Pause feedback.');
         &message('Then measure offsets or setpoints.');
         &message('Or tweak setpoints in control panel.');
      }

      # Check for Keithley changes
      $del_kl = &kl_check;
      if ($del_kl) {
	 $act_rqs_clc = 0;
	 &putPV($act_rqs_clc_pv,$act_rqs_clc,1);
	 if ($postab_v_driver) {&check_mostab(2)};
         # Set voltages to last good value
         &message('Keithley gains have changed!');
         &message('Feedback has been stopped.');
         &message('Setting voltages to last good value.');
         &putPV($vV_pv,$last_arr_v[0],1);
         &putPV($hV_pv,$last_arr_h[0],1);
         &message('Offsets have not been measured.');
         &message('Setpoints have not been measured.');
         &message('Measure offsets, setpoints.');
         &message('Then continue.');
         $zero_measured = 0;
         $stpt_measured = 0;
         &putPV($postab_zero_chk_pv,$zero_measured,1);
         &putPV($postab_stpt_chk_pv,$stpt_measured,1);
	 $postab_mode = 1;
	 return;
      }

      # Check mostab
      if ($postab_v_driver) {&check_mostab(1)};

      # Check attenuators
      $del_att = &att_check($att);
      if ($del_att) {
	 $act_rqs_clc = 0;
	 &putPV($act_rqs_clc_pv,$act_rqs_clc,1);
	 if ($postab_v_driver) {&check_mostab(2)};
         &message('Attenuators have been changed.');
         &message('Feedback has been stopped.');
         # Set voltages to last good value
         &message('Setting voltages to last good value.');
         &putPV($vV_pv,$last_arr_v[0],1);
         &putPV($hV_pv,$last_arr_h[0],1);
         &message('Check beam alignment.');
         &message('Remeasure setpoints.');
         $stpt_measured = 0;
         &putPV($postab_stpt_chk_pv,$stpt_measured,1);
         $postab_mode = 1;
	 return;
      }

      #Check DAC voltages
      $dac_err = &dac_check;
      if ($dac_err == 1) {
         $postab_mode = 1;
	 return;
      }
      elsif ($dac_err == 2) {
         &message('Cause for limit is undetermined.');
         &message('Consider shifting BD');
         &message('   and remeasuring setpoints.');
         $postab_mode = 1;
	 return;
      }
      elsif ($dac_err == 3) {
         &message('Keithley gains have changed!');
         &message('Offsets have not been measured.');
         &message('Setpoints have not been measured.');
         &message('Measure offsets, setpoints.');
         &message('Then continue.');
         $zero_measured = 0;
         $stpt_measured = 0;
         &putPV($postab_zero_chk_pv,$zero_measured,1);
         &putPV($postab_stpt_chk_pv,$stpt_measured,1);
	 $postab_mode = 1;
	 return;
      }
      elsif ($dac_err == 4) {
         &message('Attenuators have been changed.');
         &message('Check beam alignment.');
         &message('Remeasure setpoints.');
         $stpt_measured = 0;
         &putPV($postab_stpt_chk_pv,$stpt_measured,1);
         $postab_mode = 1;
	 return;
      }
      elsif ($dac_err == 5) {
         $postab_mode = 7;
         return;
      }

      # Check for beam
      $v_lo_val = &getPV($v_lo_val_pv,1);
      $h_lo_val = &getPV($h_lo_val_pv,1);
      # Emergency stop if no beam, else record 'good' piezo voltages
      if ($v_lo_val||$h_lo_val) {
	 $act_rqs_clc = 0;
	 &putPV($act_rqs_clc_pv,$act_rqs_clc,1);
	 if ($postab_v_driver) {&check_mostab(2)};
         &message('Intensity out of range.');
         &message('Investigating.');
	 $postab_mode = 7;
	 return;
      }
      else {
         $last_pos_v = &getPV($vV_pv,1);
         $last_pos_h = &getPV($hV_pv,1);
         shift @last_arr_v;
         shift @last_arr_h;
         push @last_arr_v, $last_pos_v;
         push @last_arr_h, $last_pos_h;
      }

      # Blink active light
      &active_light;
   }
}

# ---------------------------------------------------------------

sub pause_no_beam {

   # README -----------------------------------------------------------
   #
   # Pause_no_beam originally just handled low intensity conditions.
   # However, high intensity is also a concern for positional feedback.
   # This sub now deals with both conditions.
   #
   # ------------------------------------------------------------------

   # Declare local variables
   my (@status_pvs);
   my ($last_mode, $permit, $err, $E_diff, $E_check);
   my ($att_err, $kl_err, $sr_err, $permit_err, $a_err, $wbsh_err, $mbsh_err, $E_mov_err, $ID_mov_err, $E_err);
   my $E_adj = 0.120;							# 120 chosen based on current ID/DCM offsets - 7 Feb 2007
   my $max_E_tol = 0.090;						# 90 = current theoretical max + 10 - 7 Feb 2007

   &mode($postab_mode);

   # Set voltages to last good value
   &message('Setting voltages to last good value.');
   &putPV($vV_pv,$last_arr_v[0],1);
   &putPV($hV_pv,$last_arr_h[0],1);

   &message('Investigating intensity problem.');

   # Check to see if intensity is too high
   $up_val = &getPV($up_val_pv,1);
   $dn_val = &getPV($dn_val_pv,1);
   $ob_val = &getPV($ob_val_pv,1);
   $ib_val = &getPV($ib_val_pv,1);
   if (($up_val > $good_hi) || ($up_val > $good_hi) || ($up_val > $good_hi) || ($up_val > $good_hi)) {
      &message('Keithley gain is too high');
      &message('Please select new gain');
      &message('Remeasure offsets, setpoints');
      $postab_mode = 1;
      return;
   }

   # Wait before determining cause of lost beam
   select(undef,undef,undef,2);
   (@pvs) = &pv_check(70);
   if ($pvs[3] == 1) {$postab_mode = 99; return;}                     # stop
   if ($pvs[0] == 3) {$postab_mode = 1;  return;}                     # pause = 3; pause
   if (($pvs[2] == 1) || ($pvs[1] == 1)) {
      $postab_zero = 0;
      $postab_stpt = 0;
      &putPV($postab_zero_pv,$postab_zero,1);
      &putPV($postab_stpt_pv,$postab_stpt,1);
      &message('Pause feedback.');
      &message('Then measure offsets or setpoints.');
   }

   # Other errors not fatal
   $E_mov	= &getPV($E_mov_pv,1);
   if ($E_mov)	{$E_mov_err = 0;}
   else		{$E_mov_err = 1;}
   $ID_mov_err	= &getPV($ID_mov_pv,1);
   $mbsh_err	= &getPV($shutter_mono_pv,1);
   $wbsh_err	= &getPV($shutter_white_pv,1);
   $a_err	= &getPV($shutter_a_pv,1);
   $permit	= &getPV($shutter_permit_pv,1);
   if ($permit)	{$permit_err = 0;}
   else		{$permit_err = 1;}
   $sr_err	= &getPV($sr_cmp_val_pv,1);
   if (defined $E_ID_pv) {
      $E_ID	= &getPV($E_ID_pv,1);
      $E	= &getPV($E_pv,1);
      $E_diff = $E_ID - $E;
      $E_check = abs($E_diff - $E_adj);
      if ($E_check > $max_E_tol) {$E_err = 1}
      else                       {$E_err = 0}
   }
   else {$E_err = 0};
   $err = $E_mov_err + $ID_mov_err + $mbsh_err + $wbsh_err + $a_err + $permit_err + $sr_err + $E_err;

   if ($E_mov_err)  {&message('Waiting for DCM.');}
   if ($ID_mov_err) {&message('Waiting for ID.');}
   if ($mbsh_err)   {&message('Waiting for MBSH.');}
   if ($wbsh_err)   {&message('Waiting for WBSH.');}
   if ($a_err)      {&message('Waiting for A shutter.');}
   if ($permit_err) {&message('Waiting for shutter permit.');}
   if ($sr_err)     {&message('Waiting for SR current.');}
   if ($E_err)      {&message('ID/DCM mismatch.');}

   if ($err > 0) {
      while ($err > 0) {
         (@pvs) = &pv_check(71);
         if ($pvs[3] == 1) {$postab_mode = 99; return;}                     # stop
         if ($pvs[0] == 3) {$postab_mode = 1;  return;}                     # pause = 3; pause
         if (($pvs[2] == 1) || ($pvs[1] == 1)) {
            $postab_zero = 0;
            $postab_stpt = 0;
            &putPV($postab_zero_pv,$postab_zero,1);
            &putPV($postab_stpt_pv,$postab_stpt,1);
            &message('Pause feedback.');
            &message('Then measure offsets or setpoints.');
         }

         $E_mov		= &getPV($E_mov_pv,1);
         if ($E_mov)	{$E_mov_err = 0;}
         else		{$E_mov_err = 1;}
         $ID_mov_err	= &getPV($ID_mov_pv,1);
         $mbsh_err	= &getPV($shutter_mono_pv,1);
         $wbsh_err	= &getPV($shutter_white_pv,1);
         $a_err		= &getPV($shutter_a_pv,1);
         $permit	= &getPV($shutter_permit_pv,1);
         if ($permit)	{$permit_err = 0;}
         else		{$permit_err = 1;}
         $sr_err	= &getPV($sr_cmp_val_pv,1);
         if (defined $E_ID_pv) {
            $E_ID	= &getPV($E_ID_pv,1);
            $E	= &getPV($E_pv,1);
            $E_diff = $E_ID - $E;
            $E_check = abs($E_diff - $E_adj);
            if ($E_check > $max_E_tol) {$E_err = 1}
            else                       {$E_err = 0}
         }
         else {$E_err = 0};

         $err = $E_mov_err + $ID_mov_err + $mbsh_err + $wbsh_err + $a_err + $permit_err + $sr_err + $E_err;
         (@pvs) = &pv_check(71);
         if ($pvs[3] == 1) {$postab_mode = 99; return;}                     # stop
         if ($pvs[0] == 3) {$postab_mode = 1;  return;}                     # pause = 3; pause
         if (($pvs[2] == 1) || ($pvs[1] == 1)) {
            $postab_zero = 0;
            $postab_stpt = 0;
            &putPV($postab_zero_pv,$postab_zero,1);
            &putPV($postab_stpt_pv,$postab_stpt,1);
            &message('Pause feedback.');
            &message('Then measure offsets or setpoints.');
         }
         &active_light;
      }
      select(undef,undef,undef,2);
      &message('All errors cleared.');
   }
   $v_lo_val = &getPV($v_lo_val_pv,1);
   $h_lo_val = &getPV($h_lo_val_pv,1);
   if ($v_lo_val||$h_lo_val) {
      &message('Intensity still out of range');
      &message('Is Keithley gain correct?');
      &message('Pausing.');
      $postab_mode = 1;
      return;
   }

   $postab_mode = 4;
   &message('Resuming feedback.');
}

# ---------------------------------------------------------------

sub stop_script {

   # Declare local variables
   my ($current_time, $stop_type, $limit1_rate_pv, $limit2_rate_pv);

   # Need to stop feedback (unless stopping due to duplicate script)
   $stop_type = shift (@_);
   $current_time = scalar localtime;
   &clear_monitors;
   if ($stop_type) {
      if (defined $aborted) {     # for repeating stop_script calls
         print STDOUT '!!! Brute-force aborting on repeated stop request!'."\n";
         exit;
      }
      $aborted = 1;
      &mode(99);			# This line stops feedback!
      &message('Clearing monitors.');

      # Set update rates
      &message('Setting update rates.');
      $limit1_rate_pv = $beamline.':SoftPID1_limits.SCAN';
      $limit2_rate_pv = $beamline.':SoftPID2_limits.SCAN';
      &putPV($limit1_rate_pv,0,1);		# Passive
      &putPV($limit2_rate_pv,0,1);		# Passive
      &putPV($calc_rate_pv,0,1);		# Passive
      &putPV($calc2_rate_pv,0,1);		# Passive
      &putPV($v_pid_rate_pv,0,1);		# Passive
      &putPV($h_pid_rate_pv,0,1);		# Passive
      &putPV($gs_up_rate_pv,6,1);		# 1 sec
      &putPV($gs_dn_rate_pv,6,1);		# 1 sec
      &putPV($gs_ob_rate_pv,6,1);		# 1 sec
      &putPV($gs_ib_rate_pv,6,1);		# 1 sec
      &putPV($vV_check_rate_pv,6,1);		# 1 sec
      &putPV($hV_check_rate_pv,6,1);		# 1 sec
      $zero_measured = 0;
      $stpt_measured = 0;
      &putPV($postab_zero_chk_pv,$zero_measured,1);
      &putPV($postab_stpt_chk_pv,$stpt_measured,1);

      &message('');
      &message('Exiting.');
   }

   # Close logfile
   if (defined $filehandle) {print DAT "\n";};
   if (defined $filehandle) {print DAT 'Script stopped at:'."\n".$current_time."\n\n";};
   close(DAT);
   exit;
}

# ---------------------------------------------------------------
# ---------------------------------------------------------------
#
#
#
#
#
#
#
#            Secondary sub routines
#
#
#
#
#
#
#
# ---------------------------------------------------------------
# ---------------------------------------------------------------

sub kl_set {

   # Declare local variables here:
   my ($gains, $offsets, $old_gain, $timeout, $return);
   my ($max_int, $init_shutter_mono, $gn_chk);
   my ($set_type, $up_del, $dn_del, $ob_del, $ib_del, $max_del);
   my ($kl_off_up_err, $kl_off_dn_err, $kl_off_ob_err, $kl_off_ib_err);
   my ($kl_gain_up, $kl_gain_dn, $kl_gain_ob, $kl_gain_ib);
   my ($rqs_exp, $gain_report);

   # Outline:
   #    Set gains
   #    Measure offsets if requested or needed
   #    Include PVchecks and timeouts

   &message('KL adjustment - checking gains, offsets');

   # Determine whether offsets will be measured
   $set_type = shift(@_);
   if ($set_type == 1) {
      $zero_measured = 0;					# Setting to zero forces measurement of offsets
      &putPV($postab_zero_chk_pv,$zero_measured,1);
   }

   # Get requested gain
   $postab_kl_rqs = &getPV($postab_kl_rqs_pv,1);
   if    ($postab_kl_rqs == 0) {$rqs_exp = '1e3';}
   elsif ($postab_kl_rqs == 1) {$rqs_exp = '1e4';}
   elsif ($postab_kl_rqs == 2) {$rqs_exp = '1e5';}
   elsif ($postab_kl_rqs == 3) {$rqs_exp = '1e6';}
   elsif ($postab_kl_rqs == 4) {$rqs_exp = '1e7';}
   elsif ($postab_kl_rqs == 5) {$rqs_exp = '1e8';}
   elsif ($postab_kl_rqs == 6) {$rqs_exp = '1e9';}
   elsif ($postab_kl_rqs == 7) {$rqs_exp = '1e10';}
   $gain_report = 'Requested gain: '.$rqs_exp;
   &message($gain_report);

   # Check current gains compared to requested gain
   $gn_chk = 0;
   $kl_gain_up = &getPV($kl_gain_up_pv,1);
   $kl_gain_dn = &getPV($kl_gain_dn_pv,1);
   $kl_gain_ob = &getPV($kl_gain_ob_pv,1);
   $kl_gain_ib = &getPV($kl_gain_ib_pv,1);
   if ($kl_gain_up != $postab_kl_rqs) {$gn_chk +=1;}
   if ($kl_gain_dn != $postab_kl_rqs) {$gn_chk +=1;}
   if ($kl_gain_ob != $postab_kl_rqs) {$gn_chk +=1;}
   if ($kl_gain_ib != $postab_kl_rqs) {$gn_chk +=1;}

   # If any gains need setting, set them all (it doesn't add too much time)
   if ($gn_chk > 0) {
      &message('Making requested change');
      $zero_measured = 0;
      &putPV($postab_zero_chk_pv,$zero_measured,1);
      $kl_gain_rqs = $postab_kl_rqs;
      &putPV($kl_gain_up_rqs_pv,$kl_gain_rqs,1);
      &putPV($kl_gain_dn_rqs_pv,$kl_gain_rqs,1);
      &putPV($kl_gain_ob_rqs_pv,$kl_gain_rqs,1);
      &putPV($kl_gain_ib_rqs_pv,$kl_gain_rqs,1);
      $kl_set = 1;
      &putPV($kl_set_pv,1,1);
      select(undef,undef,undef,0.1);
      &timeout(0);
      $timeout = 0;
      while ($kl_set == 1) {
         $kl_set = &getPV($kl_set_pv,1);
         $timeout = &timeout(30);				# If KL changes aren't complete within 30 sec, script times out.
	 if ($timeout) {
            $stpt_measured = 0;
            $zero_measured = 0;
            &putPV($postab_zero_chk_pv,$zero_measured,1);
            &putPV($postab_stpt_chk_pv,$stpt_measured,1);
            &message('KL timeout.');
            &message('Please investigate.');
	    $postab_pause = 3;
	    &putPV($postab_pause_pv,$postab_pause,1);
	    return(0);
	 }
         (@pvs) = &pv_check(81);
         if ($pvs[3] == 1) {$postab_mode = 99; return(0);}		# stop
         if ($pvs[0] == 3) {$postab_mode = 1;  return(0);}		# pause = 3; pause
         if (($pvs[2] == 1) || ($pvs[1] == 1)) {
            $postab_zero = 0;
            $postab_stpt = 0;
            &putPV($postab_zero_pv,$postab_zero,1);
            &putPV($postab_stpt_pv,$postab_stpt,1);
            &message('Pause feedback.');
            &message('Then measure offsets or setpoints.');
         }
         select(undef,undef,undef,0.5);
   }  }
   else {&message('No gain changes required');}

   # PV Check - Look for user input
   (@pvs) = &pv_check(81);
   if ($pvs[3] == 1) {$postab_mode = 99; return(0);}		# stop
   if ($pvs[0] == 3) {$postab_mode = 1;  return(0);}		# pause = 3; pause
   if (($pvs[2] == 1) || ($pvs[1] == 1)) {
      $postab_zero = 0;
      $postab_stpt = 0;
      &putPV($postab_zero_pv,$postab_zero,1);
      &putPV($postab_stpt_pv,$postab_stpt,1);
      &message('Please wait for pause.');
      &message('Then measure offsets or setpoints.');
   }



   # Determine if offsets need to be measured
   $zero_measured = &getPV($postab_zero_chk_pv,1);
   if ($zero_measured == 0) {$offsets = 1;}



   # If offsets requested or needed, measure here
   if ($offsets) {

      # Close shutter if necessary
      &message('Checking MBSH.');
      $shutter_mono = &getPV($shutter_mono_pv,1);
      $init_shutter_mono = $shutter_mono;
      if ($shutter_mono == 0) {
         $timeout = 0;
         if (($beamline eq '23i') || ($beamline eq '23o')) {
            &message('Closing shutter.');
	    Pezca::PutDouble($mbsh_close_pv,1);		# Added due to unknown EPICS problem.  PutDouble returns error, but error isn't obvious
            #&putPV($mbsh_close_pv,1,1);
	    &timeout(0);
       	    $shutter_mono = &getPV($shutter_mono_pv,1);
	    while (($shutter_mono == 0)  && ($timeout == 0)) {
	       $shutter_mono = &getPV($shutter_mono_pv,1);
               (@pvs) = &pv_check(51);
               if ($pvs[3] == 1) {$postab_mode = 99; return(0);}		# stop
               if ($pvs[0] == 3) {$postab_mode = 1;  return(0);}		# pause = 3; pause
               if (($pvs[2] == 1) || ($pvs[1] == 1)) {
                  $postab_zero = 0;
                  $postab_stpt = 0;
                  &putPV($postab_zero_pv,$postab_zero,1);
                  &putPV($postab_stpt_pv,$postab_stpt,1);
                  &message('Pause feedback.');
                  &message('Then measure offsets or setpoints.');
               }
	       $timeout = &timeout(30);
               if ($timeout) {&message('Problem while closing shutter');}
	       select(undef,undef,undef,0.5);
	 }  }
         if ((($beamline ne '23i') && ($beamline ne '23o')) || ($timeout)) {
            &message('Please close MBSH manually!');
      	    while ($shutter_mono == 0) {
	       $shutter_mono = &getPV($shutter_mono_pv,1);
               (@pvs) = &pv_check(51);
               if ($pvs[3] == 1) {$postab_mode = 99; return(0);}		# stop
               if ($pvs[0] == 3) {$postab_mode = 1;  return(0);}		# pause = 3; pause
               if (($pvs[2] == 1) || ($pvs[1] == 1)) {
                  $postab_zero = 0;
                  $postab_stpt = 0;
                  &putPV($postab_zero_pv,$postab_zero,1);
                  &putPV($postab_stpt_pv,$postab_stpt,1);
                  &message('Pause feedback.');
                  &message('Then measure offsets or setpoints.');
               }
	       select(undef,undef,undef,0.5);
      }  }  }

      # Actual adjustment of offsets here
      $max_del = 1;
      while ($max_del > 0.015) {

         # Calculate deviation from 'ideal' offset of 0.025 V
         &message('Measuring offsets.');
	 $gs_up = &getPV($gs_up_pv,1);
         $gs_dn = &getPV($gs_dn_pv,1);
         $gs_ob = &getPV($gs_ob_pv,1);
         $gs_ib = &getPV($gs_ib_pv,1);
         $up_del  = $gs_up - 0.025;
         $dn_del  = $gs_dn - 0.025;
         $ob_del  = $gs_ob - 0.025;
         $ib_del  = $gs_ib - 0.025;

         # Calculate change in KL offset that is needed to produce desired voltage offset
         $kl_off_up = &getPV($kl_off_up_pv,1);
         $kl_off_dn = &getPV($kl_off_dn_pv,1);
         $kl_off_ob = &getPV($kl_off_ob_pv,1);
         $kl_off_ib = &getPV($kl_off_ib_pv,1);
         $kl_off_up_rqs = $kl_off_up - $up_del;
         $kl_off_dn_rqs = $kl_off_dn - $dn_del;
         $kl_off_ob_rqs = $kl_off_ob - $ob_del;
         $kl_off_ib_rqs = $kl_off_ib - $ib_del;

         # Change KL offset
         &message('Setting offsets to 0.025 V.');
         &putPV($kl_off_up_rqs_pv,$kl_off_up_rqs,1);
         &putPV($kl_off_dn_rqs_pv,$kl_off_dn_rqs,1);
         &putPV($kl_off_ob_rqs_pv,$kl_off_ob_rqs,1);
         &putPV($kl_off_ib_rqs_pv,$kl_off_ib_rqs,1);
         $kl_set = 1;
         &putPV($kl_set_pv,1,1);
         select(undef,undef,undef,0.1);
         &timeout(0);
         while ($kl_set == 1) {
            $kl_set = &getPV($kl_set_pv,1);
            $timeout = &timeout(30);				# If KL changes aren't complete within 30 sec, script times out.
            if ($timeout == 1) {
               $stpt_measured = 0;
               $zero_measured = 0;
               &putPV($postab_zero_chk_pv,$zero_measured,1);
               &putPV($postab_stpt_chk_pv,$stpt_measured,1);
               &message('KL timeout.');
               &message('Please investigate.');
	       $postab_pause = 3;
	       &putPV($postab_pause_pv,$postab_pause,1);
	       return(0);
            }
            (@pvs) = &pv_check(50);
            if ($pvs[3] == 1) {$postab_mode = 99; return(0);}		# stop
            if ($pvs[0] == 3) {$postab_mode = 1;  return(0);}		# pause = 3; pause
            if (($pvs[2] == 1) || ($pvs[1] == 1)) {
               $postab_zero = 0;
               $postab_stpt = 0;
               &putPV($postab_zero_pv,$postab_zero,1);
               &putPV($postab_stpt_pv,$postab_stpt,1);
               &message('Pause feedback.');
               &message('Then measure offsets or setpoints.');
            }
	    select(undef,undef,undef,0.5);
         }

         # Verify results
         $gs_up = 0;
         $gs_dn = 0;
         $gs_ob = 0;
         $gs_ib = 0;

         for (my $samples = 0; $samples < 20; $samples++) {
	    $gs_up += &getPV($gs_up_pv,1);
            $gs_dn += &getPV($gs_dn_pv,1);
            $gs_ob += &getPV($gs_ob_pv,1);
            $gs_ib += &getPV($gs_ib_pv,1);
	    select(undef,undef,undef,0.1);
            print STDOUT $samples.' '.$gs_up."\n";
         }

         $gs_up /= 20;
         $gs_dn /= 20;
         $gs_ob /= 20;
         $gs_ib /= 20;

	 $up_del  = $gs_up - 0.025;
         $dn_del  = $gs_dn - 0.025;
         $ob_del  = $gs_ob - 0.025;
         $ib_del  = $gs_ib - 0.025;
         $max_del = List::Util::max(abs($up_del),abs($dn_del),abs($ob_del),abs($ib_del));

         # Verify KL set correctly
         $kl_off_up = &getPV($kl_off_up_pv,1);
         $kl_off_dn = &getPV($kl_off_dn_pv,1);
         $kl_off_ob = &getPV($kl_off_ob_pv,1);
         $kl_off_ib = &getPV($kl_off_ib_pv,1);
         $kl_off_up_err = abs($kl_off_up - $kl_off_up_rqs);
         $kl_off_dn_err = abs($kl_off_dn - $kl_off_dn_rqs);
         $kl_off_ob_err = abs($kl_off_ob - $kl_off_ob_rqs);
         $kl_off_ib_err = abs($kl_off_ib - $kl_off_ib_rqs);
         if (($kl_off_up_err>0.011)||
             ($kl_off_dn_err>0.011)||
      	     ($kl_off_ob_err>0.011)||
	     ($kl_off_ib_err>0.011)) {
                &message('Error setting Keithleys.');
                $postab_pause = 3;
                &putPV($postab_pause_pv,$postab_pause,1);
                return(0);
         }
      }

      # Record offsets
      &message('Recording offsets.');
      $up_0_clc = $gs_up;
      $dn_0_clc = $gs_dn;
      $ob_0_clc = $gs_ob;
      $ib_0_clc = $gs_ib;
      &putPV($up_0_clc_pv,$up_0_clc,0);
      &putPV($dn_0_clc_pv,$dn_0_clc,0);
      &putPV($ob_0_clc_pv,$ob_0_clc,0);
      &putPV($ib_0_clc_pv,$ib_0_clc,0);

      # Restore shutter to original state
      if ($init_shutter_mono == 0) {
         $timeout = 0;
         if (($beamline eq '23i') || ($beamline eq '23o')) {
            &message('Opening shutter.');
            Pezca::PutDouble($mbsh_open_pv,1);		# Added due to unknown EPICS problem.  PutDouble returns error, but error isn't obvious
	    #&putPV($mbsh_open_pv,1,1);
            $shutter_mono = &getPV($shutter_mono_pv,1);
	    &timeout(0);
	    while (($shutter_mono == 1) && ($timeout == 0)) {
	       $shutter_mono = &getPV($shutter_mono_pv,1);
               (@pvs) = &pv_check(51);
               if ($pvs[3] == 1) {$postab_mode = 99; return(0);}		# stop
               if ($pvs[0] == 3) {$postab_mode = 1;  return(0);}		# pause = 3; pause
               if (($pvs[2] == 1) || ($pvs[1] == 1)) {
                  $postab_zero = 0;
                  $postab_stpt = 0;
                  &putPV($postab_zero_pv,$postab_zero,1);
                  &putPV($postab_stpt_pv,$postab_stpt,1);
                  &message('Pause feedback.');
                  &message('Then measure offsets or setpoints.');
               }
	       $timeout = &timeout(30);
               if ($timeout) {&message('Problem while opening shutter');}
	       select(undef,undef,undef,0.5);
	    }
         }
         if ((($beamline ne '23i') && ($beamline ne '23o')) || ($timeout)) {
            &message('Please open MBSH manually!');
	    while ($shutter_mono == 1) {
	       $shutter_mono = &getPV($shutter_mono_pv,1);
               (@pvs) = &pv_check(51);
               if ($pvs[3] == 1) {$postab_mode = 99; return(0);}		# stop
               if ($pvs[0] == 3) {$postab_mode = 1;  return(0);}		# pause = 3; pause
               if (($pvs[2] == 1) || ($pvs[1] == 1)) {
                  $postab_zero = 0;
                  $postab_stpt = 0;
                  &putPV($postab_zero_pv,$postab_zero,1);
                  &putPV($postab_stpt_pv,$postab_stpt,1);
                  &message('Pause feedback.');
                  &message('Then measure offsets or setpoints.');
               }
	    select(undef,undef,undef,0.5);
            }
	 }
         &message('Shutter open.');
      }

      # Change PV to reflect that offsets are measured
      $zero_measured = 1;
      &putPV($postab_zero_chk_pv,$zero_measured,1);
      &message('Offsets complete.');
   }

   else {&message('Offset measurement unneeded');}
   $return = 1;
}

# ---------------------------------------------------------------

sub init_vars {

   # Set initial variable values
   #
   # As of v. 4.01, initialization of vars related directly to the vertical
   # has been moved to another sub

      ##############################################################################################
      ##############################################################################################

      # Configuration variables (these are most likely to be changed by staff)

         # Beamline information (may want to use Sergey's common subs or make my own?)
         # $beamline            = '23i';        # Now using Sergey's common subs

         # Logfile information
         $outdir                = '/home/gmca/shared/logfiles/';
         $outfile               = 'pstb_'.$beamline;

         # Connections
         $gs_up_ch              = 5;            # ADC channel for: GS up
         $gs_dn_ch              = 6;            #                  GS down
         $gs_ob_ch              = 7;            #                  GS outboard
         $gs_ib_ch              = 8;            #                  GS inboard
         $hV_check_ch           = 15;           #                  voltage verification - piezo horz
         $hV_ch                 = 3;            #                  voltage - piezo horz
         #$vV_check_ch          = 14;           #                  voltage verification - piezo vert    #moved
         #$vV_ch                = 2;            # DAC channel for: voltage - piezo vert                 #moved

         # Feedback frequency
         $rate			= 9;		# 9 corresponds to 10 Hz

         # Low intensity shutdown level (under which feedback will halt)
         $lo_cutoff             = .25;          # Express as string
         $hi_cutoff             = 8;            # Express as string
         # The 'Good' range for the KL output of GS BPM diodes
	 $good_lo 		= 0.25;		# Express as number
	 $good_hi 		= 8;		# Express as number

	 # DAC/ADC mismatch tolerance
	 $dac_adc_err		= 0.05;

	 # Low ring current cutoff (under which an error *may* be triggered)
         $low_sr_cutoff         = 50;           # Express as a string

	 # In addition, the SR current may be read directly or indirectly.
	 # If directly, uncomment the next line.
         # $sr_current_pv       = 'S:SRcurrentAI.VAL';

	 # Time between messages
	 $message_time = 0.25;

      ##############################################################################################
      ##############################################################################################

      # Misc variables
         $host 			= $ENV{HOSTNAME};
         $user 			= $ENV{USER};
         @last_arr_v		= (5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5);
         @last_arr_h		= (5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5);
	 $zero_measured		= 0;
	 $stpt_measured		= 0;
	 $quitting		= 0;					# This is changed to one after a ca error; helps script avoid endless loops

      # Variables for PV names
         $postab_pv             = $beamline.':postab:';
         $pid_pv                = $beamline.':SoftPID';
         $pid_v_pv              = $pid_pv.'1.';
         $pid_h_pv              = $pid_pv.'2.';
         $pid_calc_pv           = $pid_pv.'1_incalc.';
         $pid_calc2_pv          = $pid_pv.'2_incalc.';

      # PVs needed for status checks
         $shutter_permit_pv     = 'ACIS:ShutterPermit';
         $shutter_a_pv          = '23ID:FES:status';
         if ($beamline eq "23i")    {$shutter_white_pv  = '23i:IBSH:status';
                                     $shutter_mono_pv   = '23i:MBSHI:status';
                                     $ID_mov_pv         = 'ID23ds:Busy.VAL';
                                     $mbsh_open_pv      = $beamline.':V2000:6';
                                     $mbsh_close_pv     = $beamline.':V2000:7';
                                     $E_ID_pv           = 'ID23ds:Energy.VAL';}
         elsif ($beamline eq "23o") {$shutter_white_pv  = '23ID:FES:status';
                                     $shutter_mono_pv   = '23o:MBSHO:status';
                                     $ID_mov_pv         = 'ID23us:Busy.VAL';
                                     $mbsh_open_pv      = $beamline.':V2000:2';
                                     $mbsh_close_pv     = $beamline.':V2000:3';
                                     $E_ID_pv           = 'ID23us:Energy.VAL';}
         $E_pv                  = $beamline.':MO:E:ActPos';
         $E_mov_pv              = $beamline.':MO:En:InPos';

      # Keithleys
         $kl_gain_up_pv         = $beamline.':keithley1:ActGain';
         $kl_gain_dn_pv         = $beamline.':keithley4:ActGain';
         $kl_gain_ob_pv         = $beamline.':keithley2:ActGain';
         $kl_gain_ib_pv         = $beamline.':keithley5:ActGain';
         $kl_gain_up_rqs_pv     = $beamline.':keithley1:RqsGain';
         $kl_gain_dn_rqs_pv     = $beamline.':keithley4:RqsGain';
         $kl_gain_ob_rqs_pv     = $beamline.':keithley2:RqsGain';
         $kl_gain_ib_rqs_pv     = $beamline.':keithley5:RqsGain';
         $kl_off_up_pv          = $beamline.':keithley1:ActOffset';
         $kl_off_dn_pv          = $beamline.':keithley4:ActOffset';
         $kl_off_ob_pv          = $beamline.':keithley2:ActOffset';
         $kl_off_ib_pv          = $beamline.':keithley5:ActOffset';
         $kl_off_up_rqs_pv      = $beamline.':keithley1:RqsOffset';
         $kl_off_dn_rqs_pv      = $beamline.':keithley4:RqsOffset';
         $kl_off_ob_rqs_pv      = $beamline.':keithley2:RqsOffset';
         $kl_off_ib_rqs_pv      = $beamline.':keithley5:RqsOffset';
         $kl_pv                 = $beamline.':keithley:Server';
         $kl_set_pv             = $beamline.':keithley:Set';

      # Attenuators
         $att_pv                = $beamline.':bo0:ch';
	 $att			= 0;
         $att_00_pv             = $att_pv.'0';
         $att_01_pv             = $att_pv.'1';
         $att_02_pv             = $att_pv.'2';
         $att_03_pv             = $att_pv.'3';
         $att_04_pv             = $att_pv.'4';
         $att_05_pv             = $att_pv.'5';
         $att_06_pv             = $att_pv.'6';
         $att_07_pv             = $att_pv.'7';
         $att_08_pv             = $att_pv.'8';
         $att_09_pv             = $att_pv.'9';
         $att_10_pv             = $att_pv.'10';
         $att_11_pv             = $att_pv.'11';
         $att_12_pv             = $att_pv.'12';
         $att_13_pv             = $att_pv.'13';
         $att_14_pv             = $att_pv.'14';
         $att_15_pv             = $att_pv.'15';

      # DAC
         # Setup of $vV_pv performed below
         $hV_pv                 = $beamline.':DAC2_'.$hV_ch.'.VAL';

      # ADC
	 # GS Up
         $gs_up_pv              = $beamline.':ADC_'.$gs_up_ch.'.VAL';
         $gs_up_rate_pv         = $beamline.':ADC_'.$gs_up_ch.'.SCAN';
         $gs_up_name_pv         = $beamline.':ADC_'.$gs_up_ch.'.DESC';
         $gs_up_name		= "GS Up";
         #GS Dn
         $gs_dn_pv              = $beamline.':ADC_'.$gs_dn_ch.'.VAL';
         $gs_dn_rate_pv         = $beamline.':ADC_'.$gs_dn_ch.'.SCAN';
         $gs_dn_name_pv         = $beamline.':ADC_'.$gs_dn_ch.'.DESC';
         $gs_dn_name            = 'GS Dn';
         # GS OB
         $gs_ob_pv              = $beamline.':ADC_'.$gs_ob_ch.'.VAL';
         $gs_ob_rate_pv         = $beamline.':ADC_'.$gs_ob_ch.'.SCAN';
         $gs_ob_name_pv         = $beamline.':ADC_'.$gs_ob_ch.'.DESC';
         $gs_ob_name            = 'GS OB';
         # GS IB
         $gs_ib_pv              = $beamline.':ADC_'.$gs_ib_ch.'.VAL';
         $gs_ib_rate_pv         = $beamline.':ADC_'.$gs_ib_ch.'.SCAN';
         $gs_ib_name_pv         = $beamline.':ADC_'.$gs_ib_ch.'.DESC';
         $gs_ib_name            = 'GS IB';
         # Check on DAC vertical voltage output set below
         # Check on DAC horizontal voltage output
         $hV_check_pv           = $beamline.':ADC_'.$hV_check_ch.'.VAL';
         $hV_check_rate_pv      = $beamline.':ADC_'.$hV_check_ch.'.SCAN';
         $hV_check_name_pv      = $beamline.':ADC_'.$hV_check_ch.'.DESC';
         $hV_check_name         = 'Hor V Check';

      # Postab PVs
         $postab_active_pv      = $postab_pv.'active';
	 $postab_active		= 1;
         $postab_status_pv      = $postab_pv.'status';
         $postab_status		= 0;
         $postab_stline_pv      = $postab_pv.'statusline';
         $postab_stline         = '';
         $postab_sttime_pv      = $postab_pv.'statustime';
         $postab_sttime         = '';
         $postab_pause_pv       = $postab_pv.'pause';
         $postab_pause		= 3;
         $postab_pause_h_pv     = $postab_pv.'pause_h';
         $postab_pause_h	= 0;
         $postab_pause_v_pv     = $postab_pv.'pause_v';
         $postab_pause_v	= 0;
         $postab_zero_pv        = $postab_pv.'zero';            # Should "zero" be measured?
	 $postab_zero		= 0;
         $postab_stpt_pv        = $postab_pv.'stpt';            # Should "setpoint" be measured?
	 $postab_stpt		= 0;
         $postab_v_driver_pv    = $postab_pv.'v_driver';
         #$postab_zero_chk	= $zero_measured;
         $postab_zero_chk_pv    = $postab_pv.'zero_chk';        # Has "zero" been measured?
	 #$postab_stpt_chk	= $stpt_measured;
         $postab_stpt_chk_pv    = $postab_pv.'stpt_chk';        # Has "setpoint" been measured?
         $postab_query_pv       = $postab_pv.'query';           # Is feedback running?
	 $postab_query		= 0;
         $postab_stop_pv        = $postab_pv.'stop';
         $postab_stop		= 0;
         $postab_mode_pv        = $postab_pv.'mode';
	 $postab_mode		= 0;
         $postab_set_v_pv       = $postab_pv.'set_v';
         $postab_set_h_pv       = $postab_pv.'set_h';
         $postab_del_v_pv       = $postab_set_v_pv.'Inc';
         $postab_del_h_pv       = $postab_set_h_pv.'Inc';
         if (!defined $sr_current_pv) {$sr_current_pv = $beamline.':SRcurrent'};                # An indirect reading of the ring current.
         $message1_pv           = $postab_pv.'message1';
         $message2_pv           = $postab_pv.'message2';
         $message3_pv           = $postab_pv.'message3';
         $message4_pv           = $postab_pv.'message4';
         $message5_pv           = $postab_pv.'message5';
         $message1              = '';
         $message2              = '';
         $message3              = '';
         $message4              = '';
         $message5              = '';
         $postab_kl_rqs_pv      = $postab_pv.'kl_rqs';
	 $postab_kl_rqs		= 8;

      # Mostab PVs
         $mostab_pv             = $beamline.':mostab:';
         $mostab_pause_pv       = $mostab_pv.'pause';
	 $mostab_pause		= 0;
	 $mostab_pause_init	= 0;

      # Vertical voltage driver vars
         # Grab user preference for vertical position driver
         # This is either the DCM piezo or the VFM piezo
         $postab_v_driver = &getPV($postab_v_driver_pv,1);

         if ($postab_v_driver == 1) {
            # These are the settings for using the DCM piezo for feedback
            $vV_check_ch        = 13;           # ADC channel for vertical piezo voltage verification
            $vV_ch              = 1;            # DAC channel for vertical piezo voltage
            $v_kp		= -0.5;		# Vertical PID parameters
            $v_ki		= 80;
            $v_kd		= 0.01;
            $vV_pv              = $beamline.':DAC1_'.$vV_ch.'.VAL';
         } else {
            $postab_v_driver	= 0;		# 0 and 1 should be the only possibilities
            # These are the settings for using the VFM piezo for feedback
            $vV_check_ch        = 14;           # ADC channel for vertical piezo voltage verification
            $vV_ch              = 2;            # DAC channel for vertical piezo voltage
            $v_kp		= -1;		# Vertical PID parameters
            $v_ki		= 20;
            $v_kd		= 0.01;
            $vV_pv              = $beamline.':DAC2_'.$vV_ch.'.VAL';
         }
         # Vertical voltage channels
         #$vV_pv                = $beamline.':DAC1_'.$vV_ch.'.VAL';
         $vV_check_pv           = $beamline.':ADC_'.$vV_check_ch.'.VAL';
         $vV_check_rate_pv      = $beamline.':ADC_'.$vV_check_ch.'.SCAN';
         $vV_check_name_pv      = $beamline.':ADC_'.$vV_check_ch.'.DESC';
         $vV_check_name = 'Ver V Check';

      # PID parameter PVs
         $v_kp_pv               = $pid_v_pv.'KP';
         $h_kp_pv               = $pid_h_pv.'KP';
         $h_kp			= -1;
         $v_ki_pv               = $pid_v_pv.'KI';
         $h_ki_pv               = $pid_h_pv.'KI';
         $h_ki			= 20;
         $v_kd_pv               = $pid_v_pv.'KD';
         $h_kd_pv               = $pid_h_pv.'KD';
         $h_kd			= 0.01;
         $v_i_pv                = $pid_v_pv.'I';
         $v_i			= 0;
         $h_i_pv                = $pid_h_pv.'I';
         $h_i			= 0;
         $v_dt_pv               = $pid_v_pv.'DT';
         $h_dt_pv               = $pid_h_pv.'DT';
         $v_ll_pv               = $pid_v_pv.'DRVL';
         $v_ll			= 0;
         $h_ll_pv               = $pid_h_pv.'DRVL';
         $h_ll			= 0;
         $v_hl_pv               = $pid_v_pv.'DRVH';
         $v_hl			= 10;
         $h_hl_pv               = $pid_h_pv.'DRVH';
         $h_hl			= 10;
         #$v_kp			= -1;		#moved
         #$v_ki			= 20;		#moved
         #$v_kd			= 0.01;		#moved

      # PID calc 1 PVs
         # Refresh rate
         $calc_rate_pv          = $pid_calc_pv.'SCAN';
         $calc_rate		= $rate;
         # A - Up zero
         $up_0_cmt_pv           = $pid_calc_pv.'CMTA';
         $up_0_cmt              = 'Up zero';
         $up_0_ipv_pv           = $pid_calc_pv.'INPA';
         $up_0_ipv              = '';
         $up_0_clc_pv           = $pid_calc_pv.'CLCA';
         $up_0_clc		= 0;
         $up_0_val_pv           = $pid_calc_pv.'A';
         $up_0_out_pv           = $pid_calc_pv.'OUTA';
         $up_0_out              = '';
         # B - Dn zero
         $dn_0_cmt_pv           = $pid_calc_pv.'CMTB';
         $dn_0_cmt              = 'Dn zero';
         $dn_0_ipv_pv           = $pid_calc_pv.'INPB';
         $dn_0_ipv              = '';
         $dn_0_clc_pv           = $pid_calc_pv.'CLCB';
         $dn_0_clc		= 0;
         $dn_0_val_pv           = $pid_calc_pv.'B';
         $dn_0_out_pv           = $pid_calc_pv.'OUTB';
         $dn_0_out              = '';
         # I - OB zero
         $ob_0_cmt_pv           = $pid_calc_pv.'CMTI';
         $ob_0_cmt              = 'OB zero';
         $ob_0_ipv_pv           = $pid_calc_pv.'INPI';
         $ob_0_ipv              = '';
         $ob_0_clc_pv           = $pid_calc_pv.'CLCI';
         $ob_0_clc		= 0;
         $ob_0_val_pv           = $pid_calc_pv.'I';
         $ob_0_out_pv           = $pid_calc_pv.'OUTI';
         $ob_0_out		= "";
         # J - IB zero
         $ib_0_cmt_pv           = $pid_calc_pv.'CMTJ';
         $ib_0_cmt		= "IB zero";
         $ib_0_ipv_pv           = $pid_calc_pv.'INPJ';
         $ib_0_ipv		= "";
         $ib_0_clc_pv           = $pid_calc_pv.'CLCJ';
         $ib_0_clc		= 0;
         $ib_0_val_pv           = $pid_calc_pv.'J';
         $ib_0_out_pv           = $pid_calc_pv.'OUTJ';
         $ib_0_out              = '';
         # C - Up value
         $up_cmt_pv             = $pid_calc_pv.'CMTC';
         $up_cmt                = 'Up';
         $up_ipv_pv             = $pid_calc_pv.'INPC';
         $up_ipv		= $gs_up_pv;
         $up_clc_pv             = $pid_calc_pv.'CLCC';
         $up_clc		= "";
         $up_val_pv             = $pid_calc_pv.'C';
         $up_out_pv             = $pid_calc_pv.'OUTC';
         $up_out		= "";
         # D - Dn value
         $dn_cmt_pv             = $pid_calc_pv.'CMTD';
         $dn_cmt                = 'Dn';
         $dn_ipv_pv             = $pid_calc_pv.'INPD';
         $dn_ipv		= $gs_dn_pv;
         $dn_clc_pv             = $pid_calc_pv.'CLCD';
         $dn_clc                = '';
         $dn_val_pv             = $pid_calc_pv.'D';
         $dn_out_pv             = $pid_calc_pv.'OUTD';
         $dn_out                = '';
         # K - OB value
         $ob_cmt_pv             = $pid_calc_pv.'CMTK';
         $ob_cmt                = 'OB';
         $ob_ipv_pv             = $pid_calc_pv.'INPK';
         $ob_ipv		= $gs_ob_pv;
         $ob_clc_pv             = $pid_calc_pv.'CLCK';
         $ob_clc                = '';
         $ob_val_pv             = $pid_calc_pv.'K';
         $ob_out_pv             = $pid_calc_pv.'OUTK';
         $ob_out                = '';
         # L - IB value
         $ib_cmt_pv             = $pid_calc_pv.'CMTL';
         $ib_cmt                = 'IB';
         $ib_ipv_pv             = $pid_calc_pv.'INPL';
         $ib_ipv		= $gs_ib_pv;
         $ib_clc_pv             = $pid_calc_pv.'CLCL';
         $ib_clc                = '';
         $ib_val_pv             = $pid_calc_pv.'L';
         $ib_out_pv             = $pid_calc_pv.'OUTL';
         $ib_out                = '';
         # E - V sum
         $vs_cmt_pv             = $pid_calc_pv.'CMTE';
         $vs_cmt                = 'V sum';
         $vs_ipv_pv             = $pid_calc_pv.'INPE';
         $vs_ipv                = '';
         $vs_clc_pv             = $pid_calc_pv.'CLCE';
         $vs_clc                = '(C-A)+(D-B)';
         $vs_val_pv             = $pid_calc_pv.'E';
         $vs_out_pv             = $pid_calc_pv.'OUTE';
         $vs_out                = '';
         # F - V pos
         $vp_cmt_pv             = $pid_calc_pv.'CMTF';
         $vp_cmt                = 'V pos';
         $vp_ipv_pv             = $pid_calc_pv.'INPF';
         $vp_ipv                = '';
         $vp_clc_pv             = $pid_calc_pv.'CLCF';
         $vp_clc                = '((C-A)-(D-B))/E';
         $vp_val_pv             = $pid_calc_pv.'F';
         $vp_out_pv             = $pid_calc_pv.'OUTF';
         $vp_out                = '';
         # M - H sum
         $hs_cmt_pv             = $pid_calc_pv.'CMTM';
         $hs_cmt                = 'H sum';
         $hs_ipv_pv             = $pid_calc_pv.'INPM';
         $hs_ipv                = '';
         $hs_clc_pv             = $pid_calc_pv.'CLCM';
         $hs_clc                = '(K-I)+(L-J)';
         $hs_val_pv             = $pid_calc_pv.'M';
         $hs_out_pv             = $pid_calc_pv.'OUTM';
         $hs_out                = '';
         # N - H pos
         $hp_cmt_pv             = $pid_calc_pv.'CMTN';
         $hp_cmt                = 'H pos';
         $hp_ipv_pv             = $pid_calc_pv.'INPN';
         $hp_ipv                = '';
         $hp_clc_pv             = $pid_calc_pv.'CLCN';
         $hp_clc                = '((K-I)-(L-J))/M';
         $hp_val_pv             = $pid_calc_pv.'N';
         $hp_out_pv             = $pid_calc_pv.'OUTN';
         $hp_out                = '';
         # G - V setpoint
         $vsp_cmt_pv            = $pid_calc_pv.'CMTG';
         $vsp_cmt               = 'V setpt';
         $vsp_ipv_pv            = $pid_calc_pv.'INPG';
         $vsp_ipv		= $postab_set_v_pv;
         $vsp_clc_pv            = $pid_calc_pv.'CLCG';
         $vsp_clc               = '';
         $vsp_val_pv            = $pid_calc_pv.'G';
         $vsp_out_pv            = $pid_calc_pv.'OUTG';
         $vsp_out               = $pid_v_pv.'VAL';
         # O - H setpoint
         $hsp_cmt_pv            = $pid_calc_pv.'CMTO';
         $hsp_cmt               = 'H setpt';
         $hsp_ipv_pv            = $pid_calc_pv.'INPO';
         $hsp_ipv		= $postab_set_h_pv;
         $hsp_clc_pv            = $pid_calc_pv.'CLCO';
         $hsp_clc               = '';
         $hsp_val_pv            = $pid_calc_pv.'O';
         $hsp_out_pv            = $pid_calc_pv.'OUTO';
         $hsp_out               = $pid_h_pv.'VAL';
         # H - Vertical cutoff calc
         $v_lo_cmt_pv           = $pid_calc_pv.'CMTH';
         $v_lo_cmt              = 'V int bad?';
         $v_lo_ipv_pv           = $pid_calc_pv.'INPH';
         $v_lo_ipv              = '';
         $v_lo_clc_pv           = $pid_calc_pv.'CLCH';
         $v_lo_clc              = '(C<'.$lo_cutoff.')|(D<'.$lo_cutoff.')|(C>'.$hi_cutoff.')|(D>'.$hi_cutoff.')';
         $v_lo_val_pv           = $pid_calc_pv.'H';
         $v_lo_out_pv           = $pid_calc_pv.'OUTH';
         $v_lo_out              = '';
         # P - Horizontal cutoff calc
         $h_lo_cmt_pv           = $pid_calc_pv.'CMTP';
         $h_lo_cmt              = 'H int bad?';
         $h_lo_ipv_pv           = $pid_calc_pv.'INPP';
         $h_lo_ipv              = '';
         $h_lo_clc_pv           = $pid_calc_pv.'CLCP';
         $h_lo_clc              = '(K<'.$lo_cutoff.')|(L<'.$lo_cutoff.')|(K>'.$hi_cutoff.')|(L>'.$hi_cutoff.')';
         $h_lo_val_pv           = $pid_calc_pv.'P';
         $h_lo_out_pv           = $pid_calc_pv.'OUTP';
         $h_lo_out              = '';

      # PID calc2 PVs
         # Refresh rate
         $calc2_rate_pv         = $pid_calc2_pv.'SCAN';
         $calc2_rate		= $rate;
         # A - SR current
         $sr_cmt_pv             = $pid_calc2_pv.'CMTA';
         $sr_cmt                = 'SR current';
         $sr_ipv_pv             = $pid_calc2_pv.'INPA';
         $sr_ipv		= $sr_current_pv;
         $sr_clc_pv             = $pid_calc2_pv.'CLCA';
         $sr_clc                = '';
         $sr_val_pv             = $pid_calc2_pv.'A';
         $sr_out_pv             = $pid_calc2_pv.'OUTA';
         $sr_out                = '';
	 # B - SR current comparison
         $sr_cmp_cmt_pv         = $pid_calc2_pv.'CMTB';
         $sr_cmp_cmt            = 'SR current comp';
         $sr_cmp_ipv_pv         = $pid_calc2_pv.'INPB';
         $sr_cmp_ipv            = '';
         $sr_cmp_clc_pv         = $pid_calc2_pv.'CLCB';
         $sr_cmp_clc            = 'a<'.$low_sr_cutoff;
         $sr_cmp_val_pv         = $pid_calc2_pv.'B';
         $sr_cmp_out_pv         = $pid_calc2_pv.'OUTB';
         $sr_cmp_out            = '';
	 # C - Vert DAC ctrl
         $vV_cmt_pv             = $pid_calc2_pv.'CMTC';
         $vV_cmt                = 'Vert DAC ctrl';
         $vV_ipv_pv             = $pid_calc2_pv.'INPC';
         $vV_ipv		= $vV_pv;
         $vV_clc_pv             = $pid_calc2_pv.'CLCC';
         $vV_clc                = '';
         $vV_val_pv             = $pid_calc2_pv.'C';
         $vV_out_pv             = $pid_calc2_pv.'OUTC';
         $vV_out                = '';
	 # D - Vert ADC rdbk
         $vV_adc_cmt_pv         = $pid_calc2_pv.'CMTD';
         $vV_adc_cmt            = 'Vert ADC rdbk';
         $vV_adc_ipv_pv         = $pid_calc2_pv.'INPD';
         $vV_adc_ipv		= $vV_check_pv;
         $vV_adc_clc_pv         = $pid_calc2_pv.'CLCD';
         $vV_adc_clc            = '';
         $vV_adc_val_pv         = $pid_calc2_pv.'D';
         $vV_adc_out_pv         = $pid_calc2_pv.'OUTD';
         $vV_adc_out            = '';
	 # E - Vert DAC-ADC comp
         $vV_err_cmt_pv         = $pid_calc2_pv.'CMTE';
         $vV_err_cmt            = 'Vert V error';
         $vV_err_ipv_pv         = $pid_calc2_pv.'INPE';
         $vV_err_ipv            = '';
         $vV_err_clc_pv         = $pid_calc2_pv.'CLCE';
         $vV_err_clc            = 'ABS(C-D)>'.$dac_adc_err;
         $vV_err_val_pv         = $pid_calc2_pv.'E';
         $vV_err_out_pv         = $pid_calc2_pv.'OUTE';
         $vV_err_out            = '';
	 # F - Horz DAC ctrl
         $hV_cmt_pv             = $pid_calc2_pv.'CMTF';
         $hV_cmt                = 'Horz DAC ctrl';
         $hV_ipv_pv             = $pid_calc2_pv.'INPF';
         $hV_ipv		= $hV_pv;
         $hV_clc_pv             = $pid_calc2_pv.'CLCF';
         $hV_clc                = '';
         $hV_val_pv             = $pid_calc2_pv.'F';
         $hV_out_pv             = $pid_calc2_pv.'OUTF';
         $hV_out                = '';
	 # G - Horz ADC rdbk
         $hV_adc_cmt_pv         = $pid_calc2_pv.'CMTG';
         $hV_adc_cmt            = 'Horz ADC rdbk';
         $hV_adc_ipv_pv         = $pid_calc2_pv.'INPG';
         $hV_adc_ipv		= $hV_check_pv;
         $hV_adc_clc_pv         = $pid_calc2_pv.'CLCG';
         $hV_adc_clc            = '';
         $hV_adc_val_pv         = $pid_calc2_pv.'G';
         $hV_adc_out_pv         = $pid_calc2_pv.'OUTG';
         $hV_adc_out            = '';
	 # H - Horz DAC-ADC comp
         $hV_err_cmt_pv         = $pid_calc2_pv.'CMTH';
         $hV_err_cmt            = 'Horz V error';
         $hV_err_ipv_pv         = $pid_calc2_pv.'INPH';
         $hV_err_ipv            = '';
         $hV_err_clc_pv         = $pid_calc2_pv.'CLCH';
         $hV_err_clc            = 'ABS(F-G)>'.$dac_adc_err;
         $hV_err_val_pv         = $pid_calc2_pv.'H';
         $hV_err_out_pv         = $pid_calc2_pv.'OUTH';
         $hV_err_out            = '';
	 # I - Active request
         $act_rqs_cmt_pv        = $pid_calc2_pv.'CMTI';
         $act_rqs_cmt           = 'Fdbk active?';
         $act_rqs_ipv_pv        = $pid_calc2_pv.'INPI';
         $act_rqs_ipv           = '';
         $act_rqs_clc_pv        = $pid_calc2_pv.'CLCI';                 # Putting 1 here actually starts feedback; user can still control V and H pause
         $act_rqs_clc           = '0';
         $act_rqs_val_pv        = $pid_calc2_pv.'I';
         $act_rqs_out_pv        = $pid_calc2_pv.'OUTI';
         $act_rqs_out           = '';
	 # J - Vertical pause
         $v_rqs_cmt_pv          = $pid_calc2_pv.'CMTJ';
         $v_rqs_cmt             = 'Vert paused?';
         $v_rqs_ipv_pv          = $pid_calc2_pv.'INPJ';
         $v_rqs_ipv		= $postab_pause_v_pv;
         $v_rqs_clc_pv          = $pid_calc2_pv.'CLCJ';
         $v_rqs_clc             = '';
         $v_rqs_val_pv          = $pid_calc2_pv.'J';
         $v_rqs_out_pv          = $pid_calc2_pv.'OUTJ';
         $v_rqs_out             = '';
	 # K - Horizontal pause
         $h_rqs_cmt_pv          = $pid_calc2_pv.'CMTK';
         $h_rqs_cmt             = 'Horz paused?';
         $h_rqs_ipv_pv          = $pid_calc2_pv.'INPK';
         $h_rqs_ipv		= $postab_pause_h_pv;
         $h_rqs_clc_pv          = $pid_calc2_pv.'CLCK';
         $h_rqs_clc             = '';
         $h_rqs_val_pv          = $pid_calc2_pv.'K';
         $h_rqs_out_pv          = $pid_calc2_pv.'OUTK';
         $h_rqs_out             = '';
	 # L - Vertical On
         $v_on_cmt_pv           = $pid_calc2_pv.'CMTL';
         $v_on_cmt              = 'Vert Cmd';
         $v_on_ipv_pv           = $pid_calc2_pv.'INPL';
         $v_on_ipv              = '';
         $v_on_clc_pv           = $pid_calc2_pv.'CLCL';
         $v_on_clc              = 'I*(!J)';
         $v_on_val_pv           = $pid_calc2_pv.'L';
         $v_on_out_pv           = $pid_calc2_pv.'OUTL';
         #$v_on_out		= $v_on_pv;				# This needs to be defined below.  It relies on PID control PVs
	 # M - Horizontal On
         $h_on_cmt_pv           = $pid_calc2_pv.'CMTM';
         $h_on_cmt              = 'Horz Cmd';
         $h_on_ipv_pv           = $pid_calc2_pv.'INPM';
         $h_on_ipv              = '';
         $h_on_clc_pv           = $pid_calc2_pv.'CLCM';
         $h_on_clc              = 'I*(!K)';
         $h_on_val_pv           = $pid_calc2_pv.'M';
         $h_on_out_pv           = $pid_calc2_pv.'OUTM';
         #$h_on_out		= $h_on_pv;				# This needs to be defined below.  It relies on PID control PVs
	 # N - Voltage limits
         $V_limit_cmt_pv        = $pid_calc2_pv.'CMTN';
         $V_limit_cmt           = 'DAC at limit?';
         $V_limit_ipv_pv        = $pid_calc2_pv.'INPN';
         $V_limit_ipv           = '';
         $V_limit_clc_pv        = $pid_calc2_pv.'CLCN';
         $V_limit_clc           = '(C<0.1)|(C>9.9)';
         $V_limit_val_pv        = $pid_calc2_pv.'N';
         $V_limit_out_pv        = $pid_calc2_pv.'OUTN';
         $V_limit_out           = '';
	 # O - Voltage limits
         $H_limit_cmt_pv        = $pid_calc2_pv.'CMTO';
         $H_limit_cmt           = 'DAC at limit?';
         $H_limit_ipv_pv        = $pid_calc2_pv.'INPO';
         $H_limit_ipv           = '';
         $H_limit_clc_pv        = $pid_calc2_pv.'CLCO';
         $H_limit_clc           = '(F<0.1)|(F>9.9)';
         $H_limit_val_pv        = $pid_calc2_pv.'O';
         $H_limit_out_pv        = $pid_calc2_pv.'OUTO';
         $H_limit_out           = '';
	 # Field P are unassigned and unmodified at this time.

      # PID control PVs
         $v_on_pv               = $pid_v_pv.'FBON';
	 $v_on			= 0;
         $h_on_pv               = $pid_h_pv.'FBON';
         $h_on			= 0;
         $v_pid_rate_pv         = $pid_v_pv.'SCAN';
         $v_pid_rate		= $rate;
         $h_pid_rate_pv         = $pid_h_pv.'SCAN';
         $h_pid_rate		= $rate;
         $v_stpt_pv             = $pid_v_pv.'VAL';
         $h_stpt_pv             = $pid_h_pv.'VAL';
         $v_rdbkpv_pv           = $pid_v_pv.'INP';
         $v_rdbkpv              = $vp_val_pv.' PP';
         $h_rdbkpv_pv           = $pid_h_pv.'INP';
         $h_rdbkpv              = $hp_val_pv.' PP';
         $v_ctrlpv_pv           = $pid_v_pv.'OUTL';
         $v_ctrlpv              = $vV_pv.' PP';
         $h_ctrlpv_pv           = $pid_h_pv.'OUTL';
         $h_ctrlpv              = $hV_pv.' PP';

      # Additional PID calc PVs defined here b/c they rely on PID control PVs
         $v_on_out		= $v_on_pv;
	 $h_on_out		= $h_on_pv;

   # Signal handler
      $SIG{INT}  = 'signal_handler';
      $SIG{QUIT} = 'signal_handler';

   &message('New script started.');
   &message('Variables initialized.');
}

# ---------------------------------------------------------------

sub logfile {

   # Declare local variables
   my (@ct, $sec, $min, $hour, $mday, $mon, $year, $current_time);

   # Open logfile with timestamp in filename;
   &message('Opening logfile.');
   @ct      = localtime(time);	# This is ($sec,$min,$hour,$mday,$mon,$year$,$wday,$yday,$isdst)
   $sec     = $ct[0];
   $min     = $ct[1];
   $hour    = $ct[2];
   $mday    = $ct[3];
   $mon     = $ct[4] + 1;
   $year    = $ct[5] + 1900;
   $outfile = sprintf $outdir.$outfile.'_%2d%2d%2d_%2d%2d%2d.dat', $year,$mon,$mday,$hour,$min,$sec;
   $outfile =~ s/ /0/g;
   $filehandle = open (DAT,'> '.$outfile);                      # open DAT file for writing
   if (!defined $filehandle) {
      &message('Could not open file '.$outfile);
      print STDOUT 'Could not open file '.$outfile."\n";
   }
   select DAT; $|=1;						# set unbuffered output

   # Write logfile header;
   $current_time = scalar localtime;
   if (defined $filehandle) {
      &message('Writing file header.');
      if (defined $filehandle) {
         print DAT 'Postab intensity feedback script started at:'."\n".$current_time."\n\n";
         print DAT 'Started by:'."\n".$host.' '.$user."\n\n";
         print DAT 'Output file:'."\n".$outfile."\n\n";
         if ($postab_v_driver) {print DAT 'Vertical feedback using DCM piezo.'."\n";}
         else                  {print DAT 'Vertical feedback using VFM piezo.'."\n";}
      }
   }
}

# ---------------------------------------------------------------

sub script_check {

   # Declare local variables
   my ($postab_query_err);

   # Check for existing scripts
   &message('Checking for existing script.');
   Pezca::PutDouble($postab_query_pv,1);
   select(undef,undef,undef,5);
   ($postab_query_err, $postab_query) = Pezca::GetDouble($postab_query_pv);
   if ($postab_query != 1) {
      &message('Previous script detected.');
      &message('Halting new script.');
      &stop_script(0);
   }
   &message('No previous scripts detected.');
}

# ---------------------------------------------------------------

sub init_PVs {

   # Set EPICS retry
      my $ca_error = Pezca::SetRetryCount(20);
      if ($ca_error != 0) {
         &message('CA error (EPICS retry)');
         &message('Please investigate.');
         &message('Feedback stopping!');
         $quitting = 1;
         &stop_script(1);
      }

   # Initial mode
   &mode($postab_mode);

   # Initialize PVs
      # Log entry
         &message('Initializing PVs.');
         if ($postab_v_driver) {&message('Vertical position driven by DCM');}
         else                  {&message('Vertical position driven by VFM');}
      # Turn off feedback (should already be off)
         &message('Ensuring feedback off.');
	 &putPV($v_on_pv,$v_on,1);
         &putPV($h_on_pv,$h_on,1);
      # Set up ADC
         &message('Setting up ADC.');
	 &putPV($gs_up_name_pv,$gs_up_name,0);
         &putPV($gs_dn_name_pv,$gs_dn_name,0);
         &putPV($gs_ob_name_pv,$gs_ob_name,0);
         &putPV($gs_ib_name_pv,$gs_ib_name,0);
         &putPV($vV_check_name_pv,$vV_check_name,0);
         &putPV($hV_check_name_pv,$hV_check_name,0);
         &putPV($gs_up_rate_pv,$rate,1);
	 &putPV($gs_dn_rate_pv,$rate,1);
	 &putPV($gs_ob_rate_pv,$rate,1);
	 &putPV($gs_ib_rate_pv,$rate,1);
	 &putPV($vV_check_rate_pv,$rate,1);
	 &putPV($hV_check_rate_pv,$rate,1);
      # Set limit calc fields blank
         &message('Setting limit calc fields blank');
         my ($line, $commentPV, $inputPV, $expressionPV, $outputPV, $basePV);
         for (my $softPID = 1; $softPID < 3; $softPID++) {
            for (my $calcline = 1; $calcline < 17; $calcline++) {
               if    ($calcline == 1)  {$line = 'A';}
               elsif ($calcline == 2)  {$line = 'B';}
               elsif ($calcline == 3)  {$line = 'C';}
               elsif ($calcline == 4)  {$line = 'D';}
               elsif ($calcline == 5)  {$line = 'E';}
               elsif ($calcline == 6)  {$line = 'F';}
               elsif ($calcline == 7)  {$line = 'G';}
               elsif ($calcline == 8)  {$line = 'H';}
               elsif ($calcline == 9)  {$line = 'I';}
               elsif ($calcline == 10) {$line = 'J';}
               elsif ($calcline == 11) {$line = 'K';}
               elsif ($calcline == 12) {$line = 'L';}
               elsif ($calcline == 13) {$line = 'M';}
               elsif ($calcline == 14) {$line = 'N';}
               elsif ($calcline == 15) {$line = 'O';}
               else                    {$line = 'P';}
               $basePV          = $beamline.':SoftPID'.$softPID.'_limits.';
               $commentPV       = $basePV.'CMT'.$line;
               $inputPV         = $basePV.'INP'.$line;
               $expressionPV    = $basePV.'CLC'.$line;
               $outputPV        = $basePV.'OUT'.$line;
               &putPV($commentPV,'',0);
               &putPV($inputPV,'',0);
               &putPV($expressionPV,'',0);
               &putPV($outputPV,'',0);
            }
         }
      # Set up PID calc fields
         &message('Setting PID1 calc fields.');
	 # Refresh rate
         &putPV($calc_rate_pv,$calc_rate,1);
         # Field A
         &putPV($up_0_cmt_pv,$up_0_cmt,0);
         &putPV($up_0_ipv_pv,$up_0_ipv,0);
         &putPV($up_0_clc_pv,$up_0_clc,0);
         &putPV($up_0_out_pv,$up_0_out,0);
         # Field B
         &putPV($dn_0_cmt_pv,$dn_0_cmt,0);
         &putPV($dn_0_ipv_pv,$dn_0_ipv,0);
         &putPV($dn_0_clc_pv,$dn_0_clc,0);
         &putPV($dn_0_out_pv,$dn_0_out,0);
         #Field C
         &putPV($up_cmt_pv,$up_cmt,0);
         &putPV($up_ipv_pv,$up_ipv,0);
         &putPV($up_clc_pv,$up_clc,0);
         &putPV($up_out_pv,$up_out,0);
         #Field D
         &putPV($dn_cmt_pv,$dn_cmt,0);
         &putPV($dn_ipv_pv,$dn_ipv,0);
         &putPV($dn_clc_pv,$dn_clc,0);
         &putPV($dn_out_pv,$dn_out,0);
         #Field E
         &putPV($vs_cmt_pv,$vs_cmt,0);
         &putPV($vs_ipv_pv,$vs_ipv,0);
         &putPV($vs_clc_pv,$vs_clc,0);
         &putPV($vs_out_pv,$vs_out,0);
         #Field F
         &putPV($vp_cmt_pv,$vp_cmt,0);
         &putPV($vp_ipv_pv,$vp_ipv,0);
         &putPV($vp_clc_pv,$vp_clc,0);
         &putPV($vp_out_pv,$vp_out,0);
         #Field G
         &putPV($vsp_cmt_pv,$vsp_cmt,0);
         &putPV($vsp_ipv_pv,$vsp_ipv,0);
         &putPV($vsp_clc_pv,$vsp_clc,0);
         &putPV($vsp_out_pv,$vsp_out,0);
         #Field H
         &putPV($v_lo_cmt_pv,$v_lo_cmt,0);
         &putPV($v_lo_ipv_pv,$v_lo_ipv,0);
         &putPV($v_lo_clc_pv,$v_lo_clc,0);
         &putPV($v_lo_out_pv,$v_lo_out,0);
         # Field I
         &putPV($ob_0_cmt_pv,$ob_0_cmt,0);
         &putPV($ob_0_ipv_pv,$ob_0_ipv,0);
         &putPV($ob_0_clc_pv,$ob_0_clc,0);
         &putPV($ob_0_out_pv,$ob_0_out,0);
         # Field J
         &putPV($ib_0_cmt_pv,$ib_0_cmt,0);
         &putPV($ib_0_ipv_pv,$ib_0_ipv,0);
         &putPV($ib_0_clc_pv,$ib_0_clc,0);
         &putPV($ib_0_out_pv,$ib_0_out,0);
         #Field K
         &putPV($ob_cmt_pv,$ob_cmt,0);
         &putPV($ob_ipv_pv,$ob_ipv,0);
         &putPV($ob_clc_pv,$ob_clc,0);
         &putPV($ob_out_pv,$ob_out,0);
         #Field L
         &putPV($ib_cmt_pv,$ib_cmt,0);
         &putPV($ib_ipv_pv,$ib_ipv,0);
         &putPV($ib_clc_pv,$ib_clc,0);
         &putPV($ib_out_pv,$ib_out,0);
         #Field M
         &putPV($hs_cmt_pv,$hs_cmt,0);
         &putPV($hs_ipv_pv,$hs_ipv,0);
         &putPV($hs_clc_pv,$hs_clc,0);
         &putPV($hs_out_pv,$hs_out,0);
         #Field N
         &putPV($hp_cmt_pv,$hp_cmt,0);
         &putPV($hp_ipv_pv,$hp_ipv,0);
         &putPV($hp_clc_pv,$hp_clc,0);
         &putPV($hp_out_pv,$hp_out,0);
         #Field O
         &putPV($hsp_cmt_pv,$hsp_cmt,0);
         &putPV($hsp_ipv_pv,$hsp_ipv,0);
         &putPV($hsp_clc_pv,$hsp_clc,0);
         &putPV($hsp_out_pv,$hsp_out,0);
         #Field P
         &putPV($h_lo_cmt_pv,$h_lo_cmt,0);
         &putPV($h_lo_ipv_pv,$h_lo_ipv,0);
         &putPV($h_lo_clc_pv,$h_lo_clc,0);
         &putPV($h_lo_out_pv,$h_lo_out,0);
      # Set up PID calc2 fields
         &message('Setting PID2 calc fields.');
         # Refresh rate
         &putPV($calc2_rate_pv,$calc2_rate,1);
         # Field A
         &putPV($sr_cmt_pv,$sr_cmt,0);
         &putPV($sr_ipv_pv,$sr_ipv,0);
         &putPV($sr_clc_pv,$sr_clc,0);
         &putPV($sr_out_pv,$sr_out,0);
         # Field B
         &putPV($sr_cmp_cmt_pv,$sr_cmp_cmt,0);
         &putPV($sr_cmp_ipv_pv,$sr_cmp_ipv,0);
         &putPV($sr_cmp_clc_pv,$sr_cmp_clc,0);
         &putPV($sr_cmp_out_pv,$sr_cmp_out,0);
         # Field C
         &putPV($vV_cmt_pv,$vV_cmt,0);
         &putPV($vV_ipv_pv,$vV_ipv,0);
         &putPV($vV_clc_pv,$vV_clc,0);
         &putPV($vV_out_pv,$vV_out,0);
         # Field D
         &putPV($vV_adc_cmt_pv,$vV_adc_cmt,0);
         &putPV($vV_adc_ipv_pv,$vV_adc_ipv,0);
         &putPV($vV_adc_clc_pv,$vV_adc_clc,0);
         &putPV($vV_adc_out_pv,$vV_adc_out,0);
         # Field E
         &putPV($vV_err_cmt_pv,$vV_err_cmt,0);
         &putPV($vV_err_ipv_pv,$vV_err_ipv,0);
         &putPV($vV_err_clc_pv,$vV_err_clc,0);
         &putPV($vV_err_out_pv,$vV_err_out,0);
         # Field F
         &putPV($hV_cmt_pv,$hV_cmt,0);
         &putPV($hV_ipv_pv,$hV_ipv,0);
         &putPV($hV_clc_pv,$hV_clc,0);
         &putPV($hV_out_pv,$hV_out,0);
         # Field G
         &putPV($hV_adc_cmt_pv,$hV_adc_cmt,0);
         &putPV($hV_adc_ipv_pv,$hV_adc_ipv,0);
         &putPV($hV_adc_clc_pv,$hV_adc_clc,0);
         &putPV($hV_adc_out_pv,$hV_adc_out,0);
         # Field H
         &putPV($hV_err_cmt_pv,$hV_err_cmt,0);
         &putPV($hV_err_ipv_pv,$hV_err_ipv,0);
         &putPV($hV_err_clc_pv,$hV_err_clc,0);
         &putPV($hV_err_out_pv,$hV_err_out,0);
	 # Field I
         &putPV($act_rqs_cmt_pv,$act_rqs_cmt,0);
         &putPV($act_rqs_ipv_pv,$act_rqs_ipv,0);
         &putPV($act_rqs_clc_pv,$act_rqs_clc,0);
         &putPV($act_rqs_out_pv,$act_rqs_out,0);
	 # Field J
         &putPV($v_rqs_cmt_pv,$v_rqs_cmt,0);
         &putPV($v_rqs_ipv_pv,$v_rqs_ipv,0);
         &putPV($v_rqs_clc_pv,$v_rqs_clc,0);
         &putPV($v_rqs_out_pv,$v_rqs_out,0);
	 # Field K
         &putPV($h_rqs_cmt_pv,$h_rqs_cmt,0);
         &putPV($h_rqs_ipv_pv,$h_rqs_ipv,0);
         &putPV($h_rqs_clc_pv,$h_rqs_clc,0);
         &putPV($h_rqs_out_pv,$h_rqs_out,0);
	 # Field L
         &putPV($v_on_cmt_pv,$v_on_cmt,0);
         &putPV($v_on_ipv_pv,$v_on_ipv,0);
         &putPV($v_on_clc_pv,$v_on_clc,0);
         &putPV($v_on_out_pv,$v_on_out,0);
	 # Field M
         &putPV($h_on_cmt_pv,$h_on_cmt,0);
         &putPV($h_on_ipv_pv,$h_on_ipv,0);
         &putPV($h_on_clc_pv,$h_on_clc,0);
         &putPV($h_on_out_pv,$h_on_out,0);
	 # Field N
         &putPV($V_limit_cmt_pv,$V_limit_cmt,0);
         &putPV($V_limit_ipv_pv,$V_limit_ipv,0);
         &putPV($V_limit_clc_pv,$V_limit_clc,0);
         &putPV($V_limit_out_pv,$V_limit_out,0);
	 # Field O
         &putPV($H_limit_cmt_pv,$H_limit_cmt,0);
         &putPV($H_limit_ipv_pv,$H_limit_ipv,0);
         &putPV($H_limit_clc_pv,$H_limit_clc,0);
         &putPV($H_limit_out_pv,$H_limit_out,0);
	 # Field P is unassigned and set blank
         $line                  = 'P';
         $basePV                = $beamline.':SoftPID2_incalc.';
         $commentPV             = $basePV.'CMT'.$line;
         $inputPV               = $basePV.'INP'.$line;
         $expressionPV          = $basePV.'CLC'.$line;
         $outputPV              = $basePV.'OUT'.$line;
         &putPV($commentPV,'',0);
         &putPV($inputPV,'',0);
         &putPV($expressionPV,'',0);
         &putPV($outputPV,'',0);

      # Set up PID control fields
         &message('Setting PID ctrl fields.');
	 # On/Off - This was already set above.
         # Setpoint - This is changed from the PID calc records.
         #          - No change needed here.
         # Readback PV
         &putPV($v_rdbkpv_pv,$v_rdbkpv,0);
         &putPV($h_rdbkpv_pv,$h_rdbkpv,0);
         # Control PV
         &putPV($v_ctrlpv_pv,$v_ctrlpv,0);
         &putPV($h_ctrlpv_pv,$h_ctrlpv,0);
         # Update rate
         &putPV($v_pid_rate_pv,$v_pid_rate,1);
         &putPV($h_pid_rate_pv,$h_pid_rate,1);
      # Set up PID feedback parameters
         &message('Setting PID params.');
	 # KP
         &putPV($v_kp_pv,$v_kp,1);
         &putPV($h_kp_pv,$h_kp,1);
         # KI
         &putPV($v_ki_pv,$v_ki,1);
         &putPV($h_ki_pv,$h_ki,1);
         # KD
         &putPV($v_kd_pv,$v_kd,1);
         &putPV($h_kd_pv,$h_kd,1);
         # I
         &putPV($v_i_pv,$v_i,1);
         &putPV($h_i_pv,$h_i,1);
         # DT - This gets changed when the 'Update rate' field gets changed.
         # LL
         &putPV($v_ll_pv,$v_ll,1);
         &putPV($h_ll_pv,$h_ll,1);
         # HL
         &putPV($v_hl_pv,$v_hl,1);
         &putPV($h_hl_pv,$h_hl,1);
      # Set up postab PVs
         &message('Setting postab PVs.');
	 &putPV($postab_status_pv,$postab_status,1);
	 &putPV($postab_stline_pv,$postab_stline,0);
	 &putPV($postab_sttime_pv,$postab_sttime,0);
	 &putPV($postab_pause_pv,$postab_pause,1);
	 &putPV($postab_pause_v_pv,$postab_pause_v,1);
	 &putPV($postab_pause_h_pv,$postab_pause_h,1);
         #print STDOUT $postab_pause_v_pv,' ',$postab_pause_v,' ',$postab_pause_h_pv,' ',$postab_pause_h,"\n";
	 &putPV($postab_stop_pv,$postab_stop,1);
	 &putPV($postab_mode_pv,$postab_mode,1);
	 &putPV($message1_pv,$message1,0);
	 &putPV($message2_pv,$message2,0);
	 &putPV($message3_pv,$message3,0);
	 &putPV($message4_pv,$message4,0);
	 &putPV($message5_pv,$message5,0);
	 &putPV($postab_zero_chk_pv,$zero_measured,1);
	 &putPV($postab_stpt_chk_pv,$stpt_measured,1);
	 &putPV($postab_active_pv,$postab_active,1);

   # Set monitors on frequently accessed PVs
      &message('Setting monitors.');
      &set_monitors;

   # Capture initial DAC values; verify DAC voltages
      &message('Initial DAC check.');
      &dac_check;
      $vV = &getPV($vV_pv,1);
      $hV = &getPV($hV_pv,1);
      if (($vV<1)||($hV<1)||($vV>9)||($hV>9)) {
         &message('WARNING!!!');
         &message('DAC voltage near limit');
         &message('Consider adjusting to mid-range');
      }

   # Capture initial KL values (this requires KL server)
      &message('Initial KL check.');
      &kl_check;
      &message('Initial attenuation check.');
      &att_check($att);

   # Finished initializing - goto standby
      &message('Done with initialization.');
      &message('Ready.');
}

# ---------------------------------------------------------------

sub putPV {

   # Declare local variables

   my ($pv2change, $new_pv_value, $ca_error, $pvtype);

   # putPV($pv2change,$new_pv_value,$pvtype)
   #
   # $pvtype = 0 => string
   #         = 1 => double
   #
   # This sub not created for arrays.

   $pv2change		= shift (@_);
   $new_pv_value	= shift (@_);
   $pvtype		= shift (@_);
   if ($pvtype == 0)   {$ca_error = Pezca::PutString($pv2change,$new_pv_value)}
   else                {$ca_error = Pezca::PutDouble($pv2change,$new_pv_value)}
   if (($ca_error != 0) && ($quitting = 0)) {
      &message('Caput error, '.$pv2change);
      &message('Please investigate.');
      &message('Feedback stopping!');
      $quitting = 1;
      &stop_script(1);
   }
}

# ---------------------------------------------------------------

sub getPV {

   # Declare local variables

   my ($pv2get, $pvtype, $ca_error, $pv_value);

   # ($pv_value) = getPV($pv2get,$pvtype)
   #
   # $pvtype = 0 => string
   #         = 1 => double
   #
   # This sub not created for arrays.

   $pv2get			= shift (@_);
   $pvtype			= shift (@_);
   if ($pvtype == 0) {($ca_error,$pv_value) = Pezca::GetString($pv2get)}
   else              {($ca_error,$pv_value) = Pezca::GetDouble($pv2get)}
   if ($ca_error != 0) {
      if ($quitting == 0) {
         &message('caget error, '.$pv2get);
         &message('Please investigate.');
         &message('Feedback stopping!');
         $quitting = 1;
         &stop_script(1);
      } else {
         print STDOUT "\n".'*** Error reading PV='.$pv2get."\n\n";
         if ($pvtype == 0) {$pv_value = '';}
         else              {$pv_value = 0;}
      }
   }
   return($pv_value);
}

# ---------------------------------------------------------------

sub set_monitors {

   # Declare local variables

   my ($set_clear);

   # adj_monitor($pv,$pv_type,$set_clear) - set or clear a monitor
   #
   # $pv_type = 0	=> string
   #          = 1	=> double
   # $set_clear = 0	=> clear monitor
   #            = 1	=> set monitor

   $set_clear = 1;

   &adj_monitor($sr_current_pv,1,$set_clear);
   &adj_monitor($shutter_permit_pv,1,$set_clear);
   &adj_monitor($shutter_a_pv,1,$set_clear);
   &adj_monitor($shutter_white_pv,1,$set_clear);
   &adj_monitor($shutter_mono_pv,1,$set_clear);
   &adj_monitor($ID_mov_pv,1,$set_clear);
   &adj_monitor($E_mov_pv,1,$set_clear);
   &adj_monitor($postab_pause_pv,1,$set_clear);
   &adj_monitor($postab_pause_v_pv,1,$set_clear);
   &adj_monitor($postab_pause_h_pv,1,$set_clear);
   &adj_monitor($postab_zero_pv,1,$set_clear);
   &adj_monitor($postab_stpt_pv,1,$set_clear);
   &adj_monitor($postab_query_pv,1,$set_clear);
   &adj_monitor($postab_stop_pv,1,$set_clear);
   &adj_monitor($v_lo_val_pv,1,$set_clear);
   &adj_monitor($h_lo_val_pv,1,$set_clear);
   &adj_monitor($att_00_pv,1,$set_clear);
   &adj_monitor($att_01_pv,1,$set_clear);
   &adj_monitor($att_02_pv,1,$set_clear);
   &adj_monitor($att_03_pv,1,$set_clear);
   &adj_monitor($att_04_pv,1,$set_clear);
   &adj_monitor($att_05_pv,1,$set_clear);
   &adj_monitor($att_06_pv,1,$set_clear);
   &adj_monitor($att_07_pv,1,$set_clear);
   &adj_monitor($att_08_pv,1,$set_clear);
   &adj_monitor($att_09_pv,1,$set_clear);
   &adj_monitor($att_10_pv,1,$set_clear);
   &adj_monitor($att_11_pv,1,$set_clear);
   &adj_monitor($att_12_pv,1,$set_clear);
   &adj_monitor($att_13_pv,1,$set_clear);
   &adj_monitor($att_14_pv,1,$set_clear);
   &adj_monitor($att_15_pv,1,$set_clear);
   &adj_monitor($kl_set_pv,1,$set_clear);
   &adj_monitor($vV_err_val_pv,1,$set_clear);
   &adj_monitor($hV_err_val_pv,1,$set_clear);
   &adj_monitor($V_limit_val_pv,1,$set_clear);
   &adj_monitor($H_limit_val_pv,1,$set_clear);
   &adj_monitor($kl_gain_up_pv,1,$set_clear);
   &adj_monitor($kl_gain_dn_pv,1,$set_clear);
   &adj_monitor($kl_gain_ob_pv,1,$set_clear);
   &adj_monitor($kl_gain_ib_pv,1,$set_clear);
   &adj_monitor($mostab_pause_pv,1,$set_clear);

}

# ---------------------------------------------------------------

sub clear_monitors {

   # Declare local variables

   my ($set_clear);

   # adj_monitor($pv,$pv_type,$set_clear)
   #
   # $pv = pv of interest
   # $pv_type = 0	=> string
   #          = 1	=> double
   # $set_clear = 0	=> clear
   #            = 1	=> set

   $set_clear = 0;

   &adj_monitor($sr_current_pv,1,$set_clear);
   &adj_monitor($shutter_permit_pv,1,$set_clear);
   &adj_monitor($shutter_a_pv,1,$set_clear);
   &adj_monitor($shutter_white_pv,1,$set_clear);
   &adj_monitor($shutter_mono_pv,1,$set_clear);
   &adj_monitor($ID_mov_pv,1,$set_clear);
   &adj_monitor($E_mov_pv,1,$set_clear);
   &adj_monitor($postab_pause_pv,1,$set_clear);
   &adj_monitor($postab_pause_v_pv,1,$set_clear);
   &adj_monitor($postab_pause_h_pv,1,$set_clear);
   &adj_monitor($postab_zero_pv,1,$set_clear);
   &adj_monitor($postab_stpt_pv,1,$set_clear);
   &adj_monitor($postab_query_pv,1,$set_clear);
   &adj_monitor($postab_stop_pv,1,$set_clear);
   &adj_monitor($v_lo_val_pv,1,$set_clear);
   &adj_monitor($h_lo_val_pv,1,$set_clear);
   &adj_monitor($att_00_pv,1,$set_clear);
   &adj_monitor($att_01_pv,1,$set_clear);
   &adj_monitor($att_02_pv,1,$set_clear);
   &adj_monitor($att_03_pv,1,$set_clear);
   &adj_monitor($att_04_pv,1,$set_clear);
   &adj_monitor($att_05_pv,1,$set_clear);
   &adj_monitor($att_06_pv,1,$set_clear);
   &adj_monitor($att_07_pv,1,$set_clear);
   &adj_monitor($att_08_pv,1,$set_clear);
   &adj_monitor($att_09_pv,1,$set_clear);
   &adj_monitor($att_10_pv,1,$set_clear);
   &adj_monitor($att_11_pv,1,$set_clear);
   &adj_monitor($att_12_pv,1,$set_clear);
   &adj_monitor($att_13_pv,1,$set_clear);
   &adj_monitor($att_14_pv,1,$set_clear);
   &adj_monitor($att_15_pv,1,$set_clear);
   &adj_monitor($kl_set_pv,1,$set_clear);
   &adj_monitor($vV_err_val_pv,1,$set_clear);
   &adj_monitor($hV_err_val_pv,1,$set_clear);
   &adj_monitor($V_limit_val_pv,1,$set_clear);
   &adj_monitor($H_limit_val_pv,1,$set_clear);
   &adj_monitor($kl_gain_up_pv,1,$set_clear);
   &adj_monitor($kl_gain_dn_pv,1,$set_clear);
   &adj_monitor($kl_gain_ob_pv,1,$set_clear);
   &adj_monitor($kl_gain_ib_pv,1,$set_clear);
   &adj_monitor($mostab_pause_pv,1,$set_clear);

}

# ---------------------------------------------------------------

sub adj_monitor {

   # Declare local variables

   my ($pv, $pvtype, $set_clear, $ca_error);

   # adj_monitor($pv,$pv_type,$set_clear)
   #
   # $pv = pv of interest
   # $pv_type = 0	=> string
   #          = 1	=> double
   # $set_clear = 0	=> clear
   #            = 1	=> set

   $pv		= shift (@_);
   $pvtype	= shift (@_);
   $set_clear	= shift (@_);
   if ($set_clear) {
      if ($pvtype) {$ca_error = Pezca::SetMonitorDouble($pv)}
      else         {$ca_error = Pezca::SetMonitorString($pv)}
   }
   else {
      if ($pvtype) {$ca_error = Pezca::ClearMonitorDouble($pv)}
      else         {$ca_error = Pezca::ClearMonitorString($pv)}
   }
   if (($ca_error != 0) && ($quitting = 0)) {
      &message('CA error, monitor');
      print STDOUT $pv."\n";
      &message($pv);
      &message('Please investigate.');
      &message('Feedback stopping!');
      $quitting = 1;
      &stop_script(1);
   }
}

# ---------------------------------------------------------------

sub pv_check {

   # Declare local variables

   my (@pvs, $query_option);

   $query_option = shift (@_);

   # This sub is passed the current feedback mode and other relevant PVs.
   # If any control PVs change, this sub will direct the script to a new mode.
   #
   # 23x:postab:query is an unused pv.  Originally, it took the place of status below
   #
   #
   #	Current		Expected values of...
   #	Mode		mode	pause	pause_v	pause_h	zero	stpt	(query	stop)
   #----------------------------------------------------------------------------
   #	Pause		1	3	1	1	0	0	(0	0)
   #	Pause_v		3	1	1	0	0	0	(0	0)
   #	Pause_h		2	2	0	1	0	0	(0	0)
   #	Feedback on	4	0	0	0	0	0	(0	0)
   #	Zero		5	4	1	1	0	0	(0	0)
   #	Setpoint	6	4	1	1	0	0	(0	0)
   #	Pause_no_beam	7	?	?	?	0	0	(0	0)
   #    Live KL change 	8	?	?	?	0	0	(0	0)

   $postab_pause	= &getPV($postab_pause_pv,  1);
   $postab_zero		= &getPV($postab_zero_pv,   1);
   $postab_stpt		= &getPV($postab_stpt_pv,   1);
   $postab_query	= &getPV($postab_query_pv,  1);
   $postab_stop		= &getPV($postab_stop_pv,   1);

   # PV changed: query
   if ($postab_query) {
      $postab_query = 0;
      &putPV($postab_query_pv,$postab_query,1);
      &status_write($query_option);
   }

   @pvs = ($postab_pause, $postab_zero, $postab_stpt, $postab_stop);
   return (@pvs);

}

# ---------------------------------------------------------------

sub att_check {

   # Declare local variables
   my ($att_00, $att_01, $att_02, $att_03);
   my ($att_04, $att_05, $att_06, $att_07);
   my ($att_08, $att_09, $att_10, $att_11);
   my ($att_12, $att_13, $att_14, $att_15);
   my ($old_att, $new_att, $del);

   $old_att = shift (@_);
   $new_att = 0;
   $del = 0;

   $att_00 = &getPV($att_00_pv,1);
   $new_att += $att_00 * 1;
   $att_01 = &getPV($att_01_pv,1);
   $new_att += $att_01 * 2;
   $att_02 = &getPV($att_02_pv,1);
   $new_att += $att_02 * 4;
   $att_03 = &getPV($att_03_pv,1);
   $new_att += $att_03 * 8;
   $att_04 = &getPV($att_04_pv,1);
   $new_att += $att_04 * 16;
   $att_05 = &getPV($att_05_pv,1);
   $new_att += $att_05 * 32;
   $att_06 = &getPV($att_06_pv,1);
   $new_att += $att_06 * 64;
   $att_07 = &getPV($att_07_pv,1);
   $new_att += $att_07 * 128;
   $att_08 = &getPV($att_08_pv,1);
   $new_att += $att_08 * 256;
   $att_09 = &getPV($att_09_pv,1);
   $new_att += $att_09 * 512;
   $att_10 = &getPV($att_10_pv,1);
   $new_att += $att_10 * 1024;
   $att_11 = &getPV($att_11_pv,1);
   $new_att += $att_11 * 2048;
   $att_12 = &getPV($att_12_pv,1);
   $new_att += $att_12 * 4096;
   $att_13 = &getPV($att_13_pv,1);
   $new_att += $att_13 * 8192;
   $att_14 = &getPV($att_14_pv,1);
   $new_att += $att_14 * 16384;
   $att_15 = &getPV($att_15_pv,1);
   $new_att += $att_15 * 32768;

   if ($new_att != $old_att) {
      $del = 1;
   }
   $att = $new_att;
   return ($del);
}

# ---------------------------------------------------------------

sub kl_check {

   # Declare local variables
   my ($old_gain, $del);

   $old_gain = $kl_gain;
   $del = 0;

   $kl_gain = &getPV($kl_gain_up_pv,1);
   if ($kl_gain != $old_gain) {$del = 1}
   $kl_gain = &getPV($kl_gain_dn_pv,1);
   if ($kl_gain != $old_gain) {$del = 1}
   $kl_gain = &getPV($kl_gain_ob_pv,1);
   if ($kl_gain != $old_gain) {$del = 1}
   $kl_gain = &getPV($kl_gain_ib_pv,1);
   if ($kl_gain != $old_gain) {$del = 1}

   return ($del);
}

# ---------------------------------------------------------------

sub mode {

   # This sub sets the values of PVs for a given mode.
   # Hidden here is the *actual* starting and stopping of feedback (not just the request).
   # One exception: if the intensity check sub detects no beam, it automatically shuts off feedback.
   # This chart defines PVs for a given mode:
   #
   #	Mode			status	pause	pause_h	pause_v	zero	stpt	stop	act_rqs
   #	---------------------------------------------------------------------------------------
   #	startup		0	0	3	1	1	0	0	0	0
   #	pause		1 	10	3	1	1	0	0	0	0
   #	feedback	4	40	0	0	0	0	0	0	1
   #	offsets		5	50	4	1	1	0	0	0	0		i.e. Reset zero_pv to 0 after starting offsets sub
   #	setpoints	6	60	4	1	1	0	0	0	0		i.e. Reset stpt_pv to 0 after starting setpoints sub
   #	pause_no_beam	7	70	4	1	1	0	0	0	0		This sub will not change PVs for this mode
   #	stop_script	99	99	4	1	1	0	0	1	0

   my $mode = shift (@_);
   my @pv_defs;
   if    ($mode == 0 ) {@pv_defs = ( 0,3,1,1,0,0,0)}
   elsif ($mode == 1 ) {@pv_defs = (10,3,1,1,0,0,0)}
   elsif ($mode == 4 ) {@pv_defs = (40,0,0,0,0,0,1)}
   elsif ($mode == 5 ) {@pv_defs = (50,4,1,1,0,0,0)}
   elsif ($mode == 6 ) {@pv_defs = (60,4,1,1,0,0,0)}
   elsif ($mode == 7 ) {@pv_defs = (70,4,1,1,0,0,0)}
   else                {@pv_defs = (99,4,1,1,0,0,0)}

   $postab_status =	$pv_defs[0];
   $postab_pause =	$pv_defs[1];
   #$postab_pause_h =	$pv_defs[2];
   #$postab_pause_v =	$pv_defs[3];
   $postab_zero =	$pv_defs[4];
   $postab_stpt =	$pv_defs[5];
   $act_rqs_clc =	$pv_defs[6];				# This value controls feedback

   &putPV($act_rqs_clc_pv,   $act_rqs_clc,   1);		# This will turn feedback on
   &status_write($postab_status);
   &putPV($postab_pause_pv,  $postab_pause,  1);
   &putPV($postab_zero_pv,   $postab_zero,   1);
   &putPV($postab_stpt_pv,   $postab_stpt,   1);
   &putPV($postab_mode_pv,   $mode,          1);
}

# ---------------------------------------------------------------

sub dac_check {

   # Declare local variables

   my ($vV_err, $hV_err, $del_kl, $del_att);
   my ($permit, $err, $E_diff, $E_check, $E_err);
   my ($att_err, $kl_err, $sr_err, $permit_err, $a_err, $wbsh_err, $mbsh_err, $E_mov_err, $ID_mov_err);
   my $E_adj = 0.120;							# 120 chosen based on current ID/DCM offsets - 7 Feb 2007
   my $max_E_tol = 0.090;						# 90 = current theoretical max + 10 - 7 Feb 2007

   # Check that DAC outputs aren't at limits
   $V_limit_val = &getPV($V_limit_val_pv,1);
   $H_limit_val = &getPV($H_limit_val_pv,1);
   if ($V_limit_val||$H_limit_val) {
      $act_rqs_clc = 0;
      &putPV($act_rqs_clc_pv,$act_rqs_clc,1);
      if ($postab_v_driver) {&check_mostab(2)};
      &message('Control voltage at limit!');
      &message('Feedback has been stopped.');
      # Set voltages to last good value
      &message('Setting voltages to last good value.');
      &putPV($vV_pv,$last_arr_v[0],1);
      &putPV($hV_pv,$last_arr_h[0],1);
      &message('Investigation underway.');

      # Investigate after brief pause
      select(undef,undef,undef,5);					# 5 sec is the polling interval of the KL server
      # KL gains changed?
      $del_kl = &kl_check;
      if ($del_kl) {return(3);}
      # Attenuators changed?
      $del_att = &att_check($att);
      if ($del_att) {return(4);}
      # Components changed?
      (@pvs) = &pv_check(70);
      if ($pvs[3] == 1) {$postab_mode = 99; return;}                     # stop
      if ($pvs[0] == 3) {$postab_mode = 1;  return;}                     # pause = 3; pause
      if (($pvs[2] == 1) || ($pvs[1] == 1)) {
         $postab_zero = 0;
         $postab_stpt = 0;
         &putPV($postab_zero_pv,$postab_zero,1);
         &putPV($postab_stpt_pv,$postab_stpt,1);
         &message('Pause feedback.');
         &message('Then measure offsets or setpoints.');
      }
      $E_mov	= &getPV($E_mov_pv,1);
      if ($E_mov)	{$E_mov_err = 0;}
      else		{$E_mov_err = 1;}
      $ID_mov_err	= &getPV($ID_mov_pv,1);
      $mbsh_err	= &getPV($shutter_mono_pv,1);
      $wbsh_err	= &getPV($shutter_white_pv,1);
      $a_err	= &getPV($shutter_a_pv,1);
      $permit	= &getPV($shutter_permit_pv,1);
      if ($permit)	{$permit_err = 0;}
      else		{$permit_err = 1;}
      $sr_err	= &getPV($sr_cmp_val_pv,1);
      if (defined $E_ID_pv) {
         $E_ID	= &getPV($E_ID_pv,1);
         $E	= &getPV($E_pv,1);
         $E_diff = $E_ID - $E;
         $E_check = abs($E_diff - $E_adj);
         if ($E_check > $max_E_tol) {$E_err = 1}
         else                       {$E_err = 0}
      }
      else {$E_err = 0};
      $err = $E_mov_err + $ID_mov_err + $mbsh_err + $wbsh_err + $a_err + $permit_err + $sr_err + $E_err;
      if ($err) {return(5);}
      return(2);
   }

   # Check to verify that DAC outputs are within tolerance bands
   $vV_err = &getPV($vV_err_val_pv,1);	# Check to make sure that requested DAC voltage is provided
   $hV_err = &getPV($hV_err_val_pv,1);
   if ($vV_err||$hV_err) {			# If actual != requested, wait 0.25 sec and repeat
      select(undef,undef,undef,0.25);
      $vV_err =	&getPV($vV_err_val_pv,1);
      $hV_err =	&getPV($hV_err_val_pv,1);
      if ($vV_err||$hV_err) {			# If still incorrect, give error message and pause.
         &message('DAC-ADC mismatch!');
         if ($vV_err) {&message('Vert ctrl voltage.')}
         if ($hV_err) {&message('Horz ctrl voltage.')}
         &message('Please investigate.');
         &message('Feedback pausing!');
         return(1);
      }
   }
}

# ---------------------------------------------------------------

sub active_light {

   $postab_active*=-1;
   &putPV($postab_active_pv,$postab_active,1);
}

# ---------------------------------------------------------------

sub message {

   # Declare local variables

   my ($message, $current_time, $statement);

   # Print message to medm screen
   $message	= shift (@_);
   $message5	= $message4;
   $message4	= $message3;
   $message3	= $message2;
   $message2	= $message1;
   $message1	= $message;

   # Write message to logfile
   $current_time = scalar localtime;
   $statement = sprintf '%-60.60s '.$current_time, $message;
   if (defined $filehandle) {print DAT $statement."\n"};

   Pezca::PutString($message1_pv, $message1);
   Pezca::PutString($message2_pv, $message2);
   Pezca::PutString($message3_pv, $message3);
   Pezca::PutString($message4_pv, $message4);
   Pezca::PutString($message5_pv, $message5);

   select(undef,undef,undef,$message_time);
}

# ---------------------------------------------------------------

sub timeout {

   # Declare local vars here
   my ($type, $timeout, $curr_time, $del_time);

   $type = shift (@_);
   if ($type == 0) {
      $init_time = time;
      $timeout = 0;
   }
   else {
      $curr_time = time;
      $del_time = $curr_time - $init_time;
      if ($del_time > $type) {$timeout = 1}
      else                   {$timeout = 0}
   }
   select(undef,undef,undef,1);
   return ($timeout);
}

# ---------------------------------------------------------------

sub check_mostab {

   # Declare local vars here
   my ($type);

   # Determine what the sub should do
   $type = shift (@_);

   if ($type == 0) {
      # Determine mostab initial state
      $mostab_pause_init = &getPV($mostab_pause_pv,1);
      # Pause if necessary
      if ($mostab_pause_init == 0) {
         &putPV($mostab_pause_pv,1,1);
         &message('Mostab has been paused.');
         &message('Postab is using DCM piezo.');
      }
   }
   elsif ($type == 1) {
      # Check mostab during active feedback
      $mostab_pause = &getPV($mostab_pause_pv,1);
      # If mostab started, pause it
      if ($mostab_pause == 0) {
         &putPV($mostab_pause_pv,1,1);
         &message('Mostab has been paused.');
         &message('Postab is using DCM piezo.');
      }
   }
   elsif ($type == 2) {
      # Restore mostab to initial state
      if ($mostab_pause_init == 0) {
         &putPV($mostab_pause_pv,$mostab_pause_init,1);
         &message('Mostab has been restarted.');
      }
   }
}

# ---------------------------------------------------------------

sub signal_handler {

   print STDOUT 'Caught sig.'."\n";
   &message('Caught sig.');
   &stop_script(1);
}

# ---------------------------------------------------------------

sub status_write {

   my $write_option = shift (@_);

   my ($stat_mess, $current_time, $statement);

   my @ct   = localtime(time);	# This is ($sec,$min,$hour,$mday,$mon,$year$,$wday,$yday,$isdst)
   my $min  = $ct[1];
   my $hour = $ct[2];
   my $mday = $ct[3];
   my $mon  = ('Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec')[$ct[4]];
   my $stdate = sprintf '%2d', $mday;
   my $sttime = sprintf '%2d:%2d', $hour,$min;
   $stdate =~ s/ /0/g;
   $sttime =~ s/ /0/g;
   $postab_sttime = $mon.' '.$stdate.' at '.$sttime;
   if    ($write_option == 0)   {$postab_status =  0; $postab_stline = 'Starting feedback';}
   elsif ($write_option == 10)  {$postab_status = 10; $postab_stline = 'Feedback paused';}
   elsif ($write_option == 40)  {$postab_status = 40; $postab_stline = 'Feedback: active';}
   elsif ($write_option == 50)  {$postab_status = 50; $postab_stline = 'Measuring offsets';}
   elsif ($write_option == 51)  {$postab_status = 51; $postab_stline = 'Measuring offsets - waiting for MBSH';}
   elsif ($write_option == 60)  {$postab_status = 60; $postab_stline = 'Measuring setpoints';}
   elsif ($write_option == 61)  {$postab_status = 61; $postab_stline = 'Measuring setpoints - waiting for MBSH';}
   elsif ($write_option == 70)  {$postab_status = 70; $postab_stline = 'Investigating lost beam';}
   elsif ($write_option == 71)  {$postab_status = 71; $postab_stline = 'No beam; waiting for components';}
   elsif ($write_option == 81)  {$postab_status = 81; $postab_stline = 'Adjusting KL amplifier';}
   elsif ($write_option == 99)  {$postab_status = 99; $postab_stline = 'Feedback: exiting';}
   &putPV($postab_stline_pv, $postab_stline,0);
   &putPV($postab_sttime_pv, $postab_sttime,0);
   &putPV($postab_status_pv, $postab_status,1);
   $stat_mess = $postab_stline.' - '.$postab_sttime;
   $current_time = scalar localtime;
   $statement = sprintf '%-60.60s '.$current_time, $stat_mess;
   if (defined $filehandle) {print DAT $statement."\n";}
}

# ---------------------------------------------------------------


















# ---------------------------------------------------------------
# ---------------------------------------------------------------
# ---------------------------------------------------------------
#
# TO-DO LIST:
#
# Set offsets and setpoints unmeasured on exit
# If KL changes outside script, beam is lost before script knows that KL was changed.
#    Then, it gives a shift BD message.  Better to check some obvious problems first.
