#!/usr/bin/perl -W
#----------------------------------------------------------------
# This script measures PPU speed and sends out email if the speed gets slow
# It is supposed to be executed by cron every 10 minutes
#
#                                  Author: Sergey Stepanov
# Version: 1.0	2014/04/02 initial implementation
#
# Example running in cron every 10 minutes:
# */10 * * * * /usr/bin/perl -W /home/gmca/epics/pezca/ppuCheckSpeed.pl
#----------------------------------------------------------------


# Load Perl modules -----------------------------------------------
  use strict;
  use MIME::Lite;

  # Declare variables ---------------------------------------------
  my ($PEZCA, $script, $email_list, $host, $threshold, $message);
  my ($state, $msg, $common, $batch, $nBad, $maxBad, $sleepTime);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;
  if ($PEZCA !~ /\/$/) {$PEZCA .='/';}

  select STDERR; $|=1;				# set unbuffered output
  select STDOUT; $|=1;				# set unbuffered output

  $common = $PEZCA.'Common/common_subs.pl';
  if (-f $common) {do $common;}
  else            {print STDERR '!!! ppuCheckSpeed.pl: cannot find '.$common."\n";}

  $script = &makeScriptName();
  $batch  = 1;

  if ($^O =~ /Win/i) {die 'This script cannot run under Windows OS';}

  $email_list = 'rfischetti@anl.gov'
  	       .',sstepanov@anl.gov'
  	       .',makarov@anl.gov'
	       .',mhilgart@anl.gov'
	       .',qxu@anl.gov'
	       .',scorcoran@anl.gov'
	       .',fergusond@anl.gov'
	       .',ogata@anl.gov'
	       .',vnagas@anl.gov'
	       .',mbecker@anl.gov'
#	       .',stefan.brandstetter@dectris.com'
               ;

  $host = $ENV{HOSTNAME};
  if (!defined $host) {$host=`/bin/hostname`; $host=~s/\n//g;}
  if (!defined $host) {print STDERR '!!! '.$script.': HOSTNAME is undefined!'."\n\n"; $host='unknown';}
  $host = lc($host);
  $host =~ s/\.gmca\.aps\.anl\.gov$//;

# $batch     = 0;						#comment this out!
  $threshold = 10;
  $maxBad    = 3;
  $nBad      = 0;
  $sleepTime = 10;
  $message   = '';
  do {
     if (! $batch) {print STDOUT "\n".$script.': measuring speed of '.$host."\n";}
     ($state,$msg) = &measureSpeed ($threshold);
     if ($state == 1) {
	$message .= $msg;
	$nBad++;
	if ($nBad < $maxBad) {
           if (! $batch) {print STDOUT "\n".$script.': '.$host.' test is negative, sleeping '.$sleepTime.'s before repeating...'."\n";}
           sleep($sleepTime);
        }
     }
  } while ($state == 1 && $nBad < $maxBad);

  if ($state == 1) {
     &email($message,$email_list);
  } else {
     if (! $batch) {print STDOUT "\n".$script.': '.$host.' speed looks good:'."\n".$msg."\n";}
  }
  exit;

# ========================================================================

# ($state,$msg) = &measureSpeed ($threshold);		#1=timeExeedsThreshols, 0=OK, -1=Error
sub measureSpeed ($) {

  my ($script, $threshold, $duration, $command, $output);
  my (@lines, $nlines, $msg, $stamp, $i);

  $script    = 'measureSpeed';
  $threshold = shift(@_);

  $command = '/bin/rpm -qa';
  $output  = `/usr/bin/time --format="real %e\nuser %e\nsys %e\n" ${command} 2>&1 1>/dev/null`;
# if (! $batch) {print STDOUT "\n".$script.': output of '.$command.':'."\n".$output."\n";}
  @lines = split (/\n/, $output);					#split into lines
  $nlines = @lines;
  if ($nlines != 3) {
     print STDERR '!!! '.$script.': unexpected output (must be 3 lines, found '.$nlines.'):'."\n";
     for ($i=0; $i<$nlines; $i++) {print STDERR $i,'.  '.$lines[$i],"\n";}
     return -1;
  } else {
     for ($i=0; $i<$nlines; $i++) {
        if    ($lines[$i] =~ /^real /) {$lines[$i] =~ s/^real //;}
	elsif ($lines[$i] =~ /^user /) {$lines[$i] =~ s/^user //;}
	elsif ($lines[$i] =~ /^sys /)  {$lines[$i] =~ s/^sys //;}
	else {
           print STDERR '!!! '.$script.': unexpected line: '.$lines[$i].' in output:'."\n".$output."\n\n";
           return -1;
        }

        if ($lines[$i] !~ /^\d*\.\d\d$/) {
           print STDERR '!!! '.$script.': unexpected time: '.$lines[$i].' in output:'."\n".$output."\n\n";
	   return -1;
        }
        elsif ($lines[$i] > $threshold) {
           $stamp = scalar localtime;
	   $msg = "\n"
                . '*** '.$script.': '.$host.' speed looks BAD at '.$stamp.'.'."\n"
	        . 'The execution of "'.$command.'" exceeded the '.$threshold.'s threshold:'."\n"
	        . $output."\n\n";
           print STDERR $msg;
	   return 1,$msg;
	}
     }
     return 0,$output;
  }
}

# ========================================================================

  sub email ($$) {

    my ($msg, $email_list, $mail);

    $msg        = shift(@_);
    $email_list = shift(@_);

    $mail = MIME::Lite->new(
      From     => 'sstepanov@anl.gov',
      To       => $email_list,
      Subject  => $script.' at '.$host.': suspected slowness!',
      Type     => 'multipart/mixed',
    );
    $mail->attach(
      Type     => 'TEXT', 			#case-sensitive TEXT/BINARY/AUTO
      Encoding => 'quoted-printable',
      Data     => $msg."\n",
    );
    $mail->send();
    print STDERR 'Mail sent to '.$email_list.' .'."\n";
  }
