#!/usr/bin/perl -W
# _USED_BY_JBLUICE_
#--------------------------------------------------------------
# This script is a server-type version of sampleEnv.pl It runs permanently
# and receives commands via the waveform-type sampleEnvPV same way as
# sampleEnv.pl receives parameters in the command line. Once the command
# is received, the PV is changed to "received" and at the end of operation
# it is changed to either "done" or "failed".
# Typically the commands are sent by JBluIce at the beginning of frames
# acquisition. The script performs the following collective actions:
#
# (1) Normal operation ("in"):
#    - closes fast shutter (unless we are not moving collimator or pin diode)
#    - puts collimator IN
#    - puts backlight out (if requested or conditioned by collision with beamstop)
#    - flips beamstop IN
#    - does nothing to pindiode, but may optionally move pindiode IN or OUT
#    - does nothing to ringlight brightness, but may optionally do whatever requested
#    - does nothing to backlight brightness, but may optionally do whatever requested
# (2) Reverse operation ("out"):
#    - closes fast shutter
#    - puts collimator OUT
#    - flips beamstop OUT (if requested or conditioned by collision with backlight)
#    - puts backlight IN
#    - does nothing to pindiode, but may optionally move pindiode IN or OUT
#    - does nothing to ringlight brightness, but may optionally do whatever requested
#    - does nothing to backlight brightness, but may optionally do whatever requested
#
# The script can also move any individual component or any individual component
# can be excluded from the above sequences.
#
# The results are optionally reported to BluIce.
#
#     	            	Author: Sergey Stepanov
#
# Version-1.0: 2006/04/14  using Pezca
# Version-2.0: 2006/05/16  added "-reverse" flag, i.e. collimatorOut
# Version-2.1: 2006/12/07  added multiple attempts to move collimator
# Version-2.2: 2007/04/16  added verifying AutoStart at collimator motion
# Version-2.3: 2007/07/06  added skipping motions if collimator is already in place
# Version-3.0: 2007/10/09  adapted to new preset scheme: multiple presets, active preset, both X&Y.
# Version-4.0: 2008/10/20  ringlight brightness controls added; peristant backlight & beamstop added.
# Version-4.1: 2009/07/29  performed code cleanup
# Version-5.0: 2009/09/10  moved backlight on digital IO; added beamstop In/Out readback
# Version-6.0: 2009/09/15  removed calls of beamstopIN script and replaced it with internal beamstop flip controls;
#                          changed many flags and renamed from collimatorIn to collimatorCtrl
# Version-6.1: 2009/09/23  added disabling shutter during the script execution;
#                          moved collimator operation up for faster execution
# Version-6.2: 2009/10/12  shutter may not be closed if we move collimator only;
# Version-6.3: 2009/11/17  added protection against multiple instances
# Version-6.4: 2010/01/05  added backlight brightness controls
# Version-6.5: 2010/01/15  added beamstop distance threshold after which it cannot
#                          not collide with backlight and hence moving one the two
#                          IN does not require removing the other.
# Version-6.6: 2010/01/16  same as v.6.5, but simplified
# Version-6.7: 2010/04/13  added processing of beamstop and backlight not flipping
#                          when someone stands on safety mat
# Version-6.8: 2010/06/08  changed treating collimator=none : always take out
# Version-6.9: 2010/12/14  added SDIS checks for backlight and beamstop
# Version-7.0: 2011/01/06  added global flag for opening popups via BluIce
# Version-7.1: 2011/01/11  added opening guard slits when pinhole is requested
#                          and restoring it back
# Version-7.2: 2011/01/17  changed handling guard slits when switching between pinhole & scatterguard:
#                          guard slits are only opened to the minimum value per each pin and
#                          never closed back.
# Version-7.3: 2011/02/08  added workaround for backlight and beamstop flip: when one
#                          tries to command them after an attemt with SDIS=1, the value
#                          is there, but not in effect CA server ignores requests (thinks
#                          that it is already there. The workaround is to clear the value
#                          (send cmd=rbk) and then send it again.
# Version-8.0: 2013/02/27  added SONICC device. In reality SONICC can only collide with beamstop, but for
#                          safety the script assumes that it may also collide with PinDiode and backlight;
#                          Introduced collision matrix;
#                          Introduced options to move any component individually (the script auto-resolves potential collisions);
#                          Renamed the script to sampleEnv;
# Version-10.0: 2013/04/23 the first server version.
# Version-10.1: 2013/05/16 changed waiting for flip-readback to use wait4pv
# Version-10.2: 2013/05/30 added shutter waiting timeout
# Version-10.3: 2014/03/28 changed functions calls to arrays pointers (preparation to Perl-5.16)
# Version-10.4: 2014/06/24 Added auto-detection if SONICC was hooked
# Version-10.5: 2015/06/05 Added support for collimator angle (if hooked)
# Version-10.6: 2015/08/14 Change collimator motion that it can move horizontally too (before: only vertically)
# Version-10.7: 2016/06/30 Added pausing FPGA of vertical positional feedback
# Version-10.8: 2019/02/04 Added $progress for printing wait times in &wait4pv
# Version-10.9: 2019/02/26 Changed to adjust GS when collimator is commanded IN and is already IN
# Version-11.0: 2019/11/05 Added "getStates" command
# Version-11.1: 2020/01/15 Allow "Jet" object as a synonym to "SONICC".
# Version-11.2: 2021/02/18 Changed zenity_error popups to 'nowait' so that they do not hold the server if users does not acknowledge.
# Version-11.3: 2021/08/11 Added checking motion program for enabling/disabling SONICC.
#--------------------------------------------------------------
  use Time::HiRes qw(gettimeofday tv_interval time sleep);	#needed for short sleeps
  use Pezca;
# use warnings::unused;
  use strict;

  our ($beamline, $shutter, $shutter_readback, $shutter_open);
  our ($IGNORE_ERRORS, $NONINTERACTIVE, $USE_ZENITY_FOR_EXIT, $ZENITY_VIA_BLUICE);
  my ($PEZCA, $script, $scr, $crossrun, $status, $component_state, $shutter_assy);
  my (%collision, %hooked, %move, @obj_all, @obj, %state);
  my ($ringbright_ctrl, $ringbright_hooked, $backbright_ctrl, $backbright_hooked);
  my ($soniccAssy_PV, $soniccAxis_PV, $sonicc_tolerance, $soName);
  my ($sonicc_in_PV, $sonicc_out_PV, $sonicc_in_pos, $sonicc_out_pos);
  my ($colAssy_PV, @colAxes_PV, @col_dest_pos, $col_tolerance);
  my (@col_in_PV, @col_out_PV, @col_in_pos, @col_out_pos);
  my ($col_act_preset_PV, $col_active_preset, $col_started);
  my ($coaAssy_PV, @coaAxes_PV, @coa_dest_pos);
  my (@coa_in_PV, @coa_out_PV, @coa_in_pos, @coa_out_pos);
  my ($gsAssyV_PV, $gsSizeV_PV, $gsAssyH_PV, $gsSizeH_PV);
  my ($gsMinima_PV, $adjust_gs);
  my ($col_start_time, $col_motionTime);
  my ($backlight_cmd_PV, $backlight_rbk_PV, $backlight_act_PV);
  my ($backlight_out_pos, $backlight_in_pos);
  my ($beamstopflip_cmd_PV, $beamstopflip_rbk_PV, $beamstopflip_act_PV);
  my ($beamstopflip_out_pos, $beamstopflip_in_pos);
  my ($backlight_sdis_PV, $beamstopflip_sdis_PV);
  my ($bsAssy_PV, @bsAxes_PV, $bs_tolerance);
  my ($bs_in_PV, $bs_out_PV, $bs_in_pos, $bs_out_pos);
  my ($pinAssy_PV, $pinAxis_PV, $pin_tolerance);
  my ($pin_in_PV, $pin_out_PV, $pin_in_pos, $pin_out_pos);
  my ($bsDistance, $bsDistance_threshold, $bs_threshold_PV);
  my ($ringbright_ctrl_PV, $ringbright_save_PV, $ringbright_val);
  my ($backbright_ctrl_PV, $backbright_save_PV, $backbright_val);
  my ($attemptsMax, $start_time, $end_time, $tell_bluice);
  my ($safetymat_hooked_PV, $safetymat_state_PV, $safetymat_state);
  my ($confirm_timeout, $flip_timeout, $flip_poll, $flip_polldelay);
  my ($flip_counter, $flip_tolerance, $progress);
  my ($shutter_timeout, $shutter_poll);
  my ($shutter_polldelay, $shutter_counter, $shutter_tolerance);
  my ($component, $a, $b, $print_collisions, $sampleEnvPV, $nelm, $st);
  my ($sleep_interval, $gotcmd, $string, $connection_lost);
  my (@states, $stamp, $buff, $l_limit, $h_limit, $motionPrg);
  my ($vertPosFeedbackFPGAstate, $vertPosFeedbackPresent);
  my ($vertPosFeedbackUsePin, $vertPosFeedbackUseBS);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  do $PEZCA.'/Common/shutter_subs.pl';
  do $PEZCA.'/Common/motor_subs.pl';
  do $PEZCA.'/Common/mostab_subs.pl';
  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  select STDOUT; $|=1;					#set unbuffered output
  select STDERR; $|=1;					#set unbuffered output

  $SIG{'HUP'}  = 'signal_handler';			#(1) hangup (see signum.h)
  $SIG{'INT'}  = 'signal_handler';			#(2) interrupt (see signum.h)
  $SIG{'QUIT'} = 'signal_handler';			#(3) quit (see signum.h)
# $SIG{'KILL'} = 'signal_handler';			#(9) this is what is sent by kill -9 -- unblockable
  $SIG{'TERM'} = 'signal_handler';			#(15) terminate -- sent by kill without paramaters
# $SIG{'STOP'} = 'signal_handler';			#(19) stop -- unblockable
#-----------------------------------------------------------------
  $IGNORE_ERRORS       = 1;				#don't exit on error
  $NONINTERACTIVE      = 1;				#don't wait for any answers
  $USE_ZENITY_FOR_EXIT = 1;
### ATTENTION: because of this setting, caller needs to
### take care of displaying popups if it not associated
### with JBluIce. The popup data need to be retrieved
### from respective PV:
  $ZENITY_VIA_BLUICE   = 1;				#this program always reports errors via JBluIce popups

### The geometry is:  Sample -> backlight -> beamstop
### If beamstop distance (measured from sample) is less than a threshold, it may collide with beamstop
### Set the threshold to a large value to presume that collisions are ALWAYS possible 
### NOTE-1: threshold will likely be overwritten below by the content of BS:D:critical PV; so set there!!!
### NOTE-2: if Beamstop & Backlight are both in & someone starts reducing the distance,
###         the two may collide (there is no protection in the server)
  $bsDistance_threshold = 35.;				#threshold for colliding beamstop with backlight (default, overwritten by BS:D:critical)

  $script = &makeScriptName();
  $scr    = $script;
  $scr    =~ s/\.pl$//i;
  @obj_all = qw/Collimator Beamstop Pindiode Backlight SONICC/;
  @states = qw/accepted finished popup failed/;

  &protect_multirun();						#protect from multiple instances
  &read_cmdline();						#read command line parameters
  &protect_designated_host();					#ensure script is started on allowed host

###-----------------------------------------------------------------
  $status = 0;
  $sampleEnvPV = $beamline.':bi:sampleEnv';
  $status += &addCaMonitors([$sampleEnvPV]);
 ($st,$nelm) = &PezcaGetNelm($sampleEnvPV);
 if ($st) {$status++;}
###=============================
  if ($shutter          !~ /^${beamline}/) {$shutter          = $beamline.':'.$shutter;}
  if ($shutter_readback !~ /^${beamline}/) {$shutter_readback = $beamline.':'.$shutter_readback;}
  $shutter_assy = $beamline.':SH:Ps';
  &setAssyAuto($shutter_assy);
  $status += &addCaMonitors([$shutter,$shutter_readback]);
###=============================
### Collimator (motor):
  $colAssy_PV        = $beamline.':COL:St';
  @colAxes_PV        =($beamline.':COL:H',$beamline.':COL:V');
  $col_act_preset_PV = $colAssy_PV.':ActivePreset';
  $col_tolerance     = 0.03;					#max|act-rqs| in mm
  $attemptsMax       = 3;					#how may times to try moving
  $hooked{Collimator}= 1;					#beamline config parameter
# if ($beamline eq '23o') {$hooked{Collimator} = 0;}		#collimator not connected at ID-out
  if ($hooked{Collimator}) {$status += &addCaMonitors([$col_act_preset_PV]);}
###- - - - - - - - - - - - - -
### Collimator angle (motor):
  $coaAssy_PV        = $beamline.':COL:An';
  @coaAxes_PV        =($beamline.':COL:P',$beamline.':COL:Y');
  $hooked{Colliangle}= 0;					#beamline config parameter
  if ($beamline eq '23i' || $beamline eq '23o') {$hooked{Colliangle}=1;}
###=============================
### Guard Slits (X&Y sizes - motors):
  $gsAssyV_PV        = $beamline.':GS:Av';
  $gsAssyH_PV        = $beamline.':GS:Ah';
  $gsSizeV_PV        = $beamline.':GS:ZS';
  $gsSizeH_PV        = $beamline.':GS:YS';
  $gsMinima_PV       = $beamline.':bi:guards_minimum';
  $hooked{Guardslits}= 1;						#beamline config parameter
  if ($hooked{Guardslits}) {$status += &addCaMonitors(undef,[$gsMinima_PV]);}
###=============================
### Backlight (digital IO):
  $backlight_cmd_PV  = $beamline.':pmac21:acc65e:bo13';
  $backlight_rbk_PV  = $beamline.':pmac21:acc65e:rb13';
  $backlight_act_PV  = $beamline.':pmac21:acc65e:bi20';
# $backlight_sdis_PV = $backlight_cmd_PV.'.DISA';		#this is direct, but no CA monitor
  $backlight_sdis_PV = $backlight_cmd_PV;
  $backlight_sdis_PV =~ s/:bo/:SDIS/;
  $backlight_out_pos = 0.;
  $backlight_in_pos  = 1.;                              	#after  Sept-2009 (digital IO)
  $hooked{Backlight} = 1;					#beamline config parameter
# if ($beamline eq '23o') {$hooked{Backlight} = 0;}		#there is no backlight at ID-out
  if ($hooked{Backlight}) {
     $status += &addCaMonitors([$backlight_cmd_PV,
                                $backlight_rbk_PV,
                                $backlight_act_PV,
                                $backlight_sdis_PV]);
  }
###=============================
### Beamstop (digital IO):
  $beamstopflip_cmd_PV  = $beamline.':pmac21:acc65e:bo14';
  $beamstopflip_rbk_PV  = $beamline.':pmac21:acc65e:rb14';
  $beamstopflip_act_PV  = $beamline.':pmac21:acc65e:bi21';
# $beamstopflip_sdis_PV = $beamstopflip_cmd_PV.'.DISA';		#this is direct, but no CA monitor
  $beamstopflip_sdis_PV = $beamstopflip_cmd_PV;
  $beamstopflip_sdis_PV =~ s/:bo/:SDIS/;
  $beamstopflip_out_pos = 0.;
  $beamstopflip_in_pos  = 1.;                              	#after  Sept-2009 (digital IO)
  $bsAssy_PV         = $beamline.':BS:Ps';
  @bsAxes_PV         = ($beamline.':BS:PT',$beamline.':BS:D',$beamline.':BS:YW');
  $bs_in_PV          = $bsAxes_PV[0].':RqsPosMem1';
  $bs_out_PV         = $bsAxes_PV[0].':RqsPosMem2';
  $bs_threshold_PV   = $bsAxes_PV[1].':critical';
  $bs_tolerance      = 0.1; 					#max |act-rqs| in mm
  $hooked{Beamstop}  = 1;					#beamline config parameter
# if ($beamline eq '23o') {$hooked{Beamstop} = 0;}		#there is no beamstop at ID-out
  if($hooked{Beamstop}) {
     $status += &addCaMonitors([$beamstopflip_cmd_PV,
                                $beamstopflip_rbk_PV,
                                $beamstopflip_act_PV,
                                $bs_in_PV,$bs_out_PV,
                                $bs_threshold_PV]);
  }
###=============================
### Pin Diode (motor):
  $pinAssy_PV        = $beamline.':PIN:St';
  $pinAxis_PV        = $beamline.':PIN:P';
  $pin_in_PV         = $pinAxis_PV.':RqsPosMem1';
  $pin_out_PV        = $pinAxis_PV.':RqsPosMem2';
  $pin_tolerance     = 0.2; 					#max |act-rqs| in mm
  $hooked{Pindiode}  = 1;					#beamline config parameter
# if ($beamline eq '23b') {$hooked{Pindiode} = 0;}    		#there is no pin diode at BM
  if ($hooked{Pindiode}) {
     $status += &addCaMonitors([$pin_in_PV,$pin_out_PV]);
  }
###=============================
### SONICC (motor):
  if ($beamline eq '23o') {
     $soniccAssy_PV     = $beamline.':SNC:Ps';
     $soniccAxis_PV     = $beamline.':SNC:P';
     $soName            = 'SONICC';
  } else {
     $soniccAssy_PV     = $beamline.':JET:Ps';
     $soniccAxis_PV     = $beamline.':JET:P';
     $soName            = 'CryoJet';
  }
  $sonicc_in_PV         = $soniccAxis_PV.':RqsPosMem1';
  $sonicc_out_PV        = $soniccAxis_PV.':RqsPosMem2';
  $sonicc_tolerance     = 0.2; 					#max |act-rqs| in mm
  $hooked{SONICC}       = 0;					#beamline config parameter
  ### Comment this out when SONICC experiment is over:
  if ($beamline eq '23o') {		#SONICC is only available at 23ID-OUT during SONICC experiments
     ($l_limit,$h_limit) = &checkMotorLimits ($soniccAxis_PV);
     $motionPrg = &getMotionPrg($soniccAssy_PV,'string',\$status);
     if (($l_limit == $h_limit && $l_limit*$h_limit != 0) || $motionPrg eq 'None') {
        $hooked{SONICC} = 0;					#if both limits are set the same, then motor is unhooked
        print STDOUT "\n".'!!! '.$scr.': SONICC is ignored because either both limits are set the same or motion program is "None"'
                         .' (lowLimit='.$l_limit.', highLimit='.$h_limit.', prg='.$motionPrg.').'."\n\n";
     } else {
        $hooked{SONICC} = 1;					#if at least one limit is not set, motor is hooked
     }
  }
  if ($hooked{SONICC}) {
     $status += &addCaMonitors([$sonicc_in_PV,$sonicc_out_PV]);
  }
###=============================
### Ringlight brightness:
  $ringbright_ctrl_PV = $beamline.':DAC1_4';
  $ringbright_save_PV = $ringbright_ctrl_PV.'_tweak.O'; 	#max=HOPR
  $hooked{Ringbrightness} = 1;					#beamline config parameter
  if ($hooked{Ringbrightness}) {$status += &addCaMonitors([$ringbright_ctrl_PV]);}
###=============================
### Backlight brightness:
  $backbright_ctrl_PV = $beamline.':DAC1_8';
  $backbright_save_PV = $backbright_ctrl_PV.'_tweak.O'; 	#max=HOPR
  $hooked{Backbrightness} = 1;					#beamline config parameter
  if ($hooked{Backbrightness}) {$status += &addCaMonitors([$backbright_ctrl_PV]);}
###=============================
### Safety mat:
  $safetymat_hooked_PV = $beamline.':pmac21:acc65e:bi23';
  $safetymat_state_PV  = $beamline.':pmac21:acc65e:bi22';
  $status += &addCaMonitors([$safetymat_hooked_PV,$safetymat_state_PV]);
###=============================
### Vertical beam position feedback (FPGA, Sioan):
  $vertPosFeedbackUsePin = 1;
  $vertPosFeedbackUseBS  = 0;
  $vertPosFeedbackFPGAstate = &getVertPosFeedbackFPGAstate($beamline);
  if ($vertPosFeedbackFPGAstate >= 0) {
     $vertPosFeedbackPresent = 1;
     $status += &addCaMonitors([$beamline.':feedback2:FPGA:feedBackMode']);
  } else {
     $vertPosFeedbackPresent = 0;
  }
###=============================

  if ($status) {
     $USE_ZENITY_FOR_EXIT = 1;
     $ZENITY_VIA_BLUICE   = 0;
     &press_key_exit('*** '.$scr.': cannot set CA monitors on '.$status.' PVs');
  }
### Clear the command PV (both will work):
# $status = &PezcaPutStringWF($sampleEnvPV,'finished','oldca',$nelm);
  $status = &PezcaPutStringWF($sampleEnvPV,'','oldca',$nelm);

###========================================================================
  &setEpicsMessages('off');
  $sleep_interval  = 0.05;
  $connection_lost = 0;
CMD:
  while(1) { ##############################################################################
### Wait for command:                                                                     #
     $gotcmd = 0;						#no command yet           #
     $stamp = scalar localtime;                                                           #
     print STDOUT "\n".$scr.': '.$stamp.'  Waiting for SampleEnv command...'."\n";        #
# --------------------------------------------------------	  		          #
     do {                                                                                 #
        select(undef,undef,undef,$sleep_interval);               #short sleep             #
        ($status,$string) = &PezcaGetStringWF($sampleEnvPV,$nelm,'noerr');                #
        if ($status) {                                           #PV read failed          #
           if (! $connection_lost) {                             #1st loss of connection  #
              &report_only($script,$status,$sampleEnvPV);                                 #
              $stamp = scalar localtime;                                                  #
              print STDOUT $script.': '.$stamp.'  PV connection lost!'."\n";              #
#             &message('PV connection lost!','error');           #message is not available#
              $connection_lost = 1;                                                       #
           }                                                                              #
           $gotcmd = 0;                                          #Connection lost         #
           $string = '';                                         #Clear the string        #
           sleep 5;                                              #wait before trying again#
        }                                                                                 #
        else {                                                   #PV read OK              #
           if ($connection_lost) {                               #connection restored     #
              $connection_lost = 0;                                                       #
              &message('PV connection restored. Restoring CA-monitors.');                 #
              &resetAllCaMonitors('list');                                                #
              &message('CA monitors are reconnected.');                                   #
           }                                                                              #
           $string =~ s/;.*$//g;                                 #trail comments after ";"#
           $string =~ s/\s+/ /g;                                 #compress spaces         #
           $string =~ s/^\s+//g;                                 #remove start spaces     #
           $string =~ s/\s+$//g;                                 #remove end spaces       #
           if (length($string) == 0 || grep(/^$string$/i, @states)) {	 #No real command #@states=qw/accepted finished popup failed/
              $gotcmd = 0;                                                                #
              if ($string =~ /^accepted$/i) {                                             #
                 &message('Changing unexpected "accepted" to "failed".','warn');          #
                 $status = &PezcaPutStringWF($sampleEnvPV,'failed','oldca',$nelm);        #
              }                                                                           #
           } else {                                                                       #
              $status = &read_command(\$string);                                          #
              if ($status == 1) {                                                         #
                 $gotcmd = 0;                                                             #
                 $status = &PezcaPutStringWF($sampleEnvPV,'failed','oldca',$nelm);        #
              }                                                                           #
              elsif ($status == 2) {                                                      #
                 $gotcmd = 0;                                                             #
                 $status = &PezcaPutStringWF($sampleEnvPV,'popup','oldca',$nelm);         #
              }                                                                           #
              else {                                                                      #
                 $gotcmd = 1;                                 	 #got valid command       #
                 $status = &PezcaPutStringWF($sampleEnvPV,'accepted','oldca',$nelm);      #
              }                                                                           #
              &report_only($script,$status,$sampleEnvPV);                                 #
           }                                                                              #
        }                                                                                 #
     } while ($gotcmd != 1);                                                              #
# --------------------------------------------------------				  #
     &message('...Got command: "'.$string.'"');                                           #
     $start_time = time();                                                                #
                                                        				  #
     if ($vertPosFeedbackPresent) {                                                       #
        $vertPosFeedbackFPGAstate = &getVertPosFeedbackFPGAstate($beamline);              #
     }                                                                                    #
###=============================                                                          #
### Check if beamstop may collide with backlight (when it is too upstream)                #
     if ($hooked{Beamstop}) {                                                             #
### Read the threshold distance for beamstop colliding with backlight:                    #
        ($status,$buff) = Pezca::GetDouble($bs_threshold_PV);                             #
        ### on err default $bsDistance_threshold will be used (see above)                 #
        if ($status) {                                  				  #
           &ezca_warn($script, $status, $bs_threshold_PV, 				  #
                      '. Using default threshold='.$bsDistance_threshold);                #
        } else {                                                                          #
           $bsDistance_threshold = $buff;                                                 #
           &message('Using beamstop distance threshold = '                                #
                    .$bsDistance_threshold.' from a PV');                                 #
        }                                                                                 #
### Read current beamstop distance:                                                       #
        $bsDistance = &getAxisAct ($bsAxes_PV[1]);                                        #
     } else {                                                                             #
        $bsDistance = 0;                                                                  #
     }                                                                                    #
###=============================                                                          #
     @obj=();                                                                             #
     foreach $a (@obj_all) {if ($hooked{$a}) {push @obj,$a;}}                             #
###=============================                                                          #
     if ($string =~ /^getStates$/) {                                                      #
        $string = ';';                                                                    #
        &message('Sample Environment states:');                                           #
        if (&find_IN_and_OUT_collimator_PVs()) {&clearErr(); next CMD;}                   #
        foreach $a (@obj) {                                                               #
           $status = &getState($a,\$component_state);                                     #
           if (! $status) {$string .= ' '.$a.'='.$component_state;}                       #
        }                                                                                 #
        &message($string);                                                                #
        print STDOUT $string,"\n";                                                        #
        $status = &PezcaPutStringWF($sampleEnvPV,$string,'oldca',$nelm);                  #
        &report_only($script,$status,$sampleEnvPV);                                       #
        $gotcmd = 0;                                                                      #
        next CMD;                                                                         #
     }                                                                                    #
###=============================                                                          #
### Build collision matrix:                                                               #
     &message('Building collision matrix');                                               #
     foreach $a (@obj) {                                                                  #
     foreach $b (@obj) {                                                                  #
        if ($a ne $b) {$collision{$a}{$b} = 0;}                                           #
        else          {$collision{$a}{$b} = 1;}                                           #
     }}                                                                                   #
     if ($hooked{Backlight}) {                                                            #
        if ($hooked{Beamstop} && $bsDistance < $bsDistance_threshold) {                   #
           $collision{Beamstop}{Backlight} = 1;                                           #
           &message('Setting Beamstop-Backlight collision to "yes"');                     #
        }                                                                                 #
        if ($hooked{Pindiode}) {                                                          #
           $collision{Pindiode}{Backlight} = 1;                                           #
           &message('Setting Pindiode-Backlight collision to "yes"');                     #
        }                                                                                 #
     }                                                                                    #
     if ($hooked{SONICC}) {                                                               #
        if ($hooked{Beamstop})  {                                                         #
           $collision{SONICC}{Beamstop}  = 1;                                             #
           &message('Setting '.$soName.'-Beamstop collision to "yes"');                   #
        }                                                                                 #
        if ($hooked{Pindiode})  {                                                         #
           $collision{SONICC}{Pindiode}  = 1;                                             #
           &message('Setting '.$soName.'-Pindiode collision to "yes"');                   #
        }                                                                                 #
        if ($hooked{Backlight}) {                                                         #
           $collision{SONICC}{Backlight} = 1;                                             #
           &message('Setting '.$soName.'-Backlight collision to "yes"');                  #
        }                                                                                 #
     }                                                                                    #
     foreach $a (@obj) {                                                                  #
     foreach $b (@obj) {                                                                  #
        if ($collision{$a}{$b} == 1) {$collision{$b}{$a} = 1;}                            #
     }}                                                                                   #
### Print collisions matrix:                                                              #
     if (defined $print_collisions) {                                                     #
        $buff = 'Collisions matrix:'."\n"                                                 #
              . '------------------'."\n";                                                #
        $buff .= sprintf ' %-10s',' ';                                                    #
        foreach $a (@obj) {$buff .= sprintf ' %-10s',$a;}                                 #
        $buff .= "\n";                                                                    #
        foreach $a (@obj) {                                                               #
           $buff .= sprintf '%10s', $a;                                                   #
           foreach $b (@obj) {$buff .= sprintf '     %d     ',$collision{$a}{$b};}        #
           $buff .= "\n";                                                                 #
        }                                                                                 #
        &zenity_info($buff);                                                              #
        $status = &PezcaPutStringWF($sampleEnvPV,'finished','oldca',$nelm);               #
        &report_only($script,$status,$sampleEnvPV);                                       #
        next CMD;                                                                         #
     }                                                                                    #
###=============================                                                          #
                                                					  #
     if (&find_IN_and_OUT_collimator_PVs()) {&clearErr(); next CMD;}                      #
                                                					  #
### Check if anyone stands on safety mat. Standing on the mat disables                    #
### the Beamstop & Backlight moves (as it disables the digital IO):                       #
     &check_safetymat();                                                                  #
                                                					  #
### Find current state of all devices:                                                    #
     foreach $a (@obj) {                                                                  #
        if (&getState($a)) {&clearErr(); next CMD;}                                       #
     }                                                                                    #
                                                					  #
     if ($vertPosFeedbackPresent && $vertPosFeedbackFPGAstate > 0) {                      #
        &message('disabling vert.pos. feedback');       				  #
        $status = &setVertPosFeedbackFPGAstate($beamline,0);                              #
     }                                          					  #
                                                					  #
     $col_started = 0;                                                                    #
### First move all elements that are requested to be moved OUT:                           #
     foreach $a (@obj) {                                                                  #
        if ($hooked{$a} && $move{$a} eq 'out' && $state{$a} ne 'out') {                   #
           if (&moveObj($a,$move{$a})) {&clearErr(); next CMD;}                           #
        }                                                                                 #
     }                                                                                    #
                                                					  #
### Now move all elements that are requested to be moved IN:                              #
     foreach $a (@obj) {                                                                  #
        if ($hooked{$a} && $move{$a} eq 'in' && $state{$a} ne 'in') {                     #
           ### Find possible collisions:                                                  #
           foreach $b (@obj) {                                                            #
              if ($b ne $a && $state{$b} ne 'out' && $collision{$a}{$b} == 1) {           #
                 if ($move{$b} ne 'in') {                                                 #
                    &message('will move '.$b.' OUT to avoid collision with '.$a,'warn');  #
                    if (&moveObj($b,'out')) {&clearErr(); next CMD;}                      #
                 } else {                                                                 #
                    &popup_error('Contradictory request: '.$b                             #
                                .' was requested to move IN, while'                       #
                                .' it must to be moved OUT in view'                       #
                                .' of potential collision with '.$a);                     #
                    &clearErr(); next CMD;                                                #
                 }                                                                        #
              }                                                                           #
           }                                                                              #
           if (&moveObj($a,$move{$a})) {&clearErr(); next CMD;}                           #
        }                                                                                 #
     }                                                                                    #
     if ($move{Collimator} eq 'in' || $state{Collimator} eq 'in') {                       #
        $status = &sortoutGuardSlits();                                                   #
     }                                                                                    #
                                                					  #
### Control ringlight brightness:                                                         #
     if ($hooked{Ringbrightness} && $ringbright_ctrl) {                                   #
        if (&controlRinglightBrightness()) {&clearErr(); next CMD;}                       #
     }                                                                                    #
                                                					  #
### Control backlight brightness:                                                         #
     if ($hooked{Backbrightness} && $backbright_ctrl) {                                   #
        if (&controlBacklightBrightness()) {&clearErr(); next CMD;}                       #
     }                                                                                    #
                                                					  #
     if ($hooked{Collimator} && $col_started) {                                           #
        if (&finishCollimator('Collimator',$move{Collimator})) {&clearErr(); next CMD;}   #
     }                                                                                    #
                                                					  #
     &setAssyAuto ($shutter_assy);                                                        #
### Print final states of all devices:                                                    #
     foreach $a (@obj) {&message($a.' final state='.uc($state{$a}).'.');}                 #
                                                					  #
     $end_time = time();                                                                  #
     $buff = sprintf 'Elapsed time=%.1f seconds',($end_time-$start_time);                 #
     &message ($buff);                                                                    #
     $status = &PezcaPutStringWF($sampleEnvPV,'finished','oldca',$nelm);                  #
     &ezca_warn($script,$status,$sampleEnvPV);                                            #
  } #######################################################################################
### We never get here:
  print STDOUT  $scr.': job finished'."\n\n";
  &clearAllCaMonitors();
  exit 0;

#=================================================================

sub protect_multirun() {
  my ($filter, $nprocesses);

### Protection from multiple instances:
  if ($^O !~ /Win/i) {
#    $filter = '/bin/ps -ef | '
     $filter = '/bin/ps -eo user:19,pid,cmd | '
              .'/bin/grep -iv "grep\|edit\|term\|onsole" | '
              .'/bin/grep ';
     $nprocesses  = `${filter} -c ${scr}`;
     chomp $nprocesses;
     $nprocesses = $nprocesses - 1;				#ignore this script
     if ($nprocesses > 0) {
        $NONINTERACTIVE      = 1;				#do not wait for any answers
        $USE_ZENITY_FOR_EXIT = 0;
        $ZENITY_VIA_BLUICE   = 0;
        &press_key_exit('*** '.$nprocesses.' instance(s) of '
                              .$script.' is already running.');
     }
  }
}

#=================================================================

sub protect_designated_host() {
  my ($host, @hosts_allowed);

### Protection from running on multiple computers:
  if (!defined $crossrun || $crossrun == 0) {
     $host = $ENV{HOSTNAME};
     if (!defined $host) {$host=`hostname`; $host=~s/\n//g;}
     if (!defined $host) {print STDOUT '!!! HOSTNAME is undefined!'."\n\n"; $host='unknown';}
     $host = lc($host);
     $host =~ s/.gmca.aps.anl.gov$//;
     if    ($beamline eq '23i') {@hosts_allowed = ('bl1keithley', 'keithley1');}
     elsif ($beamline eq '23o') {@hosts_allowed = ('bl2keithley', 'keithley2');}
     elsif ($beamline eq '23b') {@hosts_allowed = ('bl3keithley', 'keithley3');}
     elsif ($beamline eq '23d') {@hosts_allowed = ('www', 'px0', 'sergey', 'x-server');}
     else                       {@hosts_allowed = 'undefined';}

     if (! grep(/^$host$/, @hosts_allowed)) {
        $NONINTERACTIVE      = 0;				#do not wait for any answers
        $USE_ZENITY_FOR_EXIT = 0;
        $ZENITY_VIA_BLUICE   = 0;
        &press_key_exit('*** '.$scr.': cannot run on host='.$host."\n\n"
                      . '*** '.$scr.': host(s) allowed for '.$beamline.':'."\n"
                      . "\'@{hosts_allowed}\'.");
     }
  }
}

#=================================================================

sub read_cmdline () {

  my ($syntax, @beamlinelist, $ln_specified, $arg, $narg, $i);

  $syntax = "\n\n".'Syntax: '.$script.' [beamline]'."\n";

  $narg = @ARGV;
  print STDOUT "\n".$script.': starting with cmdline: ';
  if ($narg > 0) {print STDOUT "@{ARGV}\n\n";}
  else           {print STDOUT '<no arguments>'."\n\n";}

### Read command line arguments:
  $crossrun = 0;
  @beamlinelist = ('23i:', '23o:', '23b:', '23d:');
  for ($i=0; $i<$narg; $i++) {
     $arg = $ARGV[$i];
     $arg =~ s/^-{0,2}//;					#trim leading dashes
### Check if the default beamline name is overridden:
     if (grep(/^$arg:?$/, @beamlinelist)) {
        if (defined $ln_specified) {
           &press_key_exit('*** '.$script
                 .': Duplicate "beamline" parameter in the command line.'
                 .$syntax);					#this will exit automatically if $NONINTERACTIVE=1
        }
        $ln_specified=1;
        $arg =~ s/:$//;						#trim ':'
        if ($beamline ne $arg) {
           $crossrun = 1;
           print STDOUT "\n".'!!! '.$scr.': Cross-beamline operation:'
                            .' controlling '.$arg.' from '.$beamline."\n\n";
           $beamline = $arg;
           do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
        }
     }
### Incorrect (unrecognized) command line argument:
     else {
        &press_key_exit('*** '.$script
                 .': Incorrect command line argument: '
                 .$ARGV[$i].$syntax);				#this will exit automatically if $NONINTERACTIVE=1
     }
  }
}

#=================================================================

sub read_command (\$) {						#reference to command string

  my ($stringRef, $syntax, @ARGS, $arg, $narg, $i);
  my ($op_specified, $so_specified, $st_specified);
  my ($cl_specified, $bs_specified, $pd_specified, $bl_specified);
  my ($rb_specified, $bb_specified, $tb_specified, $gs_specified);

  $stringRef = shift(@_);

  $$stringRef =~ s/[\n\r]/ /g;
  @ARGS = split(/\s+/,$$stringRef);

### Defaults (if no command line):
  foreach $a (@obj_all) {$move{$a} = 'none';}
  $adjust_gs       = 1;				#0=no, 1=adjust when collimator/scatterguard is moved in
  $ringbright_ctrl = 0;                 	#0=no_ctrl (default), 1=set a value
  $ringbright_val  = 0;
  $backbright_ctrl = 0;                 	#0=no_ctrl (default), 1=set a value
  $backbright_val  = 0;
  $flip_timeout    = 10;			#max time for flipping in seconds
  $confirm_timeout = 3;				#timeout for digital IO rbk confirmatin in seconds
  $flip_poll       = 0.05;			#polling interval in seconds
  $flip_polldelay  = 0.2;			#initial delay in seconds before starting polling
  $flip_counter    = 1;				#how many times to check that flip occurred
  $flip_tolerance  = 0.01;			#flip destination tolerance
  $shutter_timeout = 3;				#timeout for shutter closing
  $shutter_poll    = 0.05;			#polling interval in seconds
  $tell_bluice     = 1;
  $progress        = 5;				#interval between prints in wait4pv

  $syntax  = "\n\n".'Syntax: [help|?] [printCollisions] [getStates]'."\n";
  $syntax .= ' [IN|OUT]'."\n";
  $syntax .= ' [noCollimator|collimatorIN|collimatorOUT]'."\n";
  $syntax .= ' [noBeamstop|beamstopIN|beamstopOUT]'."\n";
  $syntax .= ' [noBacklight|backlightIN|backlightOUT]'."\n";
  $syntax .= ' [noPindiode|pindiodeIN|pindiodeOUT]'."\n";
  $syntax .= ' [noSonicc|soniccIN|soniccOUT] [noJet|jetIN|jetOut]'."\n";
  $syntax .= ' [ringlight=on|ringlight=value|ringlight=off|ringlight=skip]'."\n";
  $syntax .= ' [backlight=on|backlight=value|backlight=off|backlight=skip]'."\n";
  $syntax .= ' [handleGS|skipGS]'."\n";

  $narg = @ARGS;
  print STDOUT "\n".$scr.': received the following command: "';
  if ($narg > 0) {print STDOUT "@{ARGS}\"\n\n";}
  else           {print STDOUT '<empty string>"'."\n\n";}

### Make IN/OUT the first command arguments:
  if ($narg > 1) {
     $i=1;
     do {
        $arg = $ARGS[$i];
        if ($arg =~ /^-{0,2}IN$/i || $arg =~ /^-{0,2}OUT$/i) {
           splice(@ARGS, $i, 1);
           unshift(@ARGS, $arg);
        } else {
           $i++;
        }
     } while ($i<$narg);
  }
  undef $print_collisions;
  undef $st_specified;
  undef $op_specified;
  undef $cl_specified;
  undef $bs_specified;
  undef $bl_specified;
  undef $pd_specified;
  undef $so_specified;
  undef $rb_specified;
  undef $bb_specified;
  undef $gs_specified;
  undef $tb_specified;
### Read command arguments:
  for ($i=0; $i<$narg; $i++) {
     $arg = $ARGS[$i];
     $arg =~ s/^-{0,2}//;					#trim leading dashes
### Help requested:
     if ($arg =~ /^help$/i || $arg =~ /^\?$/) {
        &zenity_info($syntax,'nowait');
        return 2;
     }
### Collision matrix requested:
     elsif ($arg =~ /^printCollisions$/i) {
        if (defined $print_collisions) {
           &popup_error('Duplicate "printCollisions" parameter in the command: '
                       .$ARGS[$i].$syntax);
           return 1;
        }
        $print_collisions = 1
     }
### Request to return states of all active components:
     elsif ($arg =~ /^getStates$/i) {
        if (defined $st_specified) {
           &popup_error('Duplicate "getStates" parameter in the command: '
                       .$ARGS[$i].$syntax);
           return 1;
        }
        $st_specified = 1
     }
### Syntax = "[in|out|componentIN(OUT)=...]";
     elsif ($arg =~ /^out$/i ||
            $arg =~ /^in$/i) {
        if (defined $op_specified) {
           &popup_error('Duplicate "IN/OUT" parameter in the command: '
                       .$ARGS[$i].$syntax);
           return 1;
        }
        $op_specified=1;
        print STDOUT '--> '.$scr.': Operation: '.uc($arg).'!'."\n";
        if    ($arg =~ /^in$/i) {
           $move{Collimator} = 'in';
           $move{Beamstop}   = 'in';
           $move{Backlight}  = 'out';
        }
        elsif ($arg =~ /^out$/i) {
           $move{Collimator} = 'out';
           $move{Beamstop}   = 'out';
           $move{Backlight}  = 'in';
        }
     }
### Syntax = "[noCollimator|collimatorIN|collimatorOUT]";
     elsif ($arg =~ /^noCollimator$/i  ||
            $arg =~ /^collimatorIN$/i  ||
            $arg =~ /^collimatorOUT$/i) {
        if (defined $cl_specified) {
           &popup_error('Duplicate "collimator" parameter in the command: '
                       .$ARGS[$i].$syntax);
           return 1;
        }
        $cl_specified=1;
        if ($arg =~ /^noCollimator$/i) {
           $move{Collimator} = 'none';
           print STDOUT '--> '.$scr.': collimator will not be moved unless possible collision!'."\n";
        }
        elsif ($arg =~ /^collimatorOUT$/i || $arg =~ /^collimatorIN$/i) {
           $arg =~ s/^collimator//i;
           print STDOUT '--> '.$scr.': collimator move-'.uc($arg).' requested!'."\n";
           $move{Collimator} = lc($arg);
        }
        else {
           &popup_error('Incorrect "collimator" parameter in the command: '
                       .$ARGS[$i].$syntax);
           return 1;
        }
     }
### Syntax = "[noBeamstop|beamstopNormal|beamstopOUT|beamstopIfneeded]";
     elsif ($arg =~ /^nobeamstop$/i  ||
            $arg =~ /^beamstopIN$/i  ||
            $arg =~ /^beamstopOUT$/i) {
        if (defined $bs_specified) {
           &popup_error('Duplicate "beamstop" parameter in the command: '
                       .$ARGS[$i].$syntax);
           return 1;
        }
        $bs_specified=1;
        if ($arg =~ /^noBeamstop$/i) {
           $move{Beamstop} = 'none';
           print STDOUT '--> '.$scr.': beamstop will not be flipped unless possible collision!'."\n";
        }
        elsif ($arg =~ /^beamstopOUT$/i || $arg =~ /^beamstopIN$/i) {
           $arg =~ s/beamstop//i;
           print STDOUT '--> '.$scr.': beamstop flip-'.uc($arg).' requested!'."\n";
           $move{Beamstop} = lc($arg);
        }
        else {
           &popup_error('Incorrect "beamstop" parameter in the command: '
                       .$ARGS[$i].$syntax);
           return 1;
        }
     }
### Syntax = "[noBacklight|backlightNormal|backlightOUT|backlightIfneeded]";
     elsif ($arg =~ /^nobacklight$/i  ||
            $arg =~ /^backlightIN$/i  ||
            $arg =~ /^backlightOUT$/i) {
        if (defined $bl_specified) {
           &popup_error('Duplicate "backlight" parameter in the command: '
                       .$ARGS[$i].$syntax);
           return 1;
        }
        $bl_specified=1;
        if ($arg =~ /^nobacklight$/i) {
           $move{Backlight} = 'none';
           print STDOUT '--> '.$scr.': backlight will not be moved unless possible collision!'."\n";
        }
        elsif ($arg =~ /^backlightOUT$/i || $arg =~ /^backlightIN$/i) {
           $arg =~ s/backlight//i;
           print STDOUT '--> '.$scr.': backlight move-'.uc($arg).' requested!'."\n";
           $move{Backlight} = lc($arg);
        }
        else {
           &popup_error('Incorrect "backlight" parameter in the command: '
                       .$ARGS[$i].$syntax);
           return 1;
        }
     }
### Syntax = [pindiodeIN|pindiodeOUT]";
     elsif ($arg =~ /^nopindiode/i   ||
            $arg =~ /^pindiodeIN$/i  ||
            $arg =~ /^pindiodeOUT$/i) {
        if (defined $pd_specified) {
           &popup_error('Duplicate "pindiode" parameter in the command: '
                       .$ARGS[$i].$syntax);
           return 1;
        }
        $pd_specified=1;
        if ($arg =~ /^nopindiode$/i) {
           $move{Pindiode} = 'none';
           print STDOUT '--> '.$scr.': pindiode will not be moved unless possible collision!'."\n";
        }
        elsif ($arg =~ /^pindiodeOUT$/i || $arg =~ /^pindiodeIN$/i) {
           $arg =~ s/pindiode//i;
           print STDOUT '--> '.$scr.': pindiode move-'.uc($arg).' is requested!'."\n";
           $move{Pindiode} = lc($arg);
        }
        else {
           &popup_error('Incorrect "pindiode" parameter in the command: '
                       .$ARGS[$i].$syntax);
           return 1;
        }
     }
### Syntax = [nosonicc|soniccIN|soniccOUT]";
     elsif ($arg =~ /^nosonicc$/i  || $arg =~ /^noJet$/i ||
            $arg =~ /^soniccIN$/i  || $arg =~ /^jetIN$/i ||
            $arg =~ /^soniccOUT$/i || $arg =~ /^jetOUT$/i) {
        if (defined $so_specified) {
           &popup_error('Duplicate "'.$soName.'" parameter in the command: '
                       .$ARGS[$i].$syntax);
           return 1;
        }
        $so_specified=1;
        if ($arg =~ /^nosonicc$/i || $arg =~ /^noJet$/i) {
           $move{SONICC} = 'none';
           print STDOUT '--> '.$scr.': '.$soName.' will not be moved unless possible collision!'."\n";
        }
        elsif ($arg =~ /^soniccOUT$/i || $arg =~ /^soniccIN$/i) {
           $arg =~ s/sonicc//i;
           print STDOUT '--> '.$scr.': '.$soName.' move-'.uc($arg).' is requested!'."\n";
           $move{SONICC} = lc($arg);
        }
        elsif ($arg =~ /^jetOUT$/i || $arg =~ /^jetIN$/i) {
           $arg =~ s/jet//i;
           print STDOUT '--> '.$scr.': '.$soName.' move-'.uc($arg).' is requested!'."\n";
           $move{SONICC} = lc($arg);
        }
        else {
           &popup_error('Incorrect "'.$soName.'" parameter in the command: '
                       .$ARGS[$i].$syntax);
           return 1;
        }
     }
### Syntax = "[ringlight=skip|ringlight=on|ringlight=value|ringlight=off]";
     elsif ($arg =~ /^ringlight=/i) {
        if (defined $rb_specified) {
           &popup_error('Duplicate "ringlight" parameter in the command: '
                       .$ARGS[$i].$syntax);
           return 1;
        }
        $rb_specified=1;
        $arg =~ s/^ringlight=//i;
        $ringbright_ctrl = 1;
        if ($arg =~ /^skip$/i) {
           $ringbright_ctrl = 0;
           print STDOUT '--> '.$scr.': ringlight brightness operation disabled!'."\n";
        }
        elsif ($arg =~ /^on$/i) {
           $ringbright_val = undef;  				#take from last saved
           print STDOUT '--> '.$script
                 .': set ringlight ON, read brightness value from EPICS'."\n";
        }
        elsif ($arg =~ /^off$/i) {
           $ringbright_val = 0;
           print STDOUT '--> '.$script
                 .': set ringlight brightness value=0'."\n";
        }
        elsif (&is_decimal($arg)) {
           $ringbright_val = $arg;
           print STDOUT '--> '.$script
                 .': set ringlight brightness value='.$ringbright_val."\n";
        }
        else {
           &popup_error('Incorrect "ringlight" parameter in the command: '
                       .$ARGS[$i].$syntax);
           return 1;
        }
     }
### Syntax = "[backlight=skip|backlight=on|backlight=value|backlight=off]";
     elsif ($arg =~ /^backlight=/i) {
        if (defined $bb_specified) {
           &popup_error('Duplicate "backlight" brightness parameter in the command: '
                       .$ARGS[$i].$syntax);
           return 1;
        }
        $bb_specified=1;
        $arg =~ s/^backlight=//i;
        $backbright_ctrl = 1;
        if ($arg =~ /^skip$/i) {
           $backbright_ctrl = 0;
           print STDOUT '--> '.$script
                 .': backlight brightness operation disabled!'."\n";
        }
        elsif ($arg =~ /^on$/i) {
           $backbright_val = undef;  				#take from last saved
           print STDOUT '--> '.$script
                 .': set backlight ON, read brightness value from EPICS'."\n";
        }
        elsif ($arg =~ /^off$/i) {
           $backbright_val = 0;
           print STDOUT '--> '.$script
                 .': set backlight brightness value=0'."\n";
        }
        elsif (&is_decimal($arg)) {
           $backbright_val = $arg;
           print STDOUT '--> '.$script
                 .': set backlight brightness value='.$backbright_val."\n";
        }
        else {
           &popup_error('Incorrect "backlight" brightness parameter in the command: '
                       .$ARGS[$i].$syntax);
           return 1;
        }
     }
### Syntax = [handleGS|skipGS]";
     elsif ($arg =~ /^handleGS$/i || $arg =~ /^skipGS$/i) {
        if (defined $gs_specified) {
           &popup_error('Duplicate "GS handling" parameter in the command: '
                       .$ARGS[$i].$syntax);
           return 1;
        }
        $gs_specified=1;
        if ($arg =~ /^handleGS$/i) {
           $adjust_gs = 1;
           print STDOUT '--> '.$scr.': GS handling is requested!'."\n";
        }
        elsif ($arg =~ /^skipGS$/i) {
           $adjust_gs = 0;
           print STDOUT '--> '.$scr.': GS handling is disabled!'."\n";
        }
     }
### Syntax = [bluice|nobluice]";
     elsif ($arg =~ /^j?bluice$/i || $arg =~ /^noj?bluice$/i) {
        if (defined $tb_specified) {
           &popup_error('Duplicate "jbluice" parameter in the command: '
                       .$ARGS[$i].$syntax);
           return 1;
        }
        $tb_specified=1;
        if ($arg =~ /^j?bluice$/i) {
           $tell_bluice = 1;
           print STDOUT '--> '.$scr.': reports to JBluIce are requested!'."\n";
        }
        elsif ($arg =~ /^noj?bluice$/i) {
           $tell_bluice = 0;
           print STDOUT '--> '.$scr.': reports to JBluIce are disabled!'."\n";
        }
     }
### Incorrect (unrecognized) command argument:
     else {
        &popup_error('Incorrect command: '.$ARGS[$i].$syntax);
        return 1;
     }
  }
  return 0;
}

#=================================================================

sub find_IN_and_OUT_collimator_PVs() {
  my ($i, $j, $status, $move_only_vertically);
  $move_only_vertically = 0;					#0=vert&horiz. 1=vert.only
  if ($hooked{Collimator}) {
### Query current active collimator preset:
### 0=out, 1=300um, 2=20um, 3=10um, 4=5um
     ($status,$col_active_preset) = Pezca::GetDouble($col_act_preset_PV);
     &ezca_err($script,$status,$col_act_preset_PV,'. Getting active preset failed');
     if ($col_active_preset < 0 || $col_active_preset > 4) {
        &popup_error('Unexpected value='.$col_active_preset
                    .' in active preset PV='.$col_act_preset_PV);
        return 1;
     }
     if ($col_active_preset != 0) {				#Collimator=IN (0=out,1=300um,2=20um,3=10um,4=5um)
        $j = $col_active_preset + 1;				#the 'RqsPosMem' PVs are numbered from 1 to 5
        ### Here: @colAxes_PV=('COL:H','COL:V');
        for ($i=0; $i<2; $i++) {				#loop over H & V
           $col_in_PV[$i] = $colAxes_PV[$i].':RqsPosMem'.$j;	#IN-position PVs
           $status = &addCaMonitor($col_in_PV[$i]);
           &report_only($script,$status,$col_in_PV[$i]);
           $col_out_PV[$i] = $colAxes_PV[$i].':RqsPosMem1';	#OUT-position PVs
           $status = &addCaMonitor($col_out_PV[$i]);
           &report_only($script,$status,$col_out_PV[$i]);
           if ($move_only_vertically && $colAxes_PV[$i] =~ /COL:H/) {
              ### Move collimator OUT only vertically
              ### (ignore COL:H-out & use COL:H-in):
              $status = &clearCaMonitor($col_out_PV[$i]);
              $col_out_PV[$i] = $col_in_PV[$i];			#OUT-position PV for horiz.
              $status = &addCaMonitor($col_out_PV[$i]);
              &report_only($script,$status,$col_out_PV[$i]);
           }
        }
        if ($hooked{Colliangle}) {
           ### Here: @coaAxes_PV=('COL:P','COL:Y');
           for ($i=0; $i<2; $i++) {				#loop over P & Y
              $coa_in_PV[$i] = $coaAxes_PV[$i].':RqsPosMem'.$j;	#IN-position PVs
              $status = &addCaMonitor($coa_in_PV[$i]);
              &report_only($script,$status,$coa_in_PV[$i]);
              ### Preserve angle when moving collimator out
              $coa_out_PV[$i] = $coa_in_PV[$i];			#OUT-position PV for horiz.
           }
        }
     }
     else {							#preset=None
#       &message('Collimator motion disabled because active preset is NONE');
#       $move{Collimator} = 'none';
### Always move Collimator-out:
        for ($i=0; $i<2; $i++) {
           $col_in_PV[$i]  = $colAxes_PV[$i].':RqsPosMem1';
           $col_out_PV[$i] = $col_in_PV[$i];
           $status = &addCaMonitor($col_in_PV[$i]);
           &report_only($script,$status,$col_in_PV[$i]);
        }
        if ($hooked{Colliangle}) {
           for ($i=0; $i<2; $i++) {
              $coa_in_PV[$i]  = $coaAxes_PV[$i].':RqsPosMem1';
              $coa_out_PV[$i] = $coa_in_PV[$i];
              $status = &addCaMonitor($coa_in_PV[$i]);
              &report_only($script,$status,$coa_in_PV[$i]);
           }
        }
     }
  }
  return 0;
}

#=================================================================

sub check_safetymat() {
  my $status = 0;
### Check if anyone stands on safety mat. Standing on the mat disables
### the Beamstop & Backlight moves (as it disables the digital IO):
  if ($hooked{Beamstop} || $hooked{Backlight}) {
     ($status,$hooked{Safetymat}) = Pezca::GetDouble($safetymat_hooked_PV);
     if ($status) {
        &ezca_warn($script,$status,$safetymat_hooked_PV,'. Ignoring safety mat');
        $hooked{Safetymat} = 0;
     }
     if ($hooked{Safetymat}) {
        ($status,$safetymat_state) = Pezca::GetDouble($safetymat_state_PV);
        if ($status) {
           &ezca_warn($script,$status,$safetymat_state_PV,'. Ignoring safety mat state');
           $safetymat_state = 0;
        }
        if ($safetymat_state && ($move{Beamstop} ne 'none' || $move{Backlight} ne 'none')) {
           &popup_error('Operation is disabled by safety mat. '
                       .'Please step away from the mat and repeat.');
           return 1;
        }
     }
  }
  return 0;
}

#=================================================================

#sub getState($object,\$state) {			#status = &getState($object,\$state);
#  If \$state is not passed, fills $state{$obj}. Possible values are "in/out/unknown"
sub getState($) {
  my ($obj, $sta, $stateRef, @col_act, $bs_act, $beamstopflip_act_pos, $beamstopflip_rbk_pos);
  my ($pin_act_pos, $backlight_act_pos, $backlight_rbk_pos, $sonicc_act_pos, $i);
  $obj      = shift(@_);
  $stateRef = shift(@_);
  if    ($obj eq 'Collimator') { ###########################################
     for ($i=0; $i<2; $i++) {						#loop over horizontal/vertical
### Get collimator actual pos (horiz. & vert.):
        $col_act[$i] = &getAxisAct ($colAxes_PV[$i]);
### Get collimator IN/OUT positions:
        if (!defined $col_in_pos[$i]) {
           $status = &addCaMonitor($col_in_PV[$i]);
           &report_only($script,$status,$col_in_PV[$i]);
        }
        ($status,$col_in_pos[$i]) = Pezca::GetDouble($col_in_PV[$i]);
        &ezca_err($script,$status,$col_in_PV[$i],'. Getting '.$obj.' settings failed');
        if (!defined $col_out_pos[$i]) {
           $status = &addCaMonitor($col_out_PV[$i]);
           &report_only($script,$status,$col_out_PV[$i]);
        }
        if ($col_out_PV[$i] eq $col_in_PV[$i]) {$col_out_pos[$i] = $col_in_pos[$i];}
        else {
          ($status,$col_out_pos[$i]) = Pezca::GetDouble($col_out_PV[$i]);
          &ezca_err($script,$status,$col_out_PV[$i],'. Getting '.$obj.' settings failed');
        }
     }
     if    ((abs($col_act[0]-$col_in_pos[0])  < $col_tolerance) &&
            (abs($col_act[1]-$col_in_pos[1])  < $col_tolerance)) {$sta='in';}
     elsif ((abs($col_act[0]-$col_out_pos[0]) < $col_tolerance) &&
            (abs($col_act[1]-$col_out_pos[1]) < $col_tolerance)) {$sta='out';}
     else {
        $sta='unknown';
        printf STDOUT '!!! '.$obj.' state='.$sta."\n"
                     .'    Horz.: abs(Act-InPos)=%-6.2f abs(Act-OutPos)=%-6.2f'."\n"
                     .'    Vert.: abs(Act-InPos)=%-6.2f abs(Act-OutPos)=%-6.2f'."\n"
                     .'    tolerance=%.2f'."\n",
                     abs($col_act[0]-$col_in_pos[0]),
                     abs($col_act[0]-$col_out_pos[0]),
                     abs($col_act[1]-$col_in_pos[1]),
                     abs($col_act[1]-$col_out_pos[1]),
                     $col_tolerance;
     }
     if ($hooked{Colliangle}) {
        for ($i=0; $i<2; $i++) {
### Get collimator IN/OUT angles:
           if (!defined $coa_in_pos[$i]) {
              $status = &addCaMonitor($coa_in_PV[$i]);
              &report_only($script,$status,$coa_in_PV[$i]);
           }
           ($status,$coa_in_pos[$i]) = Pezca::GetDouble($coa_in_PV[$i]);
           &ezca_err($script,$status,$coa_in_PV[$i],'. Getting '.$obj.' settings failed');
           if (!defined $coa_out_pos[$i]) {
              $status = &addCaMonitor($coa_out_PV[$i]);
              &report_only($script,$status,$coa_out_PV[$i]);
           }
           if ($coa_out_PV[$i] eq $coa_in_PV[$i]) {$coa_out_pos[$i] = $coa_in_pos[$i];}
           else {
             ($status,$coa_out_pos[$i]) = Pezca::GetDouble($coa_out_PV[$i]);
             &ezca_err($script,$status,$coa_out_PV[$i],'. Getting '.$obj.' settings failed');
           }
        }
     }
  }
  elsif ($obj eq 'Beamstop') { #############################################
 ### Get Beamstop actual position:
     $bs_act = &getAxisAct ($bsAxes_PV[0]);			#BS:PT
### Get beamstop IN vertical motor position:
     if (!defined $bs_in_pos) {
        $status = &addCaMonitor($bs_in_PV);
        &report_only($script,$status,$bs_in_PV);
     }
     ($status,$bs_in_pos) = Pezca::GetDouble($bs_in_PV);
     &ezca_err($script,$status,$bs_in_PV,'. Getting '.$obj.' settings failed');
### Read beamstop current flip:
     ($status, $beamstopflip_act_pos) = Pezca::GetDouble($beamstopflip_act_PV);
     &ezca_err($script,$status,$beamstopflip_act_PV,'. Getting '.$obj.' settings failed');
### Read beamstop current flip readback:
     ($status, $beamstopflip_rbk_pos) = Pezca::GetDouble($beamstopflip_rbk_PV);
     &ezca_err($script,$status,$beamstopflip_rbk_PV,' Getting '.$obj.' settings failed');
     if ($beamstopflip_act_pos != $beamstopflip_rbk_pos) {
        &popup_error('Current '.$obj.' cmd & rbk disagree'
                    .' while getting '.$obj.' settings.');
        return 1;
     }
     if    (abs($beamstopflip_act_pos - $beamstopflip_in_pos)  < 0.01) {$sta='in';}
     elsif (abs($beamstopflip_act_pos - $beamstopflip_out_pos) < 0.01) {$sta='out';}
     else                                                              {$sta='unknown';}
     if (abs($bs_act - $bs_in_pos) > $bs_tolerance) {$sta .= '/notinplace';}
     if ($sta =~ /unknown/ || $sta =~ /notinplace/) {
        printf STDOUT '!!! '.$obj.' state='.$sta."\n"
                     .'    Flip:  abs(Act-InPos)=%-6.2f abs(Act-OutPos)=%-6.2f'."\n"
                     .'    Pitch: abs(Act-InPos)=%-6.2f       tolerance=%-6.2f'."\n",
                     abs($beamstopflip_act_pos - $beamstopflip_in_pos),
                     abs($beamstopflip_act_pos - $beamstopflip_out_pos),
                     abs($bs_act - $bs_in_pos), $bs_tolerance;
     }
  }
  elsif ($obj eq 'Pindiode') { #############################################
### Read pin diode current position:
     $pin_act_pos = &getAxisAct ($pinAxis_PV);
### Read pindiode IN/OUT positions:
     if (!defined $pin_in_pos) {
        $status = &addCaMonitor($pin_in_PV);
        &report_only($script,$status,$pin_in_PV);
     }
     ($status,$pin_in_pos) = Pezca::GetDouble($pin_in_PV);
     &ezca_err($script,$status,$pin_in_PV,'. Getting '.$obj.' settings failed');
     if (!defined $pin_out_pos) {
        $status = &addCaMonitor($pin_out_PV);
        &report_only($script,$status,$pin_out_PV);
     }
     ($status,$pin_out_pos) = Pezca::GetDouble($pin_out_PV);
     &ezca_err($script,$status,$pin_out_PV,'. Getting '.$obj.' settings failed');
     if    (abs($pin_act_pos - $pin_in_pos)  < $pin_tolerance) {$sta='in';}
     elsif (abs($pin_act_pos - $pin_out_pos) < $pin_tolerance) {$sta='out';}
     else {
        $sta='unknown';
        printf STDOUT '!!! '.$obj.' state='.$sta."\n"
                     .'    abs(Act-InPos)=%.2f abs(Act-OutPos)=%.2f'."\n"
                     .'    tolerance=%.2f'."\n",
                     abs($pin_act_pos - $pin_in_pos),
                     abs($pin_act_pos - $pin_out_pos),
                     $pin_tolerance;
     }
  }
  elsif ($obj eq 'Backlight') { ############################################
### Read backlight current position:
     ($status, $backlight_act_pos) = Pezca::GetDouble($backlight_act_PV);
     &ezca_err($script,$status,$backlight_act_PV,'. Getting '.$obj.' settings failed');
### Read backlight current position readback:
     ($status, $backlight_rbk_pos) = Pezca::GetDouble($backlight_rbk_PV);
     &ezca_err($script,$status,$backlight_rbk_PV,'. Getting '.$obj.' settings failed');
     if ($backlight_act_pos != $backlight_rbk_pos) {
        &popup_error('Current '.$obj.' cmd & rbk disagree'
                    .' while getting '.$obj.' settings.');
        return 1;
     }
     if    (abs($backlight_act_pos - $backlight_in_pos)  < 0.01) {$sta='in';}
     elsif (abs($backlight_act_pos - $backlight_out_pos) < 0.01) {$sta='out';}
     else {
        $sta='unknown';
        printf STDOUT '!!! '.$obj.' state='.$sta."\n"
                     .'    abs(Act-InPos)=%.2f abs(Act-OutPos)=%.2f'."\n",
                     abs($backlight_act_pos - $backlight_in_pos),
                     abs($backlight_act_pos - $backlight_out_pos);
     }
  }
  elsif ($obj eq 'SONICC') { ###############################################
### Read SONICC current position:
     $sonicc_act_pos = &getAxisAct ($soniccAxis_PV);
### Read pindiode IN/OUT positions:
     if (!defined $sonicc_in_pos) {
        $status = &addCaMonitor($sonicc_in_PV);
        &report_only($script,$status,$sonicc_in_PV);
     }
     ($status,$sonicc_in_pos) = Pezca::GetDouble($sonicc_in_PV);
     &ezca_err($script,$status,$sonicc_in_PV,'. Getting '.$obj.' settings failed');
     if (!defined $sonicc_out_pos) {
        $status = &addCaMonitor($sonicc_out_PV);
        &report_only($script,$status,$sonicc_out_PV);
     }
     ($status,$sonicc_out_pos) = Pezca::GetDouble($sonicc_out_PV);
     &ezca_err($script,$status,$sonicc_out_PV,'. Getting '.$obj.' settings failed');
     if    (abs($sonicc_act_pos - $sonicc_in_pos)  < $sonicc_tolerance) {$sta='in';}
     elsif (abs($sonicc_act_pos - $sonicc_out_pos) < $sonicc_tolerance) {$sta='out';}
     else {
        $sta='unknown';
        printf STDOUT '!!! '.$obj.' state='.$sta."\n"
                     .'    abs(Act-InPos)=%.2f abs(Act-OutPos)=%.2f'."\n"
                     .'    tolerance=%.2f'."\n",
                     abs($sonicc_act_pos - $sonicc_in_pos),
                     abs($sonicc_act_pos - $sonicc_out_pos),
                     $sonicc_tolerance;
     }
  }
  else { ###################################################################
     &popup_error($scr.': Unknown sample environment object "'.$obj.'"');
     return 1;
  } ########################################################################
  if (! defined $stateRef) {
     $state{$obj} = $sta;
     print STDOUT $scr.': '.$obj.' initial state='.uc($sta).".\n";
  } else {
     $$stateRef = $sta;
  }
  return 0;
}

#=================================================================

#sub moveObj($object,$destination) {
sub moveObj($$) {
  my ($obj, $dest, $status);
  $obj  = shift(@_);
  $dest = shift(@_);
  if ($dest ne 'in' && $dest ne 'out') {
     &popup_error('Unknown destination "'.$dest
                 .'" for sample environment object "'.$obj.'"');
     return 1;
  }
  if    ($obj eq 'Collimator') { ###########################################
     ($status,$col_started,$col_motionTime) = &startCollimator($obj,$dest);
     if ($status) {
        &message('moveObj got status='.$status.' from startCollimator','warn');
        return $status;
     }
     $col_start_time = time();
#    if ($dest eq 'in' && $hooked{Guardslits} && $adjust_gs && $col_started) {  #before 2019: adjust GS only if move collimator IN
        ###  Check if we need to open/close guard slits:
#       $status = &sortoutGuardSlits();
#    }
  }
  elsif ($obj eq 'Beamstop') { #############################################
     $status = &moveBeamstop($obj,$dest);
  }
  elsif ($obj eq 'Pindiode') { #############################################
     $status = &movePindiode($obj,$dest);
  }
  elsif ($obj eq 'Backlight') { ############################################
     $status = &moveBacklight($obj,$dest);
  }
  elsif ($obj eq 'SONICC') { ###############################################
     $status = &moveSONICC($obj,$dest);
  }
  else { ###################################################################
     &popup_error('Unknown sample environment object "'.$obj.'"');
     $status = 1;
  } ########################################################################
  return $status;
}

#=================================================================

# sub checkIfAssyMovable ($assy,$obj) {
sub checkIfAssyMovable ($$) {
  my ($assy, $obj, $motion, $homed);
  $assy = shift(@_);
  $obj  = shift(@_);

### Verify that motion program is not 'none':
  $motion = &getMotionPrg($assy);
  if ($motion == 0) {
     &popup_error($obj.' ['.$assy.'] motion is disabled.'
                 .' Cannot move the device.');
     return 1;
  }

### Verify that assembly is homed:
  $homed = &getHomed($assy);
  if ($homed == 0) {
     &popup_error($obj.' is not homed. Cannot move the device.');
     return 1;
  }
  return 0;
}

#=================================================================

# sub moveBacklight ($obj,$dest) {
sub moveBacklight ($$) {
  my ($obj, $dest, $backlight_cmd_pos, $backlight_rbk_pos);
  my ($backlight_act_pos, $backlight_sdis);
  my ($attempt, @pvlist, $pv, $val);

  $obj  = shift(@_);
  $dest = shift(@_);

  if    ($dest eq 'in')  {$backlight_cmd_pos = $backlight_in_pos;}
  elsif ($dest eq 'out') {$backlight_cmd_pos = $backlight_out_pos;}
  else {
     &popup_error('Unknown destination "'.$dest.'" for '.$obj
                 .' while preparing to move.');
     return 1;
  }
  @pvlist = ($backlight_act_PV, $backlight_cmd_PV,
             $beamstopflip_act_PV, $beamstopflip_cmd_PV,
             $bsAxes_PV[0].':ActPos', $bs_threshold_PV,
             $beamline.':PIN:stateCALC');

### If backlight is not where we need it:
  if ($state{$obj} ne $dest) {
     if (&close_disable_shutter()) {return 1;}
     $attempt = 0;
     do {
        $attempt++;
### Start flipping backlight IN or OUT:
        &message('Moving '.$obj.' '.uc($dest).', attempt '.$attempt.'/'.$attemptsMax);
        $status = Pezca::PutDouble($backlight_cmd_PV,$backlight_cmd_pos);	#digital I/O
        &ezca_err($script,$status,$backlight_cmd_PV,
                  '. Flipping '.$obj.' '.uc($dest).' failed');
### Check if the command was accepted (SDIS?)
        $status = &wait4pv($backlight_rbk_PV, $backlight_cmd_pos,
                           $confirm_timeout, $flip_poll, $flip_polldelay,
                           $flip_counter, $flip_tolerance, $progress);
        if ($status == -1) {							#PV error
           &ezca_warn($script,$status,$backlight_rbk_PV,
                      '. Taking the commanded position');
           $backlight_rbk_pos = $backlight_cmd_pos;
        }
        elsif ($status == 1) {							#timeout on waiting for expected value
           $backlight_rbk_pos = 1 - $backlight_cmd_pos;
           &message('Operation "'.$obj.'-'.uc($dest).'" failed: readback did not'
                   .' confirm the command within '.$confirm_timeout.'s','warn');
           ($status, $backlight_sdis) = Pezca::GetDouble($backlight_sdis_PV);
           if ($status) {
              &ezca_warn($script,$status,$backlight_sdis_PV,'. Presuming SDIS=0');
              $backlight_sdis = 0;
           }
           if ($backlight_sdis) {
              &message($obj.' flip disabled by these conditions (SDIS):','error');
              foreach $pv (@pvlist) {
                 ($status, $val) = Pezca::GetDouble($pv);
                 if ($status) {&report_only($script,$status,$pv); $val='undef';}
                 &message($pv.' = '.$val,'error');
              }
              select(undef,undef,undef,0.5);
           } else {
              &message('Clearing previous cmd to move '.$obj
                      .' while it was disabled (SDIS)');
              $status = Pezca::PutDouble($backlight_cmd_PV,1-$backlight_cmd_pos);
              &ezca_warn($script,$status,$backlight_cmd_PV);
              select(undef,undef,undef,0.5);
           }
        }
        else {
           $backlight_rbk_pos = $backlight_cmd_pos;				#OK: readback confirmed the command
        }
     } while ($attempt < $attemptsMax && $backlight_cmd_pos != $backlight_rbk_pos);
### Wait for backlight flip to complete:
     if ($attempt >= $attemptsMax) {$status = 1;}
     else {
        $status = &wait4pv($backlight_act_PV, $backlight_cmd_pos,
                           $flip_timeout, $flip_poll, $flip_polldelay,
                           $flip_counter, $flip_tolerance, $progress);
     }
     if ($status == -1) {							#PV error
        &ezca_warn($script,$status,$backlight_act_PV,
                   '. Taking the commanded position');
        $backlight_act_pos = $backlight_cmd_pos;
     }
     elsif ($status == 1) {							#timeout on waiting for expected value
### Bring the request and the actual in sync (safety mat issue):
        Pezca::PutDouble($backlight_cmd_PV,1-$backlight_cmd_pos);		#digital I/O
        &popup_error('Timeout on flipping '.$obj.' '.uc($dest));
        return 1;
     } else {
        select(undef,undef,undef,1.5);     					#sleep a bit extra;
        &message($obj.' has been flipped '.uc($dest));
        $state{$obj} = $dest;
     }
  }
  return 0;
}

#=================================================================

# sub moveBeamstop ($obj,$dest) {
sub moveBeamstop ($$) {
  my ($obj, $dest, $beamstopflip_cmd_pos, $beamstopflip_rbk_pos);
  my ($beamstopflip_sdis, $bs_act, $attempt, @pvlist, $pv, $val, $msg);

  $obj  = shift(@_);
  $dest = shift(@_);

  if    ($dest eq 'in')  {$beamstopflip_cmd_pos = $beamstopflip_in_pos;}
  elsif ($dest eq 'out') {$beamstopflip_cmd_pos = $beamstopflip_out_pos;}
  else {
     &popup_error('Unknown destination "'.$dest.'" for '.$obj
                 .' while preparing to move.');
     return 1;
  }
  @pvlist = ($beamstopflip_act_PV, $beamstopflip_cmd_PV,
             $backlight_act_PV, $backlight_cmd_PV,
             $bsAxes_PV[0].':ActPos', $bs_threshold_PV,
             $beamline.':PIN:stateCALC');

  if ($dest eq 'in' && $state{$obj} =~ /notinplace/) {
### Make sure that beamstop *motors* are in the proper position:
     &message('Synchronizing Act->Rqs for '.$obj);
     $status = Pezca::PutDouble($bsAssy_PV.':Sync.PROC',1);
     &ezca_warn($script,$status,$bsAssy_PV.':Sync.PROC','. Cannot sync Act->Rqs');
### Since the beamstop vertical motor is not in place, then move it there:
     if (&close_disable_shutter()) {return 1;};
     $bs_act = &getAxisAct ($bsAxes_PV[0]);
     $msg = sprintf 'Moving '.$obj.' from %.3f to %.3f',$bs_act,$bs_in_pos;
     &message($msg);
     $status = &checkIfAssyMovable ($bsAssy_PV,$obj);
     if ($status) {return 1;}
     for ($attempt=1; $attempt<=$attemptsMax; $attempt++) {
        $status = &runMotor ($bsAssy_PV, $bsAxes_PV[0], $bs_in_pos, 'checkAuto');
        if ($status) {
           &message($obj.' move failed (err='.$status.'), attempt '.$attempt.'/'.$attemptsMax.'.','warn');
           next;
        }
        last;						#operation successful; we can quit the loop.
     }
     if ($status) {
        &popup_error('All attempts to move '.$obj.' motors'
                    .' to the "IN" position failed.');
        return 1;
     }
# } else {
#    $msg = sprintf 'Did not have to move '.$obj.' IN to %.3f (device in place)',$bs_in_pos;
  }
  $state{$obj} =~ s/\/notinplace//i;			#now motor is in place or it does not matter.

### If beamstop flip not where we need it:
  if ($state{$obj} ne $dest) {
     if (&close_disable_shutter()) {return 1;};
     $attempt = 0;
     do {
        $attempt++;
### Start flipping beamstop IN or OUT
        &message('Flipping '.$obj.' '.uc($dest).', attempt '.$attempt.'/'.$attemptsMax);
        $status = Pezca::PutDouble($beamstopflip_cmd_PV,$beamstopflip_cmd_pos);	#digital I/O
        &ezca_err($script,$status,$beamstopflip_cmd_PV,
                  '. Flipping '.$obj.' '.uc($dest).' failed');
### Check if the command was accepted (SDIS?)
        $status = &wait4pv($beamstopflip_rbk_PV, $beamstopflip_cmd_pos,
                           $confirm_timeout, $flip_poll, $flip_polldelay,
                           $flip_counter, $flip_tolerance, $progress);
        if ($status == -1) {							#PV error
           &ezca_warn($script,$status,$beamstopflip_rbk_PV,
                      '. Taking the commanded position');
           $beamstopflip_rbk_pos = $beamstopflip_cmd_pos;
        }
        elsif ($status == 1) {							#timeout on waiting for expected value
           $beamstopflip_rbk_pos = 1 - $beamstopflip_cmd_pos;
           &message('Operation "'.$obj.'-'.uc($dest).'" failed: readback did not'
                   .' confirm the command within '.$confirm_timeout.'s','warn');
           ($status, $beamstopflip_sdis) = Pezca::GetDouble($beamstopflip_sdis_PV);
           if ($status) {
              &ezca_warn($script,$status,$beamstopflip_sdis_PV,'. Presuming SDIS=0');
              $beamstopflip_sdis = 0;
           }
           if ($beamstopflip_sdis) {
              &message($obj.' flip disabled by these conditions (SDIS):','error');
              foreach $pv (@pvlist) {
                 ($status, $val) = Pezca::GetDouble($pv);
                 if ($status) {&report_only($script,$status,$pv); $val='undef';}
                 &message($pv.' = '.$val,'error');
              }
              select(undef,undef,undef,0.5);
           } else {
              &message('Clearing previous cmd to flip '.$obj
                      .' while it was disabled (SDIS)');
              $status = Pezca::PutDouble($beamstopflip_cmd_PV,1-$beamstopflip_cmd_pos);
              &ezca_warn($script,$status,$beamstopflip_cmd_PV);
              select(undef,undef,undef,0.5);
           }
        }
        else {
           $beamstopflip_rbk_pos = $beamstopflip_cmd_pos;			#OK: readback confirmed the command
        }
     } while ($attempt < $attemptsMax && $beamstopflip_cmd_pos != $beamstopflip_rbk_pos);
### Wait for beamstop flip to finish:
     if ($attempt >= $attemptsMax) {$status = 1;}
     else {
        $status = &wait4pv($beamstopflip_act_PV,$beamstopflip_cmd_pos,
                           $flip_timeout, $flip_poll, $flip_polldelay,
                           $flip_counter, $flip_tolerance, $progress);
     }
     if ($status) {
### Bring the request and the actual in sync (safety mat issue):
        Pezca::PutDouble($beamstopflip_cmd_PV,1-$beamstopflip_cmd_pos);		#digital I/O
        &popup_error('Timeout on flipping '.$obj.' '.uc($dest));
        return 1;
     } else {
        select(undef,undef,undef,1.5);						#sleep a bit extra;
        &message($obj.' has been flipped '.uc($dest));
        $state{$obj} = $dest;
     }
  }
  return 0;
}

#=================================================================

# sub movePindiode($obj,$dest) {
sub movePindiode ($$) {
  my ($obj, $dest, $pin_dest_pos, $pin_act_pos);
  my ($motionTime, $attempt, $msg);

  $obj  = shift(@_);
  $dest = shift(@_);

  if    ($dest eq 'in')  {$pin_dest_pos = $pin_in_pos;}
  elsif ($dest eq 'out') {$pin_dest_pos = $pin_out_pos;}
  else {
     &popup_error('Unknown destination "'.$dest.'" for '.$obj
                 .' while preparing to move.');
     return 1;
  }

### If pin diode is not where we want it to be:
  if ($state{$obj} ne $dest) {
     if (&close_disable_shutter()) {return 1;}
     $pin_act_pos = &getAxisAct ($pinAxis_PV);
     $msg = sprintf 'Moving '.$obj.' '.uc($dest).' from %.3f to %.3f',$pin_act_pos,$pin_dest_pos;
     &message($msg);
     $status = &checkIfAssyMovable($pinAssy_PV,$obj);
     if ($status) {return 1;}
### Start moving pin diode:
     for ($attempt=1; $attempt<=$attemptsMax; $attempt++) {
### Setup motion:
        $status = &configAssy ($pinAssy_PV);
        if ($status) {
           &message($obj.' motor setup err='.$status.', attempt '.$attempt.'/'.$attemptsMax,'warn');
           next;
        }
        $motionTime = &calcMotorTime($pinAssy_PV, $pinAxis_PV, $pin_dest_pos);
#       $status = &runMotor ($pinAssy_PV, $pinAxis_PV, $pin_dest_pos, 'checkAuto');
        $status = &startMotor ($pinAxis_PV, $pin_dest_pos, $pinAssy_PV, 'checkAuto');
        if ($status) {
           &message($obj.' motor start err='.$status.', attempt '.$attempt.'/'.$attemptsMax,'warn');
           next;
        }
### Track motion:
        $status = &trackMotor ($pinAssy_PV, $motionTime, $pinAxis_PV);
        if ($status) {
           &message($obj.' motion track err='.$status.', attempt '.$attempt.'/'.$attemptsMax,'warn');
           next;
        }
        last;						#operation successful; we can quit the loop.
     }
     if ($status) {
        &popup_error('All attempts to move '.$obj.' '.uc($dest).' failed');
        return 1;
     } else {
        &message($obj.' has been moved '.uc($dest));
        $state{$obj} = $dest;
     }
  } else {
     $msg = sprintf 'Did not have to move '.$obj.' '.uc($dest)
                   .' to %.3f (device already in place)', $pin_dest_pos;
     &message($msg)
  }
  return 0;
}

#=================================================================

# sub moveSONICC($obj,$dest) {
sub moveSONICC ($$) {
  my ($obj, $dest, $sonicc_dest_pos, $sonicc_act_pos);
  my ($motionTime, $attempt, $msg);

  $obj  = shift(@_);
  $dest = shift(@_);

  if    ($dest eq 'in')  {$sonicc_dest_pos = $sonicc_in_pos;}
  elsif ($dest eq 'out') {$sonicc_dest_pos = $sonicc_out_pos;}
  else {
     &popup_error('Unknown destination "'.$dest.'" for '.$obj
                 .' while preparing to move.');
     return 1;
  }

### If SONICC is not where we want it to be:
  if ($state{$obj} ne $dest) {
     if (&close_disable_shutter()) {return 1;}
     $sonicc_act_pos = &getAxisAct ($soniccAxis_PV);
     $msg = sprintf 'Moving '.$obj.' '.uc($dest).' from %.3f to %.3f',$sonicc_act_pos,$sonicc_dest_pos;
     &message($msg);
     $status = &checkIfAssyMovable($soniccAssy_PV,$obj);
     if ($status) {return 1;}
### Start moving SONICC:
     for ($attempt=1; $attempt<=$attemptsMax; $attempt++) {
### Setup motion:
        $status = &configAssy ($soniccAssy_PV);
        if ($status) {
           &message($obj.' motor setup err='.$status.', attempt '.$attempt.'/'.$attemptsMax,'warn');
           next;
        }
        $motionTime = &calcMotorTime($soniccAssy_PV, $soniccAxis_PV, $sonicc_dest_pos);
#       $status = &runMotor ($soniccAssy_PV, $soniccAxis_PV, $sonicc_dest_pos, 'checkAuto');
        $status = &startMotor ($soniccAxis_PV, $sonicc_dest_pos, $soniccAssy_PV, 'checkAuto');
        if ($status) {
           &message($obj.' motor start err='.$status.', attempt '.$attempt.'/'.$attemptsMax,'warn');
           next;
        }
### Track motion:
        $status = &trackMotor ($soniccAssy_PV, $motionTime, $soniccAxis_PV);
        if ($status) {
           &message($obj.' motion track err='.$status.', attempt '.$attempt.'/'.$attemptsMax,'warn');
           next;
        }
        last;						#operation successful; we can quit the loop.
     }
     if ($status) {
        &popup_error('All attempts to move '.$obj.' '.uc($dest).' failed');
        return 1;
     } else {
        &message($obj.' has been moved '.uc($dest));
        $state{$obj} = $dest;
     }
  } else {
     $msg = sprintf 'Did not have to move '.$obj.' '.uc($dest)
                   .' to %.3f (device already in place)', $sonicc_dest_pos;
     &message($msg);
  }
  return 0;
}

#=================================================================

# sub startCollimator($obj,$dest) {			#returns ($status,$col_started,$col_motionTime)
sub startCollimator ($$) {
  my ($obj, $dest, $attempt, $motionTime, $msg, $i);
  my (@col_act, @coa_act, $mt2);
  $obj  = shift(@_);
  $dest = shift(@_);

  if    ($dest eq 'in')  {@col_dest_pos = @col_in_pos;}
  elsif ($dest eq 'out') {@col_dest_pos = @col_out_pos;}
  else {
     &popup_error('Unknown destination "'.$dest.'" for '.$obj
                 .' while preparing to move.');
     return 1,0,0;
  }
  if ($hooked{Colliangle}) {
     if    ($dest eq 'in')  {@coa_dest_pos = @coa_in_pos;}
     elsif ($dest eq 'out') {@coa_dest_pos = @coa_out_pos;}
  }
### Is motion is needed?
  if ($state{$obj} ne $dest) {
### Get actual pos:
     for ($i=0; $i<2; $i++) {$col_act[$i] = &getAxisAct ($colAxes_PV[$i]);}
     print  STDOUT $scr.': Moving '.$obj.'-'.uc($dest).":\n";
     printf STDOUT '   --> from V=%.3f to V=%.3f'."\n", $col_act[1], $col_dest_pos[1];
     printf STDOUT '   --> from H=%.3f to H=%.3f'."\n", $col_act[0], $col_dest_pos[0];
     if ($hooked{Colliangle}) {
        for ($i=0; $i<2; $i++) {$coa_act[$i] = &getAxisAct ($coaAxes_PV[$i]);}
        printf STDOUT '   --> from P=%.3f to P=%.3f'."\n", $coa_act[0], $coa_dest_pos[0];
        printf STDOUT '   --> from Y=%.3f to Y=%.3f'."\n", $coa_act[1], $coa_dest_pos[1];
     }

###==========Start collimator motion: Loop over attempts ===================
     $status = &checkIfAssyMovable($colAssy_PV,$obj);
     if ($status) {return 1,0,0;}
     $motionTime = &waitMotor ($colAssy_PV,5,0);			#give collimator 5s to finish prev.motion
     if ($hooked{Colliangle}) {
        $status = &checkIfAssyMovable($coaAssy_PV,$obj);
        if ($status) {return 1,0,0;}
        $motionTime = &waitMotor ($coaAssy_PV,5,0);			#give collimator 5s to finish prev.motion
     }
     for ($attempt=1; $attempt<=$attemptsMax; $attempt++) {
### Setup motion:
        $status = &configAssy ($colAssy_PV);
        if ($status) {
           &message($obj.' assembly ['.$colAssy_PV.'] setup err='.$status.', attempt '.$attempt.'/'.$attemptsMax,'warn');
           next;
        }
        if ($hooked{Colliangle}) {
           $status = &configAssy ($coaAssy_PV);
           if ($status) {
              &message($obj.' assembly ['.$coaAssy_PV.'] setup err='.$status.', attempt '.$attempt.'/'.$attemptsMax,'warn');
              next;
           }
        }
### Start motion:
        $status = &startNMotors ($colAssy_PV, 2, \@colAxes_PV, \@col_dest_pos, 'checkAuto');
        if ($status) {
           &message($obj.' assembly ['.$colAssy_PV.'] start err='.$status.', attempt '.$attempt.'/'.$attemptsMax,'warn');
           next;
        }
        if ($hooked{Colliangle}) {
           $status = &startNMotors ($coaAssy_PV, 2, \@coaAxes_PV, \@coa_dest_pos, 'checkAuto');
           if ($status) {
              &message($obj.' assembly ['.$coaAssy_PV.'] start err='.$status.', attempt '.$attempt.'/'.$attemptsMax,'warn');
              next;
           }
        }
        $motionTime = &calcNMotorsTime($colAssy_PV, 2, \@colAxes_PV, \@col_dest_pos);
        if ($hooked{Colliangle}) {
           $mt2 = &calcNMotorsTime($coaAssy_PV, 2, \@coaAxes_PV, \@coa_dest_pos);
           if ($mt2 > $motionTime) {$motionTime = $mt2;}
        }
        last;							#operation successful; we can quit the loop.
     }
###============== End of Loop over attempts to start collimator ============

     if ($status) {
        &popup_error('All attempts to start '.$obj.' '.uc($dest).' failed');
        return 1,0,0;
     } else {
        &message($obj.' started motion to '.uc($dest).' OK');
        $state{$obj} = 'unknown';
        return 0,1,$motionTime;
     }
  }
  else  {
     $msg = sprintf 'Did not move '.$obj.' '.uc($dest).' to V=%.3f  H=%.3f'
                   .' as it had been in place', $col_dest_pos[1], $col_dest_pos[0];
     &message($msg);
     if ($hooked{Colliangle}) {
        ### Just in case adjust angle (we have no way to know if it is proper)
        $status = &runNMotors ($coaAssy_PV, 2, \@coaAxes_PV, \@coa_dest_pos, 'checkAuto');
        if ($status) {
           &message($obj.' assembly ['.$coaAssy_PV.'] move to position err='.$status,'warn');
        }
     }
     return 0,0,0;
  }
}

#=================================================================

# sub finishCollimator($obj,$dest) {
sub finishCollimator ($$) {
  my ($obj, $dest, $motionTime, $mt2, $attempt);
  $obj  = shift(@_);
  $dest = shift(@_);

  $col_motionTime -= time() - $col_start_time;
  if ($col_motionTime < 0) {$col_motionTime = 0;}
  $status = &trackNMotors($colAssy_PV, $col_motionTime, 2, \@colAxes_PV);
  if ($status) {
     &message('Initial attempts to move '.$obj.' err='.$status.'. Making '.$attemptsMax.' more attempts','warn');
  } else {
     if ($hooked{Colliangle}) {
        $status = &trackNMotors($coaAssy_PV, $col_motionTime, 2, \@coaAxes_PV);
        if ($status) {
           &message('Failed to move '.$obj.'/Angle, err='.$status.'. Making '.$attemptsMax.' more attempts','warn');
        }
     } else {
        $status = 0;
     }
     if ($status == 0) {
        ### Motion finished OK. Check new state one more time:
        &getState($obj);
        if ($state{$obj} eq $dest) {
           &message($obj.' has been moved '.uc($dest));
           return 0;
        } else {
           &message($obj.' has not moved '.uc($dest).', trying more');
        }
     }
  }

###==========Start collimator motion: Loop over attempts ====================
  for ($attempt=1; $attempt<=$attemptsMax; $attempt++) {
### Setup motion:
     $status = &configAssy ($colAssy_PV);
     if ($status) {
        &message($obj.' motor setup err='.$status.', attempt '.$attempt.'/'.$attemptsMax,'warn');
        next;
     }
     if ($hooked{Colliangle}) {
        $status = &configAssy ($coaAssy_PV);
        if ($status) {
           &message($obj.' assembly ['.$coaAssy_PV.'] setup err='.$status.', attempt '.$attempt.'/'.$attemptsMax,'warn');
           next;
        }
     }
### Start motion:
     $status = &startNMotors ($colAssy_PV, 2, \@colAxes_PV, \@col_dest_pos, 'checkAuto');
     if ($status) {
        &message($obj.' motor start err='.$status.', attempt '.$attempt.'/'.$attemptsMax,'warn');
        next;
     }
     if ($hooked{Colliangle}) {
        $status = &startNMotors ($coaAssy_PV, 2, \@coaAxes_PV, \@coa_dest_pos, 'checkAuto');
        if ($status) {
           &message($obj.' assembly ['.$coaAssy_PV.'] start err='.$status.', attempt '.$attempt.'/'.$attemptsMax,'warn');
           next;
        }
     }
     $motionTime = &calcNMotorsTime($colAssy_PV, 2, \@colAxes_PV, \@col_dest_pos);
     if ($hooked{Colliangle}) {
        $mt2 = &calcNMotorsTime($coaAssy_PV, 2, \@coaAxes_PV, \@coa_dest_pos);
        if ($mt2 > $motionTime) {$motionTime = $mt2;}
     }
### Track motion:
     $status = &trackNMotors($colAssy_PV, $motionTime, 2, \@colAxes_PV);
     if ($status) {
        &message($obj.' assembly ['.$colAssy_PV.'] track err='.$status.', attempt '.$attempt.'/'.$attemptsMax,'warn');
        next;
     }
     if ($hooked{Colliangle}) {
        $status = &trackNMotors($coaAssy_PV, $col_motionTime, 2, \@coaAxes_PV);
        if ($status) {
           &message($obj.' assembly ['.$coaAssy_PV.'] track err='.$status.', attempt '.$attempt.'/'.$attemptsMax,'warn');
           next;
        }
     }
     last;						#operation successful; we can quit the loop.
  }
###============== End of Loop over attempts to move collimator =============

  if ($status) {
     &popup_error('All attempts to move '.$obj.' '.uc($dest).' failed');
     return 1;
  }
  else {
     &message($obj.' has been moved '.uc($dest));
     $state{$obj} = $dest;
  }
  return 0;
### End of Collimator motion with pre-checks #################
}

#=================================================================

# sub sortoutGuardSlits() { 				#see if we need to open/close guard slits
sub sortoutGuardSlits() {

  my (@gsMinima_um, @pinholes, $nval, $gsSizeV, $gsSizeH, $size, $string, $msg, $i);

  @pinholes = (20, 10, 5);				#in um

  if (! $hooked{Guardslits} || ! $adjust_gs) {return;}
# &message('Entering sortoutGuardSlits','warn');

### 0=none, 1=300um, 2=20um, 3=10um, 4=5um
  if ($col_active_preset >= 2) { #------------------------------#when pin is IN

     ($status,$string) = Pezca::GetString($gsMinima_PV);
     if ($status) {
        &ezca_warn($script, $status, $gsMinima_PV,'. Ignoring the Guard Slits');
        return 0;
     }
     elsif (length($string) < 5) {
        &message('"'.$gsMinima_PV.'" contains incorrect record="'.$string.'". Skipping GS','warn');
        return 0;
     }
     $string  =~ s/,/ /g;
     @gsMinima_um = split ' ', $string;
     $nval = @gsMinima_um;
     if ($nval != 3 && $nval != 1) {
        &message('"'.$gsMinima_PV.'" contains incorrect record="'.$string.'". Skipping GS','warn');
        return 0;
     }
     if ($nval == 1) {
        $gsMinima_um[1] = $gsMinima_um[0];
        $gsMinima_um[2] = $gsMinima_um[0];
        $nval = 3;
     }
     for ($i=0; $i<$nval; $i++) {
        $gsMinima_um[$i] =~ s/um//;
        if (! &is_decimal($gsMinima_um[$i]) || $gsMinima_um[$i] < 0) {
           &message('"'.$gsMinima_PV.'" contains incorrect record="'.$string.'". Skipping GS','warn');
           return 0;
        }
        elsif ($gsMinima_um[$i] < $pinholes[$i]) {
           &message('Too small values "'.$string.'" in "'.$gsMinima_PV.'". Skipping GS','warn');
           return 0;
        }

     }

     $size = 0.001*($gsMinima_um[$col_active_preset-2] + 1.0);	#0=none, 1=300um, 2=20um, 3=10um, 4=5um

     $gsSizeV = &getAxisAct ($gsSizeV_PV);
     if ($gsSizeV < $size) {
        $status = &checkIfAssyMovable($gsAssyV_PV,'GS-vertical size');
        if ($status) {return 1;}
        ### Open guard slits-vertical:
        $msg = sprintf 'Opening GS-vertical size to %.3fum', 1000*$size;
        &message($msg);
        $status = &startMotor($gsSizeV_PV, $size, $gsAssyV_PV, 'checkauto');
        if ($status) {
           $msg = sprintf 'Failed to open GS-vertical size "'.$gsSizeV_PV.'" to %fum', 1000*$size;
           &message($msg,'warn');
        }
     } else {
        $msg = sprintf 'Not touching GS-vertical because GS-size=%.3fum > pinhole=%.3fum', 1000*$gsSizeV, 1000*$size;
        &message($msg);
     }

     $gsSizeH = &getAxisAct ($gsSizeH_PV);
     if ($gsSizeH < $size) {
        $status = &checkIfAssyMovable($gsAssyH_PV,'GS-horizontal size');
        if ($status) {return 1;}
        ### Open guard slits-horizontal:
        $msg = sprintf 'Opening GS-horizontal size to %.3fum', 1000*$size;
        &message($msg);
        $status = &startMotor($gsSizeH_PV, $size, $gsAssyH_PV, 'checkauto');
        if ($status) {
           $msg = sprintf 'Failed to open GS-horizontal size "'.$gsSizeH_PV.'" to %fum', 1000*$size;
           &message($msg,'warn');
        }
     } else {
        $msg = sprintf 'Not touching GS-horizontal because GS-size=%.3fum > pinhole=%.3fum', 1000*$gsSizeH, 1000*$size;
        &message($msg);
     }

  } elsif ($col_active_preset == 1) { #-------------------------#when scatterguard is IN

### Do nothing -- do not close Guard Slits back
        &message('Not touching GS because 300um scatterguard is IN');

  } else {
        &message('Not touching GS -- unexpected','warn');
  }                              #------------------------------#end of pin or scatterguard
  return 0;
}

#=================================================================

# sub controlRinglightBrightness() {
sub controlRinglightBrightness () {
  my ($msg);

### Ringlight brightness: determine destination or save current intensity:
  if (!defined $ringbright_val) {				#val not passed via cmd line
### Read destination from previously saved:
     ($status,$ringbright_val) = Pezca::GetDouble($ringbright_save_PV);
     if ($status) {
        &ezca_warn($script,$status,$ringbright_save_PV,'. Presuming value=0');
        $ringbright_val = 0;
     }
  }
  elsif ($ringbright_val == 0) {				#request to dim ringbright
### Read current intensity:
     ($status,$ringbright_val) = Pezca::GetDouble($ringbright_ctrl_PV);
     if ($status) {
        &ezca_warn($script,$status,$ringbright_ctrl_PV,'. Operation skipped');
        $ringbright_ctrl = 0;
     } else {
### Save current intensity for future restoration:
        $status = Pezca::PutDouble($ringbright_save_PV,$ringbright_val);
        &ezca_warn($script,$status,$ringbright_save_PV,'. Ignoring');
     }
     $ringbright_val = 0;
  }

  if ($ringbright_ctrl) {
### Control ringbright:
     $status = Pezca::PutDouble($ringbright_ctrl_PV,$ringbright_val);
     if ($status) {
        &ezca_warn($script,$status,$ringbright_ctrl_PV,'. Failed to set ringlight brightness');
     } else {
        $msg = sprintf 'Ringlight brightness set to %.3f',$ringbright_val;
        &message($msg);
     }
  }
}

#=================================================================

# sub controlBacklightBrightness() {
sub controlBacklightBrightness () {
  my ($msg);

### Backlight brightness: determine destination or save current intensity:
  if (!defined $backbright_val) {				#val not passed via cmd line
### Read destination from previously saved:
     ($status,$backbright_val) = Pezca::GetDouble($backbright_save_PV);
     if ($status) {
        &ezca_warn($script,$status,$backbright_save_PV,'. Presuming saved backlight brightness=0');
        $backbright_val = 0;
     }
  }
  elsif ($backbright_val == 0) {				#request to dim backbright
### Read current intensity:
     ($status,$backbright_val) = Pezca::GetDouble($backbright_ctrl_PV);
     if ($status) {
        &ezca_warn($script,$status,$backbright_ctrl_PV,'. Operation skipped');
        $backbright_ctrl = 0;
     } else {
### Save current intensity for future restoration:
        $status = Pezca::PutDouble($backbright_save_PV,$backbright_val);
        &ezca_warn($script,$status,$backbright_save_PV,'. Current backlight brightness was not saved');
     }
     $backbright_val = 0;
  }

  if ($backbright_ctrl) {
### Control backbright:
     $status = Pezca::PutDouble($backbright_ctrl_PV,$backbright_val);
     if ($status) {
        &ezca_warn($script,$status,$backbright_ctrl_PV,'. Setting backlight brightness failed');
     } else {
        $msg = sprintf 'Backlight brightness set to %.3f',$backbright_val;
        &message($msg);
     }
  }
}

#=================================================================

# sub close_disable_shutter () {
sub close_disable_shutter () {
  my ($shutter_polldelay, $shutter_counter, $shutter_tolerance, $status);

  if (&getShutter($shutter_readback)) {			#if shutter is open
     &closeShutter($shutter,$shutter_readback,$shutter_assy);
     if (!defined $shutter_open) {			#must be in beamline.pl
        &press_key_exit('*** '.$script.': global variable "shutter_open" is undefined.');
     }
     $shutter_polldelay = 0.05;
     $shutter_counter   = 1;
     $shutter_tolerance = 0.01;			#destination tolerance
     $status = &wait4pv($shutter_readback, 1-$shutter_open,
                        $shutter_timeout, $shutter_poll, $shutter_polldelay,
                        $shutter_counter, $shutter_tolerance, $progress);
     if ($status) {					#if shutter is open
        &popup_error('Could not close shutter.');
        return 1;
     }
  }
  &setAssyManual ($shutter_assy);                       #disable shutter
  return 0;
}

#=================================================================

# sub ezca_err ($script, $status, $PV, [$msg]) {
sub ezca_err ($$$;$) {
  my ($script, $status, $PV, $statusDescr, $msg);
  $script = shift(@_);
  $status = shift(@_);
  $PV     = shift(@_);
  $msg    = shift(@_);
  if (!defined $msg) {$msg = '';}
  if ($msg !~ /\.$/) {$msg .= '.';}
  if ($status) {
     $statusDescr  = &ezcaStatusCode($status);
     &popup_error('EZCA error='.$status.' ['.$statusDescr.']'.' on PV=['.$PV.']'.$msg);
     return 1;
  } else {
     return 0;
  }
}

#=================================================================

# sub ezca_warn ($script, $status, $PV, [$msg]) {
sub ezca_warn ($$$;$) {
  my ($script, $status, $PV, $statusDescr, $msg);
  $script = shift(@_);
  $status = shift(@_);
  $PV     = shift(@_);
  $msg    = shift(@_);
  if (!defined $msg) {$msg = '';}
  if ($msg !~ /\.$/) {$msg .= '.';}
  if ($status) {
     $statusDescr = &ezcaStatusCode($status);
     &message('EZCA error='.$status.' ['.$statusDescr.'] on PV=['.$PV.']'.$msg, 'warn');
  }
}

#=================================================================

# sub popup_error ($msg) {
sub popup_error ($) {
  my ($msg, $type, $prefix, $stamp);
  $msg  = shift(@_);
  &message($msg, 'error');
  &zenity_error('*** '.$scr.': '.$msg, undef, 'nowait');
}

#=================================================================

# sub message ($msg,[$type]) {		#type=log_note,error,warn
sub message ($;$) {
  my ($msg, $type, $prefix, $stamp);
  $msg  = shift(@_);
  $type = shift(@_);
  if (!defined $type) 	      {$type = 'log_note';}
  if    ($type =~ /^error$/i) {$prefix = '*** ';}
  elsif ($type =~ /^warn/i)   {$prefix = '!!! ';}
  else                        {$prefix = '';}
  if ($msg !~ /\.$/) {$msg .= '.';}
  $stamp = scalar localtime;                                                           #
  $msg = $scr.': '.$stamp.' '.$msg;
  print STDOUT $prefix.$msg."\n";
  if ($tell_bluice) {&tellBluIce($msg,$type);}
}

#=================================================================

# sub clearErr () {
sub clearErr () {
  my ($status);
  $status = &PezcaPutStringWF($sampleEnvPV,'failed','oldca',$nelm);
  &report_only($script,$status,$sampleEnvPV);
  &setAssyAuto ($shutter_assy);
  &sortout_vertPosFeedback();
}

#=================================================================

# sub sortout_vertPosFeedback () {
sub sortout_vertPosFeedback () {
  my ($status, $PINstate, $BSstate, $resume);
### Decide if we should resume vertical beam position feedback (FPGA/Sioan)
  if ($vertPosFeedbackPresent && $vertPosFeedbackFPGAstate > 0) {
     $resume = 0;
     if ($vertPosFeedbackUsePin && $hooked{'Pindiode'}) {
        $status = &getState('Pindiode',\$PINstate);
        if ($status) {$PINstate = 'unknown';}
        if ($PINstate eq 'in') {$resume = 1;}
     }
     if ($vertPosFeedbackUseBS && $hooked{'Beamstop'}) {
        $status = &getState('Beamstop',\$BSstate);
        if ($status) {$BSstate = 'unknown';}
        if ($BSstate eq 'in') {$resume = 1;}
     }
     if ($resume) {
        &message('resuming vert.pos. feedback');
        $status = &setVertPosFeedbackFPGAstate($beamline,$vertPosFeedbackFPGAstate);
        if ($status) {
           &message('failed to resume vert.pos. feedback', 'error');
        }
     }
  }
}

#=================================================================

sub signal_handler {

  our ($ABORT_IN_PROCESS);
  my $signame = shift(@_);
  print STDOUT "\n".$script.': *** Caught a SIG'.$signame."\n";

  if (defined $ABORT_IN_PROCESS) {return;}
  $ABORT_IN_PROCESS = 1;
  &abort_external_script();

  &setAssyAuto ($shutter_assy);
  &closeShutter($shutter);
  &clearAllCaMonitors();
  print STDOUT 'Done!'."\n\n";
  exit 1;
}
