#!/usr/bin/perl -W
#--------------------------------------------------------------
#
# This script is a GUI for saving "good" CRL positions in mySQL.
# Data then can be used for plotting statistics and in bigEnergyMove
# to retune CRL when energy is changed (e.g. using slope of pitch,
# yaw, height, or downstream changes vs energy).
#
#                       Author: Sergey Stepanov
#
# Version-1.0:  2019/02/05	initial implementation
#--------------------------------------------------------------
# Subs: &main
#       &refresh_gui();
#       &dumpPositions();
#       &savePoositions();
#       &updateLogBox($msg,[$tag]);
#       &disable_buttons();
#       &enable_buttons();
#       &goodbye();
#--------------------------------------------------------------
# use if ! defined &DBI,   'DBI';
# use if ! defined &Tk,    'Tk';
# use if ! defined &Pezca, 'Pezca';
  use DBI;
  use Tk;
  use Tk::ROText;
  use Pezca;
  use strict;
  use warnings;
# use warnings::unused;

  our ($NONINTERACTIVE, $USE_ZENITY_FOR_EXIT, $IGNORE_ERRORS, $nRetry);
  our ($beamline, $mw, $repeater, $N, $fBtns, $fLog, $logBox);
  our ($saveBtn, $dumpBtn, $fNrec, $lblNrec, $entryNrec);

  my ($PEZCA, $script, $btnWidth);
  my ($btnbkgcolor, $bactivcolor, $bdisblcolor, $fborderwidth, $frelief);
  my ($greencolor, $goldcolor, $redcolor, $yellowcolor);
  my ($msg, $status, $host, @hostlist, $col);
  my ($Mono_En_assy, $Mono_Ekev_PV);
  my ($CRL_Ang_assy, $CRL_StB_assy, $CRL_StS_assy);
  my ($CRL_pitch_PV, $CRL_yaw_PV);
  my ($CRL_vertS_PV, $CRL_horzS_PV);
  my ($CRL_vertB_PV, $CRL_dwstB_PV);
  my ($header, $format);

  if (!defined $ENV{PEZCA}) {
     die 'Please define the PEZCA environment. It must point to the PERL scripts repository';
  }
  $PEZCA = $ENV{PEZCA};
  $PEZCA =~ s/\\/\//g;
  if ($PEZCA !~ /\/$/) {$PEZCA .='/';}

  do $PEZCA.'Common/common_subs.pl';
  do $PEZCA.'Common/beamline.pl';

  select(STDERR);  $|=1;							#Flush output after each print
  select(STDOUT);  $|=1;							#Flush output after each print

# local *STDOUT;								#I am not sure this is needed
  open STDOUT, '>/dev/null' or warn "Can't open /dev/null: $!";			#redirect all prints to /dev/null
  
  $script = &makeScriptName();

  $NONINTERACTIVE      = 1;
  $USE_ZENITY_FOR_EXIT = 1;
  $IGNORE_ERRORS       = 1;

  $btnbkgcolor = '#0097FF';
  $bactivcolor = '#00D8FF';
  $bdisblcolor = '#A3A3A3';
  $greencolor  = '#00BC00';
  $goldcolor   = '#D0D000';
  $redcolor    = '#FF0000';
  $yellowcolor = '#E0E000';

######################################################################################
  @hostlist = qw(bl1ws2 bl1ws3 bl1keithley bl1upper);
  $host = &gethost('short');
  if (! grep {$host =~ /^$_$/} @hostlist) {
     &zenity_error_exit($script.': cannot run on host='.$host.'.'."\n".'Please use one of: '."[@hostlist]");
  }
  $nRetry  = 5;
  &setEpicsRetry();
  Pezca::AutoErrorMessageOff();
  $N = 5;                                                                       # how many last positions to save
######################################################################################
  $fborderwidth = 0; $frelief = 'flat';						# this is normal setting
# $fborderwidth = 2; $frelief = 'groove';					# this is for debugging (comment out for normal operations)
######################################################################################
  $Mono_En_assy = $beamline.':MO:En:InPos';
  $Mono_Ekev_PV = $beamline.':MO:E:ActPos';
  $CRL_Ang_assy = $beamline.':CRL:An:InPos';
  $CRL_StS_assy = $beamline.':CRL:StS:InPos';
  $CRL_StB_assy = $beamline.':CRL:StB:InPos';
  $CRL_pitch_PV = $beamline.':CRL:AP:ActPos';
  $CRL_yaw_PV   = $beamline.':CRL:AW:ActPos';
  $CRL_vertS_PV = $beamline.':CRL:YS:ActPos';
  $CRL_horzS_PV = $beamline.':CRL:XS:ActPos';
  $CRL_vertB_PV = $beamline.':CRL:YB:ActPos';
  $CRL_dwstB_PV = $beamline.':CRL:ZB:ActPos';
 
  $header = '__Energy_______Date___________Pitch_______Yaw____Height__Outboard___Dwnstrm_BigHeight';
#           '12345678 1234567890123456 123456789 123456789 123456789 123456789 123456789 123456789';
  $format = '%8.3f %16s %9.3f %9.3f %9.3f %9.3f %9.3f %9.3f';
  
######################################################################################
### Create a GUI:
  $mw         = MainWindow->new;
# $mw->geometry('1600x1000');

  $btnWidth = 45;
  $fBtns      = $mw->Frame(-borderwidth=>$fborderwidth,
                           -relief=>$frelief,                        )->grid(-row=>0, -sticky=>'n',);
  $col = 0;
  $saveBtn   = $fBtns->Button(-text=>'Save good CRL positions',
                               -background=>$btnbkgcolor,
                               -activebackground=>$bactivcolor,
                               -disabledforeground=>$bdisblcolor,
                               -width=>$btnWidth,
                               -command=>\&savePositions,            )->grid(-row=>0, -column=>$col++, -pady=>2, -padx=>12,);
###---------------------------
  $dumpBtn   = $fBtns->Button(-text=>'Print last N CRL positions',
                               -background=>$btnbkgcolor,
                               -activebackground=>$bactivcolor,
                               -disabledforeground=>$bdisblcolor,
                               -width=>$btnWidth,
                               -command=>\&dumpPositions,            )->grid(-row=>0, -column=>$col++, -pady=>2, -padx=>12);
###---------------------------
  $fNrec      = $fBtns->Frame(-borderwidth=>$fborderwidth,
                               -relief=>$frelief,                    )->grid(-row=>0, -column=>$col++, -pady=>5, -padx=>12);
  $lblNrec    = $fNrec->Label(-text =>'N or "a" for all ',           )->grid(-row=>0, -column=>0, -pady=>2, -padx=>2);
  $entryNrec  = $fNrec->Entry(-textvariable=>\$N,
                              -width=>10,
			      -justify=>'right',                     )->grid(-row=>0, -column=>1, -pady=>2, -padx=>2);
###===========================
  $fLog = $mw->Frame(-borderwidth=>$fborderwidth,
                     -relief=>$frelief,                              )->grid(-row=>1, -columnspan=>2, -sticky=>'nwe');
# $logBox = $fLog->Scrolled('Listbox',
# $logBox = $fLog->Scrolled('Text',                                       			#see: http://docstore.mik.ua/orelly/perl3/tk/ch06_02.htm
  $logBox = $fLog->Scrolled('ROText',                                       			#see: http://docstore.mik.ua/orelly/perl3/tk/ch06_02.htm
                            -scrollbars=>'se',
                            -height=>'20',
#			    -font=>[-family=>'courier',-size=>12,-weight=>'normal'],		#see: http://docstore.mik.ua/orelly/perl3/tk/ch03_01.htm
			    -font=>['courier',12,'normal'],					#in reality only size option works
 			    -foreground=>'#000077',
                            -background=>'#FFFFFF',                  )->pack(-side => 'top',-fill => 'x');
  $logBox->Subwidget("xscrollbar")->configure(-background=>$btnbkgcolor,);
  $logBox->Subwidget("yscrollbar")->configure(-background=>$btnbkgcolor,);
###===========================

### Set up binding:
  $mw->protocol('WM_DELETE_WINDOW', \&goodbye);

  $mw->resizable( 0, 0 );
  $mw->focus();

  $repeater = $mw->repeat(1000,\&refresh_gui);

  MainLoop;
  &goodbye;

#==========================================================================
sub refresh_gui() {
  if (Exists($mw)) {$mw->update;}
}

#==========================================================================
sub savePositions() {
  our ($mysqlHost, $mysqlUser, $mysqlPass, $mysqlDb);
  my ($mysqlTable, $dsn, $dbh, $sth, $nrecords, $hashref);
  my ($energy, $epoch, $date, $pitch, $yaw, $height, $outboard, $downstream, $bigheight);
  my (@ct, $min, $hour, $mday, $mon, $year, $status, $inpos, $msg, @keys, @values);
  $mysqlTable = 'CRLpositions';
  if (&addCaMonitor($Mono_En_assy)) {&updateLogBox('Error setting up monitor on PV='.$Mono_En_assy,'error'); return 1;}
  if (&addCaMonitor($Mono_Ekev_PV)) {&updateLogBox('Error setting up monitor on PV='.$Mono_Ekev_PV,'error'); return 1;}
  if (&addCaMonitor($CRL_Ang_assy)) {&updateLogBox('Error setting up monitor on PV='.$CRL_Ang_assy,'error'); return 1;}
  if (&addCaMonitor($CRL_StS_assy)) {&updateLogBox('Error setting up monitor on PV='.$CRL_StS_assy,'error'); return 1;}
  if (&addCaMonitor($CRL_StB_assy)) {&updateLogBox('Error setting up monitor on PV='.$CRL_StB_assy,'error'); return 1;}
  if (&addCaMonitor($CRL_pitch_PV)) {&updateLogBox('Error setting up monitor on PV='.$CRL_pitch_PV,'error'); return 1;}
  if (&addCaMonitor($CRL_yaw_PV))   {&updateLogBox('Error setting up monitor on PV='.$CRL_yaw_PV,'error');   return 1;}
  if (&addCaMonitor($CRL_vertS_PV)) {&updateLogBox('Error setting up monitor on PV='.$CRL_vertS_PV,'error'); return 1;}
  if (&addCaMonitor($CRL_horzS_PV)) {&updateLogBox('Error setting up monitor on PV='.$CRL_horzS_PV,'error'); return 1;}
  if (&addCaMonitor($CRL_vertB_PV)) {&updateLogBox('Error setting up monitor on PV='.$CRL_vertB_PV,'error'); return 1;}
  if (&addCaMonitor($CRL_dwstB_PV)) {&updateLogBox('Error setting up monitor on PV='.$CRL_dwstB_PV,'error'); return 1;}
  &disable_buttons();
# if (&zenity_question('Only good focus positions should be saved. Are you sure?','Yes','No') != 1) {
  if ($mw->messageBox(-icon=>'question',
#                     -background=>$btnbkgcolor,
#                     -activebackground=>$btnbkgcolor,
                      -foreground=>$btnbkgcolor,
                      -activeforeground=>$btnbkgcolor,
                      -message=>'Only good focus positions should be saved. Are you sure?',
                      -wrap=>0,
                      -title=>'Confirmation',
                      -type=>'YesNo',
                      -default=>'No') ne 'Yes') {
     &enable_buttons(); return -1;
  }
  ($status,$inpos) = Pezca::GetDouble($Mono_En_assy);
  if ($status || $inpos != 1) {
     &updateLogBox('Monochromator energy assembly is not positioned.','error'); 
     &enable_buttons(); return 1;
  }  
  ($status,$inpos) = Pezca::GetDouble($CRL_Ang_assy);
  if ($status || $inpos != 1) {
     &updateLogBox('CRL angular assembly is not positioned.','error'); 
     &enable_buttons(); return 1;
  }  
  ($status,$inpos) = Pezca::GetDouble($CRL_StS_assy);
  if ($status || $inpos != 1) {
     &updateLogBox('CRL small positioner assembly is not positioned.','error'); 
     &enable_buttons(); return 1;
  }  
  ($status,$inpos) = Pezca::GetDouble($CRL_StB_assy);
  if ($status || $inpos != 1) {
     &updateLogBox('CRL big positioner assembly is not positioned.','error'); 
     &enable_buttons();
     return 1;
  }  
  ($status,$energy) = Pezca::GetDouble($Mono_Ekev_PV);
  if ($status) {&updateLogBox('Error reading PV='.$Mono_Ekev_PV,'error'); &enable_buttons(); return 1;}
  else {$energy = sprintf '%.3f',$energy;}

  ($status,$pitch) = Pezca::GetDouble($CRL_pitch_PV);
  if ($status) {&updateLogBox('Error reading PV='.$CRL_pitch_PV,'error'); &enable_buttons(); return 1;}
  else {$pitch = sprintf '%.3f',$pitch;}

  ($status,$yaw) = Pezca::GetDouble($CRL_yaw_PV);
  if ($status) {&updateLogBox('Error reading PV='.$CRL_yaw_PV,'error');   &enable_buttons(); return 1;}
  else {$yaw = sprintf '%.3f',$yaw;}

  ($status,$height)     = Pezca::GetDouble($CRL_vertS_PV);
  if ($status) {&updateLogBox('Error reading PV='.$CRL_vertS_PV,'error'); &enable_buttons(); return 1;}
  else {$height = sprintf '%.3f',$height;}

  ($status,$outboard)   = Pezca::GetDouble($CRL_horzS_PV);
  if ($status) {&updateLogBox('Error reading PV='.$CRL_horzS_PV,'error'); &enable_buttons(); return 1;}
  else {$outboard = sprintf '%.3f',$outboard;}

  ($status,$downstream) = Pezca::GetDouble($CRL_dwstB_PV);
  if ($status) {&updateLogBox('Error reading PV='.$CRL_dwstB_PV,'error'); &enable_buttons(); return 1;}
  else {$downstream = sprintf '%.3f',$downstream;}

  ($status,$bigheight)  = Pezca::GetDouble($CRL_vertB_PV);
  if ($status) {&updateLogBox('Error reading PV='.$CRL_vertB_PV,'error'); &enable_buttons(); return 1;}
  else {$bigheight = sprintf '%.3f',$bigheight;}

  $epoch = time();
  @ct = localtime($epoch);
  ($min,$hour,$mday,$mon,$year) = ($ct[1],$ct[2],$ct[3],$ct[4]+1,$ct[5]+1900);
  $date = sprintf '%04s-%02s-%02s %02s:%02s', $year,$mon,$mday,$hour,$min;

  @keys   = qw(energy epoch date pitch yaw height outboard downstream bigheight);
  @values = ($energy,$epoch,'"'.$date.'"',$pitch,$yaw,$height,$outboard,$downstream,$bigheight);

# $msg = '('.join(',',@keys).')';    &updateLogBox($msg);
# $msg = '('.join(',',@values).')';  &updateLogBox($msg);
# &enable_buttons(); return 0;

  $dsn = 'DBI:mysql:host='.$mysqlHost.';database='.$mysqlDb;
  &updateLogBox('Connecting to mySQL database='.$mysqlDb.' on host='.$mysqlHost);
  $dbh = DBI->connect($dsn,$mysqlUser,$mysqlPass,{PrintError=>0,});
  if (!defined $dbh) {
     &updateLogBox('Failed connecting to mySQL DB='.$mysqlDb.' on '.$mysqlHost,'error'); 
     &enable_buttons(); return -1;
  }
  $sth = $dbh->prepare('INSERT INTO '.$mysqlTable.'('.join(',',@keys).') VALUES('.join(',',@values).')');
  if (!defined $sth) {
     &updateLogBox('Error on mySQL table='.$mysqlTable.'. -- '.$dbh->errstr,'error');
     $dbh->disconnect(); &enable_buttons(); return -1;
  }
  $status = $sth->execute();				#this is only defined/undefined
  if (!defined $status) {
     &updateLogBox('Failed executing INSERT for table='.$mysqlTable.'. -- '.$sth->errstr,'error');
     $sth->finish(); $dbh->disconnect(); &enable_buttons(); return -1;
  }
  &updateLogBox('CRL positions saved successfully.');
  &updateLogBox($header);
  $msg = sprintf $format,$energy,$date,$pitch,$yaw,$height,$outboard,$downstream,$bigheight;
  &updateLogBox($msg);
  &enable_buttons();
  return 0;
}

#==========================================================================
sub dumpPositions() {
  our ($mysqlHost, $mysqlUser, $mysqlPass, $mysqlDb);
  my ($mysqlTable, $dsn, $dbh, $sth, $nrecords, $hashref);
  my ($energy, $epoch, $date, $pitch, $yaw, $height, $outboard, $downstream, $bigheight);
  my ($istart, $i, $msg, @hashref);

  &disable_buttons();
  $mysqlTable = 'CRLpositions';
  $dsn = 'DBI:mysql:host='.$mysqlHost.';database='.$mysqlDb;
  &updateLogBox('Connecting to mySQL database='.$mysqlDb.' on host='.$mysqlHost);
  $dbh = DBI->connect($dsn,$mysqlUser,$mysqlPass,{PrintError=>0,});
  if (!defined $dbh) {
     &updateLogBox('Failed connecting to mySQL DB='.$mysqlDb.' on '.$mysqlHost,'error');
     &enable_buttons(); return -1;
  }
  $sth = $dbh->prepare('SELECT * FROM '.$mysqlTable);
  if (!defined $sth) {
     &updateLogBox('Error on mySQL table='.$mysqlTable.'. -- '.$dbh->errstr,'error');
     $dbh->disconnect(); &enable_buttons(); return -1;
  }
  $nrecords = $sth->execute();		#nrecords works for MySQL only. For SQLite result is def/undef only.
  if (!defined $nrecords) {
     &updateLogBox('Failed executing SELECT for table='.$mysqlTable.'. -- '.$sth->errstr,'error');
     $sth->finish(); $dbh->disconnect(); &enable_buttons(); return -1;
  }
  if ($nrecords < 1) {
     &updateLogBox('No saved CRL positions yet in table='.$mysqlTable);
     $sth->finish(); $dbh->disconnect(); &enable_buttons();
     return 0;
  }
  for ($i=0; $i<$nrecords; $i++) {$hashref[$i] = $sth->fetchrow_hashref();}
  $sth->finish(); 
  $dbh->disconnect();
  &sort_hashes_by_epoch(\@hashref);
  if ($N eq 'a' || $N eq 'all' || length($N) == 0) {
     &updateLogBox('All saved CRL positions:');
     $istart = 0;
  }
  elsif (&is_int($N)) {
     $msg = 'Last '.$N.' saved CRL positions';
     $istart = $nrecords - $N;
     if ($istart < 0) {
        $msg .= ' (only '.($N+$istart).' is available)';
        $istart = 0;
     }
  }
  else {
     &updateLogBox('Incorrect N=['.$N.']. Should be "a" or number.','error');
     &enable_buttons(); return -1;
  }
  &updateLogBox($msg.':');
  &updateLogBox($header);
  for ($i=$istart; $i<$nrecords; $i++) {
     $energy     = $hashref[$i]->{'energy'};
     $epoch      = $hashref[$i]->{'epoch'};
     $date       = $hashref[$i]->{'date'};
     $pitch      = $hashref[$i]->{'pitch'};
     $yaw        = $hashref[$i]->{'yaw'};
     $height     = $hashref[$i]->{'height'};
     $outboard   = $hashref[$i]->{'outboard'};
     $downstream = $hashref[$i]->{'downstream'};
     $bigheight  = $hashref[$i]->{'bigheight'};
     $msg = sprintf $format,$energy,$date,$pitch,$yaw,$height,$outboard,$downstream,$bigheight;
     &updateLogBox($msg);
  }
  &enable_buttons();
  return 0;
}

#==========================================================================
sub sort_hashes_by_epoch(\@) {
  my ($arref, $key, @sorted);
  $arref  = shift(@_);
  $key    = 'epoch';
  @sorted = sort { $a->{$key} <=> $b->{$key} } @$arref;
  @$arref = @sorted;
}

#==========================================================================
#sub updateLogBox($msg,[$tag]) {
sub updateLogBox($;$) {
  my ($msg, $tag);
  $msg = shift(@_);
  $tag = shift(@_);
  if (!defined $tag) {				#auto-detect tag, if not specified
     if ($msg =~ /^\*\*\*/ ||
         $msg =~ / error/  ||
         $msg =~ /Error/) {$tag = 'error';}
     else                 {$tag = 'info';}
  }
  if (Exists($mw) && defined $logBox && Exists($logBox)) {
     $logBox->insert('end', $msg."\n", $tag);
     if    ($tag eq 'error') {$logBox->tagConfigure($tag, -foreground=>$redcolor);}
     elsif ($tag eq 'warn')  {$logBox->tagConfigure($tag, -foreground=>$goldcolor);}
     $logBox->see('end');
     if (Exists($mw)) {$mw->update;}
  } else {
     print STDOUT $msg."\n";
  }
}

#==========================================================================
sub disable_buttons() {
  $saveBtn->configure(-state=>'disabled');
  $dumpBtn->configure(-state=>'disabled');
}

#==========================================================================
sub enable_buttons() {
  my ($i);
  $saveBtn->configure(-state=>'normal',-relief=>'raised');
  $dumpBtn->configure(-state=>'normal',-relief=>'raised');
}

#==========================================================================
sub goodbye() {
  require Tk::ErrorDialog;
  &updateLogBox('Goodbye!'."\n");
  &clearAllCaMonitors();
# $mw->destroy;
# Tk::exit;
  exit 0;
}
