#!/usr/bin/perl -W
#--------------------------------------------------------------
#
#                Author: Sergey Stepanov
#
#  This script sets requested mono Bragg reflection, is available.
#
#
# Version-1.0:  2021/09/15 using Pezca
#--------------------------------------------------------------
# use Math::Trig;		# needed for mono Theta <-> Energy calculations
  use Pezca;
  use warnings::unused;
  use strict;

  our ($IGNORE_ERRORS, $NONINTERACTIVE);
  our ($USE_ZENITY_FOR_EXIT, $ZENITY_VIA_BLUICE);
  our ($beamline, $shutter, $shutter_readback);
  my ($PEZCA, $script, $syntax, @beamlinelist, @HKLlist, $status);
  my ($hkl, $hkl_ini, $narg, $arg, $i, $tell_bluice);
  my ($bln_specified, $hkl_specified, $blu_specified);
  my ($monoAssyPV, $monoAxisPV, $energy);
  my ($linkID, $gapCheckPV, $gapCheck, $gapcheck_disabled);
  my ($mostab_initial, $mostab_disabled);
  my ($postab_initial, $postab_disabled);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  $IGNORE_ERRORS          = 1;							#GLOBAL: do not break script if error occurred
  $NONINTERACTIVE         = 1;							#GLOBAL: do not wait for any answers
  $USE_ZENITY_FOR_EXIT    = 1;							#GLOBAL: popup exit messages
 
  do $PEZCA.'/Common/motor_subs.pl';
  do $PEZCA.'/Common/mono_subs.pl';
  do $PEZCA.'/Common/mostab_subs.pl';
  do $PEZCA.'/Common/shutter_subs.pl';
  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  select STDERR; $|=1;								#set unbuffered output
  select STDOUT; $|=1;								#set unbuffered output

  $script = &makeScriptName();
  $syntax = "\n\n".'Syntax: '.$script.' hkl [beamline] [bluice|nobluice]'."\n";
  @beamlinelist = ('23i:', '23o:', '23b:');
  @HKLlist      = ('111', '333');
# ##############################

  $narg = @ARGV;
  print STDOUT "\n".'Date: '.scalar(localtime)."\n";
  if ($narg > 0) {print STDOUT $script.': started with cmdline = '."@ARGV\n\n";}
  else           {print STDOUT $script.': started with no cmdline arguments.'."\n\n";}
  if ($narg == 0) {
     &press_key_exit('*** '.$script.': no HKL in the cmdline'.$syntax,1);
  }
#---------------------------------------------------------------------------
### Scan command line arguments:
  for ($i=0; $i<$narg; $i++) {
     $arg = $ARGV[$i];
     if (grep(/^$arg:?$/, @beamlinelist)) {
        if (!defined $bln_specified) {
           $bln_specified = 1;
           $arg =~ s/:$//;
           if ($arg ne $beamline) {
              print STDOUT "\n".'!!! '.$script.': Cross-beamline operation:'
                             .' controlling '.$arg.' from '.$beamline."\n\n";
              $beamline = $arg;
              do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
           }
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif (grep(/^$arg$/, @HKLlist)) {
        if (!defined $hkl_specified) {
           $hkl_specified = 1;
           $hkl = $arg;
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^bluice$/i || $arg =~ /^nobluice$/i) {
        if (!defined $blu_specified) {
           $blu_specified = 1;
           if    ($arg =~ /^bluice$/i)   {$tell_bluice = 1;}
           elsif ($arg =~ /^nobluice$/i) {$tell_bluice = 0;}
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     else {
        &press_key_exit('*** '.$script.': unrecognized command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
     }
  }
#---------------------------------------------------------------------------
  if (!defined $tell_bluice) {$tell_bluice = 0;}
  if ($tell_bluice) {$ZENITY_VIA_BLUICE = 1;}
  if (!defined $hkl) {
     &zenity_error_exit('*** '.$script.': no HKL in the cmdline'.$syntax,,1);
  }
  if ($beamline eq '23i' || $beamline eq '23o') {$linkID=1;} else {$linkID=0;}
  if (defined $shutter) {
     if ($shutter !~ /^${beamline}:/) {$shutter   = $beamline.':'.$shutter;}
  }
  if (defined $shutter_readback) {
     if ($shutter_readback !~ /^${beamline}:/) {
                                $shutter_readback = $beamline.':'.$shutter_readback;}
  }
  $monoAssyPV = $beamline.':MO:En';
  $monoAxisPV = $beamline.':MO:E';
  $gapCheckPV = $beamline.':bi:enableEnergyGapCheck';

  $mostab_disabled   = 0;
  $postab_disabled   = 0;
  $gapcheck_disabled = 0;
###------------------------------------------------------------------------
  $SIG{HUP}  = 'signal_handler';						#(1) hangup (see signum.h)
  $SIG{INT}  = 'signal_handler';						#(2) interrupt (see signum.h)
  $SIG{QUIT} = 'signal_handler';						#(3) quit (see signum.h)
# $SIG{KILL} = 'signal_handler';						#(9) this is what is sent by kill -9 -- unblockable
  $SIG{TERM} = 'signal_handler';						#(15) terminate -- sent by kill without paramaters
# $SIG{STOP} = 'signal_handler';						#(19) stop -- unblockable
### ------------------------------------------------------------------------
### Determine where are we now (at start):
  $energy = &getAxisAct ($monoAxisPV);

  $hkl_ini = &getMonoHKL();

  if ($hkl eq $hkl_ini) {
     print STDOUT $script.': mono is already using the requsted hkl='.$hkl.'. Nothing to do, exiting.'."\n";
     &signal_handler('quit');
  }

  print STDOUT $script.': closing fast shutter to protect sample.'."\n";
  &closeShutter ($shutter,$shutter_readback);

### ---------------------------------------------------
### Disable beamline feedback:
### What mode are we running (0=software,1=hybrid,2=hardware(lockin))
  $mostab_initial = &getMostabState ($beamline);				#0=run, 1=pause, -1=error
  if ($mostab_initial == -1) {
     print STDOUT $script.': Cannot talk to MOSTAB.'."\n";
     $mostab_initial = 1;							#presume paused if failure
  }
  $postab_initial = &getPostabState ($beamline);
  if ($postab_initial == -1) {
     print STDOUT $script.': Cannot talk to POSTAB.'."\n";
     $postab_initial = 3;							#presume paused if failure
  }
  if ($mostab_initial != 1) {							#if not paused
     print STDOUT $script.': Pausing I0 feedback'."\n";
     $status = &pauseMostab ($beamline);
     $mostab_disabled = 1;
  }
  if ($postab_initial != 3) {							#if not paused
     print STDOUT $script.': Pausing positional feedback'."\n";
     $status = &pausePostab ($beamline);
     $postab_disabled = 1;
  }
  if ($linkID) {
     ($status,$gapCheck) = Pezca::GetDouble($gapCheckPV);
     if ($status) {
        &report_only($script, $status, $gapCheckPV);
        print STDOUT '!!! '.$script.': Presuming no gap check in JBluIce.'."\n\n";
        $gapCheck = 0;
     }
  } else {$gapCheck = 0;}
  if ($gapCheck) {
     print STDOUT $script.': Disabling JBluIce gap check.'."\n";
     $gapcheck_disabled = 1;
     $status = Pezca::PutDouble($gapCheckPV,0);
     &report_only($script, $status, $gapCheckPV);
  }

### ---------------------------------------------------
# print STDOUT $script.': setting mono Bragg reflection to hkl='.$hkl."\n";
  $status = &setMonoHKL($hkl);
  if ($status) {
     &zenity_error('*** '.$script.': setting mono Bragg reflection to hkl='.$hkl.' failed!',undef,'nowait');
     &signal_handler('quit',2);
  }

### ---------------------------------------------------
### Restore gap check:
  if ($gapCheck) {
     print STDOUT $script.': Re-enabling BluIce gap check.'."\n";
     $gapcheck_disabled = 0;
     $status = Pezca::PutDouble($gapCheckPV,1);
     &report_only($script, $status, $gapCheckPV);
  }
### Re-enable feedback:
  if ($mostab_initial != 1 && $mostab_disabled != 0) {				#if not paused initially (#0=run, 1=pause, -1=error)
     $mostab_disabled = 0;
     print STDOUT $script.': Restoring I0 feedback...'."\n";
     $status = &restoreMostab ($beamline);
     if ($status) {
        print STDOUT "\n".'*** '.$script.': Failed to restart MOSTAB'."\n";
     }
  }
  if ($postab_initial != 3 && $mostab_disabled != 0) {				#if not paused initially (#0=run, 1=pause, -1=error)
     $postab_disabled = 0;
     print STDOUT $script.': Restoring positional feedback...'."\n";
     $status = &restorePostab ($beamline);
     if ($status) {
        print STDOUT "\n".'*** '.$script.': Failed to restart POSTAB'."\n";
     }
  }

  &clearAllCaMonitors();
  exit 0;

#=================================================================

sub signal_handler ($;$) {
  my $sig  = shift(@_);
  my $code = shift(@_);
  if ($sig !~ /^quit$/) {print STDOUT "\n".'*** '.$script.': Caught a SIG'.$sig."\n";}
  if (!defined $code) {
     if ($sig =~ /^quit$/) {$code=0;} else {$code = 1;}
  }
  if (defined $monoAssyPV) {
     print STDOUT $script.': Aborting Energy motion.'."\n" ;
     &abortAssy ($monoAssyPV, 0);
  }
  if ($gapcheck_disabled) {
     print STDOUT $script.': Re-enabling JBluIce gap check.'."\n";
     $status = Pezca::PutDoubleOldCa($gapCheckPV,1);
  }
  if ($mostab_disabled) {
     print STDOUT $script.': Re-enabling intensity feedback.'."\n";
     &restoreMostab ($beamline);
  }
  if ($postab_disabled) {
     print STDOUT $script.': Re-enabling positional feedback.'."\n";
     &restorePostab ($beamline);
  }
  print STDOUT $script.': Closing monitors and exiting.'."\n" ;
  &clearAllCaMonitors();
  exit $code;									#this will return "256*n" to calling perl script (when n is small)
}
