#!/usr/bin/perl -W
#--------------------------------------------------------------
#
# This script fills EPICS DESC fields for transfocator lense 
# groups from respective SQL labels. Only non-existing labels
# are filled unless "force" or "overwrite" is specified in 
# the command line.
#
#                       Author: Sergey Stepanov
#
# Version-1.0:  2020/09/22	initial implementation
#--------------------------------------------------------------
  use Pezca;
  use DBI;
  use strict;
  use warnings;
  use warnings::unused;

  select STDERR; $|=1;			# set unbuffered output
  select STDOUT; $|=1;			# set unbuffered output

  our (@CRLconfig);					#cached array of the latest CRL config (what is normally read from SQL)
  our $beamline = '23i';
  my ($PEZCA, $script, $device, $label, $status);
  my ($nrecords, $i, $name, $material, $ctrl_ch);
  my ($PV_ctrl, $PV_rb_in, $PV_rb_out, @mismatch);
  my (@beamlinelist, $syntax, $arg, $force);

  if (!defined $ENV{PEZCA}) {
     die 'Please define the PEZCA environment. It must point to the PERL scripts repository';
  }
  $PEZCA = $ENV{PEZCA};
  $PEZCA =~ s/\\/\//g;
  if ($PEZCA !~ /\/$/) {$PEZCA .='/';}

  if (-e $PEZCA.'Common/common_subs.pl') {
     do $PEZCA.'Common/common_subs.pl';
     do $PEZCA.'Common/crl_subs.pl';
     do $PEZCA.'/Common/beamline.pl';
  } else {die '*** Cannot fined PEZCA subs in '.$PEZCA."\n";}
  
  $script = &makeScriptName('noext');
  @beamlinelist  = qw(23i:);

  $syntax = "\n\n".'Syntax: '.$script;
  foreach $arg (@beamlinelist) {$syntax .=' ['.$arg.']';}

  if (@ARGV == 0) {print STDOUT $script.': started without command line arguments.'."\n".$syntax;}
  else            {print STDOUT $script.': started with command line arguments ['."@ARGV".']'."\n";}

  foreach $arg (@ARGV) {
     if (grep(/^$arg:?$/, @beamlinelist)) {
        $arg=~s/:$//g;
        if ($beamline ne $arg) {
           $beamline = $arg;
           print STDOUT "\n".'!!! '.$script.': Cross operation for beamline='.$beamline."\n";
           do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
        }
     }
     elsif ($arg =~ /^force$/i || $arg =~ /^overwrite$/i) {$force = 1;}
     else {&zenity_error_exit('*** '.$script.': unrecornized command line argument '.$arg.$syntax);}
  }
  if (!defined $force) {$force = 0;}

  $device = 'CRL-2020';
  @CRLconfig = ();
  @mismatch = ();
# $nrecords = &readCRLconfigSQLpopulated($device,\@CRLconfig);
  $nrecords = &readCRLconfigSQL($device,\@CRLconfig);
  print STDOUT 'Number of configured lenses in SQL for device='.$device.' is '.$nrecords."\n";
  printf STDOUT '%-12s %-8s %-8s %-9s %8s'."\n",'Name','PV_ctrl','PV_rb_in','PV_rb_out','Material';
  for ($i=0; $i<$nrecords; $i++) {
     $name       = $CRLconfig[$i]{'Name'};		#Name of lenses subset driven by $PV_ctrl
     $name =~ s/^\d{2} //;
     $PV_ctrl    = $CRLconfig[$i]{'PV_ctrl'};		#Digital output for lenses In/Out
     $PV_rb_in   = $CRLconfig[$i]{'PV_rb_in'};		#Digital input for lenses in the beam
     $PV_rb_out  = $CRLconfig[$i]{'PV_rb_out'};		#Digital input for lenses out of the beam
     $material   = $CRLconfig[$i]{'Material'};		#Lenses material, normally Be. If $material is blank, no lenses are installed in the subset
     $ctrl_ch    = $PV_ctrl;
     $ctrl_ch   =~ s/^${beamline}:pmac20:acc65e1:bo//;
     if (! &is_int($ctrl_ch)) {
        printf STDOUT 'Unexpected channel=[%s] for control PV=[%s]'."\n",$ctrl_ch,$PV_ctrl;
        $ctrl_ch = '';
     }
     if (length($material) == 0 && $name !~ /YAG/i) {$name = 'no lenses';}

     ($status,$label) = Pezca::GetString($PV_ctrl.'.DESC');
     if ($status) {&report_only($script, $status, $PV_ctrl); $label='';}
     if (length($label) == 0 || $force) {				#do not overwrite existing labels
        $label = $name.' Ctrl';
        if (length($label) > 39) {$label = substr($label,0,39);}
        $status = Pezca::PutStringOldCa($PV_ctrl.'.DESC', $label);
        &report_only($script, $status, $PV_ctrl);
     } 
     elsif (length($label) != 0 && ! $force) {
        if ($label ne $name.' Ctrl') {push (@mismatch, $label.'  !=  '.$name.' Ctrl');}
     }

     ($status,$label) = Pezca::GetString($PV_rb_in.'.DESC');
     if ($status) {&report_only($script, $status, $PV_rb_in); $label='';}
     if (length($label) == 0 || $force) {				#do not overwrite existing labels
        $label = 'IN-'.$ctrl_ch.' '.$name;
        if (length($label) > 39) {$label = substr($label,0,39);}
        $status = Pezca::PutStringOldCa($PV_rb_in.'.DESC', $label);
        &report_only($script, $status, $PV_rb_in);
     }
     elsif (length($label) != 0 && ! $force) {
        if ($label ne 'IN-'.$ctrl_ch.' '.$name) {push (@mismatch, $label.'  !=  IN-'.$ctrl_ch.' '.$name);}
     }

     ($status,$label) = Pezca::GetString($PV_rb_out.'.DESC');
     if ($status) {&report_only($script, $status, $PV_rb_out); $label='';}
     if (length($label) == 0 || $force) {				#do not overwrite existing labels
        $label = 'OUT-'.$ctrl_ch.' '.$name;
        if (length($label) > 39) {$label = substr($label,0,39);}
        $status = Pezca::PutStringOldCa($PV_rb_out.'.DESC', $label);
        &report_only($script, $status, $PV_rb_out);
     }
     elsif (length($label) != 0 && ! $force) {
        if ($label ne 'OUT-'.$ctrl_ch.' '.$name) {push (@mismatch, $label.'  !=  OUT-'.$ctrl_ch.' '.$name);}
     }

     $PV_ctrl    =~ s/^${beamline}:pmac20:acc65//;
     $PV_rb_in   =~ s/^${beamline}:pmac20:acc65//;
     $PV_rb_out  =~ s/^${beamline}:pmac20:acc65//;
     printf STDOUT '%-12s %-8s %-8s %-9s %8s'."\n",$name,$PV_ctrl,$PV_rb_in,$PV_rb_out,$material;
  }
  if (@mismatch > 0) {
     print STDOUT 'Total of '.@mismatch.' differemces between SQL and EPICS labels found.'."\n"
                 .'Use the "force" command to fix:'."\n\n";
     foreach $label (@mismatch) {print STDOUT $label,"\n";}
  }

  &clearAllCaMonitors();
  exit;
