#!/usr/bin/perl -W
#--------------------------------------------------------------
# This script sets Photon Energy and Threshold for Eiger detector.
# It is meant to be started from the Eiger IOC startup script on bl2epu.
#
#                     Author: Sergey Stepanov
#
# Version-1.0:  2005/11/23 using Pezca
# Version-1.1:  2021/07/13 added bl1ws11 (CdTe detector FPE)
#--------------------------------------------------------------
  use strict;
  use Pezca;

  our ($NONINTERACTIVE, $IGNORE_ERRORS, $USE_ZENITY_FOR_EXIT);
  our ($beamline);
  my ($PEZCA, $script, $host, $energy, $threshold);
  my ($monoEn_PV, $eigerEn_PV, $eigerTh_PV, $status, $fpe);
  my ($eigerEnRB_PV, $eigerThRB_PV, $rb);
  my ($eigerStatus_PV, $eigerStatus, $setThreshold);
  my ($wait_step, $wait_max, $wait_);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  select STDERR; $|=1;				# set unbuffered output
  select STDOUT; $|=1;				# set unbuffered output

  $setThreshold        = 0;			# set to "1" for setting energy threshold too
  $NONINTERACTIVE      = 1;
  $IGNORE_ERRORS       = 0;
  $USE_ZENITY_FOR_EXIT = 1;
  $wait_step = 0.2;				# 200 ms
  $wait_max  = 60;				# 20 s
#----------------------------------------------------------------
  $script = &makeScriptName();

  $host= &gethost('short');
  if ($host !~ /^bl2epu/ && $host !~ /^bl1ws11/) {die $script.': can only run on bl2epu or bl1ws11';}

  if (! defined $ENV{EPICS_CA_ADDR_LIST} || ($host =~ /^bl2epu/ && $ENV{EPICS_CA_ADDR_LIST} !~ /\Q.154\E/)) {
     $ENV{EPICS_CA_ADDR_LIST}='10.20.103.154 164.54.103.180';		#bl2epu & bl2ioc1
  }
  if (! defined $ENV{EPICS_CA_ADDR_LIST} || ($host =~ /^bl1ws11/ && $ENV{EPICS_CA_ADDR_LIST} !~ /\Q.121\E/)) {
     $ENV{EPICS_CA_ADDR_LIST}='10.20.103.121 164.54.103.97';		#bl1ws11 & bl1ioc1
  }

  $monoEn_PV      = $beamline.':MO:E:RqsPos';
# $fpe =$beamline.':eiger:';
  $fpe =$beamline.':fpe:';
  $eigerEn_PV     = $fpe.'PhotonEnergy';
  $eigerEnRB_PV   = $fpe.'PhotonEnergy_RBV';
  $eigerTh_PV     = $fpe.'ThresholdEnergy';
  $eigerThRB_PV   = $fpe.'ThresholdEnergy_RBV';
  $eigerStatus_PV = $fpe.'StatusMessage_RBV';
  
  print STDOUT "\n\n".'-----------------------'."\n";

  ($status,$energy) = Pezca::GetDouble($monoEn_PV);
  &report_exit($script, $status, $monoEn_PV);
  $energy    = 1000*$energy;
  $threshold = $energy / 2.;
  
  $status = &addCaMonitor($eigerStatus_PV);
  &report_exit($script, $status, $eigerStatus_PV);
  $status = &addCaMonitor($eigerEnRB_PV);
  &report_exit($script, $status, $eigerEnRB_PV);
  if ($setThreshold) {  
     $status = &addCaMonitor($eigerThRB_PV);
     &report_exit($script, $status, $eigerThRB_PV);
  }
  
  printf STDOUT $script.': setting up Eiger Photon Energy to %.3f eV'."\n",$energy;
  $status = Pezca::PutDoubleOldCa($eigerEn_PV,$energy);
  &report_exit($script, $status, $eigerEn_PV);
  select(undef,undef,undef,0.2);         	#short sleep

  print STDOUT $script.': waiting for Eiger Photon Energy to set.'."\n";
  $eigerStatus = '';
  $wait_       = 0;
  do {
     select(undef,undef,undef,$wait_step);	# sleep($wait_step);
     $wait_ += $wait_step;
     ($status,$eigerStatus) = &PezcaGetStringWF($eigerStatus_PV);
     ($status,$rb)          = Pezca::GetDouble($eigerEnRB_PV);
  } while ($eigerStatus ne 'Photon Energy set' && $wait_ < $wait_max);
  if ($eigerStatus eq 'Photon Energy set' || abs($rb-$energy) < 1) {
     printf STDOUT $script.': Eiger Photon Energy is set OK at %.3f eV'."\n", $rb;
  } else {
     print STDOUT '*** '.$script.': failed to set Eiger Photon Energy, status=['.$eigerStatus.']'."\n";
  }

  if ($setThreshold) {  
     printf STDOUT $script.': setting up Eiger Energy Threshold to %.3f eV'."\n",$threshold;
     $status = Pezca::PutDoubleOldCa($eigerTh_PV,$threshold);
     &report_exit($script, $status, $eigerTh_PV);
     select(undef,undef,undef,0.2);         	#short sleep

     print STDOUT $script.': waiting for Eiger Energy Threshold to set.'."\n";
     $eigerStatus = '';
     $wait_       = 0;
     do {
        select(undef,undef,undef,$wait_step);	# sleep($wait_step);
        $wait_ += $wait_step;
        ($status,$eigerStatus) = &PezcaGetStringWF($eigerStatus_PV);
        ($status,$rb)          = Pezca::GetDouble($eigerThRB_PV);
     } while ($eigerStatus ne 'Threshold Energy set' && $wait_ < $wait_max);
     if ($eigerStatus eq 'Threshold Energy set' || abs($rb-$threshold) < 1) {
        printf STDOUT $script.': Eiger Energy Threshold is set OK at %.3f eV'."\n", $rb;
     } else {
        print STDOUT '*** '.$script.': failed to set Eiger Energy Threshold, status=['.$eigerStatus.']'."\n";
     }
  }

  &clearAllCaMonitors();
  exit 0;
