#!/usr/bin/perl -W
#--------------------------------------------------------------
#    This is a script checks and syncs attenuation foils data
#    between mySQL (master) and EPICS (slave)
#
#       		Author: Sergey Stepanov
#
# Version-1.0:  2015/03/25 using Pezca
#--------------------------------------------------------------
  use strict;
  use Pezca;
  use DBI;

  our ($beamline, $nAttenuators, $attenuatorPV);
  our ($NONINTERACTIVE, $USE_ZENITY_FOR_EXIT);
  my ($PEZCA, $script, $arg, $i, @beamlinelist, $status, $nerr, $msg, $str);
  my (@Attenuators_material_mysql, @Attenuators_thickness_mysql);
  my (@attenuatorExceptions_mysql, $nAttenuatorExceptions_mysql);
  my (@Attenuators_material_epics, @Attenuators_thickness_epics);
  my (@attenuatorExceptions_epics, $nAttenuatorExceptions_epics);
  my ($pv, $description, $backup, $file, $handle, $syntax);
  my (@timearray, $day, $month, $year);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}

  do $PEZCA.'/Common/attenuation_subs_v2.pl';
  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  select STDERR; $|=1;					#set unbuffered output
  select STDOUT; $|=1;					#set unbuffered output

  $script  = &makeScriptName();
  $syntax = "\n".'Syntax: '.$script.' [beamline] [backup|backup=filename] [noninteractive]'."\n\n";
  @beamlinelist = ('23i:', '23o:', '23b:', '23d:');
  $NONINTERACTIVE = 0;
  $USE_ZENITY_FOR_EXIT = 1;
  $backup = 0;
#----------------------------------------------------------------
  foreach $arg (@ARGV) {
     $arg =~ s/^-{0,2}//;					#trim leading dashes
     if (grep(/^$arg:?$/, @beamlinelist)) {
        if ($arg !~ /^$beamline/) {
           $beamline = $arg;
           $beamline =~ s/:$//g;
           print STDOUT "\n".'!!! '.$script.': Cross-beamline operation for '.$beamline."\n\n";
           do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
	}
     }
     elsif ($arg =~ /^backup/) {
	$backup = 1;
	if ($arg =~ /^backup=/) {$file=$arg; $file=~s/^backup=//;}
     }
     elsif ($arg =~ /^noninteractive/i) {
        $NONINTERACTIVE = 1;
     }
     else {
        &press_key_exit('*** '.$script.': incorrect cmdline.'.$syntax);
     }
  }
#----------------------------------------------------------------
  ($status,$msg) = &makeAttenuatorDB_mySQL(\@Attenuators_material_mysql,
                                           \@Attenuators_thickness_mysql,
                                           \@attenuatorExceptions_mysql);
  if ($status) {&zenity_error_exit('Error='.$status.' from makeAttenuatorDB_mySQL'."\n".$msg);}

  ($status,$msg) = &makeAttenuatorDB_DESC(\@Attenuators_material_epics,
                                          \@Attenuators_thickness_epics,
                                          \@attenuatorExceptions_epics);
  if ($status) {&zenity_error_exit('Error='.$status.' from makeAttenuatorDB_DESC'."\n".$msg);}

  $nAttenuatorExceptions_mysql = @attenuatorExceptions_mysql;
  $nAttenuatorExceptions_epics = @attenuatorExceptions_epics;
  $nerr = 0;
  $msg = '';
  for ($i=0; $i<$nAttenuators; $i++) {
     if ($Attenuators_material_mysql[$i] ne $Attenuators_material_epics[$i]) {
	$nerr++;
	$str = sprintf '*** Material  mismatch for foil-%02d: SQL=%s,  EPICS=%s'."\n",$i+1,
                       $Attenuators_material_mysql[$i],$Attenuators_material_epics[$i];
	$msg .= $str;   print STDOUT $str;
     }
     if ($Attenuators_thickness_mysql[$i] != $Attenuators_thickness_epics[$i]) {
	$nerr++;
	$str = sprintf '*** Thickness mismatch for foil-%02d: SQL=%.2f,  EPICS=%.2f'."\n",$i+1,
                       $Attenuators_thickness_mysql[$i],$Attenuators_thickness_epics[$i];
        $msg .= $str;   print STDOUT $str;
     }
  }
  if ($nAttenuatorExceptions_mysql != $nAttenuatorExceptions_epics) {
     $nerr++;
     $str = sprintf '*** Mismatching number of exceptions: SQL=%d,  EPICS=%d'."\n"
                  . "      SQL = (@{attenuatorExceptions_mysql})\n"
                  . "    EPICS = (@{attenuatorExceptions_epics})\n",
                    $nAttenuatorExceptions_mysql, $nAttenuatorExceptions_epics;
     $msg .= $str;   print STDOUT $str;
  } else {
     for ($i=0; $i<$nAttenuatorExceptions_mysql; $i++) {
        if ($attenuatorExceptions_mysql[$i] != $attenuatorExceptions_epics[$i]) {
	   $nerr++;
	   $str = sprintf '*** Exception mismatch-%02d: SQL=%d,  EPICS=%02d'."\n",$i+1,
                       $attenuatorExceptions_mysql[$i],$attenuatorExceptions_epics[$i];
           $msg .= $str;   print STDOUT $str;
        }
     }
  }
  if ($nerr > 0) {
     if ( &zenity_question($msg."\n".'Fix?') != 0) {
        if ($backup) {
	   if (!defined $file) {
              @timearray = localtime(time);                                      # This is (sec,min,hour,mday,mon,year,wday,yday,isdst)
              $day   = $timearray[3];
              $month = $timearray[4] + 1;
              $year  = $timearray[5] + 1900;
              $file  = sprintf $beamline.'_foils_%04d_%02d_%02d.sav',$year,$month,$day;
	   }
	   if (-f $file) {&zenity_error_exit('*** '.$script.': file='.$file.' exists!');}
	   $handle = open (LOG,'>  '.$file);
	   if (!defined $handle) {&zenity_error_exit('*** '.$script.': failed to open file='.$file);}
           select LOG; $|=1;
           for ($i=0; $i<$nAttenuators; $i++) {
              $pv = $attenuatorPV.$i.'.DESC';
	      ($status,$description) = Pezca::GetString($pv);
              &report_exit($script,$status,$pv);
	      print LOG 'caput -s '.$pv.' "'.$description.'"'."\n";
	   }
	   close LOG;
        }
        for ($i=0; $i<$nAttenuators; $i++) {
           $pv = $attenuatorPV.$i.'.DESC';
	   $description = $Attenuators_thickness_mysql[$i].' '.$Attenuators_material_mysql[$i];
           $status = Pezca::PutStringOldCa($pv,$description);
           &report_exit($script,$status,$pv);
        }
        for ($i=0; $i<$nAttenuatorExceptions_mysql; $i++) {
           $pv = $attenuatorPV.$attenuatorExceptions_mysql[$i].'.DESC';
	   $description = 'excluded';
           $status = Pezca::PutStringOldCa($pv,$description);
           &report_exit($script,$status,$pv);
        }
     }
  } else {
     print STDOUT "\n\n".'No Attenuation DB mismatches'."\n\n";
  }
  &clearAllCaMonitors();
  exit;
