#!/usr/bin/perl -W
#--------------------------------------------------------------
#       This script is a beam transfocator server
#
#     		Author: Sergey Stepanov
#
# Version-1.0:  2020/09/24 using Pezca and the simplex recursive algorithm 
#                          similar to the one used for matching attenuation.
# Version-1.1:  2020/10/13 added Transfocator shift by positioner.
# Version-1.2:  2020/11/20 added calculation of Transfocator weighted shift from center.
# Version-2.0:  2021/02/12 changed the algoritm for matching focus from simplex
#                          to SQL-cached sum if 1/f and then sorting by best match; 
#                          implemented minimization of beam attenuation in TF 
#                          using focus matching by least lens blocks in the beam;
#                          implemented optional account for the change in the TF
#                          weighted center when lenses go In/Out.
# Version-3.0:  2021/05/19 changed the algoritm for matching focus to calculating
#                          exact focus by Brul's recursive method and then
#                          sorting by best match;
#                          changed minimization of beam attenuation in TF from
#                          focus matching by least lens blocks to least number
#                          of lenses in the beam.
# Version-4.0:  2021/06/07 modified the Bruls algorithm to calculate the exact focus
#                          as a distance from the rear principal plane instead of
#                          gravity. This way "gravity" is always included.
# Version-4.1:  2021/06/11 made transfocator upstream shift always included.
# Version-4.2:  2021/08/11 added $IGNORE_ERRORS.
# Version-5.0:  2021/08/19 excluded shiftIncluded and centerIncluded PVs and added source distance
# Version-5.1:  2021/09/16 added init function to set bo from rb on reboot
# Version-5.2:  2021/09/26 added auto-excluding 50um lenses below 26 keV
# Version-5.3:  2021/12/15 added fast a.k.a. "tetris" mode for searching best match
#--------------------------------------------------------------
  use Pezca;
  use DBI;
  use warnings::unused;
  use strict;

  our ($beamline);
  our ($NONINTERACTIVE, $USE_ZENITY_FOR_EXIT, $IGNORE_ERRORS);
  our (@CRLconfig, @CRLconfig_full);
  my ($PEZCA, $script, $syntax, $narg, $arg, $i, @beamlinelist, $status);
  my ($srchModePV,  $srchMode,  $srchEnabled);
  my ($rangeMaxPV,  $rangeMax);
  my ($focusRqsPV,  $focusRqs);
  my ($focusActPV,  $focusAct);
  my ($focusSrcPV,  $focusSrc);
  my ($focusCalcPV, $focusCalc, $focusCalcRqs);
  my ($focusCapXPV, $focusCapX);
  my ($focusSetPV,  $focusSet);
  my ($focusStsPV,  $focusSts);
  my ($focusRunPV,  $focusRun);
  my ($focusMsgPV,  $focusMsg);
  my ($energyPV,    $energy, $connection_lost);
  my ($device, @InOut, $tmp, $tmp2, $nrecords);
  my ($TF_axis_PV, $TF_assy_PV, $TF_act, $TF_rqs);
  my ($TF_drvl, $TF_drvh, $TF_zero, $ncorrected);
  my ($E_exclude, $r_exclude, $exclude);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  do $PEZCA.'/Common/crl_subs.pl';
  do $PEZCA.'/Common/motor_subs.pl';
  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  select STDOUT; $|=1;			# set unbuffered output
  select STDERR; $|=1;			# set unbuffered output

  $script  = &makeScriptName();
  $syntax = "\n".'Syntax: '.$script.' [beamline]'."\n";
###-----------------------------------------------------------------
### Protection from running on multiple computers:
  &check_Nrunning_and_host_allowed($script,['bl1ws2']);
#----------------------------------------------------------------
  $IGNORE_ERRORS  = 1;
  $NONINTERACTIVE = 1;
  $USE_ZENITY_FOR_EXIT = 1;
  $srchEnabled = 1;			#only allow if motor soft limits are set
  $srchMode    = 0;			#0=best match, 1=min.attenuation within $rangeMax
  $rangeMax    = 25;			# +-25mm
  $E_exclude   = 26;                    #exclude small-radius lenses below this energy in keV
  $r_exclude   = 50;                    #any lenses with this or smaller radius will be excluded an smaller energies
  @beamlinelist = qw(23i);
  $narg = @ARGV;
  if ($narg > 1) {&press_key_exit($syntax);}

###------------------------------------------------------------------------
### Parse command line arguments:

  for ($i=0; $i<$narg; $i++ ) {
     $arg = lc $ARGV[$i];
     if (grep(/^$arg$/, @beamlinelist)) {
        if ($arg ne $beamline) {
           $beamline = $arg;
           print STDOUT "\n".'!!! Cross-operation for beamline '.$beamline."\n\n";
           do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
        }
     }
     else {&press_key_exit("\n".'*** '.$script.': Incorrect command line.'.$syntax);}
  }

  $SIG{INT}  = 'signal_handler';
  $SIG{QUIT} = 'signal_handler';

  $device       = 'CRL-2020';				#transfocator data in SQL
  $energyPV     = $beamline.':MO:E';
  $focusActPV   = $beamline.':focusAct';
  $focusRqsPV   = $beamline.':focusRqs';
  $focusSrcPV   = $beamline.':focusSrcDist';
  $focusCalcPV  = $beamline.':focusCalc';
  $focusCapXPV  = $beamline.':focusCalcPlusX';
  $focusSetPV   = $beamline.':focusSet';
  $focusStsPV   = $beamline.':focusSts';
  $focusRunPV   = $beamline.':focusRun';
  $srchModePV   = $beamline.':focusSearchMode';
  $rangeMaxPV   = $beamline.':focusMaxRange';
  $focusMsgPV   = $beamline.':focusMsg';
  $TF_assy_PV   = $beamline.':CRL:Cl';
  $TF_axis_PV   = $beamline.':CRL:X';

  if (&addCaMonitor($focusRqsPV))    {&press_key_exit('*** '.$script.': failed to set CA monitor on '.$focusRqsPV);}
  if (&addCaMonitor($focusSrcPV))    {&press_key_exit('*** '.$script.': failed to set CA monitor on '.$focusSrcPV);}
  if (&addCaMonitor($focusSetPV))    {&press_key_exit('*** '.$script.': failed to set CA monitor on '.$focusSetPV);}
  if (&addCaMonitor($focusStsPV))    {&press_key_exit('*** '.$script.': failed to set CA monitor on '.$focusStsPV);}
  if (&addCaMonitor($focusRunPV))    {&press_key_exit('*** '.$script.': failed to set CA monitor on '.$focusRunPV);}
  if (&addCaMonitor($srchModePV))    {&press_key_exit('*** '.$script.': failed to set CA monitor on '.$srchModePV);}
  if (&addCaMonitor($rangeMaxPV))    {&press_key_exit('*** '.$script.': failed to set CA monitor on '.$rangeMaxPV);}
  if (&setAxisMonitors($energyPV))   {&press_key_exit('*** '.$script.': failed to set axis CA monitor on '.$energyPV);}
  if (&setAssyMonitors($TF_assy_PV)) {&press_key_exit('*** '.$script.': failed to set assy CA monitor on '.$TF_assy_PV);}
  if (&setAxisMonitors($TF_axis_PV)) {&press_key_exit('*** '.$script.': failed to set axis CA monitor on '.$TF_axis_PV);}

###------------------------------------------------------------------------

### Initialization:
  $focusMsg = '';
# &outputMsg($focusMsg);
  &outputMsg('Reading CRLconfig populated for device='.$device);
  #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  $nrecords = &readCRLconfigSQLpopulated($device,\@CRLconfig);
  #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  &outputMsg('found '.$nrecords.' populated records in CRLconfig for device='.$device, 0.8);

  &outputMsg('Copying digital out RB fields to BO fields for device='.$device.'.');
  ($status,$ncorrected) = &setTransfocatorRB2BO($device);
  if ($status) {&outputMsg('Copying RB to BO failed.', 2);}
  else         {&outputMsg($ncorrected.' digital out BO fields required correction.');}

  @CRLconfig_full = @CRLconfig;				#full set including 50um lenses
  $energy = &getAxisAct($energyPV);
  if ($energy <= $E_exclude) {				#exclude small lenses
     &outputMsg('Eliminating from TF config lenses with radius <= '.$r_exclude);
     $nrecords = &excludeCRLradius($r_exclude,$device,\$focusMsg);
     if ($nrecords > 0) {
        $exclude = 1;
     } else {
        &outputMsg('Eliminating small lenses failed.', 2);
        @CRLconfig = @CRLconfig_full;			#restore full set including 50um lenses
        $exclude = 0;
        &signal_handler();
     }
  } else {
     $exclude = 0;
  }

  ($TF_drvl, $TF_drvh) = &getAxisDRVLH ($TF_axis_PV);
  if ($TF_drvl == 0 && $TF_drvh == 0) {
     &outputMsg('UP/DOWNstream motion soft limits are not set. Minimizing attenuation & tetris are disabled');
     $srchEnabled = 0;					#only allow if motor soft limits are set
     $srchMode    = 0;
  } else {
#    $TF_zero = 0.;					#which one to use?
     $TF_zero = 0.5*($TF_drvl + $TF_drvh);		#which one to use?
     if (!defined $rangeMax || $rangeMax == 0) {$rangeMax = 0.25*abs($TF_drvh - $TF_drvl);}
  }
  $status = Pezca::PutDoubleOldCa($srchModePV,$srchMode);
  if ($status) {
     &report_only($script,$status,$srchModePV); 
     $srchMode    = 0;
     $srchEnabled = 0;					#only allow if motor soft limits are set
     &outputMsg('Cannot write to PV='.$srchModePV.'. Minimizing attenuation & tetris are disabled');
  }
  ($status,$tmp) = Pezca::GetDouble($rangeMaxPV);
  if ($status) {&report_only($script,$status,$rangeMaxPV);  $tmp=$rangeMax;}
  if ($tmp == 0) {
     $status = Pezca::PutDoubleOldCa($rangeMaxPV,$rangeMax);
     &report_only($script,$status,$rangeMaxPV);
  } else {
     $rangeMax = $tmp;
  }
  $energy = &getAxisAct($energyPV);
  $TF_act = &getAxisAct($TF_axis_PV);
# $focusAct = &getTransfocatorFocus4($energy,$device,\$focusMsg,\@InOut); #versions 4 & 5 give same results
  $focusAct = &getTransfocatorFocus5($energy,$device,\$focusMsg,\@InOut); #versions 4 & 5 give same results
  if ($focusAct < 0) {
     &outputMsg('*** '.$script.': failed to get transfocator focus.'."\n".$focusMsg);
     $focusAct = 0;
  } else {
     $focusAct += ($TF_act - $TF_zero);
  }
  ($status,$focusSrc) = Pezca::GetDouble($focusSrcPV);			#$focusSrcPV='focusSrcDist'
  if ($status) {&report_only($script,$status,$focusSrcPV);  $focusSrc=0;}
  if ($focusSrc != 0 && $focusSrc < 100) {
     if ($focusSrc > 0) {$focusSrc = 1000 * $focusSrc;}			#corect m to mm 
     else               {$focusSrc = 0;}
     $status = Pezca::PutDoubleOldCa($focusSrcPV,$focusSrc);		#focusSrcDist
  }
  $focusRqs = $focusAct;
  if    ($focusRqs > 0 && $focusSrc > 0) {$focusCapX = 1./(1./$focusRqs+1./$focusSrc);}		#$focusCapX='focusCalcPlusX'
  elsif ($focusRqs > 0)                  {$focusCapX = $focusRqs;}				#$focusCapX='focusCalcPlusX'
  elsif ($focusSrc > 0)                  {$focusCapX = $focusSrc;}
  else                                   {$focusCapX = 0;}
  $focusCalc = $focusCapX - ($TF_act-$TF_zero);
  $focusSet = 0;							#0=idle, 1=Setting
  $focusSts = 2;							#0=stopped,1=processing,2=ready
  $focusRun = 1;							#0=Stopped, 1=Running
  $focusMsg = '';
  $status = Pezca::PutDoubleOldCa($focusActPV,$focusAct);
  $status = Pezca::PutDoubleOldCa($focusRqsPV,$focusRqs);
  $status = Pezca::PutDoubleOldCa($focusSrcPV,$focusSrc);
  $status = Pezca::PutDoubleOldCa($focusCalcPV,$focusCalc);
  $status = Pezca::PutDoubleOldCa($focusCapXPV,$focusCapX);
  $status = Pezca::PutDoubleOldCa($focusSetPV,$focusSet);
  $status = Pezca::PutDoubleOldCa($focusStsPV,$focusSts);
  $status = Pezca::PutDoubleOldCa($focusRunPV,$focusRun);
  $status = &PezcaPutStringWF($focusMsgPV,$focusMsg,'oldca');

### Monitor loop
# $focusCntr = 0;
  &setEpicsMessages('off');
  $connection_lost = 0;

CMD:
  while(1) { #################################################################
     select(undef,undef,undef,0.2);					#sleep(..);
     $focusMsg = '';
## Check if there is a request to exit:
     ($status,$focusRun) = Pezca::GetDouble($focusRunPV);
     if ($status) {
        if (! $connection_lost) {
           &report_only($script,$status,$focusRunPV);
           print STDOUT scalar(localtime).' '.$script.': PV connection lost!'."\n";
           $connection_lost = 1;
           $focusRun = 1;
        }
        sleep 5;							#wait before trying again
        next CMD;
     }
     if ($connection_lost) {
        print STDOUT scalar(localtime).' '.$script.': PV connection restored!'."\n";
        $connection_lost = 0;
        $focusRun = 1;
        $status = Pezca::PutDoubleOldCa($focusRunPV,$focusRun);
        $status = Pezca::PutDoubleOldCa($focusSrcPV,$focusSrc);
        $status = Pezca::PutDoubleOldCa($focusRqsPV,$focusRqs);
        &resetAllCaMonitors();
     }
     if (! $focusRun) {&signal_handler();}
### Confirm current status:
     $focusSts = 2;							#ready
     $status = Pezca::PutDoubleOldCa($focusStsPV,$focusSts);
### Get current lenses In/Out states:
     #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     $status = &getTransfocatorInOut(\@InOut,$device,\$focusMsg);	#returns current transfocator lenses In/Out state
     #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     if ($status) {
### Output getTransfocatorInOut message, if exist:
        &outputMsg($focusMsg);
        for ($i=0; $i<$nrecords; $i++) {$InOut[$i]=0;}
     }
### Recalculate actual focus in case if energy or lenses In/Out have changed:
     ($status,$focusSrc) = Pezca::GetDouble($focusSrcPV);
     if ($status) {&report_only($script,$status,$focusSrcPV);  $focusSrc=0;}
     if ($focusSrc != 0 && $focusSrc < 100) {
        if ($focusSrc > 0) {$focusSrc = 1000 * $focusSrc;}		#corect m to mm 
        else               {$focusSrc = 0;}
        $status = Pezca::PutDoubleOldCa($focusSrcPV,$focusSrc);
     }
     $TF_act = &getAxisAct($TF_axis_PV);
     $energy = &getAxisAct($energyPV);
     #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     $tmp = &getTransfocatorFocus4($energy,$device,\$focusMsg,\@InOut);
     #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     if ($tmp >= 0) {$focusCalc = $tmp;}
     $focusCapX = $focusCalc + ($TF_act-$TF_zero);			#focusCalcPlusX
     if    ($focusCapX > 0 && $focusSrc > 0) {$focusAct = 1./(1./$focusCapX-1./$focusSrc);}
     elsif ($focusCapX > 0)                  {$focusAct = $focusCapX;}
     elsif ($focusSrc  > 0)                  {$focusAct = -$focusSrc;}
     else                                    {$focusAct = 0;}
     if ($focusAct < 0) {$focusAct = 0;}
### Output recalculated actual focus:
     $status = Pezca::PutDoubleOldCa($focusCalcPV,$focusCalc);
     &report_only($script,$status,$focusCalcPV);
     $status = Pezca::PutDoubleOldCa($focusCapXPV,$focusCapX);
     &report_only($script,$status,$focusCapXPV);
     $status = Pezca::PutDoubleOldCa($focusActPV,$focusAct);
     &report_only($script,$status,$focusActPV);
### Output focus calculation message, if exist
     &outputMsg($focusMsg);
###  - - - - - - - - - - - - - - - - - - - - - - - 
### Check the mode of calculating focus: best match or min.attenuation:
     ($status,$tmp) = Pezca::GetDouble($srchModePV);
     if ($status) {&report_only($script,$status,$srchModePV); $tmp=$srchMode;}
     if ($tmp >= 0 && $tmp <= 2) {$srchMode = $tmp;}
     if ($srchMode && ! $srchEnabled) {
        $srchMode = 0;
        $status = Pezca::PutDoubleOldCa($srchModePV,$srchMode);
        &report_only($script,$status,$srchModePV);
     }
### If there is a request to change focus:    
     ($status,$focusSet) = Pezca::GetDouble($focusSetPV);
     if ($status) {&report_only($script,$status,$focusSetPV); $focusSet=0;}
     if ($focusSet) {
### Read the requested focus value:
        ($status,$tmp) = Pezca::GetDouble($focusRqsPV);
        if ($status || $tmp < 0) {
           &outputMsg('*** Cannot get new focus from '.$focusRqsPV);
        } else {
### We need to change focus, but before doing that, verify if we need
### to include or exclude 50um lenses:
           if ($energy <= $E_exclude && $exclude == 0) {			#need to exclude small lenses
              &outputMsg('Eliminating from TF config small lenses with radius <= '.$r_exclude);
              $nrecords = &excludeCRLradius($r_exclude,$device,\$focusMsg);
              if ($nrecords > 0) {
                 $exclude = 1;
              } else {
                 &outputMsg('Eliminating small lenses failed.', 2);
                 @CRLconfig = @CRLconfig_full;					#restore full set including 50um lenses
                 $exclude = 0;
              }
           }
           elsif ($energy > $E_exclude && $exclude == 1) {			#need to add back small lenses
              &outputMsg('Adding back to TF config small lenses with radius <= '.$r_exclude);
              @CRLconfig = @CRLconfig_full;					#restore full set including 50um lenses
              $exclude = 0;
           }
           $nrecords = @CRLconfig;
### Calculate new actual focus for the requested value:
           $focusRqs = $tmp;
           if    ($focusRqs > 0 && $focusSrc > 0) {$focusCapX = 1./(1./$focusRqs+1./$focusSrc);}
           elsif ($focusRqs > 0)                  {$focusCapX = $focusRqs;}
           elsif ($focusSrc > 0)                  {$focusCapX = $focusSrc;}
           else                                   {$focusCapX = 0;}
#          $focusCalcRqs = $focusCapX - ($TF_act-$TF_zero);
           $focusCalcRqs = $focusCapX;						#2021/09/30
           #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
           if ($srchMode < 2) {
              ($tmp,@InOut) = &setTransfocatorFocus4($focusCalcRqs,$energy,$device,\$focusMsg,	#full search mode
                                                     $srchMode,$rangeMax,$TF_axis_PV);
           } else {
              ($tmp,@InOut) = &setTransfocatorFocus5($focusCalcRqs,$energy,$device,\$focusMsg);	#fast tetris mode
              $tmp2 = $TF_zero + ($focusCalcRqs - $tmp);
              if ($tmp < 0 || $tmp2 < $TF_drvl || $tmp2 > $TF_drvh) {
                 if ($tmp < 0) {$focusMsg = '*** Tetris mode failed to calculate focal distance.';}
                 else          {$focusMsg = sprintf 'Tetris mode requires TF shift to %.2fmm beyond travel limits.',$tmp2;}
                 $focusMsg .= ' Switching to full search.'; 
                 &outputMsg($focusMsg);
                ($tmp,@InOut) = &setTransfocatorFocus4($focusCalcRqs,$energy,$device,\$focusMsg, #full search mode
                                                       0,$rangeMax,$TF_axis_PV);
              }
           }
           #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
           if ($tmp < 0) {						#if calc failed
              $tmp = $focusCalc;
              if (length($focusMsg) == 0) {$focusMsg = '*** Failed to match focus';}
              &outputMsg($focusMsg);
           } else {							#if calc succeeded
              $focusCalc = $tmp;
              $focusMsg = sprintf 'focus(CRL-Rqs)=%.1fmm, focus(CRL-calc)=%.1fmm, diff=%.1fmm',$focusCalcRqs,$focusCalc,($focusCalcRqs-$focusCalc);
              &outputMsg($focusMsg);
           }
           $TF_rqs = $TF_zero + ($focusCalcRqs - $focusCalc);		#new requested position
           if ($TF_drvl != 0 || $TF_drvh != 0) {			#if limits are set
              if ($TF_rqs < $TF_drvl) {
                 $focusMsg = sprintf 'Needed TF shift to %.2fmm is beyond lower travel limit=%.2fmm. Restricting..',$TF_rqs,$TF_drvl;
                 &outputMsg($focusMsg);
                 $TF_rqs = $TF_drvl;
              }
              if ($TF_rqs > $TF_drvh) {
                 $focusMsg = sprintf 'Needed TF shift to %.2fmm is beyond upper travel limit=%.2fmm. Restricting..',$TF_rqs,$TF_drvh;
                 &outputMsg($focusMsg);
                 $TF_rqs = $TF_drvl;
              }
           }
           if (abs($TF_rqs-$TF_act)> 0.01) {
              $focusMsg = sprintf 'Shifting transfocator from %.2fmm to %.2fmm by %.2fmm',$TF_act,$TF_rqs,($TF_rqs-$TF_act);
              &outputMsg($focusMsg);
              $status = &runNMotors($TF_assy_PV, 1, [$TF_axis_PV], [$TF_rqs], 'checkAuto');
              if ($status) {&outputMsg('*** Shifting Transfocator failed with status='.$status);}
              else         {&outputMsg('Shifting Transfocator finished OK.');}
              $TF_act = &getAxisAct($TF_axis_PV);
           }
           $focusCapX = $focusCalc + ($TF_act-$TF_zero);
           if    ($focusCapX > 0 && $focusSrc > 0) {$focusAct = 1./(1./$focusCapX-1./$focusSrc);}
           elsif ($focusCapX > 0)                  {$focusAct = $focusCapX;}
           elsif ($focusSrc  > 0)                  {$focusAct = -$focusSrc;}
           else                                    {$focusAct = 0;}
           if ($focusAct < 0) {$focusAct = 0;}
           ### Output actual focus:
           $status = Pezca::PutDoubleOldCa($focusCalcPV,$focusCalc);	#calculated focus (lenses only)
           &report_only($script,$status,$focusCalcPV);
           $status = Pezca::PutDoubleOldCa($focusCapXPV,$focusCapX);	#calculated focus + X-shift
           &report_only($script,$status,$focusCapXPV);
           Pezca::PutDoubleOldCa($focusActPV,$focusAct);		#actual focus (same as $focusCapX if source at infinity, i.e. $focusSrc=0)
           &report_only($script,$status,$focusActPV);
        }
        $focusSet = 0;
        $status = Pezca::PutDoubleOldCa($focusSetPV,$focusSet);
        select(undef,undef,undef,.5);					#sleep(..);
        $status = &PezcaPutStringWF($focusMsgPV,'','oldca');
     }
  } ##########################################################################

#=================================================================

sub outputMsg($;$) {
  my ($msg, $status, $slp);
  $msg = shift(@_);
  $slp = shift(@_);  if (!defined $slp || $slp < 0.2) {$slp = 0.2;}
  if (length($msg) > 0) {
     print STDOUT scalar(localtime).' '.$msg."\n";
     $status = &PezcaPutStringWF($focusMsgPV,$msg,'oldca'); 
     &report_only($script,$status,$focusMsgPV);
     select(undef,undef,undef,$slp);					#sleep(..);
  }
}

#=================================================================

sub signal_handler (;$) {
  my $signame = shift;
  if (defined $signame) {print STDOUT "\n".scalar(localtime).' *** '.$script.': Caught a SIG'.$signame."\n";}
  $status = Pezca::PutDoubleOldCa($focusSetPV,0);			#not set
  $status = Pezca::PutDoubleOldCa($focusStsPV,0);			#0=stopped
  $status = Pezca::PutDoubleOldCa($focusRunPV,0);			#0=Stopped
  $focusMsg = 'Exiting...';
  $status = &PezcaPutStringWF($focusMsgPV,$focusMsg,'oldca'); print STDOUT $focusMsg."\n";
  &clearAllCaMonitors();
  exit;
}
