package gov.anl.gmca.analysis;

import gov.anl.gmca.analysis.PipelineOptions.PipelineBuilder;
import gov.anl.gmca.analysis.helpers.FileListToTemplateStr;
import gov.anl.gmca.analysis.helpers.FindImagePrefixes;
import gov.anl.gmca.analysis.helpers.MiscUtils;
import gov.anl.gmca.analysis.helpers.RUNNERS;
import gov.anl.gmca.collect.DevFPE;
import gov.anl.gmca.collect.ModelCollectRun;
import gov.anl.gmca.collect.ModelCollectRun.XDSProcMode;
import gov.anl.gmca.collect.ModelCollectRunList;
import gov.anl.gmca.collect.RunCalculator.Frame;
import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.gui.Main;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PV_ROBOT;
import gov.anl.gmca.io.PVLookup.PV_SCR;
import gov.anl.gmca.screening.CmdScreening;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.Util;

import java.io.*;
import java.util.*;


// start data processing with data collection
public class autoProcessData implements Runnable {
	
	private final static String[] pipelinesRunAfterCollect = Util.isUserNonProprietary() ? new String[] {"fastdp", "autoproc", "xia2_dials"}
			: new String[] { "fastdp", "xia2_xds", "xia2_dials"};

	private final static String[] pipelinesRunAfterScreen = { "mosflm_strategy", "labelit_strategy", "dials_strategy",
			"xds_strategy" };
	private final static String[] pipelinesRunOnCollect = { "gmcaproc" };
	private int[] stages = { 25, 50, 100 };
	final Double percentIgnore = 0.05;
	final Double percentFinal = 0.30; 
	private int nSweeps = 1;

	public enum PROCESSING_MODE {
		AFTER_COLLECTION, WITH_COLLECTION, CUSTOM_COLLECTION, AFTER_SCREENING, AFTER_RASTERING;
	}

	// modify this line if change runner
	public static String RUNNER = RUNNERS.SLURM.valueOf();

	private boolean canOverwrite=false;
	
	ModelCollectRun run;
	private String beamline;
	private String sampleName;
	private String sampleNumber = "0";
	private String imageDirectory;
	private String filenamePrefix;
	private String outputRootDirectory;
	private boolean anomalous;
	private int currentFrame = 1;
	private ArrayList<String> imgfilelist;
	private double exposureTime = 0.01;
	private LinkedHashMap<Integer, String> filelists;
	private PROCESSING_MODE mode;
	private double eta;
	private String spreadSheetFile;
	private String mountedCrystal;
	private Integer nEnergies = 1;

	@Override
	public void run() {
		if (this.mode == PROCESSING_MODE.AFTER_COLLECTION) {
			processAfterCollection();
		} else if (this.mode == PROCESSING_MODE.WITH_COLLECTION) {
			processWithCollection();
		} else if (this.mode == PROCESSING_MODE.AFTER_SCREENING) {
			processAfterScreen();
		} else if (this.mode == PROCESSING_MODE.CUSTOM_COLLECTION) {
			processCustom();
		} else {
			Log.warn("processing mode not set.");
			return;
		}
	}

	// constructor 1
	public autoProcessData(String imageDirectory, String beamline, String sampleName, String sampleNumber,
			String filenamePrefix, Integer runLabel, String[] imgfilelist, PROCESSING_MODE mode) {
		this.beamline = beamline;
		this.sampleName = sampleName;
		this.sampleNumber = sampleNumber;
		this.imageDirectory = imageDirectory;
		this.mode = mode;
		this.filenamePrefix = filenamePrefix;
		this.anomalous = true;

		if (imgfilelist != null) {
			this.imgfilelist = new ArrayList<String>(Arrays.asList(imgfilelist));
		}
		setOutputRootDir(false);
	}

	public static ArrayList<String> getImageFilenames(int runNumber) {
		ModelCollectRun run = (ModelCollectRun) ModelCollectRunList.get().getRun(runNumber);
		ArrayList<String> imgFileList = new ArrayList<String>();
		for (Frame f : run.seq().getFrameListCopy()) {
			if (f.cbfFilename != null) {
				imgFileList.add(f.cbfFilename);
			} else {
				imgFileList.add(f.hdfFilename); // master file
			}
		}
		return imgFileList;
	}

	// constructor 2
	public autoProcessData(ModelCollectRun run, PROCESSING_MODE mode) {
		this.run = run;
		this.mode = mode;
		this.beamline = Util.getBID(); // return 23o--idb, 23i--idd, 23b
		this.imageDirectory = run.getFullDirectory();

		this.eta = DevFPE.get().getETA();
		this.exposureTime = run.getExposureTime();
		this.currentFrame = run.getNextFrame();
		FileListToTemplateStr flt = new FileListToTemplateStr(run.getRunIndex());
		nSweeps = flt.getNSweeps();
		this.nEnergies = run.getEnergies().size();

		if (Main.FAKE_DATA) {
			setupFakeData();
			return;
		}

		try {
			String sfile = Epics.getString(PV_SCR.spreadsheetFile);
			String sdir = Epics.getString(PV_SCR.spreadsheetDir256);
			String ffile = String.format("%s/%s", sdir, sfile); 

			mountedCrystal = Epics.getString(PV_ROBOT.mounted);
			
			if (!sdir.equals("") 
					&& !sfile.equals("") 
					&& new File(ffile).exists() 
					&& !mountedCrystal.equals("none")) {
				this.spreadSheetFile = ffile;
			} else {
				this.mountedCrystal = null;
				this.spreadSheetFile = null;
				Log.warn("auto processing: spreadsheet or mounted crystal wasn't defined.");
			}
		} catch (Exception e) {
			e.printStackTrace();
			this.spreadSheetFile = null;
			this.mountedCrystal = null;
		}
		
		//run.getDirectorySuffix();
		this.filenamePrefix = String.format("%s_%d", run.getFileRoot(), run.getRunLabel());
		// requested by Mike, reassign it to same as prefix
		this.sampleName = this.filenamePrefix; // run.getFileRoot();

		try {
			this.sampleNumber = String.valueOf(CmdScreening.labelToindex(Epics.getString(PV_ROBOT.mounted)));
		} catch (Exception e) {
			this.sampleNumber = "0";
			Log.note("not able to get sampleNumber from Epics.");
		}

		//run.getRunLabel();
		if (run.getXDSProcMode() == XDSProcMode.NATIVE) {
			this.anomalous = false;
		} else {
			this.anomalous = true;
		}

		ArrayList<String> imgFileList = new ArrayList<String>();
		for (Frame f : run.seq().getFrameListCopy()) {
			if (f.cbfFilename != null) {
				imgFileList.add(f.cbfFilename);
			} else {
				imgFileList.add(f.hdfFilename); // master file
			}
		}
		this.imgfilelist = imgFileList;
		setOutputRootDir(false);
	}

	protected void setupFakeData() {
		String imgDir = DBConfig.get("processing_test_dir");

		File idir = new File(imgDir);
		if (!idir.exists() || !idir.canRead()) {
			Log.error(imgDir + " does not exist.");
			return;
		}
		Log.note("processing data in FAKEDATA mode, reading images from directory: "+imgDir);

		this.imageDirectory = imgDir;
		this.filenamePrefix = null;
		this.spreadSheetFile = null;
		this.mountedCrystal = null;
		this.sampleName = "fakedata";
		this.sampleNumber = "0";

		this.imgfilelist = FindImagePrefixes.listOfImageFiles(this.imageDirectory);

		this.outputRootDirectory = String.format("%s/%s/%s", System.getenv("HOME"), "process", "fakedata");
		String odir = newDir(this.outputRootDirectory, false);
		if (!new File(odir).exists()) {
			new File(odir).mkdirs();
		}
		this.outputRootDirectory = odir;
		Log.note("FAKEDATA mode: processing results will be written to: " + this.outputRootDirectory);
	}

	// constructor 3
	public autoProcessData(int runNumber, PROCESSING_MODE mode) {
		this(((ModelCollectRun) ModelCollectRunList.get().getRun(runNumber)), mode);
	}
	
	//processing after pause or complete, for fastdp during pause or completion of data collection
	public void processCustom() {
		String[] pipelinesToRun = {"fastdp"};

		try {
			Integer numTotalFrames = run.seq().getTotalFrameCount();
			Integer numFramesNotCollected = run.seq().getRemainingFrameCount();
			
			Double progress = 1.*(numTotalFrames - numFramesNotCollected)/numTotalFrames;
			
			if (progress < percentIgnore) {
				Log.note("data processing skipped @ collection pause, less the 5% data collected.");
				return;
			} else if (progress >= percentFinal) {
				pipelinesToRun = pipelinesRunAfterCollect;
			}
		} catch (Exception e) {
			Log.error("user paused collection, but unable to calculate progress.");
			e.printStackTrace();
		}
		
		for (String p : pipelinesToRun) {
			// replace fastdp with gmcaproc since fastdp only process single sweep
			if (nSweeps > 1 && p.equals("fastdp")) {  
				p = "gmcaproc";
				Log.note("replace fastdp with gmcaproc for pause since fastdp only deals with single sweep");
			}
			
			String outputdir = String.format("%s/%s/%s-%05d", this.outputRootDirectory, this.filenamePrefix, p,
					currentFrame);
			outputdir = newDir(outputdir, canOverwrite); // if directory exists, generate a new one
			
			PipelineOptions po = new PipelineBuilder(imageDirectory, outputdir)
					.beamline(this.beamline)
					.sampleName(this.sampleName)
					.sampleNumber(this.sampleNumber)
					.filenamePrefix(this.filenamePrefix)
					.pipeline(p)
					.njobs("16")
					.nproc("4")
					.anomalous(this.anomalous)
					.runner(autoProcessData.RUNNER)
					.xlsFile(spreadSheetFile)
					.mountedCrystal(mountedCrystal)
					.build();
			String jlabel = String.format("%s-%05d", p, currentFrame);
			CmdProcessData cp = new CmdProcessData(po, jlabel);
			new Thread(cp).start();
		}		
	}
	
	// processing job running along data collection, must set filelists
	public void processWithCollection() {
		if (this.imgfilelist != null) {
			prepareFilelist(this.imgfilelist);
			if (this.filelists == null) {
				Log.warn("processWithCollection option requires --filelist, which does not exist.");
				return;
			}
		}
		
		String to = calTimeout();

		for (String p : autoProcessData.pipelinesRunOnCollect) {
			for (Map.Entry<Integer, String> entry : this.filelists.entrySet()) {
				int endIndex = entry.getKey();
				String filelist = entry.getValue();
				Log.note("Start data processing using " + p + ", for images with no up to: " + endIndex);

				String finalTag = "";
				if (endIndex == this.imgfilelist.size()) {
					finalTag ="-Final";
				}
					
				String outputdir = String.format("%s/%s/%s-%05d%s", this.outputRootDirectory, this.filenamePrefix, p, endIndex, finalTag);

				// skip if directory already exists
				if (new File(outputdir).exists()) {
					Log.warn("directory already exists: " + outputdir + ", skipped pipeline=" + p);
					continue;
				}
				// alternatively, if directory exists, generate a new one
				outputdir = newDir(outputdir, canOverwrite); 
				
				// for final processing, turn on optimization option
				String jlabel = String.format("%s-%05d", p, endIndex);

				String pipeline = p;
				if (endIndex == this.imgfilelist.size()) {
					jlabel = String.format("%s-%s", p, "Final");
					if (p.equals("gmcaproc")) {
						pipeline = "gmcaproc w optimization";
					}
				}
				
				PipelineOptions po = new PipelineBuilder(this.imageDirectory, outputdir)
						.filelist(filelist)
						.beamline(this.beamline)
						.sampleName(sampleName)
						.sampleNumber(sampleNumber)
						.filenamePrefix(this.filenamePrefix)
						.anomalous(this.anomalous)
						.pipeline(pipeline)
						.timeout(to)
						.njobs("24")
						.nproc("4")
						.runner(autoProcessData.RUNNER)
						.xlsFile(spreadSheetFile)
						.mountedCrystal(mountedCrystal)
						.build();

				CmdProcessData cp = new CmdProcessData(po, jlabel);
				new Thread(cp).start();
				
				// some artificial delay between jobs, so that job started earlier get a head start
				// for example 25% job should have higher priority than 50% job
				try {
					Thread.sleep(4000);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		}
	}

	public void processAfterCollection() {
		for (String p : autoProcessData.pipelinesRunAfterCollect) {
			// skip fastdp since fastdp only process single sweep/partition
			if (nSweeps > 1 && p.equals("fastdp")) {
				Log.note("more than 1 partitions, end of run fastdp skipped.");
				continue;
			}
			
			String outputdir = String.format("%s/%s/%s-%s", this.outputRootDirectory, this.filenamePrefix, p, "Final");
			outputdir = newDir(outputdir, canOverwrite); // if directory exists, generate a new one
			
			PipelineOptions po = new PipelineBuilder(imageDirectory, outputdir)
					.beamline(this.beamline)
					.sampleName(this.sampleName)
					.sampleNumber(this.sampleNumber)
					.filenamePrefix(this.filenamePrefix)
					.anomalous(this.anomalous)
					.pipeline(p)
					.njobs("8")
					.nproc("4")
					.runner(autoProcessData.RUNNER)
					.xlsFile(spreadSheetFile)
					.mountedCrystal(mountedCrystal)
					.build();
			String jlabel = String.format("%s-%s", p, "Final");
			CmdProcessData cp = new CmdProcessData(po, jlabel);
			new Thread(cp).start();
		}
	}

	public void processAfterScreen() {
		// set filelist so that a user can reuse run 0 multiple times
		String filelist = null;
		if (this.imgfilelist != null) {
			String filenames = String.join("\n", this.imgfilelist);
			filelist = this.imageDirectory + "/filelist-index";
			writeFile(filenames, filelist);
		}

		for (String p : autoProcessData.pipelinesRunAfterScreen) {
			String outputdir = String.format("%s/%s/indexing/%s", this.outputRootDirectory, this.filenamePrefix, p);
			outputdir = newDir(outputdir, canOverwrite); // if directory exists, generate a new one
			
			PipelineOptions po = new PipelineBuilder(imageDirectory, outputdir)
					.beamline(this.beamline)
					.sampleName(this.sampleName)
					.sampleNumber(this.sampleNumber)
					.filenamePrefix(this.filenamePrefix)
					.filelist(filelist)
					.anomalous(this.anomalous)
					.pipeline(p)
					.njobs("1")
					.nproc("1")
					.runner(autoProcessData.RUNNER)
					.build();
			String jlabel = String.format("%s", p);
			CmdProcessData cp = new CmdProcessData(po, jlabel);
			new Thread(cp).start();
			Log.note(String.format("screening processing jobs %s started in %s.", this.sampleName, this.imageDirectory));
		}
	}

	// given a list of files to be collected, generate sublist of files for partial
	// data processing
	// i.e. --filelist=...
	// assuming that the files in FrameListCopy is ordered
	public void prepareFilelist(ArrayList<String> fl) {
		int len = fl.size();
		ArrayList<Integer> endIndices = new ArrayList<Integer>();
	
		// run thrice if total exposure time > 5min, twice if more than 60 sec, otherwise just once
		double totalExposureTime = this.exposureTime * len; 
		if (totalExposureTime >= 300.) {
			stages = new int[] { 25, 50, 100 };
		} else if (totalExposureTime >= 60. && totalExposureTime < 300) { 
			stages = new int[] { 50, 100 };
		} else {
			stages = new int[] { 100 };
		}

		for (int percent : stages) {
			int endIndex = (int) (len * percent / 100.0);
			if (endIndex != 0) {
				endIndices.add(endIndex);
			}
		}
		Set<Integer> uniqueEndIndices = new LinkedHashSet<Integer>(endIndices);

		this.filelists = new LinkedHashMap<Integer, String>();
		for (int endIndex : uniqueEndIndices) {
			String subFileList = String.join("\n", new ArrayList<String>(fl.subList(0, endIndex)));
			String outfile = String.format("%s/filelist%s", this.imageDirectory, endIndex);
			writeFile(subFileList, outfile);
			this.filelists.put(endIndex, outfile);
		}
	}

	public void writeFile(String text, String outfile) {
		try (Writer writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(outfile), "utf-8"))) {
			Log.note("Save filelist to " + outfile);
			writer.write(text);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	private String calTimeout() {
		double timeout = this.eta;

		if (timeout <= 60 && this.imgfilelist != null ) {
			timeout = this.exposureTime * this.imgfilelist.size();
		}

		timeout = timeout * nSweeps;

		// switch between energies take very long, so adding more time
		if (nEnergies > 1 ) {
			timeout += 3600.;
		}

		return Integer.toString((int)timeout + 60);
	}

	private void setOutputRootDir(boolean separate) {
		// if separate, processing will happen in a separate ./process directory under
		// dataRoot
		String subdir = String.format("%s/%s", MiscUtils.analysisNGSubDirectory, "auto");
		this.outputRootDirectory = MiscUtils.getProcessDir(this.imageDirectory, subdir).replaceAll("//", "/");
		File odir = new File(this.outputRootDirectory);
		if (!odir.exists()) {
			try {
				odir.mkdirs();
			} catch (SecurityException e) {
				Log.error("Failed to create directory " + odir);
			}
		}
	}
	
	// return a versioned directory if odir exists
	// expect that odir is absolute path
	private String newDir(String odir, boolean allowOverwrite) {
		if (allowOverwrite) {
			return odir.replaceAll("//", "/");
		}
		
		String newDir = odir;
		for (int i=0; i < 100; i++) {
			if (new File(newDir).exists()) {
				Log.warn("processing output directory already exists: " + newDir);
				newDir = String.format("%s-%02d", odir, i);
			} else {
				Log.note("new processing output directory: " + newDir);
				return newDir.replaceAll("//", "/");
			}
		}
		Log.warn("Something is wrong, too many duplicated directories found.");
		return odir.replaceAll("//", "/");
	}
}
