package gov.anl.gmca.collect;

import gov.anl.gmca.analysis.autoProcessData;
import gov.anl.gmca.collect.DevBinMode.BinMode;
import gov.anl.gmca.collect.DevBinMode.ModeChange;
import gov.anl.gmca.collect.DevFPE.FrameListResult;
import gov.anl.gmca.collect.DevFPE.XYZMotion;
import gov.anl.gmca.collect.FramePartitioner.FramePartition;
import gov.anl.gmca.collect.ModelCollectRun.Point3D;
import gov.anl.gmca.collect.ModelCollectRun.RunState;
import gov.anl.gmca.collect.ModelCollectRun.ScanMode;
import gov.anl.gmca.collect.ModelCollectRun.SetupStep;
import gov.anl.gmca.collect.ModelCollectRunList.CollectState;
import gov.anl.gmca.collect.RunCalculator.Frame;
import gov.anl.gmca.collect.ui.DlgVerifyEnergies;
import gov.anl.gmca.common.*;
import gov.anl.gmca.common.DevMove.MoveCmd;
import gov.anl.gmca.common.DevMove.MoveResult;
import gov.anl.gmca.common.ui.NBDlgDelete;
import gov.anl.gmca.gui.*;
import gov.anl.gmca.gui.options.DevUserConfig;
import gov.anl.gmca.hutch.*;
import gov.anl.gmca.io.Disk;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PV_BLUICE;
import gov.anl.gmca.raster.calc.RasterGeometry;
import gov.anl.gmca.sample.DevCollimatorCtrl;
import gov.anl.gmca.sample.DevSlider;
import gov.anl.gmca.sample.ModelBeamsize;
import gov.anl.gmca.screening.DevRobotWarmup;
import gov.anl.gmca.spotfinder.*;
import gov.anl.gmca.spotfinder.DevSFType.SFConfig;
import gov.anl.gmca.spotfinder.SFData.SFType;
import gov.anl.gmca.strategy.CmdCollectStrategy;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.util.ActionResult;
import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.Pair;
import gov.anl.gmca.util.Util;
import org.eclipse.swt.SWT;

import java.io.File;
import java.util.ArrayList;

public class CmdCollectSequence extends Command
{
	public boolean isRun0Done = false;
	private ModelCollectRunList runList;
	private boolean abortFlag = false;
	private volatile int currentRunIndex, currentFrameCount, currentRunLabel;
	private volatile String currentRunFullDir, currentRunPrefix;
	private volatile double currentExposureTimeSec;
	private volatile ArrayList<Frame> currentFrameListCopy;
	protected ListenerList.Listener<DevFPE> frameListener;
	private XYZMotion standardMotion;
	private volatile int lastFrameIndex=-1; // reset for each partition to what JBluIce sends in the FPE command
	public final static int MAX_FILENAME_LENGTH=38;
	
	public CmdCollectSequence(ModelCollectRunList _runList)
	{
		runList=_runList;
		frameListener = new ListenerList.Listener<DevFPE>()
		{
			@Override public void handleEvent(DevFPE dev, Object frameNumber)
			{
				handleCellFrameAdvance();
			}
		};
		DevFPE.get().frameListener.addListener(frameListener);
	}
	
	private void handleCellFrameAdvance()
	{
		// keep sending pause each frame to make sure FPE does in fact pause
		boolean inPausingState = runList.getCollectState() == ModelCollectRunList.CollectState.PAUSING;
		if (inPausingState)
		{
			DevFPE.get().pause(true);
		}
		
		// mid-run fix, long-term we should create the list in the constructor if possible
		if (null == currentFrameListCopy) return;
		
		// if collection is being paused, don't send the next frame which won't be collected yet
		// if collection completes, the 'done' frame index doesn't correspond to a file
		// long term: if FPE can set opStat to 0 before updating the final frame, this line would
		//            better be set as: if (DevFPE.get().getOpStat() != 0)
		boolean sfEnabled = DBConfig.getBoolean("spotfinder_enable_collect");
		boolean inCollectState = runList.getCollectState() == ModelCollectRunList.CollectState.COLLECTING;
		if (sfEnabled && inCollectState)
		{
			int currentFrameIndex = DevFPE.get().getCurrentFrameIndex();

			int numFramesAvailable = currentFrameIndex-lastFrameIndex;

			if (numFramesAvailable < 1 && currentFrameIndex != currentFrameCount) return;

			for (int frameIndex=lastFrameIndex; frameIndex<currentFrameIndex; frameIndex += 1)
			{
				if (!shouldProcess(frameIndex)) continue;

				ArrayList<String> fileList = new ArrayList<>();
				ArrayList<HeaderCheck.Values> headerValuesList = new ArrayList<>();
				for (int i=0;i<1 && frameIndex+i<currentFrameCount;i++) {
					Frame frame = currentFrameListCopy.get(frameIndex+i);
					fileList.add(frame.cbfFilename);
					HeaderCheck.Values headerValues = new HeaderCheck.Values();
					headerValues.startAngle_deg = frame.angle;
					headerValues.delta_deg = frame.delta;
					headerValues.energy_keV = frame.energy;
					headerValues.expTime_sec = currentExposureTimeSec;
					headerValuesList.add(headerValues);
				}
				for (SFConfig sfConfig : DevSFType.configList)
				{
					CmdSpotFinder cmdProcess = new CmdSpotFinderRemote
					(
						SFType.COLLECT,
						DevSFType.SFRunType.COLLECT,
						currentRunIndex,
						frameIndex,
						currentRunFullDir,
						fileList,
						currentRunPrefix,
						currentRunLabel,
						50.0,
						0.1,
						-1,
						sfConfig.id,
						headerValuesList,
						CmdSpotFinder.TIMEOUT_DEFAULT_MS
					);
					cmdProcess.saveToRedis=true;
					SpotFinderQueue.submitTask(cmdProcess);
				}
			}
			
			lastFrameIndex = currentFrameIndex;
		}
	}
	
	private static enum SkipMode { NONE, FOLLOW_PROCESSING_RATE, USE_FIXED_RATE };
	private int currentSkipNumber = 1;
	private boolean shouldProcess(int frameIndex)
	{
		boolean ret = false;
		
		Frame frame = currentFrameListCopy.get(frameIndex);
		int partitionIndex=frame.indexInPartition;
		int partitionSize=frame.partitionSize;
		
		if (partitionIndex%currentSkipNumber==0) ret = true;
		if (partitionIndex==0 || partitionIndex==partitionSize-1) ret = true;
		
		//System.out.println("Checking whether to process frame "+frameIndex+", skipNum="+currentSkipNumber+" frameCount="+currentFrameCount+" ret="+ret);
		return ret;
	}
	private int getCurrentSkipNumber()
	{
		double procRate = DetectorConfig.isEiger() ? 50.0 : 100.0;
		return (int)Math.ceil(1.0/(currentExposureTimeSec * procRate));
	}
	
	@Override
	public void run()
	{
		// check FPE connection
		if (!DevFPE.get().checkFPEState(null, false)) return;
		
		// save motion for standard mode
		Point3D currentPos = RasterGeometry.getCurrentMotorPositions().to3D();
		standardMotion = new XYZMotion(currentPos,currentPos);
		
		// determine which runs will be collected
		ArrayList<ModelCollectRun> runsToCollect = new ArrayList<ModelCollectRun>();
		if (0 == runList.getCurrentRunIndex())
		{
			runsToCollect.add((ModelCollectRun)runList.getRun(0));
		}
		else
		{
			// reminder: run count is equal to the highest active run index
			// e.g. runCount==1 means run indices 0 and 1 are active
			for (int i=runList.getCurrentRunIndex(); i<=runList.getRunCount(); i++)
			{
				ModelCollectRun run = ((ModelCollectRun)runList.getRun(i));
				boolean runComplete = RunState.COMPLETE == run.getRunState();
				boolean inSetupMode = (ScanMode.SINGLE != run.getScanMode() && SetupStep.SITE_SETUP==run.getSetupStep());
				if (!runComplete && !inSetupMode)
				{
					runsToCollect.add(run);
				}
			}
			if (runsToCollect.size()==0) return;
		}
		
		// check filename lengths
		if (DetectorConfig.isEiger())
		{
			for (ModelCollectRun run : runsToCollect)
			{
				if (run.getRunState() == ModelCollectRun.RunState.COMPLETE) continue;
				
				int maxLengthThisRun=0;
				for (Frame f : run.seq().getFrameListCopy())
				{
					int length = f.hdfFilename.length() + 5; // filenames are for master files, data filenames are 5 characters longer
					if (length>maxLengthThisRun) maxLengthThisRun=length;
				}
				if (maxLengthThisRun>MAX_FILENAME_LENGTH)
				{
					int exceedsBy = maxLengthThisRun-MAX_FILENAME_LENGTH;
					infoPopupAfterExit("Filename too long", "Run "+run.getRunLabel()+"'s filenames exceed the limit.  Please shorten the prefix ("+run.fileRoot+") by "+exceedsBy+" characters.");
					return;
				}
			}
		}

		// verify energy moves with user
		double currentEnergy = DevEnergy.getDevice().getPosition();
		ArrayList<Pair<Integer,Double>> energiesToVerify = new ArrayList<Pair<Integer,Double>>();
		for (ModelCollectRun run : runsToCollect)
		{
			if (run.getScanMode()==ScanMode.JET) continue;
			for (Double energy : run.getEnergies())
			{
				double delta = Math.abs(energy - currentEnergy);
				if (delta >= DBConfig.getDouble("energy_warn_keV"))
				{
					energiesToVerify.add(new Pair<Integer,Double>(run.getRunLabel(),energy));
				}
			}
		}
		if (energiesToVerify.size()>0 && DevCollectSettings.getEnableEnergyMove())
		{
			DlgVerifyEnergies dlg = new DlgVerifyEnergies(GUI.getShell(), energiesToVerify);
			if (dlg.open() != SWT.OK)
			{
				Log.note("Energy dialog returned cancel, aborting collection.");
				return;
			}
		}

		// verify beam size changes
		boolean multipleEnergies = false;
		for (ModelCollectRun run : runsToCollect) {
			if (run.getEnergies().size() > 1) multipleEnergies = true;
		}
		if (multipleEnergies && Util.beamSizeCalculated()) {
			boolean ok = Util.popupOkCancel(
					"Beam size changes",
					"The beam size will change with energy changes.  Is this OK?"
			);
			if (!ok) {
				Log.note("User declined beam size change popup.  Exiting.");
				return;
			}
		}

		// check that angles are within limits
		for (ModelCollectRun run : runsToCollect)
		{
			if (run.getScanMode()==ScanMode.JET) continue;

			double startAngle = run.getStartAngle();
			double endAngle = run.getEndAngle();
			double lowerLimit = DevPmacMotor.get("gonio_omega").getLimits().getLower();
			double upperLimit = DevPmacMotor.get("gonio_omega").getLimits().getUpper();
			
			if (startAngle < lowerLimit)
			{
				infoPopupAfterExit("Angle outside limits", "Run "+run.getRunLabel()+"'s start angle, "+Util.printDouble(startAngle, 2)+", is below the lower omega limit "+Util.printDouble(lowerLimit,2));
				return;
			}
			if (endAngle > upperLimit)
			{
				infoPopupAfterExit("Angle outside limits", "Run "+run.getRunLabel()+"'s end angle, "+Util.printDouble(startAngle, 2)+", is above the upper omega limit "+Util.printDouble(upperLimit,2));
				return;
			}
		}

		DevFPE.get().updateETA();

		// first, for Eiger resume, move last frame first
		ModelCollectRun firstRunToCollect = runsToCollect.get(0);
		if (DetectorConfig.isEiger())
		for (ModelCollectRun run : runsToCollect)
		{
			// TODO: update for jet mode
			if (run.getRunState() == RunState.PAUSED && run.getScanMode()!=ScanMode.JET)
			{
				int nextFrame = run.getNextFrame();
				String dir = run.getFullDirectory();
				String lastCollectedFrame = run.seq().getFrameListCopy().get(nextFrame - 1).cbfFilename;
				String lastCollectedFramePath = dir+"/"+lastCollectedFrame;

				if (!Disk.exists(lastCollectedFramePath))
				{
					boolean okToProceed = Util.popupOkCancel("Error","Expected to find last collected frame "+lastCollectedFramePath+" on resume.  Continue?");
					if (!okToProceed)
					{
						endCollectRuns(firstRunToCollect);
						infoPopupAfterExit("Collection cancelled", "User cancelled collection due to missing file on resume: "+lastCollectedFrame);
						return;
					}
				}
				else
				{
					String overwriteDir = dir+"/.overwrite";
					boolean success = true;
					if (!Disk.isDirectory(overwriteDir)) success = Disk.mkdirs(overwriteDir);
					if (success) success = Disk.mv(lastCollectedFramePath, overwriteDir+"/"+lastCollectedFrame);
					if (!success)
					{
						boolean okToProceed = Util.popupOkCancel("Error","Error moving partial frame "+lastCollectedFrame+", continue and possibly overwrite?");
						if (!okToProceed)
						{
							endCollectRuns(firstRunToCollect);
							infoPopupAfterExit("Collection cancelled", "User cancelled collection due to error moving partial frame on resume: "+lastCollectedFrame);
							return;
						}
					}
					run.dev().setNextFrame(nextFrame-1);
					run.seq().recalc();
				}
			}
		}

		// set states for collecting
		runList.devRunList.setCollectState(CollectState.COLLECTING,true);
		firstRunToCollect.dev().setRunStatus(RunState.COLLECTING, true);
		runList.devRunList.setCurrentRun(firstRunToCollect.runIndex, true);

		// check for files
		// TODO: update for jet mode
		ArrayList<SafeImageDelete.FileToDelete> existingFiles = new ArrayList<>();
		for (ModelCollectRun run : runsToCollect)
		{
			if (run.getRunState() == ModelCollectRun.RunState.COMPLETE) continue;

			for (Frame frame : new ArrayList<Frame>(run.seq().getRemainingFrameList()))
			{
				if (frame.frameIndex < run.getNextFrame()) continue;

				long modTime=(new File(run.getFullDirectory()+"/"+frame.cbfFilename)).lastModified() / 1000;
				SafeImageDelete.FileToDelete fileToDelete = new SafeImageDelete.FileToDelete(run.getFullDirectory(), frame.cbfFilename, modTime);

				if ( Disk.exists(fileToDelete.getFullPath()) )
				{
					existingFiles.add(fileToDelete);
				}
			}
		}

		if (existingFiles.size()!=0)
		{
			NBDlgDelete dlgDelete = new NBDlgDelete(existingFiles);
			int okToDelete = dlgDelete.open();

			if (SWT.OK == okToDelete)
			{
				SafeImageDelete.delete(existingFiles, true, DetectorConfig.isPilatus());
			}
			else
			{
				endCollectRuns(firstRunToCollect);
				infoPopupAfterExit("Deletion declined, collection aborted.","Deletion declined, collection aborted.");
				return;
			}
		}

		// TODO: generate log
		
		// send robot warm-up if this collection will overlap with the warm-up time
		if (!Main.SIM_FPE) {
			int etaSec = ETA.calcActiveETA();
			DevRobotWarmup.get().sendWarmupIfNeeded(etaSec);

			// close shutter
			Log.note("Closing shutter");
			DevShutter.get("shutter").close();
			DevShutter.get("shutter").waitFor();

			// pre-collect checks
			Log.note("Pre-collect checks");

			CmdCheck preCollectChecks = new CmdCheck(CmdCheck.FLAGS_PRECOLLECT);
			preCollectChecks.run();
			if (!preCollectChecks.checksPassed()) {
				endCollectRuns(firstRunToCollect);
				errorPopupAfterExit("Checks failed", "Pre-collect checks failed, collection aborted.");
				return;
			}

			// ringlight to max
			DevSlider ringLight = DevSlider.get("ring");
			ringLight.setIntensity(ringLight.getHigherLimit());
		}

		////////////////////////////////////////////////////////////////////////////////
		//////////////////////////////// start run iteration ///////////////////////////
		////////////////////////////////////////////////////////////////////////////////
		
		for (ModelCollectRun run : runsToCollect)
		{
			boolean fpeCommandSent = false;
			currentRunLabel = run.getRunLabel();
			currentRunIndex = run.getRunIndex();
			currentExposureTimeSec = run.getExposureTime();
			currentRunFullDir = DevDataDir.addToPrimary(run.getDirectorySuffix());
			currentRunPrefix = run.getFileRoot();
			currentFrameCount = run.seq().getFrameListCopy().size();
			currentSkipNumber = getCurrentSkipNumber();
			Log.note("Skip number = "+currentSkipNumber);
			
			// these operations are already done for the first run
			if (run != firstRunToCollect)
			{
				run.dev().setRunStatus(RunState.COLLECTING, true);
				runList.devRunList.setCurrentRun(run.runIndex, true);
			}
			
			// create the directory
			String collectDir = run.getFullDirectory();
			if (!Disk.exists(collectDir))
			{
				if (!Disk.mkdirs(collectDir))
				{
					endCollectRuns(run);
					errorPopupAfterExit("Directory error", "Couldn't create directory "+collectDir+", collection aborted.");
    				return;
				}
			}
			
			// set the detector binning mode
			if (!Main.SIM_FPE) {
				BinMode reqBinMode = ModelCollectRunList.get().getBinMode();
				ModeChange binModeChange = DevBinMode.get().startChange(reqBinMode);
				// move detector
				if (Math.abs(DevPmacMotor.get("detector_z").getPosition() - run.getdistance_mm()) >= 0.001) {
					Log.note("Moving detector to " + Util.sprintf("%.3f", run.getdistance_mm()) + " mm");
					MoveCmd detectorMove = DevPmacMotor.get("detector_z").startMove(run.getdistance_mm());
					MoveResult detectorResult = detectorMove.waitFor();
					if (!detectorResult.isOK()) {
						endCollectRuns(run);
						errorPopupAfterExit("Detector move error", "Detector move to " + Util.sprintf("%.3f", run.getdistance_mm()) + " failed, collection aborted.");
						return;
					}
				}

				// set up guardslits and collimator/backlight/beamstop/pin-diode
				Object[] setupRetVal = ModelBeamsize.setUpBeamsize
						(
								run.getBeamWidth_um(),
								run.getBeamHeight_um(),
								false
						);
				ActionResult setupResult = (ActionResult) setupRetVal[2];

				if (!setupResult.succeeded()) {
					endCollectRuns(run);
					errorPopupAfterExit("Beam size setup error", setupResult.getError());
					return;
				}

				// wait for the binning mode change
				ActionResult binModeResult;
				try {
					binModeResult = binModeChange.waitFor();
				} catch (Exception e) {
					e.printStackTrace();
					binModeResult = ActionResult.FAILURE;
				}
				if (!binModeResult.succeeded()) {
					endCollectRuns(run);
					errorPopupAfterExit
							(
									"CCD Bin Mode Error",
									"Changing the CCD binning mode to " + reqBinMode + " failed, collection aborted.\n\n" + DevBinMode.RESTART_MESSAGE
							);
					return;
				}

				// set attenuation
				Log.note("Setting attenuation to " + Util.printDouble(run.getAttenuation_factors(), 4));
				MoveCmd attenMove2 = DevAttenuation.get().startMove(run.getAttenuation_factors());
				MoveResult attenResult2 = attenMove2.waitFor();
				if (!attenResult2.isOK()) {
					endCollectRuns(run);
					errorPopupAfterExit("Attenuation move error", "Attenuation move to " + Util.printDouble(run.getAttenuation_factors(), 4) + " (factors) failed, collection aborted.");
					return;
				}
			}

			// clear existing SF stats
			if (run.getNextFrame()==0) {
				DBImageStats.initialize(Redis.image_stats(run.getDirectorySuffix(), run.getFileRoot(), run.getRunLabel(), ""), currentFrameCount);
			}
			
			int totalFrameCount = run.seq().getFrameListCopy().size();
			boolean startedOnce = false; // QX: make sure the job only started once inside the loop
			
			////////////////////////////////////////////////////////////////////////////////
			//////////////////////////// start partition iteration /////////////////////////
			////////////////////////////////////////////////////////////////////////////////

			currentFrameListCopy = new ArrayList<Frame>(run.seq().getFrameListCopy());
			ArrayList<FramePartition> partitionListCopy = new ArrayList<FramePartition>(run.seq().getFramePartitions());
			for (FramePartition partition : partitionListCopy)
			{
				if (partition.partitionIndex < 0) continue; // frames already taken are in partition -1
				if (!activeCheck()) return;

				DevFPE.get().updateETA();

				// check for pause
				if (runList.getCollectState() == ModelCollectRunList.CollectState.PAUSING)
				{
					Log.note("Pause received, quitting sequence thread");
					endCollectRun(run, totalFrameCount, fpeCommandSent);
					endCollectRuns(null);
					return;
				}

				//TODO QX processing jobs at start of collecting, only run once
				// handle pauses in two cases:
				// 1) pause/resume before anything is collected (setup period), does not start any jobs
				// 2) pause/resume during data collection, does not start any new jobs
				// By moving the code after pause handling code, new jobs won't started for 1)
				// if user pause in the middle of collection, nextFrame would be not 0, so 2) should be OK
				if (!startedOnce && run.getNextFrame() == 0) {
					execStartofRunProcessingJobs(run);
					startedOnce = true;
				}

				if (!Main.SIM_FPE && DevPmacMotor.get("mo_En_E").isMoveNeededTo(partition.energy))
				{
					// move energy
					Log.note("Moving energy to "+Util.printDouble(partition.energy,4)+" keV");
					MoveCmd energyMove = DevEnergy.getDevice().startMove(partition.energy);
					MoveResult energyResult = energyMove.waitFor();
					if (!energyResult.isOK())
					{
						endCollectRun(run, totalFrameCount, fpeCommandSent);
						endCollectRuns(null);
						errorPopupAfterExit("Energy move error", "Energy move to "+Util.printDouble(partition.energy,4)+" keV failed, collection aborted.");
						return;
					}

					// set attenuation
					Log.note("Setting attenuation to "+Util.printDouble(run.getAttenuation_factors(),4));
					MoveCmd attenMove = DevAttenuation.get().startMove(run.getAttenuation_factors());
					MoveResult attenResult = attenMove.waitFor();
					if (!attenResult.isOK())
					{
						endCollectRun(run, totalFrameCount, fpeCommandSent);
						endCollectRuns(null);
						errorPopupAfterExit("Attenuation move error", "Attenuation move to "+Util.printDouble(run.getAttenuation_factors(),4)+" (factors) failed, collection aborted.");
						return;
					}
				}

				if (!activeCheck()) return;

				// move gonio x-y-z
				boolean isNonHelicalVector = run.getScanMode()==ScanMode.VECTOR && partition.siteIncrement==0;
				if
				(
					null != partition.position &&
					(
						run.scanMode == ScanMode.RASTER ||
						run.scanMode == ScanMode.IMPORT ||
						isNonHelicalVector
					)
				)
				{
					Log.note("Moving gonio x-y-z to "+Util.sprintf("%.2f,%.2f,%.2f",partition.position.x,partition.position.y,partition.position.z));

					MoveResult xyzResult = DevXYZ.syncMove(partition.position, true);
					if (!xyzResult.isOK())
					{
						endCollectRun(run, totalFrameCount, fpeCommandSent);
						endCollectRuns(null);
						errorPopupAfterExit("Sample XYZ move error", "Sample XYZ move to "+Util.printDouble(partition.position.x,2)+","+Util.printDouble(partition.position.y,2)+","+Util.printDouble(partition.position.z,2)+" failed, collection aborted.");
						return;
					}
				}

				if (!activeCheck()) return;

				// reset frame list so handleFrameAdvance knows where to start from
				lastFrameIndex = partition.firstFrameIndex;

				// figure out the FPE mode
				boolean isStandard = run.getScanMode()==ScanMode.SINGLE;

				// TODO: support just passing in the ModelCollectSequence.Frame object
				DevFPE.FrameList fpeFrame 	= DevFPE.get().newFrameList();
				fpeFrame.delta 			= run.getDelta();
				fpeFrame.prefix			= partition.prefix;
				fpeFrame.energy 		= DevPmacMotor.get("mo_En_E").getPosition();
				fpeFrame.exposureTime 	= run.getExposureTime();
				fpeFrame.startAngle 	= partition.startAngle;
				fpeFrame.nextAngle 		= partition.nextAngle;
				fpeFrame.endAngle		= partition.startAngle+partition.frameCount*run.getDelta();
				fpeFrame.firstFrameIndex= partition.firstFrameIndex;
				fpeFrame.runNumber 		= run.getRunLabel();
				fpeFrame.filename 		= partition.getFilenameToSendToFPE();
				fpeFrame.firstFrameLabel = partition.firstFrameLabel;
				fpeFrame.frameCount		= partition.frameCount;
				fpeFrame.directory 		= DevDataDir.addToFPEDir(run.getDirectorySuffix());
				fpeFrame.axis 			= run.getAxis();
				fpeFrame.fpeMode		= partition.fpeMode;
				fpeFrame.shutteredMotion = isStandard ? standardMotion : partition.shutteredMotion;
				fpeFrame.shutterlessMotion = isStandard ? standardMotion : partition.shutterlessMotion;
				fpeFrame.start();
				fpeCommandSent = true;

				do
				{
					// check for abort
					if (abortFlag) fpeFrame.requestStop();
					if (!activeCheck()) return;
					Util.sleep(10);
				}
				while (fpeFrame.isRunning());

				// send run 0 filenames
				if (0==run.getRunIndex())
				{
					if (currentFrameListCopy.size()>0)
						Epics.put(PV_BLUICE.run0LastFrame0, run.getFullDirectory()+"/"+currentFrameListCopy.get(0).cbfFilename);
					if (currentFrameListCopy.size()>1)
						Epics.put(PV_BLUICE.run0LastFrame1, run.getFullDirectory()+"/"+currentFrameListCopy.get(1).cbfFilename);
				}

				FrameListResult result = fpeFrame.getResult();
				System.out.println("Collect result = "+result);

				// check for pause
				if (runList.getCollectState() == ModelCollectRunList.CollectState.PAUSING)
				{
					Log.note("Pause received, quitting sequence thread");
					endCollectRun(run, totalFrameCount, fpeCommandSent);
					endCollectRuns(null);
					return;
				}

				// check for frame error
				if (!result.isOK())
				{
					endCollectRun(run, totalFrameCount, fpeCommandSent);
					endCollectRuns(null);
					return;
				}

				// TODO: review - does this need to be done between partitions?
				//nextFrameIndex++;

			} // end partition

			// start strategy calc
			if (run.getStrategyEnabled())
			{
				Log.note("Starting strategy calculation");

				try
				{
					CmdCollectStrategy cmdStrategy = new CmdCollectStrategy
					(
						run.getFullDirectory(),
						currentFrameListCopy.get(0).hdfFilename,
						currentFrameListCopy.get(1).hdfFilename,
						currentFrameListCopy.get(0).cbfFilename,
						currentFrameListCopy.get(1).cbfFilename
					);
					cmdStrategy.setSampleName(run.fileRoot);
					cmdStrategy.start();
				}
				catch (Exception e)
				{
					e.printStackTrace();
					Log.error("Error starting strategy calculation, aborting.");
				}
			}

			//TODO test end of collection run processing jobs
			execEndofRunProcessingJobs(run);
    		endCollectRun(run, totalFrameCount, fpeCommandSent);
		} // end run
		
        endCollectRuns(null);
	}

	// launch processing jobs for each run at the end
	private void execEndofRunProcessingJobs(ModelCollectRun run) {
		try {
			Integer bgProcMode = DevUserConfig.get().getXDSProc(); // Epics.getInteger(PV_RUNS.bgProc);
			
			if ( bgProcMode != 0) {
				if (run.getSiteCheckEnable()) {
					Log.note("Skipped processing for site-check run.");
					return;
				}
				
				int runNumber = run.getRunIndex();
				if (runNumber == 0) {
					autoProcessData indexjob = new autoProcessData(run, autoProcessData.PROCESSING_MODE.AFTER_SCREENING);
					new Thread(indexjob).start();
					Log.note("Started end of run indexing jobs for run: " + runNumber);
				} else {
					autoProcessData r = new autoProcessData(run, autoProcessData.PROCESSING_MODE.AFTER_COLLECTION);
					new Thread(r).start();
					Log.note("Started end of run processing jobs for run: " + runNumber);
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
			Log.warn("Failed to start post collection processing job.");
		}
	}

	// launch processing jobs for each run at the start
	private void execStartofRunProcessingJobs(ModelCollectRun run) {
		try {
			Integer bgProcMode = DevUserConfig.get().getXDSProc();
			if (bgProcMode != 0) {

				if (run.getSiteCheckEnable()) {
					Log.note("Skipped processing for site-check run.");
					return;
				}

				int runNumber = run.getRunIndex();
				if (runNumber != 0) {
					autoProcessData r = new autoProcessData(run, autoProcessData.PROCESSING_MODE.WITH_COLLECTION);
					new Thread(r).start();
					Log.note("Started gmcaproc processing jobs for run: " + runNumber);
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
			Log.warn("failed to start gmcaproc job.");
		}
	}
	
	private boolean activeCheck()
	{
		if (!Client.isMaster())
		{
			Log.note("Client became slave during collection, exiting collection thread.");
			return false;
		}
		return true;
	}
	
	protected void endCollectRun(ModelCollectRun run, int totalFrameCount, boolean fpeCommandSent)
	{
		System.out.println("endCollectRun: run#="+run.getRunLabel()+" index="+DevFPE.get().getCurrentFrameIndex()+" size="+totalFrameCount);

		// screenshot of the graph for the run that just finished
		// TODO: enable
		//TabCollect.get().leftPane.diffTab.spotFinderGraph.saveScreenshot(false);
		
		if (0 == run.getRunIndex())
		{
			// TODO: replace this with DevFPE change to wait for final counter update
			//if (DetectorConfig.isEiger()) Util.sleep(300);
			
			int newStartLabel =
				run.getStrategyEnabled()
				?run.getStartFrameLabel()+3
				:run.getStartFrameLabel()+1;
			
			isRun0Done = true;
			run.dev().setNextFrame(0);
			run.dev().setStartFrameLabel(newStartLabel,false);
			run.dev().setRunStatus(RunState.INACTIVE, true);
		}
		else
		{
			RunState state = null;
			if (fpeCommandSent) {
				int nextFrame = DevFPE.get().getCurrentFrameIndex();
				state = ModelCollectRun.getCorrectState(nextFrame, totalFrameCount);
			} else {
				state = RunState.INACTIVE;
			}
			run.dev().setRunStatus(state, true);
		}
	}
	
	protected void endCollectRuns(ModelCollectRun collectingRun)
	{
		if (!Main.SIM_FPE) {
			// ringlight off
			DevSlider ringLight = DevSlider.get("ring");
			ringLight.setIntensity(ringLight.getLowerLimit());

			// backlight in
			DevCollimatorCtrl.CCMove ccMove = DevCollimatorCtrl.get().newMove();
			ccMove.setArguments("OUT backlightIN bluice");
			//ccMove.setDest("backlight",		InOut.IN);
			//ccMove.setDest("collimator",	InOut.OUT);
			ccMove.start();
		}

		// restore states
		if (null != collectingRun)
		{
			RunState runState = ModelCollectRun.getCorrectState(DevFPE.get().getCurrentFrameIndex(), collectingRun.seq().getFrameListCopy().size());
			collectingRun.dev().setRunStatus(runState, true);
		}
		DevFPE.get().frameListener.removeListener(frameListener);
		runList.devRunList.setCollectState(CollectState.INACTIVE,true);
		DevFPE.get().setETA(0);
		
		Log.note("Collection inactive.");
	}
	
	protected void errorPopupAfterExit(String title, String msg)
	{
		final String f_title = title;
		final String f_msg = msg;
		
		GUI.asyncExec(new Runnable() { public void run()
		{
			while (runList.isCollectThreadAlive()) Util.sleep(50);
			Log.error(f_msg);
			Util.popupError(f_title, f_msg);
		}});
	}
	
	protected void infoPopupAfterExit(String title, String msg)
	{
		final String f_title = title;
		final String f_msg = msg;
		
		GUI.asyncExec(new Runnable() { public void run()
		{
			while (runList.isCollectThreadAlive()) Util.sleep(50);
			Log.note(f_msg);
			Util.popupOk(f_title, f_msg);
		}});
	}
	
	void abort()
	{
		abortFlag = true;
		DevFPE.get().abort();
	}
}
