package gov.anl.gmca.collect;

import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.hutch.DevHutchShutter;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PV_FEEDBACK;
import gov.anl.gmca.io.PVLookup.PV_MISC;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.Util;
import gov.aps.jca.dbr.DBR;

public class CmdFeedbackCheck
{
	private static enum LOCKIN_MODE { SOFTWARE, HYBRID, HARDWARE };
	private static final int HW_MODE_CHECK_TIMEOUT_MS = 5000;
	private static String aoutValue = "";
	private static Epics.EpicsValueMonitor aoutMonitor = null; 
	
	public static boolean runCheck(boolean checkIfRunning, boolean checkHWMode)
	{
		if (!DevCollectSettings.getFeedbackCheckEnabled())
		{
			System.out.println("Wait for feedback disabled, returning proceed");
			return true;
		}
		
		System.out.println("In waitForFeedbackCheck, checkIfRunning="+checkIfRunning+" checkHWMode="+checkHWMode);
		
		// run different checks depending on lock-in mode
		//LOCKIN_MODE lockinMode = LOCKIN_MODE.values()[Epics.getInteger(PV_FEEDBACK.uselockin)];
		//if (lockinMode==LOCKIN_MODE.HARDWARE) return runHardwareCheck(checkHWMode);
		//else return runSoftwareCheck(checkIfRunning);
		return runSoftwareCheck(checkIfRunning);
	}

	private static boolean runHardwareCheck(boolean checkHWMode)
	{
		// start the AOUT monitor if needed
		if (null == aoutMonitor)
		{
			aoutMonitor = new Epics.EpicsValueMonitor()
			{
				@Override public void valueChanged(String pvName, DBR value)
				{
					aoutValue = Epics.dbrToString(value);
				}
			};
			Epics.addMonitor(PV_FEEDBACK.serialLIA$AOUT, aoutMonitor);
		}

		if (checkHWMode)
		{
			// check that hardware lockin hardware is in mode 2
			String queryString = Util.getBID().equals("23o")?"DDEF?":"DDEF?1";
			aoutValue = "";
			Epics.put(PV_FEEDBACK.serialLIA$TINP, queryString);
			long sendTime = System.currentTimeMillis();
			while (aoutValue.equals(""))
			{
				long now = System.currentTimeMillis();
				if (now-sendTime>HW_MODE_CHECK_TIMEOUT_MS)
				{
					Log.error("Hardware check timed out after "+HW_MODE_CHECK_TIMEOUT_MS+" ms");
					return false;
				}
				Util.sleep(10);
			}
			if (!aoutValue.equals("0,0"))
			{
				// TODO: finish error message
				Log.error("Invalid hardware check response: ");
				return false;
			}
			else
			{
				System.out.println("Feedback check: hardware mode OK, response="+aoutValue);
			}
		}
		
		// check that piezo mirrors are in the range 1-9
		// if mirrors are not in range, run Sergey's script
		double v_high = DBConfig.getDouble("check_feedback_v_low");
		double v_low = DBConfig.getDouble("check_feedback_v_high");
		double mostab_readback = Epics.getDouble(PV_MISC.mostab_readback);
		if (mostab_readback<v_low || mostab_readback>v_high)
		{
			Log.note(Util.sprintf("DCM voltage (%.2f) out of range (%.2f,%.2f), running correction script",mostab_readback, v_low, v_high));
			return false;
		}
		else
		{
			System.out.println("Feedback check: DCM voltage OK, value="+Util.printDouble(mostab_readback, 4));
		}
		
		// check that APS current is OK
		double apsCurrent = Epics.getDouble(PV_MISC.apsCurrent);
		if (apsCurrent < DBConfig.getDouble("check_feedback_aps_min_current"))
		{
			Log.error(Util.sprintf("APS current is below the minimum %.1fV, check failed.",apsCurrent));
			return false;
		}
		
		// check that A shutter is open
		boolean aShutterOpen = DevHutchShutter.get("aShutter").isOpen();
		if (!aShutterOpen)
		{
			Log.error("A shutter is not open, check failed.");
			return false;
		}
		
		return true;
	}
	
	private static boolean runSoftwareCheck(boolean checkIfRunning)
	{
		int statusCode = 0;
		String statusMsg = "";
		
		while (true)
		{
	        // Get the status code and message either using polling or monitoring
			if (checkIfRunning)
			{
				DevCollectSettings.setMostabStatus(0);
				
				Util.sleep(50);

				// wait up to 5 seconds
				long startTimeMs = System.currentTimeMillis();
				while (true)
				{
					statusCode = DevCollectSettings.getMostabStatus();
					if (0 != statusCode) break;

					Util.sleep(100);
					long waitTimeMs = System.currentTimeMillis() - startTimeMs;
					if (waitTimeMs > 15000) break;
				}
				
				statusMsg = DevCollectSettings.getMostabStatusLine();
			}
			else
			{
			    // Assumes feedback is running
				statusCode = DevCollectSettings.getMostabStatus();
				statusMsg = DevCollectSettings.getMostabStatusLine();
			}
    		
    		switch (statusCode)
    		{
    		case 0: // Feedback is not running
    			Util.popupOk("Intensity Feedback Error",
    						 "Intensity feedback is not running. Please ask your host to start it. Operation aborted.");
    			return false;
    			
    		case 1: // Feedback is active
    			Log.note( "Feedback check: software mode reports OK, returning proceed. "+statusMsg);
    			return true;

    		case 2: // Feedback is busy, check back soon
    			Log.note("Waiting for intensity feedback: "+statusMsg);
    			Util.sleep(1000);
    			continue;
    			
    		case 3: // Requires user intervention
    			boolean okToProceed = 
    				Util.popupOkCancel("Intensity Feedback Error",
    								   "Intensity feedback reports the following error: "+statusMsg+" "+
    							   	   "Click OK if the condition has been fixed, Cancel to abort.");
    			if (okToProceed)
    			{
    				// check again...
    				continue;
    			}
    			else
    			{
    				return false;
    			}
    			
    		case 4: // Serious error, contact host
    			Util.popupOk("Intensity Feedback Error", 
    						 "Intensity feedback reports the following error: "+statusMsg+". "+
    						 "Operation aborted.  Please contact your host.");
    			return false;
    			
    		default:
    			Util.popupOk("Error",
    						 "Unknown intensity feedback code $statusCode, message: "+statusMsg+". "+
    						 "Aborting.");
    			return false;
    		}
		}
	}
}
