package gov.anl.gmca.collect;

import gov.anl.gmca.collect.ModelCollectRun.Point3D;
import gov.anl.gmca.common.SQLDataType;
import gov.anl.gmca.io.Database;
import gov.anl.gmca.io.Database.*;
import gov.anl.gmca.io.PVLookup.PVName;
import gov.anl.gmca.io.PVLookup.PV_RUNS;
import gov.anl.gmca.io.hw.Epics;
import gov.anl.gmca.util.Util;
import gov.aps.jca.dbr.DBR;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.concurrent.Semaphore;

public class DBImportSites
{
	public static final DBName DATABASE_NAME = DBName.USERDATA;
	public static final String TABLE_NAME="ImportSites";
	public static final PVName updatePV = PV_RUNS.importDBUpdated;
	private static boolean updatePending = false;
	
	public static class ImportSite
	{
		public int collectRun=-1;
		public int index=-1;
		public Point3D xyz;
	}
	
	private static ArrayList<ImportSite> allImportSites = new ArrayList<ImportSite>();
	public static ArrayList<ImportSite> getAllImportSites()
	{
		while (updatePending) Util.sleep(50);
		lock.acquireUninterruptibly();
		
		ArrayList<ImportSite> ret = null;
		try
		{
			ret = new ArrayList<ImportSite>(allImportSites);
		}
		catch (Exception e) { e.printStackTrace(); }
		finally
		{
			lock.release();
		}
		return ret;
	}
	private static Semaphore lock = new Semaphore(1);
	
	public static void init()
	{
		Epics.addMonitor(updatePV, new Epics.EpicsValueMonitor() {
			@Override public void valueChanged(String pvName, DBR value) {
				refresh();
			}
		});
		refresh();
	}
	
	public static void refresh()
	{
		lock.acquireUninterruptibly();
		try
		{
			allImportSites.clear();
			loadSitesFromSQL();
		}
		catch (Exception e) { e.printStackTrace(); }
		finally
		{
			lock.release();
			updatePending = false;
		}
	}

	private static void createTable()
	{
		ArrayList<DBField> fields = new ArrayList<DBField>();
		fields.add(new DBField("collectRun", SQLDataType.INT, false, true, null));
		fields.add(new DBField("siteIndex", SQLDataType.INT, false, true, null));
		fields.add(new DBField("crystalX", SQLDataType.FLOAT, false, true, null));
		fields.add(new DBField("crystalY", SQLDataType.FLOAT, false, true, null));
		fields.add(new DBField("crystalZ", SQLDataType.FLOAT, false, true, null));
		Database.create(DATABASE_NAME, TABLE_NAME, fields);
	}
	
	private static void loadSitesFromSQL()
	{
		Result sqlResult = Database.select(DATABASE_NAME, TABLE_NAME, null, null, null);
		if (sqlResult.error == Database.RESULT_ERROR.TABLE_DOESNT_EXIST)
		{
			createTable();
		}
		if (null==sqlResult || null == sqlResult.getRows()) return;
		
		// iterate through SQL result rows
		System.out.println("Loaded "+sqlResult.getRows().size()+" import sites");
		for (HashMap<String,String> row : sqlResult.getRows())
		{
			try
			{
				ImportSite site = new ImportSite();
				site.collectRun = Integer.parseInt(row.get("collectRun"));
				site.index = Integer.parseInt(row.get("siteIndex"));
				double x = Double.parseDouble(row.get("crystalX"));
				double y = Double.parseDouble(row.get("crystalY"));
				double z = Double.parseDouble(row.get("crystalZ"));
				site.xyz = new Point3D(x,y,z);
				
				allImportSites.add(site);
			}
			catch (NumberFormatException nfe) { nfe.printStackTrace(); }
		}
	}
	
	// sending null for the siteList clears all sites
	// sending -1 for the runIndex clears the entire table
	public static void setSitesForRun(int runIndex, ArrayList<ImportSite> siteList)
	{
		try
		{
			// delete existing sites for this run
			WhereList whereList = new WhereList();
			whereList.add("collectRun", WhereOp.EQUALS, ""+runIndex);
			Database.delete(DATABASE_NAME, TABLE_NAME, whereList);
	
			// add new sites
			if (null != siteList && siteList.size()>0)
			{
				ArrayList<HashMap<String,String>> rows = new ArrayList<HashMap<String,String>>();
				for (ImportSite site : siteList)
				{
					HashMap<String,String> row = new HashMap<>();
					row.put("collectRun", 	""+site.collectRun);
					row.put("siteIndex", 	""+site.index);
					row.put("crystalX", 	""+site.xyz.x);
					row.put("crystalY", 	""+site.xyz.y);
					row.put("crystalZ", 	""+site.xyz.z);
					rows.add(row);
				}
				Database.insert(DATABASE_NAME, TABLE_NAME, rows);
			}
		}
		catch (Exception e) { e.printStackTrace(); }
		
		updatePending=true;
		Epics.putInteger(updatePV, 0);
		Epics.putInteger(updatePV, 1);
	}
}
