package gov.anl.gmca.collect;

import gov.anl.gmca.hutch.DetectorConfig;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PV_MAR;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.ActionResult;
import gov.anl.gmca.util.Util;
import gov.aps.jca.dbr.DBR;

import java.util.StringTokenizer;

public class DevBinMode
{
	// for asynchronous use
	public ModeChange startChange(BinMode reqMode)
	{
		ModeChange mc = new ModeChange(reqMode);
		mc.start();
		return mc;
	}
	
	// set the mode through the detector interface of FPE
	// returns true on success
	public synchronized boolean setMode(BinMode newMode)
	{
		// Pilatus doesn't need binning mode changes
		if (DetectorConfig.isDectris()) return true;
		
		Log.note("Setting CCD bin mode to "+newMode);
		
		// query the detector to get the current mode, return if no change is needed
		BinMode currentMode = queryBinMode();
		if (null == currentMode) return false;
		if (currentMode == newMode) return true;
		
		// change mode
		sendCmd("set_bin,"+modeToStr(newMode));

		// wait until the timeout for the mode to change
		long setTime = System.currentTimeMillis();
		long waitTime = 0L;
		do
		{
			long now = System.currentTimeMillis();
			waitTime = now-setTime;
			if (waitTime>MODE_SET_TIMEOUT_MS)
			{
				Log.error("Tried to change mode to "+newMode+" but current mode is "+currentMode+".");
				return false;
			}
			
			currentMode = queryBinMode();
			System.out.println("Waiting for bin mode change: current="+currentMode+" new="+newMode+".");
			Util.sleep(MODE_SET_POLL_MS);
		}
		while (currentMode != newMode);
			
		Log.note("Binning mode succeeded in "+waitTime+"ms");
		return true;
	}
	
	/////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////// settings //////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	
	public 	static 	enum	BinMode
	{
		UM_36, UM_72, UM_144;
		public int getIntSize() { return Integer.parseInt(this.toString().substring(3)); }
	};
	private static 	int 	CMD_TIMEOUT_MS 		= 5000;
	private static 	int 	CMD_SEND_TIMEOUT_MS	= 2000;
	private static 	int 	CMD_SEND_SLEEP_MS	= 10;
	private static 	int 	CMD_CLEAR_SLEEP_MS	= 10;
	private static 	int 	MODE_SET_TIMEOUT_MS	= 5000;
	private static 	int 	MODE_SET_POLL_MS	= 100;
	private static 	int 	RESPONSE_SLEEP_MS	= 10;
	private 		String 	lastResponse 		= null;
	private 		String 	lastCmdSent 		= null;
	private volatile boolean cmdClear			= true;

	public static final String RESTART_MESSAGE =
		"Please try again in a moment.  If the problem persists, restart JBluIce.";
	
	public static DevBinMode get() { return inst; }
	private static DevBinMode inst;

	/////////////////////////////////////////////////////////////////////////////////////////
	////////////////// convert between mode enum and the detector mode string ///////////////
	/////////////////////////////////////////////////////////////////////////////////////////

	public static String modeToStr(BinMode mode)
	{
		int binNumber = DetectorConfig.getBinArgument(mode);
		return binNumber+","+binNumber;
	}
	
	public static BinMode strToMode(String s)
	{
		try
		{
			StringTokenizer st = new StringTokenizer(s,",");
			if (st.countTokens()!=2)
			{
				Log.error("Unknown bin mode string: "+s);
				return null;
			}
			int a = Integer.parseInt(st.nextToken());
			int b = Integer.parseInt(st.nextToken());
			if (a!=b)
			{
				Log.error("Bin arguments don't match: "+a+","+b);
				return null;
			}
			
			for (BinMode mode : BinMode.values())
			{
				Integer binNumber = DetectorConfig.getBinArgument(mode);
				if (null != binNumber && binNumber==a) return mode;
			}
			
			Log.error("Unknown bin argument: "+a);
			return null;
		}
		catch (Exception e)
		{
			e.printStackTrace();
			Log.error("Unknown bin mode string: "+s);
			return null;
		}
	}
	
	/////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////// asynchronous access /////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////

	public class ModeChange
	{
		private BinMode reqMode;
		private Thread changeThread;
		private volatile Boolean succeeded;
		
		public ModeChange(BinMode reqMode)
		{
			this.reqMode = reqMode;
		}

		public void start()
		{
			changeThread = new Thread(new Runnable(){public void run()
			{
				succeeded = setMode(reqMode);
			}});
			changeThread.start();
		}
		
		public ActionResult waitFor() throws Exception
		{
			if (null == changeThread) throw new Exception("waitFor() called on mode change object that was not started");
			
			while (changeThread.isAlive()) Util.sleep(10);
			return succeeded ? ActionResult.SUCCESS : ActionResult.FAILURE;
		}
	}
	
	/////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////// ctor and synchronous access methods /////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	
	public DevBinMode()
	{
		inst = this;
		
		Epics.addMonitor(PV_MAR.DetCmdStr, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				String str = Epics.dbrToString(value);
				//System.out.println("camonitor "+pvName+" "+str);
				if (str.length()==0) cmdClear = true;
			}
		});
		Epics.addMonitor(PV_MAR.DetInStr, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				String str = Epics.dbrToString(value);
				//System.out.println("camonitor "+pvName+" "+str);
				if (str.length()>0)
				{
					lastResponse = str;
				}
			}
		});
		Epics.addMonitor(PV_MAR.DetOutStr, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				String str = Epics.dbrToString(value);
				//System.out.println("camonitor "+pvName+" "+str);
				if (str.length()>0)
				{
					lastCmdSent = str;
				}
			}
		});
		lastCmdSent = "";
	}

	// waits for the cmd PV to be clear before sending a new cmd
	// new for 2013-3: waits for this command to show up in DetOutStr
	// should only be called from queryBinMode or setMode, otherwise additional locks are needed
	private void sendCmd(String cmd)
	{
		long start = System.currentTimeMillis();
		while (!cmdClear)
		{
			long now = System.currentTimeMillis();
			if (now-start>CMD_TIMEOUT_MS)
			{
				Log.error(PV_MAR.DetCmdStr+" PV did not clear after "+CMD_TIMEOUT_MS+"ms, could not send command "+cmd+".");
				return;
			}
			System.out.println("Waiting for DetCmdStr to clear: value="+Epics.getString(PV_MAR.DetCmdStr)+".");
			Util.sleep(CMD_CLEAR_SLEEP_MS);
		}
		
		cmdClear = false;
		Epics.put(PV_MAR.DetCmdStr, cmd);
		
		start = System.currentTimeMillis();
		while (!lastCmdSent.equals(cmd))
		{
			long now = System.currentTimeMillis();
			if (now-start>CMD_SEND_TIMEOUT_MS)
			{
				Log.error(PV_MAR.DetCmdStr+" PV did not show the command '"+cmd+"' being sent after "+CMD_TIMEOUT_MS+"ms.");
				return;
			}
			System.out.println("Waiting for DetOutStr to verify command sent: value="+Epics.getString(PV_MAR.DetOutStr)+".");
			Util.sleep(CMD_SEND_SLEEP_MS);
		}
	}
	
	// check the mode directly from the detector
	// should only be called from setMode, otherwise a lock should be added around both methods
	public BinMode queryBinMode()
	{
		lastResponse = null;
		long queryTime = System.currentTimeMillis();
		
		sendCmd("get_bin");
		while (null == lastResponse)
		{
			long now = System.currentTimeMillis();
			long waitTime = now-queryTime;
			if (waitTime>CMD_TIMEOUT_MS)
			{
				Log.error("No response from detector for bin mode query after "+CMD_TIMEOUT_MS+"ms");
				return null;
			}
			
			Util.sleep(RESPONSE_SLEEP_MS);
		}
	
		BinMode mode = strToMode(lastResponse);
		if (null != mode) return mode;
		
		Log.error("Unrecognized bin mode response: "+lastResponse);
		return null;
	}
}
