package gov.anl.gmca.collect;

import gov.anl.gmca.collect.ModelCollectRun.*;
import gov.anl.gmca.common.DevEpicsUpdate;
import gov.anl.gmca.common.DevRunList;
import gov.anl.gmca.common.DevRunList.RUN_LABEL_SETTING;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PV_RUN;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.Util;
import gov.aps.jca.dbr.DBR;

import java.util.ArrayList;

public class DevCollectRun extends DevEpicsUpdate implements DevRunList.DevRun
{
	protected ModelCollectRun run;
	protected int ri;
	private Epics.EpicsMonitor nextFrameListener;
	
	protected static final int MAX_SITES = 10000;
	protected static final int SITES_PER_ELEMENT = 8;

	public DevCollectRun(int _runIndex, ModelCollectRun _run)
	{
		run=_run;
		ri=_runIndex;
		
		for (PV_RUN pvName : PV_RUN.values())
		{
			registerPV(pvName.run(ri));
		}
		setUpdatePV(PV_RUN.updated.run(ri));
		
		nextFrameListener = new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				run.setNextFrame(Epics.dbrToInt(value));
			}
		};
		Epics.addMonitor(PV_RUN.nextFrame.run(ri), nextFrameListener);
		run.setNextFrame(Epics.getInteger(PV_RUN.nextFrame.run(ri)));
	}

	// PV connection is not done in the constructor to give the caller time to set up listeners
	// before events are sent
	public void connect()
	{
		registrationComplete();
	}
	
	@Override public void stopEpicsMonitors()
	{
		super.stopEpicsMonitors();
		Epics.removeMonitor(PV_RUN.nextFrame.run(ri), nextFrameListener);
	}
	
	// used by the parent class
	@Override
	protected void sendNotification()
	{
		run.setFramesPerDir(		Epics.getInteger(	PV_RUN.framesPerDir.run(ri)));
		run.setRunLabel(			Epics.getInteger(	PV_RUN.runLabel.run(ri)));
		run.setFileRoot(			Epics.getString(	PV_RUN.fileroot.run(ri)));
		run.setdistance_mm(			Epics.getDouble(	PV_RUN.distance.run(ri)));
		run.setBeamstop_mm(			Epics.getDouble(	PV_RUN.beamstop_mm.run(ri)));
		run.setAttenuation_factors(	Epics.getDouble(	PV_RUN.atten_factor.run(ri)));
		run.setDelta(				Epics.getDouble(	PV_RUN.delta.run(ri)));
		run.setMeshFilter(			Epics.getDouble(	PV_RUN.meshFilter.run(ri)));
		run.setMeshParam(			Epics.getInteger(	PV_RUN.meshParam.run(ri)));
		run.setMeshSpotFinder(		Epics.getInteger(	PV_RUN.meshSpotFinder.run(ri)));
		run.setRasterNumbering(		Epics.getInteger(	PV_RUN.rasterNumbering.run(ri)));
		run.setMeshCombine(			Epics.getBoolean(	PV_RUN.meshCombine.run(ri)));
		run.setExposureTime(		Epics.getDouble(	PV_RUN.exposureTime.run(ri)));
		run.setBeamHeight_um(		Epics.getDouble(	PV_RUN.beamHeight_um.run(ri)));
		run.setBeamWidth_um(		Epics.getDouble(	PV_RUN.beamWidth_um.run(ri)));
		run.setStartAngle(			Epics.getDouble(	PV_RUN.startAngle.run(ri)));
		run.setStartAngle2(			Epics.getDouble(	PV_RUN.startAngle2.run(ri)));
		run.setEndAngle(			Epics.getDouble(	PV_RUN.endAngle.run(ri)));
		run.setStartFrameLabel(		Epics.getInteger(	PV_RUN.startFrameLabel.run(ri)));
		run.setReverseDir(			Epics.getBoolean(	PV_RUN.reverseDir.run(ri)));
		run.setStepSize_mm(			Epics.getDouble(	PV_RUN.stepSize_um.run(ri)) / 1000.0);
		run.setFramesPerSite(		Epics.getInteger(	PV_RUN.framesPerSite.run(ri)));
		run.setFramesPerDir(		Epics.getInteger(	PV_RUN.framesPerDir.run(ri)));
		run.setOverlap(				Epics.getInteger(	PV_RUN.overlap.run(ri)));
		run.setSiteDelta_deg(		Epics.getDouble(	PV_RUN.siteDelta_deg.run(ri)));
		run.setSiteCheckEnable(		Epics.getBoolean(	PV_RUN.siteDeltaEnable.run(ri)));
		run.setSiteSkip(			Epics.getInteger(	PV_RUN.siteSkip.run(ri)));
		run.setSavedFPS(			Epics.getInteger(	PV_RUN.savedFPS.run(ri)));
		run.setSavedOverlap(		Epics.getInteger(	PV_RUN.savedOverlap.run(ri)));
		run.setSavedInverseBeam(	Epics.getBoolean(	PV_RUN.savedInverseBeam.run(ri)));
		run.setSavedEndAngle(		Epics.getDouble(	PV_RUN.savedEndAngle.run(ri)));
		run.setInverseOn(			Epics.getBoolean(	PV_RUN.inverseOn.run(ri)));
		run.setWedgeSize_deg(		Epics.getDouble(	PV_RUN.wedgeSize.run(ri)));
		run.setAxis(				Epics.getString(	PV_RUN.axisMotor.run(ri)).toLowerCase().equals("gonio_phi") ? Axis.GONIO_PHI : Axis.GONIO_OMEGA);
		run.setDirectorySuffix(		Epics.getString(	PV_RUN.directory256.run(ri)));
		run.setStrategyEnabled(		Epics.getBoolean(	PV_RUN.strategyEnabled.run(ri)));
		run.setXDSProcMode(			XDSProcMode.values()[Epics.getInteger(PV_RUN.xdsProc.run(ri))]);
		run.setScanMode(			ScanMode.values()[Epics.getInteger(PV_RUN.scanMode.run(ri))]);
		run.setOptionTab(			OptionTab.values()[Epics.getInteger(PV_RUN.optionTab.run(ri))]);
		run.setSetupStep(			SetupStep.values()[Epics.getInteger(PV_RUN.setupStep.run(ri))]);
		run.setHelicalMode(			Epics.getBoolean(	PV_RUN.helicalMode.run(ri)));
		run.setRasterRun(			Epics.getInteger(	PV_RUN.rasterRun.run(ri)));
		run.setPotentialVectorSites(Epics.getInteger(	PV_RUN.potVecSiteCount.run(ri)));
		run.setImportSiteCount(		Epics.getInteger(	PV_RUN.importSiteCount.run(ri)));
		run.setImportPath(			Epics.getString(	PV_RUN.importPath256.run(ri)));
		
		RunState runState = RunState.INACTIVE;
		String stateString = Epics.getString(PV_RUN.runStatus.run(ri)).toLowerCase();
		if (stateString.equals("collecting")) 	runState = RunState.COLLECTING;
		if (stateString.equals("paused")) 		runState = RunState.PAUSED;
		if (stateString.equals("complete")) 	runState = RunState.COMPLETE;
		run.setRunState(runState);
		
		run.setPt1
		(
			Epics.getBoolean(PV_RUN.pt1_set.run(ri))
			
			?new Point3D
			(
				Epics.getDouble(PV_RUN.pt1_x_mm.run(ri)),
				Epics.getDouble(PV_RUN.pt1_y_mm.run(ri)),
				Epics.getDouble(PV_RUN.pt1_z_mm.run(ri))
			)
			
			:null
		);
				
		run.setPt2
		(
			Epics.getBoolean(PV_RUN.pt2_set.run(ri))
			
			?new Point3D
			(
				Epics.getDouble(PV_RUN.pt2_x_mm.run(ri)),
				Epics.getDouble(PV_RUN.pt2_y_mm.run(ri)),
				Epics.getDouble(PV_RUN.pt2_z_mm.run(ri))
			)
			
			:null
		);

		run.setPt3
		(
			Epics.getBoolean(PV_RUN.pt3_set.run(ri))
			
			?new Point3D
			(
				Epics.getDouble(PV_RUN.pt3_x_mm.run(ri)),
				Epics.getDouble(PV_RUN.pt3_y_mm.run(ri)),
				Epics.getDouble(PV_RUN.pt3_z_mm.run(ri))
			)
			
			:null
		);

		run.setPt4
		(
			Epics.getBoolean(PV_RUN.pt4_set.run(ri))
			
			?new Point3D
			(
				Epics.getDouble(PV_RUN.pt4_x_mm.run(ri)),
				Epics.getDouble(PV_RUN.pt4_y_mm.run(ri)),
				Epics.getDouble(PV_RUN.pt4_z_mm.run(ri))
			)
			
			:null
		);

		ArrayList<Double> energies = new ArrayList<Double>();
		for (int i=1; i<=Epics.getInteger(PV_RUN.numEnergy.run(ri)); i++)
		{
			energies.add(Epics.getDouble(PV_RUN.energy(i).run(ri)));
		}
		run.setEnergies(energies);
		
		ArrayList<Boolean> selectedSites = getSelectedSites(ri);
		run.setSelectedSites(selectedSites);

		System.out.println("Sending run "+ri+" update");
		run.handleChange();
	}
	
	public void setRunStatus(RunState state, boolean update)	{ run.setRunState(state);			Epics.put			(PV_RUN.runStatus.run(ri),			run.getRunState().toString().toLowerCase());	if(update)sendUpdate();	}
	public void setRunLabel(int label, boolean update)			{ run.setRunLabel(label);			Epics.putInteger	(PV_RUN.runLabel.run(ri),			run.getRunLabel());								if(update)sendUpdate();	}
	public void setFileroot(String root, boolean update)		{ run.setFileRoot(root);			Epics.put			(PV_RUN.fileroot.run(ri),			run.getFileRoot());								if(update)sendUpdate();	}
	public void setDistance(double d, boolean update)			{ run.setdistance_mm(d);			Epics.putDouble		(PV_RUN.distance.run(ri),			run.getdistance_mm());							if(update)sendUpdate();	}
	public void setBeamstop_mm(double d, boolean update)		{ run.setBeamstop_mm(d);			Epics.putDouble		(PV_RUN.beamstop_mm.run(ri),		run.getBeamstop_mm());							if(update)sendUpdate();	}
	public void setAttenuation_factors(double d, boolean update){ run.setAttenuation_factors(d);	Epics.putDouble		(PV_RUN.atten_factor.run(ri),		run.getAttenuation_factors());					if(update)sendUpdate();	}
	public void setAxisMotor(Axis axis, boolean update)			{ run.setAxis(axis);				Epics.put			(PV_RUN.axisMotor.run(ri),			run.getAxis().toString().toLowerCase()); 		if(update)sendUpdate(); }
	public void setDelta(double deg, boolean update)			{ run.setDelta(deg);				Epics.putDouble		(PV_RUN.delta.run(ri),				run.getDelta());								if(update)sendUpdate();	}
	public void setMeshFilter(double d, boolean update)			{ run.setMeshFilter(d);				Epics.putDouble		(PV_RUN.meshFilter.run(ri),			run.getMeshFilter());							if(update)sendUpdate();	}
	public void setMeshParam(int i, boolean update)				{ run.setMeshParam(i);				Epics.putInteger	(PV_RUN.meshParam.run(ri),			run.getMeshParam());							if(update)sendUpdate();	}
	public void setMeshSpotFinder(int i, boolean update)		{ run.setMeshSpotFinder(i);			Epics.putInteger	(PV_RUN.meshSpotFinder.run(ri),		run.getMeshSpotFinder());						if(update)sendUpdate();	}
	public void setRasterNumbering(int i, boolean update)		{ run.setRasterNumbering(i);		Epics.putInteger	(PV_RUN.rasterNumbering.run(ri),	run.getRasterNumbering());						if(update)sendUpdate();	}
	public void setMeshCombine(boolean b, boolean update)		{ run.setMeshCombine(b);			Epics.putInteger	(PV_RUN.meshCombine.run(ri),		run.getMeshCombine()?1:0);						if(update)sendUpdate();	}
	public void setExposureTime(double sec, boolean update) 	{ run.setExposureTime(sec);			Epics.putDouble		(PV_RUN.exposureTime.run(ri),		run.getExposureTime());							if(update)sendUpdate();	}
	public void setBeamHeight_um(double height, boolean update) { run.setBeamHeight_um(height);		Epics.putDouble		(PV_RUN.beamHeight_um.run(ri),		run.getBeamHeight_um());						if(update)sendUpdate();	}
	public void setBeamWidth_um(double Width, boolean update) 	{ run.setBeamWidth_um(Width);		Epics.putDouble		(PV_RUN.beamWidth_um.run(ri),		run.getBeamWidth_um());							if(update)sendUpdate();	}
	public void setStartAngle(double d, boolean update)			{ run.setStartAngle(d);				Epics.putDouble		(PV_RUN.startAngle.run(ri),			run.getStartAngle());							if(update)sendUpdate();	}
	public void setStartAngle2(double d, boolean update)		{ run.setStartAngle2(d);			Epics.putDouble		(PV_RUN.startAngle2.run(ri),		run.getStartAngle2());							if(update)sendUpdate();	}
	public void setEndAngle(double d, boolean update)			{ run.setEndAngle(d);				Epics.putDouble		(PV_RUN.endAngle.run(ri),			run.getEndAngle());								if(update)sendUpdate();	}
	public void setStartFrameLabel(int label, boolean update)	{ run.setStartFrameLabel(label);	Epics.putInteger	(PV_RUN.startFrameLabel.run(ri),	run.getStartFrameLabel());						if(update)sendUpdate();	}
	public void setReverseDir(boolean b, boolean update)		{ run.setReverseDir(b);				Epics.putInteger	(PV_RUN.reverseDir.run(ri),			run.getReverseDir()?1:0);						if(update)sendUpdate();	}
	public void setStepSize_mm(double d, boolean update)		{ run.setStepSize_mm(d);			Epics.putDouble		(PV_RUN.stepSize_um.run(ri),		run.getStepSize_mm()*1000.0);					if(update)sendUpdate();	}
	public void setFramesPerSite(int i, boolean update)			{ run.setFramesPerSite(i);			Epics.putInteger	(PV_RUN.framesPerSite.run(ri),		run.getFramesPerSite());						if(update)sendUpdate();	}
	public void setFramesPerDir(int i, boolean update)			{ run.setFramesPerDir(i);			Epics.putInteger	(PV_RUN.framesPerDir.run(ri),		run.getFramesPerDir());							if(update)sendUpdate();	}
	public void setOverlap(int i, boolean update)				{ run.setOverlap(i);				Epics.putInteger	(PV_RUN.overlap.run(ri),			run.getOverlap());								if(update)sendUpdate();	}
	public void setXDSProcMode(XDSProcMode m, boolean update)	{ run.setXDSProcMode(m);			Epics.putInteger	(PV_RUN.xdsProc.run(ri),			run.getXDSProcMode().ordinal());				if(update)sendUpdate(); }
	public void setSiteDelta_deg(double d, boolean update)		{ run.setSiteDelta_deg(d);			Epics.putDouble		(PV_RUN.siteDelta_deg.run(ri),		run.getSiteDelta_deg());						if(update)sendUpdate(); }
	public void setSiteCheckEnable(boolean b, boolean update)	{ run.setSiteCheckEnable(b);		Epics.putInteger	(PV_RUN.siteDeltaEnable.run(ri),	run.getSiteCheckEnable()?1:0);					if(update)sendUpdate(); }
	public void setSiteSkip(int i, boolean update)				{ run.setSiteSkip(i);				Epics.putInteger	(PV_RUN.siteSkip.run(ri),			run.getSiteSkip());								if(update)sendUpdate(); }
	public void setSavedFPS(int i, boolean update)				{ run.setSavedFPS(i);				Epics.putInteger	(PV_RUN.savedFPS.run(ri),			run.getSavedFPS());								if(update)sendUpdate(); }
	public void setSavedOverlap(int i, boolean update)			{ run.setSavedOverlap(i);			Epics.putInteger	(PV_RUN.savedOverlap.run(ri),		run.getSavedOverlap());							if(update)sendUpdate(); }
	public void setSavedInverseBeam(boolean b, boolean update)	{ run.setSavedInverseBeam(b);		Epics.putInteger	(PV_RUN.savedInverseBeam.run(ri),	run.getSavedInverseBeam()?1:0);					if(update)sendUpdate(); }
	public void setSavedEndAngle(double d, boolean update)		{ run.setSavedEndAngle(d);			Epics.putDouble		(PV_RUN.savedEndAngle.run(ri),		run.getSavedEndAngle());						if(update)sendUpdate();	}
	public void setInverseOn(boolean b, boolean update)			{ run.setInverseOn(b);				Epics.putInteger	(PV_RUN.inverseOn.run(ri),			run.getInverseOn()?1:0);						if(update)sendUpdate();	}
	public void setWedgeSize(double d, boolean update)			{ run.setWedgeSize_deg(d);			Epics.putDouble		(PV_RUN.wedgeSize.run(ri),			run.getWedgeSize_deg());						if(update)sendUpdate();	}
	public void setDirectorySuffix(String dir, boolean update)  { run.setDirectorySuffix(dir);		Epics.put         	(PV_RUN.directory256.run(ri),     	run.getDirectorySuffix());                      if(update)sendUpdate(); }
	public void setStrategyEnabled(boolean b, boolean update)	{ run.setStrategyEnabled(b);		Epics.putInteger	(PV_RUN.strategyEnabled.run(ri),	run.getStrategyEnabled()?1:0);					if(update)sendUpdate(); }
	public void setScanMode(ScanMode m, boolean update)			{ run.setScanMode(m);				Epics.putInteger	(PV_RUN.scanMode.run(ri),			run.getScanMode().ordinal()); 					if(update)sendUpdate(); }
	public void setOptionTab(OptionTab ot, boolean update)		{ run.setOptionTab(ot);				Epics.putInteger	(PV_RUN.optionTab.run(ri),			run.getOptionTab().ordinal()); 					if(update)sendUpdate(); }
	public void setSetupStep(SetupStep ss, boolean update)		{ run.setSetupStep(ss);				Epics.putInteger	(PV_RUN.setupStep.run(ri),			run.getSetupStep().ordinal()); 					if(update)sendUpdate(); }
	public void setHelicalMode(boolean hm, boolean update)		{ run.setHelicalMode(hm);			Epics.putInteger	(PV_RUN.helicalMode.run(ri),		run.getHelicalMode()?1:0);						if(update)sendUpdate();	}
	public void setRasterRun(int i, boolean update)				{ run.setRasterRun(i);				Epics.putInteger	(PV_RUN.rasterRun.run(ri),			run.getRasterRun());							if(update)sendUpdate();	}
	public void setPotentialVectorSites(int i, boolean update)	{ run.setPotentialVectorSites(i);	Epics.putInteger	(PV_RUN.potVecSiteCount.run(ri),	run.getPotentialVectorSites());					if(update)sendUpdate(); }
	public void setNextFrame(int frame)							{ run.setNextFrame(frame);			Epics.putInteger	(PV_RUN.nextFrame.run(ri),			run.getNextFrame()); }
	public void setImportPath(String s, boolean update)			{ run.setImportPath(s);				Epics.put         	(PV_RUN.importPath256.run(ri),     	run.getImportPath());                      		if(update)sendUpdate(); }
	public void setImportSiteCount(int i, boolean update)		{ run.setImportSiteCount(i);		Epics.putInteger   	(PV_RUN.importSiteCount.run(ri),    run.getImportSiteCount());                 		if(update)sendUpdate(); }
	
	public void setEnergies(ArrayList<Double> energies, boolean update)
	{
		if (null == energies || energies.size()<1 || energies.size()>5)
		{
			Log.error("Invalid energies passed to DevCollectRun.setEnergies()");
			return;
		}

		run.setEnergies(energies);
		
		ArrayList<Double> runEnergies = run.getEnergies();
		
		Epics.putInteger(PV_RUN.numEnergy.run(ri),runEnergies.size());
		for (int i=0; i<runEnergies.size(); i++)
		{
			Epics.putDouble(PV_RUN.energy(i+1).run(ri), runEnergies.get(i));
		}
		for (int i=runEnergies.size(); i<5; i++)
		{
			Epics.putDouble(PV_RUN.energy(i+1).run(ri), Double.NaN);
		}
		
		if (update) sendUpdate();
	}
	
	public void setPt1(ModelCollectRun.Point3D point, boolean update)
	{
		run.setPt1(point);
		
		Point3D runPoint = run.getPt1();
		
		if (null == runPoint)
		{
    		Epics.putDouble(PV_RUN.pt1_x_mm.run(ri), 0);
    		Epics.putDouble(PV_RUN.pt1_y_mm.run(ri), 0);
    		Epics.putDouble(PV_RUN.pt1_z_mm.run(ri), 0);
			Epics.putInteger(PV_RUN.pt1_set.run(ri), 0);
			if (update) sendUpdate();
		}
		else
		{
    		Epics.putDouble(PV_RUN.pt1_x_mm.run(ri), runPoint.x);
    		Epics.putDouble(PV_RUN.pt1_y_mm.run(ri), runPoint.y);
    		Epics.putDouble(PV_RUN.pt1_z_mm.run(ri), runPoint.z);
    		Epics.putInteger(PV_RUN.pt1_set.run(ri), 1);
    		if (update) sendUpdate();
		}
	}
	
	public void setPt2(ModelCollectRun.Point3D point, boolean update)
	{
		run.setPt2(point);
		
		
		Point3D runPoint = run.getPt2();
		
		if (null == runPoint)
		{
			Epics.putDouble(PV_RUN.pt2_x_mm.run(ri), 0);
			Epics.putDouble(PV_RUN.pt2_y_mm.run(ri), 0);
			Epics.putDouble(PV_RUN.pt2_z_mm.run(ri), 0);
			Epics.putInteger(PV_RUN.pt2_set.run(ri), 0);
			if (update) sendUpdate();
		}
		else
		{
			Epics.putDouble(PV_RUN.pt2_x_mm.run(ri), runPoint.x);
			Epics.putDouble(PV_RUN.pt2_y_mm.run(ri), runPoint.y);
			Epics.putDouble(PV_RUN.pt2_z_mm.run(ri), runPoint.z);
			Epics.putInteger(PV_RUN.pt2_set.run(ri), 1);
			if (update) sendUpdate();
		}
	}
	
	public void setPt3(ModelCollectRun.Point3D point, boolean update)
	{
		run.setPt3(point);
		
		Point3D runPoint = run.getPt3();
		
		if (null == runPoint)
		{
			Epics.putDouble(PV_RUN.pt3_x_mm.run(ri), 0);
			Epics.putDouble(PV_RUN.pt3_y_mm.run(ri), 0);
			Epics.putDouble(PV_RUN.pt3_z_mm.run(ri), 0);
			Epics.putInteger(PV_RUN.pt3_set.run(ri), 0);
			if (update) sendUpdate();
		}
		else
		{
			Epics.putDouble(PV_RUN.pt3_x_mm.run(ri), runPoint.x);
			Epics.putDouble(PV_RUN.pt3_y_mm.run(ri), runPoint.y);
			Epics.putDouble(PV_RUN.pt3_z_mm.run(ri), runPoint.z);
			Epics.putInteger(PV_RUN.pt3_set.run(ri), 1);
			if (update) sendUpdate();
		}
	}
	
	public void setPt4(ModelCollectRun.Point3D point, boolean update)
	{
		run.setPt4(point);
		
		Point3D runPoint = run.getPt4();
		
		if (null == runPoint)
		{
			Epics.putDouble(PV_RUN.pt4_x_mm.run(ri), 0);
			Epics.putDouble(PV_RUN.pt4_y_mm.run(ri), 0);
			Epics.putDouble(PV_RUN.pt4_z_mm.run(ri), 0);
			Epics.putInteger(PV_RUN.pt4_set.run(ri), 0);
			if (update) sendUpdate();
		}
		else
		{
			Epics.putDouble(PV_RUN.pt4_x_mm.run(ri), runPoint.x);
			Epics.putDouble(PV_RUN.pt4_y_mm.run(ri), runPoint.y);
			Epics.putDouble(PV_RUN.pt4_z_mm.run(ri), runPoint.z);
			Epics.putInteger(PV_RUN.pt4_set.run(ri), 1);
			if (update) sendUpdate();
		}
	}
	
	// addOneToRunLabel is used when creating a new tab; all settings are copied from the
	// last active tab, but the run label is incremented by one
	// resetState is true on star tab click
	@Override
	public void copyRunContents(int destRunIndex, RUN_LABEL_SETTING runLabelSetting, boolean resetState, boolean doUpdate)
	{
		System.out.println("Copying run index "+ri+" contents to run index "+destRunIndex);
		
		// determine the destination run's new label
		int newRunLabel;
		switch (runLabelSetting)
		{
		case SAME: 				newRunLabel = Epics.getInteger(PV_RUN.runLabel.run(ri)); 							break;
		case INCREMENT_SRC:		newRunLabel = Epics.getInteger(PV_RUN.runLabel.run(ri))+1;							break;
		case INCREMENT_LAST:	newRunLabel = Epics.getInteger(PV_RUN.runLabel.run(run.runList.getRunCount()))+1;	break;
		default:				throw new IllegalArgumentException("Invalid run label setting: "+runLabelSetting);
		}
		
		// JCA doesn't have a generic ch.put() function, so converting to and from a String
		// allows a simple loop here.  The alternative is to have a line for each PV.
		for (PV_RUN pvName : PV_RUN.values())
		{
			if (resetState)
			{
				switch (pvName)
				{
				case fileroot:			Epics.put		(PV_RUN.fileroot.run(destRunIndex), 		Epics.getString(PV_RUN.fileroot.run(0))); break;
				case directory256:		Epics.put		(PV_RUN.directory256.run(destRunIndex), 	Epics.getString(PV_RUN.directory256.run(0))); break;
				case runStatus:			Epics.put		(PV_RUN.runStatus.run(destRunIndex),		"inactive"); 	break;
				case nextFrame: 		Epics.putInteger(PV_RUN.nextFrame.run(destRunIndex), 		0); 			break;
				case pt1_x_mm:			Epics.putDouble (PV_RUN.pt1_x_mm.run(destRunIndex), 		0.0); 			break;
				case pt1_y_mm:			Epics.putDouble (PV_RUN.pt1_y_mm.run(destRunIndex), 		0.0); 			break;
				case pt1_z_mm:			Epics.putDouble (PV_RUN.pt1_z_mm.run(destRunIndex), 		0.0); 			break;
				case pt1_set:			Epics.putInteger(PV_RUN.pt1_set.run(destRunIndex), 		0); 			break;
				case pt2_x_mm: 			Epics.putDouble (PV_RUN.pt2_x_mm.run(destRunIndex), 		0.0); 			break;
				case pt2_y_mm:			Epics.putDouble (PV_RUN.pt2_y_mm.run(destRunIndex), 		0.0); 			break;
				case pt2_z_mm:			Epics.putDouble (PV_RUN.pt2_z_mm.run(destRunIndex), 		0.0); 			break;
				case pt2_set:			Epics.putInteger(PV_RUN.pt2_set.run(destRunIndex), 			0); 			break;
				case pt3_x_mm: 			Epics.putDouble (PV_RUN.pt3_x_mm.run(destRunIndex), 		0.0); 			break;
				case pt3_y_mm:			Epics.putDouble (PV_RUN.pt3_y_mm.run(destRunIndex), 		0.0); 			break;
				case pt3_z_mm:			Epics.putDouble (PV_RUN.pt3_z_mm.run(destRunIndex), 		0.0); 			break;
				case pt3_set:			Epics.putInteger(PV_RUN.pt3_set.run(destRunIndex), 			0); 			break;
				case pt4_x_mm: 			Epics.putDouble (PV_RUN.pt4_x_mm.run(destRunIndex), 		0.0); 			break;
				case pt4_y_mm:			Epics.putDouble (PV_RUN.pt4_y_mm.run(destRunIndex), 		0.0); 			break;
				case pt4_z_mm:			Epics.putDouble (PV_RUN.pt4_z_mm.run(destRunIndex), 		0.0); 			break;
				case pt4_set:			Epics.putInteger(PV_RUN.pt4_set.run(destRunIndex), 			0); 			break;
				case strategyEnabled:	Epics.putInteger(PV_RUN.strategyEnabled.run(destRunIndex), 	0); 			break;
				case scanMode:			Epics.putInteger(PV_RUN.scanMode.run(destRunIndex), 		0); 			break;
				case optionTab:			Epics.putInteger(PV_RUN.optionTab.run(destRunIndex),		0); 			break;
				case startAngle2:		Epics.putInteger(PV_RUN.startAngle2.run(destRunIndex), 		0); 			break;
				case runLabel:			Epics.putInteger(PV_RUN.runLabel.run(destRunIndex), 		newRunLabel);	break;
				case siteDelta_deg:		Epics.putDouble (PV_RUN.siteDelta_deg.run(destRunIndex),	0.0);			break;
				case siteDeltaEnable:	Epics.putInteger(PV_RUN.siteDeltaEnable.run(destRunIndex),	0);				break;
				case siteSkip:			Epics.putInteger(PV_RUN.siteSkip.run(destRunIndex),			0);				break;
				case savedFPS:			Epics.putInteger(PV_RUN.savedFPS.run(destRunIndex),			0);				break;
				case savedOverlap:		Epics.putInteger(PV_RUN.savedOverlap.run(destRunIndex),		0);				break;
				case savedEndAngle:		Epics.putDouble (PV_RUN.savedEndAngle.run(destRunIndex),	0.0);			break;
				case savedInverseBeam:	Epics.putInteger(PV_RUN.savedInverseBeam.run(destRunIndex),	0);				break;
				case helicalMode:		Epics.putInteger(PV_RUN.helicalMode.run(destRunIndex),		0);				break;
				case rasterRun:			Epics.putInteger(PV_RUN.rasterRun.run(destRunIndex),		0);				break;
				case potVecSiteCount:	Epics.putInteger(PV_RUN.potVecSiteCount.run(destRunIndex),	0);				break;
				case importSiteCount:	Epics.putInteger(PV_RUN.importSiteCount.run(destRunIndex),	0);				break;
				case importPath256:		Epics.put		(PV_RUN.importPath256.run(destRunIndex),	"");			break;
				case stepSize_um:		Epics.putDouble	(PV_RUN.stepSize_um.run(destRunIndex),		10.0);			break;
				case xdsProc:			Epics.putInteger(PV_RUN.xdsProc.run(destRunIndex),			Epics.getInteger(PV_RUN.inverseOn.run(ri))==1 ? XDSProcMode.ANOMALOUS.ordinal() : XDSProcMode.NATIVE.ordinal()); break;
				case selectedSites:		Epics.putIntegerArray(PV_RUN.selectedSites.run(destRunIndex),new int[1000]);break;
				case meshCombine:		Epics.putInteger(PV_RUN.meshCombine.run(destRunIndex),		1);				break;
				case updated:			break;
				case dbUpdated:			break;
				default:
				{
	    			String srcValue = Epics.getString(pvName.run(ri), 5);
	    			Epics.put(pvName.run(destRunIndex),srcValue);
				}}
			}
			else
			{
				switch (pvName)
				{
				case runLabel:			Epics.putInteger(PV_RUN.runLabel.run(destRunIndex), newRunLabel); break;
				case selectedSites:
					int[] i = Epics.getIntegerArray(pvName.run(ri));
					Epics.putIntegerArray(PV_RUN.selectedSites.run(destRunIndex), i);
					break;
				default:
        			String srcValue = Epics.getString(pvName.run(ri), 5);
        			Epics.put(pvName.run(destRunIndex),srcValue);
				}
			}
		}
		
		if (doUpdate)
		{
			DevCollectRunList.getIndex(destRunIndex).sendUpdate();
		}
	}
	
	// direct static methods for pre-setting a run before creation
	public static void setFramesPerSite(int newIndex, int fps)					{ Epics.putInteger	(PV_RUN.framesPerSite.run(newIndex),	fps);				}
	public static void setFramesPerDir(int newIndex, int fpd)					{ Epics.putInteger	(PV_RUN.framesPerDir.run(newIndex),		fpd);				}
	public static void setStartFrameLabel(int newIndex, int startFrameLabel)	{ Epics.putInteger	(PV_RUN.startFrameLabel.run(newIndex), 	startFrameLabel); 	}
	public static void setStartAngle(int newIndex, double oscStart)				{ Epics.putDouble	(PV_RUN.startAngle.run(newIndex), 		oscStart); 			} 
	public static void setEndAngle(int newIndex, double oscEnd)					{ Epics.putDouble	(PV_RUN.endAngle.run(newIndex),			oscEnd); 			} 
	public static void setDelta(int newIndex, double oscDelta)					{ Epics.putDouble	(PV_RUN.delta.run(newIndex),			oscDelta); 			}
	public static void setMeshFilter(int newIndex, double filter)				{ Epics.putDouble	(PV_RUN.meshFilter.run(newIndex),		filter); 			}
	public static void setMeshCombine(int newIndex, boolean b)					{ Epics.putInteger	(PV_RUN.meshCombine.run(newIndex),		b?1:0); 			}
	public static void setDistance(int newIndex, double detDist)				{ Epics.putDouble	(PV_RUN.distance.run(newIndex),			detDist);			}
	public static void setInverseOn(int newIndex, boolean inverse)				{ Epics.putInteger	(PV_RUN.inverseOn.run(newIndex),		inverse?1:0);		}
	public static void setXDSProcMode(int newIndex, int xdsMode)				{ Epics.putInteger	(PV_RUN.xdsProc.run(newIndex),			xdsMode);			}
	public static void setWedgeSize(int newIndex, double wedge)					{ Epics.putDouble	(PV_RUN.wedgeSize.run(newIndex),		wedge);				}
	public static void setRunState(int newIndex, RunState rs)					{ Epics.putInteger	(PV_RUN.runStatus.run(newIndex),		rs.ordinal());		}
	public static void setNextFrame(int newIndex, int nextFrame)				{ Epics.putInteger	(PV_RUN.nextFrame.run(newIndex),		nextFrame);			}
	public static void setScanMode(int newIndex, ScanMode m)					{ Epics.putInteger	(PV_RUN.scanMode.run(newIndex),			m.ordinal()); 		}
	public static void setSetupStep(int newIndex, SetupStep ss)					{ Epics.putInteger	(PV_RUN.setupStep.run(newIndex),		ss.ordinal());		}
	public static void setRunLabel(int newIndex, int runLabel)					{ Epics.putInteger	(PV_RUN.runLabel.run(newIndex),			runLabel);			}
	public static void setFileroot(int newIndex, String fileRoot)				{ Epics.put	(PV_RUN.fileroot.run(newIndex),					fileRoot);			}
	public static void setDirectorySuffix(int newIndex, String dirSuffix)		{ Epics.put	(PV_RUN.directory256.run(newIndex),				dirSuffix);			}
	
	@Override public void sendUpdate()
	{
		run.handleChange();
		super.sendUpdate();
	}

	public static int getRunLabel(int runIndex)
	{
		return Epics.getInteger(PV_RUN.runLabel.run(runIndex));
	}

	public static void sendUpdate(int runIndex)
	{
		Epics.putInteger(PV_RUN.updated.run(runIndex), 0);
		Epics.putInteger(PV_RUN.updated.run(runIndex), 1);
	}
	
	/////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////// selected sites methods ///////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////
	
	// convenience method
	// 1-based site number
	public void setSiteSelected(int siteNumber, boolean selected, boolean updated)
	{
		ArrayList<Boolean> selectedSites = run.getSelectedSites();
		selectedSites.set(siteNumber-1, selected);
		setSelectedSites(selectedSites, updated);
	}
	
	// convenience method
	public void unselectSiteGroup(int startSite, int endSite, boolean update)
	{
		ArrayList<Boolean> selectedSites = run.getSelectedSites();
		for (int i=startSite-1; i<=endSite-1; i++)
		{
			selectedSites.set(i, false);
		}
		setSelectedSites(selectedSites, update);
	}
	
	// convenience method
	public void clearSelectedSites(boolean update)
	{
		setPotentialVectorSites(0, false);
		setImportSiteCount(0, false);
		setSelectedSites(new ArrayList<Boolean>(), update);
	}
	
	// public object member method - this is the one commonly used
	public void setSelectedSites(ArrayList<Boolean> selectedSites, boolean update)
	{
		if (null == selectedSites)
		{
			throw new IllegalArgumentException();
		}
		run.setSelectedSites(selectedSites);

		setSelectedSites(run.getRunIndex(), selectedSites);
		
		if (update) sendUpdate();
	}
	
	// static method - used by the object member method and in cases where the object
	// is not available
	public static void setSelectedSites(int newIndex, ArrayList<Boolean> selectedSites)
	{
		if (null == selectedSites)
		{
			throw new IllegalArgumentException();
		}
		
		int[] selectedSitesArray = new int[MAX_SITES/SITES_PER_ELEMENT];
		int booleanArrayIndex=0;
		for (int selectedSiteElementIndex=0; selectedSiteElementIndex<selectedSitesArray.length; selectedSiteElementIndex++)
		{
			if (booleanArrayIndex>selectedSites.size()-1) break;
			
			int pvElementValue = 0;
			for (int bitIndex=SITES_PER_ELEMENT-1; bitIndex>=0; bitIndex--)
			{
				if (booleanArrayIndex>selectedSites.size()-1) break;
				
				if (selectedSites.get(booleanArrayIndex++))
				{
					pvElementValue |= 1<<bitIndex;
				}
			}
			selectedSitesArray[selectedSiteElementIndex] = pvElementValue;
		}
		
		String firstElem="";
		for (int i=7;i>=0;i--)
		{
			int pvVal = selectedSitesArray[0];
			int testBit = 1<<i;
			boolean bitOn = (pvVal & testBit) != 0;
			firstElem += bitOn ? "1" : "0";
		}
		System.out.println("Writing selected sites, first element = "+firstElem);
		
		Epics.putIntegerArray(PV_RUN.selectedSites.run(newIndex), selectedSitesArray);
	}
	
	// static get method
	public static ArrayList<Boolean> getSelectedSites(int runIndex)
	{
		ArrayList<Boolean> ret = new ArrayList<Boolean>();
		int[] selectedSitesArray = Epics.getIntegerArray(PV_RUN.selectedSites.run(runIndex));
		
		for (int pvElementValue : selectedSitesArray)
		{
			for (int bit=0; bit<SITES_PER_ELEMENT; bit++)
			{
				boolean bitValue = (pvElementValue & (1<<(SITES_PER_ELEMENT-1-bit))) != 0;
				ret.add(bitValue);
			}
		}
		
		String firstElem = "";
		for (int i=0; i<8;i++) firstElem += ret.get(i)?"1":"0";
		System.out.println("Reading selected sites, first element = "+firstElem);
		return ret;
	}
}