package gov.anl.gmca.collect;

import gov.anl.gmca.collect.ModelCollectRun.Axis;
import gov.anl.gmca.collect.ModelCollectRun.Point3D;
import gov.anl.gmca.collect.RunCalculator.Frame;
import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.common.DevDataDir;
import gov.anl.gmca.common.EpicsDevice;
import gov.anl.gmca.common.Redis;
import gov.anl.gmca.common.client_server.CounterState;
import gov.anl.gmca.common.ui.NBMessageBox;
import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.gui.Main;
import gov.anl.gmca.gui.state.SWServer;
import gov.anl.gmca.hutch.DetectorConfig;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.hutch.DevShutter;
import gov.anl.gmca.hutch.DevShutter.Listener;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.*;
import gov.anl.gmca.raster.ModelRasterCell;
import gov.anl.gmca.raster.ModelRasterRun;
import gov.anl.gmca.raster.ModelRasterState;
import gov.anl.gmca.raster.calc.FramePartitioner.FramePartition;
import gov.anl.gmca.raster.calc.RasterGeometry;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.users.ModelUserTable.ClientListener;
import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.Util;
import gov.aps.jca.dbr.DBR;
import org.eclipse.swt.SWT;
import redis.clients.jedis.Jedis;

import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

// The interface between DevFPE1/2 and users of DevFPE
// The actual protocol between JBluIce and FPE is not here, but in DevFPE1/2
//
// Example usage:
//
// DevFPE.FrameList frameList	= DevFPE.get().newFrameList();
// frameList.delta 				= 1.0;
// frameList.energy 			= DevPmacMotor.get("mo_En_E").getPosition();
// frameList.exposureTime 		= 1.0;
// frameList.startAngle 		= DevPmacMotor.get("gonio_omega").getPosition();
// frameList.firstFrameNumber	= 1;
// frameList.frameCount			= 10;
// frameList.runNumber 			= 0;
// frameList.filename 			= "test_1.0001";
// frameList.directory 			= "/mnt/share1/user0/";
// frameList.axis 				= Axis.GONIO_OMEGA;
// fpeFrameResult fpeResult 	= frameList.start().waitFor();
// if (!fpeResult()) 			{ Log.error("Frame failed."); }
//
public abstract class DevFPE extends EpicsDevice
{
	////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////// interface //////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////

	public static Pattern dirPattern = Pattern.compile("/mnt/beegfs/(\\w+)/(23ID\\w+)/(.*)");
	protected volatile boolean _requestStop = false;
	protected final static int PAUSE_SPACING_MS = 5000;
	protected long lastModeCmdSent=0;
	protected int lastFrameCountSent=0;
	protected int currentFirstFrameIndex=0, currentLastFrameIndex=0;
	public int[] getCurrentFrameRange() { return new int[]{currentFirstFrameIndex,currentLastFrameIndex}; }
	public ListenerList<DevFPE> lastFileListener, imageViewerListener, commandListener, etaListener, frameListener;
	static DevFPE inst = null;
	public static DevFPE get() { return inst; }

	public FrameList currentFrameList = null;
	protected PVName fpeStatePV, connectPV, frameNumberGetPV, frameNumberPutPV;
	public ListenerList<DevFPE> llFPEState; // TODO: add trigger for DevFPE1
	
	public abstract void abort();
	public abstract boolean pause(boolean enforcePauseSpacing);
	public abstract void collectPartition(FrameList frameList);
	public abstract boolean isIdle();
	public abstract boolean isCacheIdle(); // have all PVs been processed up to the idle state?
	
	public static enum FPEMode
	{
		STANDARD,
		SHUTTERED_VECTOR,
		SHUTTERED_RASTER,
		SHUTTERLESS_ARM_DISARM,
		SHUTTERLESS_ARM_NODISARM,
		SHUTTERLESS_NOARM_NODISARM,
		SHUTTERLESS_NOARM_DISARM
	}

	public static class XYZMotion
	{
		public double startX, endX, startY, endY, startZ, endZ;
		public XYZMotion(XYZMotion xyzm)
		{
			startX=xyzm.startX; endX=xyzm.endX; 
			startY=xyzm.startY; endX=xyzm.endY;
			startZ=xyzm.startZ; endX=xyzm.endZ; 
		}
		public XYZMotion(Point3D start, Point3D end)
		{
			startX=start.x; startY=start.y; startZ=start.z;
			endX=end.x; endY=end.y; endZ=end.z;
		}
		public XYZMotion(double sx, double ex, double sy, double ey, double sz, double ez)
		{
			startX=sx; startY=sy; startZ=sz;
			endX=ex; endY=ey; endZ=ez;
		}
		public void setStartpoint(Point3D start)
		{
			startX=start.x; startY=start.y; startZ=start.z;
		}
		public void setEndpoint(Point3D end)
		{
			endX=end.x; endY=end.y; endZ=end.z;
		}
		public static XYZMotion current()
		{
			Point3D currentPos = RasterGeometry.getCurrentMotorPositions().to3D();
			return new XYZMotion(currentPos,currentPos);
		}
	}

	public class FrameList implements Runnable, Listener
	{
		public FrameListResult getResult() { return result; }
		protected FrameListResult result;

		public String toKey = "";
		
		protected Thread moveThread = null;
		
		protected boolean shutterOpened, shutterClosed;
		
		public double delta, energy, exposureTime, startAngle, nextAngle;
		public Double endAngle = null; // temporary override for end angle
		public DevFPE.XYZMotion shutteredMotion, shutterlessMotion;
		public int firstFrameIndex; // e.g. for a 5x5 raster, this index will be 0, then 5, then 10..
		public int firstFrameLabel; // the number in the first filename, e.g. 3 in A1_00003.cbf
		public int runNumber;
		public int frameCount = 1;
		public boolean disableEigerPerImageTrigger=false;
		public String filename, directory, prefix;
		public Axis axis;
		public FramePartition partition;
		public DevFPE.FPEMode fpeMode = DevFPE.FPEMode.STANDARD;
		public boolean threadStarted = false;
		
		public FrameList start()
		{
			try
			{
				_requestStop = false;
				shutterOpened = false;
				shutterClosed = false;
				
				DevShutter.get("shutter").addListener(this);
				
				moveThread = new Thread(this,"FPE Frame");
				moveThread.start();
			}
			catch (Exception e) { e.printStackTrace(); }

			threadStarted=true;
			
			return this;
		}
		
		public void requestStop()
		{
			_requestStop = true;
		}
		
		public boolean isRunning()
		{
			return (null != moveThread && moveThread.isAlive());
		}

		public void waitForShutterClose()
		{
			while
			(
				null != moveThread &&
				moveThread.isAlive() &&
				!shutterClosed
			)
			{
				Util.sleep(10);
			}
		}
		
		public FrameListResult waitFor()
		{
			while (null != moveThread && moveThread.isAlive()) Util.sleep(10);
			return result;
		}

		@Override public void run()
		{
			long startMs=System.currentTimeMillis();
			collectPartition(this);
			Log.note("collectPartition took "+(System.currentTimeMillis()-startMs));
		}

		public void disconnectListeners()
		{
			DevShutter.get("shutter").removeListener(this);
		}

		@Override public void handleOpenChange(DevShutter shutter, boolean open)
		{
			if (open) 					shutterOpened = true;
			if (shutterOpened && !open) shutterClosed = true;
		}

		private void fpeCommandError(String msg)
		{
			if (null==msg) {
				msg = "There was an error sending the FPE collect command.  Please try\n" +
						"restarting JBluIce, or deleting and recreating the collect or\n" +
						"raster run.";
			}
			if (DBConfig.serverMode)
				gov.anl.gmca.server.messaging.Popup.syncPopup("note",msg,null,0);
			else
				Util.popupError("Error", msg);
		}
		
		public boolean sendToFPE()
		{
			currentFrameList = this;
			try (Jedis jedis = Redis.get()) {
				toKey = jedis.hget(Redis.server_fpe, "copy_frame_to_key");
			}

			// first determine xyzMotion and abort if it's null
			DevFPE.XYZMotion xyzMotion = null;
			try
			{
    			switch (fpeMode)
    			{
    			case STANDARD:
    				xyzMotion = shutterlessMotion;
    				break;
    			case SHUTTERED_VECTOR:
    			case SHUTTERED_RASTER:
    				xyzMotion = shutteredMotion;
    				break;
    			case SHUTTERLESS_ARM_DISARM:
    			case SHUTTERLESS_ARM_NODISARM:
    			case SHUTTERLESS_NOARM_NODISARM:
    			case SHUTTERLESS_NOARM_DISARM:
    				xyzMotion = shutterlessMotion;
    				break;
    			default:
    				throw new IllegalArgumentException("Unsupported FPE mode: "+fpeMode.toString());
    			}
			}
			catch (Exception e)
			{
				e.printStackTrace();
				fpeCommandError(null);
				return false;
			}
			if (null==xyzMotion)
			{
				fpeCommandError(null);
				return false;
			}
			
			/////////////////////////////////////////////////////////////////////////////
			/////////////////////////////// Eiger setup /////////////////////////////////
			/////////////////////////////////////////////////////////////////////////////

			if (DetectorConfig.isEiger()) {
				boolean disableAutosumFlag = DBConfig.getBoolean("eiger_disable_autosum_over5Hz");
				boolean disableSinglearmFlag = DBConfig.getBoolean("eiger_disable_singlearm_over5Hz");
				boolean isOver5Hz = exposureTime <= 0.19;
				boolean modeIncludesArm = fpeMode == FPEMode.SHUTTERLESS_NOARM_DISARM || fpeMode != FPEMode.SHUTTERLESS_NOARM_NODISARM;
				boolean modeIsSinglearm = fpeMode.ordinal()>3;

				if (disableSinglearmFlag && isOver5Hz && modeIsSinglearm) {
					fpeMode = FPEMode.SHUTTERLESS_ARM_DISARM; // multi-arm
				}

				if (modeIncludesArm) {
					boolean disableAutoSummation = disableAutosumFlag && isOver5Hz;
					String errMsg = DevEiger.setupForCollection(disableEigerPerImageTrigger, disableAutoSummation);
					if (null != errMsg) {
						fpeCommandError(errMsg);
					}
				}
			}

			/////////////////////////////////////////////////////////////////////////////
			///////////////////////////// send parameters ///////////////////////////////
			/////////////////////////////////////////////////////////////////////////////

			if (DetectorConfig.isEiger()) {
				boolean continuingSeries =
						fpeMode==FPEMode.SHUTTERLESS_NOARM_DISARM||
						fpeMode==FPEMode.SHUTTERLESS_NOARM_NODISARM;

				if (!DevStreamWriter.setConfig
						(
							directory,
							prefix,
							firstFrameLabel,
							firstFrameIndex,
							continuingSeries,
							startAngle,
							delta,
							energy
						)) {
					return false;
				}
			}

			Epics.putDouble(	PV_FPE.delta,			delta);
			Epics.putDouble(	PV_FPE.energy, 			energy);
			Epics.putDouble(	PV_FPE.exposureTime, 	exposureTime);
			Epics.putDouble(	PV_FPE.startAngle, 		startAngle);
			Epics.putDouble(	PV_FPE.nextAngle, 		nextAngle);
			if (null!=frameNumberPutPV) Epics.putInteger(frameNumberPutPV, firstFrameIndex);
			Epics.putInteger(	PV_FPE.modeIndex, 		0);
			Epics.putInteger(	PV_FPE.runNumber, 		runNumber);
			Epics.put(			PV_FPE.filename, 		filename);
			Epics.put(			PV_SW.filename,			filename);
			Epics.put(			PV_FPE.axisMotor, 		axis==Axis.GONIO_OMEGA?"omega":"phi");
			String dirToWrite = directory;
			if (DetectorConfig.isDectris()) {
				Matcher m = dirPattern.matcher(dirToWrite);
				if (m.matches()) {
					dirToWrite = String.format("./%s/%s",m.group(2),m.group(3));
				}
			}
			Epics.put(			PV_FPE.directory256, 	dirToWrite);

			/////////////////////////////////////////////////////////////////////////////
			//////////////////////////// send mode256 command ///////////////////////////
			/////////////////////////////////////////////////////////////////////////////

			int fpeModeInt = fpeMode.ordinal();
			if (DetectorConfig.isPilatus() && fpeModeInt>3) fpeModeInt=3;
			Epics.put(PV_FPE.mode256, Util.sprintf
			(
				//  (m=mode, fr=frame, st=start)
				//   m  fr stA  endA stX  endX stY  endY stZ  endZ
					"%d %d %.4f %.4f %.4f %.4f %.4f %.4f %.4f %.4f %d",
				fpeModeInt,
				frameCount,
				startAngle,
				endAngle,
				xyzMotion.startX, xyzMotion.endX,
				xyzMotion.startY, xyzMotion.endY,
				xyzMotion.startZ, xyzMotion.endZ,
				1
			));
			
			// save frame info
			currentFirstFrameIndex = firstFrameIndex;
			currentLastFrameIndex = firstFrameIndex +frameCount-1;
			lastFrameCountSent = frameCount;
			
			return true;
		}

		@Override public void handleStateChange(DevShutter shutter, State state) {}
	}
	
	public FrameList newFrameList() { return new FrameList(); }
	
	public enum FrameListResult
	{
		CREATED, ABORTED, FAILED, SUCCEEDED, SUCCEEDED_WITH_WARNING;
		public boolean isOK() { return (this == SUCCEEDED || this == SUCCEEDED_WITH_WARNING); }
	};

	////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////// ETA /////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////
	
	public void updateETA()
	{
		setETA(ETA.calcActiveETA());
	}
	
	////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////// direct PV access //////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////
	
	public static class ViewerFileInfo
	{
		public ViewerFileInfo(String filename, String directory, int i) { this.filename=filename; this.directory=directory; imageIndex=i; }
		public String getFullPath() { return directory+"/"+filename; }
		public String filename;
		public String directory;
		public int imageIndex;
	}
	
	// Non-HDF5 file: "imagename.cbf"
	// HDF5 file: "imagename.h5:3" for the fourth image in that file
	public ViewerFileInfo getImageViewerFile()
	{
		String filename = Epics.getString(PV_FPE.viewFile256);
		String dir = Epics.getString(PV_FPE.viewDir256);
		int colonIndex = filename.indexOf(':');
		if (colonIndex<0) return new ViewerFileInfo(filename,dir,0);
		else return new ViewerFileInfo
		(
			filename.substring(0, colonIndex),
			dir,
			Integer.parseInt(filename.substring(colonIndex+1))
		);
	}

	private Integer currentFrameIndex=null;
	public int getCurrentFrameIndex()
	{
		if (null==currentFrameIndex)
		{
			if (DetectorConfig.isEiger())
			{
				try {
					currentFrameIndex = CounterState.current.counter.latestCounter;
				} catch (NullPointerException e) {
					currentFrameIndex = 0;
				}
			}
			else
			{
				currentFrameIndex = Epics.getInteger(frameNumberGetPV);
				System.out.println("Current frame is "+currentFrameIndex+" from "+frameNumberGetPV);

			}
		}
		return currentFrameIndex;
	}

	public String getImageViewerDir()		{ return Epics.getString(PV_FPE.viewDir256); }
	public void setETA(double etaSec)		{ Epics.putDouble(PV_FPE.etc, etaSec); }
	public void setCmdOutput(String cmd) 	{ Epics.put(PV_FPE.cmd_output256, cmd); }
	public double getETA() 					{ return Epics.getDouble(PV_FPE.etc); }
	public void setImageViewerFilePath(String dir, String file)
	{
		String lowercaseFile = file.toLowerCase();
		if (lowercaseFile.endsWith(".jpg"))
		{
			Log.error("JPEG not supported in image viewer, skipping viewer update for filename: "+file);
			Util.printStackTrace();
			return;
		}
		else if (lowercaseFile.endsWith(".h5"))
		{
			Log.error("HDF5 not supported in image viewer, skipping viewer update for filename: "+file);
			Util.printStackTrace();
			return;
		}
		
		Epics.put(PV_FPE.viewDir256, dir);
		Epics.put(PV_FPE.viewFile256, file);
		Epics.putInteger(PV_FPE.viewUpdate, 0);
		Epics.putInteger(PV_FPE.viewUpdate, 1);
	}
	public String getLastFilename()			{ return Epics.getString(PV_FPE.lastfilename); }
	public String getLastFileDir()
	{
		String dir = Epics.getString(PV_FPE.lastfiledir256);
		if (DetectorConfig.isDectris() && dir.charAt(0)=='.')
		{
			dir = System.getenv("HOME")+dir.substring(1);
		}
		return dir;
	}
	
	// it's important these are not public since they work differently for DevFPE1 vs DevFPE2
	// they are only for use by those subclasses
	protected int getOperationStatus()		{ return Epics.getInteger(PV_FPE.operationStatus);	}
	protected int getStart()				{ return Epics.getInteger(PV_FPE.start);			}
	protected int getCacheOpStat()			{ return Epics.getCacheInteger(PV_FPE.operationStatus);	}
	protected int getCacheStart()			{ return Epics.getCacheInteger(PV_FPE.start);			}
	protected boolean getFPERunning() 		{ return Epics.checkConnection(fpeStatePV, 3000); }
	protected boolean getDetectorConnected() { return 1==Epics.getInteger(connectPV); }
	
	////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////
	
	protected DevFPE.XYZMotion getCurrentPosition()
	{
		double x = DevPmacMotor.get("sample_x").getPosition();
		double y = DevPmacMotor.get("sample_y").getPosition();
		double z = DevPmacMotor.get("sample_z").getPosition();
		return new DevFPE.XYZMotion(x,x,y,y,z,z);
	}
	
	protected boolean activeCheck(FrameList frameList)
	{
		if
		(
			!Client.isMaster() &&
			(
				ModelCollectRunList.get().isCollectThreadAlive() ||
				ModelRasterState.isRasterRunning()
			)
		)
		{
			Log.note("Client is passive during collect/raster, ending FPE command.");
			if (null != frameList)
			{
				frameList.result = FrameListResult.FAILED;
				frameList.disconnectListeners();
			}
			return false;
		}
		return true;
	}
	
	public boolean checkFPEState(FrameList frameList, boolean nonFPEChecks)
	{
		if (!getFPERunning())
		{
			if (DetectorConfig.isDectris())
			{
				fpeErrorPopup
				(
					"FPE not running",
					"The frame processing engine (FPE) is not running. \n"+
					"Please close the existing FPE window if it is still open,\n"+
					"and then click the Pilatus icon to restart the FPE.\n",
					false
				);
			}
			else
			{
				fpeErrorPopup
				(
					"FPE not running",
					"The frame processing engine (FPE), which \n"+
					"runs on the MAR workstation, is not running. \n"+
					"Please close all open windows on the MAR workstation,\n"+
					"then click the MARCCD icon to restart the program.",
					false
				);
			}
			if (null != frameList)
			{
				frameList.result = FrameListResult.FAILED;
				frameList.disconnectListeners();
			}
			return false;
		}
		if (!getDetectorConnected())
		{
			if (DetectorConfig.isDectris())
			{
				fpeErrorPopup
				(
					"Lost MAR GUI connection",
					
					"The frame processing engine (FPE), which \n"+
					"runs on the MAR workstation, has lost its connection\n"+
					"to the MARCCD application. Please close all open\n"+
					"windows on the MAR workstation, then click the MARCCD\n"+
					"icon to restart the program.",
					false
				);
			}
			else
			{
				fpeErrorPopup
				(
					"Lost Pilatus camserver connection",
					
					"The frame processing engine (FPE) has lost its \n"+
					"connection to the Pilatus camserver.  Please restart \n"+
					"the server by choosing 'Start Pilatus camserver' from \n"+
					"the Tools menu in the MEDM toolbar, which can be opened \n"+
					"by clicking the driving wheel icon.",
					false
				);
			}
			if (null != frameList)
			{
				frameList.result = FrameListResult.FAILED;
				frameList.disconnectListeners();
			}
			return false;
		}
		
		// non-FPE checks
		if (nonFPEChecks)
		{
			CmdCheck cc = new CmdCheck(CmdCheck.FLAGS_COLLECTFRAME);
			cc.run();
			if (!cc.checksPassed())
			{
				Log.error("Checks failed, aborting frame");
				if (null!=frameList)
				{
					frameList.result = FrameListResult.FAILED;
					frameList.disconnectListeners();
				}
				return false;
			}
		}
		
		return true;
	}
	
	protected void fpeErrorPopup(String title, String errorText, boolean addStandardMessage)
	{
		Log.error("FPE not responding: "+errorText);
		if (DBConfig.serverMode) return;

		if (addStandardMessage)
		{
			if (DetectorConfig.isDectris())
			{
				errorText =
					"The frame processing engine (FPE) is not responding. \n"+
					"Please close the existing FPE window if it is still open,\n"+
					"and then click the Pilatus icon to restart the FPE.\n"+
					"\n"+
					"The specific error was: \n"+
					errorText;
			}
			else
			{
				errorText =
					"The frame processing engine (FPE), which \n"+
					"runs on the MAR workstation, is not responding. \n"+
					"Please close all open windows on the MAR workstation,\n"+
					"then click the MARCCD icon to restart the program.\n"+
					"\n"+
					"The specific error was: \n"+
					errorText;
			}
		}
		
		NBMessageBox mb = new NBMessageBox(GUI.getShell(), SWT.APPLICATION_MODAL | SWT.ICON_ERROR | SWT.OK);
		mb.setText(title);
		mb.setMessage
		(
			errorText
		);
		mb.open();
	}
	
	private Epics.EpicsValueMonitor currentFrameHandler = null;
	public DevFPE(String motorID)
	{
		super(motorID);
		fpeStatePV = DetectorConfig.isEiger() ? PV_EIGER.FPE_state : PV_MAR.FPE_state;
		connectPV = DetectorConfig.isEiger() ? PV_EIGER.AsynIO$CNCT : PV_MAR.AsynIO$CNCT;
		llFPEState = new ListenerList<DevFPE>(this);
		inst = this;

		lastFileListener = new ListenerList<DevFPE>(this);
		imageViewerListener = new ListenerList<DevFPE>(this);
		commandListener = new ListenerList<DevFPE>(this);
		etaListener = new ListenerList<DevFPE>(this);
		frameListener = new ListenerList<DevFPE>(this);
		
		if (DBConfig.get("fpe_counter").equals("adeiger"))
		{
			frameNumberGetPV = PV_EIGER.ArrayCounter_RBV;
			frameNumberPutPV = PV_EIGER.ArrayCounter;
		}
		else if (DBConfig.get("fpe_counter").equals("latest"))
		{
			// should no longer be used
			frameNumberGetPV = null;
			frameNumberPutPV = null;
		}
		else
		{
			frameNumberGetPV = PV_FPE.frameNumber;
			frameNumberPutPV = PV_FPE.frameNumber;
		}
		
		currentFrameHandler = new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				String str = Epics.dbrToString(value);
				System.out.println("camonitor "+pvName+" "+str);
				
				handleFramePV(Epics.dbrToInt(value), false);
			}
		};
		if (DetectorConfig.isEiger()) {
            SWServer.llChange.addListener((obj,data)->{handleFramePV(((CounterState.CounterInfo)data).latestCounter, false);});
		} else {
            boolean getPVOK = Epics.checkConnection(frameNumberGetPV, 5000); // pre-connect to avoid the initial callback
			System.out.println("Connecting "+frameNumberGetPV+": connection="+getPVOK);
            Epics.addMonitor(frameNumberGetPV, currentFrameHandler);
		}
		
		/*Epics.addMonitor(PV_FPE.lastfileupdate, new Epics.EpicsValueMonitor()
		{
			@Override
			public void valueChanged(String pvName, DBR value)
			{
				String filename = getLastFilename();
				boolean isBlank = filename.trim().length()==0;
				
				if (Client.isMaster() && Epics.dbrToInt(value)==1)
				{
					// acknowledge update
					Epics.putInteger(PV_FPE.lastfileupdate, 0);
					
					// copy update to image viewer PVs
					if (!isBlank) setImageViewerFilePath(getLastFileDir(), filename);
					
		        	// start a thread to check the header to make sure values match expected
					try
					{
    					ModelCollectRun currentCollectRun = ModelCollectRunList.get().getCurrentRunModel();
    					if (null!=currentCollectRun && currentCollectRun.seq().getTotalFrameCount()>0)
    		        	{
    						int frameCheckIndex=-1;
    						for (Frame frame : currentCollectRun.seq().getFrameListCopy())
    						{
    							if (frame.filename.equals(filename)) frameCheckIndex = frame.frameIndex;
    						}
    						if (-1 != frameCheckIndex)
    						{
    			        		HeaderCheck.check(currentCollectRun, frameCheckIndex);
    			        	}
    		        	}
					} catch (Exception e) { e.printStackTrace(); }
				}
				
				if (!isBlank && Epics.dbrToInt(value)==1)
				{
					lastFileListener.notifyListeners();
				}
			}});*/
		
		Epics.addMonitor(PV_FPE.viewUpdate, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				imageViewerListener.notifyListeners();
			}
		});
		
		Epics.addMonitor(fpeStatePV, new Epics.EpicsMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				llFPEState.notifyListeners();
			}

			@Override public void connectionChanged(String pvName, boolean isConnected)
			{
				llFPEState.notifyListeners();
			}
		});
		
		Epics.addMonitor(PV_FPE.operationStatus, new Epics.EpicsMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				llFPEState.notifyListeners();
			}

			@Override public void connectionChanged(String pvName, boolean isConnected)
			{
				llFPEState.notifyListeners();
			}
		});
		
		Epics.addMonitor(PV_FPE.cmd_output256, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				String cmd = Epics.dbrToString(value);
				if (cmd.length()>0)
				{
					commandListener.notifyListeners(cmd);
				}
			}
		});
		
		Epics.addMonitor(PV_FPE.logMessage256, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				String cmd = Epics.dbrToString(value);
				if (cmd.length()>0)
				{
					commandListener.notifyListeners("log_note "+cmd);
				}
			}
		});
		
		Epics.addMonitor(PV_FPE.etc, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				Double eta = Epics.dbrToDouble(value);
				etaListener.notifyListeners(eta);
			}
		});
		
		Client.addListener(new ClientListener()
		{
			@Override public void masterChanged(boolean isMaster)
			{
				// reset lastfileupdate PV
				if (Epics.getInteger(PV_FPE.lastfileupdate) == 1)
				{
					Epics.putInteger(PV_FPE.lastfileupdate, 0);
				}
					
				// update frame index if FPE is running
				if (!DetectorConfig.isEiger() && Epics.checkConnection(frameNumberGetPV, 3000))
				{
					handleFramePV(getCurrentFrameIndex(), false);
				}
			}
		});
	}
	
	////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////// copy frame progress to run PV ////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////
	
	public void handleFramePV(int fpeCurrentFrame, boolean updateOutsideOfCollect)
	{
		// don't relay updates unless collection is going on
		int opStat = getOperationStatus();
		if (opStat>1 && !updateOutsideOfCollect)
		{
			System.out.println("Skipping frame update to "+fpeCurrentFrame+" because opStat="+opStat);
		}
		
		System.out.println("handleFramePV "+fpeCurrentFrame);
		currentFrameIndex = fpeCurrentFrame;
		frameListener.notifyListeners(fpeCurrentFrame);
		
		if (!DBConfig.serverMode && !Client.isMaster()) return;
		
		// if collection is running..
		if (CmdCollectJet.thread.isAlive()) {
			if (currentFrameIndex>0) {
				String filename = CmdCollectJet.getFilename(currentFrameIndex);
				setImageViewerFilePath(CmdCollectJet.getFullDir(), filename);
				Epics.put(PV_FPE.lastfilename, filename);

				// set nextFrame in Redis, and copy to run-specific key if requested
				try (Jedis jedis = Redis.get()) {
					jedis.hset(Redis.server_fpe, "nextFrame", "" + currentFrameIndex);
					if (currentFrameList.toKey.contains("$")) {
						StringTokenizer st = new StringTokenizer(currentFrameList.toKey, "$");
						jedis.hset(st.nextToken(), st.nextToken(), "" + currentFrameIndex);
					}
				}
			}
		}
		else if (ModelCollectRunList.get().isCollectThreadAlive())
		{
			ModelCollectRun currentCollectRun = ModelCollectRunList.get().getCurrentRunModel();
			if (null==currentCollectRun) { System.out.println("currentCollectRun is null, skipping update."); return; }
			
			// and the frame counter has moved...
	        int nextFrame = currentCollectRun.getNextFrame();
			if (fpeCurrentFrame > nextFrame)
	        {
	        	// update the frame PV for the current run
	        	currentCollectRun.dev().setNextFrame(fpeCurrentFrame);
	        	
	        	// update the image viewer path
	        	Frame collectFrame = currentCollectRun.seq().getFrameListCopy().get(fpeCurrentFrame-1);
	        	if (DetectorConfig.isEiger())
	        	{
	        		String filename = Main.FAKE_DATA ? String.format("AEF_Se_MalA_1_2_1_%06d.cbf",collectFrame.frameIndex+1) : collectFrame.cbfFilename;
	        		String dir = Main.FAKE_DATA ? "/mnt/beegfs/testdata/QXU" : currentCollectRun.getFullDirectory();
	        		setImageViewerFilePath(dir, filename);
	        		
        			Epics.put(PV_FPE.lastfilename, filename);
	        	}
	        	else if (DetectorConfig.isPilatus())
	        	{
	        		// use the filename from the previous frame, not the one currently being taken
	        		String filename = collectFrame.getDisplayFilename();
	        		setImageViewerFilePath(currentCollectRun.getFullDirectory(), filename);
	        	}
	        } else {
	        	System.out.println("fpeCurrentFrame "+fpeCurrentFrame+" <= currentCollectRun.getNextFrame()" + nextFrame + ", skipping update.");
			}
		}
		// if raster is running..
		else if (ModelRasterState.isRasterThreadAlive())
		{
			ModelRasterRun currentRasterRun = ModelRasterState.get().getCurrentRunModel();
			if (null==currentRasterRun) return;
			
			// and the frame counter has moved...
	        if (fpeCurrentFrame > currentRasterRun.getNextFrame())
	        {
	        	// update the frame PV for the current run
	        	currentRasterRun.dev().setNextFrame(fpeCurrentFrame, true);

        		// use the filename from the previous frame, not the one currently being taken
        		ModelRasterCell rasterCell = currentRasterRun.grid().getGridCalc().getCellListCopy().get(fpeCurrentFrame-1);
        		
	        	// update the image viewer path
        		if (DetectorConfig.isEiger())
        		{
	        		String filename = rasterCell.getCBFImageName();
	        		setImageViewerFilePath(currentRasterRun.getFullDirectory(), filename);
        		}
        		else if (DetectorConfig.isPilatus())
	        	{
	        		String filename = rasterCell.getImageName();
	        		setImageViewerFilePath(currentRasterRun.getFullDirectory(), filename);
	        	}
	        }
		}
		else
		{
			System.out.println("Didn't update for frame advance, neither collect nor raster running");
		}
		
		updateETA();
	}
	
	////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////// FPE state ////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////
	
	public static enum FPEState
	{
		cxp_speed,cyp_speed,czp_speed,gonio_speed,enc_home,pad_connect,idle,
		st_pos,raster_speed,gonio_still,set_speed_low,start_omega,set_speed_high,
		blu_ice_updates,stop_omega,next_point,init_1,set_go_speed,stop_motion,unknown, arm_eiger;
		
		public static FPEState lookup(String s)
		{
			for (FPEState state : FPEState.values())
			{
				if (state.toString().equals(s)) return state;
			}
			//System.out.println("Unknown FPE state: "+s);
			return unknown;
		}
	}
	
	public FPEState getCacheFPEState()
	{
		if (!Epics.checkConnection(fpeStatePV, 50)) return null;
		return FPEState.lookup(Epics.getCacheString(fpeStatePV));
	}
	
	public FPEState getFPEState()
	{
		if (!Epics.checkConnection(fpeStatePV, 50)) return null;
		return FPEState.lookup(Epics.getString(fpeStatePV));
	}
	
	public boolean isRunning()
	{
		return null!=getFPEState();
	}
	
	public boolean usingServerCounter()
	{
		return DBConfig.get("fpe_counter").equals("adeiger")||
			   DBConfig.get("fpe_counter").equals("latest");
	}
	
	@Override protected void onAllPVsConnected() {}
	@Override protected void valueChangedAction(String pvName, DBR value) {}

	public static String getPathAfterCommonDir(String fullPath)
	{
		String ret = fullPath;
		if (ret.indexOf(DevDataDir.getPrimary())==0)
		{
			ret = ret.substring(DevDataDir.getPrimary().length());
			for (int i=0;i<2;i++) if (ret.indexOf('/')==0) ret=ret.substring(1);
		}
		return ret;
	}
}
