package gov.anl.gmca.collect.graph;

import gov.anl.gmca.collect.DevFPE;
import gov.anl.gmca.collect.ModelCollectRun;
import gov.anl.gmca.collect.ModelCollectRun.RunState;
import gov.anl.gmca.collect.ModelCollectRunList;
import gov.anl.gmca.collect.RunCalculator.Frame;
import gov.anl.gmca.common.Redis;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PV_BLUICE;
import gov.anl.gmca.spotfinder.DBImageStats;
import gov.anl.gmca.spotfinder.DevSFType;
import gov.anl.gmca.spotfinder.DevSFType.SFRunType;
import gov.anl.gmca.spotfinder.DevSFType.SPOTFINDER_OUTPUT;
import gov.anl.gmca.spotfinder.DevSFType.SPOTFINDER_TYPE;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Transform;

import java.nio.FloatBuffer;
import java.util.ArrayList;

public class SFGraph
{
	private static final double MIN_DELTA = Double.MAX_VALUE;
	WdgSFGraph wdg;
	SFAxes axes;
	SFPlot plot;
	SFHover hover;
	SFImageLoad imgLoad;
	
	// contains extra info for use in graphing
	public class SpotFinderPoint
	{
		public SpotFinderPoint(float data) { this.data = data; }
		float data;
		int frameIndex;
		boolean isInverse;
		double angle, delta;
		int xPlotAreaCoord;
		int yPlotAreaParam1Coord, yPlotAreaParam2Coord;
		String filename, directory;
		SpotFinderPoint inversePoint;
	}
	
	static Object[][] selectedOutputs =
	{
		{SPOTFINDER_TYPE.DIALS_SPOTFINDER, SPOTFINDER_OUTPUT.DIALS_SPOT_TOTAL, SPOTFINDER_OUTPUT.DIALS_TOTAL_INT},
		{SPOTFINDER_TYPE.DOZOR, SPOTFINDER_OUTPUT.DOZOR_MAIN_SCORE, SPOTFINDER_OUTPUT.DOZOR_RES}
	};
	public static Object[] getOutputInfo()
	{
		for (Object[] o : selectedOutputs) if ((SPOTFINDER_TYPE)o[0]==DevSFType.getConfig(SFRunType.COLLECT).id) return o;
		return null;
	}
	public static SPOTFINDER_OUTPUT getOutput1() { return (SPOTFINDER_OUTPUT)getOutputInfo()[1]; }
	public static SPOTFINDER_OUTPUT getOutput2() { return (SPOTFINDER_OUTPUT)getOutputInfo()[2]; }
	
	ArrayList<SpotFinderPoint> points;
	int currentRunIndex;
	public int marginX=5, marginY=5;
	private Transform transform;
	double minAngle = 0, maxAngle = 1, minParam1=0,maxParam1=1, minParam2=0, maxParam2=0;
	
	int yAxisWidth=50, xAxisHeight=40;
	int plotAreaXMin,plotAreaXMax,plotAreaYMax,plotAreaHeight,plotAreaWidth,yAxisHeight;
	
	public SFGraph(WdgSFGraph wdg)
	{
		this.wdg = wdg;
		plot = new SFPlot(this);
		axes = new SFAxes(this);
		hover = new SFHover(this);
		imgLoad = new SFImageLoad(this);
	}
	
	///////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////// conversions ///////////////////////////////
	///////////////////////////////////////////////////////////////////////////////
	
	public int angleToXPixelCoord(double angle, boolean isInverse)
	{
		// find the relative angle to the starting angle
		double relativeAngle = angle - minAngle;
		
		// subtract 180 for inverse frames
		double correctedAngle = relativeAngle - (isInverse ? 180 : 0);

		// calculate the pixel
		return (int)( correctedAngle / (maxAngle - minAngle) * plotAreaWidth);
	}
	
	public int outputToYPixelCoord(double value, double min, double max, boolean reverseDisplay)
	{
		if (Double.isNaN(value)) return -1;
		if (reverseDisplay) {
			return (int)((value-min)/(max-min)*plotAreaHeight);
		} else {
			return plotAreaHeight - (int)((value-min) / (double)(max-min) * (double)plotAreaHeight);
		}
	}
	
	///////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////// transforms ////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////
	
	private void resetTransform(GC gc)
	{
		if (null != transform && !transform.isDisposed()) transform.dispose();
		transform = new Transform(gc.getDevice());
		//gc.setClipping(fullSize);
	}
	
	public void setXAxisTransform(GC gc)
	{
		resetTransform(gc);
		transform.translate(plotAreaXMin, plotAreaYMax);
		gc.setTransform(transform);
		
		//gc.setClipping(0, 0, plotAreaWidth, xAxisHeight);
	}
	
	public void setLeftYAxisTransform(GC gc)
	{
		resetTransform(gc);
		transform.translate(marginX,marginY);
		gc.setTransform(transform);
		
		//gc.setClipping(0, 0, yAxisWidth, plotAreaHeight);
	}
	
	public void setRightYAxisTransform(GC gc)
	{
		resetTransform(gc);
		transform.translate(plotAreaXMin+plotAreaWidth,marginY);
		gc.setTransform(transform);
		
		//gc.setClipping(0, 0, yAxisWidth, plotAreaHeight);
	}
	
	public void setPlotTransform(GC gc)
	{
		resetTransform(gc);
		transform.translate(marginX+yAxisWidth, marginY);
		gc.setTransform(transform);
		
		//gc.setClipping(0, 0, plotAreaWidth, plotAreaHeight);
	}

	///////////////////////////////////////////////////////////////////////////////
	////////////////////////////// reload & update ////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////
	
	private void resetRanges()
	{
		minParam1 = 0;
		maxParam1 = 1;
		minAngle = 0;
		maxAngle = 1;
		minParam2 = 0;
		maxParam2 = 1;
	}
	
	public void reloadData()
	{
		// get frame list for current run
		ModelCollectRun currentRun = ModelCollectRunList.get().getCurrentRunModel();
		ArrayList<Frame> frameList = currentRun.seq().getFrameListCopy();

		// load data from the database
		currentRunIndex = ModelCollectRunList.get().getCurrentRunIndex();
		String dir = currentRun.getDirectorySuffix();
		String prefix = currentRun.getFileRoot();
		int runLabel = currentRun.getRunLabel();

		String param1, param2;
		switch (DevSFType.get(DevSFType.SFRunType.COLLECT)._getConfig().id)
		{
		case DIALS_SPOTFINDER:
			param1 = "dials:spots";
			param2 = "dials:res";
			break;
		case DOZOR:
			param1 = "dozor:score";
			param2 = "dozor:res";
			break;
		default:
			throw new IllegalStateException("Unexpected spot finder type: "+DevSFType.get(DevSFType.SFRunType.RASTER)._getConfig().id.toString());
		}

		FloatBuffer data = DBImageStats.load(Redis.image_stats(dir, prefix, runLabel, param1));
		FloatBuffer data2 = DBImageStats.load(Redis.image_stats(dir, prefix, runLabel, param2));

		// create empty point array
		points = new ArrayList<SpotFinderPoint>();
		
		// check sizes
		if (currentRun.getRunIndex()==0)
		{
			if (data.capacity()==0 || frameList.size()==0 || !DevFPE.get().isIdle()) { resetRanges(); return; }
		}
		else
		{
			if (data.capacity()==0 || frameList.size()==0 || currentRun.getRunState()==RunState.INACTIVE) { resetRanges(); return; }
		}
		if (data.capacity()>frameList.size())
		{
			System.out.println("SpotFinder data size ("+data.capacity()+") > frame list size ("+frameList.size()+"), not reloading.");
			resetRanges();
			return;
		}
		
		// initialize range variables
		minAngle = maxAngle = frameList.get(0).angle;
		float firstVal1 = data.get(0);
		maxParam1=firstVal1;
		maxParam2 = 50;//data2.get(0);

		// calc range variables
		for (int i=0;i<data.capacity();i++)
		{
			float f = data.get(i);
			if (f < minParam1) minParam1 = f;
			if (f > maxParam1) maxParam1 = f;
			float f2 = 50;//data2.get(i);
			if (f2 < minParam2) minParam2 = f2;
			if (f2 > maxParam2) maxParam2 = f2;
		}
		for (Frame frame : frameList)
		{
			if (frame.angle < minAngle && !frame.isInverse) minAngle = frame.angle;
			if (frame.angle > maxAngle && !frame.isInverse) maxAngle = frame.angle;
		}
		
		// now that the range variables are set, create the point array which uses them
		for (int frameIndex=0;frameIndex<frameList.size();frameIndex++)
		{
			// get the frame object for this SQL row
			Frame frame = frameList.get(frameIndex);

			// copy info from frame list
			float dataValue = Float.NaN;
			if (frame.frameIndex < data.capacity() && data.get(frame.frameIndex)>=0) dataValue = data.get(frame.frameIndex);
			SpotFinderPoint point = new SpotFinderPoint(dataValue);
			point.frameIndex = frameIndex;
			point.angle = frame.angle;
			point.delta = frame.delta;
			point.isInverse = frame.isInverse;
			point.xPlotAreaCoord = angleToXPixelCoord(frame.angle, frame.isInverse);
			point.yPlotAreaParam1Coord = outputToYPixelCoord(dataValue, minParam1, maxParam1, false);
			point.yPlotAreaParam2Coord = outputToYPixelCoord(50/*data2.get(frameIndex)*/, minParam2, maxParam2, true);

			// fill in directory and filename - run 0 is filled in from PVs
			if (currentRunIndex==0)
			{
				String imagePath = Epics.getString(frameIndex==0 ? PV_BLUICE.run0LastFrame0 : PV_BLUICE.run0LastFrame1);
				int lastSlashIndex = imagePath.lastIndexOf('/');
				if (lastSlashIndex>0)
				{
					point.directory = imagePath.substring(0, lastSlashIndex);
					point.filename = imagePath.substring(lastSlashIndex+1);
				}
			}
			else
			{
				point.directory = currentRun.getFullDirectory();
				point.filename = frame.cbfFilename;
			}
			
			points.add(point);
		}
		
		// copy the inverse frame references
		for (SpotFinderPoint point : points)
		{
			int frameIndex = point.frameIndex;
			Frame frame = frameList.get(frameIndex);
			//System.out.println("Frame "+frame.frameIndex+" inverse="+frame.inverseFrame.frameIndex);
			
			if (frame.isInverse && null != frame.inverseFrame)
			{
				int inverseIndex = frame.inverseFrame.frameIndex;
				if (inverseIndex<points.size())
				{
					SpotFinderPoint inversePoint = points.get(inverseIndex);
					//inversePoint.inversePoint = point;
					point.inversePoint = inversePoint;
				}
			}
		}
	}
	
	public void updateLayout(GC gc, int width, int height)
	{
		plotAreaXMin = marginX + yAxisWidth;
		plotAreaYMax = height - marginY - xAxisHeight;
		plotAreaWidth = width - marginX*2 - yAxisWidth*2;
		plotAreaHeight = height - marginY*2 - xAxisHeight;
	}

	public SpotFinderPoint getClosestPt(Point mousePos)
	{
		SpotFinderPoint closestPoint = null;
		double deltaMin = Double.MAX_VALUE;
		for (SpotFinderPoint point : points)
		{
			if (point.isInverse) continue;
			
			int x = point.xPlotAreaCoord;
			int dx = Math.abs(mousePos.x-x);
			
			if (dx<MIN_DELTA && dx<deltaMin) { closestPoint=point; deltaMin=dx; }
		}
		
		return closestPoint;
	}

	public Point mousePosToGraphPos(int mx, int my)
	{
		Point ret = new Point(mx - plotAreaXMin, my - marginY);
		return ret;
	}
}
