package gov.anl.gmca.collect.ui;

import gov.anl.gmca.collect.ModelCollectRun;
import gov.anl.gmca.collect.ModelCollectRun.ScanMode;
import gov.anl.gmca.collect.ModelCollectRunList;
import gov.anl.gmca.collect.ModelCollectRunList.CollectLeftTab;
import gov.anl.gmca.collect.viewer.WdgDiffTab;
import gov.anl.gmca.common.ModelRunList.CurrentRunListener;
import gov.anl.gmca.common.ModelRunList.ModelRun;
import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.strategy.WdgStrategy;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.util.ListenerList;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.CTabFolder;
import org.eclipse.swt.custom.CTabItem;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;

// Contains a tab folder which contains:
//
// 1) diffraction viewer
// 2) either the vector video, raster video, repeat video, or none of those depending on the 
//    current run's mode
// 3) strategy tab
// 4) multi-crystal strategy tab
//
// The second sub-tab, which may not be visible, needs to be updated based on the current run.
// When the run's type changes, the old "visibleSubTab" is disposed if it existed and the new one,
// if needed, is created.  The controls contained in the tabs are created once at startup and
// are not disposed of.  They are simply attached to the sub-tab widgets as they are created.
//
public class WdgLeftPane extends Composite
{
	private CTabFolder 		leftFolder;
	private CTabItem 		diffractionItem, vectorItem, rasterItem;
	private CTabItem 		repeatItem, strategyItem;
	//private CTabItem 		MultiCrystalItem;
	public	WdgDiffTab		diffTab;
	private WdgVectorTab 	vectorTab;
			WdgRasterTab 	rasterTab;
	//private WdgImportTab	repeatTab;
	private WdgStrategy 	wdgStrategy;
	//private WdgMultiStrategy wdgMultiCrystal;
	private	boolean			leftPaneVisible=false;	
	private CTabItem		visibleSubTab=null;		// null, vectorItem, rasterItem, or repeatItem
	
	public WdgLeftPane(Composite parent)
	{
		super(parent, SWT.NONE);
		
		setBackground(GUI.BGCOLOR);
		setLayout(new FillLayout());

		leftFolder = new CTabFolder(this, SWT.TOP);
		leftFolder.setBackground(GUI.BGCOLOR);
		leftFolder.setFont(GUI.FONT_HUTCH_GROUP);
		leftFolder.setMinimumCharacters(12);
		leftFolder.setEnabled(true);

		diffTab	  		= new WdgDiffTab(leftFolder, true);
		vectorTab 		= new WdgVectorTab(leftFolder);
		rasterTab 		= new WdgRasterTab(leftFolder);
		//repeatTab		= new WdgImportTab(leftFolder);
		wdgStrategy 	= new WdgStrategy(leftFolder);
		//wdgMultiCrystal = new WdgMultiStrategy(leftFolder);
		
		diffractionItem = new CTabItem(leftFolder,SWT.NONE);
		diffractionItem.setText("  Diffraction  ");
		diffractionItem.setControl(diffTab);

		strategyItem = new CTabItem(leftFolder,SWT.NONE);
		strategyItem.setText("  Legacy Strategy  ");
		strategyItem.setControl(wdgStrategy);
		
		//multiCrystalItem = new CTabItem(leftFolder,SWT.NONE);
		//multiCrystalItem.setText("  Multi Crystal Strategy  ");
		//multiCrystalItem.setControl(wdgMultiCrystal);
		
		leftFolder.setSelection(0);
	}

	public void connectListeners()
	{
		diffTab.connectListeners();
		vectorTab.connectListeners();
		rasterTab.connectListeners();
		//repeatTab.connectListeners();
		wdgStrategy.connectListeners();
		//wdgMultiCrystal.connectListeners();
		
		// notify sub-tabs when they are visible
		leftFolder.addSelectionListener(new SelectionListener()
		{
			@Override public void widgetSelected(SelectionEvent e)
			{
				updateVisibility();
			}
			@Override public void widgetDefaultSelected(SelectionEvent e) {}
		});

		// show/hide sub-tabs based on run state
		ModelCollectRunList.get().addCurrentRunListener(new CurrentRunListener()
		{
			@Override public void currentRunChanged(final ModelRun currentRun)
			{
				GUI.asyncExec(new Runnable(){public void run()
				{
					updateTabState((ModelCollectRun)currentRun);
				}});
			}
		});
		GUI.asyncExec(new Runnable(){public void run()
		{
			ModelCollectRun currentRun = ModelCollectRunList.get().getCurrentRunModel();
			if (null!=currentRun) updateTabState(currentRun);
		}});

		// synchronize the selected sub-tab with the PV
		ModelCollectRunList.get().llLeftTab.addListener(new ListenerList.Listener<ModelCollectRunList>()
		{
			@Override public void handleEvent(ModelCollectRunList object, Object data)
			{
				if (!Client.isMaster()) return;
				
				final CollectLeftTab leftTab = ModelCollectRunList.get().getLeftTab();
				if (null != leftTab)
				{
    				GUI.asyncExec(new Runnable(){public void run()
    				{
    					switch (leftTab)
    					{
    					case DIFFRACTION:
    						leftFolder.setSelection(diffractionItem);
    						break;
    					case VECTOR:
    						if (null != visibleSubTab)
    							leftFolder.setSelection(visibleSubTab);
    						break;
    					case STRATEGY:
							leftFolder.setSelection(diffractionItem);
    						//leftFolder.setSelection(strategyItem);
							Log.note("Switch to strategy view");
    						break;
    					}
    				}});
				}
			}
		});
	}

	public void notifyIsVisible(boolean isVisible)
	{
		leftPaneVisible=isVisible;
		updateVisibility();
	}
	
	private void updateVisibility()
	{
		CTabItem selectedSubTab = leftFolder.getSelection();
		
		diffTab.notifyIsVisible(leftPaneVisible 	&& diffractionItem 	== selectedSubTab);
		vectorTab.notifyIsVisible(leftPaneVisible 	&& vectorItem 		== selectedSubTab);
		rasterTab.notifyIsVisible(leftPaneVisible 	&& rasterItem		== selectedSubTab);
		//repeatTab.notifyIsVisible(leftPaneVisible 	&& repeatItem		== selectedSubTab);
	}
	
	private void updateTabState(ModelCollectRun currentRun)
	{
		ScanMode scanMode = currentRun.getScanMode();
		
		switch (scanMode)
		{
		case SINGLE:
		case JET:
			
			if (null == visibleSubTab) break;
			else
			{
				// creating other tabs shows them, but here we need to specifically select
				// the diffraction tab to be visible
				selectItem(diffractionItem);
				
				hideItem(visibleSubTab);
				vectorItem=rasterItem=repeatItem=null;
				visibleSubTab = null;
			}
			break;
			
		case VECTOR:
			
			if (null != vectorItem && vectorItem == visibleSubTab) break;
			else
			{
				hideItem(visibleSubTab);
				visibleSubTab = vectorItem = showItem(vectorTab);
				rasterItem=repeatItem=null;
			}
			break;
			
		case RASTER:
		case MESH:
			
			if (null != rasterItem && rasterItem == visibleSubTab) break;
			else
			{
				hideItem(visibleSubTab);
				visibleSubTab = rasterItem = showItem(rasterTab);
				vectorItem=repeatItem=null;
			}
			break;
			
		/*case IMPORT:
			
			if (null != repeatItem && repeatItem == visibleSubTab) break;
			else
			{
				hideItem(visibleSubTab);
				visibleSubTab = repeatItem = showItem(repeatTab);
				vectorItem=rasterItem=null;
			}
			break;*/
		}
		
		updateVisibility();
	}
	
	private void selectItem(CTabItem item)
	{
		CTabItem selectedSubTab = leftFolder.getSelection();
		if (selectedSubTab != item) leftFolder.setSelection(item);
	}
	
	private CTabItem hideItem(CTabItem item)
	{
		if (null==item) return null;
		item.setControl(null); // unset the control so the comp isn't disposed too!
		item.dispose();
		return null;
	}
	
	private CTabItem showItem(Control control)
	{
		String tabTitle = "";
		if (control == vectorTab)		tabTitle = "  Vector  ";
		else if (control == rasterTab)	tabTitle = "  Mesh  ";
		//else if (control == repeatTab)	tabTitle = "  Repeat  ";
		else throw new IllegalArgumentException("Invalid control");
		
		CTabItem item = new CTabItem(leftFolder,SWT.NONE,1);
		item.setText(tabTitle);
		item.setControl(control);
		
		leftFolder.setSelection(item);
		
		return item;
	}
}
