package gov.anl.gmca.common;

import gov.anl.gmca.collect.DevCollectSettings;
import gov.anl.gmca.collect.ModelCollectRunList;
import gov.anl.gmca.collect.ModelCollectRunList.CollectState;
import gov.anl.gmca.common.ModelMotorStatus.MotorStatus;
import gov.anl.gmca.common.client_server.StateMap;
import gov.anl.gmca.io.Database;
import gov.anl.gmca.io.Database.DBName;
import gov.anl.gmca.io.Database.Result;
import gov.anl.gmca.raster.ModelRasterState;
import gov.anl.gmca.raster.ModelRasterState.RasterState;
import gov.anl.gmca.sample.DevCollimatorCtrl;
import gov.anl.gmca.scan.ModelScan;
import gov.anl.gmca.scan.ModelScan.AutoStep;
import gov.anl.gmca.scan.ModelScan.InterStep;
import gov.anl.gmca.scan.ModelScan.ScanListener;
import gov.anl.gmca.screening.ui.WdgScreeningTasks;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.users.ModelUserTable.ClientListener;
import gov.anl.gmca.util.ListenerList;
import redis.clients.jedis.Jedis;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.concurrent.Semaphore;

public class AppState
{
	protected static final String exceptionTable = "DisableExceptions";
	
	//////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////// enums ////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////

	// These states...
	public static enum State
	{
		COLLECT, 		RASTER, 		SCAN, 			SCREENING,
		MTR_COLL_CTRL, 	MTR_DETECTOR, 	MTR_ENERGY, 	MTR_SAMPLE, 	MTR_OTHER,
		ROBOT, 			AUTOCENTER, 	ALIGN_MINIBEAM, ENTER_HUTCH,
		MASTER_SLAVE;
		
		public static State stringToEnum(String s)
		{
			for (State state : State.values())
			{
				if (state.toString().toLowerCase().equals(s.toLowerCase())) return state;
			}
			return null;
		}
	}
	
	// can prevent these operations from starting.
	public static enum Operation
	{
		START_COLLECT, 	START_RASTER, 	START_SCAN, 		START_SCREENING,
		MOVE_COLL_CTRL, MOVE_DETECTOR, 	MOVE_ENERGY, 		MOVE_SAMPLE, 		MOVE_OTHER,			MOVE_ZOOM,
		OP_ANNEALING, 	OP_AUTOCENTER, 	OP_ALIGN_MINIBEAM, 	OP_ENABLE_VECTOR, 	OP_ENTER_HUTCH, 	OP_RINSE_WASH,
		MOVE_BACKLIGHT;
		
		public static Operation stringToEnum(String s)
		{
			for (Operation operation : Operation.values())
			{
				if (operation.toString().toLowerCase().equals(s.toLowerCase())) return operation;
			}
			return null;
		}
	}
	
	public static HashMap<State,Boolean> getBusyMap() { return isBusyMap; }
	protected static HashMap<State,Boolean> isBusyMap = new HashMap<>(); // which states are currently busy?
	protected static HashMap<Operation,Boolean> isAllowedMap; // which operations are currently allowed?
	protected static HashMap<Operation,HashMap<State,Boolean>> allowedLookupMap; // the database mapping of which states disallow which operations
	
	protected final static Object[][] motorGroupingArray =
	{
		{ State.MTR_DETECTOR,	"detector_z", "detector_2t" },
		{ State.MTR_SAMPLE,		"sample_x", "sample_y", "sample_z", "gonio_omega" },
		{ State.MTR_OTHER, 		"camera_zoom","camera_zoom2","beamstop_z","beam_size_x","beam_size_y" }
	};
	
	public static State getMotorState(DevMove dev)
	{
		for (Object[] group : motorGroupingArray)
		{
			State state = (State)group[0];
			for (int i=1;i<group.length;i++)
			{
				String s = (String)group[i];
				if (dev.getDeviceName().equals(s)) return state;
			}
		}
		return null;
	}
	
	//////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////// listener //////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////

	public interface Listener
	{
		public abstract void stateChanged(Operation op, boolean allowed);
	}
	protected static Semaphore listenerMapLock = new Semaphore(1);
	
	protected static HashMap<Operation,ArrayList<Listener>> listenerMap = new HashMap<Operation,ArrayList<Listener>>();
	public static void addListener(Operation op, Listener listener)
	{
		listenerMapLock.acquireUninterruptibly();
		try
		{
    		if (listenerMap.containsKey(op))
    		{
    			listenerMap.get(op).add(listener);
    		}
    		else
    		{
    			ArrayList<Listener> al = new ArrayList<Listener>();
    			al.add(listener);
    			listenerMap.put(op, al);
    		}
		}
		finally { listenerMapLock.release(); }
	}
	public static void removeListener(Operation op, Listener listener)
	{
		listenerMapLock.acquireUninterruptibly();
		try
		{
    		if (listenerMap.containsKey(op))
    		{
    			while (listenerMap.get(op).remove(listener)) {}
    		}
		}
		finally { listenerMapLock.release(); }
	}
	
	protected static void notifyListeners(Operation op, boolean allowed)
	{
		if (!listenerMap.containsKey(op)) return;
		for (Listener l : new ArrayList<Listener>(listenerMap.get(op))) { l.stateChanged(op, allowed); }
	}
	
	// listener for state changes
	public static ListenerList<AppState> stateLL = new ListenerList<AppState>(null);

	//////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////// main methods ///////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////
	
	public static boolean isOperationAllowed(Operation op)
	{
		if (null == isAllowedMap) return false;
		
		return isAllowedMap.get(op);
	}
	
	public static String getActiveStates()
	{
		ArrayList<String> busyStates = new ArrayList<>();
		for (State s : isBusyMap.keySet())
		{
			if (isBusyMap.get(s))
			{
				busyStates.add(s.toString());
			}
		}
		return String.join(" ", busyStates);
	}

	//////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////// end public interface //////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////

	public static void sendStateToRedis()
	{
		try (Jedis jedis = Redis.get()) {

			HashMap<String,String> stateMap = new HashMap<>();
			for (State state : State.values()) {
				if (state==State.MASTER_SLAVE) continue; // this is client-specific and will be phased out
				stateMap.put(state.toString().toLowerCase(), isBusyMap.get(state)?"1":"0");
			}
			jedis.hmset(Redis.busy_states, stateMap);

			HashMap<String,String> opMap = new HashMap<>();
			for (Operation op : Operation.values()) {
				opMap.put(op.toString().toLowerCase(), isAllowedMap.get(op)?"1":"0");
			}
			jedis.hmset(Redis.busy_allowed_op, opMap);
		}
	}

	public static void setState(State newState, boolean busy)
	{
		if (null==isBusyMap || !isBusyMap.containsKey(newState) || isBusyMap.get(newState)==busy) return;

		// update the busy map
		isBusyMap.put(newState, busy);

		// update the allowed map and notify listeners of any changes
		if (null != allowedLookupMap)
		{
    		for (Operation op : Operation.values())
    		{
    			boolean wasOperationAllowed = isAllowedMap.get(op);
    			
    			boolean isOperationNowAllowed = true;
    			for (State s : State.values())
    			{
    				// if state 's' is currently busy, and the lookup map says a busy state
    				// for state 's' will disable operation 'op', then operation 'op' is
    				// not currently allowed
        			boolean isStateBusy = isBusyMap.get(s);
        			boolean doesStateDisallowOp = !allowedLookupMap.get(op).get(s);
    				if (isStateBusy && doesStateDisallowOp)
        			{
        				isOperationNowAllowed = false;
        			}
    			}
    			
    			if (wasOperationAllowed != isOperationNowAllowed)
    			{
    				isAllowedMap.put(op, isOperationNowAllowed);
    				notifyListeners(op, isOperationNowAllowed);
    			}
    		}
		}

		stateLL.notifyListeners();
	}
	
	public static void init()
	{
		//WdgAppState.open();
		
		// load the static map from the database
		loadTable();
		
		// initialize the two dynamic maps
		for (State state : State.values())
		{
			isBusyMap.put(state,false);
		}
		isAllowedMap = new HashMap<Operation,Boolean>();
		for (Operation operation : Operation.values())
		{
			isAllowedMap.put(operation, false);
		}

		// set up listeners
		Client.addListener(new ClientListener()
		{
			@Override public void masterChanged(boolean isMaster)
			{
				setState(State.MASTER_SLAVE, !isMaster);
			}
		});
		setState(State.MASTER_SLAVE, !Client.isMaster());
	
		// screening status
		if (WdgScreeningTasks.globalScreeningStatus == 1) {
			setState(State.SCREENING, true);
		}
		
		// motor listener
		ModelMotorStatus.init();
		ModelMotorStatus.ll.addListener(new ListenerList.Listener<ModelMotorStatus>()
		{
			// TODO: simplify this method by using getMotorState()
			@Override public void handleEvent(ModelMotorStatus motorStatus, Object motorStatusObject)
			{
				DevMove dev = ((MotorStatus)motorStatusObject).dev;
				for (Object[] motorGrouping : motorGroupingArray)
				{
					State state = (State)motorGrouping[0];
					
					// ignore detector busy state if detector motion is disabled
					if (!DevCollectSettings.getEnableDetectorMove() && state==State.MTR_DETECTOR) continue;
					
					for (int i=1; i<motorGrouping.length; i++)
					{
						// search all groups for the motor that changed state
						String groupedDeviceName = (String)motorGrouping[i];
						if (dev.getDeviceName().equals(groupedDeviceName))
						{
							// if the motor is in a group, calculate that group's overall state
							boolean groupIsMoving = false;
							for (int j=1; j<motorGrouping.length; j++)
							{
								DevMove device = (DevMove)EpicsDevice.getDevice((String)motorGrouping[i]);
								if (null != device && device.getMoving())
								{
									groupIsMoving=true;
								}
							}
							
							// and set it
							setState(state, groupIsMoving);
							return;
						}
					}
				}
			}
		});
		
		ModelScriptStatus.init();

		DevCollimatorCtrl.get().ll.addListener(new ListenerList.Listener<DevCollimatorCtrl>()
		{
			@Override public void handleEvent(DevCollimatorCtrl dev, Object data)
			{
				boolean isRunning = dev.isRunning();
				setState(State.MTR_COLL_CTRL, isRunning);
			}
		});
		
		if (null != ModelScan.get())
		{
			ModelScan.get().addListener(new ScanListener()
			{
    			@Override public void scanChanged(ModelScan settings)
    			{
    				setState(State.SCAN, isScanBusy(settings));
    			}
    		});
			setState(State.SCAN, isScanBusy(ModelScan.get()));
		}
		if (null != ModelRasterState.get())
		{
			ModelRasterState.get().llCurrentRun.addListener(new ListenerList.Listener<ModelRasterState>()
			{
				@Override public void handleEvent(ModelRasterState state, Object data)
				{
					setState(State.RASTER, isRasterBusy(state));
				}
			});
			setState(State.RASTER, isRasterBusy(ModelRasterState.get()));
		}
		if (null != ModelCollectRunList.get())
		{
			ModelCollectRunList.get().addListener(new ModelCollectRunList.Listener()
			{
				@Override public void runListChanged(ModelRunList runList)
				{
					setState(State.COLLECT, isCollectBusy((ModelCollectRunList)runList));
				}
			});
			setState(State.COLLECT, isCollectBusy(ModelCollectRunList.get()));
		}
	}

	public static void handleStateUpdate()
	{
		boolean energyEnabled = DevCollectSettings.getEnableEnergyMove();
		boolean energyBusy = energyEnabled && StateMap.current.motorList.get("energy").state == EpicsDevice.State.MOVING;
		setState(State.MTR_ENERGY, energyBusy);
		setState(State.AUTOCENTER, StateMap.current.autoCenter.cmdRunning);
		setState(State.ENTER_HUTCH, StateMap.current.enterHutch.cmdRunning);
	}
	
	public static void setScreeningState(boolean busy) {
			setState(State.SCREENING, busy);
	}
	
	protected static boolean isScanBusy(ModelScan settings)
	{
		boolean ret =
    	(
    		InterStep.IDLE != settings.getInterStep() ||
    		AutoStep.NONE != settings.getAutoStep()
    	);
		return ret;
	}
	
	protected static boolean isRasterBusy(ModelRasterState runList)
	{
		return RasterState.INACTIVE != runList.getRasterState();
	}
	
	protected static boolean isCollectBusy(ModelCollectRunList runList)
	{
		return CollectState.INACTIVE != runList.getCollectState();
	}
	
	protected static void loadTable()
	{
		Result result = Database.exec(DBName.CONFIG, "SELECT * FROM "+exceptionTable);
		
		allowedLookupMap = new HashMap<Operation,HashMap<State,Boolean>>();
		
		for (HashMap<String, String> row : result.getRows())
		{
			String operationString = row.get("operation");
			Operation operation = Operation.stringToEnum(operationString);

			allowedLookupMap.put(operation, new HashMap<State,Boolean>());
			
			for (String stateString : row.keySet())
			{
				State state = State.stringToEnum(stateString);
				
				if (null != state) // ID and operation columns will be null
				{
    				boolean isAllowed = (row.get(stateString).length()==0);
    				allowedLookupMap.get(operation).put(state, isAllowed);
				}
			}

			// for clients: no operations are allowed when slave
			boolean opsAllowedWhenSlave = DBConfig.serverMode;
			allowedLookupMap.get(operation).put(State.MASTER_SLAVE, opsAllowedWhenSlave);
		}
	}
}
