package gov.anl.gmca.common;

import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.Util;

import java.util.HashSet;
import java.util.concurrent.ConcurrentHashMap;

// This class adds the ability for a Runnable to return the thread that it's running in
// It has no static members and is not intended for tasks that should only be run one at a time
//
// Correct usage:
//
// ExampleCommand cmd = new ExampleCommand();
// cmd.start();
//
// Incorrect usage: (thread will not be tracked)
//
// Thread t = new Thread(new ExampleCommand());
// t.start()
//
// If possible this class should be modified to disallow the incorrect usage.
//
public abstract class Command implements Runnable
{
	private static Thread listenerThread = null;
	private static ConcurrentHashMap<Thread, ListenerList<Command>> threadMap = new ConcurrentHashMap<Thread, ListenerList<Command>>();
	
	protected Thread thread = null;
	protected ListenerList<Command> ll = new ListenerList<Command>(null);
	
	public boolean isAlive()
	{
		if (null == thread) return false;
		if (!thread.isAlive()) return false;
		return true;
	}

	public synchronized void start()
	{
		verifyListenerThread();
		
		thread = new Thread(this, "Command");
		thread.start();
		ll.notifyListeners();
		
		threadMap.put(thread,ll);
	}
	
	public void waitFor()
	{
		while (isAlive()) Util.sleep(10);
	}
	
	public abstract void run();
	public void requestStop(){};
	
	private void verifyListenerThread()
	{
		if (null != listenerThread && listenerThread.isAlive()) return;
		
		listenerThread = new Thread(new Runnable(){public void run()
		{
			for (Thread thread : new HashSet<Thread>(threadMap.keySet()))
			{
				if (!thread.isAlive())
				{
					threadMap.get(thread).notifyListeners();
					threadMap.remove(thread);
				}
			}
		}}, "CommandListener");
	}
}
