package gov.anl.gmca.common;

import gov.anl.gmca.io.Database;
import gov.anl.gmca.io.Database.DBName;
import gov.anl.gmca.io.Database.Result;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PV_BLUICE;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.ListenerList;
import gov.aps.jca.dbr.DBR;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.StringTokenizer;
import java.util.concurrent.Semaphore;

// TODO: read rows in to actual variables, so both missing or invalid values can be checked for at startup
public class DBConfig
{
	public static boolean serverMode = false;
	public static boolean sim5 = false;

	public static enum ConfigKey
	{
		alignMinibeamFlags,
		alignpin_location,
		app_beamlineID,
		app_beamlineName,
		app_collimatorSet,
		app_imgsrvIP,
		autoRasterZMin_um,
		beamsizebox_H,
		beamsizebox_V,
		beam_size_factor,
		bigEnergyMove_flags,
		check_feedback_aps_min_current,
		check_feedback_v_high,
		check_feedback_v_low,
		detector_back_pos,
		detector_serial,
		edgeScan_fakeFlag,
		edgeScan_idFlag,
		edgeScan_scanFlag,
		enableCollimatorMove,
		enableDetectorMove,
		enableEnergyMove,
		energy1,
		filter_unitcell,
		fluoRaster_flags,
		header_check_enable,
		header_check_timeout_sec,
		header_check_tolerance_deg,
		header_check_tolerance_sec,
		hitfinder_c,
		hitfinder_p,
		host_ppu,
		host_spotfinder,
		monooffset,
		monooffset3,
		ppu_clear_ramdisk,
		processing_mindegrees,
		robotWarmupMarginSec,
		processing_scaling_software,
		speed_limit_gonio_omega,
		speed_limit_sample_x,
		speed_limit_sample_y,
		speed_limit_sample_z,
		spotfinder_enable_collect,
		spotfinder_enable_remote,
		spotfinder_local_skip_rate,
		spotfinder_remote_skip_rate,
		spotfinder_skip_mode,
		unduflag,
		yag_location;
		
		@Override public String toString()
		{
			return this.toString().replace('$', '.');
		}
	};
	
	private static Semaphore mapLock = new Semaphore(1);
	public static ListenerList<DBConfig> llValueChange = new ListenerList<DBConfig>(null);
	
	public static enum Type { FLOAT, INT, ENUM, BOOLEAN, STRING, HIDE };
	
	public static class Format
	{
		public Type type;
		public int precision;
		public String[] enums;
		
		public Format(Type type) 		{ this.type=type; }
		public Format(int precision) 	{ this.type=Type.FLOAT; this.precision=precision; }
		public Format(String[] enums) 	{ this.type=Type.ENUM; this.enums=enums; }
		
		public Format(String formatString)
		{
			try
			{
				char c = formatString.charAt(0);
				switch (c)
				{
				case 'f':
					type=Type.FLOAT;
					precision=Integer.parseInt(formatString.substring(1));
					break;
				case 'i':
					type=Type.INT;
					break;
				case 'e':
					type=Type.ENUM;
					StringTokenizer st = new StringTokenizer(formatString.substring(1),"(),");
					ArrayList<String> values = new ArrayList<String>();
					while (st.hasMoreTokens()) values.add(st.nextToken());
					enums = values.toArray(new String[values.size()]);
					break;
				case 'b':
					type=Type.BOOLEAN;
					break;
				case 's':
					type=Type.STRING;
					break;
				case 'h':
					type=Type.HIDE;
					break;
				}
			}
			catch (Exception e)
			{
				type=Type.STRING;
			}
		}
	}
	
	public static boolean isConnected()
	{
		return (null != valueMap);
	}
	
    protected static HashMap<String,String> valueMap = null;
    protected static HashMap<String,Format> formatMap = null;

	public static String get(String key)
	{
		//if (key.equals("bluice_server_ip")) return "10.20.103.207";

		if (null == valueMap) update();
		
		String ret = null;
		
		getLock();
		try 				{ ret = valueMap.get(key); 	}
		catch (Throwable t) { t.printStackTrace(); 		}
		finally				{ releaseLock(); 			}
		
		return ret;
	}
	
	public static double getDouble(String key)
	{
		return Double.parseDouble(get(key));
	}
	
	public static boolean getBoolean(String key)
	{
		return get(key).equals("1") || get(key).equals("true");
	}
	
	public static boolean set(String key, String value)
	{
		Result result = null;
		getLock();
		try
		{
			result = Database.exec(DBName.CONFIG, "update Config set Value='"+value+"' where ID='"+key+"'");
			valueMap.put(key, value);
		}
		catch (Throwable t) { t.printStackTrace(); 		}
		finally				{ releaseLock(); 			}

		Epics.putInteger(PV_BLUICE.reloadSQL, 0);
		Epics.putInteger(PV_BLUICE.reloadSQL, 1);
		
		return result.isOK();
	}
	
	public static Format getFormat(String key)
	{
		Format ret = null;

		getLock();
		try 				{ ret = formatMap.get(key); }
		catch (Throwable t) { t.printStackTrace(); 		}
		finally				{ releaseLock(); 			}
		
		return ret;
	}
	
	public static ArrayList<String> getKeyListCopy()
	{
		ArrayList<String> keyList = null;
		
		getLock();
		try
		{
			keyList = new ArrayList<String>(valueMap.keySet());
			Collections.sort(keyList);
		}
		catch (Throwable t) { t.printStackTrace(); 		}
		finally				{ releaseLock(); 			}
		
		return keyList;
	}
	
	private static void getLock()
	{
		mapLock.acquireUninterruptibly();
	}
	
	private static void releaseLock()
	{
		mapLock.release();
	}
	
	public static void update()
	{
		getLock();
		
		try
		{
			if (null == valueMap) valueMap = new HashMap<String,String>();
			else valueMap.clear();
			
			if (null == formatMap) formatMap = new HashMap<String,Format>();
			else formatMap.clear();
	
			System.out.println("Loading Config table from SQL");
	    	Result pmaResult = Database.exec(DBName.CONFIG, "select * from Config");
			for (int i=0;i<pmaResult.numRows();i++)
			{
				String key = pmaResult.getRow(i).get("ID");
				String value = pmaResult.getRow(i).get("Value");
				String formatString = pmaResult.getRow(i).get("TypeInfo");
				
				valueMap.put(key, value);
				formatMap.put(key, new Format(formatString));
			}
		}
		catch (Throwable t) { t.printStackTrace(); 	}
		finally				{ releaseLock();		}
		
		llValueChange.notifyListeners();
	}
	
	public static void connectListener()
	{
		Epics.addMonitor(PV_BLUICE.reloadSQL, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				if (Epics.dbrToInt(value)==1)
				{
					Log.note("Reloading SQL config");
					update();
				}
			}
		});
	}
}
