package gov.anl.gmca.common;

import gov.anl.gmca.common.AppState.Operation;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.Util;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * These devices can move to a position specified by a floating point number.
 *
 * A waitFor() method waits for the move to complete.
 * 
 * All hutch motor widgets are tied to a DevMove device.  An example of a non-DevMove device is the FPE.
 */
public abstract class DevMove extends EpicsDevice
{
	///////////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////// move handle & result /////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////////
	
	public static class MoveCmd implements Runnable
	{
		protected Thread moveThread = null;
		protected DevMove dev = null;
		protected Object dest;
		protected MoveResult result = new MoveResult(ResultType.NOT_STARTED);
		public MoveResult getResult() { return result; }
		
		public MoveCmd(DevMove _dev, Object _dest)
		{
			dev=_dev;
			dest=_dest;
		}
		
		public MoveCmd start()
		{
			moveThread = new Thread(this, dev.getDeviceName()+" move");
			moveThread.start();
			
			dev.llMoveStartStop.notifyListeners();
			
			return this;
		}
		
		public void run()
		{
			if (!dev.moveChecks(dest))
			{
				result = new MoveResult(ResultType.INVALID);
				System.out.println("Device "+dev.getDeviceName()+" move failed");
			}
			else
			{
				dev.moveCmd(dest,result);
			}
			
			// TODO: do this more efficiently, without requiring a second thread per move
			new Thread(new Runnable(){public void run()
			{
				while (moveThread.isAlive()) Util.sleep(10);
				moveThread = null;
				System.out.println("Notifying listeners for device "+dev.getDeviceName());
				dev.llMoveStartStop.notifyListeners();
			}}, "Move Completion").start();
		}
		
		public MoveResult waitFor()
		{
			while (null != moveThread)
			{
				Util.sleep(10);
			}
			return result;
		}
		
		public boolean isComplete()
		{
			return (null == moveThread || !moveThread.isAlive());
		}
	}
	
	public MoveCmd startMove(Object dest)
	{
		moveCmd = new MoveCmd(this, dest);
		moveCmd.start();
		return moveCmd;
	}
	
	public enum ResultType { NOT_STARTED, INVALID, SUCCESS, ABORTED, ERROR };
	public static class MoveResult
	{
		public MoveResult(ResultType _type) { type = _type; msg = null; }
		public MoveResult(ResultType _type, String _msg) { type = _type; msg = _msg; }
		
		public boolean isOK() { return ResultType.SUCCESS == type; }
		
		public ResultType type;
		public String msg;
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////// Interface ////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////////////

	protected long lastMoveTime = 0;
	public long getLastMoveTime() { return lastMoveTime; }
	
	MoveCmd moveCmd = null;
	
	public ListenerList<DevMove> llMoveStartStop;
	
	public boolean inLocalMove()
	{
		return (null != moveCmd &&
				null != moveCmd.moveThread &&
				moveCmd.moveThread.isAlive());
	}
	
	/**
	 * When moveTo() returns true, the device should be in the moving state.  This
	 * guarantees that an immediately subsequent waitFor() call will in fact
	 * wait for the move and not return before the move starts.
	 */
	public abstract boolean moveChecks(Object dest);
	public abstract void 	moveCmd(Object dest, MoveResult result);
	
	public abstract boolean getMoving();
	
	// convenience function used by moveCmd
	protected void setStateAndResult(MoveResult result, State state, ResultType resultType, String resultMsg)
	{
		setState(state);
		result.type = resultType;
		result.msg = resultMsg;
	}
	
	protected void setStateAndResult(MoveResult result, State state, ResultType resultType)
	{
		setStateAndResult(result, state, resultType, null);
	}
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////// Public methods ///////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////////////
	
	public boolean isMoving()
	{
		// check if the move thread is active
		if (null != moveCmd && null != moveCmd.moveThread && moveCmd.moveThread.isAlive())
		{
			return true;
		}
		
		// make sure EPICS updates are processed
		try
		{
			Epics.poll();
		}
		catch (Exception e) { e.printStackTrace(); }
		
		// check the InPos PV
		return getMoving();
	}

	// this relies on child classes updating lastMoveTime
	public boolean wasMoving(double secondsAgo)
	{
		long now = System.currentTimeMillis();
		long msSinceLastMove = now - lastMoveTime;
		double secSinceLastMove = (double)msSinceLastMove / 1000.0;
		return (secSinceLastMove <= secondsAgo);
	}
	
	// loop until isMoving() returns false
	public void waitForMove()
	{
		do
		{
			try
			{
				Thread.sleep(10);
				Epics.flushIO();
			}
			catch (Exception e) { e.printStackTrace(); }
		}
		while (isMoving());
	}
	
	public abstract void abort();
	
	public static ArrayList<DevMove> getAll()
	{
		return new ArrayList<DevMove>(devList);
	}
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////// End Interface ///////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// get the AppState operation for this device
	protected static final Object[][] operationGroups =
	{
		{ Operation.MOVE_ENERGY,	"energy",		"attenuation" },
		{ Operation.MOVE_DETECTOR,	"detector_z",	"detector_2t" },
		{ Operation.MOVE_SAMPLE,	"sample_x",		"sample_y",		"sample_z",		"gonio_omega" },
		{ Operation.MOVE_OTHER,		"beam_size_x",	"beam_size_y",	"beamstop_z" },
		{ Operation.MOVE_ZOOM,		"camera_zoom",	"camera_zoom_2" }
	};
	
	public Operation getOperation()
	{
		for (Object[] group : operationGroups)
		{
			Operation op = (Operation)group[0];
			for (int i=1;i<group.length;i++)
			{
				String s = (String)group[i];
				if (s.equals(getDeviceName())) return op;
			}
		}
		return null;
	}
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////// End Interface ///////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////////////

	protected static List<DevMove> devList = Collections.synchronizedList(new ArrayList<DevMove>());
	
	public DevMove(String motorID)
	{
		super(motorID);
		devList.add(this);
		llMoveStartStop = new ListenerList<DevMove>(this);
	}
}
