package gov.anl.gmca.common;

import gov.anl.gmca.common.ui.NBMessageBox;
import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PV_BLUICE;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.users.ModelUserTable.ClientListener;
import gov.anl.gmca.util.Util;
import gov.aps.jca.dbr.DBR;
import org.eclipse.swt.SWT;

import java.util.ArrayList;

// The JBluIce request/response PVs allow querying the master client and getting a response.
// Subsequent requests while a request is being processed are queued, but to sort out the responses
// unique keys are needed.
//
// Only the master client processes requests.  If there is no master, then the next JBI client to
// become master will process the request when it becomes master.  If a second client becomes master
// by force, the first client will cease processing (popups are hidden), and the new master client
// starts handling the request (shows a popup for example).
//
// The protocol to submit a query is as follows:
//
// 1) Wait until the request PV is clear.  Requests sent while another is being processed are ignored.
// 2) Send a request to the request PV - this will clear the response PV
// 3) Watch the response PV for your response
//
// Supported queries:
//
// 1) popup_question [key=...] "Title text" "Message text" "First choice" "Second choice" ["Third choice"...]
// >> Quotes are optional for single word arguments like OK
// >> Escaped quotes inside quoted arguments are not supported
// >> Response is an integer corresponding to the selected choice
// >> If a key is provided then the response is in the form "unique-key 0", with quotes omitted
//
// 2) popup_error [key=...] "Title text" "Message text" "First choice" "Second choice" ["Third choice"...]
// >> Same format as #1 but the icon is an exclamation mark instead of a question mark
//
public class DevScriptInterface
{
	private static class Request
	{
		public boolean isError;
		public String title;
		public String message;
		public ArrayList<String> choices = new ArrayList<String>();
		public String key;
		public Thread thread;
		
		// override in other request types
		public void process()
		{
			requests.add(this);
			openPopup();
		}
		
		public void openPopup()
		{
			messageBox = new NBMessageBox(GUI.getShell(), isError?SWT.ICON_ERROR:SWT.ICON_QUESTION);
			messageBox.setText(title);
			messageBox.setMessage(message);
			messageBox.setCustomButtons(choices.toArray(new String[]{}));
			
			thread = new Thread(new Runnable(){public void run()
			{
				int retVal = messageBox.open();
				
				String response = "";
				if (null!=key) response += key+" ";
				response += retVal;
				
				Epics.put(PV_BLUICE.request, "");
				Epics.put(PV_BLUICE.response, response);
			}});
			thread.start();
		}
		
		public void cancelPopup()
		{
			messageBox.close();
		}
	}
	
	private static ArrayList<Request> requests = new ArrayList<Request>();
	
	private static NBMessageBox messageBox = null;
	
	private static void handleRequest(String reqStr)
	{
		// parse the request type
		ArrayList<String> reqTokens = Util.parseStringWithQuotes(reqStr);
		if (reqTokens.size()==0) return;
		String cmd = reqTokens.get(0);
		
		// fully parse the request into a request object
		Request req = null;
		if (cmd.equals("popup_question"))
		{
			req = parsePopup(reqTokens, true);
		}
		else if (cmd.equals("popup_error"))
		{
			req = parsePopup(reqTokens, false);
		}
		
		// process the request
		if (null!=req) req.process();
	}
	
	private static Request parsePopup(ArrayList<String> requestTokens, boolean isError)
	{
		Request req = new Request();

		// get unique key if present
		if (requestTokens.size()<2) return null;
		boolean hasKey = requestTokens.get(1).indexOf("key=") == 0;
		if (hasKey) req.key = requestTokens.get(1).substring(4);

		// check argument count
		if (requestTokens.size()<(hasKey?5:4)) return null;
		
		// parse remaining arguments
		req.title = requestTokens.get(hasKey?2:1);
		req.message = requestTokens.get(hasKey?3:2);
		for (int i=hasKey?4:3; i<requestTokens.size(); i++)
		{
			req.choices.add(requestTokens.get(i));
		}

		// return
		return req;
	}

	private static void handleResponse(String response)
	{
	}
	
	public static void init()
	{
		Epics.addMonitor(PV_BLUICE.request, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				handleRequest(Epics.dbrToString(value));
			}
		});
		Epics.addMonitor(PV_BLUICE.response, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				handleResponse(Epics.dbrToString(value));
			}
		});
		Client.addListener(new ClientListener()
		{
			@Override public void masterChanged(boolean isMaster)
			{
				if (!isMaster)
				{
					// cancel popups?
				}
			}
		});
	}
}
