package gov.anl.gmca.common.ui;

import gov.anl.gmca.common.DevMoveFloat;
import gov.anl.gmca.common.DevMoveFloat.PositionListener;
import gov.anl.gmca.common.EpicsDevice;
import gov.anl.gmca.common.Limits;
import gov.anl.gmca.common.Limits.LimitsResult;
import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.Util;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.*;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Text;

public class WdgControlEntry extends WdgControl implements TraverseListener, ModifyListener, FocusListener, SelectionListener
{
	public 		enum 		EntryType 	{ RW_COMBO, RO_COMBO, TEXT };
	protected 	EntryType 	entryType;
	protected	boolean		popupOnError;
	protected	boolean		highlighted;
	protected	boolean		entryEnabled = false;
	
	public Control entry;
	
	@Override public void addTraverseListener(TraverseListener tl)
	{
		entry.addTraverseListener(tl);
	}
	
	public static class ValidateResult
	{
		public ValidateResult(boolean isValid) { this.isValid=isValid; this.errorMsg=null; }
		public ValidateResult(boolean isValid, String errorMsg) { this.isValid=isValid; this.errorMsg=errorMsg; }
		
		public boolean isValid;
		public String errorMsg;
	}
	
	public static abstract class ModelInterface extends WdgControl.ModelInterface
	{
		public boolean doHighlight(double d) 		{ return !Util.nearlyEquals((Double)getDeviceValue(),d,getDecimals()); }
		public ValidateResult validate(String s)	{ return defaultValidate(s,getLimits()); }
	}
	
	public WdgControlEntry(Composite parent, int style, EntryType _entryType, boolean _popupOnError, WdgControlEntry.ModelInterface _mi)
	{
		super(parent, style, _mi);
		init(parent, style, _entryType, _mi);
		popupOnError=_popupOnError;
	}
	
	public WdgControlEntry(Composite parent, int style, EntryType _entryType, WdgControlEntry.ModelInterface _mi)
	{
		super(parent, style, _mi);
		init(parent, style, _entryType, _mi);
		popupOnError=false;
	}
	
	protected void init(Composite parent, int style, EntryType _entryType, WdgControlEntry.ModelInterface _mi)
	{
		entryType=_entryType;
		
		setLayout(new FillLayout());

		entry = new_entry(this);
		if (entryType==EntryType.RO_COMBO) entry.setBackground(GUI.COLOR_WHITE);
		entry.addTraverseListener(this);
		entry.addFocusListener(this);
		entry_addModifyListener(entry, this);
		entry_addSelectionListener(entry, this);
		
		EpicsDevice dev = mi.getDevice();
		if (null != dev && dev instanceof DevMoveFloat)
		{
			DevMoveFloat dmf = (DevMoveFloat)dev;
			dmf.addListener(new PositionListener()
			{
				@Override
				public void handleActPosChange(DevMoveFloat dev, double newPos)
				{
					GUI.asyncExec(new Runnable(){public void run()
					{
						updateHighlight();
					}});
				}
			});
		}
	}

	@Override
	public void setEnabled(boolean enabled)
	{
		entry.setEnabled(enabled);
		entryEnabled=enabled;
		updateHighlight();
	}
	
	@Override
	public void setFont(Font font)
	{
		entry.setFont(font);
	}
	
	public String getText()
	{
		return entry_getText(entry);
	}
	
	public void setText(Object newVal)
	{
		entry_setText(entry,mi.formatValue(newVal));
	}
	
	public int getSelection()
	{
		if (EntryType.RO_COMBO==entryType)
		{
			return ((Combo)entry).getSelectionIndex();
		}
		return -1;
	}
	
	public void setSelection(int index)
	{
		if (EntryType.RO_COMBO==entryType)
		{
			((Combo)entry).select(index);
		}
	}
	
	@Override
	public void keyTraversed(TraverseEvent te)
	{
		switch (te.detail)
		{
		// return to the model's value
		case SWT.TRAVERSE_ESCAPE:
			entry_setText(entry, mi.formatValue(mi.getModelValue()));
			break;
			
		// validate and submit the new value to the device
		case SWT.TRAVERSE_RETURN:
			submitToDevice();
			break;
		}
	}

	@Override
	public void focusLost(FocusEvent fe)
	{
		submitToDevice();
	}

	@Override
	public void modifyText(ModifyEvent arg0)
	{
		updateHighlight();
	}
	
	@Override public void widgetSelected(SelectionEvent se)
	{
		submitToDevice();
	}
	
	protected Object validateText()
	{
		String inputText = entry_getText(entry);
		ValidateResult result = ((WdgControlEntry.ModelInterface)mi).validate(inputText);
		
		if (result.isValid)
		{
			Object modelValue = mi.getModelValue();
			if (modelValue instanceof Double)
			{
				try
				{
					double doubleVal = Double.parseDouble(inputText);
					
					if (!Util.nearlyEquals(doubleVal, (Double)modelValue, mi.getDecimals()))
					{
						return doubleVal;
					}
				}
				catch (NumberFormatException nfe)
				{
					return inputText;
				}
			}
			else if (modelValue instanceof Integer)
			{
				try
				{
					int intVal = Integer.parseInt(inputText);
					
					if (intVal != ((Integer)modelValue).intValue())
					{
						return intVal;
					}
				}
				catch (NumberFormatException nfe)
				{
					return inputText;
				}
			}
			else if (modelValue instanceof String && ((String)modelValue).equals("\u221e"))
			{
				return Double.POSITIVE_INFINITY;
			}
			else
			{
				return inputText;
			}
		}
		else
		{
			if (popupOnError)
			{
				Util.popupError("Entry error", result.errorMsg);
			}
			else
			{
				Log.warn(result.errorMsg);
			}
			return null;
		}
		return null;
	}
	
	protected void submitToDevice()
	{
		Object value = validateText();
		if (null != value)
		{
			mi.setValue(value);
			entry_setText(entry, mi.formatValue(value));
		}
		else
		{
			entry_setText(entry, mi.formatValue(mi.getModelValue()));
		}
	}
	
	@Override public boolean updateModel()
	{
		Object value = validateText();
		if (null != value)
		{
			mi.updateModel(value);
			entry_setText(entry, mi.formatValue(value));
			return true;
		}
		else
		{
			entry_setText(entry, mi.formatValue(mi.getModelValue()));
			return false;
		}
	}
	
	// is the text a valid number within limits?
	public static ValidateResult defaultValidate(String s, Limits limits)
	{
		if (s.equals("\u221e")) return new ValidateResult(true);
		
		try
		{
			double d = Double.parseDouble(s);
			
			if (null != limits)
			{
				LimitsResult result = limits.testValue(d);
				if (!result.isOK())
				{
					return new ValidateResult(false, result.getReason());
				}
			}
			return new ValidateResult(true);
		}
		catch (NumberFormatException nfe)
		{
			return new ValidateResult(false, "\""+s+"\" is not a valid number");
		}
	}

	public void updateHighlight()
	{
		if (entry.isDisposed()) return;
		
		if (!entryEnabled)
		{
			entry.setForeground(GUI.COLOR_DARK_GRAY);
			return;
		}
		
		try
		{
			// read the widget text
			String inputText = entry_getText(entry);
	
			// if there's no text there's nothing to color
			if (inputText.length()==0) return;
			
			// widgets with no device value are never red
			if (null == mi.getDeviceValue())
			{
				entry.setForeground(GUI.COLOR_BLACK);
				highlighted=false;
				return;
			}
			
			// invalid? red color
			if (!((WdgControlEntry.ModelInterface)mi).validate(inputText).isValid)
			{
				entry.setForeground(GUI.COLOR_RED);
				highlighted=true;
				return;
			}
			
			try
			{
				double widgetValue;
				if (inputText.equalsIgnoreCase("true")) widgetValue=1;
				else if (inputText.equalsIgnoreCase("false")) widgetValue=0;
				else if (inputText.equalsIgnoreCase("none")) widgetValue=Double.NaN;
				else widgetValue = Double.parseDouble(inputText);
				highlighted =
					((WdgControlEntry.ModelInterface)mi).doHighlight(widgetValue);
				entry.setForeground(highlighted ? GUI.COLOR_RED : GUI.COLOR_BLACK);
			}
			catch (NumberFormatException nfe)
			{
				entry.setForeground(GUI.COLOR_RED);
			}
		}
		catch (Exception e) { e.printStackTrace(); }
	}

	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////// methods for multiple control/data types /////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	protected Control new_entry(Composite parent)
	{
		Control ret = null;
		
		switch (entryType)
		{
		case RW_COMBO:
		case RO_COMBO:
			ret = new Combo(this, (EntryType.RW_COMBO==entryType) ? SWT.NONE : SWT.READ_ONLY);
			ret.setEnabled(false);
			
			reloadChoices((Combo)ret);
			break;
			
		case TEXT:
			ret = new Text(this, SWT.BORDER);
			ret.setEnabled(false);
			break;
		}
		
		ret.setFont(GUI.FONT_COLLECTRUN_ENTRY);
		return ret;
	}
	
	protected void entry_addModifyListener(Control entry, ModifyListener ml)
	{
		switch (entryType)
		{
		case RW_COMBO:
		case RO_COMBO:
			((Combo)entry).addModifyListener(ml);
			break;
		case TEXT:
			((Text)entry).addModifyListener(ml);
			break;
		}
	}
	
	protected void entry_addSelectionListener(Control entry, SelectionListener sl)
	{
		switch (entryType)
		{
		case RW_COMBO:
		case RO_COMBO:
			((Combo)entry).addSelectionListener(sl);
			break;
		case TEXT:
			break;
		}
	}
	
	protected String entry_getText(Control entry)
	{
		switch (entryType)
		{
		case RW_COMBO:
		case RO_COMBO:
			return ((Combo)entry).getText();
			
		case TEXT:
			return ((Text)entry).getText();
			
		default:
			return null;
		}
	}
	protected void entry_setText(Control entry, String text)
	{
		switch (entryType)
		{
		case RW_COMBO:
		case RO_COMBO:
			((Combo)entry).setText(text);
			break;
			
		case TEXT:
			((Text)entry).setText(text);
			break;
		}
	}

	@Override public void focusGained(FocusEvent fe) {}
	@Override public void widgetDefaultSelected(SelectionEvent arg0) {}

	protected void reloadChoices(Combo combo)
	{
		if (null != mi.getChoices())
		{
			combo.remove(0,combo.getItemCount()-1); // removeAll() clears the field too
			for (String choice : mi.getChoices())
			{
				combo.add(choice);
			}
		}
		if (null != mi.getFloatChoices())
		{
			combo.remove(0,combo.getItemCount()-1);
			for (double d : mi.getFloatChoices())
			{
				combo.add(mi.formatValue(d));
			}
		}
	}
	
	public void reloadChoices()
	{
		if (entry instanceof Combo) reloadChoices((Combo)entry);
	}

	public String[] getComboItems()
	{
		Combo combo = (Combo)entry;
		return combo.getItems();
	}

}
