package gov.anl.gmca.gui.options;

import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.ListenerList;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.*;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.*;

public class WdgDBEntry
{
	private DBConfig.Format format;
	private Control control;
	private String key;
	private Composite parent;
	private String currentValue;
	private Composite leftSpacer;
	private boolean scrollDetected=false;
	private ListenerList.Listener<DBConfig> dbListener = null;
	
	public WdgDBEntry(Composite parent, final String key)
	{
		// init
		this.key=key;
		this.format=DBConfig.getFormat(key);
		this.parent=parent;
		this.currentValue=DBConfig.get(key);
		
		// left spacer
		leftSpacer = WdgStaffOptions.addLeftSpacer(parent);
		
		// field label
		Label lblKey = new Label(parent, SWT.LEFT);
		lblKey.setLayoutData(new GridData());
		lblKey.setText(key);

		// field control widget
		createEntryWidget();
		GridData gd = new GridData(SWT.FILL, SWT.DEFAULT, true, false);
		gd.widthHint = 200;
		control.setLayoutData(gd);

		// right spacer
		WdgStaffOptions.addRightSpacer(parent);
		
		// listen for DB updates
		dbListener = new ListenerList.Listener<DBConfig>()
		{
			@Override public void handleEvent(DBConfig object, Object data)
			{
				final String updatedValue=DBConfig.get(key);
				if (!updatedValue.equals(currentValue))
				{
					GUI.asyncExec(new Runnable(){public void run()
					{
						currentValue = updatedValue;
						setComboToCurrentValue();
					}});
				}
			}
		};
		DBConfig.llValueChange.addListener(dbListener);
		parent.addDisposeListener(new DisposeListener()
		{
			@Override public void widgetDisposed(DisposeEvent de)
			{
				DBConfig.llValueChange.removeListener(dbListener);
			}
		});
	}

	private void handleUIChange()
	{
		boolean matchesDBValue = currentValue.equals(DBConfig.get(key));
		leftSpacer.setBackground(matchesDBValue ? GUI.COLOR_LIGHT_GRAY : GUI.COLOR_RED);
	}
	
	private void commitChange(String value)
	{
		boolean success = DBConfig.set(key, value);
		if (success)
			WdgStaffOptions.get().dbLog.println("Wrote to SQL: "+key+"='"+value+"'", Log.STYLE_NOTE);
		else
			WdgStaffOptions.get().dbLog.println("Error writing to SQL ("+key+"='"+value+"')", Log.STYLE_ERROR);
	}
	
	private void createEntryWidget()
	{
		switch (format.type)
		{
		case FLOAT:
		case INT:
		case BOOLEAN:
		case STRING:
			control = new Text(parent, SWT.BORDER);
			((Text)control).setText(currentValue);
			((Text)control).addVerifyListener(new VerifyListener()
			{
				@Override public void verifyText(VerifyEvent ve)
				{
					ve.doit = !ve.text.contains("\"");
					if (!ve.doit) Log.error("Quotes are not allowed in SQL strings");
				}
			});
			
			((Text)control).addModifyListener(new ModifyListener()
			{
				@Override public void modifyText(ModifyEvent me)
				{
					currentValue = ((Text)control).getText();
					handleUIChange();
				}
			});
			
			((Text)control).addTraverseListener(new TraverseListener()
			{
				@Override public void keyTraversed(TraverseEvent te)
				{
					if (te.detail == SWT.TRAVERSE_RETURN)
					{
						commitChange(currentValue);
						handleUIChange();
					}
				}
			});
			
			break;
			
		case ENUM:
			control = new Combo(parent, SWT.READ_ONLY);
			((Combo)control).addMouseWheelListener(new MouseWheelListener()
			{
				@Override public void mouseScrolled(MouseEvent me)
				{
					scrollDetected=true;
				}
			});
			
			// fill the combo with enums from the format object
			for (int i=0;i<format.enums.length;i++)
			{
				((Combo)control).add(format.enums[i]);
			}
			setComboToCurrentValue();
			
			((Combo)control).addSelectionListener(new SelectionListener()
			{
				@Override public void widgetSelected(SelectionEvent arg0)
				{
					if (scrollDetected)
					{
						// don't let scroll change values
						setComboToCurrentValue();
						scrollDetected=false;
					}
					else
					{
						int index = ((Combo)control).getSelectionIndex();
						currentValue = format.enums[index];
						commitChange(currentValue);
					}
				}
				@Override public void widgetDefaultSelected(SelectionEvent arg0) {}
			});
			break;
		}
	}
	
	private void setComboToCurrentValue()
	{
		for (int i=0;i<format.enums.length;i++)
		{
			if (format.enums[i].equals(currentValue))
			{
				((Combo)control).select(i);
			}
		}
	}
}
