package gov.anl.gmca.hutch;

import gov.anl.gmca.common.AppState;
import gov.anl.gmca.common.AppState.State;
import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.common.DevScript;
import gov.anl.gmca.gui.WdgEmergencyStop;
import gov.anl.gmca.io.Disk;
import gov.anl.gmca.io.Disk.File;
import gov.anl.gmca.io.ExecTask;
import gov.anl.gmca.io.PVLookup.PV_BLUICE;
import gov.anl.gmca.io.PVLookup.PV_MISC;
import gov.anl.gmca.io.hw.Epics;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.Point2D;
import gov.anl.gmca.util.Util;
import gov.aps.jca.dbr.DBR;
import org.eclipse.swt.SWT;

import java.io.IOException;
import java.text.Format;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;

public class DevAlignMinibeam extends DevScript
{
	public ListenerList<DevAlignMinibeam>
		llData, 	// sends ArrayList<Point2D> with new points, or null for clear
		llInfo;
	
	private static final String optimizeBeam_dir	= System.getProperty("user.home")+"/bluice_temp/";
	private static final String cmdFile 			= "alignMiniBeam.pl";
	
	private boolean scriptReportsRunning = false;
	protected static DevAlignMinibeam inst = null;
	public static DevAlignMinibeam get() { return inst; }

	private String logPath;
	private Format dateFormat;
	
	//////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////// end variables //////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////
	
	public DevAlignMinibeam()
	{
		super("alignMinibeam");
		inst = this;
		llData = new ListenerList<DevAlignMinibeam>(this);
		llInfo = new ListenerList<DevAlignMinibeam>(this);

		determineLogPath();
		
		dateFormat = new SimpleDateFormat("yyyy-MM-dd hh:mmaa");
	    WdgEmergencyStop.registerProcess("alignMiniBeam.pl");

	    connectListeners();
	}

	private void determineLogPath()
	{
		// if we used a new Calendar for each loop, there is a very tiny chance
		// that we would roll over midnight and not check the correct paths
		Calendar now = Calendar.getInstance();
		
		// look for log for 1-2 days ago
		for (int daysAgo = 1; daysAgo <= 2; daysAgo++)
		{
			String testPath = generatePath(now, daysAgo);
			if (Disk.exists(testPath))
			{
				logPath = testPath;
				break;
			}
		}
		
		// if none, then make a new log or append to an existing log for today
		if (null == logPath)
		{
			logPath = generatePath(now, 0);
		}
	}
	
	private String generatePath(Calendar cal, int daysAgo)
	{
		// get date object for daysAgo parameter
		cal.add(Calendar.DAY_OF_MONTH, -1*daysAgo);
		Date testDate = cal.getTime();
		
		// generate path for that date
		Format filenameFormat = new SimpleDateFormat("yyyy-MM-dd");
		String dateString = filenameFormat.format(testDate);
		String path = System.getProperty("user.home")+"/mono-feedback-"+dateString+".log";
		
		// add back the days for the next loop
		cal.add(Calendar.DAY_OF_MONTH, daysAgo);
		
		return path;
	}
	
	private void initTask()
	{
		while (null!=task && task.isRunning()) { System.out.println("Waiting for existing align minibeam script to finish..."); Util.sleep(100); }

		Epics.putDouble(PV_BLUICE.alignLastMirrorV, -1000.0);
		Epics.putDouble(PV_BLUICE.alignLastMirrorH, -1000.0);
		
		String args = DBConfig.get("alignMinibeamFlags");
		task = ExecTask.new_ExecTask(Util.getPerlPrefix()+cmdFile+" "+args, cmdFile, true);
		//task = ExecTask.new_ExecTask("/mnt/beegfs/mhilgart/test-script.sh", cmdFile, true);
	}
	
	public void asyncExec() { initTask(); _asyncExec(); }
	public void syncExec() { initTask(); _syncExec(); }
	
	public Double getVMirrorResult()
	{
		double vMirrorPos = Epics.getDouble(PV_BLUICE.alignLastMirrorV);
		if (vMirrorPos == -1000.0) return null;
		return vMirrorPos;
	}
	
	public Double getHMirrorResult()
	{
		double hMirrorPos = Epics.getDouble(PV_BLUICE.alignLastMirrorH);
		if (hMirrorPos == -1000.0) return null;
		return hMirrorPos;
	}
	
	public static void saveScanInfo()
	{
		try
		{
			// get post-scan values
			long 	javaTime 	= System.currentTimeMillis();
			int 	unixTime 	= (int)(javaTime / 1000L);
			double 	energyPos 	= DevEnergy.getDevice().getPosition();
			double 	p2actPos 	= Epics.getDouble(PV_MISC.p2Pos);
			double 	p2rqsPos 	= Epics.getDouble(PV_MISC.p2rqsPos);
			double 	ctrlVoltage = Epics.getDouble(PV_MISC.ctrlVoltage);
			double	vMirrorPos	= Epics.getDouble(PV_MISC.verticalPiezo);
			double	hMirrorPos	= Epics.getDouble(PV_MISC.horizontalPiezo);
			double 	bpm1		= Epics.getDouble(PV_MISC.bpm1);

			// determine if there was an energy change
			//Double lastOptEnergy = getLastOptEnergy();
			//Boolean energyChange = null==lastOptEnergy ? null : Math.abs(lastOptEnergy-energyPos) > 0.0005;
			
			// save info to EPICS
			Epics.putInteger(PV_BLUICE.alignLastRunTime, unixTime);
			Epics.putDouble(PV_BLUICE.alignLastEnergy, energyPos);
			Epics.putDouble(PV_BLUICE.alignLastMirrorV, vMirrorPos);
			Epics.putDouble(PV_BLUICE.alignLastMirrorH, hMirrorPos);
			
			// write info to log
			Date timestamp = new Date(javaTime);
			String timeString = get().dateFormat.format(timestamp);
			
			ArrayList<String> header = new ArrayList<String>();
			header.add("Time                 En-keV  p2ac-mr  p2rq-mr  ctrl_v  bpm1  mirrorV  mirrorH");
			header.add("-----------------------------------------------------------------------------");
			//          2012-01-01 12:00AM  12.0000  10.7135  10.7141  4.2467  0.78    5.315    4.781
			//          [------18--------]  [--7--] [--8---] [--8---] [--7--] [-5-] [--8---] [--8---]
			//          12345678901234567890123456789012345678901234567890123456789012345678901234567890
			String msg = timeString+"  ";
			msg += Util.sprintf("%7.4f ", energyPos);
			msg += Util.sprintf("%8.4f ", p2actPos);
			msg += Util.sprintf("%8.4f ", p2rqsPos);
			msg += Util.sprintf("%7.4f ", ctrlVoltage);
			msg += Util.sprintf("%5.2f ", bpm1);
			msg += Util.sprintf("%8.3f ", vMirrorPos);
			msg += Util.sprintf("%8.3f ", hMirrorPos);
			//msg += (null==energyChange) ? "---" : energyChange?"Yes":"No";
			
			writeLog(msg, header);
		}
		catch (Exception e) { e.printStackTrace(); }
	}

	/*private Double getLastOptEnergy()
	{
		try
		{
			// check for the file
			if (!Disk.exists(logPath) || !Disk.canRead(logPath) )
			{
				System.out.println("Couldn't read log file "+logPath+", returning null for last opt. energy");
				return null;
			}
	
			// read the file and make sure it's not empty
			ArrayList<String> log = Disk.readText(logPath);
			if (log.size()==0)
			{
				System.out.println("Log file "+logPath+" is empty, returning null for last opt. energy");
				return null;
			}
			
			String lastLine = log.get(log.size()-1);
			StringTokenizer st = new StringTokenizer(lastLine, " ");
			st.nextToken(); // date
			st.nextToken(); // time
			String energyStr = st.nextToken();
			
			return Double.parseDouble(energyStr);
		}
		catch (Exception e)
		{
			e.printStackTrace();
			return null;
		}
	}*/
	
	private static void writeLog(String msg, ArrayList<String> header) throws IOException
	{
		boolean writeHeader = false;
		File logFile = new File(get().logPath);
		
		if (!logFile.exists())
		{
			try
			{
				logFile.createNewFile();
			}
			catch (gov.anl.gmca.io.Disk.IOException ioe)
			{
				throw new IOException("Can't create file");
			}
			writeHeader = true;
		}
		
		if (!logFile.canRead())
		{
			logFile.setReadable(true);
			if (!logFile.canRead()) throw new IOException("Can't make readable");
		}
		if (!logFile.canWrite())
		{
			logFile.setWritable(true);
			if (!logFile.canWrite()) throw new IOException("Can't make writable");
		}
		
		ArrayList<String> contentsToAppend = new ArrayList<String>();
		if (writeHeader)
		{
			for (String line : header) contentsToAppend.add(line+"\n");
		}
		
		contentsToAppend.add(msg+"\n");
		Disk.appendText(get().logPath, contentsToAppend);
	}
	
	public Date getLastRunTime()
	{
		int unixTime = Epics.getInteger(PV_BLUICE.alignLastRunTime);
		if (0 == unixTime) return null;
		
		long javaTime = unixTime*1000L;
		Date d = new Date(javaTime);
		return d;
	}
	
	public double getLastEnergy()
	{
		return Epics.getDouble(PV_BLUICE.alignLastEnergy);
	}
	
	private AlignData lastDataH = null, lastDataV = null;
	public AlignData getDataH() { return lastDataH; }
	public AlignData getDataV() { return lastDataV; }
	
	public static class AlignData
	{
		public int scanFlag, pointCount, orientation;
		public double vLow, vHigh;
		public ArrayList<Point2D> points = new ArrayList<Point2D>();
		public double[] xPoints, yPoints;
		
		public AlignData(int _orientation, double[] dataPV)
		{
			orientation	= _orientation;
			scanFlag 	= dataPV.length>0?(int)dataPV[0]:0;
			pointCount	= dataPV.length>1?(int)dataPV[1]:0;
			vLow		= dataPV.length>2?dataPV[2]:0.0;
			vHigh		= dataPV.length>3?dataPV[3]:0.0;
			xPoints		= new double[pointCount];
			yPoints		= new double[pointCount];
			
			int headerSize = 4;
			for (int pointIndex=0; pointIndex<pointCount; pointIndex++)
			{
				int xIndex = headerSize + pointIndex;
				int yIndex = headerSize + pointCount + pointIndex;
				double x = dataPV[xIndex];
				double y = dataPV[yIndex];
				points.add(new Point2D(x,y));
				xPoints[pointIndex]=x;
				yPoints[pointIndex]=y;
			}
		}
	}
	
	private void handleData(int orientation, double[] dataPV)
	{
		AlignData data = new AlignData(orientation, dataPV);
		if (orientation==SWT.VERTICAL) 	 lastDataV = data;
		if (orientation==SWT.HORIZONTAL) lastDataH = data;
		llData.notifyListeners(data);
	}
	 
	private void connectListeners()
	{
		// set up PV listeners
		Epics.addMonitor(PV_BLUICE.alignDataV, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				handleData(SWT.VERTICAL, Epics.dbrToDoubleArray(value));
			}
		});
		Epics.addMonitor(PV_BLUICE.alignDataH, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				handleData(SWT.HORIZONTAL, Epics.dbrToDoubleArray(value));
			}
		});
		Epics.addMonitor(PV_BLUICE.alignState, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				boolean newVal = Epics.dbrToBoolean(value);
				
				if (!newVal && scriptReportsRunning && Client.isMaster())
				{
					// save only when the PV goes from running to not running
					saveScanInfo();
				}
				
				scriptReportsRunning = newVal;
				handleStateChange();
			}
		});
		Epics.addMonitor(PV_BLUICE.alignLastRunTime, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				llInfo.notifyListeners();
			}
		});
		Epics.addMonitor(PV_BLUICE.alignLastEnergy, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				llInfo.notifyListeners();
			}
		});
		Epics.addMonitor(PV_BLUICE.alignLastMirrorH, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				llInfo.notifyListeners();
			}
		});
		Epics.addMonitor(PV_BLUICE.alignLastMirrorV, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				llInfo.notifyListeners();
			}
		});

		ll.addListener(new ListenerList.Listener<DevScript>()
		{
			@Override public void handleEvent(DevScript object, Object data)
			{
				handleStateChange();
			}
		});
	}
	
	private void handleStateChange()
	{
		boolean isRunning = scriptReportsRunning || isRunning();
		AppState.setState(State.ALIGN_MINIBEAM, isRunning);
	}
}
