package gov.anl.gmca.hutch;

import gov.anl.gmca.collect.DevCollectSettings;
import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.common.DevMoveFloat;
import gov.anl.gmca.common.EpicsDevice;
import gov.anl.gmca.common.Limits;
import gov.anl.gmca.io.ExecTask;
import gov.anl.gmca.io.PVLookup.PV_EIGER;
import gov.anl.gmca.io.hw.Epics;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.Util;
import gov.aps.jca.dbr.DBR;

public class DevEnergy extends DevMoveFloat
{
	protected final static String bigEnergyScript = "bigEnergyMove.pl";
	
	DevPmacMotor monoDev;
	DevUndulator undDev;
	double smallEnergyCutoff_keV;

	public enum MoveType { SCRIPT, DEVICE, NONE };
	protected MoveType moveType = MoveType.NONE;
	public MoveType getMoveType() { return moveType; }
	
	ExecTask moveScript = null;
	private boolean isScriptRunning()
	{
		try
		{
			return (null!=moveScript && moveScript.isRunning());
		}
		catch (NullPointerException npe) { return false; }
	}
	
	@Override public double getInPosBand()
	{
		if (null==monoDev) return super.getInPosBand();
		return monoDev.getInPosBand();
	}
	
	public DevEnergy()
	{
		super("energy");
		
		monoDev = DevPmacMotor.get("mo_En_E");
		undDev = DevUndulator.getDevice();
		smallEnergyCutoff_keV = DBConfig.getDouble("bigEnergyMove_minimum_keV");
		System.out.println(String.format("Started DevEnergy with script minimum move of %.3f keV", smallEnergyCutoff_keV));

		monoDev.addListener(new Listener()
		{
			@Override public void handleStateChange(State state)
			{
				lastMoveTime = monoDev.getLastMoveTime();
				updateCombinedState();
			}

			// simply forward actpos, rqspos and units changes
			@Override public void handleActPosChange(DevMoveFloat dev, double newPos)
			{
				lastMoveTime = monoDev.getLastMoveTime();
				
				DevAttenuation.get().handleActPosChange(false);
				
				for (Listener listener : listenerSet)
				{
					listener.handleActPosChange(dev, newPos);
				}
			}
			@Override public void handleRqsPosChange(double newPos) { for (Listener listener : listenerSet) listener.handleRqsPosChange(newPos);	}
			@Override public void handleUnitsChange(String units) 	{ for (Listener listener : listenerSet) listener.handleUnitsChange(units); 		}
		});
		
		if (null != undDev)
		{
    		undDev.addListener(new Listener()
    		{
    			@Override public void handleStateChange(State state)
    			{
    				updateCombinedState();
    			}
    
    			// ignore actpos, rqspos and units because those will follow the mono device
    			@Override public void handleActPosChange(DevMoveFloat dev, double newPos) {}
    			@Override public void handleRqsPosChange(double newPos) {}
    			@Override public void handleUnitsChange(String units) {}
    		});
		}
	}
	public static DevEnergy getDevice() { return (DevEnergy)EpicsDevice.getDevice("energy"); };

	protected void updateCombinedState()
	{
		State combinedState = null;
		
		if (isScriptRunning())																					combinedState = State.MOVING;
		else if (monoDev.getState() == State.ERROR 	|| (null!=undDev && undDev.getState() == State.ERROR)) 		combinedState = State.ERROR;
		else if (monoDev.getState() == State.OFFLINE || (null!=undDev && undDev.getState() == State.OFFLINE)) 	combinedState = State.OFFLINE;
		else if (monoDev.getState() == State.MOVING 	|| (null!=undDev && undDev.getState() == State.MOVING)) combinedState = State.MOVING;
		else 																									combinedState = State.READY;
		
		if (combinedState == State.READY && moveType == MoveType.DEVICE)
		{
			moveType = MoveType.NONE;
		}

		setState(combinedState);
		
		for (Listener listener : listenerSet) listener.handleStateChange(state);
	}
	
	@Override
	public void moveCmd(double dest, final MoveResult result)
	{
		if (!DevCollectSettings.getEnableEnergyMove())
		{
			System.out.println("Energy moves disabled, returning move success for DevEnergy");
			result.type = ResultType.SUCCESS;
			return;
		}
		
		setState(State.MOVING);
		//updateDetectorEnergy(dest*1000.0);
		
		if ( Math.abs(monoDev.getPosition() - dest) <= smallEnergyCutoff_keV )
		{
			Log.note("Moving energy in BluIce mode");
			
			double startAttenuation = DevAttenuation.get().getFactors();
			Log.note("Saving attenuation="+startAttenuation);
			
			moveType = MoveType.DEVICE;
			
    		MoveCmd undMove = null;
    		
    		MoveCmd monoMove = monoDev.startMove(dest);
    		if (null != undDev)
    		{
    			undMove = undDev.startMove(dest);
    		}
    		else
    		{
    			System.out.println("Undulator move disabled (device doesn't exist)");
    		}
    		
    		monoMove.waitFor();
    		if (null != undDev) { undMove.waitFor(); }
    		
    		// state is set by the mono listener which forwards the state change

    		Log.note("Restoring attenuation after energy move to "+startAttenuation);
    		DevAttenuation.get().startMove(startAttenuation).waitFor();
    		
    		result.type = monoMove.getResult().type;
    		result.msg = monoMove.getResult().msg;
		}
		else
		{
			Log.note("Moving energy in script mode");
			
			moveType = MoveType.SCRIPT;
			
			String idFlag = (null == DevUndulator.getDevice())?"noID":"";
			String dbFlags = DBConfig.get("bigEnergyMove_flags");
			if (null == dbFlags) dbFlags = "";

			String cmdStr = Util.getPerlPrefix()+bigEnergyScript+" "+dest+" "+idFlag+" "+dbFlags+" bluice";
			
			moveScript = ExecTask.new_ExecTask(cmdStr, bigEnergyScript, true);
			moveScript.start();
			
			updateCombinedState();
			
			while (null != moveScript && moveScript.isRunning())
			{
				Util.sleep(10);
			}
			
			state = State.READY;
			if (0 == moveScript.exitValue)
			{
				result.type = ResultType.SUCCESS;
			}
			else
			{
				result.type = ResultType.ERROR;
				result.msg = moveScript.stderrString();
			}

			moveType = MoveType.NONE;
			moveScript = null;
			
			updateCombinedState();
		}
		//updateDetectorEnergy(getPosition()*1000.0);
	}
	
	private void updateDetectorEnergy(double energy_eV)
	{
		if (DetectorConfig.isEiger())
		{
			// only update the detector energy setting if it's more than the allowed difference
			double current_eV = Epics.getDouble(PV_EIGER.PhotonEnergy);
			double change_eV = Math.abs(energy_eV-current_eV);
			if (change_eV >= DBConfig.getDouble("eiger_tolerance_eV"))
			{
				Epics.putDouble(PV_EIGER.PhotonEnergy, energy_eV);
			}
		}
	}

	@Override
	public Limits getLimits()
	{
		return monoDev.getLimits();
	}
	
	@Override
	public boolean getMoving()
	{
		if (null == moveType) return (monoDev.isMoving() || (null!=undDev && undDev.isMoving()));
		
		switch (moveType)
		{
		case DEVICE:
			if (monoDev.isMoving()) return true;
			if (null!=undDev && undDev.isMoving()) return true;
			return false;
			
		case SCRIPT:
			return (null != moveScript);
			
		case NONE:
			return false;
		
		default:
			return false;
		}
	}
	
	@Override
	public double getPosition()
	{
		return monoDev.getPosition();
	}
	
	@Override
	public boolean moveChecks(double dest)
	{
		System.out.println("DevEnergy: move checks");
		if (!monoDev.moveChecks(dest)) return false;
		if (null != undDev && !undDev.moveChecks(dest)) return false;
		return true;
	}
	
	@Override
	public void abort()
	{
		monoDev.abort();
		if (null != undDev) undDev.abort();
		if (null != moveScript && moveScript.isRunning()) moveScript.asyncStop();
	}
	
	@Override protected void onAllPVsConnected() {}
	@Override protected void valueChangedAction(String pvName, DBR value) {}
	@Override
	public double getRqsPos()
	{
		if (null != monoDev) return monoDev.getRqsPos();
		return 0;
	}
}
