package gov.anl.gmca.hutch;

import gov.anl.gmca.common.EpicsDevice;
import gov.anl.gmca.io.Database;
import gov.anl.gmca.io.Database.DBName;
import gov.anl.gmca.io.Database.WhereList;
import gov.anl.gmca.io.Database.WhereOp;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PVName;
import gov.anl.gmca.io.PVLookup.PV_HUTCH_SHUTTER;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.Util;
import gov.aps.jca.CAException;
import gov.aps.jca.dbr.DBR;

import java.util.ArrayList;

public class DevHutchShutter extends EpicsDevice
{
	protected PVName shutterStatPV, shutterOpenPV, shutterClosePV;
	private int openValue=0;
	
	public ListenerList<DevHutchShutter> ll;
	
	protected static enum OpenState { OPEN, CLOSED };
	protected OpenState requestedState = null;
	protected boolean closeAllowed = false;
	
	protected static final int timeoutMs = 10000;
	
	public static DevHutchShutter get(String deviceName) { return (DevHutchShutter)getDevice(deviceName); }
	
	public DevHutchShutter(String motorID)
	{
		super(motorID);
		
		ll = new ListenerList<DevHutchShutter>(this);
		
		this.shutterStatPV=PV_HUTCH_SHUTTER.stat.shutter(motorID);
		this.shutterOpenPV=PV_HUTCH_SHUTTER.open.shutter(motorID);
		this.shutterClosePV=PV_HUTCH_SHUTTER.close.shutter(motorID);

		// find out from SQL whether 1 is open or 0
		ArrayList<String> columns = new ArrayList<String>();
		columns.add("openValue");
		WhereList whereList = new WhereList();
		whereList.add("Shutter", WhereOp.EQUALS, motorID);
		Database.Result result = Database.select(DBName.CONFIG, "HutchShutter", columns, whereList, null);
		this.openValue = Integer.parseInt(result.get(0, "openValue"));
		System.out.println("Got openvalue="+openValue+" for device "+motorID);
		
		closeAllowed = (shutterClosePV.toString().length()!=0);
		
		registerDevicePV(shutterStatPV);
		registerDevicePV(shutterOpenPV);
		if (closeAllowed) registerDevicePV(shutterClosePV);
		
		System.out.println("Completing shutter registration");
		registrationComplete();
		System.out.println("Shutter registration complete");
	}
	
	private boolean isIntValueOpen(int i)
	{
	    return i==openValue;
	}

	public boolean isOpen()
	{
	    return isIntValueOpen(Epics.getInteger(shutterStatPV));
	}
	
	protected OpenState getOpenState()
	{
		if (isOpen()) return OpenState.OPEN;
		return OpenState.CLOSED;
	}
	
	public void open() throws CAException
	{
		requestedState = OpenState.OPEN;
		Epics.putInteger(shutterOpenPV, 1);
	}
	
	public void close() throws CAException
	{
		if (!closeAllowed)
		{
			Log.error("Shutter "+getDeviceName()+" does not support close().");
			return;
		}
		requestedState = OpenState.CLOSED;
		Epics.putInteger(shutterClosePV, 1);
	}
	
	public void waitFor()
	{
		Util.sleep(20);
		
		long waitStartMs = System.currentTimeMillis();
		
		while (getOpenState() != requestedState)
		{
			Util.sleep(10);
			if (System.currentTimeMillis() - waitStartMs > timeoutMs)
			{
				Log.error("Shutter "+motorID+" timed out");
				return;
			}
		}
	}

	@Override protected void onAllPVsConnected()
	{
		requestedState = isOpen()?OpenState.OPEN:OpenState.CLOSED;
		
		Epics.addMonitor(shutterStatPV, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				Boolean isOpen = isIntValueOpen(Epics.dbrToInt(value));
				ll.notifyListeners(isOpen);
			}
		});
	}
	
	@Override protected void valueChangedAction(String pvName, DBR value) {}
	@Override public void abort() {}
}
