package gov.anl.gmca.hutch;

import gov.anl.gmca.collect.DevCollectSettings;
import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.common.DevMoveFloat;
import gov.anl.gmca.common.EpicsDevice;
import gov.anl.gmca.common.Limits;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup;
import gov.anl.gmca.io.PVLookup.PVName;
import gov.anl.gmca.io.PVLookup.PV_UNDULATOR;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.Util;
import gov.aps.jca.dbr.DBR;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.atomic.AtomicBoolean;

public class DevUndulator extends DevMoveFloat
{
	protected 	int 	harmonic 		= 0;
	protected 	double 	offset 			= Double.NaN;
	protected 	double 	position;
	protected	boolean	isMoving		= false;
	
	public 		int 	getHarmonic() 	{ return harmonic; 	}
	public 		double 	getOffset()		{ return offset; 	}
	public 		double 	getPosition() 	{ return Epics.getDouble(PV_UNDULATOR.Energy_VAL); }
	public		double	getRqsPos()     { return Epics.getDouble(PV_UNDULATOR.EnergySet_VAL); }
	public		boolean	getMoving()		{ return (1==Epics.getInteger(PV_UNDULATOR.Busy_VAL)); }

	protected	AtomicBoolean			moveAborted = new AtomicBoolean(false);
	
	public static DevUndulator getDevice() { return (DevUndulator)EpicsDevice.getDevice("undulator"); }
	
	public DevUndulator()
	{
		super("undulator");
		
		for (PVName pv : PV_UNDULATOR.values())
		{
			registerDevicePV(pv);
		}
		registrationComplete();
	}
	
	// from PmacUndulatorMotorAssembly::moveTo()
	@Override
	public boolean moveChecks(double dest)
	{
		if (!DevCollectSettings.getEnableEnergyMove())
		{
			System.out.println("Undulator moves are disabled, returning OK for checks.");
			return true;
		}
		
		// dest is within soft limits
		if (Util.sigLessThan(dest, getLimits().getLower()) || Util.sigGreaterThan(dest,getLimits().getUpper()))
		{
			Log.error("Invalid reqested position "+dest+", low limit="+getLimits().getLower()+" high limit="+getLimits().getUpper());
			return false;
		}
		
		System.out.println("Undulator checks pass.");
		return true;
	}
	
	// updates both position and harmonic
	@Override
	public void moveCmd(double dest, MoveResult result)
	{
		if (!DevCollectSettings.getEnableEnergyMove())
		{
			System.out.println("Undulator moves are disabled, returning move success.");
			result.type = ResultType.SUCCESS;
			return;
		}
		
		// is the undulator enabled in SQL?
		
		// wait for undulator to stop moving
	
		setState(State.MOVING);
    	for (Listener listener : listenerSet) listener.handleStateChange(getState());
		
		moveAborted.set(false);
		
		HarmonicAndOffset hao = calcHarmonicAndOffset(dest).get(0);
		double destWithOffset = dest + hao.offset;

		Epics.putInteger(PV_UNDULATOR.HarmonicValue, hao.harmonic);
		
		Epics.putDouble(PV_UNDULATOR.EnergySet_VAL, destWithOffset);
		Epics.putInteger(PV_UNDULATOR.Start_VAL, 1);
		
		//long moveStartMs = System.currentTimeMillis();
		int inPosCount = 0;
		while (true)
		{
			Util.sleep(10);
			Epics.poll();
			
			if (getMoving())
			{
				inPosCount = 0;
			}
			else
			{
				inPosCount++;
			}
			
			double distFromDestination = Math.abs( getPosition() - destWithOffset ); 
			
			System.out.println("track_motion: undulator inPosCount="+inPosCount+" distFromDest="+distFromDestination+" rqspos="+destWithOffset+" actpos="+getPosition());
			
			if ( distFromDestination < 0.01 && inPosCount >= 3 )
			{
				break;
			}
			
			if (moveAborted.get())
			{
				Log.warn("Undulator move aborted, tracking stopped.");
				break;
			}
		}
		
		setState(State.READY);
    	for (Listener listener : listenerSet) listener.handleStateChange(getState());
    	
		result.type = ResultType.SUCCESS;
	}
	
	public void abort()
	{
		moveAborted.set(true);
		Epics.putInteger(PV_UNDULATOR.Stop_VAL,1);
	}

	public Limits getLimits()
	{
		Limits undulatorLimits = new Limits
		(
			Epics.getDouble(PVLookup.PV_PMAC.RqsPos_DRVL.motor("mo_En_E")),
			Epics.getDouble(PVLookup.PV_PMAC.RqsPos_DRVH.motor("mo_En_E")),
			"soft limit for undulator"
		);
		System.out.println("Undulator limits are "+undulatorLimits.getLower()+" "+undulatorLimits.getUpper());
		return undulatorLimits;
	}
	
	public static class HarmonicAndOffset
	{
		public int harmonic;
		public double offset;
	}
	
	// calculates the correct undulator harmonic and position for a given mono position
	// (no side effects)
	public static List<HarmonicAndOffset> calcHarmonicAndOffset(double monoPosition)
	{
		List<HarmonicAndOffset> ret = new ArrayList<HarmonicAndOffset>();

		// read cutoff values from SQL
		double h3h1cutoff = Double.parseDouble(DBConfig.get("offset.h3h1Cutoff(keV)"));
		String h5h3cutoff_s = DBConfig.get("offset.h5h3Cutoff(keV)");
		String h7h5cutoff_s = DBConfig.get("offset.h7h5Cutoff(keV)");

		// generate strict ranges
		List<Double[]> strictRanges = new ArrayList<>();
		strictRanges.add(new Double[]{0.0, h3h1cutoff, 1.0});
		if (h5h3cutoff_s != null && h7h5cutoff_s != null) {
			double h5h3cutoff = Double.parseDouble(h5h3cutoff_s);
			double h7h5cutoff = Double.parseDouble(h7h5cutoff_s);
			strictRanges.add(new Double[]{h3h1cutoff, h5h3cutoff, 3.0});
			strictRanges.add(new Double[]{h5h3cutoff, h7h5cutoff, 5.0});
			strictRanges.add(new Double[]{h7h5cutoff, 1000.0, 7.0});
		} else {
			strictRanges.add(new Double[]{h3h1cutoff, 1000.0, 3.0});
		}

		// generate loose ranges
		List<Double[]> looseRanges = new ArrayList<>();
		double margin_keV = 0.5;
		for (Double[] range : strictRanges) {
			looseRanges.add(new Double[]{range[0]-margin_keV, range[1]+margin_keV, range[2]});
		}

		// search ranges
		int harmonic = 1, nearby_harmonic=-1;
		for (Double[] range : strictRanges) {
			if (monoPosition >= range[0] && monoPosition <= range[1]) {
				harmonic = (int) (Math.round(range[2]));
				break;
			}
		}
		for (Double[] range : looseRanges) {
			if (monoPosition >= range[0] && monoPosition <= range[1]) {
				int h = (int) (Math.round(range[2]));
				if (h != harmonic) {
					nearby_harmonic = h;
					break;
				}
			}
		}

		// create return value with harmonics found
		ret.add(generateHarmonicAndOffset(harmonic,monoPosition));
		if (nearby_harmonic != -1) ret.add(generateHarmonicAndOffset(nearby_harmonic, monoPosition));

		return ret;
	}
	
	private static HarmonicAndOffset generateHarmonicAndOffset(int harmonic, double monoPosition)
	{
		HarmonicAndOffset ret = new HarmonicAndOffset();
		
		ret.harmonic = harmonic;
		double eLo = Double.parseDouble(DBConfig.get("offset.h"+harmonic+"_eLo(keV)"));
		double oLo = Double.parseDouble(DBConfig.get("offset.h"+harmonic+"_oLo(keV)"));
		double eHi = Double.parseDouble(DBConfig.get("offset.h"+harmonic+"_eUp(keV)"));
		double oHi = Double.parseDouble(DBConfig.get("offset.h"+harmonic+"_oUp(keV)"));
		ret.offset = oLo + (oHi-oLo)*(monoPosition-eLo)/(eHi-eLo);
		
		return ret;
	}

	public static void main(String[] args) {
		List<Double> test_vals = new ArrayList<Double>();
		for (double v=-1; v < 40; v += .1) test_vals.add(v);
		for (double test_val : test_vals) {
			String s = String.format("Input %.3f", test_val);
			List<HarmonicAndOffset> result = calcHarmonicAndOffset(test_val);
			for (HarmonicAndOffset r : result) {
				s += String.format(" Output offset=%.3f harmonic=%d", r.offset, r.harmonic);
			}
			System.out.println(s);
		}
	}
	
	public void waitFor()
	{
		while (isMoving) Util.sleep(10);
	}
	
    @Override protected void onAllPVsConnected()
    {
	    for (Listener listener : listenerSet) listener.handleStateChange(getState());
	}

	@Override protected void valueChangedAction(String pvName, DBR value)
	{
		if 		(pvName.equals(PV_UNDULATOR.Energy_VAL.toString())) 	position = Epics.dbrToDouble(value);
		else if (pvName.equals(PV_UNDULATOR.HarmonicValue.toString())) harmonic = Epics.dbrToInt(value);
		else if (pvName.equals(PV_UNDULATOR.Busy_VAL.toString())) 		isMoving = (1==Epics.dbrToInt(value));
	}
}
