package gov.anl.gmca.hutch;

import gov.anl.gmca.common.client_server.StateMap;
import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.scan.ui.ModelGraph;
import gov.anl.gmca.util.Bounds;
import gov.anl.gmca.util.Pair;
import gov.anl.gmca.util.Point2D;
import gov.anl.gmca.util.Util;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Composite;

import java.util.ArrayList;

public class WdgAlignPlot extends Composite implements PaintListener
{
	private final static Color	BG_COLOR 	= GUI.COLOR_SAMPLE_BG;
	private final static Color 	FG_COLOR 	= GUI.COLOR_RED;
	private final static int 	LINE_WIDTH 	= 2;
	private final static int 	AA_SETTING	= SWT.ON;
	
	private int orientation, widgetWidth, widgetHeight;
	private double xScale, yScale, minYValue, maxYValue;
	private Bounds plotBounds = new Bounds();
	private GC gc;
	protected StateMap.StateOptimizeBeamGraph data;
	
	public WdgAlignPlot(Composite parent, int orientation)
	{
		super(parent, SWT.NONE);
		this.orientation=orientation;
		addPaintListener(this);
	}
	
	@Override public void paintControl(PaintEvent e)
	{
		gc = e.gc;
		calcBounds();
		draw();
	}

	private void calcBounds()
	{
		// for convenience
		widgetWidth = getSize().x-1;
		widgetHeight = getSize().y-1;
		
		// calculate plot area
		int scaleHeight = 20;
		int xMargin = 20;
		plotBounds.setXBounds(xMargin, widgetWidth-xMargin);
		plotBounds.setYBounds(10, widgetHeight-scaleHeight);
	}
	
	private void calcXCoords()
	{
		xScale = plotBounds.width()/(data.vHigh-data.vLow);
	}
	
	private void calcYCoords()
	{
		minYValue = Double.POSITIVE_INFINITY;
		maxYValue = Double.NEGATIVE_INFINITY;
		
		for (double yVal : data.yPoints)
		{
			if (maxYValue<yVal) maxYValue=yVal;
			if (minYValue>yVal) minYValue=yVal;
		} 
		
		yScale = 1.0/(maxYValue-minYValue)*plotBounds.height(); // .95 = 5% margin on top
	}
	
	private int getScreenX(double x)
	{
		double 	xPastMin 	= x-data.vLow;
		double 	scaledX 	= xPastMin*xScale;
		int 	screenX		= plotBounds.xmin + (int)scaledX;
		
		return screenX;
	}
	
	private int getScreenY(double y)
	{
		double 	yAboveMin 	= y-minYValue;
		double 	scaledY 	= yAboveMin*yScale;
		int 	screenY 	= plotBounds.ymax - (int)scaledY;
		
		return screenY;
	}
	
	private void draw()
	{
		// background
		gc.setBackground(GUI.COLOR_WHITE);
		gc.fillRectangle(0,0,widgetWidth,widgetHeight);

		// vert/horiz. label
		gc.setLineWidth(1);
		gc.setForeground(GUI.COLOR_DARK_BLUE);
		gc.setAntialias(SWT.ON);
		gc.setFont(GUI.FONT_10_BOLD);
		int x=10,y=10;
		String str = orientation==SWT.VERTICAL?"Vertical":"Horizontal";
		gc.drawString(str, x, y, true);

		// plot
		if (null != data)
		{
			calcXCoords();
			calcYCoords();

			drawTrace();
			drawTicks();
			drawMirrorPos();
		}
		
		// border
		gc.setForeground(GUI.COLOR_BLACK);
		gc.setLineWidth(1);
		gc.drawRectangle(0,0,widgetWidth,widgetHeight);		
	}

	private void drawTrace()
	{
		gc.setBackground(BG_COLOR);
		gc.setForeground(FG_COLOR);
		gc.setLineWidth(LINE_WIDTH);
		gc.setAntialias(AA_SETTING);
		
		for (int i=1; i<data.xPoints.length; i++)
		{
			Point2D pt = new Point2D(data.xPoints[i], data.yPoints[i]);
			Point2D lastPt = new Point2D(data.xPoints[i-1], data.yPoints[i-1]);
			
			int lastX = getScreenX(lastPt.x);
			int ptX = getScreenX(pt.x);
			
			int lastY = getScreenY(lastPt.y);
			int ptY = getScreenY(pt.y);

			gc.fillPolygon(new int[]{lastX,plotBounds.ymax,lastX,lastY,ptX,ptY,ptX,plotBounds.ymax});
			gc.drawLine(lastX, lastY, ptX, ptY);
		}
	}
	
	private void drawTicks()
	{
		Pair<ArrayList<Double>,Integer> ret = ModelGraph.looseLabel(data.vLow, data.vHigh, 5);
		ArrayList<Double> ticks = ret.a;
		int numDecimals = ret.b;
		
		gc.setFont(GUI.FONT_11);
		gc.setForeground(GUI.COLOR_BLACK);
		gc.setLineWidth(1);
		
		for (double tick : ticks)
		{
			int tickX = getScreenX(tick);
			String tickStr = Util.sprintf("%."+numDecimals+"f", tick);
			Point extent = gc.textExtent(tickStr);
			if (tickX-extent.x/2 >= 0 && tickX+extent.x/2 <= widgetWidth)
			{
				gc.drawLine(tickX, plotBounds.ymax, tickX, plotBounds.ymax+3);
				gc.drawString(tickStr, tickX-extent.x/2, plotBounds.ymax+2, true);
			}
		}
		
		gc.setLineWidth(1);
		gc.drawLine(0, plotBounds.ymax, widgetWidth, plotBounds.ymax);
	}
	
	private void drawMirrorPos()
	{
		if (null == DevAlignMinibeam.get().getLastRunTime()) return;
		
		Double mirrorResult = orientation==SWT.VERTICAL ? DevAlignMinibeam.get().getVMirrorResult() : DevAlignMinibeam.get().getHMirrorResult();
		if (null==mirrorResult) return;
		
		int mirrorX = getScreenX(mirrorResult);
		int indWidth = 6;

		int[] triangle = new int[]{mirrorX-indWidth, 0, mirrorX, indWidth, mirrorX+indWidth, 0};
		
		gc.setLineWidth(1);
		gc.setForeground(GUI.COLOR_BLACK);
		gc.setBackground(GUI.COLOR_LIGHT_GRAY);
		
		gc.fillPolygon(triangle);
		gc.drawPolygon(triangle);
		
		/* alternate indicator
		 
		int[] triangle = new int[]{mirrorX-indWidth, plotBounds.ymax, mirrorX, plotBounds.ymax-indWidth*2, mirrorX+indWidth, plotBounds.ymax};
		
		gc.setLineWidth(1);
		gc.setForeground(GUI.COLOR_BLACK);
		gc.setBackground(GUI.COLOR_WHITE);
		
		gc.fillPolygon(triangle);
		gc.drawPolygon(triangle);*/
	}
}
