package gov.anl.gmca.io;

import gov.anl.gmca.common.SQLDataType;
import gov.anl.gmca.gui.Main;

import java.util.ArrayList;
import java.util.HashMap;

public abstract class Database
{
    protected static Database inst = null;
    public static Database get()
    {
    	if (null == inst)
    	{
    		if (Main.SIM_LEVEL==2)	inst = new gov.anl.gmca.io.sim.Database();
    		else		   			inst = new gov.anl.gmca.io.hw.Database();
    	}
		return inst;
	}
    
	///////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////// classes //////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////

    public static class DBField
    {
    	public String fieldName;
    	public SQLDataType dataType;
    	public boolean allowNull;
    	public boolean primaryKey;
    	public ForeignKey foreignKey;
    	
    	public DBField(String fieldName, SQLDataType dataType, boolean allowNull, boolean primaryKey, ForeignKey foreignKey)
    	{
    		this.fieldName=fieldName; this.dataType=dataType; this.allowNull=allowNull; this.primaryKey=primaryKey; this.foreignKey=foreignKey;
    	}
    }
    
    public static class ForeignKey
    {
    	public String tableName;
    	public String fieldName;
    	public ForeignKey(String tableName, String fieldName) { this.tableName=tableName; this.fieldName=fieldName; }
    }
    
    public static class WhereList
    {
    	public ArrayList<Where> whereList = new ArrayList<Where>();
    	public void add(String fieldName, WhereOp whereOp, String value)
    	{
    		whereList.add(new Where(fieldName, whereOp, value));
    	}
    	
		public HashMap<String, String> toMap()
		{
			HashMap<String,String> whereMap = new HashMap<String,String>();
			for (Where where : whereList)
			{
				if (where.whereOp != WhereOp.EQUALS) throw new IllegalStateException("Where list contained non-equals operation! "+where.fieldName+" "+where.whereOp+" "+where.value);
				whereMap.put(where.fieldName, where.value);
			}
			return whereMap;
		}
    }
    
    public static enum WhereOp
    {
    	EQUALS("="),
    	LESS_THAN("<"),
    	GREATER_THAN(">");

    	private String symbol;
    	public String getSymbol() { return symbol; }
    	WhereOp(String symbol) { this.symbol = symbol; }
	}
    
    public static class Where
    {
    	public String fieldName;
    	public WhereOp whereOp;
    	public String value;
    	
    	public Where(String fieldName, WhereOp whereOp, String value)
    	{
    		this.fieldName=fieldName; this.whereOp=whereOp; this.value=value;
    	}
    	
    	public boolean valueMatches(String valueToCompare)
    	{
    		switch (whereOp)
    		{
    		case EQUALS: return valueToCompare.equals(value);
    		
    		case LESS_THAN:
    		case GREATER_THAN:
    		{
	    		double thisDoubleVal=Double.NaN, compareDoubleVal=Double.NaN;
	    		try
	    		{
	    			thisDoubleVal = Double.parseDouble(value);
	    			compareDoubleVal = Double.parseDouble(valueToCompare);
	    		}
	    		catch (NumberFormatException nfe) { return false; }
	    		
	    		if (whereOp==WhereOp.LESS_THAN) return compareDoubleVal < thisDoubleVal;
	    		if (whereOp==WhereOp.GREATER_THAN) return compareDoubleVal > thisDoubleVal;
    		}
    		default: throw new IllegalArgumentException("Unsupported where op: "+whereOp);
    		}
    	}
    }
    
    public static class DBRows
    {
    	public String[] columns;
    	public String[][] values;
    }
    
	///////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////// static redirectors ////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////

    // create
    public static boolean create(DBName dbName, String tableName, ArrayList<DBField> fields)
    {
    	return get().i_create(dbName, tableName, fields);
    }
    
    // delete
    public static boolean delete(DBName dbName, String tableName, WhereList whereList)
    {
    	if (null==whereList) whereList = new WhereList();
    	return get().i_delete(dbName, tableName, whereList);
    }
    
    // update existing rows
    public static boolean update(DBName dbName, String tableName, HashMap<String,String> setMap, WhereList whereList)
    { return get().i_update(dbName, tableName, setMap, whereList); }

    // multi-row insert
    public static Result insert(DBName dbName, String table, ArrayList<HashMap<String,String>> rows)
    { return get().i_insert(dbName, table, rows); }
    
    // multi-row insert (new)
    public static Result insert(DBName dbName, String table, DBRows rows)
    { return get().i_insert(dbName, table, rows); }
    
    // single row insert
    public static Result insert(DBName dbName, String table, HashMap<String,String> row)
    {
    	ArrayList<HashMap<String,String>> rows = new ArrayList<HashMap<String,String>>();
    	rows.add(row);
    	return get().i_insert(dbName, table, rows);
    }
    
    // select
    public static Result select(DBName dbName, String tableName, ArrayList<String> requestedColumns, WhereList whereList, String orderBy)
    {
    	if (null==whereList) whereList = new WhereList();
    	return get().i_select(dbName, tableName, requestedColumns, whereList, orderBy);
    }

    // generic exec
    public static Result exec(DBName dbName, String query)
    { return get().i_exec(dbName, query); }
    
    // one-time initialization
    public static void init()
    { get().i_init(); }
    
	///////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////// begin interface //////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////

    abstract protected boolean	i_create	(DBName dbName, String tableName, ArrayList<DBField> fields);
    abstract protected boolean	i_delete	(DBName dbName, String tableName, WhereList whereList);
    abstract protected boolean 	i_update	(DBName dbName, String tableName, HashMap<String,String> setMap, WhereList whereList);
	abstract protected Result 	i_insert	(DBName dbName, String table, ArrayList<HashMap<String,String>> rows);
	abstract protected Result 	i_insert	(DBName dbName, String table, DBRows rows);
    abstract protected Result 	i_select	(DBName dbName, String tableName, ArrayList<String> requestedColumns, WhereList whereList, String orderBy);
	abstract protected Result 	i_exec		(DBName dbName, String query);
	abstract protected void 	i_init		();
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////// end interface ///////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////
	
	public enum DBName
	{
		CONFIG("blc2004"),
		USERDATA("user_data"),
		GMCADATA("gmca_data");
		
		protected String sqlName;
		DBName(String sqlName)
		{
			this.sqlName = sqlName;
		}
		public String getSQLName() { return sqlName; }
	}
	
	public static enum RESULT_ERROR { NONE, TABLE_DOESNT_EXIST, SYNTAX };
	
	public class Result
	{
		private ArrayList<HashMap<String,String>> res;
		private Integer generatedKey = null;
		
		public Result(RESULT_ERROR error)						{ this.error = error;			}
		public Result(ArrayList<HashMap<String,String>> res)	{ this.res=res; 				}
		public String get(int row, String col)					{ return res.get(row).get(col); }
		public Integer getGeneratedKey()						{ return generatedKey;			}
		public HashMap<String,String> getRow(int row)			{ return res.get(row);			}
		public ArrayList<HashMap<String,String>> getRows() 		{ return res; 					}
		public int numRows() 									{ return null==res?0:res.size(); 			}
		public boolean isOK()									{ return null==error || error == RESULT_ERROR.NONE; }
		public RESULT_ERROR error								= RESULT_ERROR.NONE;
		
		public Result(ArrayList<HashMap<String,String>> res, Integer generatedKey)
		{
			this.res = res;
			this.generatedKey = generatedKey;
		}
		
		@Override public String toString()
		{
			String s = "{";

			for (HashMap<String,String> row : res)
			{
				for (String key : row.keySet())
				{
					s += "("+key+","+row.get(key)+")";
				}
			}
			s += "}";
			return s;
		}
	}
}
