package gov.anl.gmca.io.hw;

import com.cosylab.epics.caj.CAJChannel;
import gov.anl.gmca.common.*;
import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.gui.GUIStatus;
import gov.anl.gmca.io.PVLookup;
import gov.anl.gmca.io.PVLookup.PVName;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.Util;
import gov.aps.jca.*;
import gov.aps.jca.Channel.ConnectionState;
import gov.aps.jca.configuration.DefaultConfiguration;
import gov.aps.jca.dbr.*;
import gov.aps.jca.event.ConnectionEvent;
import gov.aps.jca.event.ConnectionListener;
import gov.aps.jca.event.MonitorEvent;
import gov.aps.jca.event.MonitorListener;

import java.nio.channels.AlreadyBoundException;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;

/**
 * This class is solely responsible for communicating with EPICS IOCs. 
 * 
 * Right now all PVs are monitored - the monitor is set up in connectionChanged
 * 
 */
public class Epics extends gov.anl.gmca.io.Epics implements ConnectionListener, MonitorListener
{
	public final static long CONNECT_TIMEOUT_MS = 3000;
	
	protected final static boolean LOG_PUTS = true;
	protected final static boolean LOG_CALLBACKS = false;
	
	private final static Set<String> logPVSet = new HashSet<String>();
	
	public Epics()
	{
		for (PVName pv : PVLookup.loggedPVs)
		{
			logPVSet.add(pv.toString());
		}
	}
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////// monitor methods /////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	@Override protected void i_addMonitor(final String pvName, final EpicsMonitor em)
	{
		try
		{
			imonMapLock.acquire();

			if (null == imonMap.get(pvName))
			{
				imonMap.put(pvName,new ArrayList<EpicsMonitor>());
			}
			else
			{
				// If the PV value is already known, then the monitor is already
				// connected, and adding this monitor to the list won't end up getting 
				// valueChanged called.  So call it here.
				if (dbrCache.containsKey(pvName))
				{
					((gov.anl.gmca.io.hw.Epics)inst).executor.submit(new Runnable(){@Override public void run()
					{
						// note that the value sent here could be out of date since it's cached
						if (null != em) em.valueChanged(pvName, dbrCache.get(pvName));
					}});
				}
			}
			
			imonMap.get(pvName).add(em);
		}
		catch (Exception e) { e.printStackTrace(); }
		finally
		{
			imonMapLock.release();
		}

		connectIfNeeded(pvName,false);
		if (PVLookup.isWaveform(pvName)) connectIfNeeded(pvName+".NORD",false);
	}
	
	@Override protected boolean i_removeMonitor(String pvName, EpicsMonitor em)
	{
		// get the monitors for this PV
		ArrayList<EpicsMonitor> monitors = imonMap.get(pvName);
		if (null == monitors) return false;
		
		// remove it
		boolean monitorRemoved = monitors.remove(em);
		
		// clear cache and disconnect PV if all monitors are removed
		/*if (0 == monitors.size())
		{
			imonMapLock.acquireUninterruptibly();
			
			try
			{
				pvMap.get(pvName).dispose();
				pvMap.remove(pvName);	
				dbrCache.remove(pvName);
			}
			catch (Exception e) { e.printStackTrace(); 	}
			finally 			{ imonMapLock.release();}
		}*/
		
		return monitorRemoved;
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////// get methods ///////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////

	@Override protected int i_getCacheInteger(String pvName)
	{
		return i_getIntegerArray(pvName)[0];
	}
	
	@Override protected double 	i_getDouble(String pvName) 	{ return i_getDoubleArray(pvName)[0]; 	}
	@Override protected int 	i_getInteger(String pvName) 	{ return i_getIntegerArray(pvName)[0]; 	}
	@Override protected short 	i_getEnum(String pvName) 		{ return i_getEnumArray(pvName)[0]; 		}
	@Override protected boolean	i_getBoolean(String pvName)	{ String value = i_getString(pvName, 0); return (value.equals("1")||value.toLowerCase().equals("true")); }
	
	@Override protected double[] i_getDoubleArray(String pvName)
	{
		DBR dbr = i_getDBR(pvName);
		if (DBRType.DOUBLE != dbr.getType())
		{
			try { dbr = dbr.convert(DBRType.DOUBLE); }
			catch (CAStatusException e) { e.printStackTrace(); }
		}
		double[] ret = ((DBR_Double)dbr).getDoubleValue();
		boolean isWaveform = PVLookup.isWaveform(pvName);
		if (isWaveform)
		{
			boolean isInitialized = i_getInteger(pvName+".NORD")!=0;
			if (!isInitialized)
			{
				for (int i=0;i<ret.length;i++) ret[i]=0;
			}
		}
		return ret;
	}
	
	@Override protected int[] i_getIntegerArray(String pvName)
	{
		DBR dbr = i_getDBR(pvName);
		if (DBRType.INT != dbr.getType())
		{
			try { dbr = dbr.convert(DBRType.INT); }
			catch (CAStatusException e) { e.printStackTrace(); }
		}
		return ((DBR_Int)dbr).getIntValue();
	}
	
	@Override protected short[] i_getEnumArray(String pvName)
	{
		DBR dbr = i_getDBR(pvName);
		if (DBRType.ENUM != dbr.getType())
		{
			try { dbr = dbr.convert(DBRType.ENUM); }
			catch (CAStatusException e) { e.printStackTrace(); }
		}
		return ((DBR_Enum)dbr).getEnumValue();
	}
	
	@Override protected String i_getString(String pvName, int decimals)
	{
		return dbrToString(i_getDBR(pvName), decimals);
	}
	@Override protected String i_getCacheString(String pvName, int decimals)
	{
		return dbrToString(i_getCacheDBR(pvName), decimals);
	}
	
	@Override protected DBR i_getCacheDBR(String pvName)
	{
		return dbrPostCache.get(pvName);
	}
	
	@Override protected synchronized DBR i_getDBR(String pvName)
	{
		if (Util.atExit()) return null;
		if (disabledPVMap.containsKey(pvName))
		{
			try
			{
				double d = Double.parseDouble((String)disabledPVMap.get(pvName));
				return new DBR_Double(new double[]{d});
			}
			catch (NumberFormatException nfe) { return new DBR_String((String)disabledPVMap.get(pvName)); }
		}
		
		connectIfNeeded(pvName,true);
		if (PVLookup.isWaveform(pvName)) connectIfNeeded(pvName+".NORD",true);
		
		long waitStart = System.currentTimeMillis();
		while (!dbrCache.containsKey(pvName))
		{
			try
			{
				// uncomment to see when PVs are being read before they're connected
				//System.out.println("Waiting for "+pvName);
				long now = System.currentTimeMillis();
				if (now-waitStart>10000)
				{
					System.out.println("ERROR: Timed out waiting for PV "+pvName+" to connect, quitting.");
					System.exit(-1);
				}
				
				if (null == ctxt()) break;
				ctxt().poll();
			}
			catch (CAException cae) { cae.printStackTrace(); }
			catch (Exception e) { e.printStackTrace(); }
		}
		
		return dbrCache.get(pvName);
	}
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////// put methods ///////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	@Override protected void i_putInteger(String pvName, int value) 	{ i_put(pvName, Integer.toString(value)); }
	@Override protected void i_putBoolean(String pvName, boolean value) { i_put(pvName, Boolean.toString(value)); }
	@Override protected void i_putDouble(String pvName, double value) 	{ i_put(pvName, Double.toString(value)); 	}
	
	@Override protected void i_putDoubleArray(String pvName, double[] values)
	{
		// bypass conversion if possible
		DBR cacheVal = dbrCache.get(pvName);
		Channel ch = pvMap.get(pvName);
		if (null != ch && ch.getConnectionState() != ConnectionState.CONNECTED && 
			null != cacheVal && cacheVal.isDOUBLE())
		{
			try								{ ch.put(values); 		}
			catch (IllegalStateException e) { e.printStackTrace(); 	}
			catch (CAException e)			{ e.printStackTrace(); 	}
		}
		else
		{
    		String[] array = new String[values.length];
    		for (int i=0; i<values.length; i++) array[i]=Double.toString(values[i]);
    		i_put(pvName,array);
		}
	}
	
	@Override protected void i_putIntegerArray(String pvName, int[] values)
	{
		DBR cacheVal = dbrCache.get(pvName);
		Channel ch = pvMap.get(pvName);
		if (null != ch && ch.getConnectionState() != ConnectionState.CONNECTED && 
			null != cacheVal && cacheVal.isINT())
		{
			try								{ ch.put(values); 		}
			catch (IllegalStateException e) { e.printStackTrace(); 	}
			catch (CAException e)			{ e.printStackTrace(); 	}
		}
		else
		{
    		String[] array = new String[values.length];
    		for (int i=0; i<values.length; i++) array[i]=Integer.toString(values[i]);
    		i_put(pvName,array);
		}
	}

	@Override protected synchronized void i_put(String pvName, String[] value)
	{
		if (Util.atExit()) return;
		if (disabledPVMap.containsKey(pvName)) { disabledPVMap.put(pvName,value[0]); return; }
		
		try
		{
			connectIfNeeded(pvName,true);
			if (PVLookup.isWaveform(pvName)) connectIfNeeded(pvName+".NORD",true);
			
			Channel ch = pvMap.get(pvName);
			if (ch.getConnectionState() != ConnectionState.CONNECTED)
			{
				System.out.println("ERROR: PV "+pvName+" not connected, put() failed");
				return;
			}
			
			DBRType type = ch.get().getType();
			
			if 		(type == DBRType.STRING)	ch.put(value);
			else if (type == DBRType.ENUM)		{ short[] 	a=new short[value.length]; 	for (int i=0; i<a.length;i++) a[i]=Short.parseShort(value[i]); 		ch.put(a); }
			else if (type == DBRType.INT)		{ int[] 	a=new int[value.length]; 	for (int i=0; i<a.length;i++) a[i]=Integer.parseInt(value[i]); 		ch.put(a); }
			else if (type == DBRType.SHORT)		{ short[] 	a=new short[value.length]; 	for (int i=0; i<a.length;i++) a[i]=Short.parseShort(value[i]); 		ch.put(a); }
			else if (type == DBRType.FLOAT)		{ float[] 	a=new float[value.length]; 	for (int i=0; i<a.length;i++) a[i]=Float.parseFloat(value[i]); 		ch.put(a); }
			else if (type == DBRType.DOUBLE)	{ double[] 	a=new double[value.length]; for (int i=0; i<a.length;i++) a[i]=Double.parseDouble(value[i]); 	ch.put(a); }
			else if (type == DBRType.BYTE)
			{
				if (1==value.length)	{ if (value[0].length()==0) ch.put(0); else ch.put(value[0].getBytes()); }
				else					{ byte[]    a=new byte[value.length];   for (int i=0; i<a.length;i++) a[i]=(byte)Short.parseShort(value[i]);ch.put(a); }
			}
			_ctxt.flushIO();
			
			if (LOG_PUTS)
			{
				String valueString;
				if (1==value.length)
				{
					valueString = value[0];
				}
				else
				{
					valueString = "[";
					for (String s : value) { valueString += s+","; }
					if (valueString.endsWith(",")) valueString = valueString.substring(0,valueString.length()-1);
					valueString += "]";
				}
				
    			String logLine = "caput "+pvName+" "+valueString;
    			System.out.println(logLine);
			}
		}
		catch (CAException cae) { cae.printStackTrace(); }
		catch (IllegalStateException ise) { ise.printStackTrace(); }
	}

	@Override protected synchronized void i_put(String pvName, String value)
	{
		i_put(pvName, new String[]{value});
	}
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////// misc methods ///////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	@Override protected boolean i_isConnected(String pvName)
	{
		Channel ch = pvMap.get(pvName);
		if (null == ch) return false;
		
		return (ConnectionState.CONNECTED == ch.getConnectionState());
	}
	
	@Override protected void i_pendIO(int ms)
	{
		try
		{
			if (null != ctxt()) ctxt().pendIO(ms/1000.0);
		}
		catch (Exception e)	{ e.printStackTrace(); }
	}
	
	@Override protected void i_flushIO()
	{
		try
		{
			if (null != ctxt()) ctxt().flushIO();
		}
		catch (Exception e)	{ e.printStackTrace(); }
	}
	
	@Override protected void i_poll()
	{
		try
		{
			if (null != ctxt()) ctxt().poll();
		}
		catch(AlreadyBoundException abe)	{ System.out.println("epics context is already bound"); }
		catch(IllegalStateException ise)	{ ise.printStackTrace();}
		catch (CAException e)				{ e.printStackTrace();	}
		catch (NullPointerException npe)	{ npe.printStackTrace();}
		catch (Exception e)					{ e.printStackTrace();	}
	}
	
	@Override protected String[] i_getPVNames()
	{
		ArrayList<String> ret = new ArrayList<String>();
		for (String key : dbrCache.keySet()) ret.add(key);
		Collections.sort(ret);
		return ret.toArray(new String[]{});
	}
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////// end public interface ///////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	protected 	static Context		 						_ctxt			 		= null;
	public 		static Context 								ctxt() 					{ return _ctxt; }
				static ConcurrentHashMap<String,DBR> 		dbrCache 				= new ConcurrentHashMap<String,DBR>();
				static ConcurrentHashMap<String,DBR> 		dbrPostCache			= new ConcurrentHashMap<String,DBR>(); // updated after all callbacks processed
	protected 	static JCALibrary 							jca 					= null;

	protected 	static Map<String,Monitor> 					monMap 					= new HashMap<String,Monitor>();
	public		static Map<String,Channel> 					pvMap 					= new HashMap<String,Channel>();
	protected	static HashMap<String,Object>				disabledPVMap			= new HashMap<String,Object>();
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////// connection methods ////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	@Override protected void i_connect()
	{
		try
		{
			jca = JCALibrary.getInstance();
			DefaultConfiguration conf = new DefaultConfiguration("context");
			conf.setAttribute("class", JCALibrary.CHANNEL_ACCESS_JAVA);
			
			String addrList = System.getenv("EPICS_CA_ADDR_LIST");
			if (DBConfig.sim5) {
				addrList+=" 127.0.0.1:8002";
			}
			System.out.println("Using EPICS_CA_ADDR_LIST = "+addrList);
			conf.setAttribute("addr_list", addrList);
			conf.setAttribute("auto_addr_list", "false");
			
			_ctxt = jca.createContext(conf);
		}
		catch(Exception ex) { ex.printStackTrace(); }

		i_startPollThread();
		
		EpicsStatus.start();
	}
	
	@Override protected void i_startPollThread()
	{
		new Thread(new Runnable(){@Override
		public void run()
		{
			long lastMsgTime = System.currentTimeMillis();
			
			while (true)
			{
				String reason = null;
				
				if (Util.atExit())
				{
					reason = "Util.atExit()==true";
				}
				if (GUIStatus.inGUIMode)
				{
					if (null != GUI.getShell() && GUI.getShell().isDisposed())
					{
						reason = "Shell is disposed";
					}
				}
				if (null == Epics.ctxt())
				{
					reason = "Context is null";
				}
				if (null != reason)
				{
					long now = System.currentTimeMillis();
					if (now-lastMsgTime>2000)
					{
						System.out.println(reason+", not polling");
						lastMsgTime=now;
					}
					Util.sleep(50);
					continue;
				}
				
				i_poll();
				Util.sleep(5);
			}
		}}, "Epics poll").start();
	}
	
	protected synchronized boolean connectIfNeeded(String pvName, boolean waitForConnect)
	{
		if (disabledPVMap.containsKey(pvName)) { disabledPVMap.put(pvName,"0"); return true; }
		
		if (null == pvName)
		{
			System.out.println("ERROR: null PV");
		}
		
		if (!pvMap.containsKey(pvName))
		{
			try
			{
				System.out.println("Connecting PV "+pvName);
				Channel ch = ctxt().createChannel(pvName, (gov.anl.gmca.io.hw.Epics)inst);
				
				i_poll();

				pvMap.put(pvName, ch);

				if (!waitForConnect)
				{
					EpicsStatus.connectionInitiated(ch);
					return true;
				}
				
				long startTime = System.currentTimeMillis();
				ConnectionState state = ch.getConnectionState();
				while ( state != ConnectionState.CONNECTED)
				{
   					_ctxt.poll();
   					state = ch.getConnectionState();
   					if ( (System.currentTimeMillis() - startTime) > CONNECT_TIMEOUT_MS) break;
				}
				if ( state != ConnectionState.CONNECTED )
				{
					Log.warn("PV "+pvName+" did not connect, state="+state);
				}
				
				return true;
			}
			catch (CAException cae) 		{ cae.printStackTrace(); }
			catch (IllegalStateException e) { e.printStackTrace(); }
			//catch (TimeoutException e) 		{ e.printStackTrace(); }
		}
		return false;
	}
	
	@Override protected void i_disconnectMonitors()
	{
		try
		{
			imonMapLock.acquire();
			
			for (String pvName : imonMap.keySet())
			{
				System.out.println("Disconnecting "+pvName);
				pvMap.get(pvName).dispose();
			}
		}
		catch (InterruptedException e)
		{
			e.printStackTrace();
		}
		finally
		{
			imonMapLock.release();
		}
	}
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////// event methods ///////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	@Override public void connectionChanged(ConnectionEvent ev)
	{
		try
		{
			CAJChannel ch = (CAJChannel)ev.getSource();
			
			// connect the monitor
			if (ev.isConnected())
			{
				Monitor mon = ch.addMonitor(Monitor.ALARM|Monitor.VALUE);
				mon.addMonitorListener(this);
				monMap.put(ch.getName(),mon);
				_ctxt.flushIO();
			}
			else // on disconnect, notify listeners
			{
				try
				{
					imonMapLock.acquire();
					try
					{
		    			ArrayList<EpicsMonitor> al = imonMap.get(ch.getName());
		    			if (null != al)
		    			{
		        			for (final EpicsMonitor imon : al)
		        			{
		        				if (imon != null && GUIStatus.inGUIMode && GUI.appState != GUI.STOPPING)
		        				{
		        					executor.submit(new EpicsTask(ch.getName(), null, imon));
		        				}
		        			}
		    			}
					}
					catch (Exception e) { e.printStackTrace(); }
					finally { imonMapLock.release(); }
				}
				catch (Exception e) { e.printStackTrace(); }
			}
		}
		catch (CAException cae) { cae.printStackTrace(); }
	}

	// when a monitor changes:
	// 1) update the DBR cache
	// 2) call all listeners via the Executor
	@Override public void monitorChanged(MonitorEvent ev)
	{
		// if the monitor is disconnected, don't do anything
		if (null == ev.getStatus())
		{
			System.out.println("ERROR: null status");
			return;
		}
		
		try 
		{
			CAJChannel ch = (CAJChannel)ev.getSource();
			if (null == ch)
			{
				System.out.println("ERROR: null channel");
				return;
			}
			final String channelName = ch.getName();
			final DBR evDBR = ev.getDBR();
			
			dbrCache.put(channelName,evDBR);
			
			if (LOG_CALLBACKS)
			{
				System.out.println("callback "+channelName+" "+dbrToString(ev.getDBR(),6));
			}
			
			dbrCache.put(channelName,evDBR);
			
			imonMapLock.acquire();
			
			try
			{
    			ArrayList<EpicsMonitor> al = imonMap.get(channelName);
    			if (null != al)
    			{
        			for (final EpicsMonitor imon : al)
        			{
        				if (imon != null)
        				{
        					if (GUIStatus.inGUIMode && GUI.appState == GUI.STOPPING) break;
        					
        					//System.out.println("Submitting event for "+channelName+" to "+imon);
        					executor.submit(new EpicsTask(channelName, evDBR, imon));
        				}
        			}
    			}
			}
			catch (Exception e) { e.printStackTrace(); }
			finally
			{
				imonMapLock.release();
			}
			
			dbrPostCache.put(channelName,evDBR);
		}
		catch (Exception e) { e.printStackTrace(); }
	}

	@Override protected void i_disablePV(PVName pvName, Object defaultValue)
	{
    	String sval = "0";
    	if (defaultValue instanceof String) sval = (String)defaultValue;
    	if (defaultValue instanceof Double) sval = Util.printDouble((Double)defaultValue, 3);
    	disabledPVMap.put(pvName.toString(), sval);
	}

	// in hardware mode, for exceptions, print an error and return an error value
	@Override protected int i_dbrToInt(DBR dbr)
	{
		try 				{ return gov.anl.gmca.io.Epics.d_dbrToInt(dbr); }
		catch (Exception e) { e.printStackTrace(); return Integer.MAX_VALUE;	}
	}
	@Override protected double i_dbrToDouble(DBR dbr)
	{
		try 				{ return gov.anl.gmca.io.Epics.d_dbrToDouble(dbr); }
		catch (Exception e) { e.printStackTrace(); return Double.NaN;	}
	}
	@Override protected int[] i_dbrToIntArray(DBR dbr)
	{
		try 				{ return gov.anl.gmca.io.Epics.d_dbrToIntArray(dbr); }
		catch (Exception e) { e.printStackTrace(); return null;	}
	}
	@Override protected double[] i_dbrToDoubleArray(DBR dbr)
	{
		try 				{ return gov.anl.gmca.io.Epics.d_dbrToDoubleArray(dbr); }
		catch (Exception e) { e.printStackTrace(); return null;	}
	}
	@Override protected String i_dbrToString(DBR dbr, int decimalsToKeep)
	{
		try 				{ return gov.anl.gmca.io.Epics.d_dbrToString(dbr,decimalsToKeep); }
		catch (Exception e) { e.printStackTrace(); return "";	}
	}

	// this method blocks until either the PV connects or the timeout is reached
	@Override protected boolean i_checkConnection(String pvName, int timeoutMs)
	{
		if (i_isConnected(pvName)) return true;
		
		long startMs = System.currentTimeMillis();
		connectIfNeeded(pvName, false);
		if (PVLookup.isWaveform(pvName)) connectIfNeeded(pvName+".NORD",false);

		int printedMsgForSec = 0;
		while (true)
		{
			Util.sleep(10);
			if (i_isConnected(pvName)) return true;
			
			// how long has elapsed?
			long nowMs = System.currentTimeMillis();
			long elapsedMs = nowMs - startMs;
			
			// print a message each second
			if (printedMsgForSec < elapsedMs/1000)
			{
				Channel ch = pvMap.get(pvName);
				String st = null==ch ? "null channel" : ch.getConnectionState().toString();
				System.out.println("In checkConnection waiting for "+pvName+" to connect, state = "+st);
				printedMsgForSec++;
			}
			
			// check for timeout
			if (elapsedMs > timeoutMs)
			{
				//System.out.println("Timed out checking connection for "+pvName);
				return false;
			}
		}
	}
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////// EpicsTask ///////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// a single static variable works because the EPICS executor is single-threaded
	public static volatile long lastCallbackStart = 0;
	public static volatile EpicsTask currentTask = null;
	private static List<EpicsTask> epicsTasks = Collections.synchronizedList(new LinkedList<EpicsTask>());
	@Override protected int i_getEpicsTaskCount()
	{
		return epicsTasks.size();
	}
	
	public static class EpicsTask implements Runnable
	{
		protected String channelName;
		protected DBR evDBR;
		protected EpicsMonitor imon;
		protected Thread runThread = null;
		
		public EpicsTask(String channelName, DBR evDBR, EpicsMonitor imon)
		{
			this.channelName = channelName;
			this.evDBR = evDBR;
			this.imon = imon;
			epicsTasks.add(this);
		}
		
		public void run()
		{
			//String threadInfo = channelName+"@"+imon+"("+Thread.currentThread().getName()+")";
			//System.out.println("Executing thread "+threadInfo);
			//PerfTimer t = new PerfTimer(threadInfo+": ");
			
			try
			{
				runThread = Thread.currentThread();
				
				currentTask = this;
				lastCallbackStart = System.currentTimeMillis();
				
				// this is called for each monitor of a given PV, but it simply
				// has no effect after the first one
				if (null != evDBR) dbrCache.put(channelName,evDBR);
				
				if (imon != null)
				{
					if (evDBR != null)
					{
						imon.valueChanged(channelName,evDBR);
						if (logPVSet.contains(channelName)) System.out.println("camonitor "+channelName+" "+dbrToString(evDBR));
					}
					else
					{
						imon.connectionChanged(channelName, false);
						if (logPVSet.contains(channelName)) System.out.println("PV disconnected: "+channelName);
					}
				}
			}
			catch (Exception e) { e.printStackTrace(); }
			finally
			{
				epicsTasks.remove(this);
				currentTask = null;
				lastCallbackStart = 0;
			}
			
			//t.printElapsedMs();
		}
		
		public void printFirstHangInfo()
		{
			System.out.println("EPICS HANG DETECTED");
			System.out.println("Monitor not returning: "+imon);
			Util.printAllStackTraces();
		}
		
		public void printContinuingHangInfo()
		{
			System.out.println("Monitor not returning: "+imon);
			Util.printAllStackTraces();
		}
	}
}
