/**
 * Parses XPLAN.LP file generated by XDS and generates startegy by maximizing the completeness
 * Typical XPLAN.LP starts line this
 * 
 ************************************************************************

 GUIDE FOR OPTIMAL SELECTION OF DATA COLLECTION STRATEGY

 ************************************************************************

 Assuming that the currently processed data satisfy

 SPACE_GROUP_NUMBER=    3
 UNIT_CELL_CONSTANTS=    48.63    51.74    83.34  90.000  92.770  90.000
 FRIEDEL'S_LAW=TRUE  
 STARTING_ANGLE=     0.000
 STARTING_FRAME=    99

 you would collect a maximum of data for the following combinations
 of starting spindle_angle and total range of rotation (degrees):

 starting at     total rotation     completeness        multiplicity of
 spindle_angle        range          of data (%)        measurements

 20.00           30.00            55.98                 1.14
 100.00           60.00            88.31                 1.46
 170.00           90.00            97.81                 1.99
 150.00          120.00            97.99                 2.68
 120.00          150.00            98.04                 3.33
 100.00          180.00            98.07                 3.98
 */
package gov.anl.gmca.multistrategy;

import gov.anl.gmca.syslog.Log;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.*;

public class ModelMultiStrategyParams {
	
	public String warning = "processing";
	public String workDir = "";
	public String sampleName = "";
	public String referencedata = "";
	public String minrange = "5";
	public String images = "";

	public String spaceGroup = "";
	public String unitCell = "";
	public String displaytext = "";
	public String strategy = "";

	public String resolution = "";
	public String lowresolution = "";

	public String highresolution = "";

	public String xplanLog = "";

	public String state = "";

	public enum XPLANStrategy {
		TOTAL_RANGE, START_ANGLE, RESOLUTION, OLD_COMPLETENESS, NEW_COMPLETENESS, NEWOROLD, NEWANDOLD;

	};

	protected Map<String, String> singleMap = new HashMap<String, String>();
	List<Map<String, String>> multiMap = new ArrayList<Map<String, String>>();

	public void processResult(String _workDir) throws Exception {
		String xdsLog = "";
		int pos1, pos2;
		workDir = _workDir;
		
		/* Read XDS log file */
		xdsLog = workDir + "/XDS.LOG";

		xplanLog = workDir + "/XPLAN.LP";

		File xdsFile = new File(xdsLog);
		if (!xdsFile.exists()) {
			Log.error(xdsLog + " not found");
			state = "ERROR";
			warning = xdsLog + " does not exists check data!!!";
			return;
		} else {
			String xdsContent = readLogFile(xdsLog, true);

			if (xdsContent.contains("!!! ERROR ")) {
				pos1 = xdsContent.lastIndexOf("!!! ERROR ");
				warning = xdsContent.substring(pos1,
						xdsContent.indexOf("\n", pos1));
				warning += "\nindexing using provided parameters may have failed.";
				state = "ERROR";
				return;
			}
		}
		
		/* Read XPLAN log file */
		String content = readLogFile(xplanLog, false);
		pos1 = content
				.indexOf("GUIDE FOR OPTIMAL SELECTION OF DATA COLLECTION STRATEGY");
		pos2 = content.indexOf(
				"DETAILED REPORT FOR EACH DATA COLLECTION SCENARIO", pos1);

		if (pos2 < 0 || pos1 < 0) {
			warning = "No strategy found in XPLAN.LOG";
			state = "ERROR";
			return;
		}

		displaytext = content.substring(pos1 - 88, pos2);
		displaytext = displaytext.replace("'", "");

		pos2 = content.indexOf("SPACE_GROUP_NUMBER=", pos1);

		if (pos2 < 0) {
			warning = "No spacegroup found in XPLAN.LOG";
			state = "ERROR";
			return;
		}

		spaceGroup = content.substring(pos2 + 20, content.indexOf("\n", pos2));

		pos2 = content.indexOf("UNIT_CELL_CONSTANTS=", pos1);

		if (pos2 < 0) {
			warning = "No spacegroup found in XPLAN.LOG";
			state = "ERROR";
			return;
		}

		unitCell = content.substring(pos2 + 20, content.indexOf("\n", pos2));

		int i = 0, j = 0;
		displaytext = "";
		String trStr = "TOTAL ROTATION RANGE =";
		String startStr = "COMPLETENESS OF DATA COLLECTED IN THE OSCILLATION RANGE";

		int pos3, pos4;
		pos2 = content.indexOf(trStr, pos2);
		do {

			String total_range = content.substring(pos2 + trStr.length() + 1,
					content.indexOf("\n", pos2));

			singleMap.put("TOTAL_RANGE", total_range);

			pos3 = content.indexOf(trStr, pos2 + trStr.length() + 1);
			pos4 = content.indexOf(startStr, pos2);

			if (pos3 < 0) {
				pos3 = content.length();
				String subr = content.substring(pos2, pos3);
				pos3 = content.lastIndexOf(startStr);
			}
			double max = 0, newValDouble;
			String maxStr = "", angle = "";
			do {
				String startAngle = content.substring(pos4 + startStr.length()
						+ 1, content.indexOf("DEGREES", pos4) - 1);
				String complete = content
						.substring(
								content.indexOf("\n", pos4) + 1,
								content.indexOf(startStr,
										pos4 + startStr.length() + 1));
				StringTokenizer tok = new StringTokenizer(complete);
				String tokh = "";

				j = 0;

				while (tok.hasMoreTokens()) {
					tokh += tok.nextToken() + " ";
					j++;
					if (j == 10) {
						/* finding maximum value of NEW or OLD */
						String newVal = tok.nextToken();
						newVal = newVal.substring(0, newVal.length() - 1);
						newValDouble = Double.parseDouble(newVal);

						if (newValDouble > max) {
							max = newValDouble;
							maxStr = complete;
							angle = startAngle;
						}
					}
				}
				pos4 = content.indexOf(startStr, pos4 + startStr.length() + 1);
			} while (pos4 < pos3);
			StringTokenizer tok = new StringTokenizer(maxStr);
			String tStr = "";
			if (i == 0)
				displaytext += "\nResolution\t" + tok.nextToken() + "\n";
			else
				tStr = tok.nextToken();

			tStr = tok.nextToken();
			tStr = tok.nextToken();
			tStr = tok.nextToken();
			if (i == 0) {
				displaytext += "OLD DATA\t" + tStr + "\n\n";

				displaytext += "Best Rot. range  \t NEW DATA \t Combined \t Common\n\n";

			}
			j = 0;
			while (tok.hasMoreTokens()) {
				j++;
				tStr = tok.nextToken();

				if (j == 1)

					strategy += angle.replaceAll("\\s+","") + "\t\t";
				if (j == 3 || j == 7 || j == 11)
					strategy += tStr + "\t\t";
			}

			displaytext += "\n";
			strategy += "\n";
			// Log.note(displaytext);
			pos2 = content.indexOf(trStr, pos2 + trStr.length() + 1);
			multiMap.add(i, singleMap);
			i++;
		} while (pos2 != -1);

	}
	public static String readLogFile(String logFile, boolean onlyheader) {
		
		byte[] buffer;
		if(onlyheader) {
			buffer = new byte[10240]; // just read 10 KB
		} else {
			buffer = new byte[(int) new File(logFile).length()];
		}
		BufferedInputStream f = null;
		try {
			f = new BufferedInputStream(new FileInputStream(logFile));
			f.read(buffer);
			f.close();
		} catch (IOException e) {
			System.out.println(e.getMessage());
			e.printStackTrace();
		} finally {
			if (f != null)
				try {
					f.close();
				} catch (IOException ignored) {
				}
		}

		return new String(buffer);
	}
}
