package gov.anl.gmca.raster;

import gov.anl.gmca.collect.DevFPE;
import gov.anl.gmca.collect.ModelCollectRun.Point3D;
import gov.anl.gmca.common.DevMove.MoveCmd;
import gov.anl.gmca.common.DevMove.MoveResult;
import gov.anl.gmca.common.DevMove.ResultType;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.hutch.DevXYZ;
import gov.anl.gmca.raster.ModelRasterRun.RefinementType;
import gov.anl.gmca.raster.ModelRasterRun.RunState;
import gov.anl.gmca.raster.ModelRasterSiteList.Site;
import gov.anl.gmca.raster.ModelRasterState.RasterListType;
import gov.anl.gmca.raster.ModelRasterState.RasterState;
import gov.anl.gmca.raster.calc.GridCalculator;
import gov.anl.gmca.raster.calc.RasterGeometry;
import gov.anl.gmca.raster.util.ClearSampleState;
import gov.anl.gmca.sample.DevCollimatorCtrl;
import gov.anl.gmca.sample.DevSlider;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.*;

import java.io.File;
import java.util.ArrayList;

// The main method handles the control flow and passing of hit information between runs
//
// The manage methods determine whether runs were already made or not, and if they were,
// it returns the existing results
//
// The individual run type methods set up each run and return their results
//
public class CmdRasterAuto extends CmdRaster
{
	protected int sitesFound = 0;
	protected int currentRun = 1;
	protected boolean resetAtStart = true;
	
	protected DevRasterRunList autoListDev;
	protected ModelRasterRunList autoList;
	protected ModelRasterRun run0, coarseRun;
	protected DevRasterRun run0Dev;
	
	public CmdRasterAuto()
	{
		super(null);
		
		autoListDev = DevRasterRunList.get(RasterListType.AUTO);
		autoList 	= ModelRasterRunList.get(RasterListType.AUTO);
		run0 		= (ModelRasterRun)autoList.getRun(0);
		run0Dev 	= (DevRasterRun)autoListDev.getRunDev(0);
	}

	/////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////// method for running in the screening tab //////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Returns null on error
	// WARNING: This will silently delete any auto raster results
	public static ArrayList<Site> runUnattended()
	{
		// change tab to auto
		DevRasterState.get().setAutoMode(RasterListType.AUTO);

		// clear any existing state
		resetAutoRaster();

		// set defaults?
		
		// run synchronously
		CmdRasterAuto cmd = new CmdRasterAuto();
		cmd.resetAtStart=false; // reset was already done above
		cmd.run();
		
		Util.sleep(1000);
		
		ArrayList<Site> allSites = ModelRasterSiteList.get().getSiteListCopy();
		ArrayList<Site> ret = new ArrayList<Site>();
		for (Site site : allSites)
		{
			if (site.orthoRunIndex>0) ret.add(site);
		}
		return ret;
	}
	
	private static void resetAutoRaster()
	{
		// reset auto raster
		ClearSampleState.deleteAutoSitesAndHits(false);
		ModelRasterRun autoRun0 = (ModelRasterRun)ModelRasterRunList.get(RasterListType.AUTO).getRun(0);
		autoRun0.resetAutoRaster();
		autoRun0.grid().clearAllContours(false, false);
		Util.sleep(1000);
	}
	
	/////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////// main method /////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////
	
	@Override public void run()
	{
		// check FPE connection
		if (!DevFPE.get().checkFPEState(null, false))
		{
			endRaster(RunState.INACTIVE);
			return;
		}
		
		startRaster();
		
		// check raster dir
		if (!checkDirectory(run0).succeeded()) { endRaster(RunState.INACTIVE); return; }

		// coarse run
		ArrayList<ModelRasterHits.Hit> coarseHits = manageCoarseRun();
		if (null == coarseHits) { endRaster(RunState.PAUSED); return; }
		
		boolean quitEarly = false;
		for (ModelRasterHits.Hit coarseHit : coarseHits)
		{
			boolean isLastCoarseHit = (coarseHit==coarseHits.get(coarseHits.size()-1));
			
			if (isSingleCellHit(coarseHit))
			{
				// fine run
				Pair<ModelRasterRun,ArrayList<ModelRasterHits.Hit>> fineResult = manageFineRun(coarseHit, isLastCoarseHit);
				if (null==fineResult || null==fineResult.b) { quitEarly=true; break; }
				
				for (ModelRasterHits.Hit fineHit : fineResult.b)
				{
					boolean isLastFineHit = (fineHit==fineResult.b.get(fineResult.b.size()-1));
					
					// ortho run
					boolean orthoSucceeded = manageOrthoRun(fineResult.a, fineHit, isLastFineHit&&isLastCoarseHit);
					if (!orthoSucceeded) { quitEarly=true; break; }
				}
			}
			else
			{
				// ortho run
				boolean orthoSucceeded = manageOrthoRun(null, coarseHit, isLastCoarseHit);
				if (!orthoSucceeded) { quitEarly=true; break; }
			}
		}
	
		endRaster(quitEarly?RunState.PAUSED:RunState.COMPLETE);
	}
	
	/////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////// manage methods /////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////
	
	private ArrayList<ModelRasterHits.Hit> manageCoarseRun()
	{
		boolean coarseRunPreviouslyCompleted =
		(
	    	autoList.getRunCount()>=1 &&
	    	((ModelRasterRun)autoList.getRun(1)).getRunState() == RunState.COMPLETE
	    );
		
		if (!coarseRunPreviouslyCompleted)
		{
			if (!coarseRun()) return null;
		}
		
		coarseRun = (ModelRasterRun)autoList.getRun(1);
		ArrayList<ModelRasterHits.Hit> hits = ModelRasterHits.get().getCopyOfHits(coarseRun);

		boolean isLastRun = hits.size()==0;
		if (!coarseRunPreviouslyCompleted) printFinishNotes("Coarse run found "+hits.size()+" hits", isLastRun);
		
		return hits;
	}
	
	private Pair<ModelRasterRun,ArrayList<ModelRasterHits.Hit>> manageFineRun(ModelRasterHits.Hit coarseHit, boolean couldBeLastRun)
	{
		if (RasterState.STOPPING == ModelRasterState.get().getRasterState()) return null;
		if (!incrementRunCount()) return null;
		
		ModelRasterRun fineRun = null;
		
		// find the fine run if it has been started
		fineRun = findRunThatRefinesHit(1, coarseHit);
		boolean fineRunPreviouslyCompleted =
		(
			null != fineRun &&
			RunState.COMPLETE == fineRun.getRunState()
		);
		
		// perform the fine run unless it is complete
		if (!fineRunPreviouslyCompleted)
		{
			Log.note("Starting/resuming fine run for hit "+coarseHit.getNumber());
			boolean fineRunSucceeded = fineRun(currentRun, coarseHit);
			if (!fineRunSucceeded) return null;
			
			fineRun = (ModelRasterRun)autoList.getRun(currentRun);
		}
		
		ArrayList<ModelRasterHits.Hit> fineHits = ModelRasterHits.get().getCopyOfHits(fineRun);
		
		boolean isLastRun = couldBeLastRun && fineHits.size()==0;
		if (!fineRunPreviouslyCompleted) printFinishNotes("Fine run found "+fineHits.size()+" hits", isLastRun);
		
		return new Pair<ModelRasterRun,ArrayList<ModelRasterHits.Hit>>(fineRun,fineHits);
	}
	
	private boolean manageOrthoRun(ModelRasterRun fineRun, ModelRasterHits.Hit refHit, boolean isLastRun)
	{
		if (RasterState.STOPPING == ModelRasterState.get().getRasterState()) return false;
		if (!incrementRunCount()) return false;
		
		ModelRasterRun refRun = (fineRun!=null)?fineRun:coarseRun;
		
		// find the ortho run if it has been started
		ModelRasterRun orthoRun = findRunThatRefinesHit(refRun.getRunIndex(), refHit);
		boolean orthoRunPreviouslyCompleted =
		(
			null != orthoRun &&
			RunState.COMPLETE == orthoRun.getRunState()
		);
		
		// perform the ortho run unless it is complete
		if (!orthoRunPreviouslyCompleted)
		{
			Log.note("Starting/resuming ortho run for hit "+refHit.getNumber());
			boolean orthoRunSucceeded = orthoRun(currentRun, refRun.getRunIndex(), refHit);
			if (!orthoRunSucceeded) return false;
			
			orthoRun = (ModelRasterRun)autoList.getRun(currentRun);
		}
		
		// add the hit here as a final site
		ArrayList<ModelRasterHits.Hit> orthoHits = ModelRasterHits.get().getCopyOfHits(orthoRun);
		if (!orthoRunPreviouslyCompleted)
		{
			for (ModelRasterHits.Hit orthoHit : orthoHits)
			{
				Point4D hitXYZO = orthoHit.getMotorXYZO();
				hitXYZO.omega = hitXYZO.omega - 90.0; // use coarse omega
				Site site = new Site
				(
					-1,
					-1,-1,
					(null==fineRun)?-1:fineRun.getRunIndex(),
					orthoRun.getRunIndex(),
					orthoHit.getNumber(),
					hitXYZO, new Point3D
					(
						refHit.getCellCountX() * refRun.grid().getCellWidth_um(),
						refHit.getCellCountY() * refRun.grid().getCellHeight_um(),
						orthoHit.getCellCountY() * orthoRun.grid().getCellHeight_um()
					)
				);
				DBRasterSiteList.get().addSite(site, true);
			}

			printFinishNotes("Ortho run found "+orthoHits.size()+" hits", isLastRun);
		}
		
		return true;
	}
	
	/////////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////// individual run types //////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////
	
	private boolean coarseRun()
	{
		deletePickleFile();
		
		int coarseIndex = 1;
		boolean runInProgress = autoList.getRunCount()>=coarseIndex;
		
		if (!runInProgress)
		{
			// set things here that should only be set at the very beginning of an auto raster sequence
			run0Dev.setHideRunParameters(true, false);
			DBRasterSiteList.get().clearSites(true);
			
    		if (!runAutoFindIfNeeded()) return false;
    		
			// set up run
			DevRasterRun.s_setAttenuation_factors	(RasterListType.AUTO, 1, run0.getAttenuation());
    		DevRasterRun.s_setBeamHeight_um			(RasterListType.AUTO, 1, run0.getBeamWidth_um());
    		DevRasterRun.s_setBeamWidth_um			(RasterListType.AUTO, 1, run0.getBeamHeight_um());
    		DevRasterRun.s_setCellWidth_um			(RasterListType.AUTO, 1, run0.grid().getCellWidth_um());
    		DevRasterRun.s_setCellHeight_um			(RasterListType.AUTO, 1, run0.grid().getCellHeight_um());
    		DevRasterRun.s_setCenter				(RasterListType.AUTO, 1, run0.grid().getCenter());
    		DevRasterRun.s_setContours				(RasterListType.AUTO, 1, run0.grid().getContours());
    		DevRasterRun.s_setDelta					(RasterListType.AUTO, 1, run0.getDelta());
    		DevRasterRun.s_setDirectory				(RasterListType.AUTO, 1, run0.getDirectorySuffix());
    		DevRasterRun.s_setDistance_mm			(RasterListType.AUTO, 1, run0.getDetectorDist());
    		DevRasterRun.s_setExposureTime			(RasterListType.AUTO, 1, run0.getExposureTime());
    		DevRasterRun.s_setFileroot				(RasterListType.AUTO, 1, run0.fileRoot);
    		DevRasterRun.s_setGridHOffsets			(RasterListType.AUTO, 1, run0.grid().getCorrectedLeftOffset_um(), run0.grid().getCorrectedRightOffset_um());
    		DevRasterRun.s_setGridVOffsets			(RasterListType.AUTO, 1, run0.grid().getCorrectedTopOffset_um(), run0.grid().getCorrectedBottomOffset_um());
    		DevRasterRun.s_setLowResLimit_A			(RasterListType.AUTO, 1, run0.getLowResLimit_A());
    		DevRasterRun.s_setHighResLimit_A		(RasterListType.AUTO, 1, run0.getHighResLimit_A());
    		DevRasterRun.s_setStartFrameLabel		(RasterListType.AUTO, 1, 1);
    		DevRasterRun.s_setNextFrame				(RasterListType.AUTO, 1, 0);
    		DevRasterRun.s_setRefinementType		(RasterListType.AUTO, 1, RefinementType.COARSE);
    		DevRasterRun.s_setRunLabel				(RasterListType.AUTO, 1, 1);
    		DevRasterRun.s_setRunState				(RasterListType.AUTO, 1, RunState.INACTIVE);
    		DevRasterRun.s_setSelectedCells			(RasterListType.AUTO, 1, run0.dev().getSelectedCells());
    		DevRasterRun.s_setStartFrameLabel		(RasterListType.AUTO, 1, 1);
    		
    		autoListDev.setRunCount(1, true);
    		
    		DevRasterRun coarseDev = (DevRasterRun)autoListDev.getRunDev(1);
    		while (!coarseDev.connected()) { Util.sleep(10); }
		}
		
		// retrieve run object
		ModelRasterRun coarseRun = (ModelRasterRun)autoList.getRun(1);

		// is this needed?
		coarseRun.grid().getGridCalc().fitGridToContour(false,true);

		// show coarse run
		autoListDev.setCurrentRun(1, true);
		
		// perform the run
		Log.note("Starting coarse run");
		CmdRasterDiff coarseCmd = new CmdRasterDiff((ModelRasterRun)autoList.getRun(1));
		ActionResult coarseResult = coarseCmd.runAll();
		
		if (RunState.INACTIVE == coarseRun.getRunState() || RunState.PAUSED == coarseRun.getRunState())
		{
			return false;
		}

		// check result
		if (!coarseResult.succeeded())
		{
			Log.error("Coarse run failed");
			DevRasterState.get().setRasterState(RasterState.INACTIVE);
			return false;
		}
		
		// find hits
		if (coarseRun.getRunState() == RunState.COMPLETE)
		{
			Log.note("Waiting for coarse run processing...");
			coarseCmd.waitForProcessing();
			
			//coarseRun.grid().getHitFinder().find(true, false);
		}
		
		return true;
	}
	
	private boolean fineRun(int fineIndex, ModelRasterHits.Hit hit)
	{
		deletePickleFile();
		
		String hitInfo =
			"#"+hit.getNumber()+
			Util.sprintf(" (%.3f",hit.getMotorXYZO().x)+
			Util.sprintf(",%.3f",hit.getMotorXYZO().y)+
			Util.sprintf(",%.3f",hit.getMotorXYZO().z)+
			Util.sprintf(",%.3f)",hit.getMotorXYZO().omega);
			
		Log.note("Fine run index "+fineIndex+" running for hit "+hitInfo);
		
		boolean runInProgress = autoList.getRunCount()>=fineIndex;
		
		// set up run if necessary
		if (!runInProgress)
		{
			// center on hit
			Log.note("Centering on coarse run hit...");
			centerSample(hit.getMotorXYZO());
			Log.note("Centering complete.");
			
    		// pre-configure run
			DevRasterRun.s_setAttenuation_factors	(RasterListType.AUTO, fineIndex, run0.getAttenuation());
    		DevRasterRun.s_setCenter				(RasterListType.AUTO, fineIndex, RasterGeometry.getCurrentMotorPositions());
    		DevRasterRun.s_setDelta					(RasterListType.AUTO, fineIndex, run0.getDelta());
    		DevRasterRun.s_setDirectory				(RasterListType.AUTO, fineIndex, run0.getDirectorySuffix());
    		DevRasterRun.s_setDistance_mm			(RasterListType.AUTO, fineIndex, run0.getDetectorDist());
    		DevRasterRun.s_setExposureTime			(RasterListType.AUTO, fineIndex, run0.getExposureTime());
    		DevRasterRun.s_setFileroot				(RasterListType.AUTO, fineIndex, run0.fileRoot);
    		DevRasterRun.s_setLowResLimit_A			(RasterListType.AUTO, fineIndex, run0.getLowResLimit_A());
    		DevRasterRun.s_setHighResLimit_A		(RasterListType.AUTO, fineIndex, run0.getHighResLimit_A());
    		DevRasterRun.s_setStartFrameLabel		(RasterListType.AUTO, fineIndex, 1);
    		DevRasterRun.s_setNextFrame				(RasterListType.AUTO, fineIndex, 0);
    		DevRasterRun.s_setReferenceRun			(RasterListType.AUTO, fineIndex, 1);
    		DevRasterRun.s_setRunLabel				(RasterListType.AUTO, fineIndex, fineIndex);
    		DevRasterRun.s_setRunState				(RasterListType.AUTO, fineIndex, RunState.INACTIVE);
    		DevRasterRun.s_setReferenceHit			(RasterListType.AUTO, fineIndex, hit.getNumber());
    		DevRasterRun.s_setRefinementType		(RasterListType.AUTO, fineIndex, RefinementType.FINE);
    		DevRasterRun.s_setStartFrameLabel		(RasterListType.AUTO, fineIndex, 1);

    		setSearchArea
			(
				fineIndex,
				RasterListType.AUTO,
				new Point2D
				(
					run0.grid().getCellWidth_um(),
					run0.grid().getCellHeight_um()
				)
			);
    		
			// create/connect the run
    		autoListDev.setRunCount(fineIndex, true);
    		
    		// wait for connection
    		DevRasterRun fineDev = (DevRasterRun)autoListDev.getRunDev(fineIndex);
    		while (!fineDev.connected()) { Util.sleep(10); }
		}
		
		// retreive run object
		ModelRasterRun fineRun = (ModelRasterRun)autoList.getRun(fineIndex);
		
		// show the run
		autoListDev.setCurrentRun(fineIndex, true);
		
		// perform the run
		CmdRasterDiff fineCmd = new CmdRasterDiff(fineRun);
		ActionResult fineResult = fineCmd.runAll();
		
		if (RunState.INACTIVE == fineRun.getRunState() || RunState.PAUSED == fineRun.getRunState())
		{
			return false;
		}

		if (!fineResult.succeeded())
		{
			DevRasterState.get().setRasterState(RasterState.INACTIVE);
			return false;
		}

		// find hits
		if (fineRun.getRunState() == RunState.COMPLETE)
		{
			Log.note("Waiting for fine run processing...");
			fineCmd.waitForProcessing();
			//fineRun.grid().getHitFinder().find(true, false);
		}
		
		return true;
	}
	
	private boolean orthoRun(int orthoIndex, int referenceIndex, ModelRasterHits.Hit hit)
	{
		deletePickleFile();
		
		Log.note("Ortho run index "+orthoIndex+" running for hit "+hit.getNumber());
		
		ModelRasterRun refRun = (ModelRasterRun)autoList.getRun(referenceIndex);

		boolean runInProgress = autoList.getRunCount()>=orthoIndex;
		
		// set up run if necessary
		if (!runInProgress)
		{
			// center on hit + 90 degrees
			Point4D orthoCenter = new Point4D
			(
				hit.getMotorXYZO().x,
				hit.getMotorXYZO().y,
				hit.getMotorXYZO().z,
				(
					hit.getMotorXYZO().omega<270.0
					?hit.getMotorXYZO().omega+90.0
					:hit.getMotorXYZO().omega-270.0
				)
			);
			centerSample(orthoCenter);
			
			// calculate ortho grid height
			double orthoGridHeight;
			if (run0.getAutoFind())
			{
				int screenX = hit.getScreenXYFromGridCenter(run0.grid().getCenter()).x;
				orthoGridHeight = DevAutoFind.get().getOrthoHeight(screenX);
			}
			else
			{
				orthoGridHeight = run0.getLoopMeasurement_um();
			}
			
    		// pre-configure run
			DevRasterRun.s_setAttenuation_factors	(RasterListType.AUTO, orthoIndex, run0.getAttenuation());
    		DevRasterRun.s_setCenter				(RasterListType.AUTO, orthoIndex, RasterGeometry.getCurrentMotorPositions());
    		DevRasterRun.s_setDelta					(RasterListType.AUTO, orthoIndex, run0.getDelta());
    		DevRasterRun.s_setDirectory				(RasterListType.AUTO, orthoIndex, run0.getDirectorySuffix());
    		DevRasterRun.s_setDistance_mm			(RasterListType.AUTO, orthoIndex, run0.getDetectorDist());
    		DevRasterRun.s_setExposureTime			(RasterListType.AUTO, orthoIndex, run0.getExposureTime());
    		DevRasterRun.s_setFileroot				(RasterListType.AUTO, orthoIndex, run0.fileRoot);
    		DevRasterRun.s_setLowResLimit_A			(RasterListType.AUTO, orthoIndex, run0.getLowResLimit_A());
    		DevRasterRun.s_setHighResLimit_A		(RasterListType.AUTO, orthoIndex, run0.getHighResLimit_A());
    		DevRasterRun.s_setStartFrameLabel		(RasterListType.AUTO, orthoIndex, 1);
    		DevRasterRun.s_setNextFrame				(RasterListType.AUTO, orthoIndex, 0);
    		DevRasterRun.s_setReferenceRun			(RasterListType.AUTO, orthoIndex, referenceIndex);
    		DevRasterRun.s_setRunLabel				(RasterListType.AUTO, orthoIndex, orthoIndex);
    		DevRasterRun.s_setRunState				(RasterListType.AUTO, orthoIndex, RunState.INACTIVE);
    		DevRasterRun.s_setReferenceHit			(RasterListType.AUTO, orthoIndex, hit.getNumber());
    		DevRasterRun.s_setRefinementType		(RasterListType.AUTO, orthoIndex, RefinementType.ORTHO);
    		DevRasterRun.s_setStartFrameLabel		(RasterListType.AUTO, orthoIndex, 1);
			setSearchArea
			(
				orthoIndex,
				RasterListType.AUTO,
				new Point2D
				(
					refRun.grid().getCellWidth_um(),
					orthoGridHeight
				)
			);
			
			// create/connect the run
    		autoListDev.setRunCount(orthoIndex, true);
    		
    		// wait for connection
    		DevRasterRun orthoDev = (DevRasterRun)autoListDev.getRunDev(orthoIndex);
    		while (!orthoDev.connected()) { Util.sleep(10); }
		}
		
		// retreive run object
		ModelRasterRun orthoRun = (ModelRasterRun)autoList.getRun(orthoIndex);
		
		// show the run
		autoListDev.setCurrentRun(orthoIndex, true);
		
		// perform the run
		CmdRasterDiff orthoCmd = new CmdRasterDiff(orthoRun);
		ActionResult orthoResult = orthoCmd.runAll();
		
		if (RunState.INACTIVE == orthoRun.getRunState() || RunState.PAUSED == orthoRun.getRunState())
		{
			return false;
		}

		if (!orthoResult.succeeded())
		{
			DevRasterState.get().setRasterState(RasterState.INACTIVE);
			return false;
		}

		// find hits
		if (orthoRun.getRunState() == RunState.COMPLETE)
		{
			Log.note("Waiting for ortho run processing...");
			orthoCmd.waitForProcessing();
			
			boolean referenceRunIsCoarse = (1 == referenceIndex);
			boolean lookForMultipleHits = referenceRunIsCoarse;
			//orthoRun.grid().getHitFinder().find(lookForMultipleHits, false);
		}
		
		return true;
	}
	
	/////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////// helper methods /////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////
	
	private void startRaster()
	{
		run0Dev.setRunStatus(RunState.RASTERING, true);
		DevRasterState.get().setRasterState(RasterState.RASTERING);
		
		// clear existing sites and hits
		if (resetAtStart)
		{
			resetAutoRaster();
		}
		
		// ringlight to max
		DevSlider ringLight = DevSlider.get("ring");
		ringLight.setIntensity(ringLight.getHigherLimit());
	}
	
	private boolean incrementRunCount()
	{
		// increment run count
		currentRun++;
		if (currentRun >= ModelRasterRunList.MAX_COUNT)
		{
			Log.note("Run maximum reached, exiting auto raster"); return false;
		}
		return true;
	}
	
	protected void endRaster(RunState run0state)
	{
		// ringlight off
		DevSlider ringLight = DevSlider.get("ring");
		ringLight.setIntensity(ringLight.getLowerLimit());
		
		// backlight in
	    ccMove = DevCollimatorCtrl.get().newMove();
	    ccMove.setArguments("OUT backlightIN bluice");
	    //ccMove.setDest("backlight",		InOut.IN);
	    //ccMove.setDest("collimator",	InOut.OUT);
	    ccMove.start();
		
		Log.note("Auto raster ended.");
		run0Dev.setRunStatus(run0state, true);
		DevRasterState.get().setRasterState(RasterState.INACTIVE);
	}
	
	private boolean isSingleCellHit(ModelRasterHits.Hit hit)
	{
		return (hit.getCellCountX() == 1 && hit.getCellCountY() == 1);
	}
	
	private boolean runAutoFindIfNeeded()
	{
		if (run0.getAutoFind())
		{
			// clear complete if loop centered is checked
			if (run0.getLoopCentered())
			{
				DevAutoFind.get().setComplete(false);
			}
			
			// run script if needed
			if (run0.getLoopCentered() || !DevAutoFind.get().getComplete())
			{
				boolean failure = false, stopRequested = false;
				DevAutoFind.get().start(run0.getLoopCentered());
				while (DevAutoFind.get().isRunning())
				{
					if (ModelRasterState.get().getRasterState() == RasterState.STOPPING)
					{
						DevAutoFind.get().stop();
						stopRequested = true;
						failure = true;
					}
				}
				if (!failure)
				{
					failure = !DevAutoFind.get().getComplete();
				}
				
				if (failure)
				{
					if (!stopRequested) Log.error("Autofind failed");
					return false;
				}
			}

			// set grid center
			run0.dev().setCenter(RasterGeometry.getCurrentMotorPositions(), false);
			
			// set contour
			ModelContour coarseContour = DevAutoFind.get().getCoarseContour();
			ArrayList<ModelContour> contourList = new ArrayList<ModelContour>();
			contourList.add(coarseContour);
			run0.dev().setContours(contourList, false);
			
			run0.grid().getGridCalc().fitGridToContour(true, true);
		}
		
		return true;
	}
	
	protected void deletePickleFile()
	{
		File pickleFile = new File(run0.getFullDirectory()+"/DISTL_pickle");
		if (pickleFile.exists())
		{
			Log.note("Deleting DISTL_pickle file");
			boolean deleted = pickleFile.delete();
			if (!deleted && pickleFile.exists())
			{
				pickleFile.setWritable(true);
				deleted = pickleFile.delete();
				if (!deleted && pickleFile.exists())
				{
					Util.popupError("Error", "Could not delete DISTL cache file "+pickleFile.getAbsolutePath()+", subsequent raster results may be incorrect");
				}
			}
		}
	}
	
	// if hit is null, just look for the run reference
	private ModelRasterRun findRunThatRefinesHit(int runIndexForHit, ModelRasterHits.Hit hit)
	{
		ModelRasterRun ret = null;

		for (int runIndex=2; runIndex <= autoList.getRunCount(); runIndex++)
		{
			ModelRasterRun r = ((ModelRasterRun)autoList.getRun(runIndex));
			if
			(
				r.getReferenceRun()==runIndexForHit &&
				(null==hit || r.getReferenceHit() == hit.getNumber())
			)
			{
				ret = r;
			}
		}
		
		return ret;
	}
	
	private void printFinishNotes(String msg, boolean isLastRun)
	{
		if (isLastRun)
		{
			Log.note(msg+".  Auto raster complete.");
		}
		else
		{
			for (int secLeft=5;secLeft>0;secLeft--)
			{
				Log.note(msg+", starting next run in "+secLeft+"...");
				Util.sleep(1000);
			}
		}
	}
	
	// returns corrected cell size
	public static Point2D setSearchArea(int runIndex, RasterListType rType, Point2D reqGridSize)
	{
		Point2D cellSize = new Point2D(20,20);
		/*switch (rType)
		{
		case COARSE: cellSize = CellCalculator.getDefaultCellSize(reqGridSize.x, reqGridSize.y, CellSizeMethod.USE_WIDTH); 	 break;
		case FINE:	 cellSize = CellCalculator.getDefaultCellSize(reqGridSize.x, reqGridSize.y, CellSizeMethod.USE_SMALLER);	 break;
		case ORTHO:	 cellSize = CellCalculator.getValidCellSize(reqGridSize.x, reqGridSize.y/10.0, CellSizeMethod.USE_HEIGHT);	 break;
		}*/

		// how many cells fit?
		int cellCountX = (int)Math.ceil((double)reqGridSize.x / (double)cellSize.x);
		int cellCountY = (int)Math.ceil((double)reqGridSize.y / (double)cellSize.y);

		// how big is the grid with the new cell size?
		double cGridW = cellCountX * cellSize.x;
		double cGridH = cellCountY * cellSize.y;

		// create a centered rectangular contour equal to reqGridSize
		ModelContour contour = new ModelContour();
		contour.points.add(new Point2D(-cGridW/2.0,-cGridH/2.0));
		contour.points.add(new Point2D(+cGridW/2.0,-cGridH/2.0));
		contour.points.add(new Point2D(+cGridW/2.0,+cGridH/2.0));
		contour.points.add(new Point2D(-cGridW/2.0,+cGridH/2.0));

		ArrayList<ModelContour> contourList = new ArrayList<>();
		contourList.add(contour);
		
		ArrayList<Boolean> selectedCells = ModelContour.getEnclosedCells
		(
			contourList,
			GridCalculator.RASTER_INITIAL_DIRECTION,
			cellSize.x,
			cellSize.y,
			cellCountX,
			cellCountY,
			-cGridW/2.0,	// corrected left offset
			-cGridH/2.0,	// corrected top offset
			cGridW,
			cGridH
		);
		
		DevRasterRun.s_setContours		(rType, runIndex, contourList);
		DevRasterRun.s_setSelectedCells	(rType, runIndex, selectedCells);
		DevRasterRun.s_setBeamHeight_um	(rType, runIndex, cellSize.x);
		DevRasterRun.s_setBeamWidth_um	(rType, runIndex, cellSize.y);
		DevRasterRun.s_setCellWidth_um	(rType, runIndex, cellSize.x);
		DevRasterRun.s_setCellHeight_um	(rType, runIndex, cellSize.y);
		DevRasterRun.s_setGridHOffsets	(rType, runIndex, -cGridW/2.0, cGridW/2.0);
		DevRasterRun.s_setGridVOffsets	(rType, runIndex, -cGridH/2.0, cGridH/2.0);
		
		return cellSize;
	}
	
	private MoveResult centerSample(Point4D pt)
	{
		// start omega
		MoveCmd omegaMove = DevPmacMotor.get("gonio_omega").startMove(pt.omega);
		
		// xyz move (wait is included)
		MoveResult xyzResult = DevXYZ.syncMove(new Point3D(pt.x,pt.y,pt.z), true);
		
		// wait for omega
		MoveResult omegaResult = omegaMove.waitFor();
		
		if (!xyzResult.isOK()) 		return xyzResult;
		if (!omegaResult.isOK()) 	return omegaResult;
		
		return new MoveResult(ResultType.SUCCESS);
	}
}
