package gov.anl.gmca.raster;

import gov.anl.gmca.collect.CmdCheck;
import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.common.DevMove.MoveCmd;
import gov.anl.gmca.common.ui.NBMessageBox;
import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.io.ExecTask;
import gov.anl.gmca.raster.ModelRasterRun.RunState;
import gov.anl.gmca.raster.ModelRasterState.RasterState;
import gov.anl.gmca.raster.calc.RasterGeometry;
import gov.anl.gmca.raster.cam.WdgRasterCamera;
import gov.anl.gmca.sample.WdgCamera.CameraType;
import gov.anl.gmca.scan.DevEdgeScan;
import gov.anl.gmca.spotfinder.SFData;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.ActionResult;
import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.ListenerList.Listener;
import gov.anl.gmca.util.Util;
import org.eclipse.swt.SWT;

import java.io.File;

// This is how data gets from the MCA to the screen:
//
// 1) Sergey's script interprets MCA data and writes one value per cell to 23x:bi:SCAN:fastscanData
// 2) Whenever that PV is updated, this thread reads it, averages if necessary, then calls DevRasterRunResults.setCellData
// 3) DevRasterRunResults writes data to the fluoValue column of user_data.RasterResults in SQL, and then sets the update PV
// 4) A listener set up in DevRasterRunResults sees the update PV and calls run.grid().getCell(cellNumber).setData(...)
// 5) WdgRasterMap doesn't use a listener, it simply sees the data on the next pass and renders it

public class CmdRasterFluo extends CmdRaster
{
	protected ExecTask task = null;

	ModelRasterGrid grid;
	protected Runnable runAtStart;
	protected int skippedRows = 0; // used when resuming a partially finished raster
	
	protected ListenerList.Listener<ModelRasterState> stateListener;
	
	public CmdRasterFluo(ModelRasterRun _run, Runnable _runAtStart)
	{
		super(_run);
		
		// abort() sends kill
		//WdgEmergencyStop.registerProcess("findFluorescenceFast.pl");

		grid = run.grid();
		runAtStart = _runAtStart;
		
		stateListener = new ListenerList.Listener<ModelRasterState>()
		{
			@Override public void handleEvent(ModelRasterState state, Object data)
			{
				if (null != task &&
					task.isRunning() &&
					RasterState.STOPPING == state.getRasterState())
				{
					Log.note("Sending Ctrl+C to fluorescence raster script...");
					task.asyncStop();
				}
			}
		};
	}

	// handle a row of data
	//
	// Note: cell objects only exist for "selected" cells, those within the polygon outlines
	//
	protected void updateResults(double[] dataRow)
	{
		boolean lastCellRead = false;
		
		// decode the "header" from the PV array
		final int headerSize = 5;
		
		int scanFlag	= (int)dataRow[0];
		int scanRow		= (int)dataRow[1] + skippedRows;
		int rowTotal	= (int)dataRow[2];
		double zCoord	= dataRow[3];
		int nElements	= (int)dataRow[4];
		
		System.out.println("updateResults: scanFlag="+scanFlag+" scanRow="+scanRow+" rowTotal="+rowTotal+" zCoord="+zCoord+" nElements="+nElements);
		
		// if scanFlag=0, this change was done by a BluIce client so ignore it
		if (0 == scanFlag) return;
		
		// if there is no data, just set this row to zeroes
		if (0 == nElements)
		{
			for (int i=0;i<grid.getCellCountX();i++)
			{
				//grid.getCell(scanRow-1,i).setFluoResult(0);
			}
		}
		else
		{
			double[] xCoords = new double[nElements];
			double[] yCoords = new double[nElements];
			
			String origXCoordString="", origYCoordString="";
			String xCoordString="", yCoordString="";
			
			for (int i=0; i<nElements; i++)
			{
				origXCoordString += dataRow[headerSize+i]+" ";
				origYCoordString += dataRow[headerSize+nElements+i]+" ";
			}
			
			// Check the array order and set the boolean to true if the order is not ascending as we want it.
			// The script guarantees the order from one element to the next is correct,
			// but it may be backwards depending on the scanning direction.
			int firstXIndex = headerSize;
			int lastXIndex = headerSize + nElements - 1;
			boolean epicsOrderIsDescending = (dataRow[firstXIndex] > dataRow[lastXIndex]);
			boolean flip = epicsOrderIsDescending;
			System.out.println("Checking index "+firstXIndex+"="+dataRow[firstXIndex]+" vs. "+lastXIndex+"="+dataRow[lastXIndex]+" flip="+flip);

			for (int i=0; i<nElements; i++)
			{
				// convert the single coordinate array of motor coordinates to
				// two arrays of relative coordinates
				int elementIndex = flip ? nElements-1-i : i; 		// the order to read from the arrays
				int xIndex = headerSize + elementIndex; 			// the x elements are first
				int yIndex = headerSize + nElements + elementIndex; // the y elements follow
				xCoords[i] = dataRow[xIndex];
				yCoords[i] = dataRow[yIndex];
			}

			for (int i=0; i<nElements; i++)
			{
				xCoordString += xCoords[i]+" ";
				yCoordString += yCoords[i]+" ";
			}
			
			System.out.println("Original xCoords:"+origXCoordString);
			System.out.println("Original yCoords:"+origYCoordString);
			System.out.println("xCoords:"+xCoordString);
			System.out.println("yCoords:"+yCoordString);
			
			for (int i=0;i<grid.getCellCountX();i++)
			{
				try
				{
    				ModelRasterCell cell = grid.getCell(scanRow-1, i);
    				if (null==cell) { System.out.println("Cell mapping for indices "+(scanRow-1)+","+i+" is null"); continue; }
    				else { System.out.println("Cell mapping for indices "+(scanRow-1)+","+i+" is "+cell.getIndex()); }
    				
    				double[] motorXYZ = cell.getMotorXYZ();
    				if (null==motorXYZ)
    				{
    					Log.warn("No grid center");
    					break;
    				}
					double expectedXCoord = motorXYZ[2];
					double averagedYValue = 0.0;
					
					System.out.println("expectedXCoord="+expectedXCoord);
					
					// following if/else statements assume xCoords are ASCENDING
					if (1 == xCoords.length)
					{
						System.out.println("only one xCoord="+xCoords[0]);
						averagedYValue = yCoords[0];
					}
					else if (xCoords[0] > expectedXCoord)
					{
						System.out.println("First xCoord "+xCoords[0]+" > "+expectedXCoord);
						averagedYValue = yCoords[0];
					}
					else if (xCoords[xCoords.length-1] < expectedXCoord)
					{
						System.out.println("Last xCoord "+xCoords[xCoords.length-1]+" < "+expectedXCoord);
						averagedYValue = yCoords[yCoords.length-1];
					}
					else
					{
    					// if we get here, there are at least two real data points
						// and expectedXCoord is somewhere between the first and last points
    					for (int j=1;j<xCoords.length;j++)
    					{
    						if (xCoords[j] > expectedXCoord)
    						{
    							double leftSideX = xCoords[j-1];
    							double rightSideX = xCoords[j];
    							double difference = rightSideX-leftSideX;
    							double distanceFromLeft = expectedXCoord - leftSideX;
    							double percentRight = distanceFromLeft / difference;
    							double percentLeft = 1.0 - percentRight;
    							
    							averagedYValue = (percentLeft * yCoords[j-1]) + (percentRight * yCoords[j]);
    							
    							System.out.println(expectedXCoord+" is between "+leftSideX+" and "+rightSideX+", "+(100.0*percentLeft)+"% * "+yCoords[j-1]+" + "+(100.0*percentRight)+"% * "+yCoords[j]+" = "+averagedYValue);
    							break;
    						}
    					}
					}
					
    				System.out.println("Setting (1-based) row "+scanRow+" col "+i+" to "+averagedYValue);
    				
    				// scanRow is 1-based
    				// set the data in the cell
    				int cellIndex = cell.getIndex();
    				SFData.SFImageData data = new SFData.SFImageData();
    				data.flValue = averagedYValue;
    				// TODO: fix
    				//DBSpotFinder.getDB(SFType.INTER_RASTER,run.getRunIndex()).setData(cellIndex, data, new AtomicBoolean(false));

    				// note if this is the last cell
    				if (cellIndex == grid.getSelectedCellCount()-1)
    				{
    					lastCellRead = true;
    				}
				}
				catch (ArrayIndexOutOfBoundsException aioobe)
				{
					aioobe.printStackTrace();
				}
			}
		}
		
		int nextFrame = scanRow * grid.getCellCountX();
		run.dev().setNextFrame(nextFrame, true);
		
		// tell the script we read the data
		DevEdgeScan.setData(new double[]{0});
		
		// write the log if the last cell was read
		if (lastCellRead)
		{
			//RasterLog.writeLogFile(run);
		}
	}
	
	// launch the Perl script and wait for completion
	@Override
	public void run()
	{
		Log.note("Starting fluorescence raster");
		
		// these two states need to be returned to inactive/paused/complete by this thread
		run.dev().setRunStatus(RunState.RASTERING, true);
		DevRasterState.get().setRasterState(RasterState.RASTERING);
		
		// fix the states if they get reset
		startStateCheck();
		
		// wait for any submitted run changes to take effect
		runAtStart.run();
		
		// pre-raster checks
		CmdCheck cc = new CmdCheck(CmdCheck.FLAGS_PRERASTER);
		cc.run();
		if (!cc.checksPassed())
		{
		    errorPopupAfterExit("Checks failed", "Checks failed, aborting rastering.");
		    endRaster();
		    return;
		}
		
		ActionResult moveResult = preRasterMoves();
		if (!moveResult.succeeded())
		{
		    endRaster();
		    return;
		}
		
		// set up PVs
		DevEdgeScan.setData(new double[]{0});
		Util.sleep(20);
		Listener<DevEdgeScan> dataMonitor = new Listener<DevEdgeScan>()
		{
			@Override public void handleEvent(DevEdgeScan object, Object data)
			{
				updateResults((double[])data);
			}
		};
		DevEdgeScan.llData.addListener(dataMonitor);

		run.dev().sendNonRasteredCellsToScript();
		skippedRows = 0;//run.grid().getCellList().get(nextFrame).getRow();
		
		ActionResult dirResult = checkDirectory(run);
		if (!dirResult.succeeded())
		{
		    endRaster();
		    return;
		}
		
		String logPrefix = grid.run().getFullDirectory()+"/"+grid.run().getFileRoot();
		
		File logFile = new File(logPrefix+".log");
		int counter = 1;
		
		while (logFile.exists())
		{
			counter++;
			if (counter>10000)
			{
				errorPopupAfterExit("Error creating log file","Error finding an available log file name in "+grid.run.getFullDirectory()+", rastering aborted.");
				endRaster();
				return;
			}
			logFile = new File(logPrefix+"_"+counter+".log");
		}
		
		// move omega to grid value
		MoveCmd omegaMove = DevPmacMotor.get("gonio_omega").startMove(run.grid().getCenter().omega);
		omegaMove.waitFor();
		if (!omegaMove.getResult().isOK())
		{
			endRaster();
			return;
		}
		
		// start with relative grid sides from the run 'center'
		double hMinEdge = grid.getHMin_um();
		double hMaxEdge = grid.getHMax_um();
		double vMinMidpoint = grid.getVMin_um() + skippedRows*grid.getCellHeight_um();
		double vMaxMidpoint = grid.getVMax_um();
		
		// now move them to be relative from the current center
		double[] currentCenterXY 	= WdgRasterCamera.motorMMtoAbsScreen(CameraType.HIGHRES, false, RasterGeometry.getCurrentMotorPositions().toArray(), 1.0, null);
		double[] gridCenterXY 		= WdgRasterCamera.motorMMtoAbsScreen(CameraType.HIGHRES, false, grid.getCenter().toArray(), 1.0, null);
		double[] gridMinusCurrent	= new double[]
        {
			gridCenterXY[0]-currentCenterXY[0],
			gridCenterXY[1]-currentCenterXY[1],
		};
		hMinEdge += gridMinusCurrent[0]*1000.0;
		hMaxEdge += gridMinusCurrent[0]*1000.0;
		vMinMidpoint += gridMinusCurrent[1]*1000.0;
		vMaxMidpoint += gridMinusCurrent[1]*1000.0;
		
		// convert units and round for string passing
		double scaleFactor = 0.001; // um -> mm
		double offsetHMin=Util.round(hMinEdge*scaleFactor,5);
		double offsetHMax=Util.round(hMaxEdge*scaleFactor,5);
		double offsetVMin=Util.round(vMinMidpoint*scaleFactor,5);
		double offsetVMax=Util.round(vMaxMidpoint*scaleFactor,5);

		// call the script
		String cmdLine =
			Util.getPerlPrefix()+"fluoRaster.pl "+
			DBConfig.get("app_beamlineID")+" "+
			"mcaTime="+grid.run().getExposureTime()+" "+
		    "file="+logFile.getAbsolutePath()+" "+
			"offsetH="+offsetHMin+":"+offsetHMax+":"+grid.getCellCountX()+" "+
			"offsetV="+offsetVMin+":"+offsetVMax+":"+grid.getCellCountY()+" "+
			"gnuplot=n "+
			"gnulog=n "+
			"medm=n "+
			"bluice "+
			"mcaROI="+grid.run().getROIMode().scriptString()+" "+
			DBConfig.get("fluoRaster_flags");
			
		task = ExecTask.new_ExecTask(cmdLine, "fluoRaster.pl", true);

		// on pause, stop the script
		ModelRasterState.get().llCurrentRun.addListener(stateListener);
		
		task.start();
		task.waitFor();
		
		ModelRasterState.get().llCurrentRun.removeListener(stateListener);

		Log.note("Fluorescence raster complete.");

		// cleanup
		DevEdgeScan.llData.removeListener(dataMonitor);
		endRaster();
	}
	
	protected void endRaster()
	{
		stopStateCheck();

		postRasterMoves();
		startRestoreBeamsize();
		waitForRestoreBeamsize();
		
		//RasterLog.writeLogFile(run);
		
		// update the run state based on how many frames were completed
		final int cellCount = run.grid().getSelectedCellCount();
		RunState newState = RunState.PAUSED;
		if (run.getNextFrame() == 0) 			newState = RunState.INACTIVE;
		if (run.getNextFrame() == cellCount)	newState = RunState.COMPLETE;
		run.dev().setRunStatus(newState, true);
		
		// update the run list state
		DevRasterState.get().setRasterState(RasterState.INACTIVE);
	}
	
	// TODO: move this up to CmdRaster
	protected void errorPopupAfterExit(String title, String msg)
	{
		final String f_title = title;
		final String f_msg = msg;
		
		new Thread(new Runnable(){public void run()
		{
			//while (run.runList.isRasterThreadAlive()) Util.sleep(50);
			
			NBMessageBox mb2 = new NBMessageBox(GUI.getShell(), SWT.APPLICATION_MODAL | SWT.ICON_INFORMATION);
			mb2.setText(f_title);
			mb2.setMessage(f_msg);
			mb2.open();
			Log.note(f_msg);
			
			run.resetInterRun(true);
		}}, "CmdRasterFluo error popup");
	}
	
	public void requestStop()
	{
		if (null != task && task.isRunning())
		{
			task.asyncStop();
		}
	}
	
	protected void setFlag(int msg) {}
}
