package gov.anl.gmca.raster;

import gov.anl.gmca.collect.ModelCollectRun.Point3D;
import gov.anl.gmca.io.Database;
import gov.anl.gmca.io.Database.DBName;
import gov.anl.gmca.io.Database.Result;
import gov.anl.gmca.io.Database.WhereList;
import gov.anl.gmca.io.Database.WhereOp;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PV_R_RUNS;
import gov.anl.gmca.raster.ModelRasterSiteList.Site;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.Point4D;
import gov.anl.gmca.util.Util;
import gov.aps.jca.dbr.DBR;

import java.util.ArrayList;
import java.util.HashMap;

public class DBRasterSiteList
{
	protected static final String tableName = "RasterSites";
	protected static enum COLS
	{
		SITE_INDEX,
		INTER_RUN,	INTER_CELL,	FINE_RUN,	ORTHO_RUN, HIT_NUMBER,
		MOTOR_X,	MOTOR_Y,	MOTOR_Z,	MOTOR_O,
		SIZE_X,		SIZE_Y,		SIZE_Z;
		
		public String dbString() { return toString().toLowerCase(); }
	}
	
	protected static boolean doesTableExist()
	{
		Database.Result result = Database.exec(DBName.USERDATA, "show tables like '"+tableName+"'");
		boolean exists = (result.numRows()>0);
		return exists;
	}
	
	public static synchronized boolean verifyTableExists()
	{
		if (!doesTableExist())
		{
			String createCmd =
				"CREATE TABLE `"+tableName+"` ("+
				"`site_index` int(11) NOT NULL AUTO_INCREMENT,"+
				"`inter_run` int(11) NOT NULL,"+
				"`inter_cell` int(11) NOT NULL,"+
				"`fine_run` int(11) NOT NULL,"+
				"`ortho_run` int(11) NOT NULL,"+
				"`hit_number` int(11) NOT NULL,"+
    			"`motor_x` double NOT NULL,"+
    			"`motor_y` double NOT NULL,"+
    			"`motor_z` double NOT NULL,"+
    			"`motor_o` double NOT NULL,"+
    			"`size_x` double NOT NULL,"+
    			"`size_y` double NOT NULL,"+
    			"`size_z` double NOT NULL,";
			
			createCmd += "PRIMARY KEY (`site_index`))";
			
			Database.exec(DBName.USERDATA, createCmd);
		}
		
		return doesTableExist();
	}
	
	public void addSite(Site site, boolean sendUpdate)
	{
		verifyTableExists();
		
		int siteNumber = site.index;
		
		if (siteNumber < 1)
		{
			Result result = Database.exec(DBName.USERDATA, "SELECT site_index from "+tableName);
			
			int lastNumber=0;
			for (HashMap<String,String> row : result.getRows())
			{
				int number = 0;
				try
				{
					number = Integer.parseInt(row.get("site_index"));
				}
				catch (NumberFormatException nfe) {}
				if (number > lastNumber) lastNumber = number;
			}
			
			siteNumber = lastNumber+1;
		}
		
		HashMap<String,String> row = new HashMap<String,String>();
		row.put("site_index", 	""+siteNumber);
		row.put("inter_run",	""+site.interRunIndex);
		row.put("inter_cell",	""+site.interCell);
		row.put("fine_run",		""+site.fineRunIndex);
		row.put("ortho_run",	""+site.orthoRunIndex);
		row.put("hit_number",	""+site.hitNumber);
		row.put("motor_x",		Util.sprintf("%.4f", site.motorPos.x));
		row.put("motor_y",		Util.sprintf("%.4f", site.motorPos.y));
		row.put("motor_z",		Util.sprintf("%.4f", site.motorPos.z));
		row.put("motor_o",		Util.sprintf("%.4f", site.motorPos.omega));
		row.put("size_x",		Util.sprintf("%.4f", site.size.x));
		row.put("size_y",		Util.sprintf("%.4f", site.size.y));
		row.put("size_z",		Util.sprintf("%.4f", site.size.z));
		Database.insert(DBName.USERDATA, tableName, row);
		
		if (sendUpdate) sendUpdate();
	}
	
	public void deleteSite(Site site, boolean update)
	{
		// first delete hits corresponding to this site
		DBRasterHits.get().deleteHitsForSite(site, update);
		
		// now delete the site
		WhereList whereList = new WhereList();
		whereList.add("site_index", WhereOp.EQUALS, ""+site.index);
		Database.delete(DBName.USERDATA, tableName, whereList);
		
		if (update) sendUpdate();
	}
	
	public void clearSites(boolean sendUpdate)
	{
		verifyTableExists();
		
		Database.delete(DBName.USERDATA, tableName, null);
		
		if (sendUpdate) sendUpdate();
	}
	
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	protected boolean connected = false;
	protected Epics.EpicsValueMonitor dbUpdatedMonitor = null;
	protected static DBRasterSiteList inst = null;
	
	public static DBRasterSiteList get() { return inst; }
	
	public DBRasterSiteList()
	{
		inst = this;
		verifyConnected();
	}
	
	public void verifyConnected()
	{
		if (connected) return;
		
		updateExistingResults();
		
		dbUpdatedMonitor = new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				int intVal = Epics.dbrToInt(value);
				if (-1 != intVal) handlePVUpdate(intVal);
			}
		};
		Epics.addMonitor(PV_R_RUNS.siteListUpdated, dbUpdatedMonitor);
		
		connected = true;
	}
	
	protected Site sqlRowToSite(HashMap<String,String> row)
	{
		try
		{
			Site site = new Site
			(
				Integer.parseInt(row.get(COLS.SITE_INDEX.dbString())),
				Integer.parseInt(row.get(COLS.INTER_RUN.dbString())),
				Integer.parseInt(row.get(COLS.INTER_CELL.dbString())),
				Integer.parseInt(row.get(COLS.FINE_RUN.dbString())),
				Integer.parseInt(row.get(COLS.ORTHO_RUN.dbString())),
				Integer.parseInt(row.get(COLS.HIT_NUMBER.dbString())),
				new Point4D
				(
					Double.parseDouble(row.get(COLS.MOTOR_X.dbString())),
					Double.parseDouble(row.get(COLS.MOTOR_Y.dbString())),
					Double.parseDouble(row.get(COLS.MOTOR_Z.dbString())),
					Double.parseDouble(row.get(COLS.MOTOR_O.dbString()))
				), new Point3D
				(
					Double.parseDouble(row.get(COLS.SIZE_X.dbString())),
					Double.parseDouble(row.get(COLS.SIZE_Y.dbString())),
					Double.parseDouble(row.get(COLS.SIZE_Z.dbString()))
				)
			);
			
			return site;
		}
		catch (NumberFormatException nfe)
		{
			nfe.printStackTrace();
			return null;
		}
	}
	
	public void updateAllRows()
	{
		updateExistingResults();
	}
	
	public int getSiteCount()
	{
		Database.Result result = Database.exec
		(
			DBName.USERDATA,
			"select * from "+tableName
		);
		return result.numRows();
	}
	
	protected void updateExistingResults()
	{
		verifyTableExists();
		
		final Database.Result result = Database.exec
		(
			DBName.USERDATA,
			"select * from "+tableName+" order by site_index"
		);
		
		ModelRasterSiteList.get().clearRows(false);
		for (HashMap<String,String> row : result.getRows())
		{
			Site site = sqlRowToSite(row);
			if (null!=site) ModelRasterSiteList.get().addSite(site);
		}
		ModelRasterSiteList.get().ll.notifyListeners();
	}
	
	protected void handlePVUpdate(int pvValue)
	{
		verifyTableExists();
		if (-2 == pvValue) updateAllRows();
	}
	
	public void sendUpdate()
	{
		sendRowUpdate(-2);
	}
	
	protected void sendRowUpdate(int updateRow)
	{
		Epics.putInteger(PV_R_RUNS.siteListUpdated, -1);
		Epics.putInteger(PV_R_RUNS.siteListUpdated, updateRow);
	}

	public void handleDeletedRun(ModelRasterRun run)
	{
		// update sites for subsequent runs
		for (int runIndex=run.getRunIndex()+1; runIndex<=run.getRunList().getRunCount(); runIndex++)
		{
			HashMap<String,String> setMap = new HashMap<String,String>();
			WhereList whereList = new WhereList();
			setMap.put("inter_run",""+(runIndex-1));
			whereList.add("inter_run",WhereOp.EQUALS,""+runIndex);
			Database.update(DBName.USERDATA, tableName, setMap, whereList);
		}
		sendUpdate();
	}

	public void updateHitNumber(Site site, int newHitNumber)
	{
		HashMap<String,String> setMap = new HashMap<String,String>();
		WhereList whereList = new WhereList();
		setMap.put("hit_number",""+newHitNumber);
		whereList.add("site_index",WhereOp.EQUALS,""+site.index);
		Database.update(DBName.USERDATA, tableName, setMap, whereList);
	}

	public void deleteSiteForHit(int hitIndex, int interRunIndex, int autoRunIndex)
	{
		ArrayList<Site> siteList = ModelRasterSiteList.get().getSiteListCopy();
		for (Site site : siteList)
		{
			if (site.interRunIndex==interRunIndex && 
				(-1==hitIndex || site.hitNumber==hitIndex))
			{
				Log.note("Deleting site "+(site.index+1));
				DBRasterSiteList.get().deleteSite(site, false);
			}
		}

		DBRasterHits.get().sendUpdate();
		DBRasterSiteList.get().sendUpdate();
	}

	public void deleteSitesForRun(int interRunIndex, int autoRunIndex, boolean update)
	{
		WhereList whereList = new WhereList();
		if (interRunIndex >= 0)
		{
			whereList.add("inter_run", WhereOp.EQUALS, ""+interRunIndex);
			Database.delete(DBName.USERDATA, tableName, whereList);
		}

		if (update) sendUpdate();
	}
}
