package gov.anl.gmca.raster;

import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.hutch.DevPinSensor;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.ExecTask;
import gov.anl.gmca.io.PVLookup.PV_AUTOFIND;
import gov.anl.gmca.sample.DevCameraSettings;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.*;
import gov.anl.gmca.util.ProcessHandling.Signal;

import java.util.ArrayList;

public class DevAutoFind
{
	private static String scriptName = "autoraster_xrec.pl";
	private static String scriptPath = Util.getPerlPrefix()+scriptName;
	private static final int STOP_TIMEOUT_MS = 1000;
	
	private ExecTask task;
	
	public static DevAutoFind inst;
	public static DevAutoFind get() { if (null==inst) new DevAutoFind(); return inst; }
	public DevAutoFind()
	{
		inst = this;
		
		DevPinSensor.llDismount.addListener(new ListenerList.Listener<DevPinSensor>()
		{
			@Override public void handleEvent(DevPinSensor object, Object data)
			{
				Epics.putInteger		(PV_AUTOFIND.complete,			0);
				Epics.putIntegerArray	(PV_AUTOFIND.coarseContour, 	new int[]{0});
				Epics.putIntegerArray	(PV_AUTOFIND.loopZ, 			new int[]{0});
				Epics.putDouble			(PV_AUTOFIND.faceOnOmega, 		0);
				System.out.println("Cleared autoFind PVs due to pin dismount.");
			}
		});
	}
	
	public void setComplete(boolean b) { Epics.putInteger(PV_AUTOFIND.complete, b?1:0); }
	public boolean getComplete() { return Epics.getBoolean(PV_AUTOFIND.complete); }

	// convert screen coordinates from script to micron offsets from grid center
	public ModelContour getCoarseContour()
	{
		int[] contourArray = Epics.getIntegerArray(PV_AUTOFIND.coarseContour);
		int screenX = contourArray[0];
		int screenY = contourArray[1];
		int numPts = contourArray[2];
		ArrayList<Point2D> points = new ArrayList<Point2D>();
		for (int i=0; i<numPts; i++)
		{
			// pixels from top left
			int xi = contourArray[3+i*2];
			int yi = contourArray[3+i*2+1];
			
			// pixels from center
			int xc = xi-(screenX/2);
			int yc = yi-(screenY/2);
			
			// microns from center
			double x=DevCameraSettings.getPx2mm_h_hi()*xc*1000.0;
			double y=DevCameraSettings.getPx2mm_v_hi()*yc*1000.0;
			
			Point2D pt = new Point2D(x,y);
			points.add(pt);
		}
		
		ModelContour contour = new ModelContour();
		contour.points = points;
		return contour;
	}
	
	// The autoFind script calculates loop Z values every 10 pixels on the image
	// This method looks up the correct height for a given X value
	public double getOrthoHeight(int screenX)
	{
		// get the loop Z array from the autoFind script
		int[] loopZArray = Epics.getIntegerArray(PV_AUTOFIND.loopZ);
		
		// convert method argument to pixels
		//int screenX = (int)Math.round(micronsFromLeft / DevCameraSettings.getPx2mm_h_hi());
		
		// calculate array index
		int index = screenX/10;
		if (index<0) index=0;
		if (index>=loopZArray.length) index = loopZArray.length-1;

		// read value and convert back to microns
		int loopZ_px = loopZArray[index];
		double loopZ_um = (DevCameraSettings.getPx2mm_v_hi() * loopZ_px) * 1000.0;

		// apply minimum
		double min_um = DBConfig.getDouble("autoRasterZMin_um");
		if (loopZ_um < min_um)
		{
			System.out.println("Z value "+loopZ_um+" is below minimum "+min_um+", returning the minimum instead.");
			loopZ_um = min_um;
		}
		
		System.out.println("Looked up "+screenX+"px from left (index "+index+"), value="+loopZ_px+"px ("+loopZ_um+" um)");
		
		return loopZ_um;
	}
	
	public ActionResult run(boolean skipCentering)
	{
		if (null != task && task.isRunning()) { Log.error("AutoFind script is already running, action aborted."); return ActionResult.FAILURE; }

		start(skipCentering);
		task.waitFor();
		
		if (0 != task.exitValue)		return new ActionResult("AutoFind script returned "+task.exitValue);
		if (!getComplete())				return new ActionResult("AutoFind script did not set complete flag");
		if (null==getCoarseContour())	return new ActionResult("AutoFind script did not set the coarse contour");
		
		return ActionResult.SUCCESS;
	}
	
	public void start(boolean skipCentering)
	{
		String cmdLine = scriptPath;
		if (skipCentering) cmdLine += " skipcentering";
		
		task = ExecTask.new_ExecTask(cmdLine, scriptName, true);
		task.start();
	}
	
	public boolean isRunning()
	{
		return (null != task && task.isRunning());
	}
	
	public void stop()
	{
		Log.note("Sending stop signal to AutoFind script");
		ProcessHandling.killall(scriptName, Signal.INT);
		long start = System.currentTimeMillis();
		do
		{
			Util.sleep(50);
			long now = System.currentTimeMillis();
			long msSinceSignal = now-start;
			if (msSinceSignal > STOP_TIMEOUT_MS)
			{
				Log.warn("AutoFind script did not respond to SIGINT after "+STOP_TIMEOUT_MS+"ms, sending SIGKILL");
				ProcessHandling.killall(scriptName, Signal.KILL);
			}
		}
		while (isRunning());
	}
}
