package gov.anl.gmca.raster;

import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.util.PerfTimer;
import gov.anl.gmca.util.Point2D;
import gov.anl.gmca.util.Util;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.*;

import java.util.ArrayList;

public class ModelContour
{
	protected static final double PIXELS_PER_UM = 1.0;
	
	public ArrayList<Point2D> points = new ArrayList<Point2D>();
	
	// contour point offsets are stored in microns from the grid center
	// the buffer has square cells regardless of camera zoom level or grid width to height ratio
	// therefore the conversion from contour offsets to buffer pixel is a straightforward scale and translate
	protected static Point micronOffsetToBufferPixel(double micronX, double micronY, double micronGridXStart, double micronGridYStart, double micronGridWidth, double micronGridHeight, int bufferWidth, int bufferHeight)
	{
		double percentX = (micronX - micronGridXStart) / micronGridWidth;
		double percentY = (micronY - micronGridYStart) / micronGridHeight;

		int px = (int)Math.round(percentX * bufferWidth);
		int py = (int)Math.round(percentY * bufferHeight);
		
		if (px<0) 				px=0;
		if (py<0) 				py=0;
		if (px>=bufferWidth) 	px=bufferWidth-1;
		if (py>=bufferHeight) 	py=bufferHeight-1;
		
		return new Point(px,py);
	}
	
	public static ArrayList<Boolean> getEnclosedCells(ArrayList<ModelContour> contours, int orientation, double cellWidth_um, double cellHeight_um, int cellCountX, int cellCountY, double correctedLeftOffset_um, double correctedTopOffset_um, double correctedWidth_um, double correctedHeight_um)
	{
		PerfTimer pt = new PerfTimer("polygon check: ");
		
		ArrayList<Boolean> selectedCells = new ArrayList<Boolean>();

		// create the buffer to draw on
		int pixelsPerCellX = (int)(PIXELS_PER_UM * cellWidth_um);
		int pixelsPerCellY = (int)(PIXELS_PER_UM * cellHeight_um);
		int imgWidth = pixelsPerCellX * cellCountX;
		int imgHeight = pixelsPerCellY * cellCountY;

		if (0 == imgWidth || 0 == imgHeight) return selectedCells;
		
		Image buf=null;
		GC bufGC=null;
		try
		{
			buf = new Image(GUI.getDisplay(), imgWidth, imgHeight);
			bufGC = new GC(buf);
			bufGC.setBackground(GUI.COLOR_BLACK);
			bufGC.fillRectangle(new Rectangle(0,0,imgWidth,imgHeight));
			
			// fill every other cell for reference
			boolean drawCell = true;
			bufGC.setBackground(GUI.COLOR_BLUE);
			for (int cellRow = 0; cellRow < cellCountY; cellRow++)
			{
				for (int cellCol = cellCountX-1; cellCol >= 0; cellCol--)
				{
					if (!drawCell) { drawCell = true; continue; }
					drawCell=false;
					
					int cellX = cellCol * pixelsPerCellX;
					int cellY = cellRow * pixelsPerCellY;
					bufGC.fillRectangle(cellX,cellY,pixelsPerCellX,pixelsPerCellY);
				}
			}
			
			// draw the contours
			bufGC.setBackground(GUI.COLOR_WHITE);
			for (ModelContour contour : contours)
			{
				int[] pointArray = new int[contour.points.size()*2];
				
				int paIndex = 0;
				for (Point2D point : contour.points)
				{
					Point bufferPixel = micronOffsetToBufferPixel(point.x, point.y, correctedLeftOffset_um, correctedTopOffset_um, correctedWidth_um, correctedHeight_um, imgWidth, imgHeight);
					pointArray[paIndex++] = bufferPixel.x;
					pointArray[paIndex++] = bufferPixel.y;
				}
				
				bufGC.fillPolygon(pointArray);
			}
			
			Util.saveImage(buf, System.getProperty("user.home")+"/bluice_temp/grid.png");
	        
			// test each cell to see if the polygon overlaps
			ImageData id = buf.getImageData();
			
			switch (orientation)
			{
			case SWT.VERTICAL: 		selectCellsInColumns(selectedCells,id,cellCountX,cellCountY,pixelsPerCellX,pixelsPerCellY); 	break;
			case SWT.HORIZONTAL: 	selectCellsInRows(selectedCells,id,cellCountX,cellCountY,pixelsPerCellX,pixelsPerCellY);		break;
			default:				throw new IllegalArgumentException("Invalid orientation: "+orientation);
			}
			
			pt.printElapsedMs();
		}
		catch (Exception e) { e.printStackTrace(); }
		finally
		{
			if (null!=buf && !buf.isDisposed()) buf.dispose();
			if (null!=bufGC && !bufGC.isDisposed()) bufGC.dispose();
		}
		return selectedCells;
	}
	
	private static void selectCellsInColumns(ArrayList<Boolean> selectedCells, ImageData id, int cellCountX, int cellCountY, int pixelsPerCellX, int pixelsPerCellY)
	{
		for (int cellCol = cellCountX-1; cellCol >= 0; cellCol--)
		{
			for (int cellRow = 0; cellRow < cellCountY; cellRow++)
			{
				selectCell(selectedCells,id,cellCountX,cellCountY,pixelsPerCellX,pixelsPerCellY, cellRow, cellCol);
			}
		}
	}

	private static void selectCellsInRows(ArrayList<Boolean> selectedCells, ImageData id, int cellCountX, int cellCountY, int pixelsPerCellX, int pixelsPerCellY)
	{
		for (int cellRow = 0; cellRow < cellCountY; cellRow++)
		{
			boolean isEvenRow = (cellRow+1)%2==0; // cellRow is 0-indexed, "even" rows are for counting from 1
			boolean reverseRow = DBConfig.getBoolean("raster_serpentine") && isEvenRow;
			if (!reverseRow) // right-to-left
			{
				for (int cellCol = cellCountX-1; cellCol >= 0; cellCol--)
				{
					selectCell(selectedCells,id,cellCountX,cellCountY,pixelsPerCellX,pixelsPerCellY, cellRow, cellCol);
				}
			}
			else // left-to-right
			{
				for (int cellCol = 0; cellCol < cellCountX; cellCol++)
				{
					selectCell(selectedCells,id,cellCountX,cellCountY,pixelsPerCellX,pixelsPerCellY, cellRow, cellCol);
				}
			}
		}
	}
	
	private static void selectCell(ArrayList<Boolean> selectedCells, ImageData id, int cellCountX, int cellCountY, int pixelsPerCellX, int pixelsPerCellY, int cellRow, int cellCol)
	{
		boolean cellCovered = false;
		
		int cellX = cellCol * pixelsPerCellX;
		int cellY = cellRow * pixelsPerCellY;
		
		for (int x = 0; x < pixelsPerCellX; x++)
		{
			for (int y=0; y < pixelsPerCellY; y++)
			{
				int pixel = id.getPixel(cellX+x, cellY+y);
				if (0!=pixel && 0x0000FF!=pixel) cellCovered = true;
			}
		}

		selectedCells.add(cellCovered);
	}
}