package gov.anl.gmca.raster;

import gov.anl.gmca.raster.ModelRasterState.RasterListType;
import gov.anl.gmca.raster.calc.CellCalculator;
import gov.anl.gmca.raster.calc.FramePartitioner.FramePartition;
import gov.anl.gmca.raster.calc.GridCalculator;
import gov.anl.gmca.raster.calc.HitFinder;
import gov.anl.gmca.raster.ui.TabRaster;
import gov.anl.gmca.sample.ModelBeamsize;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.*;

import java.util.ArrayList;
import java.util.concurrent.atomic.AtomicInteger;

// Everything is accessed here - grid info, dimensions, cell dimensions, state, listeners.
// But many helpers are used by this class, notably GridCalculator and CellCalculator.
//
public class ModelRasterGrid
{
	public static final double DEFAULT_GRID_ROWS	= 4;
	public static final double DEFAULT_GRID_COLUMNS	= 4;
	public static final double MIN_BEAM_HEIGHT_UM	= 25;
	public static final double CELLWIDTHMIN_UM 		= 2.0;
	public static final double CELLWIDTHMAX_UM  	= 125.0;
	public static final double CELLHEIGHTMIN_UM 	= 2.0; 
	public static final double CELLHEIGHTMAX_UM 	= 35.0;
	public static final int    CELLCOUNTMAX  		= 10000;

	public static enum CellSizeMethod { USE_WIDTH, USE_SMALLER, USE_HEIGHT };
	
	public ListenerList<ModelRasterGrid> llGridChanged;
	public ListenerList<ModelRasterGrid> llCellChanged;

	private	GridCalculator 	gridCalc 	= new GridCalculator(this);
	private	CellCalculator 	cellCalc	= new CellCalculator(this);
	private HitFinder		hitFinder	= new HitFinder(this, null, true);
	
	public	GridCalculator	getGridCalc()	{ return gridCalc; }
	public	CellCalculator	getCellCalc()	{ return cellCalc; }
	public	HitFinder		getHitFinder()	{ return hitFinder; }
	
	////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////// grid config //////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////
	
	protected 	ArrayList<ModelContour> 	contours 								= new ArrayList<ModelContour>();
	public 		ArrayList<ModelContour> 	getContours() 							{ return contours; }
	public 		void 						setContours(ArrayList<ModelContour> c) 	{ contours = c; }
	
	protected double leftOffset_um=Double.NaN, rightOffset_um=Double.NaN, topOffset_um=Double.NaN, bottomOffset_um=Double.NaN;
	public void setHOffsets_um(double left, double right)
	{
		leftOffset_um = left;
		rightOffset_um = right;
	}
	public void setVOffsets_um(double top, double bottom)
	{
		topOffset_um = top;
		bottomOffset_um = bottom;
	}
	public double 	getLeftOffset_um() 		{ return leftOffset_um; 	}
	public double 	getRightOffset_um() 	{ return rightOffset_um;	}
	public double 	getTopOffset_um() 		{ return topOffset_um;		}
	public double 	getBottomOffset_um() 	{ return bottomOffset_um;	}
	public double 	getMinGridWidth_um()	{ return rightOffset_um-leftOffset_um; }
	public double 	getMinGridHeight_um()	{ return bottomOffset_um-topOffset_um; }
	public void 	setLeftOffset_um(double d) 		{ leftOffset_um = d; 	}
	public void 	setRightOffset_um(double d) 	{ rightOffset_um = d;	}
	public void 	setTopOffset_um(double d) 		{ topOffset_um = d;		}
	public void 	setBottomOffset_um(double d) 	{ bottomOffset_um = d;	}
	
	protected double cellWidth_um			= Double.NaN;
	protected double cellHeight_um			= Double.NaN;
	public double getCellWidth_um() 		{ return cellWidth_um; 									}
	public double getCellHeight_um() 		{ return cellHeight_um; 								}
	public void setCellWidth_um(double d) 	{ cellWidth_um = d;										}
	public void setCellHeight_um(double d) 	{ cellHeight_um = d;									}

	protected Point4D center 				= null;
	public void setCenter(Point4D _center) 	{ center = _center; }
	public Point4D getCenter() 				{ return center; }
	
	protected String errorMessage 			= null;
	public String getErrorMessage() 		{ return errorMessage; }
	public void setErrorMessage(String s)	{ errorMessage=s; }
	
	protected 	AtomicInteger 				numCollectedCells 						= new AtomicInteger(0);
	public 		void 						setNumCollectedCells(int newCount) 		{ numCollectedCells.set(newCount); }
	public 		boolean 					allCellsCollected() 					{ return (selectedCellCount==numCollectedCells.get()); }
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////// calculated variables ////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////
	
	protected double correctedLeftOffset_um=0, correctedRightOffset_um=0, correctedTopOffset_um=0, correctedBottomOffset_um=0;
	
	public double getCorrectedLeftOffset_um() 	{ return correctedLeftOffset_um; 	}
	public double getCorrectedRightOffset_um() 	{ return correctedRightOffset_um; 	}
	public double getCorrectedTopOffset_um() 	{ return correctedTopOffset_um; 	}
	public double getCorrectedBottomOffset_um() { return correctedBottomOffset_um; 	}
	
	public void setCorrectedLeftOffset_um(double d) 	{ correctedLeftOffset_um = d; 	}
	public void setCorrectedRightOffset_um(double d) 	{ correctedRightOffset_um = d; 	}
	public void setCorrectedTopOffset_um(double d) 		{ correctedTopOffset_um = d; 	}
	public void setCorrectedBottomOffset_um(double d) 	{ correctedBottomOffset_um = d; }
	
	// relative grid limits
	public double getHMin_um()					{ return correctedLeftOffset_um; }
	public double getHMid_um()					{ return correctedLeftOffset_um + getCorrectedWidth_um()/2.0; }
	public double getHMax_um()					{ return correctedRightOffset_um; }
	public double getVMin_um()					{ return correctedTopOffset_um; }
	public double getVMid_um()					{ return correctedTopOffset_um + getCorrectedHeight_um()/2.0; }
	public double getVMax_um()					{ return correctedBottomOffset_um; }
	
	public double getCorrectedWidth_um()		{ return correctedRightOffset_um - correctedLeftOffset_um; }
	public double getCorrectedHeight_um()		{ return correctedBottomOffset_um - correctedTopOffset_um; }
	
	protected int cellCountX=0, cellCountY=0, totalCellCount=0, selectedCellCount=0;
	public int getCellCountX() 					{ return cellCountX; 		}
	public int getCellCountY() 					{ return cellCountY; 		}
	public int getTotalCellCount() 				{ return totalCellCount; 	}
	public int getSelectedCellCount() 			{ return selectedCellCount; }
	public void setCellCountX(int i) 			{ cellCountX=i; 		}
	public void setCellCountY(int i) 			{ cellCountY=i; 		}
	public void setTotalCellCount(int i) 		{ totalCellCount=i; 	}
	public void setSelectedCellCount(int i)		{ selectedCellCount=i;	}

	///////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////// accessors //////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////
	
	public ModelRasterCell getCellCopy(int cellIndex)
	{
		return gridCalc.getCellCopy(cellIndex);
	}

	public ModelRasterCell getCellOriginal(int cellIndex)
	{
		return gridCalc.getCellOriginal(cellIndex);
	}

	public void setCell(int cellIndex, ModelRasterCell updatedCell)
	{
		gridCalc.setCell(cellIndex, updatedCell);
		
		Pair<Integer,Integer> cellPair = new Pair<Integer,Integer>(updatedCell.getRow(),updatedCell.getCol());
		gridCalc.getCellMap().put(cellPair, updatedCell);
	}

	public ArrayList<FramePartition> getFramePartitions()
	{
		ArrayList<FramePartition> ret = null;

		gridCalc.getLock().acquireUninterruptibly();
		try 				{ ret = new ArrayList<FramePartition>(gridCalc.getPartitionList()); }
		catch (Throwable t)	{ t.printStackTrace();									}
		finally				{ gridCalc.getLock().release();									}
		
		return ret;
	}
	
	// this method uses 'cellMap' which should only contain selected cells
	public ModelRasterCell getCell(int row, int col)
	{
		ModelRasterCell ret;
		ret = gridCalc.getCellMap().get(new Pair<Integer,Integer>(row,col));
		
		if (null==ret) return null;
		if (ret.getRow()==row && ret.getCol()==col) return ret;
		return null;
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////// change methods ///////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////////
	
	public void changeCellWidth_um(double cellWidth_um)
	{
		if (!gridCalc.isCellWidthOK(cellWidth_um))
		{
			// determine how much to increase attenuation by
			double minExpTime_sec = gridCalc.getMinExposureTime_sec(cellWidth_um);
			double currentAtten = run.getAttenuation();
			double timeIncreaseFactor = minExpTime_sec / run.getExposureTime();
			double compensatedAttenuation_factors = currentAtten * timeIncreaseFactor;
			
			// OK to modify exposure time and attenuation to compensate for cell width?
			String msg = Util.sprintf
			(
				"The cell width of %.0f\u00B5m requires a horizontal speed greater than the\n"+
				"limit.  Would you like to increase the exposure time to %.3f seconds and the \n"+
				"attenuation to %.3f to compensate?",
				cellWidth_um,
				minExpTime_sec,
				compensatedAttenuation_factors
			);
			boolean okToCompensate = Util.popupOkCancel("Requested speed not achievable",msg);

			// if not, return and print note
			if (!okToCompensate)
			{
				Log.note("Field correction declined, returning cell width to "+Util.printDouble(getCellWidth_um(),3)+"\u00B5m");
				run.dev().setCellWidth_um(getCellWidth_um(), true);
				return;
			}
			
			run.dev().setExposureTime(minExpTime_sec, false);
			run.dev().setAttenuation_factors(compensatedAttenuation_factors, false);
		}
		
		run.dev().setCellWidth_um(cellWidth_um, false);
		if (!Util.beamSizeCalculated()) setBeamSizeFromCellSize();
		gridCalc.fitGridToContour(false, true); // sends update
	}
	
	// Note that validation should be done before calling this method
	public void changeCellHeight_um(double cellHeight)
	{
		run.dev().setCellHeight_um(cellHeight, false);
		
		if (!Util.beamSizeCalculated()) setBeamSizeFromCellSize();
		
		gridCalc.fitGridToContour(false, true); // sends update
	}
	
	protected void setBeamSizeFromCellSize()
	{
		double cellWidth = getCellWidth_um();
		double cellHeight = getCellHeight_um();
		double minCellDim = Math.min(cellWidth,cellHeight);

		double beamWidth=0,beamHeight=0;
		
		if (minCellDim < 7.5)			{ beamWidth = beamHeight = 5.0; }
		else if (minCellDim < 15.0) 	{ beamWidth = beamHeight = 10.0; }
		else if (minCellDim < 25.0)		{ beamWidth = beamHeight = 20.0; }
		else
		{
			beamWidth = cellWidth;
			beamHeight = cellHeight;
		}
		
		Point2D beamMax = ModelBeamsize.getActualBeamsize_um();
		if (beamWidth > beamMax.x) 	beamWidth = beamMax.x;
		if (beamHeight > beamMax.y) beamHeight = beamMax.y;
		
		run.dev().setBeamWidth_um(beamWidth, false);
		run.dev().setBeamHeight_um(beamHeight, false);
	}

	// returns the corrected requested dimension in ret.x and the other dimension in ret.y
	// therefore ret.x may be the height
	public static Point2D calcBeamSize(double requestedDim, double otherDim, double otherLimit)
	{
		Point2D ret = new Point2D(0,0);

		if (requestedDim < 7.5)			ret.x = 5.0;
		else if (requestedDim < 15.0) 	ret.x = 10.0;
		else if (requestedDim < 25.0)	ret.x = 20.0;
		else							ret.x = requestedDim;
		
		if (ret.x < 25.0)
		{
			ret.y = ret.x;
		}
		else if (ret.x >= 25.0 && otherDim < 25.0)
		{
			ret.y = ret.x;
			if (ret.y > otherLimit) ret.y = otherLimit;
		}
		else
		{
			ret.y = otherDim;
		}
		
		return ret;
	}
	
	public void changeBeamWidth_um(double beamWidth, boolean update)
	{
		Point2D beamSize = calcBeamSize(beamWidth, run.getBeamHeight_um(), ModelBeamsize.getActualBeamsize_um().y);
		run.dev().setBeamWidth_um(beamSize.x, false);
		run.dev().setBeamHeight_um(beamSize.y, false);
		if (update) run.dev().sendUpdate();
	}
	
	public void changeBeamHeight_um(double beamHeight, boolean update)
	{
		Point2D beamSize = calcBeamSize(beamHeight, run.getBeamWidth_um(), ModelBeamsize.getActualBeamsize_um().x);
		run.dev().setBeamHeight_um(beamSize.x, false);
		run.dev().setBeamWidth_um(beamSize.y, false);
		if (update) run.dev().sendUpdate();
	}
	
	//////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////// contour methods ///////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////////

	public void addContour(ModelContour newContour)
	{
		if (!deleteHitsAndSitesForRun(true)) return;

		// add the contour to the list
		ArrayList<ModelContour> newContours = new ArrayList<ModelContour>(contours);
		newContours.add(newContour);
		run.dev().setContours(newContours, false);
		
		gridCalc.fitGridToContour(true, true); // this sends the run update
	}
	

	private boolean deleteHitsAndSitesForRun(boolean prompt)
	{
		if (ModelRasterHits.get().getCopyOfHits(run()).size()>0)
		{
			if (prompt)
			{
				boolean okToDelete = Util.popupOkCancel("OK to delete hits?", "Modifying contours will delete hits for this run, OK to proceed?");
				if (!okToDelete)
				{
					Log.note("Deletion declined.");
					return false;
				}
			}
			int interRunIndex = run.getRunList().getListType()==RasterListType.INTER?run.getRunIndex():-1;
			int autoRunIndex = run.getRunList().getListType()==RasterListType.AUTO?run.getRunIndex():-1;
			DBRasterSiteList.get().deleteSitesForRun(interRunIndex, autoRunIndex, true);
			DBRasterHits.get().deleteHitsForRun(run, true);
		}
		return true;
	}
		
	public void removeLastContour()
	{
		if (0 == getContours().size())
		{
			Log.note("There are no polygons to delete");
			return;
		}

		if (!deleteHitsAndSitesForRun(true)) return;
		
		// send new contour list with one less contour to the device
		ArrayList<ModelContour> newContours = new ArrayList<ModelContour>(getContours());
		newContours.remove(newContours.size()-1);
		run().dev().setContours(newContours, false);

		// clear center if all contours are deleted
		if (0 == newContours.size())
		{
			run().dev().setCenter(null, false);
			run().dev().setSelectedCells(new ArrayList<Boolean>(), false);
		}

		// fit grid to new contours
		gridCalc.fitGridToContour(true, true);
		
		// TODO: only deselect if the selected cell was in the deleted contour
		TabRaster.get().cam.model.vcViewImages.setSelectedCell(null);
	}
	
	public void clearAllContours(boolean promptForHits, boolean clearHitsAndSites)
	{
		if (clearHitsAndSites)
		{
			if (!deleteHitsAndSitesForRun(promptForHits)) return;
		}
		
		ArrayList<ModelContour> newContours = new ArrayList<ModelContour>();
		run().dev().setContours(newContours, false);
		run().dev().setCenter(null, false);
		run().dev().setSelectedCells(new ArrayList<Boolean>(), false);
		
		gridCalc.fitGridToContour(true, true);
	}

	////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////// constructor /////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////////
	
	protected ModelRasterRun run = null;
	public ModelRasterRun run() { return run; }
	
	public ModelRasterGrid(ModelRasterRun _run)
	{
		run = _run;
		llGridChanged = new ListenerList<ModelRasterGrid>(this);
		llCellChanged = new ListenerList<ModelRasterGrid>(this);
	}
	
	protected void stopMonitors() {}
}
