package gov.anl.gmca.raster;

import gov.anl.gmca.collect.DevFPE;
import gov.anl.gmca.collect.ModelCollectRun.Point3D;
import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.common.DevDataDir;
import gov.anl.gmca.common.DevMove;
import gov.anl.gmca.common.DevMove.MoveCmd;
import gov.anl.gmca.common.DevMove.MoveResult;
import gov.anl.gmca.common.ModelRunList;
import gov.anl.gmca.gui.options.DevUserConfig;
import gov.anl.gmca.hutch.*;
import gov.anl.gmca.io.Disk;
import gov.anl.gmca.raster.ModelRasterSiteList.Site;
import gov.anl.gmca.raster.ModelRasterState.RasterListType;
import gov.anl.gmca.raster.ModelRasterState.RasterState;
import gov.anl.gmca.raster.calc.GridCalculator;
import gov.anl.gmca.raster.sonicc.CmdBinSONICC;
import gov.anl.gmca.raster.ui.TabRaster;
import gov.anl.gmca.raster.util.ClearSampleState;
import gov.anl.gmca.sample.DBBeamSize;
import gov.anl.gmca.sample.DevCollimatorCtrl;
import gov.anl.gmca.sample.ModelBeamsize;
import gov.anl.gmca.server.services.ResPredictor;
import gov.anl.gmca.spotfinder.DBImageStats;
import gov.anl.gmca.spotfinder.SFData.SFType;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.users.ModelUserTable.ClientListener;
import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.ListenerList.Listener;
import gov.anl.gmca.util.Point4D;
import gov.anl.gmca.util.Util;
import org.eclipse.swt.SWT;

import java.io.IOException;
import java.util.ArrayList;

/**
 * There should be one of these run models per raster run tab.  Each of these
 * contains one grid model, and each of those in turn contains a list of
 * cell models.
 * 
 * This class, which handles both diffraction and fluorescence rastering runs,
 * is responsible for updating listeners, handling the raster thread, and
 * reporting the state of the run.
 */
public class ModelRasterRun implements ModelRunList.ModelRun
{
	public ListenerList<ModelRasterRun> llRunChanged;
	public ListenerList<ModelRasterRun> llRunReset;
	
	//////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////// Change methods ///////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////

	// TODO: make handleChange() synchronized to prevent multiple overlapping recalc() calls
	protected Thread changeThread = null;
	private Object handleChangeLock = new Object();
	protected void handleChange()
	{
		synchronized(handleChangeLock)
		{
			if (!validateAndFix()) return;
			
			boolean gridChange = didGridVariablesChange();
			//boolean parameterChange = didParameterVariablesChange();
			// detect modifications during collection
			//if (ModelRasterState.isRasterRunning() && parameterChange) 	System.out.println("WARNING: Parameter changed while collect thread is running!");
			//if (ModelRasterState.isRasterRunning() && gridChange)  		System.out.println("WARNING: Grid changed while collect thread is running!");
	
			boolean frameChange = didFrameChange();
			if (frameChange)
			{
				grid().getGridCalc().repartition();
			}
			
			// delay sequence recalc during collection
			if (gridChange)
			{
	    		if (!runList.isRasterThreadAlive())
	    		{
	    			grid().getGridCalc().generateCellList();
	    			grid().llGridChanged.notifyListeners();
	    		}
	    		else
	    		{
	    			if (null == changeThread || !changeThread.isAlive())
	    			{
	        			changeThread = new Thread(new Runnable(){public void run()
	        			{
	            			while (ModelRasterState.isRasterRunning()) Util.sleep(10);
	        				
	            			grid().getGridCalc().generateCellList();
	            			grid().llGridChanged.notifyListeners();
	        			}}, "Raster grid change");
	        			changeThread.start();
	    			}
	    		}
			}
			
			llRunChanged.notifyListeners();
			runList.notifyCurrentRunListeners();
		}
	}
	
	// grid variables
	// note: the only variables that aren't grid or parameter variables are runState and nextFrame
	protected int lastRunLabel, lastStartFrame, lastNextFrame;
	protected Boolean lastShutterless = null;
	protected RunType lastRunType;
	protected String lastPrefix = null;
	protected double lastLeftOffset, lastRightOffset, lastTopOffset, lastBottomOffset, lastCellWidth, lastCellHeight;
	protected ArrayList<Boolean> lastSelectedCells = new ArrayList<Boolean>();
	
	private boolean didFrameChange()
	{
		boolean change = false;
		int nextFrame = getNextFrame();
		if (nextFrame != lastNextFrame) change = true;
		lastNextFrame = nextFrame;
		return change;
	}
	
	// TODO: add detection for offset changes and selectedCells changes
	protected boolean didGridVariablesChange()
	{
		String prefix = getFileRoot();
		int runLabel = getRunLabel();
		int startFrame = getStartFrame();
		RunType runType = getRunType();
		double leftOffset = grid().getLeftOffset_um();
		double rightOffset = grid().getRightOffset_um();
		double topOffset = grid().getTopOffset_um();
		double bottomOffset = grid().getBottomOffset_um();
		double cellWidth = grid().getCellWidth_um();
		double cellHeight = grid().getCellHeight_um();
		Boolean shutterless = isShutterless();
		
		boolean changed = false;
		String whatChanged = "";

		if (null==lastPrefix || !prefix.equals(lastPrefix))		{ changed = true; whatChanged += "fileroot ";	}
		if (runLabel		!= lastRunLabel)					{ changed = true; whatChanged += "runLabel ";	}
		if (startFrame		!= lastStartFrame)					{ changed = true; whatChanged += "startFrame ";	}
		if (runType 		!= lastRunType)						{ changed = true; whatChanged += "runType ";	}
		if (shutterless		!= lastShutterless)					{ changed = true; whatChanged += "shutterless "; }
		if (!Util.nearlyEquals(leftOffset, lastLeftOffset, 4))	{ changed = true; whatChanged += "leftOffset ";	}
		if (!Util.nearlyEquals(rightOffset, lastRightOffset, 4)){ changed = true; whatChanged += "rightOffset ";	}
		if (!Util.nearlyEquals(topOffset, lastTopOffset, 4))	{ changed = true; whatChanged += "topOffset ";	}
		if (!Util.nearlyEquals(bottomOffset, lastBottomOffset, 4)){ changed = true; whatChanged += "bottomOffset ";	}
		if (!Util.nearlyEquals(cellWidth, lastCellWidth, 4))	{ changed = true; whatChanged += "cellWidth ";	}
		if (!Util.nearlyEquals(cellHeight, lastCellHeight, 4))	{ changed = true; whatChanged += "cellHeight ";	}
		
		if (lastSelectedCells.size() != dev().getSelectedCells().size()) { changed = true; whatChanged += "selectedCells ";	}
		else for (int i=0; i<lastSelectedCells.size(); i++)
		{
			if (lastSelectedCells.get(i).booleanValue() != dev().getSelectedCells().get(i).booleanValue())
			{
				changed = true; whatChanged += "selectedCells "; break;
			}
		}
		
		if (!changed) return false;

		lastPrefix		= prefix;
		lastRunType		= runType;
		lastRunLabel 	= runLabel;
		lastStartFrame 	= startFrame;
		lastShutterless = shutterless;
		lastLeftOffset 	= leftOffset;
		lastRightOffset = rightOffset;
		lastTopOffset	= topOffset;
		lastBottomOffset= bottomOffset;
		lastCellWidth 	= cellWidth;
		lastCellHeight 	= cellHeight;
		lastSelectedCells=new ArrayList<Boolean>(dev().getSelectedCells());
		
		return true;
	}

	// parameter variables
	// note: the only variables that aren't grid or parameter variables are runState and nextFrame
	protected String lastDirectory;
	protected boolean lastAutoFind, lastLoopCentered;
	protected ROIMode lastRoiMode;
	protected double lastExposureTime, lastDelta, lastDetectorDist, lastEnergy, lastAtten;
	
	protected boolean didParameterVariablesChange()
	{
		String directory = getDirectorySuffix();
		boolean autoFind = getAutoFind();
		boolean loopCentered = getLoopCentered();
		ROIMode roiMode = getROIMode();
		double exposureTime = getExposureTime();
		double delta = getDelta();
		double detectorDist = getDetectorDist();
		double energy = getEnergy();
		double atten = getAttenuation();
		
		boolean changed = false;
		String whatChanged = "";

		if (directory			!= 			lastDirectory)			{ changed = true; whatChanged += "directory ";			}
		if (autoFind			!= 			lastAutoFind)			{ changed = true; whatChanged += "autoFind ";			}
		if (loopCentered		!= 			lastLoopCentered)		{ changed = true; whatChanged += "loopCentered ";		}
		if (roiMode				!= 			lastRoiMode)			{ changed = true; whatChanged += "roiMode ";			}
		if (!Util.nearlyEquals(exposureTime,lastExposureTime,4))	{ changed = true; whatChanged += "exposureTime ";		}
		if (!Util.nearlyEquals(delta,		lastDelta,4))			{ changed = true; whatChanged += "delta ";				}
		if (!Util.nearlyEquals(detectorDist,lastDetectorDist,4))	{ changed = true; whatChanged += "detectorDist ";		}
		if (!Util.nearlyEquals(energy,		lastEnergy,4))			{ changed = true; whatChanged += "energy ";				}
		if (!Util.nearlyEquals(atten,		lastAtten,4))			{ changed = true; whatChanged += "atten ";				}

		//System.out.println("didParameterVariablesChange() returning "+changed+" changed="+whatChanged);
		
		if (!changed) return false;

		lastDirectory 			= directory;
		lastAutoFind			= autoFind;
		lastLoopCentered		= loopCentered;
		lastRoiMode 			= roiMode;
		lastExposureTime 		= exposureTime;
		lastDelta 				= delta;
		lastDetectorDist 		= detectorDist;
		lastEnergy 				= energy;
		lastAtten 				= atten;
		
		return true;
	}
	
	//////////////////////////////////////////////// Variables /////////////////////////////////////////////
	
	// grid setup and results
	protected ModelRasterGrid _grid = null;
	public ModelRasterGrid grid() { return _grid; }
	
	// collection parameters
	public enum RunState
	{
		INACTIVE	("Inactive", 	"Idle"),
		RASTERING	("Rastering",	"Busy"),
		PAUSED		("Paused",		"Pause"),
		COMPLETE	("Complete",	"Done");

		protected String longString, shortString;
		@Override public String toString()	{ return longString; }
		public String toShortString()		{ return shortString; }
		
		RunState(String longString, String shortString)
		{
			this.longString=longString;
			this.shortString=shortString;
		}

	}
	protected	RunState	runState 						= RunState.INACTIVE;
	public		RunState	getRunState()					{ return runState;												}
	public 		void 		setRunState(RunState rs)		{ if (runState!=rs) { runState = rs; }		}
	public 		boolean 	isStarted()   					{ return (RunState.INACTIVE != runState);						}
	public 		boolean 	isComplete()  					{ return RunState.COMPLETE == runState;							}
	public 		boolean 	isValid()     					{ return (null == grid().errorMessage);							}
	
	protected	RunType		runType 						= RunType.DIFFRACTION;
	public		enum		RunType							{ DIFFRACTION, FLUO, SONICC};
	public		RunType		getRunType()					{ return runType;												}
	public 		void 		setRunType(RunType rt)			{ if(runType!=rt) { runType = rt; }			}
	
	protected	int			nextFrame 						= 0;
	public		int			getNextFrame()					{ return nextFrame;												}
	public 		void 		setNextFrame(int nf)			{ if (nextFrame!=nf) { nextFrame = nf; }		}
	
	protected	int			runLabel 						= -1;
	public		int			getRunLabel()					{ return runLabel;												}
	public 		void 		setRunLabel(int rl)				{ if (runLabel!=rl) { runLabel = rl; }		}
	
	protected	String		fileRoot 						= "";
	public		String		getFileRoot()					{ return fileRoot;												}
	public 		void 		setFileRoot(String fr)			{ if (fileRoot!=fr) { fileRoot = fr; }		}
	
	protected	String		directorySuffix 				= "";
	public		String		getDirectorySuffix()			{ return directorySuffix;												}
	public 		void 		setDirectorySuffix(String dir)	{ if (directorySuffix!=dir) { directorySuffix = dir;	}	}
	public		String		getFullDirectory()
	{
		java.io.File f = new java.io.File(DevDataDir.addToPrimary(directorySuffix));
		try
		{
			return f.getCanonicalPath().toString();
		}
		catch (IOException ioe)
		{
			ioe.printStackTrace();
			return DevDataDir.addToPrimary(directorySuffix);
		}
	}
	
	protected	double		beamWidth_um 					= 0.0;
	public		double		getBeamWidth_um()				{ return beamWidth_um;												}
	public 		void 		setBeamWidth_um(double gw)		{ if (beamWidth_um!=gw) { beamWidth_um=gw; }		}
	
	protected	double		beamHeight_um 					= 0.0;
	public		double		getBeamHeight_um()				{ return beamHeight_um;												}
	public 		void 		setBeamHeight_um(double gw)		{ if (beamHeight_um!=gw) { beamHeight_um=gw; }		}
	
	public 		ModelBeamsize.BeamsizeSetup getBeamsizeSetup()	{ return ModelBeamsize.getBeamsizeSetup(beamWidth_um, beamHeight_um); }
	
	protected	double		lowResLimit_A 					= 50.0;
	public		double		getLowResLimit_A()				{ return lowResLimit_A;												}
	public 		void 		setLowResLimit_A(double d)		{ lowResLimit_A=d; }
	
	protected	double		highResLimit_A 					= 1.0;
	public		double		getHighResLimit_A()				{ return highResLimit_A;												}
	public 		void 		setHighResLimit_A(double d)		{ highResLimit_A=d; }
	
	protected	Double		savedZoom	 					= null;
	public		Double		getSavedZoom()					{ return savedZoom;												}
	public 		void 		setSavedZoom(Double d)			{ savedZoom=d; }
	
	protected	int			startFrame 						= 0;
	public		int			getStartFrame()					{ return startFrame;											}
	public 		void 		setStartFrame(int sf)			{ if (startFrame!=sf) { startFrame = sf; }	}
	
	protected	boolean		autoFind		 				= true;
	public		boolean		getAutoFind()					{ return getRunList().getListType()==RasterListType.AUTO && autoFind; }
	public 		void 		setAutoFind(boolean af)			{ if (autoFind!=af) { autoFind = af; } }

	protected	boolean		soniccBin		 				= false;
	public		boolean		getSONICCBin()					{ return soniccBin; }
	public 		void 		setSONICCBin(boolean b)			{ soniccBin=b; }

	protected	Point4D		soniccRefPos		 			= null;
	public		Point4D		getSONICCRefPos()				{ return soniccRefPos; }
	public		double[]	getSONICCRefPosArray()			{ if (null==soniccRefPos) return new double[]{0,0,0,0}; else return new double[]{1,soniccRefPos.x,soniccRefPos.y,soniccRefPos.z,soniccRefPos.omega}; }
	public 		void 		setSONICCRefPos(Point4D p)		{ soniccRefPos=p; }

	protected	boolean		loopCentered		 				= true;
	public		boolean		getLoopCentered()					{ return getRunList().getListType()==RasterListType.AUTO && loopCentered; }
	public 		void 		setLoopCentered(boolean af)			{ if (loopCentered!=af) { loopCentered = af; } }
	
	public int getGridOrientation()
	{
		return (getRunType() == RunType.DIFFRACTION) ? GridCalculator.RASTER_INITIAL_DIRECTION : SWT.HORIZONTAL;
	}
	
	public		enum		ROIMode
	{
		ROUGH	("rough", 	"Any fluorescence"),
		BLUICE	("BluIce",	"Use scan tab element-edge"),
		READY	("ready",	"MCA optimized on scan tab");
		
		private String scriptString;
		private String guiString;
		
		ROIMode(String _scriptString, String _guiString)
		{
			scriptString=_scriptString;
			guiString=_guiString;
		}
		
		public String scriptString()	{ return scriptString;	}
		public String guiString()		{ return guiString;		}
	}
	protected	ROIMode		roiMode 						= ROIMode.READY;
	public		ROIMode		getROIMode()					{ return roiMode;												}
	public 		void 		setROIMode(ROIMode rm)			{ if (roiMode!=rm) { roiMode = rm; }			}
	
	protected	double		exposureTime 					= Double.NaN;
	public		double		getExposureTime() 				{ return exposureTime; 											}
	public 		void 		setExposureTime(double d)		{ if (!nearlyEquals(exposureTime,d)) { exposureTime = d; } }
	
	public		boolean		isShutterless()
	{
		boolean shutterlessEnable = DevUserConfig.get().getShutterlessEnable();
		return DetectorConfig.isDectris() &&
			   shutterlessEnable &&
			   (DBConfig.getBoolean("raster_helical") || getDelta()==0.0);
	}
	
	protected	double		delta 							= Double.NaN;
	public		double		getDelta() 						{ return delta; 												}
	public 		void 		setDelta(double d)				{ if (!nearlyEquals(delta,d)) { delta = d; }				}

	protected	double		detectorDist 					= Double.NaN;
	public		double		getDetectorDist()				{ return detectorDist;	 										}
	public 		void 		setDetectorDist(double d)		{ if (!nearlyEquals(detectorDist,d)) { detectorDist = d; } }

	protected	double		energy 							= Double.NaN;
	public		double		getEnergy()						{ return energy; 												}
	public 		void 		setEnergy(double d)				{ if (!nearlyEquals(energy,d)) { energy = d; }				}

	protected	double		attenuation 					= Double.NaN;
	public		double		getAttenuation()				{ return attenuation;											}
	public 		void 		setAttenuation(double d)		{ if (!nearlyEquals(attenuation,d)) { attenuation = d; }	}

	protected	int			referenceRun 					= 0;
	public		int			getReferenceRun()				{ return referenceRun;											}
	public 		void 		setReferenceRun(int i)			{ referenceRun = i; }

	protected	int			referenceHit 					= 0;
	public		int			getReferenceHit()				{ return referenceHit;											}
	public 		void 		setReferenceHit(int i)			{ referenceHit = i; }

	protected	int			loopMeasurement_um				= 0;
	public		int			getLoopMeasurement_um()			{ return loopMeasurement_um;											}
	public 		void 		setLoopMeasurement_um(int i)	{ loopMeasurement_um = i; }

	protected	boolean		hideRunParameters 				= false;
	public		boolean		getHideRunParameters()			{ return hideRunParameters;	}
	public 		void 		setHideRunParameters(boolean hp){ hideRunParameters = hp;   }
	
	protected	Point4D		userPosition					= null;
	public		void		setUserPosition(Point4D pt)		{ userPosition = pt; }
	public		Point4D		getUserPosition()				{ return userPosition; }
	
	public static enum RefinementType
	{
		COARSE, FINE, ORTHO;
		@Override public String toString() { return Util.capitalizeFirstLetters(super.toString()); }
	};
	protected	RefinementType	refinementType 						= RefinementType.COARSE;
	public		RefinementType	getRefinementType()					{ return refinementType;											}
	public 		void 			setRefinementType(RefinementType rt){ refinementType = rt; }

	public static final double EPSILON = 0.0001;
	public static boolean nearlyEquals(double d1, double d2) { return Math.abs(d1-d2) < EPSILON; }
	
	protected ModelRasterRunList runList;
	public ModelRasterRunList getRunList() { return runList; }
	protected int runIndex;
	public int getRunIndex() { return runIndex; }
	public DevRasterRun dev() { return (DevRasterRun)runList.devRunList.getRunDev(runIndex); }
	protected ClientListener clientListener;
	protected ListenerList.Listener<DevMove> energyListener, twoThetaListener;
	private Listener<DevUserConfig> userConfigListener;
	
	public int getInterRunIndex()
	{
		if (runList.getListType()==RasterListType.INTER) return runIndex;
		else return -1;
	}
	public int getAutoRunIndex()
	{
		if (runList.getListType()==RasterListType.AUTO) return runIndex;
		else return -1;
	}
	
	///////////////////////////////////////////// Methods ///////////////////////////////////////////
	
	// returns false if this run is invalid
	protected boolean lastValidateFailed = false;
	protected boolean validateAndFix()
	{
		if (lastValidateFailed)
		{
			lastValidateFailed = false;
			return true;
		}

		boolean isValid = true;
		
		if (lowResLimit_A<0.1)
		{
			dev().setLowResLimit_A(50.0, false);
			isValid = false;
		}

		if (highResLimit_A<0.1)
		{
			dev().setHighResLimit_A(1.0, false);
			isValid = false;
		}

		if (fileRoot.length()==0)
		{
			dev().setFileroot("raster", false);
			isValid = false;
		}
		
		if (runLabel != 0 && runIndex == 0)
		{
			runLabel = 0;
			dev().setRunLabel(0, false);
			isValid = false;
		}
		
		if (loopMeasurement_um < 5)
		{
			dev().setLoopMeasurement_um(5, false);
			isValid = false;
		}
		
		if (delta < 0.0)
		{
			delta = 0.0;
			dev().setDelta(0.0, false);
			isValid = false;
		}
		
		if (nextFrame < 0)
		{
			nextFrame = 0;
			dev().setNextFrame(0, false);
			isValid = false;
		}

		if (grid().getCellWidth_um() < 1.0)
		{
			grid().setCellWidth_um(1.0);
			dev().setCellWidth_um(1.0, false);
			isValid = false;
		}
		
		if (grid().getCellHeight_um() < 1.0)
		{
			grid().setCellHeight_um(1.0);
			dev().setCellHeight_um(1.0, false);
			isValid = false;
		}
		
		/*if (grid().getMinGridWidth_um() < grid().getCellWidth_um())
		{
			grid().setMinGridWidth(grid().getCellWidth_um());
			dev().setGridWidth_um(grid().getCellWidth_um(), false);
			isValid = false;
		}
		
		if (grid().getMinGridHeight_um() < grid().getCellHeight_um())
		{
			grid().setMinGridHeight(grid().getCellHeight_um());
			dev().setGridHeight_um(grid().getCellHeight_um(), false);
			isValid = false;
		}*/
		
		if (!isValid)
		{
			dev().sendUpdate();
		}
		
		lastValidateFailed = !isValid;
		return isValid;
	}
	
	public ModelRasterRun(ModelRasterRunList _runList, int _runIndex)
	{
		runList=_runList;
		runIndex=_runIndex;
		llRunChanged = new ListenerList<ModelRasterRun>(this);
		llRunReset = new ListenerList<ModelRasterRun>(this);
		
		_grid = new ModelRasterGrid(this);
		
		clientListener = new ClientListener()
		{
			@Override public void masterChanged(boolean isMaster)
			{
				if (isMaster &&	!runList.isRasterThreadAlive())
				{
					boolean devChange = recalcHighResLimit(false);
					
        			if (RasterListType.AUTO == runList.getListType() && 0 == getRunIndex())
        			{
        				if (0 == runList.getCurrentRunIndex())
        				{
    						dev().setRunStatus(RunState.INACTIVE, false);
    						devChange = true;
        				}
        				else
        				{
        					if (RunState.INACTIVE == getRunState() || RunState.RASTERING == getRunState())
        					{
        						RunState correctedRunState = RunState.COMPLETE;
        						
        						for (int i=1;i<=runList.getRunCount();i++)
        						{
        							if (RunState.COMPLETE != ((ModelRasterRun)runList.getRun(i)).getRunState())
        							{
        								correctedRunState = RunState.PAUSED;
        							}
        						}
        						
        						dev().setRunStatus(correctedRunState, false);
        						devChange = true;
        					}
        				}
        			}
        			else
        			{
        				if (RunState.RASTERING == getRunState())
        				{
        					if (nextFrame == 0)
        					{
        						dev().setRunStatus(RunState.INACTIVE, false);
        						devChange = true;
        					}
        					else if (nextFrame == grid().getSelectedCellCount())
        					{
        						dev().setRunStatus(RunState.COMPLETE, false);
        						devChange = true;
        					}
        					else
        					{
        						dev().setRunStatus(RunState.PAUSED, false);
        						devChange = true;
        					}
        				}
        			}
        			
        			if (devChange) dev().sendUpdate();
        			DevFPE.get().setETA(0);
				}
			}
		};
		Client.addListener(clientListener);
		
		twoThetaListener = new ListenerList.Listener<DevMove>()
		{
			@Override public void handleEvent(DevMove dev, Object data)
			{
				if (!dev.isMoving()) recalcHighResLimit(true);
			}
		};
		energyListener = new ListenerList.Listener<DevMove>()
		{
			@Override public void handleEvent(DevMove dev, Object data)
			{
				if (!dev.isMoving()) recalcHighResLimit(true);
				if (getRunState()==RunState.INACTIVE) dev().setEnergy_keV(DevEnergy.getDevice().getPosition(), true);
			}
		};
		
		DevPmacMotor.get("detector_2t").llMoveStartStop.addListener(twoThetaListener);
		DevPmacMotor.get("mo_En_E").llMoveStartStop.addListener(energyListener);
		
		userConfigListener = new Listener<DevUserConfig>()
		{
			@Override public void handleEvent(DevUserConfig object, Object data)
			{
				handleChange();
			}
		};
		DevUserConfig.get().ll.addListener(userConfigListener);
	}
	
	public void stopMonitors()
	{
		DevPmacMotor.get("detector_2t").llMoveStartStop.removeListener(twoThetaListener);
		DevPmacMotor.get("mo_En_E").llMoveStartStop.removeListener(energyListener);
		Client.removeListener(clientListener);
		DevUserConfig.get().ll.removeListener(userConfigListener);
		
		grid().stopMonitors();
	}

	public void start()
	{
		if (null != grid().getErrorMessage())
		{
			Util.popupError("Grid invalid", "Grid definition is invalid.  Please fix before starting rastering.");
			return;
		}
		
		//grid().saveOrigin();

		switch (runType)
		{
		case DIFFRACTION:
			
			runList.cmdRaster = new CmdRasterDiff(this);
			runList.cmdRaster.start();
			
			break;
			
		case SONICC:
			
			runList.cmdRaster = new CmdRasterFluo(this, new Runnable(){public void run(){Util.sleep(1000);}});
			runList.cmdRaster.start();
			
			break;
		}
		
		//notifyStart();
	}
	
	public void abort()
	{
		if (RasterState.RASTERING == ModelRasterState.get().getRasterState())
		{
			DevRasterState.get().setRasterState(RasterState.STOPPING);
		}
		
		// abort isn't yet supported between clients
		if (runList.isRasterThreadAlive())
		{
    		runList.cmdRaster.requestStop();
		}
	}
	
	protected static Thread gridCenterThread = null;
	protected volatile static boolean stopCenteringThread = false;
	
	public static void centerSite(Site site, RefinementType rType)
	{
		if (null == site || null == site.motorPos) return; 
		
		Point4D motorXYZO = new Point4D(site.motorPos);
		if (RefinementType.ORTHO == rType) motorXYZO.omega += 90.0;
		
		startCenter(motorXYZO, null, "site "+site.index+" "+rType);
	}
	
	public void centerGrid(Point4D center, Double highResZoom)
	{
		if (null == center) return;
		
		String desc = "grid center for raster run index "+getRunIndex();
		Log.note("Moving to "+desc);
		
		startCenter(center, null, desc);
	}
	
	public static synchronized void startCenter(final Point4D center, final Double highResZoom, final String description)
	{
		new Thread(new Runnable(){public void run()
		{
    		while (null != gridCenterThread && gridCenterThread.isAlive())
    		{
    			stopCenteringThread = true;
    			Util.sleep(10);
    		}
    		stopCenteringThread = false;
    		
    		gridCenterThread = new Thread(new Runnable(){public void run()
    		{
    			if (stopCenteringThread) return;

    			// start omega move
    			MoveCmd omegaMove = DevPmacMotor.get("gonio_omega").startMove(center.omega);
    			
    			// start zoom move if specified
    			MoveCmd zoomMove = (null==highResZoom) ? null : DevPmacMotor.get("camera_zoom_2").startMove(highResZoom);
    			
    			// do xyz move (wait is included)
    			Point3D xyzDest = new Point3D(center.x,center.y,center.z);
    			MoveResult xyzResult = DevXYZ.syncMove(xyzDest, true);
    
    			// wait for other moves
    			MoveResult omegaResult	= omegaMove.waitFor();
    			MoveResult zoomResult	= (null==zoomMove) ? null : zoomMove.waitFor();
    			
    			if (stopCenteringThread) { return; }
        		
        		if (xyzResult.isOK() && omegaResult.isOK() && (null==zoomResult || zoomResult.isOK()))
        		{
        			Log.note("Move to "+description+" completed successfully.");
        		}
        		else
        		{
        			Log.warn("Move to "+description+" failed.");
        		}
    		}}, "Raster grid center");
    		gridCenterThread.start();
    		
		}}).start();
	}
	
	@Override public void notifyExistingRunClick()
	{
		if (null == TabRaster.get()) return;
		
		if (getRunState() == RunState.INACTIVE)
		{
			Point4D userPosition = getUserPosition();
			if (null != userPosition)
			{
				Log.note("Centering user position");
				centerGrid(userPosition, null);
			}
		}
		else
		{
			Point4D gridCenter = grid().getCenter();
			if (null != gridCenter)
			{
				Log.note("Centering grid position");
				centerGrid(gridCenter, savedZoom);
			}
		}
	}
	
	@Override public void notifyStarTabClick() {}
	@Override public void notifySyncEnabled() {}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////// run button handlers /////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////
	
	public void handleDeleteButton()
	{
		if (getRunList().getListType()==RasterListType.INTER)
		{
			boolean okToDelete = ClearSampleState.deleteInterSites(this, "Deleting", true, true, true);
			if (okToDelete)
			{
				DBRasterHits.get().handleDeletedRun(this);
				DBRasterSiteList.get().handleDeletedRun(this);
				TabRaster.get().runTabs.deleteTab(getRunIndex(), SFType.INTER_RASTER, true);
			}
		}
		else
		{
			boolean okToDelete = ClearSampleState.deleteAutoSitesAndHits(true);
			if (okToDelete)
			{
				// anything to do?
			}
		}
	}
	
	public void handleReprocessButton()
	{
		final ModelRasterRun run = this;
		new Thread(new Runnable(){public void run()
		{
    		grid().getGridCalc().stopProcessing();
			DBImageStats.delete(directorySuffix, fileRoot, runLabel, null);
    		for (ModelRasterCell cell : grid().getGridCalc().getCellListCopy())
    		{
    			Util.popupOk("","Not yet implemented");
    			//cell.process();
    		}
		}},"Reprocess raster results").start();
	}
	
	public void handleUpdateButton()
	{
		if (Util.beamSizeCalculated()) {
			int colli_um = DevCollimatorCtrl.getCollimatorSize().size_um;
			if (colli_um <= 0) colli_um = 10; // choose 10um collimator if none is set
			DBBeamSize.BeamSize bs = DBBeamSize.getBeamSize(colli_um, DevEnergy.getDevice().getPosition());
			dev().setBeamWidth_um(colli_um, false);
			dev().setBeamHeight_um(colli_um, false);
			dev().setCellWidth_um(bs.width_um, false);
			dev().setCellHeight_um(bs.height_um, false);
			grid().getGridCalc().fitGridToContour(false, false);
		}

		dev().setAttenuation_factors(DevAttenuation.get().getPosition(), false);
		dev().setDistance_mm(DevPmacMotor.get("detector_z").getPosition(), false);
		dev().setEnergy_keV(DevEnergy.getDevice().getPosition(), false);
		recalcHighResLimit(false);
		
		dev().sendUpdate();
	}
	
	public void handleResetButton()
	{
		if (getRunType()==RunType.DIFFRACTION) {
			boolean okToReset = ClearSampleState.deleteInterSites(this, "Resetting", true, true, true);
			if (okToReset) {
				resetInterRun(true);
			}
		} else {
			resetInterRun(true);
		}
		llRunReset.notifyListeners();
	}
	
	public void resetInterRun(boolean update)
	{
		// reset run state and delete SpotFinder data
		if (getSONICCBin()) CmdBinSONICC.unbin(this, false);
		dev().setSONICCRefPos(null, false);
		dev().setRunStatus(RunState.INACTIVE, false);
		dev().setUserPosition(null, false);
		dev().setNextFrame(0, false);
		grid().getGridCalc().resetData();
		DBRasterHits.get().deleteHitsForRun(this, true);
		if (update) dev().sendUpdate();
	}

	public void resetAutoRaster()
	{
		// clear grid center
		if (null != grid().getCenter())
		{
			centerGrid(grid().getCenter(), null);
		}

		// clear contours if autoFind is on
		if (getAutoFind() && grid().getContours().size()>0)
		{
			dev().setContours(new ArrayList<ModelContour>(), false);
			dev().setCenter(null, false);
			dev().setSelectedCells(new ArrayList<Boolean>(), false);
			grid().getGridCalc().fitGridToContour(false, false);
		}

		// reset run state and delete SpotFinder data
		dev().setRunStatus(RunState.INACTIVE, false);
		dev().setUserPosition(null, false);
		dev().setNextFrame(0, false);
		grid().getGridCalc().resetData();
		dev().setHideRunParameters(false, false);
		DevRasterRunList.get(RasterListType.AUTO).setCurrentRun(0,false);
		DevRasterRunList.get(RasterListType.AUTO).setRunCount(0,true);

		dev().sendUpdate();
	}
	
	public boolean recalcHighResLimit(boolean update)
	{
		if (Client.isMaster() && !runList.isRasterThreadAlive() && !ModelRasterState.get().getFixLimits())
		{
			// gather info
			double detectorZ_mm = getDetectorDist();
			double detector2t_deg = DevPmacMotor.get("detector_2t").getPosition();
			double beamEnergy_eV = DevEnergy.getDevice().getPosition()*1000.0;
			double ringRadius_mm = DetectorConfig.getDimensionsMM().x/2;
			
			// calculate and send to device
			double highResLimit_A = ResPredictor.calcLimit(detectorZ_mm, detector2t_deg, beamEnergy_eV, ringRadius_mm);
			double diff = Math.abs(this.highResLimit_A-highResLimit_A);
			if (diff >= 0.01)
			{
				dev().setHighResLimit_A(highResLimit_A, update);
				return true;
			}
			else
			{
				return false;
			}
		}
		else
		{
			return false;
		}
	}
	
	public void changeDetectorDist(Double newDist)
	{
		if (!ModelRasterState.get().getFixLimits())
		{
			// gather info
			double detectorZ_mm = newDist;
			double detector2t_deg = DevPmacMotor.get("detector_2t").getPosition();
			double beamEnergy_eV = DevEnergy.getDevice().getPosition()*1000.0;
			double ringRadius_mm = DetectorConfig.getDimensionsMM().x/2;
			
			// calculate and send to device
			double highResLimit_A = ResPredictor.calcLimit(detectorZ_mm, detector2t_deg, beamEnergy_eV, ringRadius_mm);
			dev().setHighResLimit_A(highResLimit_A, false);
		}
		
		dev().setDistance_mm(newDist,true);
	}

	public String generateRasterScreenshotFilename()
	{
		String imageFilenameNoExtension = getFullDirectory()+"/"+getFileRoot();
		int imageExtraNumber = 0;
		while(Disk.exists(imageFilenameNoExtension+".png"))
		{
			imageFilenameNoExtension = getFullDirectory()+"/"+getFileRoot()+"_"+(++imageExtraNumber);
		}
		return imageFilenameNoExtension+".png";
	}
}
