package gov.anl.gmca.raster.calc;

import gov.anl.gmca.collect.DevFPE;
import gov.anl.gmca.collect.ModelCollectRun.Point3D;
import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.hutch.DetectorConfig;
import gov.anl.gmca.raster.ModelRasterCell;
import org.eclipse.swt.SWT;

import java.util.ArrayList;

public class FramePartitioner
{
	public static class FramePartition
	{
		// start a partition based on a given first frame
		// (motion is set for stationary partitions but not used)
		public FramePartition(ModelRasterCell cell, int partitionIndex)
		{
			this.firstFilename		= cell.getImageName();
			this.prefix				= cell.getImagePrefix()+"_";
			this.position 			= new Point3D(cell.getMotorXYZ());
			this.firstFrameIndex	= cell.getIndex();
			this.column				= cell.getCol();
			this.row				= cell.getRow();
			this.frameCount			= 1;
			this.partitionIndex		= partitionIndex;
			this.lastCellRow		= cell.getRow();
			this.lastCellCol		= cell.getCol();
			
			double[] pos 			= cell.getMotorXYZ();
			double cellWidth_mm		= cell.grid().getCellWidth_um()/1000.0;
			reverseRow 				= DBConfig.getBoolean("raster_serpentine") && (cell.getRow()%2==1);
			firstCellToGridEdge 	= reverseRow ? +cellWidth_mm/2.0 : -cellWidth_mm/2.0;
			lastCellToGridEdge		= reverseRow ? -cellWidth_mm/2.0 : +cellWidth_mm/2.0;
			
			this.shutteredMotion	= new DevFPE.XYZMotion(pos[0],pos[0],pos[1],pos[1],pos[2],pos[2]);
			this.shutterlessMotion	= new DevFPE.XYZMotion(pos[0],pos[0],pos[1],pos[1],pos[2]+firstCellToGridEdge,pos[2]+lastCellToGridEdge);
		}
		
		// should only be called during frame calculations!
		void addFrameToPartition(ModelRasterCell cell)
		{
			frameCount++;

			if (GridCalculator.RASTER_INITIAL_DIRECTION == SWT.HORIZONTAL)
			{
				if (cell.getRow() != row)
				{
					throw new IllegalArgumentException("Tried to add cell "+cell.getIndex()+" which has row "+cell.getRow()+" to existing partition "+partitionIndex+" which has row "+row);
				}
				if (cell.getCol() != lastCellCol + (reverseRow ? +1 : -1))
				{
					throw new IllegalArgumentException("Tried to add cell "+cell.getIndex()+" which has column "+cell.getCol()+" to existing partition "+partitionIndex+" which ends in column "+lastCellCol);
				}
			}
			if (GridCalculator.RASTER_INITIAL_DIRECTION == SWT.VERTICAL)
			{
				if (cell.getCol() != column)
				{
					throw new IllegalArgumentException("Tried to add cell "+cell.getIndex()+" which has column "+cell.getCol()+" to existing partition "+partitionIndex+" which has column "+column);
				}
				if (cell.getRow() != lastCellRow+1)
				{
					throw new IllegalArgumentException("Tried to add cell "+cell.getIndex()+" which has row "+cell.getRow()+" to existing partition "+partitionIndex+" which ends in row "+lastCellRow);
				}
			}
			
			double[] pos	= cell.getMotorXYZ();
			shutteredMotion.endX	= pos[0];
			shutteredMotion.endY	= pos[1];
			shutteredMotion.endZ	= pos[2];
			shutterlessMotion.endX 	= pos[0];
			shutterlessMotion.endY 	= pos[1];
			shutterlessMotion.endZ 	= pos[2]+lastCellToGridEdge;
			
			lastCellRow = cell.getRow();
			lastCellCol = cell.getCol();
		}
		
		public String 	firstFilename, prefix;
		public Point3D 	position;
		public int 		firstFrameIndex;
		public int		frameCount;
		public int		column, row;
		public boolean	reverseRow;
		public int		partitionIndex;		// -1 means it was already collected
		public int		lastCellRow;		// row the last cell is in, for computing partitions
		public int		lastCellCol;		// col the last cell is in, for computing partitions
		public double	firstCellToGridEdge=0, lastCellToGridEdge=0; // distance from each cell to the grid edge. equal to half a cell width and depends on raster direction.
		public DevFPE.XYZMotion shutteredMotion, shutterlessMotion;
		
		public String getFilenameToSendToFPE()
		{
			if (DetectorConfig.isEiger())
			{
				int prefixEndIndex = firstFilename.indexOf("_master.h5");
				return firstFilename.substring(0, prefixEndIndex);
			}
			else
			{
				return firstFilename;
			}
		}
	}
	
	// This is the key to this class - it looks at a frame and sees if it lines up
	// with the last frame in the partition.
	//
	private static boolean frameBelongs(ModelRasterCell cell, FramePartition partition)
	{
		if (null==cell || null==partition) return false;
		
		try
		{
			if (!cell.grid().run().isShutterless()) return false;
		}
		catch (Exception e) { e.printStackTrace(); }

		// make sure we're in the same row/column
		// and this isn't the first cell after a gap in selected cells
		if (GridCalculator.RASTER_INITIAL_DIRECTION == SWT.HORIZONTAL)
		{
			if (cell.getRow() != partition.row) return false;
			if (cell.getCol() != partition.lastCellCol + (partition.reverseRow ? +1 : -1)) return false;
		}
		if (GridCalculator.RASTER_INITIAL_DIRECTION == SWT.VERTICAL)
		{
			if (cell.getCol() != partition.column) return false;
			if (cell.getRow() != partition.lastCellRow+1) return false;
		}
		
		return true;
	}
	
	// the only public method in this class
	//
	// nextCellIndex indicates which frame is ready to be collected next - frames before this
	// one are already taken and should be given negative partition numbers
	//
	public static synchronized ArrayList<FramePartition> partition(ArrayList<ModelRasterCell> cellList, int nextCellIndex)
	{
		// check for 0 length frame list
		if (0 == cellList.size()) return new ArrayList<FramePartition>();
		
		// copy frame list and set up blank partition list
		ArrayList<ModelRasterCell> _cellList = new ArrayList<ModelRasterCell>(cellList);
		ArrayList<FramePartition> partitionList = new ArrayList<FramePartition>();
		
		/////////////////////////////////// partition the frames /////////////////////////////////

		FramePartition lastPartition = null;
		int notTakenPartitionIndex=0, takenPartitionIndex=-1;
		for (ModelRasterCell cell : _cellList)
		{
			boolean frameTaken 			= (cell.getIndex() <  nextCellIndex);
			boolean firstFrameNotTaken 	= (cell.getIndex() == nextCellIndex);

			// for the first frame, create a partition for it
			if (null == lastPartition)
			{
				lastPartition = new FramePartition(cell, frameTaken ? takenPartitionIndex-- : notTakenPartitionIndex++);
			}
			else if (!firstFrameNotTaken && frameBelongs(cell, lastPartition)) // add to an existing part. if possible
			{
				lastPartition.addFrameToPartition(cell);
			}
			else // otherwise create a new partition and add the existing one to the list
			{
				partitionList.add(lastPartition);
				lastPartition = new FramePartition(cell, frameTaken ? takenPartitionIndex-- : notTakenPartitionIndex++);
			}
		}
		
		// add the last partition to the list
		if (null != lastPartition)
		{
			partitionList.add(lastPartition);
		}

		// reverse the negative partition numbers
		//System.out.println("=== raster repartition ===");
		for (FramePartition partition : partitionList)
		{
			if (partition.partitionIndex < 0)
			{
				partition.partitionIndex = ++takenPartitionIndex;
			}
			//System.out.println("Partition "+partition.partitionIndex+" has "+partition.frameCount+" frames");
		}

		// done!
		return partitionList;
	}
}
