package gov.anl.gmca.raster.calc;

import gov.anl.gmca.collect.DevCollectSettings;
import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.hutch.DetectorConfig;
import gov.anl.gmca.hutch.DevEnergy;
import gov.anl.gmca.raster.ModelContour;
import gov.anl.gmca.raster.ModelRasterCell;
import gov.anl.gmca.raster.ModelRasterGrid;
import gov.anl.gmca.raster.ModelRasterGrid.CellSizeMethod;
import gov.anl.gmca.raster.ModelRasterRun.RunType;
import gov.anl.gmca.raster.ModelRasterState.RasterListType;
import gov.anl.gmca.raster.calc.FramePartitioner.FramePartition;
import gov.anl.gmca.raster.calc.RasterGeometry.Rectangle;
import gov.anl.gmca.sample.DBBeamSize;
import gov.anl.gmca.sample.WdgCamera;
import gov.anl.gmca.spotfinder.SFData.SFType;
import gov.anl.gmca.spotfinder.SpotFinderQueue;
import gov.anl.gmca.util.Pair;
import gov.anl.gmca.util.Point2D;
import gov.anl.gmca.util.Point4D;
import gov.anl.gmca.util.Util;
import org.eclipse.swt.SWT;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.concurrent.Semaphore;

public class GridCalculator
{
	private ArrayList<ModelRasterCell> cellList = new ArrayList<ModelRasterCell>();
	public static int RASTER_INITIAL_DIRECTION = SWT.HORIZONTAL;
	
	public ModelRasterCell getCellOriginal(int cellIndex)
	{
		ModelRasterCell orig = null;
		
		gridLock.acquireUninterruptibly();
		try
		{
			if (cellIndex >= cellList.size())
			{
				System.out.println("Cell index "+cellIndex+" > list size "+cellList.size());
			}
			else
			{
				orig = cellList.get(cellIndex);
			}
		}
		catch (Exception e) { e.printStackTrace(); 	}
		finally 			{ gridLock.release(); 	}
		
		return orig;
	}
	
	public ModelRasterCell getCellCopy(int cellIndex)
	{
		ModelRasterCell copy = null;
		
		gridLock.acquireUninterruptibly();
		try
		{
			if (cellIndex >= cellList.size())
			{
				System.out.println("Cell index "+cellIndex+" > list size "+cellList.size());
			}
			else
			{
				copy = new ModelRasterCell(cellList.get(cellIndex), true);
			}
		}
		catch (Exception e) { e.printStackTrace(); 	}
		finally 			{ gridLock.release(); 	}
		
		return copy;
	}
	
	public ArrayList<ModelRasterCell> getCellListCopy()
	{
		ArrayList<ModelRasterCell> copy = new ArrayList<ModelRasterCell>();

		gridLock.acquireUninterruptibly();
		try					{ for (ModelRasterCell cell : cellList) copy.add(new ModelRasterCell(cell, true)); }
		catch (Exception e) { e.printStackTrace(); 	}
		finally 			{ gridLock.release(); 	}
		
		return copy;
	}
	
	public void setCell(int cellIndex, ModelRasterCell newCell)
	{
		gridLock.acquireUninterruptibly();
		try					{ cellList.set(cellIndex, newCell); }
		catch (Exception e) { e.printStackTrace(); 	}
		finally 			{ gridLock.release(); 	}
	}
	
	private	long											listVersion			= 0;
	public	long											getListVersion()	{ return listVersion; }
	
	private HashMap<Pair<Integer,Integer>,ModelRasterCell> 	cellMap 			= new HashMap<Pair<Integer,Integer>,ModelRasterCell>();
	public  HashMap<Pair<Integer,Integer>,ModelRasterCell> 	getCellMap()		{ return cellMap; }

	private ArrayList<FramePartition> 						partitionList 		= new ArrayList<FramePartition>();
	public  ArrayList<FramePartition> 						getPartitionList()	{ return partitionList; }
	
	private ModelRasterGrid 								grid				= null;
	public 	GridCalculator(ModelRasterGrid grid) 								{ this.grid = grid; }
	
	private Semaphore 										gridLock 			= new Semaphore(1);
	public 	Semaphore 										getLock() 			{ return gridLock; }

	///////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////// grid calculation /////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////
	
	public void fitGridToContour(boolean setDefaultCellAndBeamSize, boolean update)
	{
		if (0 == grid.getContours().size() || 0 == grid.getContours().get(0).points.size())
		{
			grid.run().dev().setGridHOffsets_um(0, 0, false);
			grid.run().dev().setGridVOffsets_um(0, 0, update);
			return;
		}
		
		// update the grid to contain the new contour
		Rectangle overallRectangle = null;
		ArrayList<Point2D> contourSizes = new ArrayList<Point2D>();
		for (ModelContour contour : grid.getContours())
		{
			Rectangle contourRectangle = RasterGeometry.getContainingRectangle(contour);
			contourSizes.add(contourRectangle.size());
			overallRectangle = RasterGeometry.getContainingRectangle(overallRectangle,contourRectangle);
		}
		
		// use the average contour size to calculate the default cell size
		if (setDefaultCellAndBeamSize)
		{
			double widthMax=0, heightMax=0;
			for (Point2D size : contourSizes)
			{
				widthMax = Math.max(widthMax, size.x);
				heightMax = Math.max(heightMax, size.y);
			}

			if (!Util.beamSizeCalculated()) {
				Point2D defaultSize = CellCalculator.getDefaultCellSize(widthMax, heightMax, CellSizeMethod.USE_WIDTH);
				grid.run().dev().setCellWidth_um(defaultSize.x, false);
				grid.run().dev().setBeamWidth_um(defaultSize.x, false);
				grid.run().dev().setCellHeight_um(defaultSize.y, false);
				grid.run().dev().setBeamHeight_um(defaultSize.y, false);
			} else {
				DBBeamSize.BeamSize bs_10um = DBBeamSize.getBeamSize(10, DevEnergy.getDevice().getPosition());
				grid.run().dev().setCellWidth_um(bs_10um.width_um, false);
				grid.run().dev().setCellHeight_um(bs_10um.height_um, false);
				grid.run().dev().setBeamWidth_um(10, false);
				grid.run().dev().setBeamHeight_um(10, false);
			}
		}

		// these updates will propogate locally
		grid.run().dev().setGridHOffsets_um(overallRectangle.xmin, overallRectangle.xmax, false);
		grid.run().dev().setGridVOffsets_um(overallRectangle.ymin, overallRectangle.ymax, false);
		generateCellList();
		
		// correct the exposure time if needed
		if (!isCellWidthOK(grid.getCellWidth_um()))
		{
			// determine how much to increase attenuation by
			double currentExpTime_sec = grid.run().getExposureTime();
			double minExpTime_sec = getMinExposureTime_sec(grid.getCellWidth_um());
			double currentAtten = grid.run().getAttenuation();
			double timeIncreaseFactor = minExpTime_sec / grid.run().getExposureTime();
			double compensatedAttenuation_factors = currentAtten * timeIncreaseFactor;

			grid.run().dev().setExposureTime(minExpTime_sec, false);
			grid.run().dev().setAttenuation_factors(compensatedAttenuation_factors, false);
			
			// asynchronous OK popup describing the changes
			String msg = Util.sprintf
			(
				"The exposure time was set to %.3f seconds, which requires a horizontal\n"+
				"speed greater than the limit.  The time has been increased to %.3f seconds\n"+
				"and the attenuation was increased to %.3f.",
				currentExpTime_sec,
				minExpTime_sec,
				compensatedAttenuation_factors
			);
			Util.asyncPopupOk("Exposure time corrected", msg);
		}
		
		// updated the selected cells
		ArrayList<Boolean> selectedCells = ModelContour.getEnclosedCells
		(
			grid.getContours(),
			grid.run().getGridOrientation(),
			grid.getCellWidth_um(),
			grid.getCellHeight_um(),
			grid.getCellCountX(),
			grid.getCellCountY(),
			grid.getCorrectedLeftOffset_um(),
			grid.getCorrectedTopOffset_um(),
			grid.getCorrectedWidth_um(),
			grid.getCorrectedHeight_um()
		);
		
		int selectedCount=0;
		for (Boolean b : selectedCells) if (b) selectedCount++;
		grid.setSelectedCellCount(selectedCount);
		
		grid.run().dev().setSelectedCells(selectedCells, update);
	}
	
	private double getMaxSpeed_mmPerSec()
	{
		return DBConfig.getDouble("speed_limit_sample_z") * 0.95;
	}
	
	public double getMinExposureTime_sec(double cellWidth_um)
	{
		if (!DetectorConfig.isDectris()) { return DevCollectSettings.getRasterExposureLimits().getLower(); }
		
		double cellWidth_mm = cellWidth_um / 1000.0;
		double minExpTime_sec = cellWidth_mm / getMaxSpeed_mmPerSec();
		return Math.max(minExpTime_sec,DevCollectSettings.getRasterExposureLimits().getLower());
	}
	
	public boolean isCellWidthOK(double cellWidth_um)
	{
		if (!DetectorConfig.isDectris()) return true;
		
		double cellWidth_mm = cellWidth_um / 1000.0;
		double reqSpeed_mmPerSec = cellWidth_mm / grid.run().getExposureTime();
		return reqSpeed_mmPerSec <= getMaxSpeed_mmPerSec();
	}
	
	@SuppressWarnings("unused")
	public void generateCellList()
	{
		Point4D gridCenter = grid.getCenter();
		if (null==gridCenter) { return; }
		
		int requestedCountX = (int)Math.ceil(grid.getMinGridWidth_um()/grid.getCellWidth_um());
		int requestedCountY = (int)Math.ceil(grid.getMinGridHeight_um()/grid.getCellHeight_um());
		
		grid.setErrorMessage(null);
		
		if (grid.getSelectedCellCount() > ModelRasterGrid.CELLCOUNTMAX)
		{
			grid.setErrorMessage("Cell count ("+(grid.getCellCountX()*grid.getCellCountY())+") must not exceed "+ModelRasterGrid.CELLCOUNTMAX);
		}

		if (null != grid.getErrorMessage())
		{
			grid.setCellCountX(0);
			grid.setCellCountY(0);
			grid.setTotalCellCount(0);
			grid.setSelectedCellCount(0);
			cellList = null;
			cellMap = new HashMap<Pair<Integer,Integer>,ModelRasterCell>();
			return;
		}

		grid.setCellCountX(requestedCountX);
		grid.setCellCountY(requestedCountY);
		grid.setTotalCellCount(requestedCountX * requestedCountY);

		double actualGridWidth_um = grid.getCellCountX() * grid.getCellWidth_um();
		double actualGridHeight_um = grid.getCellCountY() * grid.getCellHeight_um();
		
		double xCorrection = (actualGridWidth_um - grid.getMinGridWidth_um())/2.0;
		double yCorrection = (actualGridHeight_um - grid.getMinGridHeight_um())/2.0;
		
		grid.setCorrectedLeftOffset_um 		(grid.getLeftOffset_um() - xCorrection);
		grid.setCorrectedRightOffset_um 	(grid.getRightOffset_um() + xCorrection);
		grid.setCorrectedTopOffset_um 		(grid.getTopOffset_um() - yCorrection);
		grid.setCorrectedBottomOffset_um 	(grid.getBottomOffset_um() + yCorrection);
		
		double gridStartX_um = grid.getCorrectedLeftOffset_um() + (grid.getCellWidth_um() / 2.0);
		double gridStartY_um = grid.getCorrectedTopOffset_um() + (grid.getCellHeight_um() / 2.0);
		
		// update the cell list
		gridLock.acquireUninterruptibly();
		try
		{
			cellList = new ArrayList<ModelRasterCell>();
			
			switch (grid.run().getRunType())
			{
			case DIFFRACTION:
				if (RASTER_INITIAL_DIRECTION == SWT.HORIZONTAL) fillCellsInRows(gridStartX_um, gridStartY_um, gridCenter);
				else											fillCellsInColumns(gridStartX_um, gridStartY_um, gridCenter);
				partitionList = FramePartitioner.partition(cellList, grid.run().getNextFrame());
				break;
			case SONICC:
				fillCellsInRows(gridStartX_um, gridStartY_um, gridCenter);
				break;
			default:
				throw new IllegalArgumentException("Unsupported run type");
			}
		}
		catch (Exception e)	{ e.printStackTrace();	}
		finally				{ gridLock.release();	}
		
		listVersion++;
		
		//Log.note("Recalc(): "+cellList.size()+" cells");
	}
	
	public void repartition()
	{
		gridLock.acquireUninterruptibly();
		try
		{
			if (grid.run().getRunType()==RunType.DIFFRACTION)
			{
				partitionList = FramePartitioner.partition(cellList, grid.run().getNextFrame());
			}
		}
		catch (Exception e)	{ e.printStackTrace();	}
		finally				{ gridLock.release();	}
	}
	
	protected void fillCellsInColumns(double gridStartX_um, double gridStartY_um, Point4D gridCenter)
	{
		cellMap = new HashMap<Pair<Integer,Integer>,ModelRasterCell>();
		
		int selectedCellNumber=0, fullGridCellIndex=0;
		for (int col=grid.getCellCountX()-1; col>=0; col--)
		{
    		for (int row=0;row<grid.getCellCountY();row++)
    		{
				if (grid.run().dev().isCellSelected(fullGridCellIndex))
				{
					addCell(row, col, selectedCellNumber++, gridStartX_um, gridStartY_um, gridCenter);
				}
				fullGridCellIndex++;
			}
		}
		grid.setSelectedCellCount(cellList.size());
	}
	
	protected void fillCellsInRows(double gridStartX_um, double gridStartY_um, Point4D gridCenter)
	{
		cellMap = new HashMap<Pair<Integer,Integer>,ModelRasterCell>();
		
		boolean reverseOddRows = DBConfig.getBoolean("raster_serpentine");
		
		int selectedCellNumber=0, fullGridCellIndex=0;
		for (int row=0;row<grid.getCellCountY();row++)
		{
			int startCol, onePastEndCol, colIncrement;
			if (reverseOddRows && row%2==1)
			{
				startCol = 0;
				onePastEndCol = grid.getCellCountX();
				colIncrement = 1;
			}
			else
			{
				startCol = grid.getCellCountX()-1;
				onePastEndCol = -1;
				colIncrement = -1;
			}
			for (int col=startCol; col!=onePastEndCol; col += colIncrement)
			{
				if (grid.run().dev().isCellSelected(fullGridCellIndex))
				{
					addCell(row, col, selectedCellNumber++, gridStartX_um, gridStartY_um, gridCenter);
				}
				fullGridCellIndex++;
			}
		}
		grid.setSelectedCellCount(cellList.size());
	}
	
	protected void addCell(int row, int col, int selectedCellNumber, double topLeftCellCenterXOffsetFromGridCenter_um, double topLeftCellCenterYOffsetFromGridCenter_um, Point4D gridCenter)
	{
		double cellXOffsetFromGridCenter_um = topLeftCellCenterXOffsetFromGridCenter_um + (col*grid.getCellWidth_um());
		double cellYOffsetFromGridCenter_um = topLeftCellCenterYOffsetFromGridCenter_um + (row*grid.getCellHeight_um());
		
		double[] xy = new double[2];
		xy[0] = cellXOffsetFromGridCenter_um / 1000.0; // cell list is in um, cell array is in mm
		xy[1] = cellYOffsetFromGridCenter_um / 1000.0;
		
		double[] xyz = new double[3];
		if (null != grid.getCenter())
		{
			double gridOmega = grid.getCenter().omega;
			xyz = WdgCamera.screenTopLeftPxToMotorMM(WdgCamera.CameraType.HIGHRES, gridOmega, xy[0], xy[1], 1.0);
			
			ModelRasterCell cell = new ModelRasterCell(grid, selectedCellNumber, row, col, xyz, gridCenter);
			cellList.add(cell);
			cellMap.put(new Pair<Integer,Integer>(row,col), cell);
		}
	}
	
	public void stopProcessing()
	{
		// stop all processing threads from a) processing, and b) posting results to the table
		SFType sfType = grid.run().getRunList().getListType()==RasterListType.AUTO?SFType.AUTO_RASTER:SFType.INTER_RASTER;
		SpotFinderQueue.stopActiveThreads(sfType, grid.run().getRunIndex());
		
		// is this delay needed to wait for updates?
		Util.sleep(200);
	}
	
	public void resetData()
	{
		stopProcessing();
	}
}
