package gov.anl.gmca.raster.calc;

import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.common.Redis;
import gov.anl.gmca.raster.DBRasterHits;
import gov.anl.gmca.raster.ModelRasterCell;
import gov.anl.gmca.raster.ModelRasterGrid;
import gov.anl.gmca.spotfinder.DBImageStats;
import gov.anl.gmca.spotfinder.DevSFType.SPOTFINDER_OUTPUT;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.PerfTimer;
import gov.anl.gmca.util.Util;

import java.nio.FloatBuffer;
import java.util.ArrayList;
import java.util.LinkedList;

// Uses flood fill to group cells in to hits
// The output is in the 'hit' field of the RasterResults table - cells within a hit
// have their hit number set
public class HitFinder
{
	protected ArrayList<ModelRasterCell> cellList;
	protected ArrayList<ArrayList<HFCell>> hitsList = new ArrayList<ArrayList<HFCell>>();
	protected int cellCountX, cellCountY;
	protected HFCell[][] hfGrid;
	protected ModelRasterGrid grid;
	private Double cutoffOverride, hitCutoff;
	private boolean saveToSQL;
	private String output;
	public ArrayList<ArrayList<Integer>> hits = new ArrayList<>();

	public static class HFOutput {
		public ArrayList<ArrayList<Integer>> hitCells;
		public HFCell[][] grid;
	}

	protected static Double HITFORMULA_P = null;
	protected static Double HITFORMULA_C = null;
	
	protected static class HFCell
	{
		boolean valid;
		boolean searched=false;
		int hitNumber=-1;
		int x,y; // row,col
		ModelRasterCell cell;
		public HFCell(ModelRasterCell cell, int x, int y) { this.valid = (null!=cell); this.x=x; this.y=y; this.cell=cell;}
	}
	
	public static Double correctBraggCount(Double braggCandidates, double exposureTime)
	{
		if (null==braggCandidates) return null;
		
		double ret = braggCandidates;
		double C = Double.parseDouble(DBConfig.get("hitfinder_c"));
		ret -= Math.round(C * exposureTime);
		if (ret < 0) ret = 0;
		return ret;
	}

	public HitFinder(ModelRasterGrid grid, Double cutoffOverride, boolean saveToSQL)
	{
		this.grid = grid;
		this.cutoffOverride=cutoffOverride;
		this.saveToSQL=saveToSQL;
		
		boolean dbOK = true;
		try
		{
    		HITFORMULA_P = Double.parseDouble(DBConfig.get("hitfinder_p"));
    		HITFORMULA_C = Double.parseDouble(DBConfig.get("hitfinder_c"));
		}
		catch (NullPointerException npe) { npe.printStackTrace(); dbOK = false; }
		catch (NumberFormatException nfe) { nfe.printStackTrace(); dbOK = false; }

		if (!dbOK)
		{
			Log.error("Error loading HitFinder parameters, using defaults.");
			HITFORMULA_P = 0.10;
			HITFORMULA_C = 10.0;
		}
	}
	
	private void init()
	{
		cellCountX = grid.getCellCountX();
		cellCountY = grid.getCellCountY();
		cellList = grid.getGridCalc().getCellListCopy();

		hfGrid = new HFCell[cellCountX][cellCountY];
		
		// find hit cutoff
		if (null!=cutoffOverride) {
			hitCutoff = cutoffOverride;
		} else {
			int maxSpotTotal = 0;
			FloatBuffer data = DBImageStats.load(Redis.image_stats(grid.run().getDirectorySuffix(), grid.run().getFileRoot(), grid.run().getRunLabel(), output.toString()));
			for (ModelRasterCell cell : cellList) {
				int spotTotal = (int)data.get(cell.index);
				if (spotTotal > maxSpotTotal) maxSpotTotal = spotTotal;
			}
			double expTime = grid.run().getExposureTime();

			double zingerTerm = HITFORMULA_C * expTime;
			double correctedSpotMax = (double) maxSpotTotal - zingerTerm;
			hitCutoff = (HITFORMULA_P * correctedSpotMax) + zingerTerm;
			Log.note("Max spot total = "+maxSpotTotal+", cutoff = "+Util.printDouble(hitCutoff,2));
		}

		// create hfGrid which is used in the search
		FloatBuffer data = DBImageStats.load(Redis.image_stats(grid.run().getDirectorySuffix(), grid.run().getFileRoot(), grid.run().getRunLabel(), output));
		for (int i=0;i<cellCountX;i++)
		{
    		for (int j=0;j<cellCountY;j++)
    		{
    			ModelRasterCell tempCell = null;

    			for (ModelRasterCell cell : cellList)
    			{
    				boolean isHit = data.get(cell.index) >= hitCutoff;
    				if (cell.getCol()==i && cell.getRow()==j && isHit) tempCell = cell;
    			}
    			
    			hfGrid[i][j] = new HFCell(tempCell, i, j);
    		}
		}
	}
	
	public void find(boolean lookForMultipleHits, boolean combineAdjacentCells, String output)
	{
		this.output=output;
		init();

		if (lookForMultipleHits)
		{
    		// try starting a flood fill at each cell
    		for (int i=0;i<cellCountX;i++)
    		{
        		for (int j=0;j<cellCountY;j++)
        		{
        			boolean foundHit = floodFill(i,j, !combineAdjacentCells);
        			if (foundHit) currentHit++;
        		}
    		}
		}
		else
		{
			// just look for the highest spot total
			int spotMax = -1;
			ModelRasterCell maxCell = null;
			FloatBuffer data = DBImageStats.load(Redis.image_stats(grid.run().getDirectorySuffix(), grid.run().getFileRoot(), grid.run().getRunLabel(), output.toString()));
			for (ModelRasterCell cell : cellList)
			{
				Integer spotTotal = (int)data.get(cell.index);
				if (null != spotTotal && spotTotal > spotMax)
				{
					spotMax = spotTotal;
					maxCell = cell;
				}
			}
			
			// set that one cell to be the one and only hit
			if (null == maxCell || spotMax==0)
			{
				Log.error("Hit finder found no spots");
			}
			else
			{
				int runIndex = grid.run().getRunIndex();
				ArrayList<Integer> cellList = new ArrayList<Integer>();
				cellList.add(maxCell.getIndex());
				if (saveToSQL) {
					DBRasterHits.get().addHit(1, -1, runIndex, cellList.toArray(new Integer[0]), true);
				} else {
					hits.add(cellList);
				}
			}
		}
	}
	
	protected int currentHit = 1;
	
	protected boolean floodFill(int x, int y, boolean skipFill)
	{
		HFCell hfcell = hfGrid[x][y];
		if (!hfcell.valid) return false;
		if (hfcell.searched) return false;

		hfcell.searched=true;
		
		LinkedList<HFCell> q = new LinkedList<HFCell>();
		q.add(hfcell);

		ArrayList<Integer> cellList = new ArrayList<Integer>();
		while (!q.isEmpty())
		{
			HFCell first = q.removeFirst();
			cellList.add(first.cell.getIndex());

			if (skipFill) break;

			if (first.x>0)
			{
				HFCell west = hfGrid[first.x-1][first.y];
				if (west.valid && !west.searched) { west.searched = true; q.add(west); }
			}
			if (first.x<cellCountX-1)
			{
				HFCell east = hfGrid[first.x+1][first.y];
				if (east.valid && !east.searched) { east.searched = true; q.add(east); }
			}
			if (first.y>0)
			{
				HFCell north = hfGrid[first.x][first.y-1];
				if (north.valid && !north.searched) { north.searched = true; q.add(north); }
			}
			if (first.y<cellCountY-1)
			{
				HFCell south = hfGrid[first.x][first.y+1];
				if (south.valid && !south.searched) { south.searched = true; q.add(south); }
			}
		}
		int runIndex = grid.run().getRunIndex();
		if (saveToSQL) {
			DBRasterHits.get().addHit(currentHit, -1, runIndex, cellList.toArray(new Integer[0]), true);
		} else {
			hits.add(cellList);
		}

		return true;
	}
}
