package gov.anl.gmca.raster.cam;

import gov.anl.gmca.raster.ModelRasterCell;
import gov.anl.gmca.raster.ModelRasterRun;
import gov.anl.gmca.raster.ModelRasterRunList;
import gov.anl.gmca.raster.ModelRasterState;
import gov.anl.gmca.raster.ModelRasterState.NumberingParameter;
import gov.anl.gmca.spotfinder.DevSFType;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.users.ModelUserTable.ClientListener;
import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.Point2D;
import gov.anl.gmca.util.Point4D;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;

// This class handles state changes and directs mouse events to the right VC object
//
// One object is instantiated separately for each camera widget.
//
public class ModelRasterVideo
{
	protected 	final static double 			OMEGA_TOLERANCE_DEG 		= 5.0;
	public 		static enum 					VideoState 					{ POINT_AND_CLICK, CREATE_CONTOUR, CREATE_SITE, VIEW_IMAGES };
	protected 	VideoState 						state;
	public 		VideoState 						getState() 					{ return state; 			}
	public 		ListenerList<ModelRasterVideo> 	llStateChange;
	protected 	volatile static boolean 		showPolygonOutlines 				= true;
	public		static void 					setEnableNumbers(boolean b) { showPolygonOutlines = b; }
	public 		double[] 						currentFrameMotorPos 		= {0,0,0,0};
	public		Point2D							current_mm2px				= new Point2D(1,1);
	public		VCCreatePolygon					vcCreatePolygon;
	public		VCCreateSite 					vcCreateSite;
	public		VCViewImages					vcViewImages;
	public		VCCellHover						vcCellHover;
	public		DrawStandard 					drawStandard;
	public		WdgRasterCamera 				cam;

	public String								displayParam				= "";
	public NumberingParameter					displayNumbering			= NumberingParameter.CELL;

	protected int								displayRunIndex				= 0;
	public void 								setDisplayRunIndex(int i)	{
		displayRunIndex=i;
		drawStandard.rasterMap.setDisplayRunIndex(i);
	}
	protected ModelRasterRun					getDisplayRun()				{ return (ModelRasterRun)ModelRasterRunList.get(ModelRasterState.RasterListType.INTER).getRun(displayRunIndex); }
	
	////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////
	
	public ModelRasterVideo(WdgRasterCamera cam, VideoState startupState)
	{
		this.cam=cam;
		llStateChange = new ListenerList<ModelRasterVideo>(this);
		state = startupState;
		
		vcCreatePolygon = new VCCreatePolygon(this);
		vcCreateSite = new VCCreateSite(this);
		vcViewImages = new VCViewImages(this);
		vcCellHover  = new VCCellHover(this);
		drawStandard = new DrawStandard(this);
		
		Client.addListener(new ClientListener()
		{
			@Override public void masterChanged(boolean isMaster)
			{
				if (!isMaster) changeToDefaultState();
			}
		});
	}
	
	private VideoState defaultState = VideoState.VIEW_IMAGES;
	public void changeToDefaultState()
	{
		if (state==defaultState) return;
		state = defaultState;
		llStateChange.notifyListeners();
	}
	
	public boolean changeState(VideoState newState)
	{
		if (state==newState) return true;
		
		state = newState;
		if (state==VideoState.VIEW_IMAGES || state==VideoState.POINT_AND_CLICK)
		{
			//Log.note("Default state is now "+state);
			defaultState = state;
		}
		
		llStateChange.notifyListeners();
		return true;
	}

	//////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////// key handling /////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////
	
	public void handleKeyPressed(WdgRasterCamera cam, KeyEvent ke)
	{
		VideoState vState = getState();
		System.out.println("Raster camera key pressed, video state = "+vState);

		switch (vState)
		{
		case POINT_AND_CLICK:
    		break;
    		
		case CREATE_CONTOUR:
    		break;
    		
		case CREATE_SITE:
    		break;
			
		case VIEW_IMAGES:
			vcViewImages.onKeyPress(ke.keyCode);
			break;
    		
		default:
    		break;
    	}
	}
	
	//////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////// mouse handling ////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////
	
	public boolean handleLeftClick(WdgRasterCamera cam, MouseEvent me)
	{
		VideoState vState = getState();
		Point relPt = new Point(me.x-cam.drawX, me.y-cam.drawY);
		Log.ui("Raster camera click, video state = "+vState);

		switch (vState)
		{
		case POINT_AND_CLICK:
    		return false;
    		
		case CREATE_CONTOUR:
    		vcCreatePolygon.addContourPoint(relPt.x, relPt.y);
    		cam.redraw();
    		return true;
    		
		case CREATE_SITE:
			vcCreateSite.onClick(relPt.x, relPt.y, 1);
    		return true;
    		
		case VIEW_IMAGES:
			vcViewImages.onClick(relPt.x, relPt.y, 1);
    		
		default:
			return true;
    	}
	}
	
	public boolean handleRightClick(WdgRasterCamera cam, MouseEvent me)
	{
		VideoState vState = getState();
		Point relPt = new Point(me.x-cam.drawX, me.y-cam.drawY);
		Log.ui("Raster camera click, video state = "+vState);

		switch (vState)
		{
		case POINT_AND_CLICK:
			vcViewImages.saveRightClickPosition(relPt.x,relPt.y);
    		return false;
    		
		case CREATE_CONTOUR:
    		return true;
    		
		case CREATE_SITE:
			vcCreateSite.onClick(relPt.x, relPt.y, 2);
    		return true;
    		
		default:
			return true;
    	}
	}
	
	public void handleScroll(int direction)
	{
		VideoState vState = getState();
		Log.ui("Raster camera click, video state = "+vState);

		switch (vState)
		{
		case POINT_AND_CLICK:
			break;
    		
		case CREATE_CONTOUR:
			break;
    		
		case CREATE_SITE:
			break;
    		
		case VIEW_IMAGES:
			vcViewImages.onScroll(direction);
			break;
    	}
	}

	public void handleMouseMove(MouseEvent me)
	{
		VideoState vState = getState();
		Point relPt = new Point(me.x-cam.drawX, me.y-cam.drawY);
		
		switch (vState)
		{
		case POINT_AND_CLICK:
    		break;
    		
		case CREATE_CONTOUR:
    		break;
    		
		case CREATE_SITE:
		{
			boolean b1Down = (me.stateMask & SWT.BUTTON1)!=0;
			boolean b2Down = (me.stateMask & SWT.BUTTON3)!=0; // SWT calls right button "3"
			int buttonDown = 0;
			if (b1Down&&!b2Down) buttonDown=1;
			if (!b1Down&&b2Down) buttonDown=2;
			vcCreateSite.onMove(relPt.x, relPt.y, buttonDown);
			break;
		}
			
		case VIEW_IMAGES:
		{
			//vcCellHover.handleMouseMove(new Point(relPt.x,relPt.y));
			boolean b1Down = (me.stateMask & SWT.BUTTON1)!=0;
			if (b1Down)
			{
				vcViewImages.onDrag(relPt.x, relPt.y, 1);
			}
		}
    	}
	}
	
	////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////// coordinate translations /////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////
	
	public Point gridRelativeMicrons2Screen(Point2D point_um)
	{
		Point4D gridCenter = cam.model.getDisplayRun().grid().getCenter();
		if (null==gridCenter) return null;
		
		Point2D gridCenterPx = cam.motorMMtoCurrentScreenPx(gridCenter);
		Point2D pointPx = new Point2D(cam.mm2pxH(point_um.x/1000.0),cam.mm2pxV(point_um.y/1000.0));
		Point2D screenPt = gridCenterPx.plus(pointPx);
		
		screenPt = cam.model.vcViewImages.applyTransforms(screenPt);
		return screenPt.round();
	}

	// returns an SWT Rectangle object
	//
	// this method performs calculations for the top left and bottom right of the cell
	// separately so that calculations for adjacent cells will come out exactly the same
	//
	// it's not the most efficient but it's clear and accurate
	//
	public Rectangle getCellScreenBounds(ModelRasterCell cell)
	{
		// the position in cell widths or heights of the top left of this cell
		double cellUnitsX 		= (double)cell.getCol();
		double cellUnitsY 		= (double)cell.getRow();
		
		double cellWidth_um 	= cell.grid().getCellWidth_um();
		double cellHeight_um 	= cell.grid().getCellHeight_um();
		
		double x1rel_um = cellUnitsX		* cellWidth_um;
		double y1rel_um = cellUnitsY		* cellHeight_um;
		double x2rel_um = (cellUnitsX+1.0)	* cellWidth_um;
		double y2rel_um = (cellUnitsY+1.0)	* cellHeight_um;

		x1rel_um += cell.grid().getCorrectedLeftOffset_um();
		y1rel_um += cell.grid().getCorrectedTopOffset_um();
		x2rel_um += cell.grid().getCorrectedLeftOffset_um();
		y2rel_um += cell.grid().getCorrectedTopOffset_um();
		
		Point scrTL = gridRelativeMicrons2Screen(new Point2D(x1rel_um,y1rel_um));
		Point scrBR = gridRelativeMicrons2Screen(new Point2D(x2rel_um,y2rel_um));
		if (null==scrTL||null==scrBR) return null;
		
		return new Rectangle(scrTL.x,scrTL.y,scrBR.x-scrTL.x,scrBR.y-scrTL.y);
	}
}
