package gov.anl.gmca.raster.cam;

import gov.anl.gmca.common.Redis;
import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.raster.*;
import gov.anl.gmca.raster.ui.TabRaster;
import gov.anl.gmca.raster.ui.WdgRasterOptions;
import gov.anl.gmca.spotfinder.DBImageStats;
import gov.anl.gmca.util.PerfTimer;
import gov.anl.gmca.util.Util;
import org.apache.commons.lang3.ArrayUtils;
import org.eclipse.swt.graphics.Color;

import java.awt.*;
import java.awt.geom.AffineTransform;
import java.awt.image.*;
import java.nio.FloatBuffer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.concurrent.atomic.AtomicBoolean;

public class RasterMap implements Runnable
{
	public		static AtomicBoolean	updatedRequested		= new AtomicBoolean(false);
	protected 	final static double 	zFloor 					= 1.0;  // lowest z scale factor for raster data
	protected 	final static int 		COLORMAP_ALPHA 			= 200; 	// transparency of the raster map
	private		final static int		MAX_WAIT_MS				= 200;
	
	protected 	BufferedImage			finalImage 				= null;
	public 		BufferedImage			getImage()
	{
		//updateImage();
		return finalImage;
	}

	private FloatBuffer data;
	private int displayRunIndex = 0;
	public void setDisplayRunIndex(int i) { displayRunIndex=i; }

	public 	static short[][]			colormap 				= null;
	public 		static Color 			getColorMapValue(int i) { if (i<0||i>255) return GUI.COLOR_BLACK; verifyColorMap(); return new Color(GUI.getDisplay(), colormap[0][i], colormap[1][i], colormap[2][i]); }
	
	private		ModelRasterVideo		model					= null;
	
	////////////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////// ctor, run() and updateImage() //////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	public RasterMap(ModelRasterVideo model)
	{
		this.model=model;
		
		verifyColorMap();
		
		Thread thread = new Thread(this, "Raster map");
		thread.start();
		
		connectListeners();
	}
	
	private void connectListeners()
	{
	}
    
	public void run()
	{
		long lastUpdateMs = 0L;
		
		while (!Util.atExit())
		{
			try
			{
				updateImage();
				lastUpdateMs = System.currentTimeMillis();
			}
			catch (Exception e) { e.printStackTrace(); }
			
			while (true)
			{
				if (updatedRequested.get()) break;
				
				long now = System.currentTimeMillis();
				long sinceLastUpdateMs = now - lastUpdateMs;
				if (sinceLastUpdateMs > MAX_WAIT_MS) break;
				
				Util.sleep(10);
			}
			updatedRequested.set(false);
		}
	}
	
	private long lastTimerPrint = 0;
	protected void updateImage()
	{
		if (!createSmallMap()) return;
		if (!interpolateAndMapColors()) return;
		if (!applyMask()) return;
	}
    
	////////////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////// main methods //////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	protected int 				fullWidth, fullHeight;
	protected BufferedImage 	smallImage, fullImage;
	protected ModelRasterGrid 	currentGrid;
	
    protected boolean createSmallMap()
    {
    	ModelRasterRun currentRun = (ModelRasterRun)ModelRasterRunList.get(ModelRasterState.RasterListType.INTER).getRun(displayRunIndex);
    	
    	if (Util.atExit() 											||
    		null == TabRaster.get() 								||
    		null == currentRun 										||
    		null == (currentGrid=currentRun.grid()) 				||
    		0    == currentGrid.getSelectedCellCount())
    	{
    		setImage(null);
    		return false;
    	}
    		
		fullWidth = (int)Math.round(model.cam.mm2pxH(currentGrid.getCorrectedWidth_um()/1000.0));
		fullHeight = (int)Math.round(model.cam.mm2pxV(currentGrid.getCorrectedHeight_um()/1000.0));

		if (0>=fullWidth || 0>=fullHeight) { setImage(null); return false; }
		
		ArrayList<ModelRasterCell> cellListCopy = currentGrid.getGridCalc().getCellListCopy();
		String outputString = model.displayParam;
		data = DBImageStats.load(Redis.image_stats(currentRun.getDirectorySuffix(), currentRun.getFileRoot(), currentRun.getRunLabel(), outputString));
		double maxZ = DBImageStats.max(data,0);
		if (maxZ<zFloor) maxZ=zFloor; // prevent divide by zero or scaling up noise too much
		
		if(Util.atExit()) return false;

		// fill image with data
		smallImage = new BufferedImage(currentGrid.getCellCountX(),currentGrid.getCellCountY(),BufferedImage.TYPE_BYTE_INDEXED);
		for (ModelRasterCell cell : cellListCopy)
		{
			float dataVal = cell.index<data.capacity()?data.get(cell.index):-1;
			if (dataVal>0)
			{
    			int val = (int)((dataVal/maxZ)*255.0); // scale value to 0-255
    			smallImage.setRGB(cell.getCol(), cell.getRow(), (val<<0) + (val<<8) + (val<<16) + (COLORMAP_ALPHA<<24));
			}
			else
			{
				smallImage.setRGB(cell.getCol(), cell.getRow(), 0);
			}
		}
		
		return true;
    }
    
    protected boolean interpolateAndMapColors()
    {
		if(Util.atExit()) return false;

		int interpolationType =
			WdgRasterOptions.getEnableBlending()
			? AffineTransformOp.TYPE_BICUBIC
			: AffineTransformOp.TYPE_NEAREST_NEIGHBOR;
		
		// set up transform
		double scaleX = (double)fullWidth/(double)currentGrid.getCellCountX();
		double scaleY = (double)fullHeight/(double)(currentGrid.getCellCountY());
		AffineTransform transform = AffineTransform.getScaleInstance(scaleX,scaleY);
    	AffineTransformOp transformOp = new AffineTransformOp(transform, interpolationType);
    	
    	// create dest image and write
    	BufferedImage interpolatedImage = new BufferedImage(fullWidth,fullHeight,BufferedImage.TYPE_INT_ARGB);
    	Graphics2D destG2 = (Graphics2D)interpolatedImage.getGraphics();
    	destG2.drawImage(smallImage,transformOp,0,0);
    	destG2.dispose();
    	
		if(Util.atExit())return false;

		// apply colormap
    	LookupTable lookupTable = new ShortLookupTable(0,colormap);
    	LookupOp lop = new LookupOp(lookupTable,null);
    	fullImage = lop.filter(interpolatedImage,null);

		if(Util.atExit())return false;

    	return true;
    }
    
    protected boolean applyMask()
    {
    	double cellWidthPx 	= model.cam.mm2pxH(currentGrid.getCellWidth_um()/1000.0);
    	double cellHeightPx = model.cam.mm2pxV(currentGrid.getCellHeight_um()/1000.0);
    	
		//System.out.println("full="+fullWidth+"x"+fullHeight+" cell="+cellWidthPx+"x"+cellHeightPx);
    	
    	// mask areas not rastered yet
		for (int y=0; y<fullHeight; y++)
		{
	    	for (int x=0; x<fullWidth; x++)
	    	{
    			int cellX = (int)((double)x/cellWidthPx);
    			int cellY = (int)((double)y/cellHeightPx);
    			if (cellX>=currentGrid.getCellCountX()) cellX=currentGrid.getCellCountX()-1;
    			if (cellY>=currentGrid.getCellCountY()) cellY=currentGrid.getCellCountY()-1;
    			try {
					ModelRasterCell cell = currentGrid.getCell(cellY, cellX);
					//if (null!=data&&null!=cell&&cell.index<data.capacity()&&data.get(cell.index)>=0)
					boolean maskCell = false;
					boolean diffMode = currentGrid.run().getRunType() == ModelRasterRun.RunType.DIFFRACTION;
					boolean runInactive = currentGrid.run().getRunState() == ModelRasterRun.RunState.INACTIVE;
					boolean negativeData = null!=cell && cell.index<data.capacity() && data.get(cell.index)<0;
					boolean nullCell = null==cell;
					ModelRasterState state = ModelRasterState.get();
					boolean noneDisplay = (null!=state && state.getDiffDisplayParameter()==null);
					boolean mask = diffMode && (nullCell||runInactive||negativeData||noneDisplay);
					if (!diffMode && ModelRasterState.get().getSONICCDisplay() == ModelRasterState.SONICCDisplay.NONE) mask=true;
					if (mask) {
						int rgb = fullImage.getRGB(x, y);
						rgb &= 0x00FFFFFF;
						fullImage.setRGB(x, y, rgb);
					}
				} catch (Exception e) {
    				e.printStackTrace();
				}
    		}
    		//System.out.println();
    	}
		//System.out.println();
    	
    	setImage(fullImage);

    	return true;
    }
    
    /////////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////// helpers //////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////////////
	
	protected static void verifyColorMap()
	{
		if (null != colormap) return;
		
    	// from http://http.developer.nvidia.com/GPUGems/gpugems_ch08.html
    	short[] red = new short[256];
    	short[] green = new short[256];
    	short[] blue = new short[256];
    	short[] alpha = new short[256];

    	final double C = 2.0;
    	for (short i = 0; i < 256; i++)
    	{
    	    double dval = (i)/255.0;
    		
    		red[i] = (short)(255.0*bump(C*(dval-0.75)));
    		green[i] = (short)(255.0*bump(C*(dval-0.50)));
    		blue[i] = (short)(255.0*bump(C*(dval-0.25)));
    		alpha[i] = 200;
    	}
    	
    	colormap = new short[][] { red, green, blue, alpha };
	}
	
    protected static double bump(double x)
    {
    	return (Math.abs(x)>1)?0:(1.0-x*x);
    }
    
    protected void setImage(BufferedImage image)
    {
    	finalImage = image;
    }
}
