package gov.anl.gmca.raster.cam;

import gov.anl.gmca.collect.ModelCollectRun.Point3D;
import gov.anl.gmca.common.AppState;
import gov.anl.gmca.common.AppState.Operation;
import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.hutch.DevXYZ;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PV_R_RUNS;
import gov.anl.gmca.raster.ModelRasterCell;
import gov.anl.gmca.raster.ModelRasterGrid;
import gov.anl.gmca.raster.ModelRasterRun;
import gov.anl.gmca.raster.ModelRasterRun.RunState;
import gov.anl.gmca.raster.ModelRasterState;
import gov.anl.gmca.raster.ModelRasterState.RasterListType;
import gov.anl.gmca.raster.cam.ModelRasterVideo.VideoState;
import gov.anl.gmca.raster.util.ClearSampleState;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.users.ModelUserTable.ClientListener;
import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.Point2D;
import gov.aps.jca.dbr.DBR;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;

// This class handles raster state when in the view images mode, which is now the default
//
// View mode updates happen this way:
// 1) Either a run state change or a user click initiates a change in the view mode by
//    calling VCViewImages.setMode(reqMode, true)
// 2) VCViewImages.mode is updated
// 3) The view mode widget is updated, and this update does not cause any other effects
// 4) The EPICS PV is updated
// 5) All clients receive the update and call VCViewImages.setMode(reqMode, false)
// 6) VCViewImages.mode is updated in slave clients, since the master client will return once
//    it sees the new mode equals the old mode
// 7) The view mode widget is updated in slave clients for the same reason as in step 6
// 8) No PV update occurs here since the second argument to the method was false
//
// Note that step 1 requires the client to be master in either case
//

public class VCViewImages
{
	private int selectedCellIndex = -2; // -2 means it hasn't been updated from EPICS yet
	public int getSelectedCellIndex() { return selectedCellIndex; }
	
	private ModelRasterCell selectedCell;
	private Point lastRightClick;
	protected ModelRasterVideo model;
	public DrawViewImages draw;
	private Point2D panPixels = new Point2D(0,0);
	private double zoomFactor = 1.0;
	private int currentRunIndex=-1;
	private Boolean currentModeIsAuto=null, currentRunIsInactive=null;
	private ViewMode mode = null;
	public static enum ViewMode { MOVE, VIEW };
	
	public ListenerList<VCViewImages> llSelectedCell, llViewMode;
	
	public void setMode(ViewMode mode, boolean sendUpdate)
	{
		if (this.mode==mode) return;
		
		setSelectedCell(null);
		this.mode=mode;
		llViewMode.notifyListeners(this.mode);
		if (sendUpdate) Epics.putInteger(PV_R_RUNS.viewMode, mode.ordinal());
	}
	public ViewMode getMode() { return mode; }
	
	public VCViewImages(final ModelRasterVideo model)
	{
		this.model=model;
		llSelectedCell = new ListenerList<VCViewImages>(this);
		llViewMode = new ListenerList<VCViewImages>(this);
		draw = new DrawViewImages(model);
	}
	
	public void connectListeners()
	{
		Epics.addMonitor(PV_R_RUNS.viewMode, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				int intVal = Epics.dbrToInt(value);
				setMode(ViewMode.values()[intVal], false);
			}
		});
		
		Epics.addMonitor(PV_R_RUNS.selectedCell, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				int intVal = Epics.dbrToInt(value);
				setSelectedCell(intVal);
			}
		});
		setSelectedCell(Epics.getInteger(PV_R_RUNS.selectedCell));
		
		ClearSampleState.llBeforeEachClear.addListener(new ListenerList.Listener<ClearSampleState>()
		{
			@Override public void handleEvent(ClearSampleState object, Object data)
			{
				// if the user is viewing images during a clear, we need to deselect the cell
				// since the grid is being deleted
				if (model.getState()==VideoState.VIEW_IMAGES)
				{
					setSelectedCell(null);
				}
			}
		});
		
		ModelRasterState.get().llCurrentRun.addListener(new ListenerList.Listener<ModelRasterState>()
		{
			@Override public void handleEvent(ModelRasterState state, Object data)
			{
				ModelRasterRun currentRun = state.getCurrentRunModel();
				int newCurrentRunIndex = currentRun.getRunIndex();
				boolean isSONICC = currentRun.getRunType() == ModelRasterRun.RunType.SONICC;
				boolean newCurrentRunIsInactive = currentRun.getRunState()==RunState.INACTIVE;
				if (!isSONICC && (newCurrentRunIndex!=currentRunIndex || newCurrentRunIsInactive != currentRunIsInactive))
				{
					// run changed
					setSelectedCell(null);
					if (Client.isMaster())
					{
						setMode(newCurrentRunIsInactive?ViewMode.MOVE:ViewMode.VIEW, true);
					}
					
					currentRunIsInactive=newCurrentRunIsInactive;
					currentRunIndex=newCurrentRunIndex;
				}
			}
		});
		try
		{
			ModelRasterRun currentRun = ModelRasterState.get().getCurrentRunModel();
			boolean isInactive = currentRun.getRunState()==RunState.INACTIVE;
			setMode(isInactive?ViewMode.MOVE:ViewMode.VIEW, false);
		}
		catch (Exception e) { e.printStackTrace(); }
		
		model.llStateChange.addListener(new ListenerList.Listener<ModelRasterVideo>()
		{
			@Override public void handleEvent(final ModelRasterVideo vc, Object data)
			{
				GUI.asyncExec(new Runnable(){public void run()
				{
					handleModelChange(vc);
				}});
			}
		});
		
		Client.addListener(new ClientListener()
		{
			@Override public void masterChanged(boolean isMaster)
			{
				handleMasterSlave(isMaster);
			}
		});
	}
	
	private boolean inMode = false;
	private void handleModelChange(ModelRasterVideo mrv)
	{
		boolean enteringViewMode = mrv.getState()==VideoState.VIEW_IMAGES;
		if (enteringViewMode && !inMode)
		{
			Log.note("Entering image viewing mode");
			
			// pass along the menu right-click to potentially select the first cell to view
			if (null!=lastRightClick) onClick(lastRightClick.x, lastRightClick.y, 1);
		}
		if (!enteringViewMode && inMode)
		{
			Log.note("Leaving image viewing mode");
			setSelectedCell(null);
		}
		inMode = enteringViewMode;
	}
	
	public Point2D applyTransforms(Point2D screenPt)
	{
		Point2D ret = new Point2D(screenPt);
		
		ret.x -= model.cam.displayWidth/2.0;
		ret.y -= model.cam.displayHeight/2.0;
		
		ret.x *= zoomFactor;
		ret.y *= zoomFactor;
		
		ret.x += model.cam.displayWidth/2.0;
		ret.y += model.cam.displayHeight/2.0;
		
		return ret;
	}
	
	private void handleMasterSlave(boolean isMaster)
	{
		if (!isMaster && model.getState()==VideoState.VIEW_IMAGES)
		{
			Log.note("Client became slave, deselecting cell.");
			setSelectedCell(null);
		}
	}
	
	public void onDrag(int mouseX, int mouseY, int mouseButton)
	{
		onClick(mouseX, mouseY, mouseButton);
	}
	
	public void onClick(int mouseX, int mouseY, int mouseButton)
	{
		if (null==mode) return;
		
		if (!Client.isMaster())
		{
			Log.note("Left clicks ignored while client is not master");
			return;
		}
		
		if (mode==ViewMode.MOVE)
		{
			model.cam.pointAndClick(mouseX+model.cam.drawX, mouseY+model.cam.drawY);
			return;
		}
		
		try
		{
			ModelRasterRun currentRun = ModelRasterState.get().getCurrentRunModel();
			if (null == currentRun) return;
			ModelRasterGrid currentGrid = currentRun.grid();
			
		    for (ModelRasterCell cell : currentGrid.getGridCalc().getCellListCopy())
		    {
		    	Rectangle cellBounds = model.getCellScreenBounds(cell);
		    	if (null==cellBounds) continue;
		    	if (cellBounds.contains(mouseX,mouseY))
		    	{
		    		if (cell.index != selectedCellIndex)
			    	{
		    			setSelectedCell(cell);
			    	}
		    		return;
		    	}
		    }
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}

		// click was outside the grid
		setSelectedCell(null);
	}
	
	public void onScroll(int direction)
	{
		//zoomFactor *= direction==SWT.UP ? 1.7 : (1.0/1.7);
	}
	
	public void onKeyPress(int keyCode)
	{
		if (null == selectedCell)
		{
			Log.warn("No cell selected, please select one before using arrow keys");
			return;
		}
		
		ModelRasterRun currentRun = ModelRasterState.get().getCurrentRunModel();
		if (null == currentRun) return;
		ModelRasterGrid currentGrid = currentRun.grid();
		if (null == currentGrid) return;
		
		ModelRasterCell newCell = null;
		switch (keyCode)
		{
		case SWT.ARROW_LEFT:
		{
			int newCol = selectedCell.getCol()-1;
			newCell = currentGrid.getCell(selectedCell.getRow(), newCol);
			break;
		}
		case SWT.ARROW_RIGHT:
		{
			int newCol = selectedCell.getCol()+1;
			newCell = currentGrid.getCell(selectedCell.getRow(), newCol);
			break;
		}
		case SWT.ARROW_UP:
		{
			int newRow = selectedCell.getRow()-1;
			newCell = currentGrid.getCell(newRow, selectedCell.getCol());
			break;
		}
		case SWT.ARROW_DOWN:
		{
			int newRow = selectedCell.getRow()+1;
			newCell = currentGrid.getCell(newRow, selectedCell.getCol());
			break;
		}
		case SWT.ESC:
			setSelectedCell(null);
			break;
		default:
			break;
		}
		
		if (null != newCell) setSelectedCell(newCell);
	}
	
	public void centerSelectedCell()
	{
		if (!Client.isMaster())
		{
			Log.error("Cell centering is not allowed when slave");
			return;
		}
		
		if (!AppState.isOperationAllowed(Operation.MOVE_SAMPLE))
		{
			Log.error("Cell centering is not allowed while other operations are in progress");
			return;
		}

		// load cell coordinates
		if (null == selectedCell) { Log.note("No cell selected"); return; }
		double[] xyz = selectedCell.getMotorXYZ();
		
		// move xyz and omega
		DevXYZ.syncMove(new Point3D(xyz), false);
		ModelRasterRun run = ModelRasterState.get().getCurrentRunModel();
		DevPmacMotor.get("gonio_omega").startMove(run.grid().getCenter().omega);
	}
	
	public void setSelectedCell(int cellIndex)
	{
		try
		{
			ModelRasterRun currentRun = ModelRasterState.get().getCurrentRunModel();
			if (null == currentRun) return;
			ModelRasterGrid currentGrid = currentRun.grid();
			ModelRasterCell currentCell = -1==cellIndex?null:currentGrid.getGridCalc().getCellListCopy().get(cellIndex);
		    setSelectedCell(currentCell);
		}
		catch (Exception e) { e.printStackTrace(); }
	}
	
	public void setSelectedCell(ModelRasterCell cell)
	{
		// no change, return (happens when callback comes back through EPICS)
		int newIndex = cell==null ? -1 : cell.index;
		if (newIndex==selectedCellIndex) return;
		
		// update EPICS
		if (Client.isMaster())
			Epics.putInteger(PV_R_RUNS.selectedCell, newIndex);

		// update local variables
		selectedCellIndex = newIndex;
		selectedCell = cell;

		// notify listeners
		if (null!=selectedCell) selectedCell.loadImage();
		llSelectedCell.notifyListeners(cell);
		GUI.asyncExec(new Runnable(){public void run()
		{
			model.cam.redraw();
		}});
	}
	
	public void saveRightClickPosition(int x, int y)
	{
		lastRightClick = new Point(x,y);
	}
}
