package gov.anl.gmca.raster.sonicc;

import gov.anl.gmca.common.Redis;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup;
import gov.anl.gmca.raster.ModelContour;
import gov.anl.gmca.raster.ModelRasterRun;
import gov.anl.gmca.raster.calc.GridCalculator;
import gov.anl.gmca.raster.calc.RasterGeometry;
import gov.anl.gmca.raster.sonicc.SONICCImage;
import gov.anl.gmca.sample.DBCamera;
import gov.anl.gmca.sample.DevCameraSettings;
import gov.anl.gmca.sample.WdgCamera;
import gov.anl.gmca.spotfinder.DBImageStats;
import gov.anl.gmca.util.Point2D;

import java.awt.*;
import java.awt.image.BufferedImage;
import java.util.ArrayList;

public class CmdBinSONICC
{
    private ModelRasterRun run;
    private Point2D binSize_um;

    public CmdBinSONICC(ModelRasterRun run, Point2D binSize_um)
    {
        this.run = run; this.binSize_um = binSize_um;
    }

    public BufferedImage bin(SONICCImage img)
    {
        double binW_um=binSize_um.x, binH_um=binSize_um.y;
        double width_um=img.cropped_size_um.x, height_um=img.cropped_size_um.y;

        int tgtW = (int)Math.ceil(width_um/binW_um);
        int tgtH = (int)Math.ceil(height_um/binH_um);
        double scaleX = (double)tgtW / (double)img.transformedImg.getWidth();
        double scaleY = (double)tgtH / (double)img.transformedImg.getHeight();

        BufferedImage binned = new BufferedImage(tgtW, tgtH, BufferedImage.TYPE_BYTE_GRAY);
        Graphics2D g2 = binned.createGraphics();
        g2.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC);
        g2.scale(scaleX, scaleY);
        g2.drawImage(img.transformedImg, null, 0, 0);
        g2.dispose();

        return binned;
    }

    public float[] binnedImageToFloatArray(BufferedImage binnedImage)
    {
        int w = binnedImage.getWidth(), h = binnedImage.getHeight();
        float[] ret = new float[w * h];
        for (int x = 0; x < w; x++) {
            for (int y = 0; y < h; y++) {
                int val = binnedImage.getRGB(x, y)&0xFF;
                ret[y * w + (w-x-1)] = val;
            }
        }
        return ret;
    }

    public ModelContour generateContour(Point imgSize, Point2D cellSize, Point2D topLeft_um)
    {
        double gridWidth = imgSize.x*cellSize.x;
        double gridHeight = imgSize.y*cellSize.y;

        double gridXLeft = topLeft_um.x;
        double gridXRight = topLeft_um.x + gridWidth;
        double gridYTop = topLeft_um.y;
        double gridYBottom = topLeft_um.y + gridHeight;

        ModelContour ret = new ModelContour();
        ret.points.add(new Point2D(gridXLeft, gridYTop));
        ret.points.add(new Point2D(gridXRight, gridYTop));
        ret.points.add(new Point2D(gridXRight, gridYBottom));
        ret.points.add(new Point2D(gridXLeft, gridYBottom));
        return ret;
    }

    public void start()
    {
        // get sonicc image, transform, and bin
        SONICCImage img = new SONICCImage("A");
        img.loadFromDisk();
        img.transform(run.getSONICCRefPos(), true, true);
        BufferedImage binnedImg = bin(img);

        // save to redis
        float[] values = binnedImageToFloatArray(binnedImg);
        DBImageStats.save(Redis.image_stats(run.getDirectorySuffix(), run.getFileRoot(), run.getRunLabel(), "sonicc"), values);

        //////////// find top left in microns from current HR center /////////

        // HR img size pixels
        org.eclipse.swt.graphics.Point hr_size_px = DBCamera.getCameraDimensions(WdgCamera.CameraType.HIGHRES, WdgCamera.CameraSize.LARGE);

        //  HR img size um
        double px2mm_h = DevCameraSettings.getPx2mm_h_hi(), px2mm_v = DevCameraSettings.getPx2mm_v_hi();
        Point2D hr_size_um = new Point2D(px2mm_h*hr_size_px.x*1000.0, px2mm_v*hr_size_px.y*1000.0);

        // sonicc img top left from HR top left
        Point2D imgTopLeft_um = new Point2D(img.hrTopLeft.x*px2mm_h*1000.0, img.hrTopLeft.y*px2mm_v*1000.0);

        // sonicc img top left from HR center
        Point2D topLeft_um = new Point2D(-1*hr_size_um.x/2.0+imgTopLeft_um.x, -1*hr_size_um.y/2.0+imgTopLeft_um.y);

        // set up grid
        Point2D cellSize = binSize_um;
        ModelContour contour = generateContour(new Point(binnedImg.getWidth(), binnedImg.getHeight()), cellSize, topLeft_um);
        ArrayList<ModelContour> contourList = new ArrayList<>();
        contourList.add(contour);
        run.dev().setCenter(RasterGeometry.getCurrentMotorPositions(), false);
        run.dev().setContours(contourList, false);
        run.dev().setCellHeight_um(cellSize.x, false);
        run.dev().setCellWidth_um(cellSize.y, false);
        run.grid().getGridCalc().fitGridToContour(false, false);

        // set bin flag
        run.dev().setSONICCBin(true, true);
    }

    public static void unbin(ModelRasterRun run, boolean update)
    {
        DBImageStats.delete(run.getDirectorySuffix(), run.getFileRoot(), run.getRunLabel(), "sonicc");

        ArrayList<ModelContour> emptyContours = new ArrayList<>();
        run.dev().setContours(emptyContours, false);
        run.dev().setSONICCBin(false, update);
    }
}
