package gov.anl.gmca.raster.util;

import gov.anl.gmca.io.Database;
import gov.anl.gmca.io.Database.DBName;
import gov.anl.gmca.io.Database.WhereList;
import gov.anl.gmca.io.Database.WhereOp;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PV_R_RUNS;
import gov.anl.gmca.raster.util.ModelMacroList.Macro;
import gov.anl.gmca.syslog.Log;
import gov.aps.jca.dbr.DBR;

import java.util.ArrayList;
import java.util.HashMap;

public class DevMacroList
{
	public static final String SQL_TABLE_NAME = "Macros";
	
	protected ModelMacroList list;
	private static DevMacroList inst;
	public static DevMacroList get() { return inst; } 
	
	public DevMacroList(final ModelMacroList list)
	{
		inst = this;
		this.list = list;
		
		verifyTableExists();
		
		Epics.addMonitor(PV_R_RUNS.macroUpdated, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				if (-2 == Epics.dbrToInt(value))
				{
					readMacros();
				}
			}
		});
		
		Epics.addMonitor(PV_R_RUNS.currentMacro, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				String newMacro = Epics.dbrToString(value);
				
				System.out.println("Got new macro: "+newMacro);
				
				if (!list.getCurrentMacro().equals(newMacro))
				{
					list.setCurrentMacro(newMacro);
					list.ll.notifyListeners();
				}
			}
		});
		
		readMacros();
	}
	
	protected static boolean doesTableExist()
	{
		Database.Result result = Database.exec(DBName.USERDATA, "show tables like '"+SQL_TABLE_NAME+"'");
		boolean exists = (result.numRows()>0);
		return exists;
	}
	
	public static synchronized boolean verifyTableExists()
	{
		if (!doesTableExist())
		{
			String createCmd =
				"CREATE TABLE `"+SQL_TABLE_NAME+"` ("+
				"`ID`					varchar(100) 	NOT NULL,"+
    			"`rasterMode` 			smallint(6) 	NOT NULL,"+
    			"`prefix` 				varchar(50)		NOT NULL,"+
    			"`dir` 					varchar(300) 	NOT NULL,"+
    			"`beamWidth` 			double 			NOT NULL,"+
    			"`beamHeight` 			double 			NOT NULL,"+
    			"`cellWidth` 			double			NOT NULL,"+
    			"`cellHeight` 			double			NOT NULL,"+
    			"`exposureTime` 		double			NOT NULL,"+
    			"`delta` 				double 			NOT NULL,"+
    			"`ccdDistance` 			double			NOT NULL,"+
    			"`attenuation` 			double 			NOT NULL,"+
    			"`energy` 				double 			NOT NULL,"+
    			"`ROI` 					smallint(6) 	NOT NULL);";
			
			Database.exec(DBName.USERDATA, createCmd);
		}
		
		return doesTableExist();
	}
	
	public void readMacros()
	{
		verifyTableExists();
		
		Database.Result macroTable = Database.exec(DBName.USERDATA, "SELECT * FROM "+DevMacroList.SQL_TABLE_NAME);
		
		ArrayList<Macro> macroList = new ArrayList<Macro>();
		for (int i=0; i < macroTable.numRows(); i++)
		{
			HashMap<String,String> row = macroTable.getRow(i);
			Macro macro = new Macro();
			
			try
			{
				macro.ID				= row.get("ID");
    			macro.prefix			= row.get("prefix");
    			macro.dir				= row.get("dir");
    			macro.rasterMode		= Integer.parseInt(row.get("rasterMode"));
    			macro.ROI				= Integer.parseInt(row.get("ROI"));
    			macro.beamWidth			= Double.parseDouble(row.get("beamWidth"));
    			macro.beamHeight		= Double.parseDouble(row.get("beamHeight"));
    			macro.cellWidth			= Double.parseDouble(row.get("cellWidth"));
    			macro.cellHeight		= Double.parseDouble(row.get("cellHeight"));
    			macro.exposureTime		= Double.parseDouble(row.get("exposureTime"));
    			macro.delta				= Double.parseDouble(row.get("delta"));
    			macro.ccdDistance		= Double.parseDouble(row.get("ccdDistance"));
    			macro.attenuation		= Double.parseDouble(row.get("attenuation"));
    			macro.energy			= Double.parseDouble(row.get("energy"));
			}
			catch (NumberFormatException nfe)
			{
				Log.error("Error reading macro "+macro.prefix+": "+nfe.toString());
				continue;
			}
			macroList.add(macro);
		}
		
		for (Macro macro : macroList)
		{
			list.macroMap.put(macro.ID, macro);
		}
		list.ll.notifyListeners();
	}
	
	public void addMacro(Macro macro)
	{
		verifyTableExists();
		
		list.macroMap.put(macro.ID,macro);
		list.setCurrentMacro(macro.ID);
		list.ll.notifyListeners();
		
		String stmt = 
			"INSERT INTO "+DevMacroList.SQL_TABLE_NAME+" VALUES ("+
			"'"+macro.ID+"',"+
			macro.rasterMode+","+
			"'"+macro.prefix+"',"+
			"'"+macro.dir+"',"+
			macro.beamWidth+","+
			macro.beamHeight+","+
			macro.cellWidth+","+
			macro.cellHeight+","+
			macro.exposureTime+","+
			macro.delta+","+
			macro.ccdDistance+","+
			macro.attenuation+","+
			macro.energy+","+
			macro.ROI+
			");";
		
		System.out.println(stmt);
		Database.exec(DBName.USERDATA, stmt);
		setCurrentMacro(macro.ID);
		sendUpdate();
	}
	
	public void removeMacro(String macroID)
	{
		verifyTableExists();

		// pick an adjacent macro to be current
		String adjacentMacroID = list.getAdjacentMacroID(macroID);
		Epics.put(PV_R_RUNS.currentMacro, adjacentMacroID);
		
		// local update
		list.macroMap.remove(macroID);
		
		// delete the macro
		WhereList whereList = new WhereList();
		whereList.add("ID", WhereOp.EQUALS, macroID);
		Database.delete(DBName.USERDATA, DevMacroList.SQL_TABLE_NAME, whereList);
		
		sendUpdate();
	}
	
	public void sendUpdate()
	{
		Epics.putInteger(PV_R_RUNS.macroUpdated, -1);
		Epics.putInteger(PV_R_RUNS.macroUpdated, -2);
	}

	public void setCurrentMacro(String macroName)
	{
		Epics.put(PV_R_RUNS.currentMacro, macroName);
	}
}
