package gov.anl.gmca.raster.util;

import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.hutch.DevAPSStatus;
import gov.anl.gmca.hutch.DevAttenuation;
import gov.anl.gmca.hutch.DevEnergy;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.io.Database;
import gov.anl.gmca.io.Database.DBName;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PV_MISC;
import gov.anl.gmca.log.TabLog;
import gov.anl.gmca.raster.ModelRasterRun;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.util.Util;
import org.apache.poi.hssf.usermodel.HSSFRow;
import org.apache.poi.hssf.usermodel.HSSFSheet;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;

public class RasterLog
{
	private static final String RASTER_LOG_TABLE = "RasterParams";
	
	//////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////// SQL Log ///////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////

	private static void createTable()
	{
		Database.exec
		(
			DBName.GMCADATA,
			"CREATE TABLE IF NOT EXISTS `RasterParamsDatabase` (" +
			"`sampleNumber` INT(11) NULL ," + 
			"`rasterNumber` INT(11) NOT NULL ," + 
			"`rasterType` VARCHAR(30) NULL ," + 
			"`searchType` VARCHAR(45) NULL ," + 
			"`referenceRaster` VARCHAR(45) NULL ," + 
			"`prefix` VARCHAR(512) NULL ," + 
			"`workdir` VARCHAR(512) NULL ," + 
			"`noofcells` VARCHAR(45) NULL ," + 
			"`cellsize` VARCHAR(30) NULL ," + 
			"`collimator` VARCHAR(30) NULL ," + 
			"`beamsize_um` VARCHAR(50) NULL ," + 
			"`siltWidth` VARCHAR(30) NULL ," + 
			"`slitHeight` VARCHAR(30) NULL ," + 
			"`izero` VARCHAR(30) NULL ," + 
			"`apsCurrent` VARCHAR(30) NULL ," + 
			"`dcm` VARCHAR(30) NULL ," + 
			"`attenuation` VARCHAR(30) NULL ," + 
			"`finalEnergy` VARCHAR(30) NULL ," + 
			"`crystalX` VARCHAR(30) NULL ," + 
			"`crystalY` VARCHAR(30) NULL ," + 
			"`crystalZ` VARCHAR(30) NULL ," + 
			"`detectorDistance` VARCHAR(30) NULL ," + 
			"`twotheta` VARCHAR(30) NULL ," + 
			"`beamstop` VARCHAR(30) NULL ," + 
			"`omegaStart` VARCHAR(30) NULL ," + 
			"`frameWidth` VARCHAR(30) NULL ," + 
			"`exposureTime` VARCHAR(30) NULL ," + 
			"`energy1` VARCHAR(30) NULL ," + 
			"`beamX` VARCHAR(30) NULL ," + 
			"`beamY` VARCHAR(30) NULL ," + 
			"`sizeX` VARCHAR(30) NULL ," + 
			"`sizeY` VARCHAR(30) NULL ," + 
			"`sizeZ` VARCHAR(30) NULL ," + 
			"`resolutionLimits` VARCHAR(45) NULL ," + 
			"`pixelSize` VARCHAR(45) NULL ," + 
			"`ROI_type` VARCHAR(45) NULL ," + 
			"`Autofind` VARCHAR(45) NULL ," + 
			"`Loopcenter` VARCHAR(45) NULL ," + 
			"`Autohits` VARCHAR(45) NULL ," + 
			"`maxReflections` VARCHAR(45) NULL ," + 
			"`timestamp` TIMESTAMP NULL ," + 
			"PRIMARY KEY (`rasterNumber`) ," + 
			"CONSTRAINT `sampleNumber`"
		); 
	}
	
	public static void writeToSQL(ModelRasterRun run)
	{
		HashMap<String, String> row = new HashMap<String, String>();
		
		row.put("sampleNumber","");
		row.put("rasterNumber","");
		row.put("rasterType",run.getRunList().getListType().toString()+"-"+run.getRunType().toString());
		row.put("prefix",run.getFileRoot());
		row.put("workdir",run.getFullDirectory());
		row.put("cellsize_um", run.grid().getCellWidth_um()+"x"+run.grid().getCellHeight_um());
		row.put("collimator", run.getBeamsizeSetup().collimatorSize.toString());
		row.put("beamsize_um", run.getBeamsizeSetup().guardslitWidth+"x"+run.getBeamsizeSetup().guardslitHeight);
		
		ArrayList<HashMap<String, String>> rows = new ArrayList<HashMap<String, String>>();
		rows.add(row);
		Database.insert(DBName.GMCADATA, RASTER_LOG_TABLE, rows);
	}
	
	//////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////
	
	private static void writeRasterParams(ModelRasterRun run, HSSFSheet paramSheet)
	{
		DateFormat dateFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
    	Date date = new Date();
    	String dateStr = dateFormat.format(date);
		
		int rowCounter=0;
		writeParam(paramSheet, rowCounter++, "DATE", 				dateStr);
		writeParam(paramSheet, rowCounter++, "USER_NAME", 			Client.getUser());
		writeParam(paramSheet, rowCounter++, "PROJECT_NAME", 		TabLog.projectName);
		writeParam(paramSheet, rowCounter++, "BEAMLINE", 			DBConfig.get("app_beamlineName"));
		writeParam(paramSheet, rowCounter++, "", 					"");
		writeParam(paramSheet, rowCounter++, "SAMPLE", 				TabLog.sampleName);
		writeParam(paramSheet, rowCounter++, "SAMPLE_SIZE", 		TabLog.sampleXSize+"X"+TabLog.sampleYSize+"X"+TabLog.sampleZSize);
		writeParam(paramSheet, rowCounter++, "", 					"");
		writeParam(paramSheet, rowCounter++, "APS_CURRENT", 		DevAPSStatus.getCurrent(), 1);
		writeParam(paramSheet, rowCounter++, "DCM", 				Epics.getDouble(PV_MISC.dcm), 4);
		writeParam(paramSheet, rowCounter++, "ATTENUATION", 		DevAttenuation.get().getFactors(), 1);
		writeParam(paramSheet, rowCounter++, "",					"");
		writeParam(paramSheet, rowCounter++, "MONO_ENERGY", 		DevEnergy.getDevice().getPosition(), 4);
		writeParam(paramSheet, rowCounter++, "", 					"");
		writeParam(paramSheet, rowCounter++, "BEAM_WIDTH", 			run.getBeamWidth_um(), 1);
		writeParam(paramSheet, rowCounter++, "BEAM_HEIGHT",			run.getBeamHeight_um(), 1);
		writeParam(paramSheet, rowCounter++, "", 					"");
		writeParam(paramSheet, rowCounter++, "DETECTOR_DISTANCE", 	run.getDetectorDist(), 4);
		writeParam(paramSheet, rowCounter++, "2THETA", 				"0.0");
		writeParam(paramSheet, rowCounter++, "BEAMSTOP_DISTANCE", 	DevPmacMotor.get("beamstop_z").getPosition(), 4);
		writeParam(paramSheet, rowCounter++, "", 					"");
		writeParam(paramSheet, rowCounter++, "OMEGA", 				DevPmacMotor.get("gonio_omega").getPosition(), 4);
		writeParam(paramSheet, rowCounter++, "FRAME_WIDTH", 		run.getDelta(), 4);
		writeParam(paramSheet, rowCounter++, "EXPOSURE_TIME", 		run.getExposureTime(), 2);
		writeParam(paramSheet, rowCounter++, "", 					"");
		writeParam(paramSheet, rowCounter++, "DIRECTORY", 			run.getFullDirectory());
		writeParam(paramSheet, rowCounter++, "", 					"");
		writeParam(paramSheet, rowCounter++, "",					"");
		writeParam(paramSheet, rowCounter++, "COMMENTS", 			TabLog.comments);
	}
	
	private static void writeParam(HSSFSheet sheet, int row, String param, String value)
	{
		HSSFRow headerRow = sheet.createRow(row);
		headerRow.createCell(0).setCellValue(param);
		headerRow.createCell(1).setCellValue(value);
	}
	
	private static void writeParam(HSSFSheet sheet, int row, String param, Double value, int decimals)
	{
		writeParam(sheet, row, param, Util.printDouble(value, decimals));
	}
}
