package gov.anl.gmca.sample;

import com.mashape.unirest.http.*;
import gov.anl.gmca.collect.ModelCollectRun.Point3D;
import gov.anl.gmca.common.AppState;
import gov.anl.gmca.common.AppState.Operation;
import gov.anl.gmca.common.ui.Tab;
import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.io.Disk;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PV_BLUICE;
import gov.anl.gmca.io.Web;
import gov.anl.gmca.io.Web.WebConnection;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.users.ModelUserTable.ClientListener;
import gov.anl.gmca.util.Point2D;
import gov.anl.gmca.util.Point4D;
import gov.anl.gmca.util.SwingUtil;
import gov.anl.gmca.util.Util;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.*;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Composite;

import javax.imageio.ImageIO;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.concurrent.atomic.AtomicBoolean;

public class WdgCamera extends Canvas
implements PaintListener, MouseListener, Tab.VisibilityListener
{
	protected final static int IMAGE_LOAD_DELAY_MS 			= 200;
	protected final static int IMAGE_LOAD_MIN_DELAY_MS		= 200;
	protected final static int IMAGE_DRAW_DELAY_MS 			= 200;
	protected final static double SLEEP_COEFFICIENT 		= 1.5;
	protected final static int MS_REQUIRED_BETWEEN_CLICKS 	= 1000;
	
	public enum 					CameraType 				{ LOWRES, HIGHRES, HUTCH, APSSTATUS };
	public enum 					CameraSize 				{ LARGE, SMALL };

	// camera config
	//
	protected String 				urlString				= null;
	protected CameraType			camType					= null;
	protected CameraSize			camSize					= null;
	
	// image parameters
	//
	public    double				ratio					= 1.0;
	public	  double				getRatio()				{ return ratio; }
	public    int 					displayWidth			= 1;
	public    int 					displayHeight			= 1;
	protected int 					imgWidth				= 0;
	protected int 					imgHeight				= 0;
	protected BufferedImage			imgFront 				= null;
	protected BufferedImage			imgBack 				= null;
	public 		int					drawX					= 0; // save where the image was last drawn, so we
	public		int					drawY					= 0; // can figure out where on the image was clicked

	// used by drawing thread
	//
	protected AtomicBoolean 		bNewImage 					= new AtomicBoolean(false);
	protected AtomicBoolean 		bLoadImages 				= new AtomicBoolean(false);
	protected Long					nextImageLoadTime			= 0L;
	protected Long					nextImageDrawTime			= 0L;
	protected Long					currentImageDrawStartTime	= null;
	protected Long					currentImageLoadStartTime	= null;
	
	// other
	//
	protected long					lastClickMs 			= 0;
	protected long					lastRedrawMs 			= 0;
	protected static int			threeClickStep 			= 0;
	protected WdgBeamsizeBox		beamsizeBox				= null;
	protected String 				statusMsg 				= "Loading image...";
	protected Point4D				currentImageMotorPos	= new Point4D(0,0,0,0); // save the sample position for each image so the grid is drawn at the position the sample was in the image instead of the current position
	public	  Point4D				getImageMotorPos()		{ return currentImageMotorPos; }
	protected Point2D				currentImagePx2mm		= new Point2D(1,1);
	public	  Point2D				getImagePx2mm()			{ return currentImagePx2mm; }
	
	// accessors
	//
	public CameraType 				getCamera() 			{ return camType; }
	public int 						getImgWidth() 			{ return imgWidth; }
	public int 						getImgHeight() 			{ return imgHeight; }
	
	// static list of Cameras for use by the CameraThread
	//
	public static ArrayList<WdgCamera> cameraList = new ArrayList<WdgCamera>();
	
	public WdgCamera(Composite parent, int style, String _urlString, CameraType camType, CameraSize camSize)
	{
		super(parent,style|SWT.NO_BACKGROUND); // NO_BACKGROUND is needed to avoid win32 flicker
		cameraList.add(this);
		urlString = _urlString;
		this.camType = camType;
		this.camSize = camSize;
		setBackground(GUI.BGCOLOR);

		addPaintListener(this);
		addMouseListener(this);
		
		addDisposeListener(new DisposeListener()
		{
			@Override public void widgetDisposed(DisposeEvent e)
			{
				cameraList.remove(this);
			}
		});
		
		Client.addListener(new ClientListener()
		{
			@Override public void masterChanged(boolean isMaster)
			{
				GUI.asyncExec(new Runnable(){public void run()
				{
					redraw();
				}});				
			}
		});
		
		beamsizeBox = new WdgBeamsizeBox();
		/*beamsizeBox.redrawLL.addListener(new ListenerList.Listener<WdgBeamsizeBox>()
		{
			@Override public void handleEvent(WdgBeamsizeBox object, Object data)
			{
			    GUI.asyncExec(new Runnable(){public void run()
			    {
			    	redraw();
			    }});
			}
		});*/
		
		CameraThread.startIfNeeded();
	}

	protected void getImage()
	{
		if (GUI.STOPPING == GUI.appState) return;

		try {
			HttpResponse resp = Unirest.get(urlString).asBinary();
			int status = resp.getStatus();
			if (200 != status) {
				Util.printf("Failed to load URL: %s (returned code %d)", urlString, status);
				setNextImageLoadTime();
				return;
			}
			imgBack = ImageIO.read(resp.getRawBody());
			imgWidth=imgBack.getWidth();
			imgHeight=imgBack.getHeight();
		} catch (Exception e) {
			e.printStackTrace();
			System.out.println("Failed to load URL: "+urlString);
			setNextImageLoadTime();
			return;
		}

		statusMsg = null;
		
		// create image
		
		// flip buffer
		imgFront=imgBack;
		imgBack=null;
		
		// save sample position for grid drawing
		currentImageMotorPos.x=DevPmacMotor.get("sample_x").getPosition();
		currentImageMotorPos.y=DevPmacMotor.get("sample_y").getPosition();
		currentImageMotorPos.z=DevPmacMotor.get("sample_z").getPosition();
		currentImageMotorPos.omega=DevPmacMotor.get("gonio_omega").getPosition();
		currentImagePx2mm.x = getConversion(camType, SWT.HORIZONTAL);
		currentImagePx2mm.y = getConversion(camType, SWT.VERTICAL);
		
		bNewImage.set(true);

		// now that the image is loaded, determine the next image load time based on the current time
		setNextImageLoadTime();
	}

	private void setNextImageLoadTime()
	{
		long now = System.currentTimeMillis();
		long delay = currentImageLoadStartTime + IMAGE_LOAD_DELAY_MS - now;
		if (delay < IMAGE_LOAD_MIN_DELAY_MS) delay = IMAGE_LOAD_MIN_DELAY_MS;
		if (camType==CameraType.APSSTATUS) delay = 10000;
		nextImageLoadTime = now + delay;
	}
	
	public void paintStatusMessage(GC gc)
	{
		String msg = statusMsg;
		
		Point size = getSize();
		Point extent = gc.textExtent(msg);
		
		gc.setBackground(GUI.COLOR_BLACK);
		gc.fillRectangle(0,0,size.x,size.y);
		
		gc.setForeground(GUI.COLOR_WHITE);
    	gc.drawText(msg,size.x/2-extent.x/2,size.y/2-extent.y/2);
	}
	
	// this method, unlike the scan method, is called from the UI thread
	// on return from this method, the file is written
	protected void saveScreenshot(final String imagePath)
	{
		int x = getSize().x, y = getSize().y;
		if (x<640) x=640;
		if (y<480) y=480;
		
		Image screenshot = new Image(GUI.getDisplay(), x, y);
		try
		{
			paintToImage(screenshot);
			
			String dir = imagePath.substring(0,imagePath.lastIndexOf('/'));
			if (!Disk.isDirectory(dir))
			{
				System.out.println("Creating directory for camera screenshot: "+dir);
				Disk.mkdirs(dir);
			}

			ImageLoader loader = new ImageLoader();
			loader.data = new ImageData[]{screenshot.getImageData()};
			loader.save(imagePath, SWT.IMAGE_PNG);
		}
		catch (Exception e) { e.printStackTrace(); }
		finally
		{
			screenshot.dispose();
		}
	}
	
	protected void paintToImage(Image target)
	{
		GC targetGC = new GC(target);
		int targetWidth=target.getBounds().width, targetHeight=target.getBounds().height;
		draw(targetGC, targetWidth, targetHeight);
   		targetGC.dispose();
	}
	
	// This is the point where the Swing->SWT conversion is done
	// A few methods call here and use GCs:
	// - paintControl, paintToImage in this class
	// - draw, drawStandard in WdgRasterCamera
	// After this class it's all Swing (Graphics2D)
	//
	public void draw(GC gc, int targetWidth, int targetHeight)
	{
		if (null==imgFront) return;
		
    	// determine scaling to fit both dimensions to the target size
		int srcWidth=getImgWidth(), srcHeight=getImgHeight();
    	float xScale = (float)targetWidth/(float)srcWidth;
    	float yScale = (float)targetHeight/(float)srcHeight;
    	ratio = Math.min(xScale, yScale);
    	displayWidth = (int)(srcWidth*ratio);
    	displayHeight = (int)(srcHeight*ratio);

		// create the Swing image
		BufferedImage buf = null;
		Graphics2D bufG2 = null;

    	// create a buffer with the scaled source image
    	try
    	{
        	buf = new BufferedImage(displayWidth, displayHeight, BufferedImage.TYPE_4BYTE_ABGR);
    		bufG2 = (Graphics2D)buf.getGraphics();
        	
    		// draw scaled imgFront to the buffer
    		bufG2.setRenderingHint(RenderingHints.KEY_INTERPOLATION,RenderingHints.VALUE_INTERPOLATION_NEAREST_NEIGHBOR);
    		bufG2.drawImage(imgFront, 0, 0, displayWidth, displayHeight, null);
        	
    		/* for testing background colors
    		double d = Epics.getDouble(PV_MISC.highlightInt);
    		int c=(int)(d/5.0*255.0);
    		bufG2.setColor(new java.awt.Color(c,c,c));
    		bufG2.fillRect(0, 0, displayWidth, displayHeight);*/
    		
	    	// draw overlays
	    	drawOverlays(bufG2);
	
	    	// determine draw range
	   		drawX = Math.max(0, (targetWidth - displayWidth)/2);
	   		drawY = Math.max(0, (targetHeight - displayHeight)/2);
	    	
	   		// fill areas outside image
	   		gc.setBackground(GUI.BGCOLOR);
	   		gc.fillRectangle(0, 0, targetWidth, drawY);													// top
	   		gc.fillRectangle(0, drawY+displayHeight, targetWidth, targetHeight-displayHeight-drawY); 	// bottom
	   		gc.fillRectangle(0, 0, drawX, targetHeight); 												// left
	   		gc.fillRectangle(drawX+displayWidth, 0, targetWidth-displayWidth-drawX, targetHeight); 		// right
	    	
	    	// paint frame centered on target image
	   		Image swtImg = null;
	   		try
	   		{
	   			swtImg = SwingUtil.AWTImageToSWTImage(buf);
		   		gc.setBackground(GUI.COLOR_GREEN);
		   		gc.drawImage(swtImg,drawX,drawY);
	   		}
	   		catch (Exception e) { e.printStackTrace(); }
	   		finally
	   		{
	   			if (null != swtImg && !swtImg.isDisposed()) swtImg.dispose();
	   		}
    	}
    	catch (Exception e) { e.printStackTrace(); }
    	finally
    	{
			if (null!=bufG2) bufG2.dispose();
    	}
	}
	
	@Override public void paintControl(PaintEvent e)
	{
		if (isDisposed()) return;
		
		try
		{
    		long startTime = System.currentTimeMillis();
    		
            if (null == statusMsg && imgFront != null && 0!=getImgWidth() && 0!=getImgHeight())
            {
            	int widgetWidth=getSize().x, widgetHeight=getSize().y;
            	draw(e.gc, widgetWidth, widgetHeight);
            }
            else
            {
            	if (null != statusMsg) paintStatusMessage(e.gc);
            }
            
            long now = System.currentTimeMillis();
            lastRedrawMs = now-startTime;
            
            long delay = IMAGE_DRAW_DELAY_MS + (int)(SLEEP_COEFFICIENT*lastRedrawMs);
            nextImageDrawTime = now + delay;
		}
		catch (NullPointerException npe) { npe.printStackTrace(); }
    }
	
	// this is the method to override in subclasses to draw the video
	public void drawOverlays(Graphics2D g2)
	{
   		drawCrosshairAndBorder(g2, displayWidth, displayHeight);
   		beamsizeBox.draw(g2, displayWidth, displayHeight, this);
	}
	
	public void drawCrosshairAndBorder(Graphics2D g2, int width, int height)
	{
    	if (camType == CameraType.LOWRES || camType == CameraType.HIGHRES)
    	{
    		int xmid = width/2;
        	int ymid = height/2;
        	
        	if (AppState.isOperationAllowed(Operation.MOVE_SAMPLE))
        	{
        		// border
            	g2.setColor(GUI.COLOR_CAM_BORDER_AWT);
    	    	g2.setStroke(new BasicStroke(3));
            	g2.drawRect(1,1,width-3,height-3);
        	}

        	// crosshair
        	g2.setColor(GUI.COLOR_CAM_CROSSHAIR_AWT);
	    	g2.setStroke(new BasicStroke(1));
        	g2.drawLine(xmid,0,xmid,height-1);
        	g2.drawLine(0,ymid,width-1,ymid);
    	}
	}
	
	// point and click
	@Override public void mouseDown (MouseEvent event)
	{
		if (event.button==1) pointAndClick(event.x, event.y);
	}
	
	public void pointAndClick(int mouseX, int mouseY)
	{
		// ignore double-clicks
		long now = System.currentTimeMillis();
		long msSinceLastClick = now - lastClickMs;
		if (msSinceLastClick < MS_REQUIRED_BETWEEN_CLICKS)
		{
			//Log.note("Click ignored: Clicks within "+MS_REQUIRED_BETWEEN_CLICKS+"ms are ignored");
			return;
		}
		
		// is sample motion allowed?
    	if (camType != CameraType.HUTCH && camType != CameraType.APSSTATUS && !AppState.isOperationAllowed(Operation.MOVE_SAMPLE))
		{
			Log.note("Click ignored, sample motion is not allowed at this time");
			return;
		}
		
		// in-bounds check (did the user click the gray area outside the image but inside the widget?)
		if (mouseX < drawX || mouseX > (displayWidth+drawX))
		{
			System.out.println("Ignored click "+mouseX+","+mouseY+" because horizontal coordinate is out of bounds "+drawX+" to "+(imgWidth+drawX));
			return;
		}
		if (mouseY < drawY || mouseY > (displayHeight+drawY))
		{
			System.out.println("Ignored click "+mouseX+","+mouseY+" because vertical coordinate is out of bounds "+drawY+" to "+(imgHeight+drawY));
			return;
		}

		// calculate click position
		int moveX = (mouseX - drawX);
		int moveY = (mouseY - drawY);

		if (camType == CameraType.HUTCH)
		{
			// construct URL
			String url=DBCamera.getPointAndClickURL(camType, camSize, moveX, moveY);
			
			// load URL
			Log.note("Sending click "+moveX+","+moveY+" to hutch camera web server");
			System.out.println(url);
			Web.get().openConnection(url);
		}
		else if (camType == CameraType.APSSTATUS)
		{
			Log.note("APS Status click ignored.");
		}
		else
		{
			// construct p+c string
			int videoFlag = (this.camType == CameraType.HIGHRES)?1:0;
			String pointclickStr = videoFlag+" "+ moveX + " "+moveY+" "+displayWidth+" "+displayHeight;
			if (threeClickStep > 0)
			{
				 pointclickStr += " R";
				 threeClickStep++;
			}
	
			// send string to p+c
			Log.note("Sending click "+moveX+","+moveY+" to point and click server");
			Epics.put(PV_BLUICE.pointclick, pointclickStr);
		}
		
		// record time
		lastClickMs = System.currentTimeMillis();
	}
	
	@Override public void notifyIsVisible(boolean isVisible)
	{
		bLoadImages.set(isVisible);
	}
	
	public void mouseDoubleClick(MouseEvent event) {}
	public void mouseUp(MouseEvent event) {}
	
	///////////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////// conversion methods ////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////////
	
	public double mm2pxH(double mm) { return s_mm2pxH(camType, ratio, currentImagePx2mm.x, mm); }
	public double mm2pxV(double mm) { return s_mm2pxV(camType, ratio, currentImagePx2mm.y, mm); }
	public double px2mmH(double px) { return s_px2mmH(camType, ratio, currentImagePx2mm.x, px); }
	public double px2mmV(double px) { return s_px2mmV(camType, ratio, currentImagePx2mm.y, px); }
	
	public static double s_mm2pxH(CameraType camera, double cameraRatio, double conversion, double mm)
	{
		return (mm / conversion) * cameraRatio;
	}
	
	public static double s_mm2pxV(CameraType camera, double cameraRatio, double conversion, double mm)
	{
		return (mm / conversion) * cameraRatio;
	}
	
	public static double s_px2mmH(CameraType camera, double cameraRatio, double conversion, double px)
	{
		return (px * conversion) / cameraRatio;
	}
	
	public static double s_px2mmV(CameraType camera, double cameraRatio, double conversion, double px)
	{
		return (px * conversion) / cameraRatio;
	}

	// Inputs:
	// X: horizontal on high-res camera, positive is right
	// Y: vertical on high-res camera, positive is down
	// Z: positive is in to the screen, negative is toward viewer
	//
	// Motor orientation:
	// sample-x at omega=0: away from high-res camera viewpoint
	// sample-y at omega=0: up on high-res camera
	// sample-z: positive is right on high-res camera
	// omega: positive rotation moves top of sample toward high-res camera
	//
	public static Point4D addXYZToMotorMM(Point4D motorPos, double x_um, double y_um, double z_um)
	{
		double x_mm = x_um / 1000.0;
		double y_mm = y_um / 1000.0;
		double z_mm = z_um / 1000.0;
		
		Point4D ret = new Point4D(motorPos);
		ret.x += -1.0 * Math.sin(Math.toRadians(motorPos.omega)) * y_mm + Math.cos(Math.toRadians(motorPos.omega)) * z_mm;
		ret.y += Math.cos(Math.toRadians(motorPos.omega)) * y_mm + Math.sin(Math.toRadians(motorPos.omega)) * z_mm;
		ret.z += -1.0 * x_mm;
		return ret;
	}
	
	public static double[] screenTopLeftPxToMotorMM(CameraType camera, double omega, double x, double y, double cameraScaleRatio)
	{
		double[] xyz = new double[3];
		
		if (camera == CameraType.LOWRES)
		{
			xyz[0] = -1.0 * Math.sin(Math.toRadians(omega)) * y;
			xyz[1] = Math.cos(Math.toRadians(omega)) * y;
			xyz[2] = -1.0 * x;
		}
		else if (camera == CameraType.HIGHRES)
		{
			xyz[0] = Math.sin(Math.toRadians(omega)) * y;
			xyz[1] = Math.cos(Math.toRadians(omega)) * y;
			xyz[2] = -1.0 * x;
		}
		else
		{
			Log.error("Unsupported camera type:"+camera);
			return null;
		}
		
		return xyz;
	}
	
	// member method for calculating points on this camera instance
	public Point2D motorMMtoCurrentScreenPx(Point4D motorPos)
	{
		Point4D currentPos = getImageMotorPos();

		Point4D relativePos = new Point4D(
			motorPos.x-currentPos.x,
			motorPos.y-currentPos.y,
			motorPos.z-currentPos.z,
			motorPos.omega);
		
		double[] xy = motorMMtoAbsScreen(camType, true, relativePos.toArray(), ratio, currentImagePx2mm);
		xy[0] += displayWidth/2;
		xy[1] += displayHeight/2;
		
		return new Point2D(xy);
	}
	
	// convert from motor coordinates to absolute screen coordinates
	// (X,Y values may be way off screen)
	public static double[] motorMMtoAbsScreen(CameraType camera, boolean scaleToPixels, double[] deviceXYZO, double cameraRatio, Point2D conversion)
	{
		if (null==conversion) conversion = new Point2D
		(
			WdgCamera.getConversion(CameraType.HIGHRES, SWT.HORIZONTAL),
			WdgCamera.getConversion(CameraType.HIGHRES, SWT.VERTICAL)
		);
		
		double[] xy = new double[2];
		xy[0] = -1.0 * deviceXYZO[2];
		
		if (CameraType.LOWRES == camera)
		{
			xy[1] = deviceXYZO[0]*Math.cos(Math.toRadians(deviceXYZO[3])) -
					deviceXYZO[1]*Math.sin(Math.toRadians(deviceXYZO[3]));
		}
		else if (CameraType.HIGHRES == camera)
		{
			xy[1] = deviceXYZO[0]*Math.sin(Math.toRadians(deviceXYZO[3])) +
					deviceXYZO[1]*Math.cos(Math.toRadians(deviceXYZO[3]));
		}
		else { Log.error("Unsupported camera type:"+camera); return null; }
		
		if (scaleToPixels)
		{
			xy[0] = s_mm2pxH(camera, cameraRatio, conversion.x, xy[0]);
			xy[1] = s_mm2pxV(camera, cameraRatio, conversion.y, xy[1]);
		}
		
		return xy;
	}

	///////////////////////////////// convenience methods //////////////////////////////////
	
	public Point motorMMtoScreenIntPx(Point4D motorPos)
	{
		Point2D xy = motorMMtoCurrentScreenPx(motorPos);
		return new Point((int)Math.round(xy.x),(int)Math.round(xy.y));
	}
	
	public Point motorMMtoScreenIntPx(Point3D p)
	{
		Point4D motorPos = new Point4D(p.x, p.y, p.z, DevPmacMotor.get("gonio_omega").getPosition());
		return motorMMtoScreenIntPx(motorPos);
	}
	
	public static double getConversion(CameraType camera, int orientation)
	{
		if (CameraType.LOWRES == camera)
		{
			if (SWT.HORIZONTAL == orientation)	return DevCameraSettings.getPx2mm_h_lo();
			else								return DevCameraSettings.getPx2mm_v_lo();
		}
		else
		{
			if (SWT.HORIZONTAL == orientation)	return DevCameraSettings.getPx2mm_h_hi();
			else								return DevCameraSettings.getPx2mm_v_hi();
		}
	}
	
	public Point2D getCenterPx() { return new Point2D(displayWidth/2.0, displayHeight/2.0); }
}
