package gov.anl.gmca.scan;

import gov.anl.gmca.common.DevMove.MoveCmd;
import gov.anl.gmca.common.DevMove.MoveResult;
import gov.anl.gmca.common.DevMove.ResultType;
import gov.anl.gmca.hutch.DevEnergy;
import gov.anl.gmca.scan.CmdCleanup.Mode;
import gov.anl.gmca.scan.ModelScan.AutoStep;
import gov.anl.gmca.scan.ModelScan.TRACE_ID;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.ActionResult;
import gov.anl.gmca.util.Pair;
import gov.anl.gmca.util.Point2D;
import gov.anl.gmca.util.Util;

import java.util.ArrayList;

public class CmdScanAuto extends CmdScan
{
	protected boolean doPrep;
	CmdScan subCommand = null;
	MoveCmd energyCmd = null;
	
	public CmdScanAuto(boolean doPrep)
	{
		this.doPrep = doPrep;
	}
	
	@Override
	public void run()
	{
		DevScan dev = DevScan.get();
		dev.setGraphTrace(TRACE_ID.edge, new ArrayList<Point2D>(), false);
		dev.setTraceVisible(TRACE_ID.edge, true, false);
		dev.setTraceVisible(TRACE_ID.fPrime, false, false);
		dev.setTraceVisible(TRACE_ID.fDoublePrime, false, false);
		dev.setGraphXMin(0, false);
		dev.setGraphXMax(0, false);
		dev.setDefaultXMin(0, false);
		dev.setDefaultXMax(0, false);
		DevScan.get().setLeftAxisTrace(null, false);
		DevScan.get().setRightAxisTrace(null, false);
		for (TRACE_ID id : TRACE_ID.values())
		{
			dev.setGraphYLabel(id, "", false);
			dev.setGraphYMin(id, 0, false);
			dev.setGraphYMax(id, 0, false);
			dev.setDefaultYMin(id, 0, false);
			dev.setDefaultYMax(id, 0, false);
		}
		dev.setGraphTitle("", false);
		dev.setGraphMessage("Starting Auto Scan...", false);
		dev.setPeak(Double.NaN, false);
		dev.setRemote(Double.NaN, false);
		dev.setInflection(Double.NaN, false);
		dev.setDefaultPeak(Double.NaN, false);
		dev.setDefaultRemote(Double.NaN, false);
		dev.setDefaultInflection(Double.NaN, false);
		dev.setROILow(Double.NaN, false);
		dev.setROIHigh(Double.NaN, false);
		dev.setRightTab(1, false);
		
		if (doPrep)
		{
            // Step one
			DevScan.get().setAutoStep(AutoStep.MOVE_ENERGY, true);

			double selectedEnergy_eV = ModelScan.get().getEnergy();
			double target_keV = ( selectedEnergy_eV / 1000.0 ) + 0.150;

			energyCmd = DevEnergy.getDevice().startMove(target_keV);
			energyCmd.waitFor();
			MoveResult energyResult = energyCmd.getResult();
			if (ResultType.SUCCESS != energyResult.type)
			{
				Log.error("Energy move failed, aborting auto scan. Reason: "+energyResult.msg);
				DevScan.get().setAutoStep(AutoStep.NONE, true);
				return;
			}
			
			if (stopRequested) { endScan("Stop requested, aborting auto scan before step 2."); return; }
			
			// Step two
			DevScan.get().setAutoStep(AutoStep.OPTIMIZE, true);
			
			subCommand = new CmdOptDetector(true);
			subCommand.start();
			subCommand.waitFor();

			// TODO: support getResult() at Command class level
			ActionResult optDetResult = ((CmdOptDetector)subCommand).getResult();
			
			if (!optDetResult.succeeded())
			{
				Log.error("Optimize fluorescence step failed, aborting auto scan. Reason: "+optDetResult.getError());
				DevScan.get().setAutoStep(AutoStep.NONE, true);
				return;
			}
			
			if (stopRequested) { endScan("Stop requested, aborting auto scan before step 3."); return; }
		}
		
		// Step three
		DevScan.get().setAutoStep(AutoStep.SPECTRUM, true);

		Pair<Double,Double> roi = ModelScan.get().setDefaultROI();
		// TODO: eliminate timer
		Util.sleep(3000);
		
		subCommand = new CmdScanSpectrum(true, ModelScan.get().getSpectrumTime());
		subCommand.start();
		subCommand.waitFor();
		ActionResult spectrumResult = ((CmdScanSpectrum)subCommand).getResult();
		if (!spectrumResult.succeeded())
		{
			endScan("Spectrum reported error, aborting auto scan. "+spectrumResult.getError());
			return;
		}
		
		if (stopRequested) { endScan("Stop requested, aborting auto scan before step 4."); return; }

		// TODO: remove timer
		Util.sleep(1000);
		
		// Ask to proceed
		boolean okToContinue = Util.popupYesNo("Continue?","Proceed with edge scan using the preset range shown?");
		
		if (!okToContinue) { endScan("User declined to continue with edge scan, aborting."); return; }

		// Step four
		DevScan.get().setAutoStep(AutoStep.EDGE_SCAN, true);
		
		subCommand = new CmdScanEdge();
		subCommand.start();
		subCommand.waitFor();
		ActionResult edgeResult = ((CmdScanEdge)subCommand).getResult();
		
		if (!edgeResult.succeeded())
		{
			endScan("Edge scan reported error, aborting auto scan. "+edgeResult.getError());
			return;
		}
		
		// TODO: remove timer
		Util.sleep(4000);
		
		// Ask to cleanup
		okToContinue = Util.popupYesNo("Done?", "Are you done with the scan tab?  Click 'yes' to retract the fluorescence detector and restore X-ray beam attenuation");

		if (!okToContinue) { endScan("User declined to run cleanup, leaving auto mode."); return; }

		if (stopRequested) { endScan("Stop requested, aborting auto scan before step 5."); return; }

		// Step five
		DevScan.get().setAutoStep(AutoStep.RETRACT, true);
		
		subCommand = new CmdCleanup(Mode.INTER);
		subCommand.start();
		subCommand.waitFor();
		
		// Done
		DevScan.get().setAutoStep(AutoStep.NONE, true);
	}
	
	protected void endScan(String msg)
	{
		Log.note(msg);
		DevScan.get().setGraphMessage("", false);
		DevScan.get().setAutoStep(AutoStep.NONE, true);
	}
	
	@Override public void requestStop()
	{
		super.requestStop();
		
		if (DevEnergy.getDevice().inLocalMove()) DevEnergy.getDevice().abort();
		if (null != subCommand) subCommand.requestStop();
	}
}
