package gov.anl.gmca.scan;

import gov.anl.gmca.collect.DevFPE;
import gov.anl.gmca.hutch.DevMisc;
import gov.anl.gmca.log.db.ScanEvent;
import gov.anl.gmca.sample.DevSlider;
import gov.anl.gmca.scan.ModelScan.InterStep;
import gov.anl.gmca.scan.ModelScan.TRACE_ID;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.ActionResult;
import gov.anl.gmca.util.ListenerList.Listener;
import gov.anl.gmca.util.Point2D;
import gov.anl.gmca.util.Util;

import java.util.ArrayList;
import java.util.StringTokenizer;

public class CmdScanEdge extends CmdScan
{
	protected 	int 				currentScanPoint	= -1;
	protected 	double 				yMin 				= Double.NaN;
	protected 	double 				yMax 				= Double.NaN;
	protected 	ArrayList<Point2D> 	trace 				= new ArrayList<Point2D>();
	protected 	ActionResult 		result 				= null;
	protected	CmdChooch			choochCmd			= null;
	protected	volatile double		dcmVsum				= 0;
	public 		ActionResult 		getResult() 		{ return result; }
	
	@Override
	public void run()
	{
		DevScan.get().setInterStep(InterStep.EDGE, false);
		DevScan.get().setChoochPDFPath("", false);
		DevScan.get().sendUpdate();

		DevMCA mca = DevMCA.get();
		
		// update MCA ROI
		double evRoiLow = ModelScan.get().getROILow();
		double evRoiHigh = ModelScan.get().getROIHigh();
		if (!Double.isNaN(evRoiLow) && !Double.isNaN(evRoiHigh))
		{
			int gain = DevMCA.get().isFalcon ? DevMCA.get().channelMax : mca.getGain();
			int offset = DevMCA.get().isFalcon ? 0 : mca.getOffset();
		    int chRoiLow = mca.energyToChannel(evRoiLow, gain, offset);
		    int chRoiHigh = mca.energyToChannel(evRoiHigh, gain, offset);
		    
		    if (chRoiLow >= chRoiHigh)
		    {
				String msg = "ROI is not optimized, aborting scan.";
				Util.popupError("ROI error", msg);
				Log.error(msg);
				result = new ActionResult(msg);
				DevScan.get().setInterStep(InterStep.IDLE, true);
				return;
		    }
		
		    mca.setROILow(chRoiLow);
		    mca.setROIHigh(chRoiHigh);
		}
		
		if (!CmdScanSpectrum.verifyCollimator())
		{
			String msg = "Collimator error, aborting scan";
			Util.popupError("ROI error", msg);
			Log.error(msg);
			result = new ActionResult(msg);
			DevScan.get().setInterStep(InterStep.IDLE, true);
			return;
		}
		
		// ringlight on
		DevSlider ringLight = DevSlider.get("ring");
		ringLight.setIntensity(ringLight.getHigherLimit());

		DevScan dev = DevScan.get();
		
		dev.setGraphTrace(TRACE_ID.edge, null, false);
		dev.setTraceVisible(TRACE_ID.edge, true, false);
		dev.setTraceVisible(TRACE_ID.fPrime, false, false);
		dev.setTraceVisible(TRACE_ID.fDoublePrime, false, false);
		DevScan.get().setLeftAxisTrace(TRACE_ID.edge, false);
		DevScan.get().setRightAxisTrace(null, false);
		dev.setGraphXMin(0, false);
		dev.setGraphXMax(0, false);
		for (TRACE_ID id : TRACE_ID.values())
		{
			dev.setGraphYLabel(id, "", false);
			dev.setGraphYMin(id, 0, false);
			dev.setGraphYMax(id, 0, false);
			dev.setDefaultYMin(id, 0, false);
			dev.setDefaultYMax(id, 0, false);
		}
		dev.setROILow(Double.NaN, false);
		dev.setROIHigh(Double.NaN, false);
		dev.setDefaultInflection(Double.NaN, false);
		dev.setDefaultPeak(Double.NaN, false);
		dev.setDefaultRemote(Double.NaN, false);
		dev.setInflection(Double.NaN, false);
		dev.setPeak(Double.NaN, false);
		dev.setRemote(Double.NaN, false);
		dev.setGraphMessage("Starting edge scan...", false);
		dev.setRightTab(1, false);
		dev.setGraphTitle("Edge scan of "+ModelScan.get().getElementEdge().toString(), false);
		dev.setGraphYLabel(TRACE_ID.edge, "Counts", false);
		dev.setGraphXLabel("Energy (eV)", false);
		dev.sendUpdate();
		
		currentScanPoint = 0;
		DevEdgeScan.setData(new double[]{0});
		
		// monitor PV for data from script
		Listener<DevEdgeScan> dataMonitor = new Listener<DevEdgeScan>()
		{
			@Override public void handleEvent(DevEdgeScan object, Object data)
			{
				double[] values = (double[])data;
				handleDataPV(values);
			}
		};
		DevEdgeScan.llData.addListener(dataMonitor);

		// monitor PV for setting range
		Listener<DevFPE> rangeMonitor = new Listener<DevFPE>()
		{
			@Override public void handleEvent(DevFPE object, Object data)
			{
				StringTokenizer st = new StringTokenizer((String)data);
				if (st.hasMoreTokens() && st.nextToken().equals("setGraphXLimit"))
				{
    				double xMin = Double.parseDouble(st.nextToken());
    				double xMax = Double.parseDouble(st.nextToken());
    				
    				DevScan.get().setGraphXMin(xMin, false);
    				DevScan.get().setGraphXMax(xMax, false);
    				DevScan.get().setDefaultXMin(xMin, false);
    				DevScan.get().setDefaultXMax(xMax, true);
    				
    				DevFPE.get().setCmdOutput("");
				}
			}
		};
		DevFPE.get().commandListener.addListener(rangeMonitor);
		
		// run script
		result = DevEdgeScan.run();
		
		// ringlight off
		ringLight.setIntensity(ringLight.getLowerLimit());
		
		// turn off monitors
		DevEdgeScan.llData.removeListener(dataMonitor);
		DevFPE.get().commandListener.removeListener(rangeMonitor);

		ScanEvent.writeEvent(dcmVsum, currentScanPoint);
		
		// reset state to idle
		DevScan.get().setGraphMessage("", false);
		DevScan.get().setInterStep(InterStep.IDLE, true);
	}
	
	protected void handleDataPV(double[] values)
	{
		System.out.println("Scan data PV = "+values[0]+" "+values[1]+" "+values[2]+"...");
		
		// save the highest dcm V sum seen
		double currentDcmVsum = DevMisc.get().getDcmVsum();
		if (currentDcmVsum > dcmVsum) dcmVsum = currentDcmVsum;
		
		if (-1 == currentScanPoint)
		{
			Log.error("currentScanPoint = -1, ignoring data");
			return;
		}
		
		int flag 	= (int)values[0];
		int nPoints = (int)values[1];

		if (0 == flag) return;
		
		if (2 == flag)
		{
			currentScanPoint = 0;
			trace.clear();
		}
		
		while (nPoints > currentScanPoint)
		{
			int energyStartIndex = 2;
			int valueStartIndex = 2+nPoints;
			
			double plotEnergy	= values[energyStartIndex+currentScanPoint];
			double plotValue	= values[valueStartIndex+currentScanPoint]*1000.0;
			
			System.out.println("Read edge scan point "+plotEnergy+","+plotValue);
			
			trace.add(new Point2D(plotEnergy, plotValue));

			if (Double.isNaN(yMin) || yMin > plotValue)
			{
				yMin = plotValue;
				DevScan.get().setGraphYMin(TRACE_ID.edge, yMin, false);
				DevScan.get().setDefaultYMin(TRACE_ID.edge, yMin, false);
			}
			if (Double.isNaN(yMax) || yMax < plotValue)
			{
				yMax = plotValue;
				DevScan.get().setGraphYMax(TRACE_ID.edge, yMax, false);
				DevScan.get().setDefaultYMax(TRACE_ID.edge, yMax, false);
			}
			
			currentScanPoint++;
		}

		DevScan.get().setGraphTrace(TRACE_ID.edge, trace, true);

		if (3 == flag)
		{
			if (null == choochCmd)
			{
				// start chooch thread
				ModelScan.ElementEdge scanEdge = ModelScan.get().getElementEdge();
				choochCmd = new CmdChooch(scanEdge, trace);
				new Thread(choochCmd,"Chooch").start();
			}
			else
			{
				Log.warn("Received finish signal (3) from edge scan script but chooch was already run");
			}
		}
		
		DevEdgeScan.setData(new double[]{0});
	}
	
	@Override
	public void requestStop()
	{
		DevEdgeScan.stop();
		stopRequested = true;
	}
}
