package gov.anl.gmca.scan;

import gov.anl.gmca.collect.CmdCheck;
import gov.anl.gmca.hutch.DevEnergy;
import gov.anl.gmca.hutch.DevShutter;
import gov.anl.gmca.log.db.ScanEvent;
import gov.anl.gmca.sample.DevCollimatorCtrl;
import gov.anl.gmca.sample.DevCollimatorCtrl.CCMove;
import gov.anl.gmca.sample.DevInOut.InOut;
import gov.anl.gmca.scan.DevMCA.OptimizeResult;
import gov.anl.gmca.scan.ModelScan.InterStep;
import gov.anl.gmca.scan.ModelScan.ScanPlotMode;
import gov.anl.gmca.scan.ModelScan.TRACE_ID;
import gov.anl.gmca.scan.ui.ModelGraph;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.ActionResult;
import gov.anl.gmca.util.Point2D;
import gov.anl.gmca.util.Util;

import java.util.ArrayList;

// Take a fluorescence spectrum

public class CmdScanSpectrum extends CmdScan
{
	protected 	double 			scanTimeSec;
	protected 	boolean 		zoomEnergyRange;
	protected	double 			dcmVsum = 0;
	protected	int				pointsRead = 0;
	protected	ActionResult	result			= null;

	public CmdScanSpectrum(boolean zoomEnergyRange, double scanTimeSec)
	{
		this.scanTimeSec	= scanTimeSec;
		this.zoomEnergyRange = zoomEnergyRange;
	}
	
	public ActionResult getResult() { return result; }
	
	@Override
	public void run()
	{
		DevScan dev = DevScan.get();
		
		dev.setInterStep(zoomEnergyRange ?InterStep.OPT_MCA:InterStep.SPECTRUM, false);
		dev.setScanPlotMode(ScanPlotMode.SPECTRUM, false);
		dev.setChoochPDFPath("", false);
		
		if (!zoomEnergyRange) ModelScan.get().setDefaultROI();
		
		// display the edge that was selected in the periodic table
		ModelScan.get().showOnlyIndicatorForElement(ModelScan.get().getElementEdge().element);
		
		for (TRACE_ID id : TRACE_ID.values())
		{
			dev.setGraphTrace(id, new ArrayList<Point2D>(), false);
		}
		dev.setTraceVisible(TRACE_ID.edge, true, false);
		dev.setTraceVisible(TRACE_ID.fPrime, false, false);
		dev.setTraceVisible(TRACE_ID.fDoublePrime, false, false);
		dev.setLeftAxisTrace(TRACE_ID.edge, false);
		dev.setRightAxisTrace(null, false);
		dev.setGraphXMin(0, false);
		dev.setGraphXMax(0, false);
		dev.setDefaultXMin(0, false);
		dev.setDefaultXMax(0, false);
		for (TRACE_ID id : TRACE_ID.values())
		{
			dev.setGraphYLabel(id, "", false);
			dev.setGraphYMin(id, 0, false);
			dev.setGraphYMax(id, 0, false);
			dev.setDefaultYMin(id, 0, false);
			dev.setDefaultYMax(id, 0, false);
		}
		dev.setPeak(Double.NaN, false);
		dev.setRemote(Double.NaN, false);
		dev.setInflection(Double.NaN, false);
		dev.setDefaultPeak(Double.NaN, false);
		dev.setDefaultRemote(Double.NaN, false);
		dev.setDefaultInflection(Double.NaN, false);
		dev.setGraphMessage("Measuring fluorescence spectrum...", false);
		
		double energy = DevEnergy.getDevice().getPosition()*1000.0;
		dev.setSpectrumEnergy(energy, false);
		dev.setGraphTitle(Util.sprintf("Fluorescence spectrum below %.1f eV", energy), false);
		dev.setGraphYLabel(TRACE_ID.edge, "Counts", false); // FIXME: add spectrum trace
		dev.setGraphXLabel("Energy (eV)", false);
		dev.setRightTab(1, true);
		
		if (!verifyCollimator())
		{
			String msg = "Collimator error, aborting scan";
			Log.error(msg);
			result = new ActionResult(msg);
			DevScan.get().setInterStep(InterStep.IDLE, true);
			return;
		}
		
		// moveBacklightOut();

		CmdCheck checks = new CmdCheck(CmdCheck.FLAGS_PRESCAN);
		checks.run();
		if (!checks.checksPassed())
		{
			String errString = "Checks failed, aborting scan";
			Log.error(errString);
			result = new ActionResult(errString);

			dev.setGraphMessage("", false);
			dev.setInterStep(InterStep.IDLE, true);
			
			return;
		}
		
		excitationScan();

		ScanEvent.writeEvent(dcmVsum, pointsRead);
		
		dev.setGraphMessage("", false);
		dev.setInterStep(InterStep.IDLE, true);
	}
	
	protected void excitationScan()
	{
		int gain=0, offset=0;
		int falconROILowChannel=0, falconROIHighChannel=0;

		if (!DevMCA.get().isFalcon)
		{
			if (zoomEnergyRange)
			{
				// get the ROI set by the user
				double roiLow = ModelScan.get().getROILow();
				double roiHigh = ModelScan.get().getROIHigh();

				// determine the gain/offset to zoom to that ROI
				OptimizeResult result = DevMCA.get().optimizeMCA_ADC(roiLow, roiHigh);

				// set ROI in MCA
				DevMCA.get().setROILow(result.N1);
				DevMCA.get().setROIHigh(result.N2);

				// save them
				gain = result.gain;
				offset = result.offset;
			}
			else
			{
				// unset ROI in MCA
				DevMCA.get().setROILow(-1);
				DevMCA.get().setROIHigh(-1);

				// this gain/offset zooms out to the full range
				gain = DevMCA.get().channelMax;
				offset = 0;
			}
			DevMCA.get().setGain(gain);
			DevMCA.get().setOffset(offset);
		}
		else
		{
			// For Falcon, just set ROI, don't set gain or offset
			// Is setting ROI even really needed here?
			falconROILowChannel = DevMCA.get().energyToChannel(ModelScan.get().getROILow(),gain,offset);
			falconROIHighChannel = DevMCA.get().energyToChannel(ModelScan.get().getROIHigh(),gain,offset);
			DevMCA.get().setROILow(falconROILowChannel);
			DevMCA.get().setROIHigh(falconROIHighChannel);
		}

		DevShutter.get("shutter").open();
		
		result = DevMCA.get().requestSpectrum(0, DevMCA.get().channelMax-1, scanTimeSec);
		dcmVsum = DevMCA.get().getLastSpectrumDcmVsum();
		
		DevShutter.get("shutter").close();
		
		if (result.succeeded())
		{
			Util.sleep(100);
			
			final int[] spectrum = DevMCA.get().readSpectrum();
			pointsRead = spectrum.length;

			double xMin, xMax;
			final int f_gain = gain, f_offset = offset;
			if (DevMCA.get().isFalcon && zoomEnergyRange)
			{
				// use ROI for x bounds, don't display the entire range
				xMin = DevMCA.get().channelToEnergy(falconROILowChannel, f_gain, f_offset);
				xMax = DevMCA.get().channelToEnergy(falconROIHighChannel, f_gain, f_offset);
			}
			else
			{
				// find x bounds using first and last channel since we're displaying the entire range
				xMin = DevMCA.get().channelToEnergy(0, f_gain, f_offset);
				xMax = DevMCA.get().channelToEnergy(spectrum.length - 1, f_gain, f_offset);
			}
			double yMin = spectrum[0];
			double yMax = spectrum[0];

			// add trace and find y bounds by checking each point
			ArrayList<Point2D> trace = new ArrayList<Point2D>();
			for (int i=0; i<spectrum.length; i++)
			{
				double energy = DevMCA.get().channelToEnergy(i, f_gain, f_offset);
				if (spectrum[i] < yMin) yMin = spectrum[i];
				if (spectrum[i] > yMax) yMax = spectrum[i];
				trace.add(new Point2D(energy, spectrum[i]));
			}

			// apply x bounds to graph
			DevScan.get().setGraphXMin(xMin, false);
			DevScan.get().setGraphXMax(xMax, false);
			DevScan.get().setDefaultXMin(xMin, false);
			DevScan.get().setDefaultXMax(xMax, false);

			// apply y bounds to graph
			DevScan.get().setGraphYMin(TRACE_ID.edge, yMin, false);
			DevScan.get().setGraphYMax(TRACE_ID.edge, yMax, false);
			DevScan.get().setDefaultYMin(TRACE_ID.edge, yMin, false);
			DevScan.get().setDefaultYMax(TRACE_ID.edge, yMax, false);

			// send points
			DevScan.get().setGraphTrace(TRACE_ID.edge, trace, true);
			
			// save plot data to disk
			String filePrefix = createDirAndGeneratePrefixWithDir(ModelScan.get().getElementEdge(), false);
			FileBIP.saveFile(filePrefix+".bip", ModelGraph.get(), ModelScan.get());
			
			// save screenshot
			ModelGraph.get().graphWidget.saveScreenshot(filePrefix+".png");
		}
	}

	static boolean verifyCollimator()
	{
		boolean isBeamstopOrPinDiodeInPlace = true;
		
		if (!isBeamstopOrPinDiodeInPlace)
		{
		    CCMove ccMove = DevCollimatorCtrl.get().newMove();
		    ccMove.setDest("collimator",	InOut.IN);
		    ccMove.setDest("backlight",		InOut.OUT);
		    ccMove.setDest("beamstop",		InOut.IN);
		    ccMove.setDest("pindiode",		InOut.OUT);
		    ccMove.start().waitFor();

			if (!ccMove.getResult().isOK())
			{
				Log.error("Collimator move failed, aborting scan.");
				return false;
			}
		}
		
		return true;
	}
	
	@Override
	public void requestStop()
	{
		DevMCA.get().abort();
		stopRequested = true;
	}
}
