package gov.anl.gmca.scan;

import java.io.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.StringTokenizer;

// ported from BIWCif.tcl
//
// In this port, defaults are not supported since they are not used in BluIce.
//
// CIF is a standard file format for information exchange in crystallography
// 
// For more info: http://www.iucr.org/resources/cif

public class FileCIF
{
	protected static class ValueKey
	{
		int		blockCount;
		int		loopLineCount;
		String	keyword;
		
		public ValueKey(int blockCount, String keyword, int loopLineCount)
		{
			this.blockCount = blockCount; this.keyword = keyword; this.loopLineCount = loopLineCount;
		}
		
		@Override public int hashCode() { return loopLineCount; }
		
		@Override
		public boolean equals(Object o)
		{
			if (!(o instanceof ValueKey)) return false;
			
			ValueKey v = (ValueKey)o;
			
			if (v.blockCount 	!= this.blockCount)		return false;
			if (v.loopLineCount != this.loopLineCount)	return false;
			if (!v.keyword.equals(this.keyword))		return false;
			
			return true;
		}
	}
	
	protected static class ValueCountKey
	{
		int blockCount;
		String keyword;
		
		public ValueCountKey(int blockCount, String keyword)
		{
			this.blockCount = blockCount; this.keyword = keyword;
		}
		
		@Override public int hashCode() { if (null == keyword) return 0; return keyword.length(); }
		
		@Override
		public boolean equals(Object o)
		{
			if (!(o instanceof ValueCountKey)) return false;
			
			ValueCountKey v = (ValueCountKey)o;
			
			if (v.blockCount 	!= this.blockCount)		return false;
			if (!v.keyword.equals(this.keyword))		return false;
			
			return true;
		}
	}
	
	protected HashMap<ValueKey,String> valueMap = new HashMap<ValueKey,String>();
	protected HashMap<ValueCountKey,Integer> valueCountMap = new HashMap<ValueCountKey,Integer>();
	
	protected int blockCount;
	public int getBlockCount() { return blockCount; }
	
	protected enum LoopState { OUTSIDE, DEFINITION, INSIDE };
	
	public FileCIF(String fullpath)
	{
		load(fullpath);
	}
	
	// should only be called once, from the constructor
	protected void load(String fullpath)
	{
		// open the file for reading 
		File file = new File(fullpath);
		BufferedReader reader = null;
		try
		{
			reader = new BufferedReader(new FileReader(file));
		}
		catch (FileNotFoundException e1) { e1.printStackTrace(); return; }

		// initialize state machine
		blockCount = 0;
		int loopLineCount = 0;
		LoopState loopState = LoopState.OUTSIDE;
		ArrayList<String> loopKeywordList = new ArrayList<String>();

		// keep reading until end of file
		String line = null;
		while ( null != (line = getNextLine(reader)) )
		{
			// extract first token on line
			StringTokenizer st = new StringTokenizer(line);
			
			if (!st.hasMoreTokens()) continue;
			String token0 = st.nextToken();
			
			// handle start of new data block
			if (token0.indexOf("data_")==0)
			{
				// increment the data block index
				blockCount++;
				
				// inidicate that outside of loop
				loopState = LoopState.OUTSIDE;

				// get next line of file
				continue;
			}

			// handle start of new loop definition block
			if (token0.indexOf("loop_")==0)
			{
				// indicate that inside of loop definition
				loopState = LoopState.DEFINITION;

				// initialize loop keyword list
				loopKeywordList.clear();

				// get next line of file
				continue;
			}

			// handle cases where first token is a keyword
			if (token0.charAt(0) == '_')
			{
				// add keyword to loop definition if inside loop definition
				if (loopState == LoopState.DEFINITION)
				{
					loopKeywordList.add(token0);
				}
				else
				{
					String keyword = token0;
					ArrayList<String> values = quoteSplit(line);
					String value = values.size()>1 ? values.get(1) : "";

					//System.out.println("("+blockCount+","+keyword+","+loopLineCount+") = "+value);
					
					// store the value in the data value array
					valueMap.put(new ValueKey(blockCount,keyword,loopLineCount), value);
					
					// count each value associated with a keyword
					valueCountMap.put(new ValueCountKey(blockCount,keyword), loopLineCount);
				}

				// get next line of file
				continue;
			}

			// if previously in loop definition then now in loop body
			if (loopState == LoopState.DEFINITION)
			{
				loopState = LoopState.INSIDE;
				loopLineCount = 0;
			}
			
			// read data into loop variables if in loop
			if (loopState == LoopState.INSIDE)
			{
				// keep track of number of lines
				loopLineCount++;

				// loop over loop keywords and data
				ArrayList<String> values = quoteSplit(line);
				
				for (int i=0; i<loopKeywordList.size(); i++)
				{
					String keyword = loopKeywordList.get(i);
					String value = values.get(i);

					//System.out.println("("+blockCount+","+keyword+","+loopLineCount+") = "+value);
					
					// store the value in the data value array
					valueMap.put(new ValueKey(blockCount,keyword,loopLineCount), value);
					
					// count each value associated with a keyword
					valueCountMap.put(new ValueCountKey(blockCount,keyword), loopLineCount);
				}
			}
		}

		// close the file
		try
		{
			reader.close();
		}
		catch (IOException e) { e.printStackTrace(); }
	}
	
	protected ArrayList<String> quoteSplit(String line)
	{
		ArrayList<String> ret = new ArrayList<String>();
		
		boolean inQuotes = false;
		StringBuilder sb = new StringBuilder();
		for (int i=0; i<line.length(); i++)
		{
			char c = line.charAt(i);
			
			if (!inQuotes && (c == ' ' || c == '\t' || c == '\n' || c == '\r'))
			{
				if (sb.length() > 0)
				{
    				ret.add(sb.toString());
    				sb = new StringBuilder();
				}
			}
			else if (c == '\"')
			{
				if (inQuotes)
				{
					ret.add(sb.toString());
					sb = new StringBuilder();
					inQuotes = false;
				}
				else
				{
					inQuotes = true;
				}
			}
			else
			{
				sb.append(c);
			}
		}
		
		if (sb.length() > 0)
		{
			ret.add(sb.toString());
		}
		
		return ret;
	}
	
	public String getValue(int block, String keyword, int instance)
	{
		if (valueMap.containsKey(new ValueKey(block,keyword,instance)))
		{
			return valueMap.get(new ValueKey(block, keyword, instance));
		}
		else
		{
			return null;
		}
	}
	
	public String getValue(int block, String keyword) { return getValue(block, keyword, 0); }

	public int getValueCount(int block, String keyword)
	{
		//System.out.println("getValueCount("+block+","+keyword+")");
		
		if (valueCountMap.containsKey(new ValueCountKey(block, keyword)))
		{
			//System.out.println("returning "+valueCountMap.get(new ValueCountKey(block, keyword)));
			return valueCountMap.get(new ValueCountKey(block, keyword));
		}
		else
		{
			//System.out.println("returning 0");
			return 0;
		}
	}

	protected String getNextLine(BufferedReader reader)
	{
		while (true)
		{
			String lineBuffer;
			try
			{
				lineBuffer = reader.readLine();
			}
			catch (IOException e) { return null; }
			
			if (null == lineBuffer) return null;
			
			// remove any comments
			if (lineBuffer.length() > 0 && lineBuffer.charAt(0) == '#')
			{
				return "";
			}
			
			// return line buffer if any non-white space found
			for (int i=0; i<lineBuffer.length(); i++)
			{
				char c = lineBuffer.charAt(i);
				if (c != ' ' && c != '\t' && c != '\n' && c != '\r') return lineBuffer;
			}
			
			return "";
		}
	}
}
