package gov.anl.gmca.scan;

import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.common.DevDataDir;
import gov.anl.gmca.common.Limits;
import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.gui.MenuActions;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.io.ExecTask;
import gov.anl.gmca.scan.CmdCleanup.Mode;
import gov.anl.gmca.scan.DevEdgeDB.EdgeData;
import gov.anl.gmca.scan.ui.ModelGraph;
import gov.anl.gmca.scan.ui.ModelGraph.Trace;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.users.ModelUserTable.ClientListener;
import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.Pair;
import gov.anl.gmca.util.Point2D;
import gov.anl.gmca.util.Util;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Color;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.StringTokenizer;

// some variables are duplicated in this class and ModelGraph
// this is to make it easy in case ModelGraph is replaced in the future with a non-custom graph class
// 
// all variables are stored in this class
// graph-related variables are simultaneously stored in the graph class, and the graph is redrawn when
// needed
public class ModelScan
{
	public static interface ScanListener
	{
		public void scanChanged(ModelScan settings);
	}

	public ListenerList<ModelScan> llDisplayedElements;
	
	ArrayList<ScanListener> listenerList = new ArrayList<ScanListener>();
	public void addListener(ScanListener sl) { listenerList.add(sl); }
	protected void notifyListeners()
	{
		for (ScanListener l : new ArrayList<ScanListener>(listenerList))
		{
			try
			{
				l.scanChanged(this);
			}
			catch (Exception e) { e.printStackTrace(); }
		}
	}

	///////////////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////// end listener ////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////////
	
	public static class ElementEdge
	{
		String element, edge;

		public double getEnergy()
		{
			return DevEdgeDB.get().getEdgeData(this).edgeEnergy;
		}
		
		public ElementEdge(String element, String edge)
		{
			this.element=element;
			this.edge=edge;
		}
		
		public ElementEdge(String elementEdge) throws IllegalArgumentException
		{
			StringTokenizer st = new StringTokenizer(elementEdge,"-");
			if (2 != st.countTokens()) throw new IllegalArgumentException("Invalid elementEdge \""+elementEdge+"\"");
			
			this.element = st.nextToken();
			this.edge = st.nextToken();
		}

		@Override public boolean equals(Object o)
		{
			if (!(o instanceof ElementEdge)) return false;
			
			ElementEdge oEdge = (ElementEdge)o;

			return (Util.doStringsMatch(element, oEdge.element) && Util.doStringsMatch(edge, oEdge.edge));
		}
		
		@Override public int hashCode() { if (null == element || element.length()==0) return 0; return element.charAt(0); }
		@Override public String toString() { return element+"-"+edge; }
	}
	
	protected 	ElementEdge 	elementEdge						= null;
	public 		ElementEdge 	getElementEdge() 				{ return elementEdge; }
	public 		void 			setElementEdge(ElementEdge ee) 	{ elementEdge=ee; }

	public static enum ScanPlotMode
	{
		NONE, SPECTRUM, EDGE;

		public static ScanPlotMode fromString(String modeString)
		{
			if (null == modeString) return NONE;
			
			for (ScanPlotMode mode : ScanPlotMode.values())
			{
				if (mode.toString().toLowerCase().equals(modeString.toLowerCase()))
				{
					return mode;
				}
			}
			return NONE;
		}
	}
	
	protected 	ScanPlotMode 	scanPlotMode					= ScanPlotMode.NONE;
	public 		ScanPlotMode 	getScanPlotMode() 				{ return scanPlotMode; }
	public 		void 			setScanPlotMode(ScanPlotMode m) { scanPlotMode=m; }

	protected 	double 			edgeTime 						= Double.NaN;
	public 		double 			getEdgeTime() 					{ return edgeTime; }
	public 		void 			setEdgeTime(double d) 			{ edgeTime=d; }

	protected 	double 			spectrumTime 					= Double.NaN;
	public 		double 			getSpectrumTime() 				{ return spectrumTime; }
	public 		void 			setSpectrumTime(double d) 		{ spectrumTime=d; }
	
	protected 	double 			energy 							= Double.NaN;
	public 		double 			getEnergy() 					{ return energy; }
	public 		void 			setEnergy(double d) 			{ energy=d; }
	
	protected 	double 			spectrumEnergy					= Double.NaN;
	public 		double 			getSpectrumEnergy() 			{ return spectrumEnergy; }
	public 		void 			setSpectrumEnergy(double d) 	{ spectrumEnergy=d; }
	
	protected 	double 			scanCenter 						= Double.NaN;
	public 		double 			getScanCenter() 				{ return scanCenter; }
	public 		void 			setScanCenter(double d) 		{ scanCenter=d; }
	
	protected 	int 			leftTab 						= 0;
	public 		int 			getLeftTab() 					{ return leftTab; }
	public 		void 			setLeftTab(int i) 				{ leftTab=i; }
	
	protected 	int 			rightTab 						= 0;
	public 		int 			getRightTab() 					{ return rightTab; }
	public 		void 			setRightTab(int i) 				{ rightTab=i; }
	
	protected	double			graphXMin						= 0.0;
	public 		double 			getGraphXMin() 					{ return graphXMin; }
	public 		void 			setGraphXMin(double d) 			{ graphXMin=d; ModelGraph.get().setXRange(d,  graphXMax); }
	
	protected	double			graphXMax						= 1.0;
	public 		double 			getGraphXMax() 					{ return graphXMax; }
	public 		void 			setGraphXMax(double d) 			{ graphXMax=d; ModelGraph.get().setXRange(graphXMin, d); }

	protected	double			defaultXMin						= 0.0;
	public 		double 			getDefaultXMin() 				{ return defaultXMin; }
	public 		void 			setDefaultXMin(double d) 		{ defaultXMin=d; }
	
	protected	double			defaultXMax						= 1.0;
	public 		double 			getDefaultXMax() 				{ return defaultXMax; }
	public 		void 			setDefaultXMax(double d) 		{ defaultXMax=d; }

	protected	TRACE_ID		leftAxisTrace					= null;
	public		void			setLeftAxisTrace(TRACE_ID id)	{ leftAxisTrace = id; updateTraceSides(); }
	public		TRACE_ID		getLeftAxisTrace()				{ return leftAxisTrace; }
	
	protected	TRACE_ID		rightAxisTrace					= null;
	public		void			setRightAxisTrace(TRACE_ID id)	{ rightAxisTrace = id; }
	public		TRACE_ID		getRightAxisTrace()				{ return rightAxisTrace; }

	private void updateTraceSides()
	{
		for (TRACE_ID id : TRACE_ID.values())
		{
			Trace trace = ModelGraph.get().getTrace(id);
			
			if (getLeftAxisTrace() == id) trace.setYAxisSide(SWT.LEFT);
			else if (getRightAxisTrace() == id) trace.setYAxisSide(SWT.RIGHT);
			else trace.setYAxisSide(SWT.NONE);
		}
	}
	
	protected 	boolean 		autoOptimized 					= false;
	public 		boolean 		getAutoOptimized() 				{ return autoOptimized; }
	public 		void 			setAutoOptimized(boolean b) 	{ autoOptimized=b; }
	
	protected 	boolean 		adaptiveScan 					= false;
	public 		boolean 		getAdaptiveScan() 				{ return adaptiveScan; }
	public 		void 			setAdaptiveScan(boolean b) 		{ adaptiveScan=b; }

	protected 	boolean 		showAllEdges	 					= false;
	public 		boolean 		getShowAllEdges() 					{ return showAllEdges; }
	public 		void 			setShowAllEdges(boolean b) 			{ showAllEdges=b; }

	public static enum EDGE_IND
	{
		Mn,Fe,Co,Ni,Cu,Zn,As,Se,Br,Ta,W,Pt,Hg,/*Pb*/Other,Other2;

		public static EDGE_IND elementToIndicator(String str)
		{
			for (EDGE_IND ei : values())
			{
				if (ei.toString().toLowerCase().equals(str.toLowerCase())) return ei;
			}
			return null;
		}
	};
	public static enum EI_STATE { OFF, TEMP, ON };
	
	protected HashMap<EDGE_IND,EI_STATE> displayedElements = new HashMap<EDGE_IND,EI_STATE>();
	public HashMap<EDGE_IND,EI_STATE> getDisplayedElements()
	{
		return displayedElements;
	}
	public void setDisplayedElements(HashMap<EDGE_IND,EI_STATE> de)
	{
		displayedElements = de;
		
		for (EDGE_IND ind : de.keySet())
		{
			ModelGraph.get().setElementPermState(ind, de.get(ind)==EI_STATE.ON);
		}
	}

	public static enum AutoStep { NONE, MOVE_ENERGY, OPTIMIZE, SPECTRUM, EDGE_SCAN, RETRACT, STOPPING };
	protected	AutoStep		autoStep						= AutoStep.NONE;
	public		AutoStep		getAutoStep()					{ return autoStep; }
	public		void			setAutoStep(AutoStep step)		{ autoStep = step; }
	
	public static enum InterStep
	{
		IDLE 		("Idle"),
		OPT_DET		("Optimizing detector"),
		SPECTRUM	("Taking spectrum"),
		OPT_MCA		("Optimizing MCA"),
		EDGE		("Taking edge scan"),
		CLEANUP		("Cleaning up"),
		STOPPING	("Stopping");
		
		protected String text;
		InterStep(String text) { this.text=text; }
		@Override public String toString() { return text; }
	};
	
	protected 	InterStep 		interStep						= InterStep.IDLE;
	public 		InterStep 		getInterStep() 					{ return interStep; }
	public 		void 			setInterStep(InterStep is)		{  interStep = is; }
	
	protected 	String 			graphTitle						= "";
	public 		String 			getGraphTitle()					{ return graphTitle; }
	public 		void 			setGraphTitle(String s) 		{ graphTitle=s; ModelGraph.get().setTitle(s); }
	
	protected 	String 			graphMessage					= "";
	public 		String 			getGraphMessage()				{ return graphMessage; }
	public 		void 			setGraphMessage(String s) 		{ graphMessage=s; ModelGraph.get().setMessage(s); }
	
	protected 	String 			graphXLabel						= "";
	public 		String 			getGraphXLabel()				{ return graphXLabel; }
	public 		void 			setGraphXLabel(String s) 		{ graphXLabel=s; ModelGraph.get().setXLabel(s); }
	
	protected 	String 			choochPDFPath					= "";
	public 		String 			getChoochPDFPath()				{ return choochPDFPath; }
	public 		void 			setChoochPDFPath(String s) 		{ choochPDFPath=s; }
	
	protected 	String 			prefix							= "";
	public 		String 			getPrefix()						{ return prefix; }
	public 		void 			setPrefix(String s) 			{ prefix=s; }
	
	protected 	String 			directory						= "";
	public		String			getDirectorySuffix()			{ return directory; }
	public 		String 			getFullDirectory()				{ return DevDataDir.getPrimary()+"/"+directory; }
	public 		void 			setDirectory(String s) 			{ directory=s; }
	
	protected 	String 			otherElement					= "";
	public 		String 			getOtherElement()				{ return otherElement; }
	public 		void 			setOtherElement(String s) 		{ otherElement=s; }
	
	protected 	String 			otherElement2					= "";
	public 		String 			getOtherElement2()				{ return otherElement2; }
	public 		void 			setOtherElement2(String s) 		{ otherElement2=s; }
	
	// the next five variables are updated only when a user is done dragging one of the bars, or enters a text value
	protected 	double 			roiLow 							= Double.NaN;
	public 		double 			getROILow() 					{ return roiLow; }
	public 		void 			setROILow(double d) 			{ roiLow=d; ModelGraph.get().barROILow.x=d; }
	
	protected 	double 			roiHigh 						= Double.NaN;
	public 		double 			getROIHigh() 					{ return roiHigh; }
	public 		void 			setROIHigh(double d) 			{ roiHigh=d; ModelGraph.get().barROIHigh.x=d; }
	
	protected 	double 			inflection 						= Double.NaN;
	public 		double 			getInflection() 				{ return inflection; }
	public 		void 			setInflection(double d) 		{ inflection=d; ModelGraph.get().barInflection.x=d; }
	
	protected 	double 			infl_fP							= Double.NaN;
	public 		double 			getInfl_fP() 					{ return infl_fP; }
	public 		void 			setInfl_fP(double d)	 		{ infl_fP=d; }
	
	protected 	double 			infl_fPP 						= Double.NaN;
	public 		double 			getInfl_fPP() 					{ return infl_fPP; }
	public 		void 			setInfl_fPP(double d) 			{ infl_fPP=d; }
	
	protected 	double 			peak 							= Double.NaN;
	public 		double 			getPeak() 						{ return peak; }
	public 		void 			setPeak(double d) 				{ peak=d; ModelGraph.get().barPeak.x=d; }
	
	protected 	double 			peak_fP 						= Double.NaN;
	public 		double 			getPeak_fP() 					{ return peak_fP; }
	public 		void 			setPeak_fP(double d) 			{ peak_fP=d; }
	
	protected 	double 			peak_fPP 						= Double.NaN;
	public 		double 			getPeak_fPP() 					{ return peak_fPP; }
	public 		void 			setPeak_fPP(double d) 			{ peak_fPP=d; }
	
	protected 	double 			remote 							= Double.NaN;
	public 		double 			getRemote() 					{ return remote; }
	public 		void 			setRemote(double d) 			{ remote=d; ModelGraph.get().barRemote.x=d; }

	protected 	double 			defaultInflection 				= Double.NaN;
	public 		double 			getDefaultInflection() 			{ return defaultInflection; }
	public 		void 			setDefaultInflection(double d) 	{ defaultInflection=d; }
	
	protected 	double 			defaultPeak 					= Double.NaN;
	public 		double 			getDefaultPeak() 				{ return defaultPeak; }
	public 		void 			setDefaultPeak(double d) 		{ defaultPeak=d; }
	
	protected 	double 			defaultRemote 					= Double.NaN;
	public 		double 			getDefaultRemote() 				{ return defaultRemote; }
	public 		void 			setDefaultRemote(double d) 		{ defaultRemote=d; }

	// ------------- trace-specific variables follow --------------
	
	public static enum TRACE_ID
	{
		edge(GUI.COLOR_DARK_GREEN),
		fPrime(GUI.COLOR_RED),
		fDoublePrime(GUI.COLOR_BLUE);
		
		private Color color;
		public Color getColor() { return color; }
		
		TRACE_ID(Color color) { this.color=color; }
	};
	
	protected	HashMap<TRACE_ID,Double> graphYMin									= new HashMap<TRACE_ID, Double>();
	public 		Double 				getGraphYMin(TRACE_ID id) 						{ Double ret=graphYMin.get(id); return null==ret?0:ret; }
	public 		void 				setGraphYMin(TRACE_ID id, double d) 			{ graphYMin.put(id,d); ModelGraph.get().getTrace(id).setYRange(d, getGraphYMax(id)); }
	
	protected	HashMap<TRACE_ID,Double> graphYMax									= new HashMap<TRACE_ID, Double>();
	public 		Double 				getGraphYMax(TRACE_ID id) 						{ Double ret=graphYMax.get(id); return null==ret?0:ret; }
	public 		void 				setGraphYMax(TRACE_ID id, double d) 			{ graphYMax.put(id,d); ModelGraph.get().getTrace(id).setYRange(getGraphYMin(id), d); }
	
	protected	HashMap<TRACE_ID,Double> defaultYMin								= new HashMap<TRACE_ID, Double>();
	public 		Double 				getDefaultYMin(TRACE_ID id) 					{ Double ret=defaultYMin.get(id); return null==ret?0:ret; }
	public 		void 				setDefaultYMin(TRACE_ID id, double d) 			{ defaultYMin.put(id,d); }
	
	protected	HashMap<TRACE_ID,Double> defaultYMax								= new HashMap<TRACE_ID, Double>();
	public 		Double 				getDefaultYMax(TRACE_ID id) 					{ Double ret=defaultYMax.get(id); return null==ret?0:ret; }
	public 		void 				setDefaultYMax(TRACE_ID id, double d) 			{ defaultYMax.put(id,d); }
	
	protected 	HashMap<TRACE_ID,String> graphYLabel								= new HashMap<TRACE_ID, String>();
	public 		String 				getGraphYLabel(TRACE_ID id)						{ String ret=graphYLabel.get(id); return null==ret?"":ret; }
	public 		void 				setGraphYLabel(TRACE_ID id, String s) 			{ graphYLabel.put(id,s); ModelGraph.get().getTrace(id).setYLabel(s); }
	
	protected	HashMap<TRACE_ID,ArrayList<Point2D>> graphTrace						= new HashMap<TRACE_ID, ArrayList<Point2D>>();
	public		ArrayList<Point2D> 	getGraphTrace(TRACE_ID id)						{ ArrayList<Point2D> ret=graphTrace.get(id); return null==ret?(new ArrayList<Point2D>()):ret; }
	public		void				setGraphTrace(TRACE_ID id, ArrayList<Point2D> a){ graphTrace.put(id,a); ModelGraph.get().getTrace(id).setPoints(a); }
	
	protected	HashMap<TRACE_ID,Boolean> traceVisible								= new HashMap<TRACE_ID,Boolean>();
	public 		Boolean 			getTraceVisible(TRACE_ID id) 					{ Boolean ret=traceVisible.get(id); return null==ret?false:ret; }
	public 		void 				setTraceVisible(TRACE_ID id, boolean b) 		{ traceVisible.put(id,b); ModelGraph.get().getTrace(id).setVisible(b); }
	
	///////////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////// end variables ////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////////
	
	protected CmdScan currentCmd = null;
	protected DevScan scanDev;
	public void setDevice(DevScan scanDev)		{ this.scanDev = scanDev; }
	
	protected static ModelScan inst 	= null;
	public static ModelScan get() 		{ return inst; }
	
	public ModelScan()
	{
		if (null != inst)
		{
			Log.error("ModelScanSettings already exists");
		}
		inst = this;
		llDisplayedElements = new ListenerList<ModelScan>(this);
		
		for (EDGE_IND element : EDGE_IND.values()) { displayedElements.put(element, EI_STATE.OFF); }
		
		Client.addListener(new ClientListener()
		{
			@Override public void masterChanged(boolean isMaster)
			{
				if (isMaster)
				{
					if (getInterStep() != InterStep.IDLE) DevScan.get().setInterStep(InterStep.IDLE, true);
					if (getAutoStep() != AutoStep.NONE) DevScan.get().setAutoStep(AutoStep.NONE, true);
				}
				
				ModelGraph.get().setEnabled(isMaster);
				ModelGraph.get().redraw(false,false,false,true,true,false);
			}
		});
	}
	
	public Pair<Double,Double> setDefaultROI()
	{
		EdgeData data = DevEdgeDB.get().getEdgeData(getElementEdge());
		
		double roiLow = data.fluoEnergy - data.fluoRange;
		double roiHigh = data.fluoEnergy + data.fluoRange;
		
		DevScan.get().setROILow(roiLow, false);
		DevScan.get().setROIHigh(roiHigh, true);
		
		return new Pair<Double,Double>(roiLow,roiHigh);
	}
	
	protected void handleChange()
	{
		if (!validateAndFix()) return;
		ModelGraph.get().redraw(true, true, true, true, true, false);
		notifyListeners();
	}

	public boolean validateAndFix()
	{
		boolean valid = true;
		
		if (null == getElementEdge() || getElementEdge().equals(""))
		{
			scanDev.setElementEdge(new ElementEdge("Se-K"), false);
			valid = false;
		}
		
		if (0 == getPrefix().length())
		{
			scanDev.setPrefix("test", false);
			valid = false;
		}
		
		if (!valid) DevScan.get().sendUpdate();
		
		return valid;
	}

	public void openChoochPDF()
	{
		String pdfPath = getChoochPDFPath();
		if (null != pdfPath && pdfPath.length()>1)
		{
			File file = new File(pdfPath);
			if (!file.exists() || !file.canRead()) { Log.error("Couldn't read Chooch PDF "+pdfPath); return; }
			
			Log.note("Opening Chooch PDF: "+pdfPath);
			ExecTask.asyncExec(MenuActions.acroreadPath+" "+pdfPath, null, false);
		}
		else
		{
			Util.popupError("Error opening PDF","There was an error opening the chooch PDF for the current graph.");
		}
	}
	
	////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////// auto scan ///////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////
	
	public void startAutoScan()
	{
		if (!verifyIdle()) return;
		
		currentCmd = new CmdScanAuto(true);
		currentCmd.start();
	}
	
	public void restartAutoScan()
	{
		if (!verifyIdle()) return;
		
		currentCmd = new CmdScanAuto(false);
		currentCmd.start();
	}
	
	public void cleanupAutoScan()
	{
		if (!verifyIdle()) return;
		
		currentCmd = new CmdCleanup(Mode.AUTO);
		currentCmd.start();
	}
	
	public void stopAutoScan()
	{
		if (null != currentCmd && currentCmd.isAlive())
		{
			DevScan.get().setAutoStep(AutoStep.STOPPING, true);
			currentCmd.requestStop();
		}
	}
	
	////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////// interactive scan ////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////
	
	public void startOptimizeDetector()
	{
		if (!verifyIdle()) return;
		
		currentCmd = new CmdOptDetector(false);
		currentCmd.start();
	}
	
	public void startSpectrum()
	{
		if (!verifyIdle()) return;
		
		currentCmd = new CmdScanSpectrum(false, ModelScan.get().getSpectrumTime());
		currentCmd.start();
	}
	
	public void startOptimizeMCA()
	{
		if (!verifyIdle()) return;
		
		currentCmd = new CmdScanSpectrum(true, ModelScan.get().getSpectrumTime());
		currentCmd.start();
	}
	
	public void startEdgeScan()
	{
		if (!verifyIdle()) return;
		
		currentCmd = new CmdScanEdge();
		currentCmd.start();
	}
	
	public void cleanupInteractiveScan()
	{
		if (!verifyIdle()) return;
		
		currentCmd = new CmdCleanup(Mode.INTER);
		currentCmd.start();
	}
	
	public void stopInteractiveScan()
	{
		if (null != currentCmd && currentCmd.isAlive())
		{
			DevScan.get().setInterStep(InterStep.STOPPING, true);
			currentCmd.requestStop();
		}
	}

	////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////// common scan methods //////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////

	public boolean inHardwareLimits(ElementEdge ee)
	{
		EdgeData data = DevEdgeDB.get().getEdgeData(ee);
		double edgeScanDelta_keV = DBConfig.getDouble("edgeScanRange_eV")/1000.0/2.0;
		double excitationScanDelta_keV = 0.150;
		double edgeScanMin_keV = data.edgeEnergy/1000.0 - edgeScanDelta_keV;
		double edgeScanMax_keV = data.edgeEnergy/1000.0 + Math.max(edgeScanDelta_keV, excitationScanDelta_keV);
		Limits monoLimits = DevPmacMotor.get("mo_En_E").getLimits();
		
		return
			monoLimits.testValue(edgeScanMin_keV).isOK() &&
			monoLimits.testValue(edgeScanMax_keV).isOK();
	}

	
	protected boolean verifyIdle()
	{
		if (null != currentCmd && currentCmd.isAlive())
		{
			Log.error("Scan command is already running, click ignored.");
			return false;
		}
		return true;
	}
	
	public void abortScan()
	{
		// TODO: review this
		if (autoStep != AutoStep.NONE) stopAutoScan();
		if (interStep != InterStep.IDLE) stopInteractiveScan();
	}
	
	public boolean canZoomOut()
	{
		double tolerance = 0.001;
		
		double xDiff = Math.abs(defaultXMin-graphXMin)/(defaultXMin+defaultXMax);
		if (xDiff>tolerance) return true;
		
		for (TRACE_ID trace : TRACE_ID.values())
		{
			if (getTraceVisible(trace))
			{
				double yDiff = Math.abs(graphYMin.get(trace)-defaultYMin.get(trace))/(graphYMin.get(trace)+defaultYMin.get(trace));
				if (yDiff>tolerance) return true;
			}
		}
		
		return false;
	}
	
	public void showAll()
	{
		DevScan.get().setGraphXMin(getDefaultXMin(), false);
		DevScan.get().setGraphXMax(getDefaultXMax(), false);
		
		for (TRACE_ID trace : TRACE_ID.values())
		{
			DevScan.get().setGraphYMin(trace, getDefaultYMin(trace), false);
			DevScan.get().setGraphYMax(trace, getDefaultYMax(trace), false);
		}
		
		DevScan.get().sendUpdate();
	}
	
	private static final double zoomFactor = 2.0;
	
	public void zoomOut()
	{
		double xRange = getGraphXMax()-getGraphXMin();
		double newXMin = getGraphXMin() - (xRange/2.0);
		double newXMax = getGraphXMax() + (xRange/2.0);
		newXMin = Math.max(newXMin, getDefaultXMin());
		newXMax = Math.min(newXMax, getDefaultXMax());
		DevScan.get().setGraphXMin(newXMin, false);
		DevScan.get().setGraphXMax(newXMax, false);
		
		for (TRACE_ID id : TRACE_ID.values())
		{
			double yRange = getGraphYMax(id)-getGraphYMin(id);
			double newYMin = getGraphYMin(id) - (yRange/2.0);
			double newYMax = getGraphYMax(id) + (yRange/2.0);
			newYMin = Math.max(newYMin, getDefaultYMin(id));
			newYMax = Math.min(newYMax, getDefaultYMax(id));
			DevScan.get().setGraphYMin(id, newYMin, false);
			DevScan.get().setGraphYMax(id, newYMax, false);
		}
		
		DevScan.get().sendUpdate();
	}
	
	public void showOnlyIndicatorForElement(String element)
	{
		EDGE_IND edge = EDGE_IND.elementToIndicator(element);
		if (null==edge)
		{
			// if the edge is not one of the defaults, set the second 'other' element to the selected one
			edge = EDGE_IND.Other2;
			DevScan.get().setOtherElement2(element, false);
		}
		DevScan.get().setElementDisplayState(edge, EI_STATE.ON, true);
	}
}
