package gov.anl.gmca.scan.ui;

import gov.anl.gmca.common.Limits;
import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.io.Resources;
import gov.anl.gmca.scan.DevScan;
import gov.anl.gmca.scan.ModelScan;
import gov.anl.gmca.scan.ModelScan.ElementEdge;
import gov.anl.gmca.scan.ModelScan.ScanListener;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.util.Util;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.*;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Text;

import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.EventListener;
import java.util.StringTokenizer;

public class WdgPeriodicTable extends Composite implements PaintListener, EventListener
{
	protected static final Point edgeTextPos = new Point(286,70);
	protected static final Point energyTextPos = new Point(286,105);
	protected static final Point textFieldSize = new Point(96,27);

	protected static final int TABLE_WIDTH = 700;
	protected static final int TABLE_HEIGHT = 500;
	protected static final int MARKER_HL_MARGIN = 2;

	protected Text txtEdge;
	protected Text txtEnergy;
	
	protected static class EdgeMarker
	{
		public Rectangle bounds;
		public ElementEdge edge;
		public boolean enabled;
		public EdgeMarker(String elementString, String edgeString, int xMin, int yMin, int xMax, int yMax, boolean _enabled)
		{
			bounds = new Rectangle(xMin, yMin, xMax-xMin, yMax-yMin);
			edge = new ElementEdge(elementString, edgeString);
			enabled = _enabled;
		}
		public boolean inBounds(int x, int y)
		{
			return
			(
				x >= bounds.x &&
				x <= bounds.x + bounds.width &&
				y >= bounds.y &&
				y <= bounds.y + bounds.height
			);
		}
	}
	
	class Edge
	{
		String type;
		String energy;
		boolean display;
		String cutoff;
		boolean enabled;
	}
	
	class Element
	{
		int atomicZ;
		int row;
		int col;
		String symbol = null;
		ArrayList<Edge> edges = null;
	}
	
	protected ArrayList<Element> elements = new ArrayList<Element>();
	protected ArrayList<EdgeMarker> markerList = new ArrayList<EdgeMarker>();
	
	protected Image tableImage;

	protected EdgeMarker selectedEdge = null;
	
	private static WdgPeriodicTable inst;
	public static WdgPeriodicTable get() { return inst; }
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////// end variables //////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////////////

	public WdgPeriodicTable(Composite parent, int style)
	{
		super(parent, style);
		inst = this;
		
		addPaintListener(this);
		
		readData();
		
		createTableImage();
		
		txtEdge = new Text(this, SWT.BORDER);
		txtEdge.setFont(GUI.FONT_DEFAULT);
		txtEdge.setEditable(false);
		txtEdge.setSize(textFieldSize);
		txtEdge.setLocation(edgeTextPos);
		
		txtEnergy = new Text(this, SWT.BORDER);
		txtEnergy.setFont(GUI.FONT_DEFAULT);
		txtEnergy.setEditable(false);
		txtEnergy.setSize(textFieldSize);
		txtEnergy.setLocation(energyTextPos);
		
		ModelScan.get().addListener(new ScanListener()
		{
			@Override public void scanChanged(final ModelScan settings)
			{
				GUI.asyncExec(new Runnable(){public void run()
				{
    				txtEdge.setText(settings.getElementEdge().toString());
    				txtEnergy.setText(Double.toString(settings.getEnergy()));
				}});
			}
		});
		
		addMouseListener(new MouseListener()
		{
			@Override public void mouseDown(MouseEvent me)
			{
				if (!Client.isMaster()) return;
				
				for (EdgeMarker marker : markerList)
				{
					if (marker.inBounds(me.x,me.y))
					{
						if (marker.enabled) {
							DevScan.get().setElementEdge(marker.edge, true);
							return;
						} else {
							Limits monoLimits = DevPmacMotor.get("mo_En_E").getLimits();
							String mirrorString = Util.getBeamlineNumberFromHostname()==0?"mirror-related ":"";
							Util.popupError
							(
								"Edge disabled",
								Util.sprintf
								(
									"The %s edge at %.3f keV can't be scanned due to the current\n"+
									mirrorString+
									"energy limits of %.3f to %.3f keV.",
									marker.edge.toString(),
									marker.edge.getEnergy()/1000.0,
									monoLimits.getLower(),
									monoLimits.getUpper()
								)
							);
							return;
						}
					}
				}
			}
			@Override public void mouseDoubleClick(MouseEvent me) {}
			@Override public void mouseUp(MouseEvent me) {}
		});
		
		addMouseMoveListener(new MouseMoveListener()
		{
			@Override public void mouseMove(MouseEvent me)
			{
				if (!Client.isMaster()) return;
				
				EdgeMarker oldSelectedEdge = selectedEdge;
				boolean edgeFound = false;
				
				for (EdgeMarker marker : markerList)
				{
					if (marker.inBounds(me.x,me.y))
					{
						selectedEdge = marker;
						edgeFound = true;
						break;
					}
				}
				if (!edgeFound) selectedEdge = null;
				
				if (oldSelectedEdge != selectedEdge)
				{
					if (null != oldSelectedEdge)
					{
						Rectangle oldB = oldSelectedEdge.bounds;
						redraw(oldB.x-MARKER_HL_MARGIN, oldB.y-MARKER_HL_MARGIN, oldB.width+2*MARKER_HL_MARGIN+1, oldB.height+2*MARKER_HL_MARGIN+1, false);
					}
					if (null != selectedEdge)
					{
						Rectangle newB = selectedEdge.bounds;
						redraw(newB.x-MARKER_HL_MARGIN, newB.y-MARKER_HL_MARGIN, newB.width+2*MARKER_HL_MARGIN+1, newB.height+2*MARKER_HL_MARGIN+1, false);
					}
				}
			}});
	}
	
	// This widget size is fixed
	@Override
	public Point computeSize(int wHint, int hHint, boolean changed)
	{
		return new Point(TABLE_WIDTH,TABLE_HEIGHT);
	}

	protected void readData()
	{
		DataInputStream dis = null;
    	String filename = "/data/periodic_bl92.dat";
		
		try
		{
        	BufferedReader br = Resources.get().getReader(filename);
        	
        	String line;
        	while ( (line = br.readLine()) != null )
        	{
        		StringTokenizer st = new StringTokenizer(line);
        		Element element = new Element();
        		element.atomicZ = Integer.parseInt(st.nextToken());
        		element.row = Integer.parseInt(st.nextToken());
        		element.col = Integer.parseInt(st.nextToken());
        		element.symbol = st.nextToken();
        		
        		int nEdges = Integer.parseInt(st.nextToken());
        		
        		if (0 != nEdges)
        		{
        			element.edges = new ArrayList<Edge>();
        		
        			for (int i=0; i < nEdges; i++)
        			{
        				line = br.readLine();
        				if (null == line) { System.out.println("ERROR: unexpected end of file in periodic table data"); dis.close(); return; }
        				
        				StringTokenizer st2 = new StringTokenizer(line);
        				
        				Edge edge = new Edge();
        				edge.type = st2.nextToken();  
        				edge.energy = st2.nextToken();  
        				edge.display = st2.nextToken().equals("Y"); 
        				edge.cutoff = st2.nextToken();  
        				
        				ElementEdge ee = new ElementEdge(element.symbol,edge.type);
        				edge.enabled = ModelScan.get().inHardwareLimits(ee);
        				
        				element.edges.add(edge);
        			}
        		}
        		
        		elements.add(element);
        	}
		}
		catch (FileNotFoundException fnfe)
		{
			System.out.println("ERROR: Couldn't find periodic table data file "+filename);
		}
		catch (NumberFormatException e)
		{
			System.out.println("ERROR: NumberFormatException while parsing periodic table data file "+filename);
		}
		catch (IOException e)
		{
			e.printStackTrace();
		}
		finally
		{
			if (null != dis) try
			{
				dis.close();
			}
			catch (IOException e)
			{
				e.printStackTrace();
			}
		}
	}
	
	@Override
	public void paintControl(PaintEvent pe)
	{
		pe.gc.drawImage(tableImage, pe.x, pe.y, pe.width, pe.height, pe.x, pe.y, pe.width, pe.height);
		
		// highlight the edge being hovered over
		if (null != selectedEdge && Client.isMaster())
		{
			pe.gc.setForeground(GUI.COLOR_BLACK);
			pe.gc.drawRectangle(selectedEdge.bounds.x-MARKER_HL_MARGIN, selectedEdge.bounds.y-MARKER_HL_MARGIN, selectedEdge.bounds.width+2*MARKER_HL_MARGIN, selectedEdge.bounds.height+2*MARKER_HL_MARGIN);
		}
	}
	
	protected void createTableImage()
	{
		if (null != tableImage)
		{
			Log.warn("Periodic table image already created");
			return;
		}
		
		tableImage = new Image(GUI.getDisplay(), TABLE_WIDTH, TABLE_HEIGHT);
		GC gc = new GC(tableImage);
		
		Rectangle bounds = tableImage.getBounds();
		int xcenter = bounds.width/2;

		// background
		gc.setBackground(GUI.COLOR_LIGHT_BLUE2);
		gc.fillRectangle(bounds);
		
		// header
		gc.setFont(GUI.FONT_PTBL_HEADER);
		String header = "Select an X-ray Absorption Edge";
		Point headerExtent = gc.textExtent(header);
		gc.drawText(header, xcenter - (headerExtent.x/2), 15);

		// edge label
		gc.setFont(GUI.FONT_PTBL_LABEL);
		String strEdge = new String("Elemental Edge: ");
		Point sEdgeExtent = gc.textExtent(strEdge);
		gc.drawText(strEdge, edgeTextPos.x - sEdgeExtent.x, edgeTextPos.y + (textFieldSize.y - sEdgeExtent.y)/2, true);

		// energy label
		String strEnergy = new String("Energy: ");
		Point sEnergyExtent = gc.textExtent(strEnergy);
		gc.drawText(strEnergy, energyTextPos.x - sEnergyExtent.x, energyTextPos.y + (textFieldSize.y - sEnergyExtent.y)/2, true);
		
		String strEnergyUnits = new String(" eV");
		Point strEnergyUnitsExtent = gc.textExtent(strEnergyUnits);
		gc.drawText(strEnergyUnits, energyTextPos.x + textFieldSize.x, energyTextPos.y + (textFieldSize.y - strEnergyUnitsExtent.y)/2, true);
		
		// elements
		for (Element element : elements)
		{
			int xcoord = 8 + 38*(element.col-1);
			int ycoord = 8 + 53*(element.row-1);
			
			if (element.row > 7)
			{
				xcoord += 5;
				ycoord += 5;
			}

			// fill the entire box
			gc.setBackground(GUI.COLOR_WHITE);
			gc.fillRectangle(xcoord, ycoord, 38, 53);
			
			// outline the box
			gc.setForeground(GUI.COLOR_PTBL_BOXOUTLINE);
			gc.drawRectangle(xcoord, ycoord, 38, 53);
			
			// background for atomicZ
			gc.setBackground(GUI.COLOR_PTBL_ATOMICZ_BG);
			gc.fillRectangle(xcoord+1, ycoord+1, 15, 14);
			
			// atomicZ text
			gc.setForeground(GUI.COLOR_DEFAULT_TEXT);
			gc.setFont(GUI.FONT_PTBL_TINY);
			gc.drawText(Integer.toString(element.atomicZ), xcoord+2, ycoord+2, true);

			// element text
			gc.setFont(GUI.FONT_PTBL_SMALL);
			gc.drawText(element.symbol, xcoord+18, ycoord+11, true); // changed xcoord+16 to xcoord+18

			// edges text
			if (null != element.edges)
			{
				for (Edge edge : element.edges)
				{
					int xin=0, yin=0;
					
					if (edge.type.equals("K"))
					{
						xin = 15;
						yin = 30;
					}
					else if (edge.type.equals("L1"))
					{
						xin = 1;
						yin = 13;
					}
					else if (edge.type.equals("L2"))
					{
						xin = 1;
						yin = 25;
					}
					else if (edge.type.equals("L3"))
					{
						xin = 1;
						yin = 37;
					}
					else if (edge.type.equals("M1"))
					{
						xin = 16;
						yin = 25;
					}
					else if (edge.type.equals("M2"))
					{
						xin = 16;
						yin = 37;
					}
					
					int x = xcoord+xin+2;
					int y = ycoord+yin+3;
					
					gc.setFont(GUI.FONT_PTBL_TINY);
					gc.setForeground(edge.enabled?GUI.COLOR_RED:GUI.COLOR_DARK_GRAY);
					gc.drawText(edge.type, x, y, true); // added +2 and +3 from SSRL
					
					Point extent = gc.textExtent(edge.type);
					
					EdgeMarker edgeMarker = new EdgeMarker(element.symbol, edge.type, x, y, x+extent.x, y+extent.y, edge.enabled);
					markerList.add(edgeMarker);
				}
			}
		}
		
		gc.dispose();
	}
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////// misc methods ///////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////////////
	
	public int getAtomicNumber(String elementString)
	{
		for (Element element : elements)
		{
			if (element.symbol.equals(elementString)) return element.atomicZ;
		}
		return 0;
	}
	
	public String getElement(int atomicNumber)
	{
		for (Element element : elements)
		{
			if (element.atomicZ == atomicNumber) return element.symbol;
		}
		return null;
	}
}
