/**
 * Handles bottom half of the screening tab including GUI, PV listeners and user input.
 * 
 * Starts with constructor with GUI elements.
 * Followed by PV listeners
 * Then some GUI helper functions
 * Finally the update helper functions
 * 
 * One monitor handles all PVs for this part of the screening tab.
 * 
 * ScreeningStatus PV changes cause updateButtons function to be called.  So starting the
 * screening operation will update that PV and disable the buttons.  ScreeningStatus
 * has 3 states: idle, running, or paused.
 * 
 * calls robot thread to move collimator, wait, then unload sample
 * 
 */
package gov.anl.gmca.screening.ui;

import gov.anl.gmca.common.*;
import gov.anl.gmca.common.AppState.Operation;
import gov.anl.gmca.common.EpicsDevice.State;
import gov.anl.gmca.common.ui.NBMessageBox;
import gov.anl.gmca.common.ui.WdgControlDir;
import gov.anl.gmca.common.ui.WdgControlDir.PrefixDir;
import gov.anl.gmca.common.ui.WdgControlEntry;
import gov.anl.gmca.common.ui.WdgControlEntry.EntryType;
import gov.anl.gmca.common.ui.WdgPositionDisplay;
import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.gui.Main;
import gov.anl.gmca.hutch.DevAttenuation;
import gov.anl.gmca.hutch.DevEnergy;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.hutch.ModelPmacMotor;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.*;
import gov.anl.gmca.raster.ModelRasterGrid;
import gov.anl.gmca.raster.ModelRasterState;
import gov.anl.gmca.raster.ModelRasterState.RasterListType;
import gov.anl.gmca.sample.DBBeamSize;
import gov.anl.gmca.sample.DBBeamSize.BeamSize;
import gov.anl.gmca.sample.DevCollimatorCtrl;
import gov.anl.gmca.sample.ModelBeamsize;
import gov.anl.gmca.screening.*;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.users.ModelUserTable.ClientListener;
import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.Point2D;
import gov.anl.gmca.util.Util;
import gov.anl.gmca.util.WidgetFactory;
import gov.aps.jca.dbr.DBR;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.CLabel;
import org.eclipse.swt.events.*;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.*;

import java.io.File;
import java.text.DecimalFormat;
import java.util.ArrayList;

public class WdgScreeningTasks extends Composite implements
		ListenerList.Listener<ModelScreeningTasks>, ClientListener {

	protected Text textDir, textPrefix, textDelta, textExpTime;
	protected FileDialog dialog;
	
	public final Button[] checkAction;
	public volatile Button[] startAction;
	protected Button directoryButton, prefixButton, readyButton, pauseButton,
						unmountButton, mountButton, updateButton;
	protected Button warmupButton, yagButton, pinButton,
						startButton, resumeButton, crystalparamsButton;
	//protected Button jpegButton;
	public static Button stopButton, collectPauseButton;
	protected Combo[] angleSpinner;
	protected Combo centerType, detectorDistanceCombo, attenuationCombo;

	protected Combo[] multipleImagesSpinner;
	protected Label robotCommand, puckLabel, commandLabel,labelRasterWarn;
	protected CLabel dirLabel;

	protected WdgControlDir rootDir;
	protected WdgControlEntry distanceEntry, beamstopEntry, attenuationEntry,
								axisEntry, deltaEntry, timeEntry;
	protected WdgControlEntry beamWidthEntry, beamHeightEntry;
	protected Combo beamsizeCombo;
	protected Group expDeltaGroup;
	
	double distance_mm, attenuation_factors;
	
	public static int globalScreeningStatus = 0;
	public static boolean takeJpeg = true;
	public static boolean incrementGroup = true;
	
	public static String globalPrefix = "";
	public static Thread screeningThread;
	
	protected boolean requestStop = false;
	private int highlightJpeg = 0;
	
	private CmdRobot cmdRobot = new CmdRobot();
	private TabScreening screeningTab = null;
	private Epics.EpicsValueMonitor monitor = null;

	public static String[] ActionNames = { "Mount Next Crystal",
			"Auto Centering", "Pause4Center", "Crystal Jpeg", "Collect Image", "Collect Image",
			"Collect Image", "Pause4Inspection", };

	public Double getDelta() {
		return (Double.parseDouble(textDelta.getText()));
	}

	public Double getExpTime() {
		return (Double.parseDouble(textExpTime.getText()));
	}

	public String getText() {
		return ModelScreening.currentRootDir;
	}

	public void setTaskImageNULL() {
		for (int i = 0; i < ActionNames.length; i++) {
			startAction[i].setImage(null);
		}
	}

	public void setTaskBackgroundNULL() {
		for (int i = 0; i < ActionNames.length; i++) {
			checkAction[i].setBackground(GUI.BGCOLOR);
		}
	}

	// QX this function does not appear to be used??
	public void setDir() {

		String dir = ModelScreening.currentRootDir;
		// remove an extra slash 
		int length = dir.length();

		if (length > 4) { // directory is starts /mnt 
			char ch = dir.charAt(length - 1);
			if (ch == '/') {
				dir = dir.substring(0, length - 1);
			}
		}
		puttoSpreadsheetPV(dir.replace("//", "/"));
		File dirFile = new File(dir);
		if (dirFile.exists() == false) {
			System.out.println("setDir mkdir is ignored.\n");
			//dirFile.mkdirs();
		}
	}

	public void setPrefix() {
		// String tprefix = textPrefix.getText();
		String tprefix = ModelScreening.currentPrefix;
		if (!tprefix.equals("") && !tprefix.equals(" ") && tprefix != null) {
			Epics.put(PV_RUN.fileroot.run(0), tprefix);
			Epics.put(PV_SCR_RUN.fileroot.run(0), tprefix);
			Epics.put(PV_SCR.prefix, tprefix);
			globalPrefix = tprefix;
			CmdScreening.group = 0;
			updateRuns();
		} else {
			Log.warn("Invalid prefix");
			rootDir.setPrefix(globalPrefix);
		}

	}

	public static void updateRuns() {

		Epics.putInteger(PV_RUN.updated.run(0), 0);
		Epics.putInteger(PV_RUN.updated.run(0), 1);

		Epics.putInteger(PV_SCR_RUN.updated.run(0), 0);
		Epics.putInteger(PV_SCR_RUN.updated.run(0), 1);

		Epics.putInteger(PV_SCAN.update, 0);
		Epics.putInteger(PV_SCAN.update, 1);
		//Util.sleep(100);
		Epics.putInteger(PV_R_RUN.updated.run(RasterListType.INTER, 0), 0);
		Epics.putInteger(PV_R_RUN.updated.run(RasterListType.INTER, 0), 1);

		Epics.putInteger(PV_R_RUN.updated.run(RasterListType.AUTO, 0), 0);
		Epics.putInteger(PV_R_RUN.updated.run(RasterListType.AUTO, 0), 1);
	}

	public void putDataCollectionParam(PVName pv, double val) {

		Epics.putDouble(pv, val);
		updateRuns();
	}

	public void updateValues() {

		double distance = Epics.getDouble(PV_PMAC.ActPos.motor("detector_z"));
		double attenuation = Epics.getDouble(PV_BLUICE.attenuation);
		final String distance_ = String.format("%.3f", distance);
		double gonio_temp = Epics
				.getDouble(PV_PMAC.ActPos.motor("gonio_omega"));
		DecimalFormat twoDForm = new DecimalFormat("#.##");
		final double gonio = Double.valueOf(twoDForm.format(gonio_temp));

		double high = Epics.getDouble(PV_PMAC.RqsPos_DRVH.motor("gonio_omega"));
		double low = Epics.getDouble(PV_PMAC.RqsPos_DRVL.motor("gonio_omega"));

		boolean updateGonio = true;
		/*if (gonio + 90 > high || gonio + 90 < low) {
			updateGonio = false;
			Log.warn("Omega value not with in limits " + low + " : " + high);
		}*/

		Util.sleep(10);
		if (updateGonio) {
			Epics.putDouble(PV_SCR.AngleImage0, gonio);
			Epics.putDouble(PV_SCR.AngleImage45, gonio + 45.0);
			Epics.putDouble(PV_SCR.AngleImage90, gonio + 90.0);
		}
		/* update beamsize */
		Point2D beamsize = new Point2D(0,0);
		if (Util.beamSizeCalculated())
		{
			int colli_um = DevCollimatorCtrl.getCollimatorSize().size_um;
			beamsize.x=beamsize.y=colli_um;
		} else {
			beamsize = ModelBeamsize.getCurrentBeamsize();
		}
		Epics.putInteger(PV_SCR_RUN.beamWidth_um.run(0), (int)(beamsize.x+.5));
		Epics.putInteger(PV_SCR_RUN.beamHeight_um.run(0), (int)(beamsize.y+.5));

		Epics.putDouble(PV_SCR_RUN.atten_factor.run(0), attenuation);
		putDataCollectionParam(PV_SCR_RUN.distance.run(0), distance);
		putDataCollectionParam(PV_RUN.distance.run(0), distance);
		final boolean updateFinal = updateGonio;
		GUI.asyncExec(new Runnable() {
			public void run() {
				distanceEntry.setText(distance_);
				if (updateFinal) {
					angleSpinner[0].setText(Double.toString(gonio));
					angleSpinner[1].setText(Double.toString(gonio + 45.0));
					angleSpinner[2].setText(Double.toString(gonio + 90.0));
				}
			}
		});
	}

	public WdgScreeningTasks(Composite parent, int style) {
		super(parent, style);
		Client.addListener(this);
		int i;
		screeningTab = (TabScreening) parent;
		setLayout(new GridLayout(3, true));
		setBackground(GUI.BGCOLOR);

		ModelScreeningTasks.get().ll.addListener(this);
		DevScreeningTasks.get().verifyConnected();

		Group parametersGroup = new Group(this, SWT.NONE);
		GridLayout paramGrid = new GridLayout(3, false);
		parametersGroup.setBackground(GUI.BGCOLOR);
		parametersGroup.setLayout(paramGrid);
		parametersGroup.setFont(GUI.FONT_COLLECT_GROUP);
		parametersGroup.setText("Collection Parameters");
		GridData gd = new GridData(SWT.FILL, SWT.FILL, true, true);

		parametersGroup.setLayoutData(gd);

		gd = new GridData(SWT.FILL, SWT.CENTER, true, false);
		gd.horizontalSpan = 3;
		rootDir = new WdgControlDir(parametersGroup, SWT.NONE, null, 0,
				new WdgControlDir.ModelInterface() {
					@Override
					public void setValue(Object value) {
						WdgControlDir.PrefixDir prefixDir = (WdgControlDir.PrefixDir) value;
						ModelScreening.currentRootDir = prefixDir.dir;
						ModelScreening.currentPrefix = prefixDir.prefix;
						setDir();
						setPrefix();
						System.out.println("rootDir object created. dir= " + prefixDir.dir + " prefix= " + prefixDir.prefix);
					}

					@Override
					public PrefixDir getPrefixDir() {
						return new PrefixDir(ModelScreening.currentPrefix, ModelScreening.currentRootDir);
					}
				});
		rootDir.setLayoutData(gd);

		dirLabel = new CLabel(parametersGroup, SWT.BORDER | SWT.WRAP);
		gd = new GridData(SWT.FILL, SWT.DEFAULT, true, false);
		gd.horizontalSpan = 3;
		dirLabel.setText("Image Dir: ");
		dirLabel.setLayoutData(gd);

		Object[][] statList = { { "Gonio", DevPmacMotor.get("gonio_omega") },
				{ "Detector", DevPmacMotor.get("detector_z") },
				{ "Attenuation", DevAttenuation.get() } };

		WdgPositionDisplay positionDisplay = new WdgPositionDisplay(
				parametersGroup, SWT.NONE, "Current position", statList,
				GUI.FONT_COLLECT_GROUP);
		positionDisplay.setBackground(GUI.BGCOLOR);

		gd = new GridData(SWT.FILL, SWT.DEFAULT, true, false);
		gd.horizontalSpan = 2;
		positionDisplay.setLayoutData(gd);

		gd = new GridData(SWT.DEFAULT, SWT.CENTER, false, false);
		gd.minimumWidth = 80;
		updateButton = new Button(parametersGroup, SWT.NONE);
		updateButton.setText("Update");
		updateButton.setBackground(GUI.COLOR_BLUE_BUTTON);
		updateButton
				.setToolTipText("Updates to current value of detector, attenuation, beamsize \ncollect image angles : current gonio position, +45, +90");
		updateButton.setLayoutData(gd);
		updateButton.setEnabled(false);
		updateButton.addSelectionListener(new SelectionListener() {

			@Override
			public void widgetDefaultSelected(SelectionEvent arg0) {
			}

			public void widgetSelected(SelectionEvent arg0) {
				updateValues();
			}

		});

		expDeltaGroup = new Group(parametersGroup, SWT.NONE);
		expDeltaGroup.setBackground(GUI.BGCOLOR);
		expDeltaGroup.setText("Settings ");
		gd = new GridData(SWT.FILL, SWT.CENTER, true, true);
		gd.horizontalSpan = 3;

		expDeltaGroup.setLayoutData(gd);
		expDeltaGroup.setLayout(new GridLayout(6, false));

		Label detectorLabel = new Label(expDeltaGroup, SWT.NONE);
		gd = new GridData(SWT.FILL, SWT.CENTER, true, false);
		gd.horizontalSpan = 2;

		detectorLabel.setBackground(GUI.BGCOLOR);
		detectorLabel.setText("Distance(mm):");
		detectorLabel.setLayoutData(gd);

		gd = new GridData(SWT.FILL, SWT.FILL, true, true);
		gd.widthHint = 130;
		gd.horizontalSpan = 4;
		distanceEntry = new WdgControlEntry(expDeltaGroup, SWT.NONE,
				EntryType.RW_COMBO, new WdgControlEntry.ModelInterface() {
					@Override
					public void setValue(Object value) {
						String distanceStr = distanceEntry.getText();
						double distance = 0.0;
						if (!distanceStr.equals("") || distanceStr != null) {
							distance = Double.parseDouble(distanceStr);

							putDataCollectionParam(PV_RUN.distance.run(0),
									distance);
							putDataCollectionParam(PV_SCR_RUN.distance.run(0),
									distance);
						}
					}

					@Override
					public Object getModelValue() {
						return Epics.getDouble(PV_SCR_RUN.distance.run(0));
					}

					@Override
					public EpicsDevice getDevice() {
						return DevPmacMotor.get("detector_z");
					}

					@Override
					public Object getDeviceValue() {
						return DevPmacMotor.get("detector_z").getPosition();
					}

					@Override
					public Limits getLimits() {
						return DevPmacMotor.get("detector_z").getLimits();
					}

					@Override
					public int getDecimals() {
						return 3;
					}

					@Override
					public double[] getFloatChoices() {
						return ModelPmacMotor.getDetectorDistanceOptions();
					}

					@Override
					public boolean doHighlight(double d) {
						return Math.abs((Double) getDeviceValue() - d) > 0.01;
					}
				});
		distanceEntry.setLayoutData(gd);

		Label lblDelta = new Label(expDeltaGroup, SWT.NONE);
		lblDelta.setBackground(GUI.BGCOLOR);
		lblDelta.setText("Delta(deg):");
		gd = new GridData(SWT.FILL, SWT.CENTER, true, false);
		gd.horizontalSpan = 2;
		lblDelta.setLayoutData(gd);

		gd = new GridData(SWT.FILL, SWT.FILL, true, true);

		deltaEntry = new WdgControlEntry(expDeltaGroup, SWT.NONE,
				EntryType.TEXT, new WdgControlEntry.ModelInterface() {
					@Override
					public void setValue(Object value) {
						String deltaStr = deltaEntry.getText();
						double delta = 0.0;
						if (!deltaStr.equals("") || deltaStr != null) {
							try {
								delta = Double.parseDouble(deltaStr);

							} catch (Exception e) {
								e.printStackTrace();
							}
							putDataCollectionParam(PV_RUN.delta.run(0), delta);
							putDataCollectionParam(PV_SCR_RUN.delta.run(0),
									delta);
						}
					}

					@Override
					public Object getModelValue() {
						return Epics.getDouble(PV_SCR_RUN.delta.run(0));
					}

					@Override
					public Limits getLimits() {
						return new Limits(Epics.pvs_getDouble(Epics
								.addBID("bi:r0:delta.DRVL")),
								Epics.pvs_getDouble(Epics
										.addBID("bi:r0:delta.DRVH")), "delta");
					}

					@Override
					public int getDecimals() {
						return 2;
					}
				});
		deltaEntry.setLayoutData(gd);

		Label lblExpTime = new Label(expDeltaGroup, SWT.NONE);
		lblExpTime.setBackground(GUI.BGCOLOR);
		lblExpTime.setText("Time(sec):");
		gd = new GridData(SWT.FILL, SWT.CENTER, true, false);
		gd.horizontalSpan = 2;
		lblExpTime.setLayoutData(gd);

		gd = new GridData(SWT.FILL, SWT.FILL, true, true);

		timeEntry = new WdgControlEntry(expDeltaGroup, SWT.NONE,
				EntryType.TEXT, new WdgControlEntry.ModelInterface() {
					@Override
					public void setValue(Object value) {
						String expTimeStr = timeEntry.getText();
						double expTime = 0.0;
						if (!expTimeStr.equals("") || expTimeStr != null) {
							try {
								expTime = Double.parseDouble(expTimeStr);
							} catch (Exception e) {
								e.printStackTrace();
							}
							putDataCollectionParam(PV_RUN.exposureTime.run(0),
									expTime);
							putDataCollectionParam(
									PV_SCR_RUN.exposureTime.run(0), expTime);
						}
					}

					@Override
					public Object getModelValue() {
						return Epics.getDouble(PV_SCR_RUN.exposureTime.run(0));
					}

					@Override
					public Limits getLimits() {
						return new Limits(Epics.pvs_getDouble(Epics
								.addBID("bi:r0:exposureTime.DRVL")),
								Epics.pvs_getDouble(Epics
										.addBID("bi:r0:exposureTime.DRVH")),
								"exposure time");
					}

					@Override
					public int getDecimals() {
						return 2;
					}
				});
		timeEntry.setLayoutData(gd);

		Label lblAttenuation = new Label(expDeltaGroup, SWT.NONE);
		lblAttenuation.setBackground(GUI.BGCOLOR);
		lblAttenuation.setText("Attenuation(factors):");
		gd = new GridData(SWT.FILL, SWT.CENTER, true, false);

		gd.horizontalSpan = 3;
		lblAttenuation.setLayoutData(gd);

		gd = new GridData(SWT.FILL, SWT.FILL, true, true);
		gd.widthHint = 130;
		gd.horizontalSpan = 3;
		attenuationEntry = new WdgControlEntry(expDeltaGroup, SWT.NONE,
				EntryType.RW_COMBO, new WdgControlEntry.ModelInterface() {
					@Override
					public void setValue(Object value) {
						String attenuationStr = attenuationEntry.getText();
						double attenuation = 0.0;
						if (!attenuationStr.equals("")
								|| attenuationStr != null) {
							try {
								attenuation = Double
										.parseDouble(attenuationStr);

							} catch (Exception e) {
								e.printStackTrace();
							}
							putDataCollectionParam(PV_RUN.atten_factor.run(0),
									attenuation);
							putDataCollectionParam(
									PV_SCR_RUN.atten_factor.run(0), attenuation);
						}
					}

					@Override
					public Object getModelValue() {
						return Epics.getDouble(PV_SCR_RUN.atten_factor.run(0));
					}

					@Override
					public EpicsDevice getDevice() {
						return DevAttenuation.get();
					}

					@Override
					public Object getDeviceValue() {
						return DevAttenuation.get().getPosition();
					}

					@Override
					public Limits getLimits() {
						return DevAttenuation.get().getLimits();
					}

					@Override
					public int getDecimals() {
						return 2;
					}

					@Override
					public double[] getFloatChoices() {
						return new double[] { 1, 5, 10, 20, 50, 100, 1000 };
					}

					@Override
					public boolean doHighlight(double w) {
						double d = (Double) getDeviceValue();
						return (Math.abs(w - d) / d) > 0.05;
					}
				});
		attenuationEntry.setLayoutData(gd);

		/* beam size */
		if (Util.beamSizeCalculated())
			createBeamSizeDropdown(expDeltaGroup);
		else
			createBeamSizeFields(expDeltaGroup);

		final Group tasksGroup = new Group(this, SWT.NONE);
		tasksGroup.setBackground(GUI.BGCOLOR);
		tasksGroup.setFont(GUI.FONT_COLLECT_GROUP);
		GridLayout actionGrid = new GridLayout(4, false);
		tasksGroup.setLayout(actionGrid);

		gd = new GridData(SWT.CENTER, SWT.FILL, true, true);

		tasksGroup.setLayoutData(gd);
		tasksGroup.setText("Tasks");

		startAction = new Button[ActionNames.length];
		checkAction = new Button[ActionNames.length];
		angleSpinner = new Combo[3];
		multipleImagesSpinner = new Combo[3];
		String[] angles = { "0.0", "45.0", "90.0", "135.0", "180.0", "225.0",
				"270.0", "315.0" };
		String[] imagesets = { "1", "2", "3", "4", "5", "6", "7", "8", "9", "10" };
		String[] centeringType = { "Loop", "Crystal"/*, "Raster"*/ };

		for (i = 0; i < ActionNames.length; i++) {
			gd = new GridData();
			gd.widthHint = 45;
			startAction[i] = new Button(tasksGroup, SWT.RADIO);
			startAction[i].setBackground(GUI.BGCOLOR);
			startAction[i].setText("");
			final Button tempButton = startAction[i];
			if (i == 0)
				startAction[0].setImage(GUI.image);

			final int start = i;
			startAction[i].addMouseListener(new MouseListener() {

				public void mouseDoubleClick(MouseEvent arg0) {
					setTaskImageNULL();
					tempButton.setImage(GUI.image);
					Epics.put(PV_SCR.CurrentTask, "Recollect " + start);
					if (start < 3)
						incrementGroup = false;
					Util.sleep(10);
				}

				public void mouseDown(MouseEvent arg0) {
				}

				public void mouseUp(MouseEvent arg0) {
					setTaskImageNULL();
					tempButton.setImage(GUI.image);
					Epics.put(PV_SCR.CurrentTask, "Recollect " + start);
					if (start < 3)
						incrementGroup = false;
					Util.sleep(10);
				}

			});

			startAction[i].setLayoutData(gd);
			gd = new GridData();

			if (i < 4 || i > 6) {
				if (i != 1)
					gd.horizontalSpan = 3;
			}
			checkAction[i] = new Button(tasksGroup, SWT.CHECK);
			checkAction[i].setBackground(GUI.BGCOLOR);
			checkAction[i].setText(ActionNames[i]);
			checkAction[i].setLayoutData(gd);
			final int temp = i;

			checkAction[i].addSelectionListener(new SelectionListener() {

				@Override
				public void widgetDefaultSelected(SelectionEvent arg0) {
				}

				public void widgetSelected(SelectionEvent arg0) {

					boolean checked = checkAction[temp].getSelection();
					PVName taskPV = null;
					if (temp == 0) {
						taskPV = PV_SCR.MountNextXtal;
					} else if (temp == 1) {
						taskPV = PV_SCR.CenterXtal;
					} else if (temp == 2) {
						taskPV = PV_SCR.Pause4ctrng;
					} else if (temp == 3) {
						taskPV = PV_SCR.takeJpeg;
					} else if (temp == 4) {
						taskPV = PV_SCR.TakeImage0;
					} else if (temp == 5) {
						taskPV = PV_SCR.TakeImage45;
					} else if (temp == 6) {
						taskPV = PV_SCR.TakeImage90;
					} else if (temp == 7) {
						taskPV = PV_SCR.Pause4collect;
					} else {
						return;
					}
					if (checked)
						Epics.putInteger(taskPV, 1);
					else
						Epics.putInteger(taskPV, 0);
				}
			});

			checkAction[0].setSelection(true);
			checkAction[0].setEnabled(false);
			Epics.putInteger(PV_SCR.MountNextXtal, 1);
			if (i == 1) {

				gd = new GridData(SWT.FILL, SWT.CENTER, true, false);
				gd.minimumWidth = 85;

				centerType = new Combo(tasksGroup, SWT.READ_ONLY);
				centerType.setItems(centeringType);
				centerType.setText("Loop");
				centerType.setEnabled(false);
				centerType.setLayoutData(gd);
				centerType.addSelectionListener(new SelectionListener() {

					public void widgetDefaultSelected(SelectionEvent arg0) {
					}

					public void widgetSelected(SelectionEvent arg0) {
						Log.note("Centering option changeed to : "
								+ centerType.getSelection());
						Epics.putInteger(PV_SCR.centerType,
								centerType.getSelectionIndex());
					}
				});
				gd = new GridData(SWT.FILL, SWT.DEFAULT, true, false);
				gd.minimumWidth = 40;
				stopButton = new Button(tasksGroup, SWT.NONE);
				stopButton.setEnabled(false);
				stopButton.setImage(GUI.STOP_SIGN);
				stopButton.setLayoutData(gd);
				stopButton.addSelectionListener(new SelectionListener() {

					public void widgetDefaultSelected(SelectionEvent arg0) {
					}

					public void widgetSelected(SelectionEvent arg0) {
						if (Client.isMaster()) {
							if (ModelScreening.centerType.equals("LOOP")
									|| ModelScreening.centerType
											.equals("CRYSTAL"))
								Epics.putInteger(PV_BLUICE.autoCenterRun, 0);
							else
								ModelRasterState.get().pause();
						}
					}
				});
			}
			/*if (i == 2) {
				gd = new GridData();

				gd.horizontalSpan = 4;
				gd.horizontalIndent = 50;
				jpegButton = new Button(tasksGroup, SWT.CHECK | SWT.BORDER);
				jpegButton.setLayoutData(gd);
				jpegButton.setText("Jpeg");
				jpegButton.setBackground(GUI.BGCOLOR);
				boolean isJpeg = Epics.getBoolean(PV_SCR.takeJpeg);
				jpegButton.setSelection(isJpeg);
				jpegButton.addSelectionListener(new SelectionListener() {

					@Override
					public void widgetDefaultSelected(SelectionEvent arg0) {
					}

					public void widgetSelected(SelectionEvent arg0) {

						Epics.putInteger(PV_SCR.takeJpeg,
								jpegButton.getSelection() ? 1 : 0);
						screeningTab.updateTaskProgress();

					}
				});
			}*/
			if (i == 4) {
				gd = new GridData();
				gd.widthHint = 85;
				angleSpinner[0] = new Combo(tasksGroup, SWT.BORDER);
				angleSpinner[0].setLayoutData(gd);
				angleSpinner[0].setItems(angles);
				angleSpinner[0].addFocusListener(new FocusListener() {
					public void focusGained(FocusEvent arg0) {
					}

					public void focusLost(FocusEvent arg0) {
						setAngle(0);
						screeningTab.updateTaskProgress();
					}
				});
				angleSpinner[0].addListener(SWT.Verify, new Listener() {
					public void handleEvent(Event e) {
						Util.allowOnlyAlpha(e);
					}
				});
				angleSpinner[0].addSelectionListener(new SelectionListener() {
					public void widgetDefaultSelected(SelectionEvent arg0) {
					}

					public void widgetSelected(SelectionEvent arg0) {
						setAngle(0);
						screeningTab.updateTaskProgress();
					}
				});

				gd = new GridData(SWT.FILL, SWT.DEFAULT, true, false);
				gd.minimumWidth = 40;
				multipleImagesSpinner[0] = new Combo(tasksGroup, SWT.BORDER);
				multipleImagesSpinner[0].setText("1");
				multipleImagesSpinner[0].setItems(imagesets);
				multipleImagesSpinner[0].setLayoutData(gd);
				multipleImagesSpinner[0]
						.addSelectionListener(new SelectionListener() {

							@Override
							public void widgetDefaultSelected(
									SelectionEvent arg0) {
							}

							public void widgetSelected(SelectionEvent arg0) {
								
								int noofimages = 1;
								try {
									noofimages = Integer.parseInt(multipleImagesSpinner[0].getText());
								} catch (NumberFormatException e) {
									e.printStackTrace();
								}
								DevScreeningTasks.setDiffImages(0, noofimages);
								screeningTab.updateTaskProgress();
							}

						});
			}
			
			if (i == 5) {
				gd = new GridData();
				gd.widthHint = 85;

				angleSpinner[1] = new Combo(tasksGroup, SWT.BORDER);
				angleSpinner[1].setLayoutData(gd);
				angleSpinner[1].setItems(angles);
				angleSpinner[1].addFocusListener(new FocusListener() {

					@Override
					public void focusGained(FocusEvent arg0) {
					}

					public void focusLost(FocusEvent arg0) {
						setAngle(1);
						screeningTab.updateTaskProgress();
					}
				});
				angleSpinner[1].addListener(SWT.Verify, new Listener() {
					public void handleEvent(Event e) {
						Util.allowOnlyAlpha(e);
					}
				});
				angleSpinner[1].addSelectionListener(new SelectionListener() {

					@Override
					public void widgetDefaultSelected(SelectionEvent arg0) {
					}

					public void widgetSelected(SelectionEvent arg0) {
						setAngle(1);
						screeningTab.updateTaskProgress();
					}

				});
				gd = new GridData(SWT.FILL, SWT.DEFAULT, true, false);
				gd.minimumWidth = 40;
				multipleImagesSpinner[1] = new Combo(tasksGroup, SWT.BORDER);
				multipleImagesSpinner[1].setText("1");
				multipleImagesSpinner[1].setItems(imagesets);
				multipleImagesSpinner[1].setLayoutData(gd);

				multipleImagesSpinner[1]
						.addSelectionListener(new SelectionListener() {

							@Override
							public void widgetDefaultSelected(
									SelectionEvent arg0) {
							}

							public void widgetSelected(SelectionEvent arg0) {

								int noofimages = 1;
								try {
									noofimages = Integer.parseInt(multipleImagesSpinner[1].getText());
								} catch (NumberFormatException e) {
									e.printStackTrace();
								}
								DevScreeningTasks.setDiffImages(1, noofimages);
								screeningTab.updateTaskProgress();
							}

						});
			}
			
			if (i == 6) {
				gd = new GridData();
				gd.widthHint = 85;
				angleSpinner[2] = new Combo(tasksGroup, SWT.BORDER);
				angleSpinner[2].setLayoutData(gd);
				angleSpinner[2].setItems(angles);
				angleSpinner[2].addFocusListener(new FocusListener() {
					public void focusGained(FocusEvent arg0) {
					}

					public void focusLost(FocusEvent arg0) {
						setAngle(2);
						screeningTab.updateTaskProgress();
					}
				});
				angleSpinner[2].addListener(SWT.Verify, new Listener() {
					public void handleEvent(Event e) {
						Util.allowOnlyAlpha(e);
					}
				});
				angleSpinner[2].addSelectionListener(new SelectionListener() {
					public void widgetDefaultSelected(SelectionEvent arg0) {
					}

					public void widgetSelected(SelectionEvent arg0) {
						setAngle(2);
						screeningTab.updateTaskProgress();
					}
				});

				gd = new GridData(SWT.FILL, SWT.DEFAULT, true, false);
				gd.minimumWidth = 40;
				multipleImagesSpinner[2] = new Combo(tasksGroup, SWT.BORDER);
				multipleImagesSpinner[2].setItems(imagesets);
				multipleImagesSpinner[0].setText("1");
				multipleImagesSpinner[2].setLayoutData(gd);
				multipleImagesSpinner[2]
						.addSelectionListener(new SelectionListener() {

							@Override
							public void widgetDefaultSelected(
									SelectionEvent arg0) {
							}

							public void widgetSelected(SelectionEvent arg0) {
								int noofimages = 1;
								try {
									noofimages = Integer.parseInt(multipleImagesSpinner[2].getText());
								} catch (NumberFormatException e) {
									e.printStackTrace();
								}
								DevScreeningTasks.setDiffImages(2, noofimages);
								screeningTab.updateTaskProgress();
							}

						});
			}
		}

		Group screenButtonGroup = new Group(tasksGroup, SWT.NONE);
		GridLayout gs = new GridLayout(2, false);
		screenButtonGroup.setBackground(GUI.BGCOLOR);
		screenButtonGroup.setLayout(gs);

		gd = new GridData();
		gd.horizontalSpan = 4;
		gd.horizontalAlignment = SWT.CENTER;
		screenButtonGroup.setLayoutData(gd);

		gd = new GridData();
		gd.widthHint = 95;
		startButton = new Button(screenButtonGroup, SWT.NONE);
		startButton.setBackground(GUI.COLOR_BLUE_BUTTON);
		startButton.setFont(GUI.FONT_HUTCH_BUTTON);
		startButton.setLocation(30, 20);
		startButton.setText("Start");
		startButton.setSize(95, 40);
		startButton.setLayoutData(gd);

		startButton.addSelectionListener(new SelectionListener() {

			@Override
			public void widgetDefaultSelected(SelectionEvent arg0) {
			}

			public void widgetSelected(SelectionEvent arg0) {
				System.out.println("Start button clicked by user on screening tab");
				if (ModelScreening.currentTaskStr.contains("Collect"))
					ModelScreening.pauseCollect = true;
				if (startButton.getText().equals("Start"))
					TabScreening.startScreening();
				else {
					startButton.setText("Pausing..");
					startButton.setEnabled(false);
					// CmdScreening.stopThread = true;
				}
			}
		});
	
		
		gd = new GridData();
		gd.widthHint = 95;
		resumeButton = new Button(screenButtonGroup, SWT.NONE);
		resumeButton.setBackground(GUI.COLOR_BLUE_BUTTON);
		resumeButton.setFont(GUI.FONT_HUTCH_BUTTON);
		resumeButton.setText("Resume");
		resumeButton.setLocation(125, 20);
		resumeButton.setSize(95, 40);
		resumeButton.setEnabled(false);
		resumeButton.setLayoutData(gd);
		resumeButton.addSelectionListener(new SelectionListener() {
			public void widgetDefaultSelected(SelectionEvent arg0) {
			}

			public void widgetSelected(SelectionEvent arg0) {

				System.out.println("Resume button clicked by user on screening tab");
				if(CmdScreening.manualRepeatFlag > 0) {
					
					if(incrementGroup) {
						CmdScreening.group++;
						incrementGroup = true;
					}
				} 
				if (!CmdScreening.stopThread) {
					Epics.putInteger(PV_SCR.ScreeningStatus, 1);
					if (null == screeningThread || !screeningThread.isAlive()) {
						TabScreening.startScreening();
					}
				} else {
					TabScreening.startScreening();
				}
			}
		});

		gd = new GridData();
		gd.horizontalSpan = 4;
		gd.horizontalAlignment = SWT.CENTER;
		Link link = new Link(tasksGroup, SWT.NONE);
		link.setBackground(tasksGroup.getBackground());
		link.setText("<a>Crystal score formula and color code help</a>");
		link.setLayoutData(gd);

		link.addSelectionListener(new SelectionListener() {

			@Override
			public void widgetDefaultSelected(SelectionEvent arg0) {
			}

			@Override
			public void widgetSelected(SelectionEvent arg0) {
				// TODO Auto-generated method stub
				WdgColorCodeExplanation dlgColorcode = new WdgColorCodeExplanation(
						GUI.getShell(), SWT.NONE, "");
				if (dlgColorcode.open(0) == SWT.CANCEL)
					return;
			}
		});

		gd = new GridData(SWT.CENTER, SWT.CENTER, true, true);
		gd.horizontalSpan = 4;
		gd.minimumWidth = 200;

		labelRasterWarn = new Label(tasksGroup, SWT.NONE);
		labelRasterWarn.setBackground(GUI.BGCOLOR);
		labelRasterWarn.setForeground(GUI.COLOR_RED);
		labelRasterWarn.setLayoutData(gd);

		Group robotGroup = new Group(this, SWT.NONE);
		robotGroup.setSize(parametersGroup.getSize());
		GridLayout taskGrid = new GridLayout(2, true);
		robotGroup.setBackground(GUI.BGCOLOR);
		robotGroup.setFont(GUI.FONT_COLLECT_GROUP);
		robotGroup.setLayout(taskGrid);
		robotGroup.setText("Robot");

		gd = new GridData(SWT.FILL, SWT.FILL, true, false);
		robotGroup.setLayoutData(gd);

		gd = new GridData(SWT.FILL, SWT.FILL, true, true);
		readyButton = new Button(robotGroup, SWT.NONE);
		readyButton.setFont(GUI.FONT_HUTCH_BUTTON);
		readyButton.setText("Ready");
		readyButton.setLayoutData(gd);
		
		//readyButton.setImage(GUI.readyImage);

		readyButton.addSelectionListener(new SelectionListener() {

			@Override
			public void widgetDefaultSelected(SelectionEvent arg0) {
			}

			public void widgetSelected(SelectionEvent arg0) {
				NBMessageBox messageBox = new NBMessageBox(GUI.getShell(),
						SWT.OK | SWT.CANCEL | SWT.ICON_WARNING);
				messageBox.setMessage("It is unsafe to change the status to 'Ready' if a pin\n"
								+ " has been mounted manually or the automounter is no longer aligned.\n"
								+ "Click OK if it is safe to proceed.");
				if (messageBox.open() == SWT.OK) {
					Log.note("Robot status : Ready");
					System.out.println("Ready button clicked by user on screening tab");
					cmdRobot.setRobotStatus(0);
				}
			}
		});

		gd = new GridData(SWT.FILL, SWT.FILL, true, false);
		pauseButton = new Button(robotGroup, SWT.NONE);
		pauseButton.setFont(GUI.FONT_HUTCH_BUTTON);
		pauseButton.setText("Pause");
		pauseButton.setLayoutData(gd);
		
		//pauseButton.setImage(GUI.pauseImage);
		pauseButton.addSelectionListener(new SelectionListener() {

			@Override
			public void widgetDefaultSelected(SelectionEvent arg0) {
			}

			public void widgetSelected(SelectionEvent arg0) {
				// TODO Auto-generated method stub
				System.out
						.println("Pause button clicked by user on screening tab");
				cmdRobot.setRobotStatus(1);
			}
		});

		gd = new GridData(SWT.FILL, SWT.FILL, true, true);
		gd.minimumWidth = 135;
		gd.horizontalAlignment = GridData.CENTER;
		unmountButton = new Button(robotGroup, SWT.NONE);
		unmountButton.setSize(200, 40);
		unmountButton.setBackground(GUI.COLOR_BLUE_BUTTON);
		unmountButton.setText("Unmount");
		unmountButton.setLayoutData(gd);
		unmountButton.addSelectionListener(new SelectionListener() {

			@Override
			public void widgetDefaultSelected(SelectionEvent arg0) {
			}

			public void widgetSelected(SelectionEvent arg0) {

				if (!checkRobotReady())
					return;

				String mounted_ = Epics.getString(PV_ROBOT.mounted);
				if (mounted_.equals("none")) {
					Log.error("No sample mounted with Robot");
					return;
				}

				unmountButton.setEnabled(false);
				CmdRobot cmdRobo = new CmdRobot("unload");
				Thread unmountThread = new Thread(cmdRobo);
				unmountThread.start();
			}
		});

		gd = new GridData(SWT.FILL, SWT.FILL, true, true);
		gd.horizontalAlignment = GridData.CENTER;
		warmupButton = new Button(robotGroup, SWT.NONE);
		warmupButton.setBackground(GUI.COLOR_BLUE_BUTTON);
		warmupButton.setText("Gripper Warmup");
		warmupButton.setLayoutData(gd);
		warmupButton.addSelectionListener(new SelectionListener() {

			@Override
			public void widgetDefaultSelected(SelectionEvent arg0) {
			}

			public void widgetSelected(SelectionEvent arg0) {
				if (!checkRobotReady())
					return;
				Epics.put(PV_ROBOT.cmd, "warm-up");
			}
		});

		new WdgRobotStatus(robotGroup, SWT.NONE);

		Group commandGroup = new Group(robotGroup, SWT.NONE);
		commandGroup.setBackground(GUI.BGCOLOR);
		GridLayout mountedGrid = new GridLayout(4, false);
		gd = new GridData(SWT.FILL, SWT.DEFAULT, true, false);
		gd.horizontalSpan = 4;
		gd.verticalSpan = 10;

		commandGroup.setLayout(mountedGrid);
		commandGroup.setLayoutData(gd);

		gd = new GridData(SWT.LEFT, SWT.CENTER, false, false);
		gd.horizontalSpan = 1;
		Label labelCommand = new Label(commandGroup, SWT.NONE);
		labelCommand.setFont(GUI.FONT_COLLECT_BUTTON2);
		labelCommand.setBackground(GUI.BGCOLOR);

		labelCommand.setText("Command : ");
		labelCommand.setLayoutData(gd);

		gd = new GridData(SWT.LEFT, SWT.CENTER, false, false);
		gd.horizontalSpan = 3;
		gd.widthHint = 130;
		robotCommand = new Label(commandGroup, SWT.NONE);
		robotCommand.setFont(GUI.FONT_14_BOLD);
		robotCommand.setForeground(GUI.COLOR_BLUE);
		robotCommand.setBackground(GUI.BGCOLOR);

		String mounted = Epics.getString(PV_ROBOT.mounted);

		robotCommand.setText(mounted);
		robotCommand.setLayoutData(gd);

		/* Crystal wash */
		Group washGroup = new Group(robotGroup, SWT.NONE);
		washGroup.setBackground(GUI.BGCOLOR);
		washGroup.setLayout(new GridLayout(3, false));
		washGroup.setText("Wash/Rinse");
		gd = new GridData(SWT.FILL, SWT.DEFAULT, true, false);

		gd.horizontalSpan = 2;
		washGroup.setLayoutData(gd);

		new WdgRinseWash(washGroup, style);

		// TODO: (optional) this object can be broken up in to individual
		// monitors for each PV specified below -MH
		monitor = new Epics.EpicsValueMonitor() {
			@Override
			public void valueChanged(String pvName, DBR value) {
				int i = 0;
				int temp = 0;
				final String _pvName = pvName;
				final DBR _value = value;
				String status;

				if (_pvName.equals(PV_BLUICE.autoCenterRun.toString())) {
					final int autoRunning = Epics.dbrToInt(_value);
					GUI.asyncExec(new Runnable() {
						public void run() {
							if (autoRunning == 0) {
								if (CmdScreening.runningTask != null) {
									CmdScreening.runningTask.asyncStop();
									stopButton.setEnabled(false);
								}
							}
						}
					});
				}
				if (_pvName.equals(PV_ROBOT.cmd.toString())) {
					final String mounted = Epics.dbrToString(_value);
					GUI.asyncExec(new Runnable() {
						public void run() {
							robotCommand.setText(mounted);
						}
					});
				}
				if (_pvName.equals(PV_SCR.prefix.toString())) {
					final String prefix = ValidateDirPrefix.validateFileName(Epics.dbrToString(_value));
					
					GUI.asyncExec(new Runnable() {
						public void run() {
							String prefixfinal = prefix;
							if (prefix.endsWith("_0")) {
								prefixfinal = prefix.replace("_0", "_1");
								rootDir.setPrefix(prefixfinal);
								ModelScreening.preIncrement = false;
								Epics.put(PV_SCR.prefix, prefixfinal);
							} else {
								rootDir.setPrefix(prefix);
							}
							
							ModelScreening.currentPrefix = prefixfinal;
						}
					});
				}
				if (_pvName.equals(PV_SCR.AngleImage0.toString())) {
					GUI.asyncExec(new Runnable() {
						public void run() {
							double angle = Epics.dbrToDouble(_value);
							angleSpinner[0].setText(Double.toString(angle));
							screeningTab.updateTaskProgress();
						}
					});
				}
				if (_pvName.equals(PV_SCR.AngleImage45.toString())) {

					GUI.asyncExec(new Runnable() {
						public void run() {
							double angle = Epics.dbrToDouble(_value);
							angleSpinner[1].setText(Double.toString(angle));
							screeningTab.updateTaskProgress();
						}
					});
				}
				if (_pvName.equals(PV_SCR.AngleImage90.toString())) {
					GUI.asyncExec(new Runnable() {
						public void run() {
							double angle = Epics.dbrToDouble(_value);
							angleSpinner[2].setText(Double.toString(angle));
							screeningTab.updateTaskProgress();
						}
					});
				}
				if (_pvName.equals(PV_SCR_RUN.distance.run(0).toString())) {
					double distance = Epics.getDouble(PV_SCR_RUN.distance
							.run(0));
					final String distance_ = String.format("%.3f", distance);
					GUI.asyncExec(new Runnable() {
						public void run() {
							ModelScreeningTasks.detectorDistance = Double
									.parseDouble(distance_);
							distanceEntry.setText(distance_);
						}
					});
				}
				if (_pvName.equals(PV_SCR_RUN.delta.run(0).toString())) {

					final double delta = Epics.dbrToDouble(_value);
					GUI.asyncExec(new Runnable() {
						public void run() {
							ModelScreeningTasks.delta = delta;
							deltaEntry.setText(Double.toString(delta));
						}
					});
				}
				if (_pvName.equals(PV_SCR_RUN.exposureTime.run(0).toString())) {
					final double expTime = Epics.dbrToDouble(_value);
					GUI.asyncExec(new Runnable() {
						public void run() {
							ModelScreeningTasks.exposureTime = expTime;
							timeEntry.setText(Double.toString(expTime));
						}
					});
				}

				if (_pvName.equals(PV_SCR_RUN.atten_factor.run(0).toString())) {
					double attenuation_double = Epics.dbrToDouble(_value);
					final String attenuation = String.format("%.2f",
							attenuation_double);
					GUI.asyncExec(new Runnable() {
						public void run() {
							ModelScreeningTasks.attenuation = Double
									.parseDouble(attenuation);
							attenuationEntry.setText(attenuation);
						}
					});
				}

				if (_pvName.equals(PV_SCR_RUN.beamWidth_um.run(0).toString())) {
					final int beamWidth = Epics.dbrToInt(_value);
					GUI.asyncExec(new Runnable() {
						public void run() {
							if (null!=beamWidthEntry) beamWidthEntry.setText(Integer.toString(beamWidth));
							if (null!=beamsizeCombo) updateBeamSizeSelection();
						}
					});
				}

				if (_pvName.equals(PV_SCR_RUN.beamHeight_um.run(0).toString())) {
					final int beamHeight = Epics.dbrToInt(_value);
					GUI.asyncExec(new Runnable() {
						public void run() {
							if (null!=beamHeightEntry) beamHeightEntry.setText(Integer.toString(beamHeight));

						}
					});
				}
				if (_pvName.equals(PV_SCR_RUN.directory256.run(0).toString())) {
					final String imageDir = Epics.dbrToString(_value);
					GUI.asyncExec(new Runnable() {
						public void run() {
							String rStr = System.getProperty("user.home");
							dirLabel.setText("Image Dir: "
									+ DevDataDir.getPrimary().replace(rStr, "~") + "/"
									+ imageDir);
						}
					});
				}
				if (_pvName.equals(PV_SCR.directory256.toString())) {
					final String screeningDirTemp = Epics
							.getString(PV_SCR.directory256);
					GUI.asyncExec(new Runnable() {
						public void run() {
							ModelScreening.currentRootDir = screeningDirTemp;
							// textDir.setText(screeningDirTemp);
							// textDir.setSelection(screeningDirTemp.length());
							rootDir.setDir(screeningDirTemp);
						}
					});
				}
				if (_pvName.equals(PV_SCR.centerType.toString())) {
					final int centertype = Epics.dbrToInt(_value);
					GUI.asyncExec(new Runnable() {
						public void run() {
							if (centertype == 0)
								ModelScreening.centerType = "LOOP";
							else if (centertype == 1)
								ModelScreening.centerType = "CRYSTAL";
							else if (centertype == 2)
								ModelScreening.centerType = "RASTER";
							else
								ModelScreening.centerType = "UNKNOWN";
							centerType.select(centertype);
						}
					});

				}

				if (_pvName.equals(PV_ROBOT.pause.toString())) {
					setRobotPauseState(true);
				}

				if (_pvName.equals(PV_ROBOT.sta.toString())) {
					status = Epics.getString(PV_ROBOT.sta);
					if (status.length() > 9)
						status = status.substring(9);
					else
						return;

					final String statustemp = status;
					GUI.asyncExec(new Runnable() {
						public void run() {
							if (statustemp.equals("sample in Dewar")) {
								Epics.put(PV_SCR.CurrentTask, "Unloaded");
							}
							if (statustemp.length() > 6) {
								boolean stat;
								int status_;
								if (Client.isMaster()) {
									if (statustemp.equals("SMR: ready")
											|| statustemp.subSequence(0, 6)
													.equals("warm_up")
											|| statustemp
													.equals("pin_base warm_up"))
										// if ( statustemp.equals("SMR: ready")
										// || statustemp.subSequence(0,
										// 6).equals("warm_up"))
										stat = true;
									else
										stat = false;
									String currStat = Epics
											.getString(PV_SCR.CurrentTask);

									if (statustemp.equals("SMR: ready")
											|| statustemp.contains("ERR:")
											|| statustemp
													.contains("safetyMat cleared")
											|| statustemp
													.contains("unload reset")
											|| statustemp
													.contains("Dewar opened")
											|| statustemp.contains("warm-up"))
										status_ = 0;
									else
										status_ = 1;

									if (statustemp.contains("warm_up")
											|| statustemp.contains("warm-up")) {
										status_ = 0;
										if (currStat.contains("MountNextXtal")
												|| currStat
														.contains("CenterXtal")
												|| currStat.contains("wash")
												|| currStat.contains("rinse")
												|| currStat
														.contains("UnloadLastSample"))
											status_ = 1;
									}
									warmupButton.setEnabled(stat);
									globalScreeningStatus = status_;
									if (status_ == 1)
										AppState.setScreeningState(true);
									else
										AppState.setScreeningState(false);
								}
							}
						}
					});
				}

				if (_pvName.equals(PV_SCR.MountNextXtal.toString())) {
					setTaskCheckBox(_value, 0);
				}
				if (_pvName.equals(PV_SCR.CenterXtal.toString())) {
					setTaskCheckBox(_value, 1);
				}
				if (_pvName.equals(PV_SCR.Pause4ctrng.toString())) {

					setTaskCheckBox(_value, 2);
				}
				if (_pvName.equals(PV_SCR.takeJpeg.toString())) {
					setTaskCheckBox(_value, 3);
				}
				if (_pvName.equals(PV_SCR.TakeImage0.toString())) {

					setTaskCheckBox(_value, 4);
				}
				if (_pvName.equals(PV_SCR.TakeImage45.toString())) {
					setTaskCheckBox(_value, 5);
				}
				if (_pvName.equals(PV_SCR.TakeImage90.toString())) {
					setTaskCheckBox(_value, 6);
				}
				if (_pvName.equals(PV_SCR.Pause4collect.toString())) {

					setTaskCheckBox(_value, 7);
				}

				if (_pvName.equals(PV_SCR.ScreeningStatus.toString())) {

					int status_ = 0;

					try {
						status_ = Epics.dbrToInt(_value);
					} catch (Exception e) {
						e.printStackTrace();
					}

					// TODO: disableExceptions
					if (Client.isMaster()) {
						globalScreeningStatus = status_;
						if (status_ == 1)
							AppState.setScreeningState(true);
						else
							AppState.setScreeningState(false);
					}

					final int statusfinallocal = status_;
					GUI.asyncExec(new Runnable() {
						public void run() {
							if (statusfinallocal == 0)
								setTaskBackgroundNULL();
							updateButtons();
						}
					});
				}
				if (_pvName.equals(PV_SCR.CurrentTask.toString())) {

					String Sample = "";
					i = 0;
					double ang1 = Epics.getDouble(PV_SCR.AngleImage0);
					double ang2 = Epics.getDouble(PV_SCR.AngleImage45);
					double ang3 = Epics.getDouble(PV_SCR.AngleImage90);

					int numang1 = Epics.getInteger(PV_SCR.diffimages0);
					int numang2 = Epics.getInteger(PV_SCR.diffimages45);
					int numang3 = Epics.getInteger(PV_SCR.diffimages90);

					int angle1 = (int) Math.ceil(ang1) + 1;
					int angle2 = (int) Math.ceil(ang2) + 1;
					int angle3 = (int) Math.ceil(ang3) + 1;

					String status_ = Epics.dbrToString(_value);
					ModelScreening.currentTaskStr = status_;

					String status2 = status_;
					int space = status_.indexOf(' ');
					if (space > 0)
						Sample = status_.substring(0, space);

					status_ = status_.substring(space + 1);
					int task = 10;
					String status_recollect = "";
					if (status2.length() > 9)
						status_recollect = status2.substring(0, 9);

					boolean collect1 = false, collect2 = false, collect3 = false;
					for (i = 0; i < numang1; i++) {

						if (status_.equals("Collect"
								+ CmdScreening.getFrameNumber(0, ang1, i)))
							collect1 = true;

					}
					for (i = 0; i < numang2; i++) {
						if (status_.equals("Collect"
								+ CmdScreening.getFrameNumber(1, ang2, i)))
							collect2 = true;

					}
					for (i = 0; i < numang3; i++) {
						if (status_.equals("Collect"
								+ CmdScreening.getFrameNumber(2, ang3, i)))
							collect3 = true;
					}

					if (status_recollect.equals("Recollect")) {
						final String num = status2.substring(10);
						int task_ = Integer.parseInt(num);
						
						System.out.println("task " + task);
						if (task_ < 1 || task_ > 8)
							CmdScreening.manualRepeatFlag = 0;
						status2 = status2.substring(0, 9);

						if (status2.equals("Recollect")) {
							if (Client.isMaster())
								CmdScreening.manualRepeatFlag = task_;
							GUI.asyncExec(new Runnable() {
								public void run() {
										setTaskImageNULL();
										startAction[Integer.parseInt(num)]
												.setImage(GUI.image);
									}
							});
							Log.note("repeatFlag " + task_);
						} else {
								CmdScreening.manualRepeatFlag = 0;
						}
					}
					if (status_.equals("Paused4Inspection")) {
						GUI.asyncExec(new Runnable() {
							public void run() {
								setTaskImageNULL();
								labelRasterWarn.setText("");
								startAction[6].setImage(GUI.image);
								for (int k = 0; k < 6; k++) {
									startAction[k].setSelection(false);
								}
							}
						});
					}
					if (status_.equals("UnloadLastSample")) {
						GUI.asyncExec(new Runnable() {
							public void run() {
								setTaskImageNULL();
								setTaskBackgroundNULL();
							}
						});
					}
					if (status_.equals("MountNextXtal")
							|| status_.equals("loading")) {

						task = 0;
					} else if (status_.equals("CenterXtal")) {

						task = 1;
						GUI.asyncExec(new Runnable() {
							public void run() {
								// if(ModelScreening.centerType.equals("LOOP")
								// ||
								// ModelScreening.centerType.equals("CRYSTAL"))
								stopButton.setEnabled(true);
								if (centerType.getSelectionIndex() == 2)
									labelRasterWarn
											.setText("Raster is in progress...");
							}
						});
					} else if (status_.equals("Paused4Center")) {
						Log.note("Paused4Center");
						GUI.asyncExec(new Runnable() {
							public void run() {
								labelRasterWarn.setText("");
							}
						});
						task = 2;
					} else if (status_.contains("JPEGsnapshot")) {
						highlightJpeg = 1;
						task = 3;
					} else if (collect1 || status_.equals("Rotate" + angle1)) {

						task = 4;
					} else if (collect2 || status_.equals("Rotate" + angle2)) {

						task = 5;
					} else if (collect3 || status_.equals("Rotate" + angle3)) {

						task = 6;
					} else if (status_.equals("Paused4Inspection")) {

						task = 7;
					} else {
						return;
					}
					final int task_final = task;
					final String fstatus_ = status_;
					GUI.asyncExec(new Runnable() {
						public void run() {
							if (task_final < 8) {
								highlightJpeg = 0;
								setTaskImageNULL();
								setTaskBackgroundNULL();
								// if(checkAction[task_final].getSelection()) {
								startAction[task_final].setImage(GUI.image);
								checkAction[task_final]
										.setBackground(GUI.COLOR_THICK_GREEN);
								// }
							} else
								setTaskBackgroundNULL();

							if (!fstatus_.equals("CenterXtal"))
								stopButton.setEnabled(false);

							updateButtons();
						}
					});
				}
			}
		};

		Epics.addMonitor(PV_SCR.MountNextXtal, monitor);
		Epics.addMonitor(PV_SCR.CenterXtal, monitor);
		Epics.addMonitor(PV_SCR.Pause4ctrng, monitor);
		Epics.addMonitor(PV_SCR.TakeImage0, monitor);
		Epics.addMonitor(PV_SCR.TakeImage45, monitor);
		Epics.addMonitor(PV_SCR.TakeImage90, monitor);
		Epics.addMonitor(PV_SCR.Pause4collect, monitor);
		Epics.addMonitor(PV_SCR.takeJpeg, monitor);

		Epics.addMonitor(PV_ROBOT.sta$2, monitor);
		Epics.addMonitor(PV_ROBOT.sta$1, monitor);
		Epics.addMonitor(PV_ROBOT.sta, monitor);

		Epics.addMonitor(PV_ROBOT.pause, monitor);
		Epics.addMonitor(PV_ROBOT.cmd, monitor);

		Epics.addMonitor(PV_PMAC.ActPos.motor("detector_z"), monitor);
		Epics.addMonitor(PV_SCR.ScreeningStatus, monitor);
		Epics.addMonitor(PV_SCR.CurrentTask, monitor);
		Epics.addMonitor(PV_SCR.directory256, monitor);
		Epics.addMonitor(PV_SCR_RUN.directory256.run(0), monitor);

		Epics.addMonitor(PV_SCR_RUN.distance.run(0), monitor);
		Epics.addMonitor(PV_SCR_RUN.delta.run(0), monitor);
		Epics.addMonitor(PV_SCR_RUN.exposureTime.run(0), monitor);
		Epics.addMonitor(PV_SCR_RUN.atten_factor.run(0), monitor);
		Epics.addMonitor(PV_SCR_RUN.beamWidth_um.run(0), monitor);
		Epics.addMonitor(PV_SCR_RUN.beamHeight_um.run(0), monitor);

		Epics.addMonitor(PV_SCR.AngleImage0, monitor);
		Epics.addMonitor(PV_SCR.AngleImage45, monitor);
		Epics.addMonitor(PV_SCR.AngleImage90, monitor);
		Epics.addMonitor(PV_BLUICE.autoCenterRun, monitor);
		Epics.addMonitor(PV_SCR.prefix, monitor);
		Epics.addMonitor(PV_SCR.centerType, monitor);

		GUI.asyncExec(new Runnable() {
			public void run() {
				updateButtons();
			}
		});
	}
	private void setTaskCheckBox (DBR value, int check) {
		int temp = 0;
		try {
			temp = Epics.dbrToInt(value);
		} catch (Exception e) {
			e.printStackTrace();
			return;
		}
		final int tempi = check;
		final int temptemp = temp;
		GUI.asyncExec(new Runnable() {
			public void run() {
				if (temptemp == 1)
					checkAction[tempi].setSelection(true);
				else
					checkAction[tempi].setSelection(false);
			}
		});
		screeningTab.updateTaskProgress();
	}

	public void updateBeamSizeDisplay()
	{
		ArrayList<BeamSize> sizes = DBBeamSize.getBeamSizes(DevEnergy.getDevice().getPosition());
		for (int i=0;i<sizes.size();i++) {
			BeamSize size = sizes.get(i);
			String text = Util.sprintf("%d x %d", Math.round(size.width_um), Math.round(size.height_um));
			beamsizeCombo.setItem(i, text);
		}
		updateBeamSizeSelection();
	}

	private int updateBeamSizeSelection()
	{
		double currBeamWidth_um = Epics.getDouble(PV_SCR_RUN.beamWidth_um.run(0));
		int colli_index = DBBeamSize.getColliIndex((int)currBeamWidth_um);
		if (beamsizeCombo.getSelectionIndex()!=colli_index) beamsizeCombo.select(colli_index);
		return colli_index;
	}

	private void createBeamSizeDropdown(Composite beamSizeComp)
	{
		Label lblBeamsize = new Label(beamSizeComp, SWT.NONE);
		lblBeamsize.setBackground(GUI.BGCOLOR);
		lblBeamsize.setText("Beam size:");
		GridData gd = new GridData(SWT.DEFAULT, SWT.CENTER, false, false, 2, 1);
		lblBeamsize.setLayoutData(gd);

		beamsizeCombo = new Combo(beamSizeComp, SWT.READ_ONLY);
		gd = new GridData(SWT.FILL, SWT.DEFAULT, true, false, 4, 1);
		beamsizeCombo.setLayoutData(gd);
		for (int i=0;i<DBBeamSize.colli_sizes.length;i++) beamsizeCombo.add("");
		updateBeamSizeDisplay();
		beamsizeCombo.select(0);
		beamsizeCombo.addSelectionListener(new SelectionListener()
		{
			@Override public void widgetSelected(SelectionEvent selectionEvent)
			{
				int selected = beamsizeCombo.getSelectionIndex();
				int colliSize_um = DBBeamSize.colli_sizes[selected];
				double currBeamWidth = Epics.getDouble(PV_SCR_RUN.beamWidth_um.run(0));
				if (currBeamWidth != colliSize_um)
				{
					Epics.putDouble(PV_SCR_RUN.beamHeight_um.run(0), colliSize_um);
					Epics.putDouble(PV_SCR_RUN.beamWidth_um.run(0), colliSize_um);
					putDataCollectionParam(PV_SCR_RUN.beamHeight_um.run(0),	colliSize_um);
					putDataCollectionParam(PV_SCR_RUN.beamWidth_um.run(0), colliSize_um);
				}
			}
			@Override public void widgetDefaultSelected(SelectionEvent selectionEvent) {}
		});

		DevMoveFloat.Listener energyMoveListener = new DevMoveFloat.Listener()
		{
			@Override public void handleStateChange(State state) {}
			@Override public void handleActPosChange(DevMoveFloat dev, double newPos)
			{
				if (!dev.isMoving()) return;
				GUI.asyncExec(()->updateBeamSizeDisplay());
			}
			@Override public void handleRqsPosChange(double newPos) {}
			@Override public void handleUnitsChange(String units) {}
		};
		DevPmacMotor.get("mo_En_E").addListener(energyMoveListener);

		ListenerList.Listener<DevMove> energyStopListener = (object,data)->
		{
			GUI.asyncExec(()->updateBeamSizeDisplay());
		};
		DevEnergy.getDevice().llMoveStartStop.addListener(energyStopListener);
	}

	private void createBeamSizeFields(Composite expDeltaGroup) {
		Composite beamSizeComp = new Composite(expDeltaGroup, SWT.NONE);
		beamSizeComp.setBackground(GUI.BGCOLOR);
		GridLayout gl = new GridLayout(4, false);
		gl.marginHeight = 0;
		gl.marginWidth = 0;
		beamSizeComp.setLayout(gl);
		GridData gd = new GridData(SWT.FILL, SWT.FILL, true, true);
		gd.horizontalSpan = 6;
		beamSizeComp.setLayoutData(gd);

		Label lblBeamsize = new Label(beamSizeComp, SWT.NONE);
		lblBeamsize.setBackground(GUI.BGCOLOR);
		lblBeamsize.setText("Beam size:");
		gd = new GridData(SWT.DEFAULT, SWT.CENTER, false, false);
		lblBeamsize.setLayoutData(gd);

		gd = new GridData(SWT.FILL, SWT.FILL, true, true);
		beamWidthEntry = new WdgControlEntry(beamSizeComp, SWT.NONE,
				EntryType.TEXT, new WdgControlEntry.ModelInterface() {
					@Override
					public void setValue(Object value) {

						double currBeamHeight = Epics
								.getDouble(PV_SCR_RUN.beamHeight_um.run(0));
						Point2D beamSize = ModelRasterGrid.calcBeamSize(
								Double.valueOf(value.toString()),
								currBeamHeight,
								ModelBeamsize.getActualBeamsize_um().y);

						Epics.putDouble(PV_SCR_RUN.beamHeight_um.run(0), 0);
						Epics.putDouble(PV_SCR_RUN.beamWidth_um.run(0), 0);

						putDataCollectionParam(PV_SCR_RUN.beamHeight_um.run(0),
								beamSize.y);
						putDataCollectionParam(PV_SCR_RUN.beamWidth_um.run(0),
								beamSize.x);

					}

					@Override
					public Object getModelValue() {
						return Epics.getDouble(PV_SCR_RUN.beamWidth_um.run(0));
					}

					@Override
					public Object getDeviceValue() {
						return ModelBeamsize.getCurrentBeamsize().x;
					}

					@Override
					public int getDecimals() {
						return 0;
					}

					@Override
					public boolean doHighlight(double d) {
						return Math.abs((Double) getDeviceValue() - d) > 1.0;
					}

					@Override
					public Limits getLimits() {
						double lowLimit = 5.0;
						double highLimit = ModelBeamsize.getActualBeamsize_um().x;
						return new Limits(lowLimit, "minimum beam width",
								highLimit, "maximum beam width");
					}
				});
		beamWidthEntry.setLayoutData(gd);

		WidgetFactory.newLabel(beamSizeComp, SWT.NONE, "x").setLayoutData(
				new GridData());

		gd = new GridData(SWT.FILL, SWT.FILL, true, true);
		beamHeightEntry = new WdgControlEntry(beamSizeComp, SWT.NONE,
				EntryType.TEXT, new WdgControlEntry.ModelInterface() {
					@Override
					public void setValue(Object value) {
						double currBeamWidth = Epics
								.getDouble(PV_SCR_RUN.beamWidth_um.run(0));
						Point2D beamSize = ModelRasterGrid.calcBeamSize(
								Double.valueOf(value.toString()),
								currBeamWidth,
								ModelBeamsize.getActualBeamsize_um().x);

						Epics.putDouble(PV_SCR_RUN.beamHeight_um.run(0), 0);
						Epics.putDouble(PV_SCR_RUN.beamWidth_um.run(0), 0);
						putDataCollectionParam(PV_SCR_RUN.beamHeight_um.run(0),
								beamSize.x);
						putDataCollectionParam(PV_SCR_RUN.beamWidth_um.run(0),
								beamSize.y);

					}

					@Override
					public Object getModelValue() {
						return Epics.getDouble(PV_SCR_RUN.beamHeight_um.run(0));
					}

					@Override
					public Object getDeviceValue() {
						return ModelBeamsize.getCurrentBeamsize().y;
					}

					@Override
					public boolean doHighlight(double d) {
						return Math.abs((Double) getDeviceValue() - d) > 1.0;
					}

					@Override
					public int getDecimals() {
						return 0;
					}

					@Override
					public Limits getLimits() {
						double lowLimit = 5.0;
						double highLimit = ModelBeamsize.getActualBeamsize_um().y;
						return new Limits(lowLimit, "minimum beam height",
								highLimit, "maximum beam height");
					}
				});
		beamHeightEntry.setLayoutData(gd);

	}

	public void startScreening(ModelCrystalList crystalList) {
		
		/* Check for the screening directory and its permissions */
		String dir = DevDataDir.getPrimary() + "/" + ModelScreening.currentRootDir;
		if (!dir.equals("")) {
			File dirFile = new File(dir);

			/* create directory if does's exist */
			if (!dirFile.exists()) {

				if (!dirFile.mkdirs()) {
					Log.error("ERROR; Couldn't create directory " + dir);
				}
			}
			if (!dirFile.canWrite()) {
				NBMessageBox messageBox = new NBMessageBox(GUI.getShell(),SWT.OK | SWT.ICON_ERROR);
				messageBox.setMessage("Screening directory" + dir + " don't have write permissions");
				messageBox.open();
				return;
			}
		}
		/* return if robot is not ready */
		if (!checkRobotReady())
			return;
		
		/* check for multiple screening threads */
		if (null != screeningThread && screeningThread.isAlive()) {
			Log.error("Another screening thread already running, Please restart JBluice if you you reached here");
			
			exitBluice();
			return;
		}

		Epics.putInteger(PV_SCR.ScreeningStatus, 1);
		ModelScreening.screeningStatus = 1;

		startButton.setEnabled(false);
		
		/* initate a screening thread, the thread will exist as long as robot not going to pause */
		
		CmdScreening screeningControl = new CmdScreening(crystalList);
		screeningThread = new Thread(screeningControl, "Screening control");
		screeningThread.start();
		screeningControl._requestStop = true;
	}

	public static void exitBluice() {
		boolean confirmed = MessageDialog.openConfirm(GUI.getShell(), "Confirm",
				"Another screening thread already running, "
						+ "jBluIce needs to be restarted to correct the problem.\n\n" + "Exit jBluIce now?");
		if (confirmed) {
			Main.shutdown();
		}		
	}
	
	public static boolean checkRobotReady() {
		int robotStatus = Epics.getInteger(PV_ROBOT.pause);
		if (robotStatus == 1) {
			NBMessageBox messageBox = new NBMessageBox(GUI.getShell(), SWT.OK | SWT.ICON_ERROR);
			messageBox.setMessage("Robot not ready, click on Ready button and proceed");
			messageBox.open();
			return false;
		} else
			return true;
	}

	public void puttoSpreadsheetPV(String Directory) {

		Epics.put(PV_SCR.directory256, Directory);
		Epics.put(PV_SCR_RUN.directory256.run(0), (Directory+"/"
				+ ModelScreening.currentSubDir).replace("//", "/"));

		updateRuns();
	}

	public void setAngle(int angleIndex) {

		String angleStr = angleSpinner[angleIndex].getText();
		String angleText = "0";
		double angle = 0.0;

		if (!angleStr.equals("") && !angleStr.equals(" ") && angleStr != null) {
			angle = Double.parseDouble(angleStr);
		} else {
			if (angleIndex == 0)
				angleText = Epics.getString(PV_SCR.AngleImage0);
			else if (angleIndex == 1)
				angleText = Epics.getString(PV_SCR.AngleImage45);
			else if (angleIndex == 2)
				angleText = Epics.getString(PV_SCR.AngleImage90);

			angleSpinner[angleIndex].setText(angleText);
			return;
		}

		if (angleIndex == 0)
			Epics.putDouble(PV_SCR.AngleImage0, angle);
		else if (angleIndex == 1)
			Epics.putDouble(PV_SCR.AngleImage45, angle);
		else if (angleIndex == 2)
			Epics.putDouble(PV_SCR.AngleImage90, angle);
	}

	private void updateLabels() {
		
		GUI.asyncExec(new Runnable() {
			public void run() {
				multipleImagesSpinner[0]
						.setText(Integer.toString(ModelScreeningTasks.DIFFIMAGES0));
				multipleImagesSpinner[1]
						.setText(Integer.toString(ModelScreeningTasks.DIFFIMAGES45));
				multipleImagesSpinner[2]
						.setText(Integer.toString(ModelScreeningTasks.DIFFIMAGES90));
			}
		});
	}

	protected void setRobotPauseState(boolean update) {
		final boolean isPaused = Epics.getBoolean(PV_ROBOT.pause);
		final boolean isUpdate = update;
		if (isPaused)
			CmdScreening.stopThread = true;
		
		GUI.asyncExec(new Runnable(){public void run()
		{
			if(Client.isMaster()) {
				pauseButton.setEnabled(!isPaused);
				readyButton.setEnabled(isPaused);
				
				if(isUpdate)
					updateButtons();
			} else {
				pauseButton.setEnabled(false);
				readyButton.setEnabled(false);
			}
			if(!isPaused) {
				//readyButton.setImage(GUI.readyImage);
				//pauseButton.setImage(GUI.pauseImage);
				readyButton.setBackground(GUI.COLOR_GREEN_BUTTON);
				pauseButton.setBackground(GUI.COLOR_BLUE_BUTTON);
			}
			else {
				if(Client.isMaster())
					Epics.putInteger(PV_SCR.ScreeningStatus,0);
				//pauseButton.setImage(GUI.readyImage);
				//readyButton.setImage(GUI.pauseImage);
				pauseButton.setBackground(GUI.COLOR_RED_BUTTON);
				//pauseButton.setForeground(GUI.COLOR_BLACK);
				readyButton.setBackground(GUI.COLOR_BLUE_BUTTON);
			}
		}});	
	}

	public void updateButtons() {

		boolean isMaster = Client.isMaster();
		boolean isScreening = Client.isMaster()
				&& AppState.isOperationAllowed(Operation.START_SCREENING);
		int statusfinal = Epics.getInteger(PV_SCR.ScreeningStatus);
		boolean isScreeningStatus2 = isScreening && (statusfinal == 2);

		for (int i = 0; i < 3; i++) {
			angleSpinner[i].setEnabled(isMaster);
			multipleImagesSpinner[i].setEnabled(isMaster);
		}

		for (int j = 0; j < ActionNames.length; j++) {
			//startAction[j].setEnabled(isMaster);
			startAction[j].setEnabled(isScreening);
			checkAction[j].setEnabled(isMaster);
		}
		checkAction[0].setEnabled(false);
		centerType.setEnabled(isMaster);
		rootDir.setEnabled(isMaster);

		updateButton.setEnabled(isMaster);
		distanceEntry.setEnabled(isMaster);
		deltaEntry.setEnabled(isMaster);
		timeEntry.setEnabled(isMaster);
		attenuationEntry.setEnabled(isMaster);

		if (null!=beamWidthEntry) beamWidthEntry.setEnabled(isMaster);
		if (null!=beamHeightEntry) beamHeightEntry.setEnabled(isMaster);
		if (null!=beamsizeCombo) beamsizeCombo.setEnabled(isMaster);
		warmupButton.setEnabled(isScreeningStatus2);
		unmountButton.setEnabled(isScreeningStatus2);
		centerType.setEnabled(isMaster);

		updateButton.setEnabled(isScreening);
		warmupButton.setEnabled(isScreening);

		setRobotPauseState(false);
		if (!isMaster) {
			pauseButton.setEnabled(false);
			readyButton.setEnabled(false);
		}


		if (isMaster) {
			
			if (statusfinal == 2) {	
				startButton.setText("Start"); startButton.setEnabled(false);
				resumeButton.setEnabled(isScreening);
			}
			else if (statusfinal == 1) {
				if(ModelScreening.currentTaskStr.contains("Collect"))
				{
					startButton.setEnabled(true); startButton.setText("Pause");
				}
				else {
					startButton.setText("Start"); startButton.setEnabled(false);
				}
				resumeButton.setEnabled(false);
			}
			else if ( statusfinal == 0 ) {
				startButton.setText("Start"); 
				startButton.setEnabled(isScreening);
				
				resumeButton.setEnabled(false);

				warmupButton.setEnabled(true);
				unmountButton.setEnabled(true);
			}
		} else {
			startButton.setEnabled(false);
			resumeButton.setEnabled(false);
			warmupButton.setEnabled(false);
			unmountButton.setEnabled(false);
		}

		if (!Client.isMaster()) {
			stopButton.setEnabled(false);
			CmdScreening.stopThread = true;
			ModelScreening.pauseCollect = false;
		}
	}

	@Override
	public void handleEvent(ModelScreeningTasks object, Object data) {
		updateLabels();
	}

	@Override
	public void masterChanged(boolean isMaster) {
		// TODO Auto-generated method stub
		GUI.asyncExec(new Runnable() {
			public void run() {
				updateButtons();
			}
		});

	}
}
