package gov.anl.gmca.server;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.common.Redis;
import gov.anl.gmca.common.client_server.StateMap;
import gov.anl.gmca.hutch.DevAutocenter;
import gov.anl.gmca.hutch.DevPmacAssembly.*;
import gov.anl.gmca.server.messaging.Popup;
import gov.anl.gmca.server.services.BeamSize;
import gov.anl.gmca.server.services.ResPredictor;
import gov.anl.gmca.server.workers.*;
import gov.anl.gmca.server.workers.AssemblyMove;
import gov.anl.gmca.server.workers.AutoCenter.CENTER_MODE;
import gov.anl.gmca.util.Util;
import redis.clients.jedis.Jedis;
import spark.Request;
import spark.Response;
import spark.Spark;

import java.util.*;

import static spark.Spark.*;

public class HTTP
{
	public static Gson gson = new GsonBuilder().serializeNulls().serializeSpecialFloatingPointValues().create();
	public static final int HTTP_TIMEOUT_MS = 3600 * 1000; // one hour
	
	public static void init()
	{
		String serverIP = Util.getServerIP();
		int port = (int)DBConfig.getDouble("bluice_server_port");
		System.out.println("Listening on "+serverIP+" port "+port);

		StateMapUpdate.init();
		ipAddress(serverIP);
		port(port);
		threadPool(16, 2, HTTP_TIMEOUT_MS);

		Spark.exception(Exception.class, (exc, req, resp) -> exc.printStackTrace());

		get("/state", (req, res) -> handleStateRequest(req, res));
		post("/state", (req, res) -> handleStateRequest(req, res));
		post("/set_config", (req, res) -> handleSetRequest(req, res));
		post("/cmd_req", (req, res) -> handleCmdRequest(req, res));
		get("/cmd_stat", (req, res) -> handleCmdStatus(req, res));
		get("/respredictor", (req, res) -> ResPredictor.handleRequest(req, res));
		get("/beamsize", (req, res) -> BeamSize.handleRequest(req, res));
		post("/popup", (req, res) -> Popup.handleRequest(req, res));
		post("/popup_response", (req,res) -> Popup.handleResponse(req, res));
		get("/check", (req, res) -> { System.out.println("Got /check request"); return "OK".getBytes(); });
	}
	
	public static String handleStateRequest(Request req, Response res)
	{
		String clientID = req.queryParams("clientID");
		ClientTable.handleRequest(req);
		StateMapUpdate.update(clientID);
		String response = gson.toJson(StateMap.current);

		//System.out.println(req.toString());
		//System.out.println(response.toString());
		res.type("application/json");
		return response;
	}

	public static String handleSetRequest(Request req, Response res)
	{
		String key1 = req.queryParams("key1");
		String key2 = req.queryParams("key2");
		String value = req.queryParams("value");
		try (Jedis jedis = Redis.get()) {
			jedis.hset(key1, key2, value);
		}

		return "OK";
	}

	public static String handleCmdRequest(Request req, Response res)
	{
		String response = "";
		HashMap<String,String> paramMap = new HashMap<>();
		for (String key : req.queryParams()) {
			paramMap.put(key, req.queryMap(key).value());
		}


		String cmd = req.queryParams("cmd");
		if (cmd.equals("MotorMove"))
		{
			String motor = req.queryParams("motor");
			double dest = Double.parseDouble(req.queryParams("dest"));
			Worker w = new MotorMove(motor, dest);
			w.start();
			response = w.initialJSONResponse();
		}
		if (cmd.equals("AssemblyMove"))
		{
			String assembly = req.queryParams("assembly");
			ArrayList<Double> dest = new ArrayList<>();
			StringTokenizer st = new StringTokenizer(req.queryParams("dest"),",");
			while (st.hasMoreTokens()) {
				dest.add(Double.parseDouble(st.nextToken()));
			}
			Worker w = new AssemblyMove(assembly, dest);
			w.start();
			response = w.initialJSONResponse();
		}
		else if (cmd.equals("OptimizeBeam"))
		{
			Worker w = new OptimizeBeam();
			w.start();
			response = w.initialJSONResponse();
		}
		else if (cmd.equals("EnterHutch"))
		{
			Worker w = new EnterHutch();
			w.start();
			response = w.initialJSONResponse();
		}
		else if (cmd.equals("AutoCenter"))
		{
			String modeString = req.queryParams("mode");
			CENTER_MODE mode = null;
			if (modeString.toLowerCase().equals("loop")) mode=CENTER_MODE.LOOP;
			if (modeString.toLowerCase().equals("crystal")) mode=CENTER_MODE.CRYSTAL;
			if (null==mode) {
				response = "ERROR";
			} else {
				Worker w = new AutoCenter(mode);
				w.start();
				response = w.initialJSONResponse();
			}
		}
		else if (cmd.equals("CollectJet"))
		{
			Worker w = new CollectJet(paramMap);
			w.start();
			response = w.initialJSONResponse();
		}
		else if (cmd.equals("SONICC"))
		{
			Worker w = new SONICC(paramMap);
			w.start();
			response = w.initialJSONResponse();
		}
		else if (cmd.equals("SONICCStop"))
		{
			SONICC.stopRequested = true;
		}
		else if (cmd.equals("CollectJetStop"))
		{
			CollectJet.get().stop();
		}
		else if (cmd.equals("Raster"))
		{
			Worker w = new Raster(paramMap);
			w.start();
			response = w.initialJSONResponse();
		}
		else if (cmd.equals("RasterPause"))
		{
			Raster.get().pause();
		}
		else if (cmd.equals("RasterAbort"))
		{
			Raster.get().abort();
		}
		else if (cmd.equals("image_handler")) {
			//ImageHandler ih = new ImageHandler(paramMap);
			//ih.start();
			//response = ih.initialJSONResponse();
		}
		else if (cmd.equals("RunFPE")) {
			RunFPE w = new RunFPE(paramMap);
			w.start();
			response = w.initialJSONResponse();
		}
		else if (cmd.equals("StopAutoCenter"))
		{
			DevAutocenter.get().stopExec();
		}
		else
		{
			response = "Unknown command: "+cmd;
		}
		
		System.out.println(req.toString());
		System.out.println(response.toString());
		return response;
	}

	public static String handleCmdStatus(Request req, Response res)
	{
		String id = req.queryParams("cmd_id");
		Worker w = WorkerList.get(id);
		String response =  w.jsonStatus();
		
		System.out.println(req.toString());
		System.out.println(response.toString());
		return response;
	}
}
