package gov.anl.gmca.server.workers;

import com.mashape.unirest.http.*;
import gov.anl.gmca.common.*;
import gov.anl.gmca.common.client_server.RPC.*;
import gov.anl.gmca.hutch.*;
import gov.anl.gmca.io.*;
import gov.anl.gmca.sample.*;
import gov.anl.gmca.server.*;
import gov.anl.gmca.syslog.*;
import gov.anl.gmca.util.*;
import redis.clients.jedis.*;

import javax.imageio.*;
import java.awt.image.*;
import java.io.*;
import java.util.*;

public class SONICC extends Worker
{
    public static String[] MODES = new String[]{"IR","GREEN"};
    public static String[] OUTPUT_CHANNELS = new String[]{"A","B","C"};
    public static String[][] OUTPUTS = new String[][]{{"SHG","TPEF","IRBF"},{"TPUVF","FLUO","GRBF"}};

    private Map<String,String> args;
    private Point4D ref_pos;
    public static boolean stopRequested = false;
    public static SONICC inst = null;
    public static SONICC get() { return inst; }
    public SONICC(Map<String,String> args)
    {
        super("SONICC");
        if (null!=inst && inst.thread.isAlive()) {
            result = RPCResult.ERROR;
            return;
        }
        this.args=args;
        inst=this;
        stopRequested = false;
    }

    @Override
    public void run()
    {
        try {
            acquire();
        } catch (Exception e) {
            e.printStackTrace();
            result = RPCResult.ERROR;
        }
        finally {
            try (Jedis jedis = Redis.get()) {
                jedis.hset(Redis.sonicc_state, "state", "IDLE");
            }
            System.out.println("SONICC acquire complete.");
        }
    }

    public void acquire()
    {
        String rel_dir = args.get("dir");
        String abs_dir = DevDataDir.addToPrimary(rel_dir);
        String prefix = args.get("prefix");
        String mode = args.get("mode");
        Integer slice_count = Integer.parseInt(args.get("slices"));
        Integer slice_width_um = Integer.parseInt(args.get("slice_width_um"));

        // save reference position
        double xref = DevPmacMotor.get("sample_x").getPosition();
        double yref = DevPmacMotor.get("sample_y").getPosition();
        double zref = DevPmacMotor.get("sample_z").getPosition();
        double omegaref = DevPmacMotor.get("gonio_omega").getPosition();
        ref_pos = new Point4D(xref,yref,zref,omegaref);
        Util.printf("Got reference position: %.2f %.2f %.2f %.2f",xref,yref,zref,omegaref);

        try (Jedis jedis = Redis.get()) {
            jedis.hset(Redis.sonicc_state, "state", "ACQUIRING");
            jedis.hset(Redis.sonicc_state, "next_slice", "0");
            jedis.hset(Redis.sonicc_state, "current_mode", mode);
            jedis.hset(Redis.sonicc_state, "ref_pos", Util.sprintf("%.3f %.3f %.3f", xref, yref, zref));
        }

        // sample env command?  anything to move?

        // any light commands?

        // directory exists?
        File abs_dir_f = new File(abs_dir);
        if (!abs_dir_f.exists()) {
            boolean file_result = abs_dir_f.mkdirs();
            if (!file_result) {
                Util.asyncPopupError("mkdir error", "Error creating directory "+abs_dir);
                result = RPCResult.ERROR;
                return;
            }
        }

        // do files already exist?

        // anything else to check?

        // iterate slices
        for (int sliceIndex=0; sliceIndex<slice_count; sliceIndex++) {

            if (stopRequested) break;

            // move to slice location
            moveToSlice(slice_count, sliceIndex, slice_width_um);

            if (stopRequested) break;

            // iterate modes
            ArrayList<Integer> modesToAcquire = new ArrayList<>();
            if (mode.equals("IR") || mode.equals("BOTH")) modesToAcquire.add(0);
            if (mode.equals("GREEN") || mode.equals("BOTH")) modesToAcquire.add(1);
            for (Integer modeIndex : modesToAcquire)
            {
                if (stopRequested) break;

                // update status
                try (Jedis jedis = Redis.get()) {
                    jedis.hset(Redis.sonicc_state, "info", "Acquiring " + MODES[modeIndex] + " slice " + (sliceIndex + 1));
                    jedis.hset(Redis.sonicc_state, "mode", modeIndex==0?"SHG":"TPE");
                    jedis.hset(Redis.sonicc_state, "next_slice", "" + sliceIndex);
                    jedis.hset(Redis.sonicc_state, "current_mode", MODES[modeIndex]);
                }

                // send acquire command
                String cmd = modeIndex==0 ? "takeSHG" : "takeTPE";
                if (sliceIndex < slice_count - 1) cmd += " stayIN"; // TODO: only send on last mode switch
                Log.note("Sending to SONICC: " + cmd);
                Epics.put(PVLookup.PV_SONICC.serverCMD, cmd);

                // wait for server to report complete
                while (true) {
                    Util.sleep(1000);
                    String serverStatus = Epics.getString(PVLookup.PV_SONICC.serverCMD);
                    Log.note("SONICC server status = " + serverStatus);

                    if (serverStatus.contains("finished")) {
                        if (serverStatus.contains("status=0")) {
                            Log.note("SONICC acquire completed successfully.");
                            break;
                        } else {
                            Util.asyncPopupError("SONICC error", "SONICC server reported an error.");
                            break;
                        }
                    }
                }

                // write images to disk
                String sonicc_web_server = DBConfig.sim5 ? "http://127.0.0.1:8007" : "http://192.168.1.50:4000";
                try {
                    for (int outputIndex : new int[]{0, 2}) {
                        String outputChannel = OUTPUT_CHANNELS[outputIndex];
                        String hr_url = sonicc_web_server + "/Channel" + outputChannel + "Last.jpg";
                        String output = OUTPUTS[modeIndex][outputIndex];
                        String destPath = String.format("%s/%s_%s_%03d.png", abs_dir, prefix, output, sliceIndex + 1);

                        HttpResponse resp = Unirest.get(hr_url).asBinary();
                        BufferedImage originalImg = ImageIO.read(resp.getRawBody());
                        ImageIO.write(originalImg, "png", new File(destPath));
                    }
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        }

        try (Jedis jedis = Redis.get()) {
            jedis.hset(Redis.sonicc_state, "info", "Acquisition complete.");
            jedis.hset(Redis.sonicc_state, "next_slice", ""+slice_count);
        }
        result = RPCResult.OK;
    }

    public void moveToSlice(int sliceCount, int sliceIndex, int sliceWidth_um) {
        double zMove_um = (sliceIndex - ((sliceCount - 1) / 2)) * sliceWidth_um;
        Point4D dest = WdgCamera.addXYZToMotorMM(ref_pos, 0, 0, zMove_um);
        DevXYZ.syncMove(dest.to3D(), true);
    }

    @Override public String jsonStatus() {return "id="+id+" result="+result;}
    public void stop() {}
}
