import gov.anl.gmca.collect.*;
import gov.anl.gmca.common.*;
import gov.anl.gmca.gui.Main;
import gov.anl.gmca.io.*;
import gov.anl.gmca.raster.*;
import gov.anl.gmca.raster.ModelRasterState.*;
import gov.anl.gmca.sample.*;
import gov.anl.gmca.spotfinder.*;
import gov.anl.gmca.spotfinder.CmdSpotFinderRemote.*;
import gov.anl.gmca.spotfinder.DevSFType.*;
import gov.anl.gmca.spotfinder.SFData.*;
import gov.anl.gmca.util.*;
import junit.framework.*;

import java.io.*;
import java.nio.*;
import java.nio.file.*;
import java.util.*;

import static org.junit.Assert.fail;

public class TestSpotFinder extends TestCase
{
	private Path tempDir;

	@Override
	protected void setUp() throws Exception
	{
		Epics.connect();
		SpotFinderQueue.init();
		new DevSFType(SFRunType.COLLECT);
		new DevSFType(SFRunType.RASTER);
		Main.FAKE_DATA = true;
	}

	@Override
	protected void tearDown() throws Exception
	{
	}

	@org.junit.Test
	public void testFooterCheck() throws Exception
	{
		Path tempDir = null;
		ArrayList<String> writtenFiles = new ArrayList<>();
		try
		{
			// directory to write partial files to
			tempDir = Files.createTempDirectory(FileSystems.getDefault().getPath("/mnt/beegfs/testdata/temp"),"bluice");

			// filename paths
			String sourcePath = "/mnt/beegfs/testdata/PC/PC_000001.cbf";
			String partialFile = "PC_1_000001.cbf";
			writtenFiles.add(tempDir.toString()+"/"+partialFile);

			// run tests
			testPartialFile(2000, sourcePath, tempDir.toString(), partialFile);
			testPartialFile(-1, sourcePath, tempDir.toString(), partialFile);
			testPartialFile(0, sourcePath, tempDir.toString(), partialFile);
		}
		catch (Exception e) { e.printStackTrace(); fail(); }
		finally
		{
			for (String file : writtenFiles) {
				Files.delete(FileSystems.getDefault().getPath(file));
			}
			if (null!=tempDir) Files.delete(tempDir);
		}
	}

	public void testPartialFile(long bytesToWrite, String sourcePath, String tempDir, String partialFile)
	{
		String prefix = "PC";
		int runLabel = 1, frameCount = 1;
		DBImageStats.initialize(Redis.image_stats(tempDir, prefix, runLabel, ""), frameCount);

		long sourceSize = new File(sourcePath).length();
		if (bytesToWrite<=0) { bytesToWrite = sourceSize+bytesToWrite; }
		ArrayList<String> cmd = new ArrayList<>(List.of(
				"dd",
				"if="+sourcePath,
				"of="+tempDir+"/"+partialFile,
				"bs="+bytesToWrite,
				"count=1"
		));
		ExecTask2.execCmd(cmd, null, false, null, 0, true);
		startSF(tempDir, partialFile, "PC", 0, 5000);
		waitForSFQueue();

		String result_key = Redis.image_stats(tempDir, prefix, runLabel, "dials:result");
		int result_int = Math.round(DBImageStats.load(result_key).get(0));
		RESULT result = RESULT.values()[result_int];
		System.out.println("Got result "+result.toString()+" for "+bytesToWrite+" bytes from key "+result_key);
	}

	public void startSF(String dir, String filename, String prefix, int frameIndex, int timeoutMs)
	{
		for (SPOTFINDER_TYPE sftype : DevSFType.getActiveTypes()) {
			CmdSpotFinder cmdProcess = new CmdSpotFinderRemote
			(
				SFType.COLLECT,
				DevSFType.SFRunType.COLLECT,
				1,
				frameIndex,
				dir,
				new ArrayList<>(List.of(filename)),
				prefix,
				1,
				50.0,
				0.1,
				0,
				sftype,
				null,
				timeoutMs
			);
			cmdProcess.saveToRedis = true;
			SpotFinderQueue.submitTask(cmdProcess);
		}
	}

	public void waitForSFQueue()
	{
		while (true) {
			Util.sleep(1000);
			if (SpotFinderQueue.getActiveThreads() == 0) return;
		}
	}

	public boolean mismatch=false;
	@org.junit.Test
	public void testSFOutput()
	{
		String fulldir = "/mnt/beegfs/testdata/PC1", prefix="PC";
		int runLabel=1, frameCount = 5, frameStart=115, timeoutMs=20000;
		float[] dials_result_expected = new float[]{2f,2f,2f,2f,2f};
		float[] dials_spots_no_ice_expected = new float[]{0f,13f,26f,0f,0f};
		float[] dials_spots_expected = new float[]{0f,13f,26f,0f,0f};
		float[] dials_total_int_expected = new float[]{0f,1286f,2823f,0f,0f};
		float[] dozor_result_expected = new float[]{2.0f,2.0f,2.0f,2.0f,2.0f};
		float[] dozor_score_expected = new float[]{0f,.03f,.07f,0f,0f};
		float[] dozor_res_expected = new float[]{40f,5.66f,6.89f,40f,40f};

		DBImageStats.initialize(Redis.image_stats(fulldir, prefix, runLabel, ""), frameCount);

		for (int frameIndex = 0; frameIndex < frameCount; frameIndex += 1) {
			String image = String.format("PC_%d_%06d.cbf", runLabel, frameIndex + frameStart);
			startSF(fulldir, image, "PC", frameIndex, timeoutMs);
		}
		waitForSFQueue();

		// check dials results
		FloatBuffer dials_result = DBImageStats.load(Redis.image_stats(fulldir, prefix, runLabel, "dials:result"));
		FloatBuffer spots_no_ice = DBImageStats.load(Redis.image_stats(fulldir, prefix, runLabel, "dials:spots_no_ice"));
		FloatBuffer spots = DBImageStats.load(Redis.image_stats(fulldir, prefix, runLabel, "dials:spots"));
		FloatBuffer total_int = DBImageStats.load(Redis.image_stats(fulldir, prefix, runLabel, "dials:total_int"));
		System.out.println("dials result: "+toString(dials_result)+check_result(dials_result,dials_result_expected));
		System.out.println("dials spots_no_ice: "+toString(spots_no_ice)+check_result(spots_no_ice,dials_spots_no_ice_expected));
		System.out.println("dials spots: "+toString(spots)+check_result(spots,dials_spots_expected));
		System.out.println("dials total_int: "+toString(total_int)+check_result(total_int,dials_total_int_expected));

		// check dozor results
		FloatBuffer dozor_result = DBImageStats.load(Redis.image_stats(fulldir, prefix, runLabel, "dozor:result"));
		FloatBuffer score = DBImageStats.load(Redis.image_stats(fulldir, prefix, runLabel, "dozor:score"));
		FloatBuffer res = DBImageStats.load(Redis.image_stats(fulldir, prefix, runLabel, "dozor:res"));
		System.out.println("dozor result: "+toString(dozor_result)+check_result(dozor_result,dozor_result_expected));
		System.out.println("dozor score: "+toString(score)+check_result(score,dozor_score_expected));
		System.out.println("dozor res: "+toString(res)+check_result(res,dozor_res_expected));
		assertEquals(mismatch, false);
	}

	public String check_result(FloatBuffer actual, float[] expected)
	{
		float tolerance = 0.1f;
		for (int i=0;i<expected.length;i++) {
			if (Math.abs(actual.get(i)-expected[i])>tolerance) {
				mismatch=true;
				return "MISMATCH";
			}
		}
		return "OK";
	}

	public static String toString(FloatBuffer buf) {
		String ret = "";
		for (int i=0;i<buf.capacity();i++) {
			ret += String.format("%.2f ",buf.get(i));
		}
		return ret;
	}
}
