#!/opt/anaconda/bin/python

# the purpose of the script is to produce dataset definition for kamo using gmca database in the following
# data structure
# [('/mnt/beegfs/qxu/test-data/3r4k-p21212/140790_0_?????.cbf', 1, 3), ('/mnt/beegfs/qxu/test-data/3r4k-p21212/140790_1_E2_?????.cbf', 1, 250), ('/mnt/beegfs/qxu/test-data/3r4k-p21212/140790_2_?????.cbf', 1, 250), ('/mnt/beegfs/qxu/test-data/3r4k-p21212/140790_1_E1_?????.cbf', 1, 250)]

import sys, os, re
from peewee import *
from socket import gethostname
from collections import namedtuple
import datetime
from epics import PV
from collections import OrderedDict
from pprint import pprint


def get_db(bl):
    if bl.startswith('bl1'):
        db = MySQLDatabase('gmca_data',
                           host='bl1upper',
                           user='dhs',
                           passwd='')
    elif bl.startswith('bl2'):
        db = MySQLDatabase('gmca_data',
                           host='bl2upper',
                           user='dhs',
                           passwd='')
    elif bl.startswith('bl3'):
        db = MySQLDatabase('gmca_data',
                           host='bl3upper',
                           user='dhs',
                           passwd='')

    else:
        raise SystemExit('Error: Unable to continue, need to run on any bl1 or bl2 or bl3 computers.')

    return db


def get_bl():
    hn = gethostname()
    if hn.startswith("bl"):
        return hn[:3]
    else:
        raise SystemExit('Error: Unable to continue, cannot figure beamline')


def estimate_shift_range():
    # shift start and end time
    current = datetime.datetime.now()
    start = current.replace(hour=0, minute=0, second=0) - datetime.timedelta(days=2)
    end = current.replace(hour=10, minute=0, second=0) + datetime.timedelta(days=2)
    return start.strftime('%Y-%m-%d %H:%M:%S'), end.strftime('%Y-%m-%d %H:%M:%S')


def db_query(bl=None, db=None, username=None, start=None, end=None):
    # example select * from CollectEvents where username='jsmith' AND timestamp between '2018-07-15' and '2018-07-18' order by timestamp;
    if not bl:
        bl = get_bl()

    if not db:
        db = get_db(bl)

    if not username:
        username = os.getenv("USER")

    if not start or not end:
        start, end = estimate_shift_range()

    query = ("select collectType, workdir, prefix, firstframe, numimages, numsites, framepersite, "
             "omegastart, omegaend, inversebeam, wedge, finalEnergy, detectorDistance,timestamp "
             "from CollectEvents "
             "where username='{}' "
             "AND timestamp between '{}' and '{}' "
             "AND numimages >=4 "
             "order by timestamp asc").format(username, start, end)

    rs = db.execute_sql(query)
    nd = namedtuple('datasets',
                    ['collect_type', 'image_dir', 'prefix', 'first_frame', 'num_images', "num_sites", "frame_per_site",
                     "omega_start", "omega_end", "inverse_beam", "wedge", "energy", "distance", "time_stamp"])

    return [nd(*r) for r in rs]


def get_primary_data_directory(beamline=None):
    if beamline not in ("23o", "23i", "23b") or not beamline:
        hn = gethostname()
        if hn.startswith("bl1"):
            beamline = "23i"
        elif hn.startswith("bl2"):
            beamline = "23o"
        elif hn.startswith("bl3"):
            beamline = "23b"
        else:
            return guess_primary_data_directory()

        p = PV("{}:bi:data_dir".format(beamline))
        pdir = p.get(as_string=True)

        if pdir:
            return pdir
        else:
            return guess_primary_data_directory()


def guess_primary_data_directory():
    user_home = os.getenv("HOME")
    directory_pattern = re.compile("^23(BM|IDB|IDD)_\d{4}_\d{2}_\d{2}(_cbf|hdf5)?$")

    candidates = []
    for f in os.listdir(user_home):
        if os.path.isdir(os.path.join(user_home,f)) and re.match(directory_pattern, f):
            date = datetime.datetime.strptime("-".join(f.split("_")[1:4]), '%Y-%m-%d')
            now = datetime.datetime.now()
            candidates.append((f, now - date))

    # return the directory closest to the current time
    return min(candidates, key=lambda f: f[1])[0] if candidates else None

def find_datasets_from_templatefile(list_of_datasets):
    def template_to_filename(template, number):
        if isinstance(number, str):
            number = int(number)

        sep = "?"
        ndigits = template.count(sep)
        if ndigits == 0:
            sep = "#"
            ndigits = template.count(sep)

        num = "{:0{}d}".format(number, ndigits)
        filename = template.replace(sep*ndigits, num)
        return filename

    def validate_entry(template, start, end):
        # check the files are actually exist
        image_nums = [i for i in range(start, end+1) if os.path.exists(template_to_filename(template, i))]
        if len(image_nums) >= 1:
            return [template, image_nums[0], image_nums[-1]]
        else:
            return None

    datasets = []
    if os.path.exists(list_of_datasets):
        with open(list_of_datasets) as f:
            lines = f.readlines()

        # remove duplicates
        new_lines = []
        for l in lines:
            if l not in new_lines:
                new_lines.append(l)

        candidates = [c for c in [l.rstrip().split(",") for l in new_lines] if len(c) == 3]
        for c in candidates:
            template, start, end = c
            if int(start) == int (end) == 0:
                continue

            validated_sweep = validate_entry(template, int(start), int(end))
            if validated_sweep:
                datasets.append(validated_sweep)

    return datasets

    def update_jobs_from_files(self, root_dir, include_dir=[], exclude_dir=[], dataset_paths_txt=None):
        from yamtbx.dataproc import dataset
        from yamtbx.dataproc.bl_logfiles import JobInfo
        from yamtbx.dataproc import XIO

        if include_dir == []: include_dir = [root_dir]
        # XXX what if include_dir has sub directories..

        for rd in include_dir:
            if dataset_paths_txt and os.path.exists(dataset_paths_txt):
                datasets = dataset.find_datasets_from_templatefile(dataset_paths_txt)
            else:
                datasets = dataset.find_data_sets(rd, skip_0=True, skip_symlinks=False, split_hdf_miniset=config.params.split_hdf_miniset)
            for ds in datasets:
                tmpl, nr = ds[0], tuple(ds[1:3])
                prefix = tmpl[:tmpl.index("_?" if "_?" in tmpl else "?")]
                    
                if not directory_included(tmpl, root_dir, [], exclude_dir):
                    mylog.info("This directory is not in topdir or in exclude_dir. Skipped: %s"%tmpl)
                    continue

                job = JobInfo(None)
                job.filename = tmpl

                images = filter(lambda x: os.path.isfile(x), dataset.template_to_filenames(*ds))

                if len(images) == 0:
                    continue

                h = XIO.Image(images[0]).header
                job.osc_end = h.get("PhiEnd", 0)

                if len(images) > 1:
                    h_next = XIO.Image(images[1]).header
                    h_last = XIO.Image(images[-1]).header
                    job.osc_end = h_last.get("PhiEnd", 0)
                    if h_next.get("PhiStart", 0) == h.get("PhiStart", 0):
                        print "This job may be scan?:",  tmpl
                        continue

                job.wavelength = h.get("Wavelength", 0)
                job.osc_start =  h.get("PhiStart", 0)
                job.osc_step = h.get("PhiWidth", 0)
                job.status = "finished"
                job.exp_time = h.get("ExposureTime", 0)
                job.distance = h.get("Distance", 0)
                job.attenuator = None, 0
                job.detector = "?"
                job.prefix = prefix

                if job.osc_step == 0 or job.osc_end - job.osc_start == 0:
                    print "This job don't look like osc data set:",  tmpl
                    continue
                    
                if config.params.split_data_by_deg is None or job.osc_step==0:
                    self.jobs[(prefix, nr)] = job
                    self.jobs_prefix_lookup.setdefault(prefix, set()).add(nr)
                else:
                    n_per_sp = int(config.params.split_data_by_deg/job.osc_step+.5)
                    for i in xrange(nr[1]//n_per_sp+1):
                        if (i+1)*n_per_sp < nr[0]: continue
                        if nr[1] < i*n_per_sp+1: continue
                        nr2 = (max(i*n_per_sp+1, nr[0]), min((i+1)*n_per_sp, nr[1]))
                        self.jobs[(prefix, nr2)] = job # This will share the same job object.. any problem??
                        self.jobs_prefix_lookup.setdefault(prefix, set()).add(nr2)

        # Dump jobs
        pickle.dump(self.jobs, open(os.path.join(config.params.workdir, "jobs.pkl"), "wb"), 2)
    # update_jobs_from_files()




if __name__ == "__main__":
    #print(get_primary_data_directory())
    print(guess_primary_data_directory())

    file = "/mnt/beegfs/melias/23IDB_2018_07_25_cbf/datasets.txt"
    datasets = find_datasets(file)
    print("datasets = ", datasets)

    sys.exit()
    print(estimate_shift_range())
    r = db_query(bl='bl2', username='jsmith')
    print(r)
