from __future__ import absolute_import
import itertools
import logging
import os
import re
from collections import Counter, defaultdict

from .filesweep import FileSweep
from .imagesweep import ImageSweep

logger = logging.getLogger(__name__)


class GenerateSweeps:
    """ given directory or directories, parse images files and organize them
    into sweeps based on file name characteristics, assume file size
    > 3 M bytes"""

    fname_pattern = r"""(\d{3,6}\.|\.\d{3,6}|master\.)(?=.{0,3}$)"""
    file_pattern = re.compile(fname_pattern, re.VERBOSE)
    min_file_size = 3 * 1024 * 1024
    excluded_suffices = ['.txt', '.jpg', '.png', '.py', '.INP', '.HKL', '.hkl', '.img']
    included_suffices = ['.cbf', '.hdf5', '.h5']

    def __init__(self, directory,
                 filelist=None,
                 images=None,
                 prefix=None,
                 suffix=None,
                 str_filter=None,
                 start=None,
                 end=None,
                 percent=None,
                 split_gap=True,
                 max_images_per_sweep=10000,
                 min_images_per_sweep=1,
                 skip_sweeps_with_nimages_ge=None,
                 excluded_patterns=["_chk_", "_raster_", "core."]
                 ):
        self.start = start
        self.end = end
        self.percent = percent
        self.prefix = prefix
        self.suffix = suffix
        self.str_filter = str_filter
        self.split_gap = split_gap
        self.max_images_per_sweep = max_images_per_sweep
        self.min_images_per_sweep = min_images_per_sweep
        self.abs_path = os.path.abspath(directory)
        self.rel_path = os.path.relpath(directory)
        self.filelist = filelist
        self.images = images
        self.excluded_patterns = excluded_patterns
        self.skip_sweeps_with_nimages_ge = skip_sweeps_with_nimages_ge

        self.file_list = []
        if self.filelist:
            self.file_list = self.__parse_filelist()
        elif self.images:
            self.file_list = [f for f in self.images if os.path.exists(os.path.join(self.abs_path, f))]

        if not self.file_list:
            self.file_list = self.__parse_img_directory()

        self.__sweeps = self.__generate_sweeps()
        self.file_sweeps = [FileSweep(s, self.abs_path)
                            for s in self.__sweeps]

    def get_file_sweeps(self):
        return self.file_sweeps

    def get_image_sweeps(self, sort_by_mtime=False):
        xs = [ImageSweep(fs, start_frame=self.start, end_frame=self.end)
              for fs in self.file_sweeps]

        if sort_by_mtime:
            xs.sort(key=lambda x: os.path.getmtime(x.get_first_file(abs_path=True)))
        return xs

    def get_image_sweeps_groupby_wavelength(self, rounded_by=4):
        """ group sweeps by wavelength, rounded to 4 digits 
         wave4 = [ round(s.get_wavelength(), rounded_by) for s in image_sweeps ] """
        res = defaultdict(list)
        for s in self.get_image_sweeps():
            key = round(s.get_wavelength(), rounded_by)
            res[key].append(s)
        return res

    def __len__(self):
        return len(self.__sweeps)

    def __parse_img_directory(self):
        """parse one directory, look for images of certain pattern defined below"""
        directory = self.abs_path
        prefix = self.prefix
        suffix = self.suffix
        str_filter = self.str_filter

        if not os.path.isdir(directory):
            logger.error("images does not exist in {}".format(directory))
            return []

        file_list = [f for f in os.listdir(directory)
                     if os.path.isfile(os.path.join(directory, f))]

        if prefix:
            # file_list = [f for f in file_list if f.startswith(prefix)]
            prefixes = prefix.replace(" ", "").split(";")
            file_list = [f for f in file_list if any(f.startswith(p) for p in prefixes)]

        if suffix:
            file_list = [f for f in file_list if f.endswith(suffix)]

        if str_filter:
            file_list = [f for f in file_list if str_filter in f]

        # only allow suffices .cbf or h5
        file_list = [f for f in file_list if any(f.endswith(suffix) for suffix in GenerateSweeps.included_suffices)]

        file_list = [f for f in file_list if not any(f.endswith(suffix) for suffix in GenerateSweeps.excluded_suffices)]

        # skip site-check images -- GMCA specific
        file_list = [f for f in file_list if not any(p in f for p in self.excluded_patterns)]

        file_list = [f for f in file_list
                     if os.path.getsize(os.path.join(directory, f))
                     >= GenerateSweeps.min_file_size]

        file_list = [f for f in file_list
                     if re.findall(GenerateSweeps.file_pattern, f)]

        file_list = sorted(file_list, key=lambda x: (len(x), x))

        return file_list

    def __parse_filelist(self):
        if os.path.isfile(self.filelist):
            file = self.filelist
        elif os.path.isfile(os.path.join(self.abs_path, self.filelist)):
            file = os.path.join(self.abs_path, self.filelist)
        else:
            logger.error("file {} does not exist, ignored.".format(self.filelist))
            return None

        with open(file) as f:
            file_list = f.readlines()

        file_list = [f.rstrip() for f in file_list]
        if self.prefix:
            prefixes = self.prefix.replace(" ", "").split(";")
            file_list = [f for f in file_list if any(f.startswith(p) for p in prefixes)]

        return file_list

    def __generate_sweeps(self):
        """ segment the images in a directory or directories"""
        fsplit = [re.split(GenerateSweeps.file_pattern, f)
                  for f in self.file_list]
        prefix_suffix_counter = Counter([(item[0], item[-1])
                                         for item in fsplit])

        # deal with case that end is specified as a fraction/percentage of images
        sweep_bounds = {}
        if self.percent:
            ranges = defaultdict(list)
            for key in prefix_suffix_counter:
                for f in fsplit:
                    if (f[0], f[-1]) == key:
                        ranges[key].append(int(f[1][:-1]))

            for key in prefix_suffix_counter:
                startnum, endnum = ranges[key][0], ranges[key][-1]
                # startnum = self.start if startnum < self.start else startnum
                new_endnum = int((endnum - startnum + 1) * self.percent / 100.0)
                # new_endnum = int(5 * round(new_endnum // 5))  # round to a more memorable number
                if new_endnum <= startnum + 5:
                    logger.warning(
                        "the requested sweep has very few images, check input parameters --percent and --start.")
                sweep_bounds[key] = (startnum, new_endnum)

        # filter by range TODO not working for hdf5 scheme!!
        sweeps = defaultdict(list)
        for key in prefix_suffix_counter:
            for f in fsplit:
                if (f[0], f[-1]) == key:
                    add = True
                    # filtering by sequence number does not work for hdf5
                    # if 'master' not in f[1]:
                    if f[-1] not in ['h5', 'hdf5']:
                        midf = int(f[1][:-1])

                        if self.start and midf < self.start:
                            add = False

                        if self.end and midf > self.end:
                            add = False

                        if sweep_bounds and (midf < sweep_bounds[key][0] or midf > sweep_bounds[key][1]):
                            add = False

                    if add:
                        sweeps[key].append(f)

        # split a sweep is there is a gap in the middle number
        sweeps2 = []
        for v in sweeps.values():
            sweeps2.extend(self.__split_middle(v))

        # merge hdf5 master and data files together
        sweeps3 = self.__merge_hdf5(sweeps2)

        # remove sweeps that has few images than specified
        sweeps3 = [s for s in sweeps3 if len(s) >= self.min_images_per_sweep]

        # further divide into smaller sweeps if the max_images_per_sweep is set
        sweeps4 = []
        for sw in sweeps3:
            if self.skip_sweeps_with_nimages_ge:
                if len(sw) >= self.skip_sweeps_with_nimages_ge:
                    continue

            if len(sw) > self.max_images_per_sweep:
                sweeps4.extend(split_max_length(sw, self.max_images_per_sweep))
            else:
                sweeps4.append(sw)

        return sweeps4

    def __split_middle(self, alist):
        """ given a list of split filenames, further divide it into
        sub lists if the numbers in the middle aren't continuous,
        return a list of list """
        if not self.split_gap or len(alist) <= 1:
            return [alist]

        numlist = [int(mid_field[1][:-1]) for mid_field in alist]

        ll, tmpl = [], [alist[0]]
        for i in range(1, len(numlist)):
            if numlist[i] - 1 == numlist[i - 1]:
                tmpl.append(alist[i])
            else:
                ll.append(tmpl)
                tmpl = [alist[i]]
        ll.append(tmpl)

        if len(ll) > 1:
            logger.info("data files have gaps in between, split into smaller sweeps, except for hdf5.")
            if 'h5' in alist[0][-1]:
                return [alist]

        return ll

    def __merge_hdf5(self, sweeps2):
        """ merge hdf5 master and datafiles in a sweep together into one list """
        merge_pairs = []
        for i, e1 in enumerate(sweeps2):
            if 'master.' in e1[0][1]:
                for j, e2 in enumerate(sweeps2):
                    if i != j:
                        if e1[0][0] + 'data_' == e2[0][0]:
                            merge_pairs.append((i, j))

        mergelist = set(list(itertools.chain(*merge_pairs)))
        nomerged = set(range(len(sweeps2))) - mergelist

        llist_merged = [sweeps2[i] for i in nomerged]
        llist_merged += [sweeps2[i] + sweeps2[j] for i, j in merge_pairs]

        return llist_merged


def split_max_length(alist, max_length):
    """ split a list into sub lists if it exceeds maximum specified length
        the sublist may be slightly bigger than max_length to avoid tiny
        sweep leftover """
    half = max_length // 2

    while 0 < len(alist) % max_length < half:
        max_length -= 1

    max_length = max(1, max_length)
    return [alist[i:i + max_length] for i in range(0, len(alist), max_length)]


if __name__ == '__main__':
    # dir = "/home/qxu/test-data/jsmith-idb-test1/cbf"
    # flist = parse_img_directory(dir)
    # print generate_sweeps(flist)

    # imgs = generate_sweeps("/home/qxu/test-data/16eiger-2/2015_11_10",
    #                   start=1, end=50, split_gap=True)

    # pprint(imgs.sweeps)
    # print "files = ", imgs.get_file_list()

    dir1 = '/home/qxu/test-data/jsmith-idb-test2/cbf'
    dir2 = '/home/qxu/test-data/16eiger-2/2015_11_10'
    dir3 = '/home/qxu/test-data/jsmith-idb-test1/cbf'
    dir4 = '../../test-data/jsmith-idb-test2/cbf'

    dir5 = '/home/qxu/data-for-tests/test-data/B3-spg-check/'
    dir5 = '/home/qxu/data-for-tests/tab11/'
    dir5 = '/home/qxu/data-for-tests/test-data/B3-spg-check/cbf'
    dir5 = '/home/qxu/junk'

    imgs = GenerateSweeps(dir5, split_gap=True)

    # print "file_list = ", imgs.get_file_list()
    print("nsweeps = ", len(imgs))
    fs = imgs.get_file_sweeps()
    for s in fs:
        print("first image = ", s.get_first_file())
        print("name template = ", s.get_name_template(marker='#'))
        print("abs dir = ", s.get_absdir())
        print("prefix, suffix = ", s.get_prefix(), s.get_suffix())
        print("nfiles = ", s.nfiles)
        print("hdf5 or cbf = ", s.is_hdf5(), s.is_cbf())
        # print s.get_all_files(abs_path=True)
        print(s.get_file(3))
        print(s.get_id())

    print("=" * 80)

    xs = imgs.get_image_sweeps()

    for s in xs:
        print("first image = ", s.get_first_file())
        print("name template = ", s.get_name_template(marker='#'))
        print("abs dir = ", s.get_absdir())
        print("prefix, suffix = ", s.get_prefix(), s.get_suffix())
        print("nfiles = ", s.nfiles)
        print("hdf5 or cbf = ", s.is_hdf5(), s.is_cbf())
        # print s.get_all_files(abs_path=True)
        print(s.get_file(3))
        print("beam= ", s.get_metadata_by_key('beam_in_pixel'))
        print("distance ", s.get_metadata_by_key('distance'))
        print("data range = ", s.get_data_range(), s.get_id())
        print("wavelength = ", round(s.get_wavelength(), 4))
        print("nimages per file", s.nimages_per_datafile())
        print("ith the image", s.get_image(1))
        print("." * 80)
        # print(imgs.get_image_sweeps_groupby_wavelength())
