from __future__ import absolute_import
import logging, os
from collections import namedtuple
from datetime import time

from ..common import utils
from ..common.header import Header, get_max_resolution_on_detector
from .filesweep import FileSweep

logger = logging.getLogger(__name__)


class ImageSweep(FileSweep):
    """ image sweep data structure for data processing """

    def __init__(self,
                 a_file_sweep,
                 start_frame=1,
                 end_frame=None):
        self.filelist = a_file_sweep.get_filelist()
        self.directory = a_file_sweep.get_absdir()
        super(ImageSweep, self).__init__(self.filelist, self.directory)
        self.metadata = self.__header()
        self.start_frame = start_frame
        self.end_frame = end_frame

        self.filesweep_id = a_file_sweep.get_id()
        self.sweep_id = self.get_sweep_id()
        self.image_numbers = list()

    def __header(self):
        first_image = self.get_first_file()
        utils.wait_until_files_exist(first_image, timeout=120)
        imgheader = Header(first_image)
        return imgheader.get_header_as_dict()

    def get_metadata_by_key(self, key):
        try:
            v = self.metadata[key]
            return v
        except KeyError:
            logger.error("requested key {} not found.".format(key))
            raise

    def get_wavelength(self):
        return self.metadata['wavelength']

    def get_distance(self):
        return self.metadata['distance']

    def get_edge_res(self):
        return self.metadata['edge_res']

    def get_corner_res(self):
        return self.metadata['corner_res']

    def get_osc(self):
        return self.metadata['phi_width']

    def get_twotheta(self):
        return self.metadata['two_theta']

    def get_start_angle(self):
        if self.is_hdf5():
            self.start_frame = 1 if not self.start_frame else self.start_frame
            return ((self.start_frame - 1) * self.metadata['phi_width']
                    + self.metadata['phi_start'])
        else:
            return self.metadata['phi_start']

    def get_metadata_as_dict(self):
        start, end = self.get_data_range()
        self.metadata['start'] = start
        self.metadata['end'] = end
        self.metadata['imagedir'] = self.get_absdir()
        self.metadata['template'] = self.get_name_template()
        self.metadata['imageSet'] = "s:{}e:{}".format(start, end)
        self.metadata['firstfile'] = self.get_first_file(abs_path=False)
        self.metadata['firstFrame'] = self.get_first_file(abs_path=False)
        self.metadata['sweepid'] = self.get_sweep_id()
        self.metadata['nframes'] = self.nimages()
        self.metadata['nfiles'] = self.nfiles
        self.metadata['prefix'] = self.get_prefix()

        return self.metadata

    def get_data_range(self):
        if self.is_hdf5():
            start = 1
            if self.start_frame and self.start_frame > 1:
                start = self.start_frame
            end = self.get_metadata_by_key('nimgs')
            if self.end_frame and self.end_frame < end:
                end = self.end_frame
            return start, end
        else:
            return self.get_start_file_number(), self.get_last_file_number()

    def get_sweep_id(self):
        start, end = self.get_data_range()
        return "{}{:05d}_{:05d}".format(self.filesweep_id, start, end)

    def get_image(self, i, abs_path=True):
        # return the ith image filename in an image sweep
        if self.is_hdf5():
            nimg_per_file = self.nimages_per_datafile()
            seq = (i - 1) // nimg_per_file + 1
            return self.get_file(seq, abs_path=abs_path)
        else:
            location = i - self.get_start_file_number()
            return os.path.join(self.directory,
                                "".join(self.filelist[location]))

    def get_phi(self, i):
        # return the start phi value for the ith image filename in an image sweep
        image = self.get_file(i, abs_path=True)
        h = Header(image)
        return h.get_item_by_key('phi_start')

    def nimgs(self):
        start, end = self.get_data_range()
        return end - start + 1

    def nimages(self):
        return self.nimgs()

    def nimages_per_datafile(self):
        if self.is_hdf5:
            return self.nimages() // (self.nfiles - 1) if self.nfiles > 1 else 1
        else:
            return 1

    def sweep_summary_as_dict(self):
        return {'prefix': self.get_prefix(),
                'data_range': self.get_data_range(),
                'seqlength': self.get_seqlength(),
                'suffix': self.get_suffix()}

    def sweep_summary_as_namedtuple(self):
        Sweep = namedtuple('Sweep', 'prefix data_range seqlength suffix')
        return Sweep(prefix=self.get_prefix(),
                     data_range=self.get_data_range(),
                     seqlength=self.get_seqlength(),
                     suffix=self.get_suffix())

    def get_autoPROC_ids(self):
        if not self.is_hdf5():
            # idN = self.get_prefix().strip('_')
            idN = self.get_id()  # self.sweep_id  # needs to be unique
            dirN = self.get_absdir()
            templateN = self.get_name_template(marker='#')
            startN, endN = self.get_data_range()
            return '-Id ' + ','.join([idN, dirN, templateN, str(startN), str(endN)])
        else:
            return '-h5 ' + self.get_first_file()

    def get_xia2_sweep_defintion(self):
        # image=/dls/i02/data/2011/mx1234-5/K5_M1S3_3_001.img:1:200
        startN, endN = self.get_data_range()
        first_file = self.get_first_file(abs_path=True)
        return 'image={}:{}:{}'.format(first_file, startN, endN)

    def get_data_dir(self):
        return self.directory

    def wait_for_all_files(self):
        utils.wait_until_files_exist(self.get_all_files(abs_path=True), timeout=600)

    def wait_for_first_file(self):
        utils.wait_until_files_exist(self.get_first_file(abs_path=True), timeout=600)

    def get_total_osc(self):
        osc = self.get_osc()
        return float(osc) * self.nimgs()

    def print_summary(self):
        super(ImageSweep, self).print_summary()
        logger.info("beam center= {}".format(self.get_metadata_by_key('beam_in_pixel')))
        logger.info("distance= {}".format(self.get_metadata_by_key('distance')))
        logger.info("wavelength= {}".format(self.get_metadata_by_key('wavelength')))
        logger.info("data range = {}".format(self.get_data_range()))

    def get_detector_distance_given_max_res(self, max_res, default_distance=350, tolerance=1.0):
        nx, ny = self.metadata['size']
        qx, qy = self.metadata['pixel_size']
        orgx, orgy = self.metadata['beam_in_pixel']
        distance0 = self.metadata['distance']
        wavelength = self.metadata['wavelength']
        two_theta = self.metadata['two_theta']

        if max_res >= 5:
            logger.info("distance set to {}mm due to resolution <= 5A".format(default_distance))
            return default_distance

        for d in range(210, 500, 10):
            max_edge_res, _ = get_max_resolution_on_detector((nx, ny), (orgx, orgy), qx, wavelength, d,
                                                             detector_tilt=two_theta, edge_only=True)
            if max_res - max_edge_res <= tolerance:
                logger.info(
                    "distance for collection={}mm, suggested detector distance = {}mm based on resolution {}A at edge.".format(
                        distance0, d, round(max_res - tolerance, 1)))
                return d

        logger.warning("failed to obtain detector distance, use default {}mm".format(default_distance))
        return default_distance
