from __future__ import absolute_import

import logging
import os, re
from collections import namedtuple
from math import exp

from ..common.spacegroup import Spacegroup
from .fileopener import FileOpener
from ..common.utils import getNumbers
from .table import Table
from .fileparser import FileParser

logger = logging.getLogger(__name__)


def screening_score(resol, rms, mosaic, mosaic_penalty=1.0):
    # score as calculated in the webice
    try:
        score = (1.0 - 0.7 * exp(-4.0 / float(resol))
                 - 1.5 * float(rms)
                 - 0.2 * float(mosaic) * float(mosaic_penalty))
        score = round(score, 3)
    except Exception:
        logger.warning('Failed to calculate score screening score.')
        score = 0

    return score


class LabelitParser(FileOpener):
    def __init__(self, wdir=".", filename='labelit_autoindex.out'):
        super(LabelitParser, self).__init__(wdir, filename)
        self.mosaic_penalty = 1.0
        self.candidates_table = None
        self.spot_stat = None
        self.labelit_suggestion = None
        self.OK = False
        self.error_msg = None

        #  now populate variables above
        self.__get_spot_statistics_asdict()
        self.__get_autoindex_table()
        if self.candidates_table:
            self.__get_mosaic_penalty()
            self.__get_labelit_choice()
            self.OK = True
        self.index_table = None

    def __get_autoindex_table(self):
        blocks = self.parser.get_blocks_between(text1='Solution  Metric fit',
                                                text2='MOSFLM Integration results:',
                                                skip1=1,
                                                skip2=2, iMatch=1)
        if not blocks:
            logger.warning('LABELIT failed, no autoindex results found.')
            return None

        tbl = Table()
        tbl.set_title('LABELIT index results, all candidates')
        tbl.set_headers(['status', 'solNum', 'Metric fit', 'rmsd', 'nspots', 'crystal_system', 'unit_cell'])
        for line in blocks[0]:
            fields = line.split()
            status = fields[0]
            solNum = fields[1]
            metric_fit = fields[2]
            rmsd = fields[4]
            nspots = fields[5]
            crystal_system = fields[7]
            unit_cell = ' '.join([str(round(float(fld), 1)) for fld in fields[8:14]])
            tbl.add_data_by_a_list([status, solNum, metric_fit, rmsd, nspots, crystal_system, unit_cell])

        self.index_table = tbl.format()
        self.candidates_table = tbl

    # extract all possible space or point groups from index solutions into the following format:
    # point_group lattice penalty cell;...
    def get_possible_space_groups(self, sgtype='spacegroup'):
        if not self.candidates_table:
            self.__get_autoindex_table()

        possible_space_groups = []
        for line in self.candidates_table.get_data():
            lattice = line[-2]
            if sgtype.startswith('space'):
                pgs = Spacegroup.get_space_groups_from_lattice(lattice, sgtype=sgtype, output_mode='symbol')
            else:
                pgs = Spacegroup.get_space_groups_from_lattice(lattice, sgtype='pointgroup', output_mode='symbol')

            for spg in pgs:
                possible_space_groups.append(' '.join([spg, 'lattice:', lattice, 'penalty:', line[2], 'cell:', line[-1].replace(' ', ',')]))
        return ';'.join(possible_space_groups)

    def get_raw_index_table(self):
        if not self.index_table:
            self.__get_autoindex_table()
        return self.index_table

    def __get_spot_statistics_asdict(self, findspots='labelit_spotfinder.out'):
        """ return labelit distl spot statistics, in a dictionary (of dictionary)"""
        fn = os.path.join(self.wdir, findspots)
        blocks = FileParser(fn).get_blocks_between(text1='File :',
                                                   text2='In-Resolution Ovrld Spots :',
                                                   iMatch=-1)  # all blocks
        spot_stat = {}
        for block in blocks:
            key = nspots = resol1 = resol2 = ice_rings = None
            for line in block:
                line = line.strip('\n')
                if 'File :' in line:
                    key = line.split(':')[-1].replace(' ', '')
                elif 'Good Bragg Candidates :' in line:
                    nspots = line.split(':')[-1].replace(' ', '')
                elif 'Method 1 Resolution :' in line:
                    resol1 = line.split(':')[-1].replace(' ', '')
                elif 'Method 2 Resolution :' in line:
                    resol2 = line.split(':')[-1].replace(' ', '')
                    resol2 = None if resol2 == 'None' else resol2
                elif 'Ice Rings :' in line:
                    ice_rings = line.split(':')[-1].replace(' ', '')
            if key:
                spot_stat[key] = dict(nspots=nspots,
                                      resol1=resol1,
                                      ice_rings = ice_rings,
                                      resol2=resol2)
        # a dictionary with image names as keys, dict of dict
        # self.spot_stat = spot_stat

        final_results = {}
        images = ';'.join(spot_stat.keys())
        final_results['software'] = 'labelit-strategy'
        final_results['state'] = 'SPOT'
        final_results['images'] = images
        final_results['n_spots'] = sum(int(spot_stat[k]['nspots']) for k in spot_stat if spot_stat[k]['nspots'])
        final_results['n_ice_rings'] = sum(int(spot_stat[k]['ice_rings']) for k in spot_stat if spot_stat[k]['ice_rings'])
        res = [float(spot_stat[k]['resol2']) for k in spot_stat if spot_stat[k]['resol2']]
        if res:
            final_results['resolution_from_spots'] = min(res)
        final_results['n_spots_ice'] = 'NA'
        final_results['avg_spotsize'] = 'NA'
        self.spot_stat = final_results
        return final_results

    def __get_labelit_choice(self):
        lines = self.parser.get_lines_after(text='Solution  SpaceGroup', nLines=10, iMatch=1)

        if not lines:
            logger.error('failed to parse LABELIT results.')
            return None

        for line in lines:
            if line.startswith(':)'):
                fields = line.split()
                if len(fields) == 9:
                    status, solNum, spg, beamx, beamy, dist, resol, mosaic, rmsd = fields

                    unit_cell = None
                    for row in self.candidates_table.get_data():
                        if solNum == row[1]:
                            unit_cell = row[-1]
                            break

                    score = screening_score(resol, rmsd, mosaic, mosaic_penalty=self.mosaic_penalty)
                    solution = namedtuple('LabelitSuggestedSolution',
                                          'status, solNum, space_group, mosaic, resol, rmsd, beamx, beamy, dist, unit_cell, score')
                    self.labelit_suggestion = solution(status=status,
                                                       solNum=solNum,
                                                       space_group=spg,
                                                       rmsd=rmsd,
                                                       resol=resol,
                                                       beamx=beamx,
                                                       beamy=beamy,
                                                       dist=dist,
                                                       mosaic=mosaic,
                                                       unit_cell=unit_cell,
                                                       score=score)
                    break

    def __get_mosaic_penalty(self):
        lines = self.parser.get_lines_contain('Beam center')
        if lines:
            if "60%(POOR)" in lines[0]:
                self.mosaic_penalty = 2.0
            return float(getNumbers(lines[0])[-1])
        return None

    def get_solution_based_on_symm(self, space_group):
        # look up solution number from the index table based on
        # the user provided symmetry
        lattice = Spacegroup.get_lattice_from_spacegroup(space_group)
        if self.candidates_table:
            for row in self.candidates_table.get_data():
                if lattice == row[5]:
                    logger.info("choose solution # {} with lattice {} "
                                "based on given space group {}.".format(row[1],
                                                                        lattice,
                                                                        space_group))

                    solution = namedtuple('LabelitSolution',
                                          'status, solNum, metric_fit, space_group, rmsd, nspots, crystal_system, unit_cell, score')

                    score = 'NA'
                    return solution(status=row[0],
                                    solNum=row[1],
                                    metric_fit=row[2],
                                    rmsd=row[3],
                                    nspots=row[4],
                                    space_group=space_group,
                                    crystal_system=row[5],
                                    score=score,
                                    unit_cell=row[6])

        return None

    def get_labelit_suggestion(self):
        return self.labelit_suggestion

    def get_spot_stat(self):
        return self.spot_stat

    def get_error_msg(self):
        # if no index table found, extract last text block as the error message
        if not self.candidates_table:
            message = []
            lines = self.parser.get_lines()
            for line in lines[::-1]:
                if line == '\n':
                    break
                message.append(line)
            self.error_msg = ''.join(message[::-1])
            return self.error_msg

        return None


if __name__ == '__main__':
    l1 = LabelitParser(wdir='/home/qxu/work/quickProcess/quickProcess/tests', filename='labelit.out.1')
    l2 = LabelitParser(wdir='/home/qxu/work/quickProcess/quickProcess/tests', filename='labelit.out.2')
