from __future__ import absolute_import

import logging, os
import re
from collections import namedtuple

from ..common.spacegroup import Spacegroup
from .fileopener import FileOpener
from ..common.utils import getNumbers
from .table import Table

logger = logging.getLogger(__name__)


class MosflmParser(FileOpener):
    def __init__(self, wdir=".", filename='mosflm-autoindex.log'):
        super(MosflmParser, self).__init__(wdir, filename)
        self.table_autoindex = None
        self.index_table = None  # raw mosflm index table

    def get_findspots_stat(self):
        mark1 = '===> findspots find '
        mark2 = 'spots were written to file'  # mark2 = '===>'
        blocks = self.parser.get_blocks_between(text1=mark1, text2=mark2, iMatch=-1, EOF=True)  # find all matches

        if not blocks:
            logging.warning('Spot finding results not found.')
            return None

        spot_stat = {}
        for block in blocks:
            key = nspots = resol1 = resol2 = nspots_ice = spotsize = threshold_autoindex = None
            for line in block:
                if 'image FILENAME:' in line:
                    key = os.path.split(line)[-1].replace('\n','')

                if 'highest resolution is' in line:
                    resol1 = getNumbers(line)[-1]

                if '99% have resolution less than' in line:
                    resol2 = getNumbers(line)[-1]

                if 'Number of spots excluded from possible ice rings:' in line:
                    nspots_ice = getNumbers(line)[-1]

                if 'spots were written to file' in line:
                    nspots = getNumbers(line)[-1]
                elif 'NO SPOTS have been found on image' in line:
                    nspots = 0

                if 'Based on a spot size of' in line:
                    spotsize = 'x'.join(getNumbers(line))

                if 'I/sig(I) threshold for autoindexing set to:' in line:
                    threshold_autoindex = getNumbers(line)[-1]
            if key:
                spot_stat[key] = dict(nspots=nspots,
                                      resol1=resol1,
                                      resol2=resol2,
                                      nspots_ice=nspots_ice,
                                      spotsize=spotsize,
                                      threshold_autoindex=threshold_autoindex)

        # return spot_stat
        # logger.info(spot_stat)
        # combine dict of dict into a single dict
        final_results = {}
        images = ';'.join(spot_stat.keys())
        final_results['software'] = 'mosflm_strategy'
        final_results['state'] = 'SPOT'
        final_results['images'] = images

        nspots = [int(spot_stat[k]['nspots']) for k in spot_stat if spot_stat[k]['nspots']]
        if nspots:
            final_results['n_spots'] = '{}'.format(sum(nspots))

        final_results['n_spots_ice'] = sum(
            int(spot_stat[k]['nspots_ice']) for k in spot_stat if spot_stat[k]['nspots_ice'])
        final_results['n_ice_rings'] = 'NA'

        # # use maximum resolution
        # reso = [float(spot_stat[k]['resol2']) for k in spot_stat if spot_stat[k]['resol2']]
        # final_results['resolution_from_spots'] = '{}'.format(min(reso)) if reso else 'NA'

        # use average resolution weighted by # of spots
        sum1 = sum2 = 0
        for k in spot_stat:
            if spot_stat[k]['nspots'] and spot_stat[k]['resol2']:
                sum1 += int(spot_stat[k]['nspots']) * float(spot_stat[k]['resol2'])
                sum2 += int(spot_stat[k]['nspots'])
        try:
            avg_resol = round(sum1/sum2,2)
        except ZeroDivisionError:
            logger.warning("cannot estimate resolution, spot {}, {}")
            avg_resol = 'NA'

        final_results['resolution_from_spots'] = '{}'.format(avg_resol)

        final_results['avg_spotsize'] = ';'.join(
            spot_stat[k]['spotsize'] for k in spot_stat if spot_stat[k]['spotsize'])
        # logger.info(final_results)
        return final_results

    def get_autoindex_table(self,
                            penalty_cutoff=100.0,
                            sd_cutoff=0.25,
                            score_cutoff=100,
                            remove_duplication=True,
                            raw_output=False):
        """
        header of the index table looks like either of the following, depending on whether 
        refine was used or not
        
        No PENALTY LATT     a        b        c     alpha  beta gamma  Possible spacegroups
        No PENALTY SDCELL FRACN LATT      a        b        c     alpha  beta gamma  Possible spacegroups
        
        if raw_output is set: return a list of lines
        otherwise, return a table object
        """
        mark1 = 'No PENALTY'
        mark2 = mark1
        block = self.parser.get_blocks_between(mark1,
                                               mark2,
                                               iMatch=1,
                                               skip1=0,
                                               skip2=1)
        if block:
            block = block[0]
        else:
            logger.error("cannot find pattern {} {} in {}".format(mark1, mark2, self.filename))
            return None

        if raw_output:
            self.index_table = "".join(block)
            return block

        # lines contains a title line stored in block[0]
        iborder = block[0].find('gamma') + 6
        candidates = []
        for line in block[1:]:
            if not (line.startswith('        ') or len(line) <= 10 or 'unrefined' in line):
                truncated_line = line[:iborder]
                fields = truncated_line.split()
                if len(fields) == 11:  # autoindex refine is on
                    # if float(fields[2]) <= sd_cutoff:
                    if "*" in fields[2] or "*" in fields[1]:
                        continue

                    if float(fields[2]) <= sd_cutoff or float(fields[1]) < score_cutoff:
                            candidates.append(truncated_line.split())
                elif len(fields) == 9:   # autoindex refine is off
                    if float(fields[1]) <= penalty_cutoff:
                        candidates.append(truncated_line.split())

        if not candidates:
            logger.error('No reliable autoindex solution found, aborted.')
            raise SystemExit(1)

        if remove_duplication:
            # filter candidates to remove redundant solution
            # sort potential solutions by lattice order and then by penalty, similar to that of xds
            # then remove solutions with higher penalty and similar cells
            lattice_tbl = Spacegroup.LATTICE
            idx_tbl = sorted(candidates, key=lambda row: (-lattice_tbl[row[-7]][0], float(row[1])))

            new_data = [idx_tbl[0]]
            for i in range(1, len(idx_tbl)):
                if all(not (idx_tbl[i][-7] == e[-7] and Spacegroup.similar_cell(idx_tbl[i][-6:], e[-6:])) for e in new_data):
                    new_data.append(idx_tbl[i])

            candidates = new_data

        table = Table()
        table.set_title('autoindex results by MOSFLM (sdxy < 0.2 or penalty < 100)')
        table.set_headers(block[0][:iborder].split())
        for row in candidates:
            table.add_data_by_a_list(row, hasRowname=False)

        self.table_autoindex = table
        table.print()
        return table

    def get_possible_space_groups(self, sgtype='spacegroup'):
        if not self.table_autoindex:
            self.get_autoindex_table(remove_duplication=True)

        possible_space_groups = []
        for line in self.table_autoindex.get_data():
            lattice = line[-7]
            if sgtype.startswith('space'):
                pgs = Spacegroup.get_space_groups_from_lattice(lattice, sgtype=sgtype, output_mode='symbol')
            else:
                pgs = Spacegroup.get_space_groups_from_lattice(lattice, sgtype='pointgroup', output_mode='symbol')

            for spg in pgs:
                possible_space_groups.append(' '.join([spg, 'lattice:', lattice, 'penalty:', line[1], 'cell:', ','.join(line[-6:])]))
        return ';'.join(possible_space_groups)

    def get_raw_index_table(self):
        if not self.index_table:
            self.get_autoindex_table(raw_output=True)
        return self.index_table

    def get_penalty(self, solution_number):
        # given the solution number, found penalty
        if self.table_autoindex:
            for row in self.table_autoindex.get_data():
                if int(solution_number) == int(row[0]):
                    return row[1]
        return None

    def get_mosflm_choice(self):
        """ generate information about the suggested solution by mosflm, return a named tuple"""
        line = self.parser.get_lines_contain('Suggested Solution:')
        slnNum = spg = None
        if line:
            line = line[0]
            logger.info('mosflm selected symmetry')
            slnNum, spg = line[line.find(':') + 1:].split()[:2]
        else:
            line = self.parser.get_lines_contain('Refining solution ')
            if line:
                logger.info('using user provided symmetry')
                rx = re.compile(r'''Refining solution #(.+) with (.+) \(number (.+)\) symmetry imposed''')
                slnNum, spg, spgnum = rx.findall(line[0])[0]
                slnNum = slnNum.replace(' ', '')

        penalty = self.get_penalty(slnNum)

        mark1 = 'Refining solution '
        mark2 = 'Final cell (after refinement) is'
        block = self.parser.get_blocks_between(text1=mark1, text2=mark2, iMatch=1)
        if block:
            block = block[0]
        else:
            logger.error("cannot find pattern \"{}\" \"{}\" in {}".format(mark1, mark2, self.filename))
            return None

        refined_cell = sdxy = None
        nref = 0
        nref_total = 1
        for line in block:
            if 'Final cell (after refinement) is' in line:
                refined_cell = " ".join([str(round(float(num), 1)) for num in getNumbers(line)])
            elif 'final sd in spot positions is' in line:
                sdxy, sdphi = getNumbers(line)[:2]
            elif line.startswith('Using'):
                nref, nref_total = getNumbers(line)[:2]
            elif line.startswith('This is a shift of'):
                beam_shift = getNumbers(line)[0]

        if sdxy and float(sdxy) >= 0.2:
            logger.warning("Positional deviation is greater than 0.2, chosen solution may not be correct.")

        percent = round(float(nref) / float(nref_total) * 100, 2)
        print("Chosen solution No: {} with spg {}, sd {}, penalty {},\n"
              "cell {},\n"
              "{} ({} %) used in refinement.".format(slnNum,
                                                     spg,
                                                     sdxy,
                                                     penalty,
                                                     refined_cell,
                                                     nref,
                                                     percent))
        solution = namedtuple("MosflmSolution", "SolutionNo, Penalty, sdxy, Spacegroup, Refined_cell")
        return solution(SolutionNo=slnNum,
                        Spacegroup=spg,
                        Refined_cell=refined_cell,
                        sdxy=sdxy,
                        Penalty=penalty)

    def get_strategy(self):
        # assuming single run, but need to deal with multiple run situation
        lines = self.parser.get_lines_after('Optimum rotation gives',
                                            includeCurrentLine=True,
                                            nLines=35,
                                            iMatch=1)
        if not lines:
            logger.error("Failed to extract strategy results from {}.".format(self.filename))
            return

        anomalous = True
        nrun = 0
        anomalous_completeness = native_completeness = start = end = multiplicity = None

        if 'anomalous' in lines[0]:
            anomalous = True
            anomalous_completeness = getNumbers(lines[0])[0]
        elif 'unique data' in lines[0]:
            anomalous = False
            anomalous_completeness = "NA"
            native_completeness = getNumbers(lines[0])[0]

        for line in lines[1:]:
            if 'This is' in line:
                native_completeness = getNumbers(line)[0]
            elif 'Run number' in line:
                nrun += 1
            elif 'Mean Multiplicity' in line:
                multiplicity = getNumbers(line)[0]
            elif 'From ' in line and line.strip('\n').endswith('degrees'):
                start, end = getNumbers(line)[:2]

        # figure out mosaic
        mosaic = self.get_mosaic()
        if not mosaic:
            logger.warning("mosaic estimate not found in mosflm output.")

        # find matrix used
        matrix = None
        line = self.parser.get_lines_contain("===> matrix ")
        if line:
            matrix = line[0].split()[-1]
        else:
            line = self.parser.get_lines_contain("===> newmat ")
            if line:
                matrix = line[0].split()[-1]

        return dict(anomalous=anomalous,
                    startAngle=start,
                    endAngle=end,
                    anomalousCompletenes=anomalous_completeness,
                    nRun=nrun,
                    mosaic=mosaic,
                    multiplicity=multiplicity,
                    matrix=matrix,
                    nativeCompleteness=native_completeness)

    def get_testgen(self):
        block = self.parser.get_blocks_between(text1='Phi start  Phi end   no of images  oscillation angle',
                                               text2='***** IMPORTANT *****',
                                               skip1=1,
                                               skip2=2,
                                               iMatch=1)
        if not block:
            logger.error("Failed to extract testgen results from {}.".format(self.filename))
            return None
        else:
            block = block[0]

        maxosc = 1.0
        for line in block:
            fields = line.split()
            if len(fields) == 6:
                col3 = fields[3]
                if float(col3) < maxosc:
                    maxosc = float(col3)

        if maxosc < 0.01:
            logger.warning("osc calculated as {} is too small, reset to 0.01.".format(maxosc))
            maxosc = 0.01

        return maxosc

    def get_mosaic(self):
        """ get results from mosaic estimate cmd or input as a parameter 
        The mosaicity has been estimated as ---> 0.20 <--- for this image only;
        """
        mosaic = None
        lines = self.parser.get_lines_contain('===> mosaic ')  # probably better off with regular expression here
        lines = [line for line in lines if '===> mosaic estimate' not in line]

        if not lines:
            lines = self.parser.get_lines_contain('The mosaicity has been estimated as --->')
        try:
            mosaic = float(getNumbers(lines[-1])[0])
        except Exception:
            logger.warning("failed to extract mosaic value")

        return mosaic


if __name__ == '__main__':
    # m1 = MosflmParser(wdir='/home/qxu/work/quickProcess/quickProcess/tests', filename='mosflm_autoindex.log')
    # m1.get_autoindex_table()
    # m2 = MosflmParser(wdir='/home/qxu/work/quickProcess/quickProcess/tests', filename='refined-autoindex.log')
    # m2.get_autoindex_table()
    #
    # print(m1.get_mosflm_choice())
    # print(m2.get_mosflm_choice())
    #
    # m3 = MosflmParser(wdir='/home/qxu/work/quickProcess/quickProcess/tests', filename='mosflm_autoindex.log.anomalous.withspg')
    # m3.get_autoindex_table()
    # print(m3.get_strategy())
    #
    # m3 = MosflmParser(wdir='/home/qxu/work/quickProcess/quickProcess/tests', filename='mosflm_autoindex.log.native.withspg')
    # m3.get_autoindex_table()
    # print(m3.get_strategy())

    m3 = MosflmParser(wdir='/tmp/qxu/j6/D1_2_0_scr_00061_00151/dials_strategy', filename='mosflm_strategy_sln1.log')
    m3.get_autoindex_table()
    print(m3.get_strategy())
    print(m3.get_mosaic())
    print(m3.get_testgen())

    #print(m.get_mosflm_choice())
