#!/usr/bin/env python
# -*- coding: utf-8 -*-
# NB some of the parser functions below are derived from xdsapp 2.0 xdsapp@helmholtz-berlin.de

import os, sys, re
from math import cos, radians, sqrt

from collections import defaultdict


def split_line(line, sep):
    cols = []
    for col in line.split(sep):
        if col.strip() != "":
            cols.append(col.strip())
    return cols


def get_integrate_stats(work_dir, filename="INTEGRATE.LP"):
    params = {}
    batch = 0

    '''    
    keylist = [
        'integrate_accepted', 'integrate_batch', 'integrate_frame_no',
        'integrate_novl', 'integrate_nrej', 'integrate_nstrong',
        'integrate_scale', 'integrate_sigmab', 'integrate_sigmar',
        'integrate_std_spindle', 'integrate_std_spotpos', 'integrate_uca',
        'integrate_ucb', 'integrate_ucc', 'distance', 'integrate_frame_range'
    ]
    '''

    if os.path.exists(os.path.join(work_dir, filename)) is True:
        file_open = open(os.path.join(work_dir, filename), 'r')
        line = file_open.readline()
        while line != "":
            cols = split_line(line, " ")
            if "!!! ERROR !!! CANNOT ALLOCATE MEMORY" in line:
                print("XDS processing failed")
                break

            if 'PROCESSING OF IMAGES' in line:
                cols = line.split()
                end = cols[-1]
                start = cols[-3]
                range = "{}-{}".format(start, end)
                params.setdefault('integrate_frame_range',[]).append(range)

            if "DEVIATION OF SPOT" in line:
                batch += 1
                params.setdefault('integrate_std_spotpos', []).append(
                    float(cols[6]))
                params.setdefault('integrate_batch', []).append(str(batch))
            if "DEVIATION OF SPINDLE" in line:
                params.setdefault('integrate_std_spindle', []).append(
                    float(cols[6]))
            if "UNIT CELL PARAMETERS" in line:
                params.setdefault('integrate_uca', []).append(float(cols[3]))
                params.setdefault('integrate_ucb', []).append(float(cols[4]))
                params.setdefault('integrate_ucc', []).append(float(cols[5]))
                params.setdefault('integrate_ucal', []).append(float(cols[6]))
                params.setdefault('integrate_ucbe', []).append(float(cols[7]))
                params.setdefault('integrate_ucga', []).append(float(cols[8]))
            if "CRYSTAL TO DETECTOR DISTANCE (mm)" in line:
                params.setdefault('distance', []).append(float(cols[-1]))
            if "REFLECTIONS ACCEPTED FOR REFINEMENT" in line:
                params.setdefault('integrate_accepted', []).append(
                    float(float(cols[0]) / float(cols[3]) * 100))
            if len(cols) == 10 and cols[1] == "0" and "-" not in line:
                params.setdefault('integrate_frame_no', []).append(cols[0])
                params.setdefault('integrate_scale', []).append(float(cols[2]))
                params.setdefault('integrate_novl', []).append(int(cols[4]))
                params.setdefault('integrate_nstrong', []).append(int(cols[6]))
                params.setdefault('integrate_nrej', []).append(int(cols[7]))
                params.setdefault('integrate_sigmab', []).append(
                    float(cols[8]))
                params.setdefault('integrate_sigmar', []).append(
                    float(cols[9]))
            line = file_open.readline()
        file_open.close()

    return params


def get_correct_stats(work_dir, filename="CORRECT.LP"):
    """
    keylist = [
        'alpha', 'anom_corr', 'anom_corr_table', 'anom_corr_totals', 'avalue',
        'b_wilson', 'beta', 'bvalue', 'cc_half', 'cc_half_table', 'chi_2',
        'chi_2_table', 'chi_corr', 'comp', 'comp_hi', 'comp_table',
        'comp_total', 'comp_totals', 'gamma', 'i_sig', 'i_sig_hi',
        'i_sig_table', 'i_sig_total', 'i_sig_totals', 'isa', 'mosaicity',
        'observed', 'observed_total', 'observed_totals', 'r_meas', 'r_meas_hi',
        'r_meas_table', 'r_meas_total', 'r_meas_totals', 'res', 'res_end',
        'res_hi', 'res_low', 'res_table', 'reslist_chi_2',
        'reslist_chi_2_table', 'reslist_wilson', 'reso_shell', 'sg', 'sg_no',
        'sig_ano', 'sig_ano_table', 'sig_ano_totals', 'uca', 'ucb', 'ucc',
        'uniques', 'uniques_total', 'uniques_totals', 'wilson_int', 'wilson_log_int',
    ]
    """

    params = {}

    if os.path.exists(os.path.join(work_dir, filename)) is True:
        cell_read = False
        file_open = open(os.path.join(work_dir, filename), 'r')
        line = file_open.readline()
        params['chi_corr'] = {}
        chi_i = 0
        while line != "":
            cols = split_line(line, " ")
            if "SPACE_GROUP_NUMBER=" in line:
                params['sg_no'] = cols[1]
            line = file_open.readline()
        file_open.close()
        file_open = open(os.path.join(work_dir, "CORRECT.LP"), 'r')
        line = file_open.readline()
        while line != "":
            cols = split_line(line, " ")
            if "CHI^2-VALUE" in line:
                params['chi_corr'][chi_i] = float(cols[6])
                chi_i += 1
            if len(cols) == 14 and cols[0] != "RESOLUTION" and cols[0] == "total":
                cols2 = split_line(cols[9], "%")
                params.setdefault('r_meas_totals', []).append(cols2[0])
                cols2 = split_line(cols[5], "%")
                params.setdefault('r_factor_totals', []).append(cols2[0])
                params.setdefault('i_sig_totals', []).append(cols[8])
                cols2 = split_line(cols[4], "%")
                params.setdefault('comp_totals', []).append(cols2[0])
                params.setdefault('uniques_totals', []).append(cols[2])
                params.setdefault('observed_totals', []).append(cols[1])
                params.setdefault('sig_ano_totals', []).append(cols[12])
                params.setdefault('anom_corr_totals', []).append(cols[11])
            if len(cols) == 14 and cols[0] != "RESOLUTION" and cols[0] != "total":
                params.setdefault('i_sig', []).append(cols[8])
                params.setdefault('uniques', []).append(cols[2])
                params.setdefault('observed', []).append(cols[1])
                params.setdefault('reso_shell', []).append(cols[0])
                params.setdefault('sig_ano', []).append(cols[12])
                params.setdefault('cc_half', []).append(cols[10])
                cols2 = split_line(cols[5], "%")
                params.setdefault('r_factor', []).append(cols2[0])
                cols2 = split_line(cols[9], "%")
                params.setdefault('r_meas', []).append(cols2[0])
                cols2 = split_line(cols[11], "%")
                params.setdefault('anom_corr', []).append(cols2[0])
                cols3 = split_line(cols[4], "%")
                params.setdefault('comp', []).append(cols3[0])
            pattern = r'[a-zA-Z]'
            if len(cols) == 9 and not re.search(pattern, line):
                params.setdefault('reslist_chi_2', []).append(cols[0])
                params.setdefault('chi_2', []).append(cols[3])
            if len(cols) == 6 and ("E+" in line or "E-0" in line):
                params.setdefault('reslist_wilson', []).append(cols[1])
                params.setdefault('wilson_int', []).append(cols[3])
                params.setdefault('wilson_log_int', []).append(cols[4])
            if "WILSON LINE" in line:
                pattern = r'[a-zA-Z]'
                if not re.search(pattern, cols[9]):
                    params['b_wilson'] = round(float(cols[9]), 2)
            if "sg_no" in params:
                if "[" + params['sg_no'] + "," in line and "]" in line:
                    cols = split_line(line, params['sg_no'] + ",")
                    cols = split_line(cols[1], "]")
                    params['sg'] = cols[0]
                    print("space group according to CORRECT: " + str(params['sg']))
            if "UNIT_CELL_CONSTANTS=" in line:
                params['uca'] = cols[1]
                params['ucb'] = cols[2]
                params['ucc'] = cols[3]
                params['alpha'] = round(float(cols[4]), 2)
                params['beta'] = round(float(cols[5]), 2)
                params['gamma'] = round(float(cols[6]), 2)
                cell_read = True
            if "CRYSTAL MOSAICITY" in line:
                params['mosaicity'] = cols[3]
            if ("E+" in line or "E-" in line) and len(cols) == 3:
                params['avalue'] = cols[0]
                params['bvalue'] = cols[1]
                params['isa'] = cols[2]
            if len(cols) == 9:
                pattern = r'[a-zA-Z]'
                if not re.search(pattern, line):
                    params['res_low'] = round(float(cols[0]), 2)
            line = file_open.readline()
        try:
            params['r_meas_total'] = params['r_meas_totals'][len(params['r_meas_totals']) - 1]
        except KeyError:
            pass

        try:
            params['r_factor_total'] = params['r_factor_totals'][len(params['r_factor_totals']) - 1]
        except KeyError:
            pass

        try:
            params['i_sig_total'] = params['i_sig_totals'][len(params['i_sig_totals']) - 1]
        except KeyError:
            pass
        try:
            params['comp_total'] = params['comp_totals'][len(params['comp_totals']) - 1]
        except KeyError:
            pass
        try:
            params['uniques_total'] = params['uniques_totals'][len(params['uniques_totals']) - 1]
        except KeyError:
            pass
        try:
            params['observed_total'] = params['observed_totals'][len(params['observed_totals']) - 1]
        except KeyError:
            pass
        try:
            params['res_hi'] = params['reso_shell'][len(params['reso_shell']) - 2] + "-" + params['reso_shell'][
                len(params['reso_shell']) - 1]
            params['res'] = float(params['reso_shell'][len(params['reso_shell']) - 1])
            params['res_table'] = params['reso_shell'][len(params['reso_shell']) - 9:len(params['reso_shell'])]
        except KeyError:
            pass

        try:
            params['r_meas_hi'] = params['r_meas'][len(params['r_meas']) - 1]
            params['r_meas_table'] = params['r_meas'][len(params['r_meas']) - 9:len(params['r_meas'])]
        except KeyError:
            pass

        try:
            params['r_factor_hi'] = params['r_factor'][len(params['r_factor']) - 1]
            params['r_factor_table'] = params['r_factor'][len(params['r_factor']) - 9:len(params['r_factor'])]
        except KeyError:
            pass

        try:
            params['i_sig_hi'] = params['i_sig'][len(params['i_sig']) - 1]
            params['i_sig_table'] = params['i_sig'][len(params['i_sig']) -  9:len(params['i_sig'])]
        except KeyError:
            pass
        try:
            params['comp_hi'] = params['comp'][len(params['comp']) - 1]
            params['comp_table'] = params['comp'][len(params['comp']) - 9:len(params['comp'])]
        except KeyError:
            pass
        try:
            params['sig_ano_table'] = params['sig_ano'][len(params['sig_ano']) - 9:len(params['sig_ano'])]
        except KeyError:
            pass
        try:
            params['anom_corr_table'] = params['anom_corr'][len(params['anom_corr']) - 9:len(params['anom_corr'])]
        except KeyError:
            pass
        try:
            params['cc_half_table'] = params['cc_half'][len(params['cc_half']) - 9:len(params['cc_half'])]
        except KeyError:
            pass
        try:
            params['reslist_chi_2_table'] = params['reslist_chi_2'][0:len(params['reslist_chi_2']) - 1]
        except KeyError:
            pass
        try:
            params['chi_2_table'] = params['chi_2'][0:len(params['chi_2']) - 1]
        except KeyError:
            pass

        file_open.close()
        if not cell_read:
            print("error - CORRECT didn't run through")
            params['fatal'] = True
    else:
        # CORRECT.LP does not exist
        print("error - CORRECT.LP does not exist")
        params['uca'] = "sorry, processing failed"

    return params


def get_xdsstat_stats(work_dir, filename="XDSSTAT.LP"):
    """
    keylist = [
        'r_d', 'r_meas_per_frame', 'xdsstat_frame_diff', 'xdsstat_frame_no', 'n_misfits', 'n_refs', 'cc', 'i_over_sigma_per_frame'
    ]
    """
    params = {}

    if os.path.exists(os.path.join(work_dir, filename)) is True:
        file_open = open(os.path.join(work_dir, filename), 'r')
        line = file_open.readline()
        while line != "":
            cols = split_line(line, " ")
            if len(cols) == 12 and " L" in line:
                params.setdefault('n_refs', []).append(float(cols[1]))
                params.setdefault('n_misfits', []).append(float(cols[2]))
                params.setdefault('i_over_sigma_per_frame', []).append(float(cols[5]))
                params.setdefault('cc_per_frame', []).append(float(cols[7]))
                params.setdefault('r_meas_per_frame', []).append(float(cols[8]) * 100)
                params.setdefault('xdsstat_frame_no', []).append(int(cols[0]))
            if "DIFFERENCE" in line:
                params.setdefault('xdsstat_frame_diff', []).append(int(cols[0]))
                params.setdefault('r_d', []).append(float(cols[2]) * 100)
                params.setdefault('r_d_friedel', []).append(float(cols[4]) * 100)
                params.setdefault('r_d_not_friedel', []).append(float(cols[6]) * 100)

            line = file_open.readline()
        file_open.close()

    return params


def get_maxcc12_stat(work_dir, filename="MAXCC12.LP"):
    """
    keylist = [
        'n_ano', 'n_iso', 'cc_half_ano', 'cc_half_iso', 'completeness', 'maxcc_frame_no', 'reso_shells'
    ]
    """
    params = defaultdict(list)

    with open(os.path.join(work_dir, filename)) as f:
        lines = f.read().splitlines()

    for i, line in enumerate(lines):
        if line.startswith(" resolution shells:"):
            cols = lines[i + 1].split()
            lower = ['inf'] + cols[:-1]
            params['reso_shells'] = ["{}-{}A".format(l, u) for l, u in zip(lower, cols)]
        if line.endswith('a'):
            linex = line.replace("-100.0", " -99.9")
            cols = linex.split()
            params['maxcc_frame_no'].append(cols[0])
            params['cc_half_iso'].append(cols[1:-1])
        elif line.endswith('b'):
            cols = line.split()
            params['n_iso'].append(cols[:-1])
        elif line.endswith('c'):
            cols = line.split()
            params['completeness'].append(cols[:-1])
        elif line.endswith('d'):
            linex = line.replace("-100.0", " -99.9")
            cols = linex.split()
            params['cc_half_ano'].append(cols[:-1])
        elif line.endswith('e'):
            cols = line.split()
            params['n_ano'].append(cols[:-1])

    return params


def get_xdscc12_stat(work_dir, filename="XDSCC12.LP"):
    result = defaultdict(list)
    with open(os.path.join(work_dir, filename)) as fh:
        lines = fh.read().splitlines()

    ibatch = 0
    nframes = 1
    osc = 0.2
    for i, line in enumerate(lines):
        if "resolution shells (for lines starting with" in line:
            cols = lines[i + 1].split()
            lower = ['inf'] + cols[:-1]
            result['xdscc12_reso_shells'] = ["{}-{}A".format(l, u) for l, u in zip(lower, cols)]

        if '!OSCILLATION_RANGE= ' in line:
            osc = float(line.split()[-1])
            nframes = int(lines[i+1].split()[0])

        if 'frames batched such that total rotation >=' in lines:
            cols = line.split(">=")
            result['xdscc12_ndegree'] = float(cols[-1])

        if line.startswith("a "):
            cols = line.split()
            result['xdscc12_deltaCC_iso'].append(float(cols[5]))
            ibatch += 1
            x = ibatch*nframes
            frame_range = "{}-{}".format(x-nframes+1, x)
            result['xdscc12_frame_range'].append(frame_range)

        if line.startswith("b "):
            cols = line.split()
            result['xdscc12_deltaCC_iso_shells'].append(cols[1:])

        if line.startswith('d '):
            cols = line.split()
            result['xdscc12_deltaCC_ano'].append(float(cols[5]))

        if line.startswith("e "):
            cols = line.split()
            result['xdscc12_deltaCC_ano_shells'].append(cols[1:])

    return result


class Report(object):
    def __init__(self, work_dir,
                 pipelinestatus_id=None,
                 tag_name=None,
                 integrate_filename="INTEGRATE.LP",
                 correct_filename="CORRECT.LP",
                 xdsstat_filename="XDSSTAT.LP",
                 xdscc12_filename="XDSCC12.LP",
                 maxcc12_filename="MAXCC12.LP"
                 ):
        self.work_dir = work_dir
        self.integrate_lp = integrate_filename
        self.correct_lp = correct_filename
        self.xdsstat_lp = xdsstat_filename
        self.maxcc12_lp = maxcc12_filename
        self.xdscc12_lp = xdscc12_filename
        self.pipelinestatus_id = pipelinestatus_id
        self.tag_name = tag_name

        self.integ_dic = None
        self.correct_dic = None
        self.xdsstat_dic = None
        self.maxcc12_dic = None
        self.xdscc12_dic = None

        self.stats = self.__get_stats()

    def __get_stats(self):
        params = {}
        if os.path.exists(os.path.join(self.work_dir, self.integrate_lp)):
            self.integ_dic = get_integrate_stats(self.work_dir, filename=self.integrate_lp)
            params = self.integ_dic.copy()

        if os.path.exists(os.path.join(self.work_dir, self.correct_lp)):
            self.correct_dic = get_correct_stats(self.work_dir, filename=self.correct_lp)
            params.update(self.correct_dic)

        if os.path.exists(os.path.join(self.work_dir, self.xdsstat_lp)):
            self.xdsstat_dic = get_xdsstat_stats(self.work_dir, filename=self.xdsstat_lp)
            params.update(self.xdsstat_dic)

        if os.path.exists(os.path.join(self.work_dir, self.maxcc12_lp)):
            self.maxcc12_dic = get_maxcc12_stat(self.work_dir, filename=self.maxcc12_lp)
            params.update(self.maxcc12_dic)

        if os.path.exists(os.path.join(self.work_dir, self.xdscc12_lp)):
            self.xdscc12_dic = get_xdscc12_stat(self.work_dir, filename=self.xdscc12_lp)
            params.update(self.xdscc12_dic)

        return params

    def get_correct_dic(self):
        return self.correct_dic

    def get_all_stats(self):
        return self.stats

    # per frame
    def get_frame_nos(self):
        return self.stats.get('integrate_frame_no')

    def get_scale_factor(self):
        return self.stats.get('integrate_scale')

    def get_mosaicity(self):
        return self.stats.get('integrate_sigmar')

    def get_beam_divergence(self):
        return self.stats.get('integrate_sigmab')

    def get_n_strong_spots(self):
        return self.stats.get('integrate_nstrong')

    def get_n_rejected_spots(self):
        return self.stats.get('integrate_nrej')

    # from xdsstat
    def get_rmeas(self):
        return self.stats.get('r_meas_per_frame')

    def get_rd(self):
        return self.stats.get('r_d')

    def get_rd_friedel(self):
        return self.stats.get('r_d_friedel')

    def get_rd_not_friedel(self):
        return self.stats.get('r_d_not_friedel')

    def get_cc(self):
        return self.stats.get('cc_per_frame')

    def get_i_over_sigma(self):
        return self.stats.get('i_over_sigma_per_frame')

    def get_n_refs(self):
        return self.stats.get('n_refs')

    def get_n_misfits(self):
        return self.stats.get('n_misfits')

    # per batch
    def get_integ_batch_nos(self):
        return self.stats.get('integrate_batch')

    def get_integ_frame_range(self):
        return self.stats.get('integrate_frame_range')

    def get_spot_position_std(self):
        return self.stats.get('integrate_std_spotpos')

    def get_no_reflection_accepted(self):
        return self.stats.get('integrate_accepted')

    def get_spindle_position_std(self):
        return self.stats.get('integrate_std_spindle')

    @staticmethod
    def cal_volume(a, b, c, al, be, ga):
        '''V = abc (1- cos2 α - cos2 β - cos2 γ + 2(cos(α) cos(β) cos(γ)))½'''

        cosa, cosb, cosc = [cos(radians(x)) for x in [al, be, ga]]
        return round(a * b * c * sqrt(1 - cosa ** 2 - cosb ** 2 - cosc ** 2 + 2.0 * cosa * cosb * cosc), 1)

    def get_cell_a(self):
        return self.stats.get("integrate_uca")

    def get_cell_b(self):
        return self.stats.get("integrate_ucb")

    def get_cell_c(self):
        return self.stats.get("integrate_ucc")

    def get_cell_al(self):
        return self.stats.get("integrate_ucal")

    def get_cell_be(self):
        return self.stats.get("integrate_ucbe")

    def get_cell_ga(self):
        return self.stats.get("integrate_ucga")

    '''
    def get_volume(self):
        return [Report.cal_volume(a, b, c, al, be, ga)
                for a, b, c, al, be, ga in
                zip(self.get_cell_a(),
                    self.get_cell_b(),
                    self.get_cell_c(),
                    self.get_cell_al(),
                    self.get_cell_be(),
                    self.get_cell_ga())]
    '''

    def get_volume(self, percent=True):
        arr_a = self.get_cell_a()
        arr_b = self.get_cell_b()
        arr_c = self.get_cell_c()
        arr_al = self.get_cell_al()
        arr_be = self.get_cell_be()
        arr_ga = self.get_cell_ga()
        if not arr_a:
            print("failed to parse INTEGRATE.LP")
            return None

        if percent:
            v0 = Report.cal_volume(arr_a[0], arr_b[0], arr_c[0], arr_al[0], arr_be[0], arr_ga[0])
            return [100 * (Report.cal_volume(a, b, c, al, be, ga) - v0)/v0 for a, b, c, al, be, ga in zip(arr_a, arr_b, arr_c, arr_al, arr_be, arr_ga)]
        else:
            return [Report.cal_volume(a, b, c, al, be, ga) for a, b, c, al, be, ga in zip(arr_a, arr_b, arr_c, arr_al, arr_be, arr_ga)]

    def get_distance(self):
        return self.stats.get('distance')

    # javascript data block definition
    def get_plot_block(self,
                       x,
                       y,
                       name="parameter name",
                       mode="lines",
                       shape='linear',
                       visible="legendonly"):
        out_str = \
            '''{{
                x: {},
                y: {},
                mode: '{}',
                visible: '{}',
                name: '{}',
                line: {{shape: '{}'}},
            }}'''.format(x, y, mode, visible, name, shape)
        return out_str

    @staticmethod
    def create_div(data_blocks, div_id="div1",
                   title="Plots against frame numbers",
                   xlabel="Frame No.",
                   caption="caption",
                   hide=False):
        div_inner = div_id+"_plot"
        script = ""
        for i, d_block in enumerate(data_blocks, 1):
            script += "\nvar trace{} = {};".format(i, d_block)

        script += "\nvar data=[{}];".format(", ".join("trace{}".format(i + 1) for i in range(len(data_blocks))))

        layout = """\nlayout={{
        title: '{title}', 
        titlefont: {{
            size: 18,
            color: '#000080',
        }},
        xaxis: {{
            title: '{xlabel}',
            titlefont: {{
                size: 16
            }},
            tickfont: {{
                size: 14
            }},
        }},
        legend: {{
            orientation: 'v',
            bgcolor: '#E2E2E2',
            font: {{
                size: 14,
            }},
        }},
        autosize: false,
        width: 950,
        height: 500,
        }};
        """.format(title=title, xlabel=xlabel)
        script += layout

        script += "\nPlotly.newPlot('{}', data, {}, {{displayModeBar: false}});".format(div_inner, "layout")

        if hide:
            hideit = "style='display:none'"
        else:
            hideit = ""

        return """
        <div class='analysis' id='{div_id}'>
            <a class='anchor' id='anchor_{div_id}'></a> 

            <div id='{div_id_inner}' {hideit}></div>
            <script> 
            {script}
            </script>
            <p class='caption'>{caption}</p>
            <!--- <button onclick='hidePlot(this)' id='{btn_id}'>Hide {title}</button> --->
        </div>
        """.format(anchor="",
                   div_id=div_id,
                   div_id_inner=div_inner,
                   script=script,
                   caption=caption,
                   btn_id="btn_" + div_id,
                   title=title,
                   hideit=hideit)

    def plots_vs_frame_nos(self):
        fnos = self.get_frame_nos()

        if not fnos:
            print("failed to parse " + self.integrate_lp)
            return ""

        lines = list()
        lines.append(self.get_plot_block(x=fnos, y=self.get_mosaicity(), name="mosaicity (deg)", visible='true'))
        lines.append(self.get_plot_block(x=fnos, y=self.get_scale_factor(), name="scale factor", visible='true'))
        lines.append(self.get_plot_block(x=fnos, y=self.get_n_strong_spots(), name="no. strong spots"))
        lines.append(self.get_plot_block(x=fnos, y=self.get_n_rejected_spots(), name="no rejected spots"))
        lines.append(self.get_plot_block(x=fnos, y=self.get_beam_divergence(), name="beam divergence (deg)"))
        lines.append(self.get_plot_block(x=fnos, y=self.get_rmeas(), name="r_meas (%)"))
        lines.append(self.get_plot_block(x=fnos, y=self.get_cc(), name="Corr"))
        lines.append(self.get_plot_block(x=fnos, y=self.get_i_over_sigma(), name="Iobs/Sigma"))
        lines.append(self.get_plot_block(x=fnos, y=self.get_n_misfits(), name="no. misfit"))

        caption = "Changes in various parameters between frames during integration. " \
                  "Large derivations may suggest something went wrong during integration. " \
                  "<a href='{}' target='_blank'><span>view INTEGRATE.LP</span></a>".format(self.integrate_lp)

        div = Report.create_div(lines,
                                 div_id="div_frmnos",
                                 xlabel="Frame No.",
                                 caption=caption,
                                 title="Parameter changes during integration")

        return div

    def plots_vs_framediff_nos(self):
        frm_diff = self.stats.get('xdsstat_frame_diff')

        if not frm_diff:
            print("failed to parse " + self.xdsstat_lp)
            return ""

        pba = self.get_plot_block(x=frm_diff, y=self.get_rd(), name="R_decay (%)", visible='true')
        pbb = self.get_plot_block(x=frm_diff, y=self.get_rd_friedel(), name="R_decay Friedel (%)")
        pbc = self.get_plot_block(x=frm_diff, y=self.get_rd_not_friedel(), name="R_decay not Friedel(%)")

        caption = "R_decay plot. R-factors as a function of frame number difference (Kay Diederichs), " \
                  "an increasing R_decay as frame number difference increases may suggest radiadtion damage. " \
                  "<a href='{}' target='_blank'><span>view XDSTAT</span></a>".format(self.xdsstat_lp)

        div = Report.create_div([pba, pbb, pbc],
                                div_id="div_R_decay",
                                title="R_decay (Rd) plot",
                                xlabel="Difference in Frame No.",
                                caption=caption
                                )
        return div

    def plots_vs_batch_nos(self):
        # bnos = self.get_integ_batch_nos()
        bnos = self.get_integ_frame_range()

        if not bnos:
            print("failed to parse " + self.integrate_lp)
            return ""

        batch1 = self.get_plot_block(x=bnos,
                                     y=self.get_cell_a(),
                                     name="cell_a (A)")
        batch2 = self.get_plot_block(x=bnos,
                                     y=self.get_cell_b(),
                                     name="cell_b (A)")
        batch3 = self.get_plot_block(x=bnos,
                                     y=self.get_cell_c(),
                                     name="cell_c (A)")
        batch4 = self.get_plot_block(x=bnos,
                                     y=self.get_volume(percent=True),
                                     name="cell volume changes (%)",
                                     visible="true")
        batch5 = self.get_plot_block(x=bnos,
                                     y=self.get_spot_position_std(),
                                     name="std spot position (pixel)")
        batch6 = self.get_plot_block(x=bnos,
                                     y=self.get_no_reflection_accepted(),
                                     name="% spots accepted for refinement")
        batch7 = self.get_plot_block(x=bnos,
                                     y=self.get_spindle_position_std(),
                                     name="std spindle position (deg)")

        batch8 = self.get_plot_block(x=bnos,
                                     y=self.get_distance(),
                                     name="detector distance (mm)")

        caption = "Changes in parameters between integration segements. " \
                  "Large derivations may suggest something went wrong during integration. " \
                  "<a href='{}' target='_blank'><span>view INTEGRATE.LP</span></a>".format(self.integrate_lp)

        div = Report.create_div([batch1, batch2, batch3, batch4, batch5, batch6, batch7, batch8],
                                 div_id="div_btchnos",
                                 title="Parameter changes during integration",
                                 caption=caption,
                                 xlabel="Integration batch (frame range)")
        return div

    def plots_xdscc12_vs_frame_range(self):
        caption = ("<a href='https://strucbio.biologie.uni-konstanz.de/xdswiki/index.php/XDSCC12' target='_blank'>"
                   "<span>Analysis of delta CC1/2 by frame range and resolution</span></a>. "
                   "Negative delta CC1/2 values may indicate issues with the corresponding frame range. "
                   "<a href='{}' target='_blank'><span>view XDSCC12 results.</span></a>".format(self.xdscc12_lp)
                   )

        fnos = self.stats.get('xdscc12_frame_range')
        if not fnos:
            print("failed to parse " + self.xdscc12_lp)
            return ""

        reso_shells = self.stats.get('xdscc12_reso_shells')

        cc12_iso_overall = self.stats.get('xdscc12_deltaCC_iso')

        cc12_iso_shells = self.stats.get('xdscc12_deltaCC_iso_shells')

        blocks = list()

        blocks.append(self.get_plot_block(x=fnos, y=cc12_iso_overall, visible='true',name="Overall"))
        for i in range(len(reso_shells)):
            deltaCC12 = [x[i] for x in cc12_iso_shells]
            blocks.append(self.get_plot_block(x=fnos, y=deltaCC12, name=reso_shells[i], visible='legendonly'))

        div_xdscc12_iso = Report.create_div(blocks,
                                            div_id="div_xdscc12_iso",
                                            title="delta CC1/2 (iso) vs frame range and resolution",
                                            xlabel="Frame range",
                                            caption=caption)

        cc12_ano_overall = self.stats.get('xdscc12_deltaCC_ano')
        cc12_ano_shells = self.stats.get('xdscc12_deltaCC_ano_shells')

        if not cc12_ano_overall:
            print("no anomalous cc in " + self.xdscc12_lp)
            return div_xdscc12_iso

        blocks = []
        blocks.append(self.get_plot_block(x=fnos, y=cc12_ano_overall, visible='true',name="Overall"))
        for i in range(len(reso_shells)):
            deltaCC12 = [x[i] for x in cc12_ano_shells]
            blocks.append(self.get_plot_block(x=fnos, y=deltaCC12, name=reso_shells[i], visible='legendonly'))

        div_xdscc12_ano = Report.create_div(blocks,
                                            div_id="div_xdscc12_ano",
                                            title="delta CC1/2 (ano) vs frame range and resolution",
                                            xlabel="Frame range",
                                            caption=caption)

        return div_xdscc12_iso + div_xdscc12_ano

    def plots_maxcc_vs_frame_nos(self):
        caption = ("<a href='https://strucbio.biologie.uni-konstanz.de/xdswiki/index.php/Xds_maxcc12' target='_blank'>"
                   "<span>Analysis of CC1/2 by frame number and resolution</span></a>, "
                   "Radiation damage often cause drop in CC1/2 values, esp in high resolution shells. "
                   "<a href='{}' target='_blank'><span>view MAXCC12 results.</span></a>".format(self.maxcc12_lp)
                   )

        fnos = self.stats.get('maxcc_frame_no')
        if not fnos:
            print("failed to parse maxcc12 results, ignore maxcc12 plots.")
            return ""

        start = fnos[0]
        fnos = ["{}-{}".format(start, i) for i in fnos]

        reso_shells = self.stats.get('reso_shells')
        cc12_iso = self.stats.get('cc_half_iso')

        blocks = []
        for i in range(len(reso_shells)):
            maxcc = [x[i] for x in cc12_iso]
            blocks.append(self.get_plot_block(x=fnos, y=maxcc, name=reso_shells[i], visible='true'))

        div_maxcc12_iso = Report.create_div(blocks,
                                            div_id="div_maxcc12_iso",
                                            title="CC1/2 (iso) vs frame and resolution plot",
                                            xlabel="Frame range",
                                            caption=caption)

        blocks = []
        cc12_ano = self.stats.get('cc_half_ano')

        for i in range(len(reso_shells)):
            maxcc = [x[i] for x in cc12_ano]
            blocks.append(self.get_plot_block(x=fnos, y=maxcc, name=reso_shells[i], visible='true'))

        div_maxcc12_ano = Report.create_div(blocks,
                                            div_id="div_maxcc12_ano",
                                            title="CC1/2 (ano) vs frame and resolution plot",
                                            xlabel="Frame range",
                                            caption=caption
                                            )

        caption = ("Analysis of completeness by frame number and resolution</span></a>, "
                   "<a href='{}' target='_blank'><span>view MAXCC12 results.</span></a>".format(self.maxcc12_lp)
                   )

        blocks = []
        completeness = self.stats.get('completeness')
        for i in range(len(reso_shells)):
            compl = [x[i] for x in completeness]
            blocks.append(self.get_plot_block(x=fnos, y=compl, name=reso_shells[i], visible='true'))

        div_maxcc12_completeness = Report.create_div(blocks,
                                                     div_id="div_maxcc12_completeness",
                                                     title="Completeness vs frames and resolution",
                                                     caption=caption,
                                                     xlabel="Frame range")

        return div_maxcc12_iso+ div_maxcc12_ano + div_maxcc12_completeness

    def plots_vs_resolution(self):
        if not self.stats.get('res_table'):
            print("CORRECT likely failed, cannot make plots.")
            return ""

        resol = [x+"A" for x in self.stats.get('res_table') if x]

        lines = []

        cc12 = [float(x.replace('*', '')) for x in self.stats.get('cc_half_table')]
        lines.append(self.get_plot_block(x=resol, y=cc12, name="CC_1/2 (%)", visible="true"))

        anom_corr = [float(x.replace('*', '')) for x in self.stats.get('anom_corr_table')]
        lines.append(self.get_plot_block(x=resol, y=anom_corr, name="Anomalous Correlation (%)", visible="true"))

        rmeas = [float(x.replace('*', '')) for x in self.stats.get('r_meas_table')]
        lines.append(self.get_plot_block(x=resol, y=rmeas, name="R_meas (%)", visible="legendonly"))

        rfactor = [float(x.replace('*', '')) for x in self.stats.get('r_factor_table')]
        lines.append(self.get_plot_block(x=resol, y=rfactor, name="R_factor (%)", visible="legendonly"))

        isigmai = [float(x.replace('*', '')) for x in self.stats.get('i_sig_table')]
        lines.append(self.get_plot_block(x=resol, y=isigmai, name="<I/Sigma>", visible="true"))

        completeness = [float(x.replace('*', '')) for x in self.stats.get('comp_table')]
        lines.append(self.get_plot_block(x=resol, y=completeness, name="Completeness (%)", visible="true"))

        sigano = [float(x.replace('*', '')) for x in self.stats.get('sig_ano_table')]
        lines.append(self.get_plot_block(x=resol, y=sigano, name="SigAno", visible="legendonly"))

        summary = "Rsym={}, Rmeas={}, completeness={}, redudancy={}, ISa={}".format(
            self.stats.get('r_factor_total'),
            self.stats.get('r_meas_total'),
            self.stats.get('comp_total'),
            round(float(self.stats.get('observed_total')) / float(self.stats.get('uniques_total')),2),
            self.stats.get('isa'))

        caption = "This plot shows various merging statistics in resolution shells. " \
                  "Overall result: {}, " \
                  "<a href='{}' target='_blank'><span>view CORRECT.LP</span></a>".format(summary, self.correct_lp)

        div_merge = Report.create_div(lines,
                                      div_id="div_data_quality",
                                      title="Data merging statistics",
                                      caption=caption,
                                      xlabel="Resolution (A)")
        return div_merge

    def plots_wilson(self):
        resol = [str(round(float(x), 2))+"A" for x in self.stats.get('reslist_wilson')]
        if not resol:
            print("CORRECT likely failed, cannot make plots.")
            return ""

        m_intensity = self.stats.get('wilson_int')
        line1 = self.get_plot_block(x=resol, y=m_intensity, name="I_mean", visible="true")

        ln_intensity = self.stats.get('wilson_log_int')
        # print(ln_intensity)
        line2 = self.get_plot_block(x=resol, y=ln_intensity, name="log(I_mean)", visible="legendonly")

        caption = "Wilson plot, average intensity again resolution.  " \
                  "wilson B: {}, " \
                  "<a href='{}' target='_blank'><span>view CORRECT.LP</span></a>".format(self.stats.get("b_wilson"), self.correct_lp)

        div = Report.create_div([line1, line2],
                                div_id="div_wilson",
                                xlabel="Resolution (A)",
                                title="Wilson plot",
                                caption=caption)
        return div

    def to_html(self, body):
        reprocess_tag = ""
        if self.pipelinestatus_id:
            reprocess_tag = "<a href='process:{}'>Reprocess</a> or ".format(self.pipelinestatus_id)

        if self.tag_name:
            title = "Data Analysis Graphical Report: {}".format(self.tag_name)
            tag_name = self.tag_name
        else:
            title = "Data Analysis Graphical Report"
            tag_name = ""

        out_html = """
            <html>
            <head>
                <!--- not working after v1.42.5, ie. broke in 1.43.0  TODO check what went wrong
                <script src="https://cdn.plot.ly/plotly-latest.min.js"></script> 
                --->
 
                <script src="https://cdn.plot.ly/plotly-1.42.5.min.js"></script>
                <script>
                    function hidePlot(button) {{
                        var id = button.id;
                        var div_id = id.replace("btn_","");
                        var x = document.getElementById(div_id);
                        if (x.style.display == "none") {{
                            x.style.display = "block";
                        }} else {{
                            x.style.display = "none";
                        }}
                        
                    }};
                </script>
                <style>
                    .titleTOC {{
                        top: 0;
                        z-index: 1;
                        position: fixed;
                        display: inline;
                    }}

                    .toc {{
                        position: relative;
                        z-index: 1;
                        float: left;
                        width: 100%;
                        max-width: 700px;
                        margin: 6 auto;
                        margin-top: 0px;
                        margin-left: 10px;
                        padding: 2px;
                        background: rgba(230, 230, 230, 0.8);
                    }}
                    
                    .caption {{
                        position: relative;
                        text-align: left;
                        margin-left: 20px;
                        margin-right: 20px;
                        padding: 4px;
                        background: rgba(230, 230, 255, 0.7);
                    }}
                    
                    a.anchor {{
                        position: relative;
                        top: -280px;
                        display: inline-block;
                    }}
                    
                    .plots {{
                        position: relative;
                        top: 850px;
                        width: 95%;
                    }}
           
                    .analysis {{
                        position: relative;
                        float: left;
                        text-align: left;
                        width: 100%;
                        max-width: 1200px;
                        border: 1px solid gray;
                        margin: 6 auto;
                        margin-top: 20px;
                        position: relative;
                        padding: 4px;
                        box-shadow: 4px 4px 2px grey;
                        z-index: -1;
                    }}


                    li {{
                        list-style-type: square;
                    }}
                    
                    li a:hover {{
                        background-color: #FFFF00;
                    }}
                    
                    button {{
                        width: auto;
                    }}
                    
                    summary {{
                        background: rgba(230, 230, 255, 0.7);
                        font-size: 150%;
                        font-weight: bold;
                    }}
                   
                </style>
                <title>Graphical data analysis report</title>
            </head>

            <body>
                <div id='titleTOC' class='titleTOC'>
                    
                    <!-- <h2>{title}</h2> -->
                    
                    <details open>
                        <summary>Table of Contents: {tag_name}</summary>
                        
                        <div class='toc' id='toc'>
                            <ul>
                                <li><a href="#anchor_div_data_quality">Overall Data Quality</a></li>
                                <li><a href="#anchor_div_R_decay">R_decay Plot</a></li>
                                <li><a href="#anchor_div_frmnos">Per-frame parameter changes during integration</a></li>
                                <li><a href="#anchor_div_btchnos">Per-batch parameter changes during integration</a></li>
                                <li><a href="#anchor_div_xdscc12_iso">delta CC1/2-iso vs frame and resolution</a></li>
                                <li><a href="#anchor_div_xdscc12_ano">delta CC1/2-ano vs frame and resolution</a></li>
                                <li><a href="#anchor_div_maxcc12_iso">CC1/2-iso vs frame (accumulative) and resolution</a></li>
                                <li><a href="#anchor_div_maxcc12_ano">CC1/2-ano vs frame (accumulative) and resolution</a></li>
                                <li><a href="#anchor_div_maxcc12_completeness">Completeness vs frame and resolution</a></li>
                                <li><a href="#anchor_div_wilson">Wilson Plot</a></li>
                                <li>{reprocess_tag}<a href="XDS.INP">Open XDS.INP (and start XDSGUI)</a></li>
                            </ul>
                        </div>
                    </details>
                </div>
                
                <div id='plots' class='plots'> 
                    {body}
                </div>
            </body>
            </html>""".format(title=title, tag_name=tag_name, reprocess_tag=reprocess_tag, body=body)
        return out_html

    def create_report(self, filename="report.html"):
        div0 = self.plots_vs_resolution()
        div1 = self.plots_vs_framediff_nos()
        div2 = self.plots_vs_frame_nos()
        div3 = self.plots_vs_batch_nos()
        div4 = self.plots_xdscc12_vs_frame_range()
        div5 = self.plots_maxcc_vs_frame_nos()
        div6 = self.plots_wilson()

        ohtml = os.path.join(self.work_dir, filename)
        with open(ohtml, "w") as f:
            f.write(self.to_html(div0 + div1 + div2 + div3 + div4 + div5 + div6))


if __name__ == '__main__':
    if sys.argv[-1] == '-h':
        print("generate graphical html report for a completed XDS job, run it in a XDS directory.")
        sys.exit(0)

    wdir = os.getcwd()
    wdir = "/mnt/beegfs/qxu/23BM_2017_08_08/GS13649A_4l8n/process/gmcaproc/135699_1_00001_00480/XDS-p1"
    # native dataset
    wdir = '/mnt/beegfs/qxu/23BM_2017_08_08/GS13649A_4l8n/process/gmcaproc/135699_2_00001_00360/XDS-p1'
    #wdir = '/mnt/beegfs/qxu/23BM_2017_08_08/14site-inv/collect/vector/process/analysisNG/gmcaproc/C2_TaBr6_4_S4_00109_00150/XDS-p1'
    #wdir = '/mnt/beegfs/qxu/jsmith-test/test1/cbf/process/analysisNG/gmcaproc/AEF_Se_MalA_1_2_1_INV_00001_01025/XDS-p1'
    #wdir = '/mnt/beegfs/qxu/jsmith-test/test2/cbf/process/analysisNG/gmcaproc/TropB_SeMet_B7_1_00001_01800/XDS-p1'
    #wdir = '/mnt/beegfs/qxu/test-data/PhqK_reduced_1/process/analysisNG/gmcaproc/PhqK_reduced_1_1_00001_01200/XDS'
    #wdir = '/mnt/beegfs/qxu/23BM_2017_08_08/14site-inv/collect/vector/process/gmcaproc/C2_TaBr6_4_INV_S4_00109_00150/XDS'

    #print(sys.argv)

    r = Report(wdir, xdsstat_filename="XDSSTAT.LP", maxcc12_filename="MAXCC12.LP", xdscc12_filename="XDSCC12.LP")
    r.create_report(filename="XDSsummary.html")

    exit(0)

    # anomalous dataset
    wdir = "/mnt/beegfs/qxu/23BM_2017_08_08/GS13649A_4l8n/process/gmcaproc/135699_1_00001_00480/XDS-p1"
    # native dataset
    wdir = '/mnt/beegfs/qxu/23BM_2017_08_08/4z2x-p21212/process/analysisNG/gmcaproc/196802_1_00001_00420/XDS-p1'
    wdir = '/mnt/beegfs/qxu/jsmith-test/test1/cbf/process/analysisNG/gmcaproc/AEF_Se_MalA_1_2_1_INV_00001_01025/XDS-p1'
    r = Report(wdir, xdsstat_filename="XDSSTAT.LP", maxcc12_filename="maxcc12.log")
    r.create_report()

    exit(0)

    print(get_maxcc12_stat(
        "/mnt/beegfs/qxu/23BM_2017_08_08/GS13649A_4l8n/process/gmcaproc/135699_1_00001_00480/XDS-p1"
    ))

    exit(0)

    print(get_integrate_stats(
        "/mnt/beegfs/qxu/23BM_2017_08_08/GS13649A_4l8n/process/gmcaproc/135699_1_00001_00480/XDS-p1"
    ))
    print(get_correct_stats(
        "/mnt/beegfs/qxu/23BM_2017_08_08/GS13649A_4l8n/process/gmcaproc/135699_1_00001_00480/XDS-p1"
    ))


    print(get_xdscc12_stat(
        "/mnt/beegfs/qxu/23BM_2017_08_08/GS13649A_4l8n/process/gmcaproc/135699_1_00001_00480/XDS-p1"
    ))
