from __future__ import absolute_import

import shutil, os
import logging, json
from collections import Counter
from sortedcollections import SortedDict

from .fileopener import FileOpener

logger = logging.getLogger(__name__)


def round1(num):
    return str(round(float(num), 1))


def round_nm2ang(index, num):
    if index < 3:
        return str(round(float(num) * 10, 1))
    else:
        return str(round(float(num), 1))


class StreamParser(FileOpener):
    def __init__(self, wdir=".",
                 filename='indexamajig.stream'):
        super(StreamParser, self).__init__(wdir, filename)
        self.all_results = list()
        self.chunks = self.get_chunks()
        self.indexed_chunks = []

        self.parse_chunks()

    def guess_lattice(self, n=1):
        lines = self.parser.get_lines_contain('lattice_type =')
        if not lines:
            return

        lattices = [l.split()[-1] for l in lines]

        return Counter(lattices).most_common(n)

    def guess_centering(self, n=1):
        lines = self.parser.get_lines_contain('centering =')
        if not lines:
            return

        centerings = [l.split()[-1] for l in lines]

        return Counter(centerings).most_common(n)

    def get_chunks(self):
        return self.parser.get_blocks_between(text1='---- Begin chunk -----',
                                              text2='----- End chunk -----', iMatch=-1)

    def parse_chunks(self):
        for chunk in self.chunks:
            self.parse_chunk(chunk)

    def get_indexed_chunks_original(self):
        return self.indexed_chunks

    def get_indexed_chunks(self):
        return "".join(self.indexed_chunks)

    def parse_chunk(self, chunk):
        results = dict()
        peak_end = None
        peak_start = None

        refl_start = None
        refl_end = None
        for i, line in enumerate(chunk):
            line = line.strip("\n")
            if "Image filename: " in line:
                full_filename = line.split()[-1]
                short_filename = full_filename[full_filename.rindex('/') + 1:]
                results['filename'] = full_filename
                results['run'] = full_filename.split('/')[-2]
                results['img_num'] = int(line[:-4].split('_')[-1])
                results['prefix'] = short_filename[:short_filename.rindex('_')]

            if "hit = " in line:
                results['hit'] = int(line.split()[-1])

            if "num_peaks = " in line:
                results['num_peaks'] = int(line.split()[-1])

            if "peak_resolution = " in line:
                peak_resol = line.split()[-2]
                results['peak_resolution'] = 0.0 if peak_resol == 'inf' else float(peak_resol)

            if "Peaks from peak search" in line:
                peak_start = i

            if "End of peak list" in line:
                peak_end = i

            # extra if indexed
            if 'Cell parameters' in line:
                cell = line.replace("nm,", "").split()[2:-1]
                results['cell'] = [round_nm2ang(i, e) for i, e in enumerate(cell)]  # convert to angstrom, instead of nm

            if 'lattice_type' in line:
                results['lattice_type'] = line.split()[-1]

            if 'centering' in line:
                results['centering'] = line.split()[-1]

            if 'unique_axis =' in line:
                results['unique_axis'] = line.split()[-1]

            if 'predict_refine/det_shift x' in line:
                flds = line.split()
                results['delta_x'] = flds[-5]
                results['delta_y'] = flds[-2]

            if 'predict_refine/final_residual' in line:
                results['fit_residual'] = line.split()[-1]

            if 'num_reflections' in line:
                results['num_reflections'] = int(line.split()[-1])

            if 'indexed_by' in line:
                results['indexed_by'] = line.split()[-1]

            if 'Reflections measured after indexing' in line:
                refl_start = i

            if 'End of reflections' in line:
                refl_end = i

        if peak_end and peak_start:
            results['list_of_peaks'] = StreamParser.get_adxv_spots(chunk[peak_start:peak_end])

        if refl_start and refl_end:
            results['list_of_reflections'] = StreamParser.get_adxv_predictions(chunk[refl_start:refl_end])

        if results['indexed_by'] != 'none':
            self.indexed_chunks.append("".join(chunk))

        self.all_results.append(results)

    def get_all_results(self):
        return self.all_results

    # reflection is shown as type 0 spots
    @staticmethod
    def get_adxv_spots(peaks, maxs=1000):
        if len(peaks) <= 2:
            return []

        xy_type = [peak.split()[:2] for peak in peaks[2:]][:maxs]
        xy_type = [list(map(round1, p)) + ['0'] for p in xy_type]
        return [" ".join(xyt) for xyt in xy_type]

    # reflection is shown as type 1 spots
    @staticmethod
    def get_adxv_predictions(reflections, maxp=3000):
        if len(reflections) <= 2:
            return []

        xy_type = [reflection.split()[-3:-1] for reflection in reflections[2:]]
        if len(xy_type) > maxp:
            mid = len(xy_type) // 2
            shift = maxp // 2
            xy_type = xy_type[mid - shift:mid + shift]

        xy_type = [list(map(round1, p)) + ['1'] for p in xy_type]
        return [" ".join(xyt) for xyt in xy_type]

    def save_to_json(self, dest):
        if self.all_results:
            with open(dest, "w") as fh:
                json.dump(self.all_results, fh)


if __name__ == '__main__':
    s = StreamParser(filename='/mnt/beegfs/qxu/test/indexing/run0001-1/run0001-1.stream')
    s = StreamParser(wdir='/mnt/beegfs/qxu/junk/', filename='t.stream')
    r = s.get_all_results()
    print(r)
    print(s.get_indexed_chunks())
    # s.save_to_json("a.json")
    # print(r['1'], type(r['1']), r['1']['filename'])
