
from __future__ import absolute_import
import logging
import os, shutil
from  collections import namedtuple

from .fileopener import FileOpener


logger = logging.getLogger(__name__)


class Xia2Parser(FileOpener):
    """ parse xia2 outputs """
    def __init__(self, wdir=".", filename="xia2.txt"):
        super(Xia2Parser, self).__init__(wdir, filename)
        self.data_dict = None
        self.anomalous_signal = 0

    def get_cell_space_group(self):
        results = {}
        lines = self.parser.get_lines_contain("Overall:")
        for line in lines:
            if line.startswith("Overall:"):
                results['unitcell'] = " ".join(lines[0].split()[-6:])

        lines = self.parser.get_lines_contain(text="Assuming spacegroup:")
        if lines:
            results['spacegroup'] = lines[0].split(":")[-1].strip().replace(" ", "")

        return results

    def get_table1(self, title=False, output='text'):
        skip1 = 0 if title else 1

        blocks = self.parser.get_blocks_between(text1="For ", text2="Assuming spacegroup:", iMatch=-1, skip1=skip1, skip2=1)
        if not blocks or not blocks[0]:
            logger.info("failed to find table 1 in {}/{}".format(self.wdir, self.filename))
            return

        table1 = blocks[-1]

        if output == 'text':
            return "".join(table1)
        elif output == 'list':
            return table1
        elif output == 'namedtuple':
            table1row = namedtuple('TABLE1', "paraName, Overall,  InnerShell,  OuterShell")
            ndtable1 = []
            for line in table1:
                values = line.rsplit()
                if "Wilson B factor" in line:
                    row = table1row("Wilson_B_factor", values[-1], "", "")
                    ndtable1.append(row)
                    continue

                if "Anomalous slope" in line:
                    row = table1row("Anomalous_slope", values[-1], "", "")
                    ndtable1.append(row)
                    continue

                if len(values) >= 4:
                    parameter_name, cols = '_'.join(values[:-3]), values[-3:]
                    row = table1row(parameter_name, *cols)
                    ndtable1.append(row)
            return ndtable1
        else:
            logger.warning("unknown output format defined in xia2parser get_table1")
            return


    def summarize(self):
        table1 = self.get_table1(output='namedtuple')
        if not table1:
            return

        # gather information for database, take the row assuming I+ = I-
        db_dict = {'table1': self.get_table1(title=True)}
        db_dict.update(self.get_cell_space_group())

        for e in table1:
            if "Low" in e.paraName and "resolution" in e.paraName:
                db_dict['lowresolution'] = e.Overall
                db_dict['lowresolution_inner'] = e.InnerShell
                db_dict['lowresolution_outer'] = e.OuterShell
                continue

            if "High" in e.paraName and "resolution" in e.paraName:
                db_dict['highresolution'] = e.Overall
                db_dict['highresolution_inner'] = e.InnerShell
                db_dict['highresolution_outer'] = e.OuterShell
                continue

            if "Rmerge(I)" in e.paraName:
                db_dict['rmerge'] = e.Overall
                db_dict['rmerge_inner'] = e.InnerShell
                db_dict['rmerge_outer'] = e.OuterShell
                continue

            if "Rmerge(I+/-)" in e.paraName:
                db_dict['anom_rmerge'] = e.Overall
                db_dict['anom_rmerge_inner'] = e.InnerShell
                db_dict['anom_rmerge_outer'] = e.OuterShell
                continue

            if "Rmeas(I)" in e.paraName:
                db_dict['rmeas'] = e.Overall
                db_dict['rmeas_inner'] = e.InnerShell
                db_dict['rmeas_outer'] = e.OuterShell
                continue

            if "Rpim(I)" in e.paraName:
                db_dict['rpim'] = e.Overall
                db_dict['rpim_inner'] = e.InnerShell
                db_dict['rpim_outer'] = e.OuterShell
                continue

            if e.paraName.startswith("I/sigma"):
                db_dict['isigmai'] = e.Overall
                db_dict['isigmai_inner'] = e.InnerShell
                db_dict['isigmai_outer'] = e.OuterShell
                continue

            if e.paraName.startswith("Completeness"):
                db_dict['completeness'] = e.Overall
                db_dict['completeness_inner'] = e.InnerShell
                db_dict['completeness_outer'] = e.OuterShell
                continue

            if e.paraName.startswith("Multiplicity"):
                db_dict['multiplicity'] = e.Overall
                db_dict['multiplicity_inner'] = e.InnerShell
                db_dict['multiplicity_outer'] = e.OuterShell
                continue

            if e.paraName.startswith("CC") and "half" in e.paraName:
                db_dict['cchalf'] = e.Overall
                db_dict['cchalf_inner'] = e.InnerShell
                db_dict['cchalf_outer'] = e.OuterShell
                continue

            if "Anomalous" in e.paraName and "completeness" in e.paraName:
                db_dict['anom_completeness'] = e.Overall
                db_dict['anom_completeness_inner'] = e.InnerShell
                db_dict['anom_completeness_outer'] = e.OuterShell
                continue

            if "Anomalous" in e.paraName and "multiplicity" in e.paraName:
                db_dict['anom_multiplicity'] = e.Overall
                db_dict['anom_multiplicity_inner'] = e.InnerShell
                db_dict['anom_multiplicity_outer'] = e.OuterShell
                continue

            if "Anomalous" in e.paraName and "correlation" in e.paraName:
                db_dict['anom_cchalf'] = e.Overall
                db_dict['anom_cchalf_inner'] = e.InnerShell
                db_dict['anom_cchalf_outer'] = e.OuterShell
                continue

            if "Total" in e.paraName and "observations" in e.paraName:
                db_dict['Nobs'] = e.Overall
                db_dict['Nobs_inner'] = e.InnerShell
                db_dict['Nobs_outer'] = e.OuterShell
                continue

            if "Total" in e.paraName and "unique" in e.paraName:
                db_dict['Nuniq'] = e.Overall
                db_dict['Nuniq_inner'] = e.InnerShell
                db_dict['Nuniq_outer'] = e.OuterShell
                continue

            if e.paraName.startswith("Anomalous") and "slope" in e.paraName:
                db_dict['anom_slope'] = e.Overall

        db_dict["data_quality"] = Xia2Parser.evaluate_data_quality(db_dict)

        db_dict.update(self.get_cell_space_group())
        db_dict.update({"data_quality": Xia2Parser.evaluate_data_quality(db_dict)})
        self.data_dict = db_dict

        # needs to be improved
        anom_cc_total = db_dict['anom_cchalf']
        anom_cc_inner = db_dict['anom_cchalf_inner']
        anom_cc_outer = db_dict['anom_cchalf_outer']
        anom_slope = db_dict['anom_slope']

        try:
            if float(anom_slope) > 1.3:
                self.anomalous_signal = 10

            if float(anom_cc_inner) > .90 and float(anom_cc_outer) > .20:
                self.anomalous_signal = 10

            if float(anom_cc_inner) < .30 and float(anom_cc_outer) < .20:
                self.anomalous_signal = 0
            else:
                self.anomalous_signal = 5

            db_dict.update({"anomalous_signal": self.anomalous_signal})
        except Exception:
            pass

        return db_dict

    def evaluate_anomalous_signal(self):
        """
        return 0: weak anomalous signal
        return num: strong anomalous signal, suggested resolution cutoff for anomalous signal
        return 10: strong anomalous signal but no cutoff found, shouldn't happen!
        """
        return self.anomalous_signal

    @staticmethod
    def evaluate_data_quality(data_dict):
        resol = data_dict.get('highresolution')
        completeness = data_dict.get('completeness')
        rmerge = data_dict.get('rmerge')

        try:
            if float(rmerge) < 0.3 and float(resol) < 3.5 and float(completeness) > 90:
                return True
        except Exception:
            return False

        return False


if __name__ == "__main__":
    '''
    xp = Xia2Parser("/mnt/beegfs/qxu/23BM_2017_08_08/4z2x-p21212/process/repo/xia2_dials.nativescale")


    print(xp.get_cell_space_group())

    print(xp.get_table1(output='namedtuple'))

    from pprint import pprint
    pprint(xp.summarize())
    '''

    xp = Xia2Parser("/mnt/beegfs/qxu/23BM_2017_08_08/SP13538A_4qrl/process1/repo/xia2_dials")


    print(xp.get_cell_space_group())

    print(xp.get_table1(output='namedtuple'))

    from pprint import pprint
    pprint(xp.summarize())
