from __future__ import absolute_import

import datetime
import logging
import os

from peewee import *

from ..process.get_args import ARGS

logger = logging.getLogger(__name__)

beamline = ARGS.beamline  # obtain beamline information from globals


def create_db():
    # start of module code
    DEBUG = False

    # preconfigured beamlines
    if beamline in ('23i', 'idd'):
        return MySQLDatabase('user_data',
                             host='bl1upper',
                             user='dhs',
                             passwd='')
        logger.info("results will be written to bl1upper")
    elif beamline in ('23o', 'idb'):
        return MySQLDatabase('user_data',
                             host='bl2upper',
                             user='dhs',
                             passwd='')
        logger.info("results will be written to bl2upper")
    elif beamline in ('23b', '23bm'):
        return MySQLDatabase('user_data',
                             host='bl3upper',
                             user='dhs',
                             passwd='',
                             thread_safe=True,
                             )

        '''
        from playhouse.pool import PooledMySQLDatabase

        db = PooledMySQLDatabase('user_data',
                                 max_connections=ARGS.nproc,
                                 stale_timeout=300,
                                 host='bl3upper',
                                 user='dhs',
                                 passwd='')
        '''
        logger.info("results will be written to bl3upper")

    elif beamline in 'config':
        from ..config import config
        return MySQLDatabase(config.MYSQLDB['dbname'],
                             host=config.MYSQLDB['host'],
                             user=config.MYSQLDB['user'],
                             passwd=config.MYSQLDB['password'])
        logger.info("results will be written to db defined by config file.")
    elif DEBUG:
        return SqliteDatabase("data_processing.db")
        logger.info("debugging mode, written to file data_processing.db")
    else:
        logger.info("db not defined.")

    return None


db = create_db()


class BaseModel(Model):
    class Meta:
        database = db


# record information for every job
class PipelineStatus(BaseModel):
    id = PrimaryKeyField()
    command = CharField(default="", max_length=1024)
    state = CharField(default="", max_length=20)
    pipeline = CharField(default="gmcaproc", index=True, max_length=20)
    imagedir = CharField(default=os.getenv('HOME'))
    # processing_root_dir = CharField(default="")
    workdir = CharField(default="")
    log = TextField(default="")
    warning = TextField(default="")
    processing_id = IntegerField(default=0)
    hostname = CharField(default="")
    sampleName = CharField(default="sampleName")
    sampleNumber = IntegerField(default=0)
    username = CharField(default=ARGS.username, index=True, max_length=20)
    beamline = CharField(default="")
    starttime = DateTimeField(default=datetime.datetime.now)
    elapsedtime = CharField(default="")
    imageSet = CharField(default="All")
    logfile = CharField(default="", max_length=1024)
    delete = IntegerField(default=0)

    class Meta:
        order_by = ('id',)


class DataProcessResults(BaseModel):
    # original columns
    sampleNumber = IntegerField(default=0)
    sampleName = CharField(default="sampleName")
    imageSet = CharField(default="All")
    state = CharField(default="")
    software = CharField(default="")
    scalingsoftware = CharField(default='xscale-aimless')
    collectType = CharField(default="")
    overlap = IntegerField(default=0)
    inverseOn = IntegerField(default=0)
    firstFrame = CharField(default="")
    workdir = CharField(default="")
    subdir = CharField(default="")
    imagedir = CharField(default="")
    images = CharField(default="")
    unitcell = CharField(default="")

    lowresolution = CharField(default="")
    lowresolution_inner = CharField(default="")
    lowresolution_outer = CharField(default="")
    highresolution = CharField(default="")
    highresolution_inner = CharField(default="")
    highresolution_outer = CharField(default="")
    rmerge = CharField(default="")
    rmerge_inner = CharField(default="")
    rmerge_outer = CharField(default="")
    rpim = CharField(default="")
    rpim_inner = CharField(default="")
    rpim_outer = CharField(default="")
    cchalf = CharField(default="")
    cchalf_inner = CharField(default="")
    cchalf_outer = CharField(default="")
    completeness = CharField(default="")
    completeness_inner = CharField(default="")
    completeness_outer = CharField(default="")
    anom_completeness = CharField(default="")
    anom_completeness_inner = CharField(default="")
    anom_completeness_outer = CharField(default="")
    multiplicity = CharField(default="")
    multiplicity_inner = CharField(default="")
    multiplicity_outer = CharField(default="")
    anom_multiplicity = CharField(default="")
    anom_multiplicity_inner = CharField(default="")
    anom_multiplicity_outer = CharField(default="")
    isigmai = CharField(default="")
    isigmai_inner = CharField(default="")
    isigmai_outer = CharField(default="")
    spacegroup = CharField(default="")
    truncate_log = CharField(default="")
    truncate_mtz = CharField(default="")  # mtz file link for truncate
    run_stats = CharField(default="")  # store run/sweep level statistics by xds
    log = CharField(default="")
    warning = TextField(default="")
    timestamp = DateTimeField(default=datetime.datetime.now)

    # new fields
    id = PrimaryKeyField()
    scale_log = CharField(default="")
    start = CharField(default="1")
    end = CharField(default="9999")
    prefix = CharField(default="")
    isa = CharField(default="")
    rmeas = CharField(default="")
    rmeas_inner = CharField(default="")
    rmeas_outer = CharField(default="")
    pipeline = CharField(default="gmcaproc")
    wavelength = CharField(default="")
    nobs = CharField(default="0")
    nuniq = CharField(default="0")
    table1 = TextField(default="")

    report_url = CharField(default="", max_length=1024)
    # store id of job status, aka pipelinestatus primary key
    # note: peewee will add append the primary key name when creating the database
    pipelinestatus = ForeignKeyField(PipelineStatus, on_delete='CASCADE')
    # a field to track manual reprocess
    reprocess = IntegerField(default=0)
    solve = CharField(default="")

    spacegroup_choices = TextField(default="")

    class Meta:
        order_by = ('id',)


# This combines the SilTable and StrategyResults
class ScreenStrategyResults(BaseModel):
    """
      `sampleNumber` int(11) NOT NULL,
      `selected` tinyint(4) DEFAULT '0',
      `ordered` tinyint(4) DEFAULT '0',
      `screened` tinyint(4) DEFAULT '0',
      `Port` varchar(30) NOT NULL,
      `CrystalID` varchar(250) NOT NULL,
      `Directory` varchar(250) NOT NULL,
      `Comment` varchar(250) NOT NULL,
      `workdir` varchar(250) DEFAULT '',
      `image1` varchar(100) DEFAULT '',
      `image2` varchar(100) DEFAULT '',
      `Images` varchar(250) DEFAULT '',
      `Score` varchar(30) DEFAULT '',
      `Resolution` varchar(30) DEFAULT '',
      `Protein` varchar(250) DEFAULT '',
      `Mosaicity` varchar(50) DEFAULT '',
      `UnitCell` varchar(50) DEFAULT '',
      `BravaisLattice` varchar(250) DEFAULT '',
      `Rmsd` varchar(50) DEFAULT '',
      `IceRings` varchar(50) DEFAULT '',
      `warning` varchar(5120) DEFAULT '',
      `currentPrefix` varchar(150) DEFAULT '',
      `selectedPrefix` varchar(150) DEFAULT '',
      `prefixes` varchar(5120) DEFAULT '',
      `timestamp` varchar(512) DEFAULT '',
      PRIMARY KEY (`sampleNumber`)
    """
    sampleNumber = PrimaryKeyField()
    sampleName = CharField(max_length=512, default="")  # from strategy
    # selected = SmallIntegerField(default=0)
    # ordered = SmallIntegerField(default=0)
    # screened = SmallIntegerField(default=0)
    # port = CharField(max_length=30, default="")
    # crystal_id = CharField(max_length=250, default="")
    # comment = TextField(default="")
    # protein = CharField(max_length=250, default="")
    # currentPrefix = CharField(max_length=150, default="")
    # selectedPrefix = CharField(max_length=150, default="")
    # timestamp = CharField(max_length=512, default="")

    # screening images
    directory = CharField(max_length=250, default="")
    images = CharField(max_length=250, default="")

    # characterizations, spots, autoindex, icerings, mosaicity etc,
    software = CharField(default="")
    state = CharField(default='SPOT')  # SPOT, AUTOINDEX, STRATEGY
    workdir = CharField(max_length=512, default="")
    index_table = TextField(default="")
    unitcell = CharField(max_length=100, default="")
    bravais_lattice = CharField(max_length=10, default="")
    rmsd = CharField(max_length=50, default="")
    ice_rings = CharField(max_length=50, default="")
    resolution_from_spots = CharField(max_length=30, default="")
    n_spots = CharField(max_length=30, default="")  # a new field
    n_spots_ice = CharField(max_length=30, default="")
    n_ice_rings = CharField(max_length=30, default="")
    avg_spotsize = CharField(max_length=20, default="")
    spacegroup = CharField(max_length=30, default="")
    solution_number = CharField(max_length=30, default="")
    penalty = CharField(max_length=30, default="")
    mosaicity = CharField(max_length=50, default="")
    score = CharField(max_length=30, default="")
    resolution_from_integ = CharField(max_length=30, default="")
    warning = TextField(default="")

    # strategy
    """
      `sampleNumber` int(11) NOT NULL,
      `sampleName` varchar(512) DEFAULT '',
      `solutionNumber` int(11) DEFAULT NULL,
      `software` varchar(50) DEFAULT '',
      `good` varchar(11) DEFAULT 'no',
      `oscstart` varchar(30) DEFAULT '',
      `anom_oscstart` varchar(30) DEFAULT '',
      `oscend` varchar(30) DEFAULT '',
      `anom_oscend` varchar(30) DEFAULT '',
      `continuous` varchar(30) DEFAULT '',
      `complete` varchar(30) DEFAULT '',
      `anom_complete` varchar(30) DEFAULT '',
      `oscdelta` varchar(30) DEFAULT '',
      `detectordistance` varchar(30) DEFAULT '',
      `detectorwarning` varchar(512) DEFAULT '',
      `spaceGroup_short` varchar(100) DEFAULT '',
      `spaceGroup` varchar(100) DEFAULT '',
      `resolution` varchar(30) DEFAULT '',
      `unitcell` varchar(100) DEFAULT '',
      `rmsd` varchar(30) DEFAULT '',
      `beamx` varchar(30) DEFAULT '',
      `beamy` varchar(30) DEFAULT ''
    """
    anomalous = SmallIntegerField(default=0)  # new field
    # good = CharField(max_length=11, default="")
    osc_start = CharField(max_length=30, default="")
    osc_end = CharField(max_length=30, default="")
    osc_delta = CharField(max_length=30, default="")
    completeness_native = CharField(max_length=30, default="")
    completeness_anomalous = CharField(max_length=30, default="")
    completeness_referencedata = CharField(max_length=30, default="")

    # continuous = CharField(max_length=30, default="")
    detectorwarning = CharField(max_length=512, default="")
    detectordistance = CharField(max_length=30, default="350.0")

    # multi-strategy
    """
      `state` varchar(30) DEFAULT '',
      `workdir` varchar(250) DEFAULT '',
      `referencedata` varchar(512) DEFAULT '',
      `images` varchar(512) DEFAULT '',
      `unitcell` varchar(512) DEFAULT '',
      `lowresolution` varchar(30) DEFAULT '',
      `highresolution` varchar(30) DEFAULT '',
      `spacegroup` varchar(512) DEFAULT '',
      `displaytext` varchar(250) DEFAULT '',
      `strategy` varchar(2048) DEFAULT '',
      `scalalog` varchar(512) DEFAULT '',
      `xplanlog` varchar(512) DEFAULT '',
    """
    referencedata = CharField(max_length=512, default="")
    displaytext = CharField(max_length=250, default="")
    # scalalog = CharField(max_length=512, default="")
    xplanlog = TextField(default="")
    strategy = TextField(default="")

    # store id of job status, aka pipelinestatus primary key
    pipelinestatus = ForeignKeyField(PipelineStatus, on_delete='CASCADE')
    # a field to track manual reprocess
    reprocess = IntegerField(default=0)
    solvent_content = TextField(default="")
    estimated_asu_content_aa = CharField(default="")
    # export 2 a new run
    export2run = IntegerField(default=0)

    pointgroup_choices = TextField(default="")


def initialize_tables():
    if not ARGS.log2db:
        return

    if not db:
        create_db()

    if not db:
        raise SystemExit("unable to create database.")

    try:
        db.connect(reuse_if_open=True)
        db.create_tables([PipelineStatus, DataProcessResults, ScreenStrategyResults, ], safe=True)
        db.close()
    except:
        logger.error("database connection or creation error.", exc_info=True)


def save_to_db(data_dict, **kwargs):
    data_dict.update(kwargs)

    '''
    # if ARGS.htmlout is true, generate html elements
    if 'table1' in data_dict:
        data_dict['table1'] = Htmlstr(data_dict['table1'], to_html=ARGS.htmlout).wrap("pre")

    if 'truncate_mtz' in data_dict:
        data_dict['truncate_mtz'] = Htmlstr(data_dict['truncate_mtz'], to_html=ARGS.htmlout).toLinkOfHtmlButton(toUri=True, btnText="Download", target='_parent')

    if 'truncate_log' in data_dict:
        data_dict['truncate_log'] = Htmlstr(data_dict['truncate_log'], to_html=ARGS.htmlout).toLinkOfHtmlButton(toUri=True, btnText="Open", target='_blan')


    data_dict['delete'] = Htmlstr('/delete/{}'.format(ARGS.pipelinestatus_id), to_html=ARGS.htmlout).toLinkOfHtmlButton(toUri=False, btnText="Delete", target='_parent')
    '''

    try:
        logger.info("saving data to db.")
        with db.atomic():
            dp = DataProcessResults.create(**data_dict)
            return dp
        logger.info("saved to db.")
    except Exception as e:
        logger.error("error saving data to db ")
        print("data= {}".format(data_dict))
        raise Exception


initialize_tables()
