#!/bin/env python
## summation of multiple files for David Akey

import argparse
import os
import re
import sys
from collections import OrderedDict
from functools import partial
from multiprocessing import Pool

import fabio
import numpy as np

from .cbfimage import CbfImage
from ...common.generatesweeps import GenerateSweeps


def add(*args):
    return sum(args) if all(e >= 0 for e in args) else min(args)


v_add = np.vectorize(add, otypes=[np.int32])


def add_np(l_matrix):
    if len(l_matrix) <= 1:
        return l_matrix[0]
    elif len(l_matrix) >= 2:
        return np.add(l_matrix[0], add_np(l_matrix[1:]))


def create_new_header(headers):
    nbin = len(headers)

    content = headers[0]['_array_data.header_contents']

    # Angle_increment 0.100000 deg.
    rx = r'''# Angle_increment (0.\d{4,6}) deg.'''
    x = re.search(rx, content)
    new_osc = '# Angle_increment {:0.6f} deg.'.format(float(x.group(1)) * nbin)
    content = re.sub(rx, new_osc, content)

    # # Count_cutoff 24006 counts
    rx = r'''# Count_cutoff (\d+) counts'''
    x = re.search(rx, content)
    new_count = '# Count_cutoff {:d} counts'.format(int(x.group(1)) * nbin)
    content = re.sub(rx, new_count, content)

    # Exposure_time 0.009990 s
    rx = r'''# Exposure_time (0.\d+) s'''
    x = re.search(rx, content)
    new_exp = '# Exposure_time {:0.6f} s'.format(float(x.group(1)) * nbin)
    content = re.sub(rx, new_exp, content)

    # Exposure_time 0.009990 s
    rx = r'''# Exposure_period (0.\d+) s'''
    x = re.search(rx, content)
    new_expp = '# Exposure_period {:0.6f} s'.format(float(x.group(1)) * nbin)
    content = re.sub(rx, new_expp, content)

    # copy all header items
    # new_header = headers[0]
    # new_header['_array_data.header_contents'] = content

    # just update the header_contents
    new_header = OrderedDict()
    new_header['_array_data.header_contents'] = content
    return new_header


def combine(i, image_dir=None, output_dir=None, binsize=5, mode=2, f_template='Z_NS1_114_3_00{:04d}.cbf',
            f_template_new='new_Z_NS1_114_3_00{:04d}.cbf'):
    images = [os.path.join(image_dir, f_template.format(j)) for j in range(i * binsize + 1, (i + 1) * binsize + 1)]
    fimages = [fabio.open(image) for image in images]
    headers = [f.header for f in fimages]
    data = [f.data for f in fimages]

    if mode == 1:
        new_data = v_add(*data)
    else:
        new_data = add_np(data)

    new_header = create_new_header(headers)

    fname = os.path.join(output_dir, f_template_new.format(i + 1))
    cbf = CbfImage(header=new_header, data=new_data)
    cbf.write(fname)
    print("done writing {}...".format(fname))


def get_args():
    parser = argparse.ArgumentParser(prog='cbfmerge')
    parser.add_argument('--data_dir', '-i',
                        required=True,
                        help='directory for image files')

    parser.add_argument('--out_dir', '-o',
                        required=False,
                        default=os.getcwd(),
                        help='output directory for image files')

    parser.add_argument('--binsize', '-n',
                        required=False,
                        type=int,
                        default=5,
                        help='how many images to combine into a single image')

    parser.add_argument('--prefix', '-p',
                        required=False,
                        default="new",
                        help='string to preprend to output filenames')

    args = parser.parse_args()
    return args


def merge(args):
    pool = Pool(5)

    sweeps = GenerateSweeps(args.data_dir)

    for sw in sweeps.get_file_sweeps():
        nimages = sw.nfiles

        if nimages % args.binsize != 0:
            print("skipped sweep {}, incorrect binsize {}, choose it so that nimages/nbin is an integer.".format(
                sw.get_name_template(), args.binsize))
            continue

        template = sw.get_name_template(marker='#')
        ndigits = template.count("#")
        f_template = template.replace(ndigits * "#", "{{:0{}d}}".format(ndigits))
        f_template_new = "{}_{}".format(args.prefix, f_template)

        nbatch = nimages // args.binsize
        func = partial(combine,
                       image_dir=args.data_dir,
                       output_dir=args.out_dir,
                       binsize=args.binsize,
                       mode=2,
                       f_template=f_template,
                       f_template_new=f_template_new)

        debug = False
        if debug:
            for i in range(nbatch):
                func(i)
        else:
            pool.map(func, range(nbatch))


if __name__ == '__main__':
    args = get_args()

    if not os.path.exists(args.data_dir):
        print("image directory does not exist.")
        sys.exit(0)

    if not os.path.exists(args.out_dir):
        os.makedirs(args.out_dir)

    merge(args=args)

    '''
    directory = '/home/qxu/data-for-tests/23IDB_2017_03_07_dla/Z_NS1_114/cbf_3'
    f_template = 'Z_NS1_114_3_00{:04d}.cbf'
    f_template_new = 'new_Z_NS1_114_3_00{:04d}.cbf'
    nimages = 3600
    nbin = 2

    if nimages % nbin != 0:
        print("incorrect bin size, choose it so that nimages/nbin is an integer.")
        sys.exit(1)

    nbatch = nimages // nbin
    mode = 2  # add except for bad pixel (mode 1) or gap or diretct add (mode 2)

    pool = Pool(3)

    func = partial(combine, directory=directory, nbin=nbin, mode=mode, f_template=f_template,
                   f_template_new=f_template_new)
    pool.map(func, range(nbatch))

    sys.exit(0)
    '''

    #
    # rdata = os.path.join(directory, 'new_Z_NS1_114_3_000001.cbf')
    # x = fabio.open(rdata)
    # print(x.data, x.header)
    #
    # count = 0
    # for i in range(nbatch):
    #     # read nbin images [i*nbin + 1 .. (i+1)*bin]
    #     combine(i, directory, nbin)
    #
    #
    #     # images = [os.path.join(directory, f_template.format(j)) for j in range(i*nbin+1, (i+1)*nbin+1)]
    #     # fimages = [fabio.open(image) for image in images]
    #     # headers = [f.header for f in fimages]
    #     # data = [f.data for f in fimages]
    #     # new_data = v_add(*data)
    #     # new_header = create_new_header(headers)
    #     #
    #     # fname = os.path.join(directory, f_template_new.format(i+1))
    #     # cbf = fabio.cbfimage.cbfimage(header=new_header,data=new_data)
    #     # cbf.write(fname)
    #
    #     count += 1
    #     if count >= 10:
    #         break

    # a1 = np.array([1,2,3])
    # a2 = np.array([1,2,3])
    # a3 = np.array([1,2,3])
    #
    # print(add_np([a1,a2, a1, a1]))
    # print(a1)

sys.exit(1)

#
# file1 = os.path.join(directory, 'D1_1_1_000001.cbf')
# file2 = os.path.join(directory, 'D1_1_1_000002.cbf')
# file3 = os.path.join(directory, 'D1_1_1_000003.cbf')
#
#
#
# i1 = fabio.open(file1)
# i2 = fabio.open(file2)
# i3 = fabio.open(file3)
#
# # print(type(i1))
# #
# # print("max=", np.max(i1.data),"mean=", np.mean(i1.data), 'median=', np.median(i1.data))
# #
# #
# # full = i1.data
# # # Get the indices for the largest `num_largest` values.
# # num_largest = 8
# #
# # indices = (-full).argpartition(num_largest, axis=None)[:num_largest]
# # # OR, if you want to avoid the temporary array created by `-full`:
# # # indices = full.argpartition(full.size - num_largest, axis=None)[-num_largest:]
# #
# # print(indices)
# # x, y = np.unravel_index(indices, full.shape)
# # print("x =", x)
# # print("y =", y)
# # print("Largest values:", full[x, y])
# #
# # print(full.shape)
# # for a, b in zip(x,y):
# #     print('x, y, v=', a,b, full[a,b])
# #
# # print("Compare to:    ", np.sort(full, axis=None)[-num_largest:])
# #
# #
# #
#
# # data12 = i1.data + i2.data
# #
# # print(dir(data12), data12.shape)
# #
# # # for (i,j), value in np.ndenumerate(data12):
# # #     if i1.data[i,j] < 0 or i2.data[i,j] < 0:
# # #         data12[i,j] = min(i1.data[i,j], i2.data[i,j])
# # #
# # # print(data12)
# #
# # # data12tuple = np.vstack(([i1.data.T], [i2.data.T])).T
# # # print(data12tuple.shape)
# #
#
#
# data123 = v_add(i1.data, i2.data, i3.data)
