#!/opt/anaconda/bin/python

import argparse
import io
import json
import os
import smtplib
import socket
import sys
import tarfile
import urllib.parse
from collections import namedtuple
from datetime import datetime, timedelta, date
from email import encoders
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from itertools import repeat
from multiprocessing import Pool

import pandas as pd
import requests
from dateutil import parser


def html_link(url, display_text='link'):
    # return '<a target="_blank" href="{}">{}</a>'.format(url, display_text)
    return '=HYPERLINK("{}", "{}")'.format(url, display_text)


def doi_url(doi_or_pmid, html=False):
    if doi_or_pmid:
        doi_or_pmid = str(doi_or_pmid)
        if "/" in doi_or_pmid:
            url = "https://doi.org/{}".format(doi_or_pmid)
        else:
            url = "http://www.ncbi.nlm.nih.gov/pubmed/{}".format(doi_or_pmid)

        if html:
            return html_link(url, doi_or_pmid)
        else:
            return url


def to_date(datetimefield):
    return parser.parse(datetimefield).date() if datetimefield else None


def date_str(adate):
    return adate.strftime("%Y-%m-%d") if adate else None


def iso_date_str(adate, start=True):
    """return an iso string of time at start or end of the day given"""
    if start:
        atime = datetime(year=adate.year, month=adate.month, day=adate.day, hour=0, minute=0, second=0, )
    else:
        atime = datetime(year=adate.year, month=adate.month, day=adate.day, hour=23, minute=59, second=59, )

    return datetime.strftime(atime, '%Y-%m-%dT%H:%M:%SZ')


def concat_authors(authors):
    n_authors = len(authors)
    if n_authors > 3:
        first_author = authors[0]
        last_two_authors = authors[n_authors - 2:]
        authors = [first_author, ' et al., '] + last_two_authors

    return ",".join(authors)


def get_query_dict_from_url():
    """
    procedure for obtaining a query_dict from RCSB website,
    1) contruct a query using the advanced query builder
    2) run search & order the results as desired
    3) copy the url from the web browswer address line, and input as string of this function
    """
    url = """https://www.rcsb.org/search?request=%7B%22query%22%3A%7B%22type%22%3A%22group%22%2C%22logical_operator%22%3A%22and%22%2C%22nodes%22%3A%5B%7B%22type%22%3A%22group%22%2C%22logical_operator%22%3A%22and%22%2C%22nodes%22%3A%5B%7B%22type%22%3A%22group%22%2C%22logical_operator%22%3A%22or%22%2C%22nodes%22%3A%5B%7B%22type%22%3A%22terminal%22%2C%22service%22%3A%22text%22%2C%22parameters%22%3A%7B%22operator%22%3A%22exact_match%22%2C%22negation%22%3AFalse%2C%22value%22%3A%22Severe%20acute%20respiratory%20syndrome%20coronavirus%202%22%2C%22attribute%22%3A%22rcsb_entity_source_organism.ncbi_scientific_name%22%7D%2C%22node_id%22%3A0%7D%2C%7B%22type%22%3A%22terminal%22%2C%22service%22%3A%22text%22%2C%22parameters%22%3A%7B%22negation%22%3AFalse%2C%22value%22%3A%22%5C%22COVID-19%5C%22%22%7D%2C%22node_id%22%3A1%7D%2C%7B%22type%22%3A%22terminal%22%2C%22service%22%3A%22text%22%2C%22parameters%22%3A%7B%22negation%22%3AFalse%2C%22value%22%3A%22%5C%22SARS-COV-2%5C%22%22%7D%2C%22node_id%22%3A2%7D%5D%7D%5D%2C%22label%22%3A%22text%22%7D%5D%2C%22label%22%3A%22query-builder%22%7D%2C%22return_type%22%3A%22entry%22%2C%22request_options%22%3A%7B%22pager%22%3A%7B%22start%22%3A0%2C%22rows%22%3A100%7D%2C%22scoring_strategy%22%3A%22combined%22%2C%22sort%22%3A%5B%7B%22sort_by%22%3A%22rcsb_accession_info.initial_release_date%22%2C%22direction%22%3A%22desc%22%7D%5D%7D%2C%22request_info%22%3A%7B%22src%22%3A%22ui%22%2C%22query_id%22%3A%22088520d56e1f8d03852cfcacd18a5425%22%7D%7D"""
    parameters = url.replace("https://www.rcsb.org/search?request=", "")
    print(parameters)
    parameters_uq = urllib.parse.unquote(parameters)
    # query_str = urllib.parse.parse_qs(url)
    query_dict = json.loads(parameters_uq, strict=False)
    print(parameters_uq)
    from pprint import pprint
    pprint(query_dict)
    qd = {"query": {"type": "group", "logical_operator": "and", "nodes": [{"type": "group", "logical_operator": "and",
                                                                           "nodes": [{"type": "group",
                                                                                      "logical_operator": "or",
                                                                                      "nodes": [{"type": "terminal",
                                                                                                 "service": "text",
                                                                                                 "parameters": {
                                                                                                     "operator": "exact_match",
                                                                                                     "negation": False,
                                                                                                     "value": "Severe acute respiratory syndrome coronavirus 2",
                                                                                                     "attribute": "rcsb_entity_source_organism.ncbi_scientific_name"},
                                                                                                 "node_id": 0},
                                                                                                {"type": "terminal",
                                                                                                 "service": "text",
                                                                                                 "parameters": {
                                                                                                     "negation": False,
                                                                                                     "value": "\"COVID-19\""},
                                                                                                 "node_id": 1},
                                                                                                {"type": "terminal",
                                                                                                 "service": "text",
                                                                                                 "parameters": {
                                                                                                     "negation": False,
                                                                                                     "value": "\"SARS-COV-2\""},
                                                                                                 "node_id": 2}]}],
                                                                           "label": "text"}], "label": "query-builder"},
          "return_type": "entry",
          "request_options": {"pager": {"start": 0, "rows": 100}, "scoring_strategy": "combined",
                              "sort": [{"sort_by": "rcsb_accession_info.initial_release_date", "direction": "desc"}]},
          "request_info": {"src": "ui", "query_id": "088520d56e1f8d03852cfcacd18a5425"}}
    pprint(qd)


def xref_lookup(doi_or_pmid):
    """
    examples:

    https://www.ncbi.nlm.nih.gov/pmc/utils/idconv/v1.0/?tool=my_tool&email=my_email@example.com&ids=10.1093/nar/gks1195

    https://www.ncbi.nlm.nih.gov/pmc/utils/idconv/v1.0/?ids=10.1038/nature14656

    json format

    https://www.ncbi.nlm.nih.gov/pmc/utils/idconv/v1.0/?ids=10.1016/j.ajhg.2010.04.006&format=json&tool=my_tool&email=my_email@example.com

    """
    doi = None
    pmid = None
    pmcid = None

    url = 'https://www.ncbi.nlm.nih.gov/pmc/utils/idconv/v1.0/'
    header = {'Content-Type': 'application/xml; charset=utf-8'}

    data = {"ids": doi_or_pmid, "format": "json"}
    response = requests.post(url, data=data)

    ret = response.json()

    if 'records' in ret:
        first_item = ret['records'][0]
        if "doi" in first_item:
            doi = first_item['doi']

        if "pmid" in first_item:
            pmid = first_item['pmid']

        if 'pmcid' in first_item:
            pmcid = first_item['pmcid']

    return doi, pmid, pmcid


class RCSB:
    def __init__(self, *args):
        self.pdb_ids = list()
        self.PARALLEL = 5
        for item in args:
            if item not in self.pdb_ids:
                self.pdb_ids.append(item)

    def lower(self):
        self.pdb_ids = [item.lower() for item in self.pdb_ids]

    def upper(self):
        self.pdb_ids = [item.upper() for item in self.pdb_ids]

    def add(self, *args):
        for item in args:
            item = item.lower()
            if item not in self.pdb_ids:
                self.pdb_ids.append(item)
            else:
                print("Warning: {} already added.".format(item))

    def search_by_text(self, text="text"):
        query_dict = {
            "query": {
                "parameters": {"value": "\"{}\"".format(text)},
                "type": "terminal",
                "service": "text"
            },
            "request_options":
                {
                    "return_all_hits": True,
                },
            "return_type": "entry"
        }
        return self.search(query_dict)

    def search(self, query_dict):
        print("Search pdb using the search API")
        header = {'Content-Type': 'application/json'}
        response = requests.post(
            "https://search.rcsb.org/rcsbsearch/v1/query", data=json.dumps(query_dict), headers=header)

        if response.status_code == 404:
            print("No entry found.")
            return

        data = response.json()
        pdb_ids = [entry['identifier'] for entry in data['result_set']]
        self.add(*pdb_ids)
        print("Found {} entries: {}".format(len(pdb_ids), " ".join(self.pdb_ids)))
        return pdb_ids

    @staticmethod
    def query_graphql(graphql_str):
        """
        ex = "https://data.rcsb.org/graphql/index.html?query=%7B%0A%20%20entries(entry_ids%3A%20%5B%226PYP%22%5D)%20%7B%0A%20%20%20%20rcsb_id%0A%20%20%20%20audit_author%20%7B%0A%20%20%20%20%20%20name%0A%20%20%20%20%7D%0A%20%20%20%20struct%20%7B%0A%20%20%20%20%20%20title%0A%20%20%20%20%7D%0A%20%20%20%20rcsb_entry_info%20%7B%0A%20%20%20%20%20%20resolution_combined%0A%20%20%20%20%7D%0A%20%20%20%20rcsb_accession_info%20%7B%0A%20%20%20%20%20%20deposit_date%0A%20%20%20%20%20%20initial_release_date%0A%20%20%20%20%7D%0A%20%20%20%20diffrn_source%20%7B%0A%20%20%20%20%20%20type%0A%20%20%20%20%20%20pdbx_wavelength%0A%20%20%20%20%20%20pdbx_wavelength_list%0A%20%20%20%20%20%20pdbx_synchrotron_beamline%0A%20%20%20%20%7D%0A%20%20%20%20diffrn_detector%20%7B%0A%20%20%20%20%20%20pdbx_collection_date%0A%20%20%20%20%20%20type%0A%20%20%20%20%7D%0A%20%20%20%20diffrn%20%7B%0A%20%20%20%20%20%20ambient_temp%0A%20%20%20%20%7D%0A%20%20%20%20reflns%20%7B%0A%20%20%20%20%20%20d_resolution_high%0A%20%20%20%20%20%20d_resolution_low%0A%20%20%20%20%20%20number_all%0A%20%20%20%20%20%20number_obs%0A%20%20%20%20%20%20pdbx_Rmerge_I_obs%0A%20%20%20%20%20%20pdbx_Rsym_value%0A%20%20%20%20%20%20pdbx_netI_over_sigmaI%0A%20%20%20%20%20%20pdbx_redundancy%0A%20%20%20%20%20%20percent_possible_obs%0A%20%20%20%20%20%20phase_calculation_details%0A%20%20%20%20%20%20pdbx_CC_half%0A%20%20%20%20%7D%0A%20%20%20%20reflns_shell%20%7B%0A%20%20%20%20%20%20Rmerge_I_obs%0A%20%20%20%20%20%20d_res_high%0A%20%20%20%20%20%20d_res_low%0A%20%20%20%20%20%20percent_possible_obs%0A%20%20%20%20%20%20meanI_over_sigI_obs%0A%20%20%20%20%20%20pdbx_CC_half%0A%20%20%20%20%7D%0A%20%20%20%20refine%20%7B%0A%20%20%20%20%20%20ls_R_factor_R_free%0A%20%20%20%20%20%20ls_R_factor_R_work%0A%20%20%20%20%20%20ls_d_res_high%0A%20%20%20%20%20%20ls_d_res_low%0A%20%20%20%20%20%20pdbx_starting_model%0A%20%20%20%20%7D%0A%20%20%20%20rcsb_primary_citation%20%7B%0A%20%20%20%20%20%20journal_volume%0A%20%20%20%20%20%20page_first%0A%20%20%20%20%20%20page_last%0A%20%20%20%20%20%20pdbx_database_id_DOI%0A%20%20%20%20%20%20rcsb_authors%0A%20%20%20%20%20%20rcsb_journal_abbrev%0A%20%20%20%20%20%20title%0A%20%20%20%20%20%20year%0A%20%20%20%20%7D%0A%20%20%20%20pubmed%20%7B%0A%20%20%20%20%20%20rcsb_pubmed_container_identifiers%20%7B%0A%20%20%20%20%20%20%20%20pubmed_id%0A%20%20%20%20%20%20%7D%0A%20%20%20%20%20%20rcsb_pubmed_doi%0A%20%20%20%20%7D%0A%20%20%20%20software%20%7B%0A%20%20%20%20%20%20classification%0A%20%20%20%20%20%20name%0A%20%20%20%20%20%20version%0A%20%20%20%20%7D%0A%20%20%20%20rcsb_entry_container_identifiers%20%7B%0A%20%20%20%20%20%20entry_id%0A%20%20%20%20%7D%0A%20%20%7D%0A%7D%0A"
        print(urllib.parse.unquote(ex))
        """

        url = 'https://data.rcsb.org/graphql'
        header = {'Content-Type': 'application/json; charset=utf-8'}
        data = {"query": graphql_str}
        response = requests.post(url, json=data, headers=header)
        json_data = None
        if response.status_code == 200:
            json_data = json.loads(response.text)
        else:
            print("Fail to obtain results, status code: {}".format(response.status_code))

        return json_data

    def downloads(self, file_type="pdb", directory="./", overwrite=True):
        if len(self.pdb_ids) < 1:
            print("nothing to download.")
            return

        pool = Pool(processes=self.PARALLEL)
        entries = zip(self.pdb_ids, repeat(directory), repeat(file_type), repeat(overwrite))

        err = None
        try:
            pool.starmap(RCSB.download, entries)
            # res = pool.map(partial(RCSB.download, directory=directory, overwrite=overwrite), self.pdb_ids)
        except KeyboardInterrupt as e:
            print('Err: got ^C while pool mapping, terminating the pool')
            err = e
        except Exception as e:
            print('ERR: an exception has occured during download.')
            err = e

        if err:
            pool.terminate()
            raise err

    @staticmethod
    def download(pdb_id, directory, file_type='pdb', overwrite=False):
        if not os.path.exists(directory):
            os.makedirs(directory)

        if file_type not in ('pdb', 'cif'):
            print("can only download pdb or cif files.")
            return

        # retrieve pdbs and extract, pdb or cif
        url_template = "https://files.rcsb.org/download/{id}.{file_type}"

        if len(pdb_id) == 4:
            filename = os.path.join(directory, "{}.{}".format(pdb_id, file_type))
            if os.path.exists(filename) and not overwrite:
                print("skipping file already existed {}".format(filename))
                return

            url = url_template.format(id=pdb_id, file_type=file_type)

            try:
                r = requests.get(url)

                if r.status_code == 200:
                    with open(filename, "wb") as f:
                        # f.write(gzip.decompress(r.content))
                        for chunk in r.iter_content(chunk_size=128):
                            f.write(chunk)
                    print("write {}".format(filename))
                elif r.status_code == 404:
                    print("ERR: failed to download {} with {}, trying an alternative".format(pdb_id, url))
                    RCSB.download_bundles(pdb_id, directory)
            except Exception as err:
                raise err

    @staticmethod
    def download_bundles(pdb_id, wdir='./'):
        """url http://files.rcsb.org/pub/pdb/compatible/pdb_bundle/v6/4v63/4v63-pdb-bundle.tar.gz"""
        subdir = pdb_id[1:3]
        url = "http://files.rcsb.org/pub/pdb/compatible/pdb_bundle/{}/{}/{}-pdb-bundle.tar.gz".format(
            subdir.lower(), pdb_id.lower(), pdb_id.lower())

        try:
            response = requests.get(url, stream=True)

            if response.status_code == 200 and response.content:
                t = tarfile.open(fileobj=io.BytesIO(response.content), mode="r:gz")
                t.extractall(path=wdir)
                print("{} bundle files downloaded".format(pdb_id))
        except Exception as err:
            print("ERR: {} failed to download bundle file @ {}".format(pdb_id, url))
            raise err

    @staticmethod
    def pdb_current(obsolete=False):
        """get a list of current pdb ids, or obsolete entries"""
        url = "https://data.rcsb.org/rest/v1/holdings/current/entry_ids"
        tag = 'current'
        if obsolete:
            url = "https://data.rcsb.org/rest/v1/holdings/removed/entry_ids"
            tag = 'obsolete'

        try:
            header = {'Content-Type': 'application/json;charset=utf-8'}
            response = requests.get(url, headers=header)

            if response.status_code == 404:
                print("Err: Not found.")
                return

            data = response.json()
            print("Total # of {} pdb entries: {}.".format(tag, len(data)))
            return data
        except Exception as err:
            pass

    @staticmethod
    def pdb_replaced_by(pdb_id):
        """ probe rcsb repository holding service to get the replacement id for an obsolete pdb entries"""

        replacement_ids = list()
        try:
            header = {'Content-Type': 'application/json;charset=utf-8'}
            response = requests.get(
                "https://data.rcsb.org/rest/v1/holdings/removed/{}".format(pdb_id), headers=header)

            if response.status_code == 404:
                print("Err: Not found.")
                return

            data = response.json()
            replacement_ids = data['rcsb_repository_holdings_removed'].get('id_codes_replaced_by', [])
        except Exception as err:
            pass

        if not replacement_ids:
            print("failed to find a replacement for {}".format(pdb_id))
        else:
            print("{} was superceded by {}".format(pdb_id, replacement_ids))

        return replacement_ids

    def get_ids(self):
        print("PDB IDs: {}".format(self.pdb_ids))
        return self.pdb_ids

    def __str__(self):
        return "PDB ids: {}".format(",".join(self.pdb_ids))


class Parser:
    # default parser for graphql returned json dictionary, return a named tuple
    def __init__(self, name):
        self.name = name

    def parse(self, entry_dict):
        COVID19_PDB_stats = namedtuple(self.name, ['pdbcode',
                                                   'beamline',
                                                   'source',
                                                   'authors',
                                                   'title',
                                                   'deposit_date',
                                                   'release_date',
                                                   'collect_date',
                                                   'resolution',
                                                   'ligands',
                                                   'doi',
                                                   'pubmed',
                                                   'article'])
        pdbcode = entry_dict['rcsb_id'].lower()
        title = entry_dict['struct']['title']
        authors = concat_authors([author['name'] for author in entry_dict['audit_author']])
        deposit_date = to_date(entry_dict['rcsb_accession_info']['deposit_date'])
        release_date = to_date(entry_dict['rcsb_accession_info']['initial_release_date'])

        collect_date = None
        source = None
        beamline = None

        expt = entry_dict['exptl'][0]
        if 'electron' in expt['method'].lower():
            source = "CryoEM"
            beamline = source
        elif 'nmr' in expt['method'].lower():
            source = "NMR"
            beamline = source
        else:
            if entry_dict['diffrn_source']:
                sources = list()
                beamlines = list()
                collect_dates = list()
                for i, xsource in enumerate(entry_dict['diffrn_source']):

                    if xsource['pdbx_synchrotron_beamline'] and xsource['pdbx_synchrotron_site']:
                        beamlines.append(xsource['pdbx_synchrotron_site'] + " " + xsource['pdbx_synchrotron_beamline'])
                        sources.append(xsource['pdbx_synchrotron_site'])
                    else:
                        if xsource['source']:
                            beamlines.append(xsource['source'])
                            sources.append("Home source")

                if 'diffrn_detector' in entry_dict and entry_dict['diffrn_detector']:
                    for i, det in enumerate(entry_dict['diffrn_detector']):
                        if det['pdbx_collection_date']:
                            cdate = to_date(det['pdbx_collection_date'])
                            collect_dates.append(date_str(cdate))

                beamline = ",".join(beamlines)
                source = ",".join(sources)
                collect_date = ",".join(collect_dates)

        resolution = None
        if source != "NMR":
            resolution = ",".join(str(r) for r in
                                  entry_dict['rcsb_entry_info']['resolution_combined'])

        _pubmed = entry_dict.get('pubmed', None)
        if _pubmed and 'rcsb_pubmed_container_identifiers' in _pubmed:
            pubmed = entry_dict.get('pubmed').get(
                'rcsb_pubmed_container_identifiers').get('pubmed_id', None)
        else:
            pubmed = _pubmed

        full_article = None
        year = None
        doi = None
        if entry_dict['rcsb_primary_citation']:
            doi = entry_dict['rcsb_primary_citation']['pdbx_database_id_DOI']

            article_authors = concat_authors(
                entry_dict['rcsb_primary_citation']['rcsb_authors'])
            article_title = entry_dict['rcsb_primary_citation']['title']
            journal = entry_dict['rcsb_primary_citation']['rcsb_journal_abbrev']
            if "to be published" in journal.lower():
                full_article = "To be published."
            else:
                year = entry_dict['rcsb_primary_citation']['year']
                other_info = "({}){}-{}".format(
                    entry_dict['rcsb_primary_citation']['journal_volume'],
                    entry_dict['rcsb_primary_citation']['page_first'],
                    entry_dict['rcsb_primary_citation']['page_last'])
                full_article = "{} ({}) {} {}:{}".format(
                    article_authors, year, article_title, journal, other_info)

        # fill in missing values using NIH services
        if doi and not pubmed:
            _, pubmed, _ = xref_lookup(doi)
        elif pubmed and not doi:
            doi, _, _, = xref_lookup(pubmed)

        ligands = None
        # find ligands
        ligands, _ = Parser.parse_ligands(entry_dict)

        result = COVID19_PDB_stats(pdbcode=pdbcode,
                                   beamline=beamline,
                                   source=source,
                                   title=title,
                                   authors=authors,
                                   deposit_date=deposit_date,
                                   release_date=release_date,
                                   collect_date=collect_date,
                                   resolution=resolution,
                                   ligands=ligands,
                                   doi=doi_url(doi, html=True),
                                   pubmed=doi_url(pubmed, html=True),
                                   article=full_article)
        return result

    @staticmethod
    def parse_ligands(entry_dict):
        ligands = list()
        smiles = list()

        if entry_dict['nonpolymer_entities']:
            for item in entry_dict['nonpolymer_entities']:
                three_letter_name = item['nonpolymer_comp']['chem_comp']['three_letter_code']

                full_name = item['nonpolymer_comp']['chem_comp']['name']
                if not three_letter_name and full_name:
                    three_letter_name = full_name

                ligands.append(three_letter_name)

                if 'pdbx_chem_comp_descriptor' in item:
                    for desc in item['pdbx_chem_comp_descriptor']:
                        if desc['type'] == 'SMILES' and desc['descriptor']:
                            smiles.append(desc['descriptor'])
                            break

        return ",".join(ligands), ",".join(smiles)


class Report(RCSB):
    def __init__(self,
                 directory=".",
                 start_date=None,
                 end_date=None,
                 xls_filename="report.xlsx",
                 user_pdbcodes=None,
                 search_text=None,
                 report_type=None,
                 download_type="pdb",
                 download_overwrite=True,
                 query_dict=None,
                 graph_ql=None,
                 json_parser=None,
                 email_addresses=["qxu@anl.gov", "qpxu007@gmail.com"]
                 ):

        RCSB.__init__(self)
        self.directory = directory

        self.xls_filename = os.path.join(self.directory, xls_filename)
        self.xls_name = xls_filename
        self.user_pdbcodes = user_pdbcodes

        self.rows = list()
        self.send_to = email_addresses
        self.parser = json_parser
        self.query_dict = query_dict
        self.search_text = search_text
        self.report_type = report_type
        self.download_filetype = download_type
        self.download_overwrite = download_overwrite
        self.json_data = None

        self.start_date = start_date
        self.end_date = end_date

        self.filtered_pdb_ids = list()

        if not self.parser:
            self.parser = Parser(name="report")

        self.graph_ql = graph_ql
        if not self.graph_ql:
            self.graph_ql = '''{
              entries(entry_ids: %s)
              {
                rcsb_id
                audit_author {
                  name
                }
    
                struct {
                  title
                }
    
                exptl {
                  method
                }
    
                rcsb_entry_info {
                  resolution_combined
                }
    
                rcsb_accession_info {
                  deposit_date
                  initial_release_date
                }
    
                diffrn_source {
                  pdbx_wavelength_list
                  pdbx_synchrotron_beamline
                  pdbx_synchrotron_site
                  source
                  type
                }
    
                diffrn_detector {
                  pdbx_collection_date
                  type
                }
    
                diffrn {
                  ambient_temp
                }
    
                rcsb_primary_citation {
                  journal_volume
                  page_first
                  page_last
                  pdbx_database_id_DOI
                  rcsb_authors
                  rcsb_journal_abbrev
                  title
                  year
                }
    
                pubmed {
                  rcsb_pubmed_container_identifiers {
                    pubmed_id
                  }
                  rcsb_pubmed_doi
                }
    
                nonpolymer_entities {
                  nonpolymer_comp {
                    chem_comp {
                      name
                      three_letter_code
                      type
                    }
    
                    pdbx_chem_comp_descriptor {
                      descriptor
                      type
                    }
                  }
                }
    
              }
            } '''

    def save_data(self):
        if len(self.rows) < 1:
            print("ERR: no records found.")
            return

        if not os.path.exists(self.directory):
            os.makedirs(self.directory)

        fields = self.rows[0]._fields
        df = pd.DataFrame.from_records(self.rows, columns=fields)

        writer = pd.ExcelWriter(self.xls_filename, engine='xlsxwriter')
        df.to_excel(writer, sheet_name='Sheet1', index=False)

        workbook = writer.book
        url_format = workbook.get_default_url_format()
        # url_format = workbook.add_format({'bold': False, 'font_color': 'blue'})

        # set column width
        worksheet = writer.sheets["Sheet1"]
        for idx, col in enumerate(df):  # loop through all columns
            series = df[col]
            max_len = max((
                series.astype(str).map(len).max(),  # len of largest item
                len(str(series.name)),  # len of column name/header
            )) + 1  # adding a little extra space
            max_len = 50 if max_len > 50 else max_len
            worksheet.set_column(idx, idx, max_len)  # set column width

            if series.name in ("doi", "pubmed"):
                worksheet.set_column(idx, idx, cell_format=url_format)

        writer.save()
        print("save result to {}".format(self.xls_filename))

    def send_mail(self,
                  subject="pdb report",
                  text="pdb report",
                  send_from="qxu@anl.gov",
                  files=None,
                  server="127.0.0.1") -> object:
        """https://stackoverflow.com/questions/3362600/how-to-send-email-attachments"""

        # PDB report
        if not self.send_to:
            print("no email addresses given, emails not sent.")
            return

        USER = os.environ.get("USER")
        HOST = socket.gethostname()

        msg = MIMEMultipart()
        msg['Subject'] = subject
        msg['From'] = send_from
        msg['To'] = ', '.join(self.send_to)

        msg.attach(MIMEText(text))

        if os.path.exists(self.xls_filename):
            part = MIMEBase('application', "octet-stream")
            part.set_payload(open(self.xls_filename, "rb").read())
            encoders.encode_base64(part)

            part.add_header('Content-Disposition', 'attachment; filename="{}"'.format(self.xls_name))

            msg.attach(part)

        server = smtplib.SMTP(server)
        server.sendmail("{}@{}".format(USER, HOST), self.send_to, msg.as_string())
        print("send email to {}".format(self.send_to))
        server.close()

    def verify(self, response):
        pdb_ins = set(id.lower() for id in self.pdb_ids)
        pdb_outs = set([d['rcsb_id'].lower() for d in response["data"]["entries"]])

        # print(pdb_ins, pdb_outs)

        replacement_ids = list()

        if pdb_ins != pdb_outs:
            set_diff = pdb_ins - pdb_outs
            print("Warning: # of queries = {}, # returned results = {}, missing {}".format(
                len(pdb_ins),
                len(pdb_outs),
                set_diff
            ), end='')
            print(", these entries may be obsolete.")

            for id in set_diff:
                new_replacement_ids = RCSB.pdb_replaced_by(id)
                replacement_ids.extend(new_replacement_ids)
                self.pdb_ids.extend(new_replacement_ids)
                self.pdb_ids.remove(id)

            return replacement_ids

    def report(self):
        if self.query_dict:
            self.search(self.query_dict)
        elif self.search_text:
            self.search_by_text(self.search_text)

        if self.user_pdbcodes:
            for pdb in self.user_pdbcodes:
                if pdb not in self.pdb_ids:
                    self.pdb_ids.append(pdb)

        if not self.pdb_ids:
            print("nothing was found.")
            return

        if self.graph_ql and self.report_type:

            self.json_data = self.query_graphql(self.graph_ql % json.dumps(self.pdb_ids))

            # handle obsolete entries
            replacement_ids = self.verify(self.json_data)
            if replacement_ids:
                new_data = self.query_graphql(self.graph_ql % json.dumps(replacement_ids))
                self.json_data["data"]["entries"].extend(new_data["data"]["entries"])

            # parse and filter data
            print("filter by date, reject: ", end='')
            for entry in self.json_data["data"]["entries"]:
                row = self.parser.parse(entry)
                if 'release_date' in row._fields and self.end_date >= row.release_date >= self.start_date:
                    self.rows.append(row)
                    if 'pdbcode' in row._fields:
                        self.filtered_pdb_ids.append(row.pdbcode)
                else:
                    print(" {}".format(row.pdbcode), end='')

            print("\n")

            if not self.filtered_pdb_ids:
                print("No entries within time frame specified.")
                return

            if self.rows:
                if len(self.filtered_pdb_ids) != len(self.pdb_ids):
                    print("{} filtered entries by date: {}".format(len(self.filtered_pdb_ids),
                                                                   " ".join(self.filtered_pdb_ids)))
                    self.pdb_ids = self.filtered_pdb_ids

                self.save_data()

                if self.send_to:
                    subject = "{} PDB release update".format(self.report_type)
                    email_text = "Date From {} to {}, {} of structures released.".format(self.start_date,
                                                                                         self.end_date,
                                                                                         len(self.rows))
                    self.send_mail(text=email_text, subject=subject)

        if self.download_filetype:
            dir = os.path.join(self.directory, self.download_filetype)
            if not os.path.exists(dir):
                os.makedirs(dir)

            self.downloads(directory=dir,
                           file_type=self.download_filetype,
                           overwrite=self.download_overwrite)


class Presets:
    covid19_query_dict = {'query': {'label': 'query-builder',
                                    'logical_operator': 'and',
                                    'nodes': [{'label': 'text',
                                               'logical_operator': 'and',
                                               'nodes': [{'logical_operator': 'or',
                                                          'nodes': [{'node_id': 0,
                                                                     'parameters': {'negation': False,
                                                                                    'value': '"SARS-COV-2"'},
                                                                     'service': 'text',
                                                                     'type': 'terminal'},
                                                                    {'node_id': 1,
                                                                     'parameters': {'negation': False,
                                                                                    'value': '"COVID-19"'},
                                                                     'service': 'text',
                                                                     'type': 'terminal'},
                                                                    {"type": "terminal",
                                                                     "service": "text",
                                                                     "parameters": {
                                                                         "operator": "exact_match",
                                                                         "negation": False,
                                                                         "value": "Severe acute respiratory syndrome coronavirus 2",
                                                                         "attribute": "rcsb_entity_source_organism.ncbi_scientific_name"},
                                                                     "node_id": 2},
                                                                    ],
                                                          'type': 'group'}],
                                               'type': 'group'}],
                                    'type': 'group'},
                          'request_info': {'query_id': '1f974c36887d0686f773b45b31925c0f', 'src': 'ui'},
                          'request_options': {
                              "return_all_hits": True,
                              'scoring_strategy': 'combined',
                              'sort': [{'direction': 'desc',
                                        'sort_by': 'rcsb_accession_info.initial_release_date'}]},
                          'return_type': 'entry'}

    @staticmethod
    def get_aps_query(start_date, end_date, APS=False):
        condition_date_range = {'logical_operator': 'and',
                                'nodes': [{'node_id': 0,
                                           'parameters': {
                                               'attribute': 'rcsb_accession_info.initial_release_date',
                                               'negation': False,
                                               'operator': 'range_closed',
                                               'value': [iso_date_str(start_date, start=True),
                                                         iso_date_str(end_date, start=False)]},
                                           'service': 'text',
                                           'type': 'terminal'}],
                                'type': 'group'}

        condition_site = {'logical_operator': 'and',
                          'nodes': [{'node_id': 1,
                                     'parameters': {
                                         'attribute': 'diffrn_source.pdbx_synchrotron_site',
                                         'negation': False,
                                         'operator': 'exact_match',
                                         'value': 'APS'},
                                     'service': 'text',
                                     'type': 'terminal'}],
                          'type': 'group'}

        condition_beamlines = {'logical_operator': 'or',
                               'nodes': [{'node_id': 2,
                                          'parameters': {
                                              'attribute': 'diffrn_source.pdbx_synchrotron_beamline',
                                              'negation': False,
                                              'operator': 'exact_match',
                                              'value': '23-ID-B'},
                                          'service': 'text',
                                          'type': 'terminal'},
                                         {'node_id': 3,
                                          'parameters': {
                                              'attribute': 'diffrn_source.pdbx_synchrotron_beamline',
                                              'negation': False,
                                              'operator': 'exact_match',
                                              'value': '23-ID-D'},
                                          'service': 'text',
                                          'type': 'terminal'}],
                               'type': 'group'}

        request_options = {"return_all_hits": True,
                           'scoring_strategy': 'combined',
                           'sort': [{'direction': 'desc',
                                     'sort_by': 'rcsb_accession_info.initial_release_date'}]}

        conditions = [condition_date_range, condition_site, condition_beamlines]
        if APS:
            conditions = [condition_date_range, condition_site]

        query_APS_GMCA_date_range = {'query': {'label': 'query-builder',
                                               'logical_operator': 'and',
                                               'nodes': [{'label': 'text',
                                                          'logical_operator': 'and',
                                                          'nodes': conditions,
                                                          'type': 'group'}],
                                               'type': 'group'},
                                     'request_options': request_options,
                                     'return_type': 'entry'}
        return query_APS_GMCA_date_range

    aps_graphql_statement = '''{
      entries(entry_ids: %s)
      {
        rcsb_id
        audit_author {
          name
        }

        struct {
          title
        }

        rcsb_entry_info {
          resolution_combined
        }

        rcsb_accession_info {
          deposit_date
          initial_release_date
        }

        diffrn_source {
          pdbx_wavelength_list
          pdbx_synchrotron_beamline
          pdbx_synchrotron_site
          source
          type
        }

        diffrn_detector {
          pdbx_collection_date
          type
        }

        diffrn {
          ambient_temp
        }

        reflns {
          d_resolution_high
          d_resolution_low
          number_all
          number_obs
          pdbx_Rmerge_I_obs
          pdbx_Rsym_value
          pdbx_netI_over_sigmaI
          pdbx_redundancy
          percent_possible_obs
          phase_calculation_details
          pdbx_CC_half
        }

        reflns_shell {
          Rmerge_I_obs
          d_res_high
          d_res_low
          percent_possible_obs
          meanI_over_sigI_obs
          pdbx_CC_half
        }

        refine {
          ls_R_factor_R_free
          ls_R_factor_R_work
          ls_d_res_high
          ls_d_res_low
          pdbx_starting_model
        }

        rcsb_primary_citation {
          journal_volume
          page_first
          page_last
          pdbx_database_id_DOI
          rcsb_authors
          rcsb_journal_abbrev
          title
          year
        }

        pubmed {
          rcsb_pubmed_container_identifiers {
            pubmed_id
          }
          rcsb_pubmed_doi
        }

        software {
          classification
          name
          version
        }

      }
    } '''


class APS_Parser(Parser):
    def __init__(self, name="report", **kwargs):
        Parser.__init__(self, name=name)

    def parse(self, entry_dict):
        GMCA_PDB_stats = namedtuple("PDB_stats", ['pdbcode',
                                                  'title',
                                                  'authors',
                                                  'deposit_date',
                                                  'release_date',
                                                  'collect_date',
                                                  'detector',
                                                  'beamline',
                                                  'wavelength',
                                                  'resolution',
                                                  'rmerge',
                                                  'cchalf',
                                                  'isigmai',
                                                  'redundancy',
                                                  'completeness',
                                                  'rmerge_shell',
                                                  'cchalf_shell',
                                                  'rfree',
                                                  'rwork',
                                                  'data_reduction',
                                                  'data_scaling',
                                                  'doi',
                                                  'pubmed',
                                                  'article_year',
                                                  'article'])

        pdbcode = entry_dict['rcsb_id'].lower()
        title = entry_dict['struct']['title']
        authors = concat_authors([author['name'] for author in entry_dict['audit_author']])
        deposit_date = to_date(entry_dict['rcsb_accession_info']['deposit_date'])
        release_date = to_date(entry_dict['rcsb_accession_info']['initial_release_date'])

        beamlines = list()
        wavelengths = list()
        detectors = list()
        collect_dates = list()
        for i, xsource in enumerate(entry_dict['diffrn_source']):
            if xsource['pdbx_synchrotron_beamline']:
                beamlines.append(xsource['pdbx_synchrotron_beamline'])
            elif xsource['source']:
                beamlines.append(xsource['source'])

            if xsource['pdbx_wavelength_list']:
                wavelengths.append(xsource['pdbx_wavelength_list'])

        if 'diffrn_detector' in entry_dict and entry_dict['diffrn_detector']:
            for i, det in enumerate(entry_dict['diffrn_detector']):
                if det['type']:
                    detectors.append(det['type'])
                if det['pdbx_collection_date']:
                    cdate = to_date(det['pdbx_collection_date'])
                    collect_dates.append(date_str(cdate))

        resolution = ",".join(str(r) for r in entry_dict['rcsb_entry_info']['resolution_combined'])

        reflns = entry_dict['reflns'][0]
        rmerge = reflns['pdbx_Rmerge_I_obs']
        rsym = reflns['pdbx_Rsym_value']

        cchalf = reflns['pdbx_CC_half']
        isigmai = reflns['pdbx_netI_over_sigmaI']
        redundancy = reflns['pdbx_redundancy']
        completeness = reflns['percent_possible_obs']

        rmerge_shell = None
        cchalf_shell = None
        if entry_dict['reflns_shell']:
            rmerge_shell = entry_dict['reflns_shell'][0]['Rmerge_I_obs']
            cchalf_shell = entry_dict['reflns_shell'][0]['pdbx_CC_half']

        rwork = entry_dict['refine'][0]['ls_R_factor_R_work']
        rfree = entry_dict['refine'][0]['ls_R_factor_R_free']

        data_reduction = None
        data_scaling = None
        for cx in entry_dict['software']:
            if 'reduction' in cx['classification']:
                data_reduction = cx['name']
            elif 'scaling' in cx['classification']:
                data_scaling = cx['name']

        _pubmed = entry_dict.get('pubmed', None)
        if _pubmed and 'rcsb_pubmed_container_identifiers' in _pubmed:
            pubmed = entry_dict.get('pubmed').get(
                'rcsb_pubmed_container_identifiers').get('pubmed_id', None)
        else:
            pubmed = _pubmed

        full_article = None
        year = None
        doi = None
        if entry_dict['rcsb_primary_citation']:
            doi = entry_dict['rcsb_primary_citation']['pdbx_database_id_DOI']

            article_authors = concat_authors(
                entry_dict['rcsb_primary_citation']['rcsb_authors'])
            article_title = entry_dict['rcsb_primary_citation']['title']
            journal = entry_dict['rcsb_primary_citation']['rcsb_journal_abbrev']
            if "to be published" in journal.lower():
                full_article = "To be published."
            else:
                year = entry_dict['rcsb_primary_citation']['year']
                other_info = "({}){}-{}".format(
                    entry_dict['rcsb_primary_citation']['journal_volume'],
                    entry_dict['rcsb_primary_citation']['page_first'],
                    entry_dict['rcsb_primary_citation']['page_last'])
                full_article = "{} ({}) {} {}:{}".format(
                    article_authors, year, article_title, journal, other_info)

        # fill in missing values using NIH services
        if doi and not pubmed:
            _, pubmed, _ = xref_lookup(doi)
        elif pubmed and not doi:
            doi, _, _, = xref_lookup(pubmed)

        return GMCA_PDB_stats(pdbcode=pdbcode,
                              title=title,
                              authors=authors,
                              deposit_date=deposit_date,
                              release_date=release_date,
                              collect_date=",".join(collect_dates) if collect_dates else None,
                              detector=",".join(detectors) if detectors else None,
                              beamline=",".join(beamlines) if beamlines else None,
                              wavelength=",".join(wavelengths) if wavelengths else None,
                              resolution=resolution,
                              rmerge=rmerge if rmerge else rsym,
                              cchalf=cchalf,
                              isigmai=isigmai,
                              redundancy=redundancy,
                              completeness=completeness,
                              rmerge_shell=rmerge_shell,
                              cchalf_shell=cchalf_shell,
                              rfree=rfree,
                              rwork=rwork,
                              data_reduction=data_reduction,
                              data_scaling=data_scaling,
                              doi=doi_url(doi, html=True),
                              pubmed=doi_url(pubmed, html=True),
                              article_year=year,
                              article=full_article)


def load_pdb_ids_from_file(filename):
    if not os.path.exists(filename):
        print("Err: file {} does not exist.".format(filename))
        return

    with open(filename, 'r') as fh:
        lines = fh.readlines()
        return [line.strip() for line in lines]


def get_date_range(start_date=None, end_date=None, days=-1):
    today_date = date.today()
    if start_date and end_date:
        start_date = start_date
        end_date = end_date
    elif start_date and not end_date:
        if days <= 0:
            end_date = today_date
        else:
            end_date = end_date + timedelta(days=days)
    elif end_date and not start_date:
        if days <= 0:
            start_date = date(1970, 1, 1)
        else:
            start_date = end_date - timedelta(days=days)
    else:
        end_date = today_date
        if days <= 0:
            start_date = date(1970, 1, 1)
        else:
            start_date = end_date - timedelta(days=days)

    return start_date, end_date


def run(args):
    if args.load:
        args.pdb_codes = load_pdb_ids_from_file(args.load)
        print("{} entries loaded from file: {}".format(len(args.pdb_codes), ",".join(args.pdb_codes)))

    today_date = date.today()
    TODAY = today_date.strftime("%m-%d-%Y")

    query = None
    graphql = None
    xls_filename = "PDBstats-{}.xlsx".format(TODAY)
    parser = None

    start_date, end_date = get_date_range(args.start_date, args.end_date, args.days)

    if args.report and 'aps' in args.report:
        graphql = Presets.aps_graphql_statement
        parser = APS_Parser(name="APS")
        query = Presets.get_aps_query(start_date, end_date=end_date, APS=True)
        xls_filename = "APSstats-{}.xlsx".format(TODAY)

    if args.report and 'gmca' in args.report:
        graphql = Presets.aps_graphql_statement
        parser = APS_Parser(name="GMCA")
        query = Presets.get_aps_query(start_date, end_date=end_date, APS=False)
        xls_filename = "GMCAstats-{}.xlsx".format(TODAY)

    if args.report and 'covid' in args.report:
        query = Presets.covid19_query_dict
        xls_filename = "COVID19stats-{}.xlsx".format(TODAY)

    r = Report(directory=args.directory,
               start_date=start_date,
               end_date=end_date,
               xls_filename=xls_filename,
               user_pdbcodes=args.pdb_codes,
               search_text=args.search,
               download_type=args.download,
               download_overwrite=not args.skip_existing_files,
               email_addresses=args.email_to,
               report_type=args.report,
               graph_ql=graphql,
               json_parser=parser,
               query_dict=query)

    r.report()


def valid_date(str):
    try:
        return datetime.strptime(str, "%Y-%m-%d").date()
    except ValueError:
        msg = "Not a valid date: '{0}'.".format(str)
        raise argparse.ArgumentTypeError(msg)


def get_args():
    arg_parser = argparse.ArgumentParser(description='RCSB report and download tool', add_help=True)

    g1 = arg_parser.add_argument_group("search")
    g1.add_argument('--search', '-f', type=str, help='define a string to search rcsb with')

    g2 = arg_parser.add_argument_group("report")
    g2.add_argument('--report', '-r', default="generic", choices=["gmca", "aps", "covid19", "generic"], type=str.lower,
                    help='generate xls report')
    g2.add_argument('--directory', '--dir', default=os.getcwd(), type=str, help='directory to store output file(s)')
    g2.add_argument('--start_date', default=None, type=valid_date, help='define a starting date YYYY-mm-dd')
    g2.add_argument('--end_date', default=None, type=valid_date, help='define an end date')
    g2.add_argument('--days', '-d', default=-1, type=int,
                    help='help define date range if either start or end date is given, '
                         'default -1: no range')
    g2.add_argument('--email-to', '-m', action='append', default=None, help='email address to send excel report to')

    g3 = arg_parser.add_argument_group("download")
    g3.add_argument('--download', '--save', default=None, choices=["pdb", "cif"], type=str.lower,
                    help='download pdb given ids or a search term')
    g3.add_argument('--skip_existing_files', '--skip', action='store_true',
                    help='skip files already existed in directory, do not overwrite')

    arg_parser.add_argument('pdb_codes', type=str, nargs='*', help='pdb codes')
    arg_parser.add_argument('--load', default=None, type=str, help='load pdb ids from a text file')

    return arg_parser.parse_intermixed_args()


if __name__ == '__main__':
    args = get_args()
    # print(args)
    # sys.exit(0)
    run(args)
    sys.exit(0)
