# Data processing using XDS and MPI


## Install OpenMPI and mpi4py

As root, on bl3ws1, bl3ws2, bl3ws3 install mpi and python binding (and hdf5), as well as dependencies:

```
yum install openmpi openmpi-devel mpi4py-openmpi mpi4py-common hdf5-openmpi hdf5-openmpi-devel
```

mpi environment is loaded by environment-modules (installed as a dependency) in CentOS:

```
module load  mpi/openmpi-x86_64
```

the above command can be put in user start up shell script such as .bashrc

MPI requires that mpi needs to access nodes using passwordless ssh login, and files shared by a common file system. These requirements are usually satisfied if the user's home directory are on a shared file system, such as NFS.


### XDS tweak

XDS allows user to parallelize COLSPOT and INTEGRATE steps using their own scripts (as long as the same name is used and common command interface is maintained). In this case, the two scripts forkcolspot and forkintegrate is simplied as the following (so it can soft linked to each other)

```
#!/bin/bash

module load mpi/openmpi-x86_64

mpirun -np 8 --host bl3ws1,bl3ws2,bl3ws3 python /mnt/share/qxu/mpi-xds/mpisubmit.py $*

```

the machines to run mpi and locations of scripts needs to be editted to match the setup and their locations. The python wrapper mpisubmit.py handles running mintegrate and mcolspot using mpi4py.


### Test results

Using a test dataset of 270 pilatus 6M images, using the three bl3ws* machines with 8 cpus, we tried parallelization of xds (JOBS=ALL):

```
NJOBS  time (s)
1 223
2 153
3 121
4 108
5 99
6 89
7 85
8 86
10 86
20 88
```

For comparison, processing using dual cpu on a single machine took 138 sec. To evaluate the speed up without complication of serials steps in xds, we tried only the INTEGRATE step:

```
NJOBS  time (s)
1 184
2 112
3 83
4 71
5 60
6 51
7 46
8 48
10 47
20 47
```
Integration on a dual cpu on a single machine took 106 sec.


### Use MPI under Grid engine

SGE supports MPI, but we need to set up a parallel environment for mpi





## mpisubmit.py script

```
#!/usr/bin/env python

"""
For basic applications, MPI is as easy to use as any other
message-passing system. The sample code below contains the complete
communications skeleton for a dynamically load balanced master/slave
application. Following the code is a description of the few functions
necessary to write typical parallel applications.

important parameters
--------------------

status = MPI.Status()               # where all info is stored

# Receive results from a slave
result = comm.recv(                 # message buffer
         source=MPI.ANY_SOURCE,     # receive from any sender (-1)
         tag=MPI.ANY_TAG,           # any type of message (-1)
         status=status)             # info about the received msg (class)

# Send the slave a new work unit
comm.send(work,                     # message buffer
         dest=status.Get_source(),  # to whom we just received from
         tag=WORKTAG)               # user chosen message tag

# Original author: Fernandoo Paolo <fpaolo@ucsd.edu> Jan 15, 2013
# Adapted for GMCA usage by Qingping Xu <qxu@anl.gov> Oct 12, 2016
"""

import os
import sys
import numpy as np
from mpi4py import MPI
from Queue import Queue

WORKTAG = 1
DIETAG = 0

class Work(object):
    def __init__(self, prog, params):
        q = Queue()
        for param in params:
            job = "echo "+param + "|" + prog
            q.put(job)
            #q.put(' '.join([prog, param]))
        self.work = q

    def get_next(self):
        if self.work.empty():
            return None
        return self.work.get()

    def qsize(self):
        return self.work.qsize()


def do_work(work):
    if '.py' in work:
        os.system('python ' + work)
    else:
        os.system(work)  # for './'
    return

def process_result(result):
    pass


def get_params():
    # generate work queue
    # print "len=", len(sys.argv), sys.argv
    if len(sys.argv) == 3:
        program = 'mcolspot_par'
        ntask = int(sys.argv[1])
        maxcpu = int(sys.argv[2])
        params = map(str, range(1, ntask+1))
    elif len(sys.argv) == 6:
        program = 'mintegrate_par'
        fframe = int(sys.argv[1])
        ni = int(sys.argv[2])
        ntask = int(sys.argv[3])
        niba0 = int(sys.argv[4])
        maxcpu = int(sys.argv[5])
        r1 = integ_range(fframe, ni, ntask, niba0, maxcpu)
        r2 = integ_range2(fframe, ni, ntask, niba0, maxcpu)
        assert(r1 == r2)
        params = integ_range(fframe, ni, ntask, niba0, maxcpu)
    else:
        print "Error: unexpected number of argument passed into the program: {}".format(sys.argv)
        exit(1)

    return program, params

def integ_range(fframe, ni, ntask, niba0, maxcpu):
    # fframe id number of the first image
    # ni     #number of images in the data set
    # niba0  #minimum number of images in a batch
    # maxcpu #maximum number of processors used by each job
    # print "integ_range parameters: {} {} {} {} {}".format(fframe, ni, ntask, niba0, maxcpu)
    # translated from xds forkintegrate script
    minitask = ni // ntask  # minimum number of images in a job
    mtask = ni % ntask      # number of jobs with minitask+1 images
    litask, itask = 0, 1

    params = []
    while itask <= ntask:
        if itask > mtask:
            nitask = minitask
        else:
            nitask = minitask + 1

        fitask = litask + 1
        litask = litask + nitask

        if nitask < niba0:
            n = nitask
        else:
            n = niba0

        n = min(nitask, niba0)
        if n < 1:
            n = 1

        nbatask = nitask // n
        image1 = fframe + fitask - 1

        params.append("{image1} {nitask} {itask} {nbatask}".format(image1=image1, nitask=nitask, itask=itask, nbatask=nbatask))

        itask += 1

    return params

def integ_range2(fframe, ni, ntask, niba0, maxcpu):
    # version 2, equivalent to abve
    minitask = ni // ntask
    mtask = ni % ntask
    itasks = [ i + 1 for i in range(ntask)]
    nitasks = [ minitask if i > mtask else minitask + 1 for i in itasks ]
    nbatasks = [ nitask // n for n in [ max(1, min(nitask, niba0)) for nitask in nitasks ] ]
    image1 = [ fframe + sum(nitasks[:i]) for i in range(ntask) ]
    return [ "{} {} {} {}".format(a, b, c, d) for a, b, c, d in zip(image1, nitasks, itasks, nbatasks) ]

def master(comm):
    num_procs = comm.Get_size()
    status = MPI.Status()

    program, params = get_params()
    wq = Work(program, params)

    # Seed the slaves, send one unit of work to each slave (rank)
    nworkers = min(num_procs, wq.qsize() + 1)
    for rank in xrange(1, nworkers):
        work = wq.get_next()
        if work:
            comm.send(work, dest=rank, tag=WORKTAG)

    # Loop over getting new work requests until there is no more work to be done
    while True:
        work = wq.get_next()
        if not work: break

        # Receive results from a slave
        result = comm.recv(source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG, status=status)
        #process_result(result)

        # Send the slave a new work unit
        comm.send(work, dest=status.Get_source(), tag=WORKTAG)

    # No more work to be done, receive all outstanding results from slaves
    for rank in xrange(1, nworkers):
        result = comm.recv(source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG, status=status)
        #process_result(result)

    # Tell all the slaves to exit by sending an empty message with DIETAG
    for rank in xrange(1, num_procs):
        comm.send(0, dest=rank, tag=DIETAG)

def slave(comm):
    my_rank = comm.Get_rank()
    status = MPI.Status()
    #print "I am slave {} {}".format(my_rank, status)

    while True:
        # Receive a message from the master
        work = comm.recv(source=0, tag=MPI.ANY_TAG, status=status)

        # Check the tag of the received message
        if status.Get_tag() == DIETAG: break

        # Do the work
        result = do_work(work)

        # Send the result back
        comm.send(result, dest=0, tag=0)


def main():
    comm = MPI.COMM_WORLD
    my_rank = comm.Get_rank()
    my_name = MPI.Get_processor_name()
    #comm.Barrier()
    #start = MPI.Wtime()

    if my_rank == 0:
        master(comm)
    else:
        slave(comm)

    #comm.Barrier()
    #end = MPI.Wtime()
    #print 'time:', end - start

if __name__ == '__main__':
    main()

```
