#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""

data collection simulation by copying images files from one directory
from one directory to another directory at an specified interval

@author: qxu
"""
from __future__ import absolute_import
import argparse
import logging
import os
import shutil
import sys
import time
from multiprocessing import Process

from quickProcess.tasks.scriptjob import ScriptJob
from quickProcess.common.generatesweeps import GenerateSweeps

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger(__name__)


def get_args():
    """ command line options """
    parser = argparse.ArgumentParser(prog='')
    parser.add_argument('--data_dir', '-i',
                        required=False, help='directory for image files')
    parser.add_argument('--dest_dir', '-d', '-o',
                        default=None, help='collect directory')
    parser.add_argument('--sleep', '--wait', type=float,
                        default=1, help='sleep between images')
    parser.add_argument('--start', type=int,
                        default=1, help='start image number')
    parser.add_argument('--end', type=int,
                        default=None, help='last image number to be processed')
    return parser.parse_args()


class CollectSimulator():
    def __init__(self, indir, outdir, interval):
        self.dest_dir = outdir
        self.indir = indir
        if not os.path.isdir(outdir):
            os.makedirs(outdir)

        self.interval = interval
        self.gen_sweeps = GenerateSweeps(indir)
        self.image_sweeps = self.gen_sweeps.get_image_sweeps()
        self.file_sweeps = self.gen_sweeps.get_file_sweeps()
        self.current_frame = None

    def collect(self):
        counter = 0
        for s in self.image_sweeps:
            for image in s.get_all_files():
                time.sleep(self.interval)
                counter += 1
                self.current_frame = image
                # self.queue.put(image)
                shutil.copy2(image, self.dest_dir)
                if counter % 20 == 0:
                    logger.info("collecting {} {}".format(image, counter))
        # self.queue.put(None)
        return

    def get_currentframe(self):
        """ return current frame being collected"""
        return self.current_frame

    def get_filelist(self):
        """ list of files to be collected"""
        with open(os.path.join(self.dest_dir, "filelist"), "w") as file:
            file.writelines("\n".join(self.gen_sweeps.file_list))

        return self.gen_sweeps.file_list


def collect_data(args):
    collect = CollectSimulator(args.data_dir, args.dest_dir, args.sleep)
    collect.get_filelist()
    collect.collect()


def process_data(args, start, end):
    subdir = "process-{:05d}-{:05d}".format(start, end)
    wdir = os.path.join(args.dest_dir, subdir)
    if not os.path.isdir(wdir):
        os.makedirs(wdir)

    job = ScriptJob(wdir, script_name='{}.sh'.format(subdir))
    job.update_cmd("module load anaconda/python3\n"
                   "export PYTHONPATH=/home/qxu/work/quickProcess:$PYTHONPATH\n"
                   "python -m quickProcess.process --pipeline=gmcaproc"
                   " --data_dir={} --proc_dir={}"
                   " --filelist={} --start={} --end={}"
                   " --log2db --nproc=1 --njobs=1"
                   .format(args.dest_dir,
                           wdir,
                           args.filelist,
                           start,
                           end))
    # job.write_script()
    job.run_sync()


def main():
    options = get_args()
    if not options.data_dir:
        options.data_dir = "/home/qxu/data-for-tests/jcsg-p3-pilatus-6m/SP16764A_4ghn"
    if not options.dest_dir:
        options.dest_dir = "/tmp/qxu/collect1"
    if not os.path.isdir(options.dest_dir):
        os.makedirs(options.dest_dir)

    cl = Process(target=collect_data, args=(options,))
    cl.start()

    options.filelist = os.path.join(options.dest_dir, "filelist")

    nframes = 280
    checkpoints = [0.25, 0.5, 1]
    jobs = []

    for point in checkpoints:
        end = int(nframes * point)
        job = Process(target=process_data, args=(options, 1, end))
        jobs.append(job)

    for job in jobs:
        job.start()
        time.sleep(5)  # NB log to db issue if multiple jobs create db at the same time

    cl.join()
    for job in jobs:
        job.join()


if __name__ == '__main__':
    main()
