from __future__ import absolute_import

import sys, os, re, string
import numpy as np
import fabio
from .smx_get_args import SMX_ARGS


class Geometry:
    common_block = \
"""photon_energy = {energy} ; in eV
adu_per_photon = 1.0
clen = {distance}  ; in m
coffset = 0.
res = {resolution}  ; in pixels

{disable}data = {data_label} ; /entry_1/data_1/data
{disable}dim0 = %
{disable}dim1 = ss
{disable}dim2 = fs

0/min_fs = 0
0/max_fs = {nx}
0/min_ss = 0
0/max_ss = {ny}
0/corner_x = -{xbeam} 
0/corner_y = -{ybeam}
0/fs = x
0/ss = y
"""

    # generate a mask from an example image of pilatus or eiger cbf
    # default data_label for cxi is: data = /entry_1/data_1/data\ndim0 = %\ndim1 = ss\ndim2 = fs\n
    #                                for cbf should be empty string: ""
    def __init__(self,
                 imagefile,
                 data_label="/entry_1/data_1/data",
                 geom_file='current_geometry.geom',
                 h5geom_file=None,
                 badpixelmap=None
                 ):
        self.image_file = imagefile
        self.data_label = data_label  # data_label is needed if input from cxi
        self.geometry_file = geom_file
        self.geometry_h5 = h5geom_file
        self.badpixelmap = badpixelmap
        self.image = None
        self.geometry_text = None

    def generate(self, disable_data_label=False):
        self.image = self.get_image()
        data = self.image['data']
        beamx, beamy = self.image['beam']

        if SMX_ARGS.beamx:
            print("use user defined beamx = {}".format(SMX_ARGS.beamx))
            beamx = SMX_ARGS.beamx

        if SMX_ARGS.beamy:
            print("use user defined beamy = {}".format(SMX_ARGS.beamy))
            beamy = SMX_ARGS.beamy

        common_block = self.generate_common_blocks(disable_data_label=disable_data_label).split("\n")
        vertial_gaps = Geometry.inter_module_gaps(data, axis=0)
        horizontal_gaps = Geometry.inter_module_gaps(data, axis=1)
        beamstop_shadow = Geometry.get_beamstop_shadow(beamx, beamy, radius=100)

        v_blocks = Geometry.generate_bad_regions(vertial_gaps, mark='v')
        h_blocks = Geometry.generate_bad_regions(horizontal_gaps, mark='h')
        b_blocks = Geometry.generate_beamstop(beamstop_shadow)

        all_blocks = ["\n".join(block) for block in (common_block, b_blocks, v_blocks, h_blocks)]

        self.geometry_text = "\n\n".join(all_blocks)
        with open(self.geometry_file, "w") as fh:
            fh.write(self.geometry_text)
        print("geometry file is written as: ", self.geometry_file, " in ", os.getcwd())

        make_pixelmap = '/mnt/software/px/serial-x/crystfel-0.8.0/bin/make_pixelmap'

        if self.geometry_h5:
            cmd1 = "{} -o {} {}".format(make_pixelmap, self.geometry_h5, self.geometry_file)
            print("generate h5 geometry file: ", self.geometry_h5)
            os.system(cmd1)

        if self.badpixelmap:
            cmd2 = "{} --badmap -o {} {}".format(make_pixelmap, self.badpixelmap, self.geometry_file)
            print("generate h5 bad pixel map file: ", self.badpixelmap)
            os.system(cmd2)

        return self.geometry_file

    def get_geometry_block(self):
        lines = list()
        lines.append('CrystFEL stream format 2.3')
        lines.append('Generated by CrystFEL 0.8.0')
        lines.append('indexamajig ...')
        lines.append('----- Begin geometry file -----')
        lines.append(self.geometry_text)
        lines.append('----- End geometry file -----\n')
        return "\n".join(lines)

    def get_image(self):
        if not os.path.exists(self.image_file):
            print("image file does not exist: ", self.image_file)
            raise Exception("image file does not exist: ", self.image_file)

        fh = fabio.open(self.image_file)
        image = Geometry.get_header_items(fh)
        image['data'] = fh.data
        return image

    @staticmethod
    def inter_module_gaps(data, axis=1):
        indexes = np.where((data == -1).all(axis=axis))[0]  # axis 0: column, axis 1: rows, -1 marks gaps
        # print(data.shape)

        d2max = data.shape[axis]

        gaps = Geometry.get_limits(indexes)
        n_gap = len(gaps) // 2

        out = []
        # each tuple contains (xmin, ymin, xmax, ymax)
        for i in range(n_gap):
            if axis == 0:
                out.append(tuple([gaps[2 * i], 0, gaps[2 * i + 1], d2max - 1]))
            else:
                out.append(tuple([0, gaps[2 * i], d2max - 1, gaps[2 * i + 1]]))

        return out

    def generate_common_blocks(self, disable_data_label=True):
        if 'wavelength' in self.image:
            energy = round(12398.42 / float(self.image['wavelength']), 2)

        res = round(1.0 / float(self.image['pixel_size']), 1)  # how many pixels per meter
        if disable_data_label:
            disable = ";"
        else:
            disable = "" if self.data_label else ";"

        return Geometry.common_block.format(energy=energy,
                                            nx=int(self.image['size1']) - 1,
                                            ny=int(self.image['size2']) - 1,
                                            distance=self.image['distance'],
                                            xbeam=self.image['beam'][0],
                                            ybeam=self.image['beam'][1],
                                            resolution=res,
                                            data_label=self.data_label,
                                            disable=disable
                                            )

    @staticmethod
    def get_beamstop_shadow(xbeam, ybeam, radius=100):
        # radius of beam stop in pixels
        xbeam = float(xbeam)
        ybeam = float(ybeam)
        return tuple([0, int(ybeam - radius), int(xbeam + radius), int(ybeam + radius)])

    @staticmethod
    def generate_beamstop(beamstop_shadow):
        ''' e.g.
            bad_beamstop/min_x = -2000
            bad_beamstop/min_y = -40
            bad_beamstop/max_x = 100
            bad_beamstop/max_y = 100
        '''
        blocks = list()
        blocks.append("bad_beamstop/min_fs = {}".format(beamstop_shadow[0]))
        blocks.append("bad_beamstop/min_ss = {}".format(beamstop_shadow[1]))
        blocks.append("bad_beamstop/max_fs = {}".format(beamstop_shadow[2]))
        blocks.append("bad_beamstop/max_ss = {}".format(beamstop_shadow[3]))
        return blocks

    @staticmethod
    def generate_bad_regions(panels, mark='v'):
        labels = [x for x in string.digits + string.ascii_letters][::-1]
        direction = ['min_fs', 'min_ss', 'max_fs', 'max_ss']
        labels.pop()  # skip 0

        blocks = []
        for panel in panels:
            label = labels.pop()
            block = ["bad_{}{}/{} = {}".format(mark, label, dir, p) for dir, p in zip(direction, panel)]
            blocks.append("\n".join(block))

        return blocks

    @staticmethod
    def get_limits(arr):
        # remove elements if the numbers are adjacent to each other [1,3,4,5,6] -->[1,3,6]
        if arr.size == 0:
            return []

        out = [arr[0]]
        for i, elem in enumerate(arr[1:], 1):
            if elem - arr[i - 1] != 1:
                out.append(arr[i - 1])
                out.append(elem)

        out.append(arr[-1])

        if len(out) % 2 != 0:
            print("something is wrong, not all boundary found.")

        return out

    @staticmethod
    def get_header_items(fh):
        arr = fh.header['_array_data.header_contents'].split('\r\n')

        header = {}
        for a in arr:
            fields = re.split(r"\s+", a, 2)
            # print(fields)

            if "detector:" in fields[1].lower():
                header['detector'] = fields[-1]

            if "pixel_size" in fields[1].lower():
                header['pixel_size'] = fields[2].split()[0]

            if "distance" in fields[1].lower():
                header['distance'] = fields[2].split()[0]

            if "beam_xy" in fields[1].lower():
                header['beam'] = re.findall(r"\d+\.\d+|\d+", fields[2])

            if "wavelength" in fields[1].lower():
                header['wavelength'] = fields[2].split()[0]

        header['size1'] = fh.shape[-1]  # fh.dim1
        header['size2'] = fh.shape[-2]  # fh.dim2

        return header


if __name__ == '__main__':
    disable_data_label = True
    if len(sys.argv) == 3 and os.path.exists(sys.argv[1]) and 'cheetah' in sys.argv[2].lower():
        print("generate geometry file for crystfel using cheetah cxi file as input")
        image_file = sys.argv[1]
        disable_data_label = False
    elif len(sys.argv) == 2 and os.path.exists(sys.argv[1]):
        print("generate geometry file for crystfel using cbf files as input directly")
        image_file = sys.argv[1]
        disable_data_label = True
    else:
        image_file = "/mnt/beegfs/qxu/23IDD-test/4z2x-p21212/run0001/196802_1_00056.cbf"
        print("Usage: generate_geometry an_example_image_from_the_run [-cheetah]")
        print("       use cheetah option if using cheetah cxi files for input into crystfel.")
        sys.exit(0)

    Geometry(image_file, h5geom_file='current_geometry.h5', badpixelmap='bad_pixelmap.h5').generate(disable_data_label=disable_data_label)
