from __future__ import absolute_import
# run indexing for a batch of images using indexamajig
import os, sys, logging

from ..tasks.scriptjob import ScriptJob

logger = logging.getLogger(__name__)


class Indexamajig:
    def __init__(self,
                 input_file_list,
                 wdir=".",
                 sub_dir=None,
                 geometry_file=None,
                 pdb_file=None,
                 nproc=4,
                 high_res=None,
                 use_cheetah_peaks=False,
                 cheetah_peak_label="/entry_1/result_1",
                 internal_peakfinding_parameters=None,
                 integration_paramters=None,
                 indexing_parameters=None,
                 output_stream="o.stream",
                 temp_dir=None,
                 debug=False
                 ):
        self.input_file_list = input_file_list
        self.geometry_file = geometry_file
        self.pdb_file = pdb_file
        self.nproc = nproc
        self.use_cheetah_peaks = use_cheetah_peaks
        self.cheetah_peak_label = cheetah_peak_label
        self.internal_peakfinding_parameters = internal_peakfinding_parameters
        self.indexing_parameters = indexing_parameters
        self.integration_parameters = integration_paramters
        self.output_stream = output_stream
        self.high_res = high_res
        self.wdir = wdir
        self.temp_dir = temp_dir
        if sub_dir:
            self.wdir = os.path.join(self.wdir, sub_dir)

        if len(self.input_file_list) <= self.nproc:
            self.nproc = len(self.input_file_list)

        self.debug = debug

    def find_peaks(self):
        # if use cxi peak
        if os.path.exists(self.input_file_list) and self.use_cheetah_peaks:
            with open(self.input_file_list) as fh:
                lines = fh.readlines()
            if all(".cxi" in line for line in lines):
                logger.info("peaks in cheetah will be used")
                return "--peaks=cxi --hdf5-peaks={}".format(self.peak_label)

        # find peaks using indexamajig
        opt = ["--peaks={}".format(self.internal_peakfinding_parameters["algorithm"])]

        if "min_pix_count" in self.internal_peakfinding_parameters:
            opt.append("--min-pix-count={}".format(self.internal_peakfinding_parameters.get("min_pix_count")))

        if "max_pix_count" in self.internal_peakfinding_parameters:
            opt.append("--max-pix-count={}".format(self.internal_peakfinding_parameters.get("max_pix_count")))

        if "min_snr" in self.internal_peakfinding_parameters:
            opt.append("--min-snr={}".format(self.internal_peakfinding_parameters.get("min_snr")))

        if "threshold" in self.internal_peakfinding_parameters:
            opt.append("--threshold={}".format(self.internal_peakfinding_parameters.get("threshold")))

        return " ".join(opt)

    def indexing(self):
        opt = []

        if self.temp_dir and os.path.exists(self.temp_dir):
            opt.append("--temp-dir={}".format(self.temp_dir))

        if self.pdb_file and os.path.exists(self.pdb_file):
            opt.append("-p {}".format(self.pdb_file))

        if "indexing_methods" in self.indexing_parameters:
            methods = self.indexing_parameters.get("indexing_methods")
            if isinstance(methods, (list, set)):
                methods = ",".join(set(methods))
            opt.append("--index={}".format(methods))

        return " ".join(opt)

    def integration(self):
        opt = []

        if "method" in self.integration_parameters:
            opt.append("--int-radius={}".format(self.integration_parameters.get("method")))

        if "int_radius" in self.integration_parameters:
            opt.append("--int-radius={}".format(self.integration_parameters.get("int_radius")))

        return " ".join(opt)

    def assemble_cmd(self):

        opt = list()

        opt.append("source /mnt/software/px/serial-x/serialx-setup.sh\n\nindexamajig")
        if os.path.exists(self.input_file_list):
            opt.append("--input={}".format(self.input_file_list))

        if os.path.exists(self.geometry_file):
            opt.append("--geometry={}".format(self.geometry_file))

        opt.append("-j {}".format(self.nproc))

        opt.append("-o {}".format(self.output_stream))

        if self.high_res:
            opt.append("--highres={}".format(self.high_res))

        opt.append(self.find_peaks())
        opt.append(self.indexing())
        opt.append(self.integration())

        return opt

    def add_indexing_method(self, indexing_method):
        self.indexing_methods.append(indexing_method)

    def run(self):
        script_text = " ".join(self.assemble_cmd())

        if self.debug:
            print("command used for indexing: ", script_text)

        j = ScriptJob(wdir=self.wdir,
                      script_name="indexamajig.sh",
                      script_text=script_text,
                      runner="slurm",
                      nproc=self.nproc,
                      )
        j.run_sync()

        create_link = False
        if create_link and os.path.exists(self.output_stream):
            stream_dir, stream_file = os.path.split(self.output_stream)
            wk_dir = os.path.split(self.wdir)[-1]
            if self.debug:
                print("DEBUG wkdir, stream_dir=", wk_dir, stream_dir, self.output_stream,
                      " cwd=", os.getcwd(),
                      " src_path=", os.path.relpath(stream_dir, wk_dir),
                      " dest_path=", os.path.join(wk_dir, os.path.basename(stream_dir)))
            os.symlink("../streams/{}".format(stream_file), os.path.join(self.wdir, stream_file))


if __name__ == '__main__':
    wdir = '/mnt/beegfs/qxu/23IDD-test/4z2x-p21212/run0001/test'
    flist = '/mnt/beegfs/qxu/23IDD-test/4z2x-p21212/run0001/test/flist'
    geom = "/mnt/beegfs/qxu/23IDD-test/4z2x-p21212/run0001/test/current_geometry.geom"
    Indexamajig(flist,
                wdir=wdir,
                geometry_file=geom).run()
