#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Fri Mar  3 17:33:56 2017

@author: qxu
"""

import argparse
import datetime
import logging
import os
import re
import sys

from ..common.utils import isFloat

logger = logging.getLogger(__name__)


def get_args():
    """ command line options """

    parser = argparse.ArgumentParser(prog='SMX_STREAMER')

    g1 = parser.add_argument_group("data processing, data input")

    g1.add_argument('--data_dir', '-i',
                    required=False,
                    help='directory for image files')
    g1.add_argument("--filelist",
                    default=None,
                    help="a list of collected files, one file per line, in absolute path")
    g1.add_argument("--proc_dir", "-d", "-o",
                    default=os.getcwd(),
                    help="root processing directory")
    g1.add_argument("--recursive",
                    action="store_true",
                    required=False,
                    help="parse directory recursively to find all images")
    g1.add_argument("--beamx",
                    required=False,
                    type=str,
                    default=None,
                    help="define beamx in pixel, overwrite header value")
    g1.add_argument("--beamy",
                    required=False,
                    type=str,
                    default=None,
                    help="define beamy in pixel, overwrite header value")

    g1.add_argument("--highres", "-hr",
                    type=str,
                    default=None,
                    help='high resolution cutoff')
    g1.add_argument('--space_group', '-symm',
                    type=str,
                    default=None,
                    help='space group')
    g1.add_argument("--unit_cell", "-cell",
                    type=str,
                    default=None,
                    help="unit cell parameter")
    g1.add_argument("--model", "--pdb_file",
                    type=str,
                    default=None,
                    help="model for dimple")
    g1.add_argument("--prefix", "--startwith",
                    help="image filename prefix; use semicolon to separate multiple prefixes, e.g. A1;B1_1")
    g1.add_argument('--native', '-noanomalous',
                    action="store_true",
                    help="turn off anomalous (native data)")
    g1.add_argument("--runner", "-r",
                    default="local",
                    choices=["local", "sge", "slurm"],
                    help="use local cpus, sge, or slurm")
    g1.add_argument("--nproc", "-np",
                    type=int,
                    default=4,
                    help="number of processors per machine")
    g1.add_argument("--njobs", "-nj",
                    type=int,
                    default=1,
                    help="number of machines to use")
    g1.add_argument("--sampleNumber",
                    type=int,
                    default=0,
                    help="jbluice specific, sample number")
    g1.add_argument("--sampleName",
                    type=str,
                    default="",
                    help="jbluice specific, sample name")
    g1.add_argument("--imageSet",
                    type=str,
                    default=None,
                    help="jbluice specific, sets of images processed")
    g1.add_argument("--beamline",
                    type=str,
                    choices=["idb",
                             "23o",
                             "idd",
                             "23i",
                             "23b",
                             "23bm",
                             "config"
                             ],
                    default="config",
                    help="jbluice specific, beamline submitted the job, decides which db to write to")
    g1.add_argument("--logfile",
                    type=str,
                    default="",
                    help="name of the logfile")
    g1.add_argument("--log2db",
                    action="store_true",
                    required=False,
                    help="jbluice specific, logging program state/results to a database")
    g1.add_argument("--disable_console_logger",
                    action="store_true",
                    required=False,
                    help="disable stream the log to stdout")
    g1.add_argument("--debug",
                    action="store_true",
                    required=False,
                    help="turn on debug mode with more verbose output")

    g2 = parser.add_argument_group("Serial data process")
    g2.add_argument('--hitlist_dir',
                    required=False,
                    default=None,
                    help='directory for hitlist files generated by David Kissick')

    g2.add_argument("--run", required=False, action='append', help="define a run to process (SMX specific)")

    g2.add_argument("--lattice_type", type=str, choices=['monoclinic',
                                                         'triclinic',
                                                         'tetragonal',
                                                         'cubic',
                                                         'hexagonal',
                                                         'rhombohedral',
                                                         'orthorhombic'],
                    required=False, default=None, help="define lattice type for crystfel indexamajig")
    g2.add_argument("--centering", choices=['P', 'C', 'F', 'I'], type=str, required=False, default=None,
                    help="define centering type for crystfel")
    g2.add_argument("--int_radius", required=False, default="4,5,7",
                    help="define int-radius for crystfel indexamajig")
    g2.add_argument("--min_snr", required=False, default="5", help="define min-snr for crystfel indexamajig")
    g2.add_argument("--min_pix", required=False, default="2", help="define min-pix-count for crystfel indexamajig")
    g2.add_argument("--max_pix", required=False, default="20", help="define max-pix-count for crystfel indexamajig")
    g2.add_argument("--threshold", required=False, default="1", help="define threshold for crystfel indexamajig")
    g2.add_argument("--peak_algorithm", required=False, default="peakfinder8",
                    help="define peak finding algorithm for crystfel indexamajig")

    g2.add_argument("--indexing_methods", "--index", required=True, action='append',
                    choices=['mosflm', 'asdf', 'dirax', 'xds', 'xgandalf', 'taketwo', 'pinkIndexer'],
                    default=[],
                    help="define indexing methods for crystfel indexamajig")

    g2.add_argument("--workers", required=False, type=int, default=80,
                    help="define # of workers to run indexamajig")

    g2.add_argument("--chunk_size", required=False, type=int, default=1, help="define # of images for one worker")
    g2.add_argument("--nimages", required=False, type=int, default=-1, help="maximum # images to process")

    g2.add_argument("--cell_file", required=False, default=None, help="define crystfel cell file")
    g2.add_argument("--use_preprocessor",
                    action="store_true",
                    required=False,
                    help="use preprocessor (DOZOR) to filter out blank images, default False")

    args = parser.parse_args()

    if not args.data_dir and not args.filelist and not args.hitlist_dir:
        print("at least one of --data_dir, --filelist or --hitlist_dir is required.")
        parser.print_help()
        sys.exit(0)

    if args.data_dir and not os.path.isabs(args.data_dir):
        args.data_dir = os.path.join(os.getenv("HOME"), args.data_dir)
        print("image directory change to absolute path: {}".format(args.data_dir))

        if not os.path.exists(args.data_dir):
            print("Data directory does not exist. Quit.")
            sys.exit(1)

    if not os.path.isabs(args.proc_dir):
        args.proc_dir = os.path.join(os.getenv("HOME"), args.proc_dir)
        print("output directory change to absolute path: {}".format(args.proc_dir))

    if args.unit_cell:
        cell = re.split(r"[,;\s]+", args.unit_cell.strip(" "))
        if len(cell) == 6:
            args.unit_cell = ",".join(cell)
        else:
            logger.error("cell dimensions is not defined correctly, ignored.")
            args.unit_cell = None

    args.hostname = os.getenv('HOSTNAME')
    args.username = os.getenv('USER')
    args.pid = os.getpid()
    args.processing_id = os.getpid()
    args.command = " ".join(sys.argv)
    args.starttime = datetime.datetime.now()

    if not args.imageSet:
        args.imageSet = "All"

    if not args.logfile:
        args.logfile = "SMX_STREAMER_{}_{}_{}.log".format("smx_streamer", args.hostname, args.pid)
    print("PID = {} on host {}".format(args.pid, args.hostname))

    args.DATA = dict()
    return args


# run function and make it available when user import this module
SMX_ARGS = get_args()
print(SMX_ARGS)


def namespace2dict(ns, values_as_string=True):
    ns_copy = argparse.Namespace(**vars(ns))  # make a copy, otherwise, original copy got changed as well
    if isinstance(ns_copy, argparse.Namespace):
        runtime_info = vars(ns_copy)
        if values_as_string:
            for k, v in runtime_info.items():
                runtime_info[k] = str(v)
    return runtime_info
