import datetime
import logging
import os, sys
import threading
from multiprocessing import Manager

from pyqtgraph import QtGui

from .dir_monitor import DirMonitor
from .hitlist_monitor import HitlistMonitor
from .plotter import Plotter
from .preprocess import PreProcessor
from .process_manager import ProcessManager
from .stream_merger import StreamMerger
from .smx_get_args import SMX_ARGS

logger = logging.getLogger(__name__)

sys.setrecursionlimit(1000000)


def get_merged_stream_filename():
    if SMX_ARGS.data_dir:
        tag = os.path.basename(SMX_ARGS.data_dir.rstrip("/"))
    else:
        tag = 'filelist'

    merged_stream_file = os.path.join(SMX_ARGS.proc_dir, "{}_merged.streams".format(tag))
    if os.path.exists(merged_stream_file):
        print("{} exist, renamed by time stamp.".format(merged_stream_file))
        timestamp = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
        os.rename(merged_stream_file, "{}_{}".format(merged_stream_file, timestamp))
    return merged_stream_file


if __name__ == '__main__':

    if SMX_ARGS.proc_dir == os.getenv('HOME'):
        print("Please avoid the use of $HOME as the output directory.")
        exit(0)

    if os.path.exists(SMX_ARGS.proc_dir):
        timestamp = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
        print("{} already exists, will rename it.".format(SMX_ARGS.proc_dir))
        os.rename(SMX_ARGS.proc_dir, "{}-{}".format(SMX_ARGS.proc_dir.rstrip("/"), timestamp))

    os.makedirs(SMX_ARGS.proc_dir)

    m = Manager()
    preprocess_q = m.Queue()
    index_q = m.Queue()
    out_q = m.Queue()
    merge_q = m.Queue()

    # launch preprocess, spot finder in this case
    if not SMX_ARGS.use_preprocessor:
        index_q = preprocess_q
        t_fm = None
    else:
        fm = PreProcessor(preprocess_q, index_q, out_q)
        t_fm = threading.Thread(target=fm.manage)
        t_fm.daemon = True
        t_fm.start()

    if SMX_ARGS.hitlist_dir:
        hit_mon = HitlistMonitor(preprocess_q, SMX_ARGS.hitlist_dir)
        t_hm = threading.Thread(target=hit_mon.monitor)
        t_hm.daemon = True
        t_hm.start()

    mon = None
    if SMX_ARGS.data_dir or SMX_ARGS.filelist:
        mon = DirMonitor(preprocess_q, directory=SMX_ARGS.data_dir, filelist=SMX_ARGS.filelist)

    pm = ProcessManager(index_q, out_q, chunk_size=SMX_ARGS.chunk_size, pool_size=SMX_ARGS.workers, merge_queue=merge_q)
    t_pm = threading.Thread(target=pm.manage)
    t_pm.daemon = True
    t_pm.start()

    # merge stream daemon
    merged_stream_file = get_merged_stream_filename()
    merger = StreamMerger(merge_q, merged_stream_file)

    try:
        app = QtGui.QApplication([])
        app.setApplicationName("GMCA SMX STREAMER with CrystFEL")
        plt = Plotter(out_q, merged_stream_file, index_queue=index_q)
        app.exec_()
    except Exception as e:
        print("Error: unable to start the application, quit & clean up.")
        print(e)
    finally:
        if mon and mon.is_alive():
            mon.stop()

        if t_pm.is_alive():
            pm.stop()

        if merge_q.empty() and merger.is_alive():
            merger.stop()

        if t_fm and t_fm.is_alive():
            fm.stop()

    print("--all done--")
