#!/usr/bin/env python
# -*- coding: utf-8 -*-

from mod_python import apache, util
from cgi import parse_qs, escape
import tempfile
import shutil
import os, sys
import subprocess
import re
import StringIO
import json

SPACE_EQUIVALENT_CHARACTERS = '#:=,()'

NON_OPTIONAL_KEYWORDS = {
    #key: (pattern, [value_indeces], type)
    'Detector_identifier': ('Detector ', [slice(1, None)], str),
    'Pixel_size': ('Pixel_size', [1, 4], float),
    'Silicon': ('Silicon', [3], float),
    'Exposure_time': ('Exposure_time', [1], float),
    'Exposure_period': ('Exposure_period', [1], float),
    'Tau': ('Tau', [1], float),
    'Count_cutoff': ('Count_cutoff', [1], int),
    'Threshold_setting': ('Threshold_setting', [1], float),
    'Gain_setting': ('Gain_setting', [1, 2], str),
    'N_excluded_pixels': ('N_excluded_pixels', [1], int),
    'Excluded_pixels': ('Excluded_pixels', [1], str),
    'Flat_field': ('Flat_field', [1], str),
    'Trim_file': ('Trim_file', [1], str),
    'Image_path': ('Image_path', [1], str),
    }

OPTIONAL_KEYWORDS = {
    'Wavelength': ('Wavelength', [1], float),
    'Energy_range': ('Energy_range', [1, 2], float),
    'Detector_distance': ('Detector_distance', [1], float),
    'Detector_Voffset': ('Detector_Voffset', [1], float),
    'Beam_xy': ('Beam_xy', [1, 2], float),
    'Beam_x': ('Beam_xy', [1], float),
    'Beam_y': ('Beam_xy', [2], float),
    'Flux': ('Flux', [1], str),
    'Filter_transmission': ('Filter_transmission', [1], float),
    'Start_angle': ('Start_angle', [1], float),
    'Angle_increment': ('Angle_increment', [1], float),
    'Detector_2theta': ('Detector_2theta', [1], float),
    'Polarization': ('Polarization', [1], float),
    'Alpha': ('Alpha', [1], float),
    'Kappa': ('Kappa', [1], float),
    'Phi': ('Phi', [1], float),
    'Phi_increment': ('Phi_increment', [1], float),
    'Chi': ('Chi', [1], float),
    'Chi_increment': ('Chi_increment', [1], float),
    'Oscillation_axis': ('Oscillation_axis', [slice(1, None)], str),
    'N_oscillations': ('N_oscillations', [1], int),
    'Start_position': ('Start_position', [1], float),
    'Position_increment': ('Position_increment', [1], float),
    'Shutter_time': ('Shutter_time', [1], float),
    }

ALL_KEYWORDS = {}
ALL_KEYWORDS.update(NON_OPTIONAL_KEYWORDS)
ALL_KEYWORDS.update(OPTIONAL_KEYWORDS)

class PilatusHeader(object):
    """
    Class for parsing contents of a Pilatus cbf header from a minimal
    or full cbf file.

    Parsing the Pilatus cbf header populates the header_dict dictionary,
    using Pilatus cbf header keywords as keys.
    """
    def __init__(self, filepath, non_binary_length=4096):
        self.filepath = filepath
        self.non_binary_length = non_binary_length
        self.header_dict = {}
        self.header_lines = []

        self.read_header_lines()
        self.parse_header()

    def _rawheader(self):
        return open(self.filepath, 'rb').read(self.non_binary_length)

    def has_pilatus_cbf_convention(self):
        # Check for the _array_data.header_convention data item
        pilatus_header_pattern = re.compile(
            r'''_array_data.header_convention +["']?(SLS|PILATUS)'''
            r'''_\d+(\.?\d*)*["']?''')
        return bool(pilatus_header_pattern.search(self._rawheader()))

    def read_header_lines(self):
        """
        Populate the self.header_lines list.
        """
        contents_pattern = re.compile(
            r'''_array_data.header_contents\s+;.*?;''',
            re.DOTALL)
        contents_match = contents_pattern.search(self._rawheader())
        assert contents_match is not None
        self.header_lines = contents_match.group().splitlines()

    def _spaced_header_lines(self):
        """
        Return header_lines with all space equivalent charecters converted
        to space.
        """
        spaced_header_lines = []
        for line in self.header_lines:
            for space_equivalent in SPACE_EQUIVALENT_CHARACTERS:
                line = line.replace(space_equivalent, ' ')
            spaced_header_lines.append(line)
        return spaced_header_lines

    def parse_header(self):
        """
        Populate self.header_dict with contents of Pilatus cbf header
        """
        assert self.has_pilatus_cbf_convention()
        if len(self.header_lines) == 0:
            self.read_header_lines()
        # parse the header lines
        for key, (pattern, valueindices, datatype) in ALL_KEYWORDS.items():
            for line in self._spaced_header_lines():
                if pattern in line:
                    values = []
                    for idx in valueindices:
                        try:
                            # handle multiple or single values
                            if isinstance(idx, slice):
                                values += line.split()[idx]
                            else:
                                values.append(line.split()[idx])
                        except IndexError:
                            print ('No value at index %d on header line: \n'
                                '%s' % (idx, line))
                    value = self._datatype_handling(values, key, datatype)
                    if value is not None:
                        self.header_dict[key] = value

    def _datatype_handling(self, values, key, datatype):
        # handle rare cases of value "not set"
        if datatype is float and values[0] == 'not':
            # NON_OPTIONAL_KEYWORDS should always have value, at least NaN
            if key in NON_OPTIONAL_KEYWORDS:
                return float('NaN')
            else:
                return None
        # do the conversion for standard cases
        if len(values) == 1:
            values = datatype(values[0])
        else:
            if datatype is str:
                values = ' '.join(values)
            else:
                values = tuple([datatype(v) for v in values])
        return values

    def get_beam_xy_mm(self, factor=1000):
        return tuple([n * size * factor for n, size in zip(
            self.header_dict['Beam_xy'], self.header_dict['Pixel_size'])])

    def get_date_time(self):
        """
        Return date and time of image acquistion.

        Works for format of current camserver versions
        2011-06-04T04:57:02.976
        or format of old camserver versions
        2011/Sep/12 09:21:27.252
        """
        date_time_pattern = re.compile(
            r'(\d{4}-\d{2}-\d{2}T|\d{4}/\D+/\d{2} )\d{2}:\d{2}:\d{2}.\d+')
        return date_time_pattern.search(self._rawheader()).group()

    def get_time(self):
        time_pattern = re.compile(r'\d{2}:\d{2}:\d{2}.\d+')
        return time_pattern.search(self.get_date_time()).group()

    date_time = property(get_date_time)
    time = property(get_time)

def get_dozor_in(hdict):
    DOZOR = ("job single\n"
        "detector {detector}\n"
        "pixel_max {cutoff}\n"
        "fraction_polarization 0.99\n"
        "detector_distance {distance}\n"
        "X-ray_wavelength {wavelength}\n"
        "orgx {orgx}\n"
        "orgy {orgy}\n"
        "spot_size 3\n"
        "exposure {exposure}\n"
        "oscillation_range {osc}\n"
        "starting_angle {start_angle}\n"
        "name_template_image {template}\n"
        "first_image_number  {n_first}\n"
        "number_images {nimages}\n"
        "pixel_min -1\n"
        "ix_min 1\n"
        "ix_max {ix_max}\n"
        "iy_min {iy_min}\n"
        "iy_max {iy_max}\n")

    if 'PILATUS3 6M' in hdict['Detector_identifier'].upper():
        detector = 'pilatus6m'
        beamstopsizepixel = 50
    elif 'EIGER 16M' in hdict['Detector_identifier'].upper():
        detector = 'eiger16m'
        beamstopsizepixel = 100
    else:
        print "Error: detector type {} not implemented.".format(hdict['Detector_identifier'])

    orgx, orgy = hdict['Beam_xy']
    ix_max = int(orgx + beamstopsizepixel)
    iy_min = int(orgy - beamstopsizepixel)
    iy_max = int(orgy + beamstopsizepixel)
    start_angle = hdict['Start_angle']
    osc = hdict['Angle_increment']
    wavelength = hdict['Wavelength']
    nimages = hdict['Nimages']
    cutoff = hdict['Count_cutoff']
    exposure = hdict['Exposure_time']
    distance = hdict['Detector_distance'] * 1000
    template = hdict['Template']
    n_first = hdict['N_first']
    return DOZOR.format(detector = detector, cutoff = cutoff,
                        distance=distance, wavelength = wavelength,
                        orgx=orgx, orgy=orgy, exposure=exposure,
                        osc=osc, start_angle=start_angle,
                        template=template, n_first=n_first,
                        nimages=nimages, ix_max=ix_max, iy_min=iy_min,
                        iy_max=iy_max)


def run_dozor(args):
    ''' run dozor directly '''
    if "besthome" not in os.environ:
        os.environ["besthome"] = "/mnt/software/px/best"

    wdir = tempfile.mkdtemp(prefix='dozor_tmp')
    filename =  args['file_name']
    if not os.path.isfile(filename):
        print "Error: file {} does not exist.".format(filename)
    filename = os.path.abspath(filename)

    nimages = args['nimages'] if 'nimages' in args else 1
    h_dict = PilatusHeader(filename, 4096).header_dict
    h_dict['Nimages'] = 1  # nimages
    # NB the following only works for abc_0xxxxx.cbf type of filenames
    fn_template = filename[0:-9]+'?????'+filename[-4:]
    h_dict['N_first'] = int(filename[-9:-4])
    h_dict['Template'] = fn_template

    dozor_input = get_dozor_in(h_dict)
    d_out = {}
    try:
        tempf = tempfile.NamedTemporaryFile(prefix='dozor_', delete=False)
        tempf.write(dozor_input)
        tempf.close()

        if 'spots' in args and args['spots']:
            cmd = ['/mnt/software/px/DOZOR/dozor', '-p', tempf.name]
        else:
            cmd = ['/mnt/software/px/DOZOR/dozor', tempf.name]

        if 'binning' in args:
            cmd.extend(['-bin', '{}'.format(args['binning'])])

        dozor_out = subprocess.check_output(cmd, stderr=subprocess.STDOUT, shell=False, cwd=wdir)
        d_out = parse_dozor_v2(dozor_out)

        if not isinstance(d_out, list) or not d_out:
            return d_out

        # only deal with 1st image for now, need to modify to deal with nimages >1
        odict = d_out[0]
        odict['filename'] = filename

        spots = parse_spots(wdir, filename[-9:-4])
        odict['list_of_peaks']= spots
        odict['peak_format'] = 'x(pixel), y(pixel), i/sigma'
        if args['spots']:
            return json.dumps(odict, indent=4)
        else:
            return "{} {} {} {}".format(filename, nimages,
                                        odict['Total Spots'],
                                        odict['Resolution Spots'])
    except subprocess.CalledProcessError as e:
        raise Exception('ERROR: {}'.format(e.output))
    finally:
        os.unlink(tempf.name)
        shutil.rmtree(wdir)

def parse_spots(wdir, spotfile):
    spotfile="{}/{}.spot".format(wdir, spotfile)
    if not os.path.exists(spotfile):
        return

    with open(spotfile, 'r') as f:
        lines = f.readlines()

    if len(lines) <= 3:
        return

    out = []
    for l in lines[3:]:
        x = l.split()
        if len(x) == 5:
            o = "{} {} {}".format(x[1], x[2], round(float(x[3])/float(x[4]), 2))
            out.append(o)

    return out

def parse_dozor(outlog):
    lines = outlog.split("\n")[6:-2]
    out = []
    if len(lines) >= 1:
        start = False
        for line in lines:
            if line.startswith('-'):
                start = not start
                continue

            if start:
                flds2 = line[7:28].split()
                out.append({'Total Spots':flds2[0], 'Resolution Spots':flds2[-1]})
    return out if out else outlog

def parse_dozor_v2(outlog):
    lines = outlog.split("\n")[6:-2]
    out = []
    if len(lines) >= 1:
        start = False
        for line in lines:
            if line.startswith('-'):
                start = not start
                continue

            if start:
                results = list()
                groups = line.split("|")
                results.append(groups[0])

                nfields_per_group = [1, 3, 5, 3]
                for i in range(1, 4):
                    if "no results" in groups[i]:
                        results.extend(['-'] * nfields_per_group[i])
                    else:
                        results.extend(groups[i].split())

                keys = [
                    'img_num',
                    'num_peaks',
                    'Average Intensity',
                    'peak_resolution',
                    'Scale',
                    'B factor',
                    'Resolution Wilson',
                    'Correlation',
                    'R-factor',
                    'Main Score',
                    'Spot Score',
                    'Resolution Visible',
                ]
                if len(results) == 12:
                    out.append(dict(zip(keys, results)))
    return out if out else outlog

def handler(req):
    d = util.FieldStorage(req)
    args = {'file_name':d.get('file_name',''), 'binning':d.get('binning', '1'), 'nimages':d.get('nimages', '1'), "spots":d.get('spots', '')}
    logfile = StringIO.StringIO()
    logfile.write(run_dozor(args))
    log = logfile.getvalue()
    req.set_content_length(len(log))
    req.write(log)
    return apache.OK
