from __future__ import absolute_import
import os
import logging

from .scriptjob import ScriptJob
from ..file_parser.aimlessparser import AimlessParser
from ..config.config import XDS_HDF5_PLUGIN
from ..process.get_args import ARGS, namespace2dict

logger = logging.getLogger(__name__)


class AutoPROC(ScriptJob):
    def __init__(self,
                 id_string,
                 wdir='.',
                 data_dir=None,
                 subdir='autoPROC',
                 script_name='autoproc.sh',
                 highres=None,
                 native=False,
                 space_group=None,
                 unit_cell=None,
                 njobs=1,
                 fast_mode=False,
                 script_text=None,
                 nproc=2,
                 difficulty=False,
                 log2db=False,
                 image_sweeps=None,
                 no_image_dir_link=True,
                 runner='slurm',
                 pe_name='smp'):

        super(AutoPROC, self).__init__(wdir, script_name, script_text, nproc, runner, pe_name)
        self.data_dir = data_dir
        self.highres = highres
        self.native = native
        self.space_group = space_group
        self.unit_cell = unit_cell
        self.njobs = njobs
        self.fast_mode = fast_mode
        self.id_string = id_string
        self.runner = runner
        self.subdir = subdir
        self.difficulty = difficulty
        self.wdir = os.path.abspath(wdir)
        self.wdir = os.path.join(os.path.abspath(self.wdir), self.subdir)
        self.no_image_dir_link = no_image_dir_link
        if not os.path.isdir(self.wdir):
            os.makedirs(self.wdir)

        self.log2db = log2db
        self.image_sweeps = image_sweeps
        self.data_dict = namespace2dict(ARGS, values_as_string=True)

    def cmd_for_process(self):
        cmd = ["export PX=/mnt/software/px\nexport LD_PRELOAD=\n. /mnt/software/px/autoPROC/setup.sh\nprocess -d .",
               "autoPROC_Setup_StopOnDateError=no"]

        if self.id_string:
            cmd.append(self.id_string)

        if self.highres:
            cmd.append("-M NoHighResCut")
            cmd.append("-R 45.0 {}".format(self.highres))
        else:
            cmd.append("-M HighResCutOnCChalf")  # default to CC>=0.3 cutoff value

        if self.native:
            cmd.append("-noANO")

        if self.space_group:
            cmd.append('symm="{}"'.format(self.space_group))

        if self.unit_cell:
            cmd.append('cell="{}"'.format(self.unit_cell))

        if self.njobs > 1:
            cmd.append("autoPROC_XdsKeyword_MAXIMUM_NUMBER_OF_JOBS={}".format(self.njobs))

        if self.nproc > 1:
            cmd.append("autoPROC_XdsKeyword_MAXIMUM_NUMBER_OF_PROCESSORS={}".format(self.nproc))
            cmd.append("-nthreads {}".format(self.nproc))

        # if self.no_image_dir_link:
        #    cmd.append("autoPROC_LinkImagesIntoSubdirMaxLength=256")

        # a temporary fix for forcing autoproc to use of the dectris plugin instead of h5toxds
        if os.path.isfile(XDS_HDF5_PLUGIN) and '-h5' in self.id_string:
            cmd.append("autoPROC_XdsKeyword_LIB={}".format(XDS_HDF5_PLUGIN))

        if self.fast_mode:
            cmd.append("-M fast")

        if self.difficulty:
            cmd.append("-M LowResOrTricky")

        if self.data_dir:
            cmd.append('-I {}'.format(self.data_dir))

        # find links for Images and remove them
        # this shouldn't be necessary since we use autoPROC_LinkImagesIntoSubdirMaxLength=256
        remove_image_link = True
        if remove_image_link and '/autoproc-' in self.wdir:
            cmd.append("\n\nfind {} -type l -name Images -exec rm {} \;\n".format(self.wdir, "{}"))

        self.script_text = " ".join(cmd)
        return self

    def report(self):
        """minimal report"""
        # find where the aimless.log is
        aimless_dirs = list()
        for root, dirs, files in os.walk(self.wdir):
            if 'aimless.log' in files:
                aimless_dirs.append(root)

        if aimless_dirs:
            aimless_dir = aimless_dirs[0]

            # gather aimless statistics
            ap = AimlessParser(wdir=aimless_dir, filename='aimless.log')
            self.data_dict.update(ap.summarize())

            if self.log2db:
                from ..log.database import save_to_db

                # need to deal with multiple sweeps case,
                self.data_dict['imagedir'] = ARGS.data_dir
                self.data_dict['firstFrame'] = ':'.join(s.get_first_file() for s in self.image_sweeps)
                self.data_dict['pipeline'] = 'autoPROC'
                self.data_dict['state'] = 'DONE'
                self.data_dict['software'] = 'autoPROC'

                self.data_dict['scalingsoftware'] = 'aimless'
                self.data_dict['workdir'] = self.wdir
                # self.data_dict['imageSet'] = self.runtime_info.get('imageSet')
                self.data_dict['imageSet'] = ARGS.imageSet
                self.data_dict['report_url'] = "{}/summary.html".format(self.wdir)
                self.data_dict['logfile'] = "{}/{}.out".format(self.wdir, self.script_name)

                if len(aimless_dirs) == 1:
                    self.data_dict['truncate_mtz'] = '{}/{}'.format(aimless_dir, 'truncate-unique.mtz')
                    self.data_dict['truncate_log'] = '{}/{}'.format(aimless_dir, 'truncate.log')
                    self.data_dict['scale_log'] = '{}/{}'.format(aimless_dir, 'aimless.log')
                else:
                    self.data_dict['truncate_mtz'] = self.data_dict['report_url']  # in case of multiple mtz, link
                    self.data_dict['truncate_log'] = self.data_dict['report_url']  # in case of multiple mtz, link
                    self.data_dict['scale_log'] = '{}/{}'.format(aimless_dir, 'aimless.log')

                # if no sampleName was set, use prefix
                sample_name = ARGS.sampleName
                if sample_name == 'sampleName' or not sample_name:
                    self.data_dict['sampleName'] = self.data_dict['prefix']

                save_to_db(self.data_dict)

            # make some plots, note that matplotlib is not threadsafe, so don't use it in a thread
            # ap.plot()
        else:
            logger.info("aimless.log does not exist, autoPROC may have failed.")

    def run_process(self):
        self.cmd_for_process().run_sync(timeout=10800)
        self.report()


if __name__ == '__main__':
    ap = AutoPROC('', wdir='/home/qxu/data-for-tests/jcsg-p3-pilatus-6m/CP20740A_4dad/reprocess2/')
    print(ap.report())
