# this module implements dials autoindex and exporting to mosflm and xds features

from __future__ import absolute_import
import os
import logging
import shutil
import tempfile

from .scriptjob import ScriptJob
from ..tasks.mosflm import Mosflm
from ..common.spacegroup import Spacegroup
from ..config.config import CCP4_SETUP, DIALS_SETUP
from ..file_parser.mosflmparser import MosflmParser
from ..file_parser.dialsparser import DialsParser
from ..tasks.matthews_coef import MatthewCoef
from ..file_parser.labelitparser import screening_score
from ..process.get_args import ARGS, namespace2dict
from ..common.pv import update_strategy_ready_PV

logger = logging.getLogger(__name__)


class DialsIndex(ScriptJob):
    def __init__(self,
                 image_sweep,
                 wdir='.',
                 subdir = 'dials_strategy',
                 script_name='dials.sh',
                 script_text=None,
                 pipeline='dials_strategy',
                 nproc=4,
                 runner='local',
                 pe_name='smp',
                 space_group=None,
                 native=False,
                 unit_cell=None,
                 lattice=None,
                 log2db=False):
        super(DialsIndex, self).__init__(wdir,
                                         script_name,
                                         script_text,
                                         nproc,
                                         runner,
                                         pe_name)
        self.image_sweep = image_sweep
        self.image_files = self.image_sweep.get_all_files(abs_path=True)
        self.template = self.image_sweep.get_name_template(marker='#')
        self.image_numbers = self.image_sweep.get_all_file_numbers()
        self.lattice = lattice

        self.space_group = space_group
        self.unit_cell = unit_cell
        self.native = native
        self.log2db = log2db
        self.pipeline = pipeline
        self.strategy = None
        self.solNum = None
        self.cb_op = None

        self.pdir = wdir
        self.subdir = subdir
        self.sweep_id = self.image_sweep.get_sweep_id()
        self.wdir = os.path.join(self.pdir, self.sweep_id, self.subdir)
        if not os.path.isdir(self.wdir):
            os.makedirs(self.wdir)
        self.index_table = ""
        self.dbrow = None

        self.sampleName = self.image_sweep.get_imageSet(mode="prefix")
        self.sampleNumber = ARGS.sampleNumber

    def cmd_dials_index(self):
        dials_cmds = list()

        #dials_cmds.append(CCP4_SETUP)
        dials_cmds.append(DIALS_SETUP)
        imgs = ' '.join(self.image_files)
        dials_cmds.append('dials.import {}'.format(imgs))
        dials_cmds.append('dials.find_spots imported.expt nproc={}'.format(self.nproc))
        dials_cmds.append('dials.index {} imported.expt strong.refl'.format('max_cell_estimation.multiplier=2'))
        dials_cmds.append('dials.spot_counts_per_image json=spot_stat.json imported.expt strong.refl')
        dials_cmds.append('dials.refine_bravais_settings indexed.expt indexed.refl nproc={}'.format(self.nproc))
        self.script_text = '\n'.join(dials_cmds)
        self.script_name = 'dials1.sh'
        return self

    def run_dials_index(self):
        self.cmd_dials_index().run_sync()

    def cmd_dials_reindex(self):
        # need  a solution number and its reindex matrix to run
        dials_cmds = list()
        dials_cmds.append('dials.reindex indexed.expt indexed.refl change_of_basis_op={}'.format(self.cb_op))
        dials_cmds.append('dials.refine bravais_setting_{}.expt reindexed.refl'.format(self.solNum))
        dials_cmds.append('dials.export refined.expt format=mosflm'.format(self.solNum))
        self.script_text = '\n'.join(dials_cmds)
        self.script_name = 'dials2.sh'
        return self

    def run_dials_reindex(self):
        self.cmd_dials_reindex().run_sync()

    def standard_strategy(self):

        # skip run if it is already done befor
        if not (os.path.exists(os.path.join(self.wdir, "bravais_summary.json")) and
                os.path.exists(os.path.join(self.wdir, "indexed.expt")) and
                os.path.exists(os.path.join(self.wdir, "indexed.refl")) and
                os.path.exists(os.path.join(self.wdir, "strong.refl"))):
            self.run_dials_index()
        else:
            logger.warning("skipped dials indexing since since it was done previously.")

        dparser = DialsParser(wdir=self.wdir, filename='dials1.sh.out')
        if self.log2db:
            from ..log.database import ScreenStrategyResults
            self.spot_stat = dparser.get_spot_stat()
            self.spot_stat['pipelinestatus_id'] = ARGS.pipelinestatus_id

            self.dbrow = ScreenStrategyResults.create(**self.spot_stat)
            self.dbrow.directory = self.image_sweep.get_absdir()
            self.dbrow.sampleName = self.sampleName
            self.dbrow.workdir = self.wdir
            self.dbrow.software = "dials_mosflm"
            self.dbrow.images = ';'.join(self.image_sweep.get_all_files(abs_path=False))
            self.dbrow.anomalous = 0 if self.native else 1
            self.dbrow.save()

        if self.space_group:
            self.solNum = dparser.choose_symm(self.space_group)
        else:
            self.solNum = dparser.choose_best

        if not self.solNum:
            logger.error("dials failed to index.")
            return

        logger.info("chosen solution # {}".format(self.solNum))
        self.cb_op = dparser.get_bravais_summary(solNum=self.solNum)['cb_op']
        self.unit_cell = ' '.join([str(round(v, 1)) for v in dparser.get_bravais_summary(solNum=self.solNum)['unit_cell']])
        self.lattice = dparser.get_bravais_summary(solNum=self.solNum)['bravais']
        rmsd = dparser.get_bravais_summary(solNum=self.solNum)['rmsd']
        pointgroup_choices = dparser.get_possible_space_groups(sgtype='pointgroup')

        if not self.space_group:
            self.space_group = Spacegroup.get_lowest_space_group_symbol_from_lattice(self.lattice)

        if self.log2db:
            self.dbrow.spacegroup = self.space_group
            self.dbrow.index_table = dparser.get_index_table()
            self.dbrow.bravais_lattice = self.lattice
            self.dbrow.unitcell = self.unit_cell
            self.dbrow.solution_number = self.solNum
            self.dbrow.rmsd = round(rmsd,4)
            self.dbrow.images = ';'.join(self.image_sweep.get_all_files(abs_path=False))
            self.dbrow.sampleName = self.sampleName

            if pointgroup_choices:
                self.dbrow.pointgroup_choices = pointgroup_choices

            # calculate matthews coef
            mc = MatthewCoef(self.space_group, self.unit_cell, wdir=self.wdir)
            solvent_content = mc.get_table()
            asu_content = mc.get_asu_content()

            self.dbrow.solvent_content = solvent_content
            self.dbrow.estimated_asu_content_aa = asu_content
            self.dbrow.save()
            logger.info('save spot stats into db with primary key {}'.format(self.dbrow.get_id()))

        self.run_dials_reindex()

        matrix1 = os.path.join(self.wdir, 'mosflm_1', 'index.mat')
        if not os.path.exists(matrix1):
            matrix1 = os.path.join(self.wdir, 'mosflm', 'index.mat')
        matrix_name = 'index{:02d}.mat'.format(int(self.solNum))
        shutil.copy(matrix1, os.path.join(self.wdir, matrix_name))

        mosflm = Mosflm(self.image_sweep,
                        wdir=tempfile.mkdtemp(),
                        native=self.native,
                        space_group=self.space_group,
                        orientation_matrix=matrix_name)

        # set the wdir to the same as labelit and then calculate strategy
        mosflm.set_wdir(wdir=self.wdir)
        fname_tag = 'strategy_sln{}'.format(self.solNum)
        mosflm.strategy_only(tag=fname_tag)

        p1 = MosflmParser(wdir=self.wdir, filename='mosflm_{}.log'.format(fname_tag))
        self.strategy = p1.get_strategy()
        mosflm.set_strategy(self.strategy)
        #
        # run testgen using results above
        fname_tag = 'testgen_sln{}'.format(self.solNum)
        mosflm.testgen(tag=fname_tag)

        p2 = MosflmParser(wdir=self.wdir, filename='mosflm_{}.log'.format(fname_tag))
        osc = p2.get_testgen()

        # report results
        self.strategy['osc'] = osc
        self.strategy['solutionNo'] = self.solNum
        self.strategy['spacegroup'] = self.space_group
        self.strategy['unitcell'] = self.unit_cell
        self.strategy['edge_resol'] = mosflm.get_highres()
        self.strategy['mosaic'] = p1.get_mosaic()

        try:
            max_res = float(self.spot_stat['resolution_from_spots'])
            distance = self.image_sweep.get_detector_distance_given_max_res(max_res, default_distance=350, tolerance=0.6)
        except:
            logger.warning("detector distance estimation failed.")
            distance = 350

        if self.log2db:
            self.dbrow.osc_start = self.strategy['startAngle']
            self.dbrow.osc_end = self.strategy['endAngle']
            self.dbrow.completeness_native = self.strategy['nativeCompleteness']
            self.dbrow.completeness_anomalous = self.strategy['anomalousCompletenes']
            self.dbrow.mosaicity = self.strategy['mosaic']
            self.dbrow.score = screening_score(resol=self.spot_stat['resolution_from_spots'],
                                               rms=rmsd,
                                               mosaic=self.strategy['mosaic'])
            self.dbrow.osc_delta = osc
            self.dbrow.detectordistance = distance
            self.dbrow.state = 'STRATEGY'
            self.dbrow.anomalous = 0 if self.native else 1

            self.dbrow.save()
            update_strategy_ready_PV()

            logger.info("strategy calculation by DIALS/MOSFLM: DONE.")

        logger.info("Final Results: \n**MOSFLM strategy recommends that start from {} to {}, using osc <={},\n"
                    "  to get overall completeness of {}% and anomalous completeness {}% \n"
                    "  assuming spg {}, mosaic of {} and <= 1% overlap, maximum resolution at the detector edge {}A."
                    .format(self.strategy['startAngle'],
                            self.strategy['endAngle'],
                            self.strategy['osc'],
                            self.strategy['nativeCompleteness'],
                            self.strategy['anomalousCompletenes'],
                            self.strategy['spacegroup'],
                            self.strategy['mosaic'],
                            self.strategy['edge_resol']
                            ))
