from __future__ import absolute_import
import os, tempfile
import logging

from .scriptjob import ScriptJob
from ..common.spacegroup import Spacegroup
from ..common.pv import update_strategy_ready_PV
from ..tasks.mosflm import Mosflm
from ..config.config import PHENIX_SETUP
from ..file_parser.labelitparser import LabelitParser
from ..file_parser.mosflmparser import MosflmParser
from ..tasks.matthews_coef import MatthewCoef
from ..process.get_args import ARGS, namespace2dict

logger = logging.getLogger(__name__)


class Labelit(ScriptJob):
    def __init__(self,
                 image_sweep,
                 wdir='.',
                 subdir='labelit_strategy',
                 script_name='labelit.sh',
                 script_text=None,
                 pipeline='labelit_strategy',
                 nproc=1,
                 runner='local',
                 pe_name='smp',
                 compatibility_file=None,
                 space_group=None,
                 solNum=None,
                 highres=None,
                 native=False,
                 unit_cell=None,
                 lattice=None,
                 log2db=False):

        super(Labelit, self).__init__(wdir,
                                      script_name,
                                      script_text,
                                      nproc,
                                      runner,
                                      pe_name)
        self.image_sweep = image_sweep
        self.image_files = self.image_sweep.get_all_files(abs_path=False)
        self.template = self.image_sweep.get_name_template(marker='#')
        self.image_numbers = self.image_sweep.get_all_file_numbers()
        self.compatibility_file = compatibility_file
        if lattice and lattice not in ('aP',
                                       'mP', 'mC',
                                       'oP', 'oC', 'oF', 'oI',
                                       'tP', 'tI',
                                       'hR', 'hP',
                                       'cP', 'cF', 'cI',
                                       'mA', 'mB', 'mI'):
            logger.warning("incorrect input lattice {}".format(lattice))

        self.lattice = lattice
        self.space_group = space_group
        self.unit_cell = unit_cell
        self.native = native
        self.log2db = log2db
        self.dbrow = None
        self.pipeline = pipeline
        self.highres = highres

        self.solNum = solNum  # a solution number given by a user based on the index table
        self.strategy = None

        self.pdir = wdir
        self.subdir = subdir
        self.sweep_id = self.image_sweep.get_sweep_id()
        self.wdir = os.path.join(self.pdir, self.sweep_id, self.subdir)
        if not os.path.isdir(self.wdir):
            os.makedirs(self.wdir)

        self.runtime_info = namespace2dict(ARGS)
        try:
            self.sampleName = self.image_sweep.get_imageSet(mode="prefix")
            self.sampleNumber = self.runtime_info.get('sampleNumber')
        except (AttributeError, KeyError) as e:
            self.sampleName = 'sampleName'
            self.sampleNumber = 'sampleNumber'

    def cmd_run_labelit(self):
        labelit_cmd = list()

        labelit_cmd.append('labelit.screen')
        abs_template = os.path.join(self.image_sweep.get_absdir(), self.template)
        labelit_cmd.append('{} {}'.format(abs_template, ' '.join(str(i) for i in self.image_numbers)))

        # if True:
        #    labelit_cmd.append('distl.bins.verbose=True')

        detector = self.image_sweep.get_metadata_by_key('detector_class').upper()
        if 'PILA' in detector:
            labelit_cmd.append('distl.minimum_signal_height=2.4 distl.minimum_spot_area=4')
        elif 'EIGER' in detector:
            labelit_cmd.append('distl.minimum_signal_height=4.0 distl.minimum_spot_area=5')

        # default, but can probably tweaked
        labelit_cmd.append('difflimit_sigma_cutoff=0.75')

        if self.compatibility_file:
            if (os.path.isfile(self.compatibility_file) and
                    (self.compatibility_file.endswith('.pdb') or self.compatibility_file.endswith('.mtz'))):
                # for mtz input, standard column names such as IMEAN, SIGIMEAN are expected
                labelit_cmd.append('compatibility_file={}'.format(self.compatibility_file))

        if len(self.image_numbers) > 2:
            logger.warning("LABELIT usually works the best with two images 90 degree apart.")
            labelit_cmd.append('wedgelimit={}'.format(len(self.image_numbers)))

        labelit_cmd.append('best_support=True')

        # reduce the beam search scope to make it faster, must >=0.1, default is 4.0, naga suggested 0.5mm
        labelit_cmd.append('beam_search_scope=0.5')

        if self.lattice:
            labelit_cmd.append('known_bravais={}'.format(self.lattice))

        if self.space_group:
            labelit_cmd.append('known_symm={}'.format(self.space_group))

        if self.unit_cell:
            labelit_cmd.append('known_cell={}'.format(self.unit_cell))

        labelit_cmd.append('> labelit_autoindex.out')

        self.script_text = '\n'.join([PHENIX_SETUP,
                                      ' '.join(labelit_cmd),
                                      'labelit.stats_distl > labelit_spotfinder.out',
                                      'labelit.best_parameters > labelit_export2best.out'
                                      ])
        self.script_name = 'labelit_screen.sh'
        return self

    def run_labelit(self):
        self.cmd_run_labelit().run_sync()

    def standard_strategy(self):
        self.run_labelit()
        lparser = LabelitParser(wdir=self.wdir, filename='labelit_autoindex.out')
        index_table = lparser.get_raw_index_table()
        logger.info(index_table)

        # saving state
        self.spot_stat = lparser.get_spot_stat()
        if self.spot_stat and self.log2db:
            from ..log.database import ScreenStrategyResults
            self.spot_stat['pipelinestatus_id'] = self.runtime_info.get('pipelinestatus_id') if self.runtime_info else 0
            self.dbrow = ScreenStrategyResults.create(**self.spot_stat)
            self.dbrow.directory = self.image_sweep.get_absdir()
            self.dbrow.workdir = self.wdir
            self.dbrow.anomalous = 0 if self.native else 1
            self.dbrow.sampleName = self.sampleName
            self.dbrow.save()
            logger.info('save spot stats into db with primary key {}'.format(self.dbrow.get_id()))

        if lparser.get_error_msg():
            logger.error('labelit failed:\n{}'.format(lparser.get_error_msg()))
            return None

        if self.space_group:
            solution = lparser.get_solution_based_on_symm(self.space_group)
        else:
            solution = lparser.get_labelit_suggestion()
            self.space_group = solution.space_group

        pointgroup_choices = lparser.get_possible_space_groups(sgtype='pointgroup')

        # figure out parameters necessary to run mosflm strategy
        solNum = int(solution.solNum)
        matrix_name = 'index{:02d}.mat'.format(solNum)
        cell = solution.unit_cell
        if 'mosaic' in solution._fields:
            mosaic = solution.mosaic
        else:
            mosaic = None

        mosflm = Mosflm(self.image_sweep,
                        wdir=tempfile.mkdtemp(),
                        mosaic=mosaic,
                        highres=self.highres,
                        unit_cell=cell,
                        native=self.native,
                        space_group=self.space_group,
                        orientation_matrix=matrix_name)

        # set the wdir to the same as labelit and then calculate strategy
        mosflm.set_wdir(wdir=self.wdir)
        fname_tag = 'strategy_sln{}'.format(solNum)
        mosflm.strategy_only(tag=fname_tag)

        p1 = MosflmParser(wdir=self.wdir, filename='mosflm_{}.log'.format(fname_tag))
        self.strategy = p1.get_strategy()
        mosflm.set_strategy(self.strategy)

        # calculate matthews coef
        mc = MatthewCoef(solution.space_group, cell, wdir=self.wdir)
        solvent_content = mc.get_table()
        asu_content = mc.get_asu_content()

        # saving indexing results
        if self.log2db and self.dbrow:
            self.dbrow.state = 'AUTOINDEX'
            # self.dbrow.workdir = self.wdir
            # self.dbrow.directory = self.image_sweep.get_absdir()
            # self.dbrow.anomalous = 0 if self.native else 1

            if pointgroup_choices:
                self.dbrow.pointgroup_choices = pointgroup_choices

            self.dbrow.spacegroup = solution.space_group
            self.dbrow.index_table = index_table
            self.dbrow.bravais_lattice = Spacegroup.get_lattice_from_spacegroup(solution.space_group)
            self.dbrow.unitcell = cell
            self.dbrow.solution_number = solution.solNum
            # self.dbrow.penalty = solution.Penalty
            self.dbrow.osc_start = self.strategy['startAngle']
            self.dbrow.osc_end = self.strategy['endAngle']
            self.dbrow.completeness_native = self.strategy['nativeCompleteness']
            self.dbrow.completeness_anomalous = self.strategy['anomalousCompletenes']
            self.dbrow.mosaicity = p1.get_mosaic()
            self.dbrow.rmsd = solution.rmsd
            self.dbrow.score = solution.score
            self.dbrow.solvent_content = solvent_content
            self.dbrow.estimated_asu_content_aa = asu_content
            self.dbrow.save()

        # run testgen using results above
        fname_tag = 'testgen_sln{}'.format(solNum)
        mosflm.testgen(tag=fname_tag)

        p2 = MosflmParser(wdir=self.wdir, filename='mosflm_{}.log'.format(fname_tag))
        osc = p2.get_testgen()

        # report results
        self.strategy['osc'] = osc
        self.strategy['solutionNo'] = solution.solNum
        self.strategy['spacegroup'] = solution.space_group
        self.strategy['unitcell'] = solution.unit_cell
        self.strategy['edge_resol'] = mosflm.get_highres()

        try:
            max_res = float(self.spot_stat['resolution_from_spots'])
            distance = self.image_sweep.get_detector_distance_given_max_res(max_res, default_distance=350, tolerance=1.0)
        except:
            logger.warning("detector distance estimation failed.")
            distance = 350

        if self.log2db and self.dbrow:
            self.dbrow.state = 'STRATEGY'
            self.dbrow.detectordistance = distance
            self.dbrow.osc_delta = osc
            self.dbrow.anomalous = 0 if self.native else 1
            self.dbrow.save()
            update_strategy_ready_PV()

            logger.info("strategy calculation by LABELIT: DONE.")

        logger.info("Final Results: \n**MOSFLM strategy recommends that start from {} to {}, using osc <={},\n"
                    "  to get overall completeness of {}% and anomalous completeness {}% \n"
                    "  assuming spg {}, mosaic of {} and <= 1% overlap, maximum resolution at the detector edge {}A."
                    .format(self.strategy['startAngle'],
                            self.strategy['endAngle'],
                            self.strategy['osc'],
                            self.strategy['nativeCompleteness'],
                            self.strategy['anomalousCompletenes'],
                            self.strategy['spacegroup'],
                            self.strategy['mosaic'],
                            self.strategy['edge_resol']
                            ))

        return self.strategy
