from __future__ import absolute_import

import datetime
import logging
import os

from ..config.config import MIN_OSC_VALUE
from .scriptjob import ScriptJob
from ..common.pv import update_strategy_ready_PV, update_strategy_job_status_PV
from ..common.spacegroup import Spacegroup
from ..file_parser.labelitparser import screening_score
from ..file_parser.mosflmparser import MosflmParser
from ..process.get_args import ARGS, namespace2dict
from ..tasks.matthews_coef import MatthewCoef

from ..log.redisdb import strategy_to_redis

logger = logging.getLogger(__name__)


class Mosflm(ScriptJob):
    def __init__(self,
                 image_sweep,
                 wdir='.',
                 subdir='mosflm_strategy',
                 script_name='mosflm.sh',
                 highres=None,
                 rmin=5.0,
                 best=True,
                 native=False,
                 space_group=None,
                 unit_cell=None,
                 orientation_matrix=None,
                 mosaic=None,
                 pipeline='mosflm_strategy',
                 script_text=None,
                 nproc=2,
                 njobs=1,
                 runner='local',
                 pe_name='smp',
                 log2db=False):
        super(Mosflm, self).__init__(wdir,
                                     script_name,
                                     script_text,
                                     nproc,
                                     runner,
                                     pe_name)

        self.image_sweep = image_sweep

        self.native = native
        self.pipeline = pipeline
        self.highres = highres
        self.space_group = space_group
        self.unit_cell = unit_cell
        self.log2db = log2db
        self.dbrow = None
        self.subdir = subdir
        self.pdir = self.wdir
        self.rmin = rmin
        self.best = best
        self.subdir = subdir
        self.sweep_id = self.image_sweep.get_sweep_id()
        self.runtime_info = namespace2dict(ARGS, values_as_string=True)

        self.sampleName = self.runtime_info.get('sampleName') if self.runtime_info.get('sampleName') else self.sweep_id

        self.sampleNumber = self.runtime_info.get('sampleNumber')

        self.wdir = os.path.join(self.pdir, self.sweep_id, self.subdir)

        if not os.path.isdir(self.wdir):
            os.makedirs(self.wdir)

        self.nproc = nproc
        self.njobs = njobs
        self.strategy = None
        self.spot_stat = None
        self.orientation_matrix = orientation_matrix
        self.mosaic = mosaic

        # NB assuming one to one correspondence, not TRUE for h5 data
        self.image_numbers = self.image_sweep.get_all_file_numbers()
        self.spot_files = [os.path.join(self.wdir, ''.join(img[:-1]) + 'spt')
                           for img in self.image_sweep.get_filelist()]

        self.nref20s = 0  # number of reflections greater than 20 sigma
        self.nref10s = 0
        self.n_merged_refs = 0
        self.dps_threshold = 5

    def __cmd_init(self):
        """ common section for mosflm, NB currently we ask mosflm to fill in the detector/bl parameter from
        images, this assumes that image header are accurate
        """
        cmd = list()
        cmd.append("PNAME {}".format(self.sampleName))
        cmd.append("DNAME {}".format(self.sampleNumber))
        cmd.append("directory {}".format(self.image_sweep.get_absdir()))
        cmd.append("template {}".format(self.image_sweep.get_name_template("#")))
        # cmd.append("separation 0.0 0.0 update close")

        if self.best:
            cmd.append("best on")

        if self.space_group:
            cmd.append("symm {}".format(self.space_group))

        if self.unit_cell:
            cmd.append("cell {}".format(self.unit_cell))

        if self.highres:
            cmd.append("resolution {}".format(self.highres))
        else:
            edge_res = round(self.image_sweep.get_edge_res(), 2)
            cmd.append("resolution {}".format(edge_res))
            self.highres = edge_res

        for i, imgnum in enumerate(self.image_numbers):
            # in case of still images
            if float(self.image_sweep.get_osc()) <= MIN_OSC_VALUE:
                phi = self.image_sweep.get_phi(i)
                logger.info("osc appears to be 0... reset it to a small value")
                cmd.append('image {} phi {} to {}'.format(imgnum, phi, phi))
            else:
                cmd.append('image {}'.format(imgnum))

            cmd.append('head brief')
            cmd.append('go')

        if abs(float(self.image_sweep.get_twotheta())) >= 0.1:
            raise NotImplementedError
            # calculate swungout beam
            # cmd.append("beam swungout {}".format("not implemented yet"))

        return cmd

    def __cmd_for_findspots(self):
        cmd = list()
        cmd.append('findspots rmin {} AUTORESLN AUTORING LOCAL ICE'.format(self.rmin))

        # finding the spot explicitly seems to make mosflm autoindex works better
        for i, imgnum in enumerate(self.image_numbers):
            spt_file = self.spot_files[i]
            cmd.append('findspots find {} file {}'.format(imgnum, spt_file))
            cmd.append('go')
        return cmd

    def __cmd_for_autoindex(self, refine=True):
        """ commands for mosflm dps autoindex and strategy calculation using the 
            mosflm selected solution """
        cmd = list()

        imgs = " ".join([str(i) for i in self.image_sweep.get_all_file_numbers()])

        # use the spot file if it exists
        merged_spotfile = os.path.join(self.wdir, 'merged.spt')
        if os.path.isfile(merged_spotfile):
            imgs += ' thresh {} file {}'.format(self.dps_threshold, merged_spotfile)

        if refine:
            cmd.append("autoindex dps image {} refine save".format(imgs))
        else:
            cmd.append("autoindex dps image {} save".format(imgs))

        cmd.append("newmat autoindex.mat")
        cmd.append("mosaic estimate")
        cmd.append("go")

        if self.native:
            cmd.append("strategy auto notanom")
        else:
            cmd.append("strategy auto anomalous")
        cmd.append("go")

        cmd.append("stats")

        self.script_name = 'autoindex.sh'
        return cmd

    def __cmd_for_testgen(self,
                          matrix_name='autoindex.mat',
                          mosaic=None,
                          start=0,
                          end=180,
                          minosc=0.05,
                          maxosc=0.2,
                          overlap=1.0):
        cmd = list()

        # invoke image to check spot size
        cmd.append('image {}'.format(self.image_numbers[0]))
        cmd.append('go')
        if mosaic:
            cmd.append('mosaic {}'.format(mosaic))

        if self.strategy:
            start = self.strategy.get('startAngle', 0)
            end = self.strategy.get('endAngle', 180)
            matrix_name = self.strategy.get('matrix', 'autoindex.mat')
            cmd.append('matrix {}'.format(matrix_name))

            if self.strategy['mosaic']:
                mosaic = self.strategy['mosaic']
                cmd.append('mosaic {}'.format(mosaic))
            else:
                cmd.append('mosaic estimate images {}'.format(self.image_numbers[0]))
                cmd.append('go')

        cmd.append('testgen start {} end {} maxosc {} overlap {}'.format(start,
                                                                         end,
                                                                         maxosc,
                                                                         overlap))
        cmd.append('go')
        return cmd

    def __cmd_for_strategy(self):
        """ strategy with known matrix """
        cmd = list()

        if self.space_group and self.orientation_matrix:
            cmd.append('matrix {}'.format(self.orientation_matrix))
        else:
            logger.error("symm or orientation not set, cannot proceed with strategy calculation.")
            return None

        if not self.mosaic:
            cmd.append('mosaic estimate images {}'.format(self.image_numbers[0]))
            cmd.append('go')
        else:
            cmd.append('mosaic {}'.format(self.mosaic))

        if self.native:
            cmd.append("strategy auto notanom")
        else:
            cmd.append("strategy auto anomalous")
        cmd.append('go')
        return cmd

    def __cmd_for_mosaic_estimation(self, matrix_name='autoindex.mat'):
        """ estimate mosaic based on an index solution"""
        cmd = list()
        cmd.append('matrix {}'.format(matrix_name))
        cmd.append('mosaic estimate images {}'.format(self.image_numbers[0]))
        cmd.append('go')
        return cmd

    def get_wdir(self):
        return self.wdir

    def set_wdir(self, wdir):
        self.wdir = wdir

    def get_highres(self):
        return self.highres

    def set_strategy(self, strategy):
        self.strategy = strategy

    def run_mosflm(self, cmds, tag=None):
        """ 
        the wrapper to run all mosflm subtasks
        """
        self.tag = tag
        if not tag:
            self.tag = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

        self.spotod = os.path.join(self.wdir, 'mosflm_{}.spotod'.format(self.tag))
        self.summary = os.path.join(self.wdir, 'mosflm_{}.sum'.format(self.tag))
        self.coords = os.path.join(self.wdir, 'mosflm_{}.coords'.format(self.tag))
        self.logfile = os.path.join(self.wdir, 'mosflm_{}.log'.format(self.tag))

        cmd = list()
        cmd.append(
            "\nipmosflm SPOTOD {} COORDS {} SUMMARY {} << eof > {} ".format(self.spotod,
                                                                            self.coords,
                                                                            self.summary,
                                                                            self.logfile))
        cmd.extend(self.__cmd_init())
        cmd.extend(cmds)
        cmd.append("exit")
        cmd.append("eof")
        self.script_text = '\n'.join(cmd)
        self.script_name = 'mosflm_{}.sh'.format(self.tag)
        self.run_sync(timeout=300)  # quit in 5 mins in case of a hang process
        return self.logfile

    def __merge_spot_files(self):
        merge_spots = []
        header = footer = ''
        for i, file_spt in enumerate(self.spot_files):
            if os.path.isfile(file_spt):
                lines = open(file_spt, 'r').readlines()
                if i == 0:
                    header = lines[:3]
                    footer = lines[-2:]

                body = lines[3:-2]
                merge_spots.extend(body)
                nreflections = len(body)

        self.n_merged_refs = len(merge_spots)

        for spot in merge_spots:
            fields = spot.split()
            i = fields[-2]
            sigi = fields[-1]
            i_over_sigma = float(i) / float(sigi)

            if i_over_sigma >= 20.0:
                self.nref20s += 1

            if i_over_sigma >= 10.0:
                self.nref10s += 1

        logger.info('mosflm findspots, {}/{} reflections > 20/10 sigma respectively, '
                    'out of {} saved reflections.'.format(self.nref20s,
                                                          self.nref10s,
                                                          self.n_merged_refs))
        if self.nref20s >= 100:
            self.dps_threshold = 20
        else:
            if self.nref10s >= 100:
                self.dps_threshold = 10
        logger.info('dps autoindex threshold is determined as {}.'.format(self.dps_threshold))

        if self.n_merged_refs > 0:
            # save merged spots into a file named merged.spt
            merge_spots = header + merge_spots + footer
            with open(os.path.join(self.wdir, 'merged.spt'), 'w') as f:
                f.writelines(merge_spots)

    def findspots(self, tag='findspots'):
        self.run_mosflm(self.__cmd_for_findspots(), tag=tag)
        self.__merge_spot_files()

    def autoindex(self, tag='autoindex'):
        self.run_mosflm(self.__cmd_for_autoindex(), tag=tag)

    def testgen(self, tag='testgen'):
        self.run_mosflm(self.__cmd_for_testgen(), tag=tag)

    def estimate_mosaic(self, tag='mosaic_estimate'):
        self.run_mosflm(self.__cmd_for_mosaic_estimation(), tag=tag)

    def strategy_only(self, tag='strategy'):
        cmds = self.__cmd_for_strategy()
        self.run_mosflm(cmds=cmds, tag=tag)

    def standard_strategy(self):
        spotlog = os.path.join(self.wdir, 'mosflm_findspots.log')
        spotfile = os.path.join(self.wdir, 'merged.spt')

        # skip finding spots for reruns
        if not (os.path.exists(spotlog) and os.path.exists(spotfile)):
            self.findspots()
        else:
            self.n_merged_refs = sum(1 for line in open(spotfile, "r+")) - 5
            logger.info("skipped mosflm spot finding since the files already exist.")

        p0 = MosflmParser(wdir=self.wdir, filename='mosflm_findspots.log')
        self.spot_stat = p0.get_findspots_stat()
        self.spot_stat['pipelinestatus_id'] = self.runtime_info.get('pipelinestatus_id') if self.runtime_info else 0

        if self.log2db:
            from ..log.database import ScreenStrategyResults
            self.dbrow = ScreenStrategyResults.create(**self.spot_stat)
            self.dbrow.sampleName = self.image_sweep.get_imageSet(mode="prefix")
            self.dbrow.directory = self.image_sweep.get_absdir()
            self.dbrow.workdir = self.wdir
            self.dbrow.anomalous = 0 if self.native else 1

            self.dbrow.save()
            if self.spot_stat.get('nspots', None):
                strategy_to_redis({'resolution': self.spot_stat['resolution_from_spots'],
                                   'images': self.spot_stat['images'],
                                   'nspots': self.spot_stat['nspots'],
                                   'program': 'mosflm',
                                   })

            if self.n_merged_refs >= 20:
                job_status = "MOSFLM found {} spots, ~{}A.".format(self.spot_stat['n_spots'],
                                                                   self.spot_stat['resolution_from_spots'])
            else:
                job_status = "MOSFLM failed finding >=20 spots."

            imagenos = ",".join(str(i) for i in self.image_numbers)
            job_status = "{} {}\n{}".format(ARGS.sampleName, imagenos, job_status)
            update_strategy_job_status_PV(job_status)

            logger.info('save spot stats into db with primary key {}'.format(self.dbrow.get_id()))

        if self.n_merged_refs < 20:
            logger.error('Not enough spots {} found to proceed (<20), aborted.'.format(self.n_merged_refs))
            raise SystemExit(1)

        self.autoindex()
        p1 = MosflmParser(wdir=self.wdir, filename='mosflm_autoindex.log')
        p1.get_autoindex_table()
        solution = p1.get_mosflm_choice()
        self.strategy = p1.get_strategy()

        if not self.strategy:
            update_strategy_job_status_PV("MOSFLM strategy failed.")
            logger.info("Strategy failed.")
            return

        pointgroup_choices = p1.get_possible_space_groups(sgtype='pointgroup')

        mc = MatthewCoef(solution.Spacegroup, solution.Refined_cell, wdir=self.wdir)
        solvent_content = mc.get_table()
        asu_content = mc.get_asu_content()

        if self.log2db and self.dbrow:
            self.dbrow.state = 'AUTOINDEX'

            if pointgroup_choices:
                self.dbrow.pointgroup_choices = pointgroup_choices

            self.dbrow.spacegroup = solution.Spacegroup
            self.dbrow.bravais_lattice = Spacegroup.get_lattice_from_spacegroup(solution.Spacegroup)
            self.dbrow.unitcell = solution.Refined_cell
            self.dbrow.index_table = p1.get_raw_index_table()
            self.dbrow.solution_number = solution.SolutionNo
            self.dbrow.penalty = solution.Penalty
            self.dbrow.osc_start = self.strategy['startAngle']
            self.dbrow.osc_end = self.strategy['endAngle']
            self.dbrow.completeness_native = self.strategy['nativeCompleteness']
            self.dbrow.completeness_anomalous = self.strategy['anomalousCompletenes']
            self.dbrow.mosaicity = self.strategy['mosaic']
            self.dbrow.rmsd = solution.sdxy
            self.dbrow.score = screening_score(resol=self.spot_stat['resolution_from_spots'],
                                               rms=solution.sdxy,
                                               mosaic=self.strategy['mosaic'])
            self.dbrow.solvent_content = solvent_content
            self.dbrow.estimated_asu_content_aa = asu_content

            self.dbrow.save()
            strategy_to_redis({'unitcell': solution.Refined_cell,
                               'laue': solution.Spacegroup,
                               'osc_start': self.strategy['startAngle'],
                               'osc_end': self.strategy['endAngle'],
                               'mosaicity': self.strategy['mosaic'],
                               'rmsd': solution.sdxy,
                               'score': self.dbrow.score,
                               })

            job_status = "MOSFLM indexed,\nsymm={}, penalty={}".format(solution.Spacegroup, solution.Penalty)
            job_status = "{} {}\n{}".format(ARGS.sampleName, imagenos, job_status)
            update_strategy_job_status_PV(job_status)

        fname_tag = 'testgen_sln{}'.format(solution.SolutionNo)
        self.testgen(tag=fname_tag)
        p2 = MosflmParser(wdir=self.wdir, filename='mosflm_{}.log'.format(fname_tag))
        osc = p2.get_testgen()

        try:
            max_res = float(self.spot_stat['resolution_from_spots'])
            distance = self.image_sweep.get_detector_distance_given_max_res(max_res, default_distance=350,
                                                                            tolerance=1.0)
        except:
            logger.warning("detector distance estimation failed.")
            distance = 350

        if self.log2db and self.dbrow:
            self.dbrow.state = 'STRATEGY'

            self.dbrow.osc_delta = osc
            self.dbrow.detectordistance = distance
            self.dbrow.anomalous = 0 if self.native else 1

            self.dbrow.save()
            strategy_to_redis(self.spot_stat)

            job_status = "MOSFLM strategy finished,\ns={} e={} osc={} d={}".format(self.strategy['startAngle'],
                                                                                   self.strategy['endAngle'],
                                                                                   osc,
                                                                                   distance, )
            job_status = "{} {}\n{}".format(ARGS.sampleName, imagenos, job_status)
            update_strategy_job_status_PV(job_status)
            strategy_to_redis({'osc_delta': osc,
                               'distance': distance})

            update_strategy_ready_PV()
            logger.info("strategy calculation by MOSFLM: DONE.")

        self.strategy['osc'] = osc
        self.strategy['solutionNo'] = solution.SolutionNo
        self.strategy['spacegroup'] = solution.Spacegroup
        self.strategy['unitcell'] = solution.Refined_cell
        self.strategy['edge_resol'] = self.highres

        logger.info("Final Results: \n**MOSFLM strategy recommends that start from {} to {}, using osc <={},\n"
                    "  to get overall completeness of {}% and anomalous completeness {}% \n"
                    "  assuming spg {}, mosaic of {} and <= 1% overlap, maximum resolution at the detector edge {}A."
                    .format(self.strategy['startAngle'],
                            self.strategy['endAngle'],
                            self.strategy['osc'],
                            self.strategy['nativeCompleteness'],
                            self.strategy['anomalousCompletenes'],
                            self.strategy['spacegroup'],
                            self.strategy['mosaic'],
                            self.strategy['edge_resol']
                            ))
        return self.strategy
