#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import
import os
import logging

from .scriptjob import ScriptJob
from ..process.get_args import ARGS

logger = logging.getLogger(__name__)


mrbump_template = """
mrbump hklin {hklin} seqin {seq_file} hklout {sample_name}.mtz xyzout {sample_name}.pdb <<eof
LABIN {labin}
JOBID {sample_name}
MRPROGRAMS MOLREP PHASER
USEACORN 0
ACORNRES 1.7
ENANT {enantiomorphic}
QSIZE {qsize}
ROOTDIR {workdir}
NCYC 30
REFTWIN 0
BUCCANEER 1
BCYC 5
ARPWARP 0
ACYC 5
SHELXE 1
SCYC 15
SXRE 0
SXRB 1
SXRA 0
TRUN 1
JOBID 1
RLEVEL 50
RESHTML mrbump.html
MRNUM 5
ENSNUM 1
ENSMODNUM 1
MAPROGRAM MAFFT
DEBUG 0
SCOPSEARCH 0
SSMSEARCH 0
PQSSEARCH 0
MDLU 0
MDLD 0
MDLM 1
MDLC 1
MDLS 1
MDLP 0
FASTALOCAL 0
UPDATE 0
DOFASTA 0
DOPHMMER 1
EVALUE 1.E-5
NMASU
IGNORE
INCLUDE
HTMLOUT 0
CHECK 1
CLUSTER 1
DOHHPRED 0
HHSCORE 1
GEST 0
GESE 0
USEENSEM 0
TRYALL 0
CLEAN 1
LITE 1
END
eof
"""


class MRBUMP(ScriptJob):
    def __init__(self,
                 wdir='.',
                 script_name='mrbump.sh',
                 script_text=None,
                 sample_name='sampleName',
                 space_group=None,
                 nproc=8,
                 nmachines=4,
                 runner='local',
                 pe_name='smp',
                 hklin='aimless.mtz',
                 labin=None,
                 seq_file=None,
                 logfile='mrbump.log'):

        super(MRBUMP, self).__init__(wdir,
                                    script_name,
                                    script_text,
                                    nproc,
                                    runner,
                                    pe_name)
        self.hklin = hklin
        self.labin = labin if labin else "F=SIGFMEAN SIGF=SIGFMEAN FreeR_flag=FreeR_flag"
        self.wdir = wdir
        self.nproc = nproc
        self.nmachines = nmachines
        self.sample_name = sample_name
        self.space_group = space_group
        self.logfile = logfile
        self.seq_file = seq_file

        # check enantiomorph unless user specified space group
        self.explore_sym = 0 if ARGS.space_group else 1

        if not os.path.exists(self.wdir):
            logger.info("{} does not exist...creating it.".format(self.wdir))
            os.makedirs(self.wdir)

    def cmd_run_mrbump(self):
        cmd = mrbump_template.format(
            hklin=self.hklin,
            seq_file=self.seq_file,
            workdir=self.wdir,
            sample_name=self.sample_name,
            labin=self.labin,
            qsize=self.nproc*self.nmachines,
            enantiomorphic=self.explore_sym,
        )

        if os.path.isfile(self.hklin) and os.path.exists(self.seq_file):
            self.script_text = cmd
            return self
        else:
            logger.error("{} does not exist".format(self.hklin))

    def run(self, async_flag=False):
        logger.info("run mrbump pipeline for structure determination.")
        if async_flag:
            self.cmd_run_mrbump().run_async()
        else:
            return self.cmd_run_mrbump().run_sync(timeout=7200)