from __future__ import absolute_import
import os
import logging

from .scriptjob import ScriptJob
from ..common.spacegroup import Spacegroup

logger = logging.getLogger(__name__)


class Pointless(ScriptJob):
    def __init__(self,
                 wdir='.',
                 script_name='pointless.sh',
                 script_text=None,
                 nproc=1,
                 runner='local',
                 pe_name='smp'):
        super(Pointless, self).__init__(wdir,
                                        script_name,
                                        script_text,
                                        nproc,
                                        runner,
                                        pe_name)

    def cmd_xds2mtz(self, xdsfiles,
                    dest_dir=None,
                    spacegroup=None,
                    mtzout='pointless_convert.mtz',
                    logfile='pointless-convert.log'):

        if not isinstance(xdsfiles, (list, tuple)):
            xdsfiles = [xdsfiles]

        if not dest_dir:
            dest_dir = self.wdir

        hklin = 'xdsin'
        if xdsfiles[0].endswith('.mtz'):
            hklin = 'hklin'

        xdsfiles = [os.path.join(self.wdir, f) for f in xdsfiles]

        if all(os.path.isfile(f) for f in xdsfiles):
            mtzout = os.path.join(os.path.abspath(dest_dir), mtzout)
            logfile = os.path.join(os.path.abspath(dest_dir), logfile)
            cmd = "pointless -copy hklout {} <<eof > {}\n".format(mtzout, logfile)
            cmd += '\n'.join('{} {}'.format(hklin, f) for f in xdsfiles)
            cmd += '\n'

            reindex = True
            if spacegroup:
                if not reindex:
                    cmd += "spacegroup {}\n".format(spacegroup.replace(" ", ""))
                    cmd += "reindex h,k,l\n"  # ask pointless not to change space group
                else:
                    new_spg, reindex_vec = Spacegroup.get_reindex_vector(spacegroup)
                    cmd += "spacegroup {}\n".format(new_spg)
                    cmd += "reindex {}\n".format(reindex_vec)  # ask pointless to change space group to stand p21212 etc

            cmd += "eof\n"
            self.script_text = cmd
            return self
        else:
            logger.error("files not exist {}, convert to mtz failed.".format(xdsfiles))

    def cmd_spg_check(self,
                      xdsfile,
                      highres=None,
                      outxml='pointless-spg.xml',
                      logfile='pointless-spg.log'):

        hklin = 'xdsin'
        if xdsfile.endswith('.mtz'):
            hklin = 'hklin'

        xdsfile = os.path.join(self.wdir, xdsfile)
        if os.path.isfile(xdsfile):
            outxml = os.path.join(self.wdir, outxml)
            logfile = os.path.join(self.wdir, logfile)
            cmd = "pointless {} {} xmlout {} > {} <<eof\n".format(hklin,
                                                                  xdsfile,
                                                                  outxml,
                                                                  logfile)
            if highres:
                cmd += 'RESOLUTION HIGH {}\n'.format(highres)
            cmd += "\neof\n"
            self.script_text = cmd
            return self
        else:
            logger.error("files not exist {}, failed space group determination.".format(xdsfile))

    def cmd_compare_to_reference(self, file_ref,
                                 file_to_check,
                                 outxml='pointless-ref.xml',
                                 logfile='pointless-ref.log',
                                 highres=None):
        hklin = 'xdsin'
        if file_to_check.endswith('.mtz'):
            hklin = 'hklin'

        if not file_ref.endswith('.mtz'):
            logger.error('pointless requires that the reference file must be in mtz format.')

        # file_ref = os.path.join(self.wdir, file_ref)
        # file_to_check = os.path.join(self.wdir, file_to_check)
        if os.path.isfile(file_ref) and os.path.isfile(file_to_check):
            cmd = "pointless hklref {} {} {} xmlout {} > {} <<eof\n".format(file_ref,
                                                                            hklin,
                                                                            file_to_check,
                                                                            outxml,
                                                                            logfile)
            if highres:
                cmd += 'RESOLUTION HIGH {}\n'.format(highres)
            cmd += "\neof\n"
            self.script_text = cmd
            return self
        else:
            logger.error("{} {} may not exist...".format(file_to_check, file_ref))

    def check_space_group(self, xdsfile,
                          outxml='pointless-spg.xml',
                          logfile='pointless-spg.log',
                          timeout=600):
        return self.cmd_spg_check(xdsfile,
                                  outxml=outxml,
                                  logfile=logfile).run_sync(timeout=timeout)
