# -*- coding: utf-8 -*-
from __future__ import absolute_import

import datetime
import logging
import os
import re
import stat
import subprocess
import threading
import time

from .script_template import get_template

logger = logging.getLogger(__name__)


class ScriptJob(object):
    def __init__(self,
                 wdir='.',
                 script_name='script.sh',
                 script_text=None,
                 nproc=1,
                 runner='local',
                 pe_name='smp'):
        self.wdir = os.path.abspath(wdir)
        self.script_name = script_name
        self.nproc = nproc
        self.process = None
        self.pe_name = pe_name
        self.runner = runner
        self.script_text = script_text
        self.OK = True
        self.QSUB = "qsub -v LD_PRELOAD="
        self.my_env = os.environ.copy()
        if 'LD_PRELOAD' in self.my_env:
            del self.my_env['LD_PRELOAD']

    def update_cmd(self, script_text):
        self.script_text = script_text
        return self

    def write_script(self, overwrite=True):
        if not os.path.exists(self.wdir):
            os.makedirs(self.wdir)

        if self.script_text:
            chdir = "cd {}".format(self.wdir)
            script = get_template(self.runner).format(joblabel=self.script_name,
                                                      chdir=chdir,
                                                      script_text=self.script_text)

            script_file = os.path.join(self.wdir, self.script_name)
            if os.path.isfile(script_file) and not overwrite:
                timeStamp = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
                file_rename = script_file + '.' + timeStamp
                os.rename(script_file, os.path.join(self.wdir, file_rename))

            with open(script_file, "w") as fh:
                fh.write(script)

            os.chmod(script_file,
                     stat.S_IXUSR
                     + stat.S_IWUSR
                     + stat.S_IRUSR
                     + stat.S_IRGRP
                     + stat.S_IROTH)
        else:
            logger.error("cannot write {}, no cmd was defined.".format(self.script_name))

    def run_sync(self, timeout=None, write_script_to_file=True):
        def target():
            if write_script_to_file:
                self.write_script()

            if not os.path.exists(os.path.join(self.wdir, self.script_name)):
                logger.error("failed to execute script {}, script not found.".format(self.wdir, self.script_name))
                return

            qsub_common = "{} -sync y -j y -V -N job{} -cwd {}".format(self.QSUB, self.script_name, self.script_name)
            qsub_ncpus = "{} -pe {} {} {}".format(qsub_common, self.pe_name, self.nproc, self.script_name)
            srun = "sbatch --wait --export=ALL --job-name=j.{} --cpus-per-task={} {}".format(self.script_name, self.nproc, self.script_name)

            if self.runner == 'local':
                cmd = os.path.join(".", self.script_name)

                # special handling for dials programs since they often hang in subprocess, need to figure out why???
                if 'dials.' in self.script_text:
                    cmd = srun
            elif 'sge' in self.runner:
                if self.nproc > 1:
                    cmd = qsub_ncpus
                else:
                    cmd = qsub_common
            elif 'slurm' in self.runner:
                cmd = srun

            if self.runner == 'local':
                logger.debug('Job thread {}/{} started locally'.format(self.wdir, self.script_name))
                self.process = subprocess.Popen(cmd, shell=True, cwd=self.wdir, env=self.my_env,
                                                stdout=open(os.path.join(self.wdir, self.script_name + ".out"), "w"),
                                                stderr=open(os.path.join(self.wdir, self.script_name + ".err"), "w"), )
            else:
                logger.debug('Job thread {}/{} started remotely'.format(self.wdir, self.script_name))
                self.process = subprocess.Popen(cmd, shell=True, cwd=self.wdir, env=self.my_env,
                                                stdout=open(os.path.join(self.wdir, "{}.out".format(self.runner)), "w"),
                                                stderr=open(os.path.join(self.wdir, "{}.err".format(self.runner)), "w"),)

            self.process.communicate()
            logger.debug('Job thread {}/{} finished'.format(self.wdir, self.script_name))

        thread = threading.Thread(target=target)
        thread.start()

        thread.join(timeout)
        if thread.is_alive():
            logger.warning(
                'Terminating process {}/{}. Timeout exceeded {}.'.format(self.wdir,
                                                                         self.script_name,
                                                                         timeout))
            self.process.terminate()
            thread.join()
        return self.process.returncode

    # start the job locally without tracking the status, need more work
    def run_async(self):
        def target():
            self.write_script()

            if 'sge' in self.runner:
                cmd = "{} -j y -N j.{} -V -cwd {}".format(self.QSUB, self.script_name, self.script_name)
            elif 'slurm' in self.runner:
                cmd = "sbatch --job-name=j.{} --chdir={} --export=ALL {}".format(self.script_name, self.wdir, self.script_name)
            else:
                cmd = "{}/{}".format(self.wdir, self.script_name)

            logger.debug('Job thread {}/{} started remotely'.format(self.wdir, self.script_name))
            self.process = subprocess.Popen(cmd, shell=True, cwd=self.wdir, env=self.my_env,
                                            stdout=open(os.path.join(self.wdir, "{}_{}.out".format(self.script_name, self.runner)), "w"),
                                            stderr=open(os.path.join(self.wdir, "{}_{}.err".format(self.script_name, self.runner)), "w"), )

            self.process.communicate()

        thread = threading.Thread(target=target)
        thread.start()
        thread.join(30)
        logger.info('Job thread {}/{} submitted.'.format(self.wdir, self.script_name))

    """
    the following was modified by qxu @ GMCA/ANL based on batchjob.py from Keitaro Yamashita @riken
    """

    def run_remote(self, timeout=None):
        self.write_script()
        script_name = self.script_name
        wdir = self.wdir
        if self.nproc > 1:
            cmd = "{} -j y -V -pe {} {} {}".format(self.QSUB, self.pe_name, self.nproc, script_name)
        else:
            cmd = "{} -j y -V {}".format(self.QSUB, script_name)

        p = subprocess.Popen(cmd, shell=True, cwd=wdir, env=self.my_env, stdout=subprocess.PIPE)
        p.wait()
        stdout = p.stdout.readlines()
        if p.returncode != 0:
            logger.error("qsub failed. returncode is {}.\n"
                         "stdout:\n{}\n".format(p.returncode, stdout))

        r = re.search(r"^Your job ([0-9]+) ", stdout[0])
        job_id = r.group(1)
        if job_id == "":
            logger.error("cannot read job-id from qsub result."
                         " please contact author. stdout is:{}\n".format(stdout))

        logger.debug("Job {}/{} started remotely. id= {}".format(self.wdir,
                                                                self.script_name,
                                                                job_id))

        acc = 0.
        while self.qstat(job_id) is not None:
            time.sleep(0.1)
            acc += 0.1
            if timeout is not None and acc > timeout:
                logger.warning(
                    'Terminating remote process {}/{}. Timeout exceeded {}.'.format(self.wdir,
                                                                                    self.script_name,
                                                                                    timeout))
                self.qdel(job_id)
                return -15
        return 0

    def qstat(self, job_id):
        cmd = "qstat -j %s" % job_id
        p = subprocess.Popen(cmd, shell=True,
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        p.wait()
        stdout = p.stdout.readlines()

        if p.returncode != 0:
            logger.info("qstat %s failed (job probably successfully finished)." % job_id)
            return None

        status = {}

        for l in filter(lambda s: ":" in s, stdout):
            splitted = l.split(":")
            key = splitted[0].strip()
            val = "".join(splitted[1:]).strip()
            status[key] = val

        return status

    def qdel(self, job_id):
        cmd = "qdel %s" % job_id
        p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        p.wait()

        if p.returncode != 0:
            logger.error("qdel %s failed." % job_id)
            return None

    def report(self):
        pass


if __name__ == '__main__':
    commands = open('/mnt/beegfs/qxu/23BM_2017_08_08/4z2x-p21212/process/analysisNG/196802_0_00181_00271/dials_strategy/dials2.sh').read()
    print(commands)

    ex = ScriptJob(wdir='/mnt/beegfs/qxu/23BM_2017_08_08/4z2x-p21212/process/analysisNG/196802_0_00181_00271/dials_strategy',
                   script_name="dials2x.sh",
                   runner='local',
                   script_text=commands)

    ex.run_sync()
