# -*- coding: utf-8 -*-
from __future__ import absolute_import
import os
import logging
import glob
from json import load

from ..process.get_args import ARGS
from ..tasks.dimple import Dimple
from ..tasks.fastep import Fast_EP
from ..tasks.mrBump import MRBUMP
import datetime
from ..tasks.morda import MoRDA
from ..common.utils import timeit, parallelize
from ..log.database import DataProcessResults

logger = logging.getLogger(__name__)

"""
ALTER TABLE dataprocessresults modify column solve varchar(255);
"""


def update_job_status(status="RUN", jobname=None):
    if ARGS.log2db:
        ARGS.pipelinestatus_table_reference.state = status
        ARGS.pipelinestatus_table_reference.elapsedtime = (datetime.datetime.now() - ARGS.starttime).total_seconds()

        if jobname:
            ARGS.pipelinestatus_table_reference.sampleName = jobname

        ARGS.pipelinestatus_table_reference.save()


class Solver:
    def __init__(self, data_info=None, wdir=None):
        self.data_stat = data_info
        self.wdir = wdir if wdir else ARGS.proc_dir
        self.data_scale = None
        self.data_truncate = None
        self.data_scalepack = None
        self.dbid = None
        self.json_chosen = None
        self.json_all = None

        self.model = ARGS.model
        self.sequence = ARGS.sequence
        self.hatom = ARGS.hatom
        self.space_group = ARGS.space_group
        self.skip_solver = False
        self.init()

        if ARGS.space_group:
            self.space_group_type = "spacegroup"

    def evaluate_anomalous_signal(self):
        # return 0, no anomalous signal, or a number for resolution for anomalous resolution cutoff
        return float(self.data_stat['anomalous_signal'])

    def evaluate_data_quality(self):
        # e.g. completeness >90% and rmerge < 30%, return True; else False
        return self.data_stat['data_quality']

    def get_data_stat_from_json(self):
        try:
            json_files = glob.glob(os.path.join(self.wdir, "*.json"))
            if len(json_files) < 1:
                self.skip_solver = True
                return

            self.json_all = sorted(json_files)
            self.json_chosen = self.json_all[-1]
            with open(self.json_chosen) as f:  # only choose last file sorted by wav4, hopefully peak
                self.data_stat = load(f)
        except Exception:
            self.skip_solver = True
            raise RuntimeError("unable to load scaling json file, skip solver")

    def init(self):
        if not self.data_stat:
            self.get_data_stat_from_json()

        if not self.data_stat:
            self.skip_solver = True
            return

        self.data_scale = self.data_stat['scale_mtz']
        self.data_truncate = self.data_stat['truncate_mtz']
        self.data_scalepack = self.data_stat.get('scalepack', None)
        self.dbid = self.data_stat['dataprocessresults_id']

        # rerun pointless to check the space group
        from ..tasks.pointless import Pointless
        from ..file_parser.pointlessparser import PointlessParser

        Pointless(wdir=ARGS.proc_dir,
                  script_name='pointless-spg.sh'
                  ).check_space_group(self.data_scale)
        spg_info = PointlessParser(wdir=ARGS.proc_dir,
                                   filename='pointless-spg.log'
                                   ).get_laue_group_from_xml(xmlfile='pointless-spg.xml')

        # for structure solution, may still need to check enantiomorphic variants
        if spg_info['type'] == 'spacegroup':
            self.space_group_type = spg_info['type']

    @timeit
    def solve(self):
        if self.skip_solver:
            return

        if not self.evaluate_data_quality():
            logger.info("{} data quality is poor, skipped solver.")
            return

        solve_wdir = self.wdir

        if self.model and not self.sequence:
            # run dimple
            update_job_status("LIGAND")

            solve_wdir = os.path.join(self.wdir, "dimple")
            if not os.path.exists(solve_wdir):
                os.makedirs(solve_wdir)
            hklin_dir, hklin_name = os.path.split(self.data_scale)
            dimple = Dimple(solve_wdir, model=self.model,
                            hklin_dir=hklin_dir,
                            hklin_name=hklin_name)
            dimple.run()
        elif self.sequence and not self.model:
            update_job_status("MR")

            """
            # create & run MRage job
            from ..tasks.mrage import MRage

            solve_wdir = os.path.join(self.wdir, "MRage")
            mrage = MRage(wdir=solve_wdir,
                          hklin=self.data_truncate,
                          seq_file=self.sequence,
                          sample_name=ARGS.sampleName,
                          )
            mrage.run()  ## skip, too time consuming
            """

            """
            # run mrbump
            mrbump = MRBUMP(wdir=os.path.join(self.wdir, "MRBump"),
                          hklin=self.data_truncate,
                          seq_file=self.sequence,
                          sample_name=ARGS.sampleName,)
            mrbump.run()
            """
            if self.space_group_type == 'spacegroup':
                solve_wdir = os.path.join(self.wdir, "MoRDA")
                MoRDA(wdir=solve_wdir,
                      hklin=self.data_truncate,
                      seq_file=self.sequence,
                      space_group_exploration=0,
                      sample_name=ARGS.sampleName, ).run()
            else:
                logger.info("skipped MR since space group isn't uniquely defined.")

        elif self.sequence and self.model:
            update_job_status("MR")
            solve_wdir = os.path.join(self.wdir, "phaser")
        elif self.evaluate_anomalous_signal():
            update_job_status("SAD")
            # run fast_ep
            '''
            if self.json_all and len(self.json_all) > 1:
                self.parallel_fast_ep_SAD()
                return  # skip saving since db saving is already done above
            else:
                hklin_dir, hklin_name = os.path.split(self.data_scale)

                if self.json_chosen:
                    sdir = "{}fast_ep".format(self.json_chosen.replace(".json", ""))
                else:
                    sdir = "fast_ep"

                solve_wdir = os.path.join(self.wdir, sdir)

                if not os.path.exists(solve_wdir):
                    os.makedirs(solve_wdir)
                fast_ep = Fast_EP(wdir=solve_wdir,
                                  machines=8,
                                  nproc=4,
                                  hklin_name=hklin_name,
                                  hklin_dir=hklin_dir)
                fast_ep.run()
            '''
            if self.json_all:
                self.parallel_fast_ep_SAD()
            else:
                Solver.fast_ep_sad(self.data_stat)

            return
        else:
            # default, do nothing
            pass

        if ARGS.log2db and self.dbid:
            query = (DataProcessResults
                     .update({DataProcessResults.solve: solve_wdir})
                     .where(DataProcessResults.id == int(self.dbid)))
            query.execute()

    def parallel_fast_ep_SAD(self):
        parallelize(Solver.fast_ep_sad, self.json_all, nproc=len(self.json_all))

    @staticmethod
    def fast_ep_sad(jsonfile):

        if isinstance(jsonfile, str):
            logger.info("anomalous signal detected, run fast_ep for data defined in {}.".format(jsonfile))

            if not jsonfile or not os.path.exists(jsonfile):
                return

            with open(jsonfile) as f:  # only choose last file sorted by wav4, hopefully peak
                data_stat = load(f)

            _, jsonfile_name = os.path.split(jsonfile)
            sdir = "{}fast_ep".format(jsonfile_name.replace(".json", ""))
            wdir = data_stat['proc_dir']
        elif isinstance(jsonfile, dict):
            logger.warning("anomalous signal detected, run fast_ep with dict input.")
            data_stat = jsonfile
            sdir = "fast_ep"
            wdir = data_stat['workdir']
        else:
            logger.warning("something weird happend with the input {}".format(jsonfile))
            return

        if not data_stat:
            return

        data_scale = data_stat['scale_mtz']
        dbid = data_stat['dataprocessresults_id']

        hklin_dir, hklin_name = os.path.split(data_scale)

        solve_wdir = os.path.join(wdir, sdir)

        if not os.path.exists(solve_wdir):
            os.makedirs(solve_wdir)

        fast_ep = Fast_EP(wdir=solve_wdir,
                          machines=8,
                          nproc=4,
                          runner='slurm',
                          hklin_name=hklin_name,
                          hklin_dir=hklin_dir)
        fast_ep.run()

        if ARGS.log2db and dbid:
            logger.info("save fast_ep dir {} to data process results id {}".format(solve_wdir, dbid))
            query = (DataProcessResults
                     .update({DataProcessResults.solve: solve_wdir})
                     .where(DataProcessResults.id == int(dbid)))
            query.execute()
