from __future__ import absolute_import
import os
import logging

from .scriptjob import ScriptJob
from ..file_parser.report import Report
from ..common.utils import timeit
from ..process.get_args import ARGS
from ..config.config import XSCALE_MAXCC12, XDSCC12, XDSSTAT

logger = logging.getLogger(__name__)


class XDSstat(ScriptJob):
    def __init__(self,
                 wdir='.',
                 tag_name=None,
                 highres=None,
                 lowres=None,
                 script_name='xdsstat.sh',
                 script_text=None,
                 nproc=2,
                 runner='local',
                 pe_name='smp'):
        super(XDSstat, self).__init__(wdir,
                                      script_name,
                                      script_text,
                                      nproc,
                                      runner,
                                      pe_name)
        self.wdir = os.path.abspath(wdir)
        if not os.path.isdir(self.wdir):
            os.makedirs(self.wdir)

        self.highres = highres
        self.lowres = lowres
        self.summary_html_name = "XDSsummary.html"
        self.tag_name = tag_name
        if not self.tag_name:
            self.tag_name = "{}/{}".format(ARGS.sampleName, ARGS.pipeline[0])

    def cmd_for_xdsstat(self):
        cmd = list()
        refl = os.path.join(self.wdir, "XDS_ASCII.HKL")
        resol_nbins = 5
        n_combine = 5  # in degree, larger value for smoother appearance

        if os.path.exists(refl):
            res_range = ""
            if self.highres and self.lowres:
                res_range = "{} {}".format(self.lowres, self.highres)
            if self.highres:
                dmax = "-dmax {}".format(self.highres)
            else:
                dmax = ""

            cmd.append("echo XDS_ASCII.HKL | {} {} > XDSSTAT.LP".format(XDSSTAT, res_range))
            cmd.append("{} {} -nbin {} -t {} XDS_ASCII.HKL > XDSCC12.LP".format(XDSCC12, res_range, resol_nbins, n_combine))
            cmd.append("{} {} -nbin {} XDS_ASCII.HKL > MAXCC12.LP".format(XSCALE_MAXCC12, dmax, resol_nbins))

            # create aliases since logview/qtrview only recognizes .log extension
            cmd.append("ln -s XDSSTAT.LP XDSSTAT.LP.log")
        else:
            logger.warning("{} does not exist.".format(refl))

        self.script_text = '\n'.join(cmd)
        return self

    @timeit
    def run(self):
        logger.info("producing xdsstat, xdscc12, and maxcc12 analysis.")
        self.cmd_for_xdsstat().run_sync()

        # generate html report
        r = Report(work_dir=self.wdir,
                   pipelinestatus_id=ARGS.pipelinestatus_id,
                   tag_name=self.tag_name,
                   xdsstat_filename="XDSSTAT.LP",
                   maxcc12_filename="MAXCC12.LP",
                   xdscc12_filename="XDSCC12.LP")
        r.create_report(filename=self.summary_html_name)

    def get_summary_html(self):
        if os.path.exists(os.path.join(self.wdir, self.summary_html_name)):
            return os.path.join(self.wdir, self.summary_html_name)
        else:
            return None