import os
import logging

from ..common.utils import isFloat
from ..common.htmlstr import Htmlstr
from ..file_parser.aimlessparser import AimlessParser
from ..tasks.pointless import Pointless
from ..tasks.aimless import Aimless
from ..tasks.truncate import Truncate
from ..process.get_args import ARGS

logger = logging.getLogger(__name__)


class XDSASCII2MTZ(object):
    def __init__(self,
                 xds_ascii_hkl,
                 wdir=".",
                 sample_name="sampleName",
                 anomalous=False,
                 data_dict=None,
                 user_spacegroup=None,
                 user_highres=None
                 ):
        self.xds_ascii_hkl = xds_ascii_hkl
        self.wdir = wdir
        self.anomalous = anomalous
        self.data_dict = data_dict
        self.user_highres = user_highres
        self.user_spacegroup = user_spacegroup
        self.sampleName = sample_name

        self.base, ext = os.path.splitext(xds_ascii_hkl)
        self.xds_ascii_mtz = self.base + '.mtz'
        self.base += '-'

    def convert(self):
        Pointless(wdir=self.wdir,
                  script_name=self.base + 'xds2mtz.sh').cmd_xds2mtz(
            self.xds_ascii_hkl, mtzout=self.xds_ascii_mtz,
            logfile=self.base + 'convertmtz.log',
            spacegroup=self.user_spacegroup,
        ).run_sync(timeout=600)

        Aimless(hklin=self.xds_ascii_mtz,
                script_name=self.base + 'aimless1.sh',
                wdir=self.wdir,
                anomalous=self.anomalous,
                hklout=self.base + 'aimless-ALLDATA.mtz',
                xmlout=self.base + 'aimless-ALLDATA.xml',
                logfile=self.base + 'aimless-ALLDATA.log'
                ).run_aimless_merge_only()

        # get high res cutoff
        aparser = AimlessParser(wdir=self.wdir, filename=self.base + 'aimless-ALLDATA.log')
        if not aparser.OK:
            logger.warning("Aimless failed, skipped the rest.")
            return

        highres = aparser.get_resolution_estimation(criterion=self.user_highres)
        if isFloat(self.user_highres):
            user_highres = float(self.user_highres)
            if user_highres < highres:
                logger.warning("user requested resolution cutoff {} A "
                               "is higher than that of automatically determined {} A.".format(user_highres, highres))
            highres = user_highres

        # run aimless again with resolution cutoff
        Aimless(hklin=self.xds_ascii_mtz, script_name=self.base + 'aimless2.sh',
                wdir=self.wdir,
                anomalous=self.anomalous,
                hklout=self.base + 'aimless-TRIMDATA.mtz',
                xmlout=self.base + 'aimless-TRIMDATA.xml',
                logfile=self.base + 'aimless-TRIMDATA.log',
                extra_output=True,
                xia2_report=False
                ).run_aimless_merge_only(highres=highres)

        aparser = AimlessParser(wdir=self.wdir, filename=self.base + 'aimless-TRIMDATA.log')
        if not aparser.OK:
            logger.warning("Aimless failed, skipped the rest.")
            return

        # run truncate
        Truncate(wdir=self.wdir,
                 script_name=self.base + 'truncate.sh',
                 hklin=self.base + 'aimless-TRIMDATA.mtz',
                 hklout=self.base + 'truncate-TRIMDATA.mtz',
                 logfile=self.base + 'truncate-TRIMDATA.log').run_ctruncate(async_flag=True)

        # gather aimless statistics
        ap = AimlessParser(wdir=self.wdir, filename=self.base + 'aimless-TRIMDATA.log')
        self.data_dict.update(ap.summarize())

        self.data_dict['state'] = 'DONE'
        self.data_dict['software'] = 'gmcaproc'
        self.data_dict['scale_dir'] = self.wdir
        self.data_dict['scale_mtz'] = "{}/{}".format(self.wdir, self.base + 'aimless-TRIMDATA.mtz')
        self.data_dict['scale_log'] = "{}/{}".format(self.wdir, self.base + 'aimless-TRIMDATA.log')
        self.data_dict['scalepack'] = "{}/{}".format(self.wdir, self.base + 'aimless-TRIMDATA.sca')
        self.data_dict['scalepack_unmerged'] = "{}/{}".format(self.wdir, self.base + 'aimless-TRIMDATA_unmerged.sca')
        self.data_dict['run_stats'] = "{}/{}".format(self.wdir, 'SWEEP_PROCESSING_SUMMARY.html')
        self.data_dict['truncate_log'] = "{}/{}".format(self.wdir, self.base + 'truncate-TRIMDATA.log')
        self.data_dict['truncate_mtz'] = "{}/{}".format(self.wdir, self.base + 'truncate-TRIMDATA.mtz')
        self.data_dict['imageSet'] = ARGS.imageSet
        self.data_dict['run_stats'] = "{}/{}".format(self.wdir, 'SWEEP_PROCESSING_SUMMARY.html')

        # if no sampleName was set, use prefix
        if self.sampleName == 'sampleName' or not self.sampleName:
            self.data_dict['sampleName'] = self.data_dict['prefix']

        if ARGS.log2db:
            from ..log.database import save_to_db
            db = save_to_db(self.get_data_dict())

        # dump json results for later stages
        if db:
            self.data_dict['dataprocessresults_id'] = db.id

        import json
        with open('{}/{}.json'.format(ARGS.proc_dir, self.base), 'w') as fp:
            json.dump(self.data_dict, fp)

    def get_data_dict(self):
        return self.data_dict
