#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Thu Dec  1 08:34:55 2016

@author: qxu
"""
from __future__ import absolute_import
import datetime
import logging
import math
import os

from ..common.spacegroup import Spacegroup
from ..common import utils
from ..common.imagesweep import ImageSweep

from .xdsinp_templates import (MASK_PILATUS_6M,
                               MASK_EIGER_16M,
                               MASK_EIGER2_CDTE_4M,
                               XDS_TEMPLATE,
                               ICE_RINGS)
from .xds_ranges import XDS_ranges
from ..process.get_args import ARGS

logger = logging.getLogger(__name__)


class XDS_INP:
    def __init__(self,
                 wdir=".",
                 filename='XDS.INP',
                 nproc=2,
                 njobs=1,
                 highres=None,
                 native=False,
                 spot_range_mode='firsthalf',
                 xdscmds=None,
                 user_refhkl=None,
                 user_unit_cell=None,
                 user_space_group=None):
        wdir = os.path.abspath(wdir)
        if not os.path.isdir(wdir):
            os.makedirs(wdir)
        self.wdir = wdir
        self.filename = filename
        self.data = []
        self.xdscmds = xdscmds
        self.user_refhkl = user_refhkl

        if user_space_group:
            self.user_space_group = Spacegroup.get_spg_number_from_symbol(user_space_group)
        else:
            self.user_space_group = user_space_group

        self.user_unit_cell = user_unit_cell
        self.highres = highres
        self.native = native
        self.nproc = nproc
        self.njobs = njobs
        self.spot_range_mode = spot_range_mode
        self.ranges = None
        self.edge_res = None
        self.corner_res = None

    def load_from_file(self, from_dir=None, filename='XDS.INP'):
        if not from_dir:
            from_dir = self.wdir

        fname = os.path.join(from_dir, self.filename)
        if os.path.isfile(fname):
            self.data = [l.lstrip(" ") for l in open(fname, "r").readlines()]
        else:
            logger.error("failed to load {}.".format(fname))
        return self

    def create_from_sweep(self, img_sweep):
        if not isinstance(img_sweep, ImageSweep):
            img_sweep = ImageSweep(img_sweep)

        detector_mask = ""
        ice_rings = ICE_RINGS
        sensor_thickness = 0.0
        two_theta = 0.0
        high_res = 0.5
        direction_of_detector_y_axis = '0.0 1.0 0.0'
        additional_keywords = []

        md = img_sweep.get_metadata_as_dict()
        detector_class = md['detector_class'].upper()
        n_profile_g = n_profile_ab = 9
        nbits = 16

        if 'EIGER' in detector_class:
            detector = 'EIGER'
            # additional_keywords.append("STRONG_PIXEL=4")
            if '16M' in detector_class:
                detector_mask = MASK_EIGER_16M
                n_profile_g = n_profile_ab = 21
                nbits = md.get('nbits', 32)
            elif 'CDTE' in detector_class and "4M" in detector_class:
                detector_mask = MASK_EIGER2_CDTE_4M
                n_profile_g = n_profile_ab = 13
                nbits = md.get('nbits', 32)

                from .siganl_CdTe import cal_silicon_CdTe
                wavelength = md.get("wavelength")
                signal = cal_silicon_CdTe(12398.0 / wavelength)
                additional_keywords.append(f"SILICON={signal:.02f}")

        elif 'PILATUS' in detector_class:
            detector = 'PILATUS'
            # additional_keywords.append("GAIN=1.0")
            if '6M' in detector_class:
                detector_mask = MASK_PILATUS_6M
                n_profile_g = n_profile_ab = 13
        elif 'ADSC' in detector_class:
            detector = 'ADSC'
        elif 'MARC' in detector_class:
            detector = 'CCDCHESS'
        else:
            logger.warning("detector not supported. ")
            detector = md['detector_class'].split()[1].upper()

        sensor_thickness = md['sensor']
        overload = md['saturation']
        nx, ny = md['size']
        qx, qy = md['pixel_size']
        orgx, orgy = md['beam_in_pixel']

        if ARGS.beamx:
            orgx = ARGS.beamx

        if ARGS.beamy:
            orgy = ARGS.beamy

        osc_range = md['phi_width']
        distance = md['distance']
        wavelength = md['wavelength']
        start_angle = img_sweep.get_start_angle()

        start, end = img_sweep.get_data_range()
        imgdir = img_sweep.get_absdir()
        template = img_sweep.get_name_template(marker="?")

        xds_ranges = XDS_ranges(start, end, osc_range, delta=5.0)
        if img_sweep.is_screening():
            spot_ranges = img_sweep.get_xds_spot_ranges_for_screening()
            bg_range = spot_ranges[0]
        else:
            if "!SPOT_RANGE=DATA_RANGE" in self.xdscmds:
                self.spot_range_mode = "full"
            elif "!SPOT_RANGE=3DEGREE_EVERY_15DEGREE" in self.xdscmds:
                self.spot_range_mode = "standard"
            elif "!SPOT_RANGE=3_WEDGES_OF_5IMAGES" in self.xdscmds:
                self.spot_range_mode = "sparse"

            spot_ranges = xds_ranges.set_spot_ranges(mode=self.spot_range_mode)
            bg_range = xds_ranges.get_bg_range()

        spot_ranges = '\n'.join(['SPOT_RANGE= {} {}'.format(s, e)
                                 for s, e in spot_ranges])
        bg_range = "{} {}".format(*bg_range)
        data_range = "{} {}".format(start, end)
        start_frame = start

        self.ranges = xds_ranges

        if img_sweep.is_hdf5():
            from ..config.config import XDS_HDF5_PLUGIN
            if os.path.isfile(XDS_HDF5_PLUGIN):
                additional_keywords.append("LIB={}".format(XDS_HDF5_PLUGIN))

        if 'two_theta' in md:
            two_theta = md['two_theta'] * math.pi / 180.0
            cost = math.cos(two_theta)
            sint = math.sin(two_theta)
            direction_of_detector_y_axis = "0. {:.6f} {:.6f}".format(cost, sint)

        corner_res = md['corner_res']
        edge_res = md['edge_res']

        nbins = 20 if edge_res < 2.0 else 15

        if self.highres and utils.isFloat(self.highres):
            high_res = self.highres
            resolution_shells = " ".join(
                '{:.2f}'.format(r) for r in utils.get_resolution_bins(float(self.highres), nbins))
        else:
            high_res = '{:.2f}'.format(corner_res)
            resolution_shells = " ".join('{:.2f}'.format(r) for r in utils.get_resolution_bins(corner_res, nbins))

        if self.user_refhkl:
            spindle_range = "5 25 5"
        else:
            spindle_range = "10 190 10"

        xdsinp = XDS_TEMPLATE.format(detector=detector,
                                     sensor_thickness=sensor_thickness,
                                     nx=nx, ny=ny, qx=qx, qy=qy,
                                     overload=overload,
                                     osc_range=osc_range,
                                     orgx=orgx, orgy=orgy,
                                     distance=distance,
                                     wavelength=wavelength,
                                     start_angle=start_angle,
                                     spot_ranges=spot_ranges,
                                     bg_range=bg_range,
                                     data_range=data_range,
                                     start_frame=start_frame,
                                     template=imgdir + "/" + template,
                                     high_res=high_res,
                                     n_profile_ab=n_profile_ab,
                                     n_profile_g=n_profile_g,
                                     resolution_shells=resolution_shells,
                                     direction_of_detector_y_axis=direction_of_detector_y_axis,
                                     ice_rings=ice_rings,
                                     total_spindle_rot_ranges=spindle_range,
                                     detector_mask=detector_mask,
                                     additional="\n".join(additional_keywords))
        self.data = xdsinp.split('\n')

        self.update('MAXIMUM_NUMBER_OF_JOBS={};MAXIMUM_NUMBER_OF_PROCESSORS={}'.format(self.njobs,
                                                                                       self.nproc))

        if self.user_unit_cell and self.user_space_group:
            self.update('SPACE_GROUP_NUMBER={};UNIT_CELL_CONSTANTS={}'.format(self.user_space_group,
                                                                              self.user_unit_cell))

        if self.native:
            logger.info("XDS assumes that I+=I-.")
            self.update("FRIEDEL'S_LAW=TRUE")

        # add reference data for strategy and correct purpose, if user provides it
        if self.user_refhkl:
            if os.path.isfile(self.user_refhkl):
                self.update("REFERENCE_DATA_SET={}".format(self.user_refhkl))
            else:
                logger.warning("reference data {} does not exist, ignored.".format(self.user_refhkl))

        # probably need to check the cmds are valid, otherwise, wrong cmds can screw things up
        if self.xdscmds:
            self.update(self.xdscmds)

        if end - start < 5:
            self.update("INDEX_ERROR=0.2")

        # try to modify DELPHI based on njobs, nproc
        try:
            import psutil

            total_osc = float(osc_range) * float(end - start + 1)
            delphi = round(total_osc / self.njobs, 1)
            if delphi >= 5.0:
                delphi = 5
            elif delphi <= 1.5:
                delphi = 1.5

            # reduce delphi is not a good idea
            delphi = 5.0

            self.update(f"DELPHI={delphi}")

            nimg_cached = int(delphi / osc_range) + 1
            imgsize = nx * ny * nbits / 8  # bytes
            total_cache_mem = imgsize * nimg_cached * self.nproc
            total_mem = psutil.virtual_memory().total
            logger.debug(
                f"delphi = {delphi}, osc = {osc_range}, nimg_cached = {nimg_cached}, imgsize = {imgsize}, "
                f"total_cache_mem = {total_cache_mem}, total_mem = {total_mem}")

            if total_cache_mem >= 0.6 * total_mem:
                logger.warning("caching will use too much memory, disable caching by sacrificing speed.")
                self.update("NUMBER_OF_IMAGES_IN_CACHE=0")
        except ImportError:
            pass

        return self

    def cleanup(self):
        data_clean = []
        for line in self.data:
            line = line.lstrip(" ")
            i = line.find('!')
            if i > 0:
                data_clean.append(line[0:i])
            elif i < 0:
                data_clean.append(line)

        self.data = data_clean
        return self

    def delete(self, key):
        # remove a key from XDS.INP
        key = key.upper()
        for i, line in enumerate(self.data):
            line = line.lstrip()
            if line.startswith(key):
                self.data[i] = "!" + self.data[i]

    def update(self, Str: object) -> object:
        # NB: key value mechanism has an issue dealing with SPOT_RANGE since it can be specified multiple times
        # update XDS.inp using str with the following format: "Key=Value;..."
        if not Str:
            return self

        items = [i.strip(" ") for i in Str.split(";")]
        for i in items:
            key, value = map(str.strip, i.split("="))
            self.delete(key)

        for i in items:
            self.data.append(i)
        return self

    def save(self, sep="", saveBackup=False, dest_dir='.', dest_fname='XDS.INP'):
        filename = self.filename
        dest_dir = os.path.abspath(dest_dir)
        if not os.path.isdir(dest_dir):
            os.makedirs(dest_dir)

        fname = os.path.join(dest_dir, dest_fname)
        if os.path.exists(fname) and saveBackup:
            # check versions and make backups
            timeStamp = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
            newname = filename + "." + timeStamp
            os.rename(fname, os.path.join(dest_dir, newname))

        with open(fname, "w") as fh:
            # fh.write(sep.join(self.lines))
            for line in self.data:
                if line.endswith("\n"):
                    fh.write(line.lstrip(" "))
                else:
                    fh.write(line.lstrip(" ") + "\n")

    def __str__(self):
        return "XDS.INP: \n" + "".join(self.data)
