#!/usr/bin/perl -W
#--------------------------------------------------------------
# These are SCAN RECORD related subs for beamline controls Perl scripts
#
#     				Author: Sergey Stepanov
#
# Version-1.0:  2001/02/05 using Pezca
# Version-2.0:  2006/02/05 added fast scans
# Version-2.1:  2007/02/05 added fast scans with PV list
# Version-2.2:  2007/02/05 added fast scans with mca and PV list
# Version-2.3:  2008/02/01 added step scans with motor record motors
# Version-2.4:  2008/09/29 added fly scans with Hytec DAC
# Version-2.5:  2011/01/10 added flag to DAC scans to watch for overload
# Version-2.6:  2012/01/31 changed the method of scan record aborting: removed scanAbort PV and added waiting.
# Version-2.7:  2013/02/06 added setupSRbeforeScan ($scanrec, $pv, $val, ['nowait|wait']);
#                          Changed threating overload in DAC scans: always report.
# Version-2.8:  2013/05/26 changed all SRscan routines to use $SRSCAN_FLAGS instead of $SRscan2ndTime
# Version-2.9:  2013/07/13 changed calculation of middle point for on-the-fly scans (because scanrec stopped recording the 1st point)
# Version-3.0:  2013/08/31 added optional DAC check (ADC "encoder") to SRflyScanDAC
# Version-3.1:  2014/03/28 changed calls of many functions from @ to \@ (i.e. to passing arrays pointers) to satisfy Perl-5.16
# Version-3.2:  2017/07/21 fixes for findPosSR (when post-reading does not confirm good points.
# Version-3.3:  2018/11/16 added driving to Gaussian and Lorentzian to SRflyScanDAC
#                                (needs PDL, PDL::Fit::LM and peakFit from common_PDL.pl).
#                                Also moved all after DAC scan operations to DAC_afterScan (see dac_subs.pl).
# Version-3.4:  2021/06/06 added &setupSRtriggers
#--------------------------------------------------------------
# List of functions:
# ^^^^^^^^^^^^^^^^^
#  1. $maxPts = &SRmaxPts($scanrec);
#  2.           &clearSR ($scanrec,[$mode]);					# $mode=[0-5], default=3 (clear all PVs and flags)
#  3.           &setupSRmotors ($scanrec, $numPts,
#                               $numMotors, \@actPosPV, \@rqsPosPV, \@begin, \@end);
#  4.           &setupSRmotorsMore ($scanrec, $numPts,
#                               $scanMode, $settleTime, $afterScan, $numMotors,
#                               \@actPosPV, \@rqsPosPV, \@begin, \@end, [$absRel]);
#  5.           &changeSRmotors ($scanrec, $numPts, $numMotors, \@begin, \@end);
#  6.           &changeSRmotorLIN ($scanrec, $numPts, $motorNo, $begin, $end);
#  7.           &changeSRmotorTBL ($scanrec, $numPts, $motorNo, \@posArray);
#  8.           &setupSRmotorsCommon ($scanrec, $numPts, $settleTime, $afterScan);
#  9.           &setupSRmotorLIN ($scanrec, $motorNo, $actPosPV, $rqsPosPV, $begin, $end, [$scanMode]);
# 10.           &setupSRmotorTBL ($scanrec, $motorNo, $actPosPV, $rqsPosPV, \@posArray);
# 11.           &setupSRstepmotors ($scanrec, $numPts,
#                               $numMotors, \@motors, \@begin, \@end);
# 12.           &setupSRstepmotorsMore ($scanrec, $numPts,
#                               $scanMode, $settleTime, $afterScan,
#                               $numMotors, \@motors, \@begin, \@end);
# 13.           &setupSRtriggers ($scanrec, \@triggerPV, [\@triggerVAL, $settleTime, $autowaitNr]);
# 14. *OLD*     &setupSRdetectors ($scanrec, $joerger, \@IntyFlag);             #returns global $numDetectors, channels[]
# 15.           &setupSRdetectors_v2 ($scanrec, $joerger, $ninputs, \@inputs,
#                                    [$triggerPV, $triggerVAL, $settleTime]);   #does not return anything
# 16.@SRdetec = &setupSRdetectors_auto ($scanrec, $joerger, $ninputs, \@inputs,
#                                    [$triggerPV, $triggerVAL, $settleTime]);   #preserves SR detectors & adds new as needed
# 17.           &setupSRdetectors_pv ($scanrec, $joerger, \@inputs, [\@pvlist],
#                                    [$triggerPV, $triggerVAL, $settleTime]);   #does not return anything
# 18. $status = &setupSRdetectors_addpvs ($scanrec, \@pvlist, [$firstdetector]);#returns status
# 19.           &setupSRdetectors_mca ($scanrec, $mca, \@ROIs, [\@pvlist],
#                                    [$triggerPV, $triggerVAL, $settleTime]);   #does not return anything
# 20.           &setupSRdetectors_scaler_mca ($scanrec, $joerger, $mca,
#                                             \@inputs, \@ROIs,
#                                    [$triggerPV, $triggerVAL, $settleTime]);   #does not return anything
# 21.           &setupSRdetectors_scaler_mca_pv ($scanrec, $joerger, $mca,
#                                             \@inputs, \@ROIs, [\@pvlist],
#                                     [$triggerPV, $triggerVAL, $settleTime]);  #does not return anything
# 22. $status = &setupSRbeforeScan ($scanrec, $bsPV, $bsVAL, [$waitFlag);	#returns status
# 23. $status = &setupSRafterScan ($scanrec, $asPV, $asVAL, [$waitFlag]);	#returns status
# 24.           &setupSRrefreshrate ($scanrec, [$rate]); 			#for refreshing .P${N}CA and .D${NN}CA
# 25.           &startSR ($scanrec,[$scanAbort]);
# 26. $currPts= &getSRcurrPts($scanrec);                                        #read current execution point
# 27.           &abortSR ($scanAbort,[$scanrec,$nPress]);                       #default $nPress=2. scanAbortPV is different from scanrec!
# 28. *ERR*     &abortSR1 ($scanrec,[$nPress,waitflag]);                        #default $nPress=2. This function does not work!
# 29. $status = &waitSRabort ($scanrec);		                        #wait up to 10s until SR is aborted
# 30. $actPts = &waitSRdata ($scanrec);
# 31.           &readSRdata ($scanrec, $numPts, $numMotors, $numDetectors);     #returns global x1[]..x$numMotors[] and y1[]..y$numDetectors[]
# 32. @x      = &readSRx ($scanrec, $numPts, $motorNumber, [$final-or-progess])
# 33. @y      = &readSRy ($scanrec, $numPts, $scalerNumber, [$final-or-progess]);
# 34. @x      = &calcSRxMiddle ($numPts, $begin, @x);				#2013/07: added new argument $begin as scanrec changed
# 35. $xcurr  = &readSRx_curr ($scanrec, $motorNumber)
# 36. $ycurr  = &readSRy_curr ($scanrec, $scalerNumber);
# 37.           &openSRmedm    ($scanrec, $ninputs, \@inputs, [$type]);         #first SR detectors, scaler only
# 38.           &openSRmedm_detec ($scanrec, $ninputs, \@inputs,
#                                                     \@SRdetectors, [$type]);  #any SR detectors, scaler only
# 39.           &openSRmedm_v2 ($scanrec, $ndetectors, [$type]);                #first SR detectors, reads labels from SR
# 40.           &openSRmedm_pv ($scanrec, \@inputs, \@pvlist, [$type]);         #faster, scaler+pvlist
# 41.           &openSRmedm_mca ($scanrec, \@ROIs, \@pvlist, [$type]);          #faster, ROI+pvlist
# 42.           &openSRmedm_scaler_mca ($scanrec, \@inputs, \@ROIs,
#                                                          \@pvlist, [$type]);  #faster, scaler+ROI+pvlist
#
# On-the-fly scans:
# ^^^^^^^^^^^^^^^^
# 43. ($status,$actPts) =
#               &SRflyScan  ($assy,$axis,$begin,$end,$npts,$scanTime,
#                            $medm,$scaler,$fileBkg,$ninputs,@inputs);
#     (Also returns GLOBALS: @x1SR,@y1SR,...@y{$ninputs}SR and @Offsets)
#     -- on-the-fly scan with SCAN RECORD, PMAC motors and Joerger scaler
#
# 44. ($status,$actPts) =
#               &SRflyScan2 ($assy,$axis,$begin,$end,$npts,$scanTime,
#                            $medm,$scaler,$fileBkg,\@inputs,[\@pvlist,
#                            $afterScan,$checkOverload,$postarrayPV,$post_interval]);
#     (Also returns GLOBALS: @x1SR,@y1SR,...@y{$ninputs+$npvs}SR and @Offsets)
#     -- on-the-fly scan with SCAN RECORD, PMAC motors and Joerger scaler + PVlist
#
# 45. ($status,$actPts) =
#               &SRflyScanMCA ($assy,$axis,$begin,$end,$npts,$scanTime,
#                              $medm,$mca,\@ROIs,[\@pvlist]);
#     (Also returns GLOBALS: @x1SR,@y1SR,...@y{$nROIs+$npvs}SR)
#     -- on-the-fly scan with SCAN RECORD, PMAC motors and MCA ROI + PVlist
#
# 46. ($status,$actPts) =
#               &SRflyScanJoergerMCA ($assy,$axis,$begin,$end,$npts,$scanTime,
#                                     $medm,$scaler,$fileBkg,$mca,\@inputs,\@ROIs,[\@pvlist]);
#     (Also returns GLOBALS: @x1SR,@y1SR,...@y{$ninputs+$nROIs+$npvs}SR and @Offsets)
#     -- on-the-fly scan with SCAN RECORD, PMAC motors and (Joerger inputs + MCA ROI + PVlist)
#
# 47.  ($status,$actPts) =
#               &SRflyScanOMS58 ($motor,$begin,$end,$npts,$scanTime,
#                                $medm,$scaler,$fileBkg,$ninputs,@inputs);
#     (Also returns GLOBALS: @x1SR,@y1SR,...@y{$ninputs}SR and @Offsets)
#     -- on-the-fly scan with SCAN RECORD, OMS58 (MotorRec) motors and Joerger scaler
#
# 48. ($status,$dac_slope) = &SRflyScanDAC ($dacPV,$begin,$end,$npts,$scanTime,$afterScan,
#                                           $medm,$scaler,$fileBkg,\@inputs,
#                                          [\@pvlist,$checkOverload,
#                                           $postarrayPV,$post_interval,
#                                           $adc_encoder,$dac_tolerance]);
#     (Also returns GLOBALS: @x1SR,@y1SR,...@y{$ninputs+$npvs}SR and @Offsets)
#     -- on-the-fly scan with SCAN RECORD, DAC and (Joerger inputs + PVlist)
#
#
#    STEP scans:
# ^^^^^^^^^^^^^^^^
# 49. ($status,$actPts) =
#               &SRstepScan ($assy,$axis,$begin,$end,$npts,$cntTime,
#                            $medm,$scaler,$cntStartPV,$fileBkg,
#                            \@inputs,[\@pvlist,$settleTime]);
#     (Also returns GLOBALS: @x1SR,@y1SR,...@y{$ninputs+$npvs}SR)
#     -- step scan with SCAN RECORD, PMAC motors and Joerger inputs + PVlist; ATTN: Requires Tim Mooney correction to PMAC DB
#
# 50. ($status,$actPts) =
#               &SRtableScan ($assy,$axis,\@positions,$cntTime,
#                            $medm,$scaler,$cntStartPV,$fileBkg,
#                            \@inputs,[\@pvlist,$settleTime]);
#     (Also returns GLOBALS: @x1SR,@y1SR,...@y{$ninputs+$npvs}SR)
#     -- step scan with SCAN RECORD, PMAC motors and Joerger inputs + PVlist; ATTN: Requires Tim Mooney correction to PMAC DB
#
# 51. ($status,$actPts) =
#               &SRstepScanMCA ($assy,$axis,$begin,$end,$npts,$cntTime,
#                               $medm,$mca,$cntStartPV,
#                               \@ROIs,[\@pvlist,$settleTime]);
#     (Also returns GLOBALS: @x1SR,@y1SR,...@y{$nROIs+$npvs}SR)
#     -- step scan with SCAN RECORD, PMAC motors and MCA ROI + PVlist; ATTN: Requires Tim Mooney correction to PMAC DB
#
# 52. ($status,$actPts) =
#               &SRtableScanMCA ($assy,$axis,\@positions,$cntTime,
#                               $medm,$mca,$cntStartPV,
#                               \@ROIs,[\@pvlist,$settleTime]);
#     (Also returns GLOBALS: @x1SR,@y1SR,...@y{$nROIs+$npvs}SR)
#     -- step scan with SCAN RECORD, PMAC motors and MCA ROI + PVlist; ATTN: Requires Tim Mooney correction to PMAC DB
#
# 53. ($status,$actPts) =
#               &SRstepScanOMS58 ($motor,$begin,$end,$npts,
#                                   $medm,$scaler,$cnt_time,$fileBkg,
#                                   $ninputs,\@inputs,[$settleTime]);
#     -- STEP SCAN with SCAN RECORD, OMS58 (MotorRec) motors and Joerger inputs
#
# 54. $status = &SRstepScanDAC ($dacPV,$begin,$end,$npts,$cntTime,
#                              $medm,$scaler,$fileBkg,
#                              \@inputs,[\@pvlist,$settleTime,$checkOverload,
#                              $postarrayPV,$post_interval,$scanindex,
#                              $adc_encoder, $dac_tolerance]);
#     (Also returns GLOBALS: @x1SR,@y1SR,...@y{$ninputs}SR and @Offsets)
#     -- STEP SCAN with SCAN RECORD, DAC and (Joerger inputs + PVlist)
#
#
#  Find-on-the-fly scans:
# ^^^^^^^^^^^^^^^^^^^^^^
# 55. ($status,$actPts,$pos) =
#               &findPosSR ($assy,$axis,$begin,$end,$npts,$scanTime,
#                           $medm,$scaler,$fileBkg,$input,$target,$direction,[\%flags]);
#     (Also returns GLOBALS: @x1SR,@y1SR, and @Offsets)
#     -- on-the-fly scan with SCAN RECORD, PMAC motors and Joerger scaler; search target signal level on the fly.
#--------------------------------------------------------------
# use Pezca;
# use List::Util qw(min max);
# use warnings::unused;
  use strict;
# no strict 'vars';
# no strict 'subs';
# no strict 'refs';

#=================================================================

# sub SRmaxPts ($scanrec) {
sub SRmaxPts ($) {

### Returns maximum number of points allowed by scanrec

  my ($script, $scanrec, $status, $maxPts);

  $script  = 'SRmaxPts';
  $scanrec = $_[0];
  $status  = 0;
  $maxPts  = 0;

  ($status,$maxPts) = Pezca::GetDouble($scanrec.'.MPTS');
  &report_only($script, $status, $scanrec.'.MPTS');
  if (!defined $maxPts) {$maxPts = 0;}
  return $maxPts;
}

#=================================================================

# sub clearSR ($scanrec,[$mode]) {
sub clearSR ($;$) {

### Clears fields of scan record depending on $mode

  my ($script, $scanrec, $mode, $mode_default, $status);

  $script  = 'clearSR';
  $scanrec = shift(@_);
  $mode    = shift(@_);		#0=clear state message (SMSG)
                		#1=exec dry run
                		#2=check limits
                		#3=clear all PVs, freeze flags, etc. -- DEFAULT
                		#4=clear all positioner-related PVs and flags
                		#5=clear positioner-name PVs
  $mode_default = 3;
  if (!defined $mode) {$mode=$mode_default;}
  if ($mode < 0 || $mode > 5) {
     print STDOUT '*** '.$script.': incorrect clearing mode='.$mode.' for scanrec='.$scanrec."\n",
                  '*** '.$script.': using default mode='.$mode_default.' (clear ALL)'."\n";
     $mode = $mode_default;
  }

### Clear scan record parameters depending on mode:
  $status = Pezca::PutDouble($scanrec.'.CMND',$mode);
  if ($status) {
     print STDOUT '*** '.$script.': failed clearing scan record, mode='.$mode."\n";
     &report_exit($script, $status, $scanrec.'.CMND');
  }
}

#=================================================================

# sub setupSRmotors ($scanrec, $numPts, $numMotors, \@actPosPV, \@rqsPosPV, \@begin, \@end) {
sub setupSRmotors ($$$\@\@\@\@) {

### This function setups SCAN RECORD positioners for *on-the-fly* scans

  my ($scanrec, $numPts, $numMotors, $actPVref, $rqsPVref, $beginRef, $endRef);
  my ($scanMode, $settleTime, $afterScan, $absRel);

  $scanrec    = shift(@_);	# scan record name (PV)
  $numPts     = shift(@_);	# should we add +1 here, like in C subs?
  $numMotors  = shift(@_);
  $actPVref   = shift(@_);
  $rqsPVref   = shift(@_);
  $beginRef   = shift(@_);
  $endRef     = shift(@_);
  $scanMode   = 2;        	# 0=linear, 1=table 2=on-the-fly)
  $settleTime = 0.0;       	# positioner settle time (delay)
  $afterScan  = 0;        	# 0=stay, 1=startpos, 2=priorpos 3=peakpos 4=valleypos 5=+edgepos 6=-edgepos
  $absRel     = 0;		# 0=abs, 1=rel

### Here @_= $numMotors, @actPosPV, @rqsPosPV, @begin, @end;
  &setupSRmotorsMore ($scanrec, $numPts, $scanMode, $settleTime, $afterScan,
                      $numMotors, $actPVref, $rqsPVref, $beginRef, $endRef, $absRel);
}

#=================================================================

# sub setupSRmotorsMore ($scanrec, $numPts, $scanMode, $settleTime, $afterScan,
#                        $numMotors, \@actPosPV, \@rqsPosPV, \@begin, \@end, [$absRel]) {
sub setupSRmotorsMore ($$$$$$\@\@\@\@;$) {

### This function setups SCAN RECORD positioners for step/fly type of scan

  my ($script, $scanrec, $numPts, $scanMode, $settleTime, $afterScan);
  my ($numMotors, $actPVref, $rqsPVref, $beginRef, $endRef, $absRel);
  my ($nActPV, $nRqsPV, $nBegin, $nEnd);
  my ($status, $maxPts, $i, $j, $current_time, $msg);

  $script     = 'setupSRmotorsMore';
  $scanrec    = shift(@_);      #0: scan record name (PV)
  $numPts     = shift(@_);      #1: should we add +1 here, like in C subs?
  $scanMode   = shift(@_);      #2: 0=linear, 1=table 2=on-the-fly
  $settleTime = shift(@_);      #3: positioner settle time (delay)
  $afterScan  = shift(@_);      #4: 0=stay, 1=startpos, 2=priorpos 3=peakpos 4=valleypos 5=+edgepos 6=-edgepos
  $numMotors  = shift(@_);      #5: number of motors
  $actPVref   = shift(@_);
  $rqsPVref   = shift(@_);
  $beginRef   = shift(@_);
  $endRef     = shift(@_);
  $absRel     = shift(@_);      if (!defined $absRel) {$absRel=0;}       #absolute

  $nActPV = @$actPVref;
  $nRqsPV = @$rqsPVref;
  $nBegin = @$beginRef;
  $nEnd   = @$endRef;

  if ($scanMode != 0 && $scanMode != 2) {
     $current_time = scalar localtime;
     $msg  = '*** '.$current_time."\n";
     $msg .= '*** '.$script.': for scanrec='.$scanrec
                   .' incorrect scan mode = '.$scanMode."\n";
     &press_key_exit($msg);
  }

  ($status,$maxPts) = Pezca::GetDouble($scanrec.'.MPTS');
  &report_exit($script, $status, $scanrec.'.MPTS');
  if ($numPts > $maxPts) {
     $current_time = scalar localtime;
     $msg  = '*** '.$current_time."\n";
     $msg .= '*** '.$script.': for scanrec='.$scanrec
                   .' requested npts='.$numPts
                   .' exceeds buffer='.$maxPts."\n";
     &press_key_exit($msg);
  }

  if ($numMotors > $nActPV ||
      $numMotors > $nRqsPV ||
      $numMotors > $nBegin ||
      $numMotors > $nEnd) {
     $current_time = scalar localtime;
     $msg  = '*** '.$current_time."\n";
     $msg .= '*** '.$script.': for scanrec='.$scanrec
                   .' requested motors='.$numMotors
                   .' exceeds one of the arrays lengths:'."\n"
                   .' @ActPosPV='.$nActPV
                   .' @RqsPosPV='.$nRqsPV
                   .' @begin='.$nBegin
                   .' @end='.$nEnd."\n";
     &press_key_exit($msg);
  }

### Upload $numPts:
  $status = Pezca::PutDouble($scanrec.'.NPTS',$numPts);
  &report_exit($script, $status, $scanrec.'.NPTS');

  for ($i=1; $i<=$numMotors; $i++) {  #=====================================
     $j = $i - 1;
### Upload @actPosPV:
     if (defined @$actPVref[$j]) {
        $status = Pezca::PutString($scanrec.'.R'.$i.'PV', @$actPVref[$j]);
        &report_exit($script, $status, $scanrec.'.R'.$i.'PV');
     }
### Upload @rqsPosPV:
     if (defined @$rqsPVref[$j]) {
        $status = Pezca::PutString($scanrec.'.P'.$i.'PV', @$rqsPVref[$j]);
        &report_exit($script, $status, $scanrec.'.P'.$i.'PV');
     }
### Upload @begin:
     if (defined @$beginRef[$j]) {
        $status = Pezca::PutDouble($scanrec.'.P'.$i.'SP', @$beginRef[$j]);
        &report_exit($script, $status, $scanrec.'.P'.$i.'SP');
     }
### Upload @end:
     if (defined @$endRef[$j]) {
        $status = Pezca::PutDouble($scanrec.'.P'.$i.'EP', @$endRef[$j]);
        &report_exit($script, $status, $scanrec.'.P'.$i.'EP');
     }

### Upload scan mode (0=linear, 1=table, 2=on-the-fly):
     $status = Pezca::PutDouble($scanrec.'.P'.$i.'SM',$scanMode);
     &report_exit($script, $status, $scanrec.'.P'.$i.'SM');
### Upload abs|rel=absolute
### (0=absolute, 1-relative)
     if ($absRel ne '1') {$absRel = 0;}
     $status = Pezca::PutDouble($scanrec.'.P'.$i.'AR',$absRel);
     &report_exit($script, $status, $scanrec.'.P'.$i.'AR');
  }  #======================================================================

### Upload after_scan=stay_at_end_pos
### (0=stay, 1=startpos, 2=priorpos 3=peakpos 4=valleypos 5=+edgepos 6=-edgepos).
  if (defined $afterScan) {
     $status = Pezca::PutDouble($scanrec.'.PASM',$afterScan);
     &report_exit($script, $status, $scanrec.'.PASM');
  }

### Upload positioner_delay=0 (settling time)
  if (defined $settleTime) {
     $status = Pezca::PutDouble($scanrec.'.PDLY',$settleTime);
     &report_exit($script, $status, $scanrec.'.PDLY');
  }

### Tell scanrec that we want to use 1 reference detector:
  $status = Pezca::PutDouble($scanrec.'.REFD',1);
  &report_exit($script, $status, $scanrec.'.REFD');
}

#=================================================================

# sub changeSRmotors ($scanrec, $numPts, $numMotors, \@begin, \@end) {
sub changeSRmotors ($$$\@\@) {

### This function changes scan range in SCAN RECORD positioners:

  my ($script, $scanrec, $numPts, $numMotors);
  my ($beginRef, $endRef, $nBegin, $nEnd);
  my ($status, $maxPts, $i, $j, $current_time, $msg);

  $script    = 'changeSRmotors';
  $scanrec    = shift(@_);      #0: scan record name (PV)
  $numPts     = shift(@_);      #1: should we add +1 here, like in C subs?
  $numMotors  = shift(@_);      #3: number of motors
  $beginRef   = shift(@_);
  $endRef     = shift(@_);

  ($status,$maxPts) = Pezca::GetDouble($scanrec.'.MPTS');
  &report_exit($script, $status, $scanrec.'.MPTS');
  if ($numPts > $maxPts) {
     $current_time = scalar localtime;
     $msg  = '*** '.$current_time."\n";
     $msg .= '*** '.$script.': for scanrec='.$scanrec.' requested npts='
                           .$numPts.' exceeds buffer='.$maxPts."\n";
     &press_key_exit($msg);
  }

  $nBegin = @$beginRef;
  $nEnd   = @$endRef;
  if ($numMotors > $nBegin || $numMotors > $nEnd) {
     $current_time = scalar localtime;
     $msg  = '*** '.$current_time."\n";
     $msg .= '*** '.$script.': for scanrec='.$scanrec
                   .' requested motors='.$numMotors
                   .' exceeds one of the arrays lengths:'."\n"
                   .' @begin='.$nBegin
                   .' @end='.$nEnd."\n";
     &press_key_exit($msg);
  }

### Upload $numPts:
  $status = Pezca::PutDouble($scanrec.'.NPTS',$numPts);
  &report_exit($script, $status, $scanrec.'.NPTS');

  for ($i=1; $i<=$numMotors; $i++) {
     $j = $i - 1;
### Upload @begin:
     if (defined @$beginRef[$j]) {
        $status = Pezca::PutDouble($scanrec.'.P'.$i.'SP', @$beginRef[$j]);
        &report_exit($script, $status, $scanrec.'.P'.$i.'SP');
     }
### Upload @end:
     if (defined @$endRef[$j]) {
        $status = Pezca::PutDouble($scanrec.'.P'.$i.'EP', @$endRef[$j]);
        &report_exit($script, $status, $scanrec.'.P'.$i.'EP');
     }
  }
}

#=================================================================

# sub changeSRmotorLIN ($scanrec, $numPts, $motorNo, $begin, $end) {
sub changeSRmotorLIN ($$$$$) {

### This function changes scan range in SCAN RECORD positioners:

  my ($script, $scanrec, $numPts, $motorNo, $begin, $end);
  my ($status, $maxPts, $current_time, $msg);

  $script    = 'changeSRmotorLIN';
  $scanrec   = $_[0];
  $numPts    = $_[1];		# should we add +1 here, like in C subs?
  $motorNo   = $_[2];
  $begin     = $_[3];
  $end       = $_[4];

  ($status,$maxPts) = Pezca::GetDouble($scanrec.'.MPTS');
  &report_exit($script, $status, $scanrec.'.MPTS');
  if ($numPts > $maxPts) {
     $current_time = scalar localtime;
     $msg  = '*** '.$current_time."\n";
     $msg .= '*** '.$script.': for scanrec='.$scanrec.' requested npts='
                           .$numPts.' exceeds buffer='.$maxPts."\n";
     &press_key_exit($msg);
  }

  if ($motorNo < 1 || $motorNo > 4) {
     $current_time = scalar localtime;
     $msg  = '*** '.$current_time."\n";
     $msg .= '*** '.$script.': for scanrec='.$scanrec.' requested motorNo='
                           .$motorNo.' is not in the 1-4 range'."\n";
     &press_key_exit($msg);
  }

### Upload $numPts:
  $status = Pezca::PutDouble($scanrec.'.NPTS',$numPts);
  &report_exit($script, $status, $scanrec.'.NPTS');

### Upload $begin:
  $status = Pezca::PutDouble($scanrec.'.P'.$motorNo.'SP',$begin);
  &report_exit($script, $status, $scanrec.'.P'.$motorNo.'SP');
### Upload $end:
  $status = Pezca::PutDouble($scanrec.'.P'.$motorNo.'EP',$end);
  &report_exit($script, $status, $scanrec.'.P'.$motorNo.'EP');
}

#=================================================================

# sub changeSRmotorTBL ($scanrec, $numPts, $motorNo, \@positions) {
sub changeSRmotorTBL ($$$\@) {

### This function changes scan range in SCAN RECORD positioners:

  my ($script, $scanrec, $numPts, $motorNo, $positionsRef);
  my ($status, $maxPts, $current_time, $msg);

  $script       = 'changeSRmotorTBL';
  $scanrec      = $_[0];
  $numPts       = $_[1];	# should we add +1 here, like in C subs?
  $motorNo      = $_[2];
  $positionsRef = $_[3];

  ($status,$maxPts) = Pezca::GetDouble($scanrec.'.MPTS');
  &report_exit($script, $status, $scanrec.'.MPTS');
  if ($numPts > $maxPts) {
     $current_time = scalar localtime;
     $msg  = '*** '.$current_time."\n";
     $msg .= '*** '.$script.': for scanrec='.$scanrec.' requested npts='
                           .$numPts.' exceeds buffer='.$maxPts."\n";
     &press_key_exit($msg);
  }

  if ($motorNo < 1 || $motorNo > 4) {
     $current_time = scalar localtime;
     $msg  = '*** '.$current_time."\n";
     $msg .= '*** '.$script.': for scanrec='.$scanrec.' requested motorNo='
                           .$motorNo.' is not in the 1-4 range'."\n";
     &press_key_exit($msg);
  }

### Upload $numPts:
  $status = Pezca::PutDouble($scanrec.'.NPTS',$numPts);
  &report_exit($script, $status, $scanrec.'.NPTS');

### Upload $positions:
  $status = Pezca::PutList($scanrec.'.P'.$motorNo.'PA','DOUBLE',@$positionsRef);
  &report_exit($script, $status, $scanrec.'.P'.$motorNo.'PA');
}

#=================================================================

# sub setupSRmotorsCommon ($scanrec, $numPts, $settleTime, $afterScan)
sub setupSRmotorsCommon ($$$$) {

### This function setups common positioners fields in SCAN RECORD for *any* type of scan

  my ($script, $scanrec, $numPts, $settleTime, $afterScan);
  my ($status, $maxPts, $current_time, $msg);

  $script     = 'setupSRmotorsCommon';
  $scanrec    = shift(@_);	#0: scan record name (PV)
  $numPts     = shift(@_);	#1: should we add +1 here, like in C subs?
  $settleTime = shift(@_);	#3: positioner settle time (delay)
  $afterScan  = shift(@_);	#4: 0=stay, 1=startpos, 2=priorpos 3=peakpos 4=valleypos 5=+edgepos 6=-edgepos

  ($status,$maxPts) = Pezca::GetDouble($scanrec.'.MPTS');
  &report_exit($script, $status, $scanrec.'.MPTS');
  if ($numPts > $maxPts) {
     $current_time = scalar localtime;
     $msg  = '*** '.$current_time."\n";
     $msg .= '*** '.$script.': for scanrec='.$scanrec.' requested npts='
                           .$numPts.' exceeds buffer='.$maxPts."\n";
     &press_key_exit($msg);
  }
### Upload $numPts:
  $status = Pezca::PutDouble($scanrec.'.NPTS',$numPts);
  &report_exit($script, $status, $scanrec.'.NPTS');

### Upload after_scan=stay_at_end_pos
### (0=stay, 1=startpos, 2=priorpos 3=peakpos 4=valleypos 5=+edgepos 6=-edgepos).
  if (defined $afterScan) {
     $status = Pezca::PutDouble($scanrec.'.PASM',$afterScan);
     &report_exit($script, $status, $scanrec.'.PASM');
  }

### Upload positioner_delay=0 (settling time)
  if (defined $settleTime) {
     $status = Pezca::PutDouble($scanrec.'.PDLY',$settleTime);
     &report_exit($script, $status, $scanrec.'.PDLY');
  }

### Tell scanrec that we want to use 1 reference detector:
  $status = Pezca::PutDouble($scanrec.'.REFD',1);
  &report_exit($script, $status, $scanrec.'.REFD');
}

#=================================================================

# sub setupSRmotorLIN ($scanrec, $motorNo, $actPosPV, $rqsPosPV, $begin, $end, [$scanMode]);
sub setupSRmotorLIN ($$$$$$;$) {

### This function setups given SCAN RECORD positioner for LINEAR/FLY type of scan
### It is complimentary: *** should be called after setupSRmotorsCommon ***

  my ($script, $scanrec, $scanMode, $motorNo);
  my ($actPosPV, $rqsPosPV, $begin, $end);
  my ($status, $current_time, $msg);

  $script     = 'setupSRmotorLIN';
  $scanrec    = shift(@_);	#0: scan record name (PV)
  $motorNo    = shift(@_);      #1: 1-4
  $actPosPV   = shift(@_);      #2:
  $rqsPosPV   = shift(@_);      #3:
  $begin      = shift(@_);      #4:
  $end        = shift(@_);      #5:
  $scanMode   = shift(@_);      #6: 0=linear, 2=fly=2, but not 1=table
  if (!defined $scanMode) {
     $scanMode = 0;		#default is 0=linear
  }

  if ($scanMode != 0 && $scanMode != 2) {
     $current_time = scalar localtime;
     $msg  = '*** '.$current_time."\n";
     $msg .= '*** '.$script.': for scanrec='.$scanrec
                           .' incorrect scan mode = '.$scanMode."\n";
     &press_key_exit($msg);
  }

  if ($motorNo < 1 || $motorNo > 4) {
     $current_time = scalar localtime;
     $msg  = '*** '.$current_time."\n";
     $msg .= '*** '.$script.': for scanrec='.$scanrec.' requested motorNo='
                           .$motorNo.' is not in the 1-4 range'."\n";
     &press_key_exit($msg);
  }

### Upload scan mode (0=linear, 1=table, 2=on-the-fly):
  $status = Pezca::PutDouble($scanrec.'.P'.$motorNo.'SM',$scanMode);
  &report_exit($script, $status, $scanrec.'.P'.$motorNo.'SM');
### Upload abs|rel (0=absolute, 1-relative)
  $status = Pezca::PutDouble($scanrec.'.P'.$motorNo.'AR',0);
  &report_exit($script, $status, $scanrec.'.P'.$motorNo.'AR');
### Upload $actPosPV:
  $status = Pezca::PutString($scanrec.'.R'.$motorNo.'PV',$actPosPV);
  &report_exit($script, $status, $scanrec.'.R'.$motorNo.'PV');
### Upload $rqsPosPV:
  $status = Pezca::PutString($scanrec.'.P'.$motorNo.'PV',$rqsPosPV);
  &report_exit($script, $status, $scanrec.'.P'.$motorNo.'PV');
### Upload $begin:
  $status = Pezca::PutDouble($scanrec.'.P'.$motorNo.'SP',$begin);
  &report_exit($script, $status, $scanrec.'.P'.$motorNo.'SP');
### Upload $end:
  $status = Pezca::PutDouble($scanrec.'.P'.$motorNo.'EP',$end);
  &report_exit($script, $status, $scanrec.'.P'.$motorNo.'EP');
}

#=================================================================

# sub setupSRmotorTBL ($scanrec, $motorNo, $actPosPV, $rqsPosPV, \@posArray);
sub setupSRmotorTBL ($$$$\@) {

### This function setups given SCAN RECORD positioner for TABLE type of scan
### It is complimentary: *** should be called after setupSRmotorsCommon ***

  my ($script, $scanrec, $motorNo);
  my ($actPosPV, $rqsPosPV, $posArray_ref);
  my ($status, $current_time, $scanMode, $msg);

  $script       = 'setupSRmotorTBL';
  $scanrec      = shift(@_);      #0: scan record name (PV)
  $motorNo      = shift(@_);      #2: 1-4
  $actPosPV     = shift(@_);      #3:
  $rqsPosPV     = shift(@_);      #4:
  $posArray_ref = shift(@_);      #5:

  $scanMode = 1;		  #0=linear, 1=table, 2=on-the-fly

  if ($motorNo < 1 || $motorNo > 4) {
     $current_time = scalar localtime;
     $msg .= '*** '.$current_time."\n";
     $msg .= '*** '.$script.': for scanrec='.$scanrec.' requested motorNo='
                           .$motorNo.' is not in the 1-4 range'."\n";
     &press_key_exit($msg);
  }

### Upload scan mode (0=linear, 1=table, 2=on-the-fly):
  $status = Pezca::PutDouble($scanrec.'.P'.$motorNo.'SM',$scanMode);
  &report_exit($script, $status, $scanrec.'.P'.$motorNo.'SM');
### Upload abs|rel (0=absolute, 1-relative)
  $status = Pezca::PutDouble($scanrec.'.P'.$motorNo.'AR',0);
  &report_exit($script, $status, $scanrec.'.P'.$motorNo.'AR');
### Upload $actPosPV:
  $status = Pezca::PutString($scanrec.'.R'.$motorNo.'PV',$actPosPV);
  &report_exit($script, $status, $scanrec.'.R'.$motorNo.'PV');
### Upload $rqsPosPV:
  $status = Pezca::PutString($scanrec.'.P'.$motorNo.'PV',$rqsPosPV);
  &report_exit($script, $status, $scanrec.'.P'.$motorNo.'PV');
### Upload @posArray:
  $status = Pezca::PutList($scanrec.'.P'.$motorNo.'PA','DOUBLE',@$posArray_ref);
  &report_exit($script, $status, $scanrec.'.P'.$motorNo.'PA');
}

#=================================================================

# sub setupSRstepmotors ($scanrec, $numPts, $numMotors, \@motors, \@begin, \@end) {
sub setupSRstepmotors ($$$\@\@\@) {

### This function setups SCAN RECORD positioners for *on-the-fly* scans (motor record / OMS58 version):

  my ($scanrec, $numPts, $numMotors, $motorsRef, $beginRef, $endRef);
  my ($scanMode, $settleTime, $afterScan);

  $scanrec    = shift(@_);	#scan record name (PV)
  $numPts     = shift(@_);	#should we add +1 here, like in C subs?
  $numMotors  = shift(@_);
  $motorsRef  = shift(@_);
  $beginRef   = shift(@_);
  $endRef     = shift(@_);

  $scanMode   = 2;        	# 0=linear, 1=table 2=on-the-fly)
  $settleTime = 0.0;       	# positioner settle time (delay)
  $afterScan  = 0;        	# 0=stay, 1=startpos, 2=priorpos 3=peakpos 4=valleypos 5=+edgepos 6=-edgepos

### Here @_= $numMotors, @motors, @begin, @end;
  &setupSRstepmotorsMore ($scanrec, $numPts, $scanMode, $settleTime, $afterScan,
                          $numMotors, $motorsRef, $beginRef, $endRef);
}

#=================================================================

# sub setupSRstepmotorsMore ($scanrec, $numPts, $scanMode, $settleTime, $afterScan,
#                        $numMotors, \@motors, \@begin, \@end) {
sub setupSRstepmotorsMore ($$$$$$\@\@\@) {

### This function setups SCAN RECORD positioners for fly/step scans (motor record / OMS58 version):

  my ($script, $scanrec, $numPts, $scanMode, $settleTime, $afterScan);
  my ($numMotors, $motorsRef, $beginRef, $endRef, $nMotors, $nBegin, $nEnd);
  my ($status, $maxPts, $i, $j, $current_time, $mtrRqsPV, $mtrActPV, $msg);

  $script     = 'setupSRstepmotorsMore';
  $scanrec    = shift(@_);      #0: scan record name (PV)                    *** do not use 'shift' !!! ***
  $numPts     = shift(@_);      #1: should we add +1 here, like in C subs?   *** do not use 'shift' !!! ***
  $scanMode   = shift(@_);      #2: 0=linear, 1=table 2=on-the-fly           *** do not use 'shift' !!! ***
  $settleTime = shift(@_);      #3: positioner settle time (delay)           *** do not use 'shift' !!! ***
  $afterScan  = shift(@_);      #4: 0=stay, 1=startpos, 2=priorpos 3=peakpos 4=valleypos 5=+edgepos 6=-edgepos
  $numMotors  = shift(@_);      #5: number of motors                         *** do not use 'shift' !!! ***
  $motorsRef  = shift(@_);
  $beginRef   = shift(@_);
  $endRef     = shift(@_);

  $nMotors = @$motorsRef;
  $nBegin  = @$beginRef;
  $nEnd    = @$endRef;

  if ($scanMode != 0 || $scanMode != 2) {
     $current_time = scalar localtime;
     $msg  = '*** '.$current_time."\n";
     $msg .= '*** '.$script.': for scanrec='.$scanrec
                           .' incorrect scan mode = '.$scanMode."\n";
     &press_key_exit($msg);
  }

  ($status,$maxPts) = Pezca::GetDouble($scanrec.'.MPTS');
  &report_exit($script, $status, $scanrec.'.MPTS');
  if ($numPts > $maxPts) {
     $current_time = scalar localtime;
     $msg  = '*** '.$current_time."\n";
     $msg .= '*** '.$script.': for scanrec='.$scanrec.' requested npts='
                           .$numPts.' exceeds buffer='.$maxPts."\n";
     &press_key_exit($msg);
  }

  if ($numMotors > $nMotors ||
      $numMotors > $nBegin ||
      $numMotors > $nEnd) {
     $current_time = scalar localtime;
     $msg  = '*** '.$current_time."\n";
     $msg .= '*** '.$script.': for scanrec='.$scanrec
                   .' requested motors='.$numMotors
                   .' exceeds one of the arrays lengths:'."\n"
                   .' @motors='.$nMotors
                   .' @begin='.$nBegin
                   .' @end='.$nEnd."\n";
     &press_key_exit($msg);
  }

### Upload $numPts:
  $status = Pezca::PutDouble($scanrec.'.NPTS',$numPts);
  &report_exit($script, $status, $scanrec.'.NPTS');

  for ($i=1; $i<=$numMotors; $i++) {
     $j = $i - 1;
     $mtrRqsPV = @$motorsRef[$j].'.VAL';
     $mtrActPV = @$motorsRef[$j].'.RBV';
### Upload @actPos:
     $status = Pezca::PutString($scanrec.'.R'.$i.'PV',$mtrActPV);
     &report_exit($script, $status, $scanrec.'.R'.$i.'PV');
### Upload @rqsPos:
     $status = Pezca::PutString($scanrec.'.P'.$i.'PV',$mtrRqsPV);
     &report_exit($script, $status, $scanrec.'.P'.$i.'PV');
### Upload @begin:
     $status = Pezca::PutDouble($scanrec.'.P'.$i.'SP',@$beginRef[$j]);
     &report_exit($script, $status, $scanrec.'.P'.$i.'SP');
### Upload @end:
     $status = Pezca::PutDouble($scanrec.'.P'.$i.'EP',@$endRef[$j]);
     &report_exit($script, $status, $scanrec.'.P'.$i.'EP');

### Upload scan mode (0=linear, 1=table, 2=on-the-fly):
     if (defined $scanMode) {
        $status = Pezca::PutDouble($scanrec.'.P'.$i.'SM',$scanMode);
        &report_exit($script, $status, $scanrec.'.P'.$i.'SM');
     }
### Upload abs|rel=absolute
### (0=absolute, 1-relative)
     $status = Pezca::PutDouble($scanrec.'.P'.$i.'AR',0);
     &report_exit($script, $status, $scanrec.'.P'.$i.'AR');
  }

### Upload after_scan=stay_at_end_pos
### (0=stay, 1=startpos, 2=priorpos 3=peakpos 4=valleypos 5=+edgepos 6=-edgepos).
  if (defined $afterScan) {
     $status = Pezca::PutDouble($scanrec.'.PASM',$afterScan);
     &report_exit($script, $status, $scanrec.'.PASM');
  }

### Upload positioner_delay=0 (settling time)
  if (defined $settleTime) {
     $status = Pezca::PutDouble($scanrec.'.PDLY',$settleTime);
     &report_exit($script, $status, $scanrec.'.PDLY');
  }

### Tell scanrec that we want to use 1 reference detector:
  $status = Pezca::PutDouble($scanrec.'.REFD',1);
  &report_exit($script, $status, $scanrec.'.REFD');
}

#=================================================================

# sub setupSRtriggers ($scanrec, \@triggerPV, [\@triggerVAL, $settleTime, $autowaitNr]) {
sub setupSRtriggers ($$\@;\@$$) {

### This function setups SCAN RECORD triggers (up to 4)

  my ($script, $scanrec, $triggerPV, $triggerVAL);
  my ($settleTime, $autowaitNr);
  my ($status, $current_time, $msg, $ntrig, $i, $j, $val);

  $script     = 'setupSRtriggers';
  $scanrec    = shift(@_);
  $triggerPV  = shift(@_);
  $triggerVAL = shift(@_);
  $settleTime = shift(@_);  if (!defined $settleTime) {$settleTime=0;}
  $autowaitNr = shift(@_);  if (!defined $autowaitNr) {$autowaitNr=0;}

  $ntrig = @$triggerPV;
  if ($ntrig > 4) {
     $current_time = scalar localtime;
     $msg  = '*** '.$current_time."\n";
     $msg .= '*** '.$script.': too many triggers for ['.$scanrec.']: '.$ntrig.' > 4.'."\n";
     &press_key_exit($msg);
  }
  if ($autowaitNr > $ntrig) {
     $current_time = scalar localtime;
     $msg  = '*** '.$current_time."\n";
     $msg .= '*** '.$script.': wait clients for ['.$scanrec.'] exceeds triggers: '.$autowaitNr.' > '.$ntrig.'.'."\n";
     &press_key_exit($msg);
  }

### Upload detector triggers:
  for ($i=0; $i<$ntrig; $i++) {
     $j = $i+1;
     $status = Pezca::PutString($scanrec.'.T'.$j.'PV',@$triggerPV[$i]);
     &report_exit($script, $status, $scanrec.'.T.'.$j.'PV');		#trigger PV
     if (defined @$triggerVAL[$j]) {$val = @$triggerVAL[$j];}
     else                          {$val = 1;}				#default if not passed
     $status = Pezca::PutDouble($scanrec.'.T'.$j.'CD',$val);		#trigger value
     &report_exit($script, $status, $scanrec.'.T'.$j.'CD');
  }
#
  $status = Pezca::PutDouble($scanrec.'.DDLY',$settleTime);	# settling time
  &report_exit($script, $status, $scanrec.'.DDLY');
#
  $status = Pezca::PutDouble($scanrec.'.AWCT',$autowaitNr);		#number of clients to wait
  &report_exit($script, $status, $scanrec.'.AWCT');
}

#=================================================================

# sub setupSRdetectors ($scanrec, $joerger, \@IntyFlag, [$triggerPV, $triggerVAL, $settleTime]) {
sub setupSRdetectors ($$\@;$$$) {

### This function setups SCAN RECORD detectors (Joerger scaler inputs):
### Only those Joerger inputs are used for which @IntyFlag[$i]=1

  our ($numDetectors, @channels);
  my ($script, $scanrec, $joerger, $IntyRef, $IntyFlag, $maxJoerger, $detectorPV);
  my ($triggerPV, $triggerVAL, $settleTime, $i, $status);

  $numDetectors = 0;		# this is GLOBAL returned by the function
  @channels     = (0);		# this is GLOBAL returned by the function (detectror inputs)
  $script       = 'setupSRdetectors';
  $scanrec      = shift(@_);
  $joerger      = shift(@_);
  $IntyRef      = shift(@_);
  $triggerPV    = shift(@_);
  $triggerVAL   = shift(@_);  if (!defined $triggerVAL) {$triggerVAL=1;}
  $settleTime   = shift(@_);  if (!defined $settleTime) {$settleTime=0;}

  if (!defined $triggerPV) {$triggerPV=$joerger.'.CNT'; $triggerVAL=1;}

### Request the number of inputs in the detector:
### (should be either 16 or 32 depending on model)
  ($status,$maxJoerger) = Pezca::GetDouble($joerger.'.NCH');
  &report_exit($script, $status, $joerger.'.NCH');

  for ($i=1; $i<=$maxJoerger; $i++) {
     $IntyFlag = @$IntyRef[$i-1];
     if (defined $IntyFlag) {
        if ($IntyFlag != 0) {
           $numDetectors++;
           $detectorPV = sprintf $scanrec.'.D%02dPV', $numDetectors;
           $status = Pezca::PutString($detectorPV,$joerger.'.S'.$i);
           &report_exit($script, $status, $detectorPV);
           $channels[$numDetectors-1] = $i;
        }
     }
  }

### Upload detector triggers:
  $status = Pezca::PutString($scanrec.'.T1PV',$triggerPV);
  &report_exit($script, $status, $scanrec.'.T1PV');
  $status = Pezca::PutDouble($scanrec.'.T1CD',$triggerVAL);	# trigger value
  &report_exit($script, $status, $scanrec.'.T1CD');
#
  $status = Pezca::PutDouble($scanrec.'.DDLY',$settleTime);	# settling time
  &report_exit($script, $status, $scanrec.'.DDLY');
  $status = Pezca::PutDouble($scanrec.'.ACQM',0);		# data acquisition: normal
  &report_exit($script, $status, $scanrec.'.ACQM');
}

#=================================================================

# sub setupSRdetectors_v2 ($scanrec, $joerger, $ninputs, \@inputs, [$triggerPV, $triggerVAL, $settleTime]) {
sub setupSRdetectors_v2 ($$$\@;$$$) {

### This function setups SCAN RECORD detectors (Joerger scaler inputs)

  my ($script, $scanrec, $joerger, $ninputs, $inputsRef, $input, $ninp);
  my ($triggerPV, $triggerVAL, $settleTime, $maxJoerger);
  my ($i, $detPV, $status, $current_time, $msg);

  $script     = 'setupSRdetectors_v2';
  $scanrec    = shift(@_);
  $joerger    = shift(@_);
  $ninputs    = shift(@_);
  $inputsRef  = shift(@_);
  $triggerPV  = shift(@_);
  $triggerVAL = shift(@_);  if (!defined $triggerVAL) {$triggerVAL=1;}
  $settleTime = shift(@_);  if (!defined $settleTime) {$settleTime=0;}

  if (!defined $triggerPV) {$triggerPV=$joerger.'.CNT'; $triggerVAL=1;}

  $ninp = @$inputsRef;
  if ($ninputs > $ninp) {
     $current_time = scalar localtime;
     $msg  = '*** '.$current_time."\n";
     $msg .= '*** '.$script.': for scanrec='.$scanrec
                   .' requested scaler inputs='.$ninputs
                   .' exceeds the array length:'."\n"
                   .' @inputs='.$ninp."\n";
     &press_key_exit($msg);
  }

### Request the number of inputs in the detector:
### (should be either 16 or 32 depending on model)
  ($status,$maxJoerger) = Pezca::GetDouble($joerger.'.NCH');
  &report_exit($script, $status, $joerger.'.NCH');

  for ($i=1; $i<=$ninputs; $i++) {
     $input = @$inputsRef[$i-1];
     if ($input < 1 || $input > $maxJoerger) {
        $current_time = scalar localtime;
        $msg  = '*** '.$current_time."\n";
        $msg .= '*** '.$script.': scaler input ['.$input.'] is not in range.'."\n";
        &press_key_exit($msg);
     }
     $detPV = sprintf $scanrec.'.D%02dPV', $i;
     $status = Pezca::PutString($detPV,$joerger.'.S'.$input);
     &report_exit($script, $status, $detPV);
  }

### Upload detector triggers:
  $status = Pezca::PutString($scanrec.'.T1PV',$triggerPV);
  &report_exit($script, $status, $scanrec.'.T1PV');
  $status = Pezca::PutDouble($scanrec.'.T1CD',$triggerVAL);	# trigger value
  &report_exit($script, $status, $scanrec.'.T1CD');
  $status = Pezca::PutDouble($scanrec.'.DDLY',$settleTime);	# settling time
  &report_exit($script, $status, $scanrec.'.DDLY');
  $status = Pezca::PutDouble($scanrec.'.ACQM',0);		# data acquisition: normal
  &report_exit($script, $status, $scanrec.'.ACQM');
}

#=================================================================

# sub changeSRdetectors_v2 ($scanrec, [$triggerPV, $triggerVAL, $settleTime]) {
sub changeSRdetectors_v2 ($$;$$$) {

### This function changes SCAN RECORD detectors (Joerger scaler inputs)

  my ($script, $scanrec, $triggerPV, $triggerVAL, $settleTime, $status);

  $script     = 'changeSRdetectors_v2';
  $scanrec    = shift(@_);
  $triggerPV  = shift(@_);
  $triggerVAL = shift(@_);
  $settleTime = shift(@_);

  if (defined $triggerPV) {
     $status = Pezca::PutString($scanrec.'.T1PV',$triggerPV);
     &report_exit($script, $status, $scanrec.'.T1PV');
  }
  if (defined $triggerVAL) {
     $status = Pezca::PutDouble($scanrec.'.T1CD',$triggerVAL);	# trigger value
     &report_exit($script, $status, $scanrec.'.T1CD');
  }
  if (defined $settleTime) {
     $status = Pezca::PutDouble($scanrec.'.DDLY',$settleTime);	# settling time
     &report_exit($script, $status, $scanrec.'.DDLY');
  }
}

#=================================================================

# sub setupSRdetectors_auto ($scanrec, $joerger, $ninputs, \@inputs, [$triggerPV, $triggerVAL, $settleTime]) {
sub setupSRdetectors_auto ($$$\@;$$$) {

### This sub plays gently with existing SR detectors: it does not
### erase existing detectors and new Joerger inputs are added only
### as needed. It returns array @SRdetectors of size $ninputs
### showing what SR detectors are assigned to @inputs.
### ATTENTION: use with &clearSR($scanrec,4) and &openSRmedm_detec.

  my ($script, $scanrec, $joerger, $ninputs, $inputsRef);
  my ($triggerPV, $triggerVAL, $settleTime, $maxInputs, $ninp);
  my (@SRdetectors, $ndetectors, $input, $candidate);
  my (@filled, $numFilled, $detPV, $detLink);
  my ($i, $j, $status, $found, $maxJoerger, $current_time, $msg);

  $script     = 'setupSRdetectors_auto';
  $scanrec    = shift(@_);
  $joerger    = shift(@_);
  $ninputs    = shift(@_);
  $inputsRef  = shift(@_);
  $triggerPV  = shift(@_);
  $triggerVAL = shift(@_);  if (!defined $triggerVAL) {$triggerVAL=1;}
  $settleTime = shift(@_);  if (!defined $settleTime) {$settleTime=0;}

  if (!defined $triggerPV) {$triggerPV=$joerger.'.CNT'; $triggerVAL=1;}

  $maxInputs  = 70;                             # max number of detectors in scanrec
  $ninp       = @$inputsRef;

### Request the number of inputs in the detector:
### (should be either 16 or 32 depending on model)
  ($status,$maxJoerger) = Pezca::GetDouble($joerger.'.NCH');
  &report_exit($script, $status, $joerger.'.NCH');

### Find what scan record detectors are already setup:
### ATTENTION: this method presumes that there are no 'holes'
### in filling ScanRec inputs, e.g. one cannot use 1, 2, and 4
### then '4' will be ignored! Should we scan all channels?
  $i = 0;
  $numFilled = 0;
  do {
     $i++;
     $detPV = sprintf $scanrec.'.D%02dPV', $i;
     ($status,$detLink) = Pezca::GetString($detPV);
     &report_exit($script, $status, $detPV);
     if (length($detLink) > 0) {
        $filled[$numFilled] = $detLink;
        $numFilled++;
     }
  } while (length($detLink)>0 && $i<=$maxInputs);

### Determine 'detectors' array - how the scaler inputs are mapped to SR detectors:
  $ndetectors = 0;
  for ($i=0; $i<$ninputs; $i++) {
     $input = @$inputsRef[$i];
     if ($input < 1 || $input > $maxJoerger) {
         $current_time = scalar localtime;
         $msg  = '*** '.$current_time."\n";
         $msg .= '*** '.$script.': scaler input ['.$input.'] is not in range.'."\n";
         &press_key_exit($msg);
     }
     $candidate = $joerger.'.S'.$input;
### First check if the candidate is already in the scan record detectors list:
     $found = 0;
     for ($j=0; $j<$numFilled; $j++) {
        if ($candidate eq $filled[$j]) {
           $SRdetectors[$ndetectors]=$j+1;
           $ndetectors++;
           $found = 1;
           last;
        }
     }
     if (! $found) {
### Add new detector if not found among existing ones:
        $filled[$numFilled] = $candidate;
        $numFilled++;
        if ($numFilled > $maxInputs) {
           $current_time = scalar localtime;
           $msg  = '*** '.$current_time."\n";
           $msg .= '*** '.$script.': all '.$maxInputs.' of SR detectors are filled.'."\n";
           &press_key_exit($msg);
        }
        $detPV = sprintf $scanrec.'.D%02dPV', $numFilled;
        $status = Pezca::PutString($detPV, $candidate);
        &report_exit($script, $status, $detPV);
        $SRdetectors[$ndetectors] = $numFilled;
        $ndetectors++;
     }
  }

### Upload detector triggers:
  $status = Pezca::PutString($scanrec.'.T1PV',$triggerPV);
  &report_exit($script, $status, $scanrec.'.T1PV');
  $status = Pezca::PutDouble($scanrec.'.T1CD',$triggerVAL);	# trigger value
  &report_exit($script, $status, $scanrec.'.T1CD');
#
  $status = Pezca::PutDouble($scanrec.'.DDLY',$settleTime);	# settling time
  &report_exit($script, $status, $scanrec.'.DDLY');
  $status = Pezca::PutDouble($scanrec.'.ACQM',0);		# data acquisition: normal
  &report_exit($script, $status, $scanrec.'.ACQM');

  return @SRdetectors;
}

#=================================================================

# sub setupSRdetectors_pv ($scanrec, $joerger, \@inputs, \@pvlist, [$triggerPV, $triggerVAL, $settleTime]) {
sub setupSRdetectors_pv ($$\@;\@$$$) {

### This function setups SCAN RECORD detectors (Joerger scaler inputs + PV list)

  my ($script, $scanrec, $joerger, $inputsRef, $pvlistRef, $ninputs, $npvs);
  my ($triggerPV, $triggerVAL, $settleTime, $input, $pv, $maxJoerger, $detPV);
  my ($status, $current_time, $msg, $i, $j);

  $script     = 'setupSRdetectors_pv';
  $scanrec    = shift(@_);
  $joerger    = shift(@_);
  $inputsRef  = shift(@_);
  $pvlistRef  = shift(@_);
  $triggerPV  = shift(@_);
  $triggerVAL = shift(@_);  if (!defined $triggerVAL) {$triggerVAL=1;}
  $settleTime = shift(@_);  if (!defined $settleTime) {$settleTime=0;}

  if (!defined $triggerPV) {$triggerPV=$joerger.'.CNT'; $triggerVAL=1;}

  $ninputs = @$inputsRef;
  if (defined $pvlistRef) {$npvs = @$pvlistRef;} else {$npvs = 0;}

### Request the number of inputs in the detector:
### (should be either 16 or 32 depending on model)
  ($status,$maxJoerger) = Pezca::GetDouble($joerger.'.NCH');
  &report_exit($script, $status, $joerger.'.NCH');

  for ($i=1; $i<=$ninputs; $i++) {
     $input = @$inputsRef[$i-1];
     if ($input < 1 || $input > $maxJoerger) {
        $current_time = scalar localtime;
        $msg  = '*** '.$current_time."\n";
        $msg .= '*** '.$script.': scaler input ['.$input.'] is not in range.'."\n";
        &press_key_exit($msg);
     }
     $detPV = sprintf $scanrec.'.D%02dPV', $i;
     $status = Pezca::PutString($detPV,$joerger.'.S'.$input);
     &report_exit($script, $status, $detPV);
  }

  for ($i=1; $i<=$npvs; $i++) {
     $pv = @$pvlistRef[$i-1];
     $j = $ninputs + $i;
     $detPV = sprintf $scanrec.'.D%02dPV', $j;
     $status = Pezca::PutString($detPV,$pv);
     &report_exit($script, $status, $detPV);
  }

### Upload detector triggers:
  $status = Pezca::PutString($scanrec.'.T1PV',$triggerPV);
  &report_exit($script, $status, $scanrec.'.T1PV');
  $status = Pezca::PutDouble($scanrec.'.T1CD',$triggerVAL);	# trigger value
  &report_exit($script, $status, $scanrec.'.T1CD');
#
  $status = Pezca::PutDouble($scanrec.'.DDLY',$settleTime);	# settling time
  &report_exit($script, $status, $scanrec.'.DDLY');
  $status = Pezca::PutDouble($scanrec.'.ACQM',0);		# data acquisition: normal
  &report_exit($script, $status, $scanrec.'.ACQM');
}

#=================================================================

# sub setupSRdetectors_addpvs ($scanrec, \@pvlist, [$firstdetector]) {
sub setupSRdetectors_addpvs ($\@;$) {

### This sub adds a list of PVs to the scan record detectors. If
### $firstdetector is undefined, it scans for already set SR detectors
### and then appends new PVs to the list.

  my ($script, $scanrec, $pvlistRef, $firstdetector, $npvs, $pv, $i);
  my ($detPV, $detLink, $maxInputs, $status, $current_time, $msg);

  $script        = 'setupSRdetectors_addpvs';
  $scanrec       = shift(@_);
  $pvlistRef     = shift(@_);
  $firstdetector = shift(@_);
  $maxInputs  = 70;				# max number of detectors in scanrec

  if (!defined $firstdetector) {
### Find what scan record detectors are already setup:
### ATTENTION: this method presumes that there are no 'holes'
### in filling ScanRec inputs, e.g. one cannot use 1, 2, and 4
### then '4' will be ignored! Should we scan all channels?
     $firstdetector = 0;
     do {
        $firstdetector++;
        $detPV = sprintf $scanrec.'.D%02dPV', $firstdetector;
        ($status,$detLink) = Pezca::GetString($detPV);
        &report_exit($script, $status, $detPV);
     } while (length($detLink)>0 && $firstdetector<=$maxInputs);
  }
  if ($firstdetector > $maxInputs) {
     $current_time = scalar localtime;
     $msg  = '*** '.$current_time."\n";
     $msg .= '*** '.$script.': All SR detectors are filled.'."\n";
     &press_key_exit($msg);
     return 1;
  }
  $npvs = @$pvlistRef;
  for ($i=0; $i<$npvs; $i++) {
     $pv = @$pvlistRef[$i];
     $detPV = sprintf $scanrec.'.D%02dPV', ($firstdetector+$i);
     $status = Pezca::PutString($detPV,$pv);
     if ($status) {&report_exit($script, $status, $detPV); return 2;}
  }
  return 0;
}

#=================================================================

# sub setupSRdetectors_mca ($scanrec, $mca, \@ROIs, \@pvlist, [$triggerPV, $triggerVAL, $settleTime]) {
sub setupSRdetectors_mca ($$\@;\@$$$) {

### This function setups SCAN RECORD detectors (MCA ROIs + PV list)

  my ($script, $scanrec, $mca, $ROIsRef, $pvlistRef, $nROIs, $npvs);
  my ($triggerPV, $triggerVAL, $settleTime, $ROI, $pv, $maxROIs, $detPV);
  my ($status, $current_time, $msg, $i, $j, $mcapv);

  $script     = 'setupSRdetectors_mca';
  $scanrec    = shift(@_);
  $mca        = shift(@_);
  $ROIsRef    = shift(@_);
  $pvlistRef  = shift(@_);
  $triggerPV  = shift(@_);
  $triggerVAL = shift(@_);  if (!defined $triggerVAL) {$triggerVAL=1;}
  $settleTime = shift(@_);  if (!defined $settleTime) {$settleTime=0;}

  if ($mca !~ /falcon/) {
     $mca =~ s/:$//;					#mca should not be ended by ":"
     $mcapv = $mca;
  } else {
     if ($mca !~ /:$/) {$mca.=':';}			#falcon must be ended by ":"
     $mcapv = $mca.'mca1';
  }

  if (!defined $triggerPV) {$triggerPV=$mcapv.'.ERST'; $triggerVAL=1;}

  $nROIs   = @$ROIsRef;
  if (defined $pvlistRef) {$npvs = @$pvlistRef;} else {$npvs = 0;}

### The max number of ROIs in the mca (currently hardcoded):
  $maxROIs = 32;				# nunbered from 0 to 31

  for ($i=1; $i<=$nROIs; $i++) {
     $ROI = @$ROIsRef[$i-1];
     if ($ROI < 0 || $ROI > $maxROIs-1) {
        $current_time = scalar localtime;
        $msg  = '*** '.$current_time."\n";
        $msg .= sprintf '*** '.$script.': ROI ['.$ROI.'] is not in range [0-%d].'."\n",$maxROIs-1;
        &press_key_exit($msg);
     }
     $detPV = sprintf $scanrec.'.D%02dPV', $i;
     $status = Pezca::PutString($detPV,$mcapv.'.R'.$ROI);		#for 'Sum'
#    $status = Pezca::PutString($detPV,$mcapv.'.R'.$ROI.'N');	#for 'Net'
     &report_exit($script, $status, $detPV);
  }

  for ($i=1; $i<=$npvs; $i++) {
     $pv = @$pvlistRef[$i-1];
     $j = $nROIs + $i;
     $detPV = sprintf $scanrec.'.D%02dPV', $j;
     $status = Pezca::PutString($detPV,$pv);
     &report_exit($script, $status, $detPV);
  }

### Upload detector triggers:
# $status = Pezca::PutString($scanrec.'.T1PV',$mca.'EraseStart');
  $status = Pezca::PutString($scanrec.'.T1PV',$triggerPV);
  &report_exit($script, $status, $scanrec.'.T1PV');
  $status = Pezca::PutDouble($scanrec.'.T1CD',$triggerVAL);	# trigger value
  &report_exit($script, $status, $scanrec.'.T1CD');
#
  $status = Pezca::PutDouble($scanrec.'.DDLY',$settleTime);	# settling time
  &report_exit($script, $status, $scanrec.'.DDLY');
  $status = Pezca::PutDouble($scanrec.'.ACQM',0);		# data acquisition: normal
  &report_exit($script, $status, $scanrec.'.ACQM');
}

#=================================================================

# sub setupSRdetectors_scaler_mca ($scanrec, $joerger, $mca, \@inputs, \@ROIs, [$triggerPV, $triggerVAL, $settleTime]) {
sub setupSRdetectors_scaler_mca ($$$\@\@;$$$) {

### This function setups SCAN RECORD detectors (Joerger inputs + MCA ROIs)

  my ($script, $scanrec, $joerger, $mca, $inputsRef, $ROIsRef);
  my ($triggerPV, $triggerVAL, $settleTime);
  my ($ninputs, $nROIs, $input, $ROI, $maxJoerger, $maxROIs);
  my ($i, $j, $detPV, $status, $current_time, $msg, $mcapv);

  $script     = 'setupSRdetectors_scaler_mca';
  $scanrec    = shift(@_);
  $joerger    = shift(@_);
  $mca        = shift(@_);
  $inputsRef  = shift(@_);
  $ROIsRef    = shift(@_);
  $triggerPV  = shift(@_);
  $triggerVAL = shift(@_);  if (!defined $triggerVAL) {$triggerVAL=1;}
  $settleTime = shift(@_);  if (!defined $settleTime) {$settleTime=0;}

  if ($mca !~ /falcon/) {
     $mca =~ s/:$//;					#mca should not be ended by ":"
     $mcapv = $mca;
  } else {
     if ($mca !~ /:$/) {$mca.=':';}			#falcon must be ended by ":"
     $mcapv = $mca.'mca1';
  }

  $ninputs = @$inputsRef;
  $nROIs   = @$ROIsRef;

### Request the number of inputs in the detector:
### (should be either 16 or 32 depending on model)
  ($status,$maxJoerger) = Pezca::GetDouble($joerger.'.NCH');
  &report_exit($script, $status, $joerger.'.NCH');
### The max number of ROIs in the mca (currently hardcoded):
  $maxROIs = 32;				# numbered from 0 to 31

  $j = 0;
  for ($i=1; $i<=$ninputs; $i++) {
     $input = @$inputsRef[$i-1];
     if ($input < 1 || $input > $maxJoerger) {
        $current_time = scalar localtime;
        $msg  = '*** '.$current_time."\n";
        $msg .= '*** '.$script.': scaler input ['.$input.'] is not in range.'."\n";
        &press_key_exit($msg);
     }
     $j++;
     $detPV = sprintf $scanrec.'.D%02dPV', $j;
     $status = Pezca::PutString($detPV,$joerger.'.S'.$input);
     &report_exit($script, $status, $detPV);
  }

  for ($i=1; $i<=$nROIs; $i++) {
     $ROI = @$ROIsRef[$i-1];
     if ($ROI < 0 || $ROI > $maxROIs-1) {
        $current_time = scalar localtime;
        $msg  = '*** '.$current_time."\n";
        $msg .= sprintf '*** '.$script.': ROI ['.$ROI.'] is not in range [0-%d].'."\n",$maxROIs-1;
        &press_key_exit($msg);
     }
     $j++;
     $detPV = sprintf $scanrec.'.D%02dPV', $j;
     $status = Pezca::PutString($detPV,$mcapv.'.R'.$ROI);		#for 'Sum'
#    $status = Pezca::PutString($detPV,$mcapv.'.R'.$ROI.'N');		#for 'Net'
     &report_exit($script, $status, $detPV);
  }

### Upload detector triggers:
  if (defined $triggerPV) {
     $status = Pezca::PutString($scanrec.'.T1PV',$triggerPV);
     &report_exit($script, $status, $scanrec.'.T1PV');
     $status = Pezca::PutDouble($scanrec.'.T1CD',$triggerVAL);
     &report_exit($script, $status, $scanrec.'.T1CD');
  } else {
     $status = Pezca::PutString($scanrec.'.T1PV',$joerger.'.CNT');
     &report_exit($script, $status, $scanrec.'.T1PV');
     $status = Pezca::PutDouble($scanrec.'.T1CD',1);			# start scaler value
     &report_exit($script, $status, $scanrec.'.T1CD');
#
#    $status = Pezca::PutString($scanrec.'.T2PV',$mca.'EraseStart');
     $status = Pezca::PutString($scanrec.'.T2PV',$mcapv.'.ERST');
     &report_exit($script, $status, $scanrec.'.T2PV');
     $status = Pezca::PutDouble($scanrec.'.T2CD',1);			# start mca value
     &report_exit($script, $status, $scanrec.'.T2CD');
  }
#
  $status = Pezca::PutDouble($scanrec.'.DDLY',$settleTime);		# settling time
  &report_exit($script, $status, $scanrec.'.DDLY');
  $status = Pezca::PutDouble($scanrec.'.ACQM',0);			# data acquisition: normal
  &report_exit($script, $status, $scanrec.'.ACQM');
}

#=================================================================

# sub setupSRdetectors_scaler_mca_pv ($scanrec, $joerger, $mca, \@inputs, \@ROIs, [\@pvlist], [$triggerPV, $triggerVAL, $settleTime]) {
sub setupSRdetectors_scaler_mca_pv ($$$\@\@;\@$$$) {

### This function setups SCAN RECORD detectors (Joerger inputs + MCA ROIs + PV list)

  my ($script, $scanrec, $joerger, $mca, $inputsRef, $ROIsRef, $pvlistRef);
  my ($ninputs, $nROIs, $npvs, $input, $ROI, $pv, $maxJoerger, $maxROIs);
  my ($triggerPV, $triggerVAL, $settleTime, $detPV);
  my ($status, $current_time, $msg, $i, $j, $mcapv);

  $script     = 'setupSRdetectors_scaler_mca_pv';
  $scanrec    = shift(@_);
  $joerger    = shift(@_);
  $mca        = shift(@_);
  $inputsRef  = shift(@_);
  $ROIsRef    = shift(@_);
  $pvlistRef  = shift(@_);
  $triggerPV  = shift(@_);
  $triggerVAL = shift(@_);  if (!defined $triggerVAL) {$triggerVAL=1;}
  $settleTime = shift(@_);  if (!defined $settleTime) {$settleTime=0;}

  if ($mca !~ /falcon/) {
     $mca =~ s/:$//;					#mca should not be ended by ":"
     $mcapv = $mca;
  } else {
     if ($mca !~ /:$/) {$mca.=':';}			#falcon must be ended by ":"
     $mcapv = $mca.'mca1';
  }

  $ninputs = @$inputsRef;
  $nROIs   = @$ROIsRef;
  if (defined $pvlistRef) {$npvs = @$pvlistRef;} else {$npvs = 0;}

### Request the number of inputs in the detector:
### (should be either 16 or 32 depending on model)
  ($status,$maxJoerger) = Pezca::GetDouble($joerger.'.NCH');
  &report_exit($script, $status, $joerger.'.NCH');
### The max number of ROIs in the mca (currently hardcoded):
  $maxROIs = 32;				# numbered from 0 to 31

  $j = 0;
  for ($i=1; $i<=$ninputs; $i++) {
     $input = @$inputsRef[$i-1];
     if ($input < 1 || $input > $maxJoerger) {
        $current_time = scalar localtime;
        $msg  = '*** '.$current_time."\n";
        $msg .= '*** '.$script.': scaler input ['.$input.'] is not in range.'."\n";
        &press_key_exit($msg);
     }
     $j++;
     $detPV = sprintf $scanrec.'.D%02dPV', $j;
     $status = Pezca::PutString($detPV,$joerger.'.S'.$input);
     &report_exit($script, $status, $detPV);
  }

  for ($i=1; $i<=$nROIs; $i++) {
     $ROI = @$ROIsRef[$i-1];
     if ($ROI < 0 || $ROI > $maxROIs-1) {
        $current_time = scalar localtime;
        $msg  = '*** '.$current_time."\n";
        $msg .= sprintf '*** '.$script.': ROI ['.$ROI.'] is not in range [0-%d].'."\n",$maxROIs-1;
        &press_key_exit($msg);
     }
     $j++;
     $detPV = sprintf $scanrec.'.D%02dPV', $j;
     $status = Pezca::PutString($detPV,$mcapv.'.R'.$ROI);		#for 'Sum'
#    $status = Pezca::PutString($detPV,$mcapv.'.R'.$ROI.'N');		#for 'Net'
     &report_exit($script, $status, $detPV);
  }

  for ($i=1; $i<=$npvs; $i++) {
     $pv = @$pvlistRef[$i-1];
     $j++;
     $detPV = sprintf $scanrec.'.D%02dPV', $j;
     $status = Pezca::PutString($detPV,$pv);
     &report_exit($script, $status, $detPV);
  }

### Upload detector triggers:
  if (defined $triggerPV) {
     $status = Pezca::PutString($scanrec.'.T1PV',$triggerPV);
     &report_exit($script, $status, $scanrec.'.T1PV');
     $status = Pezca::PutDouble($scanrec.'.T1CD',$triggerVAL);
     &report_exit($script, $status, $scanrec.'.T1CD');
  } else {
     $status = Pezca::PutString($scanrec.'.T1PV',$joerger.'.CNT');
     &report_exit($script, $status, $scanrec.'.T1PV');
     $status = Pezca::PutDouble($scanrec.'.T1CD',1);			# start scaler value
     &report_exit($script, $status, $scanrec.'.T1CD');
#
#    $status = Pezca::PutString($scanrec.'.T2PV',$mca.'EraseStart');
     $status = Pezca::PutString($scanrec.'.T2PV',$mcapv.'.ERST');
     &report_exit($script, $status, $scanrec.'.T2PV');
     $status = Pezca::PutDouble($scanrec.'.T2CD',1);			# start mca value
     &report_exit($script, $status, $scanrec.'.T2CD');
  }
#
  $status = Pezca::PutDouble($scanrec.'.DDLY',$settleTime);		# settling time
  &report_exit($script, $status, $scanrec.'.DDLY');
  $status = Pezca::PutDouble($scanrec.'.ACQM',0);			# data acquisition: normal
  &report_exit($script, $status, $scanrec.'.ACQM');
}

#=================================================================

#  sub setupSRbeforeScan ($scanrec, $bsPV, $bsVAL, [$waitFlag]);	#returns status
sub setupSRbeforeScan ($$$;$) {

### This function setups ScanRec beforescan operations

  my ($script, $scanrec, $bsPV, $bsVAL, $waitFlag, $status);

  $script   = 'setupSRbeforeScan';
  $scanrec  = shift(@_);
  $bsPV     = shift(@_);
  $bsVAL    = shift(@_);
  $waitFlag = shift(@_);

  $status = Pezca::PutString($scanrec.'.BSPV',$bsPV);
  &report_exit($script, $status, $scanrec.'.BSPV');

  $status = Pezca::PutDouble($scanrec.'.BSCD',$bsVAL);
  &report_exit($script, $status, $scanrec.'.BSCD');

  if (defined $waitFlag) {
     if ($waitFlag ne '1' && $waitFlag ne '0') {
        print STDOUT '*** '.$script.': incorrect after-scan wait flag='.$waitFlag."\n";
        $status++;
     } else {
        $status += Pezca::PutDouble($scanrec.'.BSWAIT',$waitFlag);
        &report_exit($script, $status, $scanrec.'.BSWAIT');
     }
  }
  return $status;
}

#=================================================================

#  sub setupSRafterScan ($scanrec, $asPV, $asVAL, [$waitFlag]);		#returns status
sub setupSRafterScan ($$$;$) {

### This function setups ScanRec afterscan operations

  my ($script, $scanrec, $asPV, $asVAL, $waitFlag, $status);

  $script   = 'setupSRafterScan';
  $scanrec  = shift(@_);
  $asPV     = shift(@_);
  $asVAL    = shift(@_);
  $waitFlag = shift(@_);				#1=wait, 0=nowait
  if (!defined $waitFlag) {$waitFlag = 0;}
  $status = 0;

  $status += Pezca::PutString($scanrec.'.ASPV',$asPV);
  &report_exit($script, $status, $scanrec.'.ASPV');

  $status += Pezca::PutDouble($scanrec.'.ASCD',$asVAL);
  &report_exit($script, $status, $scanrec.'.ASCD');

  if (defined $waitFlag) {
     if ($waitFlag ne '1' && $waitFlag ne '0') {
        print STDOUT '*** '.$script.': incorrect after-scan wait flag='.$waitFlag."\n";
        $status++;
     } else {
        $status += Pezca::PutDouble($scanrec.'.ASWAIT',$waitFlag);
        &report_exit($script, $status, $scanrec.'.ASWAIT');
     }
  }
  return $status;
}

#=================================================================

#  sub setupSRrefreshrate ($scanrec, [$rate]);		#does not return anything
sub setupSRrefreshrate ($;$) {

### This function setups ScanRec afterscan operations

  my ($script, $scanrec, $rate, $status);

  $script   = 'setupSRrefreshrate';
  $scanrec  = shift(@_);
  $rate     = shift(@_);
  if (!defined $rate) {$rate = 0;}
  if ($rate < 0.1)    {$rate = 0;}

  $status = Pezca::PutString($scanrec.'.ATIME',$rate);
  &report_exit($script, $status, $scanrec.'.ATIME');
}

#=================================================================

# sub startSR ($scanrec,[$scanAbort]) {
sub startSR ($;$) {

### This function starts SCAN RECORD operation (scan)

  my ($script, $scanrec, $scanAbort,$ status, $status1, $busy);

  $script    = 'startSR';
  $scanrec   = shift(@_);
  $scanAbort = shift(@_);

### push 'SCAN' button:
### (should we verify that scanrec is not in 'PAUSE' mode?)
  $status = Pezca::PutDoubleOldCa($scanrec.'.EXSC',1);				# only use OldCa for start!
  if ($status != 0) {
     if ($status != 5) {							# if error, but PV connected
        &report_only($script, $status, $scanrec.'.EXSC');			# report (warn), but try to fix
        ($status1,$busy) = Pezca::GetDouble($scanrec.'.BUSY');
        if ($status1) {
           &report_exit($script, $status1, $scanrec.'.BUSY');
        }
        elsif ($busy == 1) {
           print STDOUT '*** '.$script.': cannot start scan record ['.$scanrec.'] as it is already running.'."\n";
           if (defined $scanAbort) {
              print STDOUT '!!! '.$script.': aborting scan record ['.$scanrec.'] previous run and retrying start...'."\n";
              &abortSR ($scanAbort,$scanrec,3);
              $status = Pezca::PutDoubleOldCa($scanrec.'.EXSC',1);		# only use OldCa for start!
           }
        }
     }
     &report_exit($script, $status, $scanrec.'.EXSC');				# PV not connected
  }
}

#=================================================================

# sub getSRcurrPts ($scanrec) {                                 # read current execution point
sub getSRcurrPts ($) {

### This function returns current execution point of scan record.

  my ($script, $scanrec, $status, $currPts);

  $script  = 'getSRcurrPts';
  $scanrec = shift(@_);

  ($status,$currPts) = Pezca::GetDouble($scanrec.'.CPT');
  if ($status) {
     &report_only($script, $status, $scanrec.'.CPT');
     $currPts = -1;
  }
  return $currPts;
}

#=================================================================

# sub abortSR ($scanAbort,[$scanrec,$nPress]) {
sub abortSR ($;$$) {

### This function aborts multi-dimensional SCAN RECORD
### (same as ABORT button on scan record MEDM screen)
### 1. A single write of 0 to the EXSC field requests a polite scan abort,
###    waiting for callbacks and data storage.
### 2. Two successive writes of 0 to EXSC request a scan abort with no wait
###    for outstanding callbacks. However, the sscan record will still wait
###    for the data-storage client.
### 3. Three successive writes of 0 to EXSC demand an immediate scan abort
###    with no regard for consequences. Scan data will be lost in this case
###    (no copy to final arrays)

  my ($script, $scanAbort, $scanrec, $nPress, $status, $i);

  $script    = 'abortSR';
  $scanAbort = shift(@_);
  $scanrec   = shift(@_);
  $nPress    = shift(@_);
  if (!defined $nPress) {$nPress=2;}				#abort unconditionally
  if    ($nPress < 1) {$nPress = 1;}
  elsif ($nPress > 2) {$nPress = 2;}

### Send aborts to the record:
  for ($i=1; $i<=$nPress; $i++) {
### The first ABORD press does a polite abort: it waits for motors to stop, etc.
### The second waits only for the data-storage client.
### The third press kills unconditionally
#    $status = Pezca::PutDoubleOldCa('nnID:AbortScans.PROC',0);
     $status = Pezca::PutDoubleOldCa($scanAbort.'.PROC',0);
#    &report_exit($script, $status, $scanAbort.'.PROC');
     if ($status) {print STDOUT '*** '.$script.': Unable to connect to ['.$scanAbort.'.PROC]'."\n";}
     select(undef,undef,undef,0.1);
  }

### Wait until record finishes aborting:
  if (defined $scanrec) {&waitSRabort ($scanrec);}
}

#=================================================================

# sub abortSR1 ($scanrec,[$nPress,waitflag]) {
sub abortSR1 ($;$$) {

### *** THIS DOES NOT WORK: EXSC returns PV write error is the record is already stopped! ***

### This function aborts single SCAN RECORD
### 1. A single write of 0 to the EXSC field requests a polite scan abort,
###    waiting for callbacks and data storage.
### 2. Two successive writes of 0 to EXSC request a scan abort with no wait
###    for outstanding callbacks. However, the sscan record will still wait
###    for the data-storage client.
### 3. Three successive writes of 0 to EXSC demand an immediate scan abort
###    with no regard for consequences. Scan data will be lost in this case.
###    (no copy to final arrays)

  my ($script, $scanrec, $busy, $nPress, $waitflag, $status, $i);

  $script   = 'abortSR1';
  $scanrec  = shift(@_);
  $nPress   = shift(@_);
  $waitflag = shift(@_);
  if (!defined $waitflag) {$waitflag = 0;}			#no wait
  if (!defined $nPress)   {$nPress   = 2;}			#abort unconditionally
  if    ($nPress < 1) {$nPress = 1;}
  elsif ($nPress > 3) {$nPress = 3;}

  ($status,$busy) = Pezca::GetDouble($scanrec.'.BUSY');
  if ($status) {
     &report_only($script, $status, $scanrec.'.BUSY');
     $busy = 1;							#presume worst
  }
  if ($busy) {
     ### Send aborts to the record:
     for ($i=1; $i<=$nPress; $i++) {
        ### The first ABORD press does a polite abort: it waits for motors to stop, etc.
        ### The second waits only for the data-storage client.
        ### The third press kills unconditionally
        $status = Pezca::PutDouble($scanrec.'.EXSC',0);
        &report_only($script, $status, $scanrec.'.EXSC');
        select(undef,undef,undef,0.1);
     }
     ### Wait until record finishes aborting:
     if ($waitflag) {&waitSRabort ($scanrec);}
  }
}

#=================================================================

# sub waitSRabort ($scanrec) {
sub waitSRabort ($) {

### This function waits until scan record finishes aborting

  my ($script, $scanrec, $status);
  my ($busy, $wait_max, $wait_, $wait_step);

  $script  = 'waitSRabort';
  $scanrec = shift(@_);

  $status = &addCaMonitor($scanrec.'.BUSY');
  &report_only($script, $status, $scanrec.'.BUSY');
  $wait_max  = 10;
  $wait_     = 0;
  $wait_step = 0.1;
  do {
     select(undef,undef,undef,$wait_step);			#sleep($wait_step);
     $wait_ += $wait_step;
     ($status,$busy) = Pezca::GetDouble($scanrec.'.BUSY');
     if ($status) {
        &report_only($script, $status, $scanrec.'.BUSY');
        $busy = 1;						#presume worst
     }
  } while ($busy == 1 && $wait_< $wait_max);
  if ($busy) {
     print STDOUT '!!! '.$script.': timeout waiting for scan record '.$scanrec.' to abort'."\n\n";
     return 1;
  } else {
     return 0;
  }
}

#=================================================================

# $actPts = &waitSRdata($scanrec);
# sub waitSRdata ($scanrec) {
sub waitSRdata ($) {

### This function waits for SCAN RECORD to upload data into arrays after
### the scan is finished or aborted

  my ($script, $scanrec, $wait_max, $wait_, $wait_step, $actPts);
  my ($status, $data_ready, $current_time);

  $script    = 'waitSRdata';
  $scanrec   = shift(@_);
  $wait_max  = 10;
  $wait_     = 0;
  $wait_step = 0.1;
  $actPts    = 0;

  $data_ready = 0;
  $status = &addCaMonitor($scanrec.'.DATA');
  &report_exit($script, $status, $scanrec.'.DATA');
  while (!$data_ready && $wait_ < $wait_max) {
     select(undef,undef,undef,$wait_step);		#sleep($wait_step);
     $wait_ += $wait_step;
     ($status,$data_ready) = Pezca::GetDouble($scanrec.'.DATA');
  }
  if ($wait_ >= $wait_max) {
     $current_time = scalar localtime;
     print STDOUT "\n";
     print STDOUT '*** '.$current_time."\n";
     print STDOUT '*** '.$script.': data not ready after '.$wait_max." s\n";
  }
### Read data:
  ($status,$actPts) = Pezca::GetDouble($scanrec.'.CPT');
  &report_exit($script, $status, $scanrec.'.CPT');

  return $actPts;
}



#=================================================================

# sub readSRdata ($scanrec, $numPts, $numMotors, $numDetectors) {
sub readSRdata ($$$$) {

### This function reads SCAN RECORD data from EPICS arrays
###
### Returns global x1[] ... x$numMotors[] and y1[] ... y$numDetectors[]
###

  our $EZCAFLOAT;
  my ($script, $scanrec, $numPts, $numMotors, $numDetectors);
  my ($status, $i, $name, $detPV);

  $script       = 'readSRdata';
  $scanrec      = shift(@_);
  $numPts       = shift(@_);
  $numMotors    = shift(@_);
  $numDetectors = shift(@_);

# print STDOUT $script.': numPts = '.$numPts."\n";
  for ($i=1; $i<=$numMotors; $i++) {
     $name = 'x'.$i;
     print STDOUT 'Reading '.$name.' from ['.$scanrec.']...'."\n";
     ($status,@$name) = Pezca::GetList($scanrec.'.P'.$i.'RA','double',$numPts);		#on 2008/10/03
     &report_exit($script, $status, $scanrec.'.P'.$i.'RA');
  }

  for ($i=1; $i<=$numDetectors; $i++) {
     $name = 'y'.$i;
     print STDOUT 'Reading '.$name.' from ['.$scanrec.']...'."\n";
     $detPV = sprintf $scanrec.'.D%02dDA', $i;
     ($status,@$name) = Pezca::GetList($detPV,$EZCAFLOAT,$numPts);
     &report_exit($script, $status, $detPV);
  }
}

#=================================================================

# @xarray = &readSRx ($scanrec, $numPts, $motorNumber,[$final-or-progress]);
# sub readSRx ($scanrec, $numPts, $motorNumber) {
sub readSRx ($$$;$) {
###
### This function reads SCAN RECORD @x data from EPICS
### Returns array @x
###

  my ($script, $scanrec, $numPts, $motorNumber, $final_or_progess);
  my ($status, @xarray, $mtrPV);

  $script           = 'readSRx';
  $scanrec          = shift(@_);
  $numPts           = shift(@_);
  $motorNumber      = shift(@_);
  $final_or_progess = shift(@_);
  if (!defined $final_or_progess) {$final_or_progess = 'final';}
  if ($final_or_progess !~ /^fly$/i &&
      $final_or_progess !~ /^progress$/i) {
     $mtrPV = $scanrec.'.P'.$motorNumber.'RA';				#final data
     print STDOUT 'Reading x'.$motorNumber.' from ['
                 .$mtrPV.'] (numPts='.$numPts.')...'."\n";
  } else {
     $mtrPV = $scanrec.'.P'.$motorNumber.'CA';				#fly data
  }
  ($status,@xarray) = Pezca::GetList($mtrPV,'double',$numPts);		#on 2008/10/03
  &report_exit($script, $status, $mtrPV);

  return @xarray;
}

#=================================================================

# @yarray = &readSRy ($scanrec, $numPts, $scalerNumber,[$final-or-progress]);
# sub readSRy ($scanrec, $numPts, $scalerNumber) {
sub readSRy ($$$;$) {

### This function reads SCAN RECORD @y data for given scaler input from EPICS
###
### Returns array @yarray
###

  our $EZCAFLOAT;
  my ($script, $scanrec, $numPts, $scalerNumber, $final_or_progess);
  my ($status, @yarray, $detPV);

  $script           = 'readSRy';
  $scanrec          = shift(@_);
  $numPts           = shift(@_);
  $scalerNumber     = shift(@_);
  $final_or_progess = shift(@_);
  if (!defined $final_or_progess) {$final_or_progess = 'final';}
  if ($final_or_progess !~ /^fly$/i &&
      $final_or_progess !~ /^progress$/i) {
     $detPV = sprintf $scanrec.'.D%02dDA', $scalerNumber;
     print STDOUT 'Reading y'.$scalerNumber.' from ['.$detPV.'] (numPts='.$numPts.')...'."\n";
  } else {
     $detPV = sprintf $scanrec.'.D%02dCA', $scalerNumber;
  }
  ($status,@yarray) = Pezca::GetList($detPV,$EZCAFLOAT,$numPts);
  &report_exit($script, $status, $detPV);
  return @yarray;
}

#=================================================================

# @xarray = &calcSRxMiddle ($numPts, $begin, @x);
# sub calcSRxMiddle ($numPts, @x) {
sub calcSRxMiddle ($$@) {

###
### Returns array @x of averaged X-coordinates for on-the-fly scans with SCAN RECORD
###

  my ($numPts, $begin, @xarray, $i, $x_prev, $x_middle);

  $numPts = shift(@_);
  $begin  = shift(@_);

  $x_prev = $begin;
  for ($i=0; $i<$numPts; $i++) {
     $xarray[$i] = shift(@_);
     $x_middle   = 0.5*($xarray[$i] + $x_prev);
     $x_prev     = $xarray[$i];
     $xarray[$i] = $x_middle;
  }
  return @xarray;
}

#=================================================================

# $xcurr = &readSRx_curr ($scanrec, $motorNumber);
# sub readSRx_curr ($scanrec, $motorNumber) {
sub readSRx_curr ($$) {

###
### Returns current $x for given motor number when SCAN RECORD scan is in progress
###

  my ($script, $scanrec, $motorNumber, $status, $xcurr, $pv);

  $script      = 'readSRx_curr';
  $scanrec     = shift(@_);
  $motorNumber = shift(@_);
  $status = 0;
  $xcurr  = 0;

  $pv = $scanrec.'.R'.$motorNumber.'CV';
  ($status,$xcurr) = Pezca::GetDouble($pv);
  &report_exit($script, $status, $pv);

  return $xcurr;
}

#=================================================================

# $ycurr = &readSRy_curr ($scanrec, $scalerNumber);
# sub readSRx_curr ($scanrec, $scalerNumber) {
sub readSRy_curr ($$) {
###
### Returns current $y for given scaler input when SCAN RECORD scan is in progress
###

  my ($script, $scanrec, $scalerNumber, $status, $ycurr, $pv);

  $script       = 'readSRy_curr';
  $scanrec      = shift(@_);
  $scalerNumber = shift(@_);
  $status       = 0;
  $ycurr        = 0;

  $pv = sprintf $scanrec.'.D%02dCV', $scalerNumber;
  ($status,$ycurr) = Pezca::GetDouble($pv);
  &report_exit($script, $status, $pv);

  return $ycurr;
}

#=================================================================

# sub openSRmedm ($scanrec, $ninputs, \@inputs, [$type]) {	#type: realtime|afterscan
sub openSRmedm ($$\@;$) {

### This subroutine opens on-the-fly plots for Scan Record
### (first $ninputs of ScanRec detectors and scaler only)

  my ($script, $scanrec, $ninputs, $inputsRef, $input, $ninp);
  my ($medmPar, $adl, $status, $i, $scrType);
  my ($current_time, $msg);

  $script    = 'openSRmedm';
  $scanrec   = shift(@_);
  $ninputs   = shift(@_);
  $inputsRef = shift(@_);
  $scrType   = shift(@_);;

  if (!defined $scrType) {$scrType = 'realtime';}
  $scrType = lc($scrType);
  if ($scrType ne 'afterscan') {$scrType = 'realtime';}

  $ninp = @$inputsRef;
  if ($ninputs > $ninp) {
     $current_time = scalar localtime;
     $msg  = '*** '.$current_time."\n";
     $msg .= '*** '.$script.': for scanrec='.$scanrec
                   .' requested scaler inputs='.$ninputs
                   .' exceeds the array length:'."\n"
                   .' @inputs='.$ninp."\n";
     &press_key_exit($msg);
  }

  if ((defined $ENV{FASTSCANS}) && $^O !~ /Win/i) {
     $adl = $ENV{FASTSCANS}.'/adl/fastscans_scanrec_'.$scrType.'.adl';
     for ($i=1; $i<=$ninputs; $i++) {
        $input = @$inputsRef[$i-1];
        $medmPar = sprintf 'P='.$scanrec.',D=D%02d,TITLE=Scan:scaler_input_'.$input, $i;
        $status = &open_medm($adl,$medmPar);
     }
  }
}

#=================================================================

# sub openSRmedm_detec ($scanrec, $ninputs, \@inputs, \@detectors, [$type]) {	#type: realtime|afterscan
sub openSRmedm_detec ($$\@\@;$) {

### This subroutine opens on-the-fly plots for Scan Record
### (given @detectors of ScanRec and scaler only)

  my ($script, $scanrec, $ninputs, $inputsRef, $detectorsRef, $scrType);
  my ($medmPar, $adl, $status, $i, $input, $detector);

  $script       = 'openSRmedm_detec';
  $scanrec      = shift(@_);
  $ninputs      = shift(@_);
  $inputsRef    = shift(@_);
  $detectorsRef = shift(@_);
  $scrType      = shift(@_);

  if (!defined $scrType) {$scrType = 'realtime';}
  $scrType = lc($scrType);
  if ($scrType ne 'afterscan') {$scrType = 'realtime';}

  if (defined $ENV{FASTSCANS} && $^O !~ /Win/i) {
     $adl = $ENV{FASTSCANS}.'/adl/fastscans_scanrec_'.$scrType.'.adl';
     for ($i=0; $i<$ninputs; $i++) {
        $input    = @$inputsRef[$i];
        $detector = @$detectorsRef[$i];
        $medmPar = sprintf 'P='.$scanrec.',D=D%02d,TITLE=Scan:scaler_input_'.$input, $detector;
        $status = &open_medm($adl,$medmPar);
     }
  }
}

#=================================================================

# sub openSRmedm_v2 ($scanrec, $ndetectors, [$type]) {	#type: realtime|afterscan
sub openSRmedm_v2 ($$;$) {

### This subroutine opens on-the-fly plots for Scan Record
### (first $ndetectors of ScanRec and any detector type)

  my ($script, $scanrec, $ndetectors, $detector, $detectorPV, $scrType);
  my ($medmPar, $adl, $status, $i);

  $script     = 'openSRmedm_v2';
  $scanrec    = shift(@_);
  $ndetectors = shift(@_);
  $scrType    = shift(@_);

  if (!defined $scrType) {$scrType = 'realtime';}
  $scrType = lc($scrType);
  if ($scrType ne 'afterscan') {$scrType = 'realtime';}

  if (defined $ENV{FASTSCANS} && $^O !~ /Win/i) {
     $adl = $ENV{FASTSCANS}.'/adl/fastscans_scanrec_'.$scrType.'.adl';
     for ($i=1; $i<=$ndetectors; $i++) {
        $detectorPV = sprintf $scanrec.'.D%02dPV', $i;
        ($status,$detector) = Pezca::GetString($detectorPV);
        if ($status == 0) {
           $medmPar = sprintf 'P='.$scanrec.',D=D%02d,TITLE='.$detector, $i;
        } else {
           $medmPar = sprintf 'P='.$scanrec.',D=D%02d,TITLE=SRdetector-%02d', $i, $i;
        }
        $status = &open_medm($adl,$medmPar);
     }
  }
}

#=================================================================

# sub openSRmedm_pv ($scanrec, \@inputs, [\@pvlist], [$type]) {	#type: realtime|afterscan
sub openSRmedm_pv ($$\@;\@$) {

### This subroutine opens on-the-fly plots for Scan Record
### loaded with scaler inputs AND list of PVs

  my ($script, $scanrec, $inputsRef, $pvlistRef, $scrType, $ninputs, $npvs);
  my ($medmPar, $status, $i, $adl, $detector);

  $script    = 'openSRmedm_pv';
  $scanrec   = shift(@_);
  $inputsRef = shift(@_);
  $pvlistRef = shift(@_);
  $scrType   = shift(@_);		#'realtime', 'afterscan', or undef

  if (defined $pvlistRef) {$npvs = @$pvlistRef;} else {$npvs = 0;}
  if (!defined $scrType)  {$scrType = 'realtime';}
  $scrType = lc($scrType);
  if ($scrType ne 'afterscan') {$scrType = 'realtime';}

  $ninputs   = @$inputsRef;

  if (defined $ENV{FASTSCANS} && $^O !~ /Win/i) {
     $adl = $ENV{FASTSCANS}.'/adl/fastscans_scanrec_'.$scrType.'.adl';
     $detector = 0;
     for ($i=0; $i<$ninputs; $i++) {
        $detector++;
        $medmPar = sprintf 'P='.$scanrec.',D=D%02d,TITLE=Scan:scaler_input_'.@$inputsRef[$i], $detector;
        $status = &open_medm($adl,$medmPar);
     }
     for ($i=0; $i<$npvs; $i++) {
        $detector++;
        $medmPar = sprintf 'P='.$scanrec.',D=D%02d,TITLE=Scan:PV_'.@$pvlistRef[$i], $detector;
        $status = &open_medm($adl,$medmPar);
     }
  }
}

#=================================================================

# sub openSRmedm_mca ($scanrec, \@ROIs, [\@pvlist], [$type]) {	#type: realtime|afterscan
sub openSRmedm_mca ($$\@;\@$) {

### This subroutine opens on-the-fly plots for Scan Record
### loaded with MCA ROIs and list of PVs

  my ($script, $scanrec, $ROIsRef, $pvlistRef, $scrType, $nROIs, $npvs);
  my ($medmPar, $status, $i, $adl, $detector);

  $script    = 'openSRmedm_mca';
  $scanrec   = shift(@_);
  $ROIsRef   = shift(@_);
  $pvlistRef = shift(@_);
  $scrType   = shift(@_);

  if (defined $pvlistRef) {$npvs = @$pvlistRef;} else {$npvs = 0;}
  if (!defined $scrType)  {$scrType = 'realtime';}
  $scrType = lc($scrType);
  if ($scrType ne 'afterscan') {$scrType = 'realtime';}

  $nROIs   = @$ROIsRef;

  if (defined $ENV{FASTSCANS} && $^O !~ /Win/i) {
     $adl = $ENV{FASTSCANS}.'/adl/fastscans_scanrec.adl';
     $detector = 0;
     for ($i=0; $i<$nROIs; $i++) {
        $detector++;
        $medmPar = sprintf 'P='.$scanrec.',D=D%02d,TITLE=Scan:mca_ROI_'.@$ROIsRef[$i], $detector;
        $status = &open_medm($adl,$medmPar);
     }
     for ($i=0; $i<$npvs; $i++) {
        $detector++;
        $medmPar = sprintf 'P='.$scanrec.',D=D%02d,TITLE=PV_'.@$pvlistRef[$i], $detector;
        $status = &open_medm($adl,$medmPar);
     }
  }
}

#=================================================================

# sub openSRmedm_scaler_mca ($scanrec, \@inputs, \@ROIs, [\@pvlist], [$type]) {	#type: realtime|afterscan
sub openSRmedm_scaler_mca ($$\@\@;\@$) {

### This subroutine opens on-the-fly plots for Scan Record
### loaded with MCA ROIs and list of PVs

  my ($script, $scanrec, $inputsRef, $ROIsRef, $pvlistRef);
  my ($scrType, $ninputs, $nROIs, $npvs);
  my ($medmPar, $status, $i, $adl, $detector);

  $script    = 'openSRmedm_mca';
  $scanrec   = shift(@_);
  $inputsRef = shift(@_);
  $ROIsRef   = shift(@_);
  $pvlistRef = shift(@_);
  $scrType   = shift(@_);

  if (defined $pvlistRef) {$npvs = @$pvlistRef;} else {$npvs = 0;}
  if (!defined $scrType)  {$scrType = 'realtime';}
  $scrType = lc($scrType);
  if ($scrType ne 'afterscan') {$scrType = 'realtime';}

  $ninputs = @$inputsRef;
  $nROIs   = @$ROIsRef;

  if (defined $ENV{FASTSCANS} && $^O !~ /Win/i) {
     $adl = $ENV{FASTSCANS}.'/adl/fastscans_scanrec.adl';
     $detector = 0;
     for ($i=0; $i<$ninputs; $i++) {
        $detector++;
        $medmPar = sprintf 'P='.$scanrec.',D=D%02d,TITLE=Scan:scaler_input_'.@$inputsRef[$i], $detector;
        $status = &open_medm($adl,$medmPar);
     }
     for ($i=0; $i<$nROIs; $i++) {
        $detector++;
        $medmPar = sprintf 'P='.$scanrec.',D=D%02d,TITLE=Scan:mca_ROI_'.@$ROIsRef[$i], $detector;
        $status = &open_medm($adl,$medmPar);
     }
     for ($i=0; $i<$npvs; $i++) {
        $detector++;
        $medmPar = sprintf 'P='.$scanrec.',D=D%02d,TITLE=PV_'.@$pvlistRef[$i], $detector;
        $status = &open_medm($adl,$medmPar);
     }
  }
}

#========================================================================

# ($status,$actPts) =
#   &SRflyScan ($assy,$axis,$begin,$end,$npts,$scanTime,$medm,$scaler,$fileBkg,$ninputs,@inputs);
#sub SRflyScan ($assy,$axis,$begin,$end,$npts,$scanTime,$medm,$scaler,$fileBkg,$ninputs,@inputs) {
sub  SRflyScan ($$$$$$$$$$@) {

### This function implements ON-THE-FLY scan with SCAN RECORD, PMAC motors and Joerger scaler
###
### Returns GLOBALS: @x1SR,@y1SR,...@y{$ninputs}SR, @Offsets
### Also returns GLOBAL $messageSRSCAN if $status != 0.

  our ($beamline, $shutter, $shutter_readback, $scanrecName, $scanrecAbort);
  our (@Offsets, $STRUCK_MODEL);
  our ($IGNORE_ERRORS, $report);
  our ($SRSCAN_FLAGS);		#reference to a hash with various flags for SRSCAN
  our ($messageSRSCAN);         #returned message if $status != 0.
  our ($EXTRA_MOTION_TIME);	#extra allowance for motion tracking
  my ($script, $assy, $axis, $begin, $end, $npts, $scanTime, $medm);
  my ($scaler, $fileBkg, $IGNORE_ERRORS_local, $extra_motion_time);
  my ($ninputs, @inputs, $i, $j, $srDeadTime, $struckOverhead, $scanTimeReal);
  my ($scalerNo, $status, $actPts, $name, $cnt_time, $cnt_min, $cnt_timeReal);
  my ($scanrec, $scanAbort, $scanRange, $l_speed, $h_speed, $offsetNr);
  my ($wait_step, $wait_, $wait_start, $busy, $currPts, $inpos, $runprg, $positioned);

  $script   = 'SRflyScan';
  $assy     = shift(@_);
  $axis     = shift(@_);
  $begin    = shift(@_);     # start scan position
  $end      = shift(@_);     # end scan position
  $npts     = shift(@_);     # number of points on fast scan
  $scanTime = shift(@_);     # scan time in seconds
  $medm     = shift(@_);     # 1=open MEDM, 0=not
  $scaler   = shift(@_);     # scaler1=Struck, scaler3=Joerger
  $fileBkg  = shift(@_);     # joerger background file ('auto' for automatic, '' for none)
  $ninputs  = shift(@_);
  for ($i=0; $i<$ninputs; $i++) {$inputs[$i]=shift(@_);}

  if (!defined $IGNORE_ERRORS) {$IGNORE_ERRORS=0;}
  $IGNORE_ERRORS_local = $IGNORE_ERRORS;
  if (!defined $STRUCK_MODEL) {$STRUCK_MODEL='SIS3801';}

  if (!defined $SRSCAN_FLAGS->{'has run'})              {$SRSCAN_FLAGS->{'has run'}              = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter close'})   {$SRSCAN_FLAGS->{'skip shutter close'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter open'})    {$SRSCAN_FLAGS->{'skip shutter open'}    = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip calc speed'})      {$SRSCAN_FLAGS->{'skip calc speed'}      = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip set speed'})       {$SRSCAN_FLAGS->{'skip set speed'}       = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip set slew'})        {$SRSCAN_FLAGS->{'skip set slew'}        = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip restore speed'})   {$SRSCAN_FLAGS->{'skip restore speed'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip restore mprg'})    {$SRSCAN_FLAGS->{'skip restore mprg'}    = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip move to start'})   {$SRSCAN_FLAGS->{'skip move to start'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'return to start'})      {$SRSCAN_FLAGS->{'return to start'}      = 0;}
#
  if (!defined $SRSCAN_FLAGS->{'speed'})                {$SRSCAN_FLAGS->{'speed'}                = 0;}

  $srDeadTime     = 0.005;   # ScanRec needs some time to finish/erase
  $struckOverhead = 0.008;   # Struck needs some time to finish/erase
  $cnt_min        = 0.1;     # minimum counting time
  $wait_start     = 4.0;
  $status         = 0;
  $actPts         = 0;
  $cnt_time       = 0;
  $messageSRSCAN  = '';
  if (!defined $STRUCK_MODEL) {$STRUCK_MODEL='SIS3801';}
  if ($STRUCK_MODEL eq 'SIS3820') {$struckOverhead=0;}

  if (!defined $fileBkg) {$fileBkg='';}
  if (lc($fileBkg) eq 'auto') {
     $scalerNo = $scaler;
     $scalerNo =~ s/^.*\D+//;	# strip out all non-numeric (\D) stuff at the beginning
     $fileBkg  = 'joerger'.$scalerNo.'_Bkg.dat';
  }

  if ($npts > 1) {$cnt_time = $scanTime / ($npts - 1);}
  else           {$cnt_time = $scanTime;              }
  if ($cnt_time < $cnt_min) {
     $messageSRSCAN = sprintf 'Too short time per point=%.3fs, the minimum is %.3fs', $cnt_time,$cnt_min;
     print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
     return 666, 0;
  }
  $cnt_timeReal = $cnt_time + $srDeadTime;
  if ($scaler eq $beamline.':scaler1') {
     $cnt_timeReal = $cnt_timeReal + $struckOverhead;	# Struck needs some extra time to finish
  }
  if ($npts > 1) {$scanTimeReal = $cnt_timeReal * ($npts - 1);}
  else           {$scanTimeReal = $cnt_timeReal;}
  if ($SRSCAN_FLAGS->{'has run'} == 0 && $scanTimeReal > 1.5*$scanTime) {
     printf STDOUT '!!! '.$script.': because of Scaler and ScanRec '
                                 .'overheads, the scantime will be '
                                 .'%1.0fs instead of %1.0f !'."\n",
                                  $scanTimeReal, $scanTime;
  }
#-----------------------------------------------------------------
### The globals below are defined in beamline.h:
  $scanrec   = $scanrecName;
  $scanAbort = $scanrecAbort;
  if (defined $shutter) {
     if ($shutter !~ /^${beamline}:/) {$shutter   = $beamline.':'.$shutter;}
  }
  if (defined $shutter_readback) {
     if ($shutter_readback !~ /^${beamline}:/) {
                                $shutter_readback = $beamline.':'.$shutter_readback;}
  }
  if ($scanrec   !~ /^${beamline}:/)  {$scanrec   = $beamline.':'.$scanrec;}
  if ($scanAbort !~ /^${beamline}:/)  {$scanAbort = $beamline.':'.$scanAbort;}
#-----------------------------------------------------------------
  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
     print STDOUT 'Closing shutter ['.$shutter.']'."\n";
     &closeShutter ($shutter,$shutter_readback);
  } else {
     if ($SRSCAN_FLAGS->{'has run'} == 0) {print STDOUT '!!! '.$script.': shutter is not used!'."\n";}
  }

  if ($SRSCAN_FLAGS->{'has run'} == 0 || $SRSCAN_FLAGS->{'skip calc speed'} == 0) {
### Calculate scan speed for the motor:
     print STDOUT 'Calculating slew-mode motor speed for ['.$axis.']'."\n";
     $scanRange = abs ($end - $begin);
     $IGNORE_ERRORS = 1;
     ($l_speed, $h_speed) = &calcAxisSpeed ($assy, $axis, $scanRange, $scanTimeReal);
     $IGNORE_ERRORS = $IGNORE_ERRORS_local;
     if ($l_speed > 1.01*$h_speed) {
        $messageSRSCAN = sprintf 'Calculated low speed=%.3f exceeds max drive speed=%.3f',$l_speed,$h_speed;
        print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
        return 777, 0;
     }
     $SRSCAN_FLAGS->{'speed'} = $l_speed;
  } else {
#    printf STDOUT '!!! Using '.$axis.' slew-mode speed=%.3f from previous scan'."\n",$SRSCAN_FLAGS->{'speed'};
#    print  STDOUT '!!! (This presumes same range and time)'."\n";
     $l_speed = $SRSCAN_FLAGS->{'speed'};
  }

### Read the scaler background offsets:
  if (!defined $Offsets[0]) {	# skip reading offsets if they were read in main
     if (length($fileBkg) > 0) {print STDOUT 'Reading scaler ['.$scaler.'] offsets from '.$fileBkg."\n";}
     else                      {print STDOUT 'No scaler ['.$scaler.'] offsets file specified. Setting zero offsets.'."\n";}
     @Offsets = &getJoergerOffsets ($fileBkg, $scaler);  # we make it GLOBAL to access from main
  }

### set scaler:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up scaler ['.$scaler.']'."\n";
     &setJoerger($scaler,$cnt_time);
  }

### Set motor:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up motor ['.$axis.']'."\n";
     &configAssy ($assy);
     &restoreAxis ($assy, $axis);		#restore high-speed
  }

### Set Scan Record:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
#    print STDOUT 'Clearing scan record ['.$scanrec.'], ['.$scanAbort.']'."\n";
     &clearSR ($scanrec);

     print STDOUT 'Setting up scan record ['.$scanrec.'] positioner'."\n";
     &setupSRmotors ($scanrec, $npts, 1, [$axis.':ActPos'], [$axis.':RqsPos'], [$begin], [$end]);

     print STDOUT 'Setting up scan record ['.$scanrec.'] detector(s): ';
     &setupSRdetectors_v2 ($scanrec, $scaler, $ninputs, \@inputs);
     print STDOUT '... nDetectors='.$ninputs."\n";
  } else {
     print STDOUT 'Changing scan record ['.$scanrec.'] drive range'."\n";
     &changeSRmotors ($scanrec, $npts, 1, [$begin], [$end]);
  }

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  if ($SRSCAN_FLAGS->{'skip move to start'} == 0) {
### Move motor to the starting position:
     printf STDOUT 'Moving motor ['.$axis.'] to the starting position (%.3g)'."\n",$begin;
### Increase expected motion time due to possibly linked motor#32:
     if (defined $EXTRA_MOTION_TIME) {$extra_motion_time = $EXTRA_MOTION_TIME;}
     $EXTRA_MOTION_TIME = 600;
     $status = &runMotor ($assy, $axis, $begin);
### Restore expected motion time to normal from local cache:
     if (defined $extra_motion_time) {$EXTRA_MOTION_TIME = $extra_motion_time;}
     else                            {undef $EXTRA_MOTION_TIME;}
     if ($status) {
        $messageSRSCAN = 'Failed moving ['.$axis.'] to the starting position.';
        print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
        return $status, 0;
     }
  }

### Setting slow speed for scanning:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
        printf STDOUT 'Setting up motor ['.$axis.'] speed=%.3g & motionPRG=slew'."\n",$l_speed;
        &setAxisSlewScan ($assy, $axis, $l_speed);
  } else {
     if ($SRSCAN_FLAGS->{'skip set slew'} == 0 && $SRSCAN_FLAGS->{'skip set speed'} == 0) {
        printf STDOUT 'Setting up motor ['.$axis.'] speed=%.3g & motionPRG=slew.'."\n",$l_speed;
        &setAxisSlewScan ($assy, $axis, $l_speed);			#setMotionPrg+setAxisSpeed
     }
     elsif ($SRSCAN_FLAGS->{'skip set slew'} && $SRSCAN_FLAGS->{'skip set speed'} == 0) {
        printf STDOUT $script.': setting scanning speed=%.3g only, but presuming motionPRG=slew.'."\n",$l_speed;
        &setAxisSpeed ($assy, $axis, $l_speed); 		#this is faster (no motion prg change)
     }
     elsif ($SRSCAN_FLAGS->{'skip set slew'} == 0 && $SRSCAN_FLAGS->{'skip set speed'}) {
        print STDOUT $script.': setting motionPRG=slew, but presuming scanning speed is set.'."\n";
        &setMotionPrg($assy, 1); 				#this is faster (no speed change): 1=slew
     }
  }

### Setting monitors on motors motion and scan record execution:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     $status = &addCaMonitor($scanrec.'.BUSY');
     $status = &addCaMonitor($scanrec.'.CPT');
     $status = &setAssyMonitors($assy);
  }

  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter open'} == 0) {
     print STDOUT 'Opening shutter ['.$shutter.']'."\n";
     $status = &openShutter ($shutter,$shutter_readback);
     if ($status) {
        &abortAssy ($assy, 1, $axis);		# also restore speed and motion prg.
        &closeShutter($shutter,$shutter_readback);
        $messageSRSCAN = 'Failed opening fast shutter ['.$shutter.'].';
        print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
        return $status;
     }
  }

  $wait_step     = 0.1;
  $wait_         = 0;
  $busy          = 1;
  $currPts       = 0;
  $inpos         = 0;
  $runprg        = 1;
  $positioned    = 0;

### Start scan record:
  print STDOUT 'Starting on-the-fly scan with scan record ['.$scanrec.']...'."\n";
  &startSR ($scanrec, $scanAbort);
# &press_key_continue();

  if ($medm && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm($scanrec,$ninputs,\@inputs,'realtime');
  }

### Wait to finish (don't give up first $wait_start seconds):
  while ((($busy ==1 || $currPts < 1) && ($runprg || $positioned<3)) || $wait_< $wait_start) {
     select(undef,undef,undef,$wait_step);	# sleep($wait_step);
     $wait_ += $wait_step;
     ($status,$busy)   = Pezca::GetDouble($scanrec.'.BUSY');
     ($status,$currPts)= Pezca::GetDouble($scanrec.'.CPT');
     ($status,$inpos)  = Pezca::GetDouble($assy.':InPos');
     ($status,$runprg) = Pezca::GetDouble($assy.':RunPrg');
     if ($inpos == 1) { $positioned++; }
     else             { $positioned=0; }
#    print STDOUT 'busy='.$busy.' currPts='.$currPts.' runprg='.$runprg.' inpos='
#                 .$inpos.' positioned='.$positioned.' wait_='.$wait_."\n";
  }
  printf STDOUT 'Scan finished! (scanBusy='.$busy.', scanCpt='.$currPts
               .', mtrRunPrg='.$runprg.', mtrPos='.$positioned.', time=%.1fs)'."\n",$wait_;
  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
     print STDOUT 'Closing shutter ['.$shutter.']'."\n";
     &closeShutter ($shutter,$shutter_readback);
  }
  &abortSR ($scanAbort, $scanrec);
  &abortAssy ($assy, 1, $axis);		# also restore speed and motion prg.

### Wait the data to be ready:
  print STDOUT 'Waiting for scan record data...'."\n";
  $actPts = &waitSRdata ($scanrec);

### Start reverting motor to the beginning position
### in order to save time on the next scan:
  if ($SRSCAN_FLAGS->{'return to start'}) {&startMotor ($axis, $begin);}

  print STDOUT 'Reading scan record data: npts='.$actPts.' nDetectors='.$ninputs."\n";

  if ($medm && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm($scanrec,$ninputs,\@inputs,'afterscan');
  }

  no strict 'refs';
  $name   = 'x1SR';
  @$name = &readSRx ($scanrec, $actPts, 1);
  @$name = &calcSRxMiddle ($actPts, $begin, @$name);

  for ($i=1; $i <= $ninputs; $i++) {
     $name   = 'y'.$i.'SR';
     @$name = &readSRy ($scanrec, $actPts, $i);
### Normalize data for time and correct for background:
     $offsetNr = $inputs[$i-1];
     for ($j=0; $j < $actPts; $j++) {
        $$name[$j] = $$name[$j]/$cnt_time - $Offsets[$offsetNr-1];
     }
  }
  use strict 'refs';
  print STDOUT 'Scan record data retrieved!'."\n";
  $SRSCAN_FLAGS->{'has run'} = 1;
  return $status,$actPts;

}	##SRflyScan

#========================================================================

# ($status,$actPts) =
#   &SRflyScan2 ($assy,$axis,$begin,$end,$npts,$scanTime,$medm,$scaler,$fileBkg,\@inputs,
#               [\@pvlist,$afterScan,$checkOverload,$postarrayPV,$post_interval]);
#sub SRflyScan2 ($assy,$axis,$begin,$end,$npts,$scanTime,$medm,$scaler,$fileBkg,\@inputs,
#               [\@pvlist,$afterScan,$checkOverload,$postarrayPV,$post_interval]) {
sub  SRflyScan2 ($$$$$$$$$\@;\@$$$$) {

### This function implements ON-THE-FLY scan with SCAN RECORD, PMAC motors and Joerger scaler + PVlist
###
### Returns GLOBALS: @x1SR,@y1SR,...@y{$ninputs+$npvs}SR, @Offsets
### Also returns GLOBAL $messageSRSCAN if $status != 0.

### This version has two differences from SRflyScan:
### (1) scaler inputs are transferred by reference
### (2) It also accepts optional PVlist to record during scan
###    (this list added to SR detectors after scaler inputs)

  our ($beamline, $shutter, $shutter_readback, $scanrecName, $scanrecAbort);
  our (@Offsets, $STRUCK_MODEL);
  our ($IGNORE_ERRORS, $report);
  our ($SRSCAN_FLAGS);		#reference to a hash with various flags for SRSCAN
  our ($messageSRSCAN);         #returned message if $status != 0.
  our ($EXTRA_MOTION_TIME);	#extra allowance for motion tracking
  our (@x1SR, @y1SR);		#GLOBAL RESULTs: positions and intensity
  my ($script, $assy, $axis, $begin, $end, $npts, $scanTime, $medm);
  my ($scaler, $fileBkg, $inputsRef, $pvlistRef, $afterScan);
  my ($checkOverload, $postarrayPV, $post_interval);
  my ($IGNORE_ERRORS_local, $extra_motion_time);
  my ($srDeadTime, $struckOverhead);
  my ($ninputs, $npvs, $ndetectors, $actPts, $cnt_time, $cnt_min);
  my ($scanrec, $scanAbort, $scalerNo, $status, $name, $i, $j);
  my ($scanRange, $l_speed, $h_speed, $offsetNr, $d0, $d1);
  my ($scanTimeReal, $cnt_timeReal);
  my ($wait_step, $wait_, $wait_start, $busy);
  my ($currPts, $inpos, $runprg, $positioned);
  my ($i_plot, $name_plot, $scan_index);
  my ($threshold, $OVERLOAD, $overload_detected);
  my ($time_epoch, $time_passed, $npts_last);
  my ($min_val, $narr, @post_array, $axisIni, $nelm);

  $script        = 'SRflyScan2';
  $assy          = shift(@_);
  $axis          = shift(@_);
  $begin         = shift(@_);	# start scan position
  $end           = shift(@_);	# end scan position
  $npts          = shift(@_);	# number of points on fast scan
  $scanTime      = shift(@_);	# scan time in seconds
  $medm          = shift(@_);	#1=open MEDM for scaler inputs, 2=for inputs+PVs, 0=not
  $scaler        = shift(@_);	#scaler1=Struck, scaler3=Joerger
  $fileBkg       = shift(@_);	#joerger background file ('auto' for automatic, '' for none)
  $inputsRef     = shift(@_);
  $pvlistRef     = shift(@_);	#optional
  $afterScan     = shift(@_);	#0=stay 1=startpos 2=priorpos 3=maxpos 4,14=peakpos(noisy) 5,15=peakpos(smooth)
                                #6=minpos 7,17=valleypos(noisy) 8,18=valleypos(smooth) 10,110=gaussian 11,111=lorentzian
  $checkOverload = shift(@_);
  $postarrayPV   = shift(@_);   #array for posting partial & final data to JBluIce
  $post_interval = shift(@_);

  if (!defined $fileBkg)       {$fileBkg='';}
  if (!defined $afterScan)     {$afterScan=0;}          #stay
  if (!defined $post_interval) {$post_interval = 2;}	#post interval in seconds
  if (!defined $checkOverload) {$checkOverload = 0;}

  if (!defined $IGNORE_ERRORS) {$IGNORE_ERRORS=0;}
  $IGNORE_ERRORS_local = $IGNORE_ERRORS;

  if (!defined $SRSCAN_FLAGS->{'has run'})              {$SRSCAN_FLAGS->{'has run'}              = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter close'})   {$SRSCAN_FLAGS->{'skip shutter close'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter open'})    {$SRSCAN_FLAGS->{'skip shutter open'}    = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip calc speed'})      {$SRSCAN_FLAGS->{'skip calc speed'}      = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip set speed'})       {$SRSCAN_FLAGS->{'skip set speed'}       = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip set slew'})        {$SRSCAN_FLAGS->{'skip set slew'}        = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip restore speed'})   {$SRSCAN_FLAGS->{'skip restore speed'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip restore mprg'})    {$SRSCAN_FLAGS->{'skip restore mprg'}    = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip move to start'})   {$SRSCAN_FLAGS->{'skip move to start'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'return to start'})      {$SRSCAN_FLAGS->{'return to start'}      = 0;}
#
  if (!defined $SRSCAN_FLAGS->{'speed'})                {$SRSCAN_FLAGS->{'speed'}                = 0;}

  $srDeadTime        = 0.005;			#ScanRec needs some time to finish/erase
  $struckOverhead    = 0.008;			#Struck needs some time to finish/erase
  $cnt_min           = 0.1;			#minimum counting time
  $wait_start        = 4.0;
  $i_plot            = 1;			#the main plot = 1st input ($inputs[0])
  $name_plot         = 'y'.$i_plot.'SR';	#intensity array to post
  $threshold         = 0.95E+6;			#threshold intensity for detecting overload (2013)
  $OVERLOAD          = 100000;		        #scan failure flag when overload detected
  $overload_detected = 0;	                #overload not detected yet
  $time_epoch        = time();           	#seconds since epoch
  $scan_index        = 0;                   	#scan index to save with the post (e.g. 11-vertical, 21-horizontal)
  $messageSRSCAN     = '';
  $npts_last         = 0;

  if (!defined $STRUCK_MODEL) {$STRUCK_MODEL='SIS3801';}
  if ($STRUCK_MODEL eq 'SIS3820') {$struckOverhead=0;}

  if (lc($fileBkg) eq 'auto') {
     $scalerNo = $scaler;
     $scalerNo =~ s/^.*\D+//;	# strip out all non-numeric (\D) stuff at the beginning
     $fileBkg  = 'joerger'.$scalerNo.'_Bkg.dat';
  }
  $ninputs = @$inputsRef;
  if (defined $pvlistRef) {$npvs = @$pvlistRef;} else {$npvs = 0;}
  $ndetectors = $ninputs + $npvs;

  if ($npts > 1) {$cnt_time = $scanTime / ($npts - 1);}
  else           {$cnt_time = $scanTime;}
  if ($cnt_time < $cnt_min) {
     $messageSRSCAN = sprintf 'Too short time per point=%.3fs, the minimum is %.3fs', $cnt_time,$cnt_min;
     print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
     return 666, 0;
  }
  $cnt_timeReal = $cnt_time + $srDeadTime;
  if ($scaler eq $beamline.':scaler1') {
     $cnt_timeReal = $cnt_timeReal + $struckOverhead;	# Struck needs some extra time to finish
  }
  if ($npts > 1) {$scanTimeReal = $cnt_timeReal * ($npts - 1);}
  else           {$scanTimeReal = $cnt_timeReal;}
  if ($SRSCAN_FLAGS->{'has run'} == 0 && $scanTimeReal > 1.5*$scanTime) {
     printf STDOUT '!!! '.$script.': because of Scaler and ScanRec '
                                 .'overheads, the scantime will be '
                                 .'%1.0fs instead of %1.0f !'."\n",
                                   $scanTimeReal, $scanTime;
  }
#-----------------------------------------------------------------
### The globals below are defined in beamline.h:
  $scanrec   = $scanrecName;
  $scanAbort = $scanrecAbort;
  if (defined $shutter) {
     if ($shutter !~ /^${beamline}:/) {$shutter   = $beamline.':'.$shutter;}
  }
  if (defined $shutter_readback) {
     if ($shutter_readback !~ /^${beamline}:/) {
                                $shutter_readback = $beamline.':'.$shutter_readback;}
  }
  if ($scanrec   !~ /^${beamline}:/)  {$scanrec   = $beamline.':'.$scanrec;}
  if ($scanAbort !~ /^${beamline}:/)  {$scanAbort = $beamline.':'.$scanAbort;}
#-----------------------------------------------------------------
  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
     print STDOUT 'Closing shutter ['.$shutter.']'."\n";
     &closeShutter ($shutter,$shutter_readback);
  } else {
     if ($SRSCAN_FLAGS->{'has run'} == 0) {print STDOUT '!!! '.$script.': shutter is not used!'."\n";}
  }

  if ($SRSCAN_FLAGS->{'has run'} == 0 || $SRSCAN_FLAGS->{'skip calc speed'} == 0) {
### Calculate scan speed for the motor:
     print STDOUT 'Calculating slew-mode motor speed for ['.$axis.']'."\n";
     $scanRange = abs ($end - $begin);
     $IGNORE_ERRORS = 1;
     ($l_speed,$h_speed) = &calcAxisSpeed ($assy,$axis,$scanRange,$scanTimeReal);
     $IGNORE_ERRORS = $IGNORE_ERRORS_local;
     if ($l_speed > 1.01*$h_speed) {
        $messageSRSCAN = sprintf 'Calculated low speed=%.3f exceeds max drive speed=%.3f',$l_speed,$h_speed;
        print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
        return 777, 0;
     }
     $SRSCAN_FLAGS->{'speed'} = $l_speed;
  } else {
#    printf STDOUT '!!! Using '.$axis.' slew-mode speed=%.3f from previous scan'."\n",$SRSCAN_FLAGS->{'speed'};
#    print  STDOUT '!!! (This presumes same range and time)'."\n";
     $l_speed = $SRSCAN_FLAGS->{'speed'};
  }

  if (defined $postarrayPV) {
     ($status,$nelm) = Pezca::GetDouble($postarrayPV.'.NELM');
     if ($status) {&report_only($script, $status, $postarrayPV.'.NELM'); $nelm = 2048;}
  }

### Read the scaler background offsets:
  if (!defined $Offsets[0]) {	# skip reading offsets if they were read in main
     if (length($fileBkg) > 0) {print STDOUT 'Reading scaler ['.$scaler.'] offsets from '.$fileBkg."\n";}
     else                      {print STDOUT 'No scaler ['.$scaler.'] offsets file specified. Setting zero offsets.'."\n";}
     @Offsets = &getJoergerOffsets ($fileBkg,$scaler);  # we make it GLOBAL to access from main
  }

### set scaler:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up scaler ['.$scaler.']'."\n";
     &setJoerger($scaler,$cnt_time);
  }

### set motor:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up motor ['.$axis.']'."\n";
     &configAssy ($assy);
     &restoreAxis ($assy, $axis);			#restore high-speed
     ### Record initial axis position:
     $axisIni = &getAxisAct($axis,\$status);
     $SRSCAN_FLAGS->{'inipos'} = $axisIni;
  } else {
     if (defined $SRSCAN_FLAGS->{'inipos'}) {$axisIni = $SRSCAN_FLAGS->{'inipos'};}
     else                                   {$axisIni = $begin;}
  }

### set Scan Record:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
#    print STDOUT 'Clearing scan record ['.$scanrec.'], ['.$scanAbort.']'."\n";
     &clearSR ($scanrec);

     print STDOUT 'Setting up scan record ['.$scanrec.'] positioner'."\n";
### This sets motors stay after scan:
     &setupSRmotors ($scanrec, $npts, 1, [$axis.':ActPos'] , [$axis.':RqsPos'], [$begin], [$end]);

     print STDOUT 'Setting up scan record ['.$scanrec.'] detector(s): ';
     &setupSRdetectors_pv ($scanrec, $scaler, $inputsRef, $pvlistRef);
     printf STDOUT '... nDetectors='.$ninputs.'+'.$npvs.'=%d'."\n",$ndetectors;
  } else {
     print STDOUT 'Changing scan record ['.$scanrec.'] drive range'."\n";
     &changeSRmotors ($scanrec, $npts, 1, [$begin], [$end]);
  }

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  if ($SRSCAN_FLAGS->{'skip move to start'} == 0) {
### Move motor to the starting position:
     printf STDOUT 'Moving motor ['.$axis.'] to the starting position (%.3g)'."\n",$begin;
### Increase expected motion time due to possibly linked motor#32:
     if (defined $EXTRA_MOTION_TIME) {$extra_motion_time = $EXTRA_MOTION_TIME;}
     $EXTRA_MOTION_TIME = 600;
     $status = &runMotor ($assy, $axis, $begin);
### Restore expected motion time to normal from local cache:
     if (defined $extra_motion_time) {$EXTRA_MOTION_TIME = $extra_motion_time;}
     else                            {undef $EXTRA_MOTION_TIME;}
     if ($status) {
        $messageSRSCAN = 'Failed moving ['.$axis.'] to the starting position.';
        print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
        return $status, 0;
     }
  }

### Setting slow speed for scanning:
  printf STDOUT 'Setting up motor ['.$axis.'] speed to slew (%.3g)'."\n",$l_speed;
  &setAxisSlewScan ($assy, $axis, $l_speed);
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
        printf STDOUT 'Setting up motor ['.$axis.'] speed=%.3g & motionPRG=slew'."\n",$l_speed;
        &setAxisSlewScan ($assy, $axis, $l_speed);
  } else {
     if ($SRSCAN_FLAGS->{'skip set slew'} == 0 && $SRSCAN_FLAGS->{'skip set speed'} == 0) {
        printf STDOUT 'Setting up motor ['.$axis.'] speed=%.3g & motionPRG=slew.'."\n",$l_speed;
        &setAxisSlewScan ($assy, $axis, $l_speed);			#setMotionPrg+setAxisSpeed
     }
     elsif ($SRSCAN_FLAGS->{'skip set slew'} && $SRSCAN_FLAGS->{'skip set speed'} == 0) {
        printf STDOUT $script.': setting scanning speed=%.3g only, but presuming motionPRG=slew.'."\n",$l_speed;
        &setAxisSpeed ($assy, $axis, $l_speed); 		#this is faster (no motion prg change)
     }
     elsif ($SRSCAN_FLAGS->{'skip set slew'} == 0 && $SRSCAN_FLAGS->{'skip set speed'}) {
        print STDOUT $script.': setting motionPRG=slew, but presuming scanning speed is set.'."\n";
        &setMotionPrg($assy, 1); 				#this is faster (no speed change): 1=slew
     }
  }

### Setting monitors on motors motion and scan record execution:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     $status = &addCaMonitor($scanrec.'.BUSY');
     $status = &addCaMonitor($scanrec.'.CPT');
     $status = &setAssyMonitors($assy);
  }

  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter open'} == 0) {
     print STDOUT 'Opening shutter ['.$shutter.']'."\n";
     $status = &openShutter ($shutter,$shutter_readback);
     if ($status) {
        &abortAssy ($assy, 1, $axis);		# also restore speed and motion prg.
        &closeShutter($shutter,$shutter_readback);
        $messageSRSCAN = 'Failed opening fast shutter ['.$shutter.'].';
        print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
        return $status;
     }
  }

  $wait_step     = 0.1;
  $wait_         = 0;
  $busy          = 1;
  $currPts       = 0;
  $inpos         = 0;
  $runprg        = 1;
  $positioned    = 0;

### Start scan record:
  print STDOUT 'Starting on-the-fly scan with Scan Record ['.$scanrec.']...'."\n";
  &startSR ($scanrec, $scanAbort);
# &press_key_continue();

  if    ($medm == 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_pv($scanrec, $inputsRef, undef, 'realtime');
  }
  elsif ($medm  > 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_pv($scanrec, $inputsRef, $pvlistRef, 'realtime');
  }

### Wait to finish (don't give up first $wait_start seconds):
  while ((($busy == 1 || $currPts < 1) && ($runprg || $positioned<3)) || $wait_< $wait_start) {
     select(undef,undef,undef,$wait_step);	# sleep($wait_step);
     $wait_ += $wait_step;
     ($status,$busy)   = Pezca::GetDouble($scanrec.'.BUSY');
     ($status,$currPts)= Pezca::GetDouble($scanrec.'.CPT');
     ($status,$inpos)  = Pezca::GetDouble($assy.':InPos');
     ($status,$runprg) = Pezca::GetDouble($assy.':RunPrg');
     if ($inpos == 1) { $positioned++; }
     else             { $positioned=0; }
#    print STDOUT 'busy='.$busy.' currPts='.$currPts.' runprg='.$runprg
#                .' inpos='.$inpos.' positioned='.$positioned.' wait_='.$wait_."\n";
     if (defined $postarrayPV || $checkOverload) {  #----Begin-the-post-array-bisiness------------
        $time_passed = time() - $time_epoch;
        if ($time_passed >= $post_interval) {
           $actPts = &getSRcurrPts($scanrec); 
           if ($actPts > $npts_last+1) {
              $time_epoch = time();			#reset seconds since epoch
              $npts_last  = $actPts;
              @x1SR = &readSRx ($scanrec, $actPts, 1, 'fly');
              @x1SR = &calcSRxMiddle ($actPts, $begin, @x1SR);
              @y1SR = &readSRy ($scanrec, $actPts, $i_plot, 'fly');
              for ($j=0; $j<$actPts; $j++) {
                 $y1SR[$j] = $y1SR[$j]/$cnt_time;
                 if ($checkOverload) {if ($y1SR[$j] >= $threshold) {$overload_detected++;}}
              }
              if (defined $postarrayPV) {
                 $min_val = &minvalRef(\@y1SR);
                 @post_array = ($scan_index,$npts_last,$begin,$end,@x1SR);
                 for ($j=0; $j<$actPts; $j++) {push @post_array,($y1SR[$j]-$min_val);}
                 $narr = @post_array;
                 if ($narr <= $nelm) {
                    $status = Pezca::PutList($postarrayPV, 'FLOAT', @post_array);
                    if ($status) {
                       &report_only($script, $status, $postarrayPV);
                       print STDOUT '!!! '.$script.': error posting '.$narr.'-element array into PV='
                                         .$postarrayPV.'. Further posting will be disabled.'."\n\n";
                       undef $postarrayPV;
                    }
                 } else {
                    print STDOUT '!!! '.$script.': cannot post '.$narr.'-element array into PV='
                                      .$postarrayPV.'. Array is too long. Further posting will be disabled.'."\n\n";
                    undef $postarrayPV;
                 }
              }
              if ($overload_detected > 0) {
                 print STDOUT '!!! '.$script.': Intensity overload detected '.$overload_detected.' times.'."\n";
                 $busy = 0; $positioned = 3;
              }
           }
        }
     }
  }
  printf STDOUT 'Scan finished! (scanBusy='.$busy.', scanCpt='.$currPts.', mtrRunPrg='
               .$runprg.', mtrPos='.$positioned.', time=%.1fs)'."\n",$wait_;
  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
     print STDOUT 'Closing shutter ['.$shutter.']'."\n";
     &closeShutter ($shutter,$shutter_readback);
  }
  &abortSR ($scanAbort, $scanrec);
  &abortAssy ($assy, 1, $axis);		# also restore speed and motion prg.

  if ($overload_detected) {
     $messageSRSCAN = 'Intensity overload detected. Returning '.$axis.' to starting position for re-scanning...';
     print STDOUT '!!! '.$script.': '.$messageSRSCAN."\n";
     &runMotor($assy,$axis,$begin,'forceAuto');
     return $OVERLOAD,$actPts;
  }

### Wait the data to be ready:
  print STDOUT 'Waiting for scan record data...'."\n";
  $actPts = &waitSRdata ($scanrec);

### Start reverting motor to the beginning position
### in order to save time on the next scan:
  if ($SRSCAN_FLAGS->{'return to start'}) {&startMotor ($axis, $begin);}

  print STDOUT 'Reading scan record data: npts='.$actPts.' nDetectors='.$ndetectors."\n";

  if    ($medm == 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_pv($scanrec, $inputsRef, undef, 'afterscan');
  }
  elsif ($medm  > 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_pv($scanrec, $inputsRef, $pvlistRef,'afterscan');
  }

  @x1SR = &readSRx ($scanrec, $actPts, 1);
  @x1SR = &calcSRxMiddle ($actPts, $begin, @x1SR);

  no strict 'refs';
  for ($i=1; $i<=$ndetectors; $i++) {
     $name  = 'y'.$i.'SR';
     @$name = &readSRy ($scanrec, $actPts, $i);
     if ($i <= $ninputs) {
### Normalize scaler data for time and correct for background:
        $offsetNr = @$inputsRef[$i-1];
        for ($j=0; $j<$actPts; $j++) {
           @$name[$j] = @$name[$j]/$cnt_time;
           if ($checkOverload && $i == $i_plot && @$name[$j] >= $threshold) {$overload_detected++;}
           @$name[$j] = @$name[$j] - $Offsets[$offsetNr-1];
        }
        if (defined $postarrayPV && $actPts > $npts_last && $i == $i_plot) {
           $min_val = &minvalRef(\@$name);
           @post_array = ($scan_index,$actPts,$begin,$end,@x1SR);
           for ($j=0; $j<$actPts; $j++) {push @post_array,(@$name[$j]-$min_val);}
           $narr = @post_array;
           if ($narr <= $nelm) {
              $status = Pezca::PutList($postarrayPV, 'FLOAT', @post_array);
              if ($status) {
                 &report_only($script, $status, $postarrayPV);
                 print STDOUT '!!! '.$script.': error posting '.$narr.'-element array into PV='
                             .$postarrayPV.'. Further posting will be disabled.'."\n\n";
                 undef $postarrayPV;
              }
           } else {
              print STDOUT '!!! '.$script.': cannot post '.$narr.'-element array into PV='
                          .$postarrayPV.'. Array is too long. Further posting will be disabled.'."\n\n";
              undef $postarrayPV;
           }
        }
     }
     else  {
### Average PV data over measurements intervals:
        $d0 = @$name[0];
        for ($j=1; $j<$actPts; $j++) {$d1=@$name[$j]; @$name[$j]=0.5*($d1+$d0); $d0=$d1;}
     }
  }
  use strict 'refs';
  print STDOUT 'Scan record data retrieved!'."\n";
  $SRSCAN_FLAGS->{'has run'} = 1;

  if ($overload_detected) {
     $messageSRSCAN = 'Intensity overload detected. Returning '.$axis.' to starting position for re-scanning...';
     print STDOUT '!!! '.$script.': '.$messageSRSCAN."\n";
     &runMotor($assy,$axis,$begin,'forceAuto');
     return $OVERLOAD,$actPts;
  }

### -----------------------------------------------------
### afterScan:
### (see motor_subs.pl)
  no strict 'refs';  #-----------------------------------------------------------------------#
  &axis_afterScan($afterScan,\@x1SR,\@$name_plot,$begin,$end,$axisIni,$axis,$assy);          #
  use strict 'refs'; #-----------------------------------------------------------------------#
  print STDOUT '..'.$script.' finished!'."\n";
  return $status,$actPts;
}	##SRflyScan2

#========================================================================

# ($status,$actPts) =
#   &SRflyScanMCA ($assy,$axis,$begin,$end,$npts,$scanTime,$medm,$mca,\@ROIs,[\@pvlist]);
#sub SRflyScanMCA ($assy,$axis,$begin,$end,$npts,$scanTime,$medm,$mca,\@ROIs,[\@pvlist]) {
sub  SRflyScanMCA ($$$$$$$$\@;\@) {

### This function implements ON-THE-FLY scan with SCAN RECORD, PMAC motors and MCA ROI + PVlist
###
### Returns GLOBALS: @x1SR,@y1SR,...@y{$nROIs+$npvs}SR
### Also returns GLOBAL $messageSRSCAN if $status != 0.

### This version works with MCA instead of Joergey-type scaler
### It also accepts optional PVlist to record during scan
###    (this list added to SR detectors after MCA ROIs)

  our ($beamline, $shutter, $shutter_readback, $scanrecName, $scanrecAbort);
  our ($IGNORE_ERRORS, $report);
  our ($SRSCAN_FLAGS);		#reference to a hash with various flags for SRSCAN
  our ($messageSRSCAN);         #returned message if $status != 0.
  our ($EXTRA_MOTION_TIME);	#extra allowance for tracking because
  my ($script, $assy, $axis, $begin, $end, $npts, $scanTime, $medm);
  my ($mca, $ROIsRef, $pvlistRef, $srDeadTime, $mcaOverhead);
  my ($nROIs, $npvs, $ndetectors, $actPts, $cnt_time, $cnt_min);
  my ($scanrec, $scanAbort, $status, $name, $i, $j, $IGNORE_ERRORS_local);
  my ($scanRange, $l_speed, $h_speed, $d0, $d1, $extra_motion_time);
  my ($ROI, $roi_lo, $roi_hi, $nROI_set, $scanTimeReal, $cnt_timeReal);
  my ($wait_step, $wait_, $wait_start, $busy, $currPts, $inpos, $runprg, $positioned);

  $script     = 'SRflyScanMCA';
  $assy       = shift(@_);
  $axis       = shift(@_);
  $begin      = shift(@_);	# start scan position
  $end        = shift(@_);	# end scan position
  $npts       = shift(@_);	# number of points on fast scan
  $scanTime   = shift(@_);	# scan time in seconds
  $medm       = shift(@_);	#1=open MEDM for scaler inputs, 2=for inputs+PVs, 0=not
  $mca        = shift(@_);	#Canberra or Falcon
  $ROIsRef    = shift(@_);
  $pvlistRef  = shift(@_);

  if (!defined $IGNORE_ERRORS) {$IGNORE_ERRORS=0;}
  $IGNORE_ERRORS_local = $IGNORE_ERRORS;

  if (!defined $SRSCAN_FLAGS->{'has run'})              {$SRSCAN_FLAGS->{'has run'}              = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter close'})   {$SRSCAN_FLAGS->{'skip shutter close'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter open'})    {$SRSCAN_FLAGS->{'skip shutter open'}    = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip calc speed'})      {$SRSCAN_FLAGS->{'skip calc speed'}      = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip set speed'})       {$SRSCAN_FLAGS->{'skip set speed'}       = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip set slew'})        {$SRSCAN_FLAGS->{'skip set slew'}        = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip restore speed'})   {$SRSCAN_FLAGS->{'skip restore speed'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip restore mprg'})    {$SRSCAN_FLAGS->{'skip restore mprg'}    = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip move to start'})   {$SRSCAN_FLAGS->{'skip move to start'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'return to start'})      {$SRSCAN_FLAGS->{'return to start'}      = 0;}
#
  if (!defined $SRSCAN_FLAGS->{'speed'})                {$SRSCAN_FLAGS->{'speed'}                = 0;}

  $srDeadTime     = 0.005;		#ScanRec needs some time to finish/erase
# $mcaOverhead    = 0.400;		#MCA needs some time to finish/erase [before 2008]
  $mcaOverhead    = 0.100;		#MCA needs some time to finish/erase [2008/09/17]
  $cnt_min        = 0.1;		#minimum counting time
  $wait_start     = 4.0;
  $messageSRSCAN  = '';

  $nROIs = @$ROIsRef;
  if (defined $pvlistRef) {$npvs = @$pvlistRef;} else {$npvs = 0;}
  $ndetectors = $nROIs + $npvs;

  if ($npts > 1) {$cnt_time = $scanTime / ($npts - 1);}
  else           {$cnt_time = $scanTime;}
  $cnt_min = &maxval($cnt_min, 2*$srDeadTime, 2*$mcaOverhead);
  if ($cnt_time < $cnt_min) {
     $messageSRSCAN = sprintf 'Too short time per point=%.3fs, the minimum is %.3fs', $cnt_time,$cnt_min;
     print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
     return 666, 0;
  }
  $cnt_timeReal = $cnt_time + $srDeadTime + $mcaOverhead;
  if ($npts > 1) {$scanTimeReal = $cnt_timeReal * ($npts - 1);}
  else           {$scanTimeReal = $cnt_timeReal;}

  if ($SRSCAN_FLAGS->{'has run'} == 0 && $scanTimeReal > 1.5*$scanTime) {
     printf STDOUT '!!! '.$script.': because of MCA and ScanRec '
                                 .'overheads, the scantime will be '
                                 .'%1.0fs instead of %1.0f !'."\n",
                                  $scanTimeReal, $scanTime;
  }
#-----------------------------------------------------------------
### The globals below are defined in beamline.h:
  $scanrec   = $scanrecName;
  $scanAbort = $scanrecAbort;
  if (defined $shutter) {
     if ($shutter !~ /^${beamline}:/) {$shutter   = $beamline.':'.$shutter;}
  }
  if (defined $shutter_readback) {
     if ($shutter_readback !~ /^${beamline}:/) {
                                $shutter_readback = $beamline.':'.$shutter_readback;}
  }
  if ($scanrec    !~ /^${beamline}:/) {$scanrec   = $beamline.':'.$scanrec;}
  if ($scanAbort  !~ /^${beamline}:/) {$scanAbort = $beamline.':'.$scanAbort;}
#-----------------------------------------------------------------
  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
     print STDOUT 'Closing shutter ['.$shutter.']'."\n";
     &closeShutter ($shutter,$shutter_readback);
  } else {
     if ($SRSCAN_FLAGS->{'has run'} == 0) {print STDOUT '!!! '.$script.': shutter is not used!'."\n";}
  }

  if ($SRSCAN_FLAGS->{'has run'} == 0 || $SRSCAN_FLAGS->{'skip calc speed'} == 0) {
### Calculate scan speed for the motor:
     print STDOUT 'Calculating slew-mode motor speed for ['.$axis.']'."\n";
     $scanRange = abs ($end - $begin);
     $IGNORE_ERRORS = 1;
     ($l_speed,$h_speed) = &calcAxisSpeed ($assy,$axis,$scanRange,$scanTimeReal);
     $IGNORE_ERRORS = $IGNORE_ERRORS_local;
     if ($l_speed > 1.01*$h_speed) {
        $messageSRSCAN = sprintf 'Calculated low speed=%.3f exceeds max drive speed=%.3f',$l_speed,$h_speed;
        print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
        return 777, 0;
     }
     $SRSCAN_FLAGS->{'speed'} = $l_speed;
  } else {
#    printf STDOUT '!!! Using '.$axis.' slew-mode speed=%.3f from previous scan'."\n",$SRSCAN_FLAGS->{'speed'};
#    print  STDOUT '!!! (This presumes same range and time)'."\n";
     $l_speed = $SRSCAN_FLAGS->{'speed'};
  }

### set MCA:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up MCA ['.$mca.']'."\n";
     &setMCA ($mca, $cnt_time);
     $nROI_set = 0;
     for ($i=0; $i<$nROIs; $i++) {
        $ROI = @$ROIsRef[$i];
        ($roi_lo, $roi_hi) = &checkMCA_ROI($mca,$ROI);
        if ($roi_lo == -1 && $roi_hi == -1) {
           print STDOUT '!!! '.$script.': ROI('.$i.')='.$ROI.' is not setup in the MCA'."\n";
        }
        else {
           $nROI_set++;
        }
     }
     if ($nROI_set == 0) {
        $messageSRSCAN = 'None of ROI are set in the MCA';
        print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
        return 222, 0;
     }
#    print STDOUT $script.':  assy='.$assy.'  axis='.$axis.'  begin='.$begin.'  end='.$end.'  npts='.$npts."\n";
#    print STDOUT $script.':  scanTime='.$scanTime.'  scanTimeReal='.$scanTimeReal.'  medm='.$medm."\n";
#    print STDOUT $script.':   mca='.$mca.'  ROi='."@$ROIsRef\n";
  }

### set motor:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up motor ['.$axis.']'."\n";
     &configAssy ($assy);
     &restoreAxis ($assy, $axis);			#restore high-speed
  }

### set Scan Record:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
#    print STDOUT 'Clearing scan record ['.$scanrec.'], ['.$scanAbort.']'."\n";
     &clearSR ($scanrec);

     print STDOUT 'Setting up scan record ['.$scanrec.'] positioner'."\n";
     &setupSRmotors ($scanrec, $npts, 1, [$axis.':ActPos'], [$axis.':RqsPos'], [$begin], [$end]);

     print STDOUT 'Setting up scan record ['.$scanrec.'] detector(s): ';
     &setupSRdetectors_mca ($scanrec, $mca, $ROIsRef, $pvlistRef);
     printf STDOUT '... nDetectors='.$nROIs.'+'.$npvs.'=%d'."\n",$ndetectors;
  } else {
     print STDOUT 'Changing scan record ['.$scanrec.'] drive range'."\n";
     &changeSRmotors ($scanrec, $npts, 1, [$begin], [$end]);
  }

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  if ($SRSCAN_FLAGS->{'skip move to start'} == 0) {
### Move motor to the starting position:
     printf STDOUT 'Moving motor ['.$axis.'] to the starting position (%.3g)'."\n",$begin;
### Increase expected motion time (just in case):
     if (defined $EXTRA_MOTION_TIME) {$extra_motion_time = $EXTRA_MOTION_TIME;}
     $EXTRA_MOTION_TIME = 600;
     $status = &runMotor ($assy, $axis, $begin);
### Restore expected motion time to normal from local cache:
     if (defined $extra_motion_time) {$EXTRA_MOTION_TIME = $extra_motion_time;}
     else                            {undef $EXTRA_MOTION_TIME;}
     if ($status) {
        $messageSRSCAN = 'Failed moving ['.$axis.'] to the starting position.';
        print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
        return $status, 0;
     }
  }

### Setting slow speed for scanning:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
        printf STDOUT 'Setting up motor ['.$axis.'] speed=%.3g & motionPRG=slew'."\n",$l_speed;
        &setAxisSlewScan ($assy, $axis, $l_speed);
  } else {
     if ($SRSCAN_FLAGS->{'skip set slew'} == 0 && $SRSCAN_FLAGS->{'skip set speed'} == 0) {
        printf STDOUT 'Setting up motor ['.$axis.'] speed=%.3g & motionPRG=slew.'."\n",$l_speed;
        &setAxisSlewScan ($assy, $axis, $l_speed);			#setMotionPrg+setAxisSpeed
     }
     elsif ($SRSCAN_FLAGS->{'skip set slew'} && $SRSCAN_FLAGS->{'skip set speed'} == 0) {
        printf STDOUT $script.': setting scanning speed=%.3g only, but presuming motionPRG=slew.'."\n",$l_speed;
        &setAxisSpeed ($assy, $axis, $l_speed); 		#this is faster (no motion prg change)
     }
     elsif ($SRSCAN_FLAGS->{'skip set slew'} == 0 && $SRSCAN_FLAGS->{'skip set speed'}) {
        print STDOUT $script.': setting motionPRG=slew, but presuming scanning speed is set.'."\n";
        &setMotionPrg($assy, 1); 				#this is faster (no speed change): 1=slew
     }
  }

### Setting monitors on motors motion and scan record execution:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     $status = &addCaMonitor($scanrec.'.BUSY');
     $status = &addCaMonitor($scanrec.'.CPT');
     $status = &setAssyMonitors($assy);
  }

  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter open'} == 0) {
     print STDOUT 'Opening shutter ['.$shutter.']'."\n";
     $status = &openShutter ($shutter,$shutter_readback);
     if ($status) {
        &abortAssy ($assy, 1, $axis);		# also restore speed and motion prg.
        &closeShutter($shutter,$shutter_readback);
        $messageSRSCAN = 'Failed opening fast shutter ['.$shutter.'].';
        print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
        return $status;
     }
  }

  $wait_step     = 0.1;
  $wait_         = 0;
  $busy          = 1;
  $currPts       = 0;
  $inpos         = 0;
  $runprg        = 1;
  $positioned    = 0;

### Start scan record:
  print STDOUT 'Starting on-the-fly scan with MCA ['.$mca.'] and scan record ['.$scanrec.']...'."\n";
  &startSR ($scanrec, $scanAbort);
# &press_key_continue();

  if    ($medm == 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_mca($scanrec, $ROIsRef, undef, 'realtime');
  }
  elsif ($medm  > 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_mca($scanrec, $ROIsRef, $pvlistRef, 'realtime');
  }

### Wait to finish (don't give up first $wait_start seconds):
  while ((($busy ==1 || $currPts < 1) && ($runprg || $positioned<3)) || $wait_< $wait_start) {
     select(undef,undef,undef,$wait_step);	# sleep($wait_step);
     $wait_ += $wait_step;
     ($status,$busy)   = Pezca::GetDouble($scanrec.'.BUSY');
     ($status,$currPts)= Pezca::GetDouble($scanrec.'.CPT');
     ($status,$inpos)  = Pezca::GetDouble($assy.':InPos');
     ($status,$runprg) = Pezca::GetDouble($assy.':RunPrg');
     if ($inpos == 1) { $positioned++; }
     else             { $positioned=0; }
#    print STDOUT 'busy='.$busy.' currPts='.$currPts.' runprg='.$runprg.' inpos='
#                .$inpos.' positioned='.$positioned.' wait_='.$wait_."\n";
  }
  printf STDOUT 'Scan finished! (scanBusy='.$busy.', scanCpt='.$currPts
               .', mtrRunPrg='.$runprg.', mtrPos='.$positioned.', time=%.1fs)'."\n",$wait_;
  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
     print STDOUT 'Closing shutter ['.$shutter.']'."\n";
     &closeShutter ($shutter,$shutter_readback);
  }
  &abortSR ($scanAbort, $scanrec);
  &abortAssy ($assy, 1, $axis);		# also restore speed and motion prg.

### Wait the data to be ready:
  print STDOUT 'Waiting for scan record data...'."\n";
  $actPts = &waitSRdata ($scanrec);

### Start reverting motor to the beginning position
### in order to save time on the next scan:
  if ($SRSCAN_FLAGS->{'return to start'}) {&startMotor ($axis, $begin);}

  print STDOUT 'Reading scan record data: npts='.$actPts.' nDetectors='.$ndetectors."\n";

  if    ($medm == 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_mca($scanrec, $ROIsRef, undef, 'afterscan');
  }
  elsif ($medm  > 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_mca($scanrec, $ROIsRef, $pvlistRef, 'afterscan');
  }

  no strict 'refs';
  $name   = 'x1SR';
  @$name = &readSRx ($scanrec, $actPts, 1);
  @$name = &calcSRxMiddle ($actPts, $begin, @$name);

  for ($i=1; $i<=$ndetectors; $i++) {
     $name  = 'y'.$i.'SR';
     @$name = &readSRy ($scanrec, $actPts, $i);
     if ($i <= $nROIs) {
### Normalize mcadata for time:
        for ($j=0; $j<$actPts; $j++) {@$name[$j] = @$name[$j]/$cnt_time;}
     }
     else  {
### Average PV data over measurements intervals:
        $d0 = @$name[0];
        for ($j=1; $j<$actPts; $j++) {$d1=@$name[$j]; @$name[$j]=0.5*($d1+$d0); $d0=$d1;}
     }
  }
  use strict 'refs';
  print STDOUT 'Scan record data retrieved!'."\n";
  $SRSCAN_FLAGS->{'has run'} = 1;
  return $status,$actPts;

}	##SRflyScanMCA

#========================================================================

# ($status,$actPts) =
#   &SRflyScanJoergerMCA ($assy,$axis,$begin,$end,$npts,$scanTime,$medm,$scaler,$fileBkg,$mca,\@inputs,\@ROIs,[\@pvlist]);
#sub SRflyScanJoergerMCA ($assy,$axis,$begin,$end,$npts,$scanTime,$medm,$scaler,$fileBkg,$mca,\@inputs,\@ROIs,[\@pvlist]) {
sub  SRflyScanJoergerMCA ($$$$$$$$$$\@\@;\@) {

### This function implements ON-THE-FLY scan with SCAN RECORD, PMAC motors and (Joerger inputs + MCA ROI + PVlist)
###
### Returns GLOBALS: @x1SR,@y1SR,...@y{$nROIs+$npvs}SR
### Also returns GLOBAL $messageSRSCAN if $status != 0.

### This version works with both Joergey-type scaler and MCA (e.g. Canberra)
### It also accepts optional PVlist to record during scan
###    (this list added to SR detectors after MCA ROIs)

  our ($beamline, $shutter, $shutter_readback, $scanrecName, $scanrecAbort, @Offsets);
  our ($IGNORE_ERRORS, $report);
  our ($SRSCAN_FLAGS);		#reference to a hash with various flags for SRSCAN
  our ($messageSRSCAN);         #returned message if $status != 0.
  our ($EXTRA_MOTION_TIME);	#extra allowance for motion tracking
  my ($script, $assy, $axis, $begin, $end, $npts, $scanTime, $medm);
  my ($scaler, $fileBkg, $mca, $inputsRef, $ROIsRef, $pvlistRef);
  my ($srDeadTime, $mcaOverhead, $IGNORE_ERRORS_local, $extra_motion_time);
  my ($ninputs, $nROIs, $npvs, $ndetectors, $actPts, $cnt_time, $cnt_min);
  my ($scanrec, $scanAbort, $scalerNo, $status, $name, $i, $j);
  my ($scanRange, $l_speed, $h_speed, $offsetNr, $d0, $d1);
  my ($ROI, $roi_lo, $roi_hi, $nROI_set, $scanTimeReal, $cnt_timeReal);
  my ($wait_step, $wait_, $wait_start, $busy, $currPts, $inpos, $runprg, $positioned);

  $script    = 'SRflyScanJoergerMCA';
  $assy      = shift(@_);
  $axis      = shift(@_);
  $begin     = shift(@_);	#start scan position
  $end       = shift(@_);	#end scan position
  $npts      = shift(@_);	#number of points on fast scan
  $scanTime  = shift(@_);	#scan time in seconds
  $medm      = shift(@_);	#1=open MEDM for scaler inputs, 2=for inputs+PVs, 0=not
  $scaler    = shift(@_);	#scaler1=Struck, scaler3=Joerger
  $fileBkg   = shift(@_);	#joerger background file ('auto' for automatic, '' for none)
  $mca       = shift(@_);	#Canberra or Falcon
  $inputsRef = shift(@_);
  $ROIsRef   = shift(@_);
  $pvlistRef = shift(@_);

  if (!defined $IGNORE_ERRORS) {$IGNORE_ERRORS=0;}
  $IGNORE_ERRORS_local = $IGNORE_ERRORS;

  if (!defined $SRSCAN_FLAGS->{'has run'})              {$SRSCAN_FLAGS->{'has run'}              = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter close'})   {$SRSCAN_FLAGS->{'skip shutter close'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter open'})    {$SRSCAN_FLAGS->{'skip shutter open'}    = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip calc speed'})      {$SRSCAN_FLAGS->{'skip calc speed'}      = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip set speed'})       {$SRSCAN_FLAGS->{'skip set speed'}       = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip set slew'})        {$SRSCAN_FLAGS->{'skip set slew'}        = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip restore speed'})   {$SRSCAN_FLAGS->{'skip restore speed'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip restore mprg'})    {$SRSCAN_FLAGS->{'skip restore mprg'}    = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip move to start'})   {$SRSCAN_FLAGS->{'skip move to start'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'return to start'})      {$SRSCAN_FLAGS->{'return to start'}      = 0;}
#
  if (!defined $SRSCAN_FLAGS->{'speed'})                {$SRSCAN_FLAGS->{'speed'}                = 0;}

  $srDeadTime     = 0.005;		#ScanRec needs some time to finish/erase
# $mcaOverhead    = 0.400;		#MCA needs some time to finish/erase [before 2008]
  $mcaOverhead    = 0.100;		#MCA needs some time to finish/erase [2008/09/17]
  $cnt_min        = 0.1;		#minimum counting time
  $wait_start     = 4.0;
  $messageSRSCAN  = '';

  if (!defined $fileBkg) {$fileBkg='';}
  if (lc($fileBkg) eq 'auto') {
     $scalerNo = $scaler;
     $scalerNo =~ s/^.*\D+//;	# strip out all non-numeric (\D) stuff at the beginning
     $fileBkg  = 'joerger'.$scalerNo.'_Bkg.dat';
  }
  $ninputs = @$inputsRef;

  $nROIs = @$ROIsRef;
  if (defined $pvlistRef) {$npvs = @$pvlistRef;} else {$npvs = 0;}
  $ndetectors = $ninputs + $nROIs + $npvs;

  if ($npts > 1) {$cnt_time = $scanTime / ($npts - 1);}
  else           {$cnt_time = $scanTime;}
  $cnt_min = &maxval($cnt_min, 2*$srDeadTime, 2*$mcaOverhead);
  if ($cnt_time < $cnt_min) {
     $messageSRSCAN = sprintf 'Too short time per point=%.3fs, the minimum is %.3fs', $cnt_time,$cnt_min;
     print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
     return 666, 0;
  }
  $cnt_timeReal = $cnt_time + $srDeadTime + $mcaOverhead;
  if ($npts > 1) {$scanTimeReal = $cnt_timeReal * ($npts - 1);}
  else           {$scanTimeReal = $cnt_timeReal;}
  if ($SRSCAN_FLAGS->{'has run'} == 0 && $scanTimeReal > 1.5*$scanTime) {
     printf STDOUT '!!! '.$script.': because of MCA, Scaler and ScanRec '
                                 .'overheads, the scantime will be '
                                 .'%1.0fs instead of %1.0f !'."\n",
                                 $scanTimeReal,$scanTime;
  }
#-----------------------------------------------------------------
### The globals below are defined in beamline.h:
  $scanrec   = $scanrecName;
  $scanAbort = $scanrecAbort;
  if (defined $shutter) {
     if ($shutter !~ /^${beamline}:/) {$shutter   = $beamline.':'.$shutter;}
  }
  if (defined $shutter_readback) {
     if ($shutter_readback !~ /^${beamline}:/) {
                                $shutter_readback = $beamline.':'.$shutter_readback;}
  }
  if ($scanrec    !~ /^${beamline}:/) {$scanrec   = $beamline.':'.$scanrec;}
  if ($scanAbort  !~ /^${beamline}:/) {$scanAbort = $beamline.':'.$scanAbort;}
#-----------------------------------------------------------------
  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
     print STDOUT 'Closing shutter ['.$shutter.']'."\n";
     &closeShutter ($shutter,$shutter_readback);
  } else {
     if ($SRSCAN_FLAGS->{'has run'} == 0) {print STDOUT '!!! '.$script.': shutter is not used!'."\n";}
  }

  if ($SRSCAN_FLAGS->{'has run'} == 0 || $SRSCAN_FLAGS->{'skip calc speed'} == 0) {
### Calculate scan speed for the motor:
     print STDOUT 'Calculating slew-mode motor speed for ['.$axis.']'."\n";
     $scanRange = abs ($end - $begin);
     $IGNORE_ERRORS = 1;
     ($l_speed,$h_speed) = &calcAxisSpeed ($assy,$axis,$scanRange,$scanTimeReal);
     $IGNORE_ERRORS = $IGNORE_ERRORS_local;
     if ($l_speed > 1.01*$h_speed) {
        $messageSRSCAN = sprintf 'Calculated low speed=%.3f exceeds max drive speed=%.3f',$l_speed,$h_speed;
        print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
        return 777, 0;
     }
     $SRSCAN_FLAGS->{'speed'} = $l_speed;
  } else {
#    printf STDOUT '!!! Using '.$axis.' slew-mode speed=%.3f from previous scan'."\n",$SRSCAN_FLAGS->{'speed'};
#    print  STDOUT '!!! (This presumes same range and time)'."\n";
     $l_speed = $SRSCAN_FLAGS->{'speed'};
  }

### Read the scaler background offsets:
  if (!defined $Offsets[0]) {			# skip reading offsets if they were read in main
     if (length($fileBkg) > 0) {print STDOUT 'Reading scaler ['.$scaler.'] offsets from '.$fileBkg."\n";}
     else                      {print STDOUT 'No scaler ['.$scaler.'] offsets file specified. Setting zero offsets.'."\n";}
     @Offsets = &getJoergerOffsets ($fileBkg,$scaler);  # we make it GLOBAL to access from main
  }

### set scaler:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up scaler ['.$scaler.']'."\n";
     &setJoerger($scaler,$cnt_time);
  }

### set MCA:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up MCA ['.$mca.']'."\n";
     &setMCA ($mca, $cnt_time);
     $nROI_set = 0;
     for ($i=0; $i<$nROIs; $i++) {
        $ROI = @$ROIsRef[$i];
        ($roi_lo, $roi_hi) = &checkMCA_ROI($mca,$ROI);
        if ($roi_lo == -1 && $roi_hi == -1) {
           print STDOUT '!!! '.$script.': ROI('.$i.')='.$ROI.' is not setup in the MCA'."\n";
        }
        else {
           $nROI_set++;
        }
     }
     if ($nROI_set == 0) {
        $messageSRSCAN = 'None of ROI are set in the MCA';
        print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
        return 222, 0;
     }
#    print STDOUT $script.':  assy='.$assy.'  axis='.$axis.'  begin='.$begin.'  end='.$end.'  npts='.$npts."\n";
#    print STDOUT $script.':  scanTime='.$scanTime.'  scanTimeReal='.$scanTimeReal.'  medm='.$medm."\n";
#    print STDOUT $script.':  mca='.$mca.'  ROi='."@$ROIsRef\n";
  }

### set motor:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up motor ['.$axis.']'."\n";
     &configAssy ($assy);
     &restoreAxis ($assy, $axis);			#restore high-speed
  }

### set Scan Record:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
#    print STDOUT 'Clearing scan record ['.$scanrec.'], ['.$scanAbort.']'."\n";
     &clearSR ($scanrec);

     print STDOUT 'Setting up scan record ['.$scanrec.'] positioner'."\n";
     &setupSRmotors ($scanrec, $npts, 1, [$axis.':ActPos'], [$axis.':RqsPos'], [$begin], [$end]);

     print STDOUT 'Setting up scan record ['.$scanrec.'] detector(s): ';
     &setupSRdetectors_scaler_mca_pv ($scanrec, $scaler, $mca, $inputsRef, $ROIsRef, $pvlistRef);
     printf STDOUT '... nDetectors='.$ninputs.'+'.$nROIs.'+'.$npvs.'=%d'."\n",$ndetectors;
  } else {
     print STDOUT 'Changing scan record ['.$scanrec.'] drive range'."\n";
     &changeSRmotors ($scanrec, $npts, 1, [$begin], [$end]);
  }

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  if ($SRSCAN_FLAGS->{'skip move to start'} == 0) {
### Move motor to the starting position:
     printf STDOUT 'Moving motor ['.$axis.'] to the starting position (%.3g)'."\n",$begin;
### Increase expected motion time (just in case):
     if (defined $EXTRA_MOTION_TIME) {$extra_motion_time = $EXTRA_MOTION_TIME;}
     $EXTRA_MOTION_TIME = 600;
     $status = &runMotor ($assy, $axis, $begin);
### Restore expected motion time to normal from local cache:
     if (defined $extra_motion_time) {$EXTRA_MOTION_TIME = $extra_motion_time;}
     else                            {undef $EXTRA_MOTION_TIME;}
     if ($status) {
        $messageSRSCAN = 'Failed moving ['.$axis.'] to the starting position.';
        print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
        return $status, 0;
     }
  }

### Setting slow speed for scanning:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
        printf STDOUT 'Setting up motor ['.$axis.'] speed=%.3g & motionPRG=slew'."\n",$l_speed;
        &setAxisSlewScan ($assy, $axis, $l_speed);
  } else {
     if ($SRSCAN_FLAGS->{'skip set slew'} == 0 && $SRSCAN_FLAGS->{'skip set speed'} == 0) {
        printf STDOUT 'Setting up motor ['.$axis.'] speed=%.3g & motionPRG=slew.'."\n",$l_speed;
        &setAxisSlewScan ($assy, $axis, $l_speed);			#setMotionPrg+setAxisSpeed
     }
     elsif ($SRSCAN_FLAGS->{'skip set slew'} && $SRSCAN_FLAGS->{'skip set speed'} == 0) {
        printf STDOUT $script.': setting scanning speed=%.3g only, but presuming motionPRG=slew.'."\n",$l_speed;
        &setAxisSpeed ($assy, $axis, $l_speed); 		#this is faster (no motion prg change)
     }
     elsif ($SRSCAN_FLAGS->{'skip set slew'} == 0 && $SRSCAN_FLAGS->{'skip set speed'}) {
        print STDOUT $script.': setting motionPRG=slew, but presuming scanning speed is set.'."\n";
        &setMotionPrg($assy, 1); 				#this is faster (no speed change): 1=slew
     }
  }

### Setting monitors on motors motion and scan record execution:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     $status = &addCaMonitor($scanrec.'.BUSY');
     $status = &addCaMonitor($scanrec.'.CPT');
     $status = &setAssyMonitors($assy);
  }

  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter open'} == 0) {
     print STDOUT 'Opening shutter ['.$shutter.']'."\n";
     $status = &openShutter ($shutter,$shutter_readback);
     if ($status) {
        &abortAssy ($assy, 1, $axis);		# also restore speed and motion prg.
        &closeShutter($shutter,$shutter_readback);
        $messageSRSCAN = 'Failed opening fast shutter ['.$shutter.'].';
        print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
        return $status;
     }
  }

  $wait_step  = 0.1;
  $wait_      = 0;
  $busy       = 1;
  $currPts    = 0;
  $inpos      = 0;
  $runprg     = 1;
  $positioned = 0;

### Start scan record:
  print STDOUT 'Starting on-the-fly scan with MCA ['.$mca.'] and scan record ['.$scanrec.']...'."\n";
  &startSR ($scanrec, $scanAbort);
# &press_key_continue();

  if    ($medm == 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_scaler_mca($scanrec, $inputsRef, $ROIsRef, undef, 'realtime');
  }
  elsif ($medm  > 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_scaler_mca($scanrec, $inputsRef, $ROIsRef, $pvlistRef, 'realtime');
  }

### Wait to finish (don't give up first $wait_start seconds):
  while ((($busy ==1 || $currPts < 1) && ($runprg || $positioned<3)) || $wait_< $wait_start) {
     select(undef,undef,undef,$wait_step);	# sleep($wait_step);
     $wait_ += $wait_step;
     ($status,$busy)   = Pezca::GetDouble($scanrec.'.BUSY');
     ($status,$currPts)= Pezca::GetDouble($scanrec.'.CPT');
     ($status,$inpos)  = Pezca::GetDouble($assy.':InPos');
     ($status,$runprg) = Pezca::GetDouble($assy.':RunPrg');
     if ($inpos == 1) { $positioned++; }
     else             { $positioned=0; }
#    print STDOUT 'busy='.$busy.' currPts='.$currPts.' runprg='.$runprg.' inpos='
#                 .$inpos.' positioned='.$positioned.' wait_='.$wait_."\n";
  }
  printf STDOUT 'Scan finished! (scanBusy='.$busy.', scanCpt='.$currPts
               .', mtrRunPrg='.$runprg.', mtrPos='.$positioned.', time=%.1fs)'."\n",$wait_;
  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
     print STDOUT 'Closing shutter ['.$shutter.']'."\n";
     &closeShutter ($shutter,$shutter_readback);
  }
  &abortSR ($scanAbort, $scanrec);
  &abortAssy ($assy, 1, $axis);		# also restore speed and motion prg.

### Wait the data to be ready:
  print STDOUT 'Waiting for scan record data...'."\n";
  $actPts = &waitSRdata ($scanrec);

### Start reverting motor to the beginning position
### in order to save time on the next scan:
  if ($SRSCAN_FLAGS->{'return to start'}) {&startMotor ($axis, $begin);}

  print STDOUT 'Reading scan record data: npts='.$actPts.' nDetectors='.$ndetectors."\n";

  if    ($medm == 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_scaler_mca($scanrec, $inputsRef, $ROIsRef, undef, 'afterscan');
  }
  elsif ($medm  > 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_scaler_mca($scanrec, $inputsRef, $ROIsRef, $pvlistRef, 'afterscan');
  }

  no strict 'refs';
  $name   = 'x1SR';
  @$name = &readSRx ($scanrec, $actPts, 1);
  @$name = &calcSRxMiddle ($actPts, $begin, @$name);

  for ($i=1; $i<=$ndetectors; $i++) {
     $name  = 'y'.$i.'SR';
     @$name = &readSRy ($scanrec, $actPts, $i);
     if ($i <= $ninputs) {
### Normalize scaler data for time and correct for background:
        $offsetNr = @$inputsRef[$i-1];
        for ($j=0; $j<$actPts; $j++) {@$name[$j] = @$name[$j]/$cnt_time - $Offsets[$offsetNr-1];}
     }
     elsif ($i > $ninputs && $i <= $nROIs) {
### Normalize mcadata for time:
        for ($j=0; $j<$actPts; $j++) {@$name[$j] = @$name[$j]/$cnt_time;}
     }
     else  {
### Average PV data over measurements intervals:
        $d0 = @$name[0];
        for ($j=1; $j<$actPts; $j++) {$d1=@$name[$j]; @$name[$j]=0.5*($d1+$d0); $d0=$d1;}
     }
  }
  use strict 'refs';
  print STDOUT 'Scan record data retrieved!'."\n";
  $SRSCAN_FLAGS->{'has run'} = 1;
  return $status,$actPts;

}	##SRflyScanJoergerMCA

#========================================================================

# ($status,$actPts) =
#   &SRflyScanOMS58 ($motor,$begin,$end,$npts,$scanTime,$medm,$scaler,$fileBkg,$ninputs,@inputs);
#sub SRflyScanOMS58 ($motor,$begin,$end,$npts,$scanTime,$medm,$scaler,$fileBkg,$ninputs,@inputs) {
sub  SRflyScanOMS58 ($$$$$$$$$@) {

### This function implements ON-THE-FLY scan with SCAN RECORD, OMS58 (MotorRec) motors and Joerger scaler
###
### Returns GLOBALS: @x1SR,@y1SR,...@y{$ninputs}SR, @Offsets
### Also returns GLOBAL $messageSRSCAN if $status != 0.

  our ($beamline, $shutter, $shutter_readback, $scanrecName, $scanrecAbort);
  our (@Offsets, $STRUCK_MODEL);
  our ($IGNORE_ERRORS, $report);
  our ($SRSCAN_FLAGS);		#reference to a hash with various flags for SRSCAN
  our ($messageSRSCAN);         #returned message if $status != 0.
  my ($script, $motor, $begin, $end, $npts, $scanTime, $medm, $scaler, $fileBkg);
  my ($ninputs, @inputs, $scalerNo, $i, $j, $srDeadTime, $struckOverhead);
  my ($scanTimeReal, $cnt_timeReal, $status, $actPts, $name, $cnt_time, $cnt_min);
  my ($scanrec, $scanAbort, $scanRange, $speedRate, $h_speed, $l_speed);
  my ($wait_step, $wait_, $wait_start, $busy, $currPts, $arrived);
  my ($positioned, $offsetNr);

  $script   = 'SRflyScanOMS58';
  $motor    = shift(@_);
  $begin    = shift(@_);	# start scan position
  $end      = shift(@_);	# end scan position
  $npts     = shift(@_);	# number of points on fast scan
  $scanTime = shift(@_);	# scan time in seconds
  $medm     = shift(@_);	# 1=open MEDM, 0=not
  $scaler   = shift(@_);	# scaler1=Struck, scaler3=Joerger
  $fileBkg  = shift(@_);	# joerger background file ('auto' for automatic, '' for none)
  $ninputs  = shift(@_);
  for ($i=0; $i<$ninputs; $i++) {$inputs[$i]=shift(@_);}

  if (!defined $IGNORE_ERRORS) {$IGNORE_ERRORS=0;}

  if (!defined $SRSCAN_FLAGS->{'has run'})              {$SRSCAN_FLAGS->{'has run'}              = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter close'})   {$SRSCAN_FLAGS->{'skip shutter close'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter open'})    {$SRSCAN_FLAGS->{'skip shutter open'}    = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip calc speed'})      {$SRSCAN_FLAGS->{'skip calc speed'}      = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip set speed'})       {$SRSCAN_FLAGS->{'skip set speed'}       = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip restore speed'})   {$SRSCAN_FLAGS->{'skip restore speed'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip move to start'})   {$SRSCAN_FLAGS->{'skip move to start'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'return to start'})      {$SRSCAN_FLAGS->{'return to start'}      = 0;}
#
  if (!defined $SRSCAN_FLAGS->{'speed'})                {$SRSCAN_FLAGS->{'speed'}                = 0;}

  $srDeadTime     = 0.005;	#ScanRec needs some time to finish/erase
  $struckOverhead = 0.008;	#Struck needs some time to finish/erase
  $cnt_min        = 0.1;        #minimum counting time
  $wait_start     = 4.0;
  $status         = 0;
  $actPts         = 0;
  $cnt_time       = 0;
  $messageSRSCAN  = '';
  if (!defined $STRUCK_MODEL) {$STRUCK_MODEL='SIS3801';}
  if ($STRUCK_MODEL eq 'SIS3820') {$struckOverhead=0;}

  if (!defined $fileBkg) {$fileBkg='';}
  if (lc($fileBkg) eq 'auto') {
     $scalerNo = $scaler;
     $scalerNo =~ s/^.*\D+//;	# strip out all non-numeric (\D) stuff at the beginning
     $fileBkg  = 'joerger'.$scalerNo.'_Bkg.dat';
  }

  if ($npts > 1) {$cnt_time = $scanTime / ($npts - 1);}
  else           {$cnt_time = $scanTime;              }
  if ($cnt_time < $cnt_min) {
     $messageSRSCAN = sprintf 'Too short time per point=%.3fs, the minimum is %.3fs', $cnt_time,$cnt_min;
     print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
     return 666, 0;
  }
  $cnt_timeReal = $cnt_time + $srDeadTime;
  if ($scaler eq $beamline.':scaler1') {
     $cnt_timeReal = $cnt_timeReal + $struckOverhead;	# Struck needs some extra time to finish
  }
  if ($npts > 1) {$scanTimeReal = $cnt_timeReal * ($npts - 1);}
  else           {$scanTimeReal = $cnt_timeReal;}
  if ($SRSCAN_FLAGS->{'has run'} == 0 && $scanTimeReal > 1.5*$scanTime) {
     printf STDOUT '!!! '.$script.': because of Scaler and ScanRec '
                                 .'overheads, the scantime will be '
                                 .'%1.0fs instead of %1.0f !'."\n",
                                   $scanTimeReal,$scanTime;
  }
#-----------------------------------------------------------------
### The globals below are defined in beamline.h:
  $scanrec   = $scanrecName;
  $scanAbort = $scanrecAbort;
  if (defined $shutter) {
     if ($shutter !~ /^${beamline}:/) {$shutter   = $beamline.':'.$shutter;}
  }
  if (defined $shutter_readback) {
     if ($shutter_readback !~ /^${beamline}:/) {
                                $shutter_readback = $beamline.':'.$shutter_readback;}
  }
  if ($scanrec   !~ /^${beamline}:/)  {$scanrec   = $beamline.':'.$scanrec;}
  if ($scanAbort !~ /^${beamline}:/)  {$scanAbort = $beamline.':'.$scanAbort;}
#-----------------------------------------------------------------
  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
     print STDOUT 'Closing shutter ['.$shutter.']'."\n";
     &closeShutter ($shutter,$shutter_readback);
  } else {
     if ($SRSCAN_FLAGS->{'has run'} == 0) {print STDOUT '!!! '.$script.': shutter is not used!'."\n";}
  }

  if ($SRSCAN_FLAGS->{'has run'} == 0 || $SRSCAN_FLAGS->{'skip calc speed'} == 0) {
### Calculate scan speed for the motor:
     print STDOUT 'Calculating slew-mode motor speed for ['.$motor.']'."\n";
     $scanRange = abs ($end - $begin);
     $speedRate = 1;
     ($h_speed, $l_speed) = &calcStepMotorSpeed($motor, $scanRange, $scanTime, $speedRate);
     if (defined $l_speed) {
        printf STDOUT '  ... low_speed=%.3g'."\n",$l_speed;
        $SRSCAN_FLAGS->{'speed'} = $l_speed;
     } else {
        $messageSRSCAN = 'Low_speed calculation failed';
        print STDOUT '  ... '.$messageSRSCAN."\n";
        return 777, 0;
      }
  } else {
#    printf STDOUT '!!! Using '.$axis.' slew-mode speed=%.3f from previous scan'."\n",$SRSCAN_FLAGS->{'speed'};
#    print  STDOUT '!!! (This presumes same range and time)'."\n";
     $l_speed = $SRSCAN_FLAGS->{'speed'};
  }

### Read the scaler background offsets:
  if (!defined $Offsets[0]) {	# skip reading offsets if they were read in main
     if (length($fileBkg) > 0) {print STDOUT 'Reading scaler ['.$scaler.'] offsets from '.$fileBkg."\n";}
     else                      {print STDOUT 'No scaler ['.$scaler.'] offsets file specified. Setting zero offsets.'."\n";}
     @Offsets = &getJoergerOffsets ($fileBkg, $scaler);  # we make it GLOBAL to access from main
  }

### set scaler:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up scaler ['.$scaler.']'."\n";
     &setJoerger($scaler,$cnt_time);
  }

### Set motor:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up motor ['.$motor.']'."\n";
     &setStepMotor ($motor);
     &restoreStepMotor ($motor);			#restore high-speed
  }

### Set Scan Record:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
#    print STDOUT 'Clearing scan record ['.$scanrec.'], ['.$scanAbort.']'."\n";
     &clearSR ($scanrec);

     print STDOUT 'Setting up scan record ['.$scanrec.'] positioner'."\n";
     &setupSRstepmotors ($scanrec, $npts, 1, [$motor], [$begin], [$end]);

     print STDOUT 'Setting up scan record ['.$scanrec.'] detector(s): ';
     &setupSRdetectors_v2 ($scanrec, $scaler, $ninputs, \@inputs);
     print STDOUT '... nDetectors='.$ninputs."\n";
  } else {
     print STDOUT 'Changing scan record ['.$scanrec.'] drive range'."\n";
     &changeSRmotors ($scanrec, $npts, 1, [$begin], [$end]);
  }

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Move motor to the starting position:
  if ($SRSCAN_FLAGS->{'skip move to start'} == 0) {
     printf STDOUT 'Moving motor ['.$motor.'] to the starting position (%.3g)'."\n",$begin;
     $status = &runStepMotor ($motor, $begin);
     if ($status) {
        $messageSRSCAN = sprintf 'Failed moving ['.$motor.'] to the starting position=%.3g',$begin;
        print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
        return $status, 0;
     }
  }

### Setting slow speed for scanning:
  if ($SRSCAN_FLAGS->{'has run'} == 0 || $SRSCAN_FLAGS->{'skip set speed'} == 0) {
     printf STDOUT 'Setting up motor ['.$motor.'] speed=%.3g'."\n",$l_speed;
     &setStepMotorSlew ($motor, $l_speed);
  }

### Setting monitors on motors motion and scan record execution:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     $status = &addCaMonitor($scanrec.'.BUSY');
     $status = &addCaMonitor($scanrec.'.CPT');
     $status = &addCaMonitor($motor.'.DMOV');
  }

  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter open'} == 0) {
     print STDOUT 'Opening shutter ['.$shutter.']'."\n";
     &openShutter ($shutter,$shutter_readback);
     if ($status) {
        &abortStepMotor ($motor, 1);			# also restore speed
        &closeShutter($shutter,$shutter_readback);
        $messageSRSCAN = 'Failed opening fast shutter ['.$shutter.'].';
        print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
        return $status;
     }
  }

  $wait_step     = 0.1;
  $wait_         = 0;
  $busy          = 1;
  $currPts       = 0;
  $arrived       = 0;
  $positioned    = 0;

### Start scan record:
  print STDOUT 'Starting on-the-fly scan with scan record ['.$scanrec.']...'."\n";
  &startSR ($scanrec, $scanAbort);
# &press_key_continue();

  if ($medm && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm($scanrec,$ninputs,\@inputs,'realtime');
  }

### Wait to finish (don't give up first $wait_start seconds):
  while ((($busy ==1 || $currPts < 1) && $positioned<3) || $wait_< $wait_start) {
     select(undef,undef,undef,$wait_step);	# sleep($wait_step);
     $wait_ += $wait_step;
     ($status,$busy)    = Pezca::GetDouble($scanrec.'.BUSY');
     ($status,$currPts) = Pezca::GetDouble($scanrec.'.CPT');
     ($status,$arrived) = Pezca::GetDouble($motor.'.DMOV');
     if ($arrived == 1) {$positioned++;}
     else               {$positioned=0;}
#    print STDOUT 'busy='.$busy.' currPts='.$currPts.' arrived='.$arrived
#                .' positioned='.$positioned.' wait_='.$wait_."\n";
  }
  printf STDOUT 'Scan finished! (scanBusy='.$busy.', scanCpt='.$currPts.', mtrArrived='
               .$arrived.', mtrPos='.$positioned.', time=%.1fs)'."\n",$wait_;
  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
     print STDOUT 'Closing shutter ['.$shutter.']'."\n";
     &closeShutter ($shutter,$shutter_readback);
  }
  &abortSR ($scanAbort, $scanrec);
  &abortStepMotor ($motor, 1);			# also restore speed

### Wait the data to be ready:
  print STDOUT 'Waiting for scan record data...'."\n";
  $actPts = &waitSRdata ($scanrec);

### Start reverting motor to the beginning position
### in order to save time on the next scan:
  if ($SRSCAN_FLAGS->{'return to start'}) {&startStepMotor($motor, $begin);}

  print STDOUT 'Reading scan record data: npts='.$actPts.' nDetectors='.$ninputs."\n";

  if ($medm && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm($scanrec,$ninputs,\@inputs,'afterscan');
  }

  no strict 'refs';
  $name   = 'x1SR';
  @$name = &readSRx ($scanrec, $actPts, 1);
  @$name = &calcSRxMiddle ($actPts, $begin, @$name);

  for ($i=1; $i<=$ninputs; $i++) {
     $name   = 'y'.$i.'SR';
     @$name = &readSRy ($scanrec, $actPts, $i);
### Normalize data for time and correct for background:
     $offsetNr = $inputs[$i-1];
     for ($j=0; $j<$actPts; $j++) {
        $$name[$j] = $$name[$j]/$cnt_time - $Offsets[$offsetNr-1];
     }
  }
  use strict 'refs';
  print STDOUT 'Scan record data retrieved!'."\n";
  $SRSCAN_FLAGS->{'has run'} = 1;
  return $status,$actPts;

}	##SRflyScanOMS58

#========================================================================

# ($status,$dac_slope) =
#   &SRflyScanDAC ($dacPV,$begin,$end,$npts,$scanTime,$afterScan,$medm,
#                  $scaler,$fileBkg,\@inputs,[\@pvlist,$checkOverload,
#                  $postarrayPV,$post_interval,
#                  $adc_encdrPV, $dac_tolerance]);
#sub SRflyScanDAC ($dacPV,$begin,$end,$npts,$scanTime,$afterScan,$medm,
#                  $scaler,$fileBkg,\@inputs,[\@pvlist,$checkOverload,
#                  $postarrayPV,$post_interval,
#                  $adc_encdrPV, $dac_tolerance]) {
sub  SRflyScanDAC ($$$$$$$$$\@;\@$$$$$$) {

### This function implements ON-THE-FLY SCAN with SCAN RECORD, DAC and (Joerger inputs + PVlist)
###
### Returns GLOBALS: @x1SR,@y1SR,...@y{$ninputs+$npvs}SR, @Offsets
### Also returns GLOBAL $messageSRSCAN if $status != 0.

  our ($beamline, $shutter, $shutter_readback, $scanrecName, $scanrecAbort);
  our (@Offsets, $STRUCK_MODEL);
  our (@x1SR, @y1SR);			#preserve $dacLo,$dacHi between scans
  our ($IGNORE_ERRORS, $report);
  our ($SRSCAN_FLAGS);			#reference to a hash with various flags for SRSCAN
  our ($messageSRSCAN);			#returned message if $status != 0.
  my ($script, $dacPV, $dacPV_common, $begin, $end, $npts, $npts_last, $scanTime);
  my ($fileBkg, $inputsRef, $pvlistRef, $settleTime, $scanMode, $afterScan);
  my ($ninputs, $npvs, $ndetectors, $actPts, $cnt_time, $cnt_min, $threshold);
  my ($scanrec, $scanAbort, $medm, $scaler, $scalerNo, $name, $name_plot);
  my ($offsetNr, $d0, $d1, $scanTimeReal, $cnt_timeReal, $dac_slope, $dac_scanning);
  my ($wait_step, $wait_start, $wait_, $busy, $currPts, $checkOverload);
  my ($srStartTime, $srDeadTime, $struckOverhead, $afterScanSR, $dac_ini);
  my ($msg, $postarrayPV, $post_interval, $scan_index, $min_val, $nelm);
  my ($i_plot, $time_epoch, $time_passed, @post_array, $narr, $status, $i, $j);
  my ($OVERLOAD, $overload_detected, $dacLo, $dacHi, $adc_offset, $adc_ini);
  my ($FOLLERR, $follerr_detected, $adc_encdrPV, $dac_tolerance, $tolerance_min, @encoder);
  
  $script        = 'SRflyScanDAC';
  $dacPV         = shift(@_);
  $begin         = shift(@_);   #start scan position
  $end           = shift(@_);   #end scan position
  $npts          = shift(@_);	#number of points on fast scan
  $scanTime      = shift(@_);	#scaler counting time per point in seconds
  $afterScan     = shift(@_);	#0=stay, 1=startpos, 2=priorpos 
                                #3=maxpos 4,14=peakpos(noisy), 5,15=peakpos(smooth)
                                #6=minpos 7,17=valleypos(noisy), 8,18=valleypos(smooth)
                                #10,110=gaussian 11,111=lorentzian
  $medm          = shift(@_);	#1=open MEDM for scaler inputs, 2=for inputs+PVs, 0=not
  $scaler        = shift(@_);	#scaler1=Struck, scaler3=Joerger
  $fileBkg       = shift(@_);	#joerger background file ('auto' for automatic, '' for none)
  $inputsRef     = shift(@_);
  $pvlistRef     = shift(@_);
  $checkOverload = shift(@_);
  $postarrayPV   = shift(@_);   #array for posting partial & final data to JBluIce
  $post_interval = shift(@_);
  $adc_encdrPV   = shift(@_);   #ADC PV to test DAC output (optional)
  $dac_tolerance = shift(@_);	#tolerance when testing DAC output (optional; default=step/2)

  if (!defined $IGNORE_ERRORS) {$IGNORE_ERRORS=0;}

  if (!defined $SRSCAN_FLAGS->{'has run'})              {$SRSCAN_FLAGS->{'has run'}              = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter close'})   {$SRSCAN_FLAGS->{'skip shutter close'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter open'})    {$SRSCAN_FLAGS->{'skip shutter open'}    = 0;}
#
  if (!defined $SRSCAN_FLAGS->{'dacLo'})                {$SRSCAN_FLAGS->{'dacLo'}                = 0;}
  if (!defined $SRSCAN_FLAGS->{'dacHi'})                {$SRSCAN_FLAGS->{'dacHi'}                = 0;}

  $scanMode          = 2;	#0=linear, 1=table 2=ON-THE-FLY
  $settleTime        = 0.0;	#positioner settle time (delay)
  $srStartTime       = 0.1;	#ScanRec needs some time to start talking with scaler  2008/10/03
  $srDeadTime        = 0.005;	#ScanRec needs some time to finish/erase
  $struckOverhead    = 0.008;	#Struck needs some time to finish/erase
  $cnt_min           = 0.1; 	#minimum counting time
  $wait_start        = 4.0;
# $threshold         = 0.95E+5;	#threshold intensity for detecting overload (before 2013)
  $threshold         = 0.95E+6;	#threshold intensity for detecting overload (2013)
  $OVERLOAD          = 100000;	#scan failure flag when overload detected
  $overload_detected = 0;	#overload not detected
  $FOLLERR           = 200000;	#scan failure flag when following error detected
  $tolerance_min     = 0.2;	#the minimum sensible DAC error because of different DAC & ADC calibrations
  $scan_index        = 0;	#scan index to save with the post (e.g. 11-vertical, 21-horizontal)
  $follerr_detected  = 0;
  $messageSRSCAN     = '';

  if (!defined $STRUCK_MODEL) {$STRUCK_MODEL='SIS3801';}
  if ($STRUCK_MODEL eq 'SIS3820') {$struckOverhead=0;}
  if (!defined $post_interval) {$post_interval = 2;}		#post interval
  if (!defined $checkOverload) {$checkOverload = 0;}
  if (!defined $fileBkg)       {$fileBkg='';}
  if (!defined $adc_encdrPV)   {$adc_encdrPV   = $dacPV;}	#ADC PV to test DAC output (optional)
  if (!defined $dac_tolerance) {$dac_tolerance = 0.;}		#ADC channel to test DAC output (optional)
  $dac_tolerance = abs($dac_tolerance);
  if ($dac_tolerance == 0.)    {$dac_tolerance = abs(0.5*($end-$begin)/$npts);}
  if ($dac_tolerance < $tolerance_min) {$dac_tolerance = $tolerance_min;}

  $ninputs = @$inputsRef;
  if (defined $pvlistRef) {$npvs = @$pvlistRef;} else {$npvs = 0;}
  $ndetectors = $ninputs + $npvs;

  if ($npts > 1) {$cnt_time = $scanTime / ($npts - 1);}
  else           {$cnt_time = $scanTime;}
  if ($cnt_time < $cnt_min) {
     $messageSRSCAN = sprintf 'Too short time per point=%.3fs, the minimum is %.3fs', $cnt_time,$cnt_min;
     print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
     return 666;
  }
  $cnt_timeReal = $cnt_time + $srDeadTime;
  if ($scaler eq $beamline.':scaler1') {
     $cnt_timeReal = $cnt_timeReal + $struckOverhead;	#Struck needs some extra time to finish
  }
  if ($npts > 1) {$scanTimeReal = $srStartTime + $cnt_timeReal * ($npts - 1);}
  else           {$scanTimeReal = $srStartTime + $cnt_timeReal;}
  if ($SRSCAN_FLAGS->{'has run'} == 0 && $scanTimeReal > 1.5*$scanTime) {
     printf STDOUT '!!! '.$script.': because of Scaler and ScanRec '
                                 .'overhead the scantime will be '
                                 .'%1.0fs instead of %1.0f !'."\n",
                                  $scanTimeReal, $scanTime;
  }
#-----------------------------------------------------------------
### The globals below are defined in beamline.h:
  $scanrec   = $scanrecName;
  $scanAbort = $scanrecAbort;
  if (defined $shutter) {
     if ($shutter !~ /^${beamline}:/) {$shutter   = $beamline.':'.$shutter;}
  }
  if (defined $shutter_readback) {
     if ($shutter_readback !~ /^${beamline}:/) {
                                $shutter_readback = $beamline.':'.$shutter_readback;}
  }
  if ($scanrec    !~ /^${beamline}:/) {$scanrec   = $beamline.':'.$scanrec;}
  if ($scanAbort  !~ /^${beamline}:/) {$scanAbort = $beamline.':'.$scanAbort;}
#-----------------------------------------------------------------
  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
     print STDOUT 'Closing shutter ['.$shutter.']'."\n";
     &closeShutter ($shutter,$shutter_readback);
  } else {
     if ($SRSCAN_FLAGS->{'has run'} == 0) {print STDOUT '!!! '.$script.': shutter is not used!'."\n";}
  }

### Read the scaler background offsets:
  if (lc($fileBkg) eq 'auto') {
     $scalerNo = $scaler;
     $scalerNo =~ s/^.*\D+//;				#strip out all non-numeric (\D) stuff at the beginning
     $fileBkg  = 'joerger'.$scalerNo.'_Bkg.dat';
  }
  if (!defined $Offsets[0]) {				#skip reading offsets if they were read in main
     if (length($fileBkg) > 0) {print STDOUT 'Reading scaler ['.$scaler.'] offsets from '.$fileBkg."\n";}
     else                      {print STDOUT 'No scaler ['.$scaler.'] offsets file specified. Setting zero offsets.'."\n";}
     @Offsets = &getJoergerOffsets ($fileBkg,$scaler);	#we make it GLOBAL to access from main
  }

  if (defined $postarrayPV) {
     ($status,$nelm) = Pezca::GetDouble($postarrayPV.'.NELM');
     if ($status) {&report_only($script, $status, $postarrayPV.'.NELM'); $nelm = 2048;}
  }

### set scaler:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up scaler ['.$scaler.']'."\n";
     &setJoerger($scaler,$cnt_time);
  }

### set DAC:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up DAC ['.$dacPV.']'."\n";
     $status = &setupDAC ($dacPV);
     if ($status) {
        $messageSRSCAN = 'Setting up DAC ['.$dacPV.'] failed.';
        print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
        return $status;
     }
     ($dacLo,$dacHi) = &limitsDAC ($dacPV);
     ($SRSCAN_FLAGS->{'dacLo'},$SRSCAN_FLAGS->{'dacHi'}) = ($dacLo,$dacHi);
     $status = &setDACslope ($dacPV,0.0);
     if ($status) {
        $messageSRSCAN = 'Setting up DAC ['.$dacPV.'] slope=0 failed.';
        print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
        return $status;
     }
  } else {
     ($dacLo,$dacHi) = ($SRSCAN_FLAGS->{'dacLo'},$SRSCAN_FLAGS->{'dacHi'});
  }

### set Scan Record:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
#    print STDOUT 'Clearing scan record ['.$scanrec.'], ['.$scanAbort.']'."\n";
     &clearSR ($scanrec);

     print STDOUT 'Setting up scan record ['.$scanrec.'] positioner'."\n";
### The DAC must stay after the scan since the slew speed is in effect
### (we better reset the speed and then move the DAC outside ScanRec)
     $afterScanSR = 0;        				#0=stay
### Even if we pre-drive DAC to the starting position, SR may still attempt
### to do so. Therefore, we exclude the rqsDAC from SR positioners and emulate
## the request to run to the end with the "before scan" functionality:
#    &setupSRmotorsMore ($scanrec, $npts, $scanMode, $settleTime, $afterScanSR, 1,		#this does not always work
#                                    ($dacPV.'rb'), ($dacPV), ($begin), ($end));
     &setupSRmotorsMore ($scanrec, $npts, $scanMode, $settleTime, $afterScanSR, 1,
                                        [$dacPV.'rb'], [''], [$begin], [$end]);

     print STDOUT 'Setting up scan record ['.$scanrec.'] detector(s): ';
     &setupSRdetectors_pv ($scanrec, $scaler, $inputsRef, $pvlistRef);
     printf STDOUT '... nDetectors='.$ninputs.'+'.$npvs.'=%d'."\n",$ndetectors;
     if ($adc_encdrPV ne $dacPV) {
        print STDOUT 'Adding ADC readback to scan record ['.$scanrec.'] detector(s): ';
        $status = &setupSRdetectors_addpvs ($scanrec, [$adc_encdrPV], $ndetectors+1);
        if ($status) {
           print STDOUT '!!! '.$script.': Adding ADC readback failed with status='.
                              $status.'. Proceeding with no readback!'."\n\n";
           $adc_encdrPV = $dacPV;
        }
     }
  } else {
     print STDOUT 'Changing scan record ['.$scanrec.'] drive range'."\n";
     &changeSRmotors ($scanrec, $npts, 1, [$begin], [$end]);
  }
  &setupSRbeforeScan ($scanrec, $dacPV, $end, 0);

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Record initial DAC:
# $dac_ini = &getDACactpos($dacPV);		#Oleg: This may be incorrect in no-scan mode
  $dac_ini = &readDAC($dacPV);
  $dacPV_common = &DAC2common($dacPV);
### Measure DAC-ADC offset:
  $adc_offset = 0.;
  if ($adc_encdrPV ne $dacPV) {
     ($status,$adc_ini) = Pezca::GetDouble($adc_encdrPV);
     if ($status) {&report_only($script, $status, $adc_encdrPV);}
     else {
        $adc_offset = $adc_ini - $dac_ini;
        $msg = sprintf "\n".$script.': DAC=%.2f, ADC(rb)=%.2f, Offset=%.2f'."\n\n",
                                       $dac_ini, $adc_ini, $adc_offset;
        print STDOUT $msg;
        if (abs($adc_offset) > $dac_tolerance) {
           &tellBluIce ($msg,'warn');
           if (abs($adc_offset) > &maxval(2*$dac_tolerance,0.5)) {
              $adc_encdrPV = $dacPV;
              $adc_offset = 0.;
              $msg = '!!! '.$script.': the initial DAC-ADC mismatch is too high.'
                   . ' Turning off DAC readback control in the fly scanning'."\n";
              print STDOUT $msg;
              &tellBluIce ($msg,'warn');
           }
        }
     }
  }
### Move DAC to the starting position:
  printf STDOUT 'Moving DAC ['.$dacPV.'] to the starting position=%.3f'."\n",$begin;
  $status = &driveDAC ($dacPV, $begin, $dacLo, $dacHi);
  if ($status) {
     $messageSRSCAN = sprintf 'Moving DAC ['.$dacPV.'] to the starting position=%.3f failed.',$begin;
     print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
     return $status;
  }

  if ($SRSCAN_FLAGS->{'has run'} == 0) {
### Setting monitors on scan record execution:
     $status = &addCaMonitor($scanrec.'.BUSY');
     $status = &addCaMonitor($scanrec.'.CPT');
### Setting monitors on DAC:
     $status = &addCaMonitor($dacPV_common.'_wfScan');
  }

### Setting DAC slew speed:
  if ($scanTimeReal > 0) {
     printf STDOUT 'Setting DAC scan time to %.3fs'."\n",$scanTimeReal;
     $dac_slope = abs($end - $begin)/$scanTimeReal;
  } else {
     $dac_slope = 0.0;
  }
  $status = &setDACslope ($dacPV,$dac_slope);
  if ($status) {
     $messageSRSCAN = sprintf 'Setting up DAC ['.$dacPV.'] slope=%.3f failed.',$dac_slope;
     print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
     return $status;
  }

  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter open'} == 0) {
     print STDOUT 'Opening shutter ['.$shutter.']'."\n";
     $status = &openShutter ($shutter,$shutter_readback);
     if ($status) {
        &abortDAC ($dacPV);				# abort DAC and restore ramp
        &closeShutter($shutter,$shutter_readback);
        $messageSRSCAN = 'Failed opening fast shutter ['.$shutter.'].';
        print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
        return $status;
     }
  } else {
     if ($SRSCAN_FLAGS->{'has run'} == 0) {print STDOUT '!!! '.$script.': shutter is not used!'."\n";}
  }

  $wait_step     = 0.1;
  $wait_         = 0;
  $busy          = 1;
  $currPts       = 0;
  $dac_scanning  = 1;
  $i_plot        = 1;				#the main plot = 1st input ($inputs[0])
  $name_plot     = 'y'.$i_plot.'SR';
  $npts_last     = 0;
  $time_epoch    = time();			#seconds since epoch

### Start scan record:
  print  STDOUT 'Starting on-the-fly scan with Scan Record ['.$scanrec.']...'."\n";
  printf STDOUT 'DAC='.$dacPV.'  range=[%.3f : %.3f]  speed=%.3f(V/s)'."\n",
                $begin, $end, $dac_slope;
  &startSR ($scanrec, $scanAbort);
# &press_key_continue();

  if    ($medm == 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_pv($scanrec, $inputsRef, undef, 'realtime');		#do not open MEDM for PVs
  }
  elsif ($medm  > 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_pv($scanrec, $inputsRef, $pvlistRef, 'realtime');
  }

  no strict 'refs';
### Wait to finish (don't give up first $wait_start seconds):
  while ((($busy == 1 || $currPts < 1) && $dac_scanning) || $wait_< $wait_start) {
     select(undef,undef,undef,$wait_step);	# sleep($wait_step);
     $wait_ += $wait_step;
     ($status,$busy)   = Pezca::GetDouble($scanrec.'.BUSY');
     ($status,$currPts)= Pezca::GetDouble($scanrec.'.CPT');
     ($status,$dac_scanning) = Pezca::GetDouble($dacPV_common.'_wfScan');
#    print STDOUT 'busy='.$busy.' currPts='.$currPts.' dacScanning='
#                 .$dac_scanning.' wait_='.$wait_."\n";
     if (defined $postarrayPV || $checkOverload) {  #- - Begin the post array bisiness - - - - - -
        $time_passed = time() - $time_epoch;
        if ($time_passed >= $post_interval && $currPts > $npts_last+1) {
           $time_epoch = time();			#reset seconds since epoch
           $npts_last  = $currPts;
           @x1SR = &readSRx ($scanrec, $currPts, 1, 'fly');
           @x1SR = &calcSRxMiddle ($currPts, $begin, @x1SR);
           @y1SR = &readSRy ($scanrec, $currPts, $i_plot, 'fly');
           for ($j=0; $j<$currPts; $j++) {
              $y1SR[$j] = $y1SR[$j]/$cnt_time;
              if ($checkOverload) {if ($y1SR[$j] >= $threshold) {$overload_detected++;}}
           }
           if (defined $postarrayPV) {
              ### Normalize scaler data for time and correct for background:
#             $offsetNr=@$inputsRef[$i_plot-1];  $min_val=$Offsets[$offsetNr-1];         #subtract measured noise (before 2019.02)
              $min_val=&minvalRef(\@y1SR);                                               #subtract min value (after 2019.02)
              @post_array = ($scan_index,$npts_last,$begin,$end,@x1SR,);
              for ($j=0; $j<$currPts; $j++) {push @post_array,($y1SR[$j]-$min_val);}
              $narr = @post_array;
              if ($narr <= $nelm) {$status = Pezca::PutList($postarrayPV, 'FLOAT', @post_array);}
              else {
                 print STDOUT '!!! '.$script.': cannot post '.$narr.'-element array into PV='
                             .$postarrayPV.'. Array is too long. Further posting will be disabled.'."\n\n";
                 $status = 0;
                 undef $postarrayPV;
              }
              if ($status) {
                 &report_only($script, $status, $postarrayPV);
                 print STDOUT '!!! '.$script.': error posting '.$narr.'-element array into PV='
                             .$postarrayPV.'. Further posting will be disabled.'."\n\n";
                 undef $postarrayPV;
              }
           }
           if ($overload_detected > 0) {
              print STDOUT '!!! '.$script.': Intensity overload detected '.$overload_detected.' times.'."\n";
              $busy = 0; $dac_scanning = 0;
           }
        }
     }                            #- - - End the post array bisiness - - - - - -
  }
  printf STDOUT 'Scan finished,  time=%.1fs'."\n", $wait_;
  print  STDOUT '(scanBusy='.$busy.', dacScanning='.$dac_scanning.', scanCpt='.$currPts.')'."\n";
  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
     print STDOUT 'Closing shutter ['.$shutter.']'."\n";
     &closeShutter ($shutter,$shutter_readback);
  }
  &abortSR ($scanAbort, $scanrec);
  &abortDAC ($dacPV);				# abort DAC and restore ramp
  &setupSRbeforeScan ($scanrec, '', 0, 0);	# cleanup before-scan

  if ($overload_detected) {
     print STDOUT '!!! '.$script.':  Restoring DAC to the beginning for re-scanning...'."\n\n";
     &driveDAC ($dacPV, $begin, $dacLo, $dacHi);
     use strict 'refs';
     $messageSRSCAN = 'Intensity overload detected.';
     print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
     return ($OVERLOAD,$dac_slope);
  }
### Wait the data to be ready:
  print STDOUT 'Waiting for scan record data...'."\n";
  $actPts = &waitSRdata ($scanrec);

  print STDOUT 'Reading scan record data: npts='.$actPts.' nDetectors='.$ndetectors."\n";

  if    ($medm == 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_pv($scanrec, $inputsRef, undef, 'afterscan');		#do not open MEDM for PVs
  }
  elsif ($medm  > 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_pv($scanrec, $inputsRef, $pvlistRef, 'afterscan');
  }

  @x1SR = &readSRx ($scanrec, $actPts, 1);
  if ($adc_encdrPV ne $dacPV) {
     @encoder = &readSRy ($scanrec, $actPts, $ndetectors+1);
     for ($i=0; $i<$actPts; $i++) {
        if (abs($encoder[$i]-$x1SR[$i]-$adc_offset) > $dac_tolerance) {
           printf STDOUT '!!! '.$script.': DAC mismatch problem at'
                        .' rqsDAC=%.3f:'
                        .' readback=%.3f,'
                        .' rbk_offset=%.3f,'
                        .' (rbk - rbk_offset)=%.3f,'
                        .' error=%.3f,'
                        .' tolerance=%.3f,'
                        .' DAC=%s,'
                        .' ADC=%s'
                        ."\n",
                        $x1SR[$i],
                        $encoder[$i],
                        $adc_offset,
                        ($encoder[$i]-$adc_offset),
                        abs($encoder[$i]-$x1SR[$i]-$adc_offset),
                        $dac_tolerance,
                        $dacPV,
                        $adc_encdrPV;
                        $follerr_detected++;
        }
     }
     if ($follerr_detected > 0) {
        $msg = sprintf "\n".'!!! '.$script.': DAC following error detected %d times of %d.'."\n"
                           .'DAC=%s ADC(readback)=%s Tolerance=%.2fV Offset=%.2fV'."\n\n",
                           $follerr_detected,$actPts,$dacPV,$adc_encdrPV,$dac_tolerance,$adc_offset;
        print STDOUT $msg;
        &tellBluIce ($msg,'warn');
        if ($follerr_detected > 0.1*$actPts) {
           $msg = sprintf '!!! '.$script.':  Rescan needed. Restoring DAC=%s to V=%.3f.'."\n\n", $dacPV, $dac_ini;
           print STDOUT $msg;
           &tellBluIce ($msg,'warn');
           &driveDAC ($dacPV, $dac_ini, $dacLo, $dacHi);
           use strict 'refs';
           $messageSRSCAN = 'DAC following error detected.';
           print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
           return ($FOLLERR,$dac_slope);
        } else {
           $msg = '!!! '.$script.':  ignoring following errors as their rate is under 10%'."\n\n";
           print STDOUT $msg;
           &tellBluIce ($msg,'warn');
        }
     }
  }
  @x1SR = &calcSRxMiddle ($actPts, $begin, @x1SR);

  for ($i=1; $i<=$ndetectors; $i++) {
     $name  = 'y'.$i.'SR';
     @$name = &readSRy ($scanrec, $actPts, $i);
     if ($i <= $ninputs) {
### Normalize scaler data for time and correct for background:
        $offsetNr = @$inputsRef[$i-1];
        for ($j=0; $j<$actPts; $j++) {
           @$name[$j] = @$name[$j] / $cnt_time;
           if (@$name[$j] >= $threshold) {$overload_detected++;}
           @$name[$j] = @$name[$j] - $Offsets[$offsetNr-1];
        }
     }
     else  {
### Average PV data over measurements intervals:
        $d0 = @$name[0];
        for ($j=1; $j<$actPts; $j++) {
           $d1=@$name[$j];
           @$name[$j]=0.5*($d1+$d0);
           $d0=$d1;
        }
     }
  }

  if (defined $postarrayPV && $actPts > $npts_last) {  #- - Begin the post array bisiness - - - - - -
     $min_val = &minvalRef(\@$name_plot);                                           #subtract min value (after 2019.02)
     @post_array = ($scan_index,$actPts,$begin,$end,@x1SR);
     for ($j=0; $j<$currPts; $j++) {push @post_array,(@$name_plot[$j]-$min_val);}
     $narr = @post_array;
     if ($narr <= $nelm) {$status = Pezca::PutList($postarrayPV, 'FLOAT', @post_array);}
     else {
        print STDOUT '!!! '.$script.': cannot post '.$narr.'-element array into PV='
                    .$postarrayPV.'. Array is too long. Further posting will be disabled.'."\n\n";
        $status = 0;
        undef $postarrayPV;
     }
     if ($status) {
        &report_only($script, $status, $postarrayPV);
        print STDOUT '!!! '.$script.': error posting '.$narr.'-element array into PV='
                    .$postarrayPV.'. Further posting will be disabled.'."\n\n";
        undef $postarrayPV;
     }
  }                            #- - - End the post array bisiness - - - - - -

  print STDOUT 'Scan record data retrieved!'."\n";
  if ($overload_detected > 0) {
     print STDOUT '!!! '.$script.': Intensity overload detected '.$overload_detected.' times.'."\n";
     if ($checkOverload) {
        print STDOUT '!!! '.$script.':  Restoring DAC to the beginning for re-scanning...'."\n\n";
        &driveDAC ($dacPV, $begin, $dacLo, $dacHi);
        use strict 'refs';
        $messageSRSCAN = 'Intensity overload detected.';
        print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
        return ($OVERLOAD,$dac_slope);
     } else {
        print STDOUT '!!! '.$script.':  Ignoring intensity overload.'."\n\n";
     }
  }

### -----------------------------------------------------
### afterScan:
### (see dac_subs.pl)
  &DAC_afterScan($afterScan,\@x1SR,\@$name_plot,$begin,$end,$dac_ini,$dacPV);

  use strict 'refs';
  print STDOUT '..DAC scan finished!'."\n";
  $SRSCAN_FLAGS->{'has run'} = 1;
  if ($status) {
     $messageSRSCAN = 'Scan failed with status='.$status;
     print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
     return ($status,$dac_slope);
  }
  elsif ($overload_detected > 0) {
     $messageSRSCAN = 'Scan detected intensity overload.';
     print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
     return ($OVERLOAD,$dac_slope);
  }
  else {
     return (0,$dac_slope);
  }
}	##SRflyScanDAC

#========================================================================

# ($status,$actPts) =
#   &SRstepScan ($assy,$axis,$begin,$end,$npts,$cntTime,$medm,$scaler,$cntStartPV,$fileBkg,\@inputs,[\@pvlist,$settleTime]);
#sub SRstepScan ($assy,$axis,$begin,$end,$npts,$cntTime,$medm,$scaler,$cntStartPV,$fileBkg,\@inputs,[\@pvlist,$settleTime]) {
sub  SRstepScan ($$$$$$$$$$\@;\@$) {

### This function implements step scan with SCAN RECORD, PMAC motors and Joerger scaler + PVlist
###
### Returns GLOBALS: @x1SR,@y1SR,...@y{$ninputs+$npvs}SR, @Offsets
###
### #########################################################
### # ATTENTION: Requires Tim Mooney correction for PMAC DB #
### #########################################################

  our ($beamline, $shutter, $shutter_readback, $shutter_closed);
  our ($scanrecName, $scanrecAbort, @Offsets);
  our ($IGNORE_ERRORS, $report);
  our ($SRSCAN_FLAGS);		#reference to a hash with various flags for SRSCAN
  my ($script, $assy, $axis, $begin, $end, $npts, $medm);
  my ($scaler, $scalerNo, $fileBkg, $offsetNr, $inputsRef, $ninputs);
  my ($cntTime, $pvlistRef, $npvs, $ndetectors, $cntStartPV);
  my ($scanrec, $scanAbort, $scanMode, $afterScan);
  my ($settleTime, $status, $name, $i, $j);
  my ($wait_start, $wait_step, $wait_, $busy, $actPts, $currPts);

  $script     = 'SRstepScan';
  $assy       = shift(@_);
  $axis       = shift(@_);
  $begin      = shift(@_);	#start scan position
  $end        = shift(@_);	#end scan position
  $npts       = shift(@_);	#number of points on fast scan
  $cntTime    = shift(@_);	#counttime per point in seconds
  $medm       = shift(@_);	#1=open MEDM for scaler inputs, 2=for inputs+PVs, 0=not
  $scaler     = shift(@_);	#scaler1=Struck, scaler3=Joerger
  $cntStartPV = shift(@_);	#sequence PV for opening shutter & starting Joerger
  $fileBkg    = shift(@_);	#joerger background file ('auto' for automatic, '' for none)
  $inputsRef  = shift(@_);
  $pvlistRef  = shift(@_);
  $settleTime = shift(@_);

  if (!defined $IGNORE_ERRORS) {$IGNORE_ERRORS=0;}

  if (!defined $SRSCAN_FLAGS->{'has run'})              {$SRSCAN_FLAGS->{'has run'}              = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter close'})   {$SRSCAN_FLAGS->{'skip shutter close'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter open'})    {$SRSCAN_FLAGS->{'skip shutter open'}    = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip move to start'})   {$SRSCAN_FLAGS->{'skip move to start'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'return to start'})      {$SRSCAN_FLAGS->{'return to start'}      = 0;}

  $ninputs = @$inputsRef;
  if (defined $pvlistRef) {$npvs = @$pvlistRef;} else {$npvs = 0;}
  $ndetectors = $ninputs + $npvs;
#-----------------------------------------------------------------
### The globals below are defined in beamline.h:
  $scanrec   = $scanrecName;
  $scanAbort = $scanrecAbort;
  if (defined $shutter) {
     if ($shutter !~ /^${beamline}:/) {$shutter   = $beamline.':'.$shutter;}
  }
  if (defined $shutter_readback) {
     if ($shutter_readback !~ /^${beamline}:/) {
                                $shutter_readback = $beamline.':'.$shutter_readback;}
  }
  if ($scanrec    !~ /^${beamline}:/) {$scanrec   = $beamline.':'.$scanrec;}
  if ($scanAbort  !~ /^${beamline}:/) {$scanAbort = $beamline.':'.$scanAbort;}
#-----------------------------------------------------------------
  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
     print STDOUT 'Closing shutter ['.$shutter.']'."\n";
     &closeShutter ($shutter,$shutter_readback);
  } else {
     if ($SRSCAN_FLAGS->{'has run'} == 0) {print STDOUT '!!! '.$script.': shutter is not used!'."\n";}
     undef $cntStartPV;
  }
  if ($SRSCAN_FLAGS->{'skip shutter open'}) {undef $cntStartPV;}

### Read the scaler background offsets:
  if (!defined $fileBkg) {$fileBkg='';}
  if (lc($fileBkg) eq 'auto') {
     $scalerNo = $scaler;
     $scalerNo =~ s/^.*\D+//;	#strip out all non-numeric (\D) stuff at the beginning
     $fileBkg  = 'joerger'.$scalerNo.'_Bkg.dat';
  }
  if (!defined $Offsets[0]) {	#skip reading offsets if they have been read before or in main
     if (length($fileBkg) > 0) {print STDOUT 'Reading scaler ['.$scaler.'] offsets from '.$fileBkg."\n";}
     else                      {print STDOUT 'No scaler ['.$scaler.'] offsets file specified. Setting zero offsets.'."\n";}
     @Offsets = &getJoergerOffsets ($fileBkg,$scaler);  # we make it GLOBAL to access from main
  }

### set scaler:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up scaler ['.$scaler.']'."\n";
     &setJoerger($scaler,$cntTime);
  }

### set motor:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up motor ['.$axis.']'."\n";
     &configAssy ($assy);
  }

### set Scan Record:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
#    print STDOUT 'Clearing scan record ['.$scanrec.'], ['.$scanAbort.']'."\n";
     &clearSR ($scanrec);

     $scanMode = 0;					#0=linear
     if (!defined $settleTime) {$settleTime = 0.0;}     #positioner settle time (delay)
     $afterScan  = 0;        				#0=stay, 1=startpos, 2=priorpos 3=peakpos 4=valleypos 5=+edgepos 6=-edgepos
     print STDOUT 'Setting up scan record ['.$scanrec.'] positioner'."\n";
     &setupSRmotorsCommon ($scanrec, $npts, $settleTime, $afterScan);
     &setupSRmotorLIN     ($scanrec, 1, $axis.':ActPos', $axis.':RqsPos', $begin, $end, $scanMode);
     if (defined $cntStartPV) {
### $cntStartPV is a sequence PV for opening shutter & starting Joerger/MCA
### This closes shutter each time we start motion; besides we record time:
        &setupSRmotorLIN  ($scanrec, 2, 'time', $shutter, $shutter_closed, $shutter_closed, $scanMode);
     } else {
### Here we only record time:
        &setupSRmotorLIN  ($scanrec, 2, 'time',    '',            0,               0,       $scanMode);
     }

     print STDOUT 'Setting up scan record ['.$scanrec.'] detector(s): ';
     &setupSRdetectors_pv ($scanrec, $scaler, $inputsRef, $pvlistRef, $cntStartPV);
     printf STDOUT '... nDetectors='.$ninputs.'+'.$npvs.'=%d'."\n",$ndetectors;

     if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
        print STDOUT 'Setting up scan record ['.$scanrec.'] afterscan'."\n";
        &setupSRafterScan ($scanrec, $shutter, $shutter_closed, 0);
     }
  } else {
     print STDOUT 'Changing scan record ['.$scanrec.'] drive range'."\n";
     &changeSRmotorLIN ($scanrec, $npts, 1, $begin, $end);
  }

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# if ($SRSCAN_FLAGS->{'skip move to start'} == 0) {             #SR will do that
### Move motor to the starting position:
#    printf STDOUT 'Moving motor ['.$axis.'] to the starting position (%.3g)'."\n",$begin;
#    $status = &runMotor ($assy, $axis, $begin);
#    if ($status) {
#       print STDOUT '*** '.$script.': failed moving ['.$axis.'] to the starting position'."\n";
#       return $status, 0;
#    }
# }

### Setting monitors on scan record execution:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     $status = &addCaMonitor($scanrec.'.BUSY');
     $status = &addCaMonitor($scanrec.'.CPT');
  }

# if (defined $shutter) {		#commented out: shutter will be opened by $cntStartPV
#    if ($SRSCAN_FLAGS->{'skip shutter open'} == 0) {
#       print STDOUT 'Opening shutter ['.$shutter.']'."\n";
#       &openShutter ($shutter,$shutter_readback);
#    }
# }

  $wait_start = 4.0;
  $wait_step  = 0.1;
  $wait_      = 0;
  $busy       = 1;
  $currPts    = 0;

### Start scan record:
  print STDOUT 'Starting step scan with Scan Record ['.$scanrec.']...'."\n";
  &startSR ($scanrec, $scanAbort);
# &press_key_continue();

  if    ($medm == 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_pv($scanrec, $inputsRef, undef, 'realtime');		#do not open MEDM for PVs
  }
  elsif ($medm  > 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_pv($scanrec, $inputsRef, $pvlistRef, 'realtime');
  }

### Wait to finish (don't give up first $wait_start seconds):
  while ($busy == 1 || $currPts < 1 || $wait_< $wait_start) {
     select(undef,undef,undef,$wait_step);	#sleep($wait_step);
     $wait_ += $wait_step;
     ($status,$busy)   = Pezca::GetDouble($scanrec.'.BUSY');
     ($status,$currPts)= Pezca::GetDouble($scanrec.'.CPT');
#    print STDOUT 'busy='.$busy.' currPts='.$currPts.' wait_='.$wait_."\n";
  }
  printf STDOUT 'Scan finished! (scanBusy='.$busy.', scanCpt='.$currPts.', time=%.1fs)'."\n",$wait_;

# if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {	#this is done in scanRec
#    print STDOUT 'Closing shutter ['.$shutter.']'."\n";
#    &closeShutter ($shutter,$shutter_readback);
# }
  &abortSR ($scanAbort, $scanrec);
  &abortAssy ($assy);

### Wait the data to be ready:
  print STDOUT 'Waiting for scan record data...'."\n";
  $actPts = &waitSRdata ($scanrec);

### Start reverting motor to the beginning position
### in order to save time on the next scan:
  if ($SRSCAN_FLAGS->{'return to start'}) {&startMotor ($axis, $begin);}

  print STDOUT 'Reading scan record data: npts='.$actPts.' nDetectors='.$ndetectors."\n";

  if    ($medm == 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_pv($scanrec, $inputsRef, undef, 'afterscan');		#do not open MEDM for PVs
  }
  elsif ($medm  > 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_pv($scanrec, $inputsRef, $pvlistRef, 'afterscan');
  }

  no strict 'refs';
  $name   = 'x1SR';
  @$name = &readSRx ($scanrec, $actPts, 1);

  for ($i=1; $i<=$ndetectors; $i++) {
     $name  = 'y'.$i.'SR';
     @$name = &readSRy ($scanrec, $actPts, $i);
     if ($i <= $ninputs) {
### Normalize scaler data for time and correct for background:
        $offsetNr = @$inputsRef[$i-1];
        for ($j=0; $j<$actPts; $j++) {@$name[$j] = @$name[$j]/$cntTime - $Offsets[$offsetNr-1];}
     }
  }
  use strict 'refs';
  print STDOUT 'Scan record data retrieved!'."\n";
  $SRSCAN_FLAGS->{'has run'} = 1;
  return $status,$actPts;

}	##SRstepScan

#========================================================================

# ($status,$actPts) =
#   &SRtableScan ($assy,$axis,\@positions,$cntTime,$medm,$scaler,$cntStartPV,$fileBkg,\@inputs,[\@pvlist,$settleTime]);
#sub SRtableScan ($assy,$axis,\@positions,$cntTime,$medm,$scaler,$cntStartPV,$fileBkg,\@inputs,[\@pvlist,$settleTime]) {
sub  SRtableScan ($$\@$$$$$\@;\@$) {

### This function implements table scan with SCAN RECORD, PMAC motors and Joerger scaler + PVlist
###
### Returns GLOBALS: @x1SR,@y1SR,...@y{$ninputs+$npvs}SR, @Offsets
###
### #########################################################
### # ATTENTION: Requires Tim Mooney correction for PMAC DB #
### #########################################################

  our ($beamline, $shutter, $shutter_readback, $shutter_closed);
  our ($scanrecName, $scanrecAbort, @Offsets);
  our ($IGNORE_ERRORS, $report);
  our ($SRSCAN_FLAGS);		#reference to a hash with various flags for SRSCAN
  my ($script, $assy, $axis, $positionsRef, $npts, $medm);
  my ($scaler, $scalerNo, $fileBkg, $offsetNr, $inputsRef, $ninputs);
  my ($cntTime, $pvlistRef, $npvs, $ndetectors, $cntStartPV);
  my ($scanrec, $scanAbort, $scanMode, $afterScan);
  my ($settleTime, $status, $name, $i, $j);
  my ($wait_start, $wait_step, $wait_, $busy, $actPts, $currPts);

  $script       = 'SRtableScan';
  $assy         = shift(@_);
  $axis         = shift(@_);
  $positionsRef = shift(@_);	#start scan position
  $cntTime      = shift(@_);	#counttime per point in seconds
  $medm         = shift(@_);	#1=open MEDM for scaler inputs, 2=for inputs+PVs, 0=not
  $scaler       = shift(@_);	#scaler1=Struck, scaler3=Joerger
  $cntStartPV   = shift(@_);	#sequence PV for opening shutter & starting Joerger
  $fileBkg      = shift(@_);	#joerger background file ('auto' for automatic, '' for none)
  $inputsRef    = shift(@_);
  $pvlistRef    = shift(@_);
  $settleTime   = shift(@_);

  if (!defined $IGNORE_ERRORS) {$IGNORE_ERRORS=0;}

  if (!defined $SRSCAN_FLAGS->{'has run'})              {$SRSCAN_FLAGS->{'has run'}              = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter close'})   {$SRSCAN_FLAGS->{'skip shutter close'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter open'})    {$SRSCAN_FLAGS->{'skip shutter open'}    = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip move to start'})   {$SRSCAN_FLAGS->{'skip move to start'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'return to start'})      {$SRSCAN_FLAGS->{'return to start'}      = 0;}

  $npts = @$positionsRef;
  $ninputs = @$inputsRef;
  if (defined $pvlistRef) {$npvs = @$pvlistRef;} else {$npvs = 0;}
  $ndetectors = $ninputs + $npvs;
#-----------------------------------------------------------------
### The globals below are defined in beamline.h:
  $scanrec   = $scanrecName;
  $scanAbort = $scanrecAbort;
  if (defined $shutter) {
     if ($shutter !~ /^${beamline}:/) {$shutter   = $beamline.':'.$shutter;}
  }
  if (defined $shutter_readback) {
     if ($shutter_readback !~ /^${beamline}:/) {
                                $shutter_readback = $beamline.':'.$shutter_readback;}
  }
  if ($scanrec    !~ /^${beamline}:/) {$scanrec   = $beamline.':'.$scanrec;}
  if ($scanAbort  !~ /^${beamline}:/) {$scanAbort = $beamline.':'.$scanAbort;}
#-----------------------------------------------------------------
  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
     print STDOUT 'Closing shutter ['.$shutter.']'."\n";
     &closeShutter ($shutter,$shutter_readback);
  } else {
     if ($SRSCAN_FLAGS->{'has run'} == 0) {print STDOUT '!!! '.$script.': shutter is not used!'."\n";}
     undef $cntStartPV;
  }
  if ($SRSCAN_FLAGS->{'skip shutter open'}) {undef $cntStartPV;}

### Read the scaler background offsets:
  if (!defined $fileBkg) {$fileBkg='';}
  if (lc($fileBkg) eq 'auto') {
     $scalerNo = $scaler;
     $scalerNo =~ s/^.*\D+//;	#strip out all non-numeric (\D) stuff at the beginning
     $fileBkg  = 'joerger'.$scalerNo.'_Bkg.dat';
  }
  if (!defined $Offsets[0]) {	#skip reading offsets if they have been read before or in main
     if (length($fileBkg) > 0) {print STDOUT 'Reading scaler ['.$scaler.'] offsets from '.$fileBkg."\n";}
     else                      {print STDOUT 'No scaler ['.$scaler.'] offsets file specified. Setting zero offsets.'."\n";}
     @Offsets = &getJoergerOffsets ($fileBkg,$scaler);  # we make it GLOBAL to access from main
  }

### set scaler:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up scaler ['.$scaler.']'."\n";
     &setJoerger($scaler,$cntTime);
  }

### set motor:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up motor ['.$axis.']'."\n";
     &configAssy ($assy);
  }

### set Scan Record:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
#    print STDOUT 'Clearing scan record ['.$scanrec.'], ['.$scanAbort.']'."\n";
     &clearSR ($scanrec);

     $scanMode = 0;					#0=linear
     if (!defined $settleTime) {$settleTime = 0.0;}     #positioner settle time (delay)
     $afterScan  = 0;        				#0=stay, 1=startpos, 2=priorpos 3=peakpos 4=valleypos 5=+edgepos 6=-edgepos
     print STDOUT 'Setting up scan record ['.$scanrec.'] positioner'."\n";
     &setupSRmotorsCommon ($scanrec, $npts, $settleTime, $afterScan);
     &setupSRmotorTBL     ($scanrec, 1, $axis.':ActPos', $axis.':RqsPos', $positionsRef);
     if (defined $cntStartPV) {
### $cntStartPV is a sequence PV for opening shutter & starting Joerger/MCA
### This closes shutter each time we start motion; besides we record time:
        &setupSRmotorLIN  ($scanrec, 2, 'time', $shutter, $shutter_closed, $shutter_closed, $scanMode);
     } else {
### Here we only record time:
        &setupSRmotorLIN  ($scanrec, 2, 'time',    '',            0,               0,       $scanMode);
     }

     print STDOUT 'Setting up scan record ['.$scanrec.'] detector(s): ';
     &setupSRdetectors_pv ($scanrec, $scaler, $inputsRef, $pvlistRef, $cntStartPV);
     printf STDOUT '... nDetectors='.$ninputs.'+'.$npvs.'=%d'."\n",$ndetectors;

     if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
        print STDOUT 'Setting up scan record ['.$scanrec.'] afterscan'."\n";
        &setupSRafterScan ($scanrec, $shutter, $shutter_closed, 0);
     }
  } else {
     print STDOUT 'Changing scan record ['.$scanrec.'] drive range'."\n";
     &changeSRmotorTBL ($scanrec, $npts, 1, $positionsRef);
  }

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# if ($SRSCAN_FLAGS->{'skip move to start'} == 0) {             #SR will do that
### Move motor to the starting position:
#    printf STDOUT 'Moving motor ['.$axis.'] to the starting position (%.3g)'."\n",@$positionsRef[0];
#    $status = &runMotor ($assy, $axis, @$positionsRef[0]);
#    if ($status) {
#       print STDOUT '*** '.$script.': failed moving ['.$axis.'] to the starting position'."\n";
#       return $status, 0;
#    }
# }

### Setting monitors on scan record execution:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     $status = &addCaMonitor($scanrec.'.BUSY');
     $status = &addCaMonitor($scanrec.'.CPT');
  }

# if (defined $shutter) {		#commented out: shutter will be opened by $cntStartPV
#    if ($SRSCAN_FLAGS->{'skip shutter open'} == 0) {
#       print STDOUT 'Opening shutter ['.$shutter.']'."\n";
#       &openShutter ($shutter,$shutter_readback);
#    }
# }

  $wait_start = 4.0;
  $wait_step  = 0.1;
  $wait_      = 0;
  $busy       = 1;
  $currPts    = 0;

### Start scan record:
  print STDOUT 'Starting TABLE scan with Scan Record ['.$scanrec.']...'."\n";
  &startSR ($scanrec, $scanAbort);
# &press_key_continue();

  if    ($medm == 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_pv($scanrec, $inputsRef, undef, 'realtime');		#do not open MEDM for PVs
  }
  elsif ($medm  > 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_pv($scanrec, $inputsRef, $pvlistRef, 'realtime');
  }

### Wait to finish (don't give up first $wait_start seconds):
  while ($busy == 1 || $currPts < 1 || $wait_< $wait_start) {
     select(undef,undef,undef,$wait_step);	# sleep($wait_step);
     $wait_ += $wait_step;
     ($status,$busy)   = Pezca::GetDouble($scanrec.'.BUSY');
     ($status,$currPts)= Pezca::GetDouble($scanrec.'.CPT');
#    print STDOUT 'busy='.$busy.' currPts='.$currPts.' wait_='.$wait_."\n";
  }
  printf STDOUT 'Scan finished! (scanBusy='.$busy.', scanCpt='.$currPts.', time=%.1fs)'."\n",$wait_;

# if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) { #this is done in scanRec
#    print STDOUT 'Closing shutter ['.$shutter.']'."\n";
#    &closeShutter ($shutter,$shutter_readback);
# }
  &abortSR ($scanAbort, $scanrec);
  &abortAssy ($assy);

### Wait the data to be ready:
  print STDOUT 'Waiting for scan record data...'."\n";
  $actPts = &waitSRdata ($scanrec);

### Start reverting motor to the beginning position
### in order to save time on the next scan:
  if ($SRSCAN_FLAGS->{'return to start'}) {&startMotor ($axis, @$positionsRef[0]);}

  print STDOUT 'Reading scan record data: npts='.$actPts.' nDetectors='.$ndetectors."\n";

  if    ($medm == 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_pv($scanrec, $inputsRef, undef, 'afterscan');		#do not open MEDM for PVs
  }
  elsif ($medm  > 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_pv($scanrec, $inputsRef, $pvlistRef, 'afterscan');
  }

  no strict 'refs';
  $name   = 'x1SR';
  @$name = &readSRx ($scanrec, $actPts, 1);

  for ($i=1; $i<=$ndetectors; $i++) {
     $name  = 'y'.$i.'SR';
     @$name = &readSRy ($scanrec, $actPts, $i);
     if ($i <= $ninputs) {
### Normalize scaler data for time and correct for background:
        $offsetNr = @$inputsRef[$i-1];
        for ($j=0; $j<$actPts; $j++) {@$name[$j] = @$name[$j]/$cntTime - $Offsets[$offsetNr-1];}
     }
  }
  use strict 'refs';
  print STDOUT 'Scan record data retrieved!'."\n";
  $SRSCAN_FLAGS->{'has run'} = 1;
  return $status,$actPts;

}	##SRtableScan

#========================================================================

# ($status,$actPts) =
#   &SRstepScanMCA ($assy,$axis,$begin,$end,$npts,$cntTime,$medm,$mca,$cntStartPV,\@ROIs,[\@pvlist,$settleTime]);
#sub SRstepScanMCA ($assy,$axis,$begin,$end,$npts,$cntTime,$medm,$mca,$cntStartPV,\@ROIs,[\@pvlist,$settleTime]) {
sub  SRstepScanMCA ($$$$$$$$$\@;\@$) {

### This function implements step scan with SCAN RECORD, PMAC motors and MCA ROI + PVlist
###
### Returns GLOBALS: @x1SR,@y1SR,...@y{$nROIs+$npvs}SR
###
### #########################################################
### # ATTENTION: Requires Tim Mooney correction for PMAC DB #
### #########################################################

### This version works with MCA instead of Joergey-type scaler
### It also accepts optional PVlist to record during scan
###    (this list added to SR detectors after MCA ROIs)

  our ($beamline, $shutter, $shutter_readback, $shutter_closed);
  our ($scanrecName, $scanrecAbort);
  our ($IGNORE_ERRORS, $report);
  our ($SRSCAN_FLAGS);		#reference to a hash with various flags for SRSCAN
  my ($script, $assy, $axis, $begin, $end, $npts, $medm);
  my ($mca, $ROIsRef, $nROIs, $ROI, $roi_lo, $roi_hi, $nROI_set);
  my ($cntTime, $pvlistRef, $npvs, $ndetectors, $cntStartPV);
  my ($scanrec, $scanAbort, $scanMode, $afterScan);
  my ($settleTime, $status, $name, $i, $j);
  my ($wait_start, $wait_step, $wait_, $busy, $actPts, $currPts);

  $script     = 'SRstepScanMCA';
  $assy       = shift(@_);
  $axis       = shift(@_);
  $begin      = shift(@_);	#start scan position
  $end        = shift(@_);	#end scan position
  $npts       = shift(@_);	#number of points on fast scan
  $cntTime    = shift(@_);	#count time per point in seconds
  $medm       = shift(@_);	#1=open MEDM for scaler inputs, 2=for inputs+PVs, 0=not
  $mca        = shift(@_);	#Canberra or Falcon
  $cntStartPV = shift(@_);	#sequence PV for opening shutter & starting Joerger/MCA
  $ROIsRef    = shift(@_);
  $pvlistRef  = shift(@_);
  $settleTime = shift(@_);

  if (!defined $IGNORE_ERRORS) {$IGNORE_ERRORS=0;}

  if (!defined $SRSCAN_FLAGS->{'has run'})              {$SRSCAN_FLAGS->{'has run'}              = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter close'})   {$SRSCAN_FLAGS->{'skip shutter close'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter open'})    {$SRSCAN_FLAGS->{'skip shutter open'}    = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip move to start'})   {$SRSCAN_FLAGS->{'skip move to start'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'return to start'})      {$SRSCAN_FLAGS->{'return to start'}      = 0;}

  $nROIs = @$ROIsRef;
  if (defined $pvlistRef) {$npvs = @$pvlistRef;} else {$npvs = 0;}
  $ndetectors = $nROIs + $npvs;
#-----------------------------------------------------------------
### The globals below are defined in beamline.h:
  $scanrec   = $scanrecName;
  $scanAbort = $scanrecAbort;
  if (defined $shutter) {
     if ($shutter !~ /^${beamline}:/) {$shutter   = $beamline.':'.$shutter;}
  }
  if (defined $shutter_readback) {
     if ($shutter_readback !~ /^${beamline}:/) {
                                $shutter_readback = $beamline.':'.$shutter_readback;}
  }
  if ($scanrec    !~ /^${beamline}:/) {$scanrec   = $beamline.':'.$scanrec;}
  if ($scanAbort  !~ /^${beamline}:/) {$scanAbort = $beamline.':'.$scanAbort;}
#-----------------------------------------------------------------
  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
     print STDOUT 'Closing shutter ['.$shutter.']'."\n";
     &closeShutter ($shutter,$shutter_readback);
  } else {
     if ($SRSCAN_FLAGS->{'has run'} == 0) {print STDOUT '!!! '.$script.': shutter is not used!'."\n";}
     undef $cntStartPV;
  }
  if ($SRSCAN_FLAGS->{'skip shutter open'}) {undef $cntStartPV;}

### set MCA:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up MCA ['.$mca.']'."\n";
     &setMCA ($mca, $cntTime);
     $nROI_set = 0;
     for ($i=0; $i<$nROIs; $i++) {
        $ROI = @$ROIsRef[$i];
        ($roi_lo, $roi_hi) = &checkMCA_ROI($mca,$ROI);
        if ($roi_lo == -1 && $roi_hi == -1) {
           print STDOUT '!!! '.$script.': ROI('.$i.')='.$ROI.' is not setup in the MCA'."\n";
        }
        else {
           $nROI_set++;
        }
     }
     if ($nROI_set == 0) {
        print STDOUT '*** '.$script.': none of ROI are set in the MCA'."\n";
        return 222, 0;
     }
#    print STDOUT $script.':  assy='.$assy.'  axis='.$axis.'  begin='.$begin.'  end='.$end.'  npts='.$npts."\n";
#    print STDOUT $script.':  cntTime='.$cntTime.'  mca='.$mca.'  ROi='."@$ROIsRef\n";
  }

### set motor:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up motor ['.$axis.']'."\n";
     &configAssy ($assy);
  }

### set Scan Record:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
#    print STDOUT 'Clearing scan record ['.$scanrec.'], ['.$scanAbort.']'."\n";
     &clearSR ($scanrec);

     $scanMode = 0;					#0=linear
     if (!defined $settleTime) {$settleTime = 0.0;}	#positioner settle time (delay)
     $afterScan  = 0;        				#0=stay, 1=startpos, 2=priorpos 3=peakpos 4=valleypos 5=+edgepos 6=-edgepos
     print STDOUT 'Setting up scan record ['.$scanrec.'] positioner'."\n";
     &setupSRmotorsCommon ($scanrec, $npts, $settleTime, $afterScan);
     &setupSRmotorLIN     ($scanrec, 1, $axis.':ActPos', $axis.':RqsPos', $begin, $end, $scanMode);
     if (defined $cntStartPV) {
### $cntStartPV is a sequence PV for opening shutter & starting Joerger/MCA
### This closes shutter each time we start motion; besides we record time:
        &setupSRmotorLIN  ($scanrec, 2, 'time', $shutter, $shutter_closed, $shutter_closed, $scanMode);
     } else {
### Here we only record time:
        &setupSRmotorLIN  ($scanrec, 2, 'time',    '',            0,               0,       $scanMode);
     }

     print STDOUT 'Setting up scan record ['.$scanrec.'] detector(s): ';
     &setupSRdetectors_mca ($scanrec, $mca, $ROIsRef, $pvlistRef, $cntStartPV);
     printf STDOUT '... nDetectors='.$nROIs.'+'.$npvs.'=%d'."\n",$ndetectors;

     if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
        print STDOUT 'Setting up scan record ['.$scanrec.'] afterscan'."\n";
        &setupSRafterScan ($scanrec, $shutter, $shutter_closed, 0);
     }
  } else {
     print STDOUT 'Changing scan record ['.$scanrec.'] drive range'."\n";
     &changeSRmotorLIN ($scanrec, $npts, 1, $begin, $end);
  }

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# if ($SRSCAN_FLAGS->{'skip move to start'} == 0) {             #SR will do that
### Move motor to the starting position:
#    printf STDOUT 'Moving motor ['.$axis.'] to the starting position (%.3g)'."\n",$begin;
#    $status = &runMotor ($assy, $axis, $begin);
#    if ($status) {
#       print STDOUT '*** '.$script.': failed moving '.$axis.' to the starting position'."\n";
#       return $status, 0;
#    }
# }

### Setting monitors on scan record execution:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     $status = &addCaMonitor($scanrec.'.BUSY');
     $status = &addCaMonitor($scanrec.'.CPT');
  }

# if (defined $shutter) {		#commented out: shutter will be opened by $cntStartPV
#    if ($SRSCAN_FLAGS->{'skip shutter open'} == 0) {
#       print STDOUT 'Opening shutter ['.$shutter.']'."\n";
#       &openShutter ($shutter,$shutter_readback);
#    }
# }

  $wait_start = 4.0;
  $wait_step  = 0.1;
  $wait_      = 0;
  $busy       = 1;
  $currPts    = 0;

### Start scan record:
  print STDOUT 'Starting STEP scan with MCA ['.$mca.'] and scan record ['.$scanrec.']...'."\n";
  &startSR ($scanrec, $scanAbort);
# &press_key_continue();

  if    ($medm == 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_mca($scanrec, $ROIsRef, undef, 'realtime');
  }
  elsif ($medm  > 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_mca($scanrec, $ROIsRef, $pvlistRef, 'realtime');
  }

### Wait to finish (don't give up first $wait_start seconds):
  while ($busy == 1 || $currPts < 1 || $wait_< $wait_start) {
     select(undef,undef,undef,$wait_step);	# sleep($wait_step);
     $wait_ += $wait_step;
     ($status,$busy)   = Pezca::GetDouble($scanrec.'.BUSY');
     ($status,$currPts)= Pezca::GetDouble($scanrec.'.CPT');
#    print STDOUT 'busy='.$busy.' currPts='.$currPts.' wait_='.$wait_."\n";
  }
  printf STDOUT 'Scan finished! (scanBusy='.$busy.', scanCpt='.$currPts.', time=%.1fs)'."\n",$wait_;

# if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {	#this is done in scanRec
#    print STDOUT 'Closing shutter ['.$shutter.']'."\n";
#    &closeShutter ($shutter,$shutter_readback);
# }
  &abortSR ($scanAbort, $scanrec);
  &abortAssy ($assy);

### Wait the data to be ready:
  print STDOUT 'Waiting for scan record data...'."\n";
  $actPts = &waitSRdata ($scanrec);

### Start reverting motor to the beginning position
### in order to save time on the next scan:
  if ($SRSCAN_FLAGS->{'return to start'}) {&startMotor ($axis, $begin);}

  print STDOUT 'Reading scan record data: npts='.$actPts.' nDetectors='.$ndetectors."\n";

  if    ($medm == 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_mca($scanrec, $ROIsRef, undef, 'afterscan');
  }
  elsif ($medm  > 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_mca($scanrec, $ROIsRef, $pvlistRef, 'afterscan');
  }

  no strict 'refs';
  $name   = 'x1SR';
  @$name = &readSRx ($scanrec, $actPts, 1);

  for ($i=1; $i<=$ndetectors; $i++) {
    $name  = 'y'.$i.'SR';
    @$name = &readSRy ($scanrec, $actPts, $i);
    if ($i <= $nROIs) {
### Normalize mcadata for time:
      for ($j=0; $j<$actPts; $j++) {@$name[$j] = @$name[$j]/$cntTime;}
    }
  }
  use strict 'refs';
  print STDOUT 'Scan record data retrieved!'."\n";
  $SRSCAN_FLAGS->{'has run'} = 1;
  return $status,$actPts;

}	##SRstepScanMCA

#========================================================================

# ($status,$actPts) =
#   &SRtableScanMCA ($assy,$axis,\@positions,$cntTime,$medm,$mca,$cntStartPV,\@ROIs,[\@pvlist,$settleTime]);
#sub SRtableScanMCA ($assy,$axis,\@positions,$cntTime,$medm,$mca,$cntStartPV,\@ROIs,[\@pvlist,$settleTime]) {
sub  SRtableScanMCA ($$\@$$$$\@;\@$) {

### This function implements table scan with SCAN RECORD, PMAC motors and MCA ROI + PVlist
###
### Returns GLOBALS: @x1SR,@y1SR,...@y{$nROIs+$npvs}SR
###
### #########################################################
### # ATTENTION: Requires Tim Mooney correction for PMAC DB #
### #########################################################

### This version works with MCA instead of Joergey-type scaler
### It also accepts optional PVlist to record during scan
###    (this list added to SR detectors after MCA ROIs)

  our ($beamline, $shutter, $shutter_readback, $shutter_closed);
  our ($scanrecName, $scanrecAbort);
  our ($IGNORE_ERRORS, $report);
  our ($SRSCAN_FLAGS);		#reference to a hash with various flags for SRSCAN
  my ($script, $assy, $axis, $positionsRef, $npts, $medm);
  my ($mca, $ROIsRef, $nROIs, $ROI, $roi_lo, $roi_hi, $nROI_set);
  my ($cntTime, $pvlistRef, $npvs, $ndetectors, $cntStartPV);
  my ($scanrec, $scanAbort, $scanMode, $afterScan);
  my ($settleTime, $status, $name, $i, $j);
  my ($wait_start, $wait_step, $wait_, $busy, $actPts, $currPts);

  $script       = 'SRtableScanMCA';
  $assy         = shift(@_);
  $axis         = shift(@_);
  $positionsRef = shift(@_);	#start scan position
  $cntTime      = shift(@_);	#count time per point in seconds
  $medm         = shift(@_);	#1=open MEDM for scaler inputs, 2=for inputs+PVs, 0=not
  $mca          = shift(@_);	#Canberra or Falcon
  $cntStartPV   = shift(@_);	#sequence PV for opening shutter & starting Joerger/MCA
  $ROIsRef      = shift(@_);
  $pvlistRef    = shift(@_);
  $settleTime   = shift(@_);

  if (!defined $IGNORE_ERRORS) {$IGNORE_ERRORS=0;}

  if (!defined $SRSCAN_FLAGS->{'has run'})              {$SRSCAN_FLAGS->{'has run'}              = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter close'})   {$SRSCAN_FLAGS->{'skip shutter close'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter open'})    {$SRSCAN_FLAGS->{'skip shutter open'}    = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip move to start'})   {$SRSCAN_FLAGS->{'skip move to start'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'return to start'})      {$SRSCAN_FLAGS->{'return to start'}      = 0;}

  $npts = @$positionsRef;
  $nROIs = @$ROIsRef;
  if (defined $pvlistRef) {$npvs = @$pvlistRef;} else {$npvs = 0;}
  $ndetectors = $nROIs + $npvs;
#-----------------------------------------------------------------
### The globals below are defined in beamline.h:
  $scanrec   = $scanrecName;
  $scanAbort = $scanrecAbort;
  if (defined $shutter) {
     if ($shutter !~ /^${beamline}:/) {$shutter   = $beamline.':'.$shutter;}
  }
  if (defined $shutter_readback) {
     if ($shutter_readback !~ /^${beamline}:/) {
                                $shutter_readback = $beamline.':'.$shutter_readback;}
  }
  if ($scanrec    !~ /^${beamline}:/) {$scanrec   = $beamline.':'.$scanrec;}
  if ($scanAbort  !~ /^${beamline}:/) {$scanAbort = $beamline.':'.$scanAbort;}
#-----------------------------------------------------------------
  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
     print STDOUT 'Closing shutter ['.$shutter.']'."\n";
     &closeShutter ($shutter,$shutter_readback);
  } else {
     if ($SRSCAN_FLAGS->{'skip shutter close'} == 0) {print STDOUT '!!! '.$script.': shutter is not used!'."\n";}
     undef $cntStartPV;
  }
  if ($SRSCAN_FLAGS->{'skip shutter open'}) {undef $cntStartPV;}

### set MCA:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up MCA ['.$mca.']'."\n";
     &setMCA ($mca, $cntTime);
     $nROI_set = 0;
     for ($i=0; $i<$nROIs; $i++) {
        $ROI = @$ROIsRef[$i];
        ($roi_lo, $roi_hi) = &checkMCA_ROI($mca,$ROI);
        if ($roi_lo == -1 && $roi_hi == -1) {
           print STDOUT '!!! '.$script.': ROI('.$i.')='.$ROI.' is not setup in the MCA'."\n";
        }
        else {
           $nROI_set++;
        }
     }
     if ($nROI_set == 0) {
        print STDOUT '*** '.$script.': none of ROI are set in the MCA'."\n";
        return 222, 0;
     }
#    print STDOUT $script.':  assy='.$assy.'  axis='.$axis.'  npts='.$npts."\n";
#    print STDOUT $script.':  cntTime='.$cntTime.'  mca='.$mca.'  ROi='."@$ROIsRef\n";
  }

### set motor:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up motor ['.$axis.']'."\n";
     &configAssy ($assy);
  }

### set Scan Record:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
#    print STDOUT 'Clearing scan record ['.$scanrec.'], ['.$scanAbort.']'."\n";
     &clearSR ($scanrec);

     $scanMode = 0;					#0=linear
     if (!defined $settleTime) {$settleTime = 0.0;}	#positioner settle time (delay)
     $afterScan  = 0;        				#0=stay, 1=startpos, 2=priorpos 3=peakpos 4=valleypos 5=+edgepos 6=-edgepos
     print STDOUT 'Setting up scan record ['.$scanrec.'] positioner'."\n";
     &setupSRmotorsCommon ($scanrec, $npts, $settleTime, $afterScan);
     &setupSRmotorTBL     ($scanrec, 1, $axis.':ActPos', $axis.':RqsPos', $positionsRef);
     if (defined $cntStartPV) {
###  $cntStartPV is a sequence PV for opening shutter & starting Joerger/MCA
### This closes shutter each time we start motion; besides we record time:
        &setupSRmotorLIN  ($scanrec, 2, 'time', $shutter, $shutter_closed, $shutter_closed, $scanMode);
     } else {
### Here we only record time:
        &setupSRmotorLIN  ($scanrec, 2, 'time',    '',            0,               0,       $scanMode);
     }
     print STDOUT 'Setting up scan record ['.$scanrec.'] detector(s): ';
     &setupSRdetectors_mca ($scanrec, $mca, $ROIsRef, $pvlistRef, $cntStartPV);
     printf STDOUT '... nDetectors='.$nROIs.'+'.$npvs.'=%d'."\n",$ndetectors;

     if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
        print STDOUT 'Setting up scan record ['.$scanrec.'] afterscan'."\n";
        &setupSRafterScan ($scanrec, $shutter, $shutter_closed, 0);
     }
  } else {
     print STDOUT 'Changing scan record ['.$scanrec.'] positions table'."\n";
     &changeSRmotorTBL ($scanrec, $npts, 1, $positionsRef);
  }

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# if ($SRSCAN_FLAGS->{'skip move to start'} == 0) {             #SR will do that
### Move motor to the starting position:
#    printf STDOUT 'Moving motor ['.$axis.'] to the starting position (%.3g)'."\n",@$positionsRef[0];
#    $status = &runMotor ($assy, $axis, @$positionsRef[0]);
#    if ($status) {
#       print STDOUT '*** '.$script.': failed moving '.$axis.' to the starting position'."\n";
#       return $status, 0;
#    }
# }

### Setting monitors on scan record execution:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     $status = &addCaMonitor($scanrec.'.BUSY');
     $status = &addCaMonitor($scanrec.'.CPT');
  }

# if (defined $shutter) {		#commented out: shutter will be opened by $cntStartPV
#    if ($SRSCAN_FLAGS->{'skip shutter open'} == 0) {
#       print STDOUT 'Opening shutter ['.$shutter.']'."\n";
#       &openShutter ($shutter,$shutter_readback);
#    }
# }

  $wait_start = 4.0;
  $wait_step  = 0.1;
  $wait_      = 0;
  $busy       = 1;
  $currPts    = 0;

### Start scan record:
  print STDOUT 'Starting step scan with MCA ['.$mca.'] and scan record ['.$scanrec.']...'."\n";
  &startSR ($scanrec, $scanAbort);
# &press_key_continue();

  if    ($medm == 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_mca($scanrec, $ROIsRef, undef, 'realtime');
  }
  elsif ($medm  > 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_mca($scanrec, $ROIsRef, $pvlistRef, 'realtime');
  }

### Wait to finish (don't give up first $wait_start seconds):
  while ($busy == 1 || $currPts < 1 || $wait_< $wait_start) {
     select(undef,undef,undef,$wait_step);	#sleep($wait_step);
     $wait_ += $wait_step;
     ($status,$busy)   = Pezca::GetDouble($scanrec.'.BUSY');
     ($status,$currPts)= Pezca::GetDouble($scanrec.'.CPT');
#    print STDOUT 'busy='.$busy.' currPts='.$currPts.' wait_='.$wait_."\n";
  }
  printf STDOUT 'Scan finished! (scanBusy='.$busy.', scanCpt='.$currPts.', time=%.1fs)'."\n",$wait_;

# if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) { #this is done in scanRec
#    print STDOUT 'Closing shutter ['.$shutter.']'."\n";
#    &closeShutter ($shutter,$shutter_readback);
# }
  &abortSR ($scanAbort, $scanrec);
  &abortAssy ($assy);

### Wait the data to be ready:
  print STDOUT 'Waiting for scan record data...'."\n";
  $actPts = &waitSRdata ($scanrec);

### Start reverting motor to the beginning position
### in order to save time on the next scan:
  if ($SRSCAN_FLAGS->{'return to start'}) {&startMotor ($axis, @$positionsRef[0]);}

  print STDOUT 'Reading scan record data: npts='.$actPts.' nDetectors='.$ndetectors."\n";

  if    ($medm == 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_mca($scanrec, $ROIsRef, undef, 'afterscan');
  }
  elsif ($medm  > 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_mca($scanrec, $ROIsRef, $pvlistRef, 'afterscan');
  }

  no strict 'refs';
  $name   = 'x1SR';
  @$name = &readSRx ($scanrec, $actPts, 1);

  for ($i=1; $i<=$ndetectors; $i++) {
    $name  = 'y'.$i.'SR';
    @$name = &readSRy ($scanrec, $actPts, $i);
    if ($i <= $nROIs) {
### Normalize mcadata for time:
      for ($j=0; $j<$actPts; $j++) {@$name[$j] = @$name[$j]/$cntTime;}
    }
  }
  use strict 'refs';
  print STDOUT 'Scan record data retrieved!'."\n";
  $SRSCAN_FLAGS->{'has run'} = 1;
  return $status,$actPts;

}	##SRtableScanMCA

#========================================================================

#                           *******  OMS-58  ********
# ($status,$actPts) =
#   &SRstepScanOMS58 ($motor,$begin,$end,$npts,$medm,$scaler,$cnt_time,$fileBkg,$ninputs,\@inputs,[$settleTime]);
#sub SRstepScanOMS58 ($motor,$begin,$end,$npts,$medm,$scaler,$cnt_time,$fileBkg,$ninputs,\@inputs,[$settleTime]) {
sub  SRstepScanOMS58 ($$$$$$$$$\@;$) {
###
### This function implements STEP SCAN with SCAN RECORD, OMS58 (MotorRec) motors and Joerger inputs
### (tested at XOR Sector 30)
###
### Returns GLOBALS: @x1SR,@y1SR,...@y{$ninputs}SR, @Offsets
###
  our ($beamline, $shutter, $shutter_readback, $shutter_closed);
  our ($scanrecName, $scanrecAbort, @Offsets);
  our (@SRdetectors);		# make @SRdetectors global to keep between calls
  our ($IGNORE_ERRORS, $report);
  our ($SRSCAN_FLAGS);		#reference to a hash with various flags for SRSCAN
  my ($script, $motor, $begin, $end, $inputsRef, $npts, $medm, $scaler);
  my ($cnt_time, $fileBkg, $ninputs, $ninp, @inputs);
  my ($current_time, $msg, $i, $j, $scalerNo, $offsetNr);
  my ($scanMode, $afterScan, $settleTime, $status, $actPts, $name);
  my ($scanrec, $scanAbort, $wait_step, $wait_, $wait_start, $busy, $currPts);

  $script     = 'SRstepScanOMS58';
  $motor      = shift(@_);
  $begin      = shift(@_);	# start scan position
  $end        = shift(@_);	# end scan position
  $npts       = shift(@_);	# number of points on fast scan
  $medm       = shift(@_);	# 1=open MEDM, 0=not
  $scaler     = shift(@_);	# Joerger scaler name
  $cnt_time   = shift(@_);	# scaler counting time
  $fileBkg    = shift(@_);	# joerger background file ('auto' for automatic, '' for none)
  $ninputs    = shift(@_);
  $inputsRef  = shift(@_);
  $settleTime = shift(@_);	# ignored if 'undef'

  $ninp = @$inputsRef;
  if ($ninputs > $ninp) {
     $current_time = scalar localtime;
     $msg  = '*** '.$current_time."\n";
     $msg .= '*** '.$script.': for scanrec='.$scanrec
                   .' requested scaler inputs='.$ninputs
                   .' exceeds the array length:'."\n"
                   .' @inputs='.$ninp."\n";
     &press_key_exit($msg);
  }
  for ($i=0; $i<$ninputs; $i++) {$inputs[$i] = @$inputsRef[$i];}

  if (!defined $IGNORE_ERRORS) {$IGNORE_ERRORS=0;}

  if (!defined $SRSCAN_FLAGS->{'has run'})              {$SRSCAN_FLAGS->{'has run'}              = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter close'})   {$SRSCAN_FLAGS->{'skip shutter close'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter open'})    {$SRSCAN_FLAGS->{'skip shutter open'}    = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip move to start'})   {$SRSCAN_FLAGS->{'skip move to start'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'return to start'})      {$SRSCAN_FLAGS->{'return to start'}      = 0;}

  $wait_start = 4.0;
  $scanMode   = 0;		# 0=linear, 1=table 2=on-the-fly
  $afterScan  = 1;		# 0=stay, 1=startpos, 2=priorpos 3=peakpos 4=valleypos 5=+edgepos 6=-edgepos
  $status = 0;
  $actPts = 0;

#-----------------------------------------------------------------
### The globals below are defined in beamline.h:
  $scanrec   = $scanrecName;
  $scanAbort = $scanrecAbort;
  if (defined $shutter) {
     if ($shutter !~ /^${beamline}:/) {$shutter   = $beamline.':'.$shutter;}
  }
  if (defined $shutter_readback) {
     if ($shutter_readback !~ /^${beamline}:/) {
                                $shutter_readback = $beamline.':'.$shutter_readback;}
  }
  if ($scanrec    !~ /^${beamline}:/) {$scanrec   = $beamline.':'.$scanrec;}
  if ($scanAbort  !~ /^${beamline}:/) {$scanAbort = $beamline.':'.$scanAbort;}
#-----------------------------------------------------------------
  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
        print STDOUT 'Closing shutter ['.$shutter.']'."\n";
        &closeShutter ($shutter,$shutter_readback);
  } else {
    if ($SRSCAN_FLAGS->{'has run'} == 0) {print STDOUT '!!! '.$script.': shutter is not used!'."\n";}
  }

  if (!defined $fileBkg) {$fileBkg='';}
  if (lc($fileBkg) eq 'auto') {
     $scalerNo = $scaler;
     $scalerNo =~ s/^.*\D+//;	# strip out all non-numeric (\D) stuff at the beginning
     $fileBkg  = 'joerger'.$scalerNo.'_Bkg.dat';
  }
### Read the scaler background offsets:
  if (!defined $Offsets[0]) {	# skip reading offsets in case it was done before
     if (length($fileBkg) > 0) {print STDOUT 'Reading scaler ['.$scaler.'] offsets from '.$fileBkg."\n";}
     else                      {print STDOUT 'No scaler ['.$scaler.'] offsets file specified. Setting zero offsets.'."\n";}
     @Offsets = &getJoergerOffsets ($fileBkg, $scaler);  # we make it GLOBAL to access from main
  }

### set scaler:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up scaler ['.$scaler.']'."\n";
     &setJoerger($scaler,$cnt_time);
  }

### Set Scan Record:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
#    print STDOUT 'Clearing scan record ['.$scanrec.'], ['.$scanAbort.']'."\n";
     &clearSR ($scanrec,4);		# mode=4 leaves detector fields uncleared.
     print STDOUT 'Setting up scan record ['.$scanrec.'] positioner'."\n";
     &setupSRstepmotorsMore ($scanrec, $npts, $scanMode, $settleTime, $afterScan, 1, [$motor], [$begin], [$end]);
     print STDOUT 'Setting up scan record ['.$scanrec.'] detector(s): ';
     @SRdetectors = &setupSRdetectors_auto ($scanrec, $scaler, $ninputs, @inputs);
     print STDOUT '... nDetectors='.$ninputs."\n";
     if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
        print STDOUT 'Setting up scan record ['.$scanrec.'] afterscan'."\n";
        &setupSRafterScan ($scanrec, $shutter, $shutter_closed, 0);
     }
### Setting monitors on the scan record execution:
     $status = &addCaMonitor($scanrec.'.BUSY');
     $status = &addCaMonitor($scanrec.'.CPT');
     if ($medm) {&openSRmedm_detec($scanrec,$ninputs,\@inputs,\@SRdetectors,'realtime');}
  } else {
     print STDOUT 'Changing scan record ['.$scanrec.'] drive range'."\n";
     &changeSRmotors ($scanrec, $npts, 1, [$begin], [$end]);
  }

### Move motor to the starting position:
# if ($SRSCAN_FLAGS->{'skip move to start'} == 0) {     #scan record will do it
#    printf STDOUT 'Moving motor ['.$motor.'] to the starting position (%.3g)'."\n",$begin;
#    $status = &runStepMotor ($motor, $begin);
#    if ($status) {
#       print STDOUT '*** '.$script.': failed moving ['.$motor.'] to the starting position'."\n";
#       return $status, 0;
#    }
# }

  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter open'} == 0) {
     print STDOUT 'Opening shutter ['.$shutter.']'."\n";
     &openShutter ($shutter,$shutter_readback);
  }

  $wait_step = 0.1;
  $wait_     = 0;
  $busy      = 1;
  $currPts   = 0;

### Start scan record:
  print STDOUT 'Starting step scan with scan record ['.$scanrec.']...'."\n";
  &startSR ($scanrec, $scanAbort);
# &press_key_continue();

### Wait to finish (don't give up first $wait_start seconds):
  while ($busy ==1 || $currPts < 1 || $wait_< $wait_start) {
    select(undef,undef,undef,$wait_step);	# sleep($wait_step);
    $wait_ += $wait_step;
    ($status,$busy)   = Pezca::GetDouble($scanrec.'.BUSY');
    ($status,$currPts)= Pezca::GetDouble($scanrec.'.CPT');
  }
  printf STDOUT 'Scan finished! (scanBusy='.$busy.', scanCpt='.$currPts.', time=%.1fs)'."\n",$wait_;

# if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) { #this is done in scanRec
#    print STDOUT 'Closing shutter ['.$shutter.']'."\n";
#    &closeShutter ($shutter,$shutter_readback);
# }
  &abortSR ($scanAbort, $scanrec);
# &abortStepMotor ($motor);

### Wait the data to be ready:
  print STDOUT 'Waiting for scan record data...'."\n";
  $actPts = &waitSRdata ($scanrec);

### Start reverting motor to the beginning position
### in order to save time on the next scan:
  if ($SRSCAN_FLAGS->{'return to start'}) {&startStepMotor ($motor, $begin);}

  print STDOUT 'Reading scan record data: npts='.$actPts.' nDetectors='.$ninputs."\n";

  if ($medm && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_detec($scanrec,$ninputs,\@inputs,\@SRdetectors,'afterscan');
  }

  no strict 'refs';
  $name   = 'x1SR';
  @$name = &readSRx ($scanrec, $actPts, 1);
# @$name = &calcSRxMiddle ($actPts, $begin, @$name);    # for fly scans

  for ($i=0; $i<$ninputs; $i++) {
    $name = sprintf 'y%dSR', $i+1;
    @$name = &readSRy ($scanrec, $actPts, $SRdetectors[$i]);
### Normalize data for time and correct for background:
    $offsetNr = $inputs[$i];
    for ($j=0; $j<$actPts; $j++) {
      $$name[$j] = $$name[$j]/$cnt_time - $Offsets[$offsetNr-1];
    }
  }
  use strict 'refs';
  print STDOUT 'Scan record data retrieved!'."\n";
  $SRSCAN_FLAGS->{'has run'} = 1;
  return $status,$actPts;

}	##SRstepScanOMS58

#========================================================================

# $status =
#   &SRstepScanDAC ($dacPV,$begin,$end,$npts,$cntTime,$medm,
#                   $scaler,$fileBkg,\@inputs,[\@pvlist,$settleTime,$checkOverload,
#                   $postarrayPV,$post_interval,
#                   $adc_encdrPV, $dac_tolerance]);
#sub SRstepScanDAC ($dacPV,$begin,$end,$npts,$cntTime,$medm,
#                   $scaler,$fileBkg,\@inputs,[\@pvlist,$settleTime,$checkOverload,
#                   $postarrayPV,$post_interval,
#                   $adc_encdrPV, $dac_tolerance]) {
sub  SRstepScanDAC ($$$$$$$$\@;\@$$$$$$) {

### This function implements STEP SCAN with SCAN RECORD, DAC and (Joerger inputs + PVlist)
###
### Returns GLOBALS: @x1SR,@y1SR,...@y{$ninputs}SR, @Offsets

  our ($beamline, $shutter, $shutter_readback, $shutter_closed);
  our ($scanrecName, $scanrecAbort);
  our (@Offsets, @x1SR, @y1SR);
  our ($IGNORE_ERRORS, $report);
  our ($SRSCAN_FLAGS);		#reference to a hash with various flags for SRSCAN
  my ($script, $npts, $npts_last);
  my ($dacPV, $dac_ini, $begin, $end, $dacLo, $dacHi);
  my ($cntTime, $medm, $scaler, $fileBkg);
  my ($inputsRef, $pvlistRef, $scanMode);
  my ($settleTime, $afterScan, $scanTime);
  my ($ninputs, $npvs, $ndetectors);
  my ($actPts, $cnt_time, $cnt_min);
  my ($scanrec, $scanAbort, $scalerNo);
  my ($status, $name, $i, $j, $min_val);
  my ($offsetNr, $busy, $currPts);
  my ($wait_step, $wait_, $wait_start);
  my ($threshold, $checkOverload);
  my ($postarrayPV, $post_interval, $scan_index);
  my ($i_plot, $time_epoch, $time_passed);
  my ($name_plot, @post_array, $narr, $nelm);
  my ($OVERLOAD, $overload_detected);
  my ($FOLLERR, $follerr_detected, $adc_offset, $adc_ini);
  my ($adc_encdrPV, $dac_tolerance, $tolerance_min, @encoder, $msg);

  $script        = 'SRstepScanDAC';
  $dacPV         = shift(@_);
  $begin         = shift(@_);				#start scan position
  $end           = shift(@_);				#end scan position
  $npts          = shift(@_);				#number of points on fast scan
  $cntTime       = shift(@_);				#scaler counting time per point in seconds
  $medm          = shift(@_);				#1=open MEDM for scaler inputs, 2=for inputs+PVs, 0=not
  $scaler        = shift(@_);				#scaler1=Struck, scaler3=Joerger
  $fileBkg       = shift(@_);				#joerger background file ('auto' for automatic, '' for none)
  $inputsRef     = shift(@_);
  $pvlistRef     = shift(@_);
  $settleTime    = shift(@_);
  $checkOverload = shift(@_);
  $postarrayPV   = shift(@_);
  $post_interval = shift(@_);
  $adc_encdrPV   = shift(@_);		#ADC PV to test DAC output (optional)
  $dac_tolerance = shift(@_);		#tolerance when testing DAC output (optional; default=step/2)

  $tolerance_min = 0.2;			#the minimum sensible DAC error because of different DAC & ADC calibrations
  if (!defined $settleTime)    {$settleTime    = 0.1;}  	#positioner settle time (delay)
  if (!defined $post_interval) {$post_interval = 2;}		#post interval
  if (!defined $checkOverload) {$checkOverload = 0;}   		#check overload flag
  if (!defined $fileBkg)       {$fileBkg='';}
  if (!defined $adc_encdrPV)   {$adc_encdrPV   = $dacPV;}	#ADC PV to test DAC output (optional)
  if (!defined $dac_tolerance) {$dac_tolerance = 0.;}		#ADC channel to test DAC output (optional)
  $dac_tolerance = abs($dac_tolerance);
  if ($dac_tolerance == 0.)    {$dac_tolerance = abs(0.5*($end-$begin)/$npts);}
  if ($dac_tolerance < $tolerance_min) {$dac_tolerance = $tolerance_min;}

  if (!defined $IGNORE_ERRORS) {$IGNORE_ERRORS=0;}

  if (!defined $SRSCAN_FLAGS->{'has run'})              {$SRSCAN_FLAGS->{'has run'}              = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter close'})   {$SRSCAN_FLAGS->{'skip shutter close'}   = 0;}
  if (!defined $SRSCAN_FLAGS->{'skip shutter open'})    {$SRSCAN_FLAGS->{'skip shutter open'}    = 0;}
#
  if (!defined $SRSCAN_FLAGS->{'dacLo'})                {$SRSCAN_FLAGS->{'dacLo'}                = 0;}
  if (!defined $SRSCAN_FLAGS->{'dacHi'})                {$SRSCAN_FLAGS->{'dacHi'}                = 0;}

  $scanMode   = 0;        				#0=linear, 1=table 2=on-the-fly
  $afterScan  = 0;        				#0=stay, 1=startpos, 2=priorpos 3=peakpos 4=valleypos 5=+edgepos 6=-edgepos
  $cnt_min    = 0.1;            			#minimum counting time
  $wait_start = 4.0;
# $threshold  = 0.95E+5;				#threshold intensity for detecting overload (before 2013)
  $threshold  = 0.95E+6;				#threshold intensity for detecting overload (2013)
  $OVERLOAD   = 100000;					#scan failure flag when overload detected
  $overload_detected = 0;				#overload not detected
  $FOLLERR    = 200000;					#scan failure flag when following error detected
  $follerr_detected = 0;
  $scan_index = 0;					#scan index to save with the post (e.g. 11-vertical, 21-horizontal)

  $scanTime = $npts*($cntTime + $settleTime + 0.2);	#scan time in seconds

  $ninputs = @$inputsRef;
  if (defined $pvlistRef) {$npvs = @$pvlistRef;}
  else                    {$npvs = 0;}
  $ndetectors = $ninputs + $npvs;

  if ($cntTime < $cnt_min) {
     printf STDOUT '*** '.$script.': Too short time per point=%.3fs; '
                                 .'the minimum is '.$cnt_min."s\n", $cntTime;
     return 666;
  }
#-----------------------------------------------------------------
### The globals below are defined in beamline.h:
  $scanrec   = $scanrecName;
  $scanAbort = $scanrecAbort;
  if (defined $shutter) {
     if ($shutter !~ /^${beamline}:/) {$shutter   = $beamline.':'.$shutter;}
  }
  if (defined $shutter_readback) {
     if ($shutter_readback !~ /^${beamline}:/) {
                                $shutter_readback = $beamline.':'.$shutter_readback;}
  }
  if ($scanrec    !~ /^${beamline}:/) {$scanrec   = $beamline.':'.$scanrec;}
  if ($scanAbort  !~ /^${beamline}:/) {$scanAbort = $beamline.':'.$scanAbort;}
#-----------------------------------------------------------------
  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
     print STDOUT 'Closing shutter ['.$shutter.']'."\n";
     &closeShutter ($shutter,$shutter_readback);
  } else {
     if ($SRSCAN_FLAGS->{'has run'} == 0) {print STDOUT '!!! '.$script.': shutter is not used!'."\n";}
  }

  if (defined $postarrayPV) {
     ($status,$nelm) = Pezca::GetDouble($postarrayPV.'.NELM');
     if ($status) {&report_only($script, $status, $postarrayPV.'.NELM'); $nelm = 2048;}
  }

### Read the scaler background offsets:
  if (lc($fileBkg) eq 'auto') {
     $scalerNo = $scaler;
     $scalerNo =~ s/^.*\D+//;				#strip out all non-numeric (\D) stuff at the beginning
     $fileBkg  = 'joerger'.$scalerNo.'_Bkg.dat';
  }
  if (!defined $Offsets[0]) {				#skip reading offsets if they were read in main
     if (length($fileBkg) > 0) {print STDOUT 'Reading scaler ['.$scaler.'] offsets from '.$fileBkg."\n";}
     else                      {print STDOUT 'No scaler ['.$scaler.'] offsets file specified. Setting zero offsets.'."\n";}
     @Offsets = &getJoergerOffsets ($fileBkg,$scaler);	#we make it GLOBAL to access from main
  }

### set scaler:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up scaler ['.$scaler.']'."\n";
     &setJoerger($scaler,$cntTime);
  }

### set DAC:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
     print STDOUT 'Setting up DAC ['.$dacPV.']'."\n";
     $status = &setupDAC ($dacPV);
     if ($status) {return $status;}
     ($dacLo,$dacHi) = &limitsDAC ($dacPV);
     ($SRSCAN_FLAGS->{'dacLo'},$SRSCAN_FLAGS->{'dacHi'}) = ($dacLo,$dacHi);
  } else {
     ($dacLo,$dacHi) = ($SRSCAN_FLAGS->{'dacLo'},$SRSCAN_FLAGS->{'dacHi'});
  }

### Record initial DAC:
# $dac_ini = &getDACactpos($dacPV);			#Oleg: this may be incorrect in no-scan mode
  $dac_ini = &readDAC($dacPV);

### Measure DAC-ADC offset:
  $adc_offset = 0.;
  if ($adc_encdrPV ne $dacPV) {
     ($status,$adc_ini) = Pezca::GetDouble($adc_encdrPV);
     if ($status) {&report_only($script, $status, $adc_encdrPV);}
     else {
        $adc_offset=$adc_ini-$dac_ini;
        $msg = sprintf "\n".$script.': DAC=%.2f, ADC(rb)=%.2f, Offset=%.2f'."\n\n",
                                   $dac_ini,$adc_ini,$adc_offset;
        print STDOUT $msg;
        if (abs($adc_offset) > $dac_tolerance) {&tellBluIce ($msg,'err'); $adc_offset = 0.;}
     }
  }

### set Scan Record:
  if ($SRSCAN_FLAGS->{'has run'} == 0) {
#    print STDOUT 'Clearing scan record ['.$scanrec.'], ['.$scanAbort.']'."\n";
     &clearSR ($scanrec);

     print STDOUT 'Setting up scan record ['.$scanrec.'] positioner'."\n";
     &setupSRmotorsMore ($scanrec, $npts, $scanMode, $settleTime, $afterScan, 1, [$dacPV], [$dacPV], [$begin], [$end]);

     print STDOUT 'Setting up scan record ['.$scanrec.'] detector(s): ';
     &setupSRdetectors_pv ($scanrec, $scaler, $inputsRef, $pvlistRef);
     printf STDOUT '... nDetectors='.$ninputs.'+'.$npvs.'=%d'."\n",$ndetectors;
     if ($adc_encdrPV ne $dacPV) {
        print STDOUT 'Adding ADC readback to scan record ['.$scanrec.'] detector(s): ';
        $status = &setupSRdetectors_addpvs ($scanrec, [$adc_encdrPV], $ndetectors+1);
        if ($status) {
           print STDOUT '!!! '.$script.': Adding ADC readback failed with status='.
                              $status.'. Proceeding with no readback!'."\n\n";
           $adc_encdrPV = $dacPV;
        }
     }

     if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) {
        print STDOUT 'Setting up scan record ['.$scanrec.'] afterscan'."\n";
        &setupSRafterScan ($scanrec, $shutter, $shutter_closed, 0);
     }
  } else {
     print STDOUT 'Changing scan record ['.$scanrec.'] drive range'."\n";
     &changeSRmotors ($scanrec, $npts, 1, [$begin], [$end]);
  }

  if ($SRSCAN_FLAGS->{'has run'} == 0) {
### Setting monitors on scan record execution:
     $status = &addCaMonitor($scanrec.'.BUSY');
     $status = &addCaMonitor($scanrec.'.CPT');
  }

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Move DAC to the starting position:
  print STDOUT 'Moving DAC ['.$dacPV.'] to the starting position ('.$begin.')'."\n";
  $status = &driveDAC ($dacPV, $begin, $dacLo, $dacHi);
  if ($status) {return $status;}

  if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter open'} == 0) {
     print STDOUT 'Opening shutter ['.$shutter.']'."\n";
     &openShutter ($shutter,$shutter_readback);
  }

  $wait_step     = 0.1;
  $wait_         = 0;
  $busy          = 1;
  $currPts       = 0;
  $i_plot        = 1;				#the main plot = 1st input ($inputs[0])
  $name_plot     = 'y'.$i_plot.'SR';
  $npts_last     = 0;
  $time_epoch    = time();			#seconds since epoch

### Start scan record:
  print STDOUT 'Starting scan with Scan Record ['.$scanrec.']...'."\n";
  &startSR ($scanrec, $scanAbort);
# &press_key_continue();

  if    ($medm == 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_pv($scanrec, $inputsRef, undef, 'realtime');
  }
  elsif ($medm  > 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_pv($scanrec, $inputsRef, $pvlistRef, 'realtime');
  }

  no strict 'refs';
### Wait to finish (don't give up first $wait_start seconds):
  while ($busy == 1 || $currPts < 1 || $wait_< $wait_start) {
     select(undef,undef,undef,$wait_step);		#sleep($wait_step);
     $wait_ += $wait_step;
     ($status,$busy)   = Pezca::GetDouble($scanrec.'.BUSY');
     ($status,$currPts)= Pezca::GetDouble($scanrec.'.CPT');
#    print STDOUT 'busy='.$busy.' currPts='.$currPts.' wait_='.$wait_."\n";
     if (defined $postarrayPV || $checkOverload) {  #- - Begin post array bisiness - - - - - -
        $time_passed = time() - $time_epoch;
        if ($time_passed >= $post_interval && $currPts > $npts_last+1) {
           $time_epoch = time();			#reset seconds since epoch
           $npts_last  = $currPts;
           @x1SR = &readSRx ($scanrec, $currPts, 1, 'fly');
           @y1SR = &readSRy ($scanrec, $currPts, $i_plot, 'fly');
           for ($j=0; $j<$currPts; $j++) {
              $y1SR[$j] = $y1SR[$j]/$cnt_time;
              if ($checkOverload) {if ($y1SR[$j] >= $threshold) {$overload_detected++;}}
           }
           if (defined $postarrayPV) {
              ### Normalize scaler data for time and correct for background:
#             $offsetNr = @$inputsRef[$i_plot-1];  $min_val=$Offsets[$offsetNr-1];         #subtract measured background (before 2019.02)
              $min_val=&minvalRef(\@y1SR);                                                 #subtract min value (after 2019.02)
              @post_array = ($scan_index,$npts_last,$begin,$end,@x1SR);
              for ($j=0; $j<$currPts; $j++) {push @post_array,($y1SR[$j]-$min_val);}
              $narr = @post_array;
              if ($narr <= $nelm) {$status = Pezca::PutList($postarrayPV, 'FLOAT', @post_array);}
              else {
                 print STDOUT '!!! '.$script.': cannot post '.$narr.'-element array into PV='
                             .$postarrayPV.'. Array is too long. Further posting will be disabled.'."\n\n";
                 $status = 0;
                 undef $postarrayPV;
              }
              if ($status) {
                 &report_only($script, $status, $postarrayPV);
                 print STDOUT '!!! '.$script.': error posting '.$narr.'-element array into PV='
                             .$postarrayPV.'. Further posting will be disabled.'."\n\n";
                 undef $postarrayPV;
              }
           }
           if ($overload_detected > 0) {
              print STDOUT '!!! '.$script.': Intensity overload detected '.$overload_detected.' times.'."\n";
              $busy = 0;
           }
           
        }
     }                            #- - - End post array bisiness - - - - - -
  }
  printf STDOUT 'Scan finished! (scanBusy='.$busy.', scanCpt='.$currPts.', time=%.1fs)'."\n",$wait_;

# if (defined $shutter && $SRSCAN_FLAGS->{'skip shutter close'} == 0) { #this is done in scanRec
#    print STDOUT 'Closing shutter [.'.$shutter.']'."\n";
#    &closeShutter ($shutter,$shutter_readback);
# }
  &abortSR ($scanAbort, $scanrec);

  if ($overload_detected > 0) {
     print STDOUT '!!! '.$script.': Setting DAC to the beginning for re-scanning...'."\n\n";
     &driveDAC ($dacPV, $begin, $dacLo, $dacHi);
     return $OVERLOAD;
  }

### Wait the data to be ready:
  print STDOUT 'Waiting for scan record data...'."\n";
  $actPts = &waitSRdata ($scanrec);

  print STDOUT 'Reading scan record data: npts='.$actPts.' nDetectors='.$ndetectors."\n";

  if    ($medm == 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_pv($scanrec, $inputsRef, undef, 'afterscan');
  }
  elsif ($medm  > 1 && $SRSCAN_FLAGS->{'has run'} == 0) {
     &openSRmedm_pv($scanrec, $inputsRef, $pvlistRef, 'afterscan');
  }

  @x1SR = &readSRx ($scanrec, $actPts, 1);
  if ($adc_encdrPV ne $dacPV) {
     @encoder = &readSRy ($scanrec, $actPts, $ndetectors+1);
     for ($i=0; $i<$actPts; $i++) {
        if (abs($encoder[$i]-$x1SR[$i]-$adc_offset) > $dac_tolerance) {
           printf STDOUT '!!! '.$script.': DAC mismatch problem at rqsDAC=%.3f:'
            .' readback=%.3f (rbk-offset=%.3f, err=%.3f, tolerance=%.3f)'."\n",
            $x1SR[$i],$encoder[$i],,($encoder[$i]-$adc_offset),
            abs($encoder[$i]-$x1SR[$i]-$adc_offset),$dac_tolerance;
           $follerr_detected++;
        }
     }
     if ($follerr_detected > 0) {
        $msg = sprintf "\n".'!!! '.$script.': DAC following error detected %d times of %d.'."\n"
                           .'DAC=%s ADC(readback)=%s Tolerance=%.2fV Offset=%.2fV'."\n\n",
                           $follerr_detected,$actPts,$dacPV,$adc_encdrPV,$dac_tolerance,$adc_offset;
        print STDOUT $msg;
        &tellBluIce ($msg,'warn');
        if ($follerr_detected > 0.1*$actPts) {
           $msg = '!!! '.$script.':  Rescan needed. Restoring DAC.'."\n\n";
           print STDOUT $msg;
           &tellBluIce ($msg,'warn');
           &driveDAC ($dacPV, $dac_ini, $dacLo, $dacHi);
           use strict 'refs';
           return $FOLLERR;
        } else {
           $msg = '!!! '.$script.':  ignoring following errors as their rate is under 10%'."\n\n";
           print STDOUT $msg;
           &tellBluIce ($msg,'warn');
        }
     }
  }

  for ($i=1; $i<=$ndetectors; $i++) {
     $name  = 'y'.$i.'SR';
     @$name = &readSRy ($scanrec, $actPts, $i);
     if ($i <= $ninputs) {
### Normalize scaler data for time and correct for background:
        $offsetNr = @$inputsRef[$i-1];
        for ($j=0; $j<$actPts; $j++) {
           @$name[$j] = @$name[$j]/$cntTime;
           if (@$name[$j] >= $threshold) {$overload_detected++;}
           @$name[$j] = @$name[$j] - $Offsets[$offsetNr-1];
        }
     }
  }

  if (defined $postarrayPV && $actPts > $npts_last) {  #- - Begin post array bisiness - - - - - -
     $min_val = &minvalRef(\@$name_plot);                         #subtract min value (after 2019.02)
     @post_array = ($scan_index,$actPts,$begin,$end,@x1SR);
     for ($j=0; $j<$currPts; $j++) {push @post_array,(@$name_plot[$j]-$min_val);}
     $narr = @post_array;
     if ($narr <= $nelm) {$status = Pezca::PutList($postarrayPV, 'FLOAT', @post_array);}
     else {
        print STDOUT '!!! '.$script.': cannot post '.$narr.'-element array into PV='
                    .$postarrayPV.'. Array is too long. Further posting will be disabled.'."\n\n";
        $status = 0;
        undef $postarrayPV;
     }
     if ($status) {
        &report_only($script, $status, $postarrayPV);
        print STDOUT '!!! '.$script.': error posting '.$narr.'-element array into PV='
                    .$postarrayPV.'. Further posting will be disabled.'."\n\n";
        undef $postarrayPV;
     }
  }                            #- - - End post array bisiness - - - - - -

  use strict 'refs';
  print STDOUT 'Scan record data retrieved!'."\n";
  $SRSCAN_FLAGS->{'has run'} = 1;

  if ($status) {
     return $status;
  }
  elsif ($overload_detected > 0) {
     print STDOUT '!!! '.$script.': Intensity overload detected '.$overload_detected.' times.'."\n";
     if ($checkOverload) {
        print STDOUT '!!! '.$script.': Setting DAC to the beginning for re-scanning...'."\n\n";
#       &driveDAC ($dacPV, $dac_ini, $dacLo, $dacHi);
        &driveDAC ($dacPV, $begin, $dacLo, $dacHi);
     } else {
        print STDOUT '!!! '.$script.': Ignoring.'."\n\n";
     }
     return $OVERLOAD;
  }
  else {
     return 0;
  }
}	##SRstepScanDAC

#========================================================================

# ($status,$actPts,$pos) =
#   &findPosSR ($assy,$axis,$begin,$end,$npts,$scanTime,$medm,$scaler,$fileBkg,$input,$target,$direction,[\%flags]);
#sub findPosSR ($assy,$axis,$begin,$end,$npts,$scanTime,$medm,$scaler,$fileBkg,$input,$target,$direction,[\%flags]) {
sub  findPosSR ($$$$$$$$$$$$;\%) {

### This function implements ON-THE-FLY scan with SCAN RECORD, PMAC motors and Joerger scaler.
### It analyzes scan data on the fly and when the intensity reaches $target (from either low
### or high depending on $direction) it aborts the scan and drives motor to the position where
### target intensity was reached.
###
### Returns GLOBALS: @x1SR,@y1SR, @Offsets
### Also returns GLOBAL $messageSRSCAN if $status != 0.

  our ($beamline, $shutter, $shutter_readback, $scanrecName, $scanrecAbort);
  our (@Offsets, $findPos2ndTime, $IGNORE_ERRORS, $messageSRSCAN);
  my ($script, $assy, $axis, $begin, $end, $npts, $scanTime, $medm);
  my ($scaler, $fileBkg, $IGNORE_ERRORS_local, $maxPts, %dir, $flagsref);
  my ($scalerNo, $input, $target, $direction, $status, $actPts, $position);
  my ($name, $cnt_time, $DECREASING, $INCREASING, $scanrec, $scanAbort, $j);
  my ($scanRange, $l_speed, $h_speed, $wait_step, $wait_, $wait_start, $busy);
  my ($runprg, $positioned, $currPts, $prevPts, $found, $found_max, $inpos);
  my ($pt_left, $pt_right, $y_left, $y_right, $y_curr, $pt_diff, $cnt_min);
  my ($x_left, $x_right, $x_min, $x_max, $pt_min, $y_min, $pt_max, $y_max);
  my (@good_int, @good_pts, $found_r, $found_p, $currMin, $hasStarted);

  $script    = 'findPosSR';
  $assy      = shift(@_);
  $axis      = shift(@_);
  $begin     = shift(@_);		#start scan position
  $end       = shift(@_);		#end scan position
  $npts      = shift(@_);		#number of points on fast scan
  $scanTime  = shift(@_);		#scan time in seconds
  $medm      = shift(@_);		#1=open MEDM, 0=not
  $scaler    = shift(@_);		#scaler1=Struck, scaler3=Joerger
  $fileBkg   = shift(@_);		#joerger background file ('auto' for automatic, '' for none)
  $input     = shift(@_);		#Joerger scaler input for detecting intensity
  $target    = shift(@_);		#target intensity to stop search
  $direction = shift(@_);		#targeted intensity behavior: 0=decreasing, 1=increasing
  $flagsref  = shift(@_);		#reference to an optional hash with program flags (options)

  if (!defined $findPos2ndTime) {$findPos2ndTime=0;}
  if (!defined $IGNORE_ERRORS)  {$IGNORE_ERRORS=0;}
  if (!defined $flagsref->{'can change step'})     {$flagsref->{'can change step'}=0;}
  if (!defined $flagsref->{'can change time ppt'}) {$flagsref->{'can change time ppt'}=0;}
  $IGNORE_ERRORS_local = $IGNORE_ERRORS;

  if (defined $shutter) {
     if ($shutter !~ /^${beamline}:/) {$shutter   = $beamline.':'.$shutter;}
  }
  if (defined $shutter_readback) {
     if ($shutter_readback !~ /^${beamline}:/) {
                                $shutter_readback = $beamline.':'.$shutter_readback;}
  }
  if (!defined $fileBkg) {$fileBkg='';}
  if ($fileBkg =~ /^auto$/i) {
     $scalerNo = $scaler;
     $scalerNo =~ s/^.*\D+//;			#strip out all non-numeric (\D) stuff at the beginning
     $fileBkg  = 'joerger'.$scalerNo.'_Bkg.dat';
  }

  $cnt_min    = 0.1;            		#minimum counting time
  $wait_start = 4.0;
  $actPts     = 0;
  $position   = 0;
  $cnt_time   = 0;
  $DECREASING = 0;
  $INCREASING = 1;
  $status     = 0;
  $messageSRSCAN = '';
  $dir{$DECREASING} = 'Decreasing';
  $dir{$INCREASING} = 'Increasing';
  print  STDOUT 'Entering '.$script.' :'."\n";
  printf STDOUT '  motor='.$axis.', range=[%.4g : %.4g],'."\n",$begin,$end;
  printf STDOUT '  scaler='.$scaler.', input='.$input.', target_intensity=%.0f'."\n",$target;
  if ($direction eq $DECREASING || $direction eq $INCREASING) {
     printf STDOUT '  search_time=%.2fs, intensity_direction=%s'."\n",$scanTime,$dir{$direction};
  } else {
     printf STDOUT '  search_time=%.2fs, intensity_direction=%s'."\n",$scanTime,'**UNKNOWN**';
     $messageSRSCAN = 'UNKNOWN search direction';
     return 1,0,0;
  }
#-----------------------------------------------------------------
### The globals below are defined in beamline.h:
  if (defined $shutter) {
     print STDOUT 'Closing shutter ['.$shutter.']'."\n";
     &closeShutter ($shutter,$shutter_readback);
  } else {
     print STDOUT '!!! '.$script.': Shutter is not used!'."\n";
  }
  $scanrec   = $scanrecName;
  $scanAbort = $scanrecAbort;
  if ($scanrec    !~ /^${beamline}:/) {$scanrec   = $beamline.':'.$scanrec;}
  if ($scanAbort  !~ /^${beamline}:/) {$scanAbort = $beamline.':'.$scanAbort;}
  $maxPts = &SRmaxPts($scanrec);
#-----------------------------------------------------------------
### Begin parameters consistency tests

  if ($npts > $maxPts) {
     if ($flagsref->{'can change step'}) {
        printf STDOUT '!!! '.$script.': Too many scan points=%d (max=%d)'."\n"
        .'  Increasing scan step by  factor of %.2f'."\n", $npts,$maxPts,$npts/$maxPts;
        $npts = $maxPts;
     } else {
        $messageSRSCAN = 'Requested npts='.$npts.' exceeds scanrec buffer='.$maxPts;
        print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
        $status = 888;
        return $status,0,0;
     }
  }
  elsif ($npts < 2) {
     $messageSRSCAN = 'Requested npts='.$npts.' is too small: no room to search';
     print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
     $status = 999;
     return $status,0,0;
  }

### Calculate time per point
  if ($npts > 1) {$cnt_time = $scanTime / ($npts - 1);}
  else           {$cnt_time = $scanTime;}
  $cnt_time = $cnt_time - 0.005;			#subtract scanrec dead time
  if ($scaler eq $beamline.':scaler1') {
     $cnt_time = $cnt_time - 0.008;			#Struck needs some extra time to finish
  }
  if ($cnt_time < $cnt_min) {
     if ($flagsref->{'can change time ppt'}) {
        $cnt_time = $cnt_min + 0.005 + 0.008;
        $y_min = $scanTime;
        if ($npts > 1) {$scanTime = $cnt_time*($npts-1);}
        else           {$scanTime = $cnt_time;}
        printf STDOUT '!!! '.$script.': Too short time per point=%.2fs (min=%.2fs)'."\n"
        .'  Changing scan time from %.2fs to %.2fs'."\n", $cnt_min,$cnt_time,$y_min,$scanTime;
     } else {
        $messageSRSCAN = sprintf 'Too short time per point=%.2fs (min=%.2fs)', $cnt_min,$cnt_time;
        print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
        $status = 666;
        return $status,0,0;
     }
  }

### Calculate scan speed for the motor:
  print STDOUT 'Calculating slew-mode motor speed for ['.$axis.']'."\n";
  $scanRange = abs ($end - $begin);
  $IGNORE_ERRORS = 1;
  ($l_speed, $h_speed) = &calcAxisSpeed ($assy, $axis, $scanRange, $scanTime);
  $IGNORE_ERRORS = $IGNORE_ERRORS_local;
  if ($l_speed > 1.01*$h_speed && $flagsref->{'can change time ppt'}) {
     $cnt_time = $cnt_time * ($l_speed/$h_speed);
     $y_min = $scanTime;
     if ($npts > 1) {$scanTime = $cnt_time*($npts-1);}
     else           {$scanTime = $cnt_time;}
     printf STDOUT '!!! '.$script.': Calculated low speed is %.0f%s higher than motor can run'."\n"
                         .'  Changing scan time from %.2fs to %.2fs'."\n",
                      100*($l_speed-$h_speed)/$h_speed,'%',$y_min,$scanTime;
     ($l_speed, $h_speed) = &calcAxisSpeed ($assy, $axis, $scanRange, $scanTime);
  }
  if ($l_speed > 1.01*$h_speed) {
     $messageSRSCAN = sprintf 'Calculated low speed is %.0f%s higher than motor can run',
                                         100*($l_speed-$h_speed)/$h_speed,'%';
     print STDOUT '*** '.$script.': '.$messageSRSCAN."\n";
     $status = 777;
     return $status,0,0;
  }
### End parameters consistency tests
#-----------------------------------------------------------------

### Read the scaler background offsets:
  if (!defined $Offsets[0]) {			#skip reading offsets if they were read in main
     if (length($fileBkg) > 0) {print STDOUT 'Reading scaler ['.$scaler.'] offsets from '.$fileBkg."\n";}
     else                      {print STDOUT 'No scaler ['.$scaler.'] offsets file specified. Setting zero offsets.'."\n";}
     @Offsets = &getJoergerOffsets ($fileBkg, $scaler);  # we make it GLOBAL to access from main
  }

### set scaler:
  if ($findPos2ndTime == 0) {
     print STDOUT 'Setting up scaler ['.$scaler.']'."\n";
     &setJoerger($scaler,$cnt_time);
  }

### set motor:
  if ($findPos2ndTime == 0) {
     print STDOUT 'Setting up motor ['.$axis.']'."\n";
     &configAssy ($assy);
     &restoreAxis ($assy, $axis);		#restore high-speed
  }

### set Scan Record:
# print STDOUT 'Clearing scan record ['.$scanrec.'], ['.$scanAbort.']'."\n";
  &abortSR ($scanAbort, $scanrec);
  &clearSR ($scanrec);

  print STDOUT 'Setting up scan record ['.$scanrec.'] positioner'."\n";
### Give scanRec some more pts since motor may move longer because of accelerating:
  $actPts = $npts + 10;
  if ($actPts > $maxPts) {$actPts = $maxPts;}
  &setupSRmotors ($scanrec, $actPts, 1, [$axis.':ActPos'], [$axis.':RqsPos'], [$begin], [$end]);

  print STDOUT 'Setting up scan record ['.$scanrec.'] detector'."\n";
  &setupSRdetectors_v2 ($scanrec, $scaler, 1, [$input]);

  &setupSRrefreshrate ($scanrec, 0.1);				#for fly posting, 0.1 is the minumum time
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Move motor to the starting position:
  printf STDOUT 'Moving motor ['.$axis.'] to the starting position (%.4g)'."\n",$begin;
  &runMotor ($assy, $axis, $begin);

### Setting slow speed for scanning:
  printf STDOUT 'Setting up motor ['.$axis.'] speed to slew (%.4g)'."\n",$l_speed;
  &setAxisSlewScan ($assy, $axis, $l_speed);
# select(undef,undef,undef,2);			#sleep to settle

### Setting monitors on motors motion and scan record execution:
  $status = &addCaMonitor($scanrec.'.BUSY');
  $status = &addCaMonitor($scanrec.'.CPT');
  $status = &setAssyMonitors($assy);

  if (defined $shutter) {
     print STDOUT 'Opening shutter ['.$shutter.']'."\n";
     $status = &openShutter ($shutter,$shutter_readback);
     if ($status) {
        &abortAssy ($assy, 1, $axis);		# also restore speed and motion prg.
        &closeShutter($shutter,$shutter_readback);
        return $status,0,0;
     }
  }

  $wait_step  = 0.1;
  $wait_      = 0;
  $busy       = 1;
  $inpos      = 0;
  $runprg     = 1;
  $positioned = 0;
  $currPts    = 0;
  $prevPts    = $currPts;
  $found_p    = 0;
  $found_max  = 3;				#max times to check that the condition is satisfied
  @good_pts   = ();
  @good_int   = ();
  $y_curr     = 0;
  $hasStarted = 0;
  $currMin    = &minval(($actPts,2));

### Start scan record:
  print STDOUT 'Starting on-the-fly search with scan record ['.$scanrec.']...'."\n";
  &startSR ($scanrec, $scanAbort);
# &press_key_continue();

  if ($medm && ($findPos2ndTime == 0)) {&openSRmedm($scanrec,1,[$input],'realtime');}

#---------------------------------------------------------------------------
### Wait to finish (don't give up first $wait_start seconds):
  while ( (($busy == 1 || $currPts < $currMin) &&
           ($runprg || $positioned<3 || $hasStarted == 0))
                                     || $wait_< $wait_start ) {
     select(undef,undef,undef,$wait_step);		#sleep($wait_step);
     $wait_ += $wait_step;
     ($status,$busy)    = Pezca::GetDouble($scanrec.'.BUSY');
     ($status,$currPts) = Pezca::GetDouble($scanrec.'.CPT');
     ($status,$inpos)   = Pezca::GetDouble($assy.':InPos');
     ($status,$runprg)  = Pezca::GetDouble($assy.':RunPrg');
     if ($runprg == 1) {$hasStarted = 1;}
     if ($inpos == 1)  {$positioned++;}
     else              {$positioned=0;}
     if (($busy == 1) && ($currPts > 0) && ($currPts != $prevPts)) {
        $prevPts = $currPts;
        $y_curr  = &readSRy_curr ($scanrec, 1);		#read current value from 1st input of scanrec
        $y_curr  = $y_curr / $cnt_time - $Offsets[$input-1];
        if (($y_curr <= $target && $direction == $DECREASING) ||
            ($y_curr >= $target && $direction == $INCREASING)) {
           if ($found_p > 0 && $y_curr == $good_int[$found_p-1]) {next;}
           $good_pts[$found_p] = $currPts;
           $good_int[$found_p] = $y_curr;
           $found_p++;					#position found 1 time
        } else {
           @good_pts = ();
           @good_int = ();
           $found_p  = 0;				#skip glitches
        }
        if ($found_p >= $found_max) {last;}
     }
#    print STDOUT 'busy='.$busy.' currPts='.$currPts.' runprg='.$runprg.' inpos='
#                 .$inpos.' positioned='.$positioned.' wait_='.$wait_."\n";
  } #-----------------------------------------------------------------------

  if (defined $shutter) {
    print STDOUT 'Closing shutter ['.$shutter.']'."\n";
    &closeShutter ($shutter,$shutter_readback);
  }
  &abortSR ($scanAbort, $scanrec);
  &abortAssy ($assy, 1, $axis);				#also restore speed and motion prg.

  if ($found_p >= $found_max) {
    printf STDOUT 'Search finished! (found='.$found_p.', scanCpt='.$currPts
                 .', time=%.1fs)'."\n",$wait_;
    printf STDOUT '  Good data points (target=%.0f):'."\n",$target;
    for ($j=0; $j<$found_max; $j++) {
       printf STDOUT '     pt=%d  intensity=%.0f'."\n",$good_pts[$j],$good_int[$j];
    }
  } else {
    printf STDOUT '!!! '.$script.': Search failed! (found='.$found_p
                 .', scanBusy='.$busy.', scanCpt='.$currPts
                 .', mtrRunPrg='.$runprg.', mtrPos='.$positioned.'
                 , time=%.1fs)'."\n",$wait_;
  }

### Wait the data to be ready:
  print STDOUT 'Waiting for scan record data...'."\n";
  $actPts = &waitSRdata ($scanrec);

  print STDOUT 'Reading scan record data: npts='.$actPts."\n";

  if ($medm && ($findPos2ndTime == 0)) {&openSRmedm($scanrec,1,[$input],'afterscan');}

  no strict 'refs';
  $name  = 'x1SR';
  @$name = &readSRx ($scanrec, $actPts, 1);		#($scanrec, $numPts, $scalerNumber)
  @$name = &calcSRxMiddle ($actPts, $begin, @$name);
  $name  = 'y1SR';
  @$name = &readSRy ($scanrec, $actPts, 1);
### Normalize data for time and correct for background:
  for ($j=0; $j<$actPts; $j++) {
     $$name[$j] = $$name[$j]/$cnt_time - $Offsets[$input-1];
  }
  print STDOUT 'Scan record data retrieved!'."\n";

### Re-analyze data again in case we missed needed points while reading SR on the fly:
  $found    = 0;
  $found_r  = 0;
  for ($j=0; $j < $actPts; $j++) {
     $y_curr = $$name[$j];
     if (($y_curr <= $target && $direction == $DECREASING) ||
         ($y_curr >= $target && $direction == $INCREASING)) {
        $found++;
        if ($found > $found_r) {$found_r = $found;}
     } else {
        $found = 0;
     }
     if ($found >= $found_max) {last;}
  }

  if ($found >= $found_max) { #- - - - - - - - - - - - -#target found

### Re-search the target intensity from the left and right sides of data array:
     $pt_left  = 0;
     $pt_right = 0;
     $y_left   = 0.;
     $y_right  = 0.;
     for ($j=0; $j < $actPts; $j++) {
        $y_curr = $$name[$j];
        if (($y_curr <= $target && $direction == $DECREASING) ||
            ($y_curr >= $target && $direction == $INCREASING)) {
           $pt_left = $j;
           $y_left  = $y_curr;
           last;
        }
     }
     for ($j=$actPts-1; $j >= 0; $j--) {
        $y_curr = $$name[$j];
        if (($y_curr >= $target && $direction == $DECREASING) ||
            ($y_curr <= $target && $direction == $INCREASING)) {
           $pt_right = $j;
           $y_right  = $y_curr;
           last;
        }
     }
     $pt_diff = abs($pt_right-$pt_left);
     if ($pt_diff > 2) {
        print STDOUT '!!! '.$script.': left and right thresholds differ by '.$pt_diff.' pts.'."\n";
        print STDOUT '               Possible noise on the intensity data!'."\n";
     }
     $name   = 'x1SR';
     $x_left  = $$name[$pt_left];
     $x_right = $$name[$pt_right];
     if ($y_left != $y_right) {
###       $pos = $x_left  * ($y_right - $target) / ($y_right - $y_left)
###            + $x_right * ($target - $y_left ) / ($y_right - $y_left);
### Same:
        $position = $x_left + ($x_right - $x_left) * ($target - $y_left )
                                                   / ($y_right - $y_left);
     } else {
        $position = 0.5*($x_left + $x_right);
     }
     printf STDOUT $script.': target_intensity=%1.0f for motor='.$axis
                          .' found at position=%-.3g'."\n",$target,$position;
     $findPos2ndTime = 1;
     use strict 'refs';
     return 0,$actPts,$position;

  } else {  #- - - - - - - - - - - - - - - - - - - - - -#target not found:

### Collect and print out statistics:
     if ($actPts > 0) {
        $name   = 'y1SR';
        ($pt_min,$y_min,$pt_max,$y_max) = &arrayStats($actPts,@$name);
        $name   = 'x1SR';
        $x_min = $$name[$pt_min];
        $x_max = $$name[$pt_max];
        $messageSRSCAN = sprintf
                         'Unsufficient scan range for ['.$axis.']!'."\n"
                        .'   Target intensity = %10.0f'."\n"
                        .'   Search direction = '.$dir{$direction}."\n"
                        .'   Min intensity found = %10.0f  (at pt =%d, coord=%.3f)'."\n"
                        .'   Max intensity found = %10.0f  (at pt =%d, coord=%.3f)'."\n"
                        .'   Max '.$found_r.' consecutive pts satisfying the'
                        .' condition found of '.$found_max.' required.'."\n",
                                $target,
                                $y_min,$pt_min,$x_min,
                                $y_max,$pt_max,$x_max;
        $messageSRSCAN .=  '  Array data on prevously found "good" data points:'."\n";
        for ($j=0; $j<$found_p; $j++) {
           $name   = 'y1SR';
           $messageSRSCAN .= sprintf '     pt=%d  intensity=%.0f'."\n",
                                          $good_pts[$j],@$name[$good_pts[$j]];
        }
     } else {
        $messageSRSCAN = sprintf
                         'Unsufficient scan range for ['.$axis.']!'."\n"
                        .'   Target intensity = %10.0f'."\n"
                        .'   Search direction = '.$dir{$direction}."\n\n"
                        .'** Zero-length data array returned by scanrec! **'."\n",
                                $target;
     }
     print STDOUT "\n".'*** '.$script.': '.$messageSRSCAN."\n";
     $findPos2ndTime = 1;
     use strict 'refs';
     return 1,$actPts,$position;
  } #- - - - - - - - - - - - - - - - - - - - - - - - - -#end analyzing data
}	##findPosSR
